Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax GeneralTyping Defs.
Require Import ConfigTyping RecordTypes HeapUpdate HeapValid Deref PreciseTyping InertContext.

(** * Closed syntax
Lemmas about proving, inverting and preserving closedness of terms, types, heap etc. *)

Lemma heap_closed_push_item Si x i:
  heap_closed (Si & x ~ i) ->
  closed i.
Proof.
  introv Hcl.
  inversions Hcl. false. eapply empty_push_inv. eauto. eapply eq_push_inv in H. destruct_all. subst. auto.
Qed.
Lemma heap_closed_push_inv Si x i:
  heap_closed (Si & x ~ i) ->
  heap_closed Si.
Proof.
  introv Hcl.
  inversions Hcl. false. eapply empty_push_inv. eauto. eapply eq_push_inv in H. destruct_all. subst. auto.
Qed.

Instance heap_closed_conjunctive: EnvConjunctive heap_closed (fun x v => closed v) := {
  env_conj_empty := heap_closed_empty ;
  env_conj_push := heap_closed_push ;
  env_conj_push_item := heap_closed_push_item ;
  env_conj_push_inv := heap_closed_push_inv
}.

Hint Resolve heap_closed_conjunctive: core.


Lemma heap_binds_closed y it Sigma:
  heap_closed Sigma ->
  binds y it Sigma ->
  closed it.
Proof.
  introv HclS Hbi.

  eapply (binds_conjunctive heap_closed_conjunctive); eauto.
Qed.

Lemma heap_binds_fun_closed y t Sigma x:
  heap_closed Sigma ->
  binds y (item_fun t) Sigma ->
  closed (open x t).
Proof.
  introv HclS Hbi.
  forwards: heap_binds_closed; eauto.
  unfold closed in *.
  intros.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x0). sympls.
    rewrite Nat.succ_pred in H by auto. injection H. intro.
    rewrite open_commut by auto. rewrite <- H0 at 2. auto.
Qed.

Lemma closed_defs_inv ds d:
  closed (defs_cons ds d) ->
  closed ds.
Proof.
  intros.
  unfold closed in *.
  intros. specialize (H n x).
  simpls. injection H. intros. eauto.
Qed.

Lemma closed_defs_def ds d:
  closed (defs_cons ds d) ->
  closed d.
Proof.
  intros.
  unfold closed in *.
  intros. specialize (H n x).
  simpls. injection H. intros. eauto.
Qed.



Lemma closed_defs_nil:
  closed (defs_nil).
Proof.
  unfold closed. simpl. auto.
Qed.

Lemma closed_defs_cons ds d:
  closed ds ->
  closed d ->
  closed (defs_cons ds d).
Proof.
  unfold closed. intros.
  specialize (H n x).
  specialize (H0 n x).
  simpls.
  fold_trm_open_rec.
  rewrite H.
  rewrite H0.
  auto.
Qed.

Instance defs_closed_conjunctive: DefsConjunctive closed closed := {
  defs_conj_nil := closed_defs_nil ;
  defs_conj_cons := closed_defs_cons ;
  defs_conj_cons_def := closed_defs_def ;
  defs_conj_cons_inv := closed_defs_inv
}.

Hint Resolve defs_closed_conjunctive: core.


Lemma defs_valid_all_has_conjunctive (P: defs -> Prop) (Q: def -> Prop) (PC: DefsConjunctive P Q) ds:
  defs_valid ds ->
  (forall d, defs_has ds d -> Q d) ->
  (P ds).
Proof.
  intros.
  induction ds.
  - apply defs_conj_nil.
  - eapply defs_conj_cons.
    + inversions H.
      eapply IHds; eauto.
      intros.
      eapply H0.
      eapply defs_has_n; eauto.
      eapply neq_sym.
      eapply defs_has_hasnt_neq; eauto.
    + apply H0. apply defs_has_cons.
Qed.

Lemma defs_update_conjunctive (P: defs -> Prop) (Q: def -> Prop) (PC: DefsConjunctive P Q) ds1 a y ds2:
  P ds1 ->
  defs_update ds1 a y ds2 ->
  Q (def_fld a (avar_f y)) ->
  P ds2.
Proof.
  introv Hcl Hu HQ.
  induction Hu.
  - apply defs_conj_cons.
    + eapply defs_conj_cons_inv; eauto.
    + auto.
  - apply defs_conj_cons.
    + eapply IHHu.
      eapply defs_conj_cons_inv; eauto.
      auto.
    + eapply defs_conj_cons_def. eauto.
Qed.

Lemma closed_item_obj ds:
  closed (item_obj ds) ->
  closed ds.
Proof.
  intros.
  unfold closed in *. intros. specialize (H n x).
  simpls. injection H. intros. eauto.
Qed.

Lemma closed_item_obj' ds:
  closed ds ->
  closed (item_obj ds).
Proof.
  intros.
  unfold closed in *. intros. specialize (H n x).
  simpls. rewrite H. auto.
Qed.


Lemma closed_item_fun' w t:
  closed (open w t) ->
  closed (item_fun t).
Proof.
  intros.
  unfold closed in *. intros. specialize (H (S n) x).
  rewrite open_commut in H.
  eapply open_twice_inv in H.
  simpls.
  rewrite H. auto.
  auto. auto.
Qed.

Lemma closed_defs_has ds d:
  closed ds ->
  defs_has ds d ->
  closed d.
Proof.
  eapply defs_has_conjunctive.
  eapply defs_closed_conjunctive.
Qed.

Lemma closed_def_met m t w1 w2:
  closed (def_met m t) ->
  closed (open_met w1 w2 t).
Proof.
  introv Hcl.
  unfold closed in *.
  intros.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - destruct (classicT(n=1)).
    + subst. rewrite open_commut by auto.
    assert ( (open_rec 1 x (open_rec 1 w2 t)) =(open_rec 1 w2 t)).
    eapply open_twice.
    rewrite H. auto.
    + specialize (Hcl (Nat.pred (Nat.pred n)) x). sympls. rewrite Nat.succ_pred in Hcl. rewrite Nat.succ_pred in Hcl by auto. injection Hcl. intro. rewrite open_commut by auto. rewrite open_commut with (m0:=n) by auto. rewrite <- H at 2. auto.
      eapply Nat.succ_inj_wd_neg.
      rewrite Nat.succ_pred. auto. auto.
Qed.

Lemma heap_binds_met_closed Sigma yd y m t w2 w1:
  heap_closed Sigma ->
  binds y (item_obj yd) Sigma ->
  defs_has yd (def_met m t) ->
  closed (open_met w2 w1 t).
Proof.
  introv HclS Hbi Hhas.
  forwards: heap_binds_closed; eauto.
  forwards: closed_item_obj; eauto.
  forwards: closed_defs_has; eauto.
  eapply closed_def_met; eauto.
Qed.

Lemma closed_trm_avar_f x:
  closed (trm_var (avar_f x)).
Proof.
  unfold closed.
  intros.
  simpl.
  auto.
Qed.

Lemma defs_update_closed ds1 a y ds2:
  closed ds1 ->
  defs_update ds1 a y ds2 ->
  closed ds2.
Proof.
  intros.
  eapply defs_update_conjunctive; eauto.
  unfold closed. intros. simpl. auto.
Qed.

Lemma heap_closed_update S1 y1 a y3 S2:
  heap_closed S1 ->
  heap_update S1 y1 a y3 S2 ->
  heap_closed S2.
Proof.
  introv Hcl Hu.
  pose proof Hu as Hu'.
  destructc Hu as ? ? Hdom Hbi ds ds' Hrest. destructc Hrest as Hbiy1 Hdsu Hbiy2.

  assert (closed ds).
  eapply closed_item_obj; eauto.
  eapply heap_binds_closed; eauto.

  assert (closed ds').
  eapply defs_update_closed; eauto.

  eapply binds_all_conjunctive.
  apply heap_closed_conjunctive.
  auto.
  simpl.
  intros.
  destruct (heap_update_binds_rev Hu' H3).
  - eapply heap_binds_closed; eauto.
  - subst. pose proof (binds_functional Hbiy2 H3). subst.
    apply closed_item_obj'. auto.
Qed.

Lemma heap_closed_subst S x y :
  heap_closed S ->
  heap_closed (subst_heap x y S).
Proof.
  induction 1; simpl; eauto using closed_subst_var.
  - unfold subst_heap. rewrite map_empty. constructor; eauto.
  - unfold subst_heap. rewrite map_concat, map_single.
    constructor; eauto. eapply closed_subst_var. easy.
Qed.

Lemma closed_trm_let_l t1 t2:
  closed (trm_let t1 t2) ->
  closed t1.
Proof.
  unfold closed.
  intros.
  simpls.
  einjection H. intros. apply H1.
Qed.

Lemma closed_trm_let_r t1 t2 x:
  closed (trm_let t1 t2) ->
  closed (open x t2).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x0).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H0. auto.
Qed.

Lemma closed_deref_def d e:
  closed d ->
  closed (deref_def e d).
Proof.
  introv Hcl.
  destruct d; simpl; eauto.

  destruct a; simpl.
  - unfold closed in Hcl.
    pick_fresh y.
    specialize (Hcl n y).
    cases_if.
  - unfold deref, deref_avar.
    revert dependent v.
    induction e using env_ind.
    + intros. rewrite empty_def. eauto.
    + intros. rewrite deref_base_push.
      unfold subst_var. unfold AvarSubstVar at 2.
      eapply IHe. cases_if; eauto. unfold closed. eauto.
Qed.

Lemma closed_deref_defs ds e:
  closed ds ->
  closed (deref_defs e ds).
Proof.
  introv Hcl.
  induction ds.
  - simpl. eauto.
  - simpl.
    eapply closed_defs_cons.
    + apply IHds.
      eapply closed_defs_inv; eauto.
    + eapply closed_deref_def.
      eapply closed_defs_def; eauto.
Qed.

Lemma closed_trm_lit_obj_l R d t w:
  closed (trm_lit (lit_obj R d) t) ->
  closed (open w d).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H1. auto.
Qed.

Lemma closed_trm_lit_obj_r R d t w:
  closed (trm_lit (lit_obj R d) t) ->
  closed (open w t).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H0. auto.
Qed.

Lemma closed_trm_lit_obj_t R d t w:
  closed (trm_lit (lit_obj R d) t) ->
  closed (open w R).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H2. auto.
Qed.


Lemma closed_trm_lit_fun_t2 R t1 t2 w:
  closed (trm_lit (lit_fun R t1) t2) ->
  closed (open w t2).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H0. auto.
Qed.

Lemma closed_trm_lit_fun_t1 R t1 t2 w:
  closed (trm_lit (lit_fun R t1) t2) ->
  closed (open w t1).
Proof.
  unfold closed.
  intros.
  simpls.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (H (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in H by auto.
    injection H.  fold_trm_open_rec. intros. rewrite open_commut by auto.
    rewrite H1. auto.
Qed.

Lemma closed_trm_lit_fun_T R t1 t2:
  closed (trm_lit (lit_fun R t1) t2) ->
  closed R.
Proof.
  unfold closed.
  intros.
  simpls.
  einjection H. intros. apply H2.
Qed.

Lemma closed_typ_bot:
  closed typ_bot.
Proof.
  unfold closed.
  simpls.
  eauto.
Qed.
Lemma closed_typ_top:
  closed typ_top.
Proof.
  unfold closed.
  simpls.
  eauto.
Qed.
Lemma closed_typ_nocap:
  closed typ_nocap.
Proof.
  unfold closed.
  simpls.
  eauto.
Qed.

Lemma closed_typ_and_l T1 T2:
  closed (typ_and T1 T2) ->
  closed T1.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_typ_and_r T1 T2:
  closed (typ_and T1 T2) ->
  closed T2.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_typ_and T1 T2:
  closed T1 ->
  closed T2 ->
  closed (typ_and T1 T2).
Proof.
  unfold closed.
  introv Hcl1 Hcl2. simpls.
  intros.
  fold_typ_open_rec.
  rewrite (Hcl1 n x).
  rewrite (Hcl2 n x).
  auto.
Qed.

Lemma closed_typ_or_l T1 T2:
  closed (typ_or T1 T2) ->
  closed T1.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_typ_or_r T1 T2:
  closed (typ_or T1 T2) ->
  closed T2.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_typ_or T1 T2:
  closed T1 ->
  closed T2 ->
  closed (typ_or T1 T2).
Proof.
  unfold closed.
  introv Hcl1 Hcl2. simpls.
  intros.
  fold_typ_open_rec.
  rewrite (Hcl1 n x).
  rewrite (Hcl2 n x).
  auto.
Qed.


Lemma closed_typ_bnd_inv T y:
  closed (typ_bnd T) ->
  closed (open y T).
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (Hcl (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in Hcl by auto.
    injection Hcl.
 fold_typ_open_rec. intros. rewrite open_commut by auto.
    rewrite H. auto.
Qed.
Lemma closed_typ_rcd_inv D:
  closed (typ_rcd D) ->
  closed D.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_typ_rcd D:
  closed D ->
  closed (typ_rcd D).
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  fold_typ_open_rec.
  rewrite (Hcl n x).
  eauto.
Qed.

Lemma closed_dec_typ_inv_r B T1 T2 y:
  closed (dec_typ B T1 T2) ->
  closed (open y T2).
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (Hcl (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in Hcl by auto.
    injection Hcl.
 fold_typ_open_rec. intros. rewrite open_commut by auto.
    rewrite H. auto.
Qed.
Lemma closed_dec_typ_inv_l B T1 T2 y:
  closed (dec_typ B T1 T2) ->
  closed (open y T1).
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  destruct (classicT(n=0)).
  - subst. eapply open_twice.
  - specialize (Hcl (Nat.pred n) x).
    simpls.
    rewrite Nat.succ_pred in Hcl by auto.
    injection Hcl.
 fold_typ_open_rec. intros. rewrite open_commut by auto.
    rewrite H0. auto.
Qed.

Lemma closed_dec_typ B T1 T2 L:
  (forall y, y \notin L -> closed (open y T1)) ->
  (forall y, y \notin L -> closed (open y T2)) ->
  closed (dec_typ B T1 T2).
Proof.
  unfold closed.
  introv Hcl1 Hcl2. simpls.
  intros.
  fold_typ_open_rec.

  pick_fresh z.
  assert (z \notin L) as HnzL by eauto.
  specialize (Hcl1 z HnzL (S n) x).
  specialize (Hcl2 z HnzL (S n) x).
  rewrite open_commut in Hcl1.
  rewrite open_commut in Hcl2.

  eapply open_fresh_injective in Hcl1.
  eapply open_fresh_injective in Hcl2.

  rewrite Hcl1.
  rewrite Hcl2.
  eauto.
  eapply notin_fv_open. eauto. eauto. eauto.
  eapply notin_fv_open. eauto. eauto. eauto.
  eauto.
  eauto.
Qed.

Lemma closed_dec_fld_l a T1 T2:
  closed (dec_fld a T1 T2) ->
  closed T1.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_dec_fld_r a T1 T2:
  closed (dec_fld a T1 T2) ->
  closed T2.
Proof.
  unfold closed.
  introv Hcl. simpls.
  intros.
  injection (Hcl n x).
  eauto.
Qed.
Lemma closed_dec_fld a T1 T2:
  closed T1 ->
  closed T2 ->
  closed (dec_fld a T1 T2).
Proof.
  unfold closed.
  introv Hcl1 Hcl2. simpls.
  intros.
  fold_typ_open_rec.
  rewrite (Hcl1 n x).
  rewrite (Hcl2 n x).
  auto.
Qed.

Lemma closed_typ_fun_invl T U:
  closed (typ_fun T U) ->
  closed T.
Proof.
  introv Hcl.
  intro.
  intro.
  forwards: Hcl n x.  
  sympls. injection H. auto.
Qed.

Lemma closed_lit_fun_invl T U:
  closed (lit_fun T U) ->
  closed T.
Proof.
  introv Hcl.
  intro.
  intro.
  forwards: Hcl n x.  
  sympls. injection H. auto.
Qed.

Lemma closed_record_has T D:
  closed T ->
  record_has T D ->
  closed D.
Proof.
  introv Hcl Hhas.
  induction Hhas.
  - eapply closed_typ_rcd_inv. eauto.
  - eapply closed_typ_and_l in Hcl. eauto.
  - eapply closed_typ_and_r in Hcl. eauto.
Qed.

Lemma closed_sealed_typdec:
  (forall (t: typ) n y, SealClosedOf n y t) /\
  (forall (d: dec) n y, SealClosedOf n y d).
Proof.
  apply typ_mutind; intros; simpls; eauto.
  - rewrite H; eauto. intros. specialize (H0 m x H1). inversions H0. rewrite H3. auto.
  - rewrite H; eauto. rewrite H0; eauto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H5. auto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto.
  - rewrite H; eauto. rewrite H0; eauto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H5. auto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto.
  - rewrite closed_sealed_rec; eauto.
    rewrite closed_sealed_rec; eauto.
    intros. specialize (H m x H0). inversions H. rewrite H3. auto.
    intros. specialize (H m x H0). inversions H. rewrite H2. auto.
  - rewrite H; eauto.
    intros. specialize (H0 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H1)).
    rewrite Nat.succ_pred_pos in H0. injection H0. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - (* typ *)
    rewrite H; eauto.
    2: {intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto. }

    rewrite H0. eauto.
    intros. specialize (H1 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H2)).
    rewrite Nat.succ_pred_pos in H1. injection H1. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.

  - rewrite H; eauto.
    2: {
      intros. specialize (H1 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H2)).
      rewrite Nat.succ_pred_pos in H1. injection H1. auto.
      eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
    }
    rewrite H0. eauto.
    intros. specialize (H1 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H2)).
    rewrite Nat.succ_pred_pos in H1. injection H1. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - rewrite H; eauto. rewrite H0; eauto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H5. auto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto.
  - rewrite H; eauto.
    2: {
      intros. specialize (H2 (Nat.pred m0) x (Nat.le_succ_le_pred _ _ H3)).
      rewrite Nat.succ_pred_pos in H2. injection H2. auto.
      eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
    }
    rewrite H0.
    2: {intros. specialize (H2 m0 x H3). inversions H2. rewrite H6. auto. }
    rewrite H1; eauto.
    {
      intros. specialize (H2 (Nat.pred (Nat.pred m0)) x (Nat.le_succ_le_pred _ _ (Nat.le_succ_le_pred _ _ H3))).
      rewrite Nat.succ_pred_pos in H2.
      rewrite Nat.succ_pred_pos in H2.
      injection H2. auto.
      eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
      eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eapply Nat.le_succ_le_pred.
      eauto.
    }
Qed.

Instance TypSealClosed: SealClosed TypAbstractSyntax TypSealable.
Proof.
  hnf; intros.
  eapply closed_sealed_typdec; eauto.
Qed.

Lemma closed_sealed_trm_defs:
  (forall (t: trm) n y, SealClosedOf n y t) /\
  (forall (t: lit) n y, SealClosedOf n y t) /\
  (forall (t: def) n y, SealClosedOf n y t) /\
  (forall (d: defs) n y, SealClosedOf n y d).
Proof.
  apply trm_mutind; intros; simpls; eauto.
  - rewrite closed_sealed_rec; eauto.
    intros. specialize (H m x H0). inversions H. rewrite H2. auto.
  - rewrite closed_sealed_rec; eauto.
    intros. specialize (H m x H0). inversions H. rewrite H2. auto.
  - rewrite closed_sealed_rec; eauto.
    2:{ intros. specialize (H m x H0). inversions H. rewrite H2. auto. }
    rewrite closed_sealed_rec; eauto.
    { intros. specialize (H m x H0). inversions H. rewrite H3. auto. }
  - (* let *)
    rewrite H; eauto.
    2: {intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto. }

    rewrite H0. eauto.
    intros. specialize (H1 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H2)).
    rewrite Nat.succ_pred_pos in H1. injection H1. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - (* lit *)
    rewrite H; eauto.
    2: {intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto. }

    rewrite H0. eauto.
    intros. specialize (H1 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H2)).
    rewrite Nat.succ_pred_pos in H1. injection H1. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - rewrite closed_sealed_rec; eauto.
    2:{ intros. specialize (H m x H0). inversions H. rewrite H2. auto. }
    rewrite closed_sealed_rec; eauto.
    { intros. specialize (H m x H0). inversions H. rewrite H3. auto. }
  - rewrite closed_sealed_rec; eauto.
    2:{ intros. specialize (H m0 x H0). inversions H. rewrite H2. auto. }
    rewrite closed_sealed_rec; eauto.
    { intros. specialize (H m0 x H0). inversions H. rewrite H3. auto. }
  - (* lit_fun *)
    rewrite closed_sealed_rec; eauto.
    2:{ intros. specialize (H0 m x H1). inversions H0. rewrite H3. auto. }

    rewrite H. eauto.
    intros. specialize (H0 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H1)).
    rewrite Nat.succ_pred_pos in H0. injection H0. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - (* lit_fun *)
    rewrite closed_sealed_rec; eauto.
    2:{  intros. specialize (H0 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H1)).
    rewrite Nat.succ_pred_pos in H0. injection H0. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto. }

    rewrite H. eauto.
    intros. specialize (H0 (Nat.pred m) x (Nat.le_succ_le_pred _ _ H1)).
    rewrite Nat.succ_pred_pos in H0. injection H0. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - rewrite closed_sealed_rec; eauto.
    intros. specialize (H (Nat.pred m) x (Nat.le_succ_le_pred _ _ H0)).
    rewrite Nat.succ_pred_pos in H. injection H. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
  - rewrite closed_sealed_rec; eauto.
    intros. specialize (H m x H0). inversions H. rewrite H2. auto.
  - (* def_met *)
    rewrite H. eauto.
    intros. specialize (H0 (Nat.pred (Nat.pred m0)) x (Nat.le_succ_le_pred _ _ (Nat.le_succ_le_pred _ _ H1))).
    rewrite Nat.succ_pred_pos in H0.
    rewrite Nat.succ_pred_pos in H0.
    injection H0. auto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eauto.
    eapply Nat.lt_lt_0. apply -> PeanoNat.Nat.le_succ_l. eapply Nat.le_succ_le_pred.
    eauto.
  - rewrite H; eauto. rewrite H0; eauto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H5. auto.
    intros. specialize (H1 m x H2). inversions H1. rewrite H4. auto.
Qed.

Instance TrmSealClosed: SealClosed TrmAbstractSyntax TrmSealable.
Proof.
  hnf; intros.
  eapply closed_sealed_trm_defs; eauto.
Qed.

Instance DefSealClosed: SealClosed DefAbstractSyntax DefSealable.
Proof.
  hnf; intros.
  eapply closed_sealed_trm_defs; eauto.
Qed.
Instance LitSealClosed: SealClosed LitAbstractSyntax LitSealable.
Proof.
  hnf; intros.
  eapply closed_sealed_trm_defs; eauto.
Qed.
Instance DefsSealClosed: SealClosed DefsAbstractSyntax DefsSealable.
Proof.
  hnf; intros.
  eapply closed_sealed_trm_defs; eauto.
Qed.

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma pf_rcd_closed F v T D:
  inert F ->
  sealed F ->
  F ⊢! v ∶ T ⪼ typ_rcd D ->
  closed D.
Proof.
  introv Hin Hfx Hpf.

  assert (closed T) as HclT. {
    forwards: pf_binds; eauto.
    forwards: ctx_closed_binds; eauto.
    unfold sealed in H0.
    lets: (sealed_closed v T).
    rewrite H0 in H1.
    auto.
  }
  forwards U S HT: pf_inert_rcd_U; eauto.
  subst.
  forwards: precise_flow_has; eauto.

  destruct H.
  + (* record *)
    destruct H.
    eapply closed_typ_and_l in HclT.
    eapply closed_typ_bnd_inv with (y:=v) in HclT.
    forwards: closed_record_has; eauto.
  + (* refcap *)
    destruct H.
    eapply closed_typ_and_r in HclT.
    forwards: closed_record_has; eauto.
Qed.

Lemma pf_typ_closed_l F v T B T1 T2 x:
  inert F ->
  sealed F ->
  F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
  closed (open x T2).
Proof.
  introv Hin Hfx Hpf.

  forwards: pf_rcd_closed; eauto.
  eapply closed_dec_typ_inv_r in H.
  eauto.
Qed.
Lemma pf_typ_closed_r F v T B T1 T2 x:
  inert F ->
  sealed F ->
  F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
  closed (open x T1).
Proof.
  introv Hin Hfx Hpf.

  forwards: pf_rcd_closed; eauto.
  eapply closed_dec_typ_inv_l in H.
  eauto.
Qed.

Lemma closed_typ_cap_adapt x TcapN:
  closed (open x TcapN) ->
  closed (typ_cap_adapt typ_refcap_mutability x TcapN).
Proof.
  introv HclN.
  hnf in *. intros.
  sympls.
  cases_if.
  specialize (HclN (S n) x0).
  rewrite open_commut in HclN.
  eapply open_twice_inv in HclN.
  rewrite HclN; auto.
  eauto.
  eauto.
Qed.

End WithTypingMode.
