Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping
        Weakening EWeakening Narrowing
        RecordTypes InertContext TFree TFreeTyping
        HeapCorrespondence ConfigVarkinds NothingFresh ConfigNotations.

(** * Typing of machine configurations *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* Part of Figure 9 in "Reference mutability for DOT" *)
(* Part of Figure 14 in "Pure methods for roDOT" *)
(** The typing of a term with a stack *)
Inductive ty_stack : ctx -> renv -> typ -> stack -> typ -> Prop :=
  | ty_stack_nil : forall G e T U,
      G⁏e ⊢ T <: U ->
      ty_stack G e T stack_nil U
  | ty_stack_let : forall L G e s t S T U,
      ty_stack G e T s U ->
      (forall x,
          x \notin L ->
          G § x / v_let ~ S ⁏e ⊢ open x t ∶ T) ->
      ty_stack G e S (stack_cons (frame_let t) s) U.

(* While for ctx, we use sealed, for configuration we use closed.
   Those two are interchangeable for types and terms. *)

Inductive heap_closed: heap -> Prop :=
| heap_closed_empty: heap_closed empty
| heap_closed_push S x i:
    heap_closed S ->
    closed i ->
    heap_closed (S & x ~ i)
.

Inductive config_closed: config -> Prop:=
  config_sealed_cfg t s Sigma e Lv:
    closed t ->
    closed s ->
    heap_closed Sigma ->
    config_closed (t; s; Sigma; e; Lv).

(* Part of Figure 9 in "Reference mutability for DOT" *)
(* Part of Figure 14 in "Pure methods for roDOT" *)
Inductive ty_config : ctx -> config -> typ -> Prop :=
| ty_config_c : forall G e s Sigma Lv t T U,
    inert G ->
    heap_correspond G e Sigma ->
    ty_stack G e T s U ->
    all_tfree_k_stack G s v_ref ->
    G⁏e ⊢ t ∶ T ->
    all_tfree_k G t v_ref ->
    ty_nothing_fresh_in_config G U (t ; s ; Sigma ; e; Lv) ->
    has_null Lv ->
    sealed G -> (* context and configuration must be closed *)
    config_closed (t; s; Sigma; e; Lv) -> (* context and configuration must be closed *)
    closed U ->
    ty_config G (t; s; Sigma; e; Lv) U.

Inductive config_typd : config -> typ -> Prop :=
| config_typd_c : forall Gamma c U,
    ty_config Gamma c U ->
    config_typd c U.

Inductive config_hastyp : config -> Prop :=
| config_hastyp_c : forall Gamma c U,
    ty_config Gamma c U ->
    config_hastyp c.


Hint Constructors ty_stack ty_config config_typd config_hastyp : core.

Notation "'⊢' c '∶' T" := (config_typd c T) (at level 40, c at level 59).

Lemma ty_stack_push : forall G e T x vk T' s U,
    ctx_ok G ->
    x §# G ->
    ty_stack G e T s U ->
    ty_stack (G § x / vk ~ T') e T s U.
Proof.
  introv Hok Fr Hts.
  induction Hts.
  + apply ty_stack_nil.
    apply weaken_push; auto.
  + specialize (IHHts Hok Fr).
    eapply (@ty_stack_let (L \u ctx_dom G)); eauto 2.
    intros. rewrite ctx_concat_double2. eapply weaken; eauto.
    rewrite <- ctx_concat_single; eauto.
Qed.
Hint Resolve ty_stack_push : core.

Lemma ty_stack_epush : forall G e T x y s U,
    ok e ->
    x # e ->
    ty_stack G e T s U ->
    ty_stack G (e & x ~ y) T s U.
Proof.
  introv Hok Fr Hts.
  induction Hts.
  + apply ty_stack_nil.
    apply ewkn_push; auto.
  + specialize (IHHts Hok Fr).
    eapply (@ty_stack_let (L \u ctx_dom G)); eauto 2.
    intros.
    apply ewkn_push; auto.
Qed.
Hint Resolve ty_stack_epush : core.

Lemma ty_stack_sub : forall Gamma e T T' s U,
  ctx_ok Gamma ->
  ty_stack Gamma e T s U ->
  Gamma⁏e ⊢ T' <: T ->
  ty_stack Gamma e T' s U.
Proof.
  introv Hok Hts Hsub.
  inversions Hts; eauto.
  eapply (@ty_stack_let (L \u (ctx_dom Gamma))); eauto.
  intros. eauto using narrow_ty.
Qed.

Lemma ty_config_env_ok G t fs S e Lv T:
  ty_config G (t ; fs ; S ; e ; Lv) T ->
  ok e.
Proof.
  introv Hty.
  inversions Hty.
  hnde H5. auto.
Qed.

Lemma ty_config_heap_ok G t fs S e Lv T:
  ty_config G (t ; fs ; S ; e ; Lv) T ->
  ok S.
Proof.
  introv Hty.
  inversions Hty.
  hnde H5. auto.
Qed.


Lemma ty_config_Lv_subset G t fs S e Lva Lvb T:
  ty_config G (t ; fs ; S ; e ; Lva) T ->
  Lva \c Lvb ->
  ty_config G (t ; fs ; S ; e ; Lvb) T.
Proof.
  introv Htyc Hsu.
  inversions Htyc.
  econstructor; eauto.
  - eapply ty_nothing_fresh_in_config_Lv_subset; eauto.
  - destruct H13. exists x. eauto.
  - inversions H15. constructor; eauto.
Qed.



Notation config_initial_mk_typ t T null := (t; stack_nil; empty; empty; \{ null } \u fv t \u fv T) (only parsing).
Inductive config_initial: trm -> config -> Prop :=
  config_initial_mk t Lv: config_initial t (t; stack_nil; empty; empty; Lv).

Lemma ty_config_initial t T:
  closed t ->
  closed T ->
  ctx_empty⁏empty ⊢ t ∶ T ->
  exists C, config_initial t C /\
  ty_config ctx_empty C T.
Proof.
  introv Hcl HclT Hty.
  pick_fresh null.
  exists (config_initial_mk_typ t T null).
  split.
  constructor.
  apply ty_config_c with (T:=T); eauto.
  - unfold all_tfree_k_trm.
    intros.
    assert (v \in ctx_dom (ctx_empty §§& objctx_empty) ).
    eapply tfree_dom; eauto.
    unfold ctx_obj in *.
    ctx_simpl_dom.
    false. eapply in_empty_inv. eauto.
  - ctx_simpl_dom.
    unfold nothing_fresh_in_config.
    split. split.
    eauto using subset_empty_l.
    eauto using subset_empty_l.
    intro. intro.
    repeat rewrite in_union.
    repeat right. auto.
  - exists null. rewrite in_union. left. rewrite in_singleton. auto.
  - constructor; eauto.
    constructor. constructor.
Qed.

End WithTypingMode.

(* Inverts typing Hty of focus of execution t, while dealing with subtyping.
   When proving typed preservation, we want to invert term typing, where the type of the term has a known shape.
   However, the type may be given by subsumption, so induction is needed to unwrap the subtyping.
   t      is the shape of the focus of execution
   Hty    gives type to the focus *)
Ltac invert_focus_typing_of t Hty :=
  let tz := fresh "tz" in
  (* Heqtz  name of a new hypothesis that will give the shape of the type by equality *)
  let Heqtz := fresh "Heqtz" in
  remember t as tz eqn: Heqtz;
  induction Hty; (* Induction on term typing *)
  try congruence; (* Removes cases where the type does not match *)
  [ inversions Heqtz (* Base case. Substitutes equal variables occurring in the type. *)
  | eauto using ty_stack_sub (* Solves subtyping case *)
  ].

Ltac invert_focus_typing Hty :=
  match goal with
    Hty2 : _ ⁏ _ ⊢ ?t ∶ _   |- _ =>
    (* Finds the term that is given type by the hypothesis *)
    pose proof Hty: _ ⁏ _ ⊢ t ∶ _ as Hty;
    invert_focus_typing_of t Hty2
  end.

(* Finds a set of variables of a typed configuration and names it LvC *)
Ltac find_Lv LvC := (match goal with
    H: ty_config _ ?G ?C ?U |- _ => set (LvC:=fv_config C)
    end).


Notation "'⊢' c '∶' T" := (config_typd _ c T) (at level 40, c at level 59).
Hint Constructors ty_stack ty_config config_typd : core.
Hint Resolve ty_stack_push : core.
Hint Resolve ty_stack_epush : core.
Arguments ty_config {typing_mode}.
Arguments ty_stack {typing_mode}.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma ty_config_fields_not_refs ws c G T it y :
  ty_config G c T ->
  ws \c dom (config_env c) ->
  binds y it (config_heap c) ->
  disjoint ws (fields it).
Proof.
  introv Htyc Hrefs HbiS. destruct c. simpl in *.
  inversions Htyc. inversions H5. destruct_all.
  eapply binds_to_dom in HbiS as HinS.
  eapply H13 in HinS. inversions HinS.
  1: { binds_eq. simpl. eapply disjoint_empty_r. }
  eapply defs_varkinds_fields_loc in H23. binds_eq.
  forwards:ctx_kdom_disjoint; fweauto.
  Unshelve. 3: exact v_ref. 3: exact v_loc. discriminate.
  eapply fset_extens; eauto using subset_empty_l.
  introv Hin. rewrite <- H20. rewrite in_inter in *.
  destructc Hin as Hinw Hinds. split.
  - eapply Hrefs in Hinw. rewrite H10. easy.
  - eapply H23. eauto.
Qed.

End WithTypingMode.
