Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import DotTactics LibExtra.
Require Import
        AbstractSyntax TypingMode GeneralTyping
        EnvCorrespondence TFree Deref .

(* Varkinds in configurations *)

(* Each variable in a typing context has a varkind.
   Configurations require that:
   1) all fields on heap contain locations
   2) all t-free variables in methods on heap are references
   3) all t-free variables in methods on heap are references
   4) all t-free variables on stack are references
   5) all t-free variables in focus are references

   Conditions 1 and 2 are required by ty_item_obj_s in HeapCorrespondence.
   Condition 3 is required by ty_item_fun_s in HeapCorrespondence.
   Conditions 4 and 5 are required by ty_config in ConfigTyping.
*)

(* States the varkinds condition for a definition *)
Inductive def_varkinds: ctx -> def -> Prop :=
(* Fields must contain a location *)
| def_varkinds_fld G x l:
  ctx_k G x v_loc ->
  def_varkinds G (def_fld l (avar_f x)) (* values of fields are locations *)
(* No requirement for types *)
| def_varkinds_typ G T l:
  def_varkinds G (def_typ l T)
(* In methods, tfree variables must be references *)
| def_varkinds_met G t l:
  all_tfree_k G t v_ref -> (* t-free variables in methods are references *)
  def_varkinds G (def_met l t)
.
Hint Constructors def_varkinds : core.

(* States the varkinds condition for definitions *)
Inductive defs_varkinds: ctx -> defs -> Prop :=
| defs_varkinds_nil G:
  defs_varkinds G (defs_nil)
| defs_varkinds_cons G ds d:
  defs_varkinds G ds ->
  def_varkinds G d ->
  defs_varkinds G (defs_cons ds d)
.
Hint Constructors defs_varkinds : core.

Lemma weaken_def_varkinds G d y vk T:
  def_varkinds G d ->
  y §# G ->
  def_varkinds (G § y / vk ~ T) d.
Proof.
  intros.
  induction H.
  - eapply def_varkinds_fld.
    apply ctx_k_push_n. auto.
    intro; subst.
    apply H0.
    assumption.
  - eapply def_varkinds_typ.
  - eapply def_varkinds_met.
    apply all_tfree_k_Gpush; auto.
Qed.

Lemma weaken_defs_varkinds G ds y vk T:
  defs_varkinds G ds ->
  y §# G ->
  defs_varkinds (G § y / vk ~ T) ds.
Proof.
  intros.
  induction H.
  - eapply defs_varkinds_nil.
  - eapply defs_varkinds_cons; auto.
    apply weaken_def_varkinds; auto.
Qed.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma renv_corr_binds_vk G e vR vL:
  ctx_ok G ->
  renv_corr G e ->
  binds vR vL e ->
  ctx_k G vR v_ref /\ ctx_k G vL v_loc.
Proof.
  introv Hok Hcor Hbi.
  destruct (renv_corr_binds_to_binds_after Hok Hcor Hbi) as [TR [TL [G2 [G3 [HbindsR [HbindsL ?]]]]]].
  split.
  - eapply ctx_binds_k_after_k; eauto.
  - eapply ctx_binds_k_after_k.
    eapply ctx_binds_k_after_trans; eauto.
Qed.

Lemma renv_corr_diff G e w w' y y' :
  ctx_ok G ->
  renv_corr G e ->
  binds w y' e ->
  binds w' y e ->
  w <> y.
Proof.
  introv Hok Hcor Hbi1 Hbi2. hnf. intros; subst.
  eapply renv_corr_binds_vk in Hbi1 as Hvk1; eauto.
  eapply renv_corr_binds_vk in Hbi2 as Hvk2; eauto.
  destruct Hvk1 as [Hvkr1 Hvkl1]. destruct Hvk2 as [Hvkr2 Hvkl2].
  pose (ctx_k_eq Hvkr1 Hvkl2). easy.
Qed.

Lemma renv_corr_env_idempotent G e :
  ctx_ok G -> ok e ->
  renv_corr G e ->
  renv_idempotent e.
Proof.
  introv HokG Hoke Hcor Hind Hini.
  eapply dom_to_binds in Hind as [y1 Hbi1].
  eapply fv_env_to_binds in Hini as [y2 Hbi2]; eauto.
  eapply renv_corr_diff with (w:=x) (y:=x); eauto.
Qed.

(* deref does not affect locations *)
Lemma renv_corr_ctx_k_loc_deref G e v:
  ctx_ok G ->
  renv_corr G e ->
  ctx_k G v v_loc ->
  (deref e v) = v.
Proof.
  introv Hok Hcor Hvk.
  induction e.
  - auto.
  - destruct a.
    rewrite cons_to_push in *.
    assert (ctx_k G v0 v_ref).
    assert (binds v0 v1 (e & v0 ~ v1)) as Hbi by auto.
    destruct (renv_corr_binds_vk Hok Hcor Hbi).
    auto.
    unfold deref, deref_var.
    rewrite deref_base_push.
    sympl.
    destruct (subst_cases2 v0 v1 v) as [[Heq Hsubst] | [Heq Hsubst]].
    + subst.
      false. eapply (ctx_k_ne_false).
      assert (v_ref <> v_loc). all: eauto.
      intro. simplify_eq H0.
    + rewrite Hsubst.
      apply IHe.
      eapply renv_corr_epush_inv; eauto.
Qed.

Lemma renv_corr_in_e_dom_deref G e v:
  ctx_ok G ->
  renv_corr G e ->
  v \in dom e ->
  ctx_k G (deref e v) v_loc.
Proof.
  introv Hok Hcor Hdom.
  destruct (dom_to_binds Hdom) as [vl Hbinds].
  destruct (renv_corr_binds_to_binds_after Hok Hcor Hbinds) as [TR [TL [G2 [G3 [HbindsR [HbindsL ?]]]]]].

  assert ((deref e v) = vl).

  induction e.
  - rewrite <- empty_def in Hbinds.
    apply binds_empty_inv in Hbinds.
    contradiction.
  - destruct a.
    rewrite cons_to_push in *.
    unfold deref, deref_var.
    rewrite deref_base_push.
    destruct (subst_cases2 v0 v1 v) as [[Heq Hsubst] | [Heq Hsubst]].
    + subst.
      assert (v1 = vl).
      assert (binds v0 vl (e & v0 ~ vl)) by auto.
      eapply binds_functional; eauto.
      subst.
      rewrite Hsubst.
      eapply renv_corr_ctx_k_loc_deref; eauto.
      eapply renv_corr_epush_inv; eauto.
      auto.
      eapply ctx_binds_k_after_k; eapply ctx_binds_k_after_trans; eauto.
    + rewrite Hsubst. apply IHe. eapply renv_corr_epush_inv; eauto.
      simpl_dom. rewrite in_union in Hdom. destruct Hdom.
      rewrite in_singleton in H0. contradiction.
      auto.
      eapply binds_push_neq_inv; eauto.
  -
  rewrite H0.
  unfold ctx_k.
  exists TR G3.
  eapply ctx_binds_k_after_trans; eauto.
Qed.


Lemma renv_corr_ctx_k_ref_deref G e v:
  ctx_ok G ->
  renv_corr G e ->
  ctx_kdom v_ref G = dom e ->
  ctx_k G v v_ref ->
  ctx_k G (deref e v) v_loc.
Proof.
  introv Hok Hcor Hdom Hvk.
  assert (v \in ctx_kdom v_ref G) by (apply ctx_k_in_kdom; auto).
  rewrite Hdom in H.
  apply renv_corr_in_e_dom_deref; auto.
Qed.

Lemma all_tfree_def_deref G Go e d T:
  ctx_ok G ->
  renv_corr G e ->
  ctx_kdom v_ref G = dom e ->
  G §§ Go ⁏e ⊢ d ∶d T -> (* Requires typing to rule out bound variables *)
  all_tfree_k_def G d v_ref ->
  def_varkinds G (deref_def e d).
Proof.
  unfold all_tfree_k_def.
  introv Hok Hecor Hdom Hty Hdvk.
  induction d.
  - apply def_varkinds_typ.
  - inversion Hty; subst.
    unfold deref_def.
    rewrite deref_avar_comm.
    apply def_varkinds_fld.

    apply renv_corr_in_e_dom_deref; auto.
    rewrite <- Hdom.
    apply ctx_k_in_kdom; auto.
  - apply def_varkinds_met.
    unfold all_tfree_k.
    auto.
Qed.

Lemma all_tfree_defs_deref G Go e (ds: defs) T:
  ctx_ok G ->
  renv_corr G e ->
  ctx_kdom v_ref G = dom e ->
  G §§ Go⁏e ⊢ ds ∶ T -> (* Requires typing to rule out bound variables *)
  all_tfree_k_defs G ds v_ref ->
  defs_varkinds G (deref e ds).
Proof.
  unfold all_tfree_k_defs.
  introv Hok Hecor Hdom Hty Hdvk.
  induction Hty.
  - rewrite deref_defs_cons_comm.
    apply defs_varkinds_cons; auto.
    apply all_tfree_def_deref with (T:=D) (Go:= Go); auto.
    unfold all_tfree_k_def. auto.
  - rewrite deref_defs_cons_comm.
    apply defs_varkinds_cons; auto.
    apply all_tfree_def_deref with (T:=D) (Go:= Go); auto.
    unfold all_tfree_k_def. auto.
Qed.

Lemma def_varkinds_fields_loc G d:
  def_varkinds G d ->
  fields d \c ctx_kdom v_loc G.
Proof.
  introv Hvk.
  induction Hvk;
  unfold fields, FieldsDef, fields_def.
  eapply subset_singleton.
  eapply ctx_k_in_kdom; eauto.
  all: eapply subset_empty_l.
Qed.
Lemma defs_varkinds_fields_loc G d:
  defs_varkinds G d ->
  fields d \c ctx_kdom v_loc G.
Proof.
  introv Hvk.
  induction Hvk;
  unfold fields, FieldsDefs, fields_defs.
  - eapply subset_empty_l.
  - fold fields_defs. eapply subset_union_l.
    eapply IHHvk. eapply def_varkinds_fields_loc; eauto.
Qed.


End WithTypingMode.
