Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses Program.Equality.

Require Import AbstractSyntax Equivalence PreciseTyping RecordTypes TypingMode InertContext Indep GeneralTyping.
Require Import Weakening EWeakening Substitution.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Environment correspondence *)

(** An environemnt e corresponding to a typing context G
   only maps references w to locations y, such that
   w and y have corresponding types in G.

   Type Tw of a reference w corresponds to a type Ty of a location y
   if Tw and Ty only differ in reference capabilities.
*)

(* Reference capabilities are always supertype of location capabilities *)
(* Note: if in basic mode, if would be relaxed to this:
     ctx_empty⁏renv_empty ⊢ Ty <: Tw /\
     ctx_empty⁏renv_empty ⊢ Tw <: Ty
   then we would not be able to show that the reference type is inert in inert_typ_corr_inert,
   which would make typed preservation proof uglier.
*)

Notation typ_refcap_corr Tw Ty := (
  match (mode_refcap_corr' typing_mode) with
  | mode_refcap_corr_same =>
    (* In basic mode, the capabilities are the same *)
    Ty = Tw
  | mode_refcap_corr_ref_mutability =>
    (* In rodot mode, the capabilities of a location a re *)
    exists Tbnd, (Tw = (typ_mut_marker Tbnd) /\ Ty = (typ_mut_marker typ_bot))
  end) (only parsing).

(** Correspondence of a reference type and location type *)
Inductive typ_corr: typ -> typ -> Prop :=
| typ_corr_obj(T S U: typ):
  typ_refcap_corr S U ->
  typ_corr (typ_objref T S) (typ_objref T U)
| typ_corr_fun(t1 t2: typ):
  mode_has_fun typing_mode ->
  typ_corr (typ_fun t1 t2) (typ_fun t1 t2)
.

(* Part of Figure 9 in "Reference mutability for DOT" *)
(* Part of Figure 14 in "Pure methods for roDOT" *)
(** Correspondence of an environment and typing context *)
Inductive renv_corr: ctx -> renv -> Prop :=
(* empty environment corresponds to any context *)
| renv_corr_empty(G: ctx):
    renv_corr G renv_empty
(* if v1 maps to v2 in e, then in G,
   - v1 must be a reference
   - v2 must be a location
   - v2 must come before v1
   - they must have corresponding types
*)
| renv_corr_ref(G1 G2 G3: ctx)(e: renv)(w y: var)(Tw Ty: typ):
    ctx_binds_k_after w v_ref Tw G1 G2 ->
    ctx_binds_k_after y v_loc Ty G2 G3 ->
    typ_corr Tw Ty ->
    renv_corr G1 e ->
    renv_corr G1 (renv_bind e w y)
.

Hint Constructors renv_corr : core.

(* Preservation of correspondence when extending the context and environment *)

Lemma renv_corr_push G x vk t e:
  renv_corr G e ->
  x §# G ->
  renv_corr (G § x / vk ~ t) e.
Proof.
  introv Hecor Hun.
  induction Hecor.
  - auto.
  - eapply renv_corr_ref; eauto 2.
    apply ctx_binds_k_n; auto.
    apply ctx_binds_k_after_binds in H.
    apply ctx_binds_dom in H.
    intro; subst.
    contradict H; auto.
Qed.

Lemma renv_corr_push_inv G x vk t e:
  renv_corr (G § x / vk ~ t) e ->
  x # e ->
  renv_corr G e.
Proof.
  introv Hecor Hn.
  dependent induction Hecor.
  - auto.
  - eapply renv_corr_ref; eauto.
    inversions H; auto.
    contradiction Hn.
    simpl_dom. rewrite in_union. left.
    rewrite in_singleton.
    auto.
Qed.

Lemma ctx_binds_k_after_corr v vk t G G' e :
  ctx_ok G -> renv_corr G' e ->
  ctx_binds_k_after v vk t G G' ->
  renv_corr G e.
Proof.
  introv Hok Hcor Hbk. induction Hbk.
  - eapply renv_corr_push; eauto. inversions Hok. auto.
  - inversions Hok. specialize (IHHbk H2 Hcor).
    eapply renv_corr_push; eauto.
Qed.

Lemma renv_corr_epush_inv G e w y:
  ctx_ok G ->
  renv_corr G (e & w ~ y) ->
  renv_corr G e.
Proof.
  intros. rename H into Hok, H0 into H. dependent induction H.
  - exfalso. eapply empty_push_inv; eauto.
  - repeat (rewrite <- cons_to_push in x); inversions x.
    eauto using ctx_binds_k_after_corr.
Qed.

Lemma renv_corr_concat G1 G2 e:
  renv_corr G1 e ->
  ctx_weakens G2 ->
  ctx_ok (G1 §& G2) ->
  renv_corr (G1 §& G2) e.
Proof.
  intros.
  induction G2; inversions H0; auto.
  - rewrite ctx_concat_push in *.
    apply renv_corr_push.
    apply IHG2.
    inversion H3; auto.
    inversion H1; auto.
    eapply ctx_ok_push_inv_unbound; eauto.
Qed.

Hint Resolve renv_corr_push : core.
Hint Resolve renv_corr_concat : core.

(* Properties of corresponding types *)

Lemma typ_refcap_corr_subtyp_empty T1 T2:
  typ_refcap_corr T1 T2 ->
  ctx_empty ⁏ empty ⊢ T2 <: T1.
Proof.
  destruct (mode_refcap_corr' typing_mode); intros;
  destruct_all; subst; eauto.
Qed.

(* The type of a location is a subtype of a type of a reference *)
Lemma typ_corr_subtyp G e T1 T2:
  ctx_ok G ->
  ok e ->
  typ_corr T1 T2 ->
  G ⁏ e ⊢ T2 <: T1.
Proof.
  introv HokG Hoke Hcor. inversion Hcor.
  - (* obj *)
    iapply subtyp_and. eauto.
    destruct (mode_refcap_corr' typing_mode); subst; eauto.
    destruct_all; subst. rename x into U.
    eapply subtyp_typ; intros; eauto.
  - (* fun *)
    eauto.
Qed.

Lemma typ_corr_rev_subtyp G e T1 T2:
  mode_refcap_corr' typing_mode = mode_refcap_corr_same ->
  ctx_ok G ->
  ctx_weakens G ->
  ok e ->
  typ_corr T1 T2 ->
  G ⁏ e ⊢ T1 <: T2.
Proof.
  introv Hm HokG HwkG Hoke Hcor. inversion Hcor.
  iapply subtyp_and.
  - eauto.
  - rewrite <- ctx_concat_empty_l at 1.
    eapply weaken_end; eauto.
    eapply ewkn_empty; eauto.
    rewrite Hm in H.
    subst; easy.
    rewrite ctx_concat_empty_l.
    eauto.
  - eauto.
Qed.

(* Note: do not move this to Context.v, in case the definition of contexts
   is extended such that this only hold for inert contexts *)
Lemma ctx_binds_k_after_trans : forall v1 v2 vk1 vk2 T1 T2 G0 G1 G2,
    (* G₀ is an initial sequence of G₁ *)
    ctx_binds_k_after v1 vk1 T1 G1 G0 ->
    (* G₁ an initial sequence of G₂ *)
    ctx_binds_k_after v2 vk2 T2 G2 G1 ->
    (* Entire context is well-formed *)
    ctx_ok G2 ->
    (* G₀ is an initial sequence of G₂, because G₀ ⊂ G₁ ⊂ G₂ *)
    ctx_binds_k_after v1 vk1 T1 G2 G0.
Proof.
  introv Hbk10 Hbk12 Hin. (* Hbk2 is the chain G₁ ⊂ G₂ *)
  assert (ctx_ok G2) as Hok by auto.
  clear Hin.
  induction Hbk12.
  - constructor; auto.
    apply ctx_binds_k_after_binds in Hbk10.
    apply ctx_binds_dom in Hbk10.
    inversions Hok. unfold not. intros. subst. apply H4. auto.
  - assert (Hok1 : ctx_ok G1) by eauto;
      specialize (IHHbk12 Hbk10 Hok1); clear Hok1.
    constructor; auto.
    apply ctx_binds_k_after_binds in IHHbk12.
    apply ctx_binds_dom in IHHbk12.
    inversions Hok. unfold not. intros. subst. apply H5. auto.
Qed.

(* Finding types of corresponding variables in the typing context *)

Lemma renv_corr_binds_to_binds_after G1 e vR vL:
  ctx_ok G1 ->
  renv_corr G1 e ->
  binds vR vL e ->
  exists TL TR G2 G3,
    ctx_binds_k_after vR v_ref TR G1 G2 /\
    ctx_binds_k_after vL v_loc TL G2 G3 /\
    typ_corr TR TL.
Proof.
  intros Hok Hcor Hbi.
  induction Hcor.
  - false. eapply binds_empty_inv; eauto.
  - pose proof (binds_push_inv Hbi); destruct_all; subst.
    + exists Ty Tw G2 G3. easy.
    + assert (Hok' : ctx_ok G2) by eauto using ctx_binds_k_after_ok;
        specialize (IHHcor Hok H3); clear Hok'.
      destruct IHHcor as [TL [TR [G0 [G4 [HbaR [HbaL Htcor]]]]]].
      exists TL TR G0 G4. repeat(split); try assumption.
Qed.

Lemma renv_corr_binds_to_ref_loc G e vR vL:
  ctx_ok G ->
  renv_corr G e ->
  binds vR vL e ->
  (exists TL TR,
    ctx_binds vR TR G /\
    ctx_k G vR v_ref /\
    ctx_binds vL TL G /\
    ctx_k G vL v_loc /\
    typ_corr TR TL).
Proof.
  introv Hok Hcor Hbi.
  pose proof (renv_corr_binds_to_binds_after Hok Hcor Hbi).
  destruct_all. rename x into TL, x0 into TR, x1 into G1, x2 into G2. exists TL, TR.
  pose proof (ctx_binds_k_after_trans H0 H Hok).
  repeat split;
  eauto using ctx_binds_k_after_binds, ctx_binds_k_after_k.
Qed.
Lemma renv_corr_binds_dom G e w y:
  ctx_ok G ->
  renv_corr G e ->
  binds w y e ->
  w \in ctx_dom G /\ y \in ctx_dom G.
Proof.
  introv Hok Hcor Hbi.
  forwards Hrl: renv_corr_binds_to_ref_loc Hcor Hbi; eauto.
  destructecs Hrl.
  split;
  eapply ctx_binds_dom; eauto.
Qed.

End WithTypingMode.
Arguments renv_corr {typing_mode}.
Arguments typ_corr {typing_mode}.
Ltac destruct_renv_corr_binds_to_ref_loc H H0 H1 TL TR :=
  destruct (renv_corr_binds_to_ref_loc H H0 H1) as [TL [TR [HbiR [HvkR [HbiL [HvkL Hcor]]]]]].

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma renv_corr_binds_ref_to_loc G e vR vL TR:
  ctx_ok G ->
  renv_corr G e ->
  binds vR vL e ->
  ctx_binds vR TR G ->
  exists TL,
    (ctx_binds vL TL G /\
     ctx_k G vL v_loc /\
     typ_corr TR TL).
Proof.
  intros.
  destruct_renv_corr_binds_to_ref_loc H H0 H1  TL TR0.
  exists TL.
  split.
  - auto.
  - ctx_binds_eq; subst; auto.
Qed.

Lemma renv_corr_binds_loc_to_ref G e vR vL TL:
  ctx_ok G ->
  renv_corr G e ->
  binds vR vL e ->
  ctx_binds vL TL G ->
  (exists TR,
      ctx_binds vR TR G /\
      ctx_k G vR v_ref /\
      typ_corr TR TL).
Proof.
  intros.
  destruct_renv_corr_binds_to_ref_loc H H0 H1  TL0 TR.
  exists TR.
  split.
  - auto.
  - ctx_binds_eq; subst; auto.
Qed.

Lemma renv_corr_binds_ref_loc_corr G e vR vL TR TL:
  ctx_ok G ->
  renv_corr G e ->
  binds vR vL e ->
  ctx_binds vR TR G ->
  ctx_binds vL TL G ->
  typ_corr TR TL.
Proof.
  intros.
  destruct_renv_corr_binds_to_ref_loc H H0 H1 TL0 TR0.
  ctx_binds_eq; subst; auto.
  ctx_binds_eq; subst; auto.
Qed.

Lemma renv_corr_tyq_obj_loc_to_ref G e vR vL TL SL:
  ctx_ok G ->
  renv_corr G e ->
  typing_equiv e vR vL ->
  (
    ctx_binds vL (typ_objref TL SL) G ->
    exists SR, ctx_binds vR (typ_objref TL SR) G
  ) /\ (
    ctx_binds vR (typ_objref TL SL) G ->
    exists SR, ctx_binds vL (typ_objref TL SR) G
  ).
Proof.
  introv Hok Hecor Htyq.
  revert dependent SL.
  induction Htyq.
  - split; intros; exists SL; auto.
  - split; intros.
    + destruct (renv_corr_binds_loc_to_ref Hok Hecor H H0) as [TR [Hbi [HvkL Hcor]]].
      inversion Hcor; subst; auto.
      exists S; auto.
    + destruct (renv_corr_binds_ref_to_loc Hok Hecor H H0) as [TR [Hbi [HvkR Hcor]]].
      inversion Hcor; subst; auto.
      exists U; auto.
  - intro. split; intros; edestruct IHHtyq; eauto.
  - split; intros.
    + destruct (IHHtyq0 Hecor SL) as [[S ?] ?]; auto; edestruct (IHHtyq); eauto.
    + destruct (IHHtyq Hecor SL) as [? [S ?]]; auto; edestruct (IHHtyq0); eauto.
Qed.

Lemma renv_corr_tyq_fun_loc_to_ref G e vR vL TL SL:
  ctx_ok G ->
  renv_corr G e ->
  typing_equiv e vR vL ->
  (
    ctx_binds vL (typ_fun TL SL) G ->
    ctx_binds vR (typ_fun TL SL) G
  ) /\ (
    ctx_binds vR (typ_fun TL SL) G ->
    ctx_binds vL (typ_fun TL SL) G
  ).
Proof.
  introv Hok Hecor Htyq.
  induction Htyq.
  - split; intros; auto.
  - split; intros.
    + destruct (renv_corr_binds_loc_to_ref Hok Hecor H H0) as [TR [Hbi [HvkL Hcor]]].
      inversion Hcor; subst; auto.
    + destruct (renv_corr_binds_ref_to_loc Hok Hecor H H0) as [TR [Hbi [HvkR Hcor]]].
      inversion Hcor; subst; auto.
  - split; intros; destruct IHHtyq; auto.
  - split; intros;
    destruct IHHtyq0;
    destruct IHHtyq;
    auto.
Qed.


Lemma inert_typ_corr_refl T vk:
  mode_refcap_corr' typing_mode = mode_refcap_corr_same ->
  inert_typ T vk ->
  typ_corr T T.
Proof.
  introv Hm Hin.
  inversion Hin.
  - apply typ_corr_fun; auto.
  - apply typ_corr_obj.
    rewrite Hm. auto.
Qed.

Lemma inert_typ_corr_loc_refl T:
  inert_typ T v_loc ->
  typ_corr T T.
Proof.
  introv Hin.
  inversion Hin.
  - apply typ_corr_fun; auto.
  - apply typ_corr_obj.
hnf in H0.
  destruct (mode_refcap_corr' typing_mode) eqn: Hmc;
  destruct (mode_refcaps' typing_mode) eqn: Hm; try typing_mode_contradiction typing_mode.

auto.
subst.
inversions H0.
eauto.
Qed.


Lemma inert_type_corr_loc TR TL :
  inert_type TL -> typ_corr TR TL -> inert_typ TL v_loc.
Proof.
  introv [vk Hin] Hcorr. inversions Hcorr; eauto.
  destruct (mode_refcap_corr' typing_mode) eqn:Heqrf; subst.
  - inversions Hin. rename H1 into Hrcd, H2 into Hrf, H4 into Hindep.
    constructor; eauto. hnf.
    destruct (mode_refcaps' typing_mode) eqn:Heqrc; subst.
    2: { destruct typing_mode; simpl in *; try discriminate. }
    hnf in Hrf. rewrite Heqrc in Hrf. eauto.
  - destruct H as [Tbnd [HeqS HeqU]]. inversions Hin.
    rename H1 into Hrcd, H2 into Hrf, H4 into Hindep.
    constructor; eauto. hnf.
    destruct (mode_refcaps' typing_mode) eqn:Heqrc; subst; try constructor.
    destruct typing_mode; simpl in *; try discriminate.
Qed.

(* Equivalent variables and precise typing *)
(* Equivalent variables have equivalent function and precise types *)
Lemma renv_corr_pf_tyq G e v1 v2 U1 T1:
  inert G ->
  renv_corr G e ->
  typing_equiv e v1 v2 ->
  G ⊢! v1 ∶ U1 ⪼ T1 ->
  (fun_type T1 \/ record_type T1)->
  exists U2 T2, (
      G ⊢! v2 ∶ U2 ⪼ T2 /\
      typing_equiv e T1 T2
    ).
Proof.
  introv Hin Hcor Htyq Hpf Hrcd.
  assert (Hok : ctx_ok G) by eauto.
  assert (Hbin1 : ctx_binds v1 U1 G) by (eapply pf_binds; eauto).

  destructc (pf_inert_T Hin Hpf) as vk HinT.
  destruct HinT.
  * (* typ_fun *)
    assert (Hbin2 : ctx_binds v2 (typ_fun S T) G).
    { destruct (renv_corr_tyq_fun_loc_to_ref S T Hok Hcor Htyq) as [Hltr Hrtl];
        auto. }
    apply precise_flow_fun_inv in Hpf; subst.
    exists (typ_fun S T); eauto.
  * (* typ_objref *)
    assert (Hbin2 : exists S2, ctx_binds v2 (typ_objref T S2) G).
    { destruct (renv_corr_tyq_obj_loc_to_ref T S Hok Hcor Htyq) as [Hltr Hrtl];
        auto. }
    destruct Hrcd as [[? [? ?]] | ?]; subst.
    ** pose proof (pf_inert_fun_U Hin Hpf) as Hfalse. inversions Hfalse.
    ** dependent induction Hpf.
      - inversions H2. inversions H4. inversions H6.
      - destruct (pf_inert_bnd_U Hin Hpf). inversions H3.
        destructc Hbin2 as S2 Hbi2.
        exists (typ_and (typ_bnd U) S2) (open v2 U).
        split; eauto.
      -
        eapply pf_bnd_same_or_rcd in Hpf; eauto.
        branches; subst.
        + inversions H3; inversion H2. inversion H3.
        + inversion H3.
        +
          specialize (IHHpf Hin Hcor Htyq T S (eq_refl _)).
          specialize (IHHpf H3 Hok Hbin1 vk H H0 H1 Hbin2).
          destructc IHHpf as U0 T2 Hpf0 Htyq0.
          destruct (tyq_typ_andor_inv _ _ typ_andor_and Htyq0)
            as [T21 [T22 [HT2 [Htyq1 Htyq2]]]]; subst.
          exists U0 T21. eauto.
        + inversions H3. inversions H4.
          false.
          eapply record_refcap_false.
          apply H2. exists ls; auto.
      -
        eapply pf_bnd_same_or_rcd in Hpf; eauto.
        branches; subst.
        + inversions H3.
          eapply refcap_inert_refcap_typ in H0.
          pose proof (record_refcap_false H2 H0); contradiction.
        + inversion H3.
        +
          specialize (IHHpf Hin Hcor Htyq T S (eq_refl _)).
          specialize (IHHpf H3 Hok Hbin1 vk H H0 H1 Hbin2).
          destructc IHHpf as U0 T2 Hpf0 Htyq0.
          destruct (tyq_typ_andor_inv _ _ typ_andor_and Htyq0)
            as [T21 [T22 [HT2 [Htyq1 Htyq2]]]]; subst.
          exists U0 T22. eauto.
        + inversions H3. inversions H4.
          false.
          eapply record_refcap_false.
          apply H2. eexists. constructor; eauto.
Qed.

Lemma renv_corr_pf_normal_typ_tyq G e v1 v2 U1 tnl T11 T12:
  inert G ->
  renv_corr G e ->
  typing_equiv e v1 v2 ->
  G ⊢! v1 ∶ U1 ⪼ typ_rcd (dec_typ (typ_label_normal tnl) T11 T12) ->
  exists U2 T21 T22, (
      G ⊢! v2 ∶ U2 ⪼ typ_rcd (dec_typ (typ_label_normal tnl) T21 T22) /\
      (forall y, typing_equiv e (open y T11) (open y T21)) /\
      (forall y, typing_equiv e (open y T12) (open y T22))
    ).
Proof.
  introv Hin Hcor Htyqv Hpf.
  assert (record_type (typ_rcd (dec_typ (typ_label_normal tnl) T11 T12))).
  { destruct (pf_inert_rcd_or_refcap Hin Hpf).
    - eauto.
    - inversions H. inversions H0. inversions H1. }
  destructc (renv_corr_pf_tyq Hin Hcor Htyqv Hpf) as T' U' Hpf' Htyq'.
  auto.
  destructc (tyq_typ_dec_typ_inv Htyq') as T21 T22 HT2 Htyq1 Htyq2.
  subst.
  exists T' T21 T22.
  auto.
Qed.

Lemma pf_tyq_two_upper_tyq G e x x' T T' L L' U U' l:
  inert G ->
  renv_corr G e ->
  typing_equiv e x x' ->
  G ⊢! x  ∶ T  ⪼ typ_rcd (dec_typ (typ_label_normal l) L  U) ->
  G ⊢! x' ∶ T' ⪼ typ_rcd (dec_typ (typ_label_normal l) L' U') ->
  (forall y, typing_equiv e (open y U) (open y U')).
Proof.
  introv Hin Hcor Htyqv Hpf Hpf'.
  destruct (renv_corr_pf_normal_typ_tyq Hin Hcor Htyqv Hpf)
    as [t [tl [tu [Hpf0 [Htyq1 Htyq2]]]]].
  destruct (pf_inert_unique_tight_bounds Hin Hpf' Hpf0); subst.
  auto.
Qed.


Lemma renv_fv_ref G e:
  ctx_ok G ->
  renv_corr G e ->
  fv_env e \c ctx_kdom v_loc G.
Proof.
  introv Hok Hcor.
  intro x. intro Hin.
  induction Hcor.
  - unfold fv_env in Hin. rewrite empty_def in Hin. rewrite values_def in Hin. compute in Hin.
    rewrite in_empty in Hin. contradiction.
  - destruct (ctx_binds_k_after_initial_concat H); subst.
    rewrite env_push_fv in Hin.
    rewrite in_union in Hin.
    destruct Hin.
    + rewrite ctx_kdom_concat. rewrite in_union.
      specialize (IHHcor Hok H2). rewrite ctx_kdom_concat in IHHcor.
      rewrite in_union in *; eauto.
    + unfold fv, VarFreeVar in H2. rewrite in_singleton in H2; subst.
      rewrite ctx_kdom_concat. rewrite in_union. left.
      eapply ctx_k_in_kdom.
      eapply ctx_binds_k_after_k; eauto.
Qed.


Lemma renv_corr_idempotent G e:
  inert G ->
  renv_corr G e ->
  dom e \c ctx_kdom v_ref G ->
  renv_idempotent e.
Proof.
  introv Hok Hcor Hdom.
  unfold renv_idempotent; intros.
  assert (fv_env e \c ctx_kdom v_loc G) by (eapply renv_fv_ref; eauto).
  intro.
  apply H0 in H1.
  apply Hdom in H.
  eapply ctx_in_kdom_k in H; eauto.
  eapply ctx_in_kdom_k in H1; eauto.
  assert (v_ref <> v_loc) as Hvk by discriminate.
  eapply (ctx_k_ne_false Hvk H H1); eauto.
Qed.

Lemma renv_corr_binds_loc G e w y :
  ctx_ok G -> renv_corr G e -> binds w y e -> ctx_k G y v_loc.
Proof.
  introv Hok Hcor Hbi. induction Hcor.
  - apply binds_empty_inv in Hbi. tauto.
  - apply binds_push_inv in Hbi.
    destruct Hbi as [[? ?] | [Hneq Hbi]]; subst.
    + hnf. pose proof (ctx_binds_k_after_trans H0 H) as Haf.
      exists Ty G3. apply Haf. auto.
    + assert (Hok2 : ctx_ok G2) by eauto using ctx_binds_k_after_ok.
      eauto.
Qed.

Lemma typ_corr_subst T1 T2 x y :
  typ_corr T1 T2 ->
  typ_corr (T1 /[ x -> y]) (T2 /[ x -> y]).
Proof.
  introv Htycor. inversions Htycor.
  - destruct (mode_refcap_corr' typing_mode) eqn:Hm; subst; eauto.
    + simpl. econstructor; eauto. rewrite Hm. auto.
    + simpl. destruct_all; subst. econstructor; eauto.
      repeat(rewrite Hm). eexists; split; simpl; eauto.
  - simpl. econstructor; eauto.
Qed.

Lemma renv_corr_eremove G e1 w y e2 :
  ctx_weakens G -> ctx_ok G ->
  renv_corr G (e1 & w ~ y & e2) ->
  renv_corr G (e1 & e2).
Proof.
  reverts G e1 w y. reverts e2.
  apply (env_ind (fun e2 => forall (G : ctx) (e1 : env var) (w y : var),
                      ctx_weakens G -> ctx_ok G ->
                      renv_corr G (e1 & w ~ y & e2) ->
                      renv_corr G (e1 & e2))).
  - introv Hwk Hok Hcor. rewrite concat_empty_r in *. inversions Hcor.
    + false. eauto using empty_push_inv.
    + apply eq_push_inv in H; destruct_all; subst.
      apply ctx_binds_k_after_initial_concat in H0. destruct_all; subst.
      rename x into G2'. auto.
  - intros e w0 y0 IH G e0 w y. introv Hwk Hok Hcor.
    rewrite concat_assoc in *. inversions Hcor.
    1: { false. eauto using empty_push_inv. }
    rename H0 into Hbkw1, H1 into Hbky1, H2 into Htycor, H4 into Hecor,
      Ty into Ty0.
    rename H into Heqenv. apply eq_push_inv in Heqenv.
    destruct_all; subst. eapply renv_corr_ref; eauto.
Qed.

Lemma renv_corr_ref_loc_fv G e w y Tw Ty :
  fresh_telescope G ->
  renv_corr G e -> binds w y e ->
  ctx_binds w Tw G -> ctx_binds y Ty G ->
  w \notin fv Ty.
Proof.
  introv HftG Hcor Hbie HbiwG HbiyG Hin.
  assert (HokG : ctx_ok G) by eauto using fresh_telescope_ok.
  assert (HwkG : ctx_weakens G) by eauto using fresh_telescope_weakens.

  eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie; eauto.
  destruct Hbie as [TL [TR [G2 [G3 [Hbkw [Hbky _]]]]]].
  eapply ctx_binds_k_after_initial_segment in Hbkw.
  destruct Hbkw as [G5 ?]; subst.
  eapply ctx_binds_k_after_initial_segment in Hbky.
  destruct Hbky as [G4 ?]; subst.

  assert (HbiwG' : ctx_binds w TR
                     (((G3 § y / v_loc ~ TL) §& G4 § w / v_ref ~ TR) §& G5)).
  { eapply ctx_binds_concat_left; eauto using ctx_weakens_concat_right.
    eapply ctx_ok_middle_inv in HokG. tauto. }
  assert (HbiyG' : ctx_binds y TL
                     (((G3 § y / v_loc ~ TL) §& G4 § w / v_ref ~ TR) §& G5)).
  { rewrite <- ctx_concat_push, <- ctx_concat_assoc in *.
    eapply ctx_binds_concat_left; eauto using ctx_weakens_concat_right.
    eapply ctx_ok_middle_inv in HokG. tauto. }

  rewrite (ctx_binds_functional HbiwG' HbiwG) in *; clear HbiwG' HbiwG.
  rewrite (ctx_binds_functional HbiyG' HbiyG) in *; clear HbiyG' HbiyG.

  apply fresh_telescope_concat_invert_l in HftG.
  inversions HftG. clear H6 H4 H3. rename H5 into Hfrw.
  rewrite ctx_fv_concat in Hfrw; simpl in Hfrw.
  repeat(rewrite notin_union in Hfrw). easy.
Qed.

End WithTypingMode.

Hint Constructors renv_corr : core.
Hint Resolve renv_corr_push : core.
Hint Resolve renv_corr_concat : core.

Notation env_binds_more e e' :=
  (forall y w, binds w y e -> binds w y e') (only parsing).

Lemma env_binds_more_permute {A} (e e' : env A) :
  env_binds_more e  e' ->
  env_binds_more e' e ->
  (dom e = dom e' /\ (forall x a, binds x a e <-> binds x a e')).
Proof.
  introv Hbm Hbm'. split.
  - apply fset_extens. all: introv Hin; apply dom_to_binds in Hin.
    all: destruct Hin as [a Hbi]. all: eapply binds_to_dom. all: eauto.
  - intros x a. split; eauto.
Qed.

Lemma ebinm_tyq_var e e' (v1 v2 : var) :
  env_binds_more e e' ->
  e ⊢ v1 ≈ v2 ->
  e' ⊢ v1 ≈ v2.
Proof. introv Hbm Htyq. induction Htyq; simpl; eauto. Qed.

Lemma ebinm_tyq_typ e e' T U :
  env_binds_more e e' ->
  e ⊢ T ≈ U ->
  e' ⊢ T ≈ U.
Proof. introv Hbm Htyq. induction Htyq; simpl; eauto using ebinm_tyq_var. Qed.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma ebinm_ts_ind :
  (forall (G : ctx) (Go : objctx) (e : env var) (v : var) (T : typ),
      ty_var typing_mode G Go e v T ->
      (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ v ∶ T))
  /\ (forall (G : ctx) (Go : objctx) (e : env var) (S T : typ),
        subtyp typing_mode G Go e S T ->
        (forall e', env_binds_more e e' -> G ⁏ e' ⊢ S <: T))
  /\ (forall (G : ctx) (Go : objctx) (e : env var) (T Ti : typ),
        ty_incap typing_mode G Go e T Ti ->
       (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ T ∶icp Ti)).
Proof. apply ts_mutind; intros; eauto using ebinm_tyq_typ. Qed.

Lemma ebinm_ty_var G Go e e' (v : var) T :
  G §§ Go ⁏ e ⊢ v ∶ T ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ v ∶ T.
Proof. intros. pose proof ebinm_ts_ind. destruct_all. eauto. Qed.

Lemma ebinm_subtyp G e e' S T :
  G ⁏ e ⊢ S <: T ->
  env_binds_more e e' ->
  G ⁏ e' ⊢ S <: T.
Proof. intros. pose proof ebinm_ts_ind. destruct_all. eauto. Qed.

Lemma ebinm_incap G Go e e' T Ti :
  G §§ Go ⁏ e ⊢ T ∶icp Ti ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ T ∶icp Ti.
Proof. intros. pose proof ebinm_ts_ind. destruct_all. eauto. Qed.

Lemma ebinm_capbnd_ind C G Go e e' T U :
  G §§ Go ⁏ e ⊢ T ∶cpb⟦C⟧ U ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ T ∶cpb⟦C⟧ U.
Proof. induction 1; intros; eauto using ebinm_ty_var. Qed.

Lemma ebinm_adapt_cap C G e e' x T U :
  G ⁏ e ⊢ x ▷⟦C⟧ T ⟶ U ->
  env_binds_more e e' ->
  G ⁏ e' ⊢ x ▷⟦C⟧ T ⟶ U.
Proof.
  induction 1; intros; constructor; eauto using ebinm_incap, ebinm_capbnd_ind.
Qed.

Lemma ebinm_adapt G e e' x T U :
  G ⁏ e ⊢ x ▷ T ⟶ U ->
  env_binds_more e e' ->
  G ⁏ e' ⊢ x ▷ T ⟶ U.
Proof.
  intros. hnf in *. pose proof ebinm_ts_ind. destruct_all.
  destruct (mode_viewpoint_adaptation' typing_mode) eqn:Hm;
    eauto using ebinm_adapt_cap.
Qed.

Lemma ebinm_rules_ind :
  (forall (G : ctx) (Go : objctx) (e : env var) (t : trm) (T : typ),
      ty_trm typing_mode G Go e t T ->
      (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ t ∶ T))
  /\ (forall (G : ctx) (Go : objctx) (e : env var) (l : lit) (T : typ),
        ty_lit typing_mode G Go e l T ->
        (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ l ∶ T))
  /\ (forall (G : ctx) (Go : objctx) (e : env var) (d : def) (D : dec),
        ty_def typing_mode G Go e d D ->
        (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ d ∶d D))
  /\ (forall (G : ctx) (Go : objctx) (e : env var) (ds : defs) (Ds : typ),
        ty_defs typing_mode G Go e ds Ds ->
        (forall e', env_binds_more e e' -> G §§ Go ⁏ e' ⊢ ds ∶ Ds)).
Proof.
  apply rules_mutind; intros;
    eauto 4 using
          ebinm_ty_var, ebinm_subtyp, ebinm_incap, ebinm_adapt, ebinm_tyq_typ;
    try solve[econstructor; eauto; intros; eauto using ebinm_ty_var].
Qed.

Lemma ebinm_ty_trm G Go e e' (t : trm) T :
  G §§ Go ⁏ e ⊢ t ∶ T ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ t ∶ T.
Proof. intros. pose proof ebinm_rules_ind. destruct_all. eauto. Qed.

Lemma ebinm_ty_lit G Go e e' (l : lit) T :
  G §§ Go ⁏ e ⊢ l ∶ T ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ l ∶ T.
Proof. intros. pose proof ebinm_rules_ind. destruct_all. eauto. Qed.

Lemma ebinm_ty_def G Go e e' (d : def) D :
  G §§ Go ⁏ e ⊢ d ∶d D ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ d ∶d D.
Proof. intros. pose proof ebinm_rules_ind. destruct_all. eauto. Qed.

Lemma ebinm_ty_defs G Go e e' (ds : defs) Ds :
  G §§ Go ⁏ e ⊢ ds ∶ Ds ->
  env_binds_more e e' ->
  G §§ Go ⁏ e' ⊢ ds ∶ Ds.
Proof. intros. pose proof ebinm_rules_ind. destruct_all. eauto. Qed.

Lemma env_binds_more_trans {A} (e1 e2 e3 : env A) :
  env_binds_more e1 e2 ->
  env_binds_more e2 e3 ->
  env_binds_more e1 e3.
Proof. intros; eauto. Qed.

Lemma env_binds_more_push {A} (e : env A) x a :
  x # e ->
  env_binds_more e (e & x ~ a).
Proof.
  introv Hfrx Hbi. apply binds_push_neq; eauto.
  apply binds_to_dom in Hbi. hnf. intros; subst. eauto.
Qed.

Lemma renv_corr_binds_more G e e' :
  ctx_ok G -> ctx_weakens G -> ok e' ->
  renv_corr G e ->
  env_binds_more e' e ->
  renv_corr G e'.
Proof.
  reverts e' e G.
  apply (env_ind (fun e' => forall e G,
                    ctx_ok G ->
                    ctx_weakens G ->
                    ok e' ->
                    renv_corr G e ->
                    env_binds_more e' e ->
                    renv_corr G e'));
    try solve [ intros; eauto ].

  intros e' w y. introv IH HokG HwkG Hoke' Hecor Hbm.

  inversions Hoke'.
  1: { false. eauto using empty_push_inv. }
  apply eq_push_inv in H. destruct_all; subst.
  rename H0 into Hoke', H1 into Hwfr.

  assert (Hbm' : env_binds_more e' e).
  { eapply env_binds_more_trans; try(eapply env_binds_more_push); eauto. }
  specialize (IH _ _ HokG HwkG Hoke' Hecor Hbm').

  specialize (Hbm _ _ (binds_push_eq w y e')).
  apply (renv_corr_binds_to_binds_after HokG Hecor) in Hbm.
  destruct Hbm as [Ty [Tw [Gw [Gy [Hbkw [Hbky Htycor]]]]]]. eauto.
Qed.

End WithTypingMode.
