(** remove printing ~ *)

(** This module contains lemmas related to invertible typing
    ([ty_var_inv], [⊢##] and [ty_item_inv], [⊢##v]). *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx.
Require Import Incap.

Require Import Coq.Bool.Bool LibBoolExtra.



(** ** Flat Invertible typing *)

(* This is the original "flat" version that requires the possibility to invert typing of union types.
   It is used in mode mode_incap_structural, because the newer layered typing only works with mode_incap_nocap *)

(** The invertible-typing relation describes the possible types that a variable or value
can be typed with in an inert context. For example, if [G] is inert, [G ⊢! x: {a: T}],
and [G ⊢ T <: T'], then [G ⊢## x: {a: T'}].

The purpose of invertible typing is to be easily invertible into a precise typing relation.
To achieve that, invertible typing avoids typing cycles that could result from, for example,
repeated applications of recursion introduction and elimination.
For this case, invertible typing defines only recursion introduction (whereas precise typing
defines only recursion elimination). *)

(** ** Invertible typing of variables [G ⊢## x: T] *)

Section WithTypingMode.
Context {typing_mode: Mode}.

Reserved Notation "G '⁏' e '⊢##f' x '∶' T" (at level 40, e at level 39, x at level 59).

Inductive ty_var_inv : ctx -> renv -> var -> typ -> Prop :=

(** [G ⊢! x∶ T]  #<br>#
    [―――――――――――] #<br>#
    [G ⊢## x∶ T]     *)
| ty_precise_inv : forall G e x T U V,
  G ⊢! x ∶ T ⪼ U ->
  typing_equiv e U V ->
  G⁏e ⊢##f x ∶ V

(** [G ⊢## x∶ {a: TL1..TU1}] #<br>#
    [G ⊢# TL2 <: TL1]     #<br>#
    [G ⊢# TU1 <: TU2]     #<br>#
    [――――――――――――――――] #<br>#
    [G ⊢## x∶ {a: TL2..TU2}]     *)
| ty_dec_fld_inv : forall G e x a TL1 TL2 TU1 TU2,
  G⁏e ⊢##f x ∶ typ_rcd (dec_fld a TL1 TU1) ->
  G⁏e ⊢# TL2 <: TL1 ->
  G⁏e ⊢# TU1 <: TU2 ->
  G⁏e ⊢##f x ∶ typ_rcd (dec_fld a TL2 TU2)

(** [G ⊢## x∶ {A: T..U}]   #<br>#
    [G ⊢# T' <: T]         #<br>#
    [G ⊢# U <: U']         #<br>#
    [―――――――――――――――――――――] #<br>#
    [G ⊢## x∶ {A: T'..U'}]     *)
| ty_dec_typ_inv : forall G e x A TL1 TL2 TU1 TU2,
  G⁏e ⊢##f x ∶ typ_rcd (dec_typ A TL1 TU1) ->
  (forall y, G⁏e ⊢# open y TL2 <: open y TL1) ->
  (forall y, G⁏e ⊢# open y TU1 <: open y TU2) ->
  G⁏e ⊢##f x ∶ typ_rcd (dec_typ A TL2 TU2)

(** [G ⊢## x∶ T^x]   #<br>#
    [T indep] #<br>#
    [―――――――――――――――] #<br>#
    [G ⊢## x∶ mu(T)] *)
| ty_bnd_inv : forall G e x T,
  typ_indep_b 0 T ->
  ((mode_has_incap typing_mode) -> (G⁏e ⊢ (open x T) ∶icp (open x T))) ->
  G⁏e ⊢##f x ∶ open x T ->
  G⁏e ⊢##f x ∶ typ_bnd T

(** [G ⊢## x∶ forall(S)T]          #<br>#
    [G ⊢# S' <: S]            #<br>#
    [G, y: S' ⊢ T^y <: T'^y]   #<br>#
    [y fresh]                  #<br>#
    [――――――――――――――――――――――]   #<br>#
    [G ⊢## x∶ forall(S')T']            *)
| ty_fun_inv : forall L G e x S T S' T',
  mode_has_fun typing_mode ->
  G⁏e ⊢##f x ∶ typ_fun S T ->
  G⁏e ⊢# S' <: S ->
  (forall y, y \notin L ->
   G § y / v_par ~ S' ⁏e ⊢ open y T <: open y T') ->
  G⁏e ⊢##f x ∶ typ_fun S' T'

(** [G ⊢## x ∶ T]     #<br>#
    [G ⊢## x ∶ U]     #<br>#
    [――――――――――――――――] #<br>#
    [G ⊢## x ∶ T /\ U]      *)
| ty_and_inv : forall G e x S1 S2,
  G⁏e ⊢##f x ∶ S1 ->
  G⁏e ⊢##f x ∶ S2 ->
  G⁏e ⊢##f x ∶ typ_and S1 S2

(** [G ⊢## x ∶ T]     #<br>#
    [――――――――――――――――] #<br>#
    [G ⊢## x ∶ T \/ U]      *)
| ty_or1_inv : forall G e x S1 S2,
  G⁏e ⊢##f x ∶ S1 ->
  G⁏e ⊢##f x ∶ typ_or S1 S2

(** [G ⊢## x ∶ U]     #<br>#
    [――――――――――――――――] #<br>#
    [G ⊢## x ∶ T \/ U]      *)
| ty_or2_inv : forall G e x S1 S2,
  G⁏e ⊢##f x ∶ S2 ->
  G⁏e ⊢##f x ∶ typ_or S1 S2


(** [G ⊢## x∶ S1^z]        #<br>#
    [G ⊢! y∶ {A: S1..S2}] #<br>#
    [――――――――――――――――――] #<br>#
    [G ⊢## x∶ y.A(z)           *)
| ty_sel_inv : forall G e x y A z S1 S2 U,
  G⁏e ⊢##f x ∶ open z S1 ->
  G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
  G⁏e ⊢##f x ∶ typ_sel (avar_f y) A (avar_f z)

(** [G ⊢## x∶ T]   #<br>#
    [―――――――――――――] #<br>#
    [G ⊢## x∶ top]     *)
| ty_top_inv : forall G e x T,
  G⁏e ⊢##f x ∶ T ->
  G⁏e ⊢##f x ∶ typ_top

| ty_dec_met_inv : forall L G e x m TC1 TZ1 TR1 TC2 TZ2 TR2,
  G⁏e ⊢##f x ∶ typ_rcd (dec_met m TC1 TZ1 TR1)  ->
  G⁏e ⊢# TZ2 <: TZ1 ->
  (forall y, y \notin L -> (* y is the parameter *)
   G § y / v_par ~ TZ2 ⁏e ⊢ open y TC2 <: open y TC1) ->
  (* If typ_indep_b 0 TC2 then typ_indep_b 0 TC1 in modes with sef *)
  (mode_has_sef typing_mode -> typ_indep_b 0 TC2 -> typ_indep_b 0 TC1) ->
  (forall y, y \notin L ->
   forall z, z <> y -> z \notin L -> (* z is the receiver *)
   G § y / v_par ~ TZ2 § z / v_recv ~ (open y TC2) ⁏e ⊢ open y (open_rec 1 z TR1) <: open y (open_rec 1 z TR2) ) ->
  G⁏e ⊢##f x ∶ typ_rcd (dec_met m TC2 TZ2 TR2)

where "G '⁏' e '⊢##f' x '∶' T" := (ty_var_inv G e x T).

Hint Constructors ty_var_inv : core.

Lemma precise_to_invertible_flat G e x T U:
  G ⊢! x ∶ T ⪼ U ->
  G⁏e ⊢##f x ∶ U.
Proof.
  eauto.
Qed.

(** ** Invertible to Precise Typing [|-## to |-!] *)

(** Invertible-to-precise typing for field declarations: #<br>#
    [G ⊢## x∶ {a: T}]            #<br>#
    [――――――――――――――――――――――]      #<br>#
    [exists T', G ⊢! x∶ {a: T'}]      #<br>#
    [G ⊢# T' <: T]. *)
Lemma invertible_flat_to_precise_fld_dec: forall G e x a S T,
  G⁏e ⊢##f x ∶ typ_rcd (dec_fld a S T) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_fld a S' T') /\
    G⁏e ⊢# T' <: T /\
    G⁏e ⊢# S <: S'.
Proof.
  introv Hinv.
  dependent induction Hinv.
  - apply tyq_typ_dec_fld_invr in H0.
    destruct H0 as [t21 [t22 [Ht [Htyq1 Htyq2]]]]; subst.
    exists t21 t22 T0.
    apply tyq_typ_sym in Htyq1.
    split; auto; split; auto.
  - specialize (IHHinv _ _ _ eq_refl). destruct IHHinv as [V [V' [V'' [Hx [Hs1 Hs2]]]]].
    exists V V' V''.
    splits; auto; eapply subtyp_trans_t; eassumption.
Qed.

Lemma invertible_flat_to_precise_typ_dec: forall G e x a S T,
  G⁏e ⊢##f x ∶ typ_rcd (dec_typ a S T) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ a S' T') /\
    (forall y, G⁏e ⊢# open y T' <: open y  T) /\
    (forall y, G⁏e ⊢# open y S <: open y  S').
Proof.
  introv Hinv.
  dependent induction Hinv.
  - symmetry in H0.
    apply tyq_typ_dec_typ_inv in H0.
    destruct H0 as [t21 [t22 [Ht [Htyq1 Htyq2]]]]; subst.
    exists t21 t22 T0.
    split; auto; split; intros.
    specialize (Htyq1 y).
    specialize (Htyq2 y).
    apply tyq_typ_sym in Htyq2; eauto.
    eauto.
  - specialize (IHHinv _ _ _ eq_refl).
    destructc IHHinv as V V' V'' Hx Hs1 Hs2.
    exists V V' V''.
    splits; auto; intros.
    eapply subtyp_trans_t; try eapply Hs1; eauto.
    eapply subtyp_trans_t; try eapply Hs2; eauto.
Qed.

Lemma invertible_flat_to_precise_met_dec: forall G e x m TC TZ TR,
  inert G -> ok e ->
  G⁏e ⊢##f x ∶ typ_rcd (dec_met m TC TZ TR) ->
  exists L S TC' TZ' TR',
    G ⊢! x ∶ S ⪼ typ_rcd (dec_met m TC' TZ' TR') /\
    G⁏e ⊢# TZ <: TZ' /\
    (forall z, z \notin L -> G § z / v_par ~ TZ ⁏e ⊢ (open z TC) <: (open z TC')) /\
    (mode_has_sef typing_mode -> typ_indep_b 0 TC -> typ_indep_b 0 TC') /\
    (forall z, z \notin L -> forall y, y <> z -> y \notin L -> ctx_met G z TZ y TC ⁏e ⊢ (open_met z y TR') <: (open_met z y TR))
.
Proof.
  introv HG He Hinv.
  dependent induction Hinv.
  - apply tyq_typ_sym in H0.
    apply tyq_typ_dec_met_inv in H0.
    destruct H0 as [t21 [t22 [t23 [Ht [Htyq1 [Htyq2 Htyq3]]]]]]; subst.
    exists (ctx_dom G) T t21 t22 t23.
    splits; eauto using typ_indep_tyq_open_all_indep.
    intros.
    specialize (Htyq1 z).
    specialize (Htyq3 z y).
    symmetry in Htyq3.
    eauto.
  - specialize (IHHinv _ _ _ _ HG He eq_refl).
    destruct IHHinv
      as [L' [S' [TC3 [TZ3 [TR3 [Hx [Hs1 [Hs2 [Hidp Hs3]]]]]]]]].
    exists (ctx_dom G \u L \u L') S' TC3 TZ3 TR3.
    splits; auto.
    + eapply subtyp_trans_t; eassumption.
    + intros z' Fr.
      assert (Hok: ctx_ok (G § z' / v_par ~ TZ)) by eauto.
      apply tight_to_general in H; auto.
      assert (Hnarrow: G § z' / v_par ~ TZ ⁏e ⊢ open z' TC <: open z' TC1).
      { eapply narrow; auto using subctx_last. }
      eauto using narrow_subtyping.
    + intros z' Frz y' Nyz Fry.
      assert (Hok: ctx_ok (ctx_met G z' TZ y' TC)) by eauto using ctx_unbound_push.
      apply tight_to_general in H; auto.
      assert (Hnarrow: (ctx_met G z' TZ y' TC) ⁏e ⊢ open_met z' y' TR3 <: open_met z' y' TR1).
      { apply narrow with (G0:= (ctx_met G z' TZ1 y' TC1)); auto using subctx_last2. }
      eauto.
Qed.

(** Invertible-to-precise typing for function types: #<br>#
    [ok G]                        #<br>#
    [G ⊢## x∶ forall(S)T]             #<br>#
    [――――――――――――――――――――――――――]  #<br>#
    [exists S', T'. G ⊢! x∶ forall(S')T']  #<br>#
    [G ⊢# S <: S']               #<br>#
    [G ⊢# T'^y <: T^y], where [y] is fresh. *)
Lemma invertible_flat_to_precise_typ_fun: forall G e x S T,
  inert G -> ok e ->
  G⁏e ⊢##f x ∶ typ_fun S T ->
  exists S' T' U L,
    G ⊢! x ∶ U ⪼ typ_fun S' T' /\
    G⁏e ⊢# S <: S' /\
    (forall y,
        y \notin L ->
            G § y / v_par ~ S ⁏e ⊢ open y T' <: open y T).
Proof.
  introv HG HE Hinv.
  dependent induction Hinv.
  - destructc (tyq_typ_fun_invr H0) as t21 t22 Ht Htyq1 Htyq2; subst.
    exists t21 t22 T0 (ctx_dom G).
    apply tyq_typ_sym in Htyq1.
    split; auto.
    split; auto.
    intros.
    specialize (Htyq2 y ).
    symmetry in Htyq2.
    eauto.
  - specialize (IHHinv _ _ HG HE eq_refl).
    destruct IHHinv as [S' [T' [V' [L' [Hpt [HSsub HTsub]]]]]].
    exists S' T' V' (ctx_dom G \u L \u L').
    split; auto.
    assert (Hsub2 : G⁏e ⊢# typ_fun S0 T0 <: typ_fun S T).
    { apply subtyp_fun_t with (L:=L); assumption. }
    split.
    + eapply subtyp_trans_t; eauto.
    + intros y Fr.
      assert (Hok: ctx_ok (G § y / v_par ~ S)) by eauto.
      apply tight_to_general in H0; auto.
      assert (Hnarrow: G § y / v_par ~ S ⁏e ⊢ open y T' <: open y T0).
      { eapply narrow; auto using subctx_last. }
      eauto.
Qed.

Lemma invertible_flat_bot_false : forall G e x,
    inert G ->
    G⁏e ⊢##f x ∶ typ_bot ->
    False.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  apply tyq_typ_sym in H0.
  pose proof (tyq_typ_nulary_inv typ_nulary_bot H0); subst.
  eapply pf_bot_false; eauto.
Qed.

Lemma invertible_flat_nocap_false : forall G e x,
    inert G ->
    G⁏e ⊢##f x ∶ typ_nocap ->
    False.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  apply tyq_typ_sym in H0.
  pose proof (tyq_typ_nulary_inv typ_nulary_nocap H0); subst.
  eapply pf_nocap_false; eauto.
Qed.

Lemma invertible_flat_and_invl G e x T1 T2:
    inert G ->
    G⁏e ⊢##f x ∶ typ_and T1 T2 ->
    G⁏e ⊢##f x ∶ T1.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  inversions_tyq H0. eapply pf_and1 in H. eauto.
  eauto.
Qed.
Lemma invertible_flat_and_invr G e x T1 T2:
    inert G ->
    G⁏e ⊢##f x ∶ typ_and T1 T2 ->
    G⁏e ⊢##f x ∶ T2.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  inversions_tyq H0. eapply pf_and2 in H. eauto.
  eauto.
Qed.



(** ** Invertible Subtyping Closure *)

(** Invertible typing is closed under equivalence. *)
Lemma invertible_flat_typing_closure_tyq G e x T U:
  (mode_incap' typing_mode = mode_incap_structural \/ mode_incap' typing_mode = mode_incap_none) ->
  inert G ->
  renv_corr G e ->
  G⁏e ⊢##f x ∶ T ->
  typing_equiv e T U ->
  G⁏e ⊢##f x ∶ U.
Proof.
  introv Hm Hi Hcor HT Htyq.
  revert Htyq.
  revert U.
  induction HT; intros U0 Htyq.
  - set (tyq_typ_trans H0 Htyq).
    eapply ty_precise_inv; eauto.
  - inversions_tyq Htyq.
    symmetry in H6.
    eapply ty_dec_fld_inv; eauto.
  - inversions_tyq Htyq.
    symmetry in H6.
    eapply ty_dec_typ_inv; eauto.
  - apply tyq_typ_bnd_inv in Htyq; auto.
    destruct Htyq as [t21 [HT2 Htyq21]]; subst.
    assert (G ⁏ e ⊢##f x ∶ open x t21).
    apply IHHT; auto.

    apply ty_bnd_inv. eapply typ_indep_tyq_open_all_indep; eauto.
    (* In rodot, the read-only condition in recursive types is a bit ugly here *)
    {
      intro.
      specialize (H0 H2).
      specialize (Htyq21 x).
      eapply ty_icp_opened_tyq; eauto.
      destruct Hm; auto. unfold mode_has_incap' in H2. rewrite H3 in H2. discriminate.
    }
    auto.
  - destruct (tyq_typ_fun_inv Htyq) as [t11 [t12 [Ht1 [Htyq11 Htyq12]]]]; subst.
    apply tyq_typ_sym in Htyq11.
    eapply ty_fun_inv; eauto.
    intros.
    eapply (subtyp_trans).
    eapply narrow.
    apply (H1 y).
    apply notin_union_r2 with (E:= ctx_dom G). eauto.
    eauto.
    eauto using subctx_last.
    apply subtyp_tyq; eauto.
  - inversions_tyq Htyq. auto.
  - inversions_tyq Htyq. auto.
  - inversions_tyq Htyq. auto.
  - apply tyq_typ_sel_inv in Htyq.
    destruct Htyq as [[v21 [v22 [tnl [Htnl [HU [? Htyq2]]]]]] | HT2]; subst.
    * simplify_eq Htnl; intro; subst.
      destruct (renv_corr_pf_normal_typ_tyq Hi Hcor Htyq2 H).
      destruct H0 as [T21 [T22 [Hpf [Htyq11 Htyq12]]]].
      eapply ty_sel_inv.
      apply IHHT. auto. auto. apply Htyq11.
      apply Hpf.
    * eapply ty_sel_inv; eauto.
  - assert (typ_top = U0). eapply (tyq_typ_nulary_inv); eauto.
    subst. eapply ty_top_inv. eauto.
  - destruct (tyq_typ_dec_met_inv Htyq) as [t11 [t12 [t13 [Ht [Htyq1 [Htyq2 Htyq3]]]]]]; subst.
    eapply ty_dec_met_inv.
    * apply HT.
    * apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans_t.
      apply subtyp_tyq_t. eauto. eauto.
    * intros.
      specialize (Htyq1 y).
      apply tyq_typ_sym in Htyq1.
      apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans.
      apply subtyp_tyq. eauto.
      eapply narrow.
      apply H0.
      apply notin_union_r2 with (E:= ctx_dom G). eauto.
      eauto. eauto.
    * (* Caused by modified method subtyping for sef *)
      intros.
      eapply H1. eauto.
      eapply typ_indep_tyq_open_all_indep; eauto. intros. symmetry. eapply Htyq1.
    * intros.
      specialize (Htyq3 y z).
      specialize (Htyq1 y).
      apply tyq_typ_sym in Htyq1.
      apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans.
      eapply narrow. auto. eauto.
      constructor; eauto.
      eauto.
Qed.

(* Lemma 14 in "Pure methods for roDOT" *)
Lemma invertible_flat_typing_or_inv G e x T U:
  (mode_incap' typing_mode = mode_incap_structural \/ mode_incap' typing_mode = mode_incap_none) ->
  inert G ->
  renv_corr G e ->
  G⁏e ⊢##f x ∶ (typ_or T U) ->
  G⁏e ⊢##f x ∶ T \/   G⁏e ⊢##f x ∶ U.
Proof.
  introv Hm Hi Hcor Hty.
  inversions Hty.
  -   inversions_tyq H0. false. eapply pf_or_false; eauto.
  - eauto.
  - eauto.
Qed.

(* Lemma 13 in "Pure methods for roDOT" *)
(** Invertible typing is closed under tight subtyping. *)
Lemma invertible_flat_typing_closure_tight G e x T U:
  (mode_incap' typing_mode = mode_incap_structural \/ mode_incap' typing_mode = mode_incap_none) ->
  inert G ->
  renv_corr G e ->
  G⁏e ⊢##f x ∶ T ->
  G⁏e ⊢# T <: U ->
  G⁏e ⊢##f x ∶ U.
Proof.
  introv Hm Hi Hcor HT Hsub.
  dependent induction Hsub; try solve [false; eapply invertible_nocap_false; eauto]; eauto.
  - false. eapply invertible_flat_bot_false; eauto.
  - inversions HT; auto.
    apply tyq_typ_sym in H0.
    inversions_tyq H0.
    apply pf_and1 in H. apply* ty_precise_inv. auto using tyq_typ_sym.
  - inversions HT; auto.
    apply tyq_typ_sym in H0.
    inversions_tyq H0.
    apply pf_and2 in H. apply* ty_precise_inv. auto using tyq_typ_sym.
  - inversions HT; auto.
    apply tyq_typ_sym in H0.
    inversions_tyq H0.
    false; eauto using pf_or_false.
  - inversions HT.
    + apply tyq_typ_sym in H1.
      apply tyq_typ_sel_inv in H1.
      destruct H1 as [[v21 [v22 [tnl [Htnl [HU [? Htyq2]]]]]] | HT2]; subst.
      false; eauto using pf_psel_false.
      false; eauto using pf_psel_false.
    + lets Hu: (x_bound_unique H H7). subst.
      pose proof (pf_inert_unique_tight_bounds Hi H H7) as Hu. destruct Hu. subst.
      destruct (pf_dec_typ_inv Hi H) as [He|Hb].
      * subst. assumption.
      * subst. false. eapply invertible_flat_bot_false; eauto.
  - eapply invertible_flat_typing_closure_tyq; eauto.
  - inversions HT.
    + apply tyq_typ_sym in H0.
      destruct (tyq_typ_andor_inv (typ_rcd (dec_typ A S1 T1)) (typ_rcd (dec_typ A S2 T2)) typ_andor_and H0) as [t1 [t2 [Ht [Htyq1 Htyq2]]]]; subst.
      destruct (tyq_typ_dec_typ_inv Htyq1) as [t11 [t12 [Ht1 [Htyq11 Htyq12]]]]; subst.
      destruct (tyq_typ_dec_typ_inv Htyq2) as [t21 [t22 [Ht2 [Htyq21 Htyq22]]]]; subst.
      pose proof (pf_inert_T Hi H).
      false. eapply pf_and_typ_same_false; eauto.
    + apply invertible_flat_to_precise_typ_dec in H4.
      destruct H4 as [S'1 [T'1 [U'1 [Pf1 [Hst11 Hst12]]]]].
      apply invertible_flat_to_precise_typ_dec in H5.
      destruct H5 as [S'2 [T'2 [U'2 [Pf2 [Hst21 Hst22]]]]].
      destruct (pf_inert_unique_tight_bounds Hi Pf1 Pf2).
      subst.
      eapply ty_dec_typ_inv.
      eapply precise_to_invertible_flat.
      apply Pf1.
      intro y. specialize (Hst12 y). specialize (Hst22 y). sympls; eauto.
      intro y. specialize (Hst12 y). specialize (Hst22 y). sympls; eauto.
  - inversions HT.
    + apply tyq_typ_sym in H0.
      destruct (tyq_typ_andor_inv T1 (typ_or T2 T3) typ_andor_and H0) as [t1 [t2 [Ht [Htyq1 Htyq2]]]]; subst.
      apply pf_and2 in H.
      destruct (tyq_typ_andor_inv T2 T3 typ_andor_or Htyq2) as [t21 [t22 [Ht2 [Htyq21 Htyq22]]]]; subst.
      false. eapply pf_or_false. eauto. apply H.
    + inversions H5.
      * symmetry in H0.
        destruct (tyq_typ_andor_inv T2 T3 typ_andor_or H0) as [t21 [t22 [Ht2 [Htyq21 Htyq22]]]]; subst.
           false. eapply pf_or_false. eauto. apply H.
      * apply ty_or1_inv. apply ty_and_inv; auto.
      * apply ty_or2_inv. apply ty_and_inv; auto.
  - false. eapply invertible_flat_nocap_false; eauto.
  - false. eapply invertible_flat_nocap_false; eauto.
  - false. eapply invertible_flat_nocap_false; eauto.
  - false. eapply invertible_flat_nocap_false; eauto.
  - false. eapply invertible_flat_nocap_false; eauto.
  - (* mode_top_nm_subtype requires mode_incap_nocap *)
    destruct Hm;
    typing_mode_contradiction typing_mode.
  - (* mode_top_nm_subtype requires mode_incap_nocap *)
    destruct Hm;
    typing_mode_contradiction typing_mode.
  -
    destruct Hm;
    typing_mode_contradiction typing_mode.
  -
    destruct Hm;
    typing_mode_contradiction typing_mode.
Qed.


(** ** Tight-to-Invertible Lemma for Variables [|-# to |-##]

       [inert G]            #<br>#
       [G ⊢# x∶ U]         #<br>#
       [―――――――――――――――]    #<br>#
       [G ⊢## x∶ U] *)
Lemma tight_to_invertible_flat G e U x:
  (mode_incap' typing_mode = mode_incap_structural \/ mode_incap' typing_mode = mode_incap_none) ->
    inert G ->
    renv_corr G e ->
    G⁏e ⊢# x ∶ U ->
    G⁏e ⊢##f x ∶ U.
Proof.
  introv Hm Hgd Hcor Hty.
  remember (trm_var (avar_f x)) as t eqn:Heq.
  induction Hty; try congruence; inversions Heq; eauto.
  - destruct Hm.
    *
    specialize (IHHty Hgd Hcor eq_refl).
    inversion IHHty; subst; eauto using ty_precise_inv.
    *
    specialize (IHHty Hgd Hcor eq_refl). eapply ty_bnd_inv; eauto.
    intro. unfold mode_has_incap' in H5. rewrite H4 in H5. discriminate.
  - specialize (IHHty Hgd Hcor eq_refl).
    inversion IHHty; subst; eauto using ty_precise_inv.
    apply tyq_typ_sym in H3.
    apply tyq_typ_bnd_inv in H3; auto.
    destruct H3 as [t21 [HT2 Htyq21]]; subst.
    specialize (Htyq21 x).
    apply tyq_typ_sym in Htyq21.
    eapply ty_precise_inv. 2: eauto.
    apply pf_open. eauto.
  - eapply invertible_flat_typing_closure_tight; auto.
  - (* In rodot, every variable has mutability *)
    destruct (tight_typing_implies_bound H0).
    destructc (pf_mutability H Hgd H2) as T2 ?H.
    eauto.
Qed.

End WithTypingMode.

Notation "G '⁏' e '⊢##f' x '∶' T" := (ty_var_inv G e x T) (at level 40, e at level 39, x at level 59).
