Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax Defs GeneralTyping TypingMode.
Require Import HeapUpdate HeapCorrespondence.

(** * Valid heaps *)

Inductive defs_valid : defs -> Prop :=
| defs_valid_nil:
    defs_valid defs_nil
| defs_valid_cons ds d:
    defs_valid ds ->
    defs_hasnt ds (label_of d) ->
    defs_valid (defs_cons ds d).

(* Heap is valid if it does not contain duplicate fields *)
(* Does not imply (ok s) *)
Inductive heap_valid: heap -> Prop :=
| heap_valid_empty: heap_valid empty
| heap_valid_push_fun s x f: heap_valid s -> heap_valid (s & x ~ item_fun (f))
| heap_valid_push_obj s x ds: heap_valid s -> defs_valid ds -> heap_valid (s & x ~ item_obj (ds)).

Lemma defs_update_has ds1 a x ds2:
    defs_update ds1 a x ds2 ->
    defs_has ds2 (def_fld a (avar_f x)).
Proof.
  introv Hu.
  induction Hu.
  - unfold defs_has. unfold get_def. cases_if. auto.
  - unfold defs_has. unfold get_def.
    cases_if.
    + inversions C. contradiction.
    + unfold defs_has in IHHu. unfold get_def in IHHu. auto.
Qed.

Lemma defs_update_has_n ds1 a x ds2 a0 x0:
    defs_update ds1 a x ds2 ->
    a <> a0 ->
    defs_has ds1 (def_fld a0 (avar_f x0)) ->
    defs_has ds2 (def_fld a0 (avar_f x0)).
Proof.
  introv Hu Hna Hhas.
  induction Hu.
  - unfold defs_has. unfold get_def. cases_if. inversions C. contradiction.
    unfold defs_has in Hhas. unfold get_def in Hhas. cases_if. auto.
  - unfold defs_has in *. unfold get_def in *.
    cases_if.
    + auto.
    + auto.
Qed.

Lemma defs_update_has_n_rev ds1 a x ds2 a0 x0:
    defs_update ds1 a x ds2 ->
    a <> a0 ->
    defs_has ds2 (def_fld a0 (avar_f x0)) ->
    defs_has ds1 (def_fld a0 (avar_f x0)).
Proof.
  introv Hu Hna Hhas.
  induction Hu.
  - unfold defs_has. unfold get_def. cases_if. inversions C. contradiction.
    unfold defs_has in Hhas. unfold get_def in Hhas. cases_if. auto.
  - unfold defs_has in *. unfold get_def in *.
    cases_if.
    + auto.
    + auto.
Qed.

Lemma heap_update_rev_inv S1 y1 a y3 S2 y0 v:
  heap_update S1 y1 a y3 S2 ->
  y0 <> y1 ->
  binds y0 v S2 ->
  binds y0 v S1.
Proof.
  introv Hu Hn Hbi.
  destructc Hu as ? ? ? ? ?.
  assert (y0 \in dom S2).
  eapply binds_to_dom; eauto.
  rewrite <- H1 in H4.
  eapply dom_to_binds in H4.
  destruct H4.
  pose proof H4.
  eapply H2 in H4.
  binds_eq.
  auto.
  auto.
Qed.

Lemma heap_update_inv_n S1 y1 a y3 S2 y0 ds0 a0 y2:
  (* If y1.a is updated to y3 *)
  heap_update S1 y1 a y3 S2 ->
  (* And y0.a0 = y2 *)
  binds y0 (item_obj ds0) S2 ->
  defs_has ds0 (def_fld a0 (avar_f y2)) ->
  (* And y2 <> y3*)
  y2 <> y3 ->
  (* Then y2 in y0.a0 is not the updated value *)
  exists ds1, (
    binds y0 (item_obj ds1) S1 /\
    defs_has ds1 (def_fld a0 (avar_f y2))
  ).
Proof.
  introv Hu Hbi Hhas Hn.
  pose proof Hu as Huu.
  destructc Hu as ? ? ? ? ds11 ds12 ?.
  destructc H3 as ? ? ?.

  pose proof H4 as Hdu.
  eapply defs_update_has in H4.
  destruct (classicT (y0 = y1)).
  - (* Different field in the same object *)
    subst.
    binds_eq.
    assert (a <> a0). { (* The field must be different *)
      intro. subst.
      eapply defs_has_inv in H4; eauto.
      simplify_eq H4. auto.
    }
    eapply defs_update_has_n_rev in Hhas; eauto.
  - (* Field in a different object *)
    exists ds0; split; eauto.

    eapply heap_update_rev_inv; eauto.
Qed.

Lemma heap_update_nfld S1 y1 a y2 S2 ds0 a0:
  heap_update S1 y1 a y2 S2 ->
  binds y1 (item_obj ds0) S2 ->
  defs_has ds0 (def_fld a0 (avar_f y2)) ->
  a <> a0 ->
  exists ds1, (
    binds y1 (item_obj ds1) S1 /\
    defs_has ds1 (def_fld a0 (avar_f y2))
  ).
Proof.
  introv Hu Hbi Hhas Hn.
  destructc Hu as ? ? ? ? ds11 ds12 ?.
  destructc H3 as ? ? ?.
  exists ds11.
  split; eauto.
  binds_eq.
  eapply defs_update_has_n_rev in Hhas; eauto.
Qed.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Typed defs are valid *)
Lemma defs_typing_valid G Go e ds T:
  G §§ Go ⁏e ⊢ ds ∶ T ->
  defs_valid ds.
Proof.
  intros.
  induction H.
  - constructor. constructor. unfold defs_hasnt. compute. auto.
  - constructor; auto.
Qed.

(* Typed heaps are valid *)
Lemma heap_correspond_valid G e s:
  heap_correspond G e s ->
  heap_valid s.
Proof.
  intros.
  destructc H as ? ? ? ? ? ? ?.
  destructc H5 as ? ?.
  clear H H0 H2 H3 H4.

  (* If we used H6 in the induction, then it would require us to show
     that the items are typed also in a smaller heap, which is not true.
     So any heap is allowed.
   *)
  assert (forall x : var, x \in dom s -> exists s2, ok (s & s2) /\ ty_item_s typing_mode G e (s & s2) x) as Hty.
  intros.
  exists (empty: heap). rewrite concat_empty_r. auto.
  clear H6.

  induction H1.
  - constructor.
  - (* The new item is typed *)
    assert (exists s2, ok (E & x ~ v & s2) /\ ty_item_s typing_mode G e (E & x ~ v & s2) x).
    { apply Hty. simpl_dom. rewrite in_union. left. rewrite in_singleton. auto. }
    assert (heap_valid E).
      eapply IHok.
      intros.
    assert (exists s2, ok (E & x ~ v & s2) /\ ty_item_s typing_mode G e (E & x ~ v & s2) x0).
    { apply Hty. simpl_dom. rewrite in_union. right. auto. }


    destructc H3 as s2 Hok2 Hbi2.
      exists (x ~ v & s2).
      rewrite concat_assoc.
      split; eauto.

    destructc H0 as s2 Hok2 Hbi2.
    pose proof Hbi2 as H00.
    inversions Hbi2.
    + eapply binds_middle_eq_inv in H4. subst.
      eapply heap_valid_push_fun. auto. auto.
    + eapply binds_middle_eq_inv in H3. subst.
      eapply heap_valid_push_obj. auto.
      eapply defs_typing_valid. apply H6.
      auto.
Qed.

End WithTypingMode.
