(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode GeneralTyping Indep Equivalence
  RecordTypes Substitution.

(* Type of an object reference in an inert context. T is a record type, S is a refcap type *)
Notation typ_objref T S := (typ_and (typ_bnd T) S) (only parsing).
(* Type of an object location in an inert context. T is a record type.
   In basic mode, locations have no reference capabilities.
   In rodot, they have a mutability flag. *)
(* Note: See also typ_objlit in GeneralTyping, which is basically the same,
   but there may be some reasons to diverge in some modes *)
Notation typ_objref_loc typing_mode T := (
  match (mode_refcaps' typing_mode) with
  | mode_refcaps_none => (typ_objref T typ_top)
  | mode_refcaps_mutability => (typ_objref T (typ_mut_marker typ_bot))
  end)(only parsing).

Inductive refcap_inert_mutability: typ -> varkind -> Prop :=
| refacp_inert_mutability_loc: refcap_inert_mutability (typ_mut_marker typ_bot) v_loc
| refacp_inert_mutability_ref Tbnd: refcap_inert_mutability (typ_mut_marker Tbnd) v_ref
.

(* Type specifying reference capabilities of inert object types, with a condition on reference capabilities, based on typing mode. *)
Definition refcap_inert_type {typing_mode: Mode} T vk :=
  match (mode_refcaps' typing_mode) with
  | mode_refcaps_none => refcap_typ T \{}
  | mode_refcaps_mutability => refcap_inert_mutability T vk
  end.

(** * Inert types
       A type is inert if it is a recursive type
       whose type declarations have equal bounds (enforced through [record_type]),
       in intersection with a mutability declaration. #<br>#
       For example, the following types are inert:
       - [mu(x: {a: T} /\ {B: U..U})  /\ {M: bot..bot}]
       - [mu(x: {C: {A: T..U}..{A: T..U}}) /\ {M: bot..top}]
       And the following types are not inert:
       - [{a: T}]
       - [{B: U..U}]
       - [top]
       - [x.A(y)]
       - [mu(x: {B: S..T})], where [S <> T]. *)

(* Notations for inert_typ constructors,
   to be able to use the cases individually *)
Notation obj_typ_is cond T S vk := (
      record_type T ->
      refcap_inert_type S vk ->
      typ_indep_b 0 T ->
      cond (typ_objref T S) vk) (only parsing).
Notation fun_typ_is cond S T := (
      cond (typ_fun S T)) (only parsing).

Notation fun_type U := (exists S T, fun_typ_is (eq U) S T) (only parsing).
Notation obj_type typing_mode U := (exists T S, obj_typ_is typing_mode (eq U) T S) (only parsing).

Section WithTypingMode.
Context (typing_mode: Mode).

Inductive inert_vk: varkind -> Prop :=
  | inert_vk_loc: inert_vk v_loc
  | inert_vk_ref: inert_vk v_ref.

Inductive inert_typ: typ -> varkind -> Prop :=
  (* Inert function types *)
  | inert_typ_fun S T vk:
    mode_has_fun typing_mode ->
    fun_typ_is inert_typ S T vk
  (* Inert object types *)
  | inert_typ_obj T S vk:
    obj_typ_is inert_typ T S vk.

(* Part of Figure 8 in "Reference mutability for DOT" *)
Definition inert_type T := exists vk, inert_typ T vk.

(* Part of Figure 8 in "Reference mutability for DOT" *)
(** An inert context is a typing context whose range consists only of inert types,
    has no duplicate keys, forward- or self- referencing entries, and all entries are locations or references. *)
Inductive inert: ctx -> Prop :=
  | inert_empty: inert ctx_empty
  | inert_push G x vk T:
      inert G ->
      inert_vk vk ->
      inert_typ T vk ->
      x §# G ->
      x \notin ctx_fv G ->
      x \notin fv T ->
      inert (G § x / vk ~ T).

(** In the proof, it is useful to be able to distinguish record types from
    other types. A record type is a concatenation of type declarations with equal
    bounds [{A: T..T}] and field declarations [{a: T}]. *)

Hint Constructors inert_typ inert_vk inert : core.

(** If a typing context is inert, then the variables in its domain are distinct. #<br>#
    Note: [ok] is defined in [TLC.LibEnv.v]. *)
Lemma inert_ok G:
    inert G ->
    ctx_ok G.
Proof.
  introv Hi. induction Hi; auto.
Qed.

Lemma inert_weakens G:
    inert G ->
    ctx_weakens G.
Proof.
  introv Hi. induction Hi; auto.
Qed.

(* Types in an inert context must be inert*)
Lemma inert_binds_inert G x T:
  inert G ->
  ctx_binds x T G ->
  inert_type T.
Proof.
  intros.
  induction H; inversion H0; auto. eexists; eauto.
Qed.

Lemma inert_binds_k_inert G G2 x T vk:
  inert G ->
  ctx_binds_k_after x T vk G G2 ->
  inert_typ vk T.
Proof.
  intros.
  induction H; inversion H0; auto.
Qed.

(* Types of objects in an inert context must be indep on the self variable *)
Lemma inert_bnd_indep G x T S:
  inert G ->
  ctx_binds x (typ_objref T S) G ->
  typ_indep_b 0 T.
Proof.
  intros. induction H. inversion H0.
  inversions H0.
  - inversions H2; auto.
  - auto.
Qed.

(* Variable bound in inert context is visible *)
Lemma inert_binds_vis G x T:
  inert G ->
  ctx_binds x T G ->
  ctx_vis G x.
Proof.
  intros.
  induction H; inversion H0; auto.
Qed.

Lemma inert_dom_vis G x:
  inert G ->
  x \in ctx_dom G ->
  ctx_vis G x.
Proof.
  intros.
  eapply ctx_dom_to_binds in H0. destruct H0.
  eapply inert_binds_vis; eauto.
  eapply inert_ok; auto.
  eapply inert_weakens; auto.
Qed.


(* Variable bound typed in inert context is visible *)
Lemma inert_ty_vis G e v T:
  inert G ->
  G ⁏ e ⊢ v ∶ T ->
  ctx_vis G v.
Proof.
  introv Hin Hty.
  eapply inert_dom_vis; eauto.
  eapply typing_implies_dom; eauto.
Qed.

(* Inert context only contains locations and references *)
Lemma inert_dom G:
  inert G ->
  ctx_dom G = (ctx_kdom v_loc G) \u (ctx_kdom v_ref G).
Proof.
  introv Hin.
  induction Hin.
  - rewrite ctx_dom_empty.
    rewrite ctx_kdom_empty.
    rewrite ctx_kdom_empty.
    auto.
  - inversion H; subst.
    rewrite ctx_dom_push.
    rewrite ctx_kdom_push_e.
    rewrite ctx_kdom_push_n.
    rewrite IHHin.
    auto.
    simplify_eq.
    rewrite ctx_dom_push.
    rewrite ctx_kdom_push_e.
    rewrite ctx_kdom_push_n.
    rewrite IHHin.
    auto.
    simplify_eq.
Qed.

Lemma inert_tyq_open_bnd_same : forall e T1 S1 T2 S2 y1 y2,
    inert_type (typ_objref T1 S1) ->
    inert_type (typ_objref T2 S2) ->
    typing_equiv e y1 y2 ->
    (forall x, typing_equiv e (open x T1) (open x T2)) ->
    typing_equiv e (open y1 T1) (open y2 T2).
Proof.
  introv Hin1 Hin2 Htyv Htyq.
  inversions Hin1; inversions Hin2.
  inversions H; inversions H0.
  apply tyq_typ_open_rec_indep; eauto.
Qed.

Lemma refcap_inert_refcap_typ T vk:
  refcap_inert_type T vk ->
  refcap_type T.
Proof.
  intros.
  unfold refcap_inert_type in H.

  destruct (mode_refcaps' typing_mode).
  - eexists; eauto.
  - inversions H.
    eexists. constructor. constructor. eauto.
    eexists. constructor. constructor. eauto.
Qed.

Lemma inert_typ_loc_ref T:
  inert_typ T v_loc ->
  inert_typ T v_ref.
Proof.
  intros.
  inversions H.
  - eauto.
  - constructor; eauto.
    unfold refcap_inert_type in *.
    destruct (mode_refcaps' typing_mode); eauto.
    inversions H1. constructor.
Qed.

Lemma inert_concat_invert_l G1 G2 : inert (G1 §& G2) -> inert G1.
Proof.
  introv Hin. induction G2; eauto; inversions Hin; eauto.
Qed.

Lemma inert_concat_invert_r G1 G2 : inert (G1 §& G2) -> inert G2.
Proof.
  revert G1. induction G2; introv Hin; eauto; inversions Hin.
  rewrite ctx_fv_concat in *. constructor; eauto.
Qed.

Lemma subst_label_of_dec x y (D : dec) :
  label_of (D /[ x -> y]) = label_of D.
Proof. induction D; simpl in *; eauto. Qed.

Lemma record_dec_subst x y (D : dec) :
  record_dec D -> record_dec (D /[ x -> y]).
Proof. introv Hrcd. induction Hrcd; simpl in *; econstructor; eauto. Qed.

Lemma record_dec_subst_conv x y (D : dec) :
  record_dec D -> record_dec (D /[ x -> y]).
Proof. introv Hrcd. induction Hrcd; simpl in *; econstructor; eauto. Qed.

Lemma record_type_subst T L x y :
  record_typ T L ->
  record_typ (T /[ x -> y]) L.
Proof.
  introv Hrcd. induction Hrcd; subst; eauto.
  - erewrite <- subst_label_of_dec. simpl. apply rt_one;
      eauto using record_dec_subst.
  - simpl in *. erewrite <- subst_label_of_dec.
    apply rt_cons; eauto using record_dec_subst.
    rewrite subst_label_of_dec. auto.
Qed.

Lemma record_typ_subst T x y :
  record_type T ->
  record_type (T /[ x -> y]).
Proof.
  introv Hrcd. inversions Hrcd. eexists. eauto using record_type_subst.
Qed.

Lemma refcap_dec_subst D x y :
  refcap_dec D -> refcap_dec (D /[ x -> y]).
Proof. introv Hrf. inversions Hrf. simpl. econstructor. Qed.

Lemma refcap_typ_subst T L x y :
  refcap_typ T L ->
  refcap_typ (T /[ x -> y]) L.
Proof.
  introv Hrf. induction Hrf; simpl in *;
    try solve [econstructor; eauto].
  - subst. simpl. erewrite <- subst_label_of_dec. econstructor;
      eauto using refcap_dec_subst.
  - subst. simpl. econstructor;
      eauto using refcap_dec_subst, subst_label_of_dec.
Qed.

Lemma refcap_inert_mutability_subst T vk x y :
  refcap_inert_mutability T vk ->
  refcap_inert_mutability (T /[ x -> y]) vk.
Proof. introv Hrf. induction Hrf; simpl in *; econstructor; eauto. Qed.

Lemma refcap_inert_type_subst T vk x y :
  refcap_inert_type T vk ->
  refcap_inert_type (T /[ x -> y]) vk.
Proof.
  introv Hrf. hnf in Hrf. hnf.
  destruct (mode_refcaps' typing_mode) eqn:Hm;
    eauto using refcap_typ_subst, refcap_inert_mutability_subst.
Qed.

Lemma inert_type_subst T vk x y :
  inert_typ T vk ->
  inert_typ (T /[ x -> y]) vk.
Proof.
  introv Hin. inversions Hin; simpl in *; eauto.
  rename H into Hrcd, H0 into Hcap, H1 into Hindep.
  apply inert_typ_obj; eauto using record_typ_subst, refcap_inert_type_subst.
Qed.

Lemma inert_binds_fv G x T :
  inert G -> ctx_binds x T G -> x \notin fv T.
Proof. introv HinG Hbi. induction HinG; inversions Hbi; eauto. Qed.

End WithTypingMode.

Arguments inert [typing_mode].
Arguments inert_typ {typing_mode}.
Arguments inert_type {typing_mode}.
(* Repeating hints from inside WithTypingMode *)
Hint Constructors inert_typ inert_vk inert : core.
Hint Resolve inert_ok : core.
Hint Resolve inert_weakens : core.
Hint Resolve inert_ty_vis: core.

(* Fresh telescope is like inert context but without the inert_typ and
   inert_vk conditions. So it is just with the freshness conditions, which
   says that a telescope G is fresh if for each decomposition
   G = G₁(x/vk~T)G₂, we have that x ∉ dom G₁ ∪ fv G₁ ∪ fv T *)
Inductive fresh_telescope : ctx -> Prop :=
  | fresh_telescope_empty : fresh_telescope ctx_empty
  | fresh_telescope_push : forall G x vk T,
      fresh_telescope G ->
      x §# G ->
      x \notin ctx_fv G ->
      x \notin fv T ->
      fresh_telescope (G § x / vk ~ T).
Hint Constructors fresh_telescope : core.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma inert_fresh_telescope G : inert G -> fresh_telescope G.
Proof. induction 1; eauto. Qed.

Lemma fresh_telescope_ok G : fresh_telescope G -> ctx_ok G.
Proof. induction 1; eauto. Qed.

Lemma fresh_telescope_weakens G : fresh_telescope G -> ctx_weakens G.
Proof. induction 1; eauto. Qed.

Lemma fresh_telescope_concat_invert_l G1 G2 :
  fresh_telescope (G1 §& G2) -> fresh_telescope G1.
Proof. introv Hfr. induction G2; eauto; inversions Hfr; eauto. Qed.

Lemma fresh_telescope_concat_invert_r G1 G2 :
  fresh_telescope (G1 §& G2) -> fresh_telescope G2.
Proof.
  revert G1. induction G2; introv Hfr; eauto; inversions Hfr.
  rewrite ctx_fv_concat in *. constructor; eauto.
Qed.

Lemma fresh_telescope_binds_fv G x T :
  fresh_telescope G -> ctx_binds x T G -> x \notin fv T.
Proof. introv HftG Hbi. induction HftG; inversions Hbi; eauto. Qed.

Lemma fresh_telescope_binds_vis G x T :
  fresh_telescope G -> ctx_binds x T G -> ctx_vis G x.
Proof. intros. induction H; inversion H0; auto. Qed.

End WithTypingMode.

Lemma ctx_binds_linear G x1 x2 T1 T2 :
  ctx_weakens G ->
  ctx_binds x1 T1 G -> ctx_binds x2 T2 G ->
  ((exists G1 vk1 G2 vk2 G3,
       (G = (((G1 § x1 / vk1 ~ T1) §& G2) § x2 / vk2 ~ T2) §& G3
        /\ (fresh_telescope G -> x2 \notin fv T1))
       \/ (G = (((G1 § x2 / vk2 ~ T2) §& G2) § x1 / vk1 ~ T1) §& G3)
         /\ (fresh_telescope G -> x1 \notin fv T2))
   /\ (ctx_ok G -> x1 <> x2))
  \/ (x1 = x2
     /\ exists G1 G2 vk T, G = ((G1 § x1 / vk ~ T) §& G2)).
Proof.
  introv Hwk Hb1 Hb2.
  eapply ctx_binds_to_ctx_binds_k_after in Hb1 as [Gx1 [vk1 Hbk1]]; eauto.
  eapply ctx_binds_to_ctx_binds_k_after in Hb2 as [Gx2 [vk2 Hbk2]]; eauto.

  eapply ctx_binds_k_after_initial_segment in Hbk1 as [G2 ?]; subst.
  eapply ctx_binds_k_after_concat_inv in Hbk2
      as [Hbk2 | [G02 [? Hbk2]]]; subst.

  - inversions Hbk2; try solve[right; split; eauto].
    rename H7 into Hne, H8 into Hbk.
    eapply ctx_binds_k_after_initial_segment
      in Hbk as [Gx3 ?]; subst. left. repeat(split).
    exists Gx2 vk1 Gx3 vk2 G2.
    + right; split; eauto. introv Hfr.
      eapply fresh_telescope_concat_invert_l in Hfr.
      inversions Hfr. clear H6 H3 H4. rename H5 into Hfr.
      rewrite ctx_fv_concat in Hfr. simpl in Hfr.
      repeat(rewrite notin_union in Hfr); tauto.
    + introv; eauto.

  - eapply ctx_binds_k_after_initial_segment
      in Hbk2 as [Gx3 ?]; subst. left. split.
    exists Gx1 vk1 G02 vk2 Gx3. left.
    + rewrite ctx_concat_assoc. simpl. split; eauto.
      introv Hfr. eapply fresh_telescope_concat_invert_l in Hfr.
      inversions Hfr. clear H6 H3 H4. rename H5 into Hfr.
      rewrite ctx_fv_concat in Hfr. simpl in Hfr.
      repeat(rewrite notin_union in Hfr); tauto.
    + introv Hok. eapply ctx_ok_concat_inv in Hok as [_ [_ Hdis]].
      rewrite ctx_dom_concat in Hdis. repeat(rewrite ctx_dom_push in Hdis).
      hnf. introv Heq; subst. erewrite <- in_empty, <- Hdis, in_inter.
      repeat(rewrite in_union). repeat(rewrite in_singleton). eauto.

Qed.
