(** remove printing ~ *)

(** This module contains lemmas related to invertible typing
    ([ty_var_inv], [⊢##] and [ty_item_inv], [⊢##v]). *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx.

Require Import Coq.Bool.Bool LibBoolExtra.

Require Import LayeredTypingMain LayeredSubtypingAtomic LayeredCapabilityBasic
    LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain LayeredCommon
    LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure
    LayeredTypingLogic LayeredTypingMainHard LayeredSubtypingMainOrd.
Require Import TypeInduction TightTypingHelpers.
Require Import FlatInvertibleTyping.
Require Import TypLogicInduction.
Require Import LayeredTypingMainCount.
Require Import LayeredSubtypingMainTree.

(** * Invertible typing *)

(** The invertible-typing relation describes the possible types that a variable or value
can be typed with in an inert context. For example, if [G] is inert, [G ⊢! x: {a: T}],
and [G ⊢ T <: T'], then [G ⊢## x: {a: T'}].

The purpose of invertible typing is to be easily invertible into a precise typing relation.
To achieve that, invertible typing avoids typing cycles that could result from, for example,
repeated applications of recursion introduction and elimination.
For this case, invertible typing defines only recursion introduction (whereas precise typing
defines only recursion elimination). *)

(** ** Invertible typing of variables [G ⊢## x: T] *)

Section WithTypingMode.
Context {typing_mode: Mode}.

Notation "G '⁏' e '⊢##m' x '∶' T" := (ty_inv_main G e x T) (at level 40, e at level 39, x at level 59).
(* Lemma 29 in "Pure methods for roDOT" *)
(** Layered-to-precise typing for field declarations: #<br>#
    [G ⊢##m x∶ {a: S..T}]            #<br>#
    [――――――――――――――――――――――]      #<br>#
    [exists T', G ⊢! x∶ {a: S'..T'}]      #<br>#
    [G ⊢# T' <: T].
    [G ⊢# S <: S']. *)
Lemma invertible_main_to_precise_fld_dec: forall G e x a S T,
  G⁏e ⊢##m x ∶ typ_rcd (dec_fld a S T) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_fld a S' T') /\
    G⁏e ⊢# T' <: T /\
    G⁏e ⊢# S <: S'.
Proof.
  introv Hinv.
  forwards Hinva: ty_inv_main_atomic_atomic Hinv.
  eauto.

  remember ((typ_rcd (dec_fld a S T))).
  reverts S T.
    induction Hinva; intros; subst.
    - apply tyq_typ_dec_fld_invr in H1.
      destructc H1 as t21 t22 Ht Htyq1 Htyq2; subst.
      exists t21 t22 T.
      apply tyq_typ_sym in Htyq1.
      split; auto; split; auto.
    -
  inversions H.
  forwards HHinva: IHHinva. 2: eauto.
  eapply ty_inv_atomic_main; eauto.

  destructecs HHinva.
  exists S' T' U.
  splits; auto; eapply subtyp_trans_t; eassumption.
Qed.


Definition ty_inv := match mode_ty_inv' typing_mode with | mode_ty_inv_layered => ty_inv_main | mode_ty_inv_flat => ty_var_inv end.

Notation "G '⁏' e '⊢##' x '∶' T" := (ty_inv G e x T) (at level 40, e at level 39, x at level 59).

(** Invertible-to-precise typing for field declarations: #<br>#
    [G ⊢## x∶ {a: S..T}]            #<br>#
    [――――――――――――――――――――――]      #<br>#
    [exists T', G ⊢! x∶ {a: S'..T'}]      #<br>#
    [G ⊢# T' <: T].
    [G ⊢# S <: S']. *)
Lemma invertible_to_precise_fld_dec: forall G e x a S T,
  G⁏e ⊢## x ∶ typ_rcd (dec_fld a S T) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_fld a S' T') /\
    G⁏e ⊢# T' <: T /\
    G⁏e ⊢# S <: S'.
Proof.
  introv Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode).
  - eapply invertible_flat_to_precise_fld_dec; eauto.
  - eapply invertible_main_to_precise_fld_dec; eauto.
Qed.


Lemma invertible_to_precise_typ_dec: forall G e x a S T,
  G⁏e ⊢## x ∶ typ_rcd (dec_typ a S T) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ a S' T') /\
    (forall y, G⁏e ⊢# open y T' <: open y  T) /\
    (forall y, G⁏e ⊢# open y S <: open y  S').
Proof.
  introv Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode).
  - eapply invertible_flat_to_precise_typ_dec; eauto.
  - eapply invertible_main_to_precise_typ_dec; eauto.
Qed.


Lemma invertible_to_precise_met_dec: forall G e x m TC TZ TR,
  inert G -> ok e ->
  G⁏e ⊢## x ∶ typ_rcd (dec_met m TC TZ TR) ->
  exists L S TC' TZ' TR',
    G ⊢! x ∶ S ⪼ typ_rcd (dec_met m TC' TZ' TR') /\
    G⁏e ⊢# TZ <: TZ' /\
    (forall z, z \notin L -> G § z / v_par ~ TZ ⁏e ⊢ (open z TC) <: (open z TC')) /\
    (mode_has_sef typing_mode -> typ_indep_b 0 TC -> typ_indep_b 0 TC') /\
    (forall z, z \notin L -> forall y, y <> z -> y \notin L -> ctx_met G z TZ y TC ⁏e ⊢ (open_met z y TR') <: (open_met z y TR))
.
Proof.
  introv HiG He Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode).
  - eapply invertible_flat_to_precise_met_dec; eauto.
  - eapply invertible_main_to_precise_met_dec; eauto.
Qed.

(** Invertible-to-precise typing for function types: #<br>#
    [ok G]                        #<br>#
    [G ⊢## x∶ forall(S)T]             #<br>#
    [――――――――――――――――――――――――――]  #<br>#
    [exists S', T'. G ⊢! x∶ forall(S')T']  #<br>#
    [G ⊢# S <: S']               #<br>#
    [G ⊢# T'^y <: T^y], where [y] is fresh. *)
Lemma invertible_to_precise_typ_fun: forall G e x S T,
  inert G -> ok e ->
  G⁏e ⊢## x ∶ typ_fun S T ->
  exists S' T' U L,
    G ⊢! x ∶ U ⪼ typ_fun S' T' /\
    G⁏e ⊢# S <: S' /\
    (forall y,
        y \notin L ->
            G § y / v_par ~ S ⁏e ⊢ open y T' <: open y T).
Proof.
  introv HiG He Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode).
  - eapply invertible_flat_to_precise_typ_fun; eauto.
  - eapply invertible_main_to_precise_typ_fun; eauto.
Qed.

Lemma invertible_bot_false : forall G e x,
    inert G ->
    G⁏e ⊢## x ∶ typ_bot ->
    False.
Proof.
  introv Hin Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_bot_false; eauto.
  - eapply invertible_main_bot_false; eauto.
Qed.

Lemma invertible_nocap_false : forall G e x,
    inert G ->
    G⁏e ⊢## x ∶ typ_nocap ->
    False.
Proof.
  introv Hin Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_nocap_false; eauto.
  - eapply invertible_main_nocap_false; eauto.
Qed.

Lemma invertible_and_invl G e x T1 T2:
    inert G ->
    G⁏e ⊢## x ∶ typ_and T1 T2 ->
    G⁏e ⊢## x ∶ T1.
Proof.
  introv Hin Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_and_invl; eauto.
  - eapply ty_inv_main_lc_and_invl with (oc:=logctx_nil); eauto.
Qed.
Lemma invertible_and_invr G e x T1 T2:
    inert G ->
    G⁏e ⊢## x ∶ typ_and T1 T2 ->
    G⁏e ⊢## x ∶ T2.
Proof.
  introv Hin Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_and_invr; eauto.
  - eapply ty_inv_main_lc_and_invr with (oc:=logctx_nil); eauto.
Qed.

Lemma invertible_typing_closure_tyq G e x T U:
  inert G ->
  renv_corr G e ->
  G⁏e ⊢## x ∶ T ->
  typing_equiv e T U ->
  G⁏e ⊢## x ∶ U.
Proof.
  introv Hin Hcor Hty Htyq.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_typing_closure_tyq; eauto with typing_mode_deps.
  - eapply ty_inv_main_tyq; eauto with typing_mode_deps.
Qed.

(** Invertible typing is closed under tight subtyping. *)
Lemma invertible_typing_closure_tight G e x T U:
  inert G ->
  renv_corr G e ->
  G⁏e ⊢## x ∶ T ->
  G⁏e ⊢# T <: U ->
  G⁏e ⊢## x ∶ U.
Proof.
  introv Hi Hcor HT Hsub.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_typing_closure_tight; eauto with typing_mode_deps.
  - eapply invertible_main_typing_closure_tight  with (oc:=logctx_nil); eauto with typing_mode_deps.
    unfold mode_has_incap'. rewrite typing_mode_dep_ty_inv_layered_incap_nocap; eauto.
Qed.


(** ** Tight-to-Invertible Lemma for Variables [|-# to |-##]

       [inert G]            #<br>#
       [G ⊢# x∶ U]         #<br>#
       [―――――――――――――――]    #<br>#
       [G ⊢## x∶ U] *)
Lemma tight_to_invertible G e U x:
    inert G ->
    renv_corr G e ->
    G⁏e ⊢# x ∶ U ->
    G⁏e ⊢## x ∶ U.
Proof.
  introv Hgd Hcor Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply tight_to_invertible_flat; eauto with typing_mode_deps.
  - eapply tight_to_invertible_main; eauto with typing_mode_deps.
Qed.


End WithTypingMode.

Notation "G '⁏' e '⊢##m' x '∶' T" := (ty_inv_main G e x T) (at level 40, e at level 39, x at level 59).
Notation "G '⁏' e '⊢##' x '∶' T" := (ty_inv G e x T) (at level 40, e at level 39, x at level 59).
