Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping TightTypingHelpers.
Require Import Orctx Logctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic LayeredCapabilityApprox.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** N and M supertypes on the basic layer. *)

(* Part of Figure 21 in "Pure methods for roDOT" *)
(** Atomic M-supertypes*)
Inductive ty_has_M G e: typ -> Prop :=
(** 
    [―――――――――――(LMB-M)]   #<br>#
    [G ⊢b M◁ {M: TL..TU}] #<br>#
*)
| ty_has_typ_mut TL TU:
  (mode_has_top_nm_complement typing_mode -> forall x : var, G⁏e ⊢# TL /[-> x] <: typ_bot) ->
  ty_has_M G e (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL TU))
.

(* Part of Figure 21 in "Pure methods for roDOT" *)
(** Atomic N-supertypes, but not M-supertypes (and not top-like) *)
Inductive ty_has_N_atomic (G: ctx) (e: renv): typ -> Prop :=
(** 
    [―――――――――――(LNB-N)]   #<br>#
    [G ⊢b N◁ N] #<br>#
*)
| ty_has_nocap:
    ty_has_N_atomic G e typ_nocap 
(** 
    [―――――――――――(LNB-Fld)]   #<br>#
    [G ⊢b N◁ {a: TL..TU}] #<br>#
*)
| ty_has_fld a TL TU:
    ty_has_N_atomic G e (typ_rcd (dec_fld a TL TU)) 
(** 
    [―――――――――――(LNB-Met)]   #<br>#
    [G ⊢b N◁ {m(TC,TZ):TR}] #<br>#
*)
| ty_has_met a TC TZ TR:
    ty_has_N_atomic G e (typ_rcd (dec_met a TC TZ TR)) 
(** 
    [―――――――――――(LNB-Typ)]   #<br>#
    [G ⊢b N◁ {A: TL..TU}] #<br>#
*)
| ty_has_typ_normal A TL TU:
    ty_has_N_atomic G e (typ_rcd (dec_typ (typ_label_normal A) TL TU)) 
(** This rule is disabled in rodot and rodot_sef typing modes. *)
| ty_has_fun T U:
    mode_has_fun typing_mode ->
    ty_has_N_atomic G e (typ_fun T U)
.

(** Top-like atomic types *)
Inductive ty_has_top_atomic (G: ctx) (e: renv): typ -> Prop :=
| ty_has_top_atomic_mut TL1 TU1:
(** This rule is disabled in rodot and rodot_sef typing modes. *)
  mode_has_mut_top typing_mode ->
  (forall x : var, G⁏e ⊢# TL1/[-> x] <: typ_bot) ->
  (forall x : var, G⁏e ⊢# typ_top <: TU1/[-> x]) ->
  ty_has_top_atomic G e (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL1 TU1))
.

(** Atomic N-supertypes, including M-supertypes *)
Inductive ty_has_N (G: ctx) (e: renv): typ -> Prop :=
| ty_has_N_has_atomic T:
  ty_has_N_atomic G e T ->
  ty_has_N G e T
| ty_has_N_has_top T:
  ty_has_top_atomic G e T ->
  ty_has_N G e T
.

End WithTypingMode.

Hint Constructors ty_has_M: core.
Hint Constructors ty_has_top_atomic: core.
Hint Constructors ty_has_N_atomic: core.
Hint Constructors ty_has_N: core.

Section WithTypingMode.
Context {typing_mode: Mode}.


(** ** Closure properties *)

Lemma ty_has_M_tyq G e U V:
  ty_has_M G e U ->
  typing_equiv e U V ->
  ty_has_M G e V.
Proof.
  introv Hh Htyq.
  inversions Hh; inversions_tyq Htyq; eauto using subtyp_bot_tyq_t.
Qed.

Lemma ty_has_N_atomic_tyq G e U V:
  ty_has_N_atomic G e U ->
  typing_equiv e U V ->
  ty_has_N_atomic G e V.
Proof.
  introv Hh Htyq.
  inversions Hh; inversions_tyq Htyq; eauto using subtyp_bot_tyq_t.
Qed.
Lemma ty_has_top_atomic_tyq G e U V:
  ty_has_top_atomic G e U ->
  typing_equiv e U V ->
  ty_has_top_atomic G e V.
Proof.
  introv Hh Htyq.
  inversions Hh; inversions_tyq Htyq; eauto using subtyp_bot_tyq_t, subtyp_top_tyq_t.
Qed.


Lemma ty_has_N_tyq G e U V:
  ty_has_N G e U ->
  typing_equiv e U V ->
  ty_has_N G e V.
Proof.
  introv Hh Htyq.
  inversions Hh.
  - eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_tyq; eauto.
  - eapply ty_has_N_has_top. eapply ty_has_top_atomic_tyq; eauto.
Qed.



Lemma ty_has_M_lc_subtyp G e oc U V:
  inert G ->
  ty_has_M G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_M G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  destruct oc; simpls.
  - induction Hsu; eauto.
    inversions Hh.
    inversions Hh.
    inversions Hh. eauto.
    inversions Hh.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
Qed.

Lemma ty_has_N_atomic_lc_subtyp G e oc U V:
  inert G ->
  ty_has_N_atomic G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_N_atomic G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  destruct oc; simpls.
  - induction Hsu; eauto.
    inversions Hh. eauto.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
Qed.

Lemma ty_has_top_atomic_lc_subtyp G e oc U V:
  inert G ->
  ty_has_top_atomic G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_top_atomic G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  destruct oc; simpls.
  - induction Hsu; 
    inversions Hh; eauto.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
Qed.

Lemma ty_has_N_lc_subtyp G e oc U V:
  inert G ->
  ty_has_N G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_N G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  inversion Hh.
  - eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_subtyp; eauto.
  - eapply ty_has_N_has_top. eapply ty_has_top_atomic_lc_subtyp; eauto.
Qed.


Lemma ty_has_N_atomic_lc_rev_subtyp G e oc U V:
  inert G ->
  ty_has_N_atomic G e (logctx_typ oc U) ->
  subtyp_atomic G e V U ->
  ty_has_N_atomic G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  destruct oc; simpls.
  - induction Hsu; eauto.
    inversions Hh. eauto.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
  - inversions Hh.
Qed.

Lemma ty_has_N_lc_rev_subtyp G e oc U V:
  inert G ->
  ty_has_N G e (logctx_typ oc U) ->
  subtyp_atomic G e V U ->
  ty_has_N G e (logctx_typ oc V) \/  ty_has_top_atomic G e U.
Proof.
  introv Hi Hh Hsu.
  inversions Hh.
  - left. eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_rev_subtyp; eauto.
  - destruct oc; simpls; inversions H; eauto.
Qed.


(** ** Inversion properties *)

Lemma ty_has_M_lc_inv G e oc T:
  ty_has_M G e (logctx_typ oc T) ->
  oc = logctx_nil.
Proof.
  introv Hty.
  induction oc; simpls; eauto;
  inversions Hty.
Qed.


Lemma ty_has_N_atomic_lc_inv G e oc T:
  ty_has_N_atomic G e (logctx_typ oc T) ->
  oc = logctx_nil.
Proof.
  introv Hh.
  induction oc; simpls; eauto;
  inversions Hh.
Qed.

Lemma ty_has_top_atomic_lc_inv G e oc T:
  ty_has_top_atomic G e (logctx_typ oc T) ->
  oc = logctx_nil.
Proof.
  introv Hh.
  induction oc; simpls; eauto;
  inversions Hh.
Qed.

Lemma ty_has_N_lc_inv G e oc T:
  ty_has_N G e (logctx_typ oc T) ->
  oc = logctx_nil.
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_inv, ty_has_top_atomic_lc_inv.
Qed.

Lemma ty_has_N_oc_inv G e oc T:
  ty_has_N G e (orctx_typ oc T) ->
  oc = orctx_nil.
Proof.
  introv Hh.
  induction oc; simpls; eauto;
  inversions Hh; inversions H.
Qed.


Lemma ty_has_N_atomic_lc_or_false G e oc T S:
  ty_has_N_atomic G e (logctx_typ oc (typ_or T S)) ->
  False.
Proof.
  introv Hh.
  induction oc; simpls; inversions Hh.
Qed.
Lemma ty_has_top_atomic_lc_or_false G e oc T S:
  ty_has_top_atomic G e (logctx_typ oc (typ_or T S)) ->
  False.
Proof.
  introv Hh.
  induction oc; simpls; inversions Hh.
Qed.
Lemma ty_has_N_lc_or_false G e oc T S:
  ty_has_N G e (logctx_typ oc (typ_or T S)) ->
  False.
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_or_false, ty_has_top_atomic_lc_or_false.
Qed.
Lemma ty_has_M_lc_or_false G e oc T S:
  ty_has_M G e (logctx_typ oc (typ_or T S)) ->
  False.
Proof.
  introv Hh.
  induction oc; simpls; inversions Hh.
Qed.

Lemma ty_has_M_lc_and_false G e T oc U:
  ty_has_M G e (logctx_typ oc (typ_and T U)) ->
  False.
Proof.
  induction oc; intros; simpls; inversions H.
Qed.

Lemma ty_has_N_atomic_lc_and_false G e T oc U:
  ty_has_N_atomic G e (logctx_typ oc (typ_and T U)) ->
  False.
Proof.
  induction oc; intros; simpls; inversions H.
Qed.
Lemma ty_has_top_atomic_lc_and_false G e T oc U:
  ty_has_top_atomic G e (logctx_typ oc (typ_and T U)) ->
  False.
Proof.
  induction oc; intros; simpls; inversions H.
Qed.

Lemma ty_has_N_lc_and_false G e T oc U:
  ty_has_N G e (logctx_typ oc (typ_and T U)) ->
  False.
Proof.
  introv Hh.
  inversions Hh.
  - eapply ty_has_N_atomic_lc_and_false; eauto.
  - eapply ty_has_top_atomic_lc_and_false; eauto.
Qed.

Lemma ty_has_N_atomic_lc_bot_false G e oc:
  ty_has_N_atomic G e (logctx_typ oc typ_bot) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_top_atomic_lc_bot_false G e oc:
  ty_has_top_atomic G e (logctx_typ oc typ_bot) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_N_lc_bot_false G e oc:
  ty_has_N G e (logctx_typ oc typ_bot) ->
  False.
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_bot_false,ty_has_top_atomic_lc_bot_false.
Qed.

Lemma ty_has_M_lc_bot_false G e oc:
  ty_has_M G e (logctx_typ oc typ_bot) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_N_atomic_lc_sel_false G e oc x A y:
  ty_has_N_atomic G e (logctx_typ oc (typ_sel x A y)) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.
Lemma ty_has_top_atomic_lc_sel_false G e oc x A y:
  ty_has_top_atomic G e (logctx_typ oc (typ_sel x A y)) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.


Lemma ty_has_N_lc_sel_false G e oc x A y:
  ty_has_N G e (logctx_typ oc (typ_sel x A y)) ->
  False.
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_sel_false, ty_has_top_atomic_lc_sel_false.
Qed.


Lemma ty_has_M_lc_sel_false G e oc x A y:
  ty_has_M G e (logctx_typ oc (typ_sel x A y)) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_N_atomic_lc_bnd_false G e oc T:
  ty_has_N_atomic G e (logctx_typ oc (typ_bnd T))  ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.
Lemma ty_has_top_atomic_lc_bnd_false G e oc T:
  ty_has_top_atomic G e (logctx_typ oc (typ_bnd T))  ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.


Lemma ty_has_N_lc_bnd_false G e oc T:
  ty_has_N G e (logctx_typ oc (typ_bnd T)) ->
  False.
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_bnd_false, ty_has_top_atomic_lc_bnd_false.
Qed.

Lemma ty_has_M_lc_bnd_false G e oc T:
  ty_has_M G e (logctx_typ oc (typ_bnd T)) ->
  False.
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.


(** ** Replacement properties *)

Lemma ty_has_N_atomic_lc_N_replace G e oc T U:
  ty_has_N_atomic G e (logctx_typ oc U) ->
  ty_has_N_atomic G e T ->
  ty_has_N_atomic G e (logctx_typ oc T).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eauto.
Qed.

Lemma ty_has_top_atomic_lc_N_replace G e oc T U:
  ty_has_top_atomic G e (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_has_top_atomic G e (logctx_typ oc T).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eauto.
Qed.
Lemma ty_has_N_lc_N_replace G e oc T U:
  ty_has_N G e (logctx_typ oc U) ->
  ty_has_N G e T ->
  ty_has_N G e (logctx_typ oc T).
Proof.
  introv HhU HhT.
  inversions HhU; inversions HhT.
  -
  eapply ty_has_N_has_atomic.
  eapply ty_has_N_atomic_lc_N_replace; eauto.
  - destruct oc; simpls; inversions H; eauto using ty_has_N_has_atomic, ty_has_N_has_top.
  - destruct oc; simpls; inversions H; eauto using ty_has_N_has_atomic, ty_has_N_has_top.
  -
  eapply ty_has_N_has_top.
  eapply ty_has_top_atomic_lc_N_replace; eauto.
Qed.
Lemma ty_has_N_atomic_lc_M_replace G e oc T:
  ty_has_N_atomic G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_N_atomic G e (logctx_typ oc T).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_top_atomic_lc_M_replace G e oc T:
  inert G ->
  renv_corr G e ->
  ty_has_top_atomic G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_top_atomic G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  destruct oc; simpls; inversions Hh.
  false.
  pick_fresh x. specialize (H3 x).
  eapply subtyp_top_bot_false; eauto.
Qed.


Lemma ty_has_N_lc_M_replace G e oc T:
  inert G ->
  renv_corr G e ->
  ty_has_N G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_N G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  inversions Hh.
  - eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_M_replace; eauto.
  - eapply ty_has_N_has_top. eapply ty_has_top_atomic_lc_M_replace; eauto.
Qed.

Lemma ty_has_M_lc_nocap_replace G e oc T:
  ty_has_M G e (logctx_typ oc typ_nocap) ->
  ty_has_M G e (logctx_typ oc T).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh.
Qed.

Lemma ty_has_N_lc_has_top_replace G e oc T U:
  ty_has_N G e (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_has_N G e (logctx_typ oc T).
Proof.
  introv Hh Htop.
  eapply ty_has_N_lc_inv in Hh. subst.
  eauto.
Qed.

Lemma ty_has_M_has_top G e T:
  ty_has_top_atomic G e T ->
  ty_has_M G e T.
Proof.
  introv Htop.
  inversions Htop.
  eauto.
Qed.

Lemma ty_has_M_lc_has_top_replace G e oc T U:
  ty_has_M G e (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_has_M G e (logctx_typ oc T).
Proof.
  introv Hh Htop.
  destruct oc; simpls; inversions Hh.
  eapply ty_has_M_has_top; eauto.
Qed.


Lemma ty_has_N_M_false G e T:
  ty_has_N_atomic G e T -> 
  ty_has_M G e T ->
  False.
Proof.
  introv HhN HhM.
  inversions HhN; inversions HhM.
Qed.

Lemma ty_has_M_lc_N_replace G e oc T U:
  ty_has_M G e (logctx_typ oc U) ->
  ty_has_N_atomic G e U ->
  ty_has_M G e (logctx_typ oc T).
Proof.
  introv Hh HhN.
  destruct oc; simpls; inversions Hh; inversions HhN.
Qed.

Lemma ty_has_N_atomic_lc_typ_replace G e oc A SL SU TL TU:
  ty_has_N_atomic G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_has_N_atomic G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh. eauto.
Qed.

(* For general labels, use ty_has_N_lc_subtyp*)
Lemma ty_has_N_lc_typ_normal_replace G e oc A SL SU TL TU:
  ty_has_N G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) SL SU))) ->
  ty_has_N G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) TL TU))).
Proof.
  introv Hh.
  inversions Hh.
  - eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_typ_replace; eauto.
  - inversions H. destruct oc; simpls; inversions H0.
Qed.

Lemma ty_has_N_lc_typ_replace G e oc A SL SU:
  ty_has_N G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (forall TL TU, ty_has_N G e (logctx_typ oc (typ_rcd (dec_typ A TL TU)))) \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hh.
  inversions Hh.
  - left. intros. eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_typ_replace; eauto.
  - forwards: ty_has_top_atomic_lc_inv H; fweauto. subst. simpl. eauto.
Qed.

Lemma ty_has_M_lc_typ_replace G e oc A SL SU TL TU:
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (mode_has_top_nm_complement typing_mode -> forall x : var, G⁏e ⊢# TL /[-> x] <: typ_bot) ->
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh. eauto.
Qed.

Lemma ty_has_M_lc_typ_replace' G e oc A SL SU TL TU:
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))) \/ (mode_has_top_nm_complement typing_mode -> forall x : var, G⁏e ⊢# SL /[-> x] <: typ_bot).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh. eauto.
Qed.

Lemma ty_has_M_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  inert G ->
  x \in ctx_dom G ->
  ty_pf_or_none G x A TL TU ->
  ty_has_M G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hh Hi Hdom Hpf.
  destruct oc; simpls; inversions Hh.
  constructor; intros.
  forwards: ty_pf_or_none_mutability_bound Hpf; eauto.
Qed.



(** ** Equivalence with tight subtyping *)
Lemma ty_has_N_atomic_tight G e T:
  ty_has_N_atomic G e T ->
  G⁏e ⊢# typ_nocap <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2.
Qed.


Lemma ty_has_top_atomic_tight G e T:
  ty_has_top_atomic G e T ->
  G⁏e ⊢# typ_top <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2.
  eapply subtyp_top_mut_t; eauto.
Qed.
Lemma ty_has_N_tight G e T:
  ty_has_N G e T ->
  G⁏e ⊢# typ_nocap <: T.
Proof.
  introv Hh.
  inversions Hh.
  eapply ty_has_N_atomic_tight; eauto.
  eapply subtyp_trans_t.
  eapply subtyp_top_t.
  eapply ty_has_top_atomic_tight; eauto.
Qed.
Lemma ty_has_M_tight G e T:
  ty_has_M G e T ->
  G⁏e ⊢# (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_top typ_bot)) <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2.
Qed.

(* Modified renv_corr_pf_normal_typ_tyq *)
Lemma ty_has_N_atomic_pf_tyq G e v1 v2 U1 T1:
  inert G ->
  renv_corr G e ->
  typing_equiv e v1 v2 ->
  G ⊢! v1 ∶ U1 ⪼ T1 ->
  ty_has_N_atomic G e T1 ->
  typ_atomic T1 ->
  exists U2 T2, (
      G ⊢! v2 ∶ U2 ⪼ T2 /\
      typing_equiv e T1 T2
    ).
Proof.
  introv Hin Hcor Htyqv Hpf Hh Ha.
  assert (Hok : ctx_ok G) by eauto.
  assert (Hbin1 : ctx_binds v1 U1 G) by (eapply pf_binds; eauto).

  destructc (pf_inert_T Hin Hpf) as vk HinT.
  destruct HinT.
  * (* typ_fun *)
    assert (Hbin2 : ctx_binds v2 (typ_fun S T) G).
    { destruct (renv_corr_tyq_fun_loc_to_ref S T Hok Hcor Htyqv) as [Hltr Hrtl];
        auto. }
    apply precise_flow_fun_inv in Hpf; subst.
    exists (typ_fun S T); eauto.
  * (* typ_objref *)

  forwards Hcases: pf_bnd_same_or_rcd; eauto.
  destructecs Hcases; subst.
  - inversions Ha.
  - inversions Ha.
  -   destructc (renv_corr_pf_tyq Hin Hcor Htyqv Hpf) as T' U' Hpf' Htyq'.
  auto. eauto.
  - (* refcap_type T1 contradicts ty_has_N T1 *)
    hnf in Hcases. destructecs Hcases. inversions Hcases; inversions Ha; inversions H2; inversions Hh.
Qed.


End WithTypingMode.