Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening.
Require Import Orctx Logctx OrctxLogctx Logpath LogpathOrder TypLogicInduction.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Ordered logic typing 

For or_diag, we want to get a type where the 
- left (primary) subtree does not contain any intersections
- right subtree contains intersections.

Therefore, intersections must be created in right subtree first, then left subtree, then above.
In logpath_le, the left side is left of or above or same as the right side.
Therefore, when creating intersection at oc, for all previous paths ocp, it must be logpath_le oc ocp.

Then it should be ty_has_N_logic <-> ty_has_N_logic_ord logpath_nil.

*)


Inductive ty_has_N_logic_ord (G: ctx) (e: renv): orctx -> typ -> Prop :=
| ty_has_N_logic_ord_or oc T:
    ty_has_N_closure G e T ->
    typ_notorand T ->
    ty_has_N_logic_ord G e oc T
| ty_has_N_logic_ord_and oc ocp T U:
    orctx_le oc ocp ->
    ty_has_N_logic_ord G e ocp (orctx_typ ocp T) ->
    ty_has_N_logic_ord G e ocp (orctx_typ ocp U) ->
    ty_has_N_logic_ord G e oc (orctx_typ ocp (typ_and T U))
.

Lemma ty_has_N_logic_ord_logic G e op T:
  ty_has_N_logic_ord G e op T ->
  ty_has_N_logic G e T.
Proof.
  introv Hh.
  induction Hh.
  - eapply ty_has_N_logic_or; eauto.
  - eapply ty_has_N_logic_oc_and; eauto.
Qed.


Lemma ty_has_N_logic_ord_le G e op1 op2 T:
  orctx_le op2 op1 ->
  ty_has_N_logic_ord G e op1 T ->
  ty_has_N_logic_ord G e op2 T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_has_N_logic_ord_or; eauto.
  - eapply ty_has_N_logic_ord_and; eauto.
    unfold orctx_le in *. eapply logpath_le_trans; eauto.
Qed.



Lemma ty_has_N_atomic_logic_ord_has G e oc T:
  ty_has_N_atomic G e T ->
  ty_has_N_logic_ord G e oc T.
Proof.
  introv Hh.
  inversions Hh; eapply ty_has_N_logic_ord_or; try eapply ty_or_closure_atomic; try constructor; eauto; discriminate.
Qed.
Lemma ty_has_top_atomic_logic_ord_has G e oc T:
  ty_has_top_atomic G e T ->
  ty_has_N_logic_ord G e oc T.
Proof.
  introv Hh.
  inversions Hh; eapply ty_has_N_logic_ord_or; try eapply ty_or_closure_atomic; try constructor; eauto; discriminate.
Qed.



Lemma ty_has_N_logic_ord_has G e oc T:
  ty_has_N G e T ->
  ty_has_N_logic_ord G e oc T.
Proof.
  introv Hh.
  inversions Hh.
  - eapply ty_has_N_atomic_logic_ord_has; eauto.
  - eapply ty_has_top_atomic_logic_ord_has; eauto.
Qed.

Lemma ty_has_N_logic_ord_has_closure G e T:
    ty_has_N_closure G e T ->
    ty_has_N_logic_ord G e (typ_orand_leftmost_orctx T) T.
Proof.
  introv Hh.
  induction T using typ_orand_ind; intros.
  - eapply ty_has_N_logic_ord_or; eauto.
  - eapply ty_has_N_logic_ord_and.
    +
    eapply typ_orand_leftmost_orctx_le.
    +
    eapply ty_has_N_logic_ord_le.
    eapply typ_orand_leftmost_orctx_notorand_le; eauto.
    eapply IHT1.
    rewrite orctx_logctx_typ in *.
    eapply ty_has_N_closure_lc_and_replace; eauto.
    +
    eapply ty_has_N_logic_ord_le.
    eapply typ_orand_leftmost_orctx_notorand_le; eauto.
    eapply IHT2.
    rewrite orctx_logctx_typ in *.
    eapply ty_has_N_closure_lc_and_replace; eauto.
Qed.

Lemma ty_has_N_logic_ord_notorand_closure G e oc T:
  ty_has_N_logic_ord G e oc T ->
  typ_notorand T ->
  ty_has_N_closure G e T.
Proof.
  introv Hh Hn.
  induction Hh; eauto.
  eapply typ_notorand_oc_notand in Hn.
  false Hn; eauto.
Qed.



(* Proof of ty_has_N_logic -> ty_has_N_logic_ord logpath_nil *)


Lemma ty_has_N_logic_ord_oc_and_symm G e oc ocp0 U0 T0:
  ty_has_N_logic_ord G e oc (orctx_typ ocp0 (typ_and U0 T0)) ->
  ty_has_N_logic_ord G e oc (orctx_typ ocp0 (typ_and T0 U0)).
Proof.
  introv Hty.
  remember (orctx_typ ocp0 (typ_and U0 T0)).
  reverts ocp0 U0 T0.
  induction Hty; intros; subst.
  - eapply ty_has_N_logic_ord_or.
    rewrite orctx_logctx_typ.
    eapply ty_has_N_closure_lc_and_replace.
    rewrite <- orctx_logctx_typ.
    eauto.
    eapply typ_notorand_oc_notand in H0. false H0; eauto.
  - 
    eapply orctx_intersect_and_and_cases in Heqt.
    destruct_all; subst.
    + eapply ty_has_N_logic_ord_and; eauto.
    + specialize (H2 (typ_and T0 U0)).
      destruct_all.
      rewrite <- H2.
      eapply ty_has_N_logic_ord_and; eauto.
      eapply orctx_le_trans. eauto. eapply orctx_follow_le. eapply orctx_follow_symm; eauto.
      *
      rewrite H3. eapply ty_has_N_logic_ord_le. eapply orctx_follow_le. eauto. eapply IHHty1. eauto.
      *
      rewrite H4. eapply ty_has_N_logic_ord_le. eapply orctx_follow_le. eauto. eapply IHHty2. eauto.
Qed.




Lemma ty_has_N_logic_ord_notorand_le G e oc ocp U:
  orctx_notorand_le ocp ->
  ty_has_N_logic_ord G e oc (orctx_typ ocp U) ->
  ty_has_N_logic_ord G e ocp (orctx_typ ocp U).
Proof.
  introv Hn Hh.
  remember (orctx_typ ocp U); reverts ocp U.
  induction Hh; intros.
  - eapply ty_has_N_logic_ord_or; eauto.
  - forwards: orctx_notorand_eq_and_le; eauto.
    eapply ty_has_N_logic_ord_and; eauto.
Qed.



Lemma ty_has_N_logic_ord_orctx_notorand_le G e oc T:
  ty_has_N_logic_ord G e oc (orctx_typ oc T) ->
  orctx_notorand_le oc.
Proof.
  introv Hh.
  remember (orctx_typ oc T).
  reverts T.
  induction Hh; intros; subst.
  - eapply orctx_notorand_le_notorand; eauto.
  -
    forwards: IHHh1; eauto.
    eapply orctx_notorand_le_le; eauto.
Qed.

Lemma ty_has_N_logic_ord_oc_notorand_le G e oc T:
  ty_has_N_logic_ord G e oc T ->
  typ_notorand T \/ ty_has_N_logic_ord G e (typ_orand_leftmost_orctx T) T /\ orctx_le oc (typ_orand_leftmost_orctx T).
Proof.
  introv Hh.
  induction Hh.
  - eauto.
  - right.
    forwards: typ_orand_leftmost_orctx_le ocp T U.
    split.
    +
      eapply ty_has_N_logic_ord_and; eauto.
    +
      forwards: ty_has_N_logic_ord_orctx_notorand_le Hh1.
      eapply orctx_le_trans.
      eauto.
      eapply typ_orand_leftmost_orctx_notorand_le; eauto.
Qed.

Lemma ty_has_N_logic_ord_oc_notorand_and_inv_l G e oc T U:
  orctx_notorand_le oc ->
  ty_has_N_logic_ord G e oc (orctx_typ oc (typ_and T U)) ->
  ty_has_N_logic_ord G e oc (orctx_typ oc T).
Proof.
  introv Hl Hh.
  inversions Hh.
  - eapply typ_notorand_oc_notand in H0. false. eapply H0. eauto.
  - forwards Hocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
    symmetry in H.
    forwards: orctx_notorand_eq_and_le H; eauto.
    forwards: orctx_le_le_eq_inv H; eauto.
    destruct_all; subst.
    inversions H5.
    eauto.
Qed.
Lemma ty_has_N_logic_ord_oc_notorand_and_inv_r G e oc T U:
  orctx_notorand_le oc ->
  ty_has_N_logic_ord G e oc (orctx_typ oc (typ_and T U)) ->
  ty_has_N_logic_ord G e oc (orctx_typ oc U).
Proof.
  introv Hl Hh.
  inversions Hh.
  - eapply typ_notorand_oc_notand in H0. false. eapply H0. eauto.
  - forwards Hocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
    symmetry in H.
    forwards: orctx_notorand_eq_and_le H; eauto.
    forwards: orctx_le_le_eq_inv H; eauto.
    destruct_all; subst.
    inversions H5.
    eauto.
Qed.


Lemma ty_has_N_logic_ord_and_lc_le G e oc ocp T U:
    orctx_le oc ocp ->
    ty_has_N_logic_ord G e oc (orctx_typ ocp T) ->
    ty_has_N_logic_ord G e oc (orctx_typ ocp U) ->
    ty_has_N_logic_ord G e oc (orctx_typ ocp (typ_and T U)).
Proof.
  introv Hl HhT HhU.
  rename ocp into ocTU.
  rename oc into ocL.
  remember (orctx_typ ocTU T).
  reverts ocTU T U.

  induction HhT; intros; subst.
  - forwards: ty_has_N_logic_ord_notorand_le HhU; eauto using orctx_notorand_le_notorand.
    eapply ty_has_N_logic_ord_and; eauto. eapply ty_has_N_logic_ord_or; eauto.
  - 
    forwards Hnlocp: ty_has_N_logic_ord_orctx_notorand_le HhT1.
    destruct (orctx_le_cases ocTU ocp).
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_has_N_logic_ord_and; eauto.
      rewrite <- Heqt.
      eapply ty_has_N_logic_ord_and; eauto.
      symmetry in Heqt.
      eapply ty_has_N_logic_ord_notorand_le; eauto.
      eapply orctx_notorand_le_le; eauto.
    + (* ocp left of ocp0  -> need induction *)

    symmetry in Heqt.
    forwards: orctx_intersect_cases_r Heqt. intros; discriminate.
    destruct_all; subst.
    * (* ocp = ocp0 -> solve without inductions same as the case above *)
      forwards: orctx_le_extend_l_inv; eauto; subst; simpls; rewrite orctx_extend_nil_l in *.
      eapply ty_has_N_logic_ord_and; eauto.
      eapply ty_has_N_logic_ord_notorand_le; eauto.
    * (* actually need induction *)
      (* Recombine types *)
      (* 1. variants according to the target intersection: T0, U0, T0&U0 *)
      forwards: orctx_separate_cross_r Heqt H1 T0.
      forwards: orctx_separate_cross_r Heqt H1 U0.
      forwards: orctx_separate_cross_r Heqt H1 (typ_and T0 U0).
      destruct_all.
      (* 2. variants according to the current intersection: T, U, T&U *)
      forwards: H9 T. (* T0 / T *)
      forwards: H9 U. (* T0 / U *)

      forwards: H7 T. (* U0 / T *)
      forwards: H7 U. (* U0 / U *)
      forwards: H7 (typ_and T U). (* U0 / T&U *)

      forwards: H5 T. (* T0&U0 / T *)
      forwards: H5 U. (* T0&U0 / U *)
      forwards: H5 (typ_and T U). (* T0&U0 / T&U *)
      destruct_all.

      (* Find equalities between variables, to have one variable for each variant *)
      forwards: orctx_follow_eq_inv x1 ocp. eauto using orctx_follow_trans, orctx_follow_symm. rewrite <- H2. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x4 x7. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H16. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x3 x6. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H14. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x2 ocTU. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H18. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x7 x9. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H16. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x6 x8. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H17. eauto. destruct_all; subst. cleanup.
      forwards: orctx_follow_eq_inv x5 ocTU. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H18. eauto. destruct_all; subst. cleanup.

      rewrite H25 in HhU.

      assert (orctx_notorand_le x0). {
        eapply orctx_notorand_replace_r; eauto.
      }

      eapply ty_has_N_logic_ord_notorand_le in HhU. 2: eauto.
      assert (ty_has_N_logic_ord G e ocp (orctx_typ x0 T )) as Htyx0T. {
        forwards: ty_has_N_logic_ord_oc_notorand_and_inv_l; eauto.
        eapply ty_has_N_logic_ord_le. 2: eauto.
        eapply orctx_follow_le; eauto.
      }
      assert (ty_has_N_logic_ord G e ocp (orctx_typ x0 U )) as Htyx0U. {
        forwards: ty_has_N_logic_ord_oc_notorand_and_inv_r; eauto.
        eapply ty_has_N_logic_ord_le. 2: eauto.
        eapply orctx_follow_le; eauto.
      }

      forwards: IHHhT1. 2: symmetry; eauto. applys orctx_le_trans ocTU; eauto using orctx_follow_le, orctx_follow_symm. rewrite H29. eauto.
      forwards: IHHhT2. 2: symmetry; eauto. applys orctx_le_trans ocTU; eauto using orctx_follow_le, orctx_follow_symm. rewrite H27. eauto.

      rewrite H19. rewrite H23 in H3. rewrite H21 in H4. 
      eapply ty_has_N_logic_ord_and.
      applys orctx_le_trans ocp; eauto using orctx_follow_le, orctx_follow_symm. 
      eapply ty_has_N_logic_ord_le; eauto. eauto using orctx_follow_le, orctx_follow_symm. 
      eapply ty_has_N_logic_ord_le; eauto. eauto using orctx_follow_le, orctx_follow_symm.
Qed.


Lemma ty_has_N_logic_ord_and_lc G e oc ocp T U:
    ty_has_N_logic_ord G e oc (orctx_typ ocp T) ->
    ty_has_N_logic_ord G e oc (orctx_typ ocp U) ->
    ty_has_N_logic_ord G e (orctx_leb ocp oc) (orctx_typ ocp (typ_and T U)).
Proof.
  introv HhT HhU.

  destruct (orctx_le_cases ocp oc).
  - (* ocp <= oc    easy case *)
  rewrite orctx_leb_l in * by auto.
  eapply ty_has_N_logic_ord_and.
  eapply orctx_le_refl.
  eapply ty_has_N_logic_ord_le; eauto.
  eapply ty_has_N_logic_ord_le; eauto.
  - (* oc <= ocp *)
  eapply ty_has_N_logic_ord_le.
  eapply orctx_leb_le_comm.
  rewrite orctx_leb_l in * by auto.
  eapply ty_has_N_logic_ord_and_lc_le; eauto.
Qed.



Lemma ty_has_N_logic_logic_ord G e T:
  ty_has_N_logic G e T ->
  ty_has_N_logic_ord G e orctx_nil T.
Proof.
  introv Hh.
  induction Hh.
  -
    eapply ty_has_N_logic_ord_le.
    2: eapply ty_has_N_logic_ord_has_closure; eauto.
    constructor.
  - eapply ty_has_N_logic_ord_le.
    2: eapply ty_has_N_logic_ord_and_lc; eauto.
    constructor.
Qed.

Lemma ty_has_N_closure_lc_or_dist_eq G e oc T U:
  ty_dist_eq T U ->
  ty_has_N_closure G e (logctx_typ oc (typ_or T U)) ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv He Hh.
  induction He.
  - eapply ty_has_N_closure_or_inv in Hh; destruct_all; eauto.
    eapply ty_has_N_logic_or; eauto.
    eapply ty_has_N_logic_or; eauto.
  - eapply ty_has_N_closure_or_inv in Hh; destruct_all; eauto.
    eapply ty_has_N_logic_or; eauto.
    eapply ty_has_N_closure_or_inv in H; destruct_all.
    eapply ty_has_N_logic_lc_orl; eapply IHHe1. 
    eapply ty_has_N_closure_lc_orr; eapply H.
    eapply ty_has_N_logic_lc_orr; eapply IHHe2. 
    eapply ty_has_N_closure_lc_orr; eapply H.
  - eapply ty_has_N_closure_or_inv in Hh; destruct_all; eauto.
    eapply ty_has_N_logic_lc_and. eapply IHHe1.
    eapply ty_has_N_closure_lc_and_replace with (U0:=(typ_or T1 U)) in H; eauto.
    eapply IHHe2.
    eapply ty_has_N_closure_lc_and_replace with (U0:=(typ_or T2 U)) in H; eauto.
    eapply ty_has_N_logic_lc_and. eapply IHHe1.
    eapply ty_has_N_closure_lc_orr; eapply H.
    eapply IHHe2.
    eapply ty_has_N_closure_lc_orr; eapply H.
  - eapply ty_has_N_closure_or_inv in Hh; destruct_all; eauto.
    eapply ty_has_N_logic_or; eauto.
    eapply IHHe.
    eapply ty_has_N_closure_lc_and_replace with (U0:=(typ_or U T1)) in H; eauto.
  - eapply ty_has_N_closure_or_inv in Hh; destruct_all; eauto.
    eapply ty_has_N_logic_or; eauto.
    eapply IHHe.
    eapply ty_has_N_closure_lc_and_replace with (U0:=(typ_or U T2)) in H; eauto.
Qed.

Lemma ty_has_N_closure_oc_or_dist_eq G e oc T U:
  ty_dist_eq T U ->
  ty_has_N_closure G e (orctx_typ oc (typ_or T U)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv He Hh.
  rewrite orctx_logctx_typ in *.
  eapply ty_has_N_closure_lc_or_dist_eq; eauto.
Qed.

(* All intersections are right of oc *)
Lemma ty_has_N_logic_ord_oc_nle_or_notorand_notorand_dist_eq G e ocp oc T U:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  typ_notorand U ->
  ty_has_N_logic_ord G e ocp (orctx_typ oc (typ_or T U)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hn He HnT HnU Hty.
  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_has_N_closure_oc_or_dist_eq; eauto.
  - 
    forwards Hnlocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
    forwards: ty_dist_eq_notorand_eq; eauto; subst.
    symmetry in HeqTT.
    forwards: typ_notorand_eq_separate HeqTT.
    eapply typ_notorand_or; eauto.

      forwards: orctx_separate_cross_r HeqTT H0 U0.
      destruct_all.

      forwards: H2 T.
      forwards: H2 U.
      destruct_all.

      assert (orctx_notorand_le x0). {
        eapply orctx_notorand_le_le. 2: eapply H5.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }
      assert (orctx_notorand_le x1).  {
        eapply orctx_notorand_le_le. 2: eapply H4.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }

      forwards: IHHty1 x1 U0 U0; eauto.
      forwards: IHHty2 x0 U0 U0; eauto.

      rewrite H1.
      rewrite H8 in H12.
      rewrite H6 in H13.
      eapply ty_has_N_logic_oc_and; eauto.
Qed.





(* All intersection are in U or right of oc *)
Lemma ty_has_N_logic_ord_oc_nle_or_notorand_dist_eq G e oc T U ocp:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  ty_has_N_logic_ord G e ocp (orctx_typ oc (typ_or T U)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hn He HnT Hty.
  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts T U oc.
  induction Hty; intros; subst.
  - eapply ty_has_N_closure_oc_or_dist_eq; eauto.
  - 
    forwards Hnlocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
    symmetry in HeqTT.
    forwards: orctx_intersect_cases_r HeqTT.
    intros; discriminate.
    destruct_all; subst.
    + (* under U -> continue induction *)
      clear HeqTT H.
      forwards: typ_notorand_eq_or_and; eauto.
      destruct_all; subst.
      clear H1.
      forwards: orctx_notorand_le_extend_invr Hnlocp.
      inversions H.

      forwards: ty_dist_eq_oc_and_cases T0; eauto.
      destruct H.
      *
      forwards: IHHty1 T0 (orctx_typ x0 T) oc0; eauto. rewrite orctx_extend_typ; simpls; eauto.
      *
      forwards: IHHty2 T0 (orctx_typ x0 U) oc0; eauto. rewrite orctx_extend_typ; simpls; eauto.
    + (* separate -> stop induction, go to next segment *)
      eapply ty_has_N_logic_ord_oc_nle_or_notorand_notorand_dist_eq; eauto.
      2: rewrite HeqTT. 2: eapply ty_has_N_logic_ord_and; eauto.
      forwards: orctx_notorand_eq_and_le. 2: symmetry; eapply HeqTT. eauto.
      forwards: typ_notorand_separate_eq_oc oc0 ocp (typ_or T0 U0); eauto.
      eapply typ_notorand_or_invr; eauto.
Qed.

(* All intersection are in T or in U or right of oc *)
Lemma ty_has_N_logic_ord_oc_nle_or_dist_eq G e oc T U ocp:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  ty_has_N_logic_ord G e ocp (orctx_typ oc (typ_or T U)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hn He Hty.

  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts T U oc.
  induction Hty; intros; subst.
  - eapply ty_has_N_closure_oc_or_dist_eq; eauto.
  - (* intersection *)
    symmetry in HeqTT.
    forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
    destruct_all; subst.
    + (* intersection under oc *)
      destruct x; simpls.
      ++ (* Cannot be right at the union *)
        discriminate.
      ++ (* Left branch -> continue induction *)
        inversions H1. clear HeqTT.
        forwards: IHHty1 (orctx_typ x T) T1 oc0.
        eapply ty_dist_eq_oc_and_invl; eauto.
        auto.
        rewrite orctx_extend_typ; simpls; eauto.
        forwards: IHHty2 (orctx_typ x U) T1 oc0.
        eapply ty_dist_eq_oc_and_invr; eauto.
        auto.
        rewrite orctx_extend_typ; simpls; eauto.
        rewrite <- orctx_extend_typ in *.
        eapply ty_has_N_logic_oc_and; eauto.
      ++ (* Right branch -> go to next segment *)
        inversions H1.

        forwards Hnlocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
        forwards: orctx_notorand_le_extend_invr Hnlocp.
        inversions H0.

        eapply ty_has_N_logic_ord_oc_nle_or_notorand_dist_eq.
        eauto.
        eauto.
        eauto.
        rewrite HeqTT. eapply ty_has_N_logic_ord_and; eauto.
    + (* intersection separate -> stop, go to last segment *)
      
      forwards Hnlocp: ty_has_N_logic_ord_orctx_notorand_le; eauto.
      forwards: orctx_notorand_eq_and_le. 2: symmetry; eauto. eauto.
      forwards: typ_notorand_separate_eq_oc oc0 ocp (typ_or T0 U0); eauto.

      eapply ty_has_N_logic_ord_oc_nle_or_notorand_notorand_dist_eq; eauto.
      eapply typ_notorand_or_invl; eauto.
      eapply typ_notorand_or_invr; eauto.
      rewrite HeqTT. eapply ty_has_N_logic_ord_and; eauto.

Qed.

Lemma ty_has_N_logic_ord_oc_nle_or_diag G e oc T:
  orctx_notorand_le oc ->
  ty_has_N_logic_ord G e oc (orctx_typ oc (typ_or T T)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hn Hty.
  eapply ty_has_N_logic_ord_oc_nle_or_dist_eq; eauto.
Qed.

Lemma ty_has_N_logic_ord_oc_or_diag G e oc T:
  ty_has_N_logic_ord G e orctx_nil (orctx_typ oc (typ_or T T)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_or T T)) as TT.
  reverts oc T.
  pose proof Hty.
  induction Hty; intros; subst.
  - eapply ty_has_N_closure_oc_or_dist_eq; eauto.
  - 
    destruct (orctx_le_cases ocp oc0).
    + (* ocp left of oc0 -> continue induction *)
      symmetry in HeqTT.
      forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
      destruct_all; subst.
      *
        eapply orctx_le_extend_l_inv in H1.
        subst; simpls. inversions H3.
      *
      forwards: orctx_separate_cross_r HeqTT H2 T0.
      destruct_all.
      forwards: H4 T.
      forwards: H4 U.
      destruct_all.

      forwards: IHHty1. eauto. symmetry; eapply H6.
      forwards: IHHty2. eauto. symmetry; eapply H7.

      rewrite H3.
      rewrite H8 in H13.
      rewrite H10 in H12.
      eapply ty_has_N_logic_oc_and; eauto.
    +
      assert (orctx_notorand_le oc0). {
      eapply orctx_notorand_le_le. eauto. symmetry; eauto.
      eapply ty_has_N_logic_ord_orctx_notorand_le; eauto.
      }
      eapply ty_has_N_logic_ord_oc_nle_or_diag.
      eauto.
      rewrite HeqTT in H.
      eapply ty_has_N_logic_ord_notorand_le; eauto.
Qed.

Lemma ty_has_N_logic_oc_or_diag G e oc T:
  ty_has_N_logic G e (orctx_typ oc (typ_or T T)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hty.
  eapply ty_has_N_logic_logic_ord in Hty.
  eapply ty_has_N_logic_ord_oc_or_diag; eauto.
Qed.

Lemma ty_has_N_logic_oclc_or_diag G e oc oc2 T:
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 (typ_or T T))) ->
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 T)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_has_N_logic_oc_or_diag; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_has_N_logic_lc_or_diag G e oc T:
  ty_has_N_logic G e (logctx_typ oc (typ_or T T)) ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv Hty.
  eapply ty_has_N_logic_oclc_or_diag with (oc := orctx_nil).
  simpls; eauto.
Qed.

End WithTypingMode.