Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath.

Require Import LayeredCommon.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Atomic subtyping *)
(* Subtyping for functions and declarations *)
(* Figure 20 in "Pure methods for roDOT" *)
Inductive subtyp_atomic: ctx -> renv -> typ -> typ -> Prop :=
(** This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_atomic_fun : forall L G e S T S' T',
  mode_has_fun typing_mode ->
  G⁏e ⊢# S' <: S ->
  (forall y, y \notin L ->
   G § y / v_par ~ S' ⁏e ⊢ open y T <: open y T') ->
  subtyp_atomic G e ( typ_fun S T ) ( typ_fun S' T') 

(** [G ⊢# T1 <: T2]           #<br>#
    [G ⊢# U1 <: U2]           #<br>#
    [――――――――――――――――――――(LSA-Fld)] #<br>#
    [G ⊢a {a: T2..U1} <: {a: T1..U2}] *)
| subtyp_atomic_fld : forall G e a TL1 TL2 TU1 TU2,
  G⁏e ⊢# TL2 <: TL1 ->
  G⁏e ⊢# TU1 <: TU2 ->
  subtyp_atomic G e (typ_rcd (dec_fld a TL1 TU1)) (typ_rcd (dec_fld a TL2 TU2))

(** [G ⊢# S2^x <: S1^x]                   #<br>#
    [G ⊢# T1^x <: T2^x]                   #<br>#
    [x fresh]                     #<br>#
    [――――――――――――――――――――――――――――――(LSA-Typ)] #<br>#
    [G ⊢a {B: S1..T1} <: {B: S2..T2}]     *)
| subtyp_atomic_typ : forall G e A TL1 TL2 TU1 TU2 ,
  (forall y, G⁏e ⊢# open y TL2 <: open y TL1) ->
  (forall y, G⁏e ⊢# open y TU1 <: open y TU2) ->
  subtyp_atomic G e  (typ_rcd (dec_typ A TL1 TU1)) (typ_rcd (dec_typ A TL2 TU2))


(** [G ⊢# TZ2 <: TZ1]                #<br>#
    [G, x: TZ2 ⊢# TC2^x <: TC1^x]     #<br>#
    [G, x: TZ2, y: TC2^x ⊢ TR1^x^y <: TR2^x^y]     #<br>#
    [x,y fresh]                     #<br>#
    [―――――――――――――――――――――――――――――――――――――――――(LSA-Met)]     #<br>#
    [G ⊢a {m(TC1, TZ1): TR1} <: {m(TC2, TZ2): TR2}}]      *)
| subtyp_atomic_met : forall L G e  m TC1 TZ1 TR1 TC2 TZ2 TR2,
  G⁏e ⊢# TZ2 <: TZ1 ->
  (forall y, y \notin L -> (* y is the parameter *)
   G § y / v_par ~ TZ2 ⁏e ⊢ open y TC2 <: open y TC1) ->
  (* If typ_indep_b 0 TC2 then typ_indep_b 0 TC1 in modes with sef *)
  (mode_has_sef typing_mode -> typ_indep_b 0 TC2 -> typ_indep_b 0 TC1) ->
  (forall y, y \notin L ->
   forall z, z <> y -> z \notin L -> (* z is the receiver *)
   G § y / v_par ~ TZ2 § z / v_recv ~ (open y TC2) ⁏e ⊢ open y (open_rec 1 z TR1) <: open y (open_rec 1 z TR2) ) ->
  subtyp_atomic G e (typ_rcd (dec_met m TC1 TZ1 TR1)) (typ_rcd (dec_met m TC2 TZ2 TR2))
.

(* Atomic subtyping only applies to atomic types,
   but not to all atomic types *)

(** LHS of atomic subtyping is an atomic type *)
Lemma subtyp_atomic_atomic_l G e T U:
  subtyp_atomic G e T U ->
  typ_atomic T.
Proof.
  introv Hty.
  induction Hty; eauto.
Qed.

(** RHS of atomic subtyping is an atomic type *)
Lemma subtyp_atomic_atomic_r G e T U:
  subtyp_atomic G e T U ->
  typ_atomic U.
Proof.
  introv Hty.
  induction Hty; eauto.
Qed.

Lemma subtyp_atomic_trans G e S T U:
  inert G ->
  ok e ->
  subtyp_atomic G e S T ->
  subtyp_atomic G e T U ->
  subtyp_atomic G e S U.
Proof.
  introv HiG Hoke Hs1 Hs2.
  induction Hs1; intros; inversions Hs2.
  - eapply subtyp_atomic_fun; eauto.
    intro_fresh y Hy.
    eapply subtyp_trans. eapply narrow; eauto. unfold ctx_obj; simpls.
    econstructor; eauto. eapply tight_to_general; eauto.
    eauto.
  - eapply subtyp_atomic_fld; eauto.
  - eapply subtyp_atomic_typ; eauto.
  - eapply subtyp_atomic_met; eauto.
    +
    intro_fresh y Hy.
    eapply subtyp_trans. eauto.
    eapply narrow; eauto. unfold ctx_obj; simpls.
    econstructor; eauto. eapply tight_to_general; eauto.
    +
    intro_fresh y Hy.
    introv Hnzy Hz.
    eapply subtyp_trans.
    eapply narrow; eauto.
    econstructor; eauto.
    econstructor; eauto. eapply tight_to_general; eauto.
    eapply narrow; eauto.
Qed.

(* Atomic subtyping is closed under equivalence on RHS *)
Lemma subtyp_atomic_tyq G e S U V:
  inert G ->
  subtyp_atomic G e S U ->
  typing_equiv e U V ->
  subtyp_atomic G e S V.
Proof.
  introv Hin Hsu Htyq.
  induction Hsu; intros.
  - destruct (tyq_typ_fun_inv Htyq) as [t11 [t12 [Ht1 [Htyq11 Htyq12]]]]; subst.
    apply tyq_typ_sym in Htyq11.
    eapply subtyp_atomic_fun; eauto.
    intros.
    eapply (subtyp_trans).
    eapply narrow.
    apply (H1 y).
    apply notin_union_r2 with (E:= ctx_dom G). eauto.
    unfold ctx_obj. constructor; eauto.
    eauto using subctx_last.
    apply subtyp_tyq; eauto.
  -
    inversions_tyq Htyq.
    symmetry in H6.
    eapply subtyp_atomic_fld; eauto.
  -
    inversions_tyq Htyq.
    symmetry in H6.
    eapply subtyp_atomic_typ; eauto.
 -  destruct (tyq_typ_dec_met_inv Htyq) as [t11 [t12 [t13 [Ht [Htyq1 [Htyq2 Htyq3]]]]]]; subst.
    eapply subtyp_atomic_met.
    * apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans_t.
      apply subtyp_tyq_t. eauto. eauto.
    * intros.
      specialize (Htyq1 y).
      apply tyq_typ_sym in Htyq1.
      apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans.
      apply subtyp_tyq. eauto.
      eapply narrow.
      apply H0.
      apply notin_union_r2 with (E:= ctx_dom G). eauto.
      eauto. eauto.
    * intros. eapply H1. eauto. eapply typ_indep_tyq_open_all_indep. eauto. intros. symmetry. eauto.
    * intros.
      specialize (Htyq3 y z).
      specialize (Htyq1 y).
      apply tyq_typ_sym in Htyq1.
      apply tyq_typ_sym in Htyq2.
      eapply subtyp_trans.
      eapply narrow. auto. eauto.
      constructor; eauto.
      eauto.
Qed.

Lemma subtyp_atomic_tight G e S T:
  subtyp_atomic G e S T ->
  G⁏e ⊢# S <: T.
Proof.
  introv Hsu.
  induction Hsu.
  eapply subtyp_fun_t; eauto. (* Need explicit eapply to avoid existential variable *)
  eauto.
  eauto.
  eauto.
Qed.

End WithTypingMode.