Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* For the purpose of finding a common path of two derivations in the main layer,
  this relation allows changing the starting point on the main layer.
  This is possible because logic-layer typing is closed under this relation.
  The rules are somewhat inverse to the rules on the main layer, but not exactly.
 *)
Inductive ty_main_prep (G: ctx)(e:renv): typ -> typ -> Prop :=
| ty_main_prep_refl T:
  ty_main_prep G e T T
| ty_main_prep_nocap T oc U:
  ty_main_prep G e T (logctx_typ oc typ_nocap) ->
  ty_has_N_logic G e U ->
  ty_main_prep G e T (logctx_typ oc U)
| ty_main_prep_sel T oc y A z U:
  ty_main_prep G e T (logctx_typ oc (typ_sel (avar_f y) A (avar_f z))) ->
  ty_main_prep G e T (logctx_typ oc U)
| ty_main_prep_bnd T oc U U1:
  ty_main_prep G e T (logctx_typ oc (typ_bnd U1)) ->
  ty_main_prep G e T (logctx_typ oc U)
.
Section WithContext.
Context (G: ctx)(e: renv).


Lemma ty_main_prep_trans T1 T2 T3: 
  ty_main_prep G e T1 T2 ->
  ty_main_prep G e T2 T3 ->
  ty_main_prep G e T1 T3.
Proof.
  introv Htr1 Htr2.
  reverts T1.
  induction Htr2; intros.
  - eauto.
  - forwards: IHHtr2 Htr1; eauto.
    eapply ty_main_prep_nocap; eauto.
  - forwards: IHHtr2 Htr1; eauto.
    eapply ty_main_prep_sel; eauto.
  - forwards: IHHtr2 Htr1; eauto.
    eapply ty_main_prep_bnd; eauto.

Qed.

Lemma ty_main_prep_lc T1 T2 oc: 
  ty_main_prep G e T1 T2 ->
  ty_main_prep G e (logctx_typ oc T1) (logctx_typ oc T2).
Proof.
  introv Htr.
  induction Htr.
  - eapply ty_main_prep_refl; eauto.
  - rewrite <- logctx_extend_typ in *. eapply ty_main_prep_nocap; eauto.
  - rewrite <- logctx_extend_typ in *. eapply ty_main_prep_sel; eauto.
  - rewrite <- logctx_extend_typ in *. eapply ty_main_prep_bnd; eauto.
Qed.

Lemma ty_main_prep_or T1 T2 T3 T4: 
  ty_main_prep G e T1 T2 ->
  ty_main_prep G e T3 T4 ->
  ty_main_prep G e (typ_or T1 T3) (typ_or T2 T4).
Proof.
  introv Htr1 Htr2.
  replace (typ_or T1 T3) with (logctx_typ logctx_nil (typ_or T1 T3)) by (simpl; eauto).
  rewrite (logctx_extend_orl logctx_nil T1).
  eapply ty_main_prep_trans.
  eapply ty_main_prep_lc. eauto.
  rewrite logctx_extend_typ.
  simpls.
  replace (typ_or T2 T4) with (logctx_typ logctx_nil (typ_or T2 T4)) by (simpl; eauto).
  replace (typ_or T2 T3) with (logctx_typ logctx_nil (typ_or T2 T3)) by (simpl; eauto).
  rewrite (logctx_extend_orr logctx_nil T2).
  rewrite (logctx_extend_orr logctx_nil T2).
  eapply ty_main_prep_lc. eauto.
Qed.

Lemma ty_main_prep_and T1 T2 T3 T4: 
  ty_main_prep G e T1 T2 ->
  ty_main_prep G e T3 T4 ->
  ty_main_prep G e (typ_and T1 T3) (typ_and T2 T4).
Proof.
  introv Htr1 Htr2.
  replace (typ_and T1 T3) with (logctx_typ logctx_nil (typ_and T1 T3)) by (simpl; eauto).
  rewrite (logctx_extend_andl logctx_nil T1).
  eapply ty_main_prep_trans.
  eapply ty_main_prep_lc. eauto.
  rewrite logctx_extend_typ.
  simpls.
  replace (typ_and T2 T4) with (logctx_typ logctx_nil (typ_and T2 T4)) by (simpl; eauto).
  replace (typ_and T2 T3) with (logctx_typ logctx_nil (typ_and T2 T3)) by (simpl; eauto).
  rewrite (logctx_extend_andr logctx_nil T2).
  rewrite (logctx_extend_andr logctx_nil T2).
  eapply ty_main_prep_lc. eauto.
Qed.

Hint Resolve ty_main_prep_or ty_main_prep_and: core.


Lemma ty_main_prep_or_inv T1 T2 U: 
  ty_main_prep G e (typ_or T1 T2) U ->
  exists T3 T4,
  U = (typ_or T3 T4) /\
  ty_main_prep G e T1 T3 /\
  ty_main_prep G e T2 T4.
Proof.
  introv Htr.
  remember  (typ_or T1 T2).
  reverts T1 T2.
  induction Htr; intros; subst.
  - exists T1 T2. splits; auto; constructor.
  - forwards: IHHtr; eauto.
    destructecs H0.
     destruct oc; simpls; inversions H1.
    + exists (logctx_typ oc U) T4. splits; eauto. econstructor; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. econstructor; eauto.
  - forwards: IHHtr; eauto.
    destructecs H.
     destruct oc; simpls; inversions H0.
    + exists (logctx_typ oc U) T4. splits; eauto. eapply ty_main_prep_sel; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. eapply ty_main_prep_sel; eauto.
  - forwards: IHHtr; eauto.
    destructecs H.
     destruct oc; simpls; inversions H0.
    + exists (logctx_typ oc U) T4. splits; eauto. eapply ty_main_prep_bnd; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. eapply ty_main_prep_bnd; eauto.
Qed.

Lemma ty_main_prep_and_inv T1 T2 U: 
  ty_main_prep G e (typ_and T1 T2) U ->
  exists T3 T4,
  U = (typ_and T3 T4) /\
  ty_main_prep G e T1 T3 /\
  ty_main_prep G e T2 T4.
Proof.
  introv Htr.
  remember  (typ_and T1 T2).
  reverts T1 T2.
  induction Htr; intros; subst.
  - exists T1 T2. splits; auto; constructor.
  - forwards: IHHtr; eauto.
    destructecs H0.
     destruct oc; simpls; inversions H1.
    + exists (logctx_typ oc U) T4. splits; eauto. econstructor; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. econstructor; eauto.
  - forwards: IHHtr; eauto.
    destructecs H.
     destruct oc; simpls; inversions H0.
    + exists (logctx_typ oc U) T4. splits; eauto. eapply ty_main_prep_sel; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. eapply ty_main_prep_sel; eauto.
  - forwards: IHHtr; eauto.
    destructecs H.
     destruct oc; simpls; inversions H0.
    + exists (logctx_typ oc U) T4. splits; eauto. eapply ty_main_prep_bnd; eauto.
    + exists T3 (logctx_typ oc U). splits; eauto. eapply ty_main_prep_bnd; eauto.

Qed.

Lemma ty_main_prep_lc_inv oc T U: 
  ty_main_prep G e (logctx_typ oc T) U ->
  exists oc2 T2,
  U = (logctx_typ oc2 T2) /\
  logctx_follow oc oc2 /\ ty_main_prep G e T T2.
Proof.
  introv Htr.
  reverts T U.
  induction oc; intros; simpls.
  - exists logctx_nil U. splits; eauto.
  - eapply ty_main_prep_or_inv in Htr. destructecs Htr; subst.
    forwards: IHoc; eauto. destructecs H; subst. exists  (logctx_orl oc2 T4) T2. simpls; splits; eauto.

  - eapply ty_main_prep_or_inv in Htr. destructecs Htr; subst.
    forwards: IHoc; eauto. destructecs H; subst. exists  (logctx_orr T3 oc2) T2. simpls; splits; eauto.
  - eapply ty_main_prep_and_inv in Htr. destructecs Htr; subst.
    forwards: IHoc; eauto. destructecs H; subst. exists  (logctx_andl oc2 T4) T2. simpls; splits; eauto.

  - eapply ty_main_prep_and_inv in Htr. destructecs Htr; subst.
    forwards: IHoc; eauto. destructecs H; subst. exists  (logctx_andr T3 oc2) T2. simpls; splits; eauto.
Qed.

Lemma ty_main_prep_lc_replace T1 T2 T3 T4 oc1 oc2: 
  ty_main_prep G e T1 T2 ->
  logctx_follow oc1 oc2 ->
  ty_main_prep G e (logctx_typ oc1 T3) (logctx_typ oc2 T4) ->
  ty_main_prep G e (logctx_typ oc1 T1) (logctx_typ oc2 T2).
Proof.
  introv Htr Hf Htroc.
  reverts T1 T2 T3 T4.
  reverts oc2.
  induction oc1; intros; subst; simpls; destruct oc2; inversions Hf; simpls.
  - auto.
  - eapply ty_main_prep_or_inv in Htroc. destructecs Htroc. inversions Htroc0. forwards: IHoc1; eauto.
  - eapply ty_main_prep_or_inv in Htroc. destructecs Htroc. inversions Htroc0. forwards: IHoc1; eauto.
  - eapply ty_main_prep_and_inv in Htroc. destructecs Htroc. inversions Htroc0. forwards: IHoc1; eauto.
  - eapply ty_main_prep_and_inv in Htroc. destructecs Htroc. inversions Htroc0. forwards: IHoc1; eauto.
Qed.

End WithContext.

End WithTypingMode.

Hint Resolve ty_main_prep_or ty_main_prep_and: core.