Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping TightTypingHelpers.
Require Import Orctx Logctx OrctxLogctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityMain.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Atomic typing.
   One layer above precise typing.
   Gives a variable declaration types, which can be equivalent or direct supertypes of its precise types.
*)
(* Figure 15 in "Pure methods for roDOT" *)
Inductive ty_inv_atomic: ctx -> renv -> var -> typ -> Prop :=
(** The precise declaration type *)
(** [G;e ⊢! x : T1]  #<br>#
    [e ⊢a T1 ≈ T2]  #<br>#
    [――――――――(LTA-Prec)]  #<br>#
    [G;e ⊢a x : T]  *)
| ty_inv_atomic_precise G e x T U V:
  G ⊢! x ∶ T ⪼ U ->
  typ_atomic U ->
  typing_equiv e U V ->
  ty_inv_atomic G e x V
(** A direct supertype of the precise declaration type *)
(** [G ⊢a x : T1]  #<br>#
    [G ⊢a T1 <: T2]  #<br>#
    [――――――――(LTA-Sub)]  #<br>#
    [G ⊢a x : T2]  *)
| ty_inv_atomic_subtyp G e x S T:
  ty_inv_atomic G e x S ->  
  subtyp_atomic G e S T ->
  ty_inv_atomic G e x T
.

(* ty_inv_atomic closed under typing_equiv *)
Lemma ty_inv_atomic_tyq G e x U V:
  inert G ->
  ty_inv_atomic G e x U ->
  typing_equiv e U V ->
  ty_inv_atomic G e x V.
Proof.
  introv Hin Hty Htyq.
  induction Hty; intros.
  - (* ty_inv_atomic_precise *)
    eapply ty_inv_atomic_precise; eauto using tyq_typ_trans. 
  - (* ty_inv_atomic_subtyp *)
    eapply ty_inv_atomic_subtyp.
    eauto.
    eapply subtyp_atomic_tyq; eauto.
Qed.

(* ty_inv_atomic only gives atomic types *)
Lemma ty_inv_atomic_atomic G e x U:
  ty_inv_atomic G e x U ->
  typ_atomic U.
Proof.
  introv Hty.
  induction Hty; eauto.
  eapply typ_atomic_tyq; eauto.
  eapply subtyp_atomic_atomic_r; eauto.
Qed.

Lemma ty_inv_atomic_oc_nil G e x oc T:
  ty_inv_atomic G e x (orctx_typ oc T) ->
  oc = orctx_nil.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply orctx_atomic_inv in H; destruct_all; assumption.
Qed.

Lemma ty_inv_atomic_lc_nil G e x oc T:
  ty_inv_atomic G e x (logctx_typ oc T) ->
  oc = logctx_nil.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all; assumption.
Qed.

Lemma ty_inv_atomic_lc_inv G e x oc T:
  ty_inv_atomic G e x (logctx_typ oc T) ->
  ty_inv_atomic G e x T.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_lc_nil Hty.
  subst; eauto.
Qed.

(* Atomic typing never gives intersection, bottom, selection, recursive, union, nocap types *)
Lemma ty_inv_atomic_oc_and_inv G e x T oc U:
  ty_inv_atomic G e x (orctx_typ oc (typ_and T U)) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply orctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.
Lemma ty_inv_atomic_lc_and_false G e x oc T S:
  ty_inv_atomic G e x (logctx_typ oc (typ_and T S)) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

Lemma ty_inv_atomic_lc_bot_false G e x oc:
  ty_inv_atomic G e x (logctx_typ oc typ_bot) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

Lemma ty_inv_atomic_lc_sel_false G e x oc z A y:
  ty_inv_atomic G e x (logctx_typ oc (typ_sel z A y)) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

Lemma ty_inv_atomic_lc_bnd_false G e x oc T:
  ty_inv_atomic G e x (logctx_typ oc (typ_bnd T)) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

Lemma ty_inv_atomic_lc_or_false G e x oc T S:
  ty_inv_atomic G e x (logctx_typ oc (typ_or T S)) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

Lemma ty_inv_atomic_lc_nocap_false G e x oc:
  ty_inv_atomic G e x (logctx_typ oc typ_nocap) ->
  False.
Proof.
  introv Hty.
  forwards: ty_inv_atomic_atomic; eauto.
  eapply logctx_atomic_inv in H; destruct_all.
  inversions H0.
Qed.

(* ty_inv_atomic closed under subtyp_atomic *)
Lemma ty_inv_atomic_lc_typ_subtyp G e x oc U V:
  inert G ->
  ty_inv_atomic G e x (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_inv_atomic G e x (logctx_typ oc V).
Proof.
  introv Hin Hty Htyq.
  forwards: ty_inv_atomic_atomic Hty.
  eapply logctx_atomic_inv in H.
  destruct_all; subst; simpls.
  eapply ty_inv_atomic_subtyp; eauto.
Qed.

(* Bounds of type members given by atomic typing can be replaced by the precise bounds *)
Lemma ty_inv_atomic_typ_pf_replace G e x oc U A SL SU TL TU:
  inert G ->
  G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hin Hpf Hty.
  forwards: ty_inv_atomic_lc_nil Hty; subst; simpls.
  dependent induction Hty.
  - eapply ty_inv_atomic_precise; eauto.
  - inversions H. eapply IHHty; eauto.
Qed.

(* Bounds of type members given by atomic typing can be replaced by any bounds if the type member is not real *)
Lemma ty_inv_atomic_typ_nopf_false G e x oc A SL SU:
  inert G ->
  ty_pf_not_typ G x A ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  False.
Proof.
  introv Hin Hpf Hty.
  forwards: ty_inv_atomic_lc_nil Hty; subst; simpls.
  dependent induction Hty.
  - inversions_tyq H1. false. eapply Hpf. eauto.
  - inversions H. eauto.
Qed.

(* Bounds of type members given by atomic typing can be replaced by by the precise bounds or any bounds if the type member is not real *)
Lemma ty_inv_atomic_typ_pfor_replace G e x oc A SL SU TL TU:
  inert G ->
  ty_pf_or_none G x A TL TU ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hin Hpf Hty.
  destruct Hpf; destruct_all.
  - eauto using ty_inv_atomic_typ_pf_replace.
  - false; eauto using ty_inv_atomic_typ_nopf_false.
Qed.

Lemma ty_inv_atomic_typ_lc_pf_inv G e x oc U A SL SU TL TU:
  inert G ->
  G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (forall y : var, G ⁏ e ⊢# open y SL <: open y TL) /\ (forall y : var, G ⁏ e ⊢# open y TU <: open y SU).
Proof.
  introv Hin Hpf Hty.
  forwards: ty_inv_atomic_lc_nil Hty; subst; simpls.
  remember (typ_rcd (dec_typ A SL SU)) as S. reverts SL SU.
  induction Hty; intros; subst.
  - inversions_tyq H1.
    lets Hu: (x_bound_unique Hpf H). subst.
    pose proof (pf_inert_unique_tight_bounds Hin Hpf H) as Hu. destruct Hu. subst.
    symmetry in H6. (* Must swap for automation to work correctly *)
    eauto.
  - inversions H.
    forwards: IHHty TL1 TU1. eauto. eauto. eauto.
    destruct_all.
    eauto.
Qed.

Lemma ty_inv_atomic_lc_typ_also_dist G e x oc A SL SU TL TU:
  inert G ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_atomic G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hin HtyT HtyU.
  forwards: ty_inv_atomic_lc_nil HtyT; subst; simpls.
  remember  (typ_rcd (dec_typ A TL TU)).
  reverts TL TU.
  induction HtyT; intros; subst.
  -
    inversions_tyq H1.
    forwards Hsu: ty_inv_atomic_typ_lc_pf_inv logctx_nil H; fweauto.
    destructecs Hsu.
    eapply ty_inv_atomic_subtyp with (S:= (typ_rcd (dec_typ A t11 t12))).
    eapply ty_inv_atomic_precise. eapply H. eauto. eauto.
    constructor; intros.
    eapply subtyp_or2_t. eapply subtyp_tyq_t; symmetry; eauto. eauto.
    eapply subtyp_and2_t. eapply subtyp_tyq_t. eauto. eauto.
  -
    inversions H.
    forwards: IHHtyT HtyU; fweauto.
    eapply ty_inv_atomic_subtyp . eauto.
    constructor; intros.
    eapply subtyp_or_tight; eauto.
    eapply subtyp_and_tight; eauto.
Qed.

Lemma ty_inv_atomic_M G e w:
  mode_has_mutability typing_mode ->
  inert G ->
  renv_corr G e ->
  w \in ctx_dom G ->
  ty_inv_atomic G e w (typ_mut_marker typ_top).
Proof.
  introv Hm Hi Hcor Hdom.
  eapply ctx_dom_to_binds in Hdom. destructecs Hdom.
  forwards Hpf: pf_mutability; eauto.
  destructecs Hpf.
  eapply ty_inv_atomic_subtyp.
  eapply ty_inv_atomic_precise. eauto.
  eauto.
  reflexivity.
  constructor; intros; simpls; eauto.
  eauto.
  eauto.
Qed.


Lemma ty_inv_atomic_has_top_atomic G e w T:
  inert G ->
  renv_corr G e ->
  w \in ctx_dom G ->
  ty_has_top_atomic G e T ->
  ty_inv_atomic G e w T.
Proof.
  introv Hi Hcor Hdom Hty.
  inversions Hty.
  eapply ty_inv_atomic_subtyp.
  eapply ty_inv_atomic_M.
  destruct (mode_has_mutability' typing_mode) eqn: Hm. auto. typing_mode_contradiction typing_mode.
  eauto. eauto.
  eauto.
  constructor; intros; eauto.
Qed.

Lemma ty_inv_atomic_ro_ref G e w y T:
  inert G ->
  renv_corr G e ->
  binds w y e ->
  ty_has_N_main G e T ->
  ty_inv_atomic G e y T ->
  ty_inv_atomic G e w T.
Proof.
  introv Hin Hcor Hbi Hh Hty.
  forwards Hbic: renv_corr_binds_dom Hcor; fweauto. destructecs Hbic.
  induction Hty.
  -
    forwards: ty_has_N_atomic_atomic Hh.
    eapply typ_atomic_tyq; eauto.
    forwards: ty_has_N_tyq. eauto. symmetry. eauto.

    inversions H3.
    +
      forwards HT2: ty_has_N_atomic_pf_tyq x w; eauto.
      destructecs HT2.
      eapply ty_inv_atomic_precise; eauto.
      eapply typ_atomic_tyq; eauto.
      eapply tyq_typ_trans. symmetry; eauto. eauto.
    +
      forwards HhV: ty_has_top_atomic_tyq H1; fweauto.
      eapply ty_inv_atomic_has_top_atomic; eauto.
  - forwards Hat: ty_has_N_atomic_atomic Hh.
    eapply subtyp_atomic_atomic_r; eauto.
    inversions Hat.
    +
      eapply ty_inv_atomic_subtyp. 2: eauto. eapply IHHty; eauto.
      eapply ty_has_N_main_logic.
      eapply LayeredCapabilityLogic.ty_has_N_logic_or.
      eapply ty_or_closure_atomic.
      eapply ty_has_N_has_atomic. eauto.
      eapply ty_has_N_atomic_lc_rev_subtyp with (oc:=logctx_nil); eauto.
    +
      eapply ty_inv_atomic_has_top_atomic; eauto.
Qed.

(* The typing is needed because function types are always atomic, but not always supertypes of typ_nocap *)
Lemma ty_inv_atomic_has_NM_cases G e w T:
  inert G ->
  ty_inv_atomic G e w T ->
  ty_has_N_atomic G e T \/ ty_has_M G e T.
Proof.
  introv Hi Hty.
  induction Hty.
  - 
    inversions H0.
    +
    inversions_tyq H1; eauto.
    destruct B; fweauto.
    destruct t; fweauto.
    right.
    constructor; intros.
    assert (G ⁏ e ⊢# t11 /[-> x0] <: typ_bot). {
      eapply ty_pf_mutability_bound. auto. eauto.
      destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.

    }
    eapply subtyp_trans_t. eapply subtyp_tyq_t. symmetry. eauto. eauto.
    
    +
    inversions_tyq H1.
    forwards: pf_fun_inv_mode; eauto.
  -
  destruct IHHty.
  auto.
  left. eapply ty_has_N_atomic_lc_subtyp with (oc:=logctx_nil); eauto.
  right. eapply ty_has_M_lc_subtyp with (oc:=logctx_nil); eauto.
Qed.

Lemma ty_inv_atomic_has_NM_replace G e w oc T U:
  inert G ->
  ty_inv_atomic G e w (logctx_typ oc T) ->
  ty_has_N_atomic G e T \/ ty_has_M G e T \/  ty_inv_atomic G e w (logctx_typ oc U).
Proof.
  introv Hi Hty.
  forwards: ty_inv_atomic_atomic; fweauto.
  eapply logctx_atomic_inv in H; destruct_all. subst. simpls.
  forwards: ty_inv_atomic_has_NM_cases; fweauto.
  destruct H; auto 3.
Qed.

Lemma ty_inv_atomic_tight G e x T:
    inert G ->
    ty_inv_atomic G e x T ->
    G⁏e ⊢# x ∶ T.
Proof.
  introv Hi Hty.
  induction Hty.
  eapply ty_sub_t.
  eapply precise_to_tight; eauto.
  eapply subtyp_tyq_t. eauto.
  forwards: subtyp_atomic_tight H; eauto.
Qed.

Lemma ty_inv_atomic_dom G e x T:
  ty_inv_atomic G e  x T ->
  x \in ctx_dom G.
Proof.
  introv Hty.
  induction Hty.
  eapply ctx_binds_dom.
  eapply pf_binds. eauto.
  eauto.
Qed.

End WithTypingMode.

