Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping TightTypingHelpers.
Require Import Orctx Logctx OrctxLogctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityMain.
Require Import LayeredTypingAtomic.

Section WithTypingMode.
Context {typing_mode: Mode}.


(** * Basic typing.
   Gives variables their declaration types,
   and in modes with NM,
   top-like types of the NM form.

   Warning: does not imply x \in ctx_dom G 
*)
(* Figure 16 in "Pure methods for roDOT" *)
Inductive ty_inv_basic: ctx -> renv -> var -> typ -> Prop :=
(** [G ⊢a x : T]  #<br>#
    [――――――――(LTB-Atom)]  #<br>#
    [G ⊢b x : T]  *)
| ty_inv_basic_atomic G e x T:
    ty_inv_atomic G e x T ->
    ty_inv_basic G e x T
(** [G;e ⊢c N◁ T1]  #<br>#
    [G;e ⊢c M◁ T2]  #<br>#
    [――――――――(LTB-NM)]  #<br>#
    [G;e ⊢b x : T1 ∨ T2]  *)
| ty_inv_basic_NM G e x  T U:
    mode_has_top_nm_subtype typing_mode ->
    ty_has_N_closure G e T ->
    ty_has_M_closure G e U ->
    ty_inv_basic G e x (typ_or T U)
(** [G;e ⊢c M◁ T1]  #<br>#
    [G;e ⊢c N◁ T2]  #<br>#
    [――――――――(LTB-MN)]  #<br>#
    [G;e ⊢b x : T1 ∨ T2]  *)
| ty_inv_basic_MN G e x T U:
    mode_has_top_nm_subtype typing_mode ->
    ty_has_M_closure G e T ->
    ty_has_N_closure G e U ->
    ty_inv_basic G e x (typ_or T U)
.

(* FORWARD CASES *)

Lemma ty_inv_basic_NM_canon G e x T1 T2:
  mode_has_top_nm_subtype typing_mode ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_inv_basic G e x (typ_NM T1 T2).
Proof.
  introv Hm Hbot.
  eapply ty_inv_basic_NM. eauto. eapply ty_or_closure_atomic. constructor. constructor.
  eapply ty_or_closure_atomic. constructor. eauto.
Qed.

(* TYPE PROPERTIES *)

Lemma ty_inv_basic_atomic_atomic G e x T:
  ty_inv_basic G e x T ->
  typ_atomic T ->
  ty_inv_atomic G e x T.
Proof.
  introv Hty Ha.
  induction Hty; eauto.
  inversions Ha.
  inversions Ha.
Qed.

(* CLOSED UNDER *)

(* ty_inv_basic closed under typing_equiv *)
Lemma ty_inv_basic_tyq G e x U V:
  inert G ->
  ty_inv_basic G e x U ->
  typing_equiv e U V ->
  ty_inv_basic G e x V.
Proof.
  introv Hin Hty Htyq.
  induction Hty.
  - eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_inv_basic_NM. eauto.
    eapply ty_has_N_closure_tyq; eauto.
    eapply ty_has_M_closure_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_inv_basic_MN. eauto.
    eapply ty_has_M_closure_tyq; eauto.
    eapply ty_has_N_closure_tyq; eauto.
Qed.

(* ty_inv_basic closed under subtyp_atomic *)
Lemma ty_inv_basic_lc_subtyp G e x oc U V:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_inv_basic G e x (logctx_typ oc V).
Proof.
  introv Hin Hty Hsu.
  remember  (logctx_typ oc U). reverts oc U.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_lc_typ_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt.
    eapply subtyp_atomic_atomic_l in Hsu. inversions Hsu.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_subtyp; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt.
    eapply subtyp_atomic_atomic_l in Hsu. inversions Hsu.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_subtyp; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_subtyp; eauto.
Qed.

(* INVERSIONS and REPLACEMENT *)

Lemma ty_inv_basic_oc_and_replace G e x T oc U S:
  ty_inv_basic G e x (orctx_typ oc (typ_and T U)) ->
  ty_inv_basic G e x (orctx_typ oc S).
Proof.
  induction oc; intros; simpls.
  - inversions H. eapply ty_inv_atomic_atomic in H0. inversions H0.
  - inversions H.
    +
      rewrite orctx_fold_l in H0.
      eapply ty_inv_atomic_oc_and_inv in H0. contradiction.
    +
      eapply ty_inv_basic_NM. eauto.
      eapply ty_has_N_closure_oc_and_inv; eauto. eauto.
    +
      eapply ty_inv_basic_MN. eauto.
      eapply ty_has_M_closure_oc_and_replace; eauto. eauto.
  - inversions H.
    +
      rewrite orctx_fold_r in H0.
      eapply ty_inv_atomic_oc_and_inv in H0. contradiction.
    +
      eapply ty_inv_basic_NM. eauto.
      eauto.
      eapply ty_has_M_closure_oc_and_replace; eauto.
    +
      eapply ty_inv_basic_MN. eauto.
      eauto.
      eapply ty_has_N_closure_oc_and_inv; eauto.
Qed.

Lemma ty_inv_basic_lc_bot_replace G e x oc U:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc typ_bot) ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  inversions Hty.
  - false. eapply ty_inv_atomic_lc_bot_false; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_bot_replace; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_bot_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_bot_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_bot_replace; eauto.
Qed.
Lemma ty_inv_basic_lc_nocap_replace G e x oc U:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc typ_nocap) ->
  ty_has_N G e U ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hin Hty HhN.
  inversions Hty.
  - false. eapply ty_inv_atomic_lc_nocap_false; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_replace; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_nocap_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_nocap_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_replace; eauto.
Qed.

Lemma ty_inv_basic_lc_M_replace G e x oc U:
  mode_has_top_nm_complement typing_mode ->
  inert G ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_M G e U ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hm Hin Hty HhN.
  inversions HhN.
  eapply ty_inv_basic_lc_subtyp; fweauto. 
  constructor; eauto.
Qed.

Lemma ty_inv_basic_lc_bnd_replace G e x oc U T:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc (typ_bnd T)) ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  inversions Hty.
  - false. eapply ty_inv_atomic_lc_bnd_false; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_bnd_replace; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_bnd_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_bnd_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_bnd_replace; eauto.
Qed.
Lemma ty_inv_basic_lc_sel_replace G e x oc U z A y:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc (typ_sel z A y)) ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  inversions Hty.
  - false. eapply ty_inv_atomic_lc_sel_false; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_sel_replace; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_sel_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_sel_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_sel_replace; eauto.
Qed.
Lemma ty_inv_basic_lc_and_replace G e x oc U S T:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc (typ_and S T)) ->
  ty_inv_basic G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  inversions Hty.
  - false. eapply ty_inv_atomic_lc_and_false; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_and_replace; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_and_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_and_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_and_replace; eauto.
Qed.

Lemma ty_inv_basic_lc_NM_replace G e x oc U T1 T2:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_inv_basic G e x (logctx_typ oc U) ->
  ty_inv_basic G e x (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hm Hin Hbot Hty.
  inversions Hty.
  - eapply ty_inv_atomic_atomic in H. eapply logctx_atomic_inv in H. destruct_all; subst; simpls.
    eapply ty_inv_basic_NM_canon; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM_canon; eauto.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_NM_replace; eauto.
    eapply ty_inv_basic_NM; eauto; eauto.
    eapply ty_has_M_closure_lc_NM_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_NM_canon; eauto. 
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_NM_replace; eauto.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_NM_replace; eauto.
Qed.

Lemma ty_inv_basic_lc_has_top_replace G e x oc U T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_inv_basic G e x (logctx_typ oc T).
Proof.
  introv Hin Hcor Hdom Hty Htop.
  inversions Hty.
  - eapply ty_inv_atomic_atomic in H. eapply logctx_atomic_inv in H. destruct_all; subst; simpls. eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
    eapply ty_inv_basic_NM; eauto. eapply ty_has_N_closure_lc_has_top_replace; eauto.
    eapply ty_inv_basic_NM; eauto. eapply ty_has_M_closure_lc_has_top_replace; eauto. 
  - destruct oc; simpls; inversions H.
    eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
    eapply ty_inv_basic_MN; eauto. eapply ty_has_M_closure_lc_has_top_replace; eauto. 
    eapply ty_inv_basic_MN; eauto.   eapply ty_has_N_closure_lc_has_top_replace; eauto.
Qed.


Lemma ty_inv_basic_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  inert G ->
  x \in ctx_dom G ->
  ty_pf_or_none G x A TL TU ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hin Hdom Hpf Hty.
  dependent induction Hty; intros; subst.
  - left. eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_typ_pfor_replace; eauto.
  - destruct oc; simpls; inversions x.
    forwards Hhc: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hhc.
    left.
    eapply ty_inv_basic_NM; eauto.
    eauto.
    left.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_typ_pfor_replace; eauto.
  - destruct oc; simpls; inversions x.
    left.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_typ_pfor_replace; eauto.

    forwards Hhc: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hhc.
    left.
    eapply ty_inv_basic_MN; eauto.
    eauto.
Qed.


Lemma ty_has_top_atomic_subtyp_distr G e A Sl Su Tl Tu:
  inert G ->
  ty_has_top_atomic G e (typ_rcd (dec_typ A Sl Su)) ->
  subtyp_atomic G e (typ_rcd (dec_typ A Tl Tu)) (typ_rcd (dec_typ A (typ_or Tl Sl) (typ_and Tu Su))).
Proof.
  introv Hin Hh.
  inversions Hh.
  constructor; intros.
  apply subtyp_or2_t. auto. eapply subtyp_trans_t. eapply H3. eauto.
  apply subtyp_and2_t. auto. eapply subtyp_trans_t. 2: eapply H4. eauto.
Qed.


Lemma ty_has_top_atomic_subtyp_distl G e A Sl Su Tl Tu:
  inert G ->
  ty_has_top_atomic G e (typ_rcd (dec_typ A Tl Tu)) ->
  subtyp_atomic G e (typ_rcd (dec_typ A Sl Su)) (typ_rcd (dec_typ A (typ_or Tl Sl) (typ_and Tu Su))).
Proof.
  introv Hin Hh.
  inversions Hh.
  constructor; intros.
  apply subtyp_or2_t.  eapply subtyp_trans_t. eapply H3. eauto. auto.
  apply subtyp_and2_t. eapply subtyp_trans_t. 2: eapply H4. eauto.  auto.
Qed.


Lemma ty_inv_basic_lc_typ_pf_inv G e x oc U A SL SU TL TU:
  inert G ->
  G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (
    (forall SL2 SU2,
      ((forall y : var, G ⁏ e ⊢# open y SL <: typ_bot) -> (forall y : var, G ⁏ e ⊢# open y SL2 <: typ_bot)) ->
      ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A SL2 SU2))))
  \/
    (forall y : var, G ⁏ e ⊢# open y SL <: open y TL) /\ (forall y : var, G ⁏ e ⊢# open y TU <: open y SU)
  ).
Proof.
    introv Hin Hpf Hty.
  dependent induction Hty; intros; subst.
  - right.
    eapply ty_inv_atomic_typ_lc_pf_inv; eauto.
  -
    destruct oc; simpls; inversions x.
    +
    forwards Hrep: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hrep.
    *
    left. 
    intros.
    eapply ty_inv_basic_NM; eauto.
    * inversions Hrep. eauto.
    +

    left. intros.
    eapply ty_inv_basic_NM; eauto.

forwards: ty_has_M_closure_lc_typ_replace' H1.
destructecs H3. eauto.

    eapply ty_has_M_closure_lc_typ_replace. eauto. eauto.


  -
    destruct oc; simpls; inversions x.
    +

    left. intros.
    eapply ty_inv_basic_MN; eauto.

forwards: ty_has_M_closure_lc_typ_replace' H0.
destructecs H3. eauto.

    eapply ty_has_M_closure_lc_typ_replace. eauto. eauto.
    
    +
    forwards Hrep: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hrep.
    *
    left. 
    intros.
    eapply ty_inv_basic_MN; eauto.
    * inversions Hrep. eauto.
Qed.

Lemma ty_inv_basic_ro_ref G e w y T:
  inert G ->
  renv_corr G e ->
  binds w y e ->
  ty_has_N_main G e T ->
  ty_inv_basic G e y T ->
  ty_inv_basic G e w T.
Proof.
  introv Hin Hcor Hbi Hh Hty.
  induction Hty.
  - eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_ro_ref; eauto.
  -     eapply ty_inv_basic_NM; eauto.
  -     eapply ty_inv_basic_MN; eauto.
Qed.

Lemma ty_inv_basic_has_NM_cases G e w T:
  inert G ->
  ty_inv_basic G e w T ->
  ty_has_N_closure G e T \/ ty_has_M_closure G e T.
Proof.
  introv Hi Hty.
  induction Hty.
  - forwards Hh: ty_inv_atomic_has_NM_cases; eauto.
    destructecs Hh.
    left. eapply ty_or_closure_atomic; eauto.
    right. eapply ty_or_closure_atomic; eauto.
  - left.
    eapply ty_or_closure_l; eauto.
  - left.
    eapply ty_or_closure_r; eauto.
Qed.


Lemma ty_inv_basic_has_NM_replace G e w oc T U:
  inert G ->
  ty_inv_basic G e w (logctx_typ oc T) ->
  ty_has_N_closure G e T \/ ty_has_M_closure G e T \/  ty_inv_basic G e w (logctx_typ oc U).
Proof.
  introv Hi Hty.
  inversions Hty.
  - forwards: ty_inv_atomic_has_NM_replace; fweauto.
    destructecs H0.
    + left. apply ty_or_closure_atomic. auto.
    + right. left. apply ty_or_closure_atomic. auto.
    + right. right. apply ty_inv_basic_atomic. eauto.
  - destruct oc; simpls; inversions H.
    + left.
      eapply ty_or_closure_l; eauto.
    + eapply ty_has_N_or_closure_lc_cases in H1.
      destruct H1.
      * auto.
      * right. right. apply ty_inv_basic_NM; eauto. 
    + eapply ty_has_M_or_closure_lc_cases in H5.
      destruct H5.
      * auto.
      * right. right. apply ty_inv_basic_NM; eauto. 
  - destruct oc; simpls; inversions H.
    + left.
      eapply ty_or_closure_r; eauto.
    + eapply ty_has_M_or_closure_lc_cases in H1.
      destruct H1.
      * auto.
      * right. right. apply ty_inv_basic_MN; eauto. 
    + eapply ty_has_N_or_closure_lc_cases in H5.
      destruct H5.
      * auto.
      * right. right. apply ty_inv_basic_MN; eauto.
Qed.

Lemma ty_has_N_M_closure_or_top_subtyp_t G e T U:
  mode_has_top_nm_subtype typing_mode ->
  ty_has_N_closure G e T ->
  ty_has_M_closure G e U ->
  G ⁏ e ⊢# typ_top <: typ_or T U.
Proof.
  introv Hm HhN HhM.
  destruct (mode_top_nm' typing_mode) eqn: Hmn.
  - typing_mode_contradiction typing_mode.
  - eapply subtyp_trans_t. eapply subtyp_nocap_mut_top_boundless_t; eauto.
    eapply subtyp_or_tight.
    eapply ty_has_N_closure_tight; eauto.
    eapply ty_has_M_closure_tight; eauto.
  - eapply subtyp_trans_t. eapply subtyp_nocap_mut_top_complement_t; eauto.
    eapply subtyp_or_tight.
    eapply ty_has_N_closure_tight; eauto.
    eapply ty_has_M_closure_tight_bot; eauto.
Qed.

Lemma ty_has_M_N_closure_or_top_subtyp_t G e T U:
  mode_has_top_nm_subtype typing_mode ->
  ty_has_M_closure G e T ->
  ty_has_N_closure G e U ->
  G ⁏ e ⊢# typ_top <: typ_or T U.
Proof.
  introv Hm HhN HhM.
  eapply subtyp_trans_t.
  eapply ty_has_N_M_closure_or_top_subtyp_t; eauto.
  eapply subtyp_or_swap_t; eauto.
Qed.

Lemma ty_inv_basic_tight G e x T:
    inert G ->
    renv_corr G e ->
    x \in ctx_dom G ->
    ty_inv_basic G e x T ->
    G⁏e ⊢# x ∶ T.
Proof.
  introv Hi Hcor Hdom Hty.
  induction Hty.
  - eapply ty_inv_atomic_tight; eauto.
  - forwards: ty_tight_dom_top e Hdom; [> eauto |].
    eapply ty_sub_t. eauto.
    eapply ty_has_N_M_closure_or_top_subtyp_t; eauto.
  - forwards: ty_tight_dom_top e Hdom; [> eauto |].
    eapply ty_sub_t. eauto.
    eapply ty_has_M_N_closure_or_top_subtyp_t; eauto.
Qed.


End WithTypingMode.