Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping TightTypingHelpers Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening.
Require Import Orctx Logctx Logpath OrctxLogctx.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure.
 
Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Logic typing *)

(**
  The previous layer [ty_inv_basic_closure] allows
  deriving union types.
  This layer allows deriving types with unions and intersections,
  by taking two types that differ in one spot and merging them
  into one type where that spot has an intersection.

  It can be thought of as converting a formula from CNF by pushing
  the intersections down, one at a time. If there are multiple intersections
  in the target type, one below some other, then the one below must be
  created before the one above.

  In this version of typing on this layer, derivations are not uniquely
  determined by the types.
  - Intersections under two separate paths can be created in any order.
  - The types from the previous layer can already contain intersections,
    if they were added in an union type, if the variable has the other type in that union.
*)

(* Figure 18 in "Pure methods for roDOT" *)
Inductive ty_inv_logic: ctx -> renv -> var -> typ -> Prop :=
(* From previous layer *)
(** [G ⊢c x : T]   #<br>#
    [x in dom(G)]   #<br>#
    [―――――――――――(LTL-Or)]   #<br>#
    [G ⊢l x : T]   *)
| ty_inv_logic_or G e x T:
    x \in ctx_dom G ->
    ty_inv_basic_closure G e x T ->
    ty_inv_logic G e x T
(* Merge into an intersection *)
(** [G ⊢l x : C[T]]    #<br>#
    [G ⊢l x : C[U]]    #<br>#
    [――――――――――――――――――(LTL-And)]   #<br>#
    [G ⊢l x : C[T ∧ U]]          *)
| ty_inv_logic_and G e x oc T U:
    ty_inv_logic G e x (orctx_typ oc T) ->
    ty_inv_logic G e x (orctx_typ oc U) ->
    ty_inv_logic G e x (orctx_typ oc (typ_and T U))
.

(** *** Simple implications *)

Lemma ty_inv_logic_dom G e x T:
  ty_inv_logic G e x T ->
  x \in ctx_dom G.
Proof.
  introv Hty.
  induction Hty; eauto.
Qed.

(** *** Closure properties *)

Lemma ty_inv_logic_atomic_basic_closure G e x T:
  ty_inv_logic G e x T ->
  typ_atomic T ->
  ty_inv_basic_closure G e x T.
Proof.
  introv Hty Ha.
  induction Hty; eauto.
  eapply orctx_atomic_inv in Ha.
  destruct_all. inversions H0.           
Qed.

Lemma ty_inv_logic_tyq G e x U V:
  inert G ->
  ty_inv_logic G e x U ->
  typing_equiv e U V ->
  ty_inv_logic G e x V.
Proof.
  introv Hin Hty Htyq.
  reverts V.
  induction Hty; intros.
  - eapply ty_inv_logic_or; eauto. eapply ty_inv_basic_closure_tyq; eauto.
  - eapply orctx_tyq_case in Htyq; destruct_all; subst.
    inversions_tyq H1.
    eapply ty_inv_logic_and.
    eauto using orctx_tyq_typ.
    eauto using orctx_tyq_typ.
Qed.

(** *** Forward derivations *)
Lemma ty_inv_logic_orl G e x T U:
  ty_inv_logic G e x T ->
  ty_inv_logic G e x (typ_or T U).
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_inv_logic_or; eauto.
    eapply ty_or_closure_l. eauto.
  - rewrite orctx_fold_l in *.
    eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_orr G e x T U:
  ty_inv_logic G e x U ->
  ty_inv_logic G e x (typ_or T U).
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_inv_logic_or; eauto.
    eapply ty_or_closure_r. eauto.
  - rewrite orctx_fold_r in *.
    eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_orctx G e x oc U:
  ty_inv_logic G e x U ->
  ty_inv_logic G e x (orctx_typ oc U).
Proof.
  introv Hty.
  induction oc; simpls; eauto using ty_inv_logic_orl, ty_inv_logic_orr.
Qed.

Lemma ty_inv_logic_and_nil G e x  T U:
    ty_inv_logic G e x T ->
    ty_inv_logic G e x U ->
    ty_inv_logic G e x (typ_and T U).
Proof.
  introv HtyT HtyU.
  forwards: ty_inv_logic_and orctx_nil T U; eauto.
Qed.

(** *** Inversions of intersections *)

Lemma ty_inv_logic_oc_and_invl G e x oc T U:
  ty_inv_logic G e x (orctx_typ oc (typ_and T U)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_and T U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_inv_basic_closure_oc_and_inv in H0.
    eapply ty_inv_logic_or; eauto.
  -
  eapply orctx_intersect_and_and_cases in Heqt.
  destruct_all; subst. eauto.
  forwards ocTU  Hrest: (H1 T0).
  rename x0 into ocT, x1 into ocU.
  destructc Hrest as Hoc0 HocT HocU HfocTU.
      forwards HtyT: IHHty1; eauto.
      forwards HtyU: IHHty2; eauto.
      rewrite <- Hoc0.
      rewrite <- HocT in HtyT.
      rewrite <- HocU in HtyU.
      eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_oc_and_invr G e x oc T U:
  ty_inv_logic G e x (orctx_typ oc (typ_and T U)) ->
  ty_inv_logic G e x (orctx_typ oc U).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_and T U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_inv_basic_closure_oc_and_inv in H0.
    eapply ty_inv_logic_or; eauto.
  -
  eapply orctx_intersect_and_and_cases in Heqt.
  destruct_all; subst. eauto.
 forwards ocTU  Hrest: (H1 U0).
  rename x0 into ocT, x1 into ocU.
  destructc Hrest as Hoc0 HocT HocU HfocTU.
      forwards HtyT: IHHty1; eauto.
      forwards HtyU: IHHty2; eauto.
      rewrite <- Hoc0.
      rewrite <- HocT in HtyT.
      rewrite <- HocU in HtyU.
      eapply ty_inv_logic_and; eauto.
Qed.


Lemma ty_inv_logic_oclc_and_invl G e x oc oc2 T U:
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and T U))) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 T)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_and_invl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_inv_logic_oclc_and_invr G e x oc oc2 T U:
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and T U))) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 U)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_and_invr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
Qed.
Lemma ty_inv_logic_lc_and_invl G e x oc2 T U:
  ty_inv_logic G e x ( (logctx_typ oc2 (typ_and T U))) ->
  ty_inv_logic G e x ( (logctx_typ oc2 T)).
Proof.
  introv Hty.
  applys ty_inv_logic_oclc_and_invl orctx_nil; eauto.
Qed.
Lemma ty_inv_logic_lc_and_invr G e x oc2 T U:
  ty_inv_logic G e x ( (logctx_typ oc2 (typ_and T U))) ->
  ty_inv_logic G e x ( (logctx_typ oc2 U)).
Proof.
  introv Hty.
  applys ty_inv_logic_oclc_and_invr orctx_nil; eauto.
Qed.


(** *** Forward derivations *)

Lemma ty_inv_logic_oclc_and G e x oc oc2 T U:
    ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 T)) ->
    ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 U)) ->
    ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and T U))).
Proof.
  introv HtyT HtyU.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_and; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl HtyT.
    forwards: ty_inv_logic_oc_and_invl HtyU.
    eauto.
    forwards: ty_inv_logic_oc_and_invr HtyT.
    eauto.
  -  eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl HtyT.
    eauto.
    forwards: ty_inv_logic_oc_and_invr HtyT.
    forwards: ty_inv_logic_oc_and_invr HtyU.
    eauto.
Qed.

Lemma ty_inv_logic_lc_and G e x oc2 T U:
    ty_inv_logic G e x  (logctx_typ oc2 T) ->
    ty_inv_logic G e x  (logctx_typ oc2 U) ->
    ty_inv_logic G e x  (logctx_typ oc2 (typ_and T U)).
Proof.
  introv HtyT HtyU.
  forwards: ty_inv_logic_oclc_and orctx_nil T U; simpls; eauto.
Qed.


Lemma ty_inv_logic_oc_orr G e x oc T U:
  ty_inv_logic G e x (orctx_typ oc U) ->
  ty_inv_logic G e x (orctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  remember (orctx_typ oc (U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_inv_basic_closure_oc_orr in H0.
    eapply ty_inv_logic_or; eauto.
  -
    symmetry in Heqt.
    pose proof Heqt as Heqt0.
    eapply orctx_intersect_cases_r in Heqt. 2: intros; discriminate.
    destruct_all; subst.
    *
      rewrite orctx_fold_r.
      rewrite <- ! orctx_extend_typ.
      eapply ty_inv_logic_and.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty1.
        rewrite ! orctx_extend_typ; eauto.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty2.
        rewrite ! orctx_extend_typ; eauto.
    *
        forwards: orctx_separate_cross_r Heqt0 H (typ_or T0 U0).
        destruct_all.
        forwards HT: H1 T.
        forwards HU: H1 U.
        destruct_all.
        rewrite H0.
        eapply ty_inv_logic_and.
        rewrite <- H7. eauto.
        rewrite <- H5. eauto.
Qed.

Lemma ty_inv_logic_oclc_orr G e x oc2 oc T U:
  ty_inv_logic G e x (orctx_typ oc2 (logctx_typ oc U)) ->
  ty_inv_logic G e x (orctx_typ oc2 (logctx_typ oc (typ_or T U))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_inv_logic_oc_orr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_inv_logic_oc_and_invl Hty. 
    forwards: ty_inv_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_inv_logic_and; eauto.
  -
    forwards: ty_inv_logic_oc_and_invl Hty. 
    forwards: ty_inv_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_lc_orr G e x oc T U:
  ty_inv_logic G e x  (logctx_typ oc U) ->
  ty_inv_logic G e x  (logctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  applys ty_inv_logic_oclc_orr orctx_nil; eauto.
Qed.

Lemma ty_inv_logic_oc_orl G e x oc T U:
  ty_inv_logic G e x  (orctx_typ oc T) ->
  ty_inv_logic G e x  (orctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_inv_basic_closure_oc_orl in H0.
    eapply ty_inv_logic_or; eauto.
  -
    symmetry in Heqt.
    pose proof Heqt as Heqt0.
    eapply orctx_intersect_cases_r in Heqt. 2: intros; discriminate.
    destruct_all; subst.
    *
      rewrite orctx_fold_l.
      rewrite <- ! orctx_extend_typ.
      eapply ty_inv_logic_and.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty1.
        rewrite ! orctx_extend_typ; eauto.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty2.
        rewrite ! orctx_extend_typ; eauto.
    *
        forwards: orctx_separate_cross_r Heqt0 H (typ_or T0 U0).
        destruct_all.
        forwards HT: H1 T.
        forwards HU: H1 U.
        destruct_all.
        rewrite H0.
        eapply ty_inv_logic_and.
        rewrite <- H7. eauto.
        rewrite <- H5. eauto.
Qed.

Lemma ty_inv_logic_oclc_orl G e x oc2 oc T U:
  ty_inv_logic G e x (orctx_typ oc2 (logctx_typ oc T)) ->
  ty_inv_logic G e x (orctx_typ oc2 (logctx_typ oc (typ_or T U))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_inv_logic_oc_orl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_inv_logic_oc_and_invl Hty. 
    forwards: ty_inv_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_inv_logic_and; eauto.
  -
    forwards: ty_inv_logic_oc_and_invl Hty. 
    forwards: ty_inv_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_lc_orl G e x oc T U:
  ty_inv_logic G e x  (logctx_typ oc T) ->
  ty_inv_logic G e x  (logctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  applys ty_inv_logic_oclc_orl orctx_nil; eauto.
Qed.

Lemma ty_inv_logic_toplike G e x T:
  x \in ctx_dom G ->
  typ_logic_toplike G e T ->
  ty_inv_logic G e x T.
Proof.
  introv Hdom Htl.
  inversions Htl.
  - (* typ_top *)
    eapply ty_inv_logic_or; eauto.
    eapply ty_or_closure_top.
  - (* typ_NM *)
    eapply ty_inv_logic_or; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM_canon; eauto.
  -
    eapply ty_inv_logic_or; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_has_top_atomic; eauto.
Qed.

(** *** Lifting from previous layer *)

(* Lifts the possibility to replace V by S from basic_closure to logic *)
Lemma ty_inv_logic_lift_or G e x oc V S P:
    inert G ->
    (forall oc2,
      ty_inv_basic_closure G e x (logctx_typ oc2 V) ->
      ty_inv_basic_closure G e x (logctx_typ oc2 S) \/ P
    ) ->
    ty_inv_logic_liftable G e V S ->
    ty_inv_logic G e x (logctx_typ oc V) ->
    ty_inv_logic G e x (logctx_typ oc S) \/ P.
Proof.
  introv Hin Hbas Hlift Hty.
  remember (logctx_typ oc V). reverts oc.
  induction Hty; intros; subst.
  - forwards Hl: Hbas H0. destructecs Hl. left. eapply ty_inv_logic_or; eauto. eauto.
  - (* AND case *)
    rewrite orctx_logctx_typ in Heqt.
    lets Heqt0: Heqt.
    (* And happened under oc, bot happened under oc0. Try cases of intersecting oc and oc0. *)
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + (* oc = oc0 - impossible if V is not and *)
      inversions Hlift.
      * (* not and *)
        false.
        forwards: H orctx_nil; simpls; eauto.
      * (* toplike *)
        rewrite <- orctx_logctx_typ.
        left.
        eapply ty_inv_logic_orctx.
        eapply ty_inv_logic_toplike; eauto.
        eapply ty_inv_logic_dom; eauto.
    + (* oc is prefix of oc0, meaning that the V happens in one of the and branches *)
      (* destruct the next step in oc0, which tells us which branch to go to *)
      destruct x0; simpls; inversions H0.
      * (* Must be a proper prefix *)
        false.
      * (* left branch *)
        forwards: IHHty1 Hin  (logctx_extend (orctx_logctx oc) x0). eauto. eauto. 
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        destructecs H.
        left; eapply ty_inv_logic_and; eauto. eauto.
      * (* right branch *)
        forwards: IHHty2 Hin  (logctx_extend (orctx_logctx oc) x0). eauto. eauto. 
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        destructecs H.
        left; eapply ty_inv_logic_and; eauto. eauto.
    + (* oc0 is a prefix of oc - impossible if V is not and *)
      eapply orctx_logctx_extend in H; destruct_all; subst.
      inversions Hlift.
      * (* not and *)
        false.
        forwards: H x2; simpls; rewrite <- orctx_logctx_typ in H0; eauto.
      * (* toplike *)
        rewrite <- orctx_logctx_typ.
        left.
        eapply ty_inv_logic_orctx.
        eapply ty_inv_logic_toplike; eauto.
        eapply ty_inv_logic_dom; eauto.
    + (* The paths diverge *)
      (* Put oc0 on the LHS *)
      symmetry in Heqt0.
      eapply logctx_separate_symm in H.
      (* Find variant of oc, where V is replaced by S *)
      forwards: logctx_separate_cross_r Heqt0 H S.
      destruct_all.
      (* Find variant of oc0, where (typ_and T U) is replaced by T *)
      forwards: H1 T.
      (* Find variant of oc0, where (typ_and T U) is replaced by U *)
      forwards: H1 U.
      destruct_all.
      rewrite <- orctx_logctx_typ in *.
      (* Recurse on the types where (typ_and T U) is replaced by T and by U individually *)
      forwards: IHHty1 Hin x2; eauto.
      forwards: IHHty2 Hin x1; eauto.
      (* Rewrite the types into a form that uses the variant of oc where V is replaced by S, therefore exposing the intersection of T and U *)
      rewrite  H7 in H9.
      rewrite  H5 in H10.
      rewrite H0.
      (* Because oc does not have intersection, then its version with V replaced by S also does not *)
      forwards Hocx0: logctx_follows_isor logctx_isor_orctx H2.
      hnf in Hocx0. destruct_all; eauto 2; subst.
      rewrite <- orctx_logctx_typ in *; simpls.
      (* Now it is in the form to apply the rule *)
      left; eapply ty_inv_logic_and; eauto.
Qed.
Lemma ty_inv_logic_lift G e x oc V S:
    inert G ->
    (forall oc2,
      ty_inv_basic_closure G e x (logctx_typ oc2 V) ->
      ty_inv_basic_closure G e x (logctx_typ oc2 S)
    ) ->
    ty_inv_logic_liftable G e V S ->
    ty_inv_logic G e x (logctx_typ oc V) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin Hbas Hlift Hty.
  forwards: ty_inv_logic_lift_or False Hlift Hty; fweauto.
  destructecs H; easy.
Qed.
Lemma ty_inv_logic_lc_bot_replace G e x oc S:
    inert G ->
    ty_inv_logic G e x (logctx_typ oc typ_bot) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_bot_replace; eauto.
  eapply ty_inv_logic_liftable_bot.
  eauto.
Qed.
Lemma ty_inv_logic_lc_bnd_replace G e x oc S T:
    inert G ->
    ty_inv_logic G e x (logctx_typ oc (typ_bnd T)) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_bnd_replace; eauto.
  eapply ty_inv_logic_liftable_not_and; intros.
  destruct oc0; simpls; discriminate.
  eauto.
Qed.
Lemma ty_inv_logic_lc_nocap_replace G e x oc S:
    inert G ->
    ty_has_N G e S ->
    ty_inv_logic G e x (logctx_typ oc typ_nocap) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin HhN Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_nocap_replace; eauto.
  eapply ty_inv_logic_liftable_nocap.
  eauto.
Qed.

Lemma ty_inv_logic_lc_nocap_closure_replace G e x oc S:
    inert G ->
    ty_has_N_closure G e S ->
    ty_inv_logic G e x (logctx_typ oc typ_nocap) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin HhN Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_nocap_closure_replace; eauto.
  eapply ty_inv_logic_liftable_nocap.
  eauto.
Qed.
Lemma ty_inv_logic_lc_nocap_logic_replace G e x oc S:
    inert G ->
    ty_has_N_logic G e S ->
    ty_inv_logic G e x (logctx_typ oc typ_nocap) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin HhN Hty.
  induction HhN.
  - eapply ty_inv_logic_lc_nocap_closure_replace; eauto.
  - rewrite orctx_logctx_typ, <- logctx_extend_typ in *.  eapply ty_inv_logic_lc_and; eauto.
Qed.
Lemma ty_inv_logic_lc_sel_replace G e x oc S z A y:
    inert G ->
    ty_inv_logic G e x (logctx_typ oc (typ_sel z A y)) ->
    ty_inv_logic G e x (logctx_typ oc S).
Proof.
  introv Hin Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_sel_replace; eauto.
  2: eauto.
  eapply ty_inv_logic_liftable_not_and; intros.
  destruct oc0; simpls; discriminate.
Qed.

Lemma ty_inv_logic_lc_top_replace G e x oc S:
    inert G ->
    ty_inv_logic G e x (logctx_typ oc S) ->
    ty_inv_logic G e x (logctx_typ oc typ_top).
Proof.
  introv Hin Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_top_replace; eauto.
  eauto.
  eauto.
Qed.
Lemma ty_inv_logic_lc_NM_replace G e x oc S T1 T2:
    mode_has_top_nm_subtype typing_mode ->
    inert G ->
    (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
    ty_inv_logic G e x (logctx_typ oc S) ->
    ty_inv_logic G e x (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hm Hin Hbot Hty.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_NM_replace; eauto.
  eapply ty_inv_logic_liftable_toplike.
  eauto.
  eauto.
Qed.

Lemma ty_inv_logic_lc_has_top_replace G e x oc S T:
    inert G ->
 renv_corr G e ->
    ty_inv_logic G e x (logctx_typ oc S) ->
    ty_has_top_atomic G e T ->
    ty_inv_logic G e x (logctx_typ oc T).
Proof.
  introv Hin Hcor Hty Htop.

  eapply ty_inv_logic_lift. 4: eauto.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_has_top_replace; eauto. eapply ty_inv_logic_dom; eauto.
  eapply ty_inv_logic_liftable_toplike.
  inversions Htop.
  eapply typ_logic_toplike_has_top; eauto.
Qed.


Lemma ty_inv_logic_lc_subtyp G e x oc U V:
  inert G ->
  ty_inv_logic G e x (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_inv_logic G e x (logctx_typ oc V).
Proof.
  introv Hin Hty Hsu.
  eapply ty_inv_logic_lift.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_subtyp; eauto.
  eapply ty_inv_logic_liftable_atomic.
  eapply subtyp_atomic_atomic_l; eauto.
  eauto.
Qed.

Lemma ty_inv_logic_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  inert G ->
  ty_pf_or_none G x A TL TU ->
  ty_inv_logic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_logic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU)))  \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hin Hpf Hty.
  eapply ty_inv_logic_lift_or.
  eauto.
  intros.
  eapply ty_inv_basic_closure_lc_typ_pfor_replace ; eauto. eapply ty_inv_logic_dom; eauto.
  eapply ty_inv_logic_liftable_atomic. eauto.
  eauto.
Qed.


Lemma ty_inv_logic_oc_dist G e x oc T1 T2 T3:
  inert G ->
  ty_inv_logic G e x (orctx_typ oc (typ_and T1 (typ_or T2 T3))) ->
  ty_inv_logic G e x (orctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv Hi Hty.
  remember (orctx_typ oc
           (typ_and T1 (typ_or T2 T3))).
  reverts oc T1 T2 T3.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_or. eauto.
    rewrite orctx_logctx_typ in *.
    eapply ty_inv_basic_closure_lc_and_replace; eauto.
  -
    lets: Heqt.
    forwards: orctx_intersect_cases_nonnil Heqt.
    destruct_all; subst.
    + inversions H1.

      assert (ty_inv_logic G e x (orctx_typ oc0 (typ_or (typ_and T1 T2) T3))).
      *
      rewrite orctx_extend_l.
      eapply ty_inv_logic_and; rewrite orctx_extend_typ; simpls.
      eapply ty_inv_logic_oc_orl; auto.
      auto.
      *
      rewrite orctx_extend_r.
      eapply ty_inv_logic_and; rewrite orctx_extend_typ; simpls.
      eapply ty_inv_logic_oc_orr; auto.
      auto.

    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    +
      symmetry in H.
      eapply orctx_separate_symm in H0.
      forwards: orctx_separate_cross_r H H0 (typ_or (typ_and T1 T2) (typ_and T1 T3)).
      destruct_all.
      forwards: H2 T; destruct_all.
      forwards: H2 U; destruct_all.

      forwards: IHHty1; eauto.
      forwards: IHHty2; eauto.

      rewrite H8 in H11.
      rewrite H5 in H10.
      rewrite H1.

      eapply ty_inv_logic_and; eauto.
Qed.

Lemma ty_inv_logic_oclc_dist  G e x oc oc2 T1 T2 T3:
  inert G ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and T1 (typ_or T2 T3)))) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_or (typ_and T1 T2) (typ_and T1 T3)))).
Proof.
  introv Hi Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_dist; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty.
    eauto.
    forwards: ty_inv_logic_oc_and_invr Hty.
    eauto.
  -  eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty.
    eauto.
    forwards: ty_inv_logic_oc_and_invr Hty.
    eauto.
Qed.

Lemma ty_inv_logic_lc_dist  G e x oc T1 T2 T3:
  inert G ->
  ty_inv_logic G e x (logctx_typ oc (typ_and T1 (typ_or T2 T3))) ->
  ty_inv_logic G e x (logctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv HtyT HtyU.
  forwards: ty_inv_logic_oclc_dist orctx_nil T1 T2 T3; simpls; eauto.
Qed.

Lemma ty_inv_logic_basic_closure_N_also_M_replace G e w oc S:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_basic_closure G e w (orctx_typ oc typ_nocap) ->
  ty_inv_logic G e w (orctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_inv_logic G e w (orctx_typ oc S).
Proof.
  introv Hm Hin Hcor HtyN HtyM.
  remember ((orctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))).
  reverts oc.
  induction HtyM; intros.
  - subst.
    eapply ty_inv_logic_or; eauto.
    rewrite orctx_logctx_typ.
    eapply ty_inv_basic_closure_N_also_M_replace with (oc0:=orctx_logctx oc); try rewrite <- orctx_logctx_typ; eauto.
  -
    forwards Hsep: orctx_intersect_cases Heqt.
    intros; discriminate.
    intros; discriminate.
    destructecs Hsep. discriminate.
    forwards HcrossT: orctx_separate_cross_r Heqt Hsep T.
    destructecs HcrossT.
    forwards HcrossU: orctx_separate_cross_r Heqt Hsep U.
    destructecs HcrossU.


    forwards HcrossTM: HcrossT1 typ_nocap.
    forwards HcrossTS: HcrossT1 S.

    forwards HcrossUM: HcrossU1 typ_nocap.
    forwards HcrossUS: HcrossU1 S.

    destructecs_all.

    forwards Heq: orctx_follow_eq_inv ocST2 ocST1.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUS0. eauto. destructecs Heq. subst. clear Heq0 HcrossUS0 HcrossUS.

    forwards Heq: orctx_follow_eq_inv ocST3 ocST0.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUM0. eauto. destructecs Heq. subst. clear Heq0 HcrossUM0 HcrossUM.

    clear HcrossT1 HcrossU1.

    rewrite <- HcrossTM0 in HtyN.
    rewrite orctx_logctx_typ in HtyN.
    forwards HtyNT: ty_inv_basic_closure_lc_and_replace T HtyN; fweauto.
    forwards HtyNU: ty_inv_basic_closure_lc_and_replace U HtyN; fweauto.
    rewrite <- orctx_logctx_typ in HtyNT, HtyNU.

    forwards HHtyM1: IHHtyM1 HcrossT0; fweauto.
    rewrite <- HcrossTM1. eauto.

    forwards HHtyM2: IHHtyM2 HcrossU0; fweauto.
    rewrite <- HcrossUM1. eauto.

    rewrite <- HcrossTS1 in HHtyM1.
    rewrite <- HcrossUS1 in HHtyM2.

    rewrite <- HcrossTS0.
    eapply ty_inv_logic_and; eauto.
Qed.

Lemma ty_inv_logic_N_also_M_replace G e w oc S:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e w (orctx_typ oc typ_nocap) ->
  ty_inv_logic G e w (orctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_inv_logic G e w (orctx_typ oc S).
Proof.
  introv Hm Hin Hcor HtyN HtyM.
  remember ((orctx_typ oc typ_nocap)).
  reverts oc.
  induction HtyN; intros.
  - subst.
    eapply ty_inv_logic_basic_closure_N_also_M_replace; eauto.
  -

    forwards Hsep: orctx_intersect_cases Heqt.
    intros; discriminate.
    intros; discriminate.
    destructecs Hsep. discriminate.
    forwards HcrossT: orctx_separate_cross_r Heqt Hsep T.
    destructecs HcrossT.
    forwards HcrossU: orctx_separate_cross_r Heqt Hsep U.
    destructecs HcrossU.


    forwards HcrossTM: HcrossT1 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)).
    forwards HcrossTS: HcrossT1 S.

    forwards HcrossUM: HcrossU1 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)).
    forwards HcrossUS: HcrossU1 S.

    destructecs_all.

    forwards Heq: orctx_follow_eq_inv ocST2 ocST1.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUS0. eauto. destructecs Heq. subst. clear Heq0 HcrossUS0 HcrossUS.

    forwards Heq: orctx_follow_eq_inv ocST3 ocST0.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUM0. eauto. destructecs Heq. subst. clear Heq0 HcrossUM0 HcrossUM.

    clear HcrossT1 HcrossU1.

    rewrite <- HcrossTM0 in HtyM.
    forwards HtyMT: ty_inv_logic_oc_and_invl HtyM.
    forwards HtyMU: ty_inv_logic_oc_and_invr HtyM.

    forwards HHtyN1: IHHtyN1 HcrossT0; fweauto.
    rewrite <- HcrossTM1. eauto.

    forwards HHtyN2: IHHtyN2 HcrossU0; fweauto.
    rewrite <- HcrossUM1. eauto.

    rewrite <- HcrossTS1 in HHtyN1.
    rewrite <- HcrossUS1 in HHtyN2.

    rewrite <- HcrossTS0.
    eapply ty_inv_logic_and; eauto.

Qed.

Lemma ty_inv_logic_oc_MN_bot_replace G e x oc:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (orctx_typ oc (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) )) ->
  ty_inv_logic G e x (orctx_typ oc typ_bot).
Proof.
  introv Hm Hi Hcor Hty.
  remember (orctx_typ oc
           (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) )).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_or. eauto.
    rewrite orctx_logctx_typ in *.
    eapply ty_inv_basic_closure_lc_and_replace; eauto.
  -
    lets: Heqt.
    forwards: orctx_intersect_cases_nonnil Heqt.
    destruct_all; subst.
    + inversions H1.
      eapply ty_inv_logic_N_also_M_replace; eauto.

    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    +
      symmetry in H.
      eapply orctx_separate_symm in H0.
      forwards: orctx_separate_cross_r H H0 typ_bot.
      destruct_all.
      forwards: H2 T; destruct_all.
      forwards: H2 U; destruct_all.

      forwards: IHHty1; eauto.
      forwards: IHHty2; eauto.

      rewrite H8 in H11.
      rewrite H5 in H10.
      rewrite H1.

      eapply ty_inv_logic_and; eauto.
Qed.


Lemma ty_inv_logic_oclc_MN_bot_replace G e x oc oc2:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) ))) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 typ_bot)).
Proof.
  introv Hm Hi Hcor Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_MN_bot_replace; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_inv_logic_lc_MN_bot_replace G e x oc:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (logctx_typ oc (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) )) ->
  ty_inv_logic G e x (logctx_typ oc typ_bot).
Proof.
  introv Hm Hi Hcor Hty.
  eapply ty_inv_logic_oclc_MN_bot_replace with (oc := orctx_nil); eauto.
Qed.

Lemma ty_inv_logic_basic_closure_oc_typ_dist G e x oc A TL TU SL SU:
  inert G ->
  renv_corr G e ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hin Hcor HtyN HtyM.
  remember (orctx_typ oc (typ_rcd (dec_typ A SL SU))).
  reverts oc.
  induction HtyM; intros.
  - subst.
    eapply ty_inv_logic_or; eauto.
    rewrite orctx_logctx_typ.
    eapply ty_inv_basic_closure_lc_typ_also_dist with (oc0:=orctx_logctx oc); try rewrite <- orctx_logctx_typ; eauto.
  -
    forwards Hsep: orctx_intersect_cases Heqt.
    intros; discriminate.
    intros; discriminate.
    destructecs Hsep. discriminate.
    forwards HcrossT: orctx_separate_cross_r Heqt Hsep T.
    destructecs HcrossT.
    forwards HcrossU: orctx_separate_cross_r Heqt Hsep U.
    destructecs HcrossU.


    forwards HcrossTM: HcrossT1 ( (typ_rcd (dec_typ A TL TU))) .
    forwards HcrossTS: HcrossT1 (((typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).

    forwards HcrossUM: HcrossU1 (( (typ_rcd (dec_typ A TL TU)))).
    forwards HcrossUS: HcrossU1 (( (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).

    destructecs_all.

    forwards Heq: orctx_follow_eq_inv ocST2 ocST1.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUS0. eauto. destructecs Heq. subst. clear Heq0 HcrossUS0 HcrossUS.

    forwards Heq: orctx_follow_eq_inv ocST3 ocST0.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUM0. eauto. destructecs Heq. subst. clear Heq0 HcrossUM0 HcrossUM.

    clear HcrossT1 HcrossU1.

    rewrite <- HcrossTM0 in HtyN.
    rewrite orctx_logctx_typ in HtyN.
    forwards HtyNT: ty_inv_basic_closure_lc_and_replace T HtyN; fweauto.
    forwards HtyNU: ty_inv_basic_closure_lc_and_replace U HtyN; fweauto.
    rewrite <- orctx_logctx_typ in HtyNT, HtyNU.

    forwards HHtyM1: IHHtyM1 HcrossT0; fweauto.
    rewrite <- HcrossTM1. eauto.

    forwards HHtyM2: IHHtyM2 HcrossU0; fweauto.
    rewrite <- HcrossUM1. eauto.

    rewrite <- HcrossTS1 in HHtyM1.
    rewrite <- HcrossUS1 in HHtyM2.

    rewrite <- HcrossTS0.
    eapply ty_inv_logic_and; eauto.
Qed.

Lemma ty_inv_logic_oc_typ_also_dist G e x oc A TL TU SL SU:
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hin Hcor HtyN HtyM.
  remember (orctx_typ oc (typ_rcd (dec_typ A TL TU))) .
  reverts oc.
  induction HtyN; intros.
  - subst.
    eapply ty_inv_logic_basic_closure_oc_typ_dist; eauto.
  -

    forwards Hsep: orctx_intersect_cases Heqt.
    intros; discriminate.
    intros; discriminate.
    destructecs Hsep. discriminate.
    forwards HcrossT: orctx_separate_cross_r Heqt Hsep T.
    destructecs HcrossT.
    forwards HcrossU: orctx_separate_cross_r Heqt Hsep U.
    destructecs HcrossU.


    forwards HcrossTM: HcrossT1 ( (typ_rcd (dec_typ A SL SU))) .
    forwards HcrossTS: HcrossT1 (((typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).

    forwards HcrossUM: HcrossU1 (( (typ_rcd (dec_typ A SL SU)))).
    forwards HcrossUS: HcrossU1 (( (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).

    destructecs_all.

    forwards Heq: orctx_follow_eq_inv ocST2 ocST1.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUS0. eauto. destructecs Heq. subst. clear Heq0 HcrossUS0 HcrossUS.

    forwards Heq: orctx_follow_eq_inv ocST3 ocST0.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite HcrossUM0. eauto. destructecs Heq. subst. clear Heq0 HcrossUM0 HcrossUM.

    clear HcrossT1 HcrossU1.

    rewrite <- HcrossTM0 in HtyM.
    forwards HtyMT: ty_inv_logic_oc_and_invl HtyM.
    forwards HtyMU: ty_inv_logic_oc_and_invr HtyM.

    forwards HHtyN1: IHHtyN1 HcrossT0; fweauto.
    rewrite <- HcrossTM1. eauto.

    forwards HHtyN2: IHHtyN2 HcrossU0; fweauto.
    rewrite <- HcrossUM1. eauto.

    rewrite <- HcrossTS1 in HHtyN1.
    rewrite <- HcrossUS1 in HHtyN2.

    rewrite <- HcrossTS0.
    eapply ty_inv_logic_and; eauto.

Qed.

Lemma ty_inv_logic_oc_typ_dist G e x  oc A TL TU SL SU:
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (orctx_typ oc (typ_and (typ_rcd (dec_typ A TL TU)) (typ_rcd (dec_typ A SL SU)) )) ->
  ty_inv_logic G e x (orctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hi Hcor Hty.
  remember (orctx_typ oc (typ_and (typ_rcd (dec_typ A TL TU)) (typ_rcd (dec_typ A SL SU)) )).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_or. eauto.
    rewrite orctx_logctx_typ in *.
    eapply ty_inv_basic_closure_lc_and_replace; eauto.
  -
    lets: Heqt.
    forwards: orctx_intersect_cases_nonnil Heqt.
    destruct_all; subst.
    + inversions H1.
      eapply ty_inv_logic_oc_typ_also_dist; eauto.

    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    + destruct x0; simpls; try contradiction.
      inversions H1.
      inversions H1.
    +
      symmetry in H.
      eapply orctx_separate_symm in H0.
      forwards: orctx_separate_cross_r H H0 ( (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
      destruct_all.
      forwards: H2 T; destruct_all.
      forwards: H2 U; destruct_all.

      forwards: IHHty1; eauto.
      forwards: IHHty2; eauto.

      rewrite H8 in H11.
      rewrite H5 in H10.
      rewrite H1.

      eapply ty_inv_logic_and; eauto.
Qed.


Lemma ty_inv_logic_oclc_typ_dist G e x oc oc2 A TL TU SL SU:
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_and (typ_rcd (dec_typ A TL TU)) (typ_rcd (dec_typ A SL SU))) )) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).
Proof.
  introv Hi Hcor Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_typ_dist; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_inv_logic_lc_typ_dist G e x oc A TL TU SL SU:
  inert G ->
  renv_corr G e ->
  ty_inv_logic G e x (logctx_typ oc (typ_and (typ_rcd (dec_typ A TL TU)) (typ_rcd (dec_typ A SL SU)) )) ->
  ty_inv_logic G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hi Hcor Hty.
  eapply ty_inv_logic_oclc_typ_dist with (oc := orctx_nil); eauto.
Qed.

(** *** Tight typing *)

Lemma ty_inv_logic_tight G e x T:
    inert G ->
    renv_corr G e ->
    ty_inv_logic G e x T ->
    G⁏e ⊢# x ∶ T.
Proof.
  introv Hi Hcor Hty.
  induction Hty.
  - eapply ty_inv_basic_closure_tight; eauto.
  - forwards HtyT: IHHty1; eauto 1.
    forwards HtyU: IHHty2; eauto 1.
    forwards HtyTU: ty_and_intro_t HtyT HtyU.
    eapply ty_sub_t. eauto.
    eapply ty_tight_oc_dist.
Qed.


End WithTypingMode.