Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening.
Require Import Orctx Logctx OrctxLogctx Logpath LogpathOrder TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityLogicOrdered LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasicClosure LayeredTypingLogic.

Section WithTypingMode.
Context {typing_mode: Mode}.

Section WithTyping.
Context (G: ctx) (e: renv) (x: var).
Context (HinG: inert G) (HGx: x \in ctx_dom G).
Context (Hcor: renv_corr G e).

(** * Ordered logic typing *)
(** A version of logic-layer typing, where the order of intersection
  is unique (bottom-to-top, right-to-left).
  In ty_inv_logic, the order is bottom-to-top,
  but arbitrary for separate paths, and allows intersections to come from
  previous layer.
  In this version, intersections are not allowed from previous layer,
  by disallowing any intersection appearing under a path of unions.
  The order is fixed right-to-left. After an intersection is added,
  the next one must be to the left or above it.
  The first parameter of ty_inv_logic_ord indicates a path where
  the next intersection can be added.
  This argument is not fixed, and can be moved left or up.
  For a type with no intersections, this argument can be arbitrary.
  The argument orctx_nil works for all types (intersections can be always added at the root).
*)

Inductive ty_inv_logic_ord: orctx -> typ -> Prop :=
(* Type from the previous layer (without intersections) *)
| ty_inv_logic_ord_or oc T:
    ty_inv_basic_closure G e x T ->
    typ_notorand T ->
    ty_inv_logic_ord oc T
(* Add a new intersection, to the left or above previous intersections *)
| ty_inv_logic_ord_and oc ocp T U:
    orctx_le oc ocp ->
    ty_inv_logic_ord ocp (orctx_typ ocp T) ->
    ty_inv_logic_ord ocp (orctx_typ ocp U) ->
    ty_inv_logic_ord oc (orctx_typ ocp (typ_and T U))
.

(* Analog of ty_inv_main_ord_main *)
Lemma ty_inv_logic_ord_logic op T:
  ty_inv_logic_ord op T ->
  ty_inv_logic G e x T.
Proof.
  introv Hh.
  induction Hh.
  - eapply ty_inv_logic_or; eauto.
  - eapply ty_inv_logic_and; eauto.
Qed.

(* Analog of ty_inv_main_ord_le *)
Lemma ty_inv_logic_ord_le op1 op2 T:
  orctx_le op2 op1 ->
  ty_inv_logic_ord op1 T ->
  ty_inv_logic_ord op2 T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_inv_logic_ord_or; eauto.
  - eapply ty_inv_logic_ord_and; eauto.
    unfold orctx_le in *. eapply logpath_le_trans; eauto.
Qed.


Lemma ty_inv_logic_ord_has_closure T:
    ty_inv_basic_closure G e x T ->
    ty_inv_logic_ord (typ_orand_leftmost_orctx T) T.
Proof.
  introv Hh.
  induction T using typ_orand_ind; intros.
  - eapply ty_inv_logic_ord_or; eauto.
  - eapply ty_inv_logic_ord_and.
    +
    eapply typ_orand_leftmost_orctx_le.
    +
    eapply ty_inv_logic_ord_le.
    eapply typ_orand_leftmost_orctx_notorand_le; eauto.
    eapply IHT1.
    rewrite orctx_logctx_typ in *.
    eapply ty_inv_basic_closure_lc_and_replace; eauto.
    +
    eapply ty_inv_logic_ord_le.
    eapply typ_orand_leftmost_orctx_notorand_le; eauto.
    eapply IHT2.
    rewrite orctx_logctx_typ in *.
    eapply ty_inv_basic_closure_lc_and_replace; eauto.
Qed.


(* Proof of ty_inv_logic -> ty_inv_logic_ord logpath_nil *)


Lemma ty_inv_logic_ord_oc_and_symm oc ocp0 U0 T0:
  ty_inv_logic_ord oc (orctx_typ ocp0 (typ_and U0 T0)) ->
  ty_inv_logic_ord oc (orctx_typ ocp0 (typ_and T0 U0)).
Proof.
  introv Hty.
  remember (orctx_typ ocp0 (typ_and U0 T0)).
  reverts ocp0 U0 T0.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_ord_or.
    rewrite orctx_logctx_typ.
    eapply ty_inv_basic_closure_lc_and_replace.
    eauto.
    rewrite <- orctx_logctx_typ.
    eauto.
    eapply typ_notorand_oc_notand in H0. false H0; eauto.
  - 
    eapply orctx_intersect_and_and_cases in Heqt.
    destructecs Heqt; subst.
    + eapply ty_inv_logic_ord_and; eauto.
    + specialize (Heqt (typ_and T0 U0)).
      destructecs Heqt.
      rewrite <- Heqt2.
      eapply ty_inv_logic_ord_and; eauto.
      eapply orctx_le_trans. eauto. eapply orctx_follow_le. eapply orctx_follow_symm; eauto.
      *
        rewrite Heqt3. eapply ty_inv_logic_ord_le. eapply orctx_follow_le. eauto.
        eapply IHHty1. eauto.
      *
        rewrite Heqt4. eapply ty_inv_logic_ord_le. eapply orctx_follow_le. eauto.
        eapply IHHty2. eauto.
Qed.



(* Analog of ty_inv_main_ord_notbndsel_le *)
Lemma ty_inv_logic_ord_notorand_le oc ocp U:
  orctx_notorand_le ocp ->
  ty_inv_logic_ord oc (orctx_typ ocp U) ->
  ty_inv_logic_ord ocp (orctx_typ ocp U).
Proof.
  introv Hn Hh.
  remember (orctx_typ ocp U); reverts ocp U.
  induction Hh; intros.
  - eapply ty_inv_logic_ord_or; eauto.
  - forwards: orctx_notorand_eq_and_le; eauto.
    eapply ty_inv_logic_ord_and; eauto.
Qed.


(* Analog of ty_inv_main_ord_orctx_notbndsel_le *)
Lemma ty_inv_logic_ord_orctx_notorand_le oc T:
  ty_inv_logic_ord oc (orctx_typ oc T) ->
  orctx_notorand_le oc.
Proof.
  introv Hh.
  remember (orctx_typ oc T).
  reverts T.
  induction Hh; intros; subst.
  - eapply orctx_notorand_le_notorand; eauto.
  -
    forwards: IHHh1; eauto.
    eapply orctx_notorand_le_le; eauto.
Qed.

Lemma ty_inv_logic_ord_oc_notorand_le oc T:
  ty_inv_logic_ord oc T ->
  typ_notorand T \/ ty_inv_logic_ord (typ_orand_leftmost_orctx T) T /\ orctx_le oc (typ_orand_leftmost_orctx T).
Proof.
  introv Hh.
  induction Hh.
  - eauto.
  - right.
    forwards: typ_orand_leftmost_orctx_le ocp T U.
    split.
    +
      eapply ty_inv_logic_ord_and; eauto.
    +
      forwards: ty_inv_logic_ord_orctx_notorand_le Hh1.
      eapply orctx_le_trans.
      eauto.
      eapply typ_orand_leftmost_orctx_notorand_le; eauto.
Qed.

(* Trying to add intersection at a place right of oc,
   but ty_inv_logic_ord allows them only in right-to-left-(to-up) order.
   (There cannot be an intersection up, because we are in an or ctx)
   We need to undo the intersections left of ocp (between oc and ocp), and then redo them.
   
   Thanks to the deterministic rules, both derivations of ocp-T and ocp-U must use the same steps until ocp.
*)


Lemma ty_inv_logic_ord_oc_notorand_and_inv_l oc T U:
  orctx_notorand_le oc ->
  ty_inv_logic_ord oc (orctx_typ oc (typ_and T U)) ->
  ty_inv_logic_ord oc (orctx_typ oc T).
Proof.
  introv Hl Hh.
  inversions Hh.
  - eapply typ_notorand_oc_notand in H0. false. eapply H0. eauto.
  - forwards Hocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
    symmetry in H.
    forwards Hlp: orctx_notorand_eq_and_le H; eauto.
    forwards Heq: orctx_le_le_eq_inv H; eauto.
    destructecs Heq; subst.
    inversions Heq.
    eauto.
Qed.
Lemma ty_inv_logic_ord_oc_notorand_and_inv_r oc T U:
  orctx_notorand_le oc ->
  ty_inv_logic_ord oc (orctx_typ oc (typ_and T U)) ->
  ty_inv_logic_ord oc (orctx_typ oc U).
Proof.
  introv Hl Hh.
  inversions Hh.
  - eapply typ_notorand_oc_notand in H0. false. eapply H0. eauto.
  - forwards Hocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
    symmetry in H.
    forwards Hlp: orctx_notorand_eq_and_le H; eauto.
    forwards Heq: orctx_le_le_eq_inv H; eauto.
    destructecs Heq; subst.
    inversions Heq.
    eauto.
Qed.


Lemma ty_inv_logic_ord_and_lc_le oc ocp T U:
    orctx_le oc ocp ->
    ty_inv_logic_ord oc (orctx_typ ocp T) ->
    ty_inv_logic_ord oc (orctx_typ ocp U) ->
    ty_inv_logic_ord oc (orctx_typ ocp (typ_and T U)).
Proof.
  introv Hl HhT HhU.
  rename ocp into ocTU.
  rename oc into ocL.
  remember (orctx_typ ocTU T).
  reverts ocTU T U.

  induction HhT; intros; subst.
  - forwards: ty_inv_logic_ord_notorand_le HhU; eauto using orctx_notorand_le_notorand.
    eapply ty_inv_logic_ord_and; eauto. eapply ty_inv_logic_ord_or; eauto.
  - 
    forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le HhT1.
    destruct (orctx_le_cases ocTU ocp).
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_logic_ord_and; eauto.
      rewrite <- Heqt.
      eapply ty_inv_logic_ord_and; eauto.
      symmetry in Heqt.
      eapply ty_inv_logic_ord_notorand_le; eauto.
      eapply orctx_notorand_le_le; eauto.
    + (* ocp left of ocp0  -> need induction *)

    symmetry in Heqt.
    forwards Hocc: orctx_intersect_cases_r Heqt. intros; discriminate.
    destructecs Hocc; subst.
    * (* ocp = ocp0 -> solve without inductions same as the case above *)
      forwards: orctx_le_extend_l_inv; eauto; subst; simpls; rewrite orctx_extend_nil_l in *.
      eapply ty_inv_logic_ord_and; eauto.
      eapply ty_inv_logic_ord_notorand_le; eauto.
    * (* actually need induction *)
      (* Recombine types *)
      (* 1. variants according to the target intersection: T0, U0, T0&U0 *)
      forwards HcrT0: orctx_separate_cross_r Heqt Hocc T0.
      forwards HcrU0: orctx_separate_cross_r Heqt Hocc U0.
      forwards HcrTU0: orctx_separate_cross_r Heqt Hocc (typ_and T0 U0).
      destructecs_all.
      (* 2. variants according to the current intersection: T, U, T&U *)
      forwards HcrT2T: HcrT2 T. (* T0 / T *)
      forwards HcrT2U: HcrT2 U. (* T0 / U *)

      forwards HcrU2T: HcrU2 T. (* U0 / T *)
      forwards HcrU2U: HcrU2 U. (* U0 / U *)
      forwards HcrU2TU: HcrU2 (typ_and T U). (* U0 / T&U *)

      forwards HcrTU2T: HcrTU2 T. (* T0&U0 / T *)
      forwards HcrTU2U: HcrTU2 U. (* T0&U0 / U *)
      forwards HcrTU2TU: HcrTU2 (typ_and T U). (* T0&U0 / T&U *)
      clear HcrT2 HcrU2 HcrTU2.
      destructecs_all.
      (* Find equalities between variables, to have one variable for each variant *)    
      orctx_eqs; cleanup.

      rewrite HcrU1 in HhU.

      assert (orctx_notorand_le ocS1T0). {
        eapply orctx_notorand_replace_r; eauto.
      }

      eapply ty_inv_logic_ord_notorand_le in HhU. 2: eauto.
      assert (ty_inv_logic_ord ocp (orctx_typ ocS1T0 T )) as Htyx0T. {
        forwards: ty_inv_logic_ord_oc_notorand_and_inv_l; eauto.
        eapply ty_inv_logic_ord_le. 2: eauto.
        eapply orctx_follow_le; eauto.
      }
      assert (ty_inv_logic_ord ocp (orctx_typ ocS1T0 U )) as Htyx0U. {
        forwards: ty_inv_logic_ord_oc_notorand_and_inv_r; eauto.
        eapply ty_inv_logic_ord_le. 2: eauto.
        eapply orctx_follow_le; eauto.
      }

      forwards HHhT1: IHHhT1. 2: symmetry; eauto.
      applys orctx_le_trans ocST1; eauto using orctx_follow_le, orctx_follow_symm. rewrite HcrU2T1. eauto.
      forwards HHhT2: IHHhT2. 2: symmetry; eauto.
      applys orctx_le_trans ocST1; eauto using orctx_follow_le, orctx_follow_symm. rewrite HcrU2U1. eauto.

      rewrite HcrTU1. rewrite HcrTU2T1 in HHhT1. rewrite HcrTU2U1 in HHhT2.
      eapply ty_inv_logic_ord_and.
      applys orctx_le_trans ocp; eauto using orctx_follow_le, orctx_follow_symm. 
      eapply ty_inv_logic_ord_le; eauto. eauto using orctx_follow_le, orctx_follow_symm. 
      eapply ty_inv_logic_ord_le; eauto. eauto using orctx_follow_le, orctx_follow_symm.
Qed.


Lemma ty_inv_logic_ord_and_lc oc ocp T U:
    ty_inv_logic_ord oc (orctx_typ ocp T) ->
    ty_inv_logic_ord oc (orctx_typ ocp U) ->
    ty_inv_logic_ord (orctx_leb ocp oc) (orctx_typ ocp (typ_and T U)).
Proof.
  introv HhT HhU.

  destruct (orctx_le_cases ocp oc).
  - (* ocp <= oc    easy case *)
  rewrite orctx_leb_l in * by auto.
  eapply ty_inv_logic_ord_and.
  eapply orctx_le_refl.
  eapply ty_inv_logic_ord_le; eauto.
  eapply ty_inv_logic_ord_le; eauto.
  - (* oc <= ocp *)
  eapply ty_inv_logic_ord_le.
  eapply orctx_leb_le_comm.
  rewrite orctx_leb_l in * by auto.
  eapply ty_inv_logic_ord_and_lc_le; eauto.
Qed.




Lemma ty_inv_basic_closure_oc_or_dist_eq oc T U:
  ty_dist_eq T U ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_or T U)) ->
  typ_notorand (orctx_typ oc (typ_or T U)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv He Hh Hn.
  forwards: typ_notorand_oc_notorand; eauto.
  forwards: typ_notorand_or_invl; eauto.
  forwards: typ_notorand_or_invr; eauto.
  forwards: ty_dist_eq_notorand_eq; eauto; subst.
  rewrite orctx_logctx_typ in *.
    eapply ty_inv_logic_or; eauto. 
  eapply ty_inv_basic_closure_lc_or_diag; eauto.
Qed.

(* All intersections are right of oc *)
Lemma ty_inv_logic_ord_oc_nle_or_notorand_notorand_dist_eq ocp oc T U:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  typ_notorand U ->
  ty_inv_logic_ord ocp (orctx_typ oc (typ_or T U)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hn He HnT HnU Hty.
  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_oc_or_dist_eq; eauto.
  - 
    forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
    forwards: ty_dist_eq_notorand_eq; eauto; subst.
    symmetry in HeqTT.
    forwards: typ_notorand_eq_separate HeqTT.
    eapply typ_notorand_or; eauto.

      forwards: orctx_separate_cross_r HeqTT H0 U0.
      destruct_all.

      forwards: H2 T.
      forwards: H2 U.
      destruct_all.

      assert (orctx_notorand_le x1). {
        eapply orctx_notorand_le_le. 2: eapply H5.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }
      assert (orctx_notorand_le x2).  {
        eapply orctx_notorand_le_le. 2: eapply H4.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }

      forwards: IHHty1 x2 U0 U0; eauto.
      forwards: IHHty2 x1 U0 U0; eauto.

      rewrite H1.
      rewrite H8 in H12.
      rewrite H6 in H13.
      eapply ty_inv_logic_and; eauto.
Qed.





(* All intersection are in U or right of oc *)
Lemma ty_inv_logic_ord_oc_nle_or_notorand_dist_eq oc T U ocp:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  ty_inv_logic_ord ocp (orctx_typ oc (typ_or T U)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hn He HnT Hty.
  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts T U oc.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_oc_or_dist_eq; eauto.
  - 
    forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
    symmetry in HeqTT.
    forwards: orctx_intersect_cases_r HeqTT.
    intros; discriminate.
    destruct_all; subst.
    + (* under U -> continue induction *)
      clear HeqTT H.
      forwards: typ_notorand_eq_or_and; eauto.
      destruct_all; subst.
      clear H1.
      forwards: orctx_notorand_le_extend_invr Hnlocp.
      inversions H.

      forwards: ty_dist_eq_oc_and_cases T0; eauto.
      destruct H.
      *
      forwards: IHHty1 T0 (orctx_typ x1 T) oc0; eauto. rewrite orctx_extend_typ; simpls; eauto.
      *
      forwards: IHHty2 T0 (orctx_typ x1 U) oc0; eauto. rewrite orctx_extend_typ; simpls; eauto.
    + (* separate -> stop induction, go to next segment *)
      eapply ty_inv_logic_ord_oc_nle_or_notorand_notorand_dist_eq; eauto.
      2: rewrite HeqTT. 2: eapply ty_inv_logic_ord_and; eauto.
      forwards: orctx_notorand_eq_and_le. 2: symmetry; eapply HeqTT. eauto.
      forwards: typ_notorand_separate_eq_oc oc0 ocp (typ_or T0 U0); eauto.
      eapply typ_notorand_or_invr; eauto.
Qed.

(* All intersection are in T or in U or right of oc *)
Lemma ty_inv_logic_ord_oc_nle_or_dist_eq oc T U ocp:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  ty_inv_logic_ord ocp (orctx_typ oc (typ_or T U)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hn He Hty.

  remember (orctx_typ oc (typ_or T U)) as TT.
  reverts T U oc.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_oc_or_dist_eq; eauto.
  - (* intersection *)
    symmetry in HeqTT.
    forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
    destruct_all; subst.
    + (* intersection under oc *)
      destruct x0; simpls.
      ++ (* Cannot be right at the union *)
        discriminate.
      ++ (* Left branch -> continue induction *)
        inversions H1. clear HeqTT.
        forwards: IHHty1 (orctx_typ x0 T) T1 oc0.
        eapply ty_dist_eq_oc_and_invl; eauto.
        auto.
        rewrite orctx_extend_typ; simpls; eauto.
        forwards: IHHty2 (orctx_typ x0 U) T1 oc0.
        eapply ty_dist_eq_oc_and_invr; eauto.
        auto.
        rewrite orctx_extend_typ; simpls; eauto.
        rewrite <- orctx_extend_typ in *.
        eapply ty_inv_logic_and; eauto.
      ++ (* Right branch -> go to next segment *)
        inversions H1.

        forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
        forwards: orctx_notorand_le_extend_invr Hnlocp.
        inversions H0.

        eapply ty_inv_logic_ord_oc_nle_or_notorand_dist_eq.
        eauto.
        eauto.
        eauto.
        rewrite HeqTT. eapply ty_inv_logic_ord_and; eauto.
    + (* intersection separate -> stop, go to last segment *)
      
      forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
      forwards: orctx_notorand_eq_and_le. 2: symmetry; eauto. eauto.
      forwards: typ_notorand_separate_eq_oc oc0 ocp (typ_or T0 U0); eauto.

      eapply ty_inv_logic_ord_oc_nle_or_notorand_notorand_dist_eq; eauto.
      eapply typ_notorand_or_invl; eauto.
      eapply typ_notorand_or_invr; eauto.
      rewrite HeqTT. eapply ty_inv_logic_ord_and; eauto.

Qed.

Lemma ty_inv_logic_ord_oc_nle_or_diag oc T:
  orctx_notorand_le oc ->
  ty_inv_logic_ord oc (orctx_typ oc (typ_or T T)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hn Hty.
  eapply ty_inv_logic_ord_oc_nle_or_dist_eq; eauto.
Qed.

Lemma ty_inv_logic_ord_oc_or_diag oc T:
  ty_inv_logic_ord orctx_nil (orctx_typ oc (typ_or T T)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_or T T)) as TT.
  reverts oc T.
  pose proof Hty.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_oc_or_dist_eq; eauto.
  - 
    destruct (orctx_le_cases ocp oc0).
    + (* ocp left of oc0 -> continue induction *)
      symmetry in HeqTT.
      forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
      destruct_all; subst.
      *
        eapply orctx_le_extend_l_inv in H1.
        subst; simpls. inversions H3.
      *
      forwards: orctx_separate_cross_r HeqTT H2 T0.
      destruct_all.
      forwards: H4 T.
      forwards: H4 U.
      destruct_all.

      forwards: IHHty1. eauto. symmetry; eapply H6.
      forwards: IHHty2. eauto. symmetry; eapply H7.

      rewrite H3.
      rewrite H8 in H13.
      rewrite H10 in H12.
      eapply ty_inv_logic_and; eauto.
    +
      assert (orctx_notorand_le oc0). {
      eapply orctx_notorand_le_le. eauto. symmetry; eauto.
      eapply ty_inv_logic_ord_orctx_notorand_le; eauto.
      }
      eapply ty_inv_logic_ord_oc_nle_or_diag.
      eauto.
      rewrite HeqTT in H.
      eapply ty_inv_logic_ord_notorand_le; eauto.
Qed.

End WithTyping.

Lemma ty_inv_logic_logic_ord G e x T:
  inert G ->
  ty_inv_logic G e x T ->
  ty_inv_logic_ord G e x orctx_nil T.
Proof.
  introv Hi Hh.
  induction Hh.
  -
    eapply ty_inv_logic_ord_le.
    2: eapply ty_inv_logic_ord_has_closure; eauto.
    constructor.
  - eapply ty_inv_logic_ord_le.
    2: eapply ty_inv_logic_ord_and_lc; eauto.
    constructor.
Qed.
Lemma ty_inv_logic_oc_or_diag G e x oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (orctx_typ oc (typ_or T T)) ->
  ty_inv_logic G e x (orctx_typ oc T).
Proof.
  introv Hi Hcor Hx Hty.
  eapply ty_inv_logic_logic_ord in Hty.
  eapply ty_inv_logic_ord_oc_or_diag; eauto.
  auto.
Qed.

Lemma ty_inv_logic_oclc_or_diag G e x oc oc2 T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 (typ_or T T))) ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2 T)).
Proof.
  introv Hi Hcor Hx Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_or_diag; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
  - eapply ty_inv_logic_and.
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_inv_logic_lc_or_diag G e x  oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (logctx_typ oc (typ_or T T)) ->
  ty_inv_logic G e x (logctx_typ oc T).
Proof.
  introv Hi Hcor Hx Hty.
  eapply ty_inv_logic_oclc_or_diag with (oc := orctx_nil); fweauto.
  simpls; eauto.
Qed.

Section WithTyping.
Context (G: ctx) (e: renv) (x: var).
Context (HinG: inert G) (HGx: x \in ctx_dom G).
Context (Hcor: renv_corr G e).

(* Analog of ty_inv_basic_closure_oc_or_dist_eq *)
Lemma ty_inv_basic_closure_oc_N_cases_dist_eq oc ocpy T U:
  ty_dist_eq T U ->
  ty_inv_basic_closure G e x (orctx_typ oc T) ->
  typ_notorand (orctx_typ oc T) ->
  typ_notorand (orctx_typ oc U) ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv He Hty HnT HnU Hh.
  forwards: typ_notorand_oc_notorand HnT; eauto.
  forwards: typ_notorand_oc_notorand HnU; eauto.
  forwards: ty_dist_eq_notorand_eq; eauto; subst.
  rewrite orctx_logctx_typ in *.
  forwards: ty_inv_basic_closure_N_cases; eauto.
  eapply ty_has_N_logic_ord_notorand_closure; eauto.
  destruct_all.
  left; eapply ty_inv_logic_or; eauto.
  right; eapply ty_inv_logic_or; eauto.
Qed.

(* Analog of ty_inv_logic_ord_oc_nle_or_notorand_notorand_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_N_cases_notorand_notorand_dist_eq ocpx ocpy oc T U:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  typ_notorand U ->
  ty_inv_logic_ord G e x ocpx (orctx_typ oc T) ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hn He HnT HnU Hty Hh.
  remember (orctx_typ oc T) as TT.
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_oc_N_cases_dist_eq; eauto.
    eapply typ_notorand_oc_replace; eauto.
  - 
    forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
    forwards: ty_dist_eq_notorand_eq; eauto; subst.
    symmetry in HeqTT.
    forwards: typ_notorand_eq_separate HeqTT.
    eauto.


      forwards: orctx_separate_cross_r HeqTT H0 typ_nocap.
      destruct_all.

      forwards: H2 T.
      forwards: H2 U.
      destruct_all.

      assert (orctx_notorand_le x1). {
        eapply orctx_notorand_le_le. 2: eapply H5.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }
      assert (orctx_notorand_le x2).  {
        eapply orctx_notorand_le_le. 2: eapply H4.
        eapply orctx_le_trans. eapply orctx_follow_le. eauto.
        eapply orctx_notorand_eq_and_le; eauto.
        eauto.
      }

      forwards: IHHty1 x2 U0 U0; eauto.
      forwards: IHHty2 x1 U0 U0; eauto.

      destruct_all; eauto.
      right.
      rewrite H1.
      rewrite H8 in H12.
      rewrite H6 in H13.
      eapply ty_inv_logic_and; eauto.
Qed.

(* Analog of ty_inv_logic_ord_oc_nle_or_notorand_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_N_cases_notorand_dist_eq oc T U ocpx ocpy:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  typ_notorand T ->
  ty_inv_logic_ord G e x ocpx (orctx_typ oc T) ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hn He HnT Hty Hh.
  reverts T oc ocpx.
  induction Hh; intros; subst.
  - eapply ty_inv_logic_ord_oc_nle_N_cases_notorand_notorand_dist_eq with (ocpy := oc0); eauto.
    eapply ty_has_N_logic_ord_or; eauto.
  -
    forwards: ty_dist_eq_oc_and_cases T0; eauto.
    destruct H0; eauto.
Qed.

(* Analog of ty_inv_logic_ord_oc_nle_or_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_N_cases_dist_eq oc T U ocpx ocpy:
  orctx_notorand_le oc ->
  ty_dist_eq T U ->
  ty_inv_logic_ord G e x ocpx (orctx_typ oc T) ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hn He Hty Hh.

  remember (orctx_typ oc T) as TT.
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_inv_logic_ord_oc_nle_N_cases_notorand_dist_eq with (ocpx := ocpy); eauto.
    eapply typ_notorand_oc_notorand; eauto.
    eapply ty_inv_logic_ord_or; eauto.
  - (* intersection *)
    symmetry in HeqTT.
    forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
    destruct_all; subst.
    + (* intersection under oc *)

        forwards: IHHty1 (orctx_typ x0 T) U0 oc0.
        eapply ty_dist_eq_oc_and_invl; eauto.
        auto. auto.
        rewrite orctx_extend_typ; simpls; eauto.
        forwards: IHHty2 (orctx_typ x0 U) U0 oc0.
        eapply ty_dist_eq_oc_and_invr; eauto.
        auto. auto.
        rewrite orctx_extend_typ; simpls; eauto.
        destruct_all; eauto.
      
        left.
        eapply ty_inv_logic_and; eauto.
    + (* intersection separate -> stop, go to last segment *)
      
      forwards Hnlocp: ty_inv_logic_ord_orctx_notorand_le; eauto.
      forwards: orctx_notorand_eq_and_le. 2: symmetry; eauto. eauto.
      forwards: typ_notorand_separate_eq_oc oc0 ocp T0; eauto.

      eapply ty_inv_logic_ord_oc_nle_N_cases_notorand_dist_eq; eauto.
      rewrite HeqTT. eapply ty_inv_logic_ord_and; eauto.
Qed.

Lemma ty_inv_logic_ord_oc_nle_N_cases oc ocpx ocpy T:
  orctx_notorand_le oc ->
  ty_inv_logic_ord G e x ocpx (orctx_typ oc T) ->
  ty_has_N_logic_ord G e ocpy T ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hn Hty Hh.
  eapply ty_inv_logic_ord_oc_nle_N_cases_dist_eq; eauto.
Qed.

Lemma ty_inv_logic_ord_oc_N_cases oc ocpx ocpy T:
  ty_inv_logic_ord G e x ocpx (orctx_typ oc T) ->
  ty_has_N_logic_ord G e ocpy T ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hty Hh.
  remember (orctx_typ oc T) as TT.
  reverts oc T.
  pose proof Hty.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_ord_oc_nle_N_cases; eauto. eapply orctx_notorand_le_notorand; eauto.
  - 
    destruct (orctx_le_cases_ne oc0 ocp).
    +
      assert (orctx_notorand_le oc0). {
      eapply orctx_notorand_le_le. eauto. symmetry; eauto.
      eapply ty_inv_logic_ord_orctx_notorand_le; eauto.
      }
      eapply ty_inv_logic_ord_oc_nle_N_cases.
      eauto.
      rewrite HeqTT in H.
      eauto. eauto.
    + (* ocp left of oc0 -> continue induction *)
      symmetry in HeqTT.
      forwards: orctx_intersect_cases_r HeqTT. intros; discriminate.
      destruct_all; subst.
      *
        eapply orctx_le_extend_l_inv in H1.
        subst; simpls. rewrite orctx_extend_nil_l in H3. contradiction.
      *
      forwards: orctx_separate_cross_r HeqTT H2 typ_nocap.
      destruct_all.
      forwards: H5 T.
      forwards: H5 U.
      destruct_all.

      forwards: IHHty1. eauto. symmetry; eapply H7. eauto.
      forwards: IHHty2. eauto. symmetry; eapply H8. eauto.

      destruct_all; eauto.
      right.

      rewrite H4.
      rewrite H9 in H14.
      rewrite H11 in H13.
      eapply ty_inv_logic_and; eauto.
Qed.

End WithTyping.

Lemma ty_inv_logic_oc_N_cases G e x oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (orctx_typ oc T) ->
  ty_has_N_logic G e T ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (typ_nocap)).
Proof.
  introv Hi Hcor Hin Hty Hh.
  eapply ty_inv_logic_logic_ord in Hty; eauto.
  eapply ty_has_N_logic_logic_ord in Hh.
  eapply ty_inv_logic_ord_oc_N_cases; eauto.
Qed.

Lemma ty_inv_logic_oclc_N_cases G e x oc oc2 T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2  T)) ->
  ty_has_N_logic G e T ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (orctx_typ oc (logctx_typ oc2  (typ_nocap))).
Proof.
  introv Hi Hcor Hin Hty Hh.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_inv_logic_oc_N_cases; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
    forwards: IHoc2; eauto.
    destruct H1; eauto.
    right.
    eapply ty_inv_logic_and; eauto.
  - 
    forwards: ty_inv_logic_oc_and_invl Hty. eauto.
    forwards: ty_inv_logic_oc_and_invr Hty. eauto.
    forwards: IHoc2; eauto.
    destruct H1; eauto.
    right.
    eapply ty_inv_logic_and; eauto.
Qed.

Lemma ty_inv_logic_lc_N_cases G e x oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_logic G e x (logctx_typ oc T) ->
  ty_has_N_logic G e T ->
  ty_inv_logic G e x T \/
  ty_inv_logic G e x (logctx_typ oc (typ_nocap)).
Proof.
  introv Hi Hcor Hin Hty Hh.
  eapply ty_inv_logic_oclc_N_cases with (oc := orctx_nil); eauto.
Qed.


Section WithTyping.
Context (G: ctx) (e: renv) (w: var) (y: var).
Context (HinG: inert G) (Hcor: renv_corr G e) (Hbi: binds w y e).
Context (Hm:   mode_has_top_nm_subtype typing_mode).

(* Analog of ty_inv_basic_closure_oc_N_cases_dist_eq, ty_inv_basic_closure_oc_or_dist_eq *)
Lemma ty_inv_basic_closure_ro_ref_dist_eq oc ocpy T U:
  ty_dist_eq T U ->
  ty_inv_basic_closure G e y T ->
  typ_notorand (orctx_typ oc T) ->
  typ_notorand (orctx_typ oc U) ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e w T.
Proof.
  introv He Hty HnT HnU Hh.

  forwards Hdom: renv_corr_binds_dom Hcor Hbi; eauto.
  destructecs Hdom.

  forwards: typ_notorand_oc_notorand HnT; eauto.
  forwards: typ_notorand_oc_notorand HnU; eauto.
  forwards: ty_dist_eq_notorand_eq; eauto; subst.
  rewrite orctx_logctx_typ in *.
  forwards: ty_inv_basic_closure_ro_ref; fweauto.
  eapply ty_has_N_logic_ord_notorand_closure; eauto.
  eapply ty_inv_logic_or; eauto.
Qed.


(* Analog of ty_inv_logic_ord_oc_nle_N_cases_notorand_notorand_dist_eq, ty_inv_logic_ord_oc_nle_or_notorand_notorand_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_ro_ref_notorand_notorand_dist_eq ocpx ocpy T U:
  ty_dist_eq T U ->
  typ_notorand T ->
  typ_notorand U ->
  ty_inv_logic_ord G e y ocpx T ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e w T.
Proof.
  introv He HnT HnU Hty Hh.
  reverts U.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_closure_ro_ref_dist_eq; eauto.
    eapply typ_notorand_oc_replace with (oc:=orctx_nil); eauto.
    simpl; eauto.
  - 
    eapply typ_notorand_oc_notand in HnT.
    false HnT; eauto.
Qed.

(* Analog of ty_inv_logic_ord_oc_nle_N_cases_notorand_dist_eq, ty_inv_logic_ord_oc_nle_or_notorand_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_ro_ref_notorand_dist_eq T U ocpx ocpy:
  ty_dist_eq T U ->
  typ_notorand T ->
  ty_inv_logic_ord G e y ocpx T ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e w T.
Proof.
  introv He HnT Hty Hh.
  reverts T ocpx.
  induction Hh; intros; subst.
  - eapply ty_inv_logic_ord_oc_nle_ro_ref_notorand_notorand_dist_eq; eauto.
    eapply ty_has_N_logic_ord_or with (oc0 := orctx_nil); eauto.
  -
    forwards: ty_dist_eq_oc_and_cases T0; eauto.
    destruct H0; eauto.
Qed.

(* Analog of ty_inv_logic_ord_oc_nle_N_cases_dist_eq, ty_inv_logic_ord_oc_nle_or_dist_eq *)
Lemma ty_inv_logic_ord_oc_nle_ro_ref_dist_eq T U ocpx ocpy:
  ty_dist_eq T U ->
  ty_inv_logic_ord G e y ocpx T ->
  ty_has_N_logic_ord G e ocpy U ->
  ty_inv_logic G e w T.
Proof.
  introv He Hty Hh.

  reverts U.
  induction Hty; intros; subst.
  -
    eapply ty_inv_logic_ord_oc_nle_ro_ref_notorand_dist_eq with (ocpx := ocpy); eauto.
    eapply ty_inv_logic_ord_or; eauto.
  - (* intersection *)

        forwards: IHHty1 U0.
        eapply ty_dist_eq_oc_and_invl; eauto.
        auto.
        forwards: IHHty2 U0.
        eapply ty_dist_eq_oc_and_invr; eauto.
        auto.
        eapply ty_inv_logic_and; eauto.
Qed.



Lemma ty_inv_logic_ord_ro_ref ocpx ocpy T:
  ty_inv_logic_ord G e y ocpx T ->
  ty_has_N_logic_ord G e ocpy T ->
  ty_inv_logic G e w T.
Proof.
  introv Hty Hh.
  pose proof Hty.
  induction Hty; intros; subst.
  - eapply ty_inv_logic_ord_oc_nle_ro_ref_dist_eq; eauto.
  - forwards: ty_inv_logic_ord_oc_nle_ro_ref_dist_eq; eauto.
Qed.

End WithTyping.


Lemma ty_inv_logic_ro_ref G e w y T:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  binds w y e ->
  ty_has_N_logic G e T ->
  ty_inv_logic G e y T ->
  ty_inv_logic G e w T.
Proof.
  introv Hm Hin Hcor Hbi Hh Hty.
  eapply ty_inv_logic_logic_ord in Hty; eauto.
  eapply ty_has_N_logic_logic_ord in Hh.
  eapply ty_inv_logic_ord_ro_ref; eauto.
Qed.

End WithTypingMode.