Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.

Require Import Logctx Logpath.
(* Reperesents an action of chaning a part of a type under a path *)
Inductive wrap_action: Set :=
  wrap_action_c(oc:logpath)(T U: typ): wrap_action.
(* Reperesents multiple of chaning a part of a type under a path. The head is applied last. *)
Definition wrap_plan := list wrap_action.

(* Extend the path of an action upwards *)
Definition wrap_action_under(ocp: logpath)(wa: wrap_action) :=
  match wa with
  | wrap_action_c oc T U => wrap_action_c (logpath_extend ocp oc) T U
  end.
Definition wrap_plan_under ocp (wp: wrap_plan) := List.map (wrap_action_under ocp) wp.

Fixpoint wrap_action_logpath(wa: wrap_action) :=
  match wa with
  | wrap_action_c oc T U => oc
  end.

Definition wrap_action_separate ocp wa := logpath_separate ocp (wrap_action_logpath wa).
Definition wrap_plan_separate ocp (wp: wrap_plan) := List.Forall (wrap_action_separate ocp) wp.

Definition wrap_action_leaf wa := 
  match wa with
  | wrap_action_c oc T U => typ_notlog U
  end.
Definition wrap_plan_leaf (wp: wrap_plan) := List.Forall (wrap_action_leaf) wp.

Definition wrap_action_direct wa := (wrap_action_logpath wa) = logpath_nil.
Definition wrap_plan_direct (wp: wrap_plan) := List.Forall (wrap_action_direct) wp.

(* Applies the action to a type without type checking *)
Inductive wrap_action_utyp: typ->wrap_action->typ->Prop :=
| wrap_action_utyp_c oc S T:
    wrap_action_utyp (logctx_typ oc S) (wrap_action_c (logctx_path oc) S T) (logctx_typ oc T)
.
Inductive wrap_plan_utyp: typ->wrap_plan->typ->Prop :=
| wrap_plan_utyp_nil T:
  wrap_plan_utyp T nil T
| wrap_plan_utyp_cons wp wa T S U:
  wrap_plan_utyp T wp S ->
  wrap_action_utyp S wa U ->
  wrap_plan_utyp T (cons wa wp) U
.
Hint Constructors wrap_plan_utyp: core.

Class WrapActionDetermined (P: wrap_action -> Prop) := 
  wrap_action_determined wa wb T1 T2 U:
  P wa ->
  P wb ->
  wrap_action_direct wa ->
  wrap_action_direct wb ->
  wrap_action_utyp T1 wa U ->
  wrap_action_utyp T2 wb U ->
  wa = wb /\ T1 = T2.

Definition wrap_action_injective wa := (
  forall T1 T2 U,
  wrap_action_utyp T1 wa U ->
  wrap_action_utyp T2 wa U ->
  T1 = T2
).
Definition wrap_plan_injective (wp: wrap_plan) := List.Forall (wrap_action_injective) wp.

Class WrapActionClosed (P: wrap_action -> Prop) :=
  wrap_action_closed wa oc:
  (P (wrap_action_under oc wa)) ->
  P wa.

Class WrapActionClosedInv (P: wrap_action -> Prop) :=
  wrap_action_closed_inv wa oc:
  P wa ->
  (P (wrap_action_under oc wa)).


Instance wrap_action_leaf_closed: WrapActionClosed wrap_action_leaf.
Proof.
  hnf; intros; destruct wa; hnf in *; auto.
Qed.

Lemma  wrap_plan_closed (P: wrap_action -> Prop) {WAC: WrapActionClosed P} wa oc:
  (List.Forall P (wrap_plan_under oc wa)) ->
  List.Forall P wa.
Proof.
  introv Hf.
  induction wa; inversions Hf; eauto using wrap_action_closed.
Qed.

Lemma wrap_plan_under_app (wp1 wp2: wrap_plan) (op: logpath):
  wrap_plan_under op (wp1 ++ wp2)%list = ((wrap_plan_under op wp1) ++ (wrap_plan_under op wp2))%list.
Proof.
  unfold wrap_plan_under.
  rewrite List.map_app. eauto.
Qed.
Lemma wrap_plan_under_cons wa (wp: wrap_plan) (op: logpath):
  wrap_plan_under op (wa :: wp)%list = ((wrap_action_under op wa) :: (wrap_plan_under op wp))%list.
Proof.
  unfold wrap_plan_under.
  rewrite List.map_cons. eauto.
Qed.
Lemma wrap_plan_under_nil  (op: logpath):
  wrap_plan_under op (nil)%list = nil%list.
Proof.
  unfold wrap_plan_under.
  simpl. auto.
Qed.

Lemma wrap_plan_utyp_app T wp1 S1 wp2 S2:
  wrap_plan_utyp T wp1 S1 ->
  wrap_plan_utyp S1 wp2 S2 ->
  wrap_plan_utyp T (List.app wp2 wp1) S2.
Proof.
  introv Hp1 Hp2.
  induction Hp2; simpls.
  -
    eauto.
  -
    econstructor; eauto.
Qed.
Lemma wrap_plan_utyp_app_inv T wp1 wp2 S2:
  wrap_plan_utyp T (List.app wp2 wp1) S2 ->
  exists S1,
    wrap_plan_utyp T wp1 S1 /\
    wrap_plan_utyp S1 wp2 S2.
Proof.
  introv Hp. reverts S2.
  induction wp2; intros; simpls.
  - exists S2; splits.
    eauto. constructor.
  - inversions Hp.
    forwards S1 Hp1 Hp2: IHwp2; eauto.
Qed.

Lemma wrap_action_separate_extend op1 op2 wp:
  wrap_plan_separate op1 wp ->
  wrap_plan_separate (logpath_extend op1 op2) wp.
Proof.
  introv Hsep.
  induction wp; hnf.
  eauto.
  inversions Hsep.
  constructor. hnf. eapply logpath_separate_symm. eapply logpath_separate_extend. eapply logpath_separate_symm. eauto.
  eapply IHwp; eauto.
Qed.

Lemma wrap_action_utyp_logpath_typ_preserve T op S wa U:
  logpath_typ T op S ->
  wrap_action_utyp T wa U ->
  wrap_action_separate op wa ->
  logpath_typ U op S.
Proof.
  introv Htyp Hut Hsep.
  inversions Hut.
  hnf in Hsep. simpls.
  forwards: logctx_path_ctx oc S0.
  forwards: logpath_ctx_typ_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ H0; eauto; subst.
  forwards: logpath_logctx_typ_separate_preserve; eauto.
Qed.

Lemma wrap_plan_utyp_logpath_typ_preserve T op S wa U:
  logpath_typ T op S ->
  wrap_plan_utyp T wa U ->
  wrap_plan_separate op wa ->
  logpath_typ U op S.
Proof.
  introv Htyp Hut Hsep.
  reverts S op.
  induction Hut; intros.
  - eauto.
  - inversions Hsep.
    forwards: IHHut; eauto.
    eapply wrap_action_utyp_logpath_typ_preserve; eauto.
Qed.

Lemma wrap_action_utyp_logpath_typ_rev_preserve T op S wa U:
  logpath_typ U op S ->
  wrap_action_utyp T wa U ->
  wrap_action_separate op wa ->
  logpath_typ T op S.
Proof.
  introv Htyp Hut Hsep.
  inversions Hut.
  hnf in Hsep. simpls.
  forwards: logctx_path_ctx oc S0.
  forwards: logpath_ctx_typ_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ H0; eauto; subst.
  forwards: logpath_logctx_typ_separate_preserve; eauto.
Qed.
Lemma wrap_plan_utyp_logpath_typ_rev_preserve T op S wa U:
  logpath_typ U op S ->
  wrap_plan_utyp T wa U ->
  wrap_plan_separate op wa ->
  logpath_typ T op S.
Proof.
  introv Htyp Hut Hsep.
  reverts op S.
  induction Hut; intros.
  - eauto.
  - inversions Hsep. 
    eapply IHHut; eauto.
    eapply wrap_action_utyp_logpath_typ_rev_preserve; eauto.
Qed.


Lemma wrap_action_utyp_logpath_ctx_preserve T op oc wa U:
  logpath_ctx T op oc ->
  wrap_action_utyp T (wrap_action_under op wa) U ->
  logpath_ctx U op oc.
Proof.
  introv Htyp Hut.
  destruct wa.
  inversions Hut.
  forwards: logctx_path_extend_case H1; destruct_all; subst.
  rewrite logctx_extend_typ in *.

  forwards: logctx_path_ctx  x (logctx_typ x0 T0).
  forwards: logpath_ctx_functional Htyp H; eauto; subst.
  eapply logctx_path_ctx.
Qed.


Lemma wrap_plan_utyp_logpath_ctx_preserve T op oc wa U:
  logpath_ctx T op oc ->
  wrap_plan_utyp T (wrap_plan_under op wa) U ->
  logpath_ctx U op oc.
Proof.
  introv Htyp Hut.
  remember (wrap_plan_under op wa).
  reverts op oc wa.
  induction Hut; intros.
  - eauto.
  - destruct wa0; simpls; inversions Heql.
    forwards: IHHut; eauto.
    eapply wrap_action_utyp_logpath_ctx_preserve; eauto.
Qed.
Lemma wrap_action_utyp_logpath_ctx_rev_preserve T op oc wa U:
  logpath_ctx U op oc ->
  wrap_action_utyp T (wrap_action_under op wa) U ->
  logpath_ctx T op oc.
Proof.
  introv Htyp Hut.
  destruct wa.
  inversions Hut.
  forwards: logctx_path_extend_case H1; destruct_all; subst.
  rewrite logctx_extend_typ in *.

  forwards: logctx_path_ctx  x (logctx_typ x0 U0).
  forwards: logpath_ctx_functional Htyp H; eauto; subst.
  eapply logctx_path_ctx.
Qed.


Lemma wrap_plan_utyp_logpath_ctx_rev_preserve T op oc wa U:
  logpath_ctx U op oc ->
  wrap_plan_utyp T (wrap_plan_under op wa) U ->
  logpath_ctx T op oc.
Proof.
  introv Htyp Hut.
  remember (wrap_plan_under op wa).
  reverts op oc wa.
  induction Hut; intros.
  - eauto.
  - destruct wa0; simpls; inversions Heql.
    eapply IHHut; eauto.
    eapply wrap_action_utyp_logpath_ctx_rev_preserve; eauto.
Qed.

Lemma wrap_action_utyp_under_case oc S wa T:
  wrap_action_utyp (logctx_typ oc S) (wrap_action_under (logctx_path oc) wa) (logctx_typ oc T) ->
  wrap_action_utyp S wa T.
Proof.
  introv Hut.
  unfold wrap_action_under in Hut.
  inversions Hut; destruct wa.
  inversions H.
  eapply logctx_path_extend_case in H3; destruct_all; subst.
  rewrite logctx_extend_typ in *.
  forwards: logctx_path_eq_inv H2; eauto; destruct_all; subst.
  forwards: logctx_path_eq_inv H0; eauto; destruct_all; subst.
  econstructor.
Qed.
Lemma wrap_plan_utyp_under_case oc S wa T:
  wrap_plan_utyp (logctx_typ oc S) (wrap_plan_under (logctx_path oc) wa) (logctx_typ oc T) ->
  wrap_plan_utyp S wa T.
Proof.
  introv Hut.
  reverts T.
  induction wa; intros.
  - simpls. inversions Hut. eapply logctx_same_typ_eq_inv in H0; subst. constructor.
  - inversions Hut.
    forwards: wrap_plan_utyp_logpath_ctx_preserve H2. eapply logctx_path_ctx.
    forwards: logpath_ctx_typ_exists H. destruct_all.
    forwards: logpath_ctx_typ H0 H; eauto; subst.
    forwards: IHwa; eauto.
    econstructor; eauto.
    eapply wrap_action_utyp_under_case; eauto.
Qed.

Lemma wrap_action_utyp_under oc S wa T:
  wrap_action_utyp S wa T ->
  wrap_action_utyp (logctx_typ oc S) (wrap_action_under (logctx_path oc) wa) (logctx_typ oc T).
Proof.
  introv Hut.
  unfold wrap_action_under.
  destruct wa; inversions Hut.
  rewrite <- ! logctx_extend_typ.
  rewrite <- ! logctx_path_extend.
  econstructor.  
Qed.
Lemma wrap_plan_utyp_under oc S wa T:
  wrap_plan_utyp S wa T ->
  wrap_plan_utyp (logctx_typ oc S) (wrap_plan_under (logctx_path oc) wa) (logctx_typ oc T).
Proof.
  introv Hut.
  reverts T.
  induction wa; intros.
  - simpls. inversions Hut. constructor.
  - inversions Hut.
    econstructor; eauto using wrap_action_utyp_under.    
Qed.

Lemma wrap_action_utyp_functional T wa S1 S2:
  wrap_action_utyp T wa S1 ->
  wrap_action_utyp T wa S2 ->
  S1 = S2.
Proof.
  introv HwaS1 HwaS2.
  induction HwaS1; inversions HwaS2.
  forwards: logctx_path_eq_inv; eauto; destruct_all; subst; auto.
Qed.

Lemma wrap_action_utyp_inv_functional T1 T2 wa S:
  wrap_action_utyp T1 wa S ->
  wrap_action_utyp T2 wa S ->
  T1 = T2.
Proof.
  introv HwaS1 HwaS2.
  induction HwaS1; inversions HwaS2.
  forwards: logctx_path_eq_inv; eauto; destruct_all; subst; auto.
Qed.

Definition wrap_action_leaf_typ (P: logpath -> Prop) (T: typ) :=
  forall op1 op2 oc,
    P (logpath_extend op1 op2) ->
    logpath_ctx T op2 oc ->
    op2 = logpath_nil.

Definition wrap_action_leaf_for (P: logpath -> Prop) wa :=
  match wa with
  | wrap_action_c op T U => (P op /\ wrap_action_leaf_typ P U)
  end.
Definition wrap_plan_leaf_for (P: logpath -> Prop)  (wp: wrap_plan) := List.Forall (wrap_action_leaf_for P) wp.

Lemma wrap_action_leaf_leaf_for wa:
  wrap_action_leaf wa -> wrap_action_leaf_for (fun op => True) wa.
Proof.
  intros.
  hnf in *; destruct wa.
  splits; auto 1.
  hnf; intros.
  hnf in H.
  destruct op2; simpls; inversions H1; auto 1; destruct_all.
    + forwards: H T0 S; contradiction.
    + forwards: H S T0; contradiction.
    + forwards: H1 T0 S; contradiction.
    + forwards: H1 S T0; contradiction.
Qed.

Lemma wrap_action_intersect_cases wa oc S T:
  wrap_action_leaf wa ->
  wrap_action_utyp S wa (logctx_typ oc T) ->
  (wrap_action_separate (logctx_path oc) wa \/ exists op2, (wrap_action_logpath wa) = (logpath_extend (logctx_path oc) op2)).
Proof.
  introv Hlf Hutyp.
  inversions Hutyp.
  inversions Hlf. hnf in H0.
  forwards: logctx_intersect_cases_nonnil H2; destruct_all; subst; simpls.
  - right. exists logpath_nil. rewrite logpath_extend_nil_l; auto.
  - destruct x; simpls.
    + contradiction.
    + forwards: H (logctx_typ x T) T0; contradiction.
    + forwards: H T0 (logctx_typ x T); contradiction.
    + forwards: H0 (logctx_typ x T) T0; contradiction.
    + forwards: H0 T0 (logctx_typ x T); contradiction.
  - right. exists (logctx_path x). rewrite logctx_path_extend. auto.
  - left. hnf; simpls. eapply logctx_separate_path_separate; eauto. eapply logctx_separate_symm; eauto.
Qed.

Lemma wrap_action_intersect_cases_for (P: logpath -> Prop) wa oc S T:
  P (logctx_path oc) ->
  wrap_action_leaf_for P wa ->
  wrap_action_utyp S wa (logctx_typ oc T) ->
  (wrap_action_separate (logctx_path oc) wa \/ exists op2, (wrap_action_logpath wa) = (logpath_extend (logctx_path oc) op2)).
Proof.
  introv Hoc Hlf Hutyp.
  inversions Hutyp.
  inversions Hlf. hnf in H0.
  forwards: logctx_intersect_cases_nonnil H2; destruct_all; subst; simpls.
  - right. exists logpath_nil. rewrite logpath_extend_nil_l; auto.
  - forwards: H0 (logctx_path oc0) (logctx_path x) x; eauto.
    rewrite <- logctx_path_extend; auto.
    eapply logctx_path_ctx.
    destruct x; simpls; try contradiction; discriminate.
  - right. exists (logctx_path x). rewrite logctx_path_extend. auto.
  - left. hnf; simpls. eapply logctx_separate_path_separate; eauto. eapply logctx_separate_symm; eauto.
Qed.

Lemma wrap_action_leaf_path_preserve wa op S T U:
  wrap_action_leaf wa ->
  wrap_action_utyp S wa T ->
  logpath_typ T op U ->
  exists SU, 
  logpath_typ S op SU.
Proof.
  introv Hlf Hwa Hop.
  forwards: logpath_typ_ctx_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ; eauto; subst.
  forwards: logpath_ctx_path H; subst; eauto.
  forwards: wrap_action_intersect_cases; eauto.
  destruct_all.
  -
    forwards: wrap_action_utyp_logpath_typ_rev_preserve; eauto.
  - destruct wa; simpls; subst.
    inversions Hwa.
    forwards: logctx_path_typ oc T.
    rewrite H2 in H0.
    forwards: logpath_typ_extend_case H0; destruct_all.
    eauto.
Qed.

Lemma wrap_action_leaf_path_preserve_for (P: logpath -> Prop) wa op S T U:
  P op ->
  wrap_action_leaf_for P wa ->
  wrap_action_utyp S wa T ->
  logpath_typ T op U ->
  exists SU, 
  logpath_typ S op SU.
Proof.
  introv HopP Hlf Hwa Hop.
  forwards: logpath_typ_ctx_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ; eauto; subst.
  forwards: logpath_ctx_path H; subst; eauto.
  forwards: wrap_action_intersect_cases_for; eauto.
  destruct_all.
  -
    forwards: wrap_action_utyp_logpath_typ_rev_preserve; eauto.
  - destruct wa; simpls; subst.
    inversions Hwa.
    forwards: logctx_path_typ oc T.
    rewrite H2 in H0.
    forwards: logpath_typ_extend_case H0; destruct_all.
    eauto.
Qed.

Lemma wrap_plan_leaf_path_preserve wa op S T U:
  wrap_plan_leaf wa ->
  wrap_plan_utyp S wa T ->
  logpath_typ T op U ->
  exists SU, 
  logpath_typ S op SU.
Proof.
  introv Hlf Hwa Hop. 
  reverts S T U.
  induction wa; intros.
  - inversions Hwa; eauto.
  - inversions Hlf. inversions Hwa.
    forwards: wrap_action_leaf_path_preserve; eauto; destruct_all.
    forwards: IHwa; eauto.
Qed.
Lemma wrap_plan_leaf_path_preserve_for (P: logpath -> Prop) wa op S T U:
  P op ->
  wrap_plan_leaf_for P wa ->
  wrap_plan_utyp S wa T ->
  logpath_typ T op U ->
  exists SU, 
  logpath_typ S op SU.
Proof.
  introv HopP Hlf Hwa Hop. 
  reverts S T U.
  induction wa; intros.
  - inversions Hwa; eauto.
  - inversions Hlf. inversions Hwa.
    forwards: wrap_action_leaf_path_preserve_for; eauto; destruct_all.
    forwards: IHwa; eauto.
Qed.

Lemma wrap_action_separate_replace wa oc1 oc2 T U:
  (logctx_path oc1) = (logctx_path oc2) ->
  wrap_action_utyp (logctx_typ oc1 T) wa (logctx_typ oc2 T) ->
  wrap_action_separate (logctx_path oc1) wa ->
  wrap_action_utyp (logctx_typ oc1 U) wa (logctx_typ oc2 U).
Proof.
  introv Hop Hut Hsep.
  inversions Hut.
  unfold wrap_action_separate in Hsep.
  simpls.
  forwards Hsepc1: logctx_path_separate Hsep. eapply logctx_path_ctx. rewrite <- H0. eapply logctx_path_ctx.
  rewrite Hop in Hsep.
  forwards Hsepc2: logctx_path_separate Hsep. eapply logctx_path_ctx. rewrite <- H. eapply logctx_path_ctx.
  symmetry in H0.
  forwards: logctx_separate_cross_r H0 Hsepc1 U. destruct_all.
  symmetry in H.
  forwards: logctx_path_eq_follow Hop.
  forwards: logctx_recombine H4 H3; eauto.
  rewrite H1, H5.
  forwards: logctx_follow_path_eq H3. rewrite H6.
  econstructor.
Qed.

Lemma wrap_plan_separate_replace wa oc1 oc2 T U:
  (logctx_path oc1) = (logctx_path oc2) ->
  wrap_plan_utyp (logctx_typ oc1 T) wa (logctx_typ oc2 T) ->
  wrap_plan_separate (logctx_path oc1) wa ->
  wrap_plan_utyp (logctx_typ oc1 U) wa (logctx_typ oc2 U).
Proof.
  introv Hop Hut Hsep.
  reverts oc2.
  induction wa; intros.
  - inversions Hut.
    forwards: logctx_path_eq_follow Hop.
    forwards: logctx_follow_eq_inv H; eauto; destruct_all; subst.
    constructor.
  - inversions Hut. inversions Hsep.
    destruct a.
    forwards: wrap_action_utyp_logpath_typ_rev_preserve H4 H1. rewrite Hop. eapply logctx_path_typ.
    forwards: logpath_typ_ctx_exists H. destruct_all.
    forwards: logpath_ctx_typ; eauto; subst.
    forwards Hoc1x: logpath_ctx_path H0.
    symmetry in Hoc1x.
    forwards: IHwa H3 Hoc1x H2.
    econstructor; eauto.
    eapply wrap_action_separate_replace; rewrite  <- ? Hoc1x; eauto.
Qed.

Lemma wrap_action_utyp_swap T wa1 wa2 T1 T2:
  logpath_separate (wrap_action_logpath wa1) (wrap_action_logpath wa2) ->
  wrap_action_utyp T wa1 T1 ->
  wrap_action_utyp T1 wa2 T2 ->
  exists S2,
    wrap_action_utyp T wa2 S2 /\
    wrap_action_utyp S2 wa1 T2.
Proof.
  introv Hsep HTwa1 HT1wa2.
  inversions HTwa1.
  inversions HT1wa2.
  simpls.

  forwards HctxS: logctx_path_ctx oc0 S0.
  forwards HctxT: logctx_path_ctx oc T0.
  (* We know the paths are separate, but not that the contexts are sepearate.
     So try all cases on the contexts and rule out all except separate *)
  forwards: logctx_intersect_cases H0; destruct_all; subst.
  - false; eapply logpath_separate_eq_false; eauto.
  - rewrite logctx_path_extend in Hsep.
    eapply logpath_separate_symm in Hsep.
    false; eapply logpath_separate_extend_false; eauto.
  - rewrite logctx_path_extend in Hsep.
    false; eapply logpath_separate_extend_false; eauto.
  - (* Now contexts are separate, cross the contexts to find the new middle type *)
    symmetry in H0.
    eapply logctx_separate_symm in H.
    forwards: logctx_separate_cross_r H0 H S.
    destruct_all.
    eapply logctx_follow_path_eq in H3.
    rewrite H1.
    rewrite H3.
    exists (logctx_typ x T).
    splits.
    constructor.
    specialize (H2 T).
    destruct_all.
    eapply logctx_follow_path_eq in H5.
    rewrite <- H4.
    rewrite <- H5.
    rewrite <- H2.
    constructor.
Qed.

Lemma wrap_plan_action_utyp_swap T wa1 wa2 T1 T2:
  wrap_plan_separate (wrap_action_logpath wa2) wa1 ->
  wrap_plan_utyp T wa1 T1 ->
  wrap_action_utyp T1 wa2 T2 ->
  exists S2,
    wrap_action_utyp T wa2 S2 /\
    wrap_plan_utyp S2 wa1 T2.
Proof.
  introv Hsep HTwa1 HT1wa2.
  reverts T2 wa2 T1 T.
  induction wa1; intros.
  - inversions HTwa1. exists T2. split. eauto. constructor.
  - inversions HTwa1. inversions Hsep.
    eapply logpath_separate_symm in H1.
    forwards: wrap_action_utyp_swap H1 H4 HT1wa2.
    destruct_all.
    forwards: IHwa1; eauto.
    destruct_all.
    exists x0. split. eauto. econstructor; eauto.
Qed.
Lemma wrap_action_plan_utyp_swap T wa1 wa2 T1 T2:
  wrap_plan_separate (wrap_action_logpath wa1) wa2 ->
  wrap_action_utyp T wa1 T1 ->
  wrap_plan_utyp T1 wa2 T2 ->
  exists S2,
    wrap_plan_utyp T wa2 S2 /\
    wrap_action_utyp S2 wa1 T2.
Proof.
  introv Hsep HTwa1 HT1wa2.
  reverts T2 wa1 T1 T.
  induction wa2; intros.
  - inversions HT1wa2. exists T. split; eauto.
  - inversions HT1wa2. inversions Hsep.

    forwards: IHwa2; eauto.
    destruct_all.

    forwards: wrap_action_utyp_swap H1 H0 H4.
    destruct_all.

    exists x0. split. econstructor; eauto. eauto.
Qed.

Lemma wrap_plan_segment_ind T wp1 wp1ex wp1sep wp1rest S op U:
  logpath_typ U op S ->
  wrap_plan_utyp T wp1 U ->
  wrap_plan_leaf wp1 ->
  wp1 = (List.app (List.app (wrap_plan_under op wp1ex) wp1sep) wp1rest) ->
  wrap_plan_separate op wp1sep ->
  exists wpex wpsep wpsepex ns,
    wpsepex = (List.app (wrap_plan_under op wpex) wpsep) /\
    list_reorder wp1 ns wpsepex /\
    LibList.noduplicates ns /\
    length ns = length wp1 /\
    wrap_plan_utyp T wpsepex U /\
    wrap_plan_separate op wpsep.
Proof.
  introv Hp Htyp Hlf.
  reverts S wp1 wp1ex wp1sep.
  induction wp1rest; intros.
  - rewrite List.app_nil_r in *.
    exists wp1ex wp1sep wp1.
    forwards: list_reorder_refl wp1; destruct_all.
    forwards: list_reorder_lengths H1.
    exists x.
    splits; eauto.
  -
    rewrite H in Htyp.
    forwards: wrap_plan_utyp_app_inv Htyp; destruct_all.
    inversions H1.
    forwards: list_forall_app_invl Hlf.
    forwards: wrap_plan_leaf_path_preserve H2; eauto. destruct_all.
    forwards: logpath_typ_ctx_exists H1. destruct_all.
    forwards: logpath_ctx_typ H1 H3; eauto; subst.
    forwards: logpath_ctx_path H3; subst.

    forwards: list_forall_app_invr Hlf. inversions H4.

    forwards: wrap_action_intersect_cases H8. eauto.
    destruct H4.
    + (* separate *)
      forwards: IHwp1rest ((wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep) ++ a :: wp1rest)%list wp1ex (wp1sep ++ (a::nil))%list.
      eauto. eauto. eauto. rewrite (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex )). rewrite <- (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep)). rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
      eapply list_forall_app; eauto.
      destruct_all.
      exists x x2 x3 x4; splits; eauto.
    + destruct_all.
      destruct a; simpls; subst.
      set (b:=(wrap_action_c x T0 U0)).
      forwards: IHwp1rest ((wrap_plan_under (logctx_path x1) (wp1ex) ++ (wrap_action_under (logctx_path x1) b):: wp1sep) ++ wp1rest)%list (wp1ex++b::nil)%list (wp1sep)%list.
      *
      eauto.
      *
      (* split Htyp *)
      forwards: wrap_plan_utyp_app_inv Htyp; destruct_all.
      inversions H4.
      forwards: wrap_plan_utyp_app_inv H5; destruct_all.
      (* swap action with wp1sep *)
      forwards: wrap_action_plan_utyp_swap H15 H4.
      hnf; simpls.
      eapply wrap_action_separate_extend; eauto.
      destruct_all.

      (* Construct the plan *)
      eapply wrap_plan_utyp_app. eauto.
      eapply wrap_plan_utyp_app.
      econstructor. eauto. eauto. eauto.
      *
      forwards: list_forall_app_invl Hlf.
      forwards: list_forall_app_invr Hlf.
      forwards: list_forall_app_invr H4.
      forwards: list_forall_app_invl H4.
      inversions H5.

      eapply list_forall_app.
      eapply list_forall_app. eauto.
      constructor. eauto. eauto. eauto.
      *       (* Move b in front of wp1sep *)
      rewrite wrap_plan_under_app.
      rewrite wrap_plan_under_cons,  wrap_plan_under_nil.
      rewrite <- (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex) (wrap_action_under (logctx_path x1) (wrap_action_c x T0 U0) ::nil)).
      rewrite <- List.app_comm_cons.
      rewrite List.app_nil_l.
      unfold b. unfold wrap_action_under. eauto.
      * eauto.
      * (* Final goal *)
        destructc H4 as wpex wpsep wpsepex Hrest; destruct_all.
        exists wpex wpsep wpsepex.
        (* Find reordering indices *)
        assert (exists ns2, list_reorder (
          (wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep) ++ wrap_action_c (logpath_extend (logctx_path x1) x) T0 U0 :: wp1rest)
          ns2
          ((wrap_plan_under (logctx_path x1) wp1ex ++ wrap_action_c (logpath_extend (logctx_path x1) x) T0 U0 :: wp1sep) ++ wp1rest) /\
          LibList.noduplicates ns2 /\
          LibList.Forall (gt (length x2)) ns2
        ).
        forwards: list_reorder_move_inv_n (length wp1ex) (length wp1sep) (length wp1rest). destruct_all.
        exists x3. splits. rewrite <- ! List.app_assoc. eapply H17; unfold wrap_plan_under; rewrite ? List.map_length; auto. auto.
        rewrite H11. rewrite ! List.app_length. simpl. unfold wrap_plan_under; rewrite ? List.map_length. rewrite <- plus_n_Sm. rewrite plus_Sn_m. auto.
        destruct_all.
        forwards: list_reorder_gt H5.
        forwards: list_reorder_exists x3 x2.
        forwards: list_reorder_lengths H14.
        rewrite <- H18. auto.

        destruct_all.

        exists x4.
  
        splits; eauto 1.
  
        eapply list_reorder_trans. eapply H14. eapply H5. eapply H18.
        eapply list_permuation_noduplicates; eauto.
        forwards: list_reorder_lengths H18. rewrite H19, H11. rewrite ! List.app_length. simpl. unfold wrap_plan_under; rewrite ? List.map_length. rewrite <- plus_n_Sm. rewrite plus_Sn_m. auto.
Qed.

Lemma wrap_plan_segment_ind_for (P: logpath -> Prop) T wp1 wp1ex wp1sep wp1rest S op U:
  P op ->
  logpath_typ U op S ->
  wrap_plan_utyp T wp1 U ->
  wrap_plan_leaf_for P wp1 ->
  wp1 = (List.app (List.app (wrap_plan_under op wp1ex) wp1sep) wp1rest) ->
  wrap_plan_separate op wp1sep ->
  exists wpex wpsep wpsepex ns,
    wpsepex = (List.app (wrap_plan_under op wpex) wpsep) /\
    list_reorder wp1 ns wpsepex /\
    LibList.noduplicates ns /\
    length ns = length wp1 /\
    wrap_plan_utyp T wpsepex U /\
    wrap_plan_separate op wpsep.
Proof.
  introv HopP Hp Htyp Hlf.
  reverts S wp1 wp1ex wp1sep.
  induction wp1rest; intros.
  - rewrite List.app_nil_r in *.
    exists wp1ex wp1sep wp1.
    forwards: list_reorder_refl wp1; destruct_all.
    forwards: list_reorder_lengths H1.
    exists x.
    splits; eauto.
  -
    rewrite H in Htyp.
    forwards: wrap_plan_utyp_app_inv Htyp; destruct_all.
    inversions H1.
    forwards: list_forall_app_invl Hlf.
    forwards: wrap_plan_leaf_path_preserve_for H2; eauto. destruct_all.
    forwards: logpath_typ_ctx_exists H1. destruct_all.
    forwards: logpath_ctx_typ H1 H3; eauto; subst.
    forwards: logpath_ctx_path H3; subst.

    forwards: list_forall_app_invr Hlf. inversions H4.

    forwards: wrap_action_intersect_cases_for HopP H8. eauto.
    destruct H4.
    + (* separate *)
      forwards: IHwp1rest ((wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep) ++ a :: wp1rest)%list wp1ex (wp1sep ++ (a::nil))%list.
      eauto. eauto. eauto. rewrite (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex )). rewrite <- (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep)). rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
      eapply list_forall_app; eauto.
      destruct_all.
      exists x x2 x3 x4; splits; eauto.
    + destruct_all.
      destruct a; simpls; subst.
      set (b:=(wrap_action_c x T0 U0)).
      forwards: IHwp1rest ((wrap_plan_under (logctx_path x1) (wp1ex) ++ (wrap_action_under (logctx_path x1) b):: wp1sep) ++ wp1rest)%list (wp1ex++b::nil)%list (wp1sep)%list.
      *
      eauto.
      *
      (* split Htyp *)
      forwards: wrap_plan_utyp_app_inv Htyp; destruct_all.
      inversions H4.
      forwards: wrap_plan_utyp_app_inv H5; destruct_all.
      (* swap action with wp1sep *)
      forwards: wrap_action_plan_utyp_swap H16 H4.
      hnf; simpls.
      eapply wrap_action_separate_extend; eauto.
      destruct_all.

      (* Construct the plan *)
      eapply wrap_plan_utyp_app. eauto.
      eapply wrap_plan_utyp_app.
      econstructor. eauto. eauto. eauto.
      *
      forwards: list_forall_app_invl Hlf.
      forwards: list_forall_app_invr Hlf.
      forwards: list_forall_app_invr H4.
      forwards: list_forall_app_invl H4.
      inversions H5.

      eapply list_forall_app.
      eapply list_forall_app. eauto.
      constructor. eauto. eauto. eauto.
      *       (* Move b in front of wp1sep *)
      rewrite wrap_plan_under_app.
      rewrite wrap_plan_under_cons,  wrap_plan_under_nil.
      rewrite <- (List.app_assoc (wrap_plan_under (logctx_path x1) wp1ex) (wrap_action_under (logctx_path x1) (wrap_action_c x T0 U0) ::nil)).
      rewrite <- List.app_comm_cons.
      rewrite List.app_nil_l.
      unfold b. unfold wrap_action_under. eauto.
      * eauto.
      * (* Final goal *)
        destructc H4 as wpex wpsep wpsepex Hrest; destruct_all.
        exists wpex wpsep wpsepex.
        (* Find reordering indices *)
        assert (exists ns2, list_reorder (
          (wrap_plan_under (logctx_path x1) wp1ex ++ wp1sep) ++ wrap_action_c (logpath_extend (logctx_path x1) x) T0 U0 :: wp1rest)
          ns2
          ((wrap_plan_under (logctx_path x1) wp1ex ++ wrap_action_c (logpath_extend (logctx_path x1) x) T0 U0 :: wp1sep) ++ wp1rest) /\
          LibList.noduplicates ns2 /\
          LibList.Forall (gt (length x2)) ns2
        ).
        forwards: list_reorder_move_inv_n (length wp1ex) (length wp1sep) (length wp1rest). destruct_all.
        exists x3. splits. rewrite <- ! List.app_assoc. eapply H18; unfold wrap_plan_under; rewrite ? List.map_length; auto. auto.
        rewrite H11. rewrite ! List.app_length. simpl. unfold wrap_plan_under; rewrite ? List.map_length. rewrite <- plus_n_Sm. rewrite plus_Sn_m. auto.
        destruct_all.
        forwards: list_reorder_gt H5.
        forwards: list_reorder_exists x3 x2.
        forwards: list_reorder_lengths H15.
        rewrite <- H19. auto.

        destruct_all.

        exists x4.
  
        splits; eauto 1.
  
        eapply list_reorder_trans. eapply H15. eapply H5. eapply H19.
        eapply list_permuation_noduplicates; eauto.
        forwards: list_reorder_lengths H19. rewrite H20, H11. rewrite ! List.app_length. simpl. unfold wrap_plan_under; rewrite ? List.map_length. rewrite <- plus_n_Sm. rewrite plus_Sn_m. auto.
Qed.

Lemma wrap_plan_segment T wp S op U:
  logpath_typ U op S ->
  wrap_plan_utyp T wp U ->
  wrap_plan_leaf wp ->
  exists wpex wpsep wpsepex ns,
    wpsepex = (List.app (wrap_plan_under op wpex) wpsep) /\
    list_reorder wp ns wpsepex /\
    LibList.noduplicates ns /\
    (length ns) = (length wp) /\
    wrap_plan_utyp T wpsepex U /\
    wrap_plan_separate op wpsep.
Proof.
  introv Hp Htyp Hlf.

  forwards: wrap_plan_segment_ind wp (nil: wrap_plan) (nil: wrap_plan) wp; eauto.
  constructor.
Qed.

Lemma wrap_plan_segment_for (P: logpath -> Prop) T wp S op U:
  P op ->
  logpath_typ U op S ->
  wrap_plan_utyp T wp U ->
  wrap_plan_leaf_for P wp ->
  exists wpex wpsep wpsepex ns,
    wpsepex = (List.app (wrap_plan_under op wpex) wpsep) /\
    list_reorder wp ns wpsepex /\
    LibList.noduplicates ns /\
    (length ns) = (length wp) /\
    wrap_plan_utyp T wpsepex U /\
    wrap_plan_separate op wpsep.
Proof.
  introv HopP Hp Htyp Hlf.

  forwards: wrap_plan_segment_ind_for wp (nil: wrap_plan) (nil: wrap_plan) wp; eauto.
  constructor.
Qed.




Lemma wrap_plan_determined_case P {WAD: WrapActionDetermined P} wp1 wp2 T1 T2 U:
  wrap_plan_utyp T1 wp1 U ->
  wrap_plan_utyp T2 wp2 U ->
  wrap_plan_direct wp1 ->
  wrap_plan_direct wp2 ->
  List.Forall P wp1 ->
  List.Forall P wp2 ->
  (
    (exists wp12, wp2 = (wp1 ++ wp12)%list /\  wrap_plan_utyp T2 wp12 T1 ) \/
    (exists wp21, wp1 = (wp2 ++ wp21)%list /\  wrap_plan_utyp T1 wp21 T2)
  ).
Proof.
  introv Hut1 Hut2 Hd1 Hd2 Hf1 Hf2.
  reverts T2 wp2.
  induction Hut1; intros.
  - left. exists wp2. rewrite List.app_nil_l. auto.
  - inversions Hut2.
    + right. exists (wa :: wp)%list. rewrite List.app_nil_l. split; auto. econstructor; eauto.
    +
      inversions Hd1. inversions Hf1.
      inversions Hd2. inversions Hf2.
      forwards: wrap_action_determined H H1; eauto.
      destruct_all; subst.
      forwards: IHHut1 wp0; eauto.
      destruct_all; subst.
      * left. exists x. rewrite List.app_comm_cons. auto.
      * right. exists x. rewrite List.app_comm_cons. auto.
Qed.

Lemma wrap_action_notlog_direct T wa S:
  typ_notlog T ->
  wrap_action_utyp S wa T ->
  wrap_action_direct wa.
Proof.
  introv Hn Hut.
  destruct wa.
  inversions Hut.
  destruct oc0; simpls.
  constructor.
  all: destructc Hn as Hn1 Hn2.
  forwards: Hn1 (logctx_typ oc0 U) T; contradiction.
  forwards: Hn1  T (logctx_typ oc0 U); contradiction.
  forwards: Hn2 (logctx_typ oc0 U) T; contradiction.
  forwards: Hn2  T (logctx_typ oc0 U); contradiction.
Qed.

Lemma wrap_action_utyp_logpath_rev_preserve S0 oc2 a S:
  wrap_action_utyp S0 a (logctx_typ oc2 S) ->
  wrap_action_separate (logctx_path oc2) a ->
  exists oc1,
    S0 = (logctx_typ oc1 S) /\
    logctx_follow oc1 oc2.
Proof.
  introv Hut Hsep.
  forwards: wrap_action_utyp_logpath_typ_rev_preserve Hut Hsep.
  eapply logctx_path_typ.

  forwards: logpath_typ_ctx_exists; eauto; destruct_all.
  forwards:  logpath_ctx_typ; eauto.
  forwards HopTS: logpath_ctx_path; eauto.
  eapply logctx_path_eq_follow in HopTS.
  eauto.
Qed.

Lemma wrap_action_opposite TB T wp U1 U2:
  wrap_action_utyp T wp ((logpath_andor_typ TB) U1 U2) ->
  wrap_action_separate (logpath_andor_l TB logpath_nil) wp ->
  wrap_action_leaf wp ->
  exists wp0,
    wp = (wrap_action_under (logpath_andor_r TB logpath_nil) wp0).
Proof.
  introv Hut Hsep Hlf.
  inversions Hut.
  inversions Hlf.
  destruct oc; destruct TB; simpls; subst; try solve [forwards: H0 U1 U2; contradiction | forwards: H U1 U2; contradiction];  inversions H2;
  inversions Hsep.
  - inversions H3.
  - forwards: H2 (logctx_path oc); destruct_all; contradiction.
  - exists (wrap_action_c (logctx_path oc) S T0). simpls; auto.
  - inversions H3.
  - forwards: H2 (logctx_path oc); destruct_all; contradiction.
  - exists (wrap_action_c (logctx_path oc) S T0). simpls; auto.  
Qed.

Lemma wrap_action_separate_lr TB wa:
  wrap_action_separate (logpath_andor_l TB logpath_nil) (wrap_action_under (logpath_andor_r TB logpath_nil) wa).
Proof.
  hnf.
  destruct TB;
  constructor; intros; destruct wa; simpls; splits; discriminate.
Qed.
Lemma wrap_action_separate_rl TB wa:
  wrap_action_separate (logpath_andor_r TB logpath_nil) (wrap_action_under (logpath_andor_l TB logpath_nil) wa).
Proof.
  hnf.
  destruct TB;
  constructor; intros; destruct wa; simpls; splits; discriminate.
Qed.


Lemma wrap_action_andor_preserve_r S TB wa U1 U2:
  wrap_action_utyp S (wrap_action_under (logpath_andor_r TB logpath_nil) wa) (logpath_andor_typ TB U1 U2) ->
  exists S2, 
    S = (logpath_andor_typ TB U1 S2) /\
    wrap_action_utyp S2 wa U2.
Proof.
  introv Hut.
  forwards Hctx: wrap_action_utyp_logpath_ctx_rev_preserve (logpath_andor_ctx_r TB U1 logctx_nil) Hut.
  destruct TB; simpls; auto.
  forwards Htyp: wrap_action_utyp_logpath_typ_rev_preserve  (logpath_andor_l TB logpath_nil) U1  Hut.
  destruct TB; simpls; auto.
  eapply wrap_action_separate_lr.
  forwards: logpath_ctx_typ_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ; eauto; subst.
  rewrite <- logpath_andor_typ_r in Hut.
  rewrite <- logpath_andor_path_r with (U2:=U1) in Hut.
  forwards: wrap_action_utyp_under_case Hut.
  rewrite logpath_andor_typ_r.
  eauto.
Qed.
Lemma wrap_action_andor_preserve_l S TB wa U1 U2:
  wrap_action_utyp S (wrap_action_under (logpath_andor_l TB logpath_nil) wa) (logpath_andor_typ TB U1 U2) ->
  exists S1, 
    S = (logpath_andor_typ TB S1 U2) /\
    wrap_action_utyp S1 wa U1.
Proof.
  introv Hut.
  forwards Hctx: wrap_action_utyp_logpath_ctx_rev_preserve (logpath_andor_ctx_l TB logctx_nil U2) Hut.
  destruct TB; simpls; auto.
  forwards Htyp: wrap_action_utyp_logpath_typ_rev_preserve (logpath_andor_r TB logpath_nil) U2  Hut.
  destruct TB; simpls; auto.
  eapply wrap_action_separate_rl.
  forwards: logpath_ctx_typ_exists; eauto; destruct_all.
  forwards: logpath_ctx_typ; eauto; subst.
  rewrite <- logpath_andor_typ_l in Hut.
  rewrite <- logpath_andor_path_l with (U2:=U2) in Hut.
  forwards: wrap_action_utyp_under_case Hut.
  rewrite logpath_andor_typ_l.
  eauto.
Qed.
Lemma wrap_plan_andor_preserve_l S TB wa U1 U2:
  wrap_plan_utyp S (wrap_plan_under (logpath_andor_l TB logpath_nil) wa) (logpath_andor_typ TB U1 U2) ->
  exists S1, 
    S = (logpath_andor_typ TB S1 U2) /\
    wrap_plan_utyp S1 wa U1.
Proof.
  introv Hut. reverts U1 U2.
  induction wa; intros.
  - simpls; inversions Hut. eauto.
  - simpls. inversions Hut.
    forwards: wrap_action_andor_preserve_l; eauto; destruct_all; subst.
    forwards: IHwa; eauto; destruct_all; subst.
    eauto.
Qed.
Lemma wrap_plan_andor_preserve_r S TB wa U1 U2:
  wrap_plan_utyp S (wrap_plan_under (logpath_andor_r TB logpath_nil) wa) (logpath_andor_typ TB U1 U2) ->
  exists S2, 
    S = (logpath_andor_typ TB U1 S2) /\
    wrap_plan_utyp S2 wa U2.
Proof.
  introv Hut. reverts U1 U2.
  induction wa; intros.
  - simpls; inversions Hut. eauto.
  - simpls. inversions Hut.
    forwards: wrap_action_andor_preserve_r; eauto; destruct_all; subst.
    forwards: IHwa; eauto; destruct_all; subst.
    eauto.
Qed.



Lemma wrap_plan_opposite TB T wp U1 U2:
  wrap_plan_utyp T wp ((logpath_andor_typ TB) U1 U2) ->
  wrap_plan_separate (logpath_andor_l TB logpath_nil) wp ->
  wrap_plan_leaf wp ->
  exists wp0,
    wp = (wrap_plan_under (logpath_andor_r TB logpath_nil) wp0).
Proof.
  introv Hut Hsep Hlf.
  reverts U1 U2.
  induction wp; intros.
  - exists (nil:wrap_plan). simpls. auto.
  - inversions Hlf.
    inversions Hut.
    inversions Hsep.
    forwards: wrap_action_opposite; eauto.
    destruct_all. subst.
    forwards: wrap_action_andor_preserve_r; eauto.
    destruct_all; subst.
    forwards: IHwp; eauto.
    destruct_all; subst.
    exists (x:: x1)%list; simpls; auto.
Qed.

Lemma wrap_plan_andor_segment TB T wp U1 U2:
  wrap_plan_utyp T wp ((logpath_andor_typ TB) U1 U2) ->
  wrap_plan_leaf wp ->
  exists wpex wpsep wpsepex ns,
    wpsepex = (List.app (wrap_plan_under (logpath_andor_l TB logpath_nil) wpex) (wrap_plan_under (logpath_andor_r TB logpath_nil) wpsep)) /\
    list_reorder wp ns wpsepex /\
    wrap_plan_utyp T wpsepex ((logpath_andor_typ TB) U1 U2) /\
    length ns = length wp.
Proof.
  introv Hut Hlf.
  (*exists (logpath_andor_ctx_l TB logctx_nil U2) (logctx_andr U1 logctx_nil).*)
    forwards: wrap_plan_segment T wp U1 (logpath_andor_l TB logpath_nil ) ((logpath_andor_typ TB) U1 U2); eauto.
    destruct TB; auto.
    destruct_all; subst.
    rename H0 into Hre, H1 into Hnd, H2 into Hl, H3 into Huts, H4 into Hsep.
    forwards: wrap_plan_utyp_app_inv Huts; destruct_all.
    forwards: wrap_plan_andor_preserve_l; eauto; destruct_all; subst.
    forwards: wrap_plan_opposite H; eauto. {
      forwards Hlfre: list_forall_reorder Hre Hlf.
      forwards Hlfre1: list_forall_app_invr Hlfre. auto.
    }
    destruct_all; subst.
    exists x x1 ( (wrap_plan_under (logpath_andor_l TB logpath_nil) x ++ wrap_plan_under (logpath_andor_r TB logpath_nil) x1)%list) x2.
    splits; eauto.
Qed.

Lemma wrap_plan_andor_split T TB Tl Tr wpl wpr:
  wrap_plan_utyp T
       (wrap_plan_under (logpath_andor_l TB logpath_nil) wpl ++ wrap_plan_under (logpath_andor_r TB logpath_nil) wpr)%list
       (logpath_andor_typ TB Tl Tr) ->
  exists Tl0 Tr0,
    T = (logpath_andor_typ TB Tl0 Tr0) /\
    wrap_plan_utyp Tl0 wpl Tl /\
    wrap_plan_utyp Tr0 wpr Tr.
Proof.
  introv Hut.
  eapply wrap_plan_utyp_app_inv in Hut. destruct_all.
  forwards Hutl: wrap_plan_andor_preserve_l H0. destruct_all; subst.
  forwards Hutr: wrap_plan_andor_preserve_r H. destruct_all; subst.
  eauto.
Qed.

Lemma wrap_plan_separate_common_same  P {WAD: WrapActionDetermined P} {WAP: WrapActionClosed P} S waT waU T U:
  wrap_plan_utyp T waT S ->
  wrap_plan_utyp U waU S ->
  wrap_plan_leaf waT ->
  wrap_plan_leaf waU ->
  List.Forall P waT ->
  List.Forall P waU ->
  exists TU waTUU waTUT nU nT,
    wrap_plan_utyp TU waTUU U /\
    wrap_plan_utyp TU waTUT T /\
    list_reorder waU nU waTUT /\
    list_reorder waT nT waTUU.
Proof.
  introv HutT HutU HlfT HlfU HPT HPU.

  (* Induction on the length of the list.
     Because the list will be also split into two parts of unknown length, 
     we will allow the list to be shorter than the number. *)
  remember (Nat.max (length waT) (length waU)) as n.

  forwards Hler: Nat.le_max_r (length waT) (length waU).
  forwards Hlel: Nat.le_max_l (length waT) (length waU).
  rewrite <- Heqn in Hler, Hlel.
  clear Heqn.

  reverts S T U.
  reverts waU waT.


  induction n using nat_lt_induction; intros.

  (* Induction on the tree of the type. In each node, we will split the lists into two parts that
      each apply to one side of the tree *)
  forwards HSlog: typ_typ_log S.

  reverts waU waT.
  reverts T U.
  induction HSlog; intros; subst.
  + (* T0 is notlog, so a must be direct *)
    destruct waT; simpls.
    - (* nil *)
      subst.
      inversions HutT.
      forwards: list_reorder_refl waU; destruct_all.
      exists U (nil:wrap_plan) waU x (nil: list nat).
      splits. constructor. auto.
      eauto.
      eapply list_reorder_nil.
    -
      destruct waU. 
      * (* nil *)
      inversions HutU.
      forwards: list_reorder_refl (w :: waT)%list; destruct_all.
      exists T0 (w::waT)%list (nil:wrap_plan) (nil: list nat) x.
      splits. auto. constructor.
      eapply list_reorder_nil.
      eauto.
      * (* both lists are non-nil *)
        inversions HutT.
        inversions HutU.
        inversions HPT.
        inversions HPU.
        inversions HlfT.
        inversions HlfU.
        forwards: wrap_action_notlog_direct H6; eauto.
        forwards: wrap_action_notlog_direct H8; eauto.
        forwards: wrap_action_determined H6 H8; eauto.
        destruct_all; subst.
        simpls.
        forwards: H (Nat.max (length waT)(length waU) ) H4 H5; eauto.
        eapply Nat.max_lub_lt; eauto.
        eapply Nat.le_max_r.
        eapply Nat.le_max_l.
        destruct_all; subst.
        (* Both plans agree, so there is no need to add to any of the lists that fix the differences *)
        exists x (x0)%list (x1)%list (List.map S x2) (List.map S x3).
        splits; eauto using list_reorder_push0.  

  + (* and/or type, so split the plans *)

    rename T1 into Sl, T2 into Sr.

    (* Split the plans *)
    forwards: typ_andor_logpath_andor H0; destruct_all; subst.
    forwards: wrap_plan_andor_segment HutT; eauto.
    forwards: wrap_plan_andor_segment HutU; eauto.
    destruct_all.

    rename x0 into waUl, x1 into waUr, x2 into waUs, x3 into nUs.
    rename x4 into waTl, x5 into waTr, x6 into waTs, x7 into nTs.
    rename x into TB.

    rename H4 into HutUs, H7 into HutTs.
    rename H3 into HreUs, H6 into HreTs.
    rename H5 into HlnUs, H8 into HlnTs.

    forwards HlwaUs: list_reorder_lengths HreUs.
    forwards HlwaTs: list_reorder_lengths HreTs.
 
    (* Apply the split plans to the subtrees *)
    subst.
    forwards HspUs: wrap_plan_andor_split HutUs; destruct_all.
    forwards HspTs: wrap_plan_andor_split HutTs; destruct_all.
    subst.
    rename x into USl, x0 into USr.
    rename x1 into TSl, x2 into TSr.

    (* Simplify length equations *)
    rewrite List.app_length in HlwaTs, HlwaUs.
    unfold wrap_plan_under in HlwaTs, HlwaUs.
    rewrite ! List.map_length in HlwaTs, HlwaUs.

    (* Recursion on left *)
    forwards Hl: IHHSlog1 waTl waUl; eauto.
    (* Preserved properties of actions *)
    eapply wrap_plan_closed;
    eauto using list_forall_app_invl, list_forall_reorder, wrap_action_leaf_closed.
    eapply wrap_plan_closed;
    eauto using list_forall_app_invl, list_forall_reorder, wrap_action_leaf_closed.
  
    eapply Nat.le_trans. 2: eapply Hlel. rewrite <- HlnTs, <- HlwaTs.
    eapply Nat.le_add_r.

    eapply wrap_plan_closed;
    eauto using list_forall_app_invl, list_forall_reorder, wrap_action_leaf_closed.
    eapply wrap_plan_closed;
    eauto using list_forall_app_invl, list_forall_reorder, wrap_action_leaf_closed.

    eapply Nat.le_trans. 2: eapply Hler. rewrite <- HlnUs, <- HlwaUs.
    eapply Nat.le_add_r.

    (* Recursion on right *)
    forwards: IHHSlog2 waTr waUr; eauto.
    (* Preserved properties of actions *)
    eapply wrap_plan_closed;
    eauto using list_forall_app_invr, list_forall_reorder, wrap_action_leaf_closed.
    eapply wrap_plan_closed;
    eauto using list_forall_app_invr, list_forall_reorder, wrap_action_leaf_closed.
  
    eapply Nat.le_trans. 2: eapply Hlel. rewrite <- HlnTs, <- HlwaTs.
    eapply Plus.le_plus_r.

    eapply wrap_plan_closed;
    eauto using list_forall_app_invr, list_forall_reorder, wrap_action_leaf_closed.
    eapply wrap_plan_closed;
    eauto using list_forall_app_invr, list_forall_reorder, wrap_action_leaf_closed.

    eapply Nat.le_trans. 2: eapply Hler. rewrite <- HlnUs, <- HlwaUs.
    eapply Plus.le_plus_r.

    (* Extract results of recursion *)
    destruct_all.

    rename x into TUSr, x0 into waUTr, x1 into waTUr, x2 into nUTr, x3 into nTUr.
    rename x4 into TUSl, x5 into waUTl, x6 into waTUl, x7 into nUTl, x8 into nTUl.

    exists (logpath_andor_typ TB TUSl TUSr).
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waTUl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waTUr)%list.
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waUTl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waUTr)%list.

    eapply list_reorder_app_case_ys in HreUs.
    eapply list_reorder_app_case_ys in HreTs.
    destruct_all.
    rename x1 into nTsl, x2 into nTsr.
    rename x into nUsl, x0 into nUsr.
    subst.

    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in H8.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in H9.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in H11.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in H12.

    forwards: list_reorder_trans_exists H16 H9.
    forwards: list_reorder_trans_exists H17 H11.
    forwards: list_reorder_trans_exists H18 H8.
    forwards: list_reorder_trans_exists H15 H12.
    destruct_all.
    rename x2 into nUTsr, x1 into nTUsl, x0 into nTUsr, x into nUTsl.


    exists (nUTsl++nUTsr)%list.
    exists (nTUsl++nTUsr)%list.

    splits.

    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB TUSl USr)).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=TUSl).
    eapply wrap_plan_utyp_under; eauto.

    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=USr).
    eapply wrap_plan_utyp_under; eauto.

    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB TUSl TSr)).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=TUSl).
    eapply wrap_plan_utyp_under; eauto.

    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=TSr).
    eapply wrap_plan_utyp_under; eauto.
    
    (* Reordering *)
    eapply list_reorder_app; eauto.
    eapply list_reorder_app; eauto.
Qed.


Lemma wrap_action_separate_under wa opsep opex:
  wrap_action_separate opsep wa ->
  wrap_action_separate (logpath_extend opex opsep) (wrap_action_under opex wa).
Proof.
  introv Hsep.
  unfold wrap_action_separate, wrap_action_under in *.
  destruct wa.
  unfold wrap_action_logpath in *.
  eapply logpath_separate_extend_left; eauto.
Qed.
Lemma wrap_plan_separate_under wa opsep opex:
  wrap_plan_separate opsep wa ->
  wrap_plan_separate (logpath_extend opex opsep) (wrap_plan_under opex wa).
Proof.
  introv Hsep.
  induction wa.
  - simpls. constructor.
  - inversions Hsep.
    forwards: IHwa; eauto.
    constructor; eauto using wrap_action_separate_under.
Qed.
Lemma wrap_action_separate_under_inv wa opsep opex:
  wrap_action_separate (logpath_extend opex opsep) (wrap_action_under opex wa) ->
  wrap_action_separate opsep wa.
Proof.
  introv Hsep.
  unfold wrap_action_separate, wrap_action_under in *.
  destruct wa.
  unfold wrap_action_logpath in *.
  eapply logpath_separate_extend_inv; eauto.
Qed.
Lemma wrap_plan_separate_under_inv wa opsep opex:
  wrap_plan_separate (logpath_extend opex opsep) (wrap_plan_under opex wa) ->
  wrap_plan_separate opsep wa.
Proof.
  introv Hsep.
  induction wa.
  - simpls. constructor.
  - inversions Hsep.
    forwards: IHwa; eauto.
    constructor; eauto using wrap_action_separate_under_inv.
Qed.

Lemma wrap_plan_separate_common P {WAD: WrapActionDetermined P} {WAP: WrapActionClosed P}  oc ocT ocU waT waU T U:
  wrap_plan_utyp (logctx_typ ocT T) waT (logctx_typ oc T) ->
  wrap_plan_utyp (logctx_typ ocU U) waU (logctx_typ oc U) ->
  wrap_plan_separate (logctx_path ocT) waT ->
  wrap_plan_separate (logctx_path ocU) waU ->
  wrap_plan_leaf waT ->
  wrap_plan_leaf waU ->
  logctx_follow ocT oc ->
  logctx_follow ocU oc ->
  List.Forall P waT ->
  List.Forall P waU ->
  exists ocTU waTUU waTUT nU nT,
    (forall S, wrap_plan_utyp (logctx_typ ocTU S) waTUU (logctx_typ ocU S)) /\
    (forall S, wrap_plan_utyp (logctx_typ ocTU S) waTUT (logctx_typ ocT S)) /\
    list_reorder waU nU waTUT /\
    list_reorder waT nT waTUU.
Proof.
  introv HutT HutU HsepT HsepU.
  introv HlfT HlfU HfT HfU HPT HPU.
  reverts ocT ocU T U.
  reverts waT waU.
  (* Induction which handles and and or together *)
  induction oc using logctx_andor_ind; intros.
  - (* Nil - the part, where it applies separately *)
    inversions HfT. inversions HfU. simpls. inversions HsepT. inversions HsepU.
    + exists logctx_nil (nil: wrap_plan) (nil: wrap_plan) (nil: list nat) (nil: list nat); simpls. splits; intros.
      constructor.
      constructor.
      eapply list_reorder_nil.
      eapply list_reorder_nil.
    + inversions H.
    + inversions H.
  - (* Left *)
    rename T into S, T0 into T.
    simpls.
    rewrite logpath_andor_typ_l' in HutT, HutU.
    forwards HsegT: wrap_plan_andor_segment HutT; eauto.
    forwards HsegU: wrap_plan_andor_segment HutU; eauto.
    destructc HsegT as wpexT wpsepT wpsepexT nsT HrestT.
    destructc HrestT as HeqTs HreT HutTs HlT (* HutsepexT HsepexT*).
    destructc HsegU as wpexU wpsepU wpsepexU nsU HrestU.
    destructc HrestU as HeqUs HreU HutUs HlU (*HutsepexU HsepexU*).
    subst.
    forwards HspUs: wrap_plan_andor_split HutUs; destruct_all.
    forwards HspTs: wrap_plan_andor_split HutTs; destruct_all.
    subst.
    rename x into USl, x0 into USr.
    rename x1 into TSl, x2 into TSr.
    rename H into HocU, H2 into HocT, H0 into HUSl, H1 into HUSr, H3 into HTSl, H4 into HTSr.

    (* Prepare separation *)
    assert (wrap_plan_separate (logctx_path oc) wpexT) as HsepexT. {
      eapply wrap_plan_separate_under_inv with (opex:=(logpath_andor_l TB logpath_nil)).
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepT).
      eapply list_forall_reorder; eauto.
      (* Equate the path with HsepT *)
      eapply logctx_follow_path_eq in HfT.
      rewrite HfT in HsepT.
      rewrite logpath_andor_path_l' in HsepT.
      destruct TB; simpls; eauto.
    }
    assert (wrap_plan_separate (logctx_path oc) wpexU) as HsepexU. {
      eapply wrap_plan_separate_under_inv with (opex:=(logpath_andor_l TB logpath_nil)).
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepU).
      eapply list_forall_reorder; eauto.
      (* Equate the path with HsepT *)
      eapply logctx_follow_path_eq in HfU.
      rewrite HfU in HsepU.
      rewrite logpath_andor_path_l' in HsepU.
      destruct TB; simpls; eauto.
    }
    assert (wrap_plan_leaf wpexT) as HlfexT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepT).
      eapply list_forall_reorder; eauto.
    }
    assert (wrap_plan_leaf wpexU) as HlfexU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepU).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpexT) as HfexT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eauto.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepT).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpexU) as HfexU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eauto.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepU).
      eapply list_forall_reorder; eauto.
    }

    forwards: logctx_follow_logpath_andor_l HfT.
    forwards: logctx_follow_logpath_andor_l HfU.
    destruct_all; subst.
    rename x into ocU, x1 into ocT.
    rewrite logpath_andor_typ_l' in HocU.
    eapply logctx_andor_typ_inv in HocU.
    rewrite logpath_andor_typ_l' in HocT.
    eapply logctx_andor_typ_inv in HocT.
    destruct_all; subst.
    clear H H4.
    rename H2 into HocT, H1 into HocU.
  

    (* Left part by induction *)
    (* Find the path *)
    forwards HocUtyp: wrap_plan_utyp_logpath_typ_rev_preserve (logctx_path oc) U HUSl. eapply logctx_path_typ. eauto.
    forwards HocTtyp: wrap_plan_utyp_logpath_typ_rev_preserve (logctx_path oc) T HTSl. eapply logctx_path_typ. eauto.

    forwards HocTctx: logpath_typ_ctx_exists HocTtyp.
    forwards HocUctx: logpath_typ_ctx_exists HocUtyp.
    destruct_all.
    rename x0 into ocTS, x into ocUS.
    rename H0 into HocUS, H into HocTS.

    forwards HocUStyp: logpath_ctx_typ HocTS; eauto.
    forwards HocTStyp: logpath_ctx_typ HocUS; eauto.
    subst.

    forwards HopTS: logpath_ctx_path HocTS.
    forwards HopUS: logpath_ctx_path HocUS.

    (* apply induction *)
    forwards Hl: IHoc wpexT wpexU HTSl HUSl; eauto.
    eapply logctx_follow_path_eq in HocT.
    rewrite HocT; eauto.
    eapply logctx_follow_path_eq in HocU.
    rewrite HocU; eauto.

    destructc Hl as ocTUl waTUUl waTUTl nUs nTs Hrestl.
    destructc Hrestl as HwaTUUl HwaTUTl HreUs HreTs.

    (* Right part is shared *)

    assert (wrap_plan_leaf wpsepT) as HlfsepT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexT).
      eapply list_forall_reorder; eauto.
    }
    assert (wrap_plan_leaf wpsepU) as HlfsepU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexU).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpsepT) as HfsepT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eauto.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexT).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpsepU) as HfsepU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eauto.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexU).
      eapply list_forall_reorder; eauto.
    }
    forwards Hr: wrap_plan_separate_common_same P HTSr HUSr; eauto.

    destructc Hr as TUr waTUUr waTUTr nTr nUr Hrestr.
    destructc Hrestr as HwaTUUr HwaTUTr HreUr HreTr.

    exists (logpath_andor_ctx_l TB ocTUl TUr).
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waTUUl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waTUUr)%list.
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waTUTl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waTUTr)%list.

    eapply list_reorder_app_case_ys in HreU.
    eapply list_reorder_app_case_ys in HreT.
    destruct_all.
    rename x1 into nTsl, x2 into nTsr.
    rename x into nUsl, x0 into nUsr.
    subst.
    rename H3 into HreTsl, H4 into HreTsr, H1 into HreUsl, H2 into HreUsr.
    (* Reorder indices *)

    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in HreTr.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in HreUr.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in HreTs.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in HreUs.

    forwards: list_reorder_trans_exists HreTsl HreTs.
    forwards: list_reorder_trans_exists HreUsl HreUs.
    forwards: list_reorder_trans_exists HreTsr HreTr.

    forwards: list_reorder_trans_exists HreUsr HreUr.
    destruct_all.
    rename x2 into nTUUl, x1 into nTUTl, x0 into nTUUr, x into nTUTr.
    rename H2 into HnTUTr, H1 into HnTUUr, H0 into HnTUTl, H into HnTUUl.


    exists (nTUTl++nTUTr)%list.
    exists (nTUUl++nTUUr)%list.
    
    (* Prove goal conditions *)
    splits; intros; simpls.

    *
    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB (logctx_typ ocTUl S0) USr)).
    **
    rewrite logpath_andor_typ_l' with (oc:=ocTUl).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=(logctx_typ ocTUl S0)).
    eapply wrap_plan_utyp_under; eauto.
    **
    rewrite logpath_andor_typ_l' with (oc:=ocU).
    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=USr).
    eapply wrap_plan_utyp_under; eauto.
    *
    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB (logctx_typ ocTUl S0) TSr)).
    **
    rewrite logpath_andor_typ_l' with (oc:=ocTUl).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=(logctx_typ ocTUl S0)).
    eapply wrap_plan_utyp_under; eauto.
    **
    rewrite logpath_andor_typ_l' with (oc:=ocT).
    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=TSr).
    eapply wrap_plan_utyp_under; eauto.
    *
    eapply list_reorder_app; eauto.
    *
    eapply list_reorder_app; eauto.
  - (* Right *)
    rename T into S, T0 into T.
    simpls.
    rewrite logpath_andor_typ_r' in HutT, HutU.
    forwards HsegT: wrap_plan_andor_segment HutT; eauto.
    forwards HsegU: wrap_plan_andor_segment HutU; eauto.
    destructc HsegT as wpexT wpsepT wpsepexT nsT HrestT.
    destructc HrestT as HeqTs HreT HutTs HlT (* HutsepexT HsepexT*).
    destructc HsegU as wpexU wpsepU wpsepexU nsU HrestU.
    destructc HrestU as HeqUs HreU HutUs HlU (*HutsepexU HsepexU*).
    subst.
    forwards HspUs: wrap_plan_andor_split HutUs; destruct_all.
    forwards HspTs: wrap_plan_andor_split HutTs; destruct_all.
    subst.
    rename x into USl, x0 into USr.
    rename x1 into TSl, x2 into TSr.
    rename H into HocU, H2 into HocT, H0 into HUSl, H1 into HUSr, H3 into HTSl, H4 into HTSr.

    (* Prepare separation *)
    assert (wrap_plan_separate (logctx_path oc) wpsepT) as HsepexT. {
      eapply wrap_plan_separate_under_inv with (opex:=(logpath_andor_r TB logpath_nil)).
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexT).
      eapply list_forall_reorder; eauto.
      (* Equate the path with HsepT *)
      eapply logctx_follow_path_eq in HfT.
      rewrite HfT in HsepT.
      rewrite logpath_andor_path_r' in HsepT.
      destruct TB; simpls; eauto.
    }
    assert (wrap_plan_separate (logctx_path oc) wpsepU) as HsepexU. {
      eapply wrap_plan_separate_under_inv with (opex:=(logpath_andor_r TB logpath_nil)).
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexU).
      eapply list_forall_reorder; eauto.
      (* Equate the path with HsepT *)
      eapply logctx_follow_path_eq in HfU.
      rewrite HfU in HsepU.
      rewrite logpath_andor_path_r' in HsepU.
      destruct TB; simpls; eauto.
    }
    assert (wrap_plan_leaf wpsepT) as HlfsepT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexT).
      eapply list_forall_reorder; eauto.
    }
    assert (wrap_plan_leaf wpsepU) as HlfsepU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexU).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpsepT) as HfsepT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eauto.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexT).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpsepU) as HfsepU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_r TB logpath_nil)). eauto.
      eapply list_forall_app_invr with (xs:= wrap_plan_under (logpath_andor_l TB logpath_nil) wpexU).
      eapply list_forall_reorder; eauto.
    }

    forwards: logctx_follow_logpath_andor_r HfT.
    forwards: logctx_follow_logpath_andor_r HfU.
    destruct_all; subst.
    rename x into ocU, x1 into ocT.
    rewrite logpath_andor_typ_r' in HocU.
    eapply logctx_andor_typ_inv in HocU.
    rewrite logpath_andor_typ_r' in HocT.
    eapply logctx_andor_typ_inv in HocT.
    destruct_all; subst.
    clear H H4.
    rename H2 into HocT, H1 into HocU.
  

    (* Right part by induction *)
    (* Find the path *)
    forwards HocUtyp: wrap_plan_utyp_logpath_typ_rev_preserve (logctx_path oc) U HUSr. eapply logctx_path_typ. eauto.
    forwards HocTtyp: wrap_plan_utyp_logpath_typ_rev_preserve (logctx_path oc) T HTSr. eapply logctx_path_typ. eauto.

    forwards HocTctx: logpath_typ_ctx_exists HocTtyp.
    forwards HocUctx: logpath_typ_ctx_exists HocUtyp.
    destruct_all.
    rename x0 into ocTS, x into ocUS.
    rename H0 into HocUS, H into HocTS.

    forwards HocUStyp: logpath_ctx_typ HocTS; eauto.
    forwards HocTStyp: logpath_ctx_typ HocUS; eauto.
    subst.

    forwards HopTS: logpath_ctx_path HocTS.
    forwards HopUS: logpath_ctx_path HocUS.

    (* apply induction *)
    forwards Hr: IHoc wpsepT wpsepU HTSr HUSr; eauto.
    eapply logctx_follow_path_eq in HocT.
    rewrite HocT; eauto.
    eapply logctx_follow_path_eq in HocU.
    rewrite HocU; eauto.

    destructc Hr as ocTUr waTUUr waTUTr nUs nTs Hrestr.
    destructc Hrestr as HwaTUUr HwaTUTr HreUs HreTs.

    (* Left part is shared *)

    assert (wrap_plan_leaf wpexT) as HlfexT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepT).
      eapply list_forall_reorder; eauto.
    }
    assert (wrap_plan_leaf wpexU) as HlfexU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eapply wrap_action_leaf_closed.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepU).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpexT) as HfexT. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eauto.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepT).
      eapply list_forall_reorder; eauto.
    }
    assert(List.Forall P wpexU) as HfexU. {
      eapply wrap_plan_closed with (oc:=(logpath_andor_l TB logpath_nil)). eauto.
      eapply list_forall_app_invl with (ys:= wrap_plan_under (logpath_andor_r TB logpath_nil) wpsepU).
      eapply list_forall_reorder; eauto.
    }
    forwards Hl: wrap_plan_separate_common_same P HTSl HUSl; eauto.

    destructc Hl as TUl waTUUl waTUTl nTl nUl Hrestl.
    destructc Hrestl as HwaTUUl HwaTUTl HreUl HreTl.

    exists (logpath_andor_ctx_r TB TUl ocTUr).
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waTUUl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waTUUr)%list.
    exists (wrap_plan_under (logpath_andor_l TB logpath_nil) waTUTl ++
        wrap_plan_under (logpath_andor_r TB logpath_nil) waTUTr)%list.

    eapply list_reorder_app_case_ys in HreU.
    eapply list_reorder_app_case_ys in HreT.
    destruct_all.
    rename x1 into nTsl, x2 into nTsr.
    rename x into nUsl, x0 into nUsr.
    subst.
    rename H3 into HreTsl, H4 into HreTsr, H1 into HreUsl, H2 into HreUsr.
    (* Reorder indices *)

    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in HreTs.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_r TB logpath_nil))) in HreUs.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in HreTl.
    eapply list_reorder_map with (f:= (wrap_action_under (logpath_andor_l TB logpath_nil))) in HreUl.

    forwards: list_reorder_trans_exists HreTsl HreTl.
    forwards: list_reorder_trans_exists HreUsl HreUl.
    forwards: list_reorder_trans_exists HreTsr HreTs.
    forwards: list_reorder_trans_exists HreUsr HreUs.
    destruct_all.
    rename x2 into nTUUl, x1 into nTUTl, x0 into nTUUr, x into nTUTr.
    rename H2 into HnTUTr, H1 into HnTUUr, H0 into HnTUTl, H into HnTUUl.


    exists (nTUTl++nTUTr)%list.
    exists (nTUUl++nTUUr)%list.
    
    (* Prove goal conditions *)
    splits; intros; simpls.

    *
    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB TUl (logctx_typ ocU S0))).

    **
    rewrite logpath_andor_typ_r' with (oc:=ocTUr).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=TUl).
    eapply wrap_plan_utyp_under; eauto.
    **
    rewrite logpath_andor_typ_r' with (oc:=ocU).
    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=(logctx_typ ocU S0)).
    eapply wrap_plan_utyp_under; eauto.

    *
    eapply wrap_plan_utyp_app with (S1:=(logpath_andor_typ TB TUl (logctx_typ ocT S0))).
    **
    rewrite logpath_andor_typ_r' with (oc:=ocTUr).
    rewrite <- ! logpath_andor_typ_r.
    rewrite <- logpath_andor_path_r with (U2:=TUl).
    eapply wrap_plan_utyp_under; eauto.
    **
    rewrite logpath_andor_typ_r' with (oc:=ocT).
    rewrite <- ! logpath_andor_typ_l.
    rewrite <- logpath_andor_path_l with (U2:=(logctx_typ ocT S0)).
    eapply wrap_plan_utyp_under; eauto.

    *
    eapply list_reorder_app; eauto.
    *
    eapply list_reorder_app; eauto.
Qed.
