(** remove printing ~ *)

(** This module reasons about the precise types of variables in inert contexts. *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode GeneralTyping RecordTypes InertContext Narrowing Indep.

(** * Precise Typing

    Precise typing is used to reason about the types of variables.
    Precise typing does not "modify" a variable's type through subtyping.
    We start out with a type [T=G(x)] (the type to which the variable is
    bound in [G]). Then we use precise typing to additionally deconstruct [T]
    by using recursion elimination and intersection elimination. #<br>#
    For example, if [G(x)=μ(x: {a: T} /\ {B: S..U})], then we can derive the following
    precise types for [x]:               #<br>#
    [G ⊢! x: μ(x: {a: T} /\ {B: S..U})] #<br>#
    [G ⊢! x: {a: T} /\ {B: S..U}]        #<br>#
    [G ⊢! x: {a: T}]                    #<br>#
    [G ⊢! x: {B: S..U}].                *)

Opaque open_rec fv.


(** * Precise Flow *)
(** We use the precise flow relation to reason about the relations between
    the precise type of a variable [G ⊢! x: T] and the type that the variable
    is bound to in the context [G(x)=T'].#<br>#
    If [G(x) = T], the [precise_flow] relation describes all the types [U] that [x] can
    derive through precise typing ([⊢!]).
    If [precise_flow x G T U], denoted as [G ⊢ x: T ⪼ U],
    then [G(x) = T] and [G ⊢! x: U].   #<br>#
    For example, if [G(x) = mu(x: {a: T} /\ {B: S..U})], then we can derive the following
    precise flows for [x]:                                                  #<br>#
    [G ⊢! x: μ(x: {a: T} /\ {B: S..U}) ⪼ μ(x: {a: T} /\ {B: S..U}]         #<br>#
    [G ⊢! x: μ(x: {a: T} /\ {B: S..U}) ⪼ {a: T} /\ {B: S..U}]               #<br>#
    [G ⊢! x: μ(x: {a: T} /\ {B: S..U}) ⪼ {a: T}]                           #<br>#
    [G ⊢! x: μ(x: {a: T} /\ {B: S..U}) ⪼ {B: S..U}]. *)

Section WithTypingMode.
Context (typing_mode: Mode).

Reserved Notation "G '⊢!' x '∶' T '⪼' U" (at level 40, x at level 59).

Inductive precise_flow : var -> ctx -> typ -> typ -> Prop :=

(** [G(x) = T]       #<br>#
    [ok G]           #<br>#
    [――――――――――――――] #<br>#
    [G ⊢! x: T ⪼ T] *)
  | pf_bind : forall x G T,
      ctx_binds x T G ->
      G ⊢! x ∶ T ⪼ T

(** [G ⊢! x: T ⪼ μ(U)] #<br>#
    [――――――――――――――――――] #<br>#
    [G ⊢! x: T ⪼ U^x]       *)
  | pf_open : forall x G T U,
      G ⊢! x ∶ T ⪼ typ_bnd U ->
      G ⊢! x ∶ T ⪼ open x U

(** [G ⊢! x: T ⪼ U1 /\ U2]   #<br>#
    [――――――――――――――――――――]   #<br>#
    [G ⊢! x: T ⪼ U1]        *)
  | pf_and1 : forall x G T U1 U2,
      G ⊢! x ∶ T ⪼ typ_and U1 U2 ->
      G ⊢! x ∶ T ⪼ U1

(** [G ⊢! x: T ⪼ U1 /\ U2]   #<br>#
    [――――――――――――――――――――]   #<br>#
    [G ⊢! x: T ⪼ U2]        *)
  | pf_and2 : forall x G T U1 U2,
      G ⊢! x ∶ T ⪼ typ_and U1 U2 ->
      G ⊢! x ∶ T ⪼ U2

where "G '⊢!' x '∶' T '⪼' U" := (precise_flow x G T U).

Hint Constructors precise_flow : core.

(** ** Precise Flow Lemmas *)

(** If [G ⊢! x: T ⪼ U] then [G(x) = T]. *)
Lemma pf_binds: forall G x T U,
    G ⊢! x ∶ T ⪼ U ->
    ctx_binds x T G.
Proof.
  introv Pf. induction Pf; auto.
Qed.

(* 5.110 PrecCut *)
Lemma pf_push_inv G y vk S x T U:
  G § y / vk ~ S ⊢! x ∶ T ⪼ U ->
  x <> y ->
  G ⊢! x ∶ T ⪼ U.
Proof.
  intros.
  dependent induction H; eauto.
  - apply ctx_binds_push_inv_n in H; eauto.
Qed.

(* A precise type of a variable cannot conain a variable that does not occur before it *)
Lemma pf_concat_notin_fv G1 G2 x T U y:
  G1 §& G2 ⊢! x ∶ T ⪼ U ->
  x §# G2 ->
  y §# G1 ->
  y \notin ctx_fv G1 ->
  y \notin fv U.
Proof.
  intros.
  assert (x <> y). {
    intro; subst. apply pf_binds in H.
    eapply ctx_binds_concat_left_inv in H; eauto.
    eapply ctx_binds_dom in H; eauto.
  }
  dependent induction H; eauto.
  - pose proof (ctx_binds_concat_left_inv G1 G2 H H0).
    eapply ctx_binds_notin_fv; eauto.
  - eapply notin_fv_open_bnd; eauto.
  - assert (y \notin fv (typ_and U1 U2)); eauto.
    rewrite fv_typ_eq in *.
    unfold fv_typ in H4. apply notin_union_r1 in H4. auto.
  - assert (y \notin fv (typ_and U1 U2)); eauto.
    rewrite fv_typ_eq in *.
    unfold fv_typ in H4. apply notin_union_r2 in H4. auto.
Qed.


(** If [G(x) = forall(S)T], then [x]'s precise type can be only [forall(S)T]. *)
Lemma precise_flow_fun_inv : forall x G S T U,
    G ⊢! x ∶ typ_fun S T ⪼ U ->
    U = typ_fun S T.
Proof.
  introv Hpf.
  dependent induction Hpf; auto;
    specialize (IHHpf S T eq_refl); inversion IHHpf.
Qed.

(** The following lemma say that the type to which a variable is bound in an inert context is inert. *)
(** See [inert_binds_inert]. *)
Lemma pf_inert_T : forall G x T U,
    inert G ->
    G ⊢! x ∶ T ⪼ U ->
    inert_type T.
Proof.
  introv Hi Pf. induction Pf; eauto.
  apply (inert_binds_inert Hi H).
Qed.

(** See [inert_typ_bnd_record] *)
Lemma pf_rcd_T : forall G x T U S,
    inert G ->
    G ⊢! x ∶ (typ_objref T S) ⪼ U ->
    record_type T.
Proof.
  introv Hi Pf.
  apply (pf_inert_T Hi) in Pf.
  destructc Pf as vk Hin.
  inversions Hin; assumption.
Qed.

Lemma pf_refcap_inert_type : forall G x T U S,
    inert G ->
    G ⊢! x ∶ (typ_objref T S) ⪼ U ->
    exists vk, refcap_inert_type S vk.
Proof.
  introv Hi Pf.
  apply (pf_inert_T Hi) in Pf.
  destructc Pf as vk Hin.
  inversions Hin.
  exists vk.
  assumption.
Qed.

Lemma pf_refcap_type : forall G x T U S,
    inert G ->
    G ⊢! x ∶ (typ_objref T S) ⪼ U ->
    refcap_type S.
Proof.
  introv Hi Pf.
  eapply pf_refcap_inert_type in Pf.
  destructc Pf as vk Hin.
  eapply refcap_inert_refcap_typ.
  eauto.
  assumption.
Qed.

(** If [G(x) = μ(x: T) /\ S], then [x]'s precise type can be only [μ(x: T) /\ S], [μ(x: T)]
     or a record type, or a refcap type. *)
Lemma pf_bnd_same_or_rcd : forall G x T S U,
    inert G ->
    G ⊢! x ∶ typ_objref T S ⪼ U ->
    U = typ_objref T S \/ U = typ_bnd T \/ record_type U \/ refcap_type U.
Proof.
  introv Hi Pf.
  dependent induction Pf; try solve [left*]; right.
  all:
    specialize (IHPf T S Hi eq_refl);
    branches; try inversions H; try inversions H0; auto.
  - (* typ_bnd *)
    branch 2 of 3. apply open_record_type.
      eapply pf_rcd_T; eauto.
  - branch 2 of 3.
    exists ls. assumption.
  - branch 3 of 3.
    exists ls. assumption.
  - branch 3 of 3.
    eapply pf_refcap_type; eauto.
  - branch 2 of 3.
    eexists. apply* rt_one.
  - (* refcap one *)
    branch 3 of 3.
    eexists.
    eapply refcap_typ_one; eauto.
Qed.

(** If [x]'s precise type is [μ(x: U) /\ S], then [G(x) = μ(x: U) /\ S] *)
Lemma pf_inert_objref_U: forall G x T U S,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_objref U S ->
    T = typ_objref U S.
Proof.
  introv Hi Pf.
  destructc (pf_inert_T Hi Pf) as vk HinT.
  destruct HinT.
  - apply precise_flow_fun_inv in Pf. congruence.
  - destructd (pf_bnd_same_or_rcd Hi Pf) as Heq Hbnd Hrcd Hrcp.
    + auto.
    + inversion Hbnd.
    + inversion Hrcd. inversion H2. inversion H5.
    + inversion Hrcp. inversion H2. inversion H5.
Qed.

Lemma pf_inert_bnd_U: forall G x T U,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_bnd U ->
    exists S, T = typ_objref U S.
Proof.
  introv Hi Pf.
  destructc (pf_inert_T Hi Pf) as vk HinT.
  destruct HinT.
  - apply precise_flow_fun_inv in Pf. congruence.
  - destructd (pf_bnd_same_or_rcd Hi Pf) as Heq Hbnd Hrcd Hrcp.
    + inversion Heq.
    + inversions Hbnd. exists S. auto.
    + inversion Hrcd. inversion H2.
    + inversion Hrcp. inversion H2.
Qed.

(** If [x]'s precise type is a field or type declaration, then [G(x)]
    has the form [μ(x: U) /\ S]. *)
Lemma pf_inert_rcd_U: forall G x T D,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_rcd D ->
    exists U S, T = typ_objref U S.
Proof.
  introv Hi Pf.
  destructc (pf_inert_T Hi Pf) as vk HinT.
  destruct HinT.
  - apply precise_flow_fun_inv in Pf. congruence.
  - exists* T.
Qed.


(** If [x]'s precise type is a record type, then [G(x)] has the form [μ(x: U) /\ S] *)
Lemma pf_inert_rcd_typ_U: forall G x T Ds,
    inert G ->
    G ⊢! x ∶ T ⪼ Ds ->
    record_type Ds ->
    exists U S, T = typ_objref U S.
Proof.
  introv Hi Pf Hr.
  destructc (pf_inert_T Hi Pf) as vk HinT.
  destruct HinT.
  - apply precise_flow_fun_inv in Pf. subst.
    inversion Hr. inversion H0.
  - exists* T.
Qed.

(* A record precise type is either a record or reference capabilities *)
Lemma pf_inert_rcd_or_refcap: forall G x T D,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_rcd D ->
    (record_type (typ_rcd D) \/ refcap_type (typ_rcd D)).
Proof.
  introv Hi Pf.
  destructc (pf_inert_rcd_U Hi Pf) as T1 S1 ?H; subst.
  destructd (pf_bnd_same_or_rcd Hi Pf) as H2 H2 H2 H2; try inversions H2.
  - left. exists x0. eauto.
  - right. exists x0. eauto.
Qed.

(* Solves inversions of precise typing *)
Ltac pf_inert_inv_solver Hi Pf HiU := (
  (* Introduce inert G and precise typing hyps *)
  introv Hi Pf;
  (* assert that U is inert, branch to fun/obj *)
  destructc (pf_inert_T Hi Pf) as vk HiU; destruct HiU;
  [ apply precise_flow_fun_inv in Pf; congruence |
    (* branch by target of object precise typing *)
    destructd (pf_bnd_same_or_rcd Hi Pf) as H2 H2 H2 H2;
    (* invert equality of function type with the part of object type, then contradict record_typ *)
    inversion H2; inversion H3]
).


(** The following two lemmas express that if [x]'s precise type is a function type,
    then [G(x)] is the same function type. *)
Lemma pf_inert_fun_U : forall x G S T U,
    inert G ->
    G ⊢! x ∶ U ⪼ typ_fun S T ->
    U = typ_fun S T.
Proof.
  pf_inert_inv_solver Hi Pf HiU.
Qed.

(** See [pf_inert_fun_U]. *)
Lemma inert_precise_fun_inv : forall x G S T U,
    inert G ->
    G ⊢! x ∶ U ⪼ typ_fun S T ->
    ctx_binds x (typ_fun S T) G.
Proof.
  introv Hi Htyp. lets H: (pf_inert_fun_U Hi Htyp). subst.
  apply* pf_binds.
Qed.

Lemma pf_fun_inv_mode : forall x G S T U,
    inert G ->
    G ⊢! x ∶ U ⪼ typ_fun S T ->
    mode_has_fun' typing_mode = true.
Proof.
  introv Hi Htyp.
  forwards Hf: pf_inert_fun_U Hi Htyp. subst.
  forwards Hif: pf_inert_T; eauto.
  inversions Hif.
  inversions H.
  auto.
Qed.


(** In an inert context, the precise type of a variable
    cannot be bottom. *)
Lemma pf_bot_false : forall G x T,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_bot ->
    False.
Proof.
  pf_inert_inv_solver Hi Pf HiU.
Qed.

Lemma pf_or_false : forall G x T T1 T2,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_or T1 T2 ->
    False.
Proof.
  pf_inert_inv_solver Hi Pf HiU.
Qed.

Lemma pf_nocap_false : forall G x T,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_nocap ->
    False.
Proof.
  pf_inert_inv_solver Hi Pf HiU.
Qed.

(** In an inert context, the precise type of
    a variable cannot be type selection. *)
Lemma pf_psel_false : forall G T x y A z,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_sel y A z ->
    False.
Proof.
  pf_inert_inv_solver Hi Pf HiU.
Qed.

Lemma pf_and_rcd_or_refcap : forall G x T U1 U2,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_and U1 U2 ->
    record_type (typ_and U1 U2) \/ refcap_type (typ_and U1 U2) \/
    (exists S vk,
      U1 = (typ_bnd S) /\ record_type S /\ refcap_inert_type U2 vk).
Proof.
  introv Hi Pf.
  destructc (pf_inert_T Hi Pf) as vk HinT.
  dependent induction Pf.
  - branch 3 of 3.
    inversions HinT.
    exists T vk; auto.
  - branch 1 of 3.
    apply pf_inert_bnd_U in Pf; auto. destruct Pf; subst.
    inversions HinT.
    apply open_record_type with (x:=x0) in H1.
    rewrite x in H1. auto.
  - edestruct (IHPf) as [? | [? | ?]]; eauto.
    + branch 1 of 3.
      inversions H; inversions H0. unfold record_type. exists ls. auto.
    + branch 2 of 3.
      inversions H; inversions H0. unfold refcap_type. exists ls. auto.
    + destructc H as ? ? ? ?. inversions H.
  - edestruct (IHPf) as [? | [? | ?]]; eauto.
    + branch 1 of 3. inversions H; inversions H0.
    + branch 2 of 3. inversions H. inversions H0.
    + branch 2 of 3. destructc H as ? ? ? ?; subst. destruct H0.
      eauto using refcap_inert_refcap_typ.
Qed.

(* For dealing with distributivity in invertible typing,
   we state that a precise type cannot be an intersection of two same type memebers *)
Lemma pf_and_typ_same_false G x T A t11 t12 t21 t22:
  inert G ->
  G ⊢! x ∶ T ⪼ typ_and (typ_rcd (dec_typ A t11 t12)) (typ_rcd (dec_typ A t21 t22)) ->
  False.
Proof.
  introv Hi Pf.
  pose proof (pf_inert_T Hi Pf).
  apply pf_and_rcd_or_refcap in Pf; auto.
  branches.
  - inversions H0.
    inversions H1.
    inversions H3.
    compute in H7.
    apply H7. rewrite in_singleton. auto.
  - inversions H0.
    inversions H1.
    inversions H3.
    compute in H7.
    apply H7. rewrite in_singleton. auto.
  - destructc H0 as S vk Heq Hrcd Hrcp.
    apply refcap_inert_refcap_typ in Hrcp.
    destruct Hrcp. inversions H0. inversions H2.
    destruct Hrcd. inversions H0; inversions Heq.
Qed.

(** If [G(x) = μ(T) /\ S], and [G ⊢! x : ... /\ D /\ ...], then [T^x = ... /\ D /\ ...] or [S = ... /\ D /\ ...]. *)
Lemma pf_record_sub : forall x G T S T' D,
    inert G ->
    G ⊢! x ∶ typ_objref T S ⪼ T' ->
    record_has T' D ->
    (record_has (open x T) D /\ record_dec D) \/ (record_has S D /\ refcap_dec D).
Proof.
  introv Hi Pf Hr.
  pose proof (pf_rcd_T Hi Pf).
  pose proof (pf_refcap_type Hi Pf).
  dependent induction Pf; auto.
  - inversions Hr.
    + left. inversions H5.
    + right. split; auto.
      destruct H1. eapply refcap_typ_dec; eauto.
  - apply (pf_inert_bnd_U Hi) in Pf.
    destruct Pf. inversions H. inversions H1.
    left; split; auto.
    eapply record_typ_dec. 2: apply Hr. eapply open_record_typ; eauto.
Qed.

(** If [G(x) = μ(T) /\ S] and [G ⊢! x : D], where [D] is a field or type declaration,
    then [T^x = ... /\ D /\ ...] or [S^x = ... /\ D /\ ...]. *)
Lemma precise_flow_has: forall T S G x D,
    inert G ->
    G ⊢! x ∶ typ_objref T S ⪼ typ_rcd D ->
    (record_has (open x T) D /\ record_dec D) \/ (record_has S D /\ refcap_dec D).
Proof.
  introv Hi Pf. eapply pf_record_sub; eauto.
Qed.

Lemma precise_flow_record_has: forall T S G x D,
    inert G ->
    G ⊢! x ∶ typ_objref T S ⪼ typ_rcd D ->
    record_dec D ->
    record_has (open x T) D.
Proof.
  introv Hi Pf.
  edestruct precise_flow_has; eauto.
  - destruct H. auto.
  - destruct H. intro. record_refcap_contradict.
Qed.

Lemma precise_flow_refcap_has: forall T S G x D,
    inert G ->
    G ⊢! x ∶ typ_objref T S ⪼ typ_rcd D ->
    refcap_dec D ->
    record_has S D.
Proof.
  introv Hi Pf.
  edestruct precise_flow_has; eauto.
  - destruct H. intro. record_refcap_contradict.
  - destruct H. auto.
Qed.

(** The type to which a variable is bound in an environment is unique. *)
Lemma x_bound_unique: forall G x T1 T2 U1 U2,
    G ⊢! x ∶ T1 ⪼ U1 ->
    G ⊢! x ∶ T2 ⪼ U2 ->
    T1 = T2.
Proof.
  introv Pf1 Pf2.
  apply pf_binds in Pf1.
  apply pf_binds in Pf2.
  apply (ctx_binds_functional Pf1 Pf2).
Qed.

(** If
    - [G ⊢! x : μ(T) /\ S ⪼ {A: T11..T12}]
    - [G ⊢! x : μ(T) /\ S ⪼ {A: T21..T22}]
    then [T11 = T21] and [T12 = T22]. *)
Lemma pf_record_unique_tight_bounds_rec : forall G x T1 S1 T2 S2 A T11 T12 T21 T22,
    inert G ->
    G ⊢! x ∶ typ_objref T1 S1 ⪼ typ_rcd (dec_typ A T11 T12) ->
    G ⊢! x ∶ typ_objref T2 S2 ⪼ typ_rcd (dec_typ A T21 T22) ->
    T11 = T21 /\ T12 = T22.
Proof.
  introv Hi Pf1 Pf2.
  assert (typ_objref T1 S1 = typ_objref T2 S2) as HTeq by (eapply x_bound_unique; eauto).
  rewrite HTeq in *.
  destruct (precise_flow_has Hi Pf1) as [[H1 Hr1] | [H1 Hr1]];
  destruct (precise_flow_has Hi Pf2) as [[H2 Hr2] | [H2 Hr2]].
  - (* record *)
    eapply unique_rcd_typ; eauto.
    apply open_record_type.
    eapply pf_rcd_T; eauto.
  - inversions HTeq. inversions Hr1; inversions Hr2.
  - inversions HTeq. inversions Hr1; inversions Hr2.
  - (* refcap *) eapply unique_refcap_typ; eauto.
    eapply pf_refcap_type; eauto.
Qed.

(** If
    - [G ⊢! x : T ⪼ {A: T11..T12}]
    - [G ⊢! x : T ⪼ {A: T21..T22}]
    then [T11 = T21] and [T12 = T22]. *)
Lemma pf_inert_unique_tight_bounds : forall G x T1 T2 T11 T12 T21 T22 A,
    inert G ->
    G ⊢! x ∶ T1 ⪼ typ_rcd (dec_typ A T11 T12) ->
    G ⊢! x ∶ T2 ⪼ typ_rcd (dec_typ A T21 T22) ->
    T11 = T21 /\ T12 = T22.
Proof.
  introv Hi Pf1 Pf2.
  assert (T1 = T2) as HTeq by (eapply x_bound_unique; eauto).
  rewrite HTeq in *.
  pose proof (pf_inert_rcd_U Hi Pf1) as [?U [?S ?H]]; subst.
  eapply pf_record_unique_tight_bounds_rec; eauto.
Qed.

(** If [G ⊢! x : {A: S..U}] then [S = U] or [S = bot]. *)
Lemma pf_dec_typ_inv : forall G x T A S U,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_rcd (dec_typ A S U) ->
    S = U \/ S = typ_bot.
Proof.
  introv Hi Pf. destruct (pf_inert_rcd_U Hi Pf) as [V1 [V2 H]]. subst.
  destructd (pf_bnd_same_or_rcd Hi Pf) as ? ? ? ?; try congruence.
  - destruct H as [?ls ?H]. inversions H. inversions* H1.
  - destruct H as [?ls ?H]. inversions H. inversions* H1.
Qed.

Lemma pf_dec_fld_inv : forall G x T a S U,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_rcd (dec_fld a S U) ->
    S = U.
Proof.
  introv Hi Pf. destruct (pf_inert_rcd_U Hi Pf) as [V1 [V2 H]]. subst.
  destructd (pf_bnd_same_or_rcd Hi Pf) as ? ? ? ?; try congruence.
  - destruct H as [?ls ?H]. inversions H. inversions* H1.
  - destruct H as [?ls ?H]. inversions H. inversions* H1.
Qed.

Lemma precise_bnd_indep G x T U:
  inert G ->
  G ⊢! x ∶ T ⪼ typ_bnd U ->
  typ_indep_b 0 U.
Proof.
  introv Hin Hpf.
  destruct (pf_inert_bnd_U Hin Hpf).
  destructc (pf_inert_T Hin Hpf) as vk H0.
  subst.
  inversion H0; subst.
  auto.
Qed.

Hint Resolve precise_bnd_indep: core.

Lemma pf_dec_typ_notin_fv_u G x T A S U y:

    y §# G ->
    y \notin ctx_fv G ->
    G ⊢! x ∶ T ⪼ typ_rcd (dec_typ A S U) ->
    y \notin fv U.
Proof.
  introv Hnd Hnf Pf.
  forwards: pf_concat_notin_fv.
  rewrite ctx_concat_empty_r.
  all: eauto.
  rewrite fv_rcd_typ in H.
  rewrite notin_union in H.
  destruct H.
  auto.
Qed.
Lemma pf_dec_typ_notin_fv_l G x T A S U y:
    y §# G ->
    y \notin ctx_fv G ->
    G ⊢! x ∶ T ⪼ typ_rcd (dec_typ A S U) ->
    y \notin fv S.
Proof.
  introv Hi Hnd Hnf Pf.
  forwards: pf_concat_notin_fv.
  rewrite ctx_concat_empty_r.
  all: eauto.
  rewrite fv_rcd_typ in H.
  rewrite notin_union in H.
  destruct H.
  auto.
Qed.


(** Precise typing implies general typing. *)
(** - for variables *)
Lemma precise_to_general: forall G e x T U,
    inert G ->
    G ⊢! x ∶ T ⪼ U ->
    G⁏e ⊢ x ∶ U.
Proof.
  introv Hin Hpf.
  induction Hpf; intros; subst; eauto.
Qed.


(* In rodot, every variable has mutability *)
Lemma pf_mutability: forall G x T,
    mode_has_mutability typing_mode ->
    inert G ->
    ctx_binds x T G ->
    exists U2,
    G ⊢! x ∶ T ⪼ typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot U2).
Proof.
  introv Hm Hin Hbi.
  destructc (inert_binds_inert Hin Hbi) as vk H.

  inversions H.
  - (* We do not allow functions in rodot *)
    typing_mode_contradiction typing_mode.
  - unfold refcap_inert_type in H1.
    destruct (mode_refcaps' typing_mode) eqn: Heq.
    typing_mode_contradiction typing_mode.
    inversions H1.
    + exists typ_bot. eauto.
    + exists Tbnd. eauto.
Qed.

End WithTypingMode.

Hint Constructors precise_flow : core.
Hint Resolve precise_bnd_indep: core.
Notation "G '⊢!' x '∶' T '⪼' U" := (precise_flow x G T U) (at level 40, x at level 59).


Section WithTypingMode.

Context (typing_mode : Mode).

(* open_proj x T U means one can reach U from T using only open x and
   projection of intersections *)
Inductive open_proj (x : var) (T : typ) : typ -> Prop :=
| open_proj_refl : open_proj x T T
| open_proj_open : forall U, open_proj x T (typ_bnd U) ->
                        open_proj x T (open x U)
| open_proj1 : forall U1 U2, open_proj x T (typ_and U1 U2) ->
                        open_proj x T U1
| open_proj2 : forall U1 U2, open_proj x T (typ_and U1 U2) ->
                        open_proj x T U2.
Hint Constructors open_proj : core.

(* If G ⊢! w ∶ μ.T ∧ {M:⊥..S} ⪼ U and U ≠ {M:⊥..S} then U is obtained
   from μ.T from arbitrary sequences of opens and projections *)
Lemma pf_dec_typ_path G w T TBnd U :
  G  ⊢! w ∶ (typ_objref T (typ_mut_marker TBnd)) ⪼ U ->
  U <> typ_objref T (typ_mut_marker TBnd) ->
  U <> typ_mut_marker TBnd ->
  open_proj w (typ_bnd T) U.
Proof.
  intros Hpf Hneq1 Hneq2.
  remember (typ_objref T (typ_mut_marker TBnd)) as S.
  induction Hpf; try tauto;
    rename T0 into S, x into w.
  - eapply open_proj_open, IHHpf; try discriminate; try easy. clear IHHpf.
    hnf. intros. subst. discriminate.
  - destruct (classicT ((typ_bnd T) = U1)) as [HeqT | HneqT];
      try solve[subst; eauto].
    eapply open_proj1, IHHpf; try discriminate; try easy. clear IHHpf.
    hnf. intros. subst. inversions H. tauto.
  - eapply open_proj2, IHHpf; try discriminate; try easy. clear IHHpf.
    hnf. intros. subst. inversions H. tauto.
Qed.

(*
   Combined with the previous lemma, it transports a precise flow of
   v₁ starting from μ.T ∧ {M:⊥.._} to a precise flow starting of v₂
   starting from μ.T ∧ {M:⊥.._} as long as v₁ ∉ fv T. *)
Lemma open_proj_pf_subst G v1 v2 T TBnd U :
  ctx_binds v2 (typ_objref T (typ_mut_marker TBnd)) G ->
  v1 \notin fv (typ_bnd T) ->
  open_proj v1 (typ_bnd T) U ->
  G ⊢! v2 ∶ (typ_objref T (typ_mut_marker TBnd)) ⪼ (U /[ v1 -> v2]).
Proof.
  intros Hbi Hfrv1 Hop. remember (typ_bnd T) as S. induction Hop; subst;
    try solve[simpl in *; eauto].
  - rewrite subst_fresh; eauto.
  - rewrite subst_open_commut. destruct (classicT (v1 = v1)); try tauto.
    simpl in *. apply pf_open. eauto.
Qed.

Lemma pf_dec_met_functional G x S m TC1 TZ1 TR1 TC2 TZ2 TR2 :
  inert G ->
  G ⊢! x ∶ S ⪼ typ_rcd (dec_met m TC1 TZ1 TR1) ->
  G ⊢! x ∶ S ⪼ typ_rcd (dec_met m TC2 TZ2 TR2) ->
  dec_met m TC1 TZ1 TR1 = dec_met m TC2 TZ2 TR2.
Proof.
  introv Hin Hpf1 Hpf2.
  apply pf_binds in Hpf1 as HbinS.
  eapply pf_inert_rcd_U in Hpf1 as HeqS; eauto.
  destruct HeqS as [U [R HeqS]]; subst.
  eapply inert_binds_inert in HbinS as [vk Hintyp]; eauto.
  inversions Hintyp. rename H1 into Hrcd, H2 into Hrc, H4 into Hindep.

  eapply pf_record_sub with (D := dec_met m TC1 TZ1 TR1) in Hpf1; eauto.
  destruct Hpf1 as [[Hhas1 _] | [_ Hcontra]].
  2: inversions Hcontra.
  eapply pf_record_sub with (D := dec_met m TC2 TZ2 TR2) in Hpf2; eauto.
  destruct Hpf2 as [[Hhas2 _] | [_ Hcontra]].
  2: inversions Hcontra.

  eapply open_vars_record_type with (n := 0) (ys := cons x nil) in Hrcd;
    simpl in Hrcd.

  eauto using record_has_functional.
Qed.

End WithTypingMode.
