(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode GeneralTyping Indep Equivalence.

(** * Record Types *)

(* Part of Figure 8 in "Reference mutability for DOT" *)
(** A record declaration is either
    a type declaration with equal bounds,
    a type declaration with a bottom lower bound,
    a field declaration,
    or a method declaration.*)
Inductive record_dec : dec -> Prop :=
| rd_typ : forall A T, record_dec (dec_typ (typ_label_normal A) T T)
| rd_typb : forall A T, record_dec (dec_typ (typ_label_normal A) typ_bot T)
| rd_fld : forall a T, record_dec (dec_fld a T T)
| rd_met : forall m TC TZ TR, record_dec (dec_met m TC TZ TR)
.

(** Given a record declaration, a [record_typ] keeps track of the declaration's
    field member labels (i.e. names of fields) and type member labels
    (i.e. names of abstract type members). [record_typ] also requires that the
    labels are distinct.  *)
Inductive record_typ : typ -> fset label -> Prop :=
| rt_one : forall D l,
  record_dec D ->
  l = label_of D ->
  record_typ (typ_rcd D) \{l}
| rt_cons: forall T ls D l,
  record_typ T ls ->
  record_dec D ->
  l = label_of D ->
  l \notin ls ->
  record_typ (typ_and T (typ_rcd D)) (union ls \{l}).

(* Part of Figure 8 in "Reference mutability for DOT" *)
(** A [record_type] is a [record_typ] with an unspecified set of labels. The meaning
    of [record_type] is an intersection of type/field declarations with distinct labels. *)
Definition record_type T := exists ls, record_typ T ls.

Inductive refcap_dec: dec -> Prop :=
  | refcap_dec_typb lt TU:
    refcap_dec (dec_typ (typ_label_refcap lt) typ_bot TU).

(* Set of reference capabilities, for types in an inert context *)
Inductive refcap_typ: typ -> fset label -> Prop :=
  | refcap_typ_top:
      refcap_typ typ_top \{}
  | refcap_typ_one D l:
      refcap_dec D ->
      l = label_of D ->
      refcap_typ (typ_rcd D) \{l}
  | refcap_typ_cons T ls D l:
      refcap_typ T ls ->
      refcap_dec D ->
      l = label_of D ->
      l \notin ls ->
      refcap_typ (typ_and T (typ_rcd D)) (union ls \{l}).

(* Type describing reference capabilities *)
Definition refcap_type T := exists ls, refcap_typ T ls.

(** Given a type [T = D1 /\ D2 /\ ... /\ Dn] and member declaration [D], [record_has T D] tells whether
    [D] is contained in the intersection of [Di]'s. *)
Inductive record_has: typ -> dec -> Prop :=
| rh_one : forall D,
    record_has (typ_rcd D) D
| rh_andl : forall T U D,
    record_has T D ->
    record_has (typ_and T U) D
| rh_andr : forall T U D,
    record_has U D ->
    record_has (typ_and T U) D.

Hint Constructors record_typ record_has : core.

(* A declaration cannot be record and refcap at the same time  *)
Ltac record_refcap_contradict := (
  match goal with
  | Hrcd: record_dec ?D, Hrcp: refcap_dec ?D  |- _ => inversions Hrcd; inversion Hrcp
  end).

Lemma record_has_no_label D ls T :
  label_of D \notin ls ->
  record_typ T ls ->
  record_has T D ->
  False.
Proof.
  introv Hnin Hrcd Hhas. induction Hrcd.
  - inversions Hhas. eauto using notin_same.
  - apply notin_union in Hnin as [Hnin1 Hnin2].
    inversions Hhas; eauto. inversions H5. eauto using notin_same.
Qed.

Lemma record_has_functional U D1 D2 :
  record_type U ->
  record_has U D1 ->
  record_has U D2 ->
  label_of D1 = label_of D2 ->
  D1 = D2.
Proof.
  introv Hrcd Hhas1 Hhas2 Hlbl.
  reverts D1 D2. destruct Hrcd as [Ds Hrcd]. induction Hrcd; intros.
  - inversions Hhas1. inversions Hhas2. auto.
  - inversions Hhas1; inversions Hhas2; eauto.
    + inversions H4. rewrite <- Hlbl in H1.
      false. eauto using record_has_no_label.
    + inversions H5. rewrite Hlbl in H1.
      false. eauto using record_has_no_label.
    + inversions H5. inversions H4. auto.
Qed.

Section WithTypingMode.
Context (typing_mode: Mode).
(** ** Lemmas About Record Types *)

(** [labels(D) = labels(D^x)] *)
Lemma open_dec_preserves_label: forall (D : dec) x i,
  label_of (D /[#i -> x]) = label_of D.
Proof.
  intros. induction D; reflexivity.
Qed.

Lemma open_rec_record_dec: forall D n x,
  record_dec D -> record_dec (D /[#n -> x]).
Proof.
  intros. inversion H; constructor.
Qed.

(** [record_dec D]   #<br>#
    [――――――――――――――] #<br>#
    [record_dec D^x] *)
Lemma open_record_dec: forall D x,
  record_dec D -> record_dec (D /[-> x]).
Proof.
  auto using open_rec_record_dec.
Qed.

Lemma open_rec_record_typ: forall T ls n x,
  record_typ T ls -> record_typ (T /[#n -> x]) ls.
Proof.
  introv H.
  induction H; simpl;
    [apply rt_one | apply rt_cons];
    try apply open_rec_record_dec;
    try rewrite open_dec_preserves_label;
    auto.
Qed.

(** [record_typ T]   #<br>#
    [――――――――――――――] #<br>#
    [record_typ T^x] *)
Lemma open_record_typ: forall T x ls,
  record_typ T ls -> record_typ (T /[-> x]) ls.
Proof.
  auto using open_rec_record_typ.
Qed.

Lemma open_rec_vars_record_typ: forall ys n T ls,
  record_typ T ls -> record_typ (T /[#n --> ys]) ls.
Proof.
  induction ys; simpls;
    intros; auto using open_rec_record_typ.
Qed.

(** [record_typ T]   #<br>#
    [――――――――――――――] #<br>#
    [record_typ T^x] *)
Lemma open_record_type: forall T x,
  record_type T -> record_type (T /[-> x]).
Proof.
  introv [ls H]. exists ls. apply open_record_typ.
  assumption.
Qed.

Lemma open_vars_record_type: forall ys n T,
  record_type T -> record_type (open_rec_vars n ys T).
Proof.
  unfold record_type.
  introv [ls H]. exists ls.
  eauto using open_rec_vars_record_typ.
Qed.

(** Opening does not affect the labels of a [record_typ]. *)
Lemma opening_preserves_labels : forall z n T ls ls',
    record_typ T ls ->
    record_typ (T /[#n -> z]) ls' ->
    ls = ls'.
Proof.
  introv Ht Hopen. gen ls'.
  dependent induction Ht; intros.
  - inversions Hopen. rewrite* open_dec_preserves_label.
  - inversions Hopen. rewrite* open_dec_preserves_label.
    specialize (IHHt ls0 H4). rewrite* IHHt.
Qed.

(** Opening does not affect the labels of a [record_type]. *)
Lemma record_dec_open_rec : forall D n z,
  z \notin fv D ->
  record_dec (D /[#n -> z]) ->
  record_dec D.
Proof.
  intros D n z Hz H.
  inversions H;
    destruct D; sympls; destruct_notin.
    all:  match goal with
      | [ H : _ = _ |- _ ] => inversions H
      end.
    all: try match goal with
      | [ H1: _ \notin ?f ?T1, H2: _ \notin ?f ?T2, H3 : _ = _ |- _ ] =>
        let H := fresh in
        pose proof (open_fresh_injective T1 T2) as H; apply H in H3
      | [ H2: typ_bot = (open_rec ?n ?z ?T) |- _ ] => (* For the typb case *)
        assert (typ_bot = (open_rec n z typ_bot)) as Htbo by auto;
        rewrite Htbo in H2;
        let H := fresh in
        pose proof (open_fresh_injective typ_bot t0) as H; apply H in H2
      end.
    all: subst; eauto.
    all: constructor.
Qed.

Lemma record_type_open_rec : forall n z T,
    z \notin fv T ->
    record_type (T /[#n -> z]) ->
    record_type T.
Proof.
  introv Hz H. destruct H as [ls H].
  exists ls. remember (open_rec n z T) as T' eqn:Heq. gen T. induction H; intros.
  - destruct T; inversion Heq; subst.
    rewrite open_dec_preserves_label.
    constructor; eauto using record_dec_open_rec.
  - destruct T0; inversion Heq; subst. rename T0_1 into T1. rename T0_2 into T2.
    destruct T2; inversion Heq; subst.
    rewrite open_dec_preserves_label in *; sympls; destruct_notin.
    apply rt_cons; eauto using record_dec_open_rec.
Qed.

Lemma record_type_open_rec_vars : forall ys n T,
    fresh (fv T) (length ys) ys ->
    record_type (T /[#n --> ys]) ->
    record_type T.
Proof.
  induction ys; eauto using record_type_open_rec.
  intros. simpl in H; destruct H.
  simpl in H0. apply (record_type_open_rec n T H).
  eapply IHys; eauto.
  pose proof (fv_open_cases a T n) as H';
    destruct H' as [H' | H']; rewrite H'; auto.
Qed.

(** If [T] is a record type with labels [ls], and [T = ... /\ D /\ ...],
    then [label(D) isin ls]. *)
Lemma record_typ_has_label_in: forall T D ls,
  record_typ T ls ->
  record_has T D ->
  label_of D \in ls.
Proof.
  introv Htyp Has. generalize dependent D. induction Htyp; intros.
  - inversion Has. subst. apply in_singleton_self.
  - inversion Has; subst; rewrite in_union.
    + left. apply* IHHtyp.
    + right. inversions H5. apply in_singleton_self.
Qed.

Lemma refcap_typ_has_label_in: forall T D ls,
  refcap_typ T ls ->
  record_has T D ->
  label_of D \in ls.
Proof.
  introv Htyp Has. generalize dependent D. induction Htyp; intros.
  - inversion Has.
  - inversion Has. subst. apply in_singleton_self.
  - inversion Has; subst; rewrite in_union.
    + left. apply* IHHtyp.
    + right. inversions H5. apply in_singleton_self.
Qed.

Lemma record_has_fv : forall T D,
    record_has T D ->
    fv D \c fv T.
Proof. induction 1; simpl in *; introv; try(rewrite in_union); eauto. Qed.

Lemma record_typ_dec: forall T D ls,
  record_typ T ls ->
  record_has T D ->
  record_dec D.
Proof.
  introv Htyp Has. generalize dependent D. induction Htyp; intros.
  - inversion Has. subst; auto.
  - inversion Has; subst; auto.
    inversions H5; auto.
Qed.

Lemma refcap_typ_dec: forall T D ls,
  refcap_typ T ls ->
  record_has T D ->
  refcap_dec D.
Proof.
  introv Htyp Has. generalize dependent D. induction Htyp; intros.
  - inversion Has.
  - inversion Has. subst; auto.
  - inversion Has; subst; auto.
    inversions H5; auto.
Qed.

Lemma refcap_type_top:
  refcap_type typ_top.
Proof.
  eexists. eapply refcap_typ_top.
Qed.


(* A type cannot be record and refcap at the same time *)
Lemma record_refcap_false T:
  record_type T ->
  refcap_type T ->
  False.
Proof.
  intros.
  inversions H; inversions H0.
  revert dependent x0.
  induction H1.
  - intros. inversions H1. record_refcap_contradict.
  - intros. inversions H3. eauto.
Qed.

(** [T = ... /\ {A: T1..T1} /\ ...] #<br>#
    [T = ... /\ {A: T2..T2} /\ ...] #<br>#
    [―――――――――――――――――――――――――――] #<br>#
    [T1 = T2] *)
Lemma unique_rcd_typ: forall T A T11 T12 T21 T22,
  record_type T ->
  record_has T (dec_typ A T11 T12) ->
  record_has T (dec_typ A T21 T22) ->
  T11 = T21 /\ T12 = T22.
Proof.
  introv Htype Has1 Has2.

  destruct Htype as [ls Htyp]. induction Htyp; intros; inversion Has1; inversion Has2; subst.
  - inversion* H3.
  - auto.
  - apply record_typ_has_label_in with (D:=dec_typ A T11 T12) in Htyp.
    + inversions H9. false* H1.
    + assumption.
  - apply record_typ_has_label_in with (D:=dec_typ A T21 T22) in Htyp.
    + inversions H5. false* H1.
    + assumption.
  - inversions H5. inversions* H9.
Qed.

Lemma unique_rcd_trm: forall T a T1 T2,
  record_type T ->
  record_has T (dec_fld a T1 T1) ->
  record_has T (dec_fld a T2 T2) ->
  T1 = T2.
Proof.
  introv [ls Htyp] Has1 Has2. gen T1 T2 a.
  induction Htyp; intros; inversion Has1; inversion Has2; subst.
  - inversion* H3.
  - inversion* H5.
  - apply record_typ_has_label_in with (D:=dec_fld a T1 T1) in Htyp.
    + inversions H9. false* H1.
    + assumption.
  - apply record_typ_has_label_in with (D:=dec_fld a T2 T2) in Htyp.
    + inversions H5. false* H1.
    + assumption.
  - inversions H5. inversions* H9.
Qed.

Lemma unique_refcap_typ: forall T A T11 T12 T21 T22,
  refcap_type T ->
  record_has T (dec_typ A T11 T12) ->
  record_has T (dec_typ A T21 T22) ->
  T11 = T21 /\ T12 = T22.
Proof.
  introv Htype Has1 Has2.

  destruct Htype as [ls Htyp]. induction Htyp; intros; inversion Has1; inversion Has2; subst.
  - inversion* H3.
  - auto.
  - apply refcap_typ_has_label_in with (D:=dec_typ A T11 T12) in Htyp.
    + inversions H9. false* H1.
    + assumption.
  - apply refcap_typ_has_label_in with (D:=dec_typ A T21 T22) in Htyp.
    + inversions H5. false* H1.
    + assumption.
  - inversions H5. inversions* H9.
Qed.

Lemma record_has_open ds d x:
  record_has ds d ->
  record_has (open x ds) (open x d).
Proof.
  introv Hhas.
  induction Hhas.
  - simpl. constructor.
  - simpl. eapply rh_andl. auto.
  - simpl. eapply rh_andr. auto.
Qed.

Lemma record_has_subst ds d x y :
  record_has ds d ->
  record_has (ds /[ x -> y]) (d /[ x -> y]).
Proof.
  introv Hhas.
  induction Hhas.
  - simpl. constructor.
  - simpl. eapply rh_andl. auto.
  - simpl. eapply rh_andr. auto.
Qed.

Lemma record_open_has_fld x R a T:
  record_has (open x R) (dec_fld a T T) ->
  exists U1 U2,
  record_has R (dec_fld a U1 U2) /\
  (open x U1 = T) /\ (open x U2 = T).
Proof.
  introv Hhas.
  induction R; simpl in Hhas; inversions Hhas.
  - destruct d; simpl in H; inversions H.
    exists t t0. eauto.
  - forwards U1 U2 HU: IHR1; auto. exists U1 U2. destruct_all. split; eauto.
  - forwards U1 U2 HU: IHR2; auto. exists U1 U2. destruct_all. split; eauto.
Qed.

Lemma record_open_has_typ n x R B T1 T2:
  record_has (open_rec n x R) (dec_typ B T1 T2) ->
  exists U1 U2,
  record_has R (dec_typ B U1 U2) /\
  (open_rec (S n) x U1 = T1) /\ (open_rec (S n) x U2 = T2).
Proof.
  introv Hhas.
  induction R; simpl in Hhas; inversions Hhas.
  - destruct d; simpl in H; inversions H.
    exists t0 t1. eauto.
  - forwards U1 U2 HU: IHR1; auto. exists U1 U2. destruct_all. split; eauto.
  - forwards U1 U2 HU: IHR2; auto. exists U1 U2. destruct_all. split; eauto.
Qed.


Lemma indep_record_has_indep n T A T1 T2:
  typ_indep_b n T ->
  record_has T (dec_typ A T1 T2) ->
  typ_indep_b (S n) T1 /\  typ_indep_b (S n) T2.
Proof.
  introv Hin Hhas.
  remember (dec_typ A T1 T2) as D.
  induction Hhas; subst;
  inversions Hin; try inversions H; eauto.
Qed.

Lemma indep_open_record_has_indep n x T A T1 T2:
  typ_indep_b n T ->
  record_has (T /[-> x]) (dec_typ A T1 T2) ->
  typ_indep_b (S n) T1 /\  typ_indep_b (S n) T2.
Proof.
  introv Hin Hhas.
  forwards: record_open_has_typ; eauto.
  destructc H as U1 U2 HhasU Ho1 Ho2.
  forwards: indep_record_has_indep; eauto.
  destruct_all.
  subst.
  split; eapply typ_indep_open; eauto.
Qed.

Lemma record_has_tyq y w ds a TL TR e:
  record_type ds ->
  typ_indep_b 0 ds ->
  record_has (open y ds) (dec_fld a TL TL) ->
  record_has (open w ds) (dec_fld a TR TR) ->
  typing_equiv e w y ->
  typing_equiv e TR TL.
Proof.
  introv Hrcd Hid HhL HhR Htyq.
  inversions Hrcd.
  induction H.
  - sympls.
    inversions HhL.
    inversions HhR.
    destruct D; sympls; inversions H3; inversions H2.
    inversions_tyq Hid.
    eapply tyq_typ_open_indep_same; auto.
  - sympls.
    inversions HhL;
    inversions HhR.
    + (* In tail *)
      inversions_tyq Hid. auto.
    + inversions H5.
      assert (record_typ (open y T) ls) by (eapply open_record_typ; eauto).
      pose proof (record_typ_has_label_in H1 H6).
      destruct D; sympls; inversions H4.
      contradiction.
    + inversions H6.
      assert (record_typ (open w T) ls) by (eapply open_record_typ; eauto).
      pose proof (record_typ_has_label_in H1 H5).
      destruct D; sympls; inversions H4.
      contradiction.
    + (* In head *)
      inversions H6. inversions H5.
      destruct D; sympls; inversions H4.
      inversions H6.
      inversions_tyq Hid.
      inversions_tyq H8.
      eapply tyq_typ_open_indep_same; auto.
Qed.

Lemma record_has_tyq_met y w ds m TCy TZy TRy TCw TZw TRw e:
  record_type ds ->
  typ_indep_b 0 ds ->
  record_has (open y ds) (dec_met m TCy TZy TRy) ->
  record_has (open w ds) (dec_met m TCw TZw TRw) ->
  typing_equiv e w y ->
  typing_equiv e
    (typ_rcd (dec_met m TCy TZy TRy)) (typ_rcd (dec_met m TCw TZw TRw)).
Proof.
  introv Hrcd Hid HhL HhR Htyq.
  inversions Hrcd.
  induction H.
  - sympls.
    inversions HhL.
    inversions HhR.
    destruct D; sympls; inversions H3; inversions H2.
    inversions_tyq Hid.
    constructor;
      eauto using tyq_typ_open, tyq_typ_open_rec, tyq_typ_open_rec_indep.
  - sympls.
    inversions HhL;
    inversions HhR.
    + (* In tail *)
      inversions_tyq Hid. auto.
    + inversions H5.
      assert (record_typ (open y T) ls) by (eapply open_record_typ; eauto).
      pose proof (record_typ_has_label_in H1 H6).
      destruct D; sympls; inversions H4.
      contradiction.
    + inversions H6.
      assert (record_typ (open w T) ls) by (eapply open_record_typ; eauto).
      pose proof (record_typ_has_label_in H1 H5).
      destruct D; sympls; inversions H4.
      contradiction.
    + (* In head *)
      inversions H6. inversions H5.
      destruct D; sympls; inversions H4.
      inversions H6.
      inversions_tyq Hid.
      inversions_tyq H6.
      constructor;
        eauto using tyq_typ_open, tyq_typ_open_rec, tyq_typ_open_rec_indep.
Qed.

End WithTypingMode.
