(** remove printing ~ *)

(** This module contains lemmas related to tight typing [G ⊢# t: T] *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping.
Require Import PreciseTyping Equivalence EWeakening Indep RecordTypes InertContext.

(** * Tight typing [G |-# t: T] *)

Class TightTyping (A : Set) := tight_typing : ctx -> renv -> A -> typ -> Prop.
Notation "G '⁏' e ⊢# t '∶' T" := (tight_typing G e t T) (at level 40, e at level 39, t at level 59).
Class DecTightTyping (A : Set) := dec_typing : ctx -> renv -> A -> dec -> Prop.
Notation "G '⁏' e ⊢# t '∶d' T" := (dec_typing G e t T) (at level 40, e at level 39, t at level 59).
Reserved Notation "G '⁏' e ⊢# t '<:' T" (at level 40, e at level 39, t at level 59).

(** *** Tight term typing [G ⊢# t: T] *)
(** Tight typing is very similar to general typing, and could be obtained by replacing
    all occurrences of [⊢] with [⊢#], except for the following:
    - in the type selection subtyping rules Sel-<: and <:-Sel ([subtyp_sel1] and [subtyp_sel2]),
      the premise is precise typing of a type declaration with equal bounds;
    - whenever a typing judgement in a premise extends the environment (for example, [subtyp_fun_t]),
      it is typed under general typing [⊢] and not tight typing [⊢#]. *)

Section WithTypingMode.
Context (typing_mode: Mode).

Inductive ty_trm_t : TightTyping var :=

(** [G(x) = T]   #<br>#
    [――――――――――(VT#-Var)] #<br>#
    [G ⊢# x: T]  *)
| ty_var_t : forall G e x T,
    ctx_binds x T G ->
    G⁏e ⊢# x ∶ T

(** [G ⊢# x: T^x]   #<br>#
    [――――――――――――――(VT#-RecI)] #<br>#
    [G ⊢# x: μ(T)] *)
| ty_rec_intro_t : forall G e x T,
    typ_indep_b 0 T ->
    
    (* NOTE: this is different from general typing, because in mode_incap_nocap, we need tight subtyping, but ty_incap_diag_nocap_sup only gives general typing,
       before the equivalence of general and tight typing is established. *)
    (mode_incap' typing_mode = mode_incap_structural -> (G⁏e ⊢ (open x T) ∶icp (open x T))) ->
    (mode_incap' typing_mode = mode_incap_nocap -> (G⁏e ⊢# typ_nocap <: (open x T))) ->
    G⁏e ⊢# x ∶ open x T ->
    G⁏e ⊢# x ∶ typ_bnd T

(** [G ⊢# x: μ(T)] #<br>#
    [――――――――――――――] #<br>#
    [G ⊢# x: T^x]       *)
| ty_rec_elim_t : forall G e x T,
    typ_indep_b 0 T ->
    G⁏e ⊢# x ∶ typ_bnd T ->
    G⁏e ⊢# x ∶ open x T

(** [G ⊢# x: T]      #<br>#
    [G ⊢# x: U]      #<br>#
    [―――――――――――――]   #<br>#
    [G ⊢# x: T /\ U]      *)
| ty_and_intro_t : forall G e x T U,
    G⁏e ⊢# x ∶ T ->
    G⁏e ⊢# x ∶ U ->
    G⁏e ⊢# x ∶ typ_and T U

(** [G ⊢# t: T]    #<br>#
    [G ⊢# T <: U]  #<br>#
    [―――――――――――――] #<br>#
    [G ⊢# t: U]        *)
| ty_sub_t : forall G e x T U,
    G⁏e ⊢# x ∶ T ->
    G⁏e ⊢# T <: U ->
    G⁏e ⊢# x ∶ U

| ty_var_mut_top_t : forall G e x T,
    mode_has_mutability typing_mode ->
    G⁏e ⊢# x ∶ T ->
    G⁏e ⊢# x ∶ (typ_mut_marker typ_top)

(** *** Tight subtyping [G ⊢# T <: U] *)
with subtyp_t : ctx -> renv -> typ -> typ -> Prop :=

(** [――――――――――(ST#-Top)] #<br>#
    [G ⊢# T <: ⊤] *)
| subtyp_top_t: forall G e T,
    G⁏e ⊢# T <: typ_top

(** [――――――――――(ST#-Bot)] #<br>#
    [G ⊢# ⊥ <: T] *)
| subtyp_bot_t: forall G e T,
    G⁏e ⊢# typ_bot <: T

(** [――――――――――(ST#-Refl)] #<br>#
    [G ⊢# T <: T] *)
| subtyp_refl_t: forall G e T,
    G⁏e ⊢# T <: T

(** [G ⊢# S <: T]     #<br>#
    [G ⊢# T <: U]     #<br>#
    [―――――――――――――(ST#-Trans)]    #<br>#
    [G ⊢# S <: U]         *)
| subtyp_trans_t: forall G e S T U,
    G⁏e ⊢# S <: T ->
    G⁏e ⊢# T <: U ->
    G⁏e ⊢# S <: U

(** [――――――――――(ST#-And1)] #<br>#
    [G ⊢# T ∧ U <: T] *)
| subtyp_and11_t: forall G e T U,
    G⁏e ⊢# typ_and T U <: T

(** [――――――――――(ST#-And2)] #<br>#
    [G ⊢# T ∧ U <: U] *)
| subtyp_and12_t: forall G e T U,
    G⁏e ⊢# typ_and T U <: U

(** [G ⊢# S <: T]       #<br>#
    [G ⊢# S <: U]       #<br>#
    [――――――――――――――――(ST#-And)]   #<br>#
    [G ⊢# S <: T ∧ U]       *)
| subtyp_and2_t: forall G e S T U,
    G⁏e ⊢# S <: T ->
    G⁏e ⊢# S <: U ->
    G⁏e ⊢# S <: typ_and T U

(** [――――――――――(ST#-Or1)] #<br>#
    [G ⊢# T <: T ∨ U] *)
| subtyp_or11_t: forall G e T U,
    G⁏e ⊢# T <: typ_or T U

(** [――――――――――(ST#-Or2)] #<br>#
    [G ⊢# U <: T ∨ U] *)
| subtyp_or12_t: forall G e T U,
    G⁏e ⊢# U <: typ_or T U

(** [G ⊢# S <: T]       #<br>#
    [G ⊢# U <: T]       #<br>#
    [――――――――――――――(ST#-Trans)]   #<br>#
    [G ⊢# S ∨ U <: T ]          *)
| subtyp_or2_t: forall G e S T U,
    G⁏e ⊢# S <: T ->
    G⁏e ⊢# U <: T ->
    G⁏e ⊢# typ_or S U <: T


(** [G ⊢# T1 <: T2]           #<br>#
    [G ⊢# U1 <: U2]           #<br>#
    [――――――――――――――――――――(ST#-Fld)] #<br>#
    [G ⊢# {a: T2..U1} <: {a: T1..U2}] *)
| subtyp_fld_t: forall G e a T1 T2 U1 U2,
    G⁏e ⊢# T1 <: T2 ->
    G⁏e ⊢# U1 <: U2 ->
    G⁏e ⊢# typ_rcd (dec_fld a T2 U1) <: typ_rcd (dec_fld a T1 U2)

(** [G ⊢ S2^x <: S1^x]                   #<br>#
    [G ⊢ T1^x <: T2^x]                   #<br>#
    [x fresh]                     #<br>#
    [――――――――――――――――――――――――――――――(ST#-Typ)] #<br>#
    [G ⊢ {B: S1..T1} <: {B: S2..T2}]     *)
| subtyp_typ_t: forall G e A S1 T1 S2 T2,
    (forall x, G⁏e ⊢# (open x S2) <: (open x S1)) ->
    (forall x, G⁏e ⊢# (open x T1) <: (open x T2)) ->
    G⁏e ⊢# typ_rcd (dec_typ A S1 T1) <: typ_rcd (dec_typ A S2 T2)

(** [G ⊢! x: {A: TL..TU}] #<br>#
    [――――――――――――――――(ST#-SelL)] #<br>#
    [G ⊢# TL^r <: x.A(r)]     *)
| subtyp_sel2_t: forall G e x A y TL TU U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
    G⁏e ⊢# (open y TL) <: typ_sel (avar_f x) A (avar_f y)

(** [G ⊢! x: {A: TL..TU}] #<br>#
    [――――――――――――――――(ST#-SelU)] #<br>#
    [G ⊢# x.A(r) <: TU^r]     *)
| subtyp_sel1_t: forall G e x A y TL TU U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
    G⁏e ⊢# typ_sel (avar_f x) A (avar_f y) <: (open y TU)

(** [G ⊢# S2 <: S1]                #<br>#
    [G, x: S2 ⊢ T1^x <: T2^x]       #<br>#
    [x fresh]                       #<br>#
    [――――――――――――――――――――――――]      #<br>#
    [G ⊢# forall(S1)T1 <: forall(S2)T2] #<br>#
    This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_fun_t: forall L G e S1 T1 S2 T2,
    mode_has_fun typing_mode ->
    G⁏e ⊢# S2 <: S1 ->
    (forall x, x \notin L ->
       G § x / v_par ~ S2 ⁏e ⊢ open x T1 <: open x T2) ->
    G⁏e ⊢# typ_fun S1 T1 <: typ_fun S2 T2

(** [G ⊢# TZ2 <: TZ1]                #<br>#
    [G, x: TZ2 ⊢# TC2^x <: TC1^x]     #<br>#
    [G, x: TZ2, y: TC2^x ⊢# TR1^x^y <: TR2^x^y]     #<br>#
    [x,y fresh]                     #<br>#
    [―――――――――――――――――――――――(ST#-Met)]     #<br>#
    [G ⊢# {m(TC1, TZ1): TR1} <: {m(TC2, TZ2): TR2}}]      *)
| subtyp_met_t: forall L G e m TC1 TZ1 TR1 TC2 TZ2 TR2,
    G⁏e ⊢# TZ2 <: TZ1 ->
    (forall x, x \notin L -> (* x is the parameter *)
    G § x / v_par ~ TZ2 ⁏e ⊢ open x TC2 <: open x TC1) ->
    (* If typ_indep_b 0 TC2 then typ_indep_b 0 TC1 in modes with sef *)
    (mode_has_sef typing_mode -> typ_indep_b 0 TC2 -> typ_indep_b 0 TC1) ->
    (forall x, x \notin L ->
     forall y, y <> x -> y \notin L -> (* y is the receiver *)
    G § x / v_par ~ TZ2 § y / v_recv ~ (open x TC2) ⁏e ⊢ open x (open_rec 1 y TR1) <: open x (open_rec 1 y TR2) ) ->
    G⁏e ⊢# typ_rcd (dec_met m TC1 TZ1 TR1) <: typ_rcd (dec_met m TC2 TZ2 TR2)

(** [e ⊢ T1 ≈ T2] #<br>#
    [――――――――――――――――(ST#-Eq)] #<br>#
    [G⁏e ⊢# T1 <: T2)]     *)
| subtyp_tyq_t: forall G e T1 T2,
    e ⊢ T1 ≈ T2 ->
    G⁏e ⊢# T1 <: T2

(** [――――――――――(ST#-TypAnd)] #<br>#
    [G ⊢# {A: S1..T1} ∧ {A: S2..T2} <: {A: S1 ∨ S2 .. T1 ∧ T2}] *)
| subtyp_typ_and_t: forall G e A S1 T1 S2 T2,
    G⁏e ⊢#
     typ_and
       (typ_rcd (dec_typ A S1 T1))
       (typ_rcd (dec_typ A S2 T2))
     <: typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))

(** [――――――――――(ST#-Dist)] #<br>#
    [G ⊢# T1 ∧ (T2 ∨ T3) <: (T1 ∧ T2) ∨ (T1 ∧ T3)] *)
| subtyp_dist_t: forall G e T1 T2 T3,
    G⁏e ⊢#
      typ_and T1 (typ_or T2 T3)
      <: typ_or (typ_and T1 T2) (typ_and T1 T3)

(** [――――――――――(ST#-N-Rec)] #<br>#
    [G ⊢# N < μ(T)] *)
| subtyp_nocap_bnd_t G e T:
    G⁏e ⊢# typ_nocap <: typ_bnd T

(** [――――――――――(ST#-N-Typ)] #<br>#
    [G ⊢# N <: {A: T1..T2}] *)
| subtyp_nocap_typ_t G e A T1 T2:
    G⁏e ⊢# typ_nocap <: typ_rcd (dec_typ (typ_label_normal A) T1 T2)

(** [――――――――――(ST#-N-Fld)] #<br>#
    [G ⊢# N <: {a: T1..T2}] *)
| subtyp_nocap_fld_t G e a T1 T2:
    G⁏e ⊢# typ_nocap <: typ_rcd (dec_fld a T1 T2)

(** [――――――――――(ST#-N-Met)] #<br>#
    [G ⊢ N <: {m(TC,TZ):TR}] *)
| subtyp_nocap_met_t G e a TC TZ TR:
    G⁏e ⊢# typ_nocap <: typ_rcd (dec_met a TC TZ TR)

(** [――――――――――(ST#-N-Fun)] #<br>#
    [G ⊢# N <: ∀(TZ)TR]
    This rule is disabled in rodot_sef typing mode.  *)
| subtyp_nocap_fun_t G e TZ TR:
    mode_has_fun typing_mode ->
    G⁏e ⊢# typ_nocap <: typ_fun TZ TR

(** [――――――――――(ST#-NM)] #<br>#
    [G ⊢# ⊤ <: N ∨ {M: T1..T2}] *)
| subtyp_nocap_mut_top_boundless_t G e T1 T2:
    mode_has_top_nm_subtype_boundless typing_mode ->
    G⁏e ⊢# typ_top <: (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) T1 T2)))

(** [G ⊢# ⊤ <: N ∨ {M: ⊥..⊥}] #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_nocap_mut_top_complement_t G e:
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢# typ_top <: (typ_or typ_nocap (typ_mut_marker typ_bot) )

(** [G ⊢# N ∧ {M: ⊥..⊥} <: ⊥] #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_nocap_mut_bot_t G e:
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢# (typ_and typ_nocap (typ_mut_marker typ_bot) ) <: typ_bot

(** [G ⊢# ⊤ <: {M: ⊥..⊤}] #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_mut_top_t G e:
    mode_has_mut_top typing_mode ->
    G⁏e ⊢# typ_top <: (typ_mut_marker typ_top)

where "G '⁏' e '⊢#' T '<:' U" := (subtyp_t G e T U).

Global Existing Instances ty_trm_t.

Hint Constructors ty_trm_t subtyp_t : core.

Scheme ts_ty_trm_t_mut := Induction for ty_trm_t Sort Prop
with   ts_subtyp_t     := Induction for subtyp_t Sort Prop.
Combined Scheme ts_t_mutind from ts_ty_trm_t_mut, ts_subtyp_t.

(** Tight typing implies general typing. *)
Lemma tight_to_general:
  (forall G e x T,
     G⁏e ⊢# x ∶ T ->
     ok e ->
     inert G ->
     G⁏e ⊢ (x) ∶ T) /\
  (forall G e S U,
     G⁏e ⊢# S <: U ->
     ok e ->
     inert G ->
     G⁏ e ⊢ S <: U).
(* The ok e hypothesis seems necessary in order to use
   ty_ewkn_var/ty_ekn_empty_trm. *)
Proof.
  intros.
  apply ts_t_mutind; intros; subst;
    eauto using precise_to_general.

  (* bnd intro *)
  eapply ty_rec_intro. eauto. eauto.
  intros.
  destruct (mode_incap' typing_mode) eqn: Hm; eauto.
  unfold mode_has_incap' in H3. rewrite Hm in H3. discriminate.
Qed.

Lemma tight_typing_implies_bound G e (x: var) T:
  G⁏e ⊢# x ∶ T ->
  exists S, ctx_binds x S G.
Proof.
  introv Ht. remember (x).
  induction Ht; try congruence;
    match goal with
    | [ H : _ = _ |- _ ] => inversions H
    end; eauto.
Qed.

Lemma ty_tight_dom_top G e x:
  inert G ->
  x \in ctx_dom G ->
  G ⁏ e ⊢# x ∶ typ_top.
Proof.
  introv Hin Hdom.
  eapply ctx_dom_to_binds in Hdom; eauto.
  destructecs Hdom.
  eauto.
Qed.


Lemma precise_to_tight G e x T U:
  inert G ->
  G ⊢! x ∶ T ⪼ U ->
  G⁏e ⊢# x ∶ U.
Proof.
  introv Hin Hpf.
  induction Hpf; eauto.
Qed.


Lemma record_type_subtyp_nocap G e T:
  record_type T ->
  G ⁏ e ⊢# typ_nocap <: T.
Proof.
  introv Hre.
  induction Hre.
  induction H; subst.
  - inversions H; eauto.
  - eapply subtyp_and2_t.
    auto.
    inversions H0; eauto.
Qed.


End WithTypingMode.

Hint Constructors ty_trm_t subtyp_t : core.
Notation "G '⁏' e ⊢# T '<:' U" := (subtyp_t _ G e T U).
