Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypeInduction TypingMode PreciseTyping Equivalence.
Require Import RecordTypes InertContext EnvCorrespondence Closedness Nosel.
Require Import Coq.Program.Equality.

(** * Approximation of types *)

Inductive polarity : Type := positive | negative.
Notation "⊕" := positive. Notation "⊝" := negative.
(* Opposite polarity *)
Notation "─ p" := match p with ⊕ => ⊝ | ⊝ => ⊕ end
                    (at level 50, p at level 49).

(* Approximation:
   ˢ - approximate selections by bounds
   ᵐ - approximate methods by extreme
   ᵉ - approximate selections by opposite extreme *)
Inductive approx_target : Type :=  ˢ | ᵐ | ᵉ.

(* Approximation target which does not have a reflexivity constructor *)
Notation approx_target_norefl atg := (atg = ᵐ \/ atg = ᵉ) (only parsing).
(* Approximation target which has subtyping defined *)
Notation approx_target_subtyping atg := (atg = ˢ \/ atg = ᵐ) (only parsing).

(* An extreme in the direction of approximaption *)
Notation typ_approx_extreme p := (match p with ⊕ => typ_top | ⊝ => typ_bot end).
Notation typ_approx_extreme_nocap p :=
  (match p with ⊕ => typ_top | ⊝ => typ_nocap end).
Notation polarity_apply ST p T1 T2 := (match p with ⊕ => ST T1 T2 | ⊝ => ST T2 T1 end).

Section WithTypingMode.
Context {typing_mode: Mode}.

Reserved Notation "F ⊢ T1 ↝ atg ⦇ p ⦈ T2"
         (at level 40, T1 at level 59, T2 at level 59,
          p at level 39, atg at level 39).

Inductive typ_approx : ctx -> typ -> approx_target -> polarity -> typ -> Prop :=

| typ_approx_refl : forall (F : ctx) (T : typ) (δ : polarity),
    F ⊢ T ↝ˢ⦇δ⦈ T

| typ_approx_s_sel_u : forall (F : ctx) (B : typ_label) (T1 T2 T3 T : typ) (v v2: var),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    F ⊢ (open v2 T2) ↝ˢ⦇⊕⦈ T3 ->
    F ⊢ typ_sel (avar_f v) B (avar_f v2) ↝ˢ⦇⊕⦈ T3

| typ_approx_s_sel_l : forall (F : ctx) (B : typ_label) (T1 T2 T3 T : typ) (v v2: var),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    F ⊢ (open v2 T1) ↝ˢ⦇⊝⦈ T3 ->
    F ⊢ typ_sel (avar_f v) B (avar_f v2) ↝ˢ⦇⊝⦈ T3

| typ_approx_andor (typ_bin : typ -> typ -> typ) atg δ:
    forall (F : ctx) (T1 T1' T2 T2' : typ),
    typ_bin = typ_and \/ typ_bin = typ_or ->
    F ⊢ T1 ↝atg⦇δ⦈ T1' ->
    F ⊢ T2 ↝atg⦇δ⦈ T2' ->
    F ⊢ (typ_bin T1 T2) ↝atg⦇δ⦈ (typ_bin T1' T2')

| typ_approx_fld : forall (F : ctx) (T1 T1' T2 T2' : typ) (a : fld_label)
                 (atg : approx_target) (δ : polarity),
    F ⊢ T1 ↝atg⦇─δ⦈ T1' ->
    F ⊢ T2 ↝atg⦇δ⦈ T2' ->
    F ⊢ (typ_rcd (dec_fld a T1 T2)) ↝atg⦇δ⦈ (typ_rcd (dec_fld a T1' T2'))

| typ_approx_typ atg δ: forall (F : ctx) (T1 T1' T2 T2' : typ) (B : typ_label),
    (forall x, F ⊢ (open x T1) ↝atg⦇─δ⦈ (open x T1')) ->
    (forall x, F ⊢ (open x T2) ↝atg⦇δ⦈ (open x T2')) ->
    F ⊢ (typ_rcd (dec_typ B T1 T2)) ↝atg⦇δ⦈ (typ_rcd (dec_typ B T1' T2'))

| typ_approx_top : forall F (δ : polarity) (atg : approx_target),
    approx_target_norefl atg ->
    F ⊢ typ_top ↝atg⦇δ⦈ typ_top

| typ_approx_bot : forall F (δ : polarity) (atg : approx_target),
    approx_target_norefl atg ->
    F ⊢ typ_bot ↝atg⦇δ⦈ typ_bot

| typ_approx_nocap : forall F (δ : polarity) (atg : approx_target),
    approx_target_norefl atg ->
    F ⊢ typ_nocap ↝atg⦇δ⦈ typ_nocap

| typ_approx_me_met : forall F (atg : approx_target) (δ : polarity) (m : met_label) (T1 T2 T3 : typ),
    atg = ᵐ \/ atg = ᵉ ->
    F ⊢ typ_rcd (dec_met m T1 T3 T2) ↝atg⦇δ⦈ (typ_approx_extreme_nocap δ)

| typ_approx_me_fun : forall F (atg : approx_target) (δ : polarity)
                        (T1 T2 : typ),
    approx_target_norefl atg ->
    F ⊢ typ_fun T1 T2
      ↝atg⦇δ⦈ match mode_has_fun' typing_mode with
              | true => (typ_approx_extreme_nocap δ)
              | false => (typ_approx_extreme δ)
              end

| typ_approx_sel : forall F (v v2: avar) (B : typ_label)
                (atg : approx_target) (δ : polarity),
    approx_target_norefl atg ->
    F ⊢ typ_sel v B v2 ↝atg⦇δ⦈ typ_sel v B v2

| typ_approx_rec : forall F (T : typ) (atg : approx_target) (δ : polarity),
    approx_target_norefl atg ->
    F ⊢ typ_bnd T ↝atg⦇δ⦈ typ_bnd T

| typ_approx_e_sel (δ : polarity): forall F (v v2: avar) (B : typ_label) ,
    F ⊢ typ_sel v B v2 ↝ᵉ⦇δ⦈ (typ_approx_extreme (─δ))

where "F ⊢ T1 ↝ atg ⦇ p ⦈ T2" := (typ_approx F T1 atg p T2).
End WithTypingMode.

Hint Constructors typ_approx : core.
(* Specialized hints *)
Hint Resolve (typ_approx_e_sel positive) (typ_approx_e_sel negative) : core.

Notation "F ⊢ T1 ↝ atg ⦇ p ⦈ T2" := (typ_approx F T1 atg p T2)
         (at level 40, T1 at level 59, T2 at level 59,
          p at level 39, atg at level 39).

Notation "T1 ↝ atg ⦇ p ⦈ T2" := (ctx_empty ⊢ T1 ↝ atg ⦇ p ⦈ T2)
         (at level 60, T2 at level 59, p at level 50, atg at level 39).

(* Tactics useful in proofs about approximation and restricted subtyping
    and possibly elsewhere. *)

(* Handles inversions where typ_approx_andor introduces disjunction equality *)
Ltac inversions_typ H := inversions H;
 repeat match goal with
 | typ_bin: typ -> typ -> typ, H: ?typ_bin = ?T1 \/ ?typ_bin = ?T2 |- _ => destruct H; subst
 | H: ˢ = ᵐ \/ ˢ = ᵉ |- _ => destruct H; discriminate (* Eliminates conflicting target *)
 | H: (@eq typ ?t1 ?t2) |- _ => inversions H
 end.

(*
 repeat match goal with
 | H: (@eq typ _ _) \/ (@eq typ _ _) |- _ => destruct H; subst; try discriminate
 | H: (@eq typ ?t1 ?t2) |- _ => inversions H
 end.*)

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma typ_approx_bot_inv atg δ G T2:
  G ⊢ typ_bot ↝ atg ⦇δ⦈ T2 ->
  T2 = typ_bot.
Proof.
  intros.
  inversions H; eauto.
  - inversions H1; inversions H0.
Qed.

End WithTypingMode.

Ltac specialize_receiver x := (
    let x:= fresh x in
    pick_fresh x;
    repeat match goal with
    H: forall x0 : var, ?P |- _ => specialize (H x)
    end
).

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Lemma 5.94 SRedMut, Lemma 5.100 TRedMut, Lemma 5.101 ERedMut *)
Lemma typ_approx_typ_bot_inv δ G l T2 atg:
  G ⊢ typ_rcd(dec_typ l typ_bot typ_bot) ↝ atg ⦇δ⦈ T2 ->
  T2 = typ_rcd(dec_typ l typ_bot typ_bot).
Proof.
  intros.
  inversions H; eauto.
  - (* binary *)
    inversions H1; inversions H0.
  - (* typ *)
    destruct δ. simpls.
    all: specialize_receiver x.
    all: apply typ_approx_bot_inv in H7; eauto.
    all: apply typ_approx_bot_inv in H8; eauto.

    all: apply typ_bot_open_inv in H7; eauto.
    all: apply typ_bot_open_inv in H8; eauto.
    all: subst; auto.
Qed.

(* Lemma 5.95 SRedTrans *)
Lemma typ_approx_s_trans δ G T1 T2 T3:
  G ⊢ T1 ↝ˢ⦇δ⦈ T2 ->
  G ⊢ T2 ↝ˢ⦇δ⦈ T3 ->
  G ⊢ T1 ↝ˢ⦇δ⦈ T3.
Proof.
  intros. revert dependent T3.
  dependent induction H; intros; eauto.
  - (* typ_bin - typ_and, typ_or *)
    destruct H; subst;
    inversions_typ H2; eauto.
  - inversions_typ H1; eauto.
  - (* typ *)
    inversions_typ H3; eauto.
  - destruct H; inversions H.
  - destruct H; inversions H.
Qed.

Lemma typ_approx_closed δ atg G T1 T2:
  inert G ->
  sealed G ->
  closed T1 ->
  G ⊢ T1 ↝atg⦇δ⦈ T2 ->
  closed T2.
Proof.
  introv Hin HclG HclT Hap12.
  dependent induction Hap12; eauto.
  - eapply IHHap12; eauto.
    eapply pf_typ_closed_l; eauto.
  - eapply IHHap12; eauto.
    eapply pf_typ_closed_r; eauto.
  - (* and or *)
    destruct H; subst.
    + forwards: closed_typ_and_l; eauto.
      forwards: closed_typ_and_r; eauto.
      eapply closed_typ_and; eauto.
    + forwards: closed_typ_or_l; eauto.
      forwards: closed_typ_or_r; eauto.
      eapply closed_typ_or; eauto.
  - eapply closed_typ_rcd_inv in HclT.
    forwards: closed_dec_fld_l; eauto.
    forwards: closed_dec_fld_r; eauto.
    eapply closed_typ_rcd.
    eapply closed_dec_fld; eauto.
  - (* typ *)
    eapply closed_typ_rcd_inv in HclT.
    assert (forall x, closed (open x T1)).
      intros. eapply closed_dec_typ_inv_l; eauto.
    assert (forall x, closed (open x T2)).
      intros. eapply closed_dec_typ_inv_r; eauto.
    assert (forall x, closed (open x T1')) by eauto.
    assert (forall x, closed (open x T2')) by eauto.
    eapply closed_typ_rcd.
    eapply closed_dec_typ with (L:=\{}); eauto.
  - destruct δ. eapply closed_typ_top. eapply closed_typ_nocap.
  - destruct δ.
    all: destruct (mode_has_fun' typing_mode); try discriminate.
    all: eauto using closed_typ_top, closed_typ_nocap, closed_typ_bot.
  - destruct δ. eapply closed_typ_bot. eapply closed_typ_top.
Qed.


Lemma typ_approx_closed_n δ atg G T1 T2 n x:
  inert G ->
  sealed G ->
  x \notin ctx_dom G \u ctx_fv G ->
  (open_rec n x T1 = T1) ->
  G ⊢ T1 ↝atg⦇δ⦈ T2 ->
  (open_rec n x T2 = T2).
Proof.
  introv Hin HclG Hn HclT Hap12.
  reverts n.
  dependent induction Hap12; intros; eauto.
  - eapply IHHap12; eauto.
    eapply pf_typ_closed_l; eauto.
  - eapply IHHap12; eauto.
    eapply pf_typ_closed_r; eauto.
  - (* and or *)
    destruct H; subst.
    + sympls. inversions HclT. rewrite IHHap12_1. rewrite IHHap12_2. all: auto; notin_solve.
    + sympls. inversions HclT. rewrite IHHap12_1. rewrite IHHap12_2. all: auto; notin_solve.
  - sympls. inversions HclT. rewrite IHHap12_1. rewrite IHHap12_2. all: auto; notin_solve.
  - (* typ *)

sympls.  inversions HclT.
pick_fresh z.
assert (open_rec (S n) x (open z T1') = open z T1' ) as Hcl1.
eapply H0; eauto.
rewrite open_commut. rewrite H4. auto. discriminate.

assert (open_rec (S n) x (open z T2') = open z T2' ) as Hcl2.
eapply H2; eauto.
rewrite open_commut. rewrite H5. auto. discriminate.


rewrite open_commut in Hcl1.
eapply open_fresh_injective in Hcl1. rewrite Hcl1.
rewrite open_commut in Hcl2.
eapply open_fresh_injective in Hcl2. rewrite Hcl2.
auto. eapply notin_fv_open; eauto. notin_solve. discriminate. eapply notin_fv_open; eauto. eauto. eauto.


  - destruct δ. eapply closed_typ_top. eapply closed_typ_nocap.
  - destruct δ.
    all: destruct (mode_has_fun' typing_mode); try discriminate.
    all: eauto using closed_typ_top, closed_typ_nocap.
  - destruct δ. eapply closed_typ_bot. eapply closed_typ_top.
Qed.



Lemma typ_approx_extreme_subst δ x y:
  (typ_approx_extreme δ) /[ x -> y] = typ_approx_extreme δ.
Proof.
  destruct δ; simpls; eauto.
Qed.
Lemma typ_approx_extreme_open δ n x:
  open_rec n x (typ_approx_extreme δ) = typ_approx_extreme δ.
Proof.
  destruct δ; simpls; eauto.
Qed.

Lemma typ_approx_extreme_nocap_open δ n x:
  open_rec n x (typ_approx_extreme_nocap δ) = typ_approx_extreme_nocap δ.
Proof.
  destruct δ; simpls; eauto.
Qed.

End WithTypingMode.


Ltac apply_reorder_classes_typ := ([> idtac|eapply TransformSubstFresh| eapply TypFvSubstCases| eapply TransformSubstCommut| eapply TypSubstId| eapply TransformSubstOpenCommut| eapply TransformOpenCommut|idtac]).
(*|eapply AvarSubstFresh| eapply AvarFvSubstCases| eapply AvarSubstCommut| eapply AvarSubstId| *)
Ltac apply_reorder_classes_avar := ([> idtac| eapply AvarOpenCommut|idtac]).
Ltac apply_reorder_classes_typ' := ([> idtac| eapply TransformOpenCommut|idtac]).

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma typ_approx_reordering G δ atg T1 T2 xs0 ys0 xs1 ys1 xs2 ys2:
  fresh (fv T1 \u fv T2 \u ctx_dom G \u ctx_fv G) (length xs0) xs0 ->
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  G ⊢ (open_vars xs1 T1) ↝ atg ⦇ δ ⦈ (open_vars xs2 T2) ->
  G ⊢ (open_vars ys1 T1) ↝ atg ⦇ δ ⦈ (open_vars ys2 T2).
Proof.
  introv Hfr Hro Hap.
  reverts ys1 ys2.
  reverts xs0 ys0.
  dependent induction Hap; intros.
(*  all: try solve [destruct H; discriminate].*)
  - (* refl *)
    erewrite open_vars_reorder_eq; eauto.
  - (* sel l *)
   invert_open. simpls. autorewrite with rewrite_open.
    assert ((avar_f v) = x). {
      destruct (avar_f_open_vars_cases _ _ _ H1). auto. false.
      destructc H0 as n Hnx Hv ?.
      eapply LibList.Nth_mem in Hv.
      inversions Hro.
      eapply list_reorder_mem in Hv; eauto.
      eapply fresh_in_notin with (X:= ctx_dom F) in Hv.
      eapply pf_binds in H.
      eapply ctx_binds_dom in H. contradiction.
      fresh_solve.
    }
    subst.
    rewrite open_rec_vars_avar_f.
    assert(fresh (fv T4) (length xs0) xs0) as HfvT4. {
      eapply fresh_forall_notin with (L:=(ctx_fv F \u ctx_dom F)). fresh_solve.
      intros x Hx.
      eapply (pf_dec_typ_notin_fv_u) with (y:=x) in H; eauto.
    }
    assert((length xs1) = (length ys1)) as Hl1. {
      eapply list_reordering_lengths in Hro. destruct Hro. auto.
    }

    destruct (avar_f_open_vars_cases _ _ _ H2).
    + subst. simpls.
      rewrite open_rec_vars_avar_f in H2. inversions H2.
      rewrite open_rec_vars_avar_f in *.
      eapply typ_approx_s_sel_u. eauto.
      pick_fresh_with z (from_list xs0).
      assert (fresh \{ z} (length xs0) xs0).
      eapply fresh_from_list; eauto.
      assert (list_reordering (z::xs0) (v2::ys0) (z::nil) (v2::nil) xs2 ys2) as Hro2.
        eapply list_reordering_new_one1; eauto.
      eapply IHHap with (T1:=(open v2 T4)) in Hro2. simpl in Hro2.
      rewrite open_twice in Hro2. eapply Hro2. simpl. rewrite open_twice. auto.  auto. auto. fresh_solve.
      eapply notin_fv_open.  notin_solve.  notin_solve. fresh_solve.
      eapply fresh_fv_open. all: fresh_solve.

    + destructc H0 as n Hx0 Hx Hl. subst.
      rewrite plus_O_n in *.
      rewrite Hl1 in Hl.
      destructc (avar_b_open_vars _ Hl) as x2 Hx2 ?. rewrite Hx2.
      eapply typ_approx_s_sel_u. eauto.
      assert (list_reordering (xs0) (ys0) (v2::nil) (x2::nil) xs2 ys2) as Hro2.
        eapply list_reordering_old_one1; eauto.
      eapply IHHap with (T1:=T4) in Hro2. simpl in Hro2.
      eapply Hro2.
      destruct (avar_f_open_vars_cases _ _ _ H2); try discriminate.
      destructc H3 as n' Hn' Hv2. subst. inversions Hn'. simpls. auto. auto.
      fresh_solve.
  - (* sel u *)
    invert_open. simpls. autorewrite with rewrite_open.
    assert ((avar_f v) = x). {
      destruct (avar_f_open_vars_cases _ _ _ H1). auto. false.
      destructc H0 as n Hnx Hv ?.
      eapply LibList.Nth_mem in Hv.
      inversions Hro.
      eapply list_reorder_mem in Hv; eauto.
      eapply fresh_in_notin with (X:= ctx_dom F) in Hv.
      eapply pf_binds in H.
      eapply ctx_binds_dom in H. contradiction.
      fresh_solve.
    }
    subst.
    rewrite open_rec_vars_avar_f.
    assert(fresh (fv T0) (length xs0) xs0) as HfvT4. {
      eapply fresh_forall_notin with (L:=(ctx_fv F \u ctx_dom F)). fresh_solve.
      intros x Hx.
      eapply (pf_dec_typ_notin_fv_l) with (y:=x) in H; eauto.
    }
    assert((length xs1) = (length ys1)) as Hl1. {
      eapply list_reordering_lengths in Hro. destruct Hro. auto.
    }

    destruct (avar_f_open_vars_cases _ _ _ H2).
    + subst. simpls.
      rewrite open_rec_vars_avar_f in H2. inversions H2.
      rewrite open_rec_vars_avar_f in *.
      eapply typ_approx_s_sel_l. eauto.
      pick_fresh_with z (from_list xs0).
      assert (fresh \{ z} (length xs0) xs0).
      eapply fresh_from_list; eauto.
      assert (list_reordering (z::xs0) (v2::ys0) (z::nil) (v2::nil) xs2 ys2) as Hro2.
        eapply list_reordering_new_one1; eauto.
      eapply IHHap with (T1:=(open v2 T0)) in Hro2. simpl in Hro2.
      rewrite open_twice in Hro2. eapply Hro2. simpl. rewrite open_twice. auto.  auto. auto. fresh_solve.
      eapply notin_fv_open.  notin_solve.  notin_solve. fresh_solve.
      eapply fresh_fv_open. all: fresh_solve.

    + destructc H0 as n Hx0 Hx Hl. subst.
      rewrite plus_O_n in *.
      rewrite Hl1 in Hl.
      destructc (avar_b_open_vars _ Hl) as x2 Hx2 ?. rewrite Hx2.
      eapply typ_approx_s_sel_l. eauto.
      assert (list_reordering (xs0) (ys0) (v2::nil) (x2::nil) xs2 ys2) as Hro2.
        eapply list_reordering_old_one1; eauto.
      eapply IHHap with (T1:=T0) in Hro2. simpl in Hro2.
      eapply Hro2.
      destruct (avar_f_open_vars_cases _ _ _ H2); try discriminate.
      destructc H3 as n' Hn' Hv2. subst. inversions Hn'. simpls. auto. auto.
      fresh_solve.
  - (* and or *)
    destruct H; subst.
    + (* and *)
      invert_open.
      invert_open.
      simpls.
      autorewrite with rewrite_open.
      constructor; eauto.
      eapply IHHap1; eauto.
      fresh_solve; unfold fv, TypFreeVar; fresh_solve.
      eapply IHHap2; eauto.
      fresh_solve; unfold fv, TypFreeVar; fresh_solve.
    + (* or *)
      invert_open.
      invert_open.
      sympls.
      autorewrite with rewrite_open.
      constructor; eauto.
      eapply IHHap1; eauto.
      fresh_solve.
      eapply IHHap2; eauto.
      fresh_solve.
  - (* fld *)
      invert_open.
      invert_open.
      sympls.
      autorewrite with rewrite_open.
      constructor; eauto.
      eapply IHHap1; eauto.
      fresh_solve.
      eapply IHHap2; eauto.
      fresh_solve.
  - (* typ *)
      invert_open.
      invert_open.
      sympls.

      rewrite open_rec_vars_typ. rewrite open_rec_vars_typ.
      constructor; eauto.
      intro.
      { pick_fresh_with z (from_list xs0).
        assert (fresh \{ z} (length xs0) xs0). {
          eapply fresh_from_list. notin_solve. eauto.
        }
        rewrite open_vars_push. rewrite open_vars_push.
        eapply H0 with (xs0:=(z::xs0)%list); auto.
        apply open_vars_push with (x0:=z).
        apply open_vars_push with (x0:=z).
        eapply list_reordering_push; eauto.
      }
      intro.
      { pick_fresh_with z (from_list xs0).
        assert (fresh \{ z} (length xs0) xs0). {
          eapply fresh_from_list. notin_solve. eauto.
        }
        rewrite open_vars_push. rewrite open_vars_push.
        eapply H2 with (xs0:=(z::xs0)%list); auto.
        apply open_vars_push with (x0:=z).
        apply open_vars_push with (x0:=z).
        eapply list_reordering_push; eauto.
      }
  -
     invert_open.
      invert_open.  autorewrite with rewrite_open. eauto.

  -
     invert_open.
      invert_open.  autorewrite with rewrite_open. eauto.
  -
     invert_open.
      invert_open.  autorewrite with rewrite_open. eauto.
  - destruct δ; simpls;
     invert_open;     invert_open.
     autorewrite with rewrite_open. eapply typ_approx_me_met. auto.
     autorewrite with rewrite_open. eapply typ_approx_me_met. auto.
  - assert (Happrox_fun : forall (F : ctx) (atg : approx_target)
                            (T1 T2 : typ),
               atg = ᵐ \/ atg = ᵉ ->
               F ⊢ typ_fun T1 T2
                 ↝atg⦇δ⦈ (if mode_has_fun' typing_mode
                          then typ_approx_extreme_nocap δ
                          else typ_approx_extreme δ)) by auto.
    destruct (mode_has_fun' typing_mode) eqn:Hhf; try discriminate.
    all: destruct δ; simpls;
        invert_open;     invert_open.
    all: autorewrite with rewrite_open; auto.

-     invert_open.
      invert_open.  autorewrite with rewrite_open.
symmetry in H1.
    eapply open_vars_reorder_eq in H1; eauto. rewrite H1.
symmetry in H2.
    eapply open_vars_reorder_eq in H2; eauto. rewrite H2.
 eauto. simpls. fresh_solve. simpls. fresh_solve.
  -
     invert_open.
      invert_open.  autorewrite with rewrite_open.
symmetry in H1.
    eapply open_vars_reorder_eq in H1; eauto. rewrite H1.
  eapply typ_approx_rec; eauto. sympls. reverts Hfr. intros. fresh_solve.
- destruct δ; simpls;
     invert_open;     invert_open.
     autorewrite with rewrite_open. eapply typ_approx_e_sel.
     autorewrite with rewrite_open. eapply typ_approx_e_sel.
Qed.

Lemma typ_approx_reopen G δ atg T1 T2 x y:
  x \notin (ctx_dom G \u ctx_fv G \u fv T1 \u fv T2) ->
  G ⊢ (open x T1) ↝ atg ⦇ δ ⦈ (open x T2) ->
  G ⊢ (open y T1) ↝ atg ⦇ δ ⦈ (open y T2).
Proof.
  repeat rewrite <- open_vars_single .
  set (xs := (x::nil)%list). set (ys := (y::nil)%list).
  introv Hn Hap.
  eapply typ_approx_reordering; eauto.
  2: eapply list_reordering_same.
  - unfold xs. fresh_solve; auto.
  - unfold xs, ys. auto.
Qed.




Lemma typ_approx_reopen_closed δ atg F x y T1 T2:
  closed T2 ->
  x \notin ctx_dom F \u ctx_fv F \u fv T1 ->
  F ⊢ open x T1 ↝ atg ⦇ δ ⦈ T2 ->
  F ⊢ open y T1 ↝ atg ⦇ δ ⦈ open y (close x T2).
Proof.
  introv Hcl Hn Hap.
  eapply typ_approx_reopen with (x:=x). notin_solve. eapply notin_fv_close.
  rewrite open_close.
  rewrite Hcl. eapply Hap.
Qed.


Lemma typ_approx_reopen_closed' δ atg F x y T1 T2:
  sealed F ->
  inert F ->
  x \notin ctx_dom F \u ctx_fv F \u fv T1 ->
  F ⊢ open x T1 ↝ atg ⦇ δ ⦈ T2 ->
  F ⊢ open y T1 ↝ atg ⦇ δ ⦈ open y (close x T2).
Proof.
  introv HclG HinG Hn Hap.
  eapply typ_approx_reopen with (x:=x). notin_solve. eapply notin_fv_close.
  rewrite open_close.
  pose proof Hap. eapply (typ_approx_closed_n 0) with (x0:=x) in H; eauto.
  rewrite H. eapply Hap. rewrite open_twice. auto.
Qed.
Hint Resolve typ_approx_reopen_closed': core.

Lemma typ_approx_open δ atg F n x (T1 T2: typ):
  inert F ->
  sealed F ->
  F ⊢ T1 ↝ atg ⦇ δ ⦈ T2 ->
  F ⊢ open_rec n x T1 ↝ atg ⦇ δ ⦈ open_rec n x T2.
Proof.
  introv Hin Hcl (*Hn *) Hap.
  reverts n.
  induction Hap; intros; simpls; eauto.
  - eapply typ_approx_s_sel_u; eauto.
    eapply pf_typ_closed_l with (x0:=v2) in H; eauto.
    specialize (IHHap Hin Hcl n).
    rewrite H in IHHap. auto.
  - eapply typ_approx_s_sel_l; eauto.
    eapply pf_typ_closed_r with (x0:=v2) in H; eauto.
    specialize (IHHap Hin Hcl n).
    rewrite H in IHHap. auto.
  - destruct H; subst; simpls; fold_typ_open_rec; eauto.
  -
    constructor. intros. fold_typ_open_rec. rewrite open_commut. rewrite (open_commut x0).
    eapply H0. eauto. eauto. discriminate. discriminate.

    intros. fold_typ_open_rec. rewrite open_commut. rewrite (open_commut x0).
    eapply H2. eauto. eauto. discriminate. discriminate.
  - fold_typ_open_rec. rewrite typ_approx_extreme_nocap_open. eauto.
  - fold_typ_open_rec.
    assert (Happrox_fun : forall (F : ctx) (atg : approx_target)
                            (T1 T2 : typ),
               atg = ᵐ \/ atg = ᵉ ->
               F ⊢ typ_fun T1 T2
                 ↝atg⦇δ⦈ (if mode_has_fun' typing_mode
                          then typ_approx_extreme_nocap δ
                          else typ_approx_extreme δ)) by auto.
    destruct (mode_has_fun' typing_mode) eqn:Hhf; try discriminate.
    + repeat(rewrite typ_approx_extreme_nocap_open); eauto.
    + repeat(rewrite typ_approx_extreme_open). eauto.
  - fold_typ_open_rec. rewrite typ_approx_extreme_open. eauto.
Qed.


(* Lemma 5.96 SRedCom *)
Lemma typ_approx_s_com δ G T1 T2 T3:
  inert G ->
  sealed G ->
  G ⊢ T1 ↝ˢ⦇δ⦈ T2 ->
  G ⊢ T1 ↝ˢ⦇δ⦈ T3 ->
  exists T4,
  G ⊢ T2 ↝ˢ⦇δ⦈ T4 /\
  G ⊢ T3 ↝ˢ⦇δ⦈ T4.
Proof.
  introv Hin HslG Hap12 Hap13.
  reverts T3.
  dependent induction Hap12; intros; eauto.
  -
      inversions_typ Hap13; eauto.
     destruct (pf_inert_unique_tight_bounds Hin H H3).
     subst; auto.
  -  inversions_typ Hap13; eauto.
     destruct (pf_inert_unique_tight_bounds Hin H H3).
     subst; auto.
  - destruct H; subst.

    + inversions_typ Hap13.
      exists (typ_and T1' T2'). auto.
      autospecialize.
      destruct (IHHap12_1) as [T3 [H13 H103]];
      destruct (IHHap12_2) as [T4 [H24 H204]].
      exists (typ_and T3 T4). auto.
    + inversions_typ Hap13.
      exists (typ_or T1' T2'). auto.
      autospecialize.
      destruct (IHHap12_1) as [T3 [H13 H103]];
      destruct (IHHap12_2) as [T4 [H24 H204]].
      exists (typ_or T3 T4). auto.
  - inversions_typ Hap13; eauto.
    autospecialize.
    destruct δ;
    destruct (IHHap12_1) as [T3 [H13 H103]];
    destruct (IHHap12_2) as [T4 [H24 H204]];
    exists (typ_rcd (dec_fld a T3 T4)); auto.
  - inversions_typ Hap13; eauto.
    pick_fresh x.
    clear H H1.
    forwards T5 H15 H105: H0; eauto.
    forwards T6 H26 H206: H2; eauto.
    exists (typ_rcd (dec_typ B (close x T5) (close x T6))).
    splits; eauto.
  - inversions_typ Hap13; eauto.
  - destruct H; inversions H.
Qed.

End WithTypingMode.

Hint Resolve typ_approx_reopen_closed': core.

(* Applies typ_approx_s_com, automatically finding the approximation hypotheses *)
Ltac typ_approx_s_com_for_type T0 T := (
  match goal with
   | Hin: inert ?F, Hcl: sealed ?F, H0: ?F ⊢ T0 ↝ ˢ ⦇ ?δ ⦈ ?T0n, H00: ?F ⊢ T0 ↝ ˢ ⦇ ?δ ⦈ ?T00n |- _ =>
      let T1 := fresh T in
      let T2 := fresh T in
      destruct (typ_approx_s_com Hin Hcl H0 H00) as [T1 [? ?]]
   end
).
Ltac typ_approx_s_com_for T := typ_approx_s_com_for_type T T.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Lemma 5.97 SRedEq, Lemma 5.106 TRedEq *)
Lemma typ_approx_tyq atg δ G e T1 T2 T3:
  inert G ->
  sealed G ->
  renv_corr G e ->
  e ⊢ T1 ≈ T2 ->
  G ⊢ T1 ↝ atg ⦇δ⦈ T3 ->
  exists T4,
  G ⊢ T2 ↝ atg ⦇δ⦈ T4 /\
  e ⊢ T3 ≈ T4.
Proof.
  introv Hin Hcl Hcor Htyq12 Hap13.
  reverts T2.
  dependent induction Hap13; intros; try solve [inversions_tyq Htyq12; eauto 6].
  - (* sel_u *)
    inversions_tyq Htyq12.
    + destruct (renv_corr_pf_normal_typ_tyq Hin Hcor H5 H) as [T4b [T41 [T42 [H4pf [Htyq41 Htyq42]]]]].
      (*eapply tyq_typ_open with (y := v2) in Htyq42.*)
      specialize (Htyq42 v2).
      destruct (IHHap13 Hin Hcl Hcor _ Htyq42) as [T5 [H5ap H5tyq]]; eauto.
    + eauto.
  - (* sel_l *)
    inversions_tyq Htyq12.
    + destruct (renv_corr_pf_normal_typ_tyq Hin Hcor H5 H) as [T4b [T41 [T42 [H4pf [Htyq41 Htyq42]]]]].
      (* eapply tyq_typ_open with (y := v2) in Htyq41.*)
      specialize (Htyq41 v2).
      destruct (IHHap13 Hin Hcl Hcor _ Htyq41) as [T5 [H5ap H5tyq]].
      eauto.
    + eauto.
  - destruct H; subst.
    inversions_tyq Htyq12.
    autospecialize.
    destruct (IHHap13_1) as [T5 [H5ap H5tyq]].
    destruct (IHHap13_2) as [T6 [H6ap H6tyq]].
    exists (typ_and T5 T6). eauto.
    inversions_tyq Htyq12.
    autospecialize.
    destruct (IHHap13_1) as [T5 [H5ap H5tyq]].
    destruct (IHHap13_2) as [T6 [H6ap H6tyq]].
    exists (typ_or T5 T6). eauto.
  - inversions_tyq Htyq12.
    autospecialize.
    destruct (IHHap13_1) as [T5 [H5ap H5tyq]].
    destruct (IHHap13_2) as [T6 [H6ap H6tyq]].
    exists (typ_rcd (dec_fld a T5 T6)).
    eauto.
  - inversions_tyq Htyq12.
    autospecialize.
    pick_fresh x.
    forwards T5 H5ap H5tyq: (H0 x Hin Hcl Hcor (open x t21)). eauto.
    forwards T6 H6ap H6tyq: (H2 x Hin Hcl Hcor (open x t22)). eauto.
    exists (typ_rcd (dec_typ B (close x T5) (close x T6))).
    split.
    + eauto.
    + eapply tyq_typ_typ.
      intros. eapply tyq_typ_reopen_closed'; eauto.
      intros. eapply tyq_typ_reopen_closed'; eauto.
Qed.

(* Lemma 5.109 ERedEq *)
Lemma typ_approx_tyq_e δ G e T1 T2 T3 y:
  inert G ->
  sealed G ->
  renv_corr G e ->
  renv_notin y e ->
  e ⊢ T1 ≈ T2 ->
  G ⊢ T1 ↝ ᵉ ⦇δ⦈ T3 ->
  nosel T3 y ->
  exists T4,
  G ⊢ T2 ↝ ᵉ ⦇δ⦈ T4 /\
  e ⊢ T3 ≈ T4 /\
  nosel T4 y.
Proof.
  introv Hin Hsl Hcor Hn Htyq12 Hap13 Hns2.
  reverts T2.
  dependent induction Hap13; intros; try solve [inversions_tyq Htyq12; eauto 6].
  - destruct H; subst.
    + inversions_tyq Htyq12.
      inversions_tyq Hns2.
      autospecialize.
      destruct (IHHap13_1) as [T5 [H5ap [H5tyq Hns5]]].
      destruct (IHHap13_2) as [T6 [H6ap [H6tyq Hns6]]].
      exists (typ_and T5 T6). repeat split; eauto.
    + inversions_tyq Htyq12.
      inversions_tyq Hns2.
      autospecialize.
      destruct (IHHap13_1) as [T5 [H5ap [H5tyq Hns5]]].
      destruct (IHHap13_2) as [T6 [H6ap [H6tyq Hns6]]].
      exists (typ_or T5 T6). repeat split; eauto.
  - inversions_tyq Htyq12.
    inversions_tyq Hns2.
    autospecialize.
    destruct (IHHap13_1) as [T5 [H5ap [H5tyq Hns5]]].
    destruct (IHHap13_2) as [T6 [H6ap [H6tyq Hns6]]].
    exists (typ_rcd (dec_fld a T5 T6)).
    repeat split; eauto using  (nosel_bin (typ_binary_rcd_fld a)).
  - (* typ *)
    inversions_tyq Htyq12.
    inversions_tyq Hns2.
    pick_fresh x.
    forwards T5 H5ap H5tyq Hns5: (H0 x); eauto.
    forwards T6 H6ap H6tyq Hns6: (H2 x); eauto.
    exists (typ_rcd (dec_typ B (close x T5) (close x T6))).
    splits.
    + (* approx *) eauto.
    + (* tyq *)
      constructor.
      intros. eapply tyq_typ_reopen_closed'; eauto.
      intros. eapply tyq_typ_reopen_closed'; eauto.
    + eapply (nosel_bin (typ_binary_rcd_typ B));
        eapply nosel_close; eauto.
  - (* sel *)
    inversions_tyq Htyq12.
    inversions_tyq Hns2.
    + exists (typ_sel (avar_f v0) (typ_label_normal l) v2). repeat split; eauto.
      constructor.
      simplify_eq. intro; subst.
      assert (v1 = y). eapply tyq_var_notin_eq; eauto.
      subst. contradiction. auto.
    + exists (typ_sel v B v2). repeat split; eauto.
Qed.

(*
  One could state the following for an arbitrary binary type constructor.
  However, then it is hard to select the right direction for the first operand.
  (Hbi: typ_binary C)
  (Hap: forall F T5 T3 T6 T4,
    F ⊢ T5 ↝ˢ⦇δl⦈ T3 ->
    F ⊢ T6 ↝ˢ⦇δu⦈ T4 ->
    F ⊢ (C T5 T6) ↝ˢ⦇δu⦈ (C T3 T4)
  )
*)

(* Lemma 5.93 SRedInv for and, or *)
Lemma typ_approx_s_inv_pos (C: typ->typ->typ) δ:
  C = typ_and \/ C = typ_or ->
  forall G T1 T2 T0,
  G ⊢ C T1 T2 ↝ ˢ ⦇ δ ⦈ T0 ->
  exists T5 T6,
  T0 = (C T5 T6) /\
  G ⊢ T1 ↝ ˢ ⦇ δ ⦈ T5 /\
  G ⊢ T2 ↝ ˢ ⦇ δ ⦈ T6.
Proof.
  intros.
  inversions H; inversions_typ H0; eauto.
Qed.

Lemma typ_approx_s_inv_typ lt δ:
  forall G T1 T2 T0,
  G ⊢ (typ_rcd (dec_typ lt T1 T2)) ↝ ˢ ⦇ δ ⦈ T0 ->
  exists T5 T6,
  T0 = (typ_rcd (dec_typ lt T5 T6)) /\
  (forall x, G ⊢ (open x T1) ↝ ˢ ⦇ ─δ ⦈ (open x T5)) /\
  (forall x, G ⊢ (open x T2) ↝ ˢ ⦇ δ ⦈ (open x T6)).
Proof.
  intros.
  inversions_typ H; eauto.
Qed.
Lemma typ_approx_s_inv_fld lf δ:
  forall G T1 T2 T0,
  G ⊢ (typ_rcd (dec_fld lf T1 T2)) ↝ ˢ ⦇ δ ⦈ T0 ->
  exists T5 T6,
  T0 = (typ_rcd (dec_fld lf T5 T6)) /\
  G ⊢ T1 ↝ ˢ ⦇ ─δ ⦈ T5 /\
  G ⊢ T2 ↝ ˢ ⦇ δ ⦈ T6.
Proof.
  intros.
  inversions_typ H; eauto.
Qed.


(* Lemma 5.102 TRedEx *)
Lemma typ_approx_m_exists δ T1:
  exists T2,
  T1 ↝ᵐ⦇δ⦈ T2.
Proof.
  revert δ.
  (* Using induction which opens nested types *)
  pose proof (typ_opening_ind_all T1).
  induction H; sympl; intros; eauto.
  -  (* dec_fld *)
    destruct (IHtyp_opening_ind1 (─δ)) as [T3 ?].
    destruct (IHtyp_opening_ind2 δ) as [T4 ?].
    exists (typ_rcd (dec_fld a T3 T4)).
    eauto.
  - (* dec_typ *)
    pick_fresh z.
    destruct (H0 z (─δ)) as [T3 ?].
    destruct (H2 z δ) as [T4 ?].
    exists (typ_rcd (dec_typ A (close z T3) (close z T4))); eauto 7.
  - (* typ_and *)
    destruct (IHtyp_opening_ind1 δ) as [T3 ?].
    destruct (IHtyp_opening_ind2 δ) as [T4 ?].
    eauto.
  - (* typ_or *)
    destruct (IHtyp_opening_ind1 δ) as [T3 ?].
    destruct (IHtyp_opening_ind2 δ) as [T4 ?].
    eauto.
Qed.

(* Lemma 5.103 ERedEx *)
Lemma typ_approx_e_exists δ T1 y:
  exists T2,
  T1 ↝ᵉ⦇δ⦈ T2 /\
  nosel T2 y.
Proof.
  revert δ.
  pose proof (typ_opening_ind_all T1).
  induction H; intros; eauto; try solve [destruct δ; eauto].
  - (* typ_fun *)
    assert (Happrox_fun : forall (F : ctx) (atg : approx_target)
                            (T1 T2 : typ),
               atg = ᵐ \/ atg = ᵉ ->
               F ⊢ typ_fun T1 T2
                 ↝atg⦇δ⦈ (if mode_has_fun' typing_mode
                          then typ_approx_extreme_nocap δ
                          else typ_approx_extreme δ)) by auto.
    destruct (mode_has_fun' typing_mode); eexists.
    all: split; eauto. all: destruct δ. all: eauto.
  - (* dec_fld *)
    intros.
    destruct (IHtyp_opening_ind1 (─δ)) as [T3 [? ?]].
    destruct (IHtyp_opening_ind2 δ) as [T4 [? ?]].
    exists (typ_rcd (dec_fld a T3 T4)).
    split; eauto.
    apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - (* dec_typ *)
    simpl; intros.
    pick_fresh z.
    destruct (H0 z (─δ)) as [T3 ?].
    destruct (H2 z δ) as [T4 ?].
    exists (typ_rcd (dec_typ A (close z T3) (close z T4))).
    destruct_all. split.
    constructor; eauto.
    apply (nosel_bin (typ_binary_rcd_typ A)); eauto using nosel_close.
  - (* typ_and *)
    intros.
    destruct (IHtyp_opening_ind1 δ) as [T3 [? ?]].
    destruct (IHtyp_opening_ind2 δ) as [T4 [? ?]].
    exists (typ_and T3 T4).
    eauto.
  - (* typ_or *)
    intros.
    destruct (IHtyp_opening_ind1 δ) as [T3 [? ?]].
    destruct (IHtyp_opening_ind2 δ) as [T4 [? ?]].
    exists (typ_or T3 T4).
    eauto.
Qed.

End WithTypingMode.

Ltac invert_approx := (match reverse goal with | H: ?T1 ↝ᵐ⦇?δ⦈ ?T3 |- _ => inversions_typ H end).

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Lemma 5.104 TRedU *)
Lemma typ_approx_m_eq δ T1 T2 T3:
  T1 ↝ᵐ⦇δ⦈ T2 ->
  T1 ↝ᵐ⦇δ⦈ T3 ->
  T2 = T3.
Proof.
  revert δ T2 T3.

  pose proof (typ_opening_ind_all T1).
  induction H;  simpl; intros; (* simpl is needed for declarations *)
    try solve [invert_approx; invert_approx; (* invert the approximation to get the target types *)
    repeat progress f_equal; eauto]. (* f_equal to go inside the type *)
  (* dec_typ *)
  invert_approx; invert_approx. pick_fresh z.
  assert((open z T1') = (open z T1'0)) by eauto.
  assert((open z T2') = (open z T2'0)) by eauto.
  eapply open_fresh_injective in H3.
  eapply open_fresh_injective in H4.
  subst. auto.
  all: notin_solve.
Qed.

Lemma typ_approx_e_com δ T1 T2 T3 y:
  T1 ↝ᵉ⦇δ⦈ T2 ->
  T1 ↝ᵉ⦇δ⦈ T3 ->
  nosel T2 y ->
  nosel T3 y ->
  exists T4,
  T2 ↝ᵉ⦇δ⦈ T4 /\
  T3 ↝ᵉ⦇δ⦈ T4 /\
  nosel T4 y.
Proof.
  introv (*Hin*) Hap12 Hap13 Hns2 Hns3.
  revert dependent T3.
  dependent induction Hap12; intros.
  - destruct H; subst.
    +
      inversions_tyq Hns2.
      inversions_typ Hap13; eauto.
      inversions_tyq Hns3.
      autospecialize.
      destruct (IHHap12_1) as [T5 [H15 [H105 Hns5]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 Hns6]]].
      exists (typ_and T5 T6).
      repeat split; eauto.
    +
      inversions_tyq Hns2.
      inversions_typ Hap13; eauto.
      inversions_tyq Hns3.
      autospecialize.
      destruct (IHHap12_1) as [T5 [H15 [H105 Hns5]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 Hns6]]].
      exists (typ_or T5 T6).
      repeat split; eauto.
  - inversions_tyq Hns2.
    inversions_typ Hap13; eauto.
    inversions_tyq Hns3.
    autospecialize.
    destruct δ.
    +
      destruct (IHHap12_1) as [T5 [H15 [H105 Hns5]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 Hns6]]].
      all: exists (typ_rcd (dec_fld a T5 T6)).
      repeat split; eauto. apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
    +
      destruct (IHHap12_1) as [T5 [H15 [H105 Hns5]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 Hns6]]].
      all: exists (typ_rcd (dec_fld a T5 T6)).
      repeat split; eauto. apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - inversions_tyq Hns2.
    inversions_typ Hap13; eauto.
    inversions_tyq Hns3.
    pick_fresh z. assert (z <> y) by eauto.
    eapply (nosel_open 0 H3) in H5.
    eapply (nosel_open 0 H3) in H6.
    eapply (nosel_open 0 H3) in H7.
    eapply (nosel_open 0 H3) in H8.
    forwards T5 H15 H105 Hns5: H0; [ | | | eapply H12 | | ]; eauto.
    forwards T6 H16 H106 Hns6: H2; [ | | | eapply H13 | | ]; eauto.
    all: exists (typ_rcd (dec_typ B (close z T5) (close z T6))).
    repeat split; eauto 6.
    apply (nosel_bin (typ_binary_rcd_typ B)); eauto using nosel_close.
  - inversions_typ Hap13; eauto 8.
  - inversions_typ Hap13; eauto 8.
  - inversions_typ Hap13; eauto 8.
  - inversions_typ Hap13; destruct δ; eauto 8.
  - inversions Hap13. destruct H1; subst; try discriminate.
    destruct (mode_has_fun' typing_mode). all: destruct δ.
    1,3: exists typ_top. 3: exists typ_nocap. 4: exists typ_bot. all: repeat(split); eauto.
  - inversions_typ Hap13; destruct δ; eauto 8.
  - inversions_typ Hap13; eauto 8; destruct δ; eauto 7.
  - inversions_typ Hap13; destruct δ; eauto 8.
Qed.

Lemma typ_approx_m_ctx δ G1 G2 T1 T2:
  G1 ⊢ T1 ↝ᵐ⦇δ⦈ T2 ->
  G2 ⊢ T1 ↝ᵐ⦇δ⦈ T2.
Proof.
  introv Hap.
  dependent induction Hap; eauto.
Qed.

Lemma typ_approx_e_ctx δ G1 G2 T1 T2:
  G1 ⊢ T1 ↝ᵉ⦇δ⦈ T2 ->
  G2 ⊢ T1 ↝ᵉ⦇δ⦈ T2.
Proof.
  introv Hap.
  dependent induction Hap; eauto.
Qed.

End WithTypingMode.
