Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import AbstractSyntax Defs  GeneralTyping EWeakening Equivalence Indep Substitution.

(** * Definition dereferencing *)

(* Dereference is replacing an object reference w by the corresponding
   object location y, as assigned by the environment e.
   It happens when putting fields on the heap.
   While in a literal, field definitions contain references,
   on the heap, fields contain locations.
   (Methods and functions on the heap still contain references )
*)
Fixpoint deref_base A (su: SubstVar A) (e: renv) (X: A): A :=
  match e with
  | nil => X
  | cons (w,y) ys => (deref_base su ys (subst_var w y X))
  end.

Class Deref A :=
  deref (e: renv) (a: A): A.

Instance deref_var: Deref var := deref_base VarSubstVar.
Instance deref_avar: Deref avar := deref_base AvarSubstVar.
Hint Resolve deref_var deref_avar: core.

Instance deref_def: Deref def :=
fun e d => match d with
| def_fld l x => def_fld l (deref e x)
| def_met l t => def_met l t
| def_typ l T => def_typ l T
end.
Fixpoint deref_defs_fix(e: renv)(ds: defs) :=
match ds with
| defs_nil => defs_nil
| defs_cons ds d => defs_cons (deref_defs_fix e ds) (deref_def e d)
end.
Instance deref_defs: Deref defs := deref_defs_fix.

Lemma deref_base_push A (AS: SubstVar A) (e: renv) x y (X: A):
  deref_base AS (e & x ~ y) X =
  deref_base AS e (subst_var x y X).
Proof.
  rewrite <- cons_to_push.
  auto.
Qed.

Lemma deref_avar_comm (e: renv) (X: var):
  deref e (avar_f X) =
  avar_f (deref e X).
Proof.
  revert X.
  induction e.
  - compute. auto.
  - intros. destruct a. rewrite cons_to_push.
    unfold deref, deref_var, deref_avar in *.
    rewrite deref_base_push.
    rewrite deref_base_push.
    unfold subst_var.
    unfold AvarSubstVar at 2.
    auto.
Qed.

Lemma deref_avar_b (e: renv) n:
  deref e (avar_b n) =
  avar_b (n).
Proof.
  induction e.
  - compute. auto.
  - intros. destruct a. rewrite cons_to_push.
    unfold deref, deref_var, deref_avar in *.
    rewrite deref_base_push.
    unfold subst_var.
    unfold AvarSubstVar at 2.
    auto.
Qed.

Lemma deref_defs_cons_comm e ds d:
  deref_defs e (defs_cons ds d) =
  defs_cons (deref_defs e ds) (deref_def e d).
Proof.
  induction ds; unfold deref_defs; auto.
Qed.

(* Free variables and dereferencing *)
Class DerefInFv (A: Set) (AD: Deref A) (AF: FreeVar A):=
  deref_in_fv x e (d: A):
    x \in fv (deref e d) ->
    x \in fv d \/ x \in fv_env e.

Instance deref_in_fv_def: DerefInFv deref_def _.
Proof.
  hnf. intros.
  destruct d; unfold deref_def in H.
  - auto.
  - revert H. revert a. induction e using env_ind.
    + rewrite empty_def in *. auto.
    + intros. unfold deref, deref_avar in H. rewrite deref_base_push in H.
      sympl in *.
      destruct (subst_cases2_avar x0 v a) as [[Heq Hsubst] | [Heq Hsubst]].
      * right. subst.
        sympl in *.
        rewrite Hsubst in H.
        rewrite fv_in_values_cons.
        rewrite in_union.
        destruct (IHe (avar_f v) H).
        --  right.
            rewrite fv_env_single. unfold fv in *.
      unfold VarFreeVar in *.
      unfold AvarFreeVar in H0. auto.
        -- left. auto.
      * rewrite Hsubst in *.
        destruct (IHe a H).
        -- left. auto.
        -- right.
        rewrite fv_in_values_cons.
        rewrite in_union.
        left. auto.
  - auto.
Qed.


Lemma deref_in_fv_defs x e (ds: defs):
  x \in fv (deref e ds) ->
  x \in fv ds \/ x \in fv_env e.
Proof.
  intros.
  induction ds.
  - auto.
  - rewrite deref_defs_cons_comm in *.
    unfold fv in *. unfold DefsFreeVar in *.
    rewrite fv_defs_eq in *.
    rewrite in_union in *.
    destruct H.
    specialize (IHds H).
    destruct IHds.
    left. auto. auto.
    apply deref_in_fv in H.
    destruct H.
    auto. auto.
Qed.

(* Dereferencing and opening a variable *)
Lemma deref_avar_open e w y (a: avar):
  deref (e & w ~ y) (open w a) =
  deref (e & w ~ y) (open y a).
Proof.
  rewrite <- cons_to_push.
  unfold deref, deref_avar, deref_base.
  rewrite subst_open_commut.
  rewrite subst_open_commut.
  rewrite (substi_eq_var w y).
  rewrite (substi_toeq_var w y).
  auto.
Qed.

Lemma deref_avar_open2 e w y (a: avar):
  a <> (avar_f w) ->
  deref (e & w ~ y) (open w a) =
  deref (e) (open y a).
Proof.
  intro.
  rewrite <- cons_to_push.
  unfold deref, deref_avar, deref_base.
  rewrite subst_open_commut.
  rewrite (substi_eq_var w y).
  unfold subst_var, AvarSubstVar. destruct a. auto. assert (v <> w). intro;subst. apply H. auto. rewrite substi_ne_var; auto.
Qed.

Lemma deref_defs_empty d:
  deref_defs empty d = d.
Proof.
  induction d; simpl; auto.
  rewrite IHd. destruct d0; simpl; auto. destruct a; simpl; rewrite empty_def; compute; auto.
Qed.

Lemma deref_notin_dom e y:
  y \notin dom e ->
  deref e (avar_f y) = (avar_f y).
Proof.
  intro.
  induction e.
  unfold deref, deref_avar, deref_base. auto.
  destruct a.
  unfold deref, deref_avar, deref_base in *. 
  rewrite cons_to_push in H.
  assert (y # e). auto.
  assert (y <> v). intro; subst. rewrite (dom_push) in H. apply H. rewrite in_union. left. rewrite in_singleton. auto.
  rewrite subst_fresh.

  auto. compute. rewrite in_singleton. intro. apply H1. auto.
Qed.
Lemma deref_var_notin_dom e y:
  y \notin dom e ->
  deref e y = y.
Proof.
  intro.
  induction e.
  unfold deref, deref_var, deref_base. auto.
  destruct a.
  unfold deref, deref_var, deref_base in *. 
  rewrite cons_to_push in H.
  assert (y # e). auto.
  assert (y <> v). intro; subst. rewrite (dom_push) in H. apply H. rewrite in_union. left. rewrite in_singleton. auto.
  rewrite subst_fresh.

  auto. compute. rewrite in_singleton. intro. apply H1. auto.
Qed.

(* Lemma 5.70 (DeInv). *)
Lemma deref_defs_has_epush_open_neq_inv e1 w1 y1 d a y:
  y <> y1 ->
  defs_has (deref_defs (e1 & w1 ~ y1) (open w1 d)) (def_fld a (avar_f y)) ->
  exists a0, deref (e1 & w1 ~ y1) (open w1 a0) = avar_f y /\ defs_has d (def_fld a a0).
Proof.
  intros.

  induction d.
  - sympls. inversions H0.
  - sympls. unfold defs_has, get_def in H0. cases_if.
    + inversions C. destruct d0.
      * sympls. inversions H2.
      * sympls. inversions H2.
        inversions H0.
        exists a0. split; eauto. compute. cases_if. auto.
      * sympls. inversions H2.
    + assert (defs_has (deref_defs (e1 & w1 ~ y1) (open w1 d)) (def_fld a (avar_f y))).
      compute; eauto.
      sympls.
      specialize (IHd H1).
      destructc IHd as w0 Hbi Hhas.
      exists w0. split; eauto.
      assert (label_of d0 <> label_fld a).
      intro. rewrite <- H2 in C.
      apply C. destruct d0; sympls; eauto.
      eapply defs_has_n; eauto.
Qed.

Lemma renv_idempotent_push_fv_notin e x v:
  renv_idempotent (e & x ~ v) ->
  v # e.
Proof.
  introv Hidm.
  unfold renv_idempotent in Hidm.
  intro.
  eapply (Hidm v). simpl_dom. rewrite in_union; auto.
  eapply renv_push_in_fv_r.
Qed.

Lemma deref_var_cases e (x: var):
  renv_idempotent e ->
  ok e ->
  (deref e x = x /\ x # e) \/
  (exists y, deref e x = y /\ binds x y e).
Proof.
  intros.
  unfold deref, deref_var.  
  induction e using env_ind; intros; sympls.
  - left. split. rewrite  empty_def. simpls; auto. eauto.
  - destruct IHe; eauto.
    eapply renv_idempotent_concat_inv; eauto.
    destruct_all.
    destruct (classicT(x=x0)).
    + subst.
      right.
      exists v.
      splits.
      rewrite deref_base_push. rewrite subst_eq_var.
      eapply deref_var_notin_dom.
      eapply renv_idempotent_push_fv_notin; eauto.
      eauto.
    + left; splits; eauto.
      rewrite deref_base_push.
      rewrite subst_ne_var; auto.
    + right. destruct_all. exists x1.
      assert (x <> x0). {
        intro. subst. eapply binds_to_dom in H2. eapply ok_push_inv in H0. destruct H0. contradiction.
      }
      splits.
      rewrite deref_base_push. rewrite subst_ne_var; eauto.
      eapply binds_concat_left_ok; eauto.
Qed.

Lemma deref_avar_cases e (a: avar):
  renv_idempotent e ->
  ok e ->
  (deref e a = a /\ ((exists n, a = (avar_b n)) \/ (exists x, a = (avar_f x) /\ x # e))) \/
  (exists x y, a = (avar_f x) /\ deref e x = y /\ binds x y e).
Proof.
  intros.
  destruct a.
  - left. splits. rewrite deref_avar_b. auto. left. exists n. auto.
  - destruct (deref_var_cases v H H0).
    + destruct_all. rewrite deref_avar_comm. rewrite H1. left. splits. auto. right. exists v. auto.
    + destruct_all. right. exists v x; auto.
Qed.

