Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.
Require Import AbstractSyntax TypingMode GeneralTyping Equivalence Context.

(** * Environment weakening *)

(* EWeakening means that typing is preserved when new variables are added
   to renv e. *)

(** ** Straight formulation of EWeakening *)
(** Variant without typing context *)
Class EWeakeningE (A B: Set) (property: EnvProperty A B) :=
  ewkne e1 e2 e3 (a: A) T:
    property (e1 & e3) a T ->
    ok (e1 & e2) ->
    property (e1 & e2 & e3) a T.
(** Variant with typing context *)
Class EWeakening (A B: Set) (property: TypingProperty A B) :=
  ewkn G Go e1 e2 e3 (a: A) T:
    property G Go (e1 & e3) a T ->
    ok (e1 & e2) ->
    property G Go (e1 & e2 & e3) a T.
Hint Unfold EWeakening EWeakeningE : core.

Section EWeakeningELemmata.
Context {A: Set} {AT: TypingEquiv A} {AW: EWeakeningE AT}.

(** Variant of EWeakening for equivalence with a single variable appended at the end *)
Lemma ewkne_bind(e: renv) (v1 v2: A):
  (typing_equiv e v1 v2) ->
  forall v3 v4,
  ok e ->
  v3 # e ->
  typing_equiv (e & v3 ~ v4) v1 v2.
Proof.
  intros.
  rewrite <- (concat_empty_r (e & v3 ~ v4)).
  eapply ewkne; eauto.
  rewrite concat_empty_r; auto.
Qed.

End EWeakeningELemmata.

Section WeakeningLemmata.
Context {A B: Set} {P: TypingProperty A B} {AW: EWeakening P}.

(** Variant of EWeakening with a single variable appended at the end *)
Lemma ewkn_push G Go e v T x y:
  P G Go e v T ->
  ok e ->
  x # e ->
  P G Go (e & x ~ y) v T
.
Proof.
  intros.
  rewrite <- concat_empty_r at 1.
  eapply ewkn; eauto.
  rewrite concat_empty_r.
  assumption.
Qed.

(** Variant of EWeakening where the initial environment is empty *)
Lemma ewkn_empty(G:ctx) Go (e: renv) (a: A) (b: B):
    P G Go renv_empty a b ->
    ok e ->
    P G Go e a b.
Proof.
  intros.
  rewrite <- (concat_empty_r e).
  rewrite <- (concat_empty_l e).
  eapply ewkn; eauto.
  rewrite concat_empty_r; auto.
  rewrite concat_empty_l; auto.
Qed.
End WeakeningLemmata.

(** Formulation of EWeakening for proofs by induction *)
(** Variant without typing context *)
Definition EWeakeningEInd (A B: Set) (property: EnvProperty A B) :=
   (forall (e: renv) (a: A) (b: B), (property e a b) -> forall e1 e2 e3,
    e = e1 & e3 ->
    ok (e1 & e2) ->
    (property (e1 & e2 & e3) a b)).
(** Variant with typing context *)
Definition EWeakeningInd (A B: Set) (property: TypingProperty A B) :=
   (forall (G:ctx) Go (e: renv) (a: A) (b: B), (property G Go e a b) -> forall e1 e2 e3,
    e = e1 & e3 ->
    ok (e1 & e2) ->
    (property G Go (e1 & e2 & e3) a b)).
(** Variants for typing and equivalence *)
Definition EWeakeningTInd (A: Set) (AT: Typing A) := EWeakeningInd obj_typing.
Definition EWeakeningIncapInd (_: ReferenceIncapacitate) :=
  EWeakeningInd obj_ref_incap.
Definition EWeakeningCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C) :=
  EWeakeningInd obj_cap_bnd.
Definition EWeakeningDInd (A: Set) (AT: DecTyping A) := EWeakeningInd obj_dec_typing.
Definition EWeakeningTqInd (A: Set) (AT: TypingEquiv A) := EWeakeningEInd typing_equiv.
Hint Unfold EWeakeningEInd EWeakeningInd EWeakeningTInd EWeakeningDInd EWeakeningTqInd: core.

(** ** Instances of EWeakening *)

(** Variable equivalence *)
Lemma ewkn_tyq_var_ind: EWeakeningTqInd tyq_var.
Proof.
  unfold EWeakeningTqInd, EWeakeningEInd.
  intros.
  induction H; subst; auto.
  - apply tyq_var_binds.
    apply binds_weaken2; auto.
  - eapply tyq_var_trans; auto.
Qed.

Instance ewkn_tyq_var: EWeakeningE tyq_var.
Proof.
  instance_by_ind EWeakeningE ewkn_tyq_var_ind.
Qed.

(** Type equivalence *)
Lemma tyq_ewkn_typ_ind: EWeakeningTqInd tyq_typ.
Proof.
  unfold EWeakeningTqInd, EWeakeningEInd.
  intros.
  induction H; subst; eauto using ewkne.
Qed.

Instance tyq_ewkn_typ: EWeakeningE tyq_typ.
Proof.
  instance_by_ind EWeakeningE tyq_ewkn_typ_ind.
Qed.

(** Variable typing and subtyping *)
Lemma ewkn_ts_ind m:
  EWeakeningTInd (ty_var m) /\
  EWeakeningInd subtyping /\
  EWeakeningIncapInd (ty_incap m).
Proof.
  apply ts_mutind; intros; eauto.
  - eapply subtyp_tyq. eapply tyq_ewkn_typ_ind; eauto.
Qed.

Instance ewkn_ty_var m: EWeakening (ty_var m).
Proof.
  instance_by_ind EWeakening ewkn_ts_ind.
Qed.
Hint Resolve ewkn_ty_var: core.

Instance ewkn_sub m: EWeakening (subtyp m).
Proof.
  instance_by_ind EWeakening ewkn_ts_ind.
Qed.
Hint Resolve ewkn_sub: core.

Instance ewkn_ty_incap m: EWeakening (ty_incap m).
Proof.
  instance_by_ind EWeakening ewkn_ts_ind.
Qed.

(** Capability bound extraction *)
Lemma ewkn_capbnd_ind m: forall C, EWeakeningCapbndInd (ty_capbnd m C).
Proof.
  intros. hnf. induction 1; eauto.
  introv Heq Hok. subst.
  apply ty_capbnd_sel with (T := T) (S := S); eauto.
  eapply ewkn_ty_var; eauto.
Qed.

Instance ewkn_ty_capbnd m: forall C, EWeakening (ty_capbnd m C).
Proof. intros. hnf. intros. eapply ewkn_capbnd_ind; eauto. Qed.

Instance ewkn_viewpoint_adapt m:
  forall x, EWeakening (viewpoint_adapt m x).
Proof.
  hnf. intros. hnf. introv Hadapt Hok.
  pose proof ewkn_ty_incap as Hicp; hnf in Hicp.
  pose proof ewkn_ty_capbnd as Hbnd; hnf in Hbnd.
  hnf in Hadapt. hnf.
  destruct (mode_viewpoint_adaptation' m); eauto.
  inversions Hadapt; econstructor.
  - eapply Hicp; eauto.
  - eapply Hbnd; eauto.
Qed.

(** Term, literal and definition typing *)
Lemma ewkn_rules_ind m:
  EWeakeningTInd (ty_trm m) /\
  EWeakeningTInd (ty_lit m) /\
  EWeakeningDInd (ty_def m) /\
  EWeakeningTInd (ty_defs m).
Proof.
  (* Bring EWeakening for ty_var and subtyp into the context
    to be used by eauto. *)
  destruct (ewkn_ts_ind m). destruct H0.
  apply rules_mutind; intros; subst; eauto using tyq_ewkn_typ_ind.
  - subst. eapply ty_trm_read; eauto. eapply ewkn_viewpoint_adapt; eauto.
  - subst. eapply ty_trm_write; eauto.
Qed.

Instance ewkn_ty_trm m: EWeakening (ty_trm m).
Proof.
  instance_by_ind EWeakening ewkn_rules_ind.
Qed.
Hint Resolve ewkn_ty_trm: core.
Instance ewkn_ty_def m: EWeakening (ty_def m).
Proof.
  instance_by_ind EWeakening ewkn_rules_ind.
Qed.
Instance ewkn_ty_defs m: EWeakening (ty_defs m).
Proof.
  instance_by_ind EWeakening ewkn_rules_ind.
Qed.
Instance ewkn_ty_lit m: EWeakening (ty_lit m).
Proof.
  instance_by_ind EWeakening ewkn_rules_ind.
Qed.
