Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import AbstractSyntax .

(** * Equivalence of variables and types *)

Class TypingEquiv (A : Set) := typing_equiv : renv -> A -> A -> Prop.
(* U+2248 ALMOST EQUAL TO *)
Notation "e ⊢ a '≈' b" := (typing_equiv e a b) (at level 40, a at level 59).
Hint Unfold typing_equiv: core.

(** Typing equivalence for variables *)
Inductive tyq_var: TypingEquiv var :=
| tyq_var_refl(e: renv)(v: var):
  typing_equiv e v v
| tyq_var_binds(e: renv)(v1 v2: var):
  renv_binds v1 v2 e ->
  typing_equiv e v1 v2
| tyq_var_sym(e: renv)(v1 v2: var):
  typing_equiv e v1 v2 ->
  typing_equiv e v2 v1
| tyq_var_trans(e: renv)(v1 v2 v3: var):
  typing_equiv e v1 v2 ->
  typing_equiv e v2 v3 ->
  typing_equiv e v1 v3
.

Hint Constructors tyq_var : core.
Existing Instance tyq_var.

Instance tyq_var_equivalence e: RelationClasses.Equivalence (tyq_var e) := {
  Equivalence_Reflexive := (@tyq_var_refl e);
  Equivalence_Symmetric := (@tyq_var_sym e);
  Equivalence_Transitive := (@tyq_var_trans e);
}.

(* Typing equivalence for types *)
Inductive tyq_typ: TypingEquiv typ :=
| tyq_typ_nulary(e: renv) T:
  typ_nulary T ->
    typing_equiv e T T
| tyq_typ_andor T (e: renv) t11 t12 t21 t22:
  typ_andor T ->
    typing_equiv e t11 t21 ->
    typing_equiv e t12 t22 ->
    typing_equiv e (T t11 t12) (T t21 t22)
| tyq_typ_fun (e: renv) t11 t12 t21 t22:
    typing_equiv e t11 t21 ->
    (forall x, typing_equiv e (open x t12) (open x t22)) ->
    typing_equiv e (typ_fun t11 t12) (typ_fun t21 t22)
| tyq_typ_fld (e: renv) a t11 t12 t21 t22:
    typing_equiv e t11 t21 ->
    typing_equiv e t12 t22 ->
    typing_equiv e (typ_rcd (dec_fld a t11 t12)) (typ_rcd (dec_fld a t21 t22))
| tyq_typ_typ (e: renv) B t11 t12 t21 t22:
    (forall x, typing_equiv e (open x t11) (open x t21)) ->
    (forall x, typing_equiv e (open x t12) (open x t22)) ->
    typing_equiv e (typ_rcd (dec_typ B t11 t12)) (typ_rcd (dec_typ B t21 t22))
| tyq_typ_sel_normal(e: renv)(v1 v2: var) (v3: avar)(l: typ_normal_label):
    typing_equiv e v1 v2 -> (* This rule must not use avar in receiver, because then opening each side with different variables would be incorrect *)
    typing_equiv e (typ_sel (avar_f v1) (typ_label_normal l) v3) (typ_sel (avar_f v2) (typ_label_normal l) v3)
| tyq_typ_sel(e: renv)(v1 v2: avar)(l: typ_label): (* This rule uses avar in order to have reflexivity. *)
    typing_equiv e (typ_sel v1 l v2) (typ_sel v1 l v2)
| tyq_typ_bnd(e: renv)(t1 t2: typ):
    (forall x, typing_equiv e (open x t1) (open x t2)) ->
    typing_equiv e (typ_bnd t1) (typ_bnd t2)
| tyq_typ_met m (e: renv) t11 t12 t13 t21 t22 t23:
    (forall x, typing_equiv e (open x t11) (open x t21)) ->
    typing_equiv e t12 t22 ->
    (forall x y, typing_equiv e (open_met x y t13) (open_met x y t23)) ->
    typing_equiv e (typ_rcd (dec_met m t11 t12 t13)) (typ_rcd (dec_met m t21 t22 t23))
.

Hint Constructors tyq_typ : core.
Existing Instance tyq_typ.

(** Variable properties *)
Lemma renv_binds_notin_false e v1 v2:
  binds v1 v2 e ->
  (renv_notin v1 e \/ renv_notin v2 e) ->
  False.
Proof.
  introv Hbi Hn.
  destruct Hn.
  + (* v1 *)
    pose proof (binds_to_dom Hbi).
    pose proof (notin_union_r H). destruct H1. contradiction.
  + (* v2 *)
    set (renv_bind_in_fv Hbi).
    pose proof (notin_union_r H). destruct H0. contradiction.
Qed.

Lemma tyq_var_notin_eq e v1 v2:
  typing_equiv e v1 v2 ->
  (renv_notin v1 e \/ renv_notin v2 e) ->
  v1 = v2.
Proof.
  intros.
  induction H.
  - apply eq_refl.
  - false. eapply renv_binds_notin_false; eauto.
  - (* sym *)
    apply eq_sym. apply IHtyq_var. destruct H0. right. assumption. left. assumption.
  - (* trans *)
    destruct H0.
    assert (v1=v2) by auto. subst. auto.
    assert (v2=v3) by auto. subst. auto.
Qed.

(** Alternative derivation of variable equivalence *)
Inductive tyq_var_alt: renv -> var -> var -> Prop :=
| tyq_var_alt_eq e v:
    tyq_var_alt e v v
| tyq_var_alt_wy e w y:
    binds w y e ->
    tyq_var_alt e w y
| tyq_var_alt_yw e w y:
    binds w y e ->
    tyq_var_alt e y w
| tyq_var_alt_ww e w1 w2 y:
    binds w1 y e ->
    binds w2 y e ->
    tyq_var_alt e w1 w2.
Hint Constructors tyq_var_alt: core.

Lemma tyq_var_to_tyq_var_alt e v1 v2:
  ok e ->
  disjoint (dom e) (fv_env e) ->
  (typing_equiv e v1 v2 <-> tyq_var_alt e v1 v2).
Proof.
  split; intros.
  - induction H1; eauto.
    + forwards: IHtyq_var; eauto.
      inversions H2; eauto.
    + (* Transitive *)
      (* Use induction on the two parts *)
      forwards Hv12: IHtyq_var; eauto.
      forwards Hv13: IHtyq_var0; eauto.
      (* Invert the alternative derivations and solve the simple cases *)
      inversions Hv12; inversions Hv13; eauto.
      (* Find where v2 is in the dom or fv of e *)
      all: try assert (v2 \in dom e) by (eapply (binds_to_dom); eauto).
      all: try assert (v2 \in fv_env e) by (eapply (renv_bind_in_fv); eauto).
      (* Find contradictions where v2 is both in dom and fv, but they should be disjoint *)
      all: forwards: disjoint_in_notin v2; eauto; try contradiction.
      (* Find cases where v2 is used on the dom side, which is unique *)
      all: try binds_eq; eauto.
  -
    induction H1; eauto.
Qed.

(* Tries to solve incompatible kinds of types *)
Ltac inversions_typs_core :=
 try match goal with
 |  H: typ_nulary _ |- _ => inversions H
 |  H: typ_andor _ |- _ => inversions H
 end.
Ltac inversions_typs := try solve [repeat inversions_typs_core; discriminate].

Ltac invert_tyq :=
 try match goal with
 |  H: typing_equiv _ _ _ |- _ => inversions H
 end.

(** Opening *)
Lemma tyq_typ_open_rec n y e (t1 t2: typ):
  typing_equiv e t1 t2 ->
  typing_equiv e (open_rec n y t1) (open_rec n y t2).
Proof.
  intros.
  revert n.
  induction H; intros; sympls; eauto.
  - (* nulary *)
    inversions H; sympls; eauto.
  - (* andor *)
    inversion H; intros; subst; simpl; auto.
  - (* fun *)
    eapply tyq_typ_fun. eauto. intros. repeat rewrite (open_commut x) by discriminate. eapply H1.
  - (* typ *)
    eapply tyq_typ_typ; intros; repeat rewrite (open_commut x) by discriminate; eauto.
  - eapply tyq_typ_bnd; intros; repeat rewrite (open_commut x) by discriminate; eauto.
  - eapply tyq_typ_met; intros;
    repeat rewrite (open_commut x) by discriminate;
    repeat rewrite (open_commut y0) by discriminate; eauto.
Qed.

Lemma tyq_typ_open y e t1 t2:
  typing_equiv e t1 t2 ->
  typing_equiv e (open y t1) (open y t2).
Proof.
  unfold open, TypOpenable.
  apply tyq_typ_open_rec.
Qed.

Ltac invert_open_rec :=
  match goal with
  | H: _ = open_rec _ _ ?t |- _ => destruct t; sympls; try discriminate
  end.

Lemma tyq_var_reopen_vars e (v1 v2: var) n xs ys (x1 x2: avar):
  fresh (fv x1 \u fv x2\u fv_renv e) (length ys) xs ->
  e ⊢ v1 ≈ v2 ->
  avar_f v1 = open_rec_vars n xs x1 ->
  avar_f v2 = open_rec_vars n xs x2 ->
  exists w1 w2,
  avar_f w1 = open_rec_vars n ys x1 /\
  avar_f w2 = open_rec_vars n ys x2 /\
  e ⊢ w1 ≈ w2.
Proof.
  introv Hfr Htyq Hv1 Hv2.
  pose proof Hfr as Hl. eapply fresh_length in Hl. symmetry in Hl.

  destruct (avar_f_reopen_vars_cases xs x1 ys n Hl Hv1); subst;
  destruct (avar_f_reopen_vars_cases xs x2 ys n Hl Hv2); subst.

  - destruct H; destruct H0; subst. exists v1 v2. splits; eauto.
  - (* Only one side changes - not possible *)
    destruct H. destructc H0 as m w Hx Hm Hmw Hw. subst.
    eapply LibList.Nth_mem in Hm.
    assert (v1 = v2). {
      eapply fresh_in_notin with (X:=fv_renv e) in Hm.
      eapply tyq_var_notin_eq; eauto.
      eauto.
    }
    subst. sympls.
    false. eapply fresh_mem_false with (l:=(length ys)); eauto. fresh_solve.
  - destruct H0. destructc H as m w Hx Hm Hmw Hw. subst.
    eapply LibList.Nth_mem in Hm.
    assert (v1 = v2). {
      eapply fresh_in_notin with (X:=fv_renv e) in Hm.
      eapply tyq_var_notin_eq; eauto.
      eauto.
    }
    subst. sympls.
    false. eapply fresh_mem_false with (l:=(length ys)); eauto. fresh_solve.
  - (* Both sides changed *)
    destructc H as m1 w1 Hx1 Hm1 Hmw1 Hw1.
    destructc H0 as m2 w2 Hx2 Hm2 Hmw2 Hw2.

    exists w1 w2; splits; eauto.
    assert (v1 = v2). {
      eapply LibList.Nth_mem in Hm1.
      eapply fresh_in_notin with (X:=fv_renv e) in Hm1.
      eapply tyq_var_notin_eq; eauto.
      eauto.
    }
    subst.
    assert (m1 = m2). {
      eapply fresh_nth_eq with (l:=length ys) (v:=v2); eauto. fresh_solve.
    }
    subst.
    assert (w1 = w2). {
      eapply LibList.Nth_functional; eauto.
    }
    subst. auto.
Qed.

Lemma tyq_var_reopen_rec e (v1 v2: var) n xs ys (x1 x2: avar):
  xs \notin (fv x1 \u fv x2\u fv_renv e) ->
  e ⊢ v1 ≈ v2 ->
  avar_f v1 = open_rec n xs x1 ->
  avar_f v2 = open_rec n xs x2 ->
  exists w1 w2,
  avar_f w1 = open_rec n ys x1 /\
  avar_f w2 = open_rec n ys x2 /\
  e ⊢ w1 ≈ w2.
Proof.
  intros.
  repeat rewrite <- open_vars_single in *.
  eapply tyq_var_reopen_vars; eauto.
  fresh_solve; eauto.
Qed.

Lemma tyq_var_reopen_tyq_rec e (v1 v2: var) n xs (y1 y2: var) (x1 x2: avar):
  xs \notin (fv x1 \u fv x2\u fv_renv e) ->
  e ⊢ v1 ≈ v2 ->
  e ⊢ y1 ≈ y2 ->
  avar_f v1 = open_rec n xs x1 ->
  avar_f v2 = open_rec n xs x2 ->
  exists w1 w2,
  avar_f w1 = open_rec n y1 x1 /\
  avar_f w2 = open_rec n y2 x2 /\
  e ⊢ w1 ≈ w2.
Proof.
  introv Hn Heqv Heqy Hv1 Hv2.
  destruct (avar_f_open_cases n xs x1 Hv1);
  destruct (avar_f_open_cases n xs x2 Hv2); subst; sympls.
  - exists v1 v2. eauto.
  - false. destruct H0. subst.
    assert (v1 = v2).
    eapply tyq_var_notin_eq; eauto.
    subst.
    rewrite notin_union in Hn. destruct Hn.
    rewrite notin_singleton in H. contradiction.
  - false. destruct H. subst.
    assert (v1 = v2).
    eapply tyq_var_notin_eq; eauto.
    subst.
    rewrite notin_union in Hn. destruct Hn.
    rewrite notin_union in H0. destruct H0.
    rewrite notin_singleton in H0. contradiction.
  - exists y1 y2.
    destruct H; destruct H0; subst.
    sympls; cases_if.
    splits; eauto.
Qed.

Lemma tyq_typ_reopen_vars xs ys e (t1 t2: typ):
  fresh ((fv t1) \u (fv t2) \u fv_renv e) (length ys) xs ->
  typing_equiv e (open_rec_vars 0 xs t1) (open_rec_vars 0 xs t2) ->
  typing_equiv e (open_rec_vars 0 ys t1) (open_rec_vars 0 ys t2).
Proof.
  introv Hn Htyq.
  remember (open_rec_vars 0 xs t1).
  remember (open_rec_vars 0 xs t2).
  reverts xs ys; reverts t1 t2.
  induction Htyq; intros.
  all: inversions_typs_core. (* Destructs nulary and binary types *)
  all: repeat (autorewrite with rewrite_open in *; try invert_open). (* Destructs types *)
  all: eauto.

  - constructor; eauto. eapply IHHtyq; eauto. sympls. fresh_solve.
    eapply IHHtyq0; eauto. sympls. fresh_solve.
  - constructor; eauto. eapply IHHtyq; eauto. sympls. fresh_solve.
    eapply IHHtyq0; eauto. sympls. fresh_solve.

  - eapply tyq_typ_fun; eauto. eapply IHHtyq; eauto. sympls. fresh_solve.
    intros.
    pick_fresh_with x4 (from_list xs).
    rewrite open_vars_push.
    rewrite open_vars_push.
    eapply (H1 x4); eauto.
    rewrite open_vars_push; eauto.
    rewrite open_vars_push; eauto.
    eapply fresh_push; eauto.

  - constructor; eauto. eapply IHHtyq; eauto. sympls. fresh_solve.
    eapply IHHtyq0; eauto. sympls. fresh_solve.

  - eapply tyq_typ_typ.
    intros.
    pick_fresh_with x4 (from_list xs).
    rewrite open_vars_push.
    rewrite open_vars_push.
    eapply (H0 x4); eauto.
    rewrite open_vars_push; eauto.
    rewrite open_vars_push; eauto.
    eapply fresh_push; eauto.
    intros.
    pick_fresh_with x4 (from_list xs).
    rewrite open_vars_push.
    rewrite open_vars_push.
    eapply (H2 x4); eauto.
    rewrite open_vars_push; eauto.
    rewrite open_vars_push; eauto.
    eapply fresh_push; eauto.

  - pose proof Hn as Hl. eapply fresh_length in Hl. symmetry in Hl.
    forwards w1 w2 Hw1 Hw2 Htyqw: tyq_var_reopen_vars; eauto. sympls. rewrite Hl. fresh_solve.
    rewrite <- Hw1.
    rewrite <- Hw2.

    forwards: (reopen_vars_unique). 2: eapply H3. sympls. rewrite Hl. fresh_solve.
    rewrite H0.
    eapply tyq_typ_sel_normal.
    auto.
  - sympls.
    eapply reopen_vars_unique with (ys:=ys) in H0.
    eapply reopen_vars_unique with (ys:=ys) in H1.
    rewrite H0. rewrite H1.
    eauto. fresh_solve. fresh_solve.
  - eapply tyq_typ_bnd.
    intros.
    pick_fresh_with x4 (from_list xs).
    rewrite open_vars_push.
    rewrite open_vars_push.
    eapply (H0 x4); eauto.
    rewrite open_vars_push; eauto.
    rewrite open_vars_push; eauto.
    eapply fresh_push; eauto.
  - eapply tyq_typ_met.
    + intros.
      pick_fresh_with x4 (from_list xs).
      rewrite open_vars_push.
      rewrite open_vars_push.
      eapply (H0 x4); eauto.
      rewrite open_vars_push; eauto.
      rewrite open_vars_push; eauto.
      eapply fresh_push; eauto.
    + eapply IHHtyq; eauto. sympls. fresh_solve.
    +
      intros.
      pick_fresh_with x4 (from_list xs).
      pick_fresh_with x5 (from_list xs).
      rewrite open_vars_push.
      rewrite open_vars_push.
      rewrite open_vars_push.
      rewrite open_vars_push.
      eapply (H3 x4 x5); eauto.
      rewrite open_vars_push; eauto.
      rewrite open_vars_push; eauto.
      rewrite open_vars_push; eauto.
      rewrite open_vars_push; eauto.
      eapply fresh_push; eauto.
      eapply fresh_push; eauto.
      rewrite from_list_cons.
      notin_solve.
Qed.
Lemma tyq_typ_reopen (x y: var) e (t1 t2: typ):
  x \notin ((fv t1) \u (fv t2) \u fv_renv e) ->
  typing_equiv e (open x t1) (open x t2) ->
  typing_equiv e (open y t1) (open y t2).
Proof.
  introv Hn Htyq.
  repeat rewrite <- open_vars_single in *.
  eapply tyq_typ_reopen_vars; eauto.
  fresh_solve; auto.
Qed.
Lemma tyq_typ_reopen_met (x1 y1 x2 y2: var) e (t1 t2: typ):
  x1 \notin ((fv t1) \u (fv t2) \u fv_renv e) ->
  x2 \notin ((fv t1) \u (fv t2) \u fv_renv e \u \{x1}) ->
  typing_equiv e (open_met x1 x2 t1) (open_met x1 x2 t2) ->
  typing_equiv e (open_met y1 y2 t1) (open_met y1 y2 t2).
Proof.
  introv Hn Htyq.
  rewrite <- ! (open_vars_single 1) in *.
  rewrite ! open_vars_push in *.
  eapply tyq_typ_reopen_vars; eauto.
Qed.

(** Equivalence *)
Lemma tyq_typ_trans(e:renv)(t1 t2 t3: typ):
  typing_equiv e t1 t2 ->
  typing_equiv e t2 t3 ->
  typing_equiv e t1 t3.
Proof.
  introv H. revert t3.
  induction H; intros; invert_tyq; inversions_typs; eauto.

  - (* andor *)
    assert (t0 = t21 /\ t1 = t22) as Hte. {
     eapply typ_binary_inv.  3: {apply H3. }
    1,2: auto using typ_andor_binary. }
    destruct Hte; subst.
    specialize (IHtyq_typ t2 H5).
    specialize (IHtyq_typ0 t4 H7).
    assert ((T0 t2 t4) = (T t2 t4)).
    eapply typ_binary_inv_ctor; eauto  using typ_andor_binary.
    rewrite H2. eauto.
Qed.

Lemma tyq_typ_refl(e: renv)(t1:typ):
  typing_equiv e t1 t1.
Proof.
  set (fun d => tyq_typ e (typ_rcd d) (typ_rcd d)).
  induction t1  using typ_mut with (P0 := P); eauto using tyq_typ_open.
  - compute. eapply tyq_typ_typ; eauto using tyq_typ_open.
  - compute. eapply tyq_typ_fld; eauto using tyq_typ_open.
  - compute. eapply tyq_typ_met; eauto using tyq_typ_open_rec.
Qed.
Hint Resolve tyq_typ_refl: core.

Lemma tyq_typ_sym(e: renv)(t1 t2: typ):
  typing_equiv e t1 t2 ->
  typing_equiv e t2 t1.
Proof.
  intros.
  induction H; auto.
Qed.

Instance tyq_typ_equivalence e: RelationClasses.Equivalence (tyq_typ e) := {
  Equivalence_Reflexive := (@tyq_typ_refl e);
  Equivalence_Symmetric := (@tyq_typ_sym e);
  Equivalence_Transitive := (@tyq_typ_trans e);
}.

Lemma tyq_typ_closed_n e T1 T2 n x:
  x \notin fv_renv e ->
  (open_rec n x T1 = T1) ->
  typing_equiv e T1 T2 ->
  (open_rec n x T2 = T2).
Proof.
  introv Hn Hcl Htyq.
  reverts n.
  induction Htyq; intros; sympls; eauto.
  - inversions H; inversions Hcl; sympls; rewrite IHHtyq by eauto; rewrite IHHtyq0 by eauto; auto.
  - inversions Hcl; sympls. rewrite IHHtyq by eauto.
    pick_fresh z. specialize (H1 z).
    forwards: H1; eauto.
    rewrite open_commut. rewrite H4. auto. auto.
    rewrite open_commut in H2 by auto.
    eapply open_fresh_injective in H2.
    rewrite H2. auto.
    eapply notin_fv_open; eauto.
    notin_solve.
  - inversions Hcl; sympls. rewrite IHHtyq by eauto. rewrite IHHtyq0 by eauto. auto.
  - inversions Hcl; sympls.
    pick_fresh z. specialize (H0 z). specialize (H2 z).
    forwards: H0; eauto.
    rewrite open_commut. rewrite H4. auto. auto.
    rewrite open_commut in H3 by auto.
    eapply open_fresh_injective in H3.
    rewrite H3. auto.
    forwards: H2; eauto.
    rewrite open_commut. rewrite H5. auto. auto.
    rewrite open_commut in H6 by auto.
    eapply open_fresh_injective in H6.
    rewrite H6. auto.
    eapply notin_fv_open; eauto.
    notin_solve.
    eapply notin_fv_open; eauto.
    notin_solve.
  - inversions Hcl; sympls. repeat rewrite H1. auto.
  - (* bnd *)
    inversions Hcl; sympls.
    pick_fresh z. specialize (H0 z).
    forwards: H0; eauto.
    rewrite open_commut. rewrite H2. auto. auto.
    rewrite open_commut in H1 by auto.
    eapply open_fresh_injective in H1.
    rewrite H1. auto.
    eapply notin_fv_open; eauto.
    notin_solve.
  - (* met *)
    inversions Hcl; sympls.
    pick_fresh z1. pick_fresh z2. specialize (H0 z1). specialize (H3 z1 z2).
    forwards: H0; eauto.
    rewrite open_commut. rewrite H5. auto. auto.
    rewrite open_commut in H4 by auto.
    eapply open_fresh_injective in H4.
    rewrite H4.
    rewrite IHHtyq; eauto.
    forwards: H3; eauto.
    rewrite open_commut. rewrite (open_commut x). rewrite H7. auto. auto. auto.
    rewrite open_commut in H8 by auto.
    rewrite (open_commut x) in H8 by auto.
    eapply open_fresh_injective in H8.
    eapply open_fresh_injective in H8.
    rewrite H8. auto.
    eapply notin_fv_open; eauto.
    notin_solve.
    eapply notin_fv_open; eauto; eapply notin_fv_open; eauto.
    eapply notin_fv_open; eauto.
    eapply notin_fv_open; eauto.
    notin_solve.
Qed.

Lemma tyq_typ_reopen_closed' e x y T1 T2:
  x \notin fv_renv e \u fv T1 ->
  e ⊢ open x T1 ≈ T2  ->
  e ⊢ open y T1 ≈ open y (close x T2).
Proof.
  introv Hn Htyq.
  eapply tyq_typ_reopen with (x:=x).
  notin_solve. eapply notin_fv_close.
  rewrite open_close.
  pose proof Htyq. eapply (tyq_typ_closed_n 0) with (x0:=x) in H; eauto.
  rewrite H. eapply Htyq. rewrite open_twice. auto.
Qed.
Hint Resolve tyq_typ_reopen_closed': core.

Lemma tyq_typ_nulary_inv T1 T2 e:
  typ_nulary T1 ->
  typing_equiv e T1 T2 ->
  T1 = T2.
Proof.
  intros.
  destruct H;
  inversion H0; subst; auto;
  destruct H1; inversion H.
  Qed.

Lemma tyq_typ_andor_inv T1 t11 t12 t2 e:
  typ_andor T1 ->
  typing_equiv e (T1 t11 t12) t2 ->
  exists t21 t22,
  t2 = (T1 t21 t22) /\ typing_equiv e t11 t21 /\ typing_equiv e t12 t22.
Proof.
  intros.
  destruct H; inversion H0; inversions_typs; subst.
  - inversions_tyq_core. exists t21 t22. eauto.
  - inversions_tyq_core. exists t21 t22. eauto.
Qed.

Lemma tyq_typ_dec_fld_inv t11 t12 t2 a e:
  typing_equiv e (typ_rcd (dec_fld a t11 t12)) t2 ->
  exists t21 t22,
  t2 = (typ_rcd (dec_fld a t21 t22)) /\ typing_equiv e t11 t21 /\ typing_equiv e t12 t22.
Proof.
  intros.
  inversions_tyq H.
  eauto.
Qed.

Lemma tyq_typ_dec_fld_invr t11 t12 t2 a e:
  typing_equiv e t2 (typ_rcd (dec_fld a t11 t12)) ->
  exists t21 t22,
  t2 = (typ_rcd (dec_fld a t21 t22)) /\ typing_equiv e t21 t11 /\ typing_equiv e t22 t12.
Proof.
  intros.
  apply tyq_typ_sym in H.
  destruct (tyq_typ_dec_fld_inv H) as [t21 [t22 [H2 [Htyq1 Htyq2]]]].
  exists t21 t22.
  split; auto; split; auto using tyq_typ_sym.
Qed.

Lemma tyq_typ_fun_inv t11 t12 t2 e:
  typing_equiv e (typ_fun t11 t12) t2 ->
  exists t21 t22,
  t2 = (typ_fun t21 t22) /\ typing_equiv e t11 t21 /\ (forall y, typing_equiv e (open y t12) (open y t22)).
Proof.
  intros.
  inversions_tyq H.
  eauto.
Qed.

Lemma tyq_typ_fun_invr t11 t12 t2 e:
  typing_equiv e t2 (typ_fun t11 t12) ->
  exists t21 t22,
  t2 = (typ_fun t21 t22) /\ typing_equiv e t21 t11 /\ (forall y, typing_equiv e (open y t12) (open y t22)).
Proof.
  intros.
  apply tyq_typ_sym in H.
  destruct (tyq_typ_fun_inv H) as [t21 [t22 [H2 [Htyq1 Htyq2]]]].
  exists t21 t22.
  split; auto; split; auto using tyq_typ_sym.
Qed.

Lemma tyq_typ_dec_typ_inv t11 t12 t2 a e:
  typing_equiv e (typ_rcd (dec_typ a t11 t12)) t2 ->
  exists t21 t22,
  t2 = (typ_rcd (dec_typ a t21 t22)) /\ (forall y, typing_equiv e (open y t11) (open y t21)) /\ (forall y, typing_equiv e (open y t12) (open y t22)).
Proof.
  intros.
  inversions_tyq H.
  eauto.
Qed.

Lemma tyq_typ_dec_met_inv t11 t12 t13 t2 a e:
  typing_equiv e (typ_rcd (dec_met a t11 t12 t13)) t2 ->
  exists t21 t22 t23,
  t2 = (typ_rcd (dec_met a t21 t22 t23)) /\ (forall y, typing_equiv e (open y t11) (open y t21)) /\ typing_equiv e t12 t22 /\ (forall x y, typing_equiv e (open_met x y t13) (open_met x y t23)).
Proof.
  intros.
  inversion H; subst; auto.
  inversion H0.
  destruct H1; subst; inversion H0; auto.
  exists t21 t22 t23.
  auto.
Qed.

Lemma tyq_typ_bnd_inv t11 t2 e:
  typing_equiv e (typ_bnd t11) t2 ->
  exists t21,
  t2 = (typ_bnd t21) /\ (forall y, typing_equiv e (open y t11) (open y t21)).
Proof.
  intros.
  inversion H; subst; auto.
  inversion H0.
  destruct H1; subst; inversion H0; auto.
  exists t0.
  auto.
Qed.

Lemma tyq_typ_sel_inv v a vr t2 e:
  typing_equiv e (typ_sel v a vr) t2 ->
  (
    exists v1 v2 l,
    v = (avar_f v1) /\
    a = (typ_label_normal l) /\
    t2 = (typ_sel (avar_f v2) a vr) /\ typing_equiv e v1 v2
  ) \/ (
    t2 = (typ_sel v a vr)
  ).
Proof.
  intros.
  inversion H; subst; auto.
  inversion H0.
  destruct H1; subst; inversion H0; auto.
  left.
  exists v1 v2 l. split; auto.
Qed.
