(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping Subenvironments Weakening Env Indep Visibility TypingContext.

Section WithTypingMode.
  Context {typing_mode: Mode}.

  (** * Narrowing Lemma *)
  (** The narrowing lemma states that typing is preserved under
      subenvironments. The proof is by mutual induction on term typing,
      definition typing, and subtyping. *)

  Lemma narrow_var : forall G e (x : var) T,
      ctx_binds x T G ->
      forall G',
        e ⊢ G' ⪯ G ->
        G'⁏e ⊢ x ∶ T.
  Proof.
    introv B.
    induction 1; inversion B; subst.
    eapply ty_var_sub.
    eapply ty_var_var.
    eapply ctx_binds_e.
    1,2: eapply weaken_push; eauto using ctx_ok_push_inv_ok,
                             ctx_ok_push_inv_unbound.
    apply ctx_binds_hide_inv in B.
    eapply hiding_hide. auto.
  Qed.

  (** The narrowing lemma, formulated for visibility *)
  Lemma narrow_vis (G : ctx) (x : var) :
    ctx_vis G x ->
    forall (G' : ctx) (e : renv),
      e ⊢ G' ⪯ G ->
      ctx_vis G' x.
  Proof.
    induction 1; introv Hnar; eauto.
    - inversions Hnar; eauto.
    - inversions Hnar.
      destruct (var_decide v2 v); subst; eauto.
  Qed.

  (** Straight formulation of Narrowing *)
  Class Narrow (A B: Set) (property: TypingProperty A B) :=
    narrow G G' Go e (a: A) (b: B):
      property G Go e a b ->
      ctx_ok (ctx_obj G' Go) ->
      (* ctx_ok G' follows from G'⪯ G, but that does not include Go *)
      e ⊢ G' ⪯ G ->
      property G' Go e a b.

End WithTypingMode.

Hint Unfold Narrow: core.

Section WithTypingMode.
  Context (typing_mode: Mode).

  Section NarrowTyLemmata.
    Context {A: Set}{AT: Typing A}{AW: Narrow AT}.

    (** [G ⊢ t: T]                 #<br>#
    [G' subG G]                #<br>#
    [ok G']                    #<br>#
    [―――――――――――――――――]        #<br>#
    [G' ⊢ t: T]
     *)

    (* Variant of narrowing for typing *)
    Lemma narrow_ty:
      forall G G' e (x : A) T,
        G⁏e ⊢ x ∶ T ->
        e ⊢ G' ⪯ G ->
        G'⁏e ⊢ x ∶ T.
    Proof.
      intros.
      eapply narrow; eauto.
    Qed.
  End NarrowTyLemmata.


Lemma narrow_subtyping{AT: Subtyping}{AW: Narrow AT}:
  forall G G' e x T,
  G⁏e ⊢ x <: T ->
  e ⊢ G' ⪯ G ->
  G'⁏e ⊢ x <: T.
Proof.
  intros.
  eapply narrow; eauto.
Qed.


(** Formulation of narrowing for proof by induction *)

Definition NarrowInd (A B: Set) (property: TypingProperty A B) :=
  (forall G Go e (x : A) T,
    (property G Go e x T -> forall G',
      ctx_ok (ctx_obj G' Go) -> (* ctx_ok G' follows from G'⪯ G, but that does not include Go *)
      e ⊢ G' ⪯ G ->
      property G' Go e x T)
  ).

Definition NarrowTInd (A: Set) (t: Typing A) := NarrowInd obj_typing.
Definition NarrowDInd (A: Set) (t: DecTyping A) := NarrowInd obj_dec_typing.
Definition NarrowIncapInd (_ : ReferenceIncapacitate)
  := NarrowInd obj_ref_incap.
Definition NarrowCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C)
  := NarrowInd obj_cap_bnd.

(** Instances of narrowing *)

Lemma narrow_ts_ind:
  NarrowTInd (ty_var _) /\
  NarrowInd subtyping /\
  NarrowIncapInd (ty_incap _).
Proof.
  apply ts_mutind; intros.
  all: match goal with
    | [ B: ctx_binds _ _ _, H: _ ⊢ ?G' ⪯ _ |- _⁏_ ⊢ (_) ∶ _ ] =>
      eauto using narrow_var
    | _ => try fresh_constructor; auto; eauto
    end.
Qed.

End WithTypingMode.

(** The narrowing lemma, formulated only for variable typing. *)
Global Instance narrow_ty_var m: Narrow (ty_var m).
Proof.
  instance_by_ind (@Narrow) narrow_ts_ind.
Qed.
Hint Resolve narrow_ty_var: core.

(** The narrowing lemma, formulated only for subtyping. *)
Global Instance narrow_subtyp m: Narrow (subtyp m).
Proof.
  instance_by_ind @Narrow narrow_ts_ind.
Qed.
Hint Resolve narrow_subtyp: core.

(** The narrowing lemma, formulated only for removing reference
    capabilities. *)
Global Instance narrow_ty_incap m: Narrow (ty_incap m).
Proof.
  instance_by_ind @Narrow narrow_ts_ind.
Qed.
Hint Resolve narrow_ty_incap: core.

(** The narrowing lemma, formulated only for extracting reference
    capabilities. *)
Lemma narrow_ty_capbnd_ind m: forall C, NarrowCapbndInd m (ty_capbnd m C).
Proof.
  intros. hnf. induction 1; eauto.
  introv Hok Hnarrow.
  apply ty_capbnd_sel with (T := T) (S := S); eauto.
  eapply narrow_ty_var; eauto.
Qed.

Instance narrow_ty_capbnd m: forall C, Narrow (ty_capbnd m C).
Proof. intros. hnf. intros. eapply narrow_ty_capbnd_ind; eauto. Qed.
Hint Resolve narrow_ty_capbnd: core.

Instance narrow_viewpoint_adapt mode : forall x,
    Narrow (viewpoint_adapt mode x).
Proof.
  intros. hnf. introv Hadapt Hwk Hok.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' mode); eauto.
  simpl in *.
  inversions Hadapt.
  apply viewpoint_adapt_cap_strip_combine.
  - eapply narrow_ty_incap; eauto.
  - eapply narrow_ty_capbnd; eauto.
Qed.
Hint Resolve narrow_viewpoint_adapt : core.

Section WithTypingMode.
  Context (typing_mode: Mode).

  Lemma narrow_rules_ind:
    NarrowTInd _ (ty_trm _) /\
    NarrowTInd _ (ty_lit _) /\
    NarrowDInd _ (ty_def _) /\
    NarrowTInd _ (ty_defs _).
  Proof.
    apply rules_mutind; intros.
    all: try objctx_destruct.
    all: match goal with
         (* ty_var_var case *)
         | [ B: binds _ _ _, H: _ ⊢ ?G' ⪯ _ |- _⁏_ ⊢ (avar_f _) ∶ _ ] =>
           eauto using narrow_var
         (* fresh_constructor handles cases which add variables to
          the context *)
         | _ => try solve[ fresh_constructor;
                          eauto using narrow_ty, narrow_subtyping,
                          subctx_grow, narrow_vis
                        | econstructor;
                          eauto using narrow_ty, narrow_subtyping,
                          subctx_grow, narrow_vis ]
         end.

    { (* Case ty_trm_read *)
      econstructor;
        eauto using narrow_ty, narrow_subtyping,
        subctx_grow, narrow_vis.
      eapply narrow_viewpoint_adapt; eauto. }

    { (* Case ty_def_met *)
      destruct Go.
      (* This comes after destruct Go so that x is already in the
         environment *)
      all:  apply_fresh_base (@ty_def_met _) gather_vars x1.
      all:  intros_fresh x2; intros.

      all:  apply H; eauto.
      all:  unfold ctx_obj_met, ctx_obj in *;
        try solve[ destruct (mode_visibility' typing_mode); eauto
                 | destruct (mode_visibility' typing_mode);
                   repeat(econstructor; eauto) ]. }

  Qed.

  (** The narrowing lemma, formulated only for term typing. *)
  Global Instance narrow_ty_trm: Narrow (ty_trm _).
  Proof.
    instance_by_ind @Narrow narrow_rules_ind.
  Qed.
  Hint Resolve narrow_ty_trm: core.

  Global Instance narrow_ty_lit: Narrow (ty_lit _).
  Proof.
    instance_by_ind @Narrow narrow_rules_ind.
  Qed.
  Hint Resolve narrow_ty_lit: core.

  Global Instance narrow_ty_def: Narrow (ty_def _).
  Proof.
    instance_by_ind @Narrow narrow_rules_ind.
  Qed.
  Hint Resolve narrow_ty_def: core.

  Global Instance narrow_ty_defs: Narrow (ty_defs _).
  Proof.
    instance_by_ind @Narrow narrow_rules_ind.
  Qed.
  Hint Resolve narrow_ty_defs: core.


  Lemma narrow_ty_trm_subtyp:
    forall G G' e (t : trm) T U,
      G⁏e ⊢ t ∶ T ->
      e ⊢ G' ⪯ G ->
      G'⁏e ⊢ T <: U ->
                  G'⁏e ⊢ t ∶ U.
  Proof.
    intros.
    eauto using ty_trm_sub, narrow_ty.
  Qed.

  Class TyOpenImpliesTyBnd (A B: Set) (property: TypingProperty A B) :=
    ty_open_implies_ty_bnd T (ti: typ_indep_b 0 T) G1 y vk G2 Go e
                           (a: A) (b: B):
      property (G1 § y / vk ~ open y T §& G2) Go e a b ->
      ctx_weakens_hide G2 ->
      property (G1 § y / vk ~ typ_bnd T §& G2) Go e a b.
  Hint Unfold TyOpenImpliesTyBnd: core.

  Section TyOpenImpliesTyBndLemmata.
    Context (A: Set) (AT: Typing A) {AW: TyOpenImpliesTyBnd obj_typing}.

    Lemma ty_open_implies_ty_bnd_push G y vk T e (t: A) U:
      typ_indep_b 0 T ->
      G § y / vk ~ (open y T) ⁏e ⊢ t ∶ U ->
      G § y / vk ~ typ_bnd T ⁏e ⊢ t ∶ U.
    Proof.
      introv Hindep Hty. rewrite <- (ctx_concat_empty_r) at 1.
      eapply (ty_open_implies_ty_bnd Hindep);
        eauto using ctx_concat_empty_r.
    Qed.

  End TyOpenImpliesTyBndLemmata.

  Definition TyOpenImpliesTyBndInd y vk T1 (A B: Set) (property: TypingProperty A B) :=
    (forall G Go e (t : A) T,
        property G Go e t T ->
        (forall G1 G2, ctx_weakens_hide G2 ->
              G = G1 § y / vk ~ open y T1 §& G2 ->
              property (G1 § y / vk ~ typ_bnd T1 §& G2) Go e t T)).
  Definition TyOpenImpliesTyBndTInd y vk T1 (A: Set) (t: Typing A) :=
    TyOpenImpliesTyBndInd y vk T1 obj_typing.
  Definition TyOpenImpliesTyBndDInd y vk T1 (A: Set) (t: DecTyping A) :=
    TyOpenImpliesTyBndInd y vk T1 obj_dec_typing.
  Definition TyOpenImpliesTyBndIncapInd y vk T1
             (_ : ReferenceIncapacitate) :=
    TyOpenImpliesTyBndInd y vk T1 obj_ref_incap.
  Definition TyOpenImpliesTyBndCapbndInd y vk T1
             (C : typ_refcap_label) (_ : CapabilityBound C) :=
    TyOpenImpliesTyBndInd y vk T1 obj_cap_bnd.

  Lemma ty_open_implies_ty_bnd_ts_ind y vk T1:
    typ_indep_b 0 T1 -> (
      TyOpenImpliesTyBndTInd y vk T1 (ty_var _) /\
      TyOpenImpliesTyBndInd y vk T1 subtyping /\
      TyOpenImpliesTyBndIncapInd y vk T1 (ty_incap _)
    ).
  Proof.
    introv Hindep.
    apply ts_mutind.
    all: try solve [ intros; subst; eauto; fresh_constructor; eauto ].
    - intros; subst.
      apply ctx_binds_middle_inv in c as [? | [? | ?]]; eauto;
        destruct_ands; subst.
      eapply ty_rec_elim; eauto.
      all: eapply ty_var_var; eauto;
        eapply ctx_binds_concat_left_hide; eauto.
    - intros; subst.
      apply_fresh (@subtyp_met _) as z; eauto.
      assert (z \notin L) by eauto.
      specialize (H0 _ H3); eauto.
      intros; assert (z \notin L) by eauto; assert (y0 <> z) by eauto;
        assert (y0 \notin L) by eauto.
      specialize (H1 _ H6 _ H7 H8).
      destruct_ands; rewrite <- ctx_concat_push; eauto.
  Qed.

  Instance ty_open_implies_ty_bnd_ty_var: TyOpenImpliesTyBnd (ty_var _).
  Proof.
    instance_by_ind TyOpenImpliesTyBnd ty_open_implies_ty_bnd_ts_ind.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_ty_var: core.

  Instance ty_open_implies_ty_bnd_subtyp: TyOpenImpliesTyBnd (subtyp _).
  Proof.
    instance_by_ind TyOpenImpliesTyBnd ty_open_implies_ty_bnd_ts_ind.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_subtyp: core.

  Instance ty_open_implies_ty_bnd_ty_incap: TyOpenImpliesTyBnd (ty_incap _).
  Proof.
    unfold TyOpenImpliesTyBnd.
    intros. eapply ty_open_implies_ty_bnd_ts_ind; eauto.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_ty_incap: core.


  Lemma ty_open_implies_ty_bnd_ty_capbnd_ind y vk T1:
    typ_indep_b 0 T1 -> (
      forall C, TyOpenImpliesTyBndCapbndInd y vk T1 (ty_capbnd _ C)
    ).
  Proof.
    introv Hindep; intros. hnf.
    induction 1;
      try solve[intros; subst; eauto].
    introv Hwk Heq; subst.
    apply ty_capbnd_sel with (T := T) (S := S); eauto.
    eapply ty_open_implies_ty_bnd_ts_ind; eauto.
  Qed.

  Instance ty_open_implies_ty_bnd_ty_capbnd:
    forall C, TyOpenImpliesTyBnd (ty_capbnd _ C).
  Proof. intros. hnf. intros.
         eapply ty_open_implies_ty_bnd_ty_capbnd_ind;
           eauto using ctx_weakens_to_hide.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_ty_capbnd: core.


  Lemma ty_open_implies_ty_bnd_ty_ind (G1: ctx) y vk T1 (A: Set)
        (AT: Typing A) {AW: TyOpenImpliesTyBnd obj_typing}:
    typ_indep_b 0 T1 ->
    (forall G e (t : A) T,
        typing G e t T ->
        (forall G2, ctx_weakens_hide G2 -> G = G1 § y / vk ~ open y T1 §& G2 ->
               typing (G1 § y / vk ~ typ_bnd T1 §& G2) e t T)).
  Proof.
    intros.
    eapply ty_open_implies_ty_bnd; subst; eauto.
  Qed.

  Instance ty_open_implies_ty_bnd_viewpoint_adapt:
    forall x, TyOpenImpliesTyBnd (viewpoint_adapt typing_mode x).
  Proof.
    hnf. intros. hnf. introv Hindep Hadapt Hwk.
    pose proof ty_open_implies_ty_bnd_ty_incap as Hicp; hnf in Hicp.
    pose proof ty_open_implies_ty_bnd_ty_capbnd as Hbnd; hnf in Hbnd.
    unfold viewpoint_adapt in *.
    destruct (mode_viewpoint_adaptation' typing_mode); eauto.
    hnf in Hadapt. inversions Hadapt; econstructor.
    - eapply Hicp; eauto.
    - eapply Hbnd; eauto.
  Qed.

  Lemma ty_open_implies_ty_bnd_rules_ind:
    forall y vk T1,
      typ_indep_b 0 T1 ->
      (TyOpenImpliesTyBndTInd y vk T1 (ty_trm _) /\
       TyOpenImpliesTyBndTInd y vk T1 (ty_lit _) /\
       TyOpenImpliesTyBndDInd y vk T1 (ty_def _) /\
       TyOpenImpliesTyBndTInd y vk T1 (ty_defs _)).
  Proof.
    introv Hindep.
    apply rules_mutind; intros; subst;
      try solve [ fresh_constructor
                | econstructor; eauto using tychange_vis_ind_emptyctx;
                  eapply ty_open_implies_ty_bnd_ty_var; eauto ].
    - eapply ty_trm_read;
        eauto using tychange_vis_ind_emptyctx.
      + eapply ty_open_implies_ty_bnd_ty_var; eauto.
      + eapply ty_open_implies_ty_bnd_viewpoint_adapt; eauto.
    - eapply ty_trm_write;
        eauto using tychange_vis_ind_emptyctx.
      intros. specialize (o H0).
      all: eapply ty_open_implies_ty_bnd_ty_var; eauto.
    - eapply ty_trm_sub;
        eauto using tychange_vis_ind_emptyctx.
      eapply ty_open_implies_ty_bnd_subtyp; eauto.
    - eapply ty_def_fld; eauto using tychange_vis_ind. unfold ctx_obj.
      destruct Go; try rewrite <- ctx_concat_push;
        eapply ty_open_implies_ty_bnd_ty_var; eauto.
    - unfold ctx_obj_met, ctx_obj in *. destruct Go.
      + eapply ty_def_met; intros; simpl;
          destruct (mode_visibility' typing_mode); 
          repeat rewrite <- ctx_concat_push; eauto.
        specialize (H _ H1 _ H2 H3
                      G1 (ctx_met (ctx_hide G2) xz TZ xc TC)).
        simpl in H; eauto.
      + eapply ty_def_met; intros; simpl;
        destruct (mode_visibility' typing_mode); 
          repeat rewrite <- ctx_concat_push; eauto.
        specialize (H _ H1 _ H2 H3
                      G1
                      (ctx_met_refined_heap
                         (ctx_hide (G2 § x / v_self ~ open x T))
                         xz TZ xc TC x T)).
        simpl in H. eapply H; eauto.
      +  eapply ty_def_met; intros; simpl;
          destruct (mode_visibility' typing_mode); 
          repeat rewrite <- ctx_concat_push; eauto.
        specialize (H _ H1 _ H2 H3
                      G1
                      (ctx_met_refined_heap
                         (ctx_hide G2 )
                         xz TZ xc TC y0 T)).
        simpl in H. eapply H; eauto.
  Qed.

  Instance ty_open_implies_ty_bnd_trm: TyOpenImpliesTyBnd (ty_trm _).
  Proof.
    instance_by_ind TyOpenImpliesTyBnd ty_open_implies_ty_bnd_rules_ind.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_trm: core.

  Instance ty_open_implies_ty_bnd_defs: TyOpenImpliesTyBnd (ty_defs _).
  Proof.
    instance_by_ind TyOpenImpliesTyBnd ty_open_implies_ty_bnd_rules_ind.
  Qed.
  Hint Resolve ty_open_implies_ty_bnd_defs: core.

End WithTypingMode.

Hint Resolve narrow_ty_trm: core.
