(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping Env.
Require Import Coq.Program.Equality.

(** * Subenvironments [G1 ⪯ G2] *)
(** [G1] is a subenvironment of [G2], denoted [e ⊢ G1 ⪯ G2],
    if [dom(G1) = dom(G2)] and for each [x],
    [G1⁏e ⊢ G1(x) <: G2(x)]. *)
Class Subctx := subctx: renv -> ctx -> ctx -> Prop.
Notation "e ⊢ G1 ⪯  G2" := (subctx e G1 G2) (at level 40, G1 at level 59).

Section WithTypingMode.
Context (typing_mode: Mode).

Inductive subctx0: Subctx :=
| subctx_empty e: e ⊢ ctx_empty ⪯  ctx_empty
| subctx_grow: forall e G G' x vk T T',
    e ⊢ G ⪯ G' ->
    ctx_ok (G § x / vk ~ T) ->
    (G⁏e ⊢ T <: T') ->
    e ⊢ G § x / vk ~ T ⪯ G' § x / vk ~ T'
| subctx_stop: forall e G G',
    e ⊢ G ⪯ G' ->
    e ⊢ ctx_stop G ⪯ ctx_stop G'
| subctx_hide: forall e G G',
    e ⊢ G ⪯ G' ->
    e ⊢ ctx_hide G ⪯ ctx_hide G'
.
End WithTypingMode.

Existing Instance subctx0.

Section WithTypingMode.
Context (typing_mode: Mode).
Local Hint Constructors subctx0 : core.

(** If [ok G], then [G ⪯ G].
    Note: [ok(G)] means that [G]'s domain consists of distinct variables.
    [ok] is defined in [TLC.LibEnv.v] *)
Lemma subctx_refl e: forall G, ctx_ok G -> e ⊢ G ⪯ G.
Proof.
  intros G H. induction H; auto.
Qed.
Hint Resolve subctx_refl : core.

Lemma subctx_push: forall e G1 G2 v vk T,
    e ⊢ G1 ⪯ G2 ->
    ctx_ok (G1 § v / vk ~ T) -> ctx_ok (G2 § v / vk ~ T) ->
    e ⊢ (G1 § v / vk ~ T) ⪯ (G2 § v / vk ~ T).
Proof.
  intros.
  rewrite ?ctx_concat_empty_r, ?ctx_concat_assoc, ?ctx_concat_push, ?ctx_concat_stop in *; auto.
Qed.
Hint Resolve subctx_push : core.

Lemma subctx_concat : forall e G1 G2 G3,
    e ⊢ G1 ⪯ G2 ->
    ctx_ok (G1 §& G3) -> ctx_ok (G2 §& G3) ->
    e ⊢ (G1 §& G3) ⪯ (G2 §& G3).
Proof.
  intros. induction G3.
  all: autorewrite with ctx_flatten in *.
  all: auto.
Qed.

Lemma subctx_concat2 : forall e G1 G2 G3 G4,
    e ⊢ G1 ⪯ G2 ->
    ctx_ok (G1 §& G3 §& G4) -> ctx_ok (G2 §& G3 §& G4) ->
    e ⊢ (G1 §& G3 §& G4) ⪯ (G2 §& G3 §& G4).
Proof.
  intros.
  apply subctx_concat.
  apply subctx_concat.
  all: auto.
  all: eapply ctx_ok_concat_l; eauto.
Qed.

(** [G ⊢ S <: U]                      #<br>#
    [ok(G, x: S)] (see [subenv_push]) #<br>#
    [――――――――――――――――――――――――――――――――――]  #<br>#
    [G', x: T subG G, x: T] *)
Lemma subctx_last: forall G e x vk S U,
  G⁏e ⊢ S <: U ->
  ctx_ok (G § x / vk ~ S) ->
  e ⊢ (G § x / vk ~ S) ⪯ (G § x / vk ~ U).
Proof. auto. Qed.
Lemma subctx_last2: forall G e x1 vk1 S1 x2 vk2 S2 U1 U2,
  G⁏e ⊢ S1 <: U1 ->
  (G § x1 / vk1 ~ S1)⁏e ⊢ (open x1 S2) <: (open x1 U2) ->
  ctx_ok (G § x1 / vk1 ~ S1 § x2 / vk2 ~ (open x1 S2)) ->
  e ⊢ (G § x1 / vk1 ~ S1 § x2 / vk2 ~ (open x1 S2)) ⪯ (G § x1 / vk1 ~ U1 § x2 / vk2 ~ (open x1 U2)).
Proof.
  intros.
  apply subctx_grow.
  apply subctx_grow.
  all: auto.
  apply subctx_refl.
  apply ctx_ok_push_inv_ok in H1.
  apply ctx_ok_push_inv_ok in H1. auto.
Qed.

Lemma subctx_ok_fresh : forall e G G',
    e ⊢ G ⪯ G' ->
    ctx_ok G' /\ (forall x, x §# G -> x §# G').
Proof.
  introv H.
  induction H.
  all: split; intros; auto.
  all: destruct IHsubctx0; auto.
  - apply ctx_ok_push. assumption.
    apply H3. inversion H0. auto.
Qed.

Lemma subctx_ok_l : forall e G1 G2,
    e ⊢ G1 ⪯ G2 -> ctx_ok G1.
Proof.
  introv H. induction H; auto.
Qed.

Lemma subctx_ok_r: forall e G1 G2,
    e ⊢ G1 ⪯ G2 -> ctx_ok G2.
Proof.
  apply subctx_ok_fresh.
Qed.

End WithTypingMode.

Hint Resolve subctx_refl : core.
Hint Resolve subctx_push : core.
Hint Resolve subctx_concat : core.
Hint Resolve subctx_concat2 : core.
Hint Resolve subctx_last : core.
Hint Resolve subctx_last2 : core.

Hint Extern 1 (ctx_ok _) =>
match goal with
| [ H : _ ⊢  ?G1 ⪯ ?G2 |- ctx_ok ?G1 ] => apply (subctx_ok_l H)
| [ H : _ ⊢ ?G1 ⪯ ?G2 |- ctx_ok ?G2 ] => apply (subctx_ok_r H)
end : core.
