Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.

(** * Typing modes *)
Inductive mode :=
  | basic
  | basic_no_fun
  | basic_incap
  | basic_sef_typing
  | rodot
  | rodot_top_nm
  | rodot_nm_compl
  | rodot_nm_muttop
  | rodot_nm_compl_muttop
  | rodot_sef.

(* Class for mode, allows declaring an instance in a section where that mode should be implicitly used *)
Class Mode := mode_m: mode. Hint Unfold Mode: core.

(** ** Feature switches. *)

(** Each of the modes defined above can have different features.
   Each feature is defined below as a definition matching on the typing mode,
   returning either a boolean or a value of an enum.

  You should never destruct typing_mode directly, because that would break when a new mode is added.
  Instead, destruct the feature switch like so:
  destruct (mode_visibility' typing_mode).
  Some of the features are not orthogonal.
  In such cases, it might be helpful to keep an equation for the feature like so:
  destruct (mode_visibility' typing_mode) eqn: Hm.
  If you destruct multiple features in this way,
  some of the combinations may not correspond to any actual mode (because the features are not orthogonal).
  In such cases, use the typing_mode_contradiction tactic to solve those impossible cases.
*)

Definition mode_has_fun' m: bool := match m with
| basic | basic_sef_typing => true
| _ => false
end.
Notation mode_has_fun m := ((mode_has_fun' m) = true)(only parsing).

(* Does the mode check reference mutability *)
Definition mode_has_mutability' m: bool := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop | rodot_sef => true
| _ => false
end.
Notation mode_has_mutability m := ((mode_has_mutability' m) = true)(only parsing).

Inductive mode_visibility := mode_visibility_all | mode_visibility_hide_method.

Definition mode_visibility' m: mode_visibility := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop | rodot_sef => mode_visibility_hide_method
| _ => mode_visibility_all
end.
Notation mode_has_visibility m := ((mode_visibility' m) = mode_visibility_hide_method)(only parsing).

Inductive mode_refcaps := mode_refcaps_none | mode_refcaps_mutability.
Definition mode_refcaps' m: mode_refcaps := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop | rodot_sef => mode_refcaps_mutability
| _ => mode_refcaps_none
end.

Inductive mode_refcap_corr := mode_refcap_corr_same | mode_refcap_corr_ref_mutability.
Definition mode_refcap_corr' m: mode_refcap_corr := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop  | rodot_sef => mode_refcap_corr_ref_mutability
| _ => mode_refcap_corr_same
end.

Inductive mode_viewpoint_adaptation := mode_viewpoint_adaptation_same | mode_viewpoint_adapt_mutability.
Definition mode_viewpoint_adaptation' m: mode_viewpoint_adaptation := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop | rodot_sef => mode_viewpoint_adapt_mutability
| _ => mode_viewpoint_adaptation_same
end.

Inductive mode_objlit_refcaps := mode_objlit_refcaps_none | mode_objlit_refcaps_mutable.
Definition mode_objlit_refcaps' m: mode_objlit_refcaps := match m with
| rodot | rodot_top_nm | rodot_nm_compl | rodot_nm_muttop | rodot_nm_compl_muttop | rodot_sef => mode_objlit_refcaps_mutable
| _ => mode_objlit_refcaps_none
end.

(* Does the mode have a Top <: (N|M) subtyping rule? *)
Inductive mode_top_nm := mode_top_nm_none | mode_top_nm_subtype | mode_top_nm_complement.
Definition mode_top_nm' m: mode_top_nm := match m with
| rodot_top_nm | rodot_nm_muttop | rodot_sef => mode_top_nm_subtype
| rodot_nm_compl | rodot_nm_compl_muttop  => mode_top_nm_complement
| _ => mode_top_nm_none
end.
Definition mode_has_top_nm_subtype' m: bool := match (mode_top_nm' m) with
| mode_top_nm_subtype | mode_top_nm_complement => true
| _ => false
end.
Notation mode_has_top_nm_subtype m := ((mode_has_top_nm_subtype' m) = true)(only parsing).
Notation mode_has_top_nm_complement m := ((mode_top_nm' m) = mode_top_nm_complement)(only parsing).
Notation mode_has_top_nm_subtype_boundless m := ((mode_top_nm' m) = mode_top_nm_subtype)(only parsing).

Inductive mode_mut_top := mode_mut_top_none | mode_mut_top_subtype.
Definition mode_mut_top' m: mode_mut_top := match m with
| rodot_nm_muttop | rodot_nm_compl_muttop => mode_mut_top_subtype
| _ => mode_mut_top_none
end.

Notation mode_has_mut_top m := ((mode_mut_top' m) = mode_mut_top_subtype)(only parsing).

(* Whether supertypes of nocap are untouched by incap.
   This option decides whether the layered invertible typing is used.
   The mode basic_incap shows that the layered typing can be used without the other rodot-related options. *)
Inductive mode_incap := mode_incap_none | mode_incap_structural | mode_incap_nocap.
Definition mode_incap' m: mode_incap := match m with
| basic_incap | rodot_top_nm | rodot_nm_muttop | rodot_nm_compl | rodot_nm_compl_muttop | rodot_sef => mode_incap_nocap (* Incap does not affect supertypes of nocap *)
| rodot => mode_incap_structural (* Incap does not use subtyping *)
| _ => mode_incap_none (* Incap does not matter *)
end.

Definition mode_has_incap' m: bool := match (mode_incap' m) with
| mode_incap_none => false
| _ => true
end.
Notation mode_has_incap m := ((mode_has_incap' m) = true)(only parsing).

(* Does the safety proof use a flat definition of invertible typing or a layered one.
   The layered version is required for mode_top_nm and requires mode_incap_nocap *)
Inductive mode_ty_inv := mode_ty_inv_flat | mode_ty_inv_layered.
Definition mode_ty_inv' m: mode_ty_inv := match m with
| basic_incap | rodot_top_nm | rodot_nm_muttop | rodot_nm_compl | rodot_nm_compl_muttop | rodot_sef => mode_ty_inv_layered
| _ => mode_ty_inv_flat
end.

(* The side-effect-free mode changes method subtyping to ensure independent mutability of the argument *)
Inductive mode_sef := mode_sef_none | mode_sef_guarantee.
Definition mode_sef' m : mode_sef := match m with
| rodot_sef |  rodot_nm_muttop | rodot_nm_compl | rodot_nm_compl_muttop  | basic_sef_typing => mode_sef_guarantee
| _ => mode_sef_none
end.
Notation mode_has_sef m := ((mode_sef' m) = mode_sef_guarantee)(only parsing).

(* Solves the goal if there are contradictory conditions on the typing mode. *)
Ltac typing_mode_contradiction typing_mode := (
  (* Look at individual typing modes *)
  destruct typing_mode;
  (* Unfold typing mode conditions.
     Each condition will eliminate some mode by resulting in hypothesis such as false = true. *)
  unfold mode_has_mutability', mode_has_fun', mode_refcaps', mode_visibility',
    mode_refcap_corr', mode_viewpoint_adaptation', mode_ty_inv', mode_incap', mode_sef', mode_objlit_refcaps' in *;
  (* Finds the contradictory hypothesis.
     If the conditions are contradictory, there will be one for each mode. *)
  discriminate
).

(* Version of typing_mode_contradiction that shows the non-contradictory results *)
Ltac info_typing_mode_contradiction typing_mode := (
  (* Look at individual typing modes *)
  destruct typing_mode eqn: ?Hm;
  (* Unfold typing mode conditions.
     Each condition will eliminate some mode by resulting in hypothesis such as false = true. *)
  unfold mode_has_mutability', mode_has_fun', mode_refcaps', mode_visibility',
    mode_refcap_corr', mode_viewpoint_adaptation', mode_ty_inv', mode_incap', mode_sef', mode_objlit_refcaps' in *;
  (* Finds the contradictory hypothesis.
     If the conditions are contradictory, there will be one for each mode. *)
  try discriminate
).

(* Destructs the typing mode and substitutes into various hypotheses.
   Tactic mostly used for backwards compatibility *)
Ltac info_typing_mode_subst :=
  match goal with
  | typing_mode : Mode |- _ => info_typing_mode_contradiction typing_mode;
                             subst; try easy; eauto
  end.

(** ** Dependencies between features *)

Lemma typing_mode_dep_ty_inv_flat_incap_structural typing_mode:
  mode_ty_inv' typing_mode = mode_ty_inv_flat ->
  (mode_incap' typing_mode = mode_incap_structural \/ mode_incap' typing_mode = mode_incap_none).
Proof.
  introv Hm.
  destruct (mode_incap' typing_mode) eqn: Hm'; auto.
  typing_mode_contradiction typing_mode.
Qed.

Lemma typing_mode_dep_ty_inv_layered_incap_nocap typing_mode:
  mode_ty_inv' typing_mode = mode_ty_inv_layered ->
  mode_incap' typing_mode = mode_incap_nocap.
Proof.
  introv Hm.
  destruct (mode_incap' typing_mode) eqn: Hm'; auto;
  typing_mode_contradiction typing_mode.
Qed.

Hint Resolve typing_mode_dep_ty_inv_flat_incap_structural typing_mode_dep_ty_inv_layered_incap_nocap: typing_mode_deps.
