Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax GeneralTyping TypingContext.

(** * Variable visibility *)

(* Removing ctx_hide from a context does not affect typing *)
Class Unhiding (A B: Set) (property: TypingProperty A B) :=
  unhiding G1 G2 Go e (a: A) (b: B):
    property ((ctx_hide G1) §& G2) Go e a b ->
    property (G1 §& G2) Go e a b.
Hint Unfold Unhiding: core.

(* Adding ctx_hide to a context does not affect variable typing and subtyping *)
Class Hiding (A B: Set) (property: TypingProperty A B) :=
  hiding G1 G2 Go e (a: A) (b: B):
    property (G1 §& G2) Go e a b ->
    property ((ctx_hide G1) §& G2) Go e a b.
Hint Unfold Hiding: core.

Section UnhidingLemmata.
Context {A B: Set}{P: TypingProperty A B}{AW: Unhiding P}.

Lemma unhiding_hide G Go e a b:
  P (ctx_hide G) Go e a b ->
  P G Go e a b.
Proof.
 rewrite <- ctx_concat_empty_r at 1.
 apply unhiding.
Qed.

Lemma unhiding_vis_hide typing_mode G1 G2 Go e a b:
  P ((vis_hide typing_mode G1) §& G2) Go e a b ->
  P (G1 §& G2) Go e a b.
Proof.
  destruct (mode_visibility' typing_mode);
  eauto.
Qed.

End UnhidingLemmata.

Section HidingLemmata.
Context {A B: Set}{P: TypingProperty A B}{AW: Hiding P}.

Lemma hiding_hide G Go e a b:
  P G Go e a b ->
  P (ctx_hide G) Go e a b.
Proof.
 rewrite <- ctx_concat_empty_r at 1.
 apply hiding.
Qed.

Lemma hiding_vis_hide typing_mode G Go e a b:
  P G Go e a b ->
  P (vis_hide typing_mode G) Go e a b.
Proof.
 destruct (mode_visibility' typing_mode); auto.
 apply hiding_hide.
Qed.


End HidingLemmata.

Lemma unhiding_vis G1 G2 x:
  ctx_vis (ctx_hide G1 §& G2) x ->
  ctx_vis (G1 §& G2) x.
Proof.
  intros.
  induction G2;
  inversions H.
  - constructor.
  - apply ctx_vis_n; eauto.
Qed.

(* Inductve definition *)
Definition UnhidingInd (A B: Set) (property: TypingProperty A B) :=
  forall G Go e (a: A) (b: B),
    property G Go e a b ->
    forall G1 G2,
      G = (ctx_hide G1) §& G2 ->
      property (G1 §& G2) Go e a b.

Hint Unfold UnhidingInd: core.
Definition UnhidingTInd (A: Set) (t: Typing A) := UnhidingInd obj_typing.
Definition UnhidingDInd (A: Set) (t: DecTyping A) := UnhidingInd obj_dec_typing.
Definition UnhidingIncapInd (_ : ReferenceIncapacitate) :=
  UnhidingInd obj_ref_incap.
Definition UnhidingCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C) :=
  UnhidingInd obj_cap_bnd.
Hint Unfold UnhidingTInd UnhidingDInd
     UnhidingIncapInd UnhidingCapbndInd: core.

Definition HidingInd (A B: Set) (property: TypingProperty A B) :=
  forall G Go e (a: A) (b: B),
    property G Go e a b ->
    forall G1 G2,
      G = G1 §& G2 ->
      property ((ctx_hide G1) §& G2) Go e a b.

Hint Unfold HidingInd: core.
Definition HidingTInd (A: Set) (t: Typing A) := HidingInd obj_typing.
Definition HidingDInd (A: Set) (t: DecTyping A) := HidingInd obj_dec_typing.
Definition HidingIncapInd (_ : ReferenceIncapacitate)
  := HidingInd obj_ref_incap.
Definition HidingCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C)
  := HidingInd obj_cap_bnd.
Hint Unfold HidingTInd HidingDInd HidingIncapInd HidingCapbndInd: core.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma unhiding_ts_ind:
  UnhidingTInd (ty_var _) /\
  UnhidingInd (subtyping) /\
  UnhidingIncapInd (ty_incap _).
Proof.
  apply ts_mutind.
  all: intros; subst.
  all: ctx_flatten in *.
  all: eauto using ctx_concat_hide.
  - (* ty_var *)
    apply ctx_binds_concat_hide_inv in c.
    auto.
  - (* subtyp_met *)
    eapply subtyp_met; intros; auto.
    rewrite <- ctx_concat_push.
    eauto.
    rewrite ctx_concat_double1.
    eauto.
Qed.

Lemma hiding_ts_ind:
  HidingTInd (ty_var _) /\
  HidingInd subtyping /\
  HidingIncapInd (ty_incap _).
Proof.
  apply ts_mutind.
  all: intros; subst.
  all: ctx_flatten in *.
  all: eauto using ctx_concat_hide.
  - (* ty_var *)
    apply ctx_binds_concat_hide in c.
    auto.
  - (* subtyp_met *)
    eapply subtyp_met; intros; auto.
    rewrite <- ctx_concat_push.
    eauto.
    rewrite ctx_concat_double1.
    eauto.
Qed.


End WithTypingMode.

Instance unhiding_ty_var m: Unhiding (ty_var m).
Proof.
  instance_by_ind Unhiding unhiding_ts_ind.
Qed.
Hint Resolve unhiding_ty_var: core.

Instance unhiding_subtyp m: Unhiding (subtyp m).
Proof.
  instance_by_ind Unhiding unhiding_ts_ind.
Qed.
Hint Resolve unhiding_subtyp: core.

Instance unhiding_ty_incap m: Unhiding (ty_incap m).
Proof.
  instance_by_ind Unhiding unhiding_ts_ind.
Qed.
Hint Resolve unhiding_ty_incap: core.


Lemma unhiding_ty_capbnd_ind m: forall C, UnhidingCapbndInd (ty_capbnd m C).
Proof.
  intros. hnf. induction 1; eauto.
  introv Heq. subst.
  apply ty_capbnd_sel with (T := T) (S := S); eauto.
  eapply unhiding_ty_var; eauto.
Qed.

Instance unhiding_ty_capbnd m: forall C, Unhiding (ty_capbnd m C).
Proof. intros. hnf. intros. eapply unhiding_ty_capbnd_ind; eauto. Qed.
Hint Resolve unhiding_ty_capbnd: core.

Instance hiding_ty_var m: Hiding (ty_var m).
Proof.
  instance_by_ind Hiding hiding_ts_ind.
Qed.
Hint Resolve hiding_ty_var: core.

Instance hiding_subtyp m: Hiding (subtyp m).
Proof.
  instance_by_ind Hiding hiding_ts_ind.
Qed.
Hint Resolve hiding_subtyp: core.

Instance hiding_ty_incap m: Hiding (ty_incap m).
Proof.
  instance_by_ind Hiding hiding_ts_ind.
Qed.
Hint Resolve hiding_ty_incap: core.

Lemma hiding_ty_capbnd_ind m: forall C, HidingCapbndInd (ty_capbnd m C).
Proof.
  intros. hnf. induction 1; eauto.
  introv Heq. subst.
  apply ty_capbnd_sel with (T := T) (S := S); eauto.
  eapply hiding_ty_var; eauto.
Qed.

Instance hiding_ty_capbnd m: forall C, Hiding (ty_capbnd m C).
Proof. intros. hnf. intros. eapply hiding_ty_capbnd_ind; eauto. Qed.
Hint Resolve hiding_ty_capbnd: core.

Instance unhiding_viewpoint_adapt m x: Unhiding (viewpoint_adapt m x).
Proof.
  unfold Unhiding.
  intros.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' m); eauto.
  inversions H.
  eapply viewpoint_adapt_cap_strip_combine.
  eapply @unhiding with (property := ty_incap _); eauto.
  eapply @unhiding with (property := ty_capbnd _ _); eauto.
Qed.
Hint Resolve unhiding_viewpoint_adapt: core.

Lemma unhiding_rules_ind m:
  UnhidingTInd (ty_trm m) /\
  UnhidingTInd (ty_lit m) /\
  UnhidingDInd (ty_def m) /\
  UnhidingTInd (ty_defs m).
Proof.
  apply rules_mutind.
  all: intros; subst.
  all: ctx_flatten in *.
  all: eauto using ctx_concat_hide.
  - (* ty_trm_var *)
    eapply ty_trm_var.
    eauto using unhiding_vis.
    eapply unhiding; eauto.
  - econstructor; eauto using unhiding_vis.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
  - econstructor; eauto using unhiding_vis.
    eapply unhiding; eauto.
    eapply unhiding_viewpoint_adapt; eauto.
  - econstructor; eauto using unhiding_vis.
    intros.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
  - econstructor; eauto using unhiding_vis.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
  - econstructor; eauto using unhiding_vis.
    eapply @unhiding with (property := subtyp _); eauto.
  - econstructor; eauto using unhiding_vis.
    intros.
    unfold ctx_obj in *.
    destruct Go.
    eauto using unhiding_vis.
    rewrite <- ctx_concat_push.
    eauto using unhiding_vis.
    eauto using unhiding_vis.
    unfold ctx_obj in *.
    destruct Go.
    eapply unhiding; eauto.
    rewrite <- ctx_concat_push.
    eapply unhiding; eauto.
    eapply unhiding; eauto.
  - econstructor; eauto using unhiding_vis.
    intros.
    unfold ctx_obj_met in *.
    destruct (mode_visibility' m).
    + destruct Go;
      ctx_split_from xz; rewrite <- ctx_concat_assoc;
      eapply H; eauto.
    + destruct Go;
      ctx_split_from xz; rewrite <- ctx_concat_hide;
      rewrite <- ctx_concat_assoc; eapply H; eauto.
Qed.

Instance unhiding_ty_trm m: Unhiding (ty_trm m).
Proof.
  instance_by_ind Unhiding unhiding_rules_ind.
Qed.
Hint Resolve unhiding_ty_trm: core.
Instance unhiding_ty_lit m: Unhiding (ty_lit m).
Proof.
  instance_by_ind Unhiding unhiding_rules_ind.
Qed.
Hint Resolve unhiding_ty_lit: core.
Instance unhiding_ty_defs m: Unhiding (ty_defs m).
Proof.
  instance_by_ind Unhiding unhiding_rules_ind.
Qed.
Hint Resolve unhiding_ty_defs: core.
Instance unhiding_ty_def m: Unhiding (ty_def m).
Proof.
  instance_by_ind Unhiding unhiding_rules_ind.
Qed.
Hint Resolve unhiding_ty_def: core.


Lemma vkchange_vis_ind: forall v vk1 vk2 T G1 G2 x,
    ctx_vis ((G1 § v / vk1 ~ T) §& G2) x ->
    ctx_vis ((G1 § v / vk2 ~ T) §& G2) x.
Proof.
  introv Hvis.
  remember ((G1 § v / vk1 ~ T) §& G2) as G.
  generalize dependent G1. generalize dependent G2.
  induction Hvis; intros; subst.
  - apply ctx_push_concat_inv in HeqG.
    destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
    inversions H0. simpl. eauto.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
    + inversions H1. simpl. eauto.
    + inversions HeqG. eauto.
Qed.

Lemma vkchange_vis_ind_hide: forall v vk1 vk2 T G1 G2 x,
    ctx_vis (ctx_hide (G1 § v / vk1 ~ T) §& G2) x ->
    ctx_vis (ctx_hide (G1 § v / vk2 ~ T) §& G2) x.
Proof.
  introv Hvis.
  remember (ctx_hide (G1 § v / vk1 ~ T) §& G2) as G.
  generalize dependent G1. generalize dependent G2.
  induction Hvis; intros; subst.
  - apply ctx_push_concat_inv in HeqG.
    destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
    inversions H0.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
    + inversions H1. + inversions HeqG. eauto.
Qed.

Lemma tychange_vis_ind: forall Go v vk T1 T2 G1 G2 x,
    ctx_vis ((G1 § v / vk ~ T1) §& G2 §§& Go) x ->
    ctx_vis ((G1 § v / vk ~ T2) §& G2 §§& Go) x.
Proof.
  introv Hvis.
  remember ((G1 § v / vk ~ T1) §& G2 §§& Go) as G.
  generalize dependent G1. generalize dependent G2.
  generalize dependent Go.
  induction Hvis; intros; subst.
  - induction Go; simpl in HeqG; subst; eauto.
    { apply ctx_push_concat_inv in HeqG.
      destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
      inversions H0. simpl. eauto. }
    { rewrite <- ctx_concat_push in HeqG.
      apply ctx_push_concat_inv in HeqG.
      destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
      + inversions H. + inversions H; simpl. eauto. }
    { apply ctx_push_concat_inv in HeqG.
      destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
      inversions H0. eauto. }
  - induction Go; simpl in HeqG; subst; eauto.
    { apply ctx_push_concat_inv in HeqG as ?.
      destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
      + inversions H1; simpl; eauto.
      + inversions HeqG. specialize (IHHvis objctx_empty).
        simpl in IHHvis. eauto. }
    { rewrite <- ctx_concat_push in HeqG. simpl.
      apply ctx_push_concat_inv in HeqG as ?.
      destruct H0 as [[? ?] | [? ?]]; subst; simpl; eauto.
      + inversions H0.
      + inversions H0; simpl. rewrite ctx_concat_push in HeqG.
        inversions HeqG.
        destruct (var_decide v0 v2); subst; eauto.
        constructor; eauto. specialize (IHHvis objctx_empty); eauto. }
    { apply ctx_push_concat_inv in HeqG as ?.
      destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
      + inversions H1; simpl; eauto.
      + inversions HeqG. specialize (IHHvis objctx_empty).
        simpl in IHHvis. eauto. }
Qed.

Corollary tychange_vis_ind_emptyctx: forall v vk T1 T2 G1 G2 x,
    ctx_vis ((G1 § v / vk ~ T1) §& G2) x ->
    ctx_vis ((G1 § v / vk ~ T2) §& G2) x.
Proof.
  intros. pose proof (tychange_vis_ind objctx_empty). simpl in H0; eauto.
Qed.

Lemma ctx_obj_met_vis_hide_inv typing_mode G Go xz TZ xc TC x :
  mode_visibility' typing_mode = mode_visibility_hide_method ->
  ctx_vis
    (G §& ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC) x ->
  x = xz \/ x = xc.
Proof.
  intros Hmode Hvis. destruct Go.
  all: simpl in Hvis; destruct (mode_visibility' typing_mode);
    try discriminate; simpl in Hvis.
  all: repeat(match goal with
              | [ H : ctx_vis _ _ |- _ ] => inversions H; try tauto
              end).
Qed.
