(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax GeneralTyping TypingContext Env.

(** * Weakening Lemma *)
(** Weakening states that typing is preserved in extended environments. *)

Class Weakening (A B: Set) (property: TypingProperty A B) :=
  weaken G1 G2 G3 Go e (ds: A) T:
    (property (G1 §& G3) Go e ds T ->
     ctx_weakens G2 ->
     ctx_ok (G1 §& G2) ->
     property (G1 §& G2 §& G3) Go e ds T).
Hint Unfold Weakening: core.

(* Tactic for applying weakening with context split into two parts *)
Ltac weaken_specialize :=
  intros;
  match goal with
  | [ Hok: ctx_ok (?G1 §& ?G2) |- _ ] =>
    rewrite <- ctx_concat_empty_r2; eapply weaken; eauto;
    (rewrite ctx_concat_empty_r || rewrite ctx_concat_empty_r2); eauto
  end.

Section WeakeningLemmata.
  Context {A B: Set} {P: TypingProperty A B} {AW: Weakening P}.

  (** Weakening lemma specialized to appending a context to the right. *)
  Lemma weaken_end:
    forall G1 G2 Go e (v: A) T,
      P G1 Go e v T ->
      ctx_weakens G2 ->
      ctx_ok (G1 §& G2) ->
      P (G1 §& G2) Go e v T.
  Proof.
    weaken_specialize.
  Qed.

  (** Weakening lemma specialized to appending a variable to the right. *)
  Lemma weaken_push:
    forall G1 x vk t Go e (v: A) T,
      P G1 Go e v T ->
      ctx_ok G1 ->
      x §# G1 ->
      P (G1 § x / vk ~ t) Go e v T.
  Proof.
    intros.
    rewrite <- (ctx_concat_empty_r G1).
    rewrite <- (ctx_concat_push).
    apply weaken_end; auto.
    rewrite (ctx_concat_push).
    rewrite (ctx_concat_empty_r).
    all: auto.
  Qed.

  Lemma ctx_met_weakens :
    forall G Go xz TZ xc TC e (v: A) T,
      P G Go e v T ->
      ctx_ok (ctx_met G xz TZ xc TC) ->
      P (ctx_met G xz TZ xc TC) Go e v T.
  Proof.
    introv Hp Hok.
    repeat (apply weaken_push; eauto);
      eauto using ctx_ok_push_inv_ok, ctx_ok_push_inv_unbound.
  Qed.

  Lemma ctx_met_refined_heap_weakens :
    forall G Go xz TZ xc TC e (v: A) y T S,
      P G Go e v S ->
      ctx_ok (ctx_met_refined_heap G xz TZ xc TC y T) ->
      P (ctx_met_refined_heap G xz TZ xc TC y T) Go e v S.
  Proof.
    introv Hp Hok.
    repeat (apply weaken_push; eauto);
      eauto using ctx_ok_push_inv_ok, ctx_ok_push_inv_unbound.
  Qed.

End WeakeningLemmata.

Section WeakeningTyLemmata.
  Context {A: Set}{AT: Typing A}{AW: Weakening AT}.

  (** [G1, G3 ⊢ t: T]                    #<br>#
    [ok(G1, G2)]                   #<br>#
    [――――――――――――――――――――]             #<br>#
    [G1, G2, G3 ⊢ t: T] #<br># #<br>#
   *)
  Lemma weaken_ty G1 G2 G3 e (ds: A) T:
    typing (G1 §& G3) e ds T ->
    ctx_weakens G2 ->
    ctx_ok (G1 §& G2) ->
    typing (G1 §& G2 §& G3) e ds T.
  Proof.
    apply weaken.
  Qed.

  Lemma weaken_obj_ty G1 G2 G3 Go e (ds: A) T:
    obj_typing (G1 §& G3) Go e ds T ->
    ctx_weakens G2 ->
    ctx_ok (G1 §& G2) ->
    obj_typing (G1 §& G2 §& G3) Go e ds T.
  Proof.
    apply weaken.
  Qed.

End WeakeningTyLemmata.

Section WeakeningSubtyLemmata.
  Context {A : Set} {AS : Subtyping} {AW : Weakening AS}.

  Lemma weaken_subtyping G1 G2 G3 Go e (ds: typ) T:
    subtyping (G1 §& G3) Go e ds T ->
    ctx_weakens G2 ->
    ctx_ok (G1 §& G2) ->
    subtyping (G1 §& G2 §& G3) Go e ds T.
  Proof.
    apply weaken.
  Qed.

End WeakeningSubtyLemmata.

Definition WeakeningIndCondition (G: ctx) (property: ctx->Prop) := (
  (property G) ->
  forall G1 G2 G3,
    (G = G1 §& G3 ->
     ctx_weakens G2 ->
     ctx_ok (G1 §& G2) ->
     (property (G1 §& G2 §& G3)))
).

(* Formulation of weakening for proof by induction *)
Definition WeakeningInd (A B: Set) (property: TypingProperty A B) :=
  (forall (G:ctx) (Go: objctx) (e:renv) (v : A) T, (WeakeningIndCondition G (fun G => property G Go e v T))).
Definition WeakeningTInd (A: Set) (t: Typing A) := WeakeningInd (obj_typing).
Definition WeakeningDInd (A: Set) (t: DecTyping A) := WeakeningInd (obj_dec_typing).
Definition WeakeningIncapInd (_ : ReferenceIncapacitate) :=
  WeakeningInd obj_ref_incap.
Definition WeakeningCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C) :=
  WeakeningInd obj_cap_bnd.

(* Instances of Weakening *)

(* The proof is by mutual induction on variable typing, and subtyping. *)
Lemma weaken_ts_ind typing_mode:
  WeakeningTInd (ty_var typing_mode) /\
  WeakeningInd subtyping /\
  WeakeningIncapInd (ty_incap typing_mode).
Proof.
  apply ts_mutind; unfold WeakeningIndCondition; intros; subst;
    intros; subst;
      eauto using ctx_binds_weaken, ctx_binds_weaken_hide.
  intros; subst; eapply subtyp_met;
    try (solve[apply H; auto]); intros; auto;
      repeat(rewrite <- ctx_concat_push); eauto.
Qed.

(** Weakening lemma specialized to variable typing. *)
Instance weaken_ty_var mode: Weakening (ty_var mode).
Proof.
  unfold Weakening. intros. eapply weaken_ts_ind; eauto.
Qed.
Hint Resolve weaken_ty_var: core.

(** Weakening lemma specialized to subtyping. *)
Instance weaken_subtyp mode: Weakening (subtyp mode).
Proof.
  unfold Weakening. intros. eapply weaken_ts_ind; eauto.
Qed.
Hint Resolve weaken_subtyp: core.

(** Weakening lemma specialized to removing capabilities. *)
Instance weaken_ty_incap mode: Weakening (ty_incap mode).
Proof.
  unfold Weakening. intros. eapply weaken_ts_ind; eauto.
Qed.
Hint Resolve weaken_ty_incap: core.

(** Weakening lemma specialized to extracting capabilities. *)
Lemma weaken_ty_capbnd_ind mode: forall C, WeakeningCapbndInd (ty_capbnd mode C).
Proof.
  intros. hnf. introv.
  induction 1; eauto; introv;
    introv Heq Hwk Hok; subst; eauto.
  apply ty_capbnd_sel with (T := T) (S := S); eauto.
  eapply weaken_ts_ind; eauto.
Qed.

Instance weaken_ty_capbnd mode: forall C, Weakening (ty_capbnd mode C).
Proof.
  intros. hnf. intros. eapply weaken_ty_capbnd_ind; eauto.
Qed.
Hint Resolve weaken_ty_capbnd: core.

Instance weaken_viewpoint_adapt mode : forall x,
    Weakening (viewpoint_adapt mode x).
Proof.
  intros. hnf. introv Hadapt Hwk Hok.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' mode); eauto. simpl in *.
  inversions Hadapt.
  apply viewpoint_adapt_cap_strip_combine.
  - eapply weaken_ty_incap; eauto.
  - eapply weaken_ty_capbnd; eauto.
Qed.
Hint Resolve weaken_viewpoint_adapt: core.

Lemma weaken_vis (G : ctx) (x : var) :
  ctx_vis G x ->
  forall G1 G2 G3 : ctx,
  (G = G1 §& G3 ->
   ctx_weakens G2 -> ctx_ok (G1 §& G2) -> ctx_vis (G1 §& G2 §& G3) x).
Proof.
  induction 1; introv Heq Hwk Hok.
  - apply ctx_push_concat_inv in Heq
      as [[? ?] | [G3' ?]]; subst; simpl; eauto using weaken_vis_push.
  - apply ctx_push_concat_inv in Heq as Heq'.
    destruct Heq' as [[? ?] | [G3' ?]];
      subst; simpl in *; eauto using weaken_vis_push.
    inversions Heq. constructor; eauto.
Qed.

(* The proof is by mutual induction on term typing and definition typing. *)
Lemma weaken_rules_ind mode:
  WeakeningTInd (ty_trm mode) /\
  WeakeningTInd (ty_lit mode) /\
  WeakeningDInd (ty_def mode) /\
  WeakeningTInd (ty_defs mode).
Proof.
  apply rules_mutind; unfold WeakeningIndCondition; intros; subst;
    try objctx_destruct;
    try solve[
        subst; simpl in *;
        repeat
          (match goal with
           | [_ : ?G1 §& ?G3 ⁏ _ ⊢ ?x ∶ _,
              Hvis : mode_has_visibility mode -> ctx_vis (?G1 §& ?G3) ?x |-
              _] =>
             assert (mode_has_visibility mode -> ctx_vis (G1 §& G2 §& G3) x)
               by (intros ?Hmodvis; specialize (Hvis Hmodvis);
                   eapply weaken_vis; eauto); clear Hvis
           end);
        econstructor; simpl in *;
        eauto using ctx_binds_weaken, weaken_ty, weaken_subtyping].

  { (* Field read *)
    assert (mode_has_visibility mode -> ctx_vis (G1 §& G2 §& G3) x)
      by (intros ?Hmodvis; specialize (c Hmodvis);
          eapply weaken_vis; eauto).
    eapply ty_trm_read;
      eauto using ctx_binds_weaken, weaken_ty, weaken_subtyping.
    eapply weaken_viewpoint_adapt; eauto. }

  { (* Field definition typing *)
    subst; simpl in *; econstructor; simpl in *; intros;
    rewrite <- ctx_concat_push;
    eauto using weaken_vis, weaken_ty, weaken_subtyping. }

  { (* Method typing: needs results about hiding *)
    subst; simpl in *.
    eapply ty_def_met. unfold ctx_obj_met.
    intros. destruct Go.
    - unfold ctx_obj_met in *. destruct (mode_visibility' mode); simpl in *; repeat rewrite <- ctx_concat_push; eauto.
      all: specialize (H _ H0 _ H3 H4).
      specialize (H G1 G2 (ctx_met (G3 §& ctx_hide ctx_empty) xz TZ xc TC)).
      simpl in H. eauto.
    - unfold ctx_obj_met in *. destruct (mode_visibility' mode); simpl in *; repeat rewrite <- ctx_concat_push; eauto.
      all: specialize (H _ H0 _ H3 H4).
      specialize (H G1 G2
                      (ctx_met_refined_heap
                         ((G3 § x / v_self ~ open x T) §& ctx_hide ctx_empty)
                         xz TZ xc TC x T)).
      simpl in H; eauto.
    - unfold ctx_obj_met in *. destruct (mode_visibility' mode); simpl in *; repeat rewrite <- ctx_concat_push; eauto.
      all: specialize (H _ H0 _ H3 H4).
      specialize (H G1 G2
                      (ctx_met_refined_heap
                         (G3 §& ctx_hide ctx_empty)
                         xz TZ xc TC y T)).
      simpl in H; eauto. }
Qed.

Instance weaken_ty_trm mode: Weakening (ty_trm mode).
Proof.
  unfold Weakening. intros.
  eapply weaken_rules_ind; eauto.
Qed.
Hint Resolve weaken_ty_trm: core.

Instance weaken_ty_lit mode: Weakening (ty_lit mode).
Proof.
  unfold Weakening. intros.
  eapply weaken_rules_ind; eauto.
Qed.
Hint Resolve weaken_ty_lit: core.

Instance weaken_ty_defs mode: Weakening (ty_defs mode).
Proof.
  unfold Weakening. intros.
  eapply weaken_rules_ind; eauto.
Qed.
Hint Resolve weaken_ty_defs: core.
Instance weaken_ty_def mode: Weakening (ty_def mode).
Proof.
  unfold Weakening. intros.
  eapply weaken_rules_ind; eauto.
Qed.
Hint Resolve weaken_ty_def: core.
