(** remove printing ~ *)

(** This proof uses the
    #<a href="http://www.chargueraud.org/softs/tlc/">TLC</a>#
    Coq library by Arthur Chargueraud. *)

Set Implicit Arguments.

Require Import Bool.
Require Import LibExtra.
Require String.
Export Coq.Strings.String.StringSyntax.

(** * Tactics *)

(** Tactics for naming cases in case analysis. *)

Open Scope string_scope.

Ltac move_to_top x :=
  match reverse goal with
  | H : _ |- _ => try move x after H
  end.

Tactic Notation "assert_eq" ident(x) constr(v) :=
  let H := fresh in
  assert (x = v) as H by reflexivity;
  clear H.

Tactic Notation "Case_aux" ident(x) constr(name) :=
  first [
    set (x := name); move_to_top x
  | assert_eq x name; move_to_top x
  | fail 1 "because we are working on a different case" ].

Tactic Notation "Case" constr(name) := Case_aux Case name.
Tactic Notation "SCase" constr(name) := Case_aux SCase name.
Tactic Notation "SSCase" constr(name) := Case_aux SSCase name.
Tactic Notation "SSSCase" constr(name) := Case_aux SSSCase name.
Tactic Notation "SSSSCase" constr(name) := Case_aux SSSSCase name.
Tactic Notation "SSSSSCase" constr(name) := Case_aux SSSSSCase name.

(** Invert Bindings *)
Ltac binds_eq :=
  match goal with
  | [Hb1: binds ?x _ ?G,
     Hb2: binds ?x _ ?G |- _] =>
     apply (binds_functional Hb1) in Hb2; inversions Hb2
  end.

Ltac binds_contra H1 H2 :=
  pose proof (binds_functional H1 H2); congruence.

(** Automatically destruct premises *)
Ltac destruct_iffs :=
  repeat match goal with
  | [ H : ?A <-> ?B |- _ ] => destruct H
  end.

Ltac destruct_ands :=
  repeat match goal with
  | [ H : ?A /\ ?B |- _ ] => destruct H
  end.

Ltac destruct_ors :=
  repeat match goal with
  | [ H : ?A \/ ?B |- _ ] => destruct H
  end.

Lemma orb_true_intro3:
  forall b1 b2 b3: bool,
  b1 = true \/ b2 = true \/ b3 = true ->
  b1 || b2 || b3 = true
.
Proof.
  intros.
  destruct H.
  apply orb_true_intro. left.  apply orb_true_intro. left. assumption.
  destruct H.
  apply orb_true_intro. left.  apply orb_true_intro. right. assumption.
  apply orb_true_intro. right. assumption.
Qed.

Lemma orb_prop3:
  forall b1 b2 b3: bool,
  b1 || b2 || b3 = true ->
  b1 = true \/ b2 = true \/ b3 = true
.
Proof.
  intros.
  apply orb_prop in H.
  destruct H.
  apply orb_prop in H.
  destruct H.
  all: auto.
Qed.

Lemma andb_prop3:
  forall b1 b2 b3: bool,
  b1 && b2 && b3 = true ->
  b1 = true /\ b2 = true /\ b3 = true
.
Proof.
  intros.
  apply andb_prop in H.
  destruct H.
  apply andb_prop in H.
  destruct H.
  all: auto.
Qed.

Lemma andb_true_intro3:
  forall b1 b2 b3: bool,
  b1 = true /\ b2 = true /\ b3 = true ->
  b1 && b2 && b3 = true
.
Proof.
  intros.
  destruct H.
  destruct H0.

  apply andb_true_intro.
  split.
  apply andb_true_intro.
  split.
  all: auto.
Qed.

(* These tactics must be able to destruct as many operands as the maximum arity of a type *)
Ltac destruct_orbs :=
  repeat
    match goal with
    | [H : _ || _ || _ = true |- _] => apply orb_prop3 in H
    | [H : _ || _ = true |- _] => apply orb_prop in H
    | [|- _ || _ || _ = true] => apply orb_true_intro3
    | [|- _ || _ = true] => apply orb_true_intro
    end.

Ltac destruct_andbs :=
  repeat
    match goal with
    | [H : _ && _ && _ = true |- _] => apply andb_prop3 in H
    | [H : _ && _ = true |- _] => apply andb_prop in H
    | [|- _ && _ && _ = true] => apply andb_true_intro3
    | [|- _ && _ = true] => apply andb_true_intro
    end.

Ltac destruct_all :=
  repeat match goal with
  | [ H : exists x, _ |- _ ]  => destruct H
  | [ H : ?A /\ ?B |- _ ] => destruct H
  | [ H : ?A \/ ?B |- _ ] => destruct H
  end.

Ltac destruct_notin :=
  repeat match goal with
         | [ H: ?z \notin ?E1 \u ?E2 |- _ ] =>
           apply notin_union in H; destruct H
         end.

Ltac specialize_notins :=
  repeat match goal with
  | [ H : (?x \notin ?F ?y -> _), H' : ?x \notin ?F ?y |- _] =>
    specialize (H H')
  end.

Ltac splitr :=
  match goal with
  | |- ?A /\ ?B => split; splits
  | _ => idtac
  end.

Ltac omega := Coq.omega.Omega.omega.

(* Tactic for proving instances,
   for which there is lemma by induction *)
Ltac instance_by_ind C ind :=
  unfold C; intros; eapply ind; eauto.

Ltac instance_by_ind0 ind := hnf; intros; eapply ind; eauto.

(* In a hypothesis H, splits a list containing x into two parts (_ ++ x :: _)
   This is useful to use with lemmas in AbstracMachine.v .
 *)
Ltac list_split_from_in x H :=  (rewrite <- (List.app_nil_l (x:: _)%list) in H; rewrite ? List.app_comm_cons in H).

(*
  These tactics try to address the issue with the forwards tactic, where the hypotheses of the lemma
  end up on an equal level with the initial goal. This means that when chaining tactics with `;`, the tactic is applied to both the hypotheses and the goal,
  but sometimes they have a very different character, and we may want to apply the tactic only to the hypotheses.
  This can be achieved using `[> tactic .. | ]` which applies the tactic to all goals except the last one.
  Also, in the proof script, the hypotheses are on the same level when using bullets and braces, which does not allow trying to make the proof more readable by separating the hypotheses.
  
  The way to separate the hypotheses is to generate them under a different goal than the initial one.
  The solution is to create a new goal at the same level as the initial one, and create the hypotheses goals under it. Then, we can use braces to focus on the hypotheses goal,
  and use bullets within the braces to focus on individual hypothesis goals.
  
  Tactic language does not allow focusing goals from within tactics, so the tactics only manipulate the goal and hypotheses, and the braces must be written by the user.
*)
(* Solution 1: 
  First, use `forward_block H` to create the temporary goal and give a name to the target hypothesis.
  Then, use `{` to focus on the temporary goal.
  Then, use `forward E0 A1` ... as you would with `forwards H: E0 A1`...
  Then, prove the goals and use `}` when done.
*) 

(*
Tactic Notation "forward_block" ident(H) := eassert _ as H. 
Tactic Notation "forward" constr(E0) := let H := fresh "H" in forwards H: E0;  [> ..| eapply H].
Tactic Notation "forward" constr(E0) constr(A1) := let H := fresh "H" in forwards H: E0 A1;  [> ..| eapply H].
Tactic Notation "forward" constr(E0) constr(A1) constr(A2) := let H := fresh "H" in forwards H: E0 A1 A2;  [> ..| eapply H].
*)

(* Solution 2: 
  First, use `forward_block H: E0 A1 A2` to create the temporary goal, give a name to the target hypothesis and specify the arguments as you would with forwards H: E0 A1 A2.
  Then, use `{` to focus on the temporary goal.
  Then, use `forward` to create the hypothesis goals.
  Then, prove the goals and use `}` when done.
*) 

Tactic Notation "forward_block" simple_intropattern(I) ":" constr(Ei) :=
  let args := forwards_build_app_arg Ei in
  let Hargs := fresh I "args" in 
  eassert _ as I; [set (Hargs := args)|].
Tactic Notation "forward_block" simple_intropattern(I) ":" constr(E0)
 constr(A1) :=
  forward_block I: (>> E0 A1).
Tactic Notation "forward_block" simple_intropattern(I) ":" constr(E0)
 constr(A1) constr(A2) :=
  forward_block I: (>> E0 A1 A2).
Tactic Notation "forward_block" simple_intropattern(I) ":" constr(E0)
 constr(A1) constr(A2) constr(A3) :=
  forward_block I: (>> E0 A1 A2 A3).
Tactic Notation "forward_block" simple_intropattern(I) ":" constr(E0)
 constr(A1) constr(A2) constr(A3) constr(A4) :=
  forward_block I: (>> E0 A1 A2 A3 A4).
Tactic Notation "forward_block" simple_intropattern(I) ":" constr(E0)
 constr(A1) constr(A2) constr(A3) constr(A4) constr(A5) :=
  forward_block I: (>> E0 A1 A2 A3 A4 A5).

Tactic Notation "forward" := match goal with X:=?args:list Boxer |- _ => clear X; lets X: args;  [> ..| eapply X] end.
