Set Implicit Arguments.

From TLC Require Import LibEnv LibVar LibLN LibTactics.
Require Import LibVarExtra LibListExtra LibSetExtra LibFreshExtra LibTactics.

(** * Lemmas about environments *)

(** Converting between binding and in dom *)
Lemma dom_to_binds : forall A (E : env A) x,
    x \in dom E ->
    exists v, binds x v E.
Proof.
  induction E using env_ind.
  - intros; false. simpl_dom.
    rewrite in_empty in H; auto.
  - intros.
    destruct (var_decide x0 x) as [?H | ?H].
    + subst x0; exists v; auto.
    + rewrite dom_concat, in_union in H.
      destruct H.
      * apply IHE in H.
        destruct H as [?v ?H].
        exists v0; apply binds_concat_left; auto.
      * false. simpl_dom; rewrite in_singleton in H.
        auto.
Qed.

Lemma binds_to_dom : forall A (E : env A) x v,
    binds x v E ->
    x \in dom E.
Proof.
  introv H. induction E using env_ind.
  - exfalso; eauto using binds_empty_inv.
  - destruct (binds_push_inv H) as [[? ?] | [? ?]]; subst; simpl_dom; rewrite in_union.
    + left. auto using in_singleton_self.
    + right. auto.
Qed.

Hint Extern 1 (_ # _) =>
match goal with
| [H : dom ?G1 = dom ?G2, _ : ?x # ?G1 |- ?x # ?G2 ] => rewrite <- H; assumption
| [H : dom ?G2 = dom ?G1, _ : ?x # ?G1 |- ?x # ?G2 ] => rewrite H; assumption
end : core.

Lemma ok_val : forall A (E : env A) x v1 v2,
    ok (E & x ~ v1) ->
    ok (E & x ~ v2).
Proof.
  introv H; apply ok_push_inv in H as [? ?]; auto 2.
Qed.
Hint Extern 1 (ok _) =>
match goal with
| [H : ok (?E & ?x ~ ?v1) |- ok (?E & ?x ~ ?v2)] =>
  apply (ok_val (v1:=v1)); assumption
end : core.

(* Like binds_weaken, but gets by with ok(E & F) instead of ok (E & F & G) *)
Lemma binds_weaken2 :
  forall (A : Type) (x : var) (a : A) (E F G : env A),
    binds x a (E & G) -> ok (E & F) -> binds x a (E & F & G).
Proof.
  intros A x a E F G Bi Hok.
  apply binds_concat_inv in Bi.
  destruct Bi as [Bi | [Fr Bi]]; auto 2.
  apply binds_concat_left.
  apply binds_concat_left_ok.
  all: auto.
Qed.

Lemma binds_ind (A: Type) (P: env A -> Prop) (Q: Prop) (E: env A) x v:
  P E ->
  binds x v E ->
  (forall (E : env A) (x1 : var) (v1 : A), x1 <> x -> (P (E & x1 ~ v1)) -> (P E)) ->
  (forall (E : env A), P (E & x ~ v) -> Q) ->
  Q.
Proof.
  induction E using env_ind.
  - intros. false binds_empty_inv; eauto.
  - intros.
    destruct (classicT(x=x0)).
    + subst. eapply binds_push_eq_inv in H0. subst. eapply H2. eauto.
    + eapply binds_push_neq_inv in H0; eauto.
Qed.

(* Conjunctive property of environment.
   A property that holds for an env exactly if it holds for all elements
*)

Unset Auto Template Polymorphism.
Class EnvConjunctive (A: Type) (P: env A -> Prop) (Q: var -> A -> Prop) := {
  (* Constructors *)
  env_conj_empty: (P empty) ;
  env_conj_push E x v: (P E) -> (Q x v) -> (P (E & x ~ v)) ;
  (* Inversion *)
  env_conj_push_item E x v: (P (E & x ~ v)) -> (Q x v) ;
  env_conj_push_inv E x v: (P (E & x ~ v)) -> (P E)
}.

(* From env property to element property *)
Lemma binds_conjunctive (A: Type) (P: env A -> Prop) (Q: var -> A -> Prop) (PC: EnvConjunctive P Q) x v E:
  P E ->
  binds x v E ->
  Q x v.
Proof.
  intros.
  eapply binds_ind.
  apply H. apply H0.
  intros.
  eapply env_conj_push_inv. eauto.
  intros. eapply env_conj_push_item. eauto.
Qed.

(* From element properties to env property *)
Lemma binds_all_conjunctive (A: Type) (P: env A -> Prop) (Q: var -> A -> Prop) (PC: EnvConjunctive P Q) E:
  ok E ->
  (forall x v, binds x v E -> Q x v) ->
  (P E).
Proof.
  introv Hok HQ.
  induction E using env_ind.
  - apply env_conj_empty.
  - eapply env_conj_push.
    + eapply IHE.
      eauto.
      intros. apply HQ. eapply binds_concat_left_ok; eauto.
    + apply HQ. apply binds_push_eq.
Qed.


Lemma ok_concat_notin_cases A x (E F: env A) :
  ok (E & F) ->
  x # E \/ x # F.
Proof.
  introv Hok.
  induction F using env_ind.
  - right. eauto.
  - rewrite concat_assoc in Hok.
    destruct (ok_push_inv Hok).
    destruct (classicT(x=x0)).
    subst; eauto.
    forwards: IHF; eauto.
    destruct H1; eauto.
Qed.


Lemma binds_concat_ok_inv A x (v: A) E F :
  ok (E & F) ->
  binds x v (E & F) ->
  (x # E /\ binds x v F) \/ (x # F /\ binds x v E).
Proof.
  introv Hok Hbi.
  destruct (binds_concat_inv Hbi);
  destruct (ok_concat_notin_cases x Hok); eauto.
  eapply binds_to_dom in H. contradiction.
Qed.

Lemma ok_swap A (E F: env A):
  ok (E & F) -> ok (F & E).
Proof.
  introv Hok.
  induction E using env_ind.
  - rewrite concat_empty_l in Hok. rewrite concat_empty_r. auto.
  - rewrite concat_assoc. eapply ok_push. eauto. lets: ok_middle_inv_l Hok. lets: ok_middle_inv_r Hok. eauto.
Qed.

Lemma ok_swap_after A (E F G: env A):
  ok (E & F & G) -> ok (E & G & F).
Proof.
  introv Hok.
  induction F using env_ind.
  - rewrite concat_empty_r in Hok. rewrite concat_empty_r. auto.
  - rewrite concat_assoc. rewrite concat_assoc in Hok. lets: ok_middle_inv_l Hok. lets: ok_middle_inv_r Hok.
    eapply ok_push. eauto. eauto.
Qed.

Lemma ok_concat A (G1 G2: env A):
  ok G1 ->
  ok G2 ->
  disjoint (dom G1) (dom G2) ->
  ok (G1 & G2).
Proof.
  introv Hok1 Hok2 Hdisj.
  induction Hok2.
  - rewrite concat_empty_r. auto.
  - 
    rewrite dom_push in Hdisj.
    eapply disjoint_union_r_inv in Hdisj. destruct Hdisj.
    rewrite concat_assoc. eapply ok_push.
    auto.
    rewrite dom_concat.
    eapply disjoint_singleton_inv_r in H0.
    notin_solve.
Qed.

Lemma ok_concat_rl A (G1 G2 G3 : env A):
  ok (G1 & (G2 & G3)) ->
  ok (G1 & G2).
Proof.
  introv Hok.
  rewrite concat_assoc in Hok.
  eapply ok_concat_inv_l; eauto.
Qed.
Lemma ok_concat_rr A (G1 G2 G3 : env A):
  ok (G1 & (G2 & G3)) ->
  ok (G1 & G3).
Proof.
  introv Hok.
  rewrite concat_assoc in Hok.
  eapply ok_remove; eauto.
Qed.
Lemma ok_concat_disjoint A (G1 G2: env A):
  ok (G1 & G2) ->
  disjoint (dom G1) (dom G2).
Proof.
  introv Hok.
  induction G2 using env_ind.
  - rewrite dom_empty. apply disjoint_empty_r.
  -
    rewrite concat_assoc in Hok.
    rewrite dom_push.
    apply ok_push_inv in Hok.
    destruct Hok.
    apply disjoint_union_r. 
    apply disjoint_singleton_r. eauto.
    auto.
Qed.

Lemma env_push_keys A (E: env A) x v:
  keys (E & x ~ v) = (x::(keys E))%list.
Proof.
  rewrite keys_def. rewrite concat_def. rewrite single_def. rewrite LibList.map_app. rewrite LibList.map_cons. rewrite LibList.map_nil.
  rewrite liblist_app. rewrite <- List.app_comm_cons. rewrite List.app_nil_l. unfold fst. auto.
Qed.
Lemma env_single_keys A (x: var) (v: A):
  keys (x ~ v) = (x::nil)%list.
Proof.
rewrite keys_def. rewrite single_def. rewrite LibList.map_cons. rewrite LibList.map_nil. unfold fst. auto.
Qed.

Lemma env_from_list_keys_dom A (E: env A):
  (from_list (keys E)) = dom E.
Proof.
  eapply fset_extens.
  +
    induction E using env_ind.
    - rewrite keys_def. rewrite empty_def. rewrite LibList.map_nil. rewrite from_list_nil. eapply subset_empty_l.
    - simpl_dom. rewrite env_push_keys. rewrite from_list_cons. eapply subset_union_2. eapply subset_refl. eauto.
  +
    induction E using env_ind.
    - simpl_dom. eapply subset_empty_l.
    - simpl_dom. rewrite env_push_keys. rewrite from_list_cons. eapply subset_union_2. eapply subset_refl. eauto.
Qed.

Lemma binds_permute_push {A} (e1 e2 : env A) x a y b :
  ok (e1 & y ~ b & e2) ->
  binds x a (e1 & y ~ b & e2) ->
  binds x a (e1 & e2 & y ~ b).
Proof.
  introv Hok Hbi.
  apply binds_middle_inv in Hbi.
  destruct Hbi as [Hbi2 | [[? [? ?]] | [Hnin [Hneq Hbi]]]]; subst; eauto.
  apply ok_middle_inv_r in Hok as Hnin.
  apply binds_to_dom in Hbi2 as Hin.
  apply binds_push_neq; eauto using binds_concat_right.
  hnf; intros; subst; eauto.
Qed.

Lemma ok_permute_push {A} (e1 e2 : env A) x a :
  ok (e1 & x ~ a & e2) ->
  ok (e1 & e2 & x ~ a).
Proof.
  introv Hok.
  generalize dependent a.
  generalize dependent x.
  generalize dependent e1.
  generalize dependent e2.
  apply (env_ind (fun e2 => forall (e1 : env A) (x : var) (a : A),
                      ok (e1 & x ~ a & e2) -> ok (e1 & e2 & x ~ a))).
  - introv Hok. rewrite concat_empty_r in *. auto.
  - intros e0 x0 a0 IH e x a Hok.
    rewrite concat_assoc in *. inversions Hok.
    1: { false. eauto using empty_push_inv. }
    apply eq_push_inv in H.
    repeat(match goal with
           | [ H : _ /\ _ |- _ ] => destruct H
           end); subst.
    rename H1 into Hfrx0, H0 into Hok.
    apply IH in Hok. constructor; eauto.
    inversions Hok.
    1: { false. eauto using empty_push_inv. }
    apply eq_push_inv in H.
    repeat(match goal with
           | [ H : _ /\ _ |- _ ] => destruct H
           end); subst.
    repeat(rewrite dom_concat in *).
    repeat(rewrite dom_single in *).
    repeat(rewrite notin_union in *).
    repeat(rewrite notin_singleton in *).
    repeat(match goal with
           | [ H : _ /\ _ |- _ ] => destruct H
           end); subst.
    repeat(split); try tauto. congruence.
Qed.

Lemma fresh_keys_ok A (G : env A):
  fresh \{} (length (keys G)) (keys G) ->
  ok G.
Proof.
  introv Hfr.
  induction G using env_ind.
  - auto.
  - rewrite env_push_keys in Hfr.
    inversions Hfr.
    rewrite union_empty_l in  H0.
    forwards: fresh_single_notin H0.
    rewrite env_from_list_keys_dom in H1.
    apply ok_push; auto.
Qed.
Lemma ok_fresh_keys A (E: env A):
  ok E ->
  fresh \{} (length (keys E)) (keys E).
Proof.
  introv Hfr.
  induction E using env_ind.
  - rewrite keys_def, empty_def. simpl. auto.
  - rewrite env_push_keys.
    forwards Hok: ok_push_inv Hfr. destruct Hok.
    constructor. apply notin_empty.
    rewrite union_empty_l.
    eapply fresh_from_list. rewrite env_from_list_keys_dom.  auto. 
    eauto.
Qed.
Lemma env_concat_keys (A: Type) (E F: env A):
  keys (E & F) = ((keys F) ++ (keys E))%list.
Proof.
  rewrite keys_def.
  rewrite concat_def.
  rewrite <- liblist_app.
  rewrite <- LibList.map_app.
  auto.
Qed.
(* For induction of two envs with the same keys *)
Lemma keys_push_eq_inv A (S1 S2: env A) x y1:
  keys (S1 & x ~ y1) = keys S2 ->
  exists y2 S21,
    S2 = S21 & x ~ y2 /\
    keys S1 = keys S21.
Proof.  
  introv Hk.
  induction S2 using env_ind.
  -
  rewrite env_push_keys, keys_def, empty_def, LibList.map_nil in Hk. inversions Hk.
  -
  rewrite ! env_push_keys, keys_def in Hk. inversions Hk.
  exists v S2. rewrite keys_def. eauto.
Qed.

Lemma env_extens A (S1 S2: env A):
  ok S1 ->
  keys S1 = keys S2 ->
  (forall (z0 : var) (v : A), binds z0 v (S1) -> binds z0 v (S2)) ->
  S1= S2.
Proof.
  introv Hok Heq.
  revert dependent S2.
  induction S1 using env_ind; intros.
  - destruct S2.
    + rewrite empty_def in *. auto.
    + rewrite empty_def, keys_def, LibList.map_cons, LibList.map_nil in Heq.
      inversions Heq.
  - destruct (env_case S2); subst.
    + rewrite concat_def, single_def, keys_def, empty_def, LibList.app_cons_one_r, LibList.map_cons, LibList.map_nil in Heq.
      inversions Heq.
    + destructecs H0; subst.
      rewrite concat_def, single_def, keys_def, ! LibList.app_cons_one_r, ! LibList.map_cons in Heq.
      inversions Heq.
      rewrite <- keys_def in H2.
      eapply ok_push_inv in Hok.
      destructecs Hok.
  
      forwards: IHS1 H2.
      auto.
      intros.
      forwards: H z0 v1.
      eapply binds_push_neq. auto.
      intro; subst. apply Hok. eapply binds_to_dom; eauto.
      eapply binds_push_neq_inv. eauto.
      intro; subst. apply Hok. eapply binds_to_dom; eauto.
      subst.

      assert (v = v0). {
      forwards: H x0 v.
          eapply binds_push_eq.
          eapply binds_push_eq_inv in H0. auto.
      }
      subst.
      auto.
Qed.

Lemma keys_from_list_dom A (e: env A):
  from_list (keys e) = dom e.
Proof.
  induction e.
  rewrite keys_def, LibList.map_nil, from_list_nil, dom_def. simpls. auto.
  rewrite keys_def, LibList.map_cons, from_list_cons, dom_def in *. rewrite  IHe. simpls. auto.
Qed.

Lemma env_dom_mem A (S: env A) a:
  a \in dom S ->
  exists p, LibList.mem (a, p) S.
Proof.
  introv Hdom.
  induction S.
  - rewrite dom_def in Hdom. simpls. rewrite in_empty in Hdom. contradiction.
  - rewrite dom_def in Hdom. simpls.  rewrite <- dom_def in Hdom.
    rewrite in_union, in_singleton in Hdom. destructecs Hdom; subst.
    + (* here *)
      exists (snd a0). destruct a0; simpl. eauto.
    + forwards Hp: IHS; fweauto.
      destructecs Hp.
      eauto.
Qed.

Lemma env_mem_dom A (S: env A) a p:
  LibList.mem (a, p) S ->
  a \in dom S.
Proof.
  introv Hmem.
  induction S.
  - inversions Hmem.
  - inversions Hmem.
    + (* here *)
      rewrite dom_def. simpls. rewrite in_union, in_singleton. auto.
    +  rewrite dom_def. rewrite LibList.fold_right_cons. rewrite <- dom_def. rewrite in_union, in_singleton. auto.
Qed.



  
Lemma env_removeat_notin A (S: env A) n x:
  x # S ->
  x # (list_removeat n S).
Proof.
  introv Hn.
  rewrite <- keys_from_list_dom in *.
  rewrite keys_def in *.
  rewrite list_removeat_map.
  eapply notin_subset.
  eapply from_list_removeat_subset.
  auto.
Qed.

Lemma env_dom_cons A (S: env A) a v:
  dom ((a, v)::S)%list = \{a} \u (dom S).
Proof.
  rewrite dom_def. rewrite ! LibList.fold_right_cons. rewrite <- dom_def. auto.
Qed. 

Lemma env_removeat_dom_remove A (S: env A) n p:
  ok S ->
  LibList.Nth n S p ->
  dom (list_removeat n S) = (dom S) \- \{fst p}.
Proof.
  introv Hok Hn.
  reverts n.

  induction Hok; intros.
  - rewrite empty_def, LibList.take_nil, LibList.drop_nil. simpl. rewrite dom_def. simpl. rewrite remove_empty_r. auto.
  - rewrite <- cons_to_push.
    destruct n.
    + rewrite LibList.take_zero.
      rewrite LibList.drop_succ.
      rewrite LibList.drop_zero.
      rewrite <- cons_to_push in Hn. inversions Hn.
      rewrite env_dom_cons. simpl.
      rewrite union_comm.
      rewrite union_remove_swap0.
      rewrite remove_same. auto. auto.
    + rewrite <- cons_to_push in Hn. inversions Hn.
      forwards HHok: IHHok H4.
      rewrite LibList.drop_succ.
      rewrite LibList.take_succ.  
      rewrite <- List.app_comm_cons.
      rewrite ! env_dom_cons.
      rewrite HHok.
      rewrite union_remove_swap0. auto.
      destruct p; simpls.
      rewrite notin_singleton.
      intro. subst.
      eapply LibList.Nth_mem in H4.
      eapply env_mem_dom in H4. contradiction.
Qed.

Lemma env_removeat_ok A (S: env A) n:
  ok S ->
  ok (list_removeat n S).
Proof.
  introv Hok.
  reverts n.
  induction Hok; intros.
  - rewrite empty_def, LibList.take_nil, LibList.drop_nil. simpl. rewrite <- empty_def. auto.
  - rewrite <- cons_to_push.
    destruct n.
    + rewrite LibList.take_zero.
      rewrite LibList.drop_succ.
      rewrite LibList.drop_zero.
      simpl. auto.
    +
      rewrite LibList.drop_succ.
      rewrite LibList.take_succ.  
      rewrite <- List.app_comm_cons.
      rewrite cons_to_push.
      constructor. eauto.
      eapply env_removeat_notin; eauto.
Qed.

Lemma env_keys_eq_ok A (S1 S2: env A):
  keys S1 = keys S2 ->
  ok S1 ->
  ok S2.
Proof.
  introv Heq Hok.
  reverts S2.
  induction Hok; intros.
   - destruct S2.
    + rewrite <- empty_def in *. auto.
    + rewrite empty_def, keys_def, LibList.map_cons, LibList.map_nil in Heq.
      inversions Heq.
  - destruct (env_case S2); subst.
    + rewrite concat_def, single_def, keys_def, empty_def, LibList.app_cons_one_r, LibList.map_cons, LibList.map_nil in Heq.
      inversions Heq.
    + destructecs H0; subst.
      rewrite concat_def, single_def, keys_def, ! LibList.app_cons_one_r, ! LibList.map_cons in Heq.
      inversions Heq.
      rewrite <- keys_def in H2.
      rewrite <- keys_from_list_dom in H. rewrite H2 in H.
      rewrite  keys_from_list_dom in H.
      eauto.
Qed.

Lemma env_binds_mem A (S: env A) x v:
  binds x v S ->
  LibList.mem (x,v) S.
Proof.
  introv Hbi.
  hnf in Hbi.
  induction S using env_ind.
  rewrite get_empty in Hbi. inversions Hbi.
  rewrite get_push in Hbi.
  cases_if.
  inversions Hbi.
  rewrite <- cons_to_push. eapply LibList.mem_here.
  rewrite <- cons_to_push. eapply LibList.mem_cons. auto.
Qed.

Lemma env_mem_binds A (S: env A) x v:
  LibList.mem (x,v) S ->
  ok S ->
  binds x v S.
Proof.
  introv Hm HokS.
  induction Hm.
  rewrite cons_to_push. eauto.
  destruct y. 
  rewrite cons_to_push in *.
  eapply ok_push_inv in HokS. destruct HokS.
  eapply binds_push_neq. eauto.
  intro; subst. eapply env_mem_dom in Hm. eauto.
Qed.

Lemma binds_middle_exists {A: Set} (S: env A) x (v: A):
  binds x v S ->
  exists S1 S2,
  S = S1 & x ~ v & S2.
Proof.  
  introv Hbi.
  
  induction S using env_ind.
  eapply binds_empty_inv in Hbi. false.
  eapply binds_push_inv in Hbi. 
  destructecs Hbi; subst.
  - exists S (empty: env A). rewrite concat_empty_r. auto.
  - forwards: IHS Hbi. destructecs H.
  subst.
  exists S1 (S2 & x0 ~ v0).
  rewrite ! concat_assoc.
  auto.
Qed.

Lemma ok_dom_remove (A: Set) (e e2: env A):
  ok (e & e2) ->
  dom (e & e2) \- dom e = dom e2.
Proof.
  introv Hok.
  eapply ok_concat_disjoint in Hok.
  simpl_dom.
  rewrite union_remove.
  rewrite remove_same.
  rewrite remove_disjoint. 
  apply union_empty_l.
  eapply disjoint_comm. auto.
Qed.
