(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Export Coq.Program.Basics.
From TLC Require Export LibVar LibLN LibEnv.

Require Export LibNatExtra LibVarExtra LibSetExtra LibEnvExtra LibListExtra LibListReorder LibFreshExtra LibTactics.

(** * Various helper definitions and lemmas *)

Open Scope program_scope.

Section Program.
  Lemma fold_compose : forall {A B C : Type} (g : B -> C) (f : A -> B) x,
      g (f x) = (g ∘ f) x.
  Proof. auto. Qed.

  Definition dec_apply {A B : Type} dec (f : A -> B) g x :=
    match dec x with
    | true => f x
    | false => g x
    end.

  Lemma dec_apply_negb : forall {A B : Type} dec (f : A -> B) g,
    dec_apply (negb ∘ dec) f g = dec_apply dec g f.
  Proof.
    intros. unfold compose, dec_apply.
    apply fun_ext_dep. intros x.
    destruct (dec x); reflexivity.
  Qed.

  Definition ignore_second {A B C : Type} (f : A -> C) (x : A) (n : B) := f x.

  Definition compose12 {A B C D : Type} (g : B -> D -> C) (f : A -> D -> B) (x : A) (n : D) :=
    g (f x n) n.

  Notation " g ∘∘ f " := (compose12 g f) (at level 40, left associativity).

  Lemma fold_compose12 : forall {A B C D : Type} (g : B -> D -> C) (f : A -> D -> B) x n,
      g (f x n) n = (g ∘∘ f) x n.
  Proof. auto. Qed.

  Lemma lift_depth_commute12 :
    forall {N N' V V' D A : Type} {p : N -> D -> N} {q : N' -> D -> N'}
      (g : N' -> V' -> A -> A) (f : N -> V -> A -> A) x y m n,
      (forall d, (f (p m d) x) ∘ (g (q n d) y) = (g (q n d) y) ∘ (f (p m d) x)) ->
      ((fun a d => f (p m d) x a)) ∘∘ ((fun a d => g (q n d) y a))
      = ((fun a d => g (q n d) y a)) ∘∘ ((fun a d => f (p m d) x a)).
  Proof.
    intros N N' V V' D A. intros.
    apply fun_ext_dep. intros a.
    apply fun_ext_dep. intros k.
    unfold compose12.
    rewrite (fold_compose (f _ _)), H.
    auto.
  Qed.
End Program.

Notation " g ∘∘ f "
  := (compose12 g f) (at level 40, left associativity) : program_scope.
Hint Unfold ignore_second compose12 : core.

Module Nat := PeanoNat.Nat.

Notation "xs ~** vs" := (singles (List.rev xs) (List.rev vs))
  (at level 27, left associativity) : env_scope.

Lemma singles_nil : forall A,
    nil ~* (@nil A) = empty.
Proof.
  intros. rewrite singles_def; auto.
Qed.

Lemma singles_push_r : forall A (v : A) (vs : list A) y ys,
    (y :: ys)%list ~* (v :: vs)%list = ys ~* vs & y ~ v.
Proof.
  intros. rewrite singles_def. auto.
Qed.

Lemma singles_push_l : forall A (v : A) (vs : list A) y ys,
    length ys = length vs ->
    (ys ++ y :: nil)%list ~* (vs ++ v :: nil)%list = y ~ v & ys ~* vs.
Proof.
  intros A v vs y ys. gen vs.
  induction ys; induction vs; intros; simpls; try congruence.
  - rewrite singles_push_r, ? singles_nil, concat_empty_l, concat_empty_r; auto.
  - rewrite ? singles_push_r. rewrite IHys; auto.
    rewrite concat_assoc. auto.
Qed.

Lemma singles_rev_push : forall A (v : A) (vs : list A) y ys,
    length ys = length vs ->
    (y :: ys)%list ~** (v :: vs)%list
    = y ~ v & ys ~** vs.
Proof.
  intros. simpl.
  rewrite singles_push_l; auto.
Qed.

Lemma singles_rev_cons : forall A (v : A) (vs : list A) y ys,
    (ys ++ y :: nil)%list ~** (vs ++ v :: nil)%list
    = ys ~** vs & y ~ v.
Proof.
  intros. simpl.
  rewrite ? List.rev_unit.
  rewrite singles_push_r; auto.
Qed.

Lemma fresh_ok_rev_ys : forall A (E : env A) vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    ok E ->
    ok (E & ys ~** vs).
Proof.
  eauto using fresh_ok_ys.
Qed.

Lemma fresh_cons_notin : forall A (E : env A) vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    x \notin dom (E & ys ~* vs).
Proof.
  intros A E vs x. induction vs; induction ys; simpls; try congruence.
  - intros. destruct H. rewrite singles_nil, concat_empty_r. auto.
  - intros. rewrite singles_push_r.
    repeat rewrite dom_concat, notin_union.
    destruct H. destruct H1.
    split; auto 1. split; auto 1.
    rewrite dom_singles by (rewrite ? liblist_length; congruence).
    apply from_list_fresh; auto.
Qed.

Lemma fresh_cons_notin_rev : forall A (E : env A) vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    x \notin dom (E & ys ~** vs).
Proof.
  auto using fresh_cons_notin.
Qed.

Lemma fresh_ok : forall A (E : env A) v vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    ok E ->
    ok (E & ys ~* vs & x ~ v).
Proof.
  intros. apply ok_push; eauto 2 using fresh_ok_ys, fresh_cons_notin.
Qed.

Lemma fresh_ok_rev : forall A (E : env A) v vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    ok E ->
    ok (E & ys ~** vs & x ~ v).
Proof.
  auto using fresh_ok.
Qed.
Hint Resolve fresh_ok_rev : core.


Hint Extern 1 (?E & ?F & ?G = ?E & (?F & ?G)) =>
  rewrite concat_assoc; reflexivity : core.
Hint Extern 1 (?E & ?F & ?G & ?H = ?E & (?F & (?G & ?H))) =>
  rewrite 2 concat_assoc; reflexivity : core.
Hint Extern 1 (?E & ?F & ?G & ?H & ?I = ?E & ?F & (?G & ?H & ?I)) =>
  rewrite 2 concat_assoc; reflexivity : core.

Lemma fv_in_values_empty : forall (A : Type) f,
    @fv_in_values A f empty = \{}.
Proof.
  intros A f. unfold fv_in_values. rewrite empty_def, values_def.
  reflexivity.
Qed.

Lemma fv_in_values_cons : forall A (E1: env A) x v f,
    fv_in_values f (E1 & x ~ v) = fv_in_values f E1 \u (fv_in_values f (x ~ v)).
Proof.
  intros. rewrite union_comm.
  unfold fv_in_values. rewrite single_def, values_def, concat_def.
  simpl. rewrite union_empty_r. reflexivity.
Qed.

Lemma fv_in_values_concat : forall A (E1 E2: env A) f,
    fv_in_values f (E1 & E2) = fv_in_values f E1 \u fv_in_values f E2.
Proof.
  intros A E1 E2. gen E1. induction E2 using env_ind; intros.
  - rewrite concat_empty_r. rewrite <- (union_empty_r (fv_in_values f E1)) at 1.
    f_equal. unfold fv_in_values. rewrite empty_def, values_def; auto.
  - rewrite concat_assoc. rewrite ? fv_in_values_cons.
    rewrite ? IHE2. rewrite <- union_assoc. auto.
Qed.

Ltac intros_fun_ext :=
  repeat (apply fun_ext_dep; intro).

Lemma or_orb_iff_left : forall P Q b1 b2,
    P <-> b1 = true ->
    Q <-> b2 = true ->
    P \/ Q <-> b1 || b2 = true.
Proof.
  intros. destruct H, H0. split; destruct b1, b2; simpl; auto.
  intros. destruct H3; auto.
Qed.

Lemma Is_true_iff_eq_true : forall bl,
    Is_true bl <-> bl = true.
Proof.
  intros; split; auto using Is_true_eq_true, Is_true_eq_left.
Qed.

Lemma list_max (X: list nat):
  forall n,
  n \in (from_list X) ->
  n <= List.fold_right max 0 X.
Proof.
  unfold from_list.
  intros.
  induction X.
  - simpls. rewrite in_empty in H. contradiction.
  - simpls. rewrite in_union in H. destruct H.
    + rewrite in_singleton in H; subst.
      eapply Max.le_max_l.
    + forwards: IHX; eauto.
      eapply Nat.le_trans; eauto.
      eapply Max.le_max_r; eauto.
Qed.

Lemma fset_max_exists (X: fset nat):
  exists m,
  forall n,
  n \in X ->
  n <= m.
Proof.
  destruct (fset_finite X) as [L HL].
  exists (List.fold_right max 0 L).
  intros.
  eapply list_max.
  subst; eauto.
Qed.

Lemma fset_bound_exists (X: fset nat):
  exists m,
  forall n,
  n \in X ->
  n < m.
Proof.
  destruct (fset_max_exists X) as [m Hm].
  exists (S m).
  intros.
  eapply Lt.le_lt_n_Sm.
  eauto.
Qed.

Lemma fun_arg_eq_1 (A B: Type) (f: A -> B) (g: A -> B) (x: A) (y: A):
  f = g ->
  x = y ->
  f x = g y.
Proof.
  intros. subst. eauto.
Qed.


