Set Implicit Arguments.

From TLC Require Import LibVar LibLN LibTactics.

Require Import LibListExtra LibListReorder LibVarExtra LibSetExtra LibTactics.

(** * Lemmas about fresh variables *)

Lemma fresh_cons : forall L n x ys,
    fresh L (S n) (x :: ys) ->
    x \notin L /\ fresh (L \u \{ x}) n ys.
Proof.
  unfold fresh; auto.
Qed.

Lemma fresh_single_in : forall l ys x F,
    x \in F ->
    fresh F l ys ->
    fresh \{ x} l ys.
Proof.
  induction l; induction ys; intros; simpl; auto.
  simpl in H0. destruct H0. split.
  - intros ?H. rewrite in_singleton in H2; subst a; auto.
  - apply fresh_union_l; auto. eapply IHl; eauto.
Qed.

Lemma fresh_from_list : forall ys x L,
    x \notin from_list ys ->
    fresh L (length ys) ys ->
    fresh \{ x} (length ys) ys.
Proof.
  induction ys; intros; simpl; auto.
  rewrite from_list_cons in H. split.
  - notin_solve.
  - simpl in H0. destruct H0.
    fresh_solve. eapply IHys; eauto.
Qed.
Hint Resolve fresh_from_list: core.

Lemma from_list_fresh : forall ys x,
    fresh \{ x} (length ys) ys ->
    x \notin from_list ys.
Proof.
  induction ys; intros; simpl; auto.
  - rewrite from_list_nil; auto.
  - rewrite from_list_cons, notin_union; auto.
Qed.

Lemma fresh_cons_app : forall x ys L,
    fresh L (length (x :: ys)) (x :: ys) ->
    fresh L (length (ys ++ x :: nil)) (ys ++ x :: nil).
Proof.
  intros x. induction ys; simpls; intros; auto.
  destruct H. destruct H0.
  split; auto.
Qed.

Lemma fresh_app : forall xs ys L,
    fresh L (length (xs ++ ys)) (xs ++ ys) ->
    fresh L (length (ys ++ xs)) (ys ++ xs).
Proof.
  induction xs; intros.
  - simpls. rewrite List.app_nil_r. auto.
  - assert ((a :: xs)%list = ((a :: nil) ++ xs)%list) by reflexivity.
    rewrite ? H0 in H. rewrite <- ? List.app_assoc in H.
    clear H0. apply fresh_cons_app in H.
    rewrite <- List.app_assoc in H. apply IHxs in H.
    rewrite <- List.app_assoc in H. apply H.
Qed.

Lemma fresh_rev : forall L xs,
    fresh L (length xs) xs ->
    fresh L (length (List.rev xs)) (List.rev xs).
Proof.
  intros L xs. gen L.
  induction xs using List.rev_ind; simpls; intros; auto.
  rewrite List.rev_app_distr. apply fresh_app.
  apply fresh_app in H. apply fresh_cons_app.
  simpls. destruct H. split; auto.
Qed.
Hint Extern 2 (fresh ?L (length (List.rev ?xs)) (List.rev ?xs)) =>
apply fresh_rev; auto 1 : core.

Lemma fresh_ok_ys : forall A (E : env A) vs x ys,
    fresh (dom E) (S (length ys)) (x :: ys) ->
    length ys = length vs ->
    ok E ->
    ok (E & ys ~* vs).
Proof.
  intros. simpl in H. destruct H.
  assert (fresh (dom E) (length vs) ys) by auto.
  eapply ok_singles; eauto.
  rewrite ? liblist_length.
  auto.
Qed.

Lemma fresh_S_split : forall L n ys',
    fresh L (S n) ys' ->
    exists x ys, ys' = (x :: ys)%list.
Proof.
  intros. unfold fresh in H; destruct ys'.
  - destruct H.
  - exists v ys'; auto.
Qed.

Lemma fresh_length : forall ys L n,
    fresh L n ys ->
    length ys = n.
Proof.
  intros. apply fresh_length in H. rewrite liblist_length in H.
  congruence.
Qed.
Hint Resolve fresh_length : core.

Lemma fresh_length_rev : forall ys L n,
    fresh L n (List.rev ys) ->
    length ys = n.
Proof.
  intros. rewrite <- List.rev_length.
  eauto 2.
Qed.
Hint Resolve fresh_length_rev : core.

Lemma fresh_length_S : forall x ys L n,
    fresh L (S n) (x :: ys) ->
    length ys = n.
Proof.
  intros; apply (fresh_length _ L).
  destruct (fresh_cons H); auto.
Qed.
Hint Resolve fresh_length_S : core.

Lemma fresh_length_rev_S : forall x ys L n,
    fresh L (S n) (x :: (List.rev ys)) ->
    length ys = n.
Proof.
  intros. rewrite <- List.rev_length.
  eauto.
Qed.
Hint Resolve fresh_length_rev_S : core.

(* weaker than notin_vars_map *)
Lemma fresh_vars_map : forall x y ys,
    fresh \{ x} (length ys) ys ->
    List.map (subst_fvar x y) ys = ys.
Proof.
  intros. induction ys.
  - auto.
  - simpls. destruct H. f_equal.
    + unfold subst_fvar; cases_if.
      * exfalso. apply H. rewrite in_singleton. auto.
      * auto.
    + apply IHys; auto.
Qed.
(* stronger than fresh_vars_map *)
Lemma notin_vars_map : forall x y ys,
    x \notin from_list ys ->
    List.map (subst_fvar x y) ys = ys.
Proof.
  intros. induction ys.
  - auto.
  - rewrite from_list_cons, notin_union, notin_singleton in H.
    destructecs H.
    simpls. f_equal.
    + unfold subst_fvar; cases_if; auto.
    + apply IHys; auto.
Qed.

Lemma fresh_remove X Y ws:
  fresh X (length ws) ws ->
  fresh (X \- Y) (length ws) ws.
Proof.
  introv Hfr.
  induction ws.
  - unfold fresh. simpls. auto. 
  - simpls.
    destruct Hfr.
    fresh_solve.
    + eapply notin_remove.
      auto.
    + fresh_solve.
      apply IHws. fresh_solve.
Qed.


Lemma fresh_remove_from_list X ws:
  fresh \{} (length ws) ws ->
  fresh (X \- (from_list ws)) (length ws) ws.
Proof.
  introv Hfr.
  induction ws.
  - unfold fresh. simpls. auto. 
  - simpls.
    destruct Hfr.
    eapply fresh_union_r2 in H0.


    rewrite ! from_list_cons.
    splits.
    eapply notin_remove. right. rewrite in_union, in_singleton; auto.
    fresh_solve.
    rewrite union_comm.
    rewrite remove_union.
    eapply fresh_remove.
    eapply IHws; eauto.
Qed.



Lemma fresh_append_r X xs ys n:
  fresh X n (xs ++ ys) ->
  fresh X (length ys) ys.
Proof.
  intro Hfr.
  replace n with (length (xs ++ ys)) in Hfr by (eapply fresh_length in Hfr; eauto).
  induction xs.
  simpls. auto.
  rewrite <- List.app_comm_cons in Hfr.
  apply IHxs.
  simpls.
  destruct Hfr.
  auto.
Qed.
Lemma fresh_append_l X xs ys n:
  fresh X n (xs ++ ys) ->
  fresh X (length xs) xs.
Proof.
  intro Hfr.
  replace n with (length (xs ++ ys)) in Hfr by (eapply fresh_length in Hfr; eauto).
  eapply fresh_app in Hfr.
  eapply fresh_append_r; eauto.
Qed.
Lemma fresh_resize_length (L : vars) (n : nat) (xs : list var):
  fresh L n xs ->
  fresh L (length xs) xs.
Proof.
  introv Hfr.
  lets: Hfr.
  eapply fresh_length in Hfr.
  rewrite Hfr.
  auto.
Qed.
Lemma fresh_append_fresh_l xs ys n:
  fresh \{} n (xs ++ ys) ->
  fresh (from_list ys) (length xs) xs.
Proof.
  introv Hfr.
  replace n with (length (xs ++ ys)) in Hfr by (eapply fresh_length in Hfr; eauto).
  induction xs.
  - simpls. auto.
  - simpls.
    destructc Hfr as Ha Hfr.
    lets: fresh_append_l Hfr.
    lets: fresh_append_r Hfr.
    split.
    eapply from_list_fresh. fresh_solve.
    fresh_solve. eapply IHxs. fresh_solve.
Qed.
Lemma fresh_append_fresh_r xs ys n:
  fresh \{} n (xs ++ ys) ->
  fresh (from_list xs) (length ys) ys.
Proof.
  introv Hfr.
  replace n with (length (xs ++ ys)) in Hfr by (eapply fresh_length in Hfr; eauto).
  eapply fresh_app in Hfr.
  eapply fresh_append_fresh_l; eauto.
Qed.

Lemma fresh_forall_notin L xs X:
  fresh L (length xs) xs ->
  (forall x, x \notin L -> x \notin X) ->
  fresh X (length xs) xs.
Proof.
  introv Hfr Hn.

  induction xs; simpls. auto.
  destruct Hfr. split; auto.
Qed.

Lemma fresh_in_notin x xs X:
  fresh X (length xs) xs ->
  LibList.mem x xs ->
  x \notin X.
Proof.
  introv Hrf Hin.
  induction Hin.
  - unfold fresh in Hrf. simpls. destruct Hrf. auto.
  - unfold fresh in Hrf. simpls. fold fresh in Hrf. destruct Hrf. apply IHHin. fresh_solve.
Qed.

Lemma fresh_push (L: vars) x y (xs ys: list var):
  fresh L (length ys) xs ->
  x \notin (L \u (from_list xs)) ->
  fresh L (length (y::ys)) (x::xs).
Proof.
  introv Hfr Hn.
  fresh_solve.
  eapply fresh_union_l.
  fresh_solve.
  pose proof Hfr as Hl.
  eapply fresh_length in Hl.
  rewrite <- Hl.
  rewrite <- Hl in Hfr.
  eapply fresh_from_list.
  notin_solve. eauto.
Qed.

Lemma fresh_nth_eq xs l n1 n2 v:
  fresh \{} l xs ->
  LibList.Nth n1 xs v ->
  LibList.Nth n2 xs v ->
  n1 = n2.
Proof.
  introv Hfr Hn1 Hn2.
  destruct (PeanoNat.Nat.lt_total n1 n2) as [? | [? | ?]].
  - false.
    eapply list_Nth_cut in Hn2.
    destructc Hn2 as xs1 xs2 Hxs Hl; subst.
    eapply list_Nth_app_invl in Hn1.
    eapply fresh_append_fresh_r in Hfr.
    eapply fresh_cons in Hfr. destruct Hfr.

    apply H0.
    eapply list_mem_in_fromlist.
    eapply LibList.Nth_mem.
    eauto.
    auto.
  - auto.
  - false.
    eapply list_Nth_cut in Hn1.
    destructc Hn1 as xs1 xs2 Hxs Hl; subst.
    eapply list_Nth_app_invl in Hn2.
    eapply fresh_append_fresh_r in Hfr.
    eapply fresh_cons in Hfr. destruct Hfr.

    apply H0.
    eapply list_mem_in_fromlist.
    eapply LibList.Nth_mem.
    eauto.
    auto.
Qed.

Lemma fresh_mem_false v l xs:
  fresh \{v} l xs ->
  LibList.mem v xs ->
  False.
Proof.
  introv Hfr Hn.
  eapply fresh_in_notin with (X:=\{v}) in Hn.
  rewrite notin_singleton in Hn; contradiction.
  eapply fresh_resize_length in Hfr.
  eauto.
Qed.

Lemma fresh_subset E F xs:
  E \c F ->
  fresh F (length xs) xs ->
  fresh E (length xs) xs.
Proof.
  intros.
  eapply fresh_resize_length in H0.
  replace F with (E \u F) in H0.
  fresh_solve.
  eapply fset_extens.  
  pose proof (subset_union_2 H (@subset_refl _ F)). rewrite union_same in H1. auto.
  eapply subset_union_weak_r.
Qed.

Lemma fresh_in_notin_list L x z:
  fresh L (length x) x ->
  z \in L ->
  z \notin (from_list x).
Proof.
  introv Hfr Hn.
  eapply fresh_single_notin.
  eapply fresh_subset.
  eapply subset_singleton. eauto. eauto.
Qed.


Inductive freshy: vars -> vars -> list var -> list var -> Prop :=
| freshy_nil L Lx:
  freshy L Lx nil nil
| freshy_fresh L Lx xs ys x y: 
  x \notin (L \u Lx) ->
  x \notin (from_list xs) ->
  freshy L Lx xs ys ->
  freshy L Lx (x::xs) (y::ys)
| freshy_same L Lx xs ys x: 
  freshy L Lx xs ys ->
  x \in Lx ->
  freshy L Lx (x::xs) (x::ys).

Lemma fresh_freshy L Lx ys xs:
  fresh (L \u Lx) (length ys) xs ->
  freshy L Lx xs ys.
Proof.
  introv Hfr.
  reverts ys.
  induction xs; intros.
  - eapply fresh_length in Hfr.
    destruct ys. eapply freshy_nil. simpls. discriminate Hfr.
  - pose proof Hfr as Hl.
    eapply fresh_length in Hl.
    destruct ys. simpls. discriminate Hl.
    simpls.
    destruct Hfr.
    eapply freshy_fresh; eauto.
    eapply fresh_union_r2 in H0.
    eapply fresh_single_notin; eauto.
Qed.

Lemma freshy_union_r1 (xs ys : list var) (L1 L2 Lx: fset var):
  freshy (L1 \u L2) Lx ys xs ->
  freshy L1 Lx ys xs.
Proof.
  introv Hfr.
  remember (L1 \u L2) as L.
  induction Hfr; intros; subst.
  - apply freshy_nil.
  - apply freshy_fresh.
    rewrite notin_union in H. destruct H. auto.
    auto.
    auto.
  - apply freshy_same; eauto.
Qed.
Lemma freshy_union_r2 (xs ys : list var) (L1 L2 Lx: fset var):
  freshy (L1 \u L2) Lx ys xs ->
  freshy L2 Lx ys xs.
Proof.
  introv Hfr.
  rewrite union_comm in Hfr.
  eapply freshy_union_r1; eauto.
Qed.

Lemma freshy_allow_singleton (xs ys : list var) (L Lx: fset var) y:
  freshy L Lx xs ys ->
  y \notin (from_list xs) ->
  freshy L (Lx \u \{y}) xs ys.
Proof.
  introv Hfr Hn.
  induction Hfr; intros; subst; simpls.
  - apply freshy_nil.
  - rewrite notin_union in H. destruct H.
    rewrite from_list_cons in Hn.
    rewrite notin_union in Hn. destruct Hn.
    apply freshy_fresh.
    repeat rewrite notin_union.
    splits; eauto.
    auto.
    auto.
  - rewrite from_list_cons in Hn.
    rewrite notin_union in Hn. destruct Hn.
    apply freshy_same; eauto.
    rewrite in_union.
    auto.
Qed.

Fixpoint list_exclude (L: vars) (xs: list var) :=
  match xs with
  | nil => nil
  | (x::xss)%list => let xse := (list_exclude L xss) in if (fv_decide x L) then (xse)%list else (x::xse)%list
  end.

Lemma list_exclude_notin x L (xs: list var):
  x \notin (from_list xs) ->
  x \notin (from_list (list_exclude L xs)).
Proof.
  introv Hn.
  induction xs; simpls.
  - rewrite from_list_nil. notin_solve.
  -
    rewrite from_list_cons in Hn.
    rewrite notin_union in Hn. destruct Hn.
    cases_if; auto.
    rewrite from_list_cons.
    rewrite notin_union.
    split; auto.
Qed.

Lemma freshy_exclude (xs ys : list var) (L Lx: fset var):
  freshy L Lx xs ys ->
  fresh L (length (list_exclude Lx xs)) (list_exclude Lx xs).
Proof.
  introv Hfr.
  induction Hfr; intros; subst; simpls.
  - auto.
  - rewrite notin_union in H. destruct H.
    cases_if.
    eapply fresh_push; eauto.
    notin_solve.
    eapply list_exclude_notin; eauto.
  - cases_if; auto.
Qed.

Lemma fresh_swap: forall (L : vars) (xs : list var) (X : list var), fresh L (length xs) xs -> fresh (from_list xs) (length X) X -> fresh (from_list X) (length xs) xs.
Proof.
  intros.
  reverts X.
  induction xs; intros; fresh_solve; eauto;
  rewrite from_list_cons in H0.
eapply fresh_union_r1 in H0. eapply from_list_fresh. eauto.
fresh_solve. 
eapply IHxs; fresh_solve.
Qed.
  

Lemma fresh_append xs ys L:
  fresh L (length (xs)) (xs) ->
  fresh (L \u (from_list xs)) (length (ys)) (ys) ->
  fresh L (length (xs ++ ys)) (xs++ys).
Proof.
  introv Hfrx Hfry.
  reverts xs L.
  induction ys; intros.
  rewrite List.app_nil_r. auto.
  eapply fresh_app.
  rewrite <- List.app_comm_cons.
  fresh_solve.
  eapply fresh_app.
  eapply IHys; eauto.
Qed.


Lemma fresh_reorder xs ns ys L:
  list_reorder xs ns ys ->
  LibList.noduplicates ns ->
  fresh L (length xs) xs ->
  fresh L (length ys) ys.
Proof.
  introv Hre Hnd Hfr.
  induction Hre.
  - eauto. 
  - fresh_solve.
    eapply fresh_in_notin; eauto. eapply LibList.Nth_mem; eauto.
    eapply fresh_union_l.  inversions Hnd. eauto. eauto.  fresh_solve.
    inversions Hnd. eapply fresh_from_list; eauto. intro. 
    apply H2.
    eapply list_in_fromlist_mem in H0.
    destruct (LibList.mem_Nth H0).
    forwards m Hx Hm: list_reorder_nth_Nth; eauto.
    eapply fresh_nth_eq in Hm. 3: apply H. 2: eauto. subst. eapply LibList.Nth_mem. eauto. 
Qed.

Lemma fresh_reorder_l xs ns ys zs:
  list_reorder xs ns ys ->
  fresh (from_list xs) (length zs) zs ->
  fresh (from_list ys) (length zs) zs.
Proof.
  introv Hre Hfr.
  induction Hre.
  - rewrite from_list_nil. eauto.
  - rewrite from_list_cons. fresh_solve;  eauto. eapply fresh_from_list; eauto. eapply fresh_in_notin_list; eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
Qed.

Lemma fresh_insert_inv L xs1 x xs2:
  fresh L (length (xs1 ++ x :: xs2)) (xs1 ++ x :: xs2) ->
  fresh L (length (xs1 ++ xs2)) (xs1 ++ xs2).
Proof.
  introv Hfr.
  lets: fresh_append_l Hfr.
  lets: fresh_append_r Hfr.
  simpls.
  destructc H0 as ? ?.
  forwards: fresh_append_fresh_l xs1; eauto using fresh_empty.
  forwards: fresh_append_fresh_r xs1; eauto using fresh_empty.
  destructc H3 as ? ?.
  rewrite from_list_cons in H2.
  eapply fresh_append; eauto.
Qed.


Lemma fresh_forall_in_from_list X l:
  fresh \{} (length l) l ->
  (forall x, x \in (from_list l) -> x \notin X) ->
  fresh X (length l) l.
Proof.
  intros.
  induction l.
  - auto.
  - fresh_solve.
    rewrite from_list_cons in H0. eapply H0. rewrite in_union. left. rewrite in_singleton; auto.
    eapply fresh_union_l; eauto.
    eapply IHl; eauto.
    intros. eapply H0. rewrite from_list_cons. rewrite in_union; eauto.
Qed.

Lemma fresh_filter X a l:
  fresh X (length l) l ->
  fresh X (length (LibList.filter a l)) (LibList.filter a l).
Proof.
  introv Hfr.
  reverts X.
  induction l; intros.
  - rewrite LibList.filter_nil; eauto.
  - rewrite LibList.filter_cons.
    cases_if.
    fresh_solve. eapply fresh_union_l; eauto. eapply IHl. simpls. destruct Hfr. eapply fresh_union_r1 in H0. auto.
Qed.

Lemma fresh_remove_r X a l:
  fresh X (length l) l ->
  fresh X (length (LibList.remove a l)) (LibList.remove a l).
Proof.
  introv Hfr.
  unfold LibList.remove.
  eapply fresh_filter; eauto.
Qed.

Lemma fresh_remove_duplicates l:
  fresh \{} (length (LibList.remove_duplicates l)) (LibList.remove_duplicates l).
Proof.
  induction l.
  - simpls. auto.
  - simpls; splits; eauto. rewrite union_empty_l.
    eapply fresh_from_list. eapply notin_from_list_remove.
    eapply fresh_remove_r; eauto.
Qed.

Lemma fresh_noduplicates l L:
  fresh L (length l) l ->
  LibList.noduplicates l.
Proof.
  introv Hfr.
  induction l.
  - constructor.
  - simpls. destructc Hfr as Hn Hfr. constructor.
    intro. eapply fresh_mem_false; eauto. eapply fresh_union_r2; eauto.
    eauto.
Qed.

Lemma list_noduplicates_fresh l:
  LibList.noduplicates l ->
  fresh \{} (length l) l.
Proof.
  introv Hnd.
  induction l.
  - constructor.
  - inversions Hnd. simpls. split. eauto. fresh_solve; eauto. eapply fresh_from_list; eauto. intro. apply H1. eapply list_in_fromlist_mem; eauto.
Qed.


Lemma fresh_in_from_list_notin x xs L:
  x \in (from_list xs) ->
  fresh L (length xs) xs ->
  x \notin L.
Proof.
  intros.
  eapply fresh_in_notin; eauto.
  eapply list_in_fromlist_mem; eauto.
Qed.

Lemma fresh_disjoint L ys:
  fresh L (length ys) ys ->
  disjoint L (from_list ys).
Proof.
  introv Hfr.
  hnf. eapply fset_extens; intro; intro.
  - false. rewrite in_inter in H. destruct H.
    eapply fresh_in_from_list_notin; eauto.
  - rewrite in_empty in H. false H.
Qed.
(* disjoint set can replace any set *)
Lemma disjoint_fresh L1 L2 ys:
  fresh L2 (length ys) ys ->
  disjoint L1 (from_list ys) ->
  fresh L1 (length ys) ys.
Proof.
  introv Hfr Hdi.
  induction ys.
  - auto.
  - simpls.
    destructecs Hfr.
    eapply fresh_union_r in Hfr.
    destructecs Hfr.
    rewrite from_list_cons in Hdi.
    eapply disjoint_union_r_inv in Hdi.
    destructecs Hdi.
    eapply disjoint_singleton_inv_r in Hdi0.
    split; auto.
Qed.

Lemma fresh_take L ys n:
  fresh L (length ys) ys ->
  fresh L (length (LibList.take n ys)) (LibList.take n ys).
Proof.
  introv Hfr.
  reverts n.
  induction ys; intros; destruct n.
  - rewrite LibList.take_nil. auto.
  - rewrite LibList.take_nil. auto.
  - rewrite LibList.take_zero. auto.
  - rewrite LibList.take_succ.
    inversions Hfr.
    forwards: IHys n. fresh_solve.
    fresh_solve.
    eapply fresh_union_l. auto.
    eapply fresh_from_list.
    eapply notin_from_list_take_from_list.
    eapply fresh_single_notin. 
    eapply fresh_union_r2; eauto.
    eauto.
Qed.
Lemma fresh_drop L ys n:
  fresh L (length ys) ys ->
  fresh L (length (LibList.drop n ys)) (LibList.drop n ys).
Proof.
  introv Hfr.
  reverts n.
  induction ys; intros; destruct n.
  - rewrite LibList.drop_nil. auto.
  - rewrite LibList.drop_nil. auto.
  - rewrite LibList.drop_zero. auto.
  - rewrite LibList.drop_succ.
    inversions Hfr.
    forwards: IHys n. fresh_solve.
    fresh_solve.
Qed.
Lemma fresh_drop_take L ys n:
  n <= length ys ->
  fresh L (length ys) ys ->
  fresh (from_list (LibList.take n ys)) (length (LibList.drop n ys)) (LibList.drop n ys).
Proof.
  introv Hl Hfr.
  erewrite <- LibList.list_eq_take_app_drop in Hfr at 2.
  -
  rewrite liblist_app in Hfr.
  eapply fresh_resize_length.
  eapply fresh_append_fresh_r.
  eapply fresh_empty; eauto.
  -
  rewrite LibNat.le_peano, liblist_length.
  auto.
Qed.
Lemma fresh_take_drop L ys n:
  n <= length ys ->
  fresh L (length ys) ys ->
  fresh (from_list (LibList.drop n ys)) (length (LibList.take n ys)) (LibList.take n ys).
Proof.
  introv Hl Hfr.
  erewrite <- LibList.list_eq_take_app_drop in Hfr at 2.
  -
  rewrite liblist_app in Hfr.
  eapply fresh_resize_length.
  eapply fresh_append_fresh_l.
  eapply fresh_empty; eauto.
  -
  rewrite LibNat.le_peano, liblist_length.
  auto.
Qed.

Lemma fresh_removeat_r L ys n:
  n < length ys ->
  fresh L (length ys) ys ->
  fresh L (length (list_removeat n ys)) (list_removeat n ys).
Proof.
  introv Hl Hfr.
  forwards: LibList.Nth_inbound_inv. rewrite LibNat.lt_peano, liblist_length. eauto. destructecs H.
  eapply fresh_insert_inv.
  rewrite <-list_Nth_cut_take_drop; eauto.
Qed.

Lemma fresh_removeat_rl xs ys m n:
  n < length ys ->
  fresh (from_list xs) (length ys) ys ->
  fresh (from_list (list_removeat m xs)) (length (list_removeat n ys)) (list_removeat n ys).
Proof.
  introv Hl Hfr.
  eapply fresh_removeat_r. auto.
  eapply fresh_subset. eapply from_list_removeat_subset.
  auto.
Qed.

Lemma list_fresh_Nth_notin_take L l xs n y:
  fresh L l xs ->
  LibList.Nth n xs y ->
  y \notin (from_list (LibList.take n xs)).
Proof.
  introv Hfr Hn.
  forwards: LibList.Nth_inbound; fweauto.
  erewrite list_Nth_cut_take_drop in Hfr; eauto.
  eapply fresh_empty in Hfr.
  forwards Hfrt: fresh_append_fresh_r Hfr.
  inversions Hfrt.
  auto.
Qed.
Lemma list_fresh_Nth_notin_drop L l xs n y:
  fresh L l xs ->
  LibList.Nth n xs y ->
  y \notin (from_list (LibList.drop (S n) xs)).
Proof.
  introv Hfr Hn.
  forwards: LibList.Nth_inbound; fweauto.
  erewrite list_Nth_cut_take_drop in Hfr; eauto.
  eapply fresh_empty in Hfr.
  forwards Hfrt: fresh_append_r Hfr.
  inversions Hfrt.
  eapply fresh_single_notin.
  eapply fresh_union_r2; eauto.
Qed.
Lemma fresh_notin_drop_S xs x n:
  fresh \{} (length xs) xs ->
  LibList.Nth n xs x ->
  x \notin from_list (LibList.drop (S n) xs).
Proof.
  introv Hfr Hn.
  reverts xs; induction n; intros.
  - inversions Hn.
    simpls. destruct Hfr. eapply from_list_fresh; eauto.
  - inversions Hn.
    simpl in Hfr. destruct Hfr.
    rewrite LibList.drop_succ.
    eauto.
Qed.
Lemma fresh_notin_take xs x n:
  fresh \{} (length xs) xs ->
  LibList.Nth n xs x ->
  x \notin from_list (LibList.take n xs).
Proof.
  introv Hfr Hn.
  reverts xs; induction n; intros.
  - inversions Hn.
    simpls. rewrite from_list_nil. eauto.
  - inversions Hn.
    simpl in Hfr. destruct Hfr.
    rewrite LibList.take_succ.
    rewrite from_list_cons.
    rewrite notin_union. split.
    eapply fresh_in_notin with (xs:=l). eauto. eapply LibList.Nth_mem; eauto.
    eauto.
Qed.

Lemma fresh_nth_ne (xs : list var) (n1 n2 : nat) (v1 v2 : var):
  fresh \{} (length xs) xs ->
  LibList.Nth n1 xs v1 ->
  LibList.Nth n2 xs v2 ->
  n1 <> n2 ->
  v1 <> v2.
Proof.
  introv Hfr Hn1 Hn2 Hn.
  intro. subst.
  eapply Hn. eapply fresh_nth_eq; eauto.
Qed.
