Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
         TightTyping RecordTypes InertContext PreciseTyping
        UnboundSubstitution Weakening Narrowing EWeakening Sealing
        ConfigTyping TFree Deref
        Closedness.

(** * Elimination of free variables *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* Instances of subst_var for typing contexts *)
Instance subst_var_ctx: SubstVar ctx := fun x y G => subst_ctx x y G.
Instance subst_var_objctx: SubstVar objctx := fun x y G => subst_objctx x y G.

(* Specifies how fresh free variables are removed *)
Inductive defree_spec: Type :=
  (* Although the target variable could be just one, here we have a list of variables.
     That is needed for transitivity, where the two specs are concatenated and can use different target variables *)
| defree_spec_empty
| defree_spec_subst (d: defree_spec) (x: var) (y: var)
| defree_spec_seal (d: defree_spec) (y: var)
.

Inductive defree_ctx: Type :=
  defree_ctx_with_null (G: ctx) (e: renv) (L: vars) (null: var): defree_ctx.

Definition defree_ctx_fv dfc :=
  match dfc with
  | (defree_ctx_with_null G e L null) =>
    ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u L \u \{ null }
  end.

Inductive defree_ctx_ok: defree_ctx -> Prop :=
  defree_ctx_ok_null G e L null:
    ctx_ok G ->
    ctx_weakens G ->
    ok e ->
    defree_ctx_ok (defree_ctx_with_null G e L null)
with defree_ctx_get: defree_ctx -> ctx -> renv -> Prop :=
  defree_ctx_get_null G e L null:
    defree_ctx_get (defree_ctx_with_null G e L null) G e
with defree_ctx_closed: defree_ctx -> Prop :=
  defree_ctx_closed_null G e L null:
    sealed G ->
    defree_ctx_closed (defree_ctx_with_null G e L null).

Inductive defree_ctx_replacement: defree_ctx -> var -> var -> Prop := 
  defree_ctx_replacement_with_null (G: ctx) (e: renv) (L: vars) (null: var) x:
    x §# G ->
    x \notin (ctx_fv G) ->
    renv_notin x e ->
    x \notin L ->
    x <> null ->
    defree_ctx_replacement (defree_ctx_with_null G e L null) x null
with defree_ctx_null: defree_ctx -> var -> Prop :=
  defree_ctx_null_with_null (G: ctx) (e: renv) (L: vars) (null: var):
    defree_ctx_null (defree_ctx_with_null G e L null) null
with defree_spec_valid: defree_ctx -> defree_spec -> Prop :=
| dfree_spec_valid_empty dfc: defree_spec_valid dfc defree_spec_empty
| dfree_spec_valid_subst (dfc: defree_ctx) (dfs: defree_spec) (x y: var):
    defree_spec_valid dfc dfs ->
    defree_ctx_replacement dfc x y ->
    defree_spec_valid dfc (defree_spec_subst dfs x y)
|   dfree_spec_valid_seal (dfc: defree_ctx) (dfs: defree_spec) (null: var):
    defree_spec_valid dfc dfs ->
    defree_ctx_closed dfc ->
    defree_ctx_null dfc null ->
    defree_spec_valid dfc (defree_spec_seal dfs null)
.
Hint Constructors defree_ctx_replacement defree_ctx_null defree_spec_valid: core.
Hint Constructors defree_spec_valid defree_ctx_ok: core.


(* Removes fresh free variables from a type according to a spec *)
Fixpoint defree_by_rec {A} {SU: SubstVar A} {AO: Sealable A} (n: nat) (dfs: defree_spec) (a: A) :=
  match dfs with
  (* Empty spec - not change *)
  | defree_spec_empty => a
  (* Remove a single variable and replace it by another *)
  | defree_spec_subst dfs x y => ((defree_by_rec n dfs a) /[x -> y])
  | defree_spec_seal dfs y => (seal_rec n y (defree_by_rec n dfs a))
  end.

Notation defree_by:= (defree_by_rec 0) (only parsing).

Fixpoint defree_spec_concat (dfs1 dfs2: defree_spec):=
  match dfs2 with
  | defree_spec_empty => dfs1
  | defree_spec_subst dfs x y => (defree_spec_subst (defree_spec_concat dfs1 dfs) x y)
  | defree_spec_seal dfs y => (defree_spec_seal (defree_spec_concat dfs1 dfs) y)
  end.
(* Removes some fresh free variables *)
Inductive defree {A} {SU: SubstVar A} {AO: Sealable A}: defree_ctx->A->A->Prop :=
| defree_by_spec dfc dfs T1:
    defree_spec_valid dfc dfs ->
    defree dfc T1 (defree_by dfs T1).
Hint Constructors defree: core.

Section Defree.
Context {A: Set} {SU: SubstVar A} {ASl: Sealable A} {AF: FreeVar A} {AO: Openable A}.
Context {AS : AbstractSyntax AO AF SU} {AFSC: FvSubstCases AS} {AFSlC: FvSealCases AS ASl} {ASF: SubstFresh AS}.
Context {AOT: OpenTwice AO} {AOC: OpenCommut AO} {ASO: SealOpen AS ASl} {ASC: SealClosed AS ASl} {AFOC: FvOpenCases AS}.

Lemma defree_by_concat m dfs1 dfs2 (T: A):
  defree_by_rec m dfs2 (defree_by_rec m dfs1 T) = defree_by_rec m (defree_spec_concat dfs1 dfs2) T.
Proof.
  induction dfs2.
  - simpl. auto.
  - simpl. rewrite IHdfs2. auto.
  - simpl. rewrite IHdfs2. auto.
Qed.

Lemma defree_spec_valid_concat dfc dfs1 dfs2:
  defree_spec_valid dfc dfs1 ->
  defree_spec_valid dfc dfs2 ->
  defree_spec_valid dfc (defree_spec_concat dfs1 dfs2).
Proof.
  introv H1 H2.
  induction H2.
  - simpl. auto.
  - simpl. constructor; eauto.
  - simpl. constructor; eauto.
Qed.

Lemma defree_trans dfc (T1 T2 T3: A):
  defree dfc T1 T2 ->
  defree dfc T2 T3 ->
  defree dfc T1 T3.
Proof.
  introv H1 H2.
  inversions H1.
  inversions H2.
  rewrite defree_by_concat.
  eapply defree_by_spec.
  eapply defree_spec_valid_concat; eauto.
Qed.

Lemma defree_refl dfc (T1: A):
  defree dfc T1 T1.
Proof.
  assert (defree_spec_valid dfc defree_spec_empty). constructor.
  epose proof (defree_by_spec T1 H). simpls. auto.
Qed.



Lemma defree_fv2 dfs dfc n (T: A) x:
  defree_spec_valid dfc dfs ->
  x \in fv (defree_by_rec n dfs T) ->
  x \in fv T \u (defree_ctx_fv dfc).
Proof.
  introv Hva Hin.
  induction Hva.
  + simpls. rewrite in_union. auto.
  +

    simpls.
    destruct (fv_subst_cases x0 y (defree_by_rec n dfs T)) as [[Hfv ?] | [? [? Hfv]]].
    - rewrite Hfv in Hin. auto.
    - rewrite Hfv in Hin.
      rewrite in_union in Hin.
      destruct Hin as [Hin | Hin].
      * rewrite in_remove in Hin. destruct Hin. auto.
      *
        inversions H.
        rewrite in_singleton in Hin; subst; simpls.
        repeat rewrite in_union. repeat right. rewrite in_singleton. auto.
  +
    inversions H0. inversions H.
    simpls.
    destruct (fv_seal_cases null (defree_by_rec n dfs T) n) as [Hfv | Hfv].
    - rewrite Hfv in Hin. auto.
    - rewrite Hfv in Hin.
      rewrite in_union in Hin.
      destruct Hin as [Hin | Hin].
      rewrite in_singleton in Hin; subst.
      repeat rewrite in_union. repeat right. rewrite in_singleton. auto.
      eauto.
Qed.

(* If all free variables are not fresh in a closed type, then defree does not change the type *)
Lemma defree_fv_same dfs dfc n (U: A):
  defree_spec_valid dfc dfs ->
  (fv U) \c (defree_ctx_fv dfc) ->
  sealed U ->
  (defree_by_rec n dfs U) = U.
Proof.
  introv Hva Hfv Hcl.
  induction Hva.
  - auto.
  - inversions H.
    unfold defree_ctx_fv in *.
    unfold renv_notin in H2. rewrite notin_union in H2. destruct H2.
    simpls.
    rewrite subst_fresh; eauto.
    rewrite (IHHva Hfv).
    intro HfvU.
    apply Hfv in HfvU.
    repeat rewrite in_union in HfvU; destruct_all; try contradiction. rewrite in_singleton in H5. contradiction.
  - unfold defree_ctx_fv in *.
    simpls.
    rewrite (IHHva Hfv).
    eapply sealed_sealed_rec; eauto.
Qed.

(* One free variable can be removed *)
Lemma defree_one m dfc (T:A) x y:
  defree_ctx_replacement dfc x y ->
  exists dfs,
    defree_spec_valid dfc dfs /\
    x \notin fv (defree_by_rec m dfs T).
Proof.
  introv Hrpl.
  inversions Hrpl.
  exists (defree_spec_subst defree_spec_empty x y).
  splits.
  econstructor; eauto.
  eapply subst_notfv.
  intro; subst.
  contradiction.
Qed.

Lemma fv_decide5 (a:var) X1 X2 X3 X4 X5:
  (a \in X1 \/ a \in X2 \/ a \in X3 \/ a \in X4 \/ a \in X5) \/
  (a \notin X1 /\ a \notin X2 /\ a \notin X3 /\ a \notin X4 /\ a \notin X5).
Proof.
  destruct (fv_decide a X1);
  destruct (fv_decide a X2);
  destruct (fv_decide a X3);
  destruct (fv_decide a X4);
  destruct (fv_decide a X5); auto 6.
Qed.


Lemma defree_fv_append x n dfs0 dfs G e L null (T: A):
  x \notin fv (defree_by_rec n dfs0 T) ->
  defree_spec_valid (defree_ctx_with_null G e L null) dfs ->
  x \notin fv (defree_by_rec n (defree_spec_concat dfs0 dfs) T) \/
  x \in defree_ctx_fv (defree_ctx_with_null G e L null).
Proof.
  introv Hn Hv.

  destruct (fv_decide x (fv (defree_by_rec n (defree_spec_concat dfs0 dfs) T))).
  2: auto.
  right.

  rewrite <- defree_by_concat in m.

  eapply defree_fv2 in m.
  2: eapply Hv.
  rewrite in_union in m.
  destruct m.
  - contradiction.
  - auto.
Qed.

Lemma defree_suffix m dfc (T: A) l1 l2:
  defree_ctx_ok dfc ->
  from_list (l1 ++ l2) = (fv T) ->
  exists dfs,
    defree_spec_valid dfc dfs /\
    forall x,
    x \in (from_list l2) -> (x \notin fv (defree_by_rec m dfs T) \/ x \in (defree_ctx_fv dfc)).
Proof.
  introv Hok Hfl.
  inversions Hok.
  revert dependent l1.
  revert dependent T.
  induction l2.
  - intros.
    exists defree_spec_empty.
    splits; eauto.
    introv Hl. rewrite from_list_nil in Hl. rewrite in_empty in Hl. contradiction.
  - intros.
    assert ((app (app l1 (cons a nil)) l2) = (app l1 (cons a l2))) as Heqlist. {
      clear Hfl.
      rewrite List.app_assoc_reverse.
      rewrite <- List.app_comm_cons.
      rewrite List.app_nil_l.
      auto.
    }
    specialize (IHl2 T (l1 ++ a :: nil)%list).
    rewrite Heqlist in IHl2.
    specialize (IHl2 Hfl).
    destructc IHl2 as dfs0 ? ?.

    destruct (fv_decide a (defree_ctx_fv (defree_ctx_with_null G e L null))).
    + (* a is in G or e *)
     exists dfs0.
     splits; eauto.
     intros.
     rename H4 into Hx.
     rewrite from_list_cons in Hx.
     rewrite in_union in Hx.
     destruct Hx as [Hx | Hx].
     * rewrite in_singleton in Hx; subst. destruct_all; auto 6.
     * auto.
    + (* a not in G or e *)
      unfold defree_ctx_fv in n. repeat rewrite notin_union in n.
      destructc n as HnGd HnGf Hned Hnef HneL Hnn.
      assert (renv_notin a e) as Hne by eauto.
      assert (defree_ctx_replacement (defree_ctx_with_null G e L null) a null) as Hrpl by eauto.
      destructc (defree_one m (defree_by_rec m dfs0 T) Hrpl) as dfs Hdf HnTN.
      rewrite defree_by_concat in HnTN.
      exists (defree_spec_concat dfs0 dfs).
      splits; eauto.
      eapply defree_spec_valid_concat; eauto.
      introv Hfl2.
      rewrite from_list_cons in Hfl2.
      rewrite in_union in Hfl2.
      destruct Hfl2.
      * rewrite in_singleton in H4; subst.
        simpl.
        eauto.
      * specialize (H3 x H4).
        destruct H3; auto.
        eapply defree_fv_append; eauto.
Qed.


Lemma defree_list dfc n (T: A) l:
    defree_ctx_ok dfc ->
    from_list l = (fv T) ->
    exists dfs,
      defree_spec_valid dfc dfs /\
      forall x,
      x \in (from_list l) -> (x \notin fv (defree_by_rec n dfs T) \/ x \in (defree_ctx_fv dfc)).
Proof.
  introv Hok Hfl.
  rewrite <- (List.app_nil_l l) in Hfl.
  eapply defree_suffix in Hfl; eauto.
Qed.


Lemma defree_typ dfc n (T: A):
    defree_ctx_ok dfc ->
    exists dfs,
      defree_spec_valid dfc dfs /\
      forall x,
        x \in (fv T) ->
        (x \notin fv (defree_by_rec n dfs T) \/ x \in (defree_ctx_fv dfc)).
Proof.
  introv Hok.
  destructc (fset_finite (fv T)) as l Hlist.
  rewrite Hlist.
  eapply defree_list; eauto.
Qed.

Lemma defree_typ_closed dfc null (T: A):
    defree_ctx_ok dfc ->
    defree_ctx_closed dfc ->
    defree_ctx_null dfc null ->
    exists dfs,
      defree_spec_valid dfc dfs /\
      (forall x,
        x \in (fv T) ->
        (x \notin fv (defree_by dfs T) \/ x \in (defree_ctx_fv dfc))) /\
      closed (defree_by dfs T).
Proof.
  introv Hok Hcl Hnull.
  forwards dfs Hva Hfv: (@defree_typ dfc 0); eauto.
  exists (defree_spec_seal dfs null).
  splits; eauto.
  - (* fv *)
    intros. specialize (Hfv x H).
    inversions Hnull.
    destruct Hfv.
    eapply defree_fv_append with (n:=0) (dfs:=defree_spec_seal defree_spec_empty null) in H0; eauto.
    auto.
  - simpls. unfold closed. intros. rewrite open_sealed. 
    auto.
    eapply Nat.le_0_l.
Qed.

Lemma defree_typ_closed_open dfc null (T: A):
    defree_ctx_ok dfc ->
    defree_ctx_closed dfc ->
    defree_ctx_null dfc null ->
    exists dfs,
      defree_spec_valid dfc dfs /\
      (forall x,
        x \in (fv T) ->
        (x \notin fv (defree_by_rec 1 dfs T) \/ x \in (defree_ctx_fv dfc))) /\
      (forall x, closed (open x (defree_by_rec 1 dfs T))).
Proof.
  introv Hok Hcl Hnull.
  forwards dfs Hva Hfv: (@defree_typ dfc 1); eauto.
  exists (defree_spec_seal dfs null).
  splits; eauto.
  - (* fv *)
    intros. specialize (Hfv x H).
    inversions Hnull.
    destruct Hfv.
    eapply defree_fv_append with (n:=1) (dfs:=defree_spec_seal defree_spec_empty null) in H0; eauto.
    auto.
  - simpls. unfold closed. intros. rewrite open_seal_down.
    rewrite open_sealed. auto. 
    eapply Nat.le_0_l.
Qed.

Lemma defree_by_rec_subst dfs (T: A) x y n:
  (defree_by_rec n dfs T) /[ x -> y] = defree_by_rec n (defree_spec_subst dfs x y) T.
Proof.
  simpl. auto.
Qed.

Lemma defree_by_rec_seal dfs (T: A) n null:
  (seal_rec n null (defree_by_rec n dfs T)) = defree_by_rec n (defree_spec_seal dfs null) T.
Proof.
  simpl. auto.
Qed.

Lemma defree_subst_open_trm dfc x y e G x0 (t: A):
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  defree_ctx_replacement dfc x y ->
  fv t \c defree_ctx_fv dfc ->
  x0 \notin defree_ctx_fv dfc \u \{ x} ->
  t /[->x0] /[ x -> y] = t /[->x0].
Proof.
  introv Hok Hget Hrpl Hfv Hnx0.
  inversions Hrpl.
  inversions Hget.
  unfold defree_ctx_fv in Hnx0.
  repeat rewrite notin_union in Hnx0.
  unfold defree_ctx_fv in Hfv.
  unfold renv_notin in H1.
  rewrite notin_union in H1.
  destruct_all.

  rewrite subst_fresh; eauto.
  eapply notin_fv_open. notin_solve.
  intro Hin.
  eapply Hfv in Hin.

  repeat rewrite in_union in Hin.
  destruct_all; try contradiction.
  rewrite in_singleton in H12.
  contradiction.
Qed.


End Defree.


Lemma subst_fresh_ty_var G e dfc (x: var) x0 y T:
  defree_ctx_ok dfc ->
  defree_ctx_replacement dfc x0 y ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ x ∶ T ->
  G ⁏ e ⊢ x ∶ (T /[x0 -> y]).
Proof.
  introv Hok Hrpl HG Hty.
  inversions HG. inversions Hrpl.

  assert (x <> x0) as Hnx. {
    intro; subst. eapply typing_implies_dom in Hty.
    contradiction.
  }

  inversions Hok.

  eapply usubst_end1 with (x1:=x0) (y0:=y) in Hty; eauto.
  rewrite subst_ne_var in Hty.
  apply Hty. auto.
Qed.


Lemma subst_fresh_ty_trm G e dfc (x: trm) x0 y T:
  defree_ctx_ok dfc ->
  defree_ctx_replacement dfc x0 y ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ x ∶ T ->
  x0 \notin fv x ->
  G ⁏ e ⊢ x ∶ (T /[x0 -> y]).
Proof.
  introv Hok Hrpl HG Hty Hfv.
  inversions HG. inversions Hrpl.

  inversions Hok.

  eapply usubst_end1 with (x1:=x0) (y0:=y) in Hty; eauto.
  rewrite subst_fresh in Hty.
  apply Hty. auto.
Qed.



Lemma defree_ty_var G e dfc (x: var) T TN:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ x ∶ T ->
  defree dfc T TN ->
  G ⁏ e ⊢ x ∶ TN.
Proof.
  introv Hok Hget Hty Hdf.
  inversions Hdf.
  induction H.
  - eauto.
  - eapply subst_fresh_ty_var; eauto.
  - forwards: IHdefree_spec_valid; eauto.
    inversions Hget.
    inversions H0.
    eapply sealing_obj with (x0:= null) in H2; eauto.
    simpls.
    rewrite H4 in H2.
    rewrite seal_var in H2; eauto.
    Unshelve. eapply sealing_ty_var.
Qed.

Lemma defree_ty_trm G e dfc (x: trm) T TN:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ x ∶ T ->
  defree dfc T TN ->
  closed x ->
  fv x \c defree_ctx_fv dfc ->
  G ⁏ e ⊢ x ∶ TN.
Proof.
  introv Hok Hget Hty Hdf Hcl Hfv.
  inversions Hdf.
  induction H.
  - eauto.
  - eapply subst_fresh_ty_trm; eauto. inversions H0. eapply notin_subset. eauto. unfold defree_ctx_fv. hnf in H3. rewrite in_union in H3. notin_solve. intro. eauto. intro. eauto.
  - forwards: IHdefree_spec_valid; eauto.
    inversions Hget.
    inversions H0.
    eapply sealing_obj with (x0:= null) in H2; eauto.
    simpls.
    rewrite H4 in H2.
    rewrite closed_sealed in H2; eauto.
Qed.


Lemma defree_capbnd G e dfc dfs T Tcap:
    defree_ctx_ok dfc ->
    defree_ctx_get dfc G e ->
    G ⁏ e ⊢ T ∶cpb ⟦ typ_refcap_mutability ⟧ Tcap ->
    defree_spec_valid dfc dfs ->
    G ⁏ e ⊢ (defree_by dfs T) ∶cpb ⟦ typ_refcap_mutability ⟧ (defree_by_rec 1 dfs Tcap).
Proof.
  introv Hok Hget Hcpb Hva.

  induction Hva.
  - eauto.
  - specialize (IHHva Hok Hget). simpls.
    inversions Hget.
    inversions Hok.
    inversions H.
    eapply usubst_end1 in IHHva; eauto.
  - specialize (IHHva Hok Hget). simpls.
    eapply @sealing_obj with (property := ty_capbnd _ typ_refcap_mutability) (x:=null) in IHHva; eauto.
    simpls.
    inversions Hget.
    inversions H.
    rewrite H2 in IHHva.
    apply IHHva.
    apply sealing_ty_capbnd.
Qed.

Lemma defree_incap G e dfc dfs T Ticp:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ T ∶icp Ticp ->
  defree_spec_valid dfc dfs ->
  G ⁏ e ⊢ (defree_by dfs T) ∶icp (defree_by dfs Ticp).
Proof.
  introv Hok Hget Hicp Hva.
  induction Hva.
  - eauto.
  - specialize (IHHva Hok Hget).
    inversions Hget.
    inversions Hok.
    inversions H.
    eapply usubst_end1 in IHHva; eauto.
  - specialize (IHHva Hok Hget).
    eapply @sealing_obj with (property := ty_incap _) (x:=null) in IHHva; eauto.
    simpls.
    inversions Hget.
    inversions H.
    rewrite H2 in IHHva.
    apply IHHva.
    apply sealing_ty_incap.
Qed. 


Lemma defree_subtyp G e dfc dfs T Ticp:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ T <: Ticp ->
  defree_spec_valid dfc dfs ->
  G ⁏ e ⊢ (defree_by dfs T) <: (defree_by dfs Ticp).
Proof.
  introv Hok Hget Hicp Hva.
  induction Hva.
  - eauto.
  - specialize (IHHva Hok Hget).
    inversions Hget.
    inversions Hok.
    inversions H.
    eapply usubst_end1 in IHHva; eauto.
  - specialize (IHHva Hok Hget).
    eapply @sealing_obj with (property := subtyp _) (x:=null) in IHHva; eauto.
    simpls.
    inversions Hget.
    inversions H.
    rewrite H2 in IHHva.
    apply IHHva.
    apply sealing_subtyp.
Qed. 

Lemma defree_replacement_notin dfc x y:
  defree_ctx_replacement dfc x y ->
  x \notin defree_ctx_fv dfc.
Proof.
  intros. inversions H. unfold defree_ctx_fv. repeat rewrite notin_union. unfold renv_notin in H2. splits; eauto.
Qed.

Lemma defree_ref dfc dfs x Tstrip Tcap:
  defree_spec_valid dfc dfs ->
  x \in (defree_ctx_fv dfc) ->
  defree_by dfs (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability x Tcap)) = (typ_and (defree_by dfs Tstrip) (typ_cap_adapt typ_refcap_mutability x (defree_by_rec 1 dfs Tcap))).
Proof.
  introv Hva Hdom.
  induction dfs.
  - simpls. auto.
  - simpls. inversions Hva.
    rewrite IHdfs by eauto.
    simpls.
    rewrite substi_ne_var; auto.
    intro; subst.
    eapply defree_replacement_notin in H4.
    contradiction.
  - simpls.
    inversions Hva.
    rewrite IHdfs by eauto.
    simpls.
    assert (~ (1 <= 0)).
    eapply Nat.nle_succ_diag_l.
    cases_if.
    eauto.
Qed.

Lemma ctx_defree_ok G dfc G1 G2:
  defree dfc G1 G2 ->
  ctx_ok (G §& G1) ->
  ctx_ok (G §& G2) .
Proof.
  introv Hdf Hok.
  inversions Hdf.
  induction H.
  - eauto.
  - eapply ctx_ok_concat_map. eauto.
  - simpls. rewrite ctx_seal_map. eapply ctx_ok_concat_map. eauto.
Qed.
Lemma objctx_defree_ok G dfs G1 Go1:
  ctx_ok (G §& G1 §§& Go1) ->
  ctx_ok (G §& (defree_by dfs G1) §§& (defree_by dfs Go1)) .
Proof.
  introv Hok.
  
  induction dfs; simpls; eauto.
  -
  remember (defree_by_rec 0 dfs Go1).
  destruct o; destruct Go1; unfold ctx_obj in *; simpls; eauto using ctx_ok_concat_map.
  (* Now remaining objctx_lit *)
  all: inversions IHdfs;
  constructor; eauto using ctx_ok_concat_map; ctx_simpl_dom; notin_solve;  replace (ctx_dom _) with (ctx_dom (defree_by_rec 0 dfs G1)); notin_solve;  eauto using ctx_dom_map. 
  -
  remember (defree_by_rec 0 dfs Go1).
  destruct o; destruct Go1; unfold ctx_obj in *; simpls; rewrite ctx_seal_map; eauto using ctx_ok_concat_map.
  (* Now remaining objctx_lit *)
  all: inversions IHdfs;
  constructor; eauto using ctx_ok_concat_map.
Qed.

Lemma ctx_defree_weakens G dfc G1 G2:
  defree dfc G1 G2 ->
  ctx_weakens (G §& G1) ->
  ctx_weakens (G §& G2) .
Proof.
  introv Hdf Hok.
  inversions Hdf.
  induction H.
  - eauto.
  - eapply ctx_weakens_concat.
    eauto using  ctx_weakens_concat_left.
    eauto using  ctx_weakens_concat_right, ctx_weakens_map.
  - eapply ctx_weakens_concat.
    eauto using  ctx_weakens_concat_left. simpls. rewrite ctx_seal_map.
    eauto using  ctx_weakens_concat_right, ctx_weakens_map.
Qed.




Lemma defree_subst_ty_let dfc x y e G x0 T2 t TN:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  defree_ctx_replacement dfc x y ->
  x0 \notin defree_ctx_fv dfc \u \{ x} ->
  G § x0 / v_let ~ T2 ⁏ e ⊢ open x0 t ∶ TN ->
  G § x0 / v_let ~ (T2/[ x -> y]) ⁏ e ⊢ ((open x0 t)/[ x -> y]) ∶ (TN/[ x -> y] ).
Proof.
  introv Hok Hget Hrpl Hnx0 Hty.
  inversions Hok.
  inversions Hget.
  inversions Hrpl.
  destruct_all.
  simpls.

  replace objctx_empty with (subst_objctx x y objctx_empty).
  replace (G § x0 / v_let ~ (T2 /[ x -> y])) with (G §& subst_ctx x y (x0 / v_let ~ T2 )). 
  eapply usubst_ty1_trm; eauto.
  ctx_flatten. eauto.
  
  compute; auto.
  compute; auto.
Qed.

Lemma defree_seal_ty_let dfc null e G x0 T2 t TN:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  defree_ctx_null dfc null ->
  defree_ctx_closed dfc ->
  x0 \notin defree_ctx_fv dfc \u \{ null} ->
  G § x0 / v_let ~ T2 ⁏ e ⊢ open x0 t ∶ TN ->
  G § x0 / v_let ~ (seal null T2) ⁏ e ⊢ (seal null (open x0 t)) ∶ (seal null TN).
Proof.
  introv Hok Hget Hnl Hcl Hnx0 Hty.
  inversions Hok.
  inversions Hget.
  inversions Hnl.
  destruct_all.
  simpls.

  replace objctx_empty with (seal null objctx_empty).
  replace (G § x0 / v_let ~ (seal null T2)) with (seal null (G § x0 / v_let ~ T2 )).
  eapply sealing_obj; eauto.
  - simpls. replace G with (seal null G) at 2. auto.
    inversions Hcl. apply H3.
  - simpls. auto.
Qed.


(* For a typed let term, we can replace the type with a defree one. Defree of the argument is input, result is output. *)
Lemma defree_ty_let G e dfc L S SN (t: trm) T:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  (fv t \c defree_ctx_fv dfc) ->
  defree dfc S SN ->
  (forall x: var, closed (open x t)) ->
  (forall x : var, x \notin L -> G § x / v_let ~ S ⁏ e ⊢ open x t ∶ T) ->
  exists TN LN,
    (defree_ctx_fv dfc) \c LN /\
    defree dfc T TN /\
    (forall x : var, x \notin LN -> G § x / v_let ~ SN ⁏ e ⊢ open x t ∶ TN).
Proof.
  introv Hok Hget Hfv Hdf Hslt Hty.
  intros.
  inversions Hdf.
  reverts T.
  (* Induction on the defree spec of the argument type *)
  induction H; intros.
  - (* empty *)
    exists T (L \u (defree_ctx_fv dfc)).
    splits; eauto.
    + intro. intro. rewrite in_union. auto.
    + eapply defree_refl.
  - (* Subst *)
    forwards TN LN ? ? ?: (IHdefree_spec_valid Hok Hget Hfv T Hty).
    exists (TN /[ x -> y]) ((defree_ctx_fv dfc) \u LN \u \{x} ).
    splits.
    + (* fv *)
      intro. intro. rewrite in_union. auto.
    + (* defree *)
      inversions H2. rewrite defree_by_rec_subst. econstructor. constructor; eauto.
    + (* typing *)
      intros.
      assert (x0 \notin LN) as Hx0LN. notin_solve.
      specialize (H3 x0 Hx0LN).
      simpls.
      remember (defree_by_rec 0 dfs S ) as T2.
      forwards Htys: defree_subst_ty_let; eauto.
      notin_solve.
      erewrite defree_subst_open_trm in Htys; eauto.
  - (* Seal *)
    forwards TN LN ? ? ?: (IHdefree_spec_valid Hok Hget Hfv T Hty).
    exists (seal null TN) ((defree_ctx_fv dfc) \u LN ).
    splits.
    + intro. intro. rewrite in_union. auto.
    + destruct H3. rewrite defree_by_rec_seal. constructor. constructor; eauto.
    + intros x0 Hnx0.
      assert (x0 \notin LN) as Hx0LN. notin_solve.
      specialize (H4 x0 Hx0LN).
      simpls.
      assert (sealed G). inversions Hget. inversions H0. auto.
      replace (open x0 t ) with (seal null (open x0 t )).
      replace (objctx_empty) with (seal null objctx_empty).
      rewrite <- H5.
      rewrite <- ctx_seal_push.
      eapply sealing_obj; eauto.
      simpls; auto.
      eapply closed_sealed in Hslt.
      eapply Hslt.
Qed.


(* For a typed let term, we can replace the type with a defree one. Defree of the argument is input, result is output. *)
Lemma defree_ty_let' G e dfc L S SN T:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  defree dfc S SN ->
  exists TN LN,
    forall (t: trm) ,
   (fv t \c defree_ctx_fv dfc) ->
  (forall x: var, closed (open x t)) ->
  (forall x : var, x \notin L -> G § x / v_let ~ S ⁏ e ⊢ open x t ∶ T) ->

    (defree_ctx_fv dfc) \c LN /\
    defree dfc T TN /\
    (forall x : var, x \notin LN -> G § x / v_let ~ SN ⁏ e ⊢ open x t ∶ TN).
Proof.
  introv Hok Hget  Hdf.
  intros.
  inversions Hdf.
  reverts T.
  (* Induction on the defree spec of the argument type *)
  induction H; intros.
  - (* empty *)
    exists T (L \u (defree_ctx_fv dfc)).
    introv  Hfv Hslt Hty.
    splits; eauto.
    + intro. intro. rewrite in_union. auto.
    + eapply defree_refl.
  - (* Subst *)
    specialize (IHdefree_spec_valid Hok Hget T).
    destructc IHdefree_spec_valid as TN LN IHrest.
    exists (TN /[ x -> y]) ((defree_ctx_fv dfc) \u LN \u \{x} ).
    intros.
    forwards: IHrest; fweauto.
    destructecs H4.
    splits.
    + (* fv *)
      intro. intro. rewrite in_union. auto.
    + (* defree *)
      inversions H6. rewrite defree_by_rec_subst. econstructor. constructor; eauto.
    + (* typing *)
      intros.
      assert (x0 \notin LN) as Hx0LN. notin_solve.
      specialize (H4 x0 Hx0LN).
      simpls.
      remember (defree_by_rec 0 dfs S ) as T2.
      forwards Htys: defree_subst_ty_let; eauto.
      notin_solve.
      erewrite defree_subst_open_trm in Htys; eauto.
  - (* Seal *)
    specialize (IHdefree_spec_valid Hok Hget T).
    destructc IHdefree_spec_valid as TN LN IHrest.
    exists (seal null TN) ((defree_ctx_fv dfc) \u LN ).
    intros.
    forwards: IHrest; fweauto.
    destructecs H5.

    splits.
    + intro. intro. rewrite in_union. auto.
    + destruct H7. rewrite defree_by_rec_seal. constructor. constructor; eauto.
    + intros x0 Hnx0.
      assert (x0 \notin LN) as Hx0LN. notin_solve.
      specialize (H5 x0 Hx0LN).
      simpls.
      assert (sealed G). inversions Hget. inversions H0. auto.
      replace (open x0 t ) with (seal null (open x0 t )).
      replace (objctx_empty) with (seal null objctx_empty).
      rewrite <- H8.
      rewrite <- ctx_seal_push.
      eapply sealing_obj; eauto.
      simpls; auto.
      eapply closed_sealed in H3.
      eapply H3.
Qed.


Lemma defree_ty_stack G e dfc T1 T2 s U:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  (fv U \u fv s) \c (defree_ctx_fv dfc) ->
  defree dfc T1 T2 ->
  closed s ->
  closed U ->
  ty_stack G e T1 s U ->
  ty_stack G e T2 s U.
Proof.
  introv Hok Hget HfvL Hdf Hcls HclU Htys.
  reverts T2.
  induction Htys.
  - intros.
    inversions Hdf.
    pose proof (defree_subtyp Hok Hget H H0).
    eapply closed_sealed0 in HclU. 
    eapply  defree_fv_same in HclU.
    2: eauto.

    2: {intro. intro. specialize (HfvL x). apply HfvL. rewrite in_union. auto. }

    rewrite <- HclU. eauto.
  - intros.
    eapply defree_ty_let in H.
    5: apply Hdf.
    destructc H as TN LN ? ? ?.

    eapply ty_stack_let.
    eapply IHHtys. eauto.  eauto.  eauto.
    intro. intro. specialize ( HfvL x).  apply HfvL. rewrite in_union. rewrite in_union in H2. destruct H2. auto. right. unfold fv. unfold StackFreeVar. compute. rewrite in_union. compute in H2. auto.
    eapply stack_closed_fs in Hcls.
    eauto. eauto.
    eauto. eauto.
    eauto. eauto.
    intro. intro. specialize (HfvL x). apply HfvL. rewrite in_union. right. compute. rewrite in_union. compute in H0. auto.
    eapply stack_closed_t; eauto.
Qed.

Lemma defree_by_andor n dfs T T1 T2:
  typ_andor T ->
  defree_by_rec n dfs (T T1 T2) =  (T (defree_by_rec n dfs T1) (defree_by_rec n dfs T2)).
Proof.
  introv HT.
  inversions HT;
  induction dfs; sympls; eauto; rewrite IHdfs; sympls; eauto.
Qed.


Lemma defree_by_fld n dfs A T1 T2:
  defree_by_rec n dfs (typ_rcd (dec_fld A T1 T2)) =  (typ_rcd (dec_fld A (defree_by_rec n dfs T1) (defree_by_rec n dfs T2))).
Proof.
  induction dfs; sympls; eauto; rewrite IHdfs; sympls; eauto.
Qed.



Lemma adapt_ref_mutability_nothing_fresh G e s null U x (w: var) a T0 T Tstrip Tcap:
  ctx_ok G ->
  ctx_weakens G ->
  sealed G ->
  ok e ->
  x \in (ctx_dom G) ->
  closed s ->
  closed U ->
  ty_stack G e (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability x Tcap)) s U ->
  G ⁏ e ⊢ w ∶ (typ_rcd (dec_fld a T0 T)) ->
  G ⁏ e ⊢ T ∶icp Tstrip ->
  G ⁏ e ⊢ T ∶cpb ⟦ typ_refcap_mutability ⟧ Tcap ->

  exists TN0 TN TstripN TcapN,
    ty_stack G e (typ_and TstripN (typ_cap_adapt typ_refcap_mutability x TcapN)) s U /\
    G ⁏ e ⊢ w ∶ (typ_rcd (dec_fld a TN0 TN)) /\
    G ⁏ e ⊢ TN ∶icp TstripN /\
    G ⁏ e ⊢ TN ∶cpb ⟦ typ_refcap_mutability ⟧ TcapN /\
    (fv TcapN \c (ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u fv U \u fv s \u \{null})) /\
    (forall w, closed (open w TcapN)).
Proof.
  introv HokG Hwk HslG Hoke HxG Hcls HclU.
  introv Htys Hfld Hstrip Hcap.
  set (dfc := defree_ctx_with_null G e (fv U \u fv s) null).
  assert (defree_ctx_ok dfc) as Hok. constructor; auto.
  assert (defree_ctx_get dfc G e) as Hget. constructor; auto.
  assert (defree_ctx_closed dfc) as Hcl. constructor; auto.
  assert (defree_ctx_null dfc null) as Hnull. constructor; auto.

  destructc (defree_typ_closed_open Tcap Hok Hcl Hnull) as dfs Hva Hfv HclTcap.
  pose proof (defree_capbnd Hok Hget Hcap Hva) as HcapN.
  pose proof (defree_incap Hok Hget Hstrip Hva) as HicpN .

  exists (defree_by_rec 0 dfs T0) (defree_by_rec 0 dfs T) (defree_by_rec 0 dfs Tstrip) (defree_by_rec 1 dfs Tcap).
  splits; eauto.

  + (* stack typing *)
    eapply defree_ty_stack; eauto.
    * intro. unfold defree_ctx_fv in *. unfold dfc in *. intro. repeat rewrite in_union in H. repeat rewrite in_union. destruct_all. right. right. auto. right. right. auto.
    * remember (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability x Tcap)) as T1.
      eapply f_equal with (f:=fun x => defree_by_rec 0 dfs x ) in HeqT1.
      erewrite defree_ref in HeqT1; eauto.
      rewrite <- HeqT1.
      eauto.
      unfold defree_ctx_fv, dfc. rewrite in_union. auto.
  + (* receiver typing *)
    eapply defree_ty_var; eauto.
    rewrite <- defree_by_fld.
    eauto.
  + (* free variables *)
    intro. intro.
    assert (x0 \in defree_ctx_fv dfc). {
      pose proof H.
      eapply defree_fv2 with (T1:=Tcap) (dfc0:=dfc) in H. 2: auto.
      repeat rewrite in_union in H.
      destruct_all; eauto.
      eapply Hfv in H. destruct H. contradiction. auto.
    }

    unfold defree_ctx_fv, dfc in H0.
    rewrite ! union_assoc in H0.
    rewrite ! union_assoc. auto.
Qed.


Lemma defree_lit_fun G e dfc dfs Targ Tfn t1:
  defree_ctx_ok dfc ->
  defree_ctx_get dfc G e ->
  G ⁏ e ⊢ lit_fun Targ t1 ∶ Tfn ->
  defree_spec_valid dfc dfs ->
  fv t1 \c (defree_ctx_fv dfc) ->
  (forall x, closed (open x t1)) ->
  G ⁏ e ⊢ lit_fun (defree_by dfs Targ) t1 ∶ (defree_by dfs Tfn).
Proof.
  introv Hok Hget Hlit Hdffn Hfv Hclt.

  reverts Targ.
  induction Hdffn.
  - intros. eauto.
  - (* subst *)
    intros.
    specialize (IHHdffn Hok Hget Hfv Targ Hlit).
    inversions Hok. inversions Hget. inversions H.
    unfold renv_notin in H11. rewrite notin_union in H11. destruct_all.
    simpls.

    replace (lit_fun ((defree_by_rec 0 dfs Targ) /[ x -> y]) t1) with ((lit_fun (defree_by_rec 0 dfs Targ) t1) /[ x -> y]).
    eapply usubst_end1; eauto.

   sympls.
    rewrite (subst_fresh y t1).
    auto.
    intro. apply Hfv in H4. repeat rewrite in_union in H4. destruct_all; try contradiction. rewrite in_singleton in H4. subst. contradiction.
  - (* seal *)
    intros.
    specialize (IHHdffn Hok Hget Hfv Targ Hlit).
    inversions Hok. inversions Hget. inversions H.
    sympls.

    replace (lit_fun (seal null (defree_by_rec 0 dfs Targ)) t1) with (seal null (lit_fun (defree_by_rec 0 dfs Targ)t1)).
    erewrite <- H5.
    replace (objctx_empty) with (seal null objctx_empty).
    eapply sealing_obj; eauto.
    sympl; auto.
    sympl.

    assert(seal_rec 1 null t1 = t1) as Hslt1. {
      pick_fresh z.
      specialize (Hclt z).
      eapply closed_seal_down; eauto.
    }
    fold_trm_seal_rec.
    rewrite Hslt1; eauto.
Qed.

Lemma fun_nothing_fresh G e s U null (t1 t2: trm) L Targ Tfn Tcont:
  ctx_ok G ->
  ctx_weakens G ->
  sealed G ->
  ok e ->
  closed s ->
  (forall x : var, closed (open x t1)) ->
  (forall x : var, closed (open x t2)) ->
  closed Targ ->
  closed U ->
  ty_stack G e Tcont s U ->
  G ⁏ e ⊢ lit_fun Targ t1 ∶ Tfn ->
  (forall x : var, x \notin L -> G § x / v_let ~ Tfn ⁏ e ⊢ open x t2 ∶ Tcont) ->

  exists LNN TfnN TcontN,
    ty_stack G e TcontN s U /\
    G ⁏ e ⊢ lit_fun Targ t1 ∶ TfnN /\
    (forall x : var, x \notin LNN -> G § x / v_let ~ TfnN ⁏ e ⊢ open x t2 ∶ TcontN)  /\
    (fv TfnN \c (ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u fv U \u fv s \u fv Targ \u fv t2 \u fv t1 \u \{null})) /\
    (closed TfnN).
Proof.
  introv HokG Hwk HslG Hoke Hcls Hclt1 Hclt2 HclTarg HclU.
  introv Htys Hfn Hcont.

  set (dfc := defree_ctx_with_null G e (fv U \u fv s \u fv Targ \u fv t2 \u fv t1) null).
  assert (defree_ctx_ok dfc) as Hok. constructor; auto.
  assert (defree_ctx_get dfc G e) as Hget. constructor; auto.
  assert (defree_ctx_null dfc null) as Hnull. constructor; auto.
  assert (defree_ctx_closed dfc) as Hcl. constructor; auto.
  destructc (defree_typ_closed Tfn Hok Hcl Hnull) as dfs Hva Hfv HclTfn.

  assert (fv t2 \c (defree_ctx_fv dfc)) as Hfvt2. intro. intro. unfold defree_ctx_fv, dfc. repeat rewrite in_union. auto 10.
  assert (fv t1 \c (defree_ctx_fv dfc)) as Hfvt1. intro. intro. unfold defree_ctx_fv, dfc. repeat rewrite in_union. auto 10.
  pose proof Hva as Hdffn. eapply defree_by_spec with (T1:= Tfn) in Hdffn.
  destructc (defree_ty_let t2 Hok Hget Hfvt2 Hdffn Hclt2 Hcont) as TcontN LN HfvLn Hdfcont HcontN.
  pose proof (defree_lit_fun Hok Hget Hfn Hva Hfvt1) as Hdfarg.
  assert ((defree_by_rec 0 dfs Targ) = Targ). {
    eapply defree_fv_same; eauto. unfold defree_ctx_fv, dfc. intro. intro. repeat rewrite in_union. auto 10.
    eapply closed_sealed0. auto.
  }
  exists LN (defree_by_rec 0 dfs Tfn) TcontN.
  splits; eauto.
  + (* stack typing *)
    eapply defree_ty_stack; eauto. intro. intro. rewrite in_union in H0. repeat rewrite in_union.
    unfold defree_ctx_fv, dfc. repeat rewrite in_union.
    destruct H0; auto 10.
  + (* lit typing *)
    rewrite <- H. eapply defree_lit_fun; eauto.
  + (* free vars *)
    intro. intro.

    assert (x \in defree_ctx_fv dfc). {
      pose proof H0.
      eapply defree_fv2 with (T:=Tfn) (dfc0:=dfc) in H0. 2: auto.
      repeat rewrite in_union in H0.
      destruct_all; eauto.
      eapply Hfv in H0. destruct H0. contradiction. auto.
    }

    unfold defree_ctx_fv, dfc in H1.
    rewrite ! union_assoc in H1.
    rewrite ! union_assoc. auto.
Qed.

Lemma fun_nothing_fresh' G e U L null Targ Tfn Tcont Lt:
  ctx_ok G ->
  ctx_weakens G ->
  sealed G ->
  ok e ->
  closed Targ ->
  closed U ->
  exists LNN TfnN TcontN,
    (fv TfnN \c (ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u fv U \u fv Targ \u Lt \u \{null})) /\
    (closed TfnN) /\ (
      forall (s: stack) (t1 t2: trm), 
      fv t1 \u fv t2 \u fv s \c Lt ->
      (forall x : var, closed (open x t1)) ->
      (forall x : var, closed (open x t2)) ->
      closed s ->
      ty_stack G e Tcont s U ->
      G ⁏ e ⊢ lit_fun Targ t1 ∶ Tfn ->
      (forall x : var, x \notin L -> G § x / v_let ~ Tfn ⁏ e ⊢ open x t2 ∶ Tcont) ->
      (
        ty_stack G e TcontN s U /\
        G ⁏ e ⊢ lit_fun Targ t1 ∶ TfnN /\
        (forall x : var, x \notin LNN -> G § x / v_let ~ TfnN ⁏ e ⊢ open x t2 ∶ TcontN)
      )
    ).
Proof.
  introv HokG Hwk HslG Hoke HclTarg HclU.

  set (dfc := defree_ctx_with_null G e (fv U \u fv Targ \u Lt) null).
  assert (defree_ctx_ok dfc) as Hok. constructor; auto.
  assert (defree_ctx_get dfc G e) as Hget. constructor; auto.
  assert (defree_ctx_null dfc null) as Hnull. constructor; auto.
  assert (defree_ctx_closed dfc) as Hcl. constructor; auto.
  destructc (defree_typ_closed Tfn Hok Hcl Hnull) as dfs Hva Hfv HclTfn.

  pose proof Hva as Hdffn. eapply defree_by_spec with (T1:= Tfn) in Hdffn.

  forwards Hdtl: defree_ty_let' Hok Hget Hdffn.
  destructecs Hdtl.

  exists LN (defree_by_rec 0 dfs Tfn) TN.
  splits.
  - (* free vars *)

    assert (fv (defree_by_rec 0 dfs Tfn) \c defree_ctx_fv dfc) as Hfvsu. {
      intro. intro.
      pose proof H.
      eapply defree_fv2 with (T:=Tfn) (dfc0:=dfc) in H0. 2: auto.
      repeat rewrite in_union in H0.
      destruct_all; eauto.
      eapply Hfv in H0. destruct H0. contradiction. auto.

    }
    unfold defree_ctx_fv, dfc in Hfvsu.
    eapply subset_trans. eapply Hfvsu.
    union_sub_solver_destruct_ls; try union_sub_solver.
  - auto.
  -

  introv Hfvts Hclt1 Hclt2 Hcls Htys Htyt Hcont.

Ltac union_sub_solver_via H := (eapply (subset_trans (subset_trans _ H)); union_sub_solver).

  assert (fv t2 \c (defree_ctx_fv dfc)) as Hfvt2. unfold defree_ctx_fv, dfc. union_sub_solver_via Hfvts.
  assert (fv t1 \c (defree_ctx_fv dfc)) as Hfvt1. unfold defree_ctx_fv, dfc. union_sub_solver_via Hfvts.


  pose proof (defree_lit_fun Hok Hget Htyt Hva Hfvt1) as Hdfarg.
  assert ((defree_by_rec 0 dfs Targ) = Targ). {
    eapply defree_fv_same; eauto. unfold defree_ctx_fv, dfc. union_sub_solver.
    eapply closed_sealed0. auto.
  }

  forwards Hdtl2: Hdtl t2; fweauto.
  destructecs Hdtl2.

  splits; eauto.
  + (* stack typing *)
    eapply defree_ty_stack; eauto.
    unfold defree_ctx_fv, dfc. union_sub_solver_destruct_ls. union_sub_solver. union_sub_solver_via Hfvts.
  + (* lit typing *)
    rewrite <- H. eapply defree_lit_fun; eauto.
Unshelve.
  union_sub_solver.
  union_sub_solver.
  union_sub_solver.
Qed.

Lemma adapt_ref_mutability_nothing_fresh' G e null U x a T0 T Tstrip Tcap Lt:
  ctx_ok G ->
  ctx_weakens G ->
  sealed G ->
  ok e ->
  x \in (ctx_dom G) ->
  closed U ->
  exists TN0 TN TstripN TcapN,
  (fv TcapN \c (ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u fv U \u Lt \u \{null})) /\
  (forall w, closed (open w TcapN)) /\
  forall (s: stack) (w: var),
  (
    closed s ->
    fv s \c Lt ->
    ty_stack G e (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability x Tcap)) s U ->
    G ⁏ e ⊢ w ∶ (typ_rcd (dec_fld a T0 T)) ->
    G ⁏ e ⊢ T ∶icp Tstrip ->
    G ⁏ e ⊢ T ∶cpb ⟦ typ_refcap_mutability ⟧ Tcap ->
    (
      ty_stack G e (typ_and TstripN (typ_cap_adapt typ_refcap_mutability x TcapN)) s U /\
      G ⁏ e ⊢ w ∶ (typ_rcd (dec_fld a TN0 TN)) /\
      G ⁏ e ⊢ TN ∶icp TstripN /\
      G ⁏ e ⊢ TN ∶cpb ⟦ typ_refcap_mutability ⟧ TcapN
    )
  ).

Proof.
  introv HokG Hwk HslG Hoke HxG HclU.

  set (dfc := defree_ctx_with_null G e (fv U \u Lt) null).
  assert (defree_ctx_ok dfc) as Hok. constructor; auto.
  assert (defree_ctx_get dfc G e) as Hget. constructor; auto.
  assert (defree_ctx_closed dfc) as Hcl. constructor; auto.
  assert (defree_ctx_null dfc null) as Hnull. constructor; auto.

  destructc (defree_typ_closed_open Tcap Hok Hcl Hnull) as dfs Hva Hfv HclTcap.

  exists (defree_by_rec 0 dfs T0) (defree_by_rec 0 dfs T) (defree_by_rec 0 dfs Tstrip) (defree_by_rec 1 dfs Tcap).
  splits; eauto.
  - (* free variables *)
    intro. intro.
    assert (x0 \in defree_ctx_fv dfc). {
      pose proof H.
      eapply defree_fv2 with (T1:=Tcap) (dfc0:=dfc) in H. 2: auto.
      repeat rewrite in_union in H.
      destruct_all; eauto.
      eapply Hfv in H. destruct H. contradiction. auto.
    }

    unfold defree_ctx_fv, dfc in H0.
    rewrite ! union_assoc in H0.
    rewrite ! union_assoc. auto.
  -

  introv Hcls Hfvsu Htys Hfld Hstrip Hcap.
  pose proof (defree_capbnd Hok Hget Hcap Hva) as HcapN.
  pose proof (defree_incap Hok Hget Hstrip Hva) as HicpN.
  splits; auto.
  + (* stack typing *)
    eapply defree_ty_stack; eauto.
    * unfold defree_ctx_fv in *. unfold dfc in *.
      union_sub_solver_destruct_ls. union_sub_solver. union_sub_solver_via Hfvsu.
    * remember (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability x Tcap)) as T1.
      eapply f_equal with (f:=fun x => defree_by_rec 0 dfs x ) in HeqT1.
      erewrite defree_ref in HeqT1; eauto.
      rewrite <- HeqT1.
      eauto.
      unfold defree_ctx_fv, dfc. rewrite in_union. auto.
  + (* receiver typing *)
    eapply defree_ty_var; eauto.
    rewrite <- defree_by_fld.
    eauto.
Unshelve.
  union_sub_solver.

Qed.


End WithTypingMode.