Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers TransformationConfig ConfigTyping.
Require Import Similarity ConfigSimilarity ContextSimilarity SyntaxSimilarity SimilarityPreservation TypingSimilarity
 TypedReductionEquivalence OperationalSemantics TypedOperationalSemantics TypedPreservation ConfigValid TransformationSwapReads TransformationSimilarity TransformationPreservation TypedProgress.

(** * Tactics and lemams about execution *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma typed_progress_noanswer : forall F  T c,
    ~ (answer c) ->
    ty_config F c T ->
    exists F' c', F ⊢ c ∶ T ↦ F' ⊢ c'.
Proof.
  intros.
  forwards: typed_progress H0; eauto.
  destruct H1. contradiction. eauto.
Qed.

(* Common premises of tred_read_inv and tred_read_fun *)
Lemma tred_read_inv t0 x1 a F fs Sigma e Lv T0 F2 C2:
  t0 = trm_read x1 a ->
  F ⊢ (t0 ; fs ; Sigma ; e ; Lv) ∶ T0 ↦  F2 ⊢ C2 ->
  exists w1 T2 w2 y2 y1 T4 T3 T5 F' yd,
  x1 = (avar_f w1) /\
  F2 = (F § w2 / v_ref ~ T2) /\
  C2 = (trm_var (avar_f w2) ; fs ; Sigma ; e & w2 ~ y2; fv_config (t0 ; fs ; Sigma ; e ; Lv)) /\
 fresh_in_config (w2 :: nil)%list (t0 ; fs ; Sigma ; e ; Lv) /\
  F ⁏ e ⊢ w1 ∶ typ_rcd (dec_fld a T4 T3) /\
  binds w1 y1 e /\
  ctx_binds_k_after y2 v_loc T5 F F' /\
  binds y1 (item_obj yd) Sigma /\
  defs_has yd (def_fld a (avar_f y2)).
Proof.
  introv Ht0 Hred.
  subst.
  inversions Hred; inversions H8; repeat eexists; simpl in Hfrw2; destruct Hfrw2; eauto.
Qed.

End WithTypingMode.

(* In a typed reduction hypothesis, finds the last element of the result configuration, and give it a name,
   to avoid growing too much *)
Ltac encapsulate_Lv Hred := (
  match type of Hred with 
  | _ ⊢ _ ∶  _ ↦ _ ⊢  (_ ; _ ; _ ; _ ; ?Lv)  =>
    let Lva := fresh "Lv" in 
    set (Lva := Lv) in Hred
  end
).
(* Same as encapsulate_Lv, but applies in another hypothesis *)
Ltac encapsulate_Lv2 Hred Hty := (
  match type of Hred with 
  | _ ⊢ _ ∶  _ ↦ _ ⊢  (_ ; _ ; _ ; _ ; ?Lv)  =>
    let Lva := fresh "Lv" in 
    set (Lva := Lv) in Hred, Hty
  end
).

(* Given a typing for a configuration, finds the reduction steps and passes the reduction and typing of the result to a tactic *)
Ltac machine_step HtyC id tac := (
  let Hprog := fresh "Hprog" id in
  let Hc := fresh "Hc" id in
  let G := fresh "G" id in
  let C := fresh "C" id in
  let Hred := fresh "Hred" id in
  let Hty := fresh "Hty" id in
  forwards Hprog: typed_progress_noanswer HtyC; [>
    intro Hc; inversions Hc |
    destructc Hprog as G C Hred;
    forwards Hty: typed_preservation Hred
  ]; tac Hred Hty
).
(* Use this as the last argument of machine_step, to do nothing after the step *)
Ltac step_stop Hred Hty := idtac.

(* Use this as the last argument of machine_step, to destruct the resulting configuration *)
Ltac step_invert Hred Hty := (
  let H := fresh "Hinv" in
    pose proof Hred as H;
    inversions H;
    match goal with
    | H: (@eq trm _ _) |- _ => inversions H
    | H: (@eq stack _ _) |- _ => inversions H
    end
).

(* Destruct the resulting configuration, but use tred_read_inv for reads, which avoids branching *)
Ltac step_invert_simple Hred Hty := (
  let H := fresh "Hinv" in
    match (type of Hred) with
    | (_ ⊢ (trm_read ?x1 _ ; _ ; _ ; _ ; _) ∶ _  ↦ ?G ⊢ ?C)  =>  
      forwards H: tred_read_inv Hred; [>
        apply eq_refl |
        let Hw := fresh "Hw" in
        destructc H as ?w1 ?T2 ?w2 ?y2 ?y1 H;
        destructc H as ?T4 ?T3 ?T5 ?F' ?yd H;
        destructc H as Hw ?HG ?HC ?Hfrw2 ?Htypw1 ?Hbwye ?Hby2G ?Hby1S ?Hhas;
        subst G; subst C; inversions Hw
      ]
    | _ =>
      pose proof Hred as H;
      inversions H;
      match goal with
      | H: (@eq trm _ _) |- _ => inversions H
      | H: (@eq stack _ _) |- _ => inversions H
      end
    end; encapsulate_Lv2 Hred Hty
).

(* Given a typing for a config, creates 0-step reduction closure *)
Ltac red_closure_init HtyC Hredc := (
  match (type of HtyC) with 
  | ty_config ?G ?C ?T => forwards Hredc: typed_red_refl G C T
  end
).

(* Tries to extend given reduction closure by one step *)
Ltac red_closure_extend_of Hredc := (
  let Hstep := fresh "Hstep" in
  match (type of Hredc) with
  | ?F1 ⊢ ?C1 ∶ ?T1 ⊣ ?n ↦* ?F2 ⊢ ?C2 =>
  match goal with 
  | Hred: ?F2 ⊢ ?C2 ∶ ?T1 ↦ ?F3 ⊢ ?C3 |- _ => forwards Hstep: typed_red_step Hredc Hred; clear Hredc; rename Hstep into Hredc
  end
  end
).

(* Tries to extend some reduction closure by one step *)
Ltac red_closure_extend := (
  let Hstep := fresh "Hstep" in
  match goal with 
  | Hredc: ?F1 ⊢ ?C1 ∶ ?T1 ⊣ ?n ↦* ?F2 ⊢ ?C2,
    Hred: ?F2 ⊢ ?C2 ∶ ?T1 ↦ ?F3 ⊢ ?C3 |- _ => forwards Hstep: typed_red_step Hredc Hred; clear Hredc; rename Hstep into Hredc
  end
).

Ltac red_closure_extends_of Hredc := repeat red_closure_extend_of Hredc.
