Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping Equivalence HeapUpdate.

(** * Evolution of heap during execution *)

(* Contents of an object evolved by possible mutation.
   Mutation can only change values of fields,
   so fields can be the same or have different values,
   all other defs must be exactly the same. *)
Inductive defs_evo : defs -> defs -> Prop :=
| defs_evo_nil:
    defs_evo defs_nil defs_nil  
| defs_evo_fld ds1 ds2 a v1 v2:
    defs_evo ds1 ds2 ->
    defs_evo (defs_cons ds1 (def_fld a v1)) (defs_cons ds2 (def_fld a v2))
| defs_update_push ds1 ds2 d:
    defs_evo ds1 ds2 ->
    defs_evo (defs_cons ds1 d) (defs_cons ds2 d)
.
Hint Resolve defs_evo_nil: core.

Lemma defs_evo_refl ds:
  defs_evo ds ds.
Proof.
  induction ds; eauto using defs_update_push.
Qed.
Hint Resolve defs_evo_refl: core.

Lemma defs_evo_symm ds1 ds2:
  defs_evo ds1 ds2 ->
  defs_evo ds2 ds1.
Proof.
  introv He.
  induction He; eauto using defs_update_push, defs_evo_fld.
Qed.

Lemma defs_evo_trans ds1 ds2 ds3:
  defs_evo ds1 ds2 ->
  defs_evo ds2 ds3 ->
  defs_evo ds1 ds3.
Proof.
  introv He12 He23.
  reverts ds3.
  induction He12; intros; inversions He23; eauto using defs_update_push, defs_evo_fld.
Qed.

Lemma defs_update_evo ds1 a x ds2:
  defs_update ds1 a x ds2 ->
  defs_evo ds1 ds2.
Proof.
  introv Hu.
  induction Hu;  eauto using defs_update_push, defs_evo_fld.
Qed.

Inductive item_evo: item -> item -> Prop :=
| item_evo_fun t:
  item_evo (item_fun t) (item_fun t)
| item_evo_obj ds1 ds2:
  defs_evo ds1 ds2 ->
  item_evo (item_obj ds1) (item_obj ds2)
.
Hint Constructors item_evo: core.


Lemma item_evo_refl it:
  item_evo it it.
Proof.
  induction it; eauto using defs_evo_refl.
Qed.
Hint Resolve item_evo_refl: core.

Lemma item_evo_symm it1 it2:
  item_evo it1 it2 ->
  item_evo it2 it1.
Proof.
  introv He.
  induction He; eauto using defs_evo_symm.
Qed.

Lemma item_evo_trans it1 it2 it3:
  item_evo it1 it2 ->
  item_evo it2 it3 ->
  item_evo it1 it3.
Proof.
  introv He12 He23.
  reverts it3.
  induction He12; intros; inversions He23; eauto using defs_evo_trans.
Qed.


(* Definition 5 in sef-step-6.pdf *)
Definition heap_evo S h1 h2 := (
  forall x, x \in S -> 
    exists it1 it2,
      binds x it1 h1 /\
      binds x it2 h2 /\
      item_evo it1 it2
).

Lemma heap_evo_empty h1 h2:
  heap_evo \{} h1 h2.
Proof.
  hnf; intros.
  rewrite in_empty in H.
  contradiction.
Qed.

Lemma heap_evo_subset S1 S2 h1 h2:
  heap_evo S1 h1 h2 ->
  S2 \c S1 ->
  heap_evo S2 h1 h2.
Proof.
  introv He Hs.
  unfold heap_evo in *.
  intros; eapply He.
  eapply Hs; eauto.
Qed.

Lemma heap_evo_dom_refl (h: heap):
  heap_evo (dom h) h h.
Proof.
  unfold heap_evo in *.
  introv Hx.
  eapply dom_to_binds in Hx.
  destructecs Hx.
  exists v v; auto. 
Qed.

Lemma heap_evo_symm S h1 h2:
  heap_evo S h1 h2 ->
  heap_evo S h2 h1.
Proof.
  introv He.
  unfold heap_evo in *.
  introv Hx.
  forwards Hitx: He; fweauto.
  destructecs Hitx.
  exists it2 it1.
  eauto using item_evo_symm.
Qed.

Lemma heap_evo_trans S h1 h2 h3:
  heap_evo S h1 h2 ->
  heap_evo S h2 h3 ->
  heap_evo S h1 h3.
Proof.
  introv He12 He23.
  unfold heap_evo in *.
  intros.
  forwards Hitx12: He12; fweauto.
  forwards Hitx23: He23; fweauto.
  destructecs Hitx12.
  destructecs Hitx23.
  binds_eq.
  exists it1 it3.
  eauto using item_evo_trans.
Qed.

Lemma heap_evo_concat S h1 h2 h1l h1r h2l h2r:
  heap_evo S h1 h2 ->
  ok (h1l & h1 & h1r) ->
  ok (h2l & h2 & h2r) ->
  heap_evo S (h1l & h1 & h1r) (h2l & h2 & h2r).
Proof.
  introv He Hok1 Hok2.
  unfold heap_evo in *.
  introv Hx.
  forwards Hitx: He; fweauto.
  destructecs Hitx.
  exists it1 it2.
  splits.
  -
  eapply binds_concat_left_ok. auto.
  eapply binds_concat_right; eauto.
  -
  eapply binds_concat_left_ok. auto.
  eapply binds_concat_right; eauto.
  - auto.
Qed.

Lemma heap_evo_push_l S h1 h2 x it:
  heap_evo S h1 h2 ->
  x # h1 ->
  heap_evo S (h1 & x ~ it) h2.
Proof.
  introv He Hx.
  unfold heap_evo in *.
  introv Hx0.
  forwards Hitx: He; fweauto.
  destructecs Hitx.
  exists it1 it2.
  splits.
  -
  eapply binds_concat_left. auto.
  intro. rewrite dom_single, in_singleton in H. subst. eapply binds_to_dom in Hitx0. contradiction.
  - auto.
  - auto.
Qed.

Lemma heap_evo_push_r S h1 h2 x it:
  heap_evo S h1 h2 ->
  x # h2 ->
  heap_evo S h1 (h2 & x ~ it).
Proof.
  introv He Hx.
  eapply heap_evo_symm.
  eapply heap_evo_push_l.
  eapply heap_evo_symm; eauto.
  auto.
Qed.

Lemma heap_evo_dom S h1 h2:
  heap_evo S h1 h2 ->
  S \c dom h1 /\ S \c dom h2.
Proof.
  introv He.
  unfold heap_evo in *.
  split; intro; intro; forwards Hx: He; fweauto; destructecs Hx; eapply binds_to_dom; eauto.
Qed.

Lemma heap_update_evo h1 x a y h2:
  heap_update h1 x a y h2 ->
  heap_evo (dom h1) h1 h2.
Proof.
  introv Hu.
  unfold heap_update, heap_evo in *.
  destructecs Hu.
  introv Hx0.
  destruct (classicT(x0 = x)).
  - (* updated item *)
    subst.
    exists (item_obj ds) (item_obj ds').
    splits; eauto using defs_update_evo.
  - (* other item *)
    forwards Hbix0: dom_to_binds Hx0.
    destructecs Hbix0.
    forwards Hbi3: Hu3 n Hbix0.
    exists v v; eauto.
Qed.

Require Import OperationalSemantics.

Lemma red_heap_evo C1 C2:
  red C1 C2 ->
  heap_evo (dom (config_heap C1)) (config_heap C1) (config_heap C2).
Proof.
  introv Hred.
  induction Hred;
  unfold config_heap in *; try eapply heap_evo_dom_refl.
  - (* write *)
    eapply heap_update_evo; eauto.
  - eapply heap_evo_push_r. eapply heap_evo_dom_refl.
    inversions H.
    notin_solve.
  - eapply heap_evo_push_r. eapply heap_evo_dom_refl.
    inversions H.
    notin_solve.
Qed.


(* Lemma 6 in sef-step-6.pdf *)

Lemma red_heap_evo_stable S C1 C2 C3 C4:
  heap_evo S (config_heap C1) (config_heap C2) ->
  red C1 C3 ->
  red C2 C4 ->
  heap_evo S (config_heap C3) (config_heap C4).
Proof.
  introv He Hred1 Hred2.
  forwards He1: red_heap_evo Hred1.
  forwards He2: red_heap_evo Hred2.
  forwards Hdom: heap_evo_dom He.
  destructecs Hdom.

  eapply heap_evo_trans.
  eapply heap_evo_symm.
  eapply heap_evo_subset; eauto.
  eapply heap_evo_trans. eauto.
  eapply heap_evo_subset; eauto.
Qed.