Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax OperationalSemantics ReductionLemmas.

(** * Definitions and lemmas about execution of a method *)

(* A "method call begin" is a configuration that calls a method *)
Inductive method_call_begin: config -> Prop :=
| method_call_begin_c w1 m w2 fs S e Lv:
  method_call_begin (trm_call (avar_f w1) m (avar_f w2); fs; S; e; Lv).

(* Projection config -> stack *)
Definition config_stack C := (
  match C with
  t1 ; fs1 ; S1 ; e1 ; Lv1 => fs1
  end
).

(* Projection config -> term *)
Definition config_trm C := (
  match C with
  t1 ; fs1 ; S1 ; e1 ; Lv1 => t1
  end
).


Lemma red_stack_cases' C1 C2 :
  C1 ↦ C2 ->
  stack_suffix (config_stack C1) (config_stack C2) \/
  exists w1 f1,
    (config_trm C1) = trm_var (avar_f w1) /\
    (config_stack C1) = (stack_cons f1 (config_stack C2)).
Proof.
  introv Hred.
  destruct C1, C2.
  eapply red_stack_cases; eauto.
Qed.

(* Execution keeps within the same frame for at least m steps *)
Definition frame_within C1 m := (
  forall k C4,
    k <= m ->
    C1 ⊣ k ↦* C4 ->
    stack_suffix (config_stack C1) (config_stack C4)
).

(* C3 is frame end for C1 in m steps *)
Definition frame_end_in C1 m C3 := (
  C1 ⊣ m ↦* C3 /\
  frame_within C1 m /\
  (config_stack C3) = (config_stack C1) /\
  exists w3,
    (config_trm C3) = (trm_var (avar_f w3))
).

(* Version allowing non-determinism *)
Definition frame_end_in' C1 m C3 := (
  C1 ⊣ m ↦* C3 /\
  (config_stack C3) = (config_stack C1) /\
  exists w3,
    (config_trm C3) = (trm_var (avar_f w3))
).

Lemma frame_within_0 C1:
  frame_within C1 0.
Proof.
  hnf; intros; destruct C1; simpl.
  eapply Le.le_n_0_eq in H; subst.
  inversions H0; simpl.
  eapply stack_suffix_refl.
Qed.
Lemma frame_within_S_inv C1 n:
  frame_within C1 (S n) ->
  frame_within C1 n.
Proof.
  introv Hfw.
  hnf in *; intros.
  eapply Hfw. eapply PeanoNat.Nat.le_le_succ_r. eauto. eauto.
Qed.

Lemma frame_within_suffix C1 C2 n:
  C1 ⊣ n ↦* C2 ->
  frame_within C1 n ->
  stack_suffix (config_stack C1) (config_stack C2).
Proof.
  introv Hred Hwi.
  hnf in Hwi.
  forwards: Hwi n Hred;
  eauto.
Qed.

(* Version allowing non-determinism *)
Lemma red_closure_stack_cases' C1 C2 n:
  C1 ⊣ n ↦* C2 ->
  stack_suffix (config_stack C1) (config_stack C2) \/
  exists m C3,
    m < n /\
    frame_end_in' C1 m C3.
Proof.
  introv Hred.
  reverts C2.
  induction n; intros; inversions Hred.
  - left. eapply stack_suffix_refl.
  - forwards: IHn H0.
    destruct H.
    eapply red_stack_cases' in H2.
    destruct H2.
    + left. eapply stack_suffix_trans; eauto.
    + destruct C3. sympls. destructc H1 as w1 f1 Ht Hs; subst.
      eapply stack_suffix_cons_inv in H.
      destruct C1, C2; simpls. destruct H; subst.
      * right.
        exists n (trm_var (avar_f w1); (stack_cons f1 s0);   S;  e;  Lv).
        split. eauto.
        hnf. splits; eauto. exists w1; simpl; eauto.
      * eauto.
    + right. destructc H as ? ? ? ? ? ? ?. exists x x0. destruct C1, x0; simpls; subst.
      split. eauto. hnf; splits; eauto.
       exists x1; simpl; eauto.
Qed.
Lemma frame_end_exists' C1:
  config_terminates C1 ->
  exists n C2,
  frame_end_in' C1 n C2.
Proof.
  introv Ht.
  destructc Ht as nt Ct HaCt HredCt.
  pose proof HredCt as Hst.
  eapply red_closure_stack_cases' in Hst.
  inversions HaCt. destruct C1; simpls.
  destruct Hst.
  - eapply stack_suffix_nil_inv in H; subst. exists nt. eexists. hnf. splits. eauto. all: simpls; eauto.
  - destruct_all. eauto.
Qed.

Lemma frame_not_end_stays_within c n :
  not (exists m ce, frame_end_in c m ce /\ m < n) ->
  frame_within c n.
Proof.
  introv Hnend. induction n.

  -introv Hlt Hstep. inversions Hlt. inversions Hstep.
    eauto using stack_suffix_refl.

  -
    assert (~ (exists m ce, frame_end_in c m ce /\ m < n)) as Hnend'. {
      intro. eapply Hnend. destructecs H.
      exists m ce. split; eauto.
    }
    introv Hlt Hstep. inversions Hlt; eauto.
    +
      inversions Hstep. eapply IHn in H0 as Hsuf; eauto.
      destruct C2. destructc Hsuf as fs ?; simpl in *; subst.
      inversions H2; simpl.
      all: try solve[exists; eauto using stack_suffix_refl].
      2: { exists (stack_cons (frame_let t0) fs). simpl. easy. }
      destruct fs; simpl in *; eauto.
      2: { inversions H3. exists. eauto. }
      false. eapply Hnend. exists; split.
      * splits; eauto. simpl. eauto.
      * eauto.
    +
      eapply IHn.
      exact Hnend'. eauto. eauto.
Qed.

Lemma frame_within_truncate_stack t fs Sigma e Lv t' fs' Sigma' e' Lv' k n :
  let c  := (t ; fs ; Sigma ; e ; Lv) in
  let c' := (t' ; stack_append fs' fs ; Sigma' ; e' ; Lv') in
  frame_within c n -> k <= n ->
  c ⊣k↦* c' ->
  exists Lv0,
    (t ; stack_nil ; Sigma ; e ; fv_config c)
      ⊣k↦* (t' ; fs' ; Sigma' ; e' ; Lv0) /\
      Lv0 \c fv_config c'.
Proof.
  introv Hwin Hlt Hsteps.
  reverts Lv' e' Sigma' fs'. reverts t'. induction k; intros;
    inversions Hsteps; eauto.
  { destruct fs'.
    - simpl. exists. splits; eauto. eauto using subset_refl.
    - simpl. false.
      forwards: stack_append_longer_r.
      2: { rewrite <- H2 in H. omega. }
      discriminate. }

  rename H0 into Hreds, H2 into Hred.
  forwards Hsf:Hwin. 2: eassumption. 1: omega.
  destruct C2. simpl in *. destructc Hsf as fs0 Heq; subst.
  forwards IH:IHk; try omega; try eassumption.
  destructc IH as Lv1 IH Hsub.

  inversions Hred.

  { Case "red_read".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_read; eauto.
      eapply fresh_subset; try eassumption. rewrite fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. rewrite ! in_union in *. tauto.
    - simpl. rewrite ! fv_stack_append, ! dom_push, fv_env_push in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. rewrite ! in_union in *. tauto. }

  { Case "red_write".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_write; eauto.
    - simpl. rewrite ! fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_call".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_call; eauto.
    - simpl. rewrite ! fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_apply".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_apply; eauto.
    - simpl. rewrite ! fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_let_fun".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_let_fun; eauto.
      eapply fresh_subset; try eassumption. rewrite fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. rewrite ! in_union in *. tauto.
    - simpl. rewrite ! fv_stack_append, ! dom_push, fv_env_push in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_let_obj".
    try erewrite <- (stack_append_l_injective fs0 fs'); eauto.
    exists. splits.
    - econstructor; try eassumption. eapply red_let_obj; eauto.
      eapply fresh_subset; try eassumption. rewrite fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. rewrite ! in_union in *. tauto.
    - simpl. rewrite ! fv_stack_append, ! dom_push, fv_env_push in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_let_var".
    forwards: (stack_append_l_injective
                 fs0 (stack_cons (frame_let t1) fs')).
    1: simpl; eauto. subst. clear H1.
    exists. splits.
    - econstructor; try eassumption. eapply red_let_var; eauto.
    - simpl. rewrite ! fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

  { Case "red_let_push".
    forwards: (stack_append_l_injective
                 fs' (stack_cons (frame_let t1) fs0)).
    1: simpl; eauto. subst. clear H6.
    exists. splits.
    - econstructor; try eassumption. eapply red_let_push; eauto.
    - simpl. rewrite ! fv_stack_append in *.
      introv Hin. rewrite ! in_union in *. destruct_all; try tauto.
      eapply Hsub in H. simpl in *. rewrite ! in_union in *. tauto. }

Qed.


Lemma frame_within_truncate_stack' t fs Sigma e Lv t' Sigma' e' Lv' k n fs2:
  let c  := (t ; fs ; Sigma ; e ; Lv) in
  let c' := (t' ; fs2 ; Sigma' ; e' ; Lv') in
  frame_within c n ->
  k <= n ->
  c ⊣k↦* c' ->
  exists Lv0 fs',
    fs2 = stack_append fs' fs /\
    (t ; stack_nil ; Sigma ; e ; fv_config c)
      ⊣k↦* (t' ; fs' ; Sigma' ; e' ; Lv0) /\
      Lv0 \c fv_config c'.
Proof.
  introv Hwin Hlt Hsteps.
  lets Hwin': Hwin.
  hnf in Hwin.
  forwards Hfs: Hwin k  (t' ; fs2 ; Sigma' ; e' ; Lv'); fweauto.
  hnf in Hfs.
  destructecs Hfs.
  simpl in Hfs. subst.

  forwards Hts: frame_within_truncate_stack Hwin' Hlt Hsteps.
  destructecs Hts.
  exists Lv0 fs3. eauto.
Qed.


