Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax Deref HeapUpdate.

(** * Operational Semantics *)

Reserved Notation "t1 ↦ t2" (at level 40, t2 at level 39).

(* Figure 10 in "Reference mutability for DOT" *)
(* Figure 13 in "Pure methods for roDOT" *)
Inductive red : config -> config -> Prop :=
(** [Σ(yObj) = ...{a=yVal}...]  #<br>#
    [(wObj↦yObj) ∈ e]  #<br>#
    [―――――――――――――――――――――――――(R-Read)]  #<br>#
    [(wObj.a; fs; Σ; e) ↦ (v wVal; fs; Σ; e ∪ (wVal↦yVal))      ]  *)
| red_read : forall yObj wObj m Sigma e Lv yVal wVal ds fs,
    fresh_in_config (wVal :: nil) (trm_read (avar_f wObj) m; fs ; Sigma; e; Lv) ->
    binds wObj yObj e ->
    binds yObj (item_obj ds) Sigma ->
    defs_has ds (def_fld m (avar_f yVal)) ->
    (trm_read (avar_f wObj) m; fs ; Sigma; e; Lv)
      ↦ (trm_var (avar_f wVal); fs; Sigma; e & wVal ~ yVal; fv_config (trm_read (avar_f wObj) m; fs ; Sigma; e; Lv))

(** [Σ(yObj) = ...{a=y}...]  #<br>#
    [(wObj↦yObj) ∈ e]  #<br>#
    [(wArg↦yArg) ∈ e]  #<br>#
    [―――――――――――――――――――――――――(R-Write)]  #<br>#
    [(wObj.a = wArg; fs; Σ; e) ↦ (v wVal; fs; Σ[yObj≔...{a=yArg}...]; e)      ]  *)
| red_write : forall yObj wObj m Sigma e Lv yArg wArg Sigma' fs,
    heap_update Sigma yObj m yArg Sigma' ->
    binds wObj yObj e ->
    binds wArg yArg e ->
    (trm_write (avar_f wObj) m (avar_f wArg); fs; Sigma; e; Lv)
      ↦ (trm_var (avar_f wArg); fs ; Sigma'; e; fv_config (trm_write (avar_f wObj) m (avar_f wArg); fs; Sigma; e; Lv))

(** [Σ(yObj) = ...{m=t}...]  #<br>#
    [(wObj↦yObj) ∈ e]  #<br>#
    [―――――――――――――――――――――――――(R-Call)]  #<br>#
    [(wObj.m(wArg); fs; Σ; e) ↦ (t^wArg^wObj; fs; Σ; e)      ]  *)
| red_call : forall yObj wObj m Sigma e Lv wArg t ds fs,
    binds yObj (item_obj ds) Sigma ->
    defs_has ds (def_met m t) ->
    binds wObj yObj e ->
    (trm_call (avar_f wObj) m (avar_f wArg); fs ; Sigma; e; Lv)
      ↦ ((open wArg (open_rec 1 wObj t)); fs ; Sigma; e; fv_config (trm_call (avar_f wObj) m (avar_f wArg); fs ; Sigma; e; Lv))

(** [Σ(yFun) = λt]  #<br>#
    [(wFun↦yFun) ∈ e]  #<br>#
    [―――――――――――――――――――――(R-Apply)]  #<br>#
    [(wFun wArg; fs; Σ; e) ↦ (t^wArg; fs; Σ; e)      ]  *)
| red_apply : forall yFun wFun wArg Sigma e Lv t fs,
    binds yFun (item_fun t) Sigma ->
    binds wFun yFun e ->
    (trm_apply (avar_f wFun) (avar_f wArg); fs; Sigma; e; Lv)
      ↦ (open wArg t; fs; Sigma; e; fv_config (trm_apply (avar_f wFun) (avar_f wArg); fs; Sigma; e; Lv))

| red_let_fun : forall T t1 t2 Sigma e Lv y w fs,
    fresh_in_config (y :: w :: nil) (trm_lit (lit_fun T t1) t2 ; fs; Sigma; e; Lv) ->
    y <> w ->
    (trm_lit (lit_fun T t1) t2 ; fs; Sigma; e; Lv)
      ↦ (open w t2 ; fs; Sigma & y ~ (item_fun t1); e & w ~ y ; fv_config (trm_lit (lit_fun T t1) t2 ; fs; Sigma; e; Lv))

(** [―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――(R-LetObj)] #<br>#
    [((let z=ν(s:R).d in t); fs; Σ; e) ↦ (open w t; fs; Σ∪(y↦open w (e d)); e∪(w↦y)] *)
| red_let_obj : forall T ds t Sigma e Lv y w fs,
    fresh_in_config (y :: w :: nil) (trm_lit (lit_obj T ds) t ; fs; Sigma; e; Lv) ->
    w <> y ->
    (trm_lit (lit_obj T ds) t ; fs; Sigma; e; Lv)
      ↦ (open w t ; fs; Sigma & y ~ (item_obj (deref_defs (e & w ~ y) (open w ds))); e & w ~ y ; fv_config (trm_lit (lit_obj T ds) t ; fs; Sigma; e; Lv))

(** [―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――(R-LetVar)] #<br>#
    [(w; (frame_let t) :: fs; Σ; e) ↦ (open w t; fs; Σ; e)] *)
| red_let_var : forall t Sigma e Lv x fs,
    (trm_var (avar_f x); (stack_cons (frame_let t) fs); Sigma; e; Lv)
      ↦ (open x t; fs; Sigma; e ; fv_config (trm_var (avar_f x); (stack_cons (frame_let t) fs); Sigma; e; Lv))

(** [―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――(R-LetPush)] #<br>#
    [((let z=t₁ in t₂) ; fs; Σ; e) ↦ (t₁; (frame_let t₂) :: fs₁; Σ; e)] *)
| red_let_push : forall s t fs Sigma e Lv,
    (trm_let s t; fs; Sigma; e; Lv)
      ↦ (s; (stack_cons (frame_let t) fs); Sigma; e ; fv_config (trm_let s t; fs; Sigma; e; Lv))

where "t1 ↦ t2" := (red t1 t2).

(** ** Normal forms *)
(** A configuration with a location and an empty stack are considered to be
    answers. *)
Inductive answer: config -> Prop :=
| nf_loc: forall x Sigma e Lv, answer (trm_var (avar_f x); stack_nil; Sigma; e; Lv).

Hint Constructors red answer : core.

(** ** Multiple step reduction *)
(** Transitive closure of the reduction relation. *)

Reserved Notation "C1 ⊣ n ↦* C2"
         (at level 40, n at level 41, C2 at level 39).

Inductive red_closure :  config -> nat -> config -> Prop :=

| red_refl : forall (C : config),
    C ⊣ 0 ↦* C

| red_trans : forall (C1 C2 C3 : config) (n : nat),
    C1 ⊣ n ↦* C2 ->
    C2      ↦  C3 ->
    C1 ⊣ S n ↦* C3

where "C1 ⊣ n ↦* C2" := (red_closure C1 n C2).

Hint Constructors red_closure : core.

(** Termination *)
Definition config_terminates_in C n :=
 (exists (Ca : config), answer Ca /\ (C ⊣n↦* Ca)).
Definition config_terminates C :=
 (exists (n : nat), config_terminates_in C n).
