Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax OperationalSemantics MethodCall.

Require Import TypingMode.
Require Import Similarity ConfigSimilarity SimilarityPreservation SyntaxSimilarity.
Require Import ConfigValid ConfigTyping TypedSimilarityPreservation TypedOperationalSemantics ReductionLemmas TypedReductionEquivalence TypedSoundness.

(** * Definitions and lemmas about typed execution of a method *)

(* Use that reduction is deterministic to show that the frames do not change *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* If two similar configurations are reduced one step and one preserves a stack suffix, then the other also does. *)
Lemma similar_red_suffix_preservation C2 C3 C5 C6:
  C2 ↦ C3 ->
  C5 ↦ C6 ->
  similar_conf_1to1 C2 C5 ->
  stack_suffix (config_stack C2) (config_stack C3) ->
  stack_suffix (config_stack C5) (config_stack C6).
Proof.
  introv Hred23 Hred56 Hsi Hsu.
  (* Consider cases where the would be preserved or would not *)
  forwards Hcases56: red_stack_cases' Hred56.
  destruct Hcases56.
  - (* Stack is preserved - trivial *)
    eauto.
  - (* Stack is not preserved - need to find a contradiciton with the premises *)
    (* If stack is not preserved, that means it must have been popped. *)
    destruct_all. destruct C5. sympls; subst.
    (* By similarity, also in C2 it has to be popped *)
    destruct C2. hnf in Hsi. destruct_all.
    destruct C3.
    sympls.
    hnf in H2. destruct_all.
    eapply similar_symm in H2; eauto.
    eapply similar_trm_var_case in H2.
    destruct_all; subst.
    inversions Hred23.
    (* But that is contradiction with the hypothesis that it is a suffix *)
    eapply stack_suffix_cons_same_false in Hsu.
    contradiction.
Qed.

(* If two similar configurations are reduced one step and one preserves a stack suffix, then the other also does. *)
Lemma similar_red_suffix_preservation' C2 C3 C5 C6 C1 C4:
  C2 ↦ C3 ->
  C5 ↦ C6 ->
  similar_conf_1to1 C2 C5 ->
  similar_conf_1to1 C1 C4 ->
  stack_suffix (config_stack C1) (config_stack C2) ->
  stack_suffix (config_stack C4) (config_stack C5) ->
  stack_suffix (config_stack C1) (config_stack C3) ->
  stack_suffix (config_stack C4) (config_stack C6).
Proof.
  introv Hred23 Hred56 Hsi25 Hsi14 Hsu12 Hsu45 Hsu13.
  (* Consider cases where the would be preserved or would not *)
  forwards Hcases56: red_stack_cases' Hred56.
  destruct Hcases56.
  - (* Stack is preserved - trivial *)
    eapply stack_suffix_trans; eauto.
  - (* Stack is not preserved - need to find a contradiciton with the premises *)
    (* If stack is not preserved, that means it must have been popped. *)
    destruct_all. rename x into w5, x0 into f5.
    destruct C5 as [t5 fs5 S5 e5 Lv5]. sympls. subst.

    eapply stack_suffix_cons_inv in Hsu45. destruct Hsu45.
    2: eauto. rewrite H.

    false. rename H into Hsu45.
    (* By similarity, also in C2 it has to be popped *)
    destruct C1 as [t1 fs1 S1 e1 Lv1]. sympls.
    destruct C2 as [t2 fs2 S2 e2 Lv2]. sympls.
    destruct C3 as [t3 fs3 S3 e3 Lv3]. sympls.
    destruct C4 as [t4 fs4 S4 e4 Lv4]. sympls.
    destruct C6 as [t6 fs6 S6 e6 Lv6]. sympls.

    (* We have (t₂;σ₂;Σ₂;e₂;L₂) = C₂ ≈ C₅ = (w₅;f₅∷σ₆;Σ₅;e₅;L₅).

       (1) And so t₂ ≈ w₅. As a result t₂ = w₂ for some w₂ variable.

       (2) Also, σ₄ = f₅∷σ₆ ≈ σ₂ = ...σ₁ and σ₁ ≈ σ₄ so by f₅∷σ₆ ≈ σ₂
       one has σ₂ = f₂∷σ₂' with f₂ ≈ f₅ and σ₆ ≈ σ₂' Moreover, σ₁ is a
       suffix of σ₂ = f₂∷σ₂' so σ₁ is either the entire list σ₁ =
       f₂∷σ₂' or a proper suffix.  Suppose σ₁ is a proper suffix so
       that σ₂ = σ₁' ++ σ₁ where σ₁' is not empty. Then σ₄ ≈ σ₂ = σ₁'
       ++ σ₁. So |σ₄| > |σ₁|. But this is impossible, because σ₁ ≈ σ₄
       so |σ₁| = |σ₄|.  This means σ₁ = f₂∷σ₂'.

       By hypothesis and (1), we see (t₂;σ₂;Σ₂;e₂;L₂) =
       (w₂;f₂∷σ₂';Σ₂;e₂;L₂) = C₂ ▷ C₃ = (t₃;σ₃;Σ₃;e₃;L₂) which means
       this reduction can only happen by popping off the frame. That
       is, f₂ = let t₂' and t₃ = open t₂' w₂ and most importantly σ₃ =
       σ₂'.

       As hypothesis, we know that σ₁ is some suffix of σ₃ = σ₂', but
       this makes σ₁ a proper suffix of f₂∷σ₂' = σ₁, which is
       absurd. This makes a required contradiction. *)

    hnf in Hsi25. destruct Hsi25 as
      [xs252 [xs255 [Hvcu252 [Hvcu255 [Hlen25 Hsim25]]]]].
    hnf in Hsim25. destruct Hsim25 as [Htsim25 [Hfssim25 [HSsim25 Hesim25]]].

    apply similar_symm in Htsim25.
    apply similar_trm_var_case in Htsim25 as [aw2 [Heqt2 Hsim52]].

    apply similar_stack_cons_inv in Hfssim25 as Hfssim25'.
    destruct Hfssim25' as [f2 [fs2' [Heqfs2 Hfssim26]]].
    rewrite Heqfs2 in Hsu12.
    eapply stack_suffix_cons_inv in Hsu12.
    destruct Hsu12 as [Heqfs1 | Hcontra].
    2: {
      assert (exists fs1', fs2 = stack_append fs1' fs1 /\ fs1' <> stack_nil)
      as Hconcat.
      { destruct Hcontra as [fs10' Hconcat].
        exists (stack_cons f2 fs10'). split; simpl. rewrite <- Hconcat.
        auto. discriminate. }
      destruct Hconcat as [fs1' [Hconcat Hnnil]].
      rewrite <- Hsu45 in Hfssim25. (* σ₂ ≈ σ₄ *)
      apply stack_similar_length in Hfssim25.
      hnf in Hsi14. destruct Hsi14 as
        [xs141 [xs144 [Hvcu141 [Hvcu144 [Hlen14 Hsim14]]]]].
      hnf in Hsim14. destruct Hsim14 as [_ [Hfssim14 _]].
      apply stack_similar_length in Hfssim14.
      pose proof (stack_append_longer_r fs1 Hnnil) as Hlonger.
      rewrite <- Hconcat in Hlonger. omega.
    }

    rewrite Heqt2 in Hred23. inversions Hred23. inversions H1.
    hnf in Hsu13. destruct Hsu13 as [fs2'' Hcontra].
    apply fun_arg_eq_1 with (f := stack_length) (g := stack_length)
      in Hcontra; auto.
    rewrite stack_append_length in Hcontra. simpl in Hcontra. omega.
Qed.

(* If _ ++ C₁ = C₂ ≈ C₅ = _ ++ C₄ for C₁ ≈ C₄ and C₂ never pops C₁
   in all of n steps then C₅ never pops C₄ in n steps. *)
Lemma similar_red_suffix_preservation'_closure n C2 C5 C1 C4 :
  config_hastyp typing_mode C2 -> config_hastyp typing_mode C5 ->
  similar_conf_1to1 C2 C5 ->
  similar_conf_1to1 C1 C4 ->
  (forall k C3, k <= n -> C2 ⊣k↦* C3 ->
           stack_suffix (config_stack C1) (config_stack C3)) ->
  stack_suffix (config_stack C4) (config_stack C5) ->
  (forall k C3 C6, k <= n -> C2 ⊣k↦* C3 -> C5 ⊣k↦* C6 ->
           stack_suffix (config_stack C4) (config_stack C6)).
Proof.
  reverts C2 C5 C1 C4.
  induction n; introv Hty2 Hty5 Hsim25 Hsim14 Hsuf12 Hsuf45 ;
    introv Hlt Hred23 Hred56.
  - inversions Hlt. inversions Hred56. inversions Hred23. auto.
  - assert (Hltcases : k <= n \/ k = S n) by omega; clear Hlt.
    destruct Hltcases as [Hlt | ?]; subst; eauto.
    inversions Hred23. rename C7 into C2', H0 into Hred22', H2 into Hred23.
    inversions Hred56. rename C7 into C5', H0 into Hred55', H2 into Hred56.
    specialize (Hsuf12 _ C2' (Nat.le_succ_diag_r _) Hred22') as Hsuf12'.
    apply similar_red_suffix_preservation'
      with (C2 := C2') (C3 := C3) (C5 := C5') (C1 := C1); auto.
    + eapply similar_conf_1to1_preservation_closure
        with (Ca1 := C2) (Cb1 := C5) (n := n); auto.
    + specialize (IHn _ _ _ _ Hty2 Hty5 Hsim25 Hsim14).
      apply IHn with (k := n) (C3 := C2'); eauto.
    + eapply Hsuf12 with (k := S n); eauto.
Qed.

(* If reduction stays within frame after n steps,
   and the (n+1) th step does not pop the stack,
   then it stays within frame after (n+1) steps.
*)
Lemma frame_within_S C1 C2 C3 n:
  config_hastyp typing_mode C1 ->
  C1 ⊣ n ↦* C2 ->
  C2 ↦ C3 ->
  stack_suffix (config_stack C2) (config_stack C3) ->
  frame_within C1 n ->
  frame_within C1 (S n).
Proof.
  introv Hht Hredc Hred Hsu Hwi.
  forwards Hva: config_hastyp_valid Hht.
  unfold frame_within in *; intros.
  apply -> Nat.le_succ_r in H.
  destruct H.
  - eauto.
  - subst.
    inversions H0.
    forwards Hsu15: Hwi H1; eauto.
    forwards Hsu12: Hwi Hredc; eauto.

    forwards Hsi11: similar_conf_1to1_refl Hva.
    forwards Hsi25: similar_conf_1to1_preservation_closure Hsi11 Hredc H1; eauto.
    forwards: similar_red_suffix_preservation Hred H3; eauto.
    eapply stack_suffix_trans; eauto.
Qed.

(* If C₂ ≈ C₅ and C₂ doesn't pop in n steps then neither does C₅.
   The small caveat is that we can only use this lemma when C₂ does have
   at least n steps of reduction left. *)
Lemma similar_red_within_preservation C2 C3 C5 n:
  config_hastyp typing_mode C2 ->
  config_hastyp typing_mode C5 ->
  C2 ⊣ n ↦* C3 ->
  similar_conf_1to1 C2 C5 ->
  frame_within C2 n ->
  frame_within C5 n.
Proof.
  introv Hht2 Hht5 Hred23 Hsi Hfw. reverts C2 C3 C5.
  induction n; intros.
  - hnf; intros. eapply Le.le_n_0_eq in H; subst. inversions H0.
    eapply stack_suffix_refl.
  - inversions Hred23. rename C0 into C2', H0 into Hred22, H2 into Hred23.
    specialize (IHn _ Hht2 (frame_within_S_inv Hfw) _ Hred22 _ Hht5 Hsi).
    hnf. intros k C6 Hlt Hred.
    assert (k <= n \/ k = S n) as Hltcases by omega; clear Hlt.
    destruct Hltcases as [Hlt | ?]; subst; eauto.
    eapply similar_red_suffix_preservation'_closure
      with (C2 := C2) (C5 := C5) (C1 := C2) (n := S n) (k := S n).
    all: eauto using stack_suffix_refl.
Qed.

(* After n steps, either a frame end has been encountered,
   or stayed within the frame. *)
Lemma red_closure_stack_cases C1 C2 n:
  config_hastyp typing_mode C1 ->
  C1 ⊣ n ↦* C2 ->
  frame_within C1 n \/
  exists m C3,
    m < n /\
    frame_end_in C1 m C3.
Proof.
  introv Hht Hred12.
  reverts C2.
  induction n; intros.
  - inversions Hred12. left. eapply frame_within_0.
  - (* By cases - stayed within frame after n steps *)
    pose proof Hred12 as H0. inversions H0.
    rename H1 into Hred13, H3 into Hred32.
    forwards: IHn Hred13. destruct H as [Hstay1 | Hend1].
    -- (* Stayed within frame *)
      pose proof Hred32 as Hred32'.
      (* By cases - popping a frame at (n+1) steps ? *)
      eapply red_stack_cases' in Hred32'.
      destruct Hred32' as [Hsu32 | Hpop32].
    + (* Still not popping a frame - staying within *)
      left. eapply frame_within_S; eauto.
    + (* Popping a frame *)
      destructc Hpop32 as w3 f3 Ht3 Hs3; subst.
      (* Up to now within frame *)
      pose proof Hstay1 as Hxx. hnf in Hxx.
      forwards Hsu13: Hxx n; eauto.
      destruct C3; simpls; subst.
      (* Is it the frame of interest? *)
      eapply stack_suffix_cons_inv in Hsu13.
      destruct Hsu13 as [Hentire | Hsu12]; subst.
      ** (* Yes, it is the one *)
        destruct C1; simpls; subst.
        right. exists n. eexists.
        splits. eauto. hnf; splits. eauto. eauto. simpls. eauto.
        exists w3; simpls; eauto.
      ** (* No, it is some smaller frame. *)
        left. hnf; introv Hlt Hred14.
        (* Staying within frame even after n+1 steps *)
        eapply Lt.le_lt_or_eq in Hlt. destruct Hlt as [Hlt | ?]; subst.
        eapply Lt.lt_n_Sm_le in Hlt; eauto.
        inversions Hred14.
        rename C3 into C3', H0 into Hred13', H2 into Hred34'.
        remember (trm_var (avar_f w3) ; stack_cons f3 (config_stack C2)
                          ; S ; e ; Lv) as C3.
        eapply similar_red_suffix_preservation'
          with (C2 := C3) (C3 := C2) (C5 := C3') (C1 := C1);
          eauto using similar_conf_1to1_refl, config_hastyp_valid,
          similar_conf_1to1_preservation_closure.
      -- (* Popped frame *)
        destruct Hend1 as [m [C3' [Hlt Hend13']]].
        right. exists m C3'. split; auto.
Qed.

Lemma frame_end_exists C1:
  config_hastyp typing_mode C1 ->
  config_terminates C1 ->
  exists C2 n,
  frame_end_in C1 n C2.
Proof.
  introv Htyp Ht.
  destructc Ht as nt Ct HaCt HredCt.
  destruct Ct. pose proof HredCt as Hst.
  eapply red_closure_stack_cases in Hst; eauto.
  inversions HaCt. destruct Hst.
  - remember (trm_var (avar_f x) ; stack_nil ; S ; e ; Lv) as C.
    exists C nt. hnf. repeat split; eauto.
    + simpl. hnf in H. specialize (H nt C (le_n _) HredCt). subst.
      sympls. hnf in H. destruct H as [fs Heq].
      destruct C1. sympls. destruct s. auto.
      destruct fs. all: sympls. all: discriminate.
    + eexists. subst. sympls. eauto.
  - destruct_all. do 2 eexists. eauto.
Qed.

(* Find call end *)
Lemma tred_call_end_exists x m y fs G S e Lv1 T:
  let C1 := (trm_call (avar_f x) m (avar_f y) ; fs ; S ; e ; Lv1) in
  ty_config G C1 T ->
  config_terminates C1 ->
  exists n w S2 e2 Lv2 G2,
       G ⊢ C1 ∶ T ⊣ n ↦* G2 ⊢ ((trm_var (avar_f w)) ; fs ; S2 ; e & e2 ; Lv2) /\
      frame_within C1 n.
Proof.
  introv HtyC1 Hct.

 
  forwards Hfe: frame_end_exists.
  econstructor. eauto. eauto.
  destructecs Hfe.
  hnf in Hfe; destructecs Hfe.
  destruct C2. sympls. subst.
  
  forwards He: red_closure_mono_e.
  eapply Hfe0. destructecs He. subst.

  forwards Htred: untyped_red_to_typed_closure Hfe0. eauto.
  destructecs Htred.

  exists n w3 S0 e3 Lv Gamma'.
  auto.
Qed.

Lemma tred_call_end_exists_or_within x m y fs G S e Lv1 T n2:
  let C1 := (trm_call (avar_f x) m (avar_f y) ; fs ; S ; e ; Lv1) in
  ty_config G C1 T ->
  (
    exists n w S2 e2 Lv2 G2,
       G ⊢ C1 ∶ T ⊣ n ↦* G2 ⊢ ((trm_var (avar_f w)) ; fs ; S2 ; e & e2 ; Lv2) /\
      frame_within C1 n
  ) \/ (
    exists C2 G2,
       G ⊢ C1 ∶ T ⊣ n2 ↦* G2 ⊢ C2 /\
      frame_within C1 n2
  ) .
Proof.
  introv HtyC1.

  forwards Hsnd: typed_red_soundness HtyC1.
  destructecs Hsnd.
  - (* Terminating case *)
    left. eapply tred_call_end_exists. eauto. exists n Ca; eauto using typed_to_untyped_red_closure.
  - (* Non-terminating case *)
    forwards Hsndn: Hsnd n2.
    destructecs Hsndn.
    forwards Hsc: red_closure_stack_cases.
    eapply config_hastyp_c; eauto.
    eapply typed_to_untyped_red_closure; eauto.
    destructecs Hsc.
    + right. eauto.
    + hnde Hsc. destruct C3. sympls; subst.
  forwards He: red_closure_mono_e.
  eapply Hsc1. destructecs He. subst.

  forwards Htred: untyped_red_to_typed_closure Hsc1. eauto.
  destructecs Htred.

  left. exists m0 w3 S0 e3 Lv. eauto.
Qed.



End WithTypingMode.
