Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes InertContext PreciseTyping
        OperationalSemantics Substitution Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
         Subenvironments Narrowing
        EnvCorrespondence EWeakening
        ConfigTyping ConfigVarkinds TFree Deref
        TypedOperationalSemantics TypedReductionEquivalence
        TypedProgress TypedPreservation TypedSoundness.

(** * Preservation *)
Section WithTypingMode.
Context (typing_mode: Mode).

(** ** Preservation Theorem *)

(** [⊢ (s, t)∶ T]           #<br>#
    [(s, t) |-> (s', t')]   #<br>#
    [―――――――――――――――――――]   #<br>#
    [⊢ (s', t')∶ T]         *)
Theorem preservation : forall t s Sigma e Lv t' s' Sigma' e' Lv' T,
    ⊢ (t; s; Sigma; e; Lv) ∶ T ->
    (t; s; Sigma; e; Lv) ↦ (t'; s'; Sigma'; e'; Lv') ->
    ⊢ (t'; s'; Sigma'; e'; Lv') ∶ T.
Proof.
  introv Htyc Hred. inversions Htyc. rename H into Htyc.
  eapply untyped_red_to_typed in Hred; eauto.
  destruct Hred as [Gamma' Hred].
  eapply typed_preservation in Hred. eauto.
Qed.

Corollary preservation_closure : forall c T n c',
    ⊢ c ∶ T ->
    c ⊣n↦* c' ->
    ⊢ c' ∶ T.
Proof.
  introv Htyc Hred. reverts c'. induction n; intros.
  - inversions Hred; easy.
  - inversions Hred. autospecialize.
    destruct c. destruct C2. destruct c'.
    forwards: preservation; eauto.
Qed.

(** * Progress *)

(** [⊢ (s, t)∶ T]           #<br>#
    [(s, t) |-> (s', t')]   #<br>#
    [―――――――――――――――――――]   #<br>#
    [t] is in normal form   #<br>#
    or [exists s', t'] such that [(s, t) |-> (s', t')] *)
Theorem progress: forall t s Sigma e Lv T,
    ⊢ (t; s; Sigma; e; Lv) ∶ T ->
    answer (t; s; Sigma; e; Lv) \/
    exists t' s' Sigma' e' Lv', (t; s; Sigma; e; Lv) ↦ (t'; s'; Sigma'; e'; Lv').
Proof.
  introv Htyc. inversions Htyc. rename H into Htyc.
  eapply typed_progress in Htyc as [? | Hstep]; eauto. right.
  destruct Hstep as [Gamma' [[t' s' Sigma' e'] Hstep]].
  exists t' s' Sigma' e'. eapply typed_red_to_untyped in Hstep; eauto.
Qed.

(** * Final Soundness Theorem *)
(** Version of the soundness theorem starting from any configuration. *)
Theorem soundness : forall (F : ctx) (C : config) (T : typ),
    (* If a configuration is well-typed *)
    ty_config F C T ->
    (* It either converges to an answer using untyped reduction *)
    (exists (n : nat) (Ca : config),
        answer Ca /\ (C ⊣n↦* Ca)) \/
    (* Or diverges using untyped reduction *)
    (forall (n : nat), exists (Cn : config),
          C ⊣n↦* Cn).
Proof.
  introv Hty.
  destruct (typed_red_soundness Hty) as [Hconv | Hdiv].
  - left. destruct_all; eauto using typed_to_untyped_red_closure.
  - right. intros. specialize (Hdiv n).
    destruct_all; eauto using typed_to_untyped_red_closure.
Qed.

Corollary terminates_or_not G c T :
  ty_config G c T ->
  (exists c', c ↦ c') \/ answer c.
Proof.
  introv Hty. forwards: soundness; eauto.
  destruct H as [Hterm | Hloop].
  - destructc Hterm as n Ca Hans Hred. destruct n.
    + inversions Hred. eauto.
    + left. clear Hans. reverts Ca. induction n; intros.
      * inversions Hred. inversions H0. eauto.
      * inversions Hred. eauto.
  - left. specialize (Hloop 1). destructc Hloop as c' Hred.
    inversions Hred. inversions H0. eauto.
Qed.

(** Version of the soundness theorem starting from an initial configuration (empty stack and heap). *)
(* Theorem 8 in "Reference mutability for DOT" *)
(* Theorem 1 in "Pure methods for roDOT" *)
Theorem soundness_initial (t : trm) (T : typ):
    (* If a term is well-typed *)
    closed t ->
    closed T ->
    ctx_empty⁏empty ⊢ t ∶ T ->
    exists C, (config_initial t C) /\ (
      (* It either converges to an answer using untyped reduction *)
      (exists (n : nat) (Ca : config),
          answer Ca /\ (C ⊣n↦* Ca)) \/
      (* Or diverges using untyped reduction *)
      (forall (n : nat), exists (Cn : config),
            C ⊣n↦* Cn)
    ).
Proof.
  introv Hclt HclT Hty.
  forwards C ? ?: ty_config_initial; eauto.
  exists C.
  split; eauto.
  eapply soundness; eauto.
Qed.

End WithTypingMode.
