Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax GeneralTyping Generalization
        Similarity SyntaxSimilarity OperationalSemantics HeapUpdate
        Deref Defs ConfigValid.
Require Import HeapCorrespondence EnvCorrespondence.
Require Import ConfigTyping ConfigVarkinds TFree.
Require Import Safety.
Require Import ReductionFreeVars.

(** * Similarity of machine configurations *)

(* Definition 46 in "Pure methods for roDOT" *)
(** In similar_conf_sef ⟨t₁,σ₁,Σ₁,ρ₁⟩ ⟨t₂,σ₂,Σ₂,ρ₂⟩ Lt₁ Lt₂ LS₁ LS₂,
    the Ltᵢ serves to identify variables that appear within a term
    while the LSᵢ serves to identify locations. And so this means:
    - The holes t₁ and t₂ must be similar up to identification Lt₁ = Lt₂.
    - We require the heap items of Σᵢ at LSᵢ[n] to be similar up to Lt₁ = Lt₂.
      In other words, Σ₁(LS₁[n]) and Σ₂(LS₂[n]) are similar up to Lt₁ = Lt₂.
*)
Definition similar_conf_sef
  (c1 c2 : config) (Lt1 Lt2 LS1 LS2 : list var) : Prop :=
  match (c1, c2) with
  | ((t1 ; s1 ; Sigma1 ; e1 ; Lv1), (t2 ; s2 ; Sigma2 ; e2 ; Lv2)) =>
      length Lt1 = length Lt2
      /\ length LS1 = length LS2
      (* Note we do not require uniqueness of the lists used to
         identify variables in terms *)
      /\ vars_conf Lt1 c1 /\ vars_conf Lt2 c2
      (* But we _do_ require uniqueness of the lists used to identify heap
         locations, in order to ensure no accidental heap update happens
         due to duplicate identifications in the context of
         conf_sim_red_append *)
      /\ vars_conf_unique LS1 c1 /\ vars_conf_unique LS2 c2
      /\ (similar Lt1 Lt2 t1 t2)
      /\ (similar Lt1 Lt2 s1 s2)
      /\ (similar_env_with similar_item Lt1 Lt2 LS1 LS2 Sigma1 Sigma2)
      (* When two references are identified then their corresponding
         locations must also be identified. *)
      /\ (forall w1 w2 y1 y2, similar Lt1 Lt2 w1 w2 ->
                        binds w1 y1 e1 ->
                        binds w2 y2 e2 ->
                        similar Lt1 Lt2 y1 y2)
      (* When two heap locations are identified in terms then they
         must also be identified in the heap *)
      /\ (forall y1 y2, (y1 \in dom Sigma1) ->
                  (y2 \in dom Sigma2) ->
                  similar Lt1 Lt2 y1 y2 ->
                  similar LS1 LS2 y1 y2)
  end.

Lemma similar_conf_sef_heap_part c1 c2 Lt1 Lt2 LS1 LS2 :
  similar_conf_sef c1 c2 Lt1 Lt2 LS1 LS2 ->
  similar_with Lt1 Lt2 LS1 LS2 (config_heap c1) (config_heap c2).
Proof. introv Hsim. destruct c1. destruct c2. simpl. hnf in Hsim. tauto. Qed.

(* Lemma 20 in "Pure methods for roDOT" *)
Lemma ref_elim_similar_conf_sef t Sigma e fs Lt w1 w2 y1 y2 :
  config_valid (t ; stack_nil ; Sigma ; e ; fs) ->
  binds w1 y1 e -> binds w2 y2 e ->
  (fv_config (t ; stack_nil ; Sigma ; e ; fs)) = from_list Lt ->
  similar_conf_sef
    (t ; stack_nil ;                                     Sigma   ; e ; fs)
    (t ; stack_nil ; (subst_heap w2 y2 (subst_heap w1 y1 Sigma)) ; e ; fs)
    (w2 :: w1 :: Lt)%list    (y2 :: y1 :: Lt)%list
    (keys Sigma)           (keys Sigma).
Proof.
  introv Hvalid Hbi1 Hbi2 HLt.
  inversions Hvalid.
  rename H3 into HokS, H5 into Hoke, H6 into Hd, H7 into Hloc.
  eapply renv_bind_in_fv in Hbi1 as Hiny1;
    eapply renv_bind_in_fv in Hbi2 as Hiny2.
  eapply binds_to_dom in Hbi1 as Hinw1;
    eapply binds_to_dom in Hbi2 as Hinw2.

  repeat(split); simpl in *; eauto.
  9: eapply similar_env_with_equiv; splits; eauto.
  all: repeat(rewrite union_empty_l in *);
       repeat(rewrite from_list_cons in *).

  3,5: eapply ok_fresh_keys; eauto.

  6: eapply similar_stack_nil; eauto.

  { rewrite union_assoc; eapply subset_union_l;
    try solve[rewrite HLt; eauto using subset_refl].
    introv Hin; rewrite in_union, ! in_singleton in Hin.
    destruct Hin as [? | ?]; subst; rewrite ! in_union; tauto. }

  { rewrite <- HLt. introv Hin.
    repeat(rewrite in_union in *); repeat(rewrite in_singleton in *).
    unfold subst_heap. repeat(rewrite dom_map in *).
    destruct_all; subst; try tauto.
    destruct (classicT (x=w2)) as [Heq2 | Hne2]; subst; try tauto.
    destruct (classicT (x=w1)) as [Heq1 | Hne1]; subst; try tauto.
    right. left. right. eapply in_fv_env_subst; eauto.
    eapply in_fv_env_subst; eauto. }

  { rewrite env_from_list_keys_dom. introv Hin.
    rewrite ! in_union; tauto. }

  { rewrite env_from_list_keys_dom. introv Hin.
    repeat(rewrite in_union in *); repeat(rewrite in_singleton in *).
    unfold subst_heap. repeat(rewrite dom_map in *).
    destruct_all; subst; try tauto. }

  { repeat(eapply similar_extend). eapply similar_refl_by_fv.
    rewrite <- HLt. eauto using subset_union_weak_l. }

  { introv Hsim. unfold subst_heap. repeat(rewrite dom_map in *).
    destructc Hsim as n Hin1 Hin2.
    eapply (LibList.Nth_functional Hin1) in Hin2; subst. firstorder. }

  { introv Hsim HbiS1 HbiS2. destructc Hsim as n Hin1 Hin2.
    eapply (LibList.Nth_functional Hin1) in Hin2; subst.
    rename x2 into x.
    assert (Hitin : fv it1 \c fv_env Sigma).
    { eapply env_binds_fv_subset in HbiS1. easy. }
    unfold subst_heap in HbiS2.
    eapply binds_map with (f := subst_var w1 y1),
        binds_map with (f := subst_var w2 y2) in HbiS1; binds_eq; clear H.
    destruct (classicT (w1=w2)) as [Heq | Hne]; subst.

    - binds_eq. rewrite subst_twice_same.
      erewrite <- subst_vars_nil_r with (xs:=nil) (a:=it1),
          subst_vars_push_left, subst_vars_nil_r.
      eapply similar_one_side_subst_item; simpl; eauto.
      + split; eauto. rewrite from_list_cons, from_list_nil, union_empty_r,
          notin_singleton. hnf; intros; subst.
        erewrite <- in_empty, <- Hd, in_inter; split; eauto.
      + introv Hin. do 2 (eapply similar_extend).
        eapply similar_refl_by_fv.
        rewrite <- HLt. unfold fv at 1, VarFreeVar.
        intros v0 Hin0. rewrite in_singleton in Hin0; subst.
        eapply Hitin in Hin. repeat(rewrite in_union). tauto.
      + introv Hnx0 Hny. destruct n0.
        2: { false. inversions Hny. inversions H4. }
        inversions Hny. inversions Hnx0.
        eapply similar_var_push; eauto.

    - erewrite <- subst_vars_nil_r with (xs:=nil) (a:=it1),
          subst_vars_push_left, subst_vars_push_left, subst_vars_nil_r.
      eapply similar_one_side_subst_item; simpl; eauto.
      + repeat(split); eauto.
        * rewrite ! from_list_cons, ! from_list_nil, union_empty_r,
            notin_union, ! notin_singleton. split; hnf; intros; subst.
          all: erewrite <- in_empty, <- Hd, in_inter; split; eauto.
        * rewrite ! from_list_cons, ! from_list_nil, union_empty_r,
            ! notin_union, ! notin_singleton. repeat(split); eauto;
          hnf; intros; subst.
          all: erewrite <- in_empty, <- Hd, in_inter; split.
          3: exact Hinw1. 1: exact Hinw1. all: eauto.
      + introv Hin. do 2 (eapply similar_extend).
        eapply similar_refl_by_fv.
        rewrite <- HLt. unfold fv at 1, VarFreeVar.
        intros v0 Hin0. rewrite in_singleton in Hin0; subst.
        eapply Hitin in Hin. repeat(rewrite in_union). tauto.
      + introv Hnx0 Hny. destruct n0.
        2: inversions Hny; inversions Hnx0; destruct n0.
        3: { inversions H3. inversions H5. }
        * inversions Hny; inversions Hnx0.
          eapply similar_var_push; eauto.
        * inversions H3; inversions H4.
          eapply similar_extend,similar_var_push; eauto. }

  { introv Hsim Hbi0 Hbi3.
    eapply similar_var_simlist_relates, simlist_relates_cons_cases in Hsim
        as [Hsim | [? ?]]; subst.
    2: { false. erewrite <- in_empty, <- Hd, in_inter; split.
         - eapply binds_to_dom in Hbi3; eauto.
         - eapply Hloc; eauto. }
    eapply simlist_relates_cons_cases in Hsim
        as [Hsim | [? ?]]; subst.
    2: { false. erewrite <- in_empty, <- Hd, in_inter; split.
         - eapply binds_to_dom in Hbi3; eauto.
         - eapply Hloc; eauto. }
    do 2 (eapply similar_extend).
    destructc Hsim as n Hin0 Hin3.
    eapply (LibList.Nth_functional Hin0) in Hin3; subst. binds_eq; clear H.
    eapply renv_bind_in_fv in Hbi0. eapply simlist_relates_similar_var;
      eauto.
    forwards: LibList.mem_Nth.
    2: { destructc H as n0 Hnth. exists n0. split; exact Hnth. }
    eapply list_in_fromlist_mem. rewrite <- HLt.
    repeat(rewrite in_union). tauto. }

  { unfold subst_heap. repeat(rewrite dom_map in *).
    introv Hbi0 Hbi3 Hsim.
    eapply similar_var_simlist_relates, simlist_relates_cons_cases in Hsim
        as [Hsim | [? ?]]; subst.
    2: { false. erewrite <- in_empty, <- Hd, in_inter; split; eauto. }
    eapply simlist_relates_cons_cases in Hsim
        as [Hsim | [? ?]]; subst.
    2: { false. erewrite <- in_empty, <- Hd, in_inter; split.
         exact Hinw1. exact Hbi0. }
    destructc Hsim as n Hin0 Hin3.
    eapply (LibList.Nth_functional Hin0) in Hin3; subst.
    eapply simlist_relates_similar_var; eauto.
    forwards: LibList.mem_Nth.
    2: { destructc H as n0 Hnth. exists n0. split; exact Hnth. }
    eapply list_in_fromlist_mem. rewrite env_from_list_keys_dom. easy. }

Qed.

Ltac destruct_sef_sim :=
  match goal with
    [ Hsim : similar_conf_sef (_ ; _ ; _ ; _ ; _) (_ ; _ ; _ ; _ ; _)
               _ _ _ _ |- _ ] =>
      destruct Hsim as [Htlen [HSlen [HvcLt1 [HvcLt2 [HvcLS1 [HvcLS2 [Hsimtrm [Hsimstk [Hsimheap [Hsimenv Hsimcor]]]]]]]]]];
      rename HvcLt1 into HinCLt1;
      rename HvcLt2 into HinCLt2;
      destruct HvcLS1 as [HfrLS1 HinCLS1];
      destruct HvcLS2 as [HfrLS2 HinCLS2]
    end.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma conf_sim_red_append_single Lt1 Lt2 LS1 LS2 c1 c2 c1' c2' G1 G2 U1 U2 :
  ty_config G1 c1 U1 -> ty_config G2 c2 U2 ->
  similar_conf_sef c1 c2 Lt1 Lt2 LS1 LS2 ->
  c1 ↦ c1' -> c2 ↦ c2' ->
   exists LS1' LS2',
     (similar_conf_sef
        c1' c2'
        (LS1' ++ Lt1) (LS2' ++ Lt2)
        (LS1' ++ LS1) (LS2' ++ LS2))
     /\ (disjoint (from_list LS1') (fv_config c1))
     /\ (disjoint (from_list LS2') (fv_config c2)).
Proof.
  introv Hty1 Hty2 Hsim Hred1 Hred2.
  destruct Hred1.

  { Case "red_read". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H7.
    rename x0 into wObj2, wObj into wObj1,
      yObj into yObj1, yVal into yVal1, wVal into wVal1,
      yObj0 into yObj2, yVal0 into yVal2, wVal0 into wVal2,
      e0 into e2, e into e1, fs into fs1, fs0 into fs2,
      Lv0 into Lv2, Lv into Lv1, m into a,
      Sigma into Sigma1, Sigma0 into Sigma2,
      ds into ds1, ds0 into ds2,
      H9 into Hsimw, H into Hfrw1, H0 into Hbiw1, H1 into Hbiy1,
      H4 into Hbiw2, H3 into Hfrw2, H5 into Hbiy2,
      H2 into Hds1, H6 into Hds2.
    exists (wVal1 :: nil)%list (wVal2 :: nil)%list.

    repeat(split); simpl;
      eauto using
        similar_trm_var, similar_avar_f, similar_var_push, similar_extend.

    3: rename HinCLS1 into HinCL. 5: rename HinCLS2 into HinCL.
    3: rename Hfrw1 into Hfrw. 5: rename Hfrw2 into Hfrw.
    3,5: rewrite union_empty_l; forwards: fresh_cons.
    4,6: destruct H as [_ Hfr]; rewrite union_empty_l in Hfr; exact Hfr.
    3,4: eapply fresh_push; eauto.
    3,4: rewrite notin_union; split; eauto using notin_empty.
    3,4: introv Hwin; specialize (HinCL _ Hwin).
    3,4: hnf in Hfrw; destruct Hfrw as [Hfrw _]; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: rewrite from_list_cons; introv Hin.
    1,2,3,4: repeat(rewrite in_union in *);
             repeat(rewrite in_singleton in *).
    1,2,3,4: destruct Hin as [? | Hin]; try tauto.
    1,2,3,4: specialize (HinCL _ Hin); simpl in *.
    1,2,3,4: repeat(rewrite in_union in *);
             repeat(rewrite in_singleton in *); tauto.

    4: rename Hfrw1 into Hfrw. 5: rename Hfrw2 into Hfrw.
    4,5: rewrite from_list_cons, from_list_nil, union_empty_r.
    4,5: eapply disjoint_singleton.
    4,5: simpl in Hfrw; tauto.

    1: eapply similar_env_with_oi_push_notin; eauto.

    - (* w₁ and w₂ are similar so either wᵢ = wValᵢ, in which case yᵢ = yValᵢ.
         By Hdsᵢ and Hbiyᵢ and Hsimcor, it follows yVal₁ = yVal₂ assuming
         Lt₁ = Lt₂. Otherwise, wᵢ ≠ wValᵢ and so (wᵢ~yᵢ) ∈ eᵢ and
         w₁ = w₂ assuming Lt₁ = Lt₂. By Hsimenv and weakening
         the result follows. *)
      introv Hsim Hbi1 Hbi2.
      eapply similar_var_simlist_relates in Hsim.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      2: { eapply binds_push_inv in Hbi1 as [[_ ?] | [Hc _]];
           try congruence; subst.
           eapply binds_push_inv in Hbi2 as [[_ ?] | [Hc _]];
             try congruence; subst.
           eapply similar_extend.
           specialize (Hsimcor _ _ (binds_to_dom Hbiy1) (binds_to_dom Hbiy2)).
           forwards: Hsimcor.
           {  eapply Hsimenv; eauto. }
           rename H into Hsimyo.
           eapply similar_env_with_binds in Hsimheap.
           2: { eapply similar_var_simlist_relates. exact Hsimyo. }
           all: eauto. similar_syntax_inverts.
           eapply similar_defs_has in Hsimheap; try eassumption; simpl; eauto.
           similar_syntax_inverts. inversions H; eauto. }
      eapply binds_push_inv in Hbi1 as [[? ?] | [_ Hbi1]]; subst.
      1: { false. eapply simlist_relates_in_list_l in Hsim.
           simpl in Hfrw1; simpl in HinCLt1.
           destructc Hfrw1 as Hfrw1 _. firstorder. }
      eapply binds_push_inv in Hbi2 as [[? ?] | [_ Hbi2]]; subst.
      1: { false. eapply simlist_relates_in_list_r in Hsim.
           simpl in Hfrw2; simpl in HinCLt2.
           destructc Hfrw2 as Hfrw2 _. firstorder. }
      eapply similar_extend. eapply Hsimenv.
      2: exact Hbi1. 2: exact Hbi2.
      eauto using simlist_relates_similar_var.

    - (* Weakening using Hsimcor. yᵢ ≠ wValᵢ because they are in the domain
         of the heap. *)
      introv Hiny1 Hiny2 Hsimy. eapply similar_var_simlist_relates in Hsimy.
      specialize (Hsimcor _ _ Hiny1 Hiny2).
      eapply simlist_relates_cons_cases in Hsimy as
          [Hsimy | [? ?]]; subst; eauto using similar_var_push.
      eapply similar_extend.
      eapply simlist_relates_similar_var in Hsimy; eauto. }

  { Case "red_write". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H5.
    rename wArg into wArg1, x1 into wArg2, fs into fs1, fs0 into fs2,
      Sigma' into Sigma'1, Sigma'0 into Sigma'2, wObj into wObj1, m into a,
      e into e1, e0 into e2,
      Lv into Lv1, Sigma into Sigma1, Sigma0 into Sigma2,
      yObj into yObj1, yArg into yArg1, yObj0 into yObj2, yArg0 into yArg2,
      x0 into wObj2, H into Hupd1, H0 into Hbiwo1, H1 into Hbiwa1,
      H2 into Hupd2, H3 into Hbiwo2, H4 into Hbiwa2,
      H7 into Hsimwo, H8 into Hsimwa.
    exists (@nil var) (@nil var).
    repeat(split); simpl in *; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: introv Hin; specialize (HinCL _ Hin); simpl in HinCL;
    repeat(rewrite in_union in *); tauto.

    4,5: rewrite from_list_nil; eauto using disjoint_empty_l.

    1: eauto using similar_trm_var, similar_avar_f.

    - (* By Hupdᵢ we know that Σᵢ' is the same as Σᵢ except
         Σᵢ(yObjᵢ).a = yArgᵢ. To show this subgoal note that by
         Hsimcor and HsimyObj we have similar LSᵢ yObjᵢ.
         Thus LSᵢ = LSᵢ,₁.yObjᵢ.LSᵢ,₂. So for a pair (y₁,y₂) ∈ LS₁ × LS₂,
         either both yᵢ = yObjᵢ or both yᵢ ≠ yObjᵢ. In the first case by heap
         update Hupdᵢ, similarity follows because of HsimyArg. In the
         second case again by heap update Hudpᵢ, similarity follows because of
         Hsimheap. *)
      specialize (Hsimenv _ _ _ _
                          Hsimwo Hbiwo1 Hbiwo2) as HsimyObj.
      forwards: Hsimcor. 3: exact HsimyObj.
      1: rename Hty1 into Hty, Hbiwo1 into Hbiwo.
      2: rename Hty2 into Hty, Hbiwo2 into Hbiwo.
      1,2: eapply renv_bind_in_fv in Hbiwo.
      1,2: forwards: heap_correspond_env_fv_heap_dom;
      try solve[inversions Hty; eassumption]; eauto.
      rename H into HsimyArgS.

      specialize (Hsimenv _ _ _ _ Hsimwa Hbiwa1 Hbiwa2) as HsimyArg.

      eapply heap_update_similar; eauto.

    - destruct Hupd1 as [_ [_ [HdomS1 _]]]; rewrite <- HdomS1.
      destruct Hupd2 as [_ [_ [HdomS2 _]]]; rewrite <- HdomS2. easy. }

  { Case "red_call". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H5.
    rename fs into fs1, fs0 into fs2, wObj into wObj1, e into e1, e0 into e2,
      Lv into Lv1, Sigma into Sigma1, Sigma0 into Sigma2, Lv0 into Lv2,
      yObj into yObj1, yObj0 into yObj2,
      wArg into wArg1, x0 into wObj2, x1 into wArg2, ds into ds1,
      ds0 into ds2, t into t1, t0 into t2, H into Hbiy1, H0 into Hhs1,
      H1 into Hbiwy1, H2 into Hbiy2, H3 into Hhs2, H4 into Hbiwy2,
      H7 into Hsimwo, H8 into Hsimwa.
    exists (@nil var) (@nil var). repeat(split); simpl; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: introv Hin; specialize (HinCL _ Hin); simpl in HinCL;
    repeat(rewrite in_union in *); tauto.

    2,3: rewrite from_list_nil; eauto using disjoint_empty_l.

    eapply similar_open_rec. eapply similar_open_rec.
    all: try eauto.

    (* By Hsimyo, there is a common index n where LSᵢ[n] = yObjᵢ. And so
       by Hsimheap along with Hbiyᵢ, we have that
       item_obj ds₁ = item_obj ds₂ assuming Lt₁ = Lt₂. By
       Hsimwo and Hsimwa, the result follows. *)
    eapply similar_env_with_equiv in Hsimheap.
    destructc Hsimheap as _ _ _ Hsimheap.
    forwards Hsimitm: Hsimheap; fweauto.
    { eapply similar_var_simlist_relates, Hsimcor; eauto using binds_to_dom. }
    forwards Hsimdef: similar_defs_has.
    3: exact Hhs1. 3: exact Hhs2.
    all: simpl; eauto. similar_syntax_inverts; eauto.
    similar_syntax_inverts; eauto. inversions H. eauto. }

  { Case "red_apply". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H3.
    rename
      fs into fs1, fs0 into fs2, wFun into wFun1, e into e1, e0 into e2,
      Lv into Lv1, Sigma into Sigma1, Sigma0 into Sigma2, Lv0 into Lv2,
      yFun into yFun1, yFun0 into yFun2,
      wArg into wArg1, x0 into wFun2, x1 into wArg2,
      t into t1, t0 into t2, H into Hbiy1, H0 into Hbiwy1, H1 into Hbiy2,
      H2 into Hbiwy2, H5 into Hsimwf, H6 into Hsimwa.
    exists (@nil var) (@nil var). repeat(split); simpl; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: introv Hin; specialize (HinCL _ Hin); simpl in HinCL;
    repeat(rewrite in_union in *); tauto.

    2,3: rewrite from_list_nil; eauto using disjoint_empty_l.

    eapply similar_open_rec; eauto.
    eapply similar_env_with_binds in Hsimheap; try eassumption.
    similar_syntax_inverts; eauto.
    eapply similar_var_simlist_relates, Hsimcor; eauto using binds_to_dom. }

  { Case "red_let_fun". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H3.
    rename
      fs into fs1, fs0 into fs2, Lv into Lv1, Sigma into Sigma1,
      Sigma0 into Sigma2, Lv0 into Lv2, e into e1, e0 into e2,
      t2 into u1, w into w1, w0 into w2, y into y1, y0 into y2,
      x1 into T2, T into T1, x0 into u2, x2 into t2,
      H6 into HsimT, H7 into Hsimt, H5 into Hsimu,
      H0 into Hnwy1, H2 into Hnwy2, H into Hfrwy1, H1 into Hfrwy2.
    exists (y1 :: w1 :: nil)%list (y2 :: w2 :: nil)%list.

    assert (Hfrye1 : y1 # e1).
    { simpl in Hfrwy1; repeat(rewrite notin_union in Hfrwy1); tauto. }
    assert (Hfrye2 : y2 # e2).
    { simpl in Hfrwy2; repeat(rewrite notin_union in Hfrwy2); tauto. }

    assert (Hsimenvp : forall w0 w3 y0 y3 : var,
               similar (y1 :: w1 :: Lt1) (y2 :: w2 :: Lt2) w0 w3 ->
               binds w0 y0 (e1 & w1 ~ y1) ->
               binds w3 y3 (e2 & w2 ~ y2) ->
               similar (y1 :: w1 :: Lt1) (y2 :: w2 :: Lt2) y0 y3).
    { introv Hsim Hbi1 Hbi2.
      eapply similar_var_simlist_relates in Hsim.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      2: { eapply binds_push_inv in Hbi1 as [[_ ?] | [Hc _]];
           try congruence; subst.
           eapply binds_push_inv in Hbi2 as [[_ ?] | [Hc _]];
             try congruence; subst.
           eapply similar_var_push; simpl; omega. }
      2: { false. eapply binds_to_dom in Hbi1.
           rewrite dom_push, in_union, in_singleton in Hbi1.
           destruct_all; try congruence. }
      eapply binds_push_inv in Hbi1 as [[? ?] | [_ Hbi1]]; subst.
      1: { false. eapply simlist_relates_in_list_l in Hsim.
           simpl in Hfrwy1; simpl in HinCLt1.
           destructc Hfrwy1 as _ Hfrwy1 _.
           rewrite notin_union in Hfrwy1;
             destructc Hfrwy1 as Hfrwy1 _. firstorder. }
      eapply binds_push_inv in Hbi2 as [[? ?] | [_ Hbi2]]; subst.
      1: { false. eapply simlist_relates_in_list_r in Hsim.
           simpl in Hfrwy2; simpl in HinCLt2.
           destructc Hfrwy2 as _ Hfrwy2 _.
           rewrite notin_union in Hfrwy2;
             destructc Hfrwy2 as Hfrwy2 _. firstorder. }
      do 2 (eapply similar_extend); eapply Hsimenv;
      eauto using simlist_relates_similar_var. }

    repeat(split); simpl;
      eauto using similar_trm_var, similar_avar_f, similar_var_push,
      similar_extend.

    3: rename HinCLS1 into HinCL. 5: rename HinCLS2 into HinCL.
    3: rename Hfrwy1 into Hfrwy. 5: rename Hfrwy2 into Hfrwy.
    3,5: rewrite union_empty_l; forwards: fresh_cons.
    4,6: destruct H as [_ Hfr]; exact Hfr.
    3,4: eapply fresh_push.
    4,6: rewrite notin_union; split; eauto using notin_empty.
    4,5: introv Hwin; specialize (HinCL _ Hwin); simpl in HinCL.
    4,5: simpl in Hfrwy.
    4,5: destruct Hfrwy as [_ [Hfrwy _]]; eauto.
    4,5: rewrite notin_union in Hfrwy; destruct Hfrwy as [Hfrwy _]; easy.
    3,4: forwards: fresh_cons.
    4,6: destruct H as [_ Hfr]; rewrite union_empty_l in Hfr; exact Hfr.
    3,4: eapply fresh_push; eauto; rewrite union_empty_l.
    3,4: simpl in Hfrwy.
    3,4: destruct Hfrwy as [Hfrwy _]; eauto.
    3,4: introv Hwin; specialize (HinCL _ Hwin); eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: repeat(rewrite from_list_cons); introv Hin.
    1,2,3,4: repeat(rewrite in_union in Hin);
    repeat(rewrite dom_push);
    repeat(rewrite in_union).
    1,2,3,4: destruct Hin as [? | [? | Hin]]; try tauto.
    1,2,3,4: specialize (HinCL _ Hin); simpl in HinCL.
    1,2,3,4: repeat(rewrite in_union in *); tauto.

    4: rename Hfrwy1 into Hfrwy. 5: rename Hfrwy2 into Hfrwy.
    4,5: repeat(rewrite from_list_cons);
    rewrite from_list_nil, union_empty_r.
    4,5: eapply disjoint_union_l; eapply disjoint_singleton.
    4,5,6,7: simpl in Hfrwy; try tauto.
    4,5: destruct Hfrwy as [_ [Hfrwy _]]; eauto.

    - eapply similar_extend, similar_open_rec; eauto using similar_var_push.
      eapply similar_extend; eauto.

    - eapply similar_env_with_push_notin.
      1: eapply similar_env_with_o_push, similar_with_i_push_notin; eauto;
      unfold envy_dom; simpl.

      1,4: rename Hfrwy1 into Hfrwy. 3,5: rename Hfrwy2 into Hfrwy.
      1,3: simpl in Hfrwy. 1,2: repeat(rewrite notin_union in Hfrwy); tauto.

      1: rename HinCLS1 into HinCL. 2: rename HinCLS2 into HinCL.
      1,2: simpl in Hfrwy; rewrite from_list_cons, notin_union,
        notin_singleton; split; eauto; introv Hin.
      1,2: specialize (HinCL _ Hin); simpl in Hfrwy, HinCL.
      1,2: destruct Hfrwy as [Hfrwy _]; eauto.

      do 2 (eapply similar_extend). eapply similar_item_fun; eauto.

    - repeat(rewrite dom_push). intros x1 x2 Hin1 Hin2 Hsim.
      eapply similar_var_simlist_relates, simlist_relates_cons_cases
        in Hsim as [Hsim | [? ?]]; subst.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      3: { eapply similar_var_push; simpl; omega. }
      2: { eapply similar_extend; eapply similar_var_push; simpl; omega. }
      do 2 (eapply similar_extend). eapply Hsimcor;
        eauto using simlist_relates_similar_var.
      1: rename Hin1 into Hin. 2: rename Hin2 into Hin.
      1: rename Hfrwy1 into Hfrwy. 2: rename Hfrwy2 into Hfrwy.
      1: rename HinCLt1 into HinCLt. 2: rename HinCLt2 into HinCLt.
      1,2: rewrite in_union, in_singleton in Hin; destruct Hin as [? | ?];
      subst; eauto.
      1,2: false; simpl in Hfrwy; simpl in HinCLt.
      eapply simlist_relates_in_list_l in Hsim; firstorder.
      eapply simlist_relates_in_list_r in Hsim; firstorder. }

  { Case "red_let_obj". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H3.
    rename fs into fs1, fs0 into fs2, Lv into Lv1, Sigma into Sigma1,
      Sigma0 into Sigma2, Lv0 into Lv2, e into e1, e0 into e2,
      t into t1, x0 into t2, w into w1, w0 into w2, y into y1, y0 into y2,
      ds into ds1, x2 into ds2, x1 into T2, T into T1,
      H6 into HsimT, H7 into Hsimds, H5 into Hsimt,
      H0 into Hnwy1, H2 into Hnwy2, H into Hfrwy1, H1 into Hfrwy2.
    exists (y1 :: w1 :: nil)%list (y2 :: w2 :: nil)%list.

    assert (Hfrye1 : y1 # e1).
    { simpl in Hfrwy1; repeat(rewrite notin_union in Hfrwy1); tauto. }
    assert (Hfrye2 : y2 # e2).
    { simpl in Hfrwy2; repeat(rewrite notin_union in Hfrwy2); tauto. }

    assert (Hsimenvp : forall w0 w3 y0 y3 : var,
               similar (y1 :: w1 :: Lt1) (y2 :: w2 :: Lt2) w0 w3 ->
               binds w0 y0 (e1 & w1 ~ y1) ->
               binds w3 y3 (e2 & w2 ~ y2) ->
               similar (y1 :: w1 :: Lt1) (y2 :: w2 :: Lt2) y0 y3).
    { introv Hsim Hbi1 Hbi2.
      eapply similar_var_simlist_relates in Hsim.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      2: { eapply binds_push_inv in Hbi1 as [[_ ?] | [Hc _]];
           try congruence; subst.
           eapply binds_push_inv in Hbi2 as [[_ ?] | [Hc _]];
             try congruence; subst.
           eapply similar_var_push; simpl; omega. }
      2: { false. eapply binds_to_dom in Hbi1.
           rewrite dom_push, in_union, in_singleton in Hbi1.
           destruct_all; try congruence. }
      eapply binds_push_inv in Hbi1 as [[? ?] | [_ Hbi1]]; subst.
      1: { false. eapply simlist_relates_in_list_l in Hsim.
           simpl in Hfrwy1; simpl in HinCLt1.
           destructc Hfrwy1 as _ Hfrwy1 _.
           rewrite notin_union in Hfrwy1;
             destructc Hfrwy1 as Hfrwy1 _. firstorder. }
      eapply binds_push_inv in Hbi2 as [[? ?] | [_ Hbi2]]; subst.
      1: { false. eapply simlist_relates_in_list_r in Hsim.
           simpl in Hfrwy2; simpl in HinCLt2.
           destructc Hfrwy2 as _ Hfrwy2 _.
           rewrite notin_union in Hfrwy2;
             destructc Hfrwy2 as Hfrwy2 _. firstorder. }
      do 2 (eapply similar_extend); eapply Hsimenv;
      eauto using simlist_relates_similar_var. }

    repeat(split); simpl; eauto using
        similar_trm_var, similar_avar_f, similar_var_push, similar_extend.

    3: rename HinCLS1 into HinCL. 5: rename HinCLS2 into HinCL.
    3: rename Hfrwy1 into Hfrwy. 5: rename Hfrwy2 into Hfrwy.
    3,5: rewrite union_empty_l; forwards: fresh_cons.
    4,6: destruct H as [_ Hfr]; exact Hfr.
    3,4: eapply fresh_push.
    4,6: rewrite notin_union; split; eauto using notin_empty.
    4,5: introv Hwin; specialize (HinCL _ Hwin); simpl in HinCL.
    4,5: simpl in Hfrwy.
    4,5: destruct Hfrwy as [_ [Hfrwy _]]; eauto.
    4,5: rewrite notin_union in Hfrwy; destruct Hfrwy as [Hfrwy _]; easy.
    3,4: forwards: fresh_cons.
    4,6: destruct H as [_ Hfr]; rewrite union_empty_l in Hfr; exact Hfr.
    3,4: eapply fresh_push; eauto; rewrite union_empty_l.
    3,4: simpl in Hfrwy.
    3,4: destruct Hfrwy as [Hfrwy _]; eauto.
    3,4: introv Hwin; specialize (HinCL _ Hwin); eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: repeat(rewrite from_list_cons); introv Hin.
    1,2,3,4: repeat(rewrite in_union in Hin);
    repeat(rewrite dom_push);
    repeat(rewrite in_union).
    1,2,3,4: destruct Hin as [? | [? | Hin]]; try tauto.
    1,2,3,4: specialize (HinCL _ Hin); simpl in HinCL.
    1,2,3,4: repeat(rewrite in_union in *); tauto.

    4: rename Hfrwy1 into Hfrwy. 5: rename Hfrwy2 into Hfrwy.
    4,5: repeat(rewrite from_list_cons);
    rewrite from_list_nil, union_empty_r.
    4,5: eapply disjoint_union_l; eapply disjoint_singleton.
    4,5,6,7: simpl in Hfrwy; try tauto.
    4,5: destruct Hfrwy as [_ [Hfrwy _]]; eauto.

    - eapply similar_extend, similar_open_rec; eauto using similar_var_push.
      eapply similar_extend; eauto.

    - eapply similar_env_with_push_notin.
      1: eapply similar_env_with_o_push, similar_with_i_push_notin; eauto;
      unfold envy_dom; simpl.

      1,4: rename Hfrwy1 into Hfrwy. 3,5: rename Hfrwy2 into Hfrwy.
      1,3: simpl in Hfrwy. 1,2: repeat(rewrite notin_union in Hfrwy); tauto.

      1: rename HinCLS1 into HinCL. 2: rename HinCLS2 into HinCL.
      1,2: simpl in Hfrwy; rewrite from_list_cons, notin_union,
        notin_singleton; split; eauto; introv Hin.
      1,2: specialize (HinCL _ Hin); simpl in Hfrwy, HinCL.
      1,2: destruct Hfrwy as [Hfrwy _]; eauto.

      pose proof (dup_lemma Hsimds) as Hsimdsop.
      eapply similar_extend with (x:=w1) (y:=w2),
          similar_open_rec with (n:=0),
            similar_extend with (x:=y1) (y:=y2) in Hsimdsop;
        eauto using similar_var_push.
      eapply similar_item_obj, deref_defs_similar; eauto;
        try solve[simpl; eauto].

      (* deref_defs replaces all fields that contain references with
         their corresponding locations. So suppose open wᵢ dsᵢ has a field
         {a=xᵢ}. By similarity Hsimds, it follows x₁ = x₂ assuming Lt₁ = Lt₂.
         Because the configuration is well-typed, xᵢ are references by
         the "all_tfree_k G t v_ref" requirement in ty_config along with the
         "ctx_kdom v_ref G = dom e" requirement in heap_correspond.
         Therefore, xᵢ is bound to y'ᵢ in eᵢ. Using Hsimenv, we know that
         y'₁ = y'₂ assuming Lt₁ = Lt₂. And so the result follows. *)

      1,2: constructor. 1: rename Hty1 into Hty. 3: rename Hty2 into Hty.
      1,3: inversions Hty; destruct H5; tauto.

      1: rename Hfrwy1 into Hfrwy. 2: rename Hfrwy2 into Hfrwy.
      1,2: simpl in Hfrwy; repeat(rewrite notin_union in Hfrwy); tauto.

      1: rename Hty1 into Hty. 2: rename Hty2 into Hty.
      1: rename Hfrwy1 into Hfrwy. 2: rename Hfrwy2 into Hfrwy.
      1,2: eapply renv_idempotent_push; try congruence.
      1,3: inversions Hty; inversions H5;
      eapply renv_corr_env_idempotent; eauto; try tauto.
      1,2: simpl in Hfrwy; repeat(rewrite notin_union in Hfrwy); tauto.

      1,2: rewrite dom_push; eapply subset_trans.
      1,3: eapply fields_open_subset_defs.
      1,2: eapply subset_union_l; eauto using subset_union_weak_l.
      1,2: eapply subset_trans. 2,4: eapply subset_union_weak_r.
      1: rename Hty1 into Hty. 2: rename Hty2 into Hty.
      1,2: inversions Hty; clear H4 H6 H7 H8 H11 H13 H14 H15 H16;
      rename H5 into Hhc, H9 into Htf.
      1,2: destructc Hhc as _ _ _ _ _ HrGe _.
      1,2: rewrite <- HrGe; hnf in Htf; introv Hin;
      eapply ctx_k_in_kdom, Htf, tfree_trm_lit_l, tfree_lit_obj,
        fields_tfree_defs; eauto.

    - repeat(rewrite dom_push). intros x1 x2 Hin1 Hin2 Hsim.
      eapply similar_var_simlist_relates, simlist_relates_cons_cases
        in Hsim as [Hsim | [? ?]]; subst.
      eapply simlist_relates_cons_cases in Hsim as [Hsim | [? ?]]; subst.
      3: { eapply similar_var_push; simpl; omega. }
      2: { eapply similar_extend; eapply similar_var_push; simpl; omega. }
      do 2 (eapply similar_extend). eapply Hsimcor;
        eauto using simlist_relates_similar_var.
      1: rename Hin1 into Hin. 2: rename Hin2 into Hin.
      1: rename Hfrwy1 into Hfrwy. 2: rename Hfrwy2 into Hfrwy.
      1: rename HinCLt1 into HinCLt. 2: rename HinCLt2 into HinCLt.
      1,2: rewrite in_union, in_singleton in Hin; destruct Hin as [? | ?];
      subst; eauto.
      1,2: false; simpl in Hfrwy; simpl in HinCLt.
      eapply simlist_relates_in_list_l in Hsim; firstorder.
      eapply simlist_relates_in_list_r in Hsim; firstorder. }

  { Case "red_let_var". destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H. inversions H0.
    rename fs into fs1, x3 into fs2, e into e1, e0 into e2,
      Lv into Lv1, Sigma into Sigma1, Sigma0 into Sigma2, Lv0 into Lv2,
      x into x1, x4 into t2, t into t1,
      H1 into Hsimx, H4 into Hsimt, H3 into Hsimfs.
    exists (@nil var) (@nil var). repeat(split); simpl; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: introv Hin; specialize (HinCL _ Hin); simpl in HinCL;
    repeat(rewrite in_union in *); tauto.

    2,3: rewrite from_list_nil; eauto using disjoint_empty_l.

    eapply similar_open_rec; eauto. }

  { Case "red_let_push".
    destruct Hred2.
    all: destruct_sef_sim; similar_syntax_inverts; try discriminate.
    inversions H.
    rename fs into fs1, fs0 into fs2, e into e1, e0 into e2,
      Lv into Lv1, Sigma into Sigma1, Sigma0 into Sigma2, Lv0 into Lv2,
      t into t1, x0 into t2, s into u1, x into u2,
      H0 into Hsimu, H1 into Hsimt.
    exists (@nil var) (@nil var). repeat(split); simpl; eauto.

    1: rename HinCLt1 into HinCL. 2: rename HinCLt2 into HinCL.
    3: rename HinCLS1 into HinCL. 4: rename HinCLS2 into HinCL.
    1,2,3,4: introv Hin; specialize (HinCL _ Hin); simpl in HinCL;
    repeat(rewrite in_union in *); tauto.

    2,3: rewrite from_list_nil; eauto using disjoint_empty_l.

    eapply similar_stack_let; eauto. }

  Unshelve.
  all: try solve[pick_fresh x; exact x].

Qed.

(* Lemma 47 in "Pure methods for roDOT" *)
Lemma conf_sim_red_append Lt1 Lt2 LS1 LS2 c1 c2 c1' c2' G1 G2 U1 U2 n :
  ty_config G1 c1 U1 -> ty_config G2 c2 U2 ->
  similar_conf_sef c1 c2 Lt1 Lt2 LS1 LS2 ->
  c1 ⊣n↦* c1' -> c2 ⊣n↦* c2' ->
   exists LS1' LS2',
     (similar_conf_sef
        c1' c2'
        (LS1' ++ Lt1) (LS2' ++ Lt2)
        (LS1' ++ LS1) (LS2' ++ LS2))
     /\ (disjoint (from_list LS1') (fv_config c1))
     /\ (disjoint (from_list LS2') (fv_config c2)).
Proof.
  introv Hty1 Hty2 Hsim Hred1 Hred2. reverts c1' c2'.
  induction n; intros.

  { inversions Hred1. inversions Hred2. exists (@nil var) (@nil var); simpl.
    repeat(rewrite from_list_nil). splits; eauto using disjoint_empty_l. }

  { inversions Hred1. inversions Hred2.
    rename C2 into C1, C0 into C2,
      H0 into Hreds1, H2 into Hred1, H1 into Hreds2, H4 into Hred2.
    forwards HtyC1:preservation_closure.
    2: exact Hreds1.
    { econstructor; eauto. }
    destruct HtyC1 as [G1' C1 U1' HtyC1].
    forwards HtyC2:preservation_closure.
    2: exact Hreds2.
    { econstructor; eauto. }
    destruct HtyC2 as [G2' C2 U2' HtyC2].
    forwards IH:IHn; fweauto.
    destructc IH as IHLS1' IHLS2' HsimIH HdisIH1 HdisIH2.
    forwards Hstep:conf_sim_red_append_single.
    3: eassumption. 3: exact Hred1. 3: exact Hred2. all: try eassumption.
    destructc Hstep as LS1' LS2' Hsims Hdis1 Hdis2.
    repeat(rewrite List.app_assoc in Hsims). exists. splits; try eassumption.
    all: rewrite from_list_append; eapply disjoint_union_l; eauto.
    1: rename Hreds1 into Hreds, Hdis1 into Hdis.
    2: rename Hreds2 into Hreds, Hdis2 into Hdis.
    1,2: eapply red_fv_config_more in Hreds.
    1,2: hnf; eapply fset_extens; eauto using subset_empty_l;
    rewrite <- Hdis; introv Hin; rewrite in_inter in *; firstorder. }
Qed.

Lemma conf_sim_red_create_single Lt1 Lt2 LS1 LS2 c1 c2 c1' G2 U2 :
  ty_config G2 c2 U2 ->
  similar_conf_sef c1 c2 Lt1 Lt2 LS1 LS2 ->
  c1 ↦ c1' -> exists c2', c2 ↦ c2'.
Proof.
  introv Hty2 Hsim Hred.
  eapply terminates_or_not in Hty2. destruct Hty2 as [? | Hans2]; eauto.
  false. inversions Hans2.
  destruct c1. destruct_sef_sim.
  destruct t; try similar_syntax_inverts; try discriminate.
  destruct s; try similar_syntax_inverts; try discriminate.
  inversions Hred.
Qed.

(* Lemma 48 in "Pure methods for roDOT" *)
Lemma conf_sim_red_create Lt1 Lt2 LS1 LS2 c1 c2 c1' G1 G2 U1 U2 n :
  ty_config G1 c1 U1 -> ty_config G2 c2 U2 ->
  similar_conf_sef c1 c2 Lt1 Lt2 LS1 LS2 ->
  c1 ⊣n↦* c1' -> exists c2', c2 ⊣n↦* c2'.
Proof.
  introv Hty1 Hty2 Hsim Hred. reverts c1'. induction n; eauto. intros.
  inversions Hred. eapply IHn in H0 as Hred2.
  destructc Hred2 as c2' Hred2.
  eapply conf_sim_red_append in Hsim; eauto.
  destructc Hsim as LS1' LS2' Hsim _ _.
  eapply preservation_closure in Hred2 as Hred2'; eauto.
  inversions Hred2'. eapply conf_sim_red_create_single in H; eauto.
  destruct H. eauto.
Qed.

End WithTypingMode.
