Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax GeneralTyping Generalization
        Similarity SyntaxSimilarity OperationalSemantics HeapUpdate
        Deref Defs ConfigValid.
Require Import HeapCorrespondence EnvCorrespondence.
Require Import ConfigTyping ConfigVarkinds TFree.
Require Import Safety.

(** * Similarity of configurations *)

Instance similar_conf: Similar config := {
  similar xs1 xs2 c1 c2 :=
    match (c1, c2) with
    | ((t1 ; s1 ; S1 ; e1 ; Lv1), (t2 ; s2 ; S2 ; e2 ; Lv2)) =>
      similar xs1 xs2 t1 t2 /\
      similar xs1 xs2 s1 s2 /\
      similar xs1 xs2 S1 S2 /\
      similar xs1 xs2 e1 e2
    end
}.

Definition similar_conf_any C1 C2 :=
  exists xs1 xs2, length xs1 = length xs2 /\ similar xs1 xs2 C1 C2.
(* Similarity is one-to-one, if there are no duplicates on either side.
   (The relation is a partial bijection.) *)
Definition similar_conf_1to1 C1 C2 :=
  exists xs1 xs2, vars_conf_unique xs1 C1 /\ vars_conf_unique xs2 C2 /\ length xs1 = length xs2 /\ similar xs1 xs2 C1 C2.

Lemma similar_conf_1to1_any C1 C2:
    similar_conf_1to1 C1 C2 ->
    similar_conf_any C1 C2.
Proof.
  introv Hs.
  unfold similar_conf_1to1, similar_conf_any in *.
  destruct_all.
  exists x x0. auto.
Qed.

Lemma vars_conf_unique_keys t s S e Lv:
  config_valid (t ; s ; S ; e ; Lv) ->
  vars_conf_unique (keys S ++ keys e)%list (t ; s ; S ; e ; Lv).
Proof.
  intros.
  inversions H.
  unfold vars_conf_unique.
  splits.
  eapply fresh_append.
  eapply ok_fresh_keys; eauto.
  rewrite union_empty_l.
  eapply fresh_forall_in_from_list.
  eapply ok_fresh_keys; eauto.
  rewrite ! env_from_list_keys_dom.
  intro. eapply disjoint_in_notin; auto.
  rewrite from_list_append.
  rewrite ! env_from_list_keys_dom.
  intro. rewrite ! in_union. intro. destruct H; right; auto.
Qed.

Lemma vars_conf_unique_fv_config_rd xs C:
  config_valid C ->
  from_list xs \c fv_config C ->
  vars_conf_unique (LibList.remove_duplicates xs) C.
Proof.
  introv Hva Hfl.
  inversions Hva.
  unfold vars_conf_unique.
  splits.
  eapply fresh_remove_duplicates.
  rewrite from_list_remove_duplicates. auto.
Qed.
Lemma similar_conf_1to1_refl_vars C:
  config_valid C ->
  exists xs,
    vars_conf_unique xs C /\
    similar xs xs C C /\
    from_list xs = fv_config C .
Proof.
  unfold similar, similar_conf.
  intros.
  inversions H.
  destruct (fset_finite ( fv_config (t ; s ; Sigma ; e ; Lv))).
  exists (LibList.remove_duplicates x).
  splits; try splits.
  - eapply vars_conf_unique_fv_config_rd; eauto. rewrite <- H. apply subset_refl.
  - rewrite from_list_remove_duplicates. rewrite <- H. eapply subset_refl.
  - eapply similar_refl_by_fv.
    rewrite from_list_remove_duplicates. rewrite <- H. intro. intro. rewrite ! in_union. auto.
  - eapply similar_refl_by_fv.
    rewrite from_list_remove_duplicates. rewrite <- H. intro. intro. rewrite ! in_union. auto.
  - eapply similar_in_refl_by_fv; auto.
    rewrite from_list_remove_duplicates. rewrite <- H. intro. intro. rewrite ! in_union. auto.
  - eapply similar_in_refl_by_fv; auto.
    rewrite from_list_remove_duplicates. rewrite <- H. intro. intro. rewrite ! in_union. right. auto.
  - rewrite from_list_remove_duplicates. auto.
Qed.
Lemma similar_conf_1to1_refl_by C xs:
  config_valid C ->
  fresh \{} (length xs) xs ->
  fv_config C \c from_list xs ->
  similar xs xs C C.
Proof.
  introv Hva Hfr Hfv.
  unfold similar, similar_conf.
  destruct C.
  splits; try splits.
  - eapply similar_refl_by_fv.
    eapply subset_trans. 2: eapply Hfv. union_sub_solver.
  - eapply similar_refl_by_fv.
    eapply subset_trans. 2: eapply Hfv. union_sub_solver.
  - eapply similar_in_refl_by_fv; auto.
    inversions Hva. auto.
    eapply subset_trans. 2: eapply Hfv. simpl. union_sub_solver.
  - eapply similar_in_refl_by_fv; auto.
    inversions Hva. auto.
    eapply subset_trans. 2: eapply Hfv. simpl. union_sub_solver.
Qed.


Lemma similar_conf_1to1_refl C1:
  config_valid C1 ->
  similar_conf_1to1 C1 C1.
Proof.
  unfold similar_conf_1to1.
  introv Hva.
  forwards Hsi: similar_conf_1to1_refl_vars Hva.
  destructecs_all.
  exists xs xs; auto.
Qed.
Lemma similar_conf_1to1_symm C1 C2:
  similar_conf_1to1 C1 C2 ->
  similar_conf_1to1 C2 C1.
Proof.
  introv Hs12.
  unfold similar_conf_1to1, similar, similar_conf, vars_conf_unique in *.
  destruct C1.
  destruct C2.
  destruct_all.
  exists x0 x; splits; eauto.
  all: eapply similar_symm; auto.
Qed.

Lemma similar_conf_length xs1 xs2 C1 C2:
  similar xs1 xs2 C1 C2 ->
  length xs1 = length xs2.
Proof.
  introv Hsi.
  unfold similar, similar_conf in Hsi.
  destruct C1; destruct C2. destruct_all.
  eapply similar_length in H.
  eauto.
Qed.

Section SimilarOut.
  Context {A: Set} {AS: Similar A} {AF: FreeVar A} {ASC: SimilarCommon AS} {ASO: SimilarOut AS}.

  Let SW: SimilarWith (env A) := similar_env_with AS.
  Existing Instance SW.

Lemma similar_trans_core_out (t t0 t1: A) (x22 x12' x02' x21 x20 x x0 x1 x2: list var) ns21 ns20:
  similar x1 x2 t t0 ->
  similar x x0 t0 t1 ->
  list_reorder x2 ns21 (x22 ++ x21) ->
  list_reorder x ns20 (x22 ++ x20) ->
  list_reorder x1 ns21 x12' ->
  list_reorder x0 ns20 x02' ->
  list_reorder x (LibList.take (LibList.length x22) ns20) x22 ->
  fresh \{} (length x) x ->
  LibList.noduplicates ns21 ->
  LibList.noduplicates ns20 ->
  LibList.noduplicates (x21 ++ x20) ->
  length ns21 = length x2 ->
  length ns20 = length x ->
  from_list x2 = from_list (x22 ++ x21) ->
  from_list x = from_list (x22 ++ x20) ->
  fresh \{} (length (x22 ++ x21)) (x22 ++ x21) ->
  fresh \{} (length (x22 ++ x20)) (x22 ++ x20) ->
  length x12' = length x22 + length x21 ->
  LibOrder.le (LibList.length x22) (LibList.length x12') ->
  length x02' = length x22 + length x20 ->
  LibOrder.le (LibList.length x22) (LibList.length x02') ->
  (LibList.take (LibList.length x22) x12' ++ LibList.drop (LibList.length x22) x12')%list = x12' ->
  (LibList.take (LibList.length x22) x02' ++ LibList.drop (LibList.length x22) x02')%list = x02' ->
  similar (LibList.take (LibList.length x22) x12') (LibList.take (LibList.length x22) x02') t t1.
Proof.
  introv Hs1 Hs0.
  introv Hre21' Hre20' Hre12' Hre02' Hre20 Hfrx.
  introv Hnd21 Hnd20 Hnd210 Hlns21 Hlns20 Hfx21 Hfx20.
  introv Hfr21 Hfr20 Hlx21 Hlx20 Hlex21 Hlex20 Hx12app Hx02app.

  forwards Hlx: similar_length Hs1.

  assert (similar x12' (x22 ++ x21) t t0) as Hs12. {
  eapply similar_reorder; eauto. rewrite Hlx. auto. }
  assert (similar (x22 ++ x20) x02' t0 t1) as Hs02. {
  eapply similar_reorder; eauto. }
  pose proof Hs1 as Hfvt0x2.
  eapply similar_out_fvr in Hfvt0x2.
  pose proof Hs0 as Hfvt0x0.
  eapply similar_out_fvl in Hfvt0x0.

  rewrite Hfx21 in Hfvt0x2.
  rewrite Hfx20 in Hfvt0x0.

  rewrite from_list_append in Hfvt0x2.
  rewrite from_list_append in Hfvt0x0.

  forwards Hfvt0: list_align_common_subset Hfvt0x2 Hfvt0x0. auto.

  assert (fresh (fv t0) (length x21) x21) as Hfr021. {
    eapply fresh_append_fresh_r in Hfr21.
    eapply fresh_subset; eauto.
  }
  assert (fresh (fv t0) (length x20) x20) as Hfr022. {
    eapply fresh_append_fresh_r in Hfr20.
    eapply fresh_subset; eauto.
  }

  eapply similar_trans with (ys:=x22) (b:=t0).
  * eapply list_noduplicates_fresh. eapply list_permuation_noduplicates; eauto. eapply fresh_noduplicates; eauto. eapply list_noduplicates_take; eauto.
  * rewrite <- Hx12app in Hs12.
    eapply similar_out_shorten_fresh_suffix_r. 3:eauto. rewrite <- liblist_length. rewrite LibList.length_drop.
    symmetry.
    eapply PeanoNat.Nat.add_sub_eq_r. rewrite liblist_length. rewrite Nat.add_comm. auto.
    auto. auto.
  * rewrite <- Hx02app in Hs02.
    eapply similar_out_shorten_fresh_suffix_l. 3:eauto. rewrite <- liblist_length. rewrite LibList.length_drop.
    eapply PeanoNat.Nat.add_sub_eq_r. rewrite liblist_length. rewrite Nat.add_comm. auto.
    auto. auto.
Qed.

Lemma similar_trans_core_env (t t0 t1: env A) (x22 x12' x02' x21 x20 x x0 x1 x2: list var) ns21 ns20:
  similar_env x1 x2 t t0 ->
  similar_env x x0 t0 t1 ->
  list_reorder x2 ns21 (x22 ++ x21) ->
  list_reorder x ns20 (x22 ++ x20) ->
  list_reorder x1 ns21 x12' ->
  list_reorder x0 ns20 x02' ->
  list_reorder x (LibList.take (LibList.length x22) ns20) x22 ->
  fresh \{} (length x) x ->
  LibList.noduplicates ns21 ->
  LibList.noduplicates ns20 ->
  LibList.noduplicates (x21 ++ x20) ->
  length ns21 = length x2 ->
  length ns20 = length x ->
  from_list x2 = from_list (x22 ++ x21) ->
  from_list x = from_list (x22 ++ x20) ->
  fresh \{} (length (x22 ++ x21)) (x22 ++ x21) ->
  fresh \{} (length (x22 ++ x20)) (x22 ++ x20) ->
  length x12' = length x22 + length x21 ->
  LibOrder.le (LibList.length x22) (LibList.length x12') ->
  length x02' = length x22 + length x20 ->
  LibOrder.le (LibList.length x22) (LibList.length x02') ->
  (LibList.take (LibList.length x22) x12' ++ LibList.drop (LibList.length x22) x12')%list = x12' ->
  (LibList.take (LibList.length x22) x02' ++ LibList.drop (LibList.length x22) x02')%list = x02' ->
  similar_env (LibList.take (LibList.length x22) x12') (LibList.take (LibList.length x22) x02') t t1.
Proof.
  introv Hs1 Hs0.
  introv Hre21' Hre20' Hre12' Hre02' Hre20 Hfrx.
  introv Hnd21 Hnd20 Hnd210 Hlns21 Hlns20 Hfx21 Hfx20.
  introv Hfr21 Hfr20 Hlx21 Hlx20 Hlex21 Hlex20 Hx12app Hx02app.

  unfold similar_env in *.

  forwards Hlx: similar_env_with_i_length Hs1.

  assert (length x1 = length ns21) as Hlx1. rewrite Hlx. rewrite Hlns21. auto.
  assert (length x0 = length ns20) as Hlx0. rewrite Hlns20. eapply similar_env_with_i_length in Hs0. auto.

  assert (similar_with x12' (x22++x21) (LibList.take (LibList.length x22) x12') x22 t t0) as Hs12. {
    eapply similar_env_with_o_reorder; eauto.
    eapply similar_with_i_append_invl with (xs12 := LibList.drop (LibList.length x22) x12') (xs22:=x21).
    rewrite <- liblist_length. rewrite LibList.length_take; eauto.
    rewrite Hx12app.
    eapply similar_env_with_i_reorder; eauto.
  }
  assert (similar_with (x22++x20)  x02' x22 (LibList.take (LibList.length x22) x02') t0 t1) as Hs02. {
    eapply similar_env_with_o_reorder; eauto.
    eapply similar_with_i_append_invl with (xs22 := LibList.drop (LibList.length x22) x02') (xs12:=x20).
    rewrite <- liblist_length. rewrite LibList.length_take; eauto.
    rewrite Hx02app.
    eapply similar_env_with_i_reorder; eauto.
  }



    remember (LibList.take (LibList.length x22) x12') as x12i in Hs12 |- * at 2.
    remember (LibList.take (LibList.length x22) x02') as x02i in Hs02 |- * at 2.
    remember x22 as x22i in Hs12 at 2, Hs02 at 2.
    clear Heqx12i.
    clear Heqx02i.
    clear Heqx22i.

    reverts x12i x02i.
    induction x22i; intros.
    - inversions Hs12. inversions Hs02. constructor. rewrite <- ! liblist_length. rewrite ! LibList.length_take; eauto.
    - inversions Hs12; inversions Hs02.
      + eapply similar_env_dom. eapply IHx22i; eauto. eauto. eauto.
        binds_eq.
        eapply similar_trans_core_out with (x:=x) (x2:=x2) (t0 := it0); eauto.

        eapply similar_reorder_inv; eauto.
        eapply similar_reorder_inv; eauto.
      + eapply binds_to_dom in H6. contradiction.
      + eapply binds_to_dom in H3. contradiction.
      + eapply similar_env_notin; eauto.
Qed.



End SimilarOut.

Lemma similar_conf_1to1_trans C1 C2 C3:
  config_valid C1 ->
  config_valid C2 ->
  config_valid C3 ->
  similar_conf_1to1 C1 C2 ->
  similar_conf_1to1 C2 C3 ->
  similar_conf_1to1 C1 C3.
Proof.
  introv Hv1 Hv2 Hv3 Hs12 Hs23.
  unfold similar_conf_1to1, similar, similar_conf, vars_conf_unique in *.
  destruct C1.
  destruct C2.
  destruct C3.
  destruct_all.

  forwards: list_align_common x2 x.
  - eapply fresh_noduplicates; eauto.
  - eapply fresh_noduplicates; eauto.
  - destructc H17 as x22 x21 x20 ns21 ns20 Hre21 Hre20 Hnd21 Hnd20 Hnd210 Hls. destructc Hls as Hlns21 Hlns20.
    pose proof Hre21 as Hre21'.
    pose proof Hre20 as Hre20'.
    eapply list_reorder_take with (m := (length x22)) in Hre21.
    eapply list_reorder_take with (m := (length x22)) in Hre20.
    rewrite <- liblist_length in Hre21, Hre20.
    rewrite <- liblist_app in Hre21, Hre20.
    rewrite LibList.take_prefix_length in Hre21, Hre20.

    forwards x12' Hre12': (list_reorder_same_length_exists x1 Hre21'). auto.
    forwards x02' Hre02': (list_reorder_same_length_exists x0 Hre20'). auto.
    pose proof Hre12' as Hre12.
    pose proof Hre02' as Hre02.
    eapply list_reorder_take with (m := (length x22)) in Hre12.
    eapply list_reorder_take with (m := (length x22)) in Hre02.
    rewrite <- liblist_length in Hre12, Hre02.

    forwards Hfx21: list_reorder_from_list Hre21'; eauto.
    forwards Hfx20: list_reorder_from_list Hre20'; eauto.

    assert (fresh \{} (length (x22++x21)) (x22 ++ x21)) as Hfr21. {
      eapply list_noduplicates_fresh. eapply list_permuation_noduplicates; eauto. eapply fresh_noduplicates; eauto.
    }
    assert (fresh \{} (length (x22++x20)) (x22 ++ x20)) as Hfr20. {
      eapply list_noduplicates_fresh. eapply list_permuation_noduplicates; eauto. eapply fresh_noduplicates; eauto.
    }

    assert (length x12' = length x22 + length x21) as Hlx12. {
      eapply list_reorder_lengths in Hre21'.
      eapply list_reorder_lengths in Hre12'.
      rewrite Hre12'. rewrite <- Hre21'. rewrite List.app_length.
      auto.
    }
    assert (LibOrder.le (LibList.length x22) (LibList.length x12')) as Hlex12. {
      rewrite liblist_length.
      rewrite Hlx12.
      rewrite LibNat.le_peano. eapply Nat.le_add_r.
    }
    assert (length x02' = length x22 + length x20) as Hlx02. {
      eapply list_reorder_lengths in Hre20'.
      eapply list_reorder_lengths in Hre02'.
      rewrite Hre02'. rewrite <- Hre20'. rewrite List.app_length.
      auto.
    }
    assert (LibOrder.le (LibList.length x22) (LibList.length x02')) as Hlex02. {
      rewrite liblist_length.
      rewrite Hlx02.
      rewrite LibNat.le_peano. eapply Nat.le_add_r.
    }
    forwards Hx12app: LibList.list_eq_take_app_drop (LibList.length x22) x12' Hlex12.
    forwards Hx02app: LibList.list_eq_take_app_drop (LibList.length x22) x02' Hlex02.
    rewrite liblist_app in Hx12app, Hx02app.

    exists (LibList.take (LibList.length x22) x12') (LibList.take (LibList.length x22) x02'); splits.
    + split. eapply list_noduplicates_fresh.
      eapply list_permuation_noduplicates. eapply Hre12. eauto.
      eapply fresh_noduplicates; eauto.
      eapply list_noduplicates_take; eauto.
      eapply subset_trans. eapply list_reorder_from_list_subset. eauto. eauto.
    + split. eapply list_noduplicates_fresh.
      eapply list_permuation_noduplicates; eauto.
      eapply fresh_noduplicates; eauto.
      eapply list_noduplicates_take; eauto.
      eapply subset_trans. eapply list_reorder_from_list_subset. eauto. eauto.
    + eapply list_reorder_lengths in Hre12.
      eapply list_reorder_lengths in Hre02.
      eapply list_reorder_lengths in Hre21.
      eapply list_reorder_lengths in Hre20.
      rewrite Hre12, Hre02.
      rewrite <- Hre21, <- Hre20. auto.
    + (* term *)
      eapply (similar_trans_core_out); eauto.
    + (* stack *)
      eapply (similar_trans_core_out); eauto.
    +
      eapply similar_trans_core_env; eauto.
    +
      eapply similar_trans_core_env; eauto.
Qed.

Lemma similar_conf_answer_case C1 C2:
  similar_conf_1to1 C1 C2 ->
  answer C1 ->
  answer C2.
Proof.
  introv Hsi Hct.
  inversions Hct.
  unfold similar_conf_1to1 in Hsi.
  destructecs Hsi.
  destruct C2.
  hnf in Hsi.
  destructecs Hsi.
  eapply similar_trm_var_case in Hsi3.
  destructecs Hsi3; subst.
  eapply similar_stack_nil_case in Hsi4.
  subst.
  eapply similar_avar_f_case in Hsi3.
  destructecs Hsi3; subst.
  constructor.
Qed.

Lemma similar_conf_1to1_heap_renv_extend
  ta fsa Sa ea Lva tb fsb Sb eb Lvb ea2 eb2 Sa2 Sb2:
  similar_conf_1to1 (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  disjoint (dom ea2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  disjoint (dom eb2) (fv_config (tb ; fsb ; Sb ; eb ; Lvb)) ->
  disjoint (dom Sa2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  disjoint (dom Sb2) (fv_config (tb ; fsb ; Sb ; eb ; Lvb)) ->
  similar_conf_1to1 (ta ; fsa ; Sa & Sa2  ; ea & ea2 ; Lva) (tb ; fsb ; Sb & Sb2  ; eb & eb2 ; Lvb).
Proof.
  introv Hsim Hdia Hdib HdiSa HdiSb.
  hnf in Hsim. destructecs Hsim.
  hnf. exists xs1 xs2; splits.
  -
  eapply vars_conf_unique_e_subset.
  eapply vars_conf_unique_S_subset.
  eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  -
  eapply vars_conf_unique_e_subset. 
  eapply vars_conf_unique_S_subset. eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  - auto.  
  -
  hnf in Hsim.
  destructecs Hsim.
  hnf. splits; auto.
  eapply similar_env_with_e_concat_l_fresh. eauto.
  destruct Hsim0.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
  destruct Hsim1.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.

  eapply similar_env_with_e_concat_l_fresh. eauto.
  destruct Hsim0.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
  destruct Hsim1.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
Qed.

Lemma similar_conf_1to1_heap_extend
  ta fsa Sa ea Lva tb fsb Sb eb Lvb Sa2 Sb2:
  similar_conf_1to1 (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  disjoint (dom Sa2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  disjoint (dom Sb2) (fv_config (tb ; fsb ; Sb ; eb ; Lvb)) ->
  similar_conf_1to1 (ta ; fsa ; Sa & Sa2; ea ; Lva) (tb ; fsb ; Sb & Sb2 ; eb ; Lvb).
Proof.
  introv Hsim Hdia Hdib.
  hnf in Hsim. destructecs Hsim.
  hnf. exists xs1 xs2; splits.
  -
  eapply vars_conf_unique_S_subset. eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  -
  eapply vars_conf_unique_S_subset. eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  - auto.  
  -
  hnf in Hsim.
  destructecs Hsim.
  hnf. splits; auto.
  eapply similar_env_with_e_concat_l_fresh. eauto.
  destruct Hsim0.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
  destruct Hsim1.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
Qed.

Lemma similar_conf_1to1_heap_extend_refl
  ta fsa Sa ea Lva Sa2:
  disjoint (dom Sa2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  config_valid (ta ; fsa ; Sa ; ea ; Lva) ->
  similar_conf_1to1 (ta ; fsa ; Sa ; ea ; Lva) (ta ; fsa ; Sa & Sa2 ; ea ; Lva).
Proof.
  introv Hdi Hva.
  forwards: similar_conf_1to1_refl Hva.
  rewrite <- concat_empty_r at 1.
  eapply similar_conf_1to1_heap_extend.
  auto.
  rewrite dom_empty. eapply disjoint_empty_l.
  auto.
Qed.


Lemma similar_conf_1to1_vars_renv_extend
  ta fsa Sa ea Lva tb fsb Sb eb Lvb ea2 eb2 xs ys:
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  fresh (dom ea2) (length xs) xs ->
  fresh (dom eb2) (length ys) ys ->
  similar xs ys (ta ; fsa ; Sa ; ea & ea2 ; Lva) (tb ; fsb ; Sb ; eb & eb2 ; Lvb).
Proof.
  introv Hsim Hdia Hdib.
  hnf in Hsim. destructecs Hsim.
  hnf. splits; auto.
  eapply similar_env_with_e_concat_l_fresh; eauto.
Qed.
Lemma similar_conf_1to1_vars_heap_extend
  ta fsa Sa ea Lva tb fsb Sb eb Lvb Sa2 Sb2 xs ys:
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  fresh (dom Sa2) (length xs) xs ->
  fresh (dom Sb2) (length ys) ys ->
  similar xs ys (ta ; fsa ; Sa & Sa2 ; ea ; Lva) (tb ; fsb ; Sb  & Sb2 ; eb ; Lvb).
Proof.
  introv Hsim Hdia Hdib.
  hnf in Hsim. destructecs Hsim.
  hnf. splits; auto.
  eapply similar_env_with_e_concat_l_fresh; eauto.
Qed.
Lemma similar_conf_1to1_vars_heap_renv_extend_r
  ta fsa Sa ea Lva tb fsb Sb eb Lvb eb2 Sb2 xs ys:
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  fresh \{} (length xs) xs ->
  fresh (dom Sb2 \u dom eb2) (length ys) ys ->
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb  & Sb2 ; eb  & eb2; Lvb).
Proof.
  introv Hsim Hdia Hdib.
  forwards: similar_conf_1to1_vars_renv_extend ea eb (empty: renv) eb2 Hsim. rewrite dom_empty. eauto. eauto.
  forwards: similar_conf_1to1_vars_heap_extend Sa Sb (empty: heap) Sb2 H. rewrite dom_empty. eauto. eauto.
  rewrite ! concat_empty_r in H0. auto.
Qed.


Lemma similar_conf_1to1_renv_extend
  ta fsa Sa ea Lva tb fsb Sb eb Lvb ea2 eb2:
  similar_conf_1to1 (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  disjoint (dom ea2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  disjoint (dom eb2) (fv_config (tb ; fsb ; Sb ; eb ; Lvb)) ->
  similar_conf_1to1 (ta ; fsa ; Sa ; ea & ea2 ; Lva) (tb ; fsb ; Sb ; eb & eb2 ; Lvb).
Proof.
  introv Hsim Hdia Hdib.
  hnf in Hsim. destructecs Hsim.
  hnf. exists xs1 xs2; splits.
  -
  eapply vars_conf_unique_e_subset. eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  -
  eapply vars_conf_unique_e_subset. eauto.
  rewrite dom_concat, fv_env_concat. union_sub_solver.
  - auto.  
  -
  hnf in Hsim.
  destructecs Hsim.
  hnf. splits; auto.
  eapply similar_env_with_e_concat_l_fresh. eauto.
  destruct Hsim0.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
  destruct Hsim1.
  eapply disjoint_fresh. eauto.
  eapply disjoint_subset_l; eauto.
Qed.

Lemma similar_conf_1to1_heap_renv_extend_refl
  ta fsa Sa ea Lva Sa2 ea2:
  disjoint (dom Sa2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  disjoint (dom ea2) (fv_config (ta ; fsa ; Sa ; ea ; Lva)) ->
  config_valid (ta ; fsa ; Sa ; ea ; Lva) ->
  similar_conf_1to1 (ta ; fsa ; Sa ; ea ; Lva) (ta ; fsa ; Sa & Sa2 ; ea & ea2 ; Lva).
Proof.
  introv HdiS Hdie Hva.
  forwards: similar_conf_1to1_refl Hva.
  rewrite <- concat_empty_r at 1.
  rewrite <- concat_empty_r with (E:=ea) at 1.
  eapply similar_conf_1to1_heap_renv_extend.
  auto.
  rewrite dom_empty. eapply disjoint_empty_l.
  auto.
  rewrite dom_empty. eapply disjoint_empty_l.
  auto.
Qed.

Lemma similar_conf_Lv xs ys
  ta fsa Sa ea Lva tb fsb Sb eb Lvb Lva2 Lvb2:
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva2) (tb ; fsb ; Sb ; eb ; Lvb2).
Proof.
  introv Hsi.
  hnf in *; auto.
Qed.

Instance similar_conf_common: SimilarCommon similar_conf.
Proof.
  econstructor; intros.
  - hnf in *. destruct a, b. destructecs H. splits. all: eapply similar_symm; auto.
  - hnf in *. destruct a, b, c. destructecs H0. destructecs H1. splits. all: eapply similar_trans; eauto.
  - hnf in *. destruct a, b. destructecs H. eapply similar_length; eauto.
  - hnf in *. destruct a, b. destructecs H. splits; eapply similar_reorder; eauto.
Qed.
