Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax Similarity.

Section SimilarCtx.
  Context {AS: Similar typ}.

(** * Similarity of contexts.
  
  It is defined in two phases.
  A "global" phase is used for configuration typing (inert contexts)
  and is defined analogously to env similarity,
  by defining which pairs of variables must be similar.

  However, to acomodate local contexts used for typing subterms,
  a second phase allows appending more context parts
*)

(** Similarity of global contexts.
   The first two lists define variable mapping allowed for the types.
   The second two lists define variable mapping required for domain,
   where either both variables are in domain and have similar types and kinds,
   or both variables are not in domain.
*)

Inductive similar_ctx_with: SimilarWith ctx :=
(* Empty list of required pairs *)
| similar_ctx_empty ys1 ys2 S1 S2:
    length ys1 = length ys2 ->
    similar_with ys1 ys2 nil nil S1 S2
(* Pair in domain *)
| similar_ctx_dom ys1 ys2 S1 S2 xs1 xs2 x1 x2 it1 it2 vk:
    similar_with ys1 ys2 xs1 xs2 S1 S2 ->
    (ctx_binds x1 it1 S1) ->
    (ctx_binds x2 it2 S2) ->
    (ctx_k S1 x1 vk) ->
    (ctx_k S2 x2 vk) ->
    similar ys1 ys2 it1 it2 ->
    similar_with ys1 ys2 (x1::xs1) (x2::xs2) S1 S2
(* Pair not in domain *)
| similar_ctx_notin ys1 ys2 S1 S2 xs1 xs2 x1 x2:
    similar_with ys1 ys2 xs1 xs2 S1 S2 ->
    (x1 §# S1) ->
    (x2 §# S2) ->
    similar_with ys1 ys2 (x1::xs1) (x2::xs2) S1 S2.

Local Existing Instance similar_ctx_with.

(** Simlarity of local contexts *)
Inductive similar_ctx: Similar ctx :=
| similar_ctx_global x1 x2 a1 a2:
    ctx_weakens a1 ->
    ctx_weakens a2 ->
    similar_with x1 x2 x1 x2 a1 a2 ->
    similar x1 x2 a1 a2
| similar_ctx_push  x1 x2 a1 a2 v1 v2 vk T1 T2:
    similar x1 x2 T1 T2 ->
    simlist_relates x1 x2 v1 v2 ->
    similar x1 x2 a1 a2 ->
    similar x1 x2 (a1 § v1 / vk ~ T1) (a2 § v2 / vk ~ T2)
| similar_ctx_hide  x1 x2 a1 a2:
    similar x1 x2 a1 a2 ->
    similar x1 x2 (ctx_hide a1) (ctx_hide a2)
.

(** Similarity of local object contexts *)
Inductive similar_objctx(AS: Similar typ): Similar objctx :=
| similar_objctx_empty xs1 xs2:
  similar xs1 xs2 objctx_empty objctx_empty
| similar_objctx_lit xs1 xs2 x1 x2 T1 T2:
  simlist_relates xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 (objctx_lit x1 T1) (objctx_lit x2 T2)
| similar_objctx_heap xs1 xs2 x1 x2 T1 T2:
  simlist_relates xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 (objctx_heap x1 T1) (objctx_heap x2 T2)
.

End SimilarCtx.

Hint Constructors similar_objctx: core.

Section SimilarCtx.
  Context {AS: Similar typ}{AF: FreeVar typ}{ASC: SimilarCommon AS}{ASO: SimilarOut AS}.

  Local Existing Instance similar_ctx_with.
  Local Existing Instance similar_ctx.
  Local Existing Instance similar_objctx.

Lemma similar_ctx_with_binds xs1 xs2 (S1 S2: ctx) x1 x2 it1 it2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  simlist_relates xs1 xs2 x1 x2 ->
  ctx_binds x1 it1 S1 ->
  ctx_binds x2 it2 S2 ->
  similar ys1 ys2 it1 it2.
Proof.
  intros.
  induction H.
  - false. eapply simlist_relates_nil_inv; eauto.
  - unfold simlist_relates in H0. destruct_all.
    inversions H0.
    + inversions H8. ctx_binds_eq. ctx_binds_eq. auto.
    + inversions H8. eapply IHsimilar_ctx_with; eauto. exists n; eauto.
  -
    unfold simlist_relates in H0. destruct_all.
    inversions H0.
    + inversions H5. eapply ctx_binds_dom in H1. contradiction.
    + inversions H5. eapply IHsimilar_ctx_with; eauto. exists n; eauto.
Qed.
Lemma similar_ctx_cases xs1 xs2 ys1 ys2 (e1 e2: ctx) x1 x2:
  simlist_relates xs1 xs2 x1 x2 ->
  similar_with ys1 ys2 xs1 xs2 e1 e2 ->
  (x1 §# e1 /\ x2 §# e2) \/ (exists b1 b2, ctx_binds x1 b1 e1 /\ ctx_binds x2 b2 e2 /\ similar ys1 ys2 b1 b2).
Proof.
  introv Hsx Hse.
  unfold simlist_relates in Hsx.
  destruct_all.
  reverts xs1 xs2.
  induction x; intros.
  - inversions Hse.
    + false. eapply LibList.Nth_nil_inv; eauto.
    + inversions H. inversions H0. eauto 6.
    + inversions H. inversions H0. eauto 6.
  - inversions H. inversions H0.
    inversions Hse.
    + forwards: IHx; eauto.
    + forwards: IHx; eauto.
Qed.
Lemma similar_ctx_binds xs1 xs2 (S1 S2: ctx) x1 x2 it1 it2:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 S1 S2 ->
  simlist_relates xs1 xs2 x1 x2 ->
  ctx_binds x1 it1 S1 ->
  ctx_binds x2 it2 S2 ->
  similar xs1 xs2 it1 it2.
Proof.
  introv Hfr1 Hfr2 HsiG Hsix Hbi1 Hbi2.
  induction HsiG.
  - eapply similar_ctx_with_binds; eauto.
  - inversions Hbi1; inversions Hbi2.
    all: try forwards: simlist_relates_noduplicates_var_eq H0 Hsix; eauto.
    all: subst; try contradiction.
    eapply simlist_relates_swap in H0.
    eapply simlist_relates_swap in Hsix.
    forwards: simlist_relates_noduplicates_var_eq H0 Hsix; eauto.
    subst. contradiction.
  - inversions Hbi1.
    inversions Hbi2.
    eauto.
Qed.

(* SimilarWithIn - most proofs are copies of the same for env *)
Lemma similar_ctx_with_push_l y1 ys1 y2 ys2 (S1 S2: ctx) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar_with (y1::ys1) (y2::ys2) xs1 xs2 S1 S2.
Proof.
  introv Hs.
  induction Hs.
  - eapply similar_ctx_empty. simpl. rewrite H. auto.
  - eapply similar_ctx_dom; eauto. eapply similar_extend; eauto.
  - eapply similar_ctx_notin; eauto.
Qed.

Lemma similar_ctx_middle_r_inv (a1 a2: ctx) (ys1 ys2 zsl zsr wsl wsr: list var) w z:
    length wsl = length zsl ->
    length wsr = length zsr ->
    similar_with ys1 ys2 (zsl ++ z :: zsr) (wsl ++ w :: wsr) a1 a2 -> (
      (similar_with ys1 ys2 (zsl ++ zsr) (wsl ++ wsr) a1 a2) /\
      ((exists b1 b2 vk,
          ctx_binds z b1 a1 /\
          ctx_binds w b2 a2 /\
          ctx_k a1 z vk /\
          ctx_k a2 w vk /\
          similar ys1 ys2 b1 b2
        )\/(
          z §# a1 /\ w §# a2
      ))) .
Proof.
  intros.
  reverts wsl.
  induction zsl; intros; destruct wsl; simpls; try discriminate.
  - inversions H1.
    +   splits. auto. left. exists it1 it2 vk. auto.
    +   splits. auto. auto.
  - inversions H.
    inversions H1.
    + splits. eapply similar_ctx_dom; eauto. eapply IHzsl; auto. eapply IHzsl; eauto.
    + splits. eapply similar_ctx_notin; eauto. eapply IHzsl; auto. eapply IHzsl; eauto.
Qed.

(* The guiding lists can be reordered *)
Lemma similar_ctx_with_reorder_r (a1 a2: ctx) ns (xs1 xs2 ys1 ys2 zs1 zs2: list var):
    list_reorder xs1 ns zs1 ->
    list_reorder xs2 ns zs2 ->
    LibList.noduplicates ns ->
    length xs1 = length ns ->
    length xs2 = length ns ->
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with ys1 ys2 zs1 zs2 a1 a2.
Proof.

  eapply (list_permutation_ind (fun (zs1 ws1 zs2 ws2: list var)=> similar_with ys1 ys2 zs1 ws1 a1 a2 -> similar_with ys1 ys2 zs2 ws2 a1 a2 )).
  - intros.
    inversions H.
    eapply similar_ctx_empty. auto.
  - introv Hl1 Hl2 Hind Hs.

    eapply similar_ctx_middle_r_inv in Hs; eauto.
    destruct_all.
    eapply similar_ctx_dom; eauto.
    eapply similar_ctx_notin; eauto.
Qed.

Lemma similar_ctx_with_reorder_l (a1 a2: ctx) ns (ys1 ys2 zs1 zs2: list var):
    list_reorder ys1 ns zs1 ->
    list_reorder ys2 ns zs2 ->
    LibList.noduplicates ns ->
    length ys1 = length ns ->
    length ys2 = length ns ->
    forall (xs1 xs2: list var),
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with zs1 zs2 xs1 xs2 a1 a2.
Proof.
  intros.
  induction H4.
  - eapply similar_ctx_empty. eapply list_reorder_lengths in H. eapply list_reorder_lengths in H0. rewrite H. rewrite H0. auto.
  - eapply similar_ctx_dom; eauto.
    eapply similar_reorder; eauto.
  - eapply similar_ctx_notin; eauto.
Qed.
Lemma similar_ctx_length xs ys zs ws a b:
  similar_with xs ys zs ws a b ->
  length zs = length ws.
Proof.
  introv Hs.
  induction Hs; simpls; auto.
Qed.
Lemma similar_ctx_with_shorten ys1 ys2 (S1 S2: ctx) xs11 xs12 xs21 xs22:
  length xs11 = length xs21 ->
  similar_with ys1 ys2 (xs11 ++ xs12) (xs21 ++ xs22) S1 S2 ->
  similar_with ys1 ys2 (xs11) (xs21) S1 S2.
Proof.
  introv Hl Hs.
  forwards: similar_ctx_length Hs.
  rewrite ! List.app_length in H. rewrite Hl in H. apply Plus.plus_reg_l in H.

  reverts xs22.
  induction xs12; intros; destruct xs22; simpls; try discriminate.
  - rewrite ! List.app_nil_r in Hs. auto.
  - inversions H. eapply IHxs12. eapply similar_ctx_middle_r_inv with (wsr:=xs22); eauto. auto.
Qed.

Global Instance similar_with_in_ctx: SimilarWithIn (EnvyCtx) (similar_with).
Proof.
  split; intros.
  - eapply similar_ctx_with_push_l; eauto.
  - eapply similar_ctx_with_shorten; eauto.
  - eapply similar_ctx_with_reorder_r; eauto.
  - eapply similar_ctx_with_reorder_l; eauto.
  - eapply similar_ctx_notin; eauto.
Qed.

Lemma similar_ctx_with_append_fresh xs1 xs2 (S1 S2: ctx) zs1 zs2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  fresh (ctx_dom S1) (length zs1) zs1 ->
  fresh (ctx_dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar_with ys1 ys2 (zs1++xs1) (zs2++xs2) S1 S2.
Proof.
  introv Hsi Hfr1 Hfr2 Hl.
  reverts zs2; induction zs1; intros; destruct zs2; inversions Hl.
  - rewrite List.app_nil_l; auto.
  - rewrite <- ! List.app_comm_cons.
    inversions Hfr1.
    eapply fresh_union_r1 in H1.
    inversions Hfr2.
    eapply fresh_union_r1 in H3.
    eapply similar_ctx_notin; eauto.
Qed.

Lemma similar_ctx_append_fresh xs1 xs2 (S1 S2: ctx) zs1 zs2:
  similar xs1 xs2 S1 S2 ->
  fresh (ctx_dom S1) (length zs1) zs1 ->
  fresh (ctx_dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar (zs1 ++ xs1) (zs2 ++ xs2) (S1) (S2).
Proof.
  introv HsS Hfr1 Hfr2 Hl.
  induction HsS.
  - constructor; eauto.
    eapply similar_ctx_with_append_fresh; eauto.
    eapply similar_with_o_append_r.
    all: eauto.
  -    ctx_simpl_dom.
    eapply similar_ctx_push; eauto.
    eapply similar_out_append; eauto.
    eapply simlist_relates_append; eauto.
  -    ctx_simpl_dom.
    eapply similar_ctx_hide; eauto.
Qed.

Lemma similar_objctx_extend xs1 xs2 (S1 S2: objctx) x1 x2:
  similar xs1 xs2 S1 S2 ->
  similar (x1::xs1) (x2::xs2) S1 S2.
Proof.
  introv HsS.
  induction HsS.
  - constructor; eauto.
  - constructor; eauto.
    eapply simlist_relates_push; eauto.
    eapply similar_extend; eauto.
  - constructor; eauto.
    eapply simlist_relates_push; eauto.
    eapply similar_extend; eauto.
Qed.

End SimilarCtx.