Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax Generalization Envs.

(** * Similarity relation *)

(* Configurations conf1 and conf2 are similar if there exist a relation R between variables in conf1 and conf2,
   where R = zip x1 x2,
   such that:

  1) t1 and t2 are structurally the same; they may differ in variables, which must be related by R
     this can be expressed b existence of term t3 and list of variables x3, such that t3[x3 --> x1] = t1 and t3[x3 --> x2] = t2
  2) fs1 and fs2 are struturally the same in the same manner
  3) For each pair v1 v2 in R, either:
     - (loc case)
       there exists object o1 at v1 in S1,
       and there exists object o2 at v2 in S2,
       and the contents are structurally the same by R
     - (ref case)
       or v1 -> y1 in e1
       and v2 -> y2 in e2
       and y1 R y2.

This means:
  - if an object is referenced by a reference w from the term or stack of conf1,
    then by (1), there must be a related counterpart in conf2.
    By (3) there must be a corresponding reference in conf1 and in conf2 and these two must be related.
    By (2) there must be an object on the heap of conf1 and conf2 and they must be equivalent.
  - if an object is reachable from the term or stack of conf2 by a sequence of field reads,
    then starting from the reference, there must be related pairs of locations of each field value,
    so an equivalent object must also exist in conf2.
  - if an object is not reachable from the term or stack of conf1,
    then it may be or may be not related to some object in conf2, because its location
    is not required to be in x1
    (see about type prefixes below)
  - two different objects in conf1 may be related to a single object in conf2
    this may be undesired because that means a result of mutation is different in conf1 and conf2

To have a stricter version preserved under mutability:
  - each location y1 (in x1) must be related to at most one variable y2 - no duplicates in x1
  - each location y2 (in x2) must be related to at most one variable y1 - no duplicates in x2
  - but we do not need such restriction for references - they can be merged, but by condition (3b) they correspond to the same variable.

Regrading type prefixes:
- we want to preserve types in the terms in order to be able to type the whole config
  - we want equivalent configs to have equivalent types.
- we want it to be loose enough to allow program transformations
  - for reordering, same rules as for terms (with uniqueness) should be ok
  - for deduplication, considering that deduplicated calls are in let-assignments
    - the method call always reduces to a single reference before being substituted into the next term
    - so it should be ok, because at that point, there is no type in the term
    - and by the deterministic condition, the references will point to the same location

A big difference in similarity of terms, types  vs  heap, env:
- In case of terms and types, which can be thought of as the output part of a conf,
  they must be structurally the same and every occurence of a variables must be related.
  If the relation is empty, then they must be the same.
  If two terms are related, then the relation can be extended with any pair of variables

  In case of heap, which can be thought of as the input part, we want to allow unrelated variables in dom,
  as long as they do not appear in the term or in a definition of a variable that is related.
  If the relation is empty, then the heap can be completely different.
  It is not in general possible to add or remove any pair of variables from the relation.

On equivalence relation:
  Similarity is in some sense an equivalence relation.
  But it cannot be an equivalence of (xs1, t1) ~ (xs2, t2), because in transitivity we would have to change the length of the variable list.
  Example: t1 = (x11, x11, x12), t2 = (x21, x21, x21), t3 = (x31, x32, x32)
    Then  ([x11, x12], t1) ~ ([x21, x21], t2) and ([x21, x21], t2) ~ ([x31, x32], t3)
    But   ([x11, x11, x12], t1) ~ ([x31, x32, x32], t3)
*)

Notation simlist := (list var) (only parsing).

Definition simlist_relates (xs1 xs2: simlist) x1 x2 :=
  exists n, LibList.Nth n xs1 x1 /\ LibList.Nth n xs2 x2.

Lemma simlist_relates_in_list_l x1 x2 (xs1 xs2: simlist) :
  simlist_relates xs1 xs2 x1 x2 -> x1 \in from_list xs1.
Proof.
  introv [n [Hin _]]. eapply LibList.Nth_mem, list_mem_in_fromlist in Hin;
    easy.
Qed.

Lemma simlist_relates_in_list_r x1 x2 (xs1 xs2: simlist) :
  simlist_relates xs1 xs2 x1 x2 -> x2 \in from_list xs2.
Proof.
  introv [n [_ Hin]]. eapply LibList.Nth_mem, list_mem_in_fromlist in Hin;
    easy.
Qed.

Lemma simlist_relates_cons_cases x1 x2 (xs1 xs2: simlist) y1 y2:
  simlist_relates (x1::xs1)%list (x2::xs2)%list y1 y2 ->
  (simlist_relates xs1 xs2 y1 y2) \/ (y1 = x1 /\ y2 = x2).
Proof.
  introv [n [Hn1 Hn2]].
  eapply LibList.Nth_cons_inv in Hn1; destruct_all;
    eapply LibList.Nth_cons_inv in Hn2; destruct_all; subst; eauto;
    try omega. replace x0 with x in * by omega. firstorder.
Qed.

Lemma simlist_relates_push x1 x2 (xs1 xs2: simlist) y1 y2:
  simlist_relates xs1 xs2 y1 y2  ->
  simlist_relates (x1::xs1)%list (x2::xs2)%list  y1 y2.
Proof.
  introv Hr.
  unfold simlist_relates in *.
  destructc Hr as ? ? ?.
  exists (S x). eauto.
Qed.
Lemma simlist_relates_push_l x1 x2 (xs1 xs2: simlist):
  simlist_relates (x1::xs1)%list (x2::xs2)%list x1 x2.
Proof.
  unfold simlist_relates in *.
  exists (0). eauto.
Qed.
Lemma simlist_relates_append zs1 zs2 (xs1 xs2: simlist) y1 y2:
  length zs1 = length zs2 ->
  simlist_relates xs1 xs2 y1 y2  ->
  simlist_relates (zs1++xs1)%list (zs2++xs2)%list  y1 y2.
Proof.
  introv Hl Hr.
  unfold simlist_relates in *.
  destructc Hr as ? ? ?.
  exists ((length zs1) + x).
  rewrite <- liblist_app.
  rewrite Nat.add_comm.
  rewrite Hl at 2.
  rewrite <- liblist_length.
  split;
  eapply LibList.Nth_app_r'; eauto.
Qed.

Lemma simlist_relates_nil_inv a b:
  simlist_relates nil nil a b ->
  False.
Proof.
  intros.
  unfold simlist_relates in H.
  destruct_all.
  eapply LibList.Nth_nil_inv; eauto.
Qed.
Lemma simlist_relates_noduplicates_var_eq xs1 xs2 (y1 y2 z2: var):
  fresh \{} (length xs1) xs1 ->
  simlist_relates xs1 xs2 y1 y2 ->
  simlist_relates xs1 xs2 y1 z2 ->
  (y2 = z2).
Proof.
  intros.
    unfold simlist_relates in *.
    destruct_all.
    lets: fresh_nth_eq H0 H1; eauto.
    subst. lets: LibList.Nth_functional H3 H2. auto.
Qed.
Lemma simlist_relates_noduplicates_var_cases xs1 xs2 (y1 y2 z1 z2: var):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  simlist_relates xs1 xs2 y1 y2 ->
  simlist_relates xs1 xs2 z1 z2 ->
  (y1 = z1) <-> (y2 = z2).
Proof.
  intros.
  split; intros.
  - subst.
    unfold simlist_relates in *.
    destruct_all.
    lets: fresh_nth_eq H1 H2; eauto. subst. lets: LibList.Nth_functional H4 H3. auto.
  - subst.
    unfold simlist_relates in *.
    destruct_all.
    lets: fresh_nth_eq H4 H3; eauto. subst. lets: LibList.Nth_functional H1 H2. auto.
Qed.
Lemma simlist_relates_swap xs1 xs2 v1 v2:
  simlist_relates xs1 xs2 v1 v2 ->
  simlist_relates xs2 xs1 v2 v1.
Proof.
  unfold simlist_relates.
  intros. destruct_all; eauto.
Qed.

(* Similarity of two elements using a relation specified by two variable lists *)
Class Similar (A: Type) := similar: simlist -> simlist -> A -> A -> Prop.
(* Similarity of two elements using two additional variable lists, one relation is for values, one for keys *)
Class SimilarWith (A: Type) := similar_with: simlist -> simlist -> simlist -> simlist -> A -> A -> Prop.

Class SimilarCommon {A: Type} (AS: Similar A) := {
  (* Commutative *)
  similar_symm xs ys a b:
    similar xs ys a b ->
    similar ys xs b a ;
  (* Transitive *)
  similar_trans xs ys zs a b c:
    fresh \{} (length ys) ys ->
    similar xs ys a b ->
    similar ys zs b c ->
    similar xs zs a c ;
  (* Length *)
  similar_length xs ys a b:
    similar xs ys a b ->
    length xs = length ys ;
  (* We can change the order of variables *)
  similar_reorder xs1 ys1 xs2 ys2 ns a b:
    similar xs1 ys1 a b ->
    list_reorder xs1 ns xs2 ->
    list_reorder ys1 ns ys2 ->
    LibList.noduplicates ns ->
    length ns = length xs1 ->
    similar xs2 ys2 a b ;
}.

(* For terms, types *)
Class SimilarOut {A: Set} (AS: Similar A) {AF: FreeVar A} := {
  (* If terms are similar with no variables related, then the terms are equal *)
  similar_by_same_eq xs a b:
    similar xs xs a b ->
    a = b ;
  (* Equal terms are similar by no variables related. *)
  similar_refl_by_fv xs a:
    (fv a) \c (from_list xs) ->
    similar xs xs a a ;
  (* The relation can be extended with any pair of variables *)
  similar_extend x xs y ys a b:
    similar xs ys a b ->
    similar (x::xs) (y::ys) a b ;
  (* Free variables *)
  similar_out_fvl xs ys a b:
    similar xs ys a b ->
    fv a \c (from_list xs) ;
  similar_out_fvr xs ys a b:
    similar xs ys a b ->
    fv b \c (from_list ys) ;
  (* Remove non-occuring variables *)
  similar_out_shorten_l x xs y ys a b:
    x \notin fv a ->
    similar (x::xs) (y::ys) a b ->
    similar xs ys a b
}.

Unset Auto Template Polymorphism.
Class Envy (E: Type) := {
  envy_ok (e: E): Prop;
  envy_dom (e: E): vars;
  envy_fv (e: E): vars;
  envy_concat(e1 e2: E): E
}.

Instance EnvyCtx: Envy ctx :={
  envy_ok:= ctx_ok;
  envy_dom:= ctx_dom;
  envy_fv := ctx_fv;
  envy_concat := ctx_concat
}.
Instance EnvyEnv (A: Set) (AF: FreeVar A) : Envy (env A):={
  envy_ok:= @ok A;
  envy_dom:= @dom A;
  envy_fv := @fv_in_values A fv;
  envy_concat := @concat A
}.

(* For heap, renv *)
Class SimilarWithIn (E: Type) (En: Envy E) (AS: SimilarWith E):= {
  similar_with_o_push y1 ys1 y2 ys2 (S1 S2: E) xs1 xs2:
    similar_with ys1 ys2 xs1 xs2 S1 S2 ->
    similar_with (y1::ys1) (y2::ys2) xs1 xs2 S1 S2 ;
  similar_with_i_append_invl ys1 ys2 (S1 S2: E) xs11 xs12 xs21 xs22:
    length xs11 = length xs21 ->
    similar_with ys1 ys2 (xs11 ++ xs12) (xs21 ++ xs22) S1 S2 ->
    similar_with ys1 ys2 (xs11) (xs21) S1 S2 ;
  similar_with_i_reorder (a1 a2: E) ns (xs1 xs2 ys1 ys2 zs1 zs2: list var):
    list_reorder xs1 ns zs1 ->
    list_reorder xs2 ns zs2 ->
    LibList.noduplicates ns ->
    length xs1 = length ns ->
    length xs2 = length ns ->
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with ys1 ys2 zs1 zs2 a1 a2 ;
  similar_with_o_reorder (a1 a2: E) ns (ys1 ys2 zs1 zs2: list var):
    list_reorder ys1 ns zs1 ->
    list_reorder ys2 ns zs2 ->
    LibList.noduplicates ns ->
    length ys1 = length ns ->
    length ys2 = length ns ->
    forall (xs1 xs2: list var),
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with zs1 zs2 xs1 xs2 a1 a2 ;
  similar_with_i_push_notin ys1 ys2 xs1 xs2 x1 x2 S1 S2 :
    similar_with ys1 ys2 xs1 xs2 S1 S2 ->
    (x1 \notin envy_dom S1) ->
    (x2 \notin envy_dom S2) ->
    similar_with ys1 ys2 (x1::xs1) (x2::xs2) S1 S2
}.

Class SimilarIn (E: Type) (En: Envy E) (AS: Similar E) := {
  similar_weaken xs ys a1 a2 b1 b2:
    envy_ok (envy_concat a1 a2) ->
    envy_ok (envy_concat b1 b2) ->
    (forall x y, simlist_relates xs ys x y -> (x \in envy_dom a2 <-> y \in envy_dom b2)) ->
    similar xs ys (envy_concat a1 a2) (envy_concat b1 b2) ->
    similar xs ys (a1) (b1) ;
  similar_in_refl_by_fv a xs:
    envy_ok a ->
    (envy_fv a) \c (from_list xs) ->
    similar xs xs a a;
  similar_append_fresh xs1 xs2 (S1 S2: E) zs1 zs2:
    similar xs1 xs2 S1 S2 ->
    fresh (envy_dom S1) (length zs1) zs1 ->
    fresh (envy_dom S2) (length zs2) zs2 ->
    length zs1 = length zs2 ->
    similar (zs1 ++ xs1) (zs2 ++ xs2) (S1) (S2)
}.


Class SimilarEnvIn (A: Set) (AS: Similar (env A)) := {
  (* Any two environments are similar by no variables related *)
  similar_by_nil a b:
    similar nil nil a b ;
  similar_extend_notin x xs y ys a b:
    similar xs ys a b ->
    x # a ->
    y # b ->
    similar (x::xs) (y::ys) a b ;
  similar_extend_dom x xs y ys a b c d:
    similar xs ys a b ->
    similar (x::xs) (y::ys) (x ~ c) (y ~ d) ->
    binds x c a ->
    binds y d b ->
    similar (x::xs) (y::ys) a b ;
  similar_swap_after xs ys a1 a2 a3 b1 b2 b3:
    ok (a1 & a2 & a3) ->
    ok (b1 & b2 & b3) ->
    similar xs ys (a1 & a2 & a3) (b1 & b2 & b3) ->
    similar xs ys (a1 & a3 & a2) (b1 & b3 & b2) ;
  similar_concat xs ys a1 b1 a2 b2:
    similar xs ys (a1) (b1) ->
    similar xs ys (a2) (b2) ->
    similar xs ys (a1 & a2) (b1 & b2)
}.

(* INSTANCES *)

Definition similar_by_subst{A: Set}{AS: SubstVar A}{AF: FreeVar A} (xs1 xs2: simlist) (a1 a2: A) :=
  exists a3 xs3,
    (length xs1) = (length xs3) /\
    (length xs2) = (length xs3) /\
    (fresh ((fv a1) \u (fv a2) \u (from_list xs1) \u (from_list xs2)) (length xs3) xs3) /\
    (fv a3) \c (from_list xs3) /\
    a3 /[xs3 --> xs1] = a1 /\
    a3 /[xs3 --> xs2] = a2.

Section SimilarBySubst.
Context  {A: Set}{AS: SubstVar A}{AF: FreeVar A}{AO: Openable A}{AAS: AbstractSyntax AO AF AS}{ASC: SubstCommut AAS}{AFC: FvSubstCases AAS}{ASF: SubstFresh AAS}.
(* Define similarity for A by subst just in this section *)
Let SA: Similar A := similar_by_subst.
Local Existing Instance SA.

Lemma similar_by_subst_L (xs1 xs2: simlist) (a1 a2: A) L:
  similar_by_subst xs1 xs2 a1 a2 ->
  exists a3 xs3,
    (length xs1) = (length xs3) /\
    (length xs2) = (length xs3) /\
    (fresh ((fv a1) \u (fv a2) \u (from_list xs1) \u (from_list xs2) \u L) (length xs3) xs3) /\
    (fv a3) \c (from_list xs3) /\
    a3 /[xs3 --> xs1] = a1 /\
    a3 /[xs3 --> xs2] = a2.
Proof.
  unfold similar_by_subst.
  intros.
  destructc H as a3 x3 ? ? ? Hfv Hrest. destructc Hrest as Ha Hb.
  sets L': (fv a1 \u fv a2 \u fv a3 \u from_list xs1 \u from_list xs2 \u from_list x3 \u L).
  destructc (var_freshes (L') (length x3)) as x0 Hx0.

  pose proof (fresh_length _ _ _ Hx0).
    eapply fresh_resize_length in Hx0.
    exists (a3 /[x3 --> x0]) (x0)%list.
    sympls.
    unfold L' in *.
    splits.
    + rewrite H.
      rewrite H2. auto.
    +
      rewrite H0.
      rewrite H2. auto.
    + fresh_solve.
    + eapply subst_vars_fv_subset; eauto.
    + assert (length xs1 = length x0). rewrite H2. auto.
      rewrite subst_vars_merge; eauto.
    + assert (length xs2 = length x0). rewrite H2. auto.
      rewrite subst_vars_merge; eauto.
Qed.
Lemma similar_by_subst_by_L (xs1 xs2 xs3: simlist) (a1 a2: A) L:
  similar_by_subst xs1 xs2 a1 a2 ->
  (length xs1) = (length xs3) ->
  (length xs2) = (length xs3) ->
  (fresh ((fv a1) \u (fv a2) \u (from_list xs1) \u (from_list xs2) \u L) (length xs3) xs3) ->
  exists a3,
    (fv a3) \c (from_list xs3) /\
    a3 /[xs3 --> xs1] = a1 /\
    a3 /[xs3 --> xs2] = a2.
Proof.
  introv Hs Hl1 Hl2 Hfr.

  forwards HsL: (similar_by_subst_L) (L \u from_list xs3) Hs; eauto.
  destructc HsL as a3 xs4 Hrest.
  destruct_all.
  exists (a3/[ xs4 --> xs3]).
  splits; eauto.
  - eapply subst_vars_fv_subset; eauto.
    eapply fresh_swap; eauto. rewrite <- H. rewrite <- Hl1.
    auto.
  - rewrite subst_vars_merge; eauto.
    + rewrite <-H. rewrite <- Hl1. auto.
    + fresh_solve. eapply fresh_subset; eauto. eapply fresh_swap; eauto.
      eapply fresh_swap; eauto.
  - rewrite subst_vars_merge; eauto.
    + rewrite <-H. rewrite <- Hl1. auto.
    + fresh_solve. eapply fresh_subset; eauto. eapply fresh_swap; eauto.
      eapply fresh_swap; eauto.
Qed.

End SimilarBySubst.

Ltac destruct_similar_by_subst L Hsimv v:=
  let v3 := fresh v "3" in
  let xv3 := fresh "x" v "3" in
  let Hav := fresh "Ha" v in
  let Hbv := fresh "Hb" v in
  let HsimvL := fresh "Hsim" v "L" in
  pose proof (similar_by_subst_L L Hsimv) as HsimvL;
  destructc HsimvL as v3 xv3 ? ? ? ? Hav Hbv.

Ltac destruct_similar_by_subst_by xs L Hsimv v:=
  let v3 := fresh v "3" in
  let Hav := fresh "Ha" v in
  let Hbv := fresh "Hb" v in
  let HsimvL := fresh "Hsim" v "L" in
  forwards HsimvL: (similar_by_subst_by_L xs L Hsimv);
  [> eauto ..| destructc HsimvL as v3 ? Hav Hbv ].

Section SimilarBySubst.
Context  {A: Set}{AS: SubstVar A}{AF: FreeVar A}{AO: Openable A}{AAS: AbstractSyntax AO AF AS}{ASC: SubstCommut AAS}{AFC: FvSubstCases AAS}{ASF: SubstFresh AAS}.
(* Define similarity for A by subst just in this section *)
Let SA: Similar A := similar_by_subst.
Local Existing Instance SA.

Lemma similar_by_subst_extend xs ys x y (a b: A):
  similar_by_subst xs ys a b ->
  similar_by_subst (x :: xs) (y :: ys) a b.
Proof.
  introv H.
    sets L: (fv a \u fv b \u from_list xs \u from_list ys \u \{x} \u \{y}).
    pick_fresh_with z L.
    pose proof (similar_by_subst_L (L \u \{z}) H) as H0.
    destructc H0 as a3 x3 ? ? Hfra3 Hfva3 Hrest; destructc Hrest as Ha Hb.

    exists (a3) (z::x3)%list.
    sympls.
    unfold L in *.
    splits; eauto.
    + rewrite ? from_list_cons. split. notin_solve. fresh_solve.
    + rewrite from_list_cons. intro. intro. rewrite in_union. auto.
    + rewrite Ha.
      rewrite subst_fresh; eauto.
    + rewrite Hb.
      rewrite subst_fresh; eauto.
Qed.
Lemma similar_by_subst_length xs ys (a b: A):
  similar_by_subst xs ys a b ->
  length xs = length ys.
Proof.
  introv Hsi.
  unfold similar_by_subst in *.
  destruct_all. rewrite H. rewrite H0. auto.
Qed.


Lemma similar_by_subst_trans {ASI: SubstId AAS} (xs ys zs: list var) (a b c: A):
  fresh \{} (length ys) ys ->
  similar xs ys a b ->
  similar ys zs b c ->
  similar xs zs a c.
Proof.
  introv Hfr Hsiab Hsibc.
  destruct_similar_by_subst (fv c \u (from_list zs)) Hsiab ab.
  assert (length zs = length xab3). {
    unfold similar, SA, similar_by_subst in *. destruct_all. rewrite H5. rewrite <- H0. auto.
  }
  destruct_similar_by_subst_by xab3 (fv a) Hsibc bc.

  unfold similar, similar_by_subst.
  exists ab3 xab3; splits; eauto.

  assert (ab3 = (b /[ys --> xab3])).
    eapply subst_vars_inverse; eauto. fresh_solve.
  assert (bc3 = (b /[ys --> xab3])).
    eapply subst_vars_inverse; eauto. fresh_solve.

  rewrite H5. rewrite <- H6. auto.
Qed.

Lemma similar_by_subst_symm (xs ys: list var) (a b: A):
  similar xs ys a b ->
  similar ys xs b a.
Proof.
  introv Hsi.
  unfold similar, SA, similar_by_subst in *; destruct_all. exists x x0. eauto 6.
Qed.

Lemma similar_by_subst_noduplicates_eq_l {ASI: SubstId AAS} xs ys (a b1 b2: A):
  fresh \{} (length ys) ys ->
  fresh \{} (length xs) xs ->
  similar xs ys a b1 ->
  similar xs ys a b2 ->
  b1 = b2.
Proof.
  introv Hfry Hfrx Hsi1 Hsi2.

  destruct_similar_by_subst (fv b2) Hsi1 ab.
  destruct_similar_by_subst_by xab3 (fv b1) Hsi2 bc.

  assert (ab3 = bc3).
  eapply subst_vars_unique with (ys0:=xs) (xs0:=xab3) (b:=a); eauto. eauto. eauto. eauto. eauto. eauto.
  subst; eauto.
Qed.
Lemma similar_by_subst_noduplicates_eq_r {ASI: SubstId AAS} xs ys (a1 a2 b: A):
  fresh \{} (length ys) ys ->
  fresh \{} (length xs) xs ->
  similar xs ys a1 b ->
  similar xs ys a2 b ->
  a1 = a2.
Proof.
  introv Hfry Hfrx Hsi1 Hsi2.
  eapply similar_by_subst_noduplicates_eq_l with (xs := ys) (ys:=xs) (a:=b); eauto using similar_by_subst_symm.
Qed.
End SimilarBySubst.

Section SimilarBySubst.
Context  {A: Set}{AS: SubstVar A}{AF: FreeVar A}{AO: Openable A}{AAS: AbstractSyntax AO AF AS}{ASC: SubstCommut AAS}{AFC: FvSubstCases AAS}{ASF: SubstFresh AAS}{ASI: SubstId AAS}.
Global Instance similar_common_by_subst: SimilarCommon similar_by_subst.
Proof.
  split; intros.
  - (* similar_out_comm *)
    unfold similar, similar_by_subst in *.
    destruct_all. exists x x0. auto 6.
  - (* similar_out_trans *)
    eapply similar_by_subst_trans; eauto.
  - (* similar_out_length *)
    eapply similar_by_subst_length; eauto.
  - (* similar_out_reorder *)
    unfold similar, similar_by_subst in *.
    destructc H as a3 x3 ? ? ? ? Hrest; destructc Hrest as Ha Hb.
    forwards xs4 Hxs4: list_reorder_same_length_exists; eauto.
    exists a3 xs4.
    splits.
    + eapply list_reorder_lengths in H0.
      eapply list_reorder_lengths in Hxs4.
      rewrite H0. rewrite Hxs4. auto.
    + eapply list_reorder_lengths in H1.
      eapply list_reorder_lengths in Hxs4.
      rewrite H1. rewrite Hxs4. auto.
    + eapply fresh_reorder; eauto. fresh_solve; eapply fresh_reorder_l; eauto.
    + erewrite <- list_reorder_from_list; eauto. rewrite H3. rewrite H. auto.
    + rewrite <- Ha. symmetry. eapply subst_vars_permutation; eauto. rewrite H3. symmetry. auto.
    + rewrite <- Hb. symmetry. eapply subst_vars_permutation; eauto. rewrite H3. symmetry. auto. rewrite H4. rewrite <- H. symmetry. auto.
Qed.


Global Instance similar_out_by_subst: SimilarOut similar_by_subst.
Proof.
  split; intros.
  - (* similar_by_same_eq *)
    unfold similar, similar_by_subst in H.
    destructc H as a3 x3 ? ? ? Hfv Hrest. destructc Hrest as Ha Hb. rewrite <- Ha. rewrite <- Hb. auto.
  - (* similar_refl_by_fv *)
    unfold similar, similar_by_subst in *.
    lets He: subst_vars_fresh_exists H.
    destructc He as b ys Hfry Hl Hfvb Hb.
    exists b ys.
    splits; sympls; eauto. fresh_solve; eapply fresh_subset; eauto.
  - (* similar_extend *)
    eapply similar_by_subst_extend; eauto.
  - (* similar_out_fvl *)
    unfold similar, similar_by_subst in *.
    destruct_all; subst.
    eapply subset_trans.
    eapply subst_vars_fv_subset'; eauto. eapply subset_refl.
  - (* similar_out_fvr *)
    unfold similar, similar_by_subst in *.
    destruct_all; subst.
    eapply subset_trans.
    eapply subst_vars_fv_subset'; eauto. eapply subset_refl.
  - (* similar_out_shorten *) pose proof True as H1.
    unfold similar, similar_by_subst in *.
    destruct_all; subst.
    destruct x1; sympls; try discriminate.
    destruct_all.
    rewrite ! from_list_cons in *.

    assert (v \notin fv (x0 /[ x1 --> xs])) as Hvnx. {
      intro. apply H.
      forwards: fv_subst_cases v x (x0 /[ x1 --> xs]). destruct_all; subst.
      + rewrite subst_id. auto.
      + contradiction.
      + rewrite H9. rewrite in_union. rewrite in_singleton. auto.
    }

    assert (v \notin fv x0). {
      intro. apply Hvnx. eapply in_fv_subst_vars; eauto. eapply from_list_fresh; eauto.
    }

    exists x0 x1; splits; eauto.
    + intro. intro Hin.
      assert (x2 <> v) by (intro; subst; contradiction).
      eapply H4 in Hin. rewrite in_union in Hin. destruct Hin. rewrite in_singleton in H8; contradiction. auto.
    +
      rewrite subst_fresh; eauto.
    +
      rewrite subst_fresh; eauto.
      eapply notin_fv_subst_vars; eauto.
Qed.
End SimilarBySubst.

Section SimilarCommon.
  Context {A: Set} {AS: Similar A} {AF: FreeVar A} {ASC: SimilarCommon AS}.
  Lemma similar_reorder_inv xs1 ys1 xs2 ys2 ns a b:
    similar xs2 ys2 a b ->
    list_reorder xs1 ns xs2 ->
    list_reorder ys1 ns ys2 ->
    LibList.noduplicates ns ->
    length ns = length xs1 ->
    length ns = length ys1 ->
    similar xs1 ys1 a b.
  Proof.
    introv Hs Hrex Hrey Hnd Hlx Hly.
    forwards Hinv: list_reorder_inverse Hnd.
    rewrite Hlx.
    eapply list_reorder_gt; eauto.
    destructc Hinv as ns2 ? ? ? ?.
    eapply similar_reorder; eauto.
    eapply list_reorder_lengths in Hrex. rewrite H0. eauto.
  Qed.
  
  Lemma similar_append_swap xs1 ys1 xs2 ys2 a b:
    similar (xs1 ++ xs2) (ys1 ++ ys2) a b ->
    length xs1 = length ys1 ->
    similar (xs2 ++ xs1) (ys2 ++ ys1) a b.
  Proof.
    introv Hsi Hl.
    assert (length xs2 = length ys2). {
      forwards: similar_length Hsi.
      rewrite ! List.app_length in H.
      rewrite Hl in H.
      eapply PeanoNat.Nat.add_cancel_l. eauto.
    }
  
    forwards: list_reorder_swap_app_n (length ys1) (length ys2).
    destructecs H0.
    forwards: H0 xs1 xs2; fweauto. 
    forwards: H0 ys1 ys2; fweauto.
    eapply similar_reorder; eauto.
      rewrite ! List.app_length . rewrite H, Hl. auto.
  Qed.

End SimilarCommon.

Section SimilarOut.
  Context {A: Set} {AS: Similar A} {AF: FreeVar A} {ASC: SimilarCommon AS} {ASO: SimilarOut AS}.

  Lemma similar_out_shorten_ls xs1 xs2 ys1 ys2 a b :
    (length xs1) = (length ys1) ->
    disjoint (from_list xs1) (fv a) ->
    similar (xs1++xs2) (ys1++ys2) a b ->
    similar xs2 ys2 a b.
  Proof.
    reverts ys1. induction xs1.
    - simpl. destruct ys1; simpl; introv Hlen Hd Hsim; try omega; eauto.
    - destruct ys1; simpl; introv Hlen Hd Hsim; try omega.
      inversions Hlen. rewrite from_list_cons in Hd.
      eapply similar_out_shorten_l in Hsim.
      2: { introv Hin. erewrite <- in_empty, <- Hd, in_inter, in_union,
          in_singleton. splits; eauto. }
      eapply disjoint_union_l_inv in Hd. eapply IHxs1; eauto; tauto.
  Qed.

  Lemma  similar_out_shorten_r x xs y ys a b:
    y \notin fv b ->
    similar (x::xs) (y::ys) a b ->
    similar xs ys a b.
  Proof.
    introv Hn Hs.
    eapply similar_symm.
    eapply similar_out_shorten_l. eauto. eapply similar_symm; eauto.
  Qed.

  Lemma similar_out_shorten_rs xs1 xs2 ys1 ys2 a b :
    (length xs1) = (length ys1) ->
    disjoint (from_list ys1) (fv b) ->
    similar (xs1++xs2) (ys1++ys2) a b ->
    similar xs2 ys2 a b.
  Proof.
    reverts xs1. induction ys1.
    - simpl. destruct xs1; simpl; introv Hlen Hd Hsim; try omega; eauto.
    - destruct xs1; simpl; introv Hlen Hd Hsim; try omega.
      inversions Hlen. rewrite from_list_cons in Hd.
      eapply similar_out_shorten_r in Hsim.
      2: { introv Hin. erewrite <- in_empty, <- Hd, in_inter, in_union,
          in_singleton. splits; eauto. }
      eapply disjoint_union_l_inv in Hd. eapply IHys1; eauto; tauto.
  Qed.

  Lemma similar_out_shorten_fresh_suffix_l xs1 xs2 ys1 ys2 a b:
    length ys2 = length xs2 ->
    fresh (fv a) (length xs2) xs2 ->
    similar (xs1++xs2) (ys1++ys2) a b ->
    similar xs1 ys1 a b.
  Proof.
    introv Hl Hfra Hs.
    reverts ys2 xs1 ys1.
    induction xs2; intros; destruct ys2; simpls; try discriminate.
    - rewrite ! List.app_nil_r in *. auto.
    - inversions Hl.
      assert (length ys1 = length xs1). {
        forwards: similar_length Hs.
        rewrite ! List.app_length in H. simpl in H.
        rewrite H0 in H. rewrite Nat.add_comm in H.  rewrite (Nat.add_comm (length ys1)) in H. eapply Plus.plus_reg_l in H. eauto.
      }

      destruct_all.
      eapply similar_out_shorten_l. eauto.
      eapply IHxs2 with (ys2 := ys2). eauto. eauto.
      rewrite <- ! List.app_comm_cons.
      forwards ns Hnd Hlns Hf Hre: (list_reorder_shift_app_n (length xs1) (length xs2)).
      rewrite <- Hlns in Hf.
      forwards Hinv: list_reorder_inverse Hnd Hf.
      destructc Hinv as nsi Hndi Hlnsi Hfi Hrei.
      eapply similar_reorder.
      eauto.
      eapply Hrei. simpls. rewrite List.app_length. eauto.
      eapply Hre; eauto.
      eapply Hrei. simpls. rewrite List.app_length. rewrite H0, H. eauto.
      eapply Hre; eauto.
      eauto.
      rewrite List.app_length. simpls. rewrite Hlnsi.
      rewrite PeanoNat.Nat.add_succ_r .
      eauto.
  Qed.

  Lemma similar_out_shorten_fresh_suffix_r xs1 xs2 ys1 ys2 a b:
    length ys2 = length xs2 ->
    fresh (fv b) (length ys2) ys2 ->
    similar (xs1++xs2) (ys1++ys2) a b ->
    similar xs1 ys1 a b.
  Proof.
    introv Hl Hn Hs.
    eapply similar_symm.
    eapply similar_symm in Hs.
    eapply similar_out_shorten_fresh_suffix_l. symmetry. eauto. eauto. eauto.
  Qed.

  Lemma similar_out_append x xs y ys a b:
    similar xs ys a b ->
    length x = length y ->
    similar (x++xs)%list (y++ys)%list a b.
  Proof.
    introv Hsi Hl.
    reverts y.
    induction x; intros; destruct y; try discriminate.
    - rewrite ! List.app_nil_l. auto.
    - rewrite  <- ! List.app_comm_cons.
      eapply similar_extend. eauto.
  Qed.
  Lemma similar_out_append_r x xs y ys a b:
    similar xs ys a b ->
    length x = length y ->
    similar (xs++x)%list (ys++y)%list a b.
  Proof.
    introv Hsi Hl.

    forwards ns Hnd: (list_reorder_swap_app_n (length x) (length xs)); eauto.
    destruct_all.
    forwards: H2 x xs; eauto.
    forwards: H2 y ys; eauto. symmetry.
    eapply similar_length; eauto.

    eapply similar_reorder.
    eapply similar_out_append; eauto.
    all: eauto. rewrite List.app_length; eauto.
  Qed.
  Lemma similar_out_append_middle x xs xs1 y ys ys1 a b:
    similar (x++xs)%list (y++ys)%list a b ->
    length x = length y ->
    length xs1 = length ys1 ->
    similar (x++xs1++xs)%list (y++ys1++ys)%list a b.
  Proof.
    introv Hsi Hl Hl1.

    eapply similar_append_swap.
    rewrite <- ! List.app_assoc.
    eapply similar_out_append.
    eapply similar_append_swap. auto.
    auto. auto.
    eapply similar_length in Hsi.
    rewrite ! List.app_length in *.
    rewrite Hl1.
    eapply PeanoNat.Nat.add_cancel_l.
    rewrite Hl in Hsi.
    eapply PeanoNat.Nat.add_cancel_l. eauto.
  Qed.


End SimilarOut.

Section SimilarWithIn.

Context {A: Type}{E: Envy A}{SW: SimilarWith A}{SWI: SimilarWithIn E SW}.

Lemma similar_with_o_append_r zs1 ys1 zs2 ys2 (S1 S2: A) xs1 xs2:
  length zs1 = length zs2 ->
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar_with (zs1 ++ ys1) (zs2 ++ ys2) xs1 xs2 S1 S2.
Proof.
  introv Hl Hs.
  reverts zs2.
  induction zs1; intros; destruct zs2; inversions Hl.
  - rewrite ! List.app_nil_l. auto.
  - rewrite <- ! List.app_comm_cons.
    eapply similar_with_o_push; eauto.
Qed.

Lemma similar_with_i_append_fresh_r  xs1 xs2 (S1 S2: A) zs1 zs2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  fresh (envy_dom S1) (length zs1) zs1 ->
  fresh (envy_dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar_with ys1 ys2 (zs1++xs1) (zs2++xs2) S1 S2.
Proof.
  introv Hsi Hfr1 Hfr2 Hl.
  reverts zs2; induction zs1; intros; destruct zs2; inversions Hl.
  - rewrite List.app_nil_l; auto.
  - rewrite <- ! List.app_comm_cons.
    inversions Hfr1.
    eapply fresh_union_r1 in H1.
    inversions Hfr2.
    eapply fresh_union_r1 in H3.
    eapply similar_with_i_push_notin; eauto.
Qed.


End SimilarWithIn.


(* Similarity for heap.
   If both variables are not in the domain, the this relation does not care about that pair of variables.
   That is because reference variables are in domain of renv, not heap.
   To ensure that there are no extra variables, say (from_list xs \c (dom S \u dom e)).
  This is by induction on the similarity lists, not on the heap.
*)
Inductive similar_env_with{A: Set}(AS: Similar A): SimilarWith (env A) (*simlist -> simlist -> simlist -> simlist -> (env A) -> (env A) -> Prop *):=
| similar_env_empty ys1 ys2 S1 S2:
    length ys1 = length ys2 ->
    similar_with ys1 ys2 nil nil S1 S2
| similar_env_dom ys1 ys2 S1 S2 xs1 xs2 x1 x2 it1 it2:
    @similar_with (env A) (@similar_env_with A AS) ys1 ys2 xs1 xs2 S1 S2 ->
    (binds x1 it1 S1) ->
    (binds x2 it2 S2) ->
    similar ys1 ys2 it1 it2 ->
    similar_with ys1 ys2 (x1::xs1) (x2::xs2) S1 S2
| similar_env_notin ys1 ys2 S1 S2 xs1 xs2 x1 x2:
    @similar_with (env A) (@similar_env_with A AS) ys1 ys2 xs1 xs2 S1 S2 ->
    (x1 # S1) ->
    (x2 # S2) ->
    similar_with ys1 ys2 (x1::xs1) (x2::xs2) S1 S2.

Section SimilarEnvWith.
Context {A: Set}{AS: Similar A}{AF: FreeVar A}{ASC: SimilarCommon AS}{ASO: SimilarOut AS}.

Let SW: SimilarWith (env A) := similar_env_with AS.
Local Existing Instance SW.

Definition similar_env x1 x2 a1 a2 := similar_with x1 x2 x1 x2 a1 a2.

(*
Naming scheme:
- prefix: similar_env_with
- relevant part: o/i/e, r/l
  - o = out, first pair of lists
  - i = in, second pair of lists
  - e = element, pair of environments
  - r = right, second of the pair
  - l = left, firts of the pair
*)

Lemma similar_env_with_o_push y1 ys1 y2 ys2 (S1 S2: env A) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar_with (y1::ys1) (y2::ys2) xs1 xs2 S1 S2.
Proof.
  introv Hs.
  induction Hs.
  - eapply similar_env_empty. simpl. rewrite H. auto.
  - eapply similar_env_dom; eauto. eapply similar_extend; eauto.
  - eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_e_swap_after ys1 ys2 (a1 a2 a3 b1 b2 b3: env A) xs1 xs2:
  ok (a1 & a2 & a3) ->
  ok (b1 & b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a2 & a3) (b1 & b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a3 & a2) (b1 & b3 & b2).
Proof.
  introv Hoka Hokb Hs.
  remember (a1 & a2 & a3) as a. remember (b1 & b2 & b3) as b.
  induction Hs; intros; subst.
  - eapply similar_env_empty. auto.
  - eapply similar_env_dom; eauto.
    + destruct (binds_concat_inv H0).
      * eapply binds_concat_left_ok; eauto using ok_swap_after.
      * destruct H3. eauto using binds_weaken2.
    + destruct (binds_concat_inv H1).
      * eapply binds_concat_left_ok; eauto using ok_swap_after.
      * destruct H3.  eauto using binds_weaken2.
  - eapply similar_env_notin; eauto.
Qed.
Lemma similar_env_with_er_swap_after ys1 ys2 (a b1 b2 b3: env A) xs1 xs2:
  ok (b1 & b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 a (b1 & b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 a (b1 & b3 & b2).
Proof.
  introv Hokb Hs.
  remember (b1 & b2 & b3) as b.
  induction Hs; intros; subst.
  - eapply similar_env_empty. auto.
  - eapply similar_env_dom; eauto.
    + destruct (binds_concat_inv H1).
      * eapply binds_concat_left_ok; eauto using ok_swap_after.
      * destruct H3.  eauto using binds_weaken2.
  - eapply similar_env_notin; eauto.
Qed.
Lemma similar_env_with_e_swap ys1 ys2 (a2 a3 b2 b3: env A) xs1 xs2:
  ok (a2 & a3) ->
  ok (b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 (a2 & a3) (b2 & b3) ->
  similar_with ys1 ys2 xs1 xs2 (a3 & a2) (b3 & b2).
Proof.
  introv Hoka Hokb Hs.
  forwards: (similar_env_with_e_swap_after).
  rewrite concat_empty_l. eapply Hoka.
  rewrite concat_empty_l. eapply Hokb.
  rewrite ! concat_empty_l. eapply Hs.
  rewrite ! concat_empty_l in H. eapply H.
Qed.


Lemma similar_env_with_i_middle_inv (a1 a2: env A) (ys1 ys2 zsl zsr wsl wsr: list var) w z:
    length wsl = length zsl ->
    length wsr = length zsr ->
    similar_with ys1 ys2 (zsl ++ z :: zsr) (wsl ++ w :: wsr) a1 a2 -> (
      (similar_with ys1 ys2 (zsl ++ zsr) (wsl ++ wsr) a1 a2) /\
      ((exists b1 b2,
          binds z b1 a1 /\
          binds w b2 a2 /\
          similar ys1 ys2 b1 b2
        )\/(
          z#a1 /\ w#a2
      ))) .
Proof.
  intros.
  reverts wsl.
  induction zsl; intros; destruct wsl; simpls; try discriminate.
  - inversions H1.
    +   splits. auto. left. exists it1 it2. auto.
    +   splits. auto. auto.
  - inversions H.
    inversions H1.
    + splits. eapply similar_env_dom; eauto. eapply IHzsl; auto. eapply IHzsl; eauto.
    + splits. eapply similar_env_notin; eauto. eapply IHzsl; auto. eapply IHzsl; eauto.
Qed.

Lemma similar_env_with_i_middle_binds_push (a1 a2: env A) (ys1 ys2 zsl zsr wsl wsr: list var) w z it1 it2 :
  length wsl = length zsl ->
  length wsr = length zsr ->
  similar_with ys1 ys2 (zsl ++ zsr) (wsl ++ wsr) a1 a2 ->
  binds z it1 a1 ->
  binds w it2 a2 ->
  similar ys1 ys2 it1 it2 ->
  similar_with ys1 ys2 (zsl ++ z :: zsr) (wsl ++ w :: wsr) a1 a2.
Proof.
  introv HlL HrL Hsim Hbi1 Hbi2 Hsit. reverts wsl.
  induction zsl; intros; destruct wsl; simpls; try discriminate; eauto.
  eapply similar_env_dom; eauto.
  inversions HlL. inversions Hsim.
  - eapply similar_env_dom; eauto.
  - eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_i_middle_notin_push (a1 a2: env A) (ys1 ys2 zsl zsr wsl wsr: list var) w z it1 it2 :
  length wsl = length zsl ->
  length wsr = length zsr ->
  similar_with ys1 ys2 (zsl ++ zsr) (wsl ++ wsr) a1 a2 ->
  z # a1 ->
  w # a2 ->
  similar ys1 ys2 it1 it2 ->
  similar_with ys1 ys2 (zsl ++ z :: zsr) (wsl ++ w :: wsr) a1 a2.
Proof.
  introv HlL HrL Hsim Hbi1 Hbi2 Hsit. reverts wsl.
  induction zsl; intros; destruct wsl; simpls; try discriminate; eauto.
  eapply similar_env_notin; eauto.
  inversions HlL. inversions Hsim.
  - eapply similar_env_dom; eauto.
  - eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_i_middle_move_front (a1 a2: env A) (ys1 ys2 zsl zsr wsl wsr: list var) w z:
    length wsl = length zsl ->
    length wsr = length zsr ->
    similar_with ys1 ys2 (z :: zsl ++ zsr) (w :: wsl ++ wsr) a1 a2 ->
    similar_with ys1 ys2 (zsl ++ z :: zsr) (wsl ++ w :: wsr) a1 a2.
Proof.
  intros. reverts wsl.
  induction zsl; intros; destruct wsl; simpls; try discriminate; eauto.
  inversions H. inversions H1.
  - inversions H8.
    1: eapply similar_env_dom; eauto.
    2: eapply similar_env_notin; eauto.
    all: eapply IHzsl; eauto;
      eapply similar_env_dom; eauto.
  - inversions H8.
    1: eapply similar_env_dom; eauto.
    2: eapply similar_env_notin; eauto.
    all: eapply IHzsl; eauto;
      eapply similar_env_notin; eauto.
Qed.

(* The guiding lists can be reordered *)
Lemma similar_env_with_i_reorder (a1 a2: env A) ns (xs1 xs2 ys1 ys2 zs1 zs2: list var):
    list_reorder xs1 ns zs1 ->
    list_reorder xs2 ns zs2 ->
    LibList.noduplicates ns ->
    length xs1 = length ns ->
    length xs2 = length ns ->
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with ys1 ys2 zs1 zs2 a1 a2.
Proof.

  eapply (list_permutation_ind (fun (zs1 ws1 zs2 ws2: list var)=> similar_with ys1 ys2 zs1 ws1 a1 a2 -> similar_with ys1 ys2 zs2 ws2 a1 a2 )).
  - intros.
    inversions H.
    eapply similar_env_empty. auto.
  - introv Hl1 Hl2 Hind Hs.

    eapply similar_env_with_i_middle_inv in Hs; eauto.
    destruct_all.
    eapply similar_env_dom; eauto.
    eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_o_reorder (a1 a2: env A) ns (ys1 ys2 zs1 zs2: list var):
    list_reorder ys1 ns zs1 ->
    list_reorder ys2 ns zs2 ->
    LibList.noduplicates ns ->
    length ys1 = length ns ->
    length ys2 = length ns ->
    forall (xs1 xs2: list var),
    similar_with ys1 ys2 xs1 xs2 a1 a2 ->
    similar_with zs1 zs2 xs1 xs2 a1 a2.
Proof.
  intros.
  induction H4.
  - eapply similar_env_empty. eapply list_reorder_lengths in H. eapply list_reorder_lengths in H0. rewrite H. rewrite H0. auto.
  - eapply similar_env_dom; eauto. eapply similar_reorder; eauto.
  - eapply similar_env_notin; eauto.
Qed.


Lemma similar_env_with_comm xs ys zs ws a b:
  similar_with xs ys zs ws a b ->
  similar_with ys xs ws zs b a.
Proof.
  introv Hs.
  induction Hs.
  + eapply similar_env_empty. auto.
  + eapply similar_env_dom; eauto using similar_symm.
  + eapply similar_env_notin; eauto using similar_symm.
Qed.

Lemma similar_env_with_i_length xs ys zs ws a b:
  similar_with xs ys zs ws a b ->
  length zs = length ws.
Proof.
  introv Hs.
  induction Hs; simpls; auto.
Qed.
Lemma similar_env_with_o_length xs ys zs ws a b:
  similar_with xs ys zs ws a b ->
  length xs = length ys.
Proof.
  introv Hs.
  induction Hs; simpls; auto.
Qed.

Lemma similar_env_with_lengths xs ys zs ws a b:
  similar_with xs ys zs ws a b ->
  length zs = length ws /\ length xs = length ys.
Proof.
  introv Hs.
  induction Hs; simpls. 
  auto.
  destructecs IHHs.  auto.
  destructecs IHHs.  auto.
Qed.

  Lemma similar_env_with_o_append_swap xs1 ys1 zs xs2 ys2 ws a b:
    similar_with (xs1 ++ xs2) (ys1 ++ ys2) zs ws a b ->
    length xs1 = length ys1 ->
    similar_with (xs2 ++ xs1) (ys2 ++ ys1) zs ws a b.
  Proof.
    introv Hsi Hl.
    assert (length xs2 = length ys2). {
      forwards: similar_env_with_lengths Hsi.
      destructecs H.
      rewrite ! List.app_length in H.
      rewrite Hl in H.
      eapply PeanoNat.Nat.add_cancel_l. eauto.
    }
  
    forwards: list_reorder_swap_app_n (length ys1) (length ys2).
    destructecs H0.
    forwards: H0 xs1 xs2; fweauto. 
    forwards: H0 ys1 ys2; fweauto.
    eapply similar_env_with_o_reorder; eauto.
      rewrite ! List.app_length . rewrite H, Hl. auto.
      rewrite ! List.app_length .  auto.
  Qed.


  Lemma similar_env_with_i_append_swap xs1 ys1 zs xs2 ys2 ws a b:
    similar_with zs ws (xs1 ++ xs2) (ys1 ++ ys2) a b ->
    length xs1 = length ys1 ->
    similar_with zs ws (xs2 ++ xs1) (ys2 ++ ys1)  a b.
  Proof.
    introv Hsi Hl.
    assert (length xs2 = length ys2). {
      forwards Hlo: similar_env_with_i_length Hsi.
      rewrite ! List.app_length in Hlo.
      rewrite Hl in Hlo.
      eapply PeanoNat.Nat.add_cancel_l. eauto.
    }
  
    forwards: list_reorder_swap_app_n (length ys1) (length ys2).
    destructecs H0.
    forwards: H0 xs1 xs2; fweauto. 
    forwards: H0 ys1 ys2; fweauto.
    eapply similar_env_with_i_reorder; eauto.
      rewrite ! List.app_length . rewrite H, Hl. auto.
      rewrite ! List.app_length .  auto.
  Qed.


Lemma similar_env_with_e_concat_invl ys1 ys2 (a1 a2 b1 b2: env A) xs1 xs2:
  ok (a1 & a2) ->
  ok (b1 & b2) ->
  (forall x y, simlist_relates xs1 xs2 x y -> (x \in dom a2 <-> y \in dom b2)) ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a2) (b1 & b2) ->
  similar_with ys1 ys2 xs1 xs2 (a1) (b1).
Proof.
  introv Hoka Hokb Hdom Hs.
  remember (a1 & a2) as a. remember (b1 & b2) as b.
  induction Hs; intros; subst.
  - eapply similar_env_empty. auto.
  - destruct (Hdom x1 x2).
    eapply simlist_relates_push_l.
    destruct (binds_concat_ok_inv Hoka H0);
    destruct (binds_concat_ok_inv Hokb H1); destruct_all.
    + eapply similar_env_notin; eauto. eapply IHHs; eauto. intros. destruct (Hdom x y). eapply simlist_relates_push; eauto.
      split; eauto.
    + false. eauto using binds_to_dom.
    + false. eauto using binds_to_dom.
    + eapply similar_env_dom; eauto. eapply IHHs; eauto. intros. destruct (Hdom x y). eapply simlist_relates_push; eauto.
      split; eauto.
  - eapply similar_env_notin; eauto. eapply IHHs; eauto. intros. destruct (Hdom x y). eapply simlist_relates_push; eauto.
      split; eauto.
Qed.
Lemma similar_env_with_e_concat_invr ys1 ys2 (a1 a2 b1 b2: env A) xs1 xs2:
  ok (a1 & a2) ->
  ok (b1 & b2) ->
  (forall x y, simlist_relates xs1 xs2 x y -> (x \in dom a1 <-> y \in dom b1)) ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a2) (b1 & b2) ->
  similar_with ys1 ys2 xs1 xs2 (a2) (b2).
Proof.
  introv Hoka Hokb Hdom Hs.

  eapply similar_env_with_e_swap in Hs.
  eapply similar_env_with_e_concat_invl.
  eapply ok_swap. eauto.
  eapply ok_swap. eauto.
  eauto.
  eauto.
  eauto.
  eauto.
Qed.

Lemma similar_env_with_e_concat ys1 ys2 (a1 a2 b1 b2: env A) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 (a1) (b1) ->
  similar_with ys1 ys2 xs1 xs2 (a2) (b2) ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a2) (b1 & b2).
Proof.
  introv Hs1 Hs2.

  induction Hs1.
  - eapply similar_env_empty. auto.
  - inversions Hs2.
    eapply similar_env_dom; eauto.
    eapply similar_env_dom; eauto.
  - inversions Hs2.
    eapply similar_env_dom; eauto.
    eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_e_concat_l_fresh ys1 ys2 (a1 a2 b1 b2: env A) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 (a1) (b1) ->
  fresh (dom a2) (length xs1) xs1 ->
  fresh (dom b2) (length xs2) xs2 ->
  similar_with ys1 ys2 xs1 xs2 (a1 & a2) (b1 & b2).
Proof.
  introv Hs1 Hfr1 Hfr2.
  eapply similar_env_with_e_concat.
  auto. eapply similar_env_with_lengths in Hs1.  destructecs Hs1.
  reverts xs2.
  induction xs1; intros.
  - inversions Hs1. destruct xs2; inversions H0. eapply similar_env_empty. auto.
    inversions Hs0.
  - inversions Hs1.
    destruct xs2; inversions H0; inversions Hs0.
    inversions Hfr2.
    inversions Hfr1.
    eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_refl_by_fv (a: env A) xs ys:
    ok a ->
    (fv_env a) \c (from_list xs) ->
    similar_with xs xs ys ys a a.
Proof.
  introv Hoka Hfv.
  induction ys.
  - constructor. auto.
  - destruct (fv_decide a0 (dom a)).
    + eapply dom_to_binds in m. destruct m. eapply similar_env_dom; eauto. eapply similar_refl_by_fv. eapply subset_trans; eauto.
      eapply env_binds_fv_subset; eauto.
    + eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_cases xs1 xs2 ys1 ys2 (e1 e2: env A) x1 x2:
  simlist_relates xs1 xs2 x1 x2 ->
  similar_with ys1 ys2 xs1 xs2 e1 e2 ->
  (x1 # e1 /\ x2 # e2) \/ (exists b1 b2, binds x1 b1 e1 /\ binds x2 b2 e2 /\ similar ys1 ys2 b1 b2).
Proof.
  introv Hsx Hse.
  unfold similar_env in Hse.
  unfold simlist_relates in Hsx.
  destruct_all.
  reverts xs1 xs2.
  induction x; intros.
  - inversions Hse.
    + false. eapply LibList.Nth_nil_inv; eauto.
    + inversions H. inversions H0. eauto 6.
    + inversions H. inversions H0. eauto 6.
  - inversions H. inversions H0.
    inversions Hse.
    + forwards: IHx; eauto.
    + forwards: IHx; eauto.
Qed.


Lemma similar_env_with_binds xs1 xs2 (S1 S2: env A) x1 x2 it1 it2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  simlist_relates xs1 xs2 x1 x2 ->
  binds x1 it1 S1 ->
  binds x2 it2 S2 ->
  similar ys1 ys2 it1 it2.
Proof.
  intros.
  unfold similar_env in H.
  induction H.
  - false. eapply simlist_relates_nil_inv; eauto.
  - unfold simlist_relates in H0. destruct_all.
    inversions H0.
    + inversions H6. binds_eq. binds_eq. auto.
    + inversions H6. eapply IHsimilar_env_with; eauto. exists n; eauto.
  -
    unfold simlist_relates in H0. destruct_all.
    inversions H0.
    + inversions H5. eapply binds_to_dom in H1. contradiction.
    + inversions H5. eapply IHsimilar_env_with; eauto. exists n; eauto.
Qed.

Lemma binds_sim_to_similar_env_with xs1 xs2 (S1 S2: env A) ys1 ys2 :
  length ys1 = length ys2 ->
  length xs1 = length xs2 ->
  (forall x1 x2, simlist_relates xs1 xs2 x1 x2 ->
            x1 \in dom S1 <-> x2 \in dom S2) ->
  (forall x1 x2 it1 it2, simlist_relates xs1 xs2 x1 x2 ->
                    binds x1 it1 S1 ->
                    binds x2 it2 S2 ->
                    similar ys1 ys2 it1 it2) ->
  similar_with ys1 ys2 xs1 xs2 S1 S2.
Proof.
  introv Hleny Hlenx Hdom Hsim. reverts xs2.
  induction xs1; intros; destruct xs2; simpls; try discriminate.
  eapply similar_env_empty; eauto.
  inversions Hlenx. rename H0 into Hlenx, a into x1, v into x2.
  specialize (Hdom _ _ (simlist_relates_push_l _ _ _ _)) as Hinff.
  destruct (fv_decide x1 (dom S1)) as [Hin1 | Hnin1].
  1: assert (Hin2 : x2 \in dom S2) by firstorder.
  2: assert (Hin2 : x2 \notin dom S2) by firstorder.

  - eapply dom_to_binds in Hin1 as [it1 Hbi1].
    eapply dom_to_binds in Hin2 as [it2 Hbi2].
    eapply similar_env_dom; eauto using simlist_relates_push_l,
      simlist_relates_push.
    eapply IHxs1; eauto using simlist_relates_push.

  - eapply similar_env_notin; eauto using simlist_relates_push_l,
      simlist_relates_push.
    eapply IHxs1; eauto using simlist_relates_push.

Qed.

Lemma similar_env_with_to_binds_sim xs1 xs2 (S1 S2: env A) ys1 ys2 :
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  ((length ys1 = length ys2)
   /\ (length xs1 = length xs2)
   /\ (forall x1 x2, simlist_relates xs1 xs2 x1 x2 ->
               x1 \in dom S1 <-> x2 \in dom S2)
   /\ (forall x1 x2 it1 it2, simlist_relates xs1 xs2 x1 x2 ->
                       binds x1 it1 S1 ->
                       binds x2 it2 S2 ->
                       similar ys1 ys2 it1 it2)).
Proof.
  induction 1; simpl.

  - repeat(split); eauto. 3: intros.
    all: destruct H0 as [n [Hcontra _]]; false;
      eauto using LibList.Nth_nil_inv.

  - destruct IHsimilar_env_with as [IHy [IHx [IHdom IHsim]]]. splits; eauto.
    + intros x1' x2' Hsim'.
      eapply binds_to_dom in H0. eapply binds_to_dom in H1.
      eapply simlist_relates_cons_cases in Hsim' as [Hsim' | [? ?]]; subst;
        eauto. split; eauto.
    + intros x1' x2' it1' it2' Hsim' Hbi1' Hbi2'.
      eapply simlist_relates_cons_cases in Hsim' as [Hsim' | [? ?]]; subst;
        eauto.
      eapply (binds_functional H0) in Hbi1'; subst.
      eapply (binds_functional H1) in Hbi2'; subst. eauto.

  - destruct IHsimilar_env_with as [IHy [IHx [IHdom IHsim]]]. splits; eauto.
    + intros x1' x2' Hsim'.
      eapply simlist_relates_cons_cases in Hsim' as [Hsim' | [? ?]]; subst;
        eauto. split; eauto; firstorder.
    + intros x1' x2' it1' it2' Hsim' Hbi1' Hbi2'.
      eapply simlist_relates_cons_cases in Hsim' as [Hsim' | [? ?]]; subst;
        eauto.
      eapply binds_to_dom in Hbi1'. eapply binds_to_dom in Hbi2'. firstorder.

Qed.

Lemma similar_env_with_equiv xs1 xs2 (S1 S2: env A) ys1 ys2 :
  similar_with ys1 ys2 xs1 xs2 S1 S2 <->
  ((length ys1 = length ys2)
   /\ (length xs1 = length xs2)
   /\ (forall x1 x2, simlist_relates xs1 xs2 x1 x2 ->
               x1 \in dom S1 <-> x2 \in dom S2)
   /\ (forall x1 x2 it1 it2, simlist_relates xs1 xs2 x1 x2 ->
                       binds x1 it1 S1 ->
                       binds x2 it2 S2 ->
                       similar ys1 ys2 it1 it2)).
Proof.
  split; eauto using similar_env_with_to_binds_sim.
  intros. destruct_all. eauto using binds_sim_to_similar_env_with.
Qed.

Lemma similar_env_with_i_append_invr xs11 xs12 xs21 xs22 (S1 S2 : env A) ys1 ys2 :
  length xs11 = length xs21 ->
  similar_with ys1 ys2 (xs11 ++ xs12) (xs21 ++ xs22) S1 S2 ->
  similar_with ys1 ys2          xs12           xs22 S1 S2.
Proof.
  introv Hlenx Hsim.
  eapply similar_env_with_equiv in Hsim.
  destructc Hsim as Hleny Hlenx12 HdomS Hsimit.
  rewrite ! List.app_length in Hlenx12.
  eapply similar_env_with_equiv; splits; try omega.
  - introv Hsim. eapply HdomS; eauto using simlist_relates_append.
  - introv Hsim. eapply Hsimit; eauto using simlist_relates_append.
Qed.

Lemma similar_env_with_o_append_invr_disjoint_fv_l
  xs1 xs2 (S1 S2 : env A) ys11 ys12 ys21 ys22 :
  length ys11 = length ys21 ->
  (forall x1 it1,
      binds x1 it1 S1 ->
      x1 \in from_list xs1 ->
      disjoint (from_list ys11) (fv it1)) ->
  similar_with (ys11 ++ ys12) (ys21 ++ ys22) xs1 xs2 S1 S2 ->
  similar_with          ys12           ys22 xs1 xs2 S1 S2.
Proof.
  introv Hleny Hd Hsim.
  eapply similar_env_with_equiv in Hsim.
  destructc Hsim as Hlenys Hlenx HdomS Hsimit.
  rewrite ! List.app_length in Hlenys.
  eapply similar_env_with_equiv; splits; try omega; eauto.
  introv Hsim Hbi1 Hbi2. forwards: Hsimit; fweauto.
  eapply similar_out_shorten_ls in H; try omega; eauto.
  eapply Hd; eauto. destructc Hsim as n Hin1 _.
  eapply LibList.Nth_mem, list_mem_in_fromlist in Hin1. eauto.
Qed.

Lemma similar_env_with_o_append_invr_disjoint_fv_r
  xs1 xs2 (S1 S2 : env A) ys11 ys12 ys21 ys22 :
  length ys11 = length ys21 ->
  (forall x2 it2,
      binds x2 it2 S2 ->
      x2 \in from_list xs2 ->
      disjoint (from_list ys21) (fv it2)) ->
  similar_with (ys11 ++ ys12) (ys21 ++ ys22) xs1 xs2 S1 S2 ->
  similar_with          ys12           ys22 xs1 xs2 S1 S2.
Proof.
  introv Hleny Hd Hsim.
  eapply similar_env_with_equiv in Hsim.
  destructc Hsim as Hlenys Hlenx HdomS Hsimit.
  rewrite ! List.app_length in Hlenys.
  eapply similar_env_with_equiv; splits; try omega; eauto.
  introv Hsim Hbi1 Hbi2. forwards: Hsimit; fweauto.
  eapply similar_out_shorten_rs in H; try omega; eauto.
  eapply Hd; eauto. destructc Hsim as n _ Hin2.
  eapply LibList.Nth_mem, list_mem_in_fromlist in Hin2. eauto.
Qed.

Lemma similar_env_with_by_fresh xs1 xs2 (S1 S2: env A) ys1 ys2:
  fresh (from_list xs1) (length(keys S1)) (keys S1) ->
  fresh (from_list xs2) (length(keys S2)) (keys S2) ->
  length xs1 = length xs2 ->
  length ys1 = length ys2 ->
  similar_with (ys1) (ys2) (xs1) (xs2) S1 S2.
Proof.
  introv Hfr1 Hfr2 Hlx Hly.
  reverts xs2.
  induction xs1; intros; destruct xs2; simpls; try discriminate.
  - eapply similar_env_empty; eauto.
  - rewrite from_list_cons in Hfr1, Hfr2.
    eapply similar_env_notin.
    eapply IHxs1; eauto.
    rewrite <- env_from_list_keys_dom. eapply fresh_union_r1 in Hfr1. eapply fresh_single_notin; eauto.
    rewrite <- env_from_list_keys_dom. eapply fresh_union_r1 in Hfr2. eapply fresh_single_notin; eauto.
Qed.

Lemma similar_env_with_oi_push_notin xs1 xs2 (S1 S2: env A) x1 x2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  x1 # S1 ->
  x2 # S2 ->
  similar_with (x1::ys1) (x2::ys2) (x1::xs1) (x2::xs2) S1 S2.
Proof.
  introv Hsim Hn1 Hn2.
  eapply similar_env_notin; eauto using similar_env_with_o_push.
Qed.

Lemma similar_env_with_push_notin  xs1 xs2 (S1 S2: env A) x1 x2 a1 a2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar (x1::ys1) (x2::ys2) a1 a2  ->
  x1 \notin (from_list xs1) ->
  x2 \notin (from_list xs2) ->
  similar_with (x1::ys1) (x2::ys2) (x1::xs1) (x2::xs2) (S1 & x1 ~ a1) (S2 & x2 ~ a2).
Proof.
  introv HsS Hsa Hn1 Hn2.
  eapply similar_env_dom; eauto.
  eapply similar_env_with_e_concat; eauto.
  - eapply similar_env_with_o_push; eauto.
  -
    forwards: similar_env_with_i_length HsS; eauto.
    forwards: similar_length Hsa; eauto.
     eapply similar_env_with_by_fresh; rewrite ? env_single_keys; eauto.
Qed.
Lemma similar_env_with_ie_push_notin  xs1 xs2 (S1 S2: env A) x1 x2 a1 a2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar ys1 ys2 a1 a2  ->
  x1 \notin (from_list xs1) ->
  x2 \notin (from_list xs2) ->
  similar_with ys1 ys2 (x1::xs1) (x2::xs2) (S1 & x1 ~ a1) (S2 & x2 ~ a2).
Proof.
  introv HsS Hsa Hn1 Hn2.
  eapply similar_env_dom; eauto.
  eapply similar_env_with_e_concat; eauto.
    forwards: similar_env_with_i_length HsS; eauto.
    forwards: similar_length Hsa; eauto.
     eapply similar_env_with_by_fresh; rewrite ? env_single_keys; eauto.
Qed.

Lemma similar_env_with_trans ys1 ys2 ys3 (S1 S2 S3: env A) xs1 xs2 xs3:
  fresh \{} (length ys2) ys2 ->
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  similar_with ys2 ys3 xs2 xs3 S2 S3 ->
  similar_with ys1 ys3 xs1 xs3 S1 S3.
Proof.
  introv Hfr Hs12 Hs23.
  reverts ys3 xs3 S3.
  induction Hs12; intros.
  - inversions Hs23.
    eapply similar_env_empty. rewrite H. auto.
  - inversions Hs23.
    + eapply similar_env_dom; eauto. binds_eq. eapply similar_trans; eauto.
    + eapply binds_to_dom in H1. contradiction.
  - inversions Hs23.
    + eapply binds_to_dom in H5. contradiction.
    + eapply similar_env_notin; eauto.
Qed.

Lemma similar_env_with_i_append_invl ys1 ys2 (S1 S2: env A) xs11 xs12 xs21 xs22:
  length xs11 = length xs21 ->
  similar_with ys1 ys2 (xs11 ++ xs12) (xs21 ++ xs22) S1 S2 ->
  similar_with ys1 ys2 (xs11) (xs21) S1 S2.
Proof.
  introv Hl Hs.
  forwards: similar_env_with_i_length Hs.
  rewrite ! List.app_length in H. rewrite Hl in H. apply Plus.plus_reg_l in H.

  reverts xs22.
  induction xs12; intros; destruct xs22; simpls; try discriminate.
  - rewrite ! List.app_nil_r in Hs. auto.
  - inversions H. eapply IHxs12. eapply similar_env_with_i_middle_inv with (wsr:=xs22); eauto. auto.
Qed.

Lemma similar_env_with_ind_list_split zs1 zs2 S1 S2  S1u S2u:
  (
    forall (ys1 ys2 xs1 xs2: list var), length ys1 = length ys2 ->
    zs1 = (ys1 ++ xs1)%list -> zs2 = (ys2 ++ xs2)%list ->
    similar_with zs1 zs2 xs1 xs2 S1 S2 ->
    similar_with zs1 zs2 xs1 xs2 S1u S2u
  ) ->
  similar_with zs1 zs2 zs1 zs2 S1 S2 ->
  similar_with zs1 zs2 zs1 zs2 S1u S2u.
Proof.
  intros.
  specialize (H nil nil zs1 zs2).
  eapply H; eauto.
Qed.

Lemma similar_env_with_ie_invr wa wb ya yb xa xb ea1 ea2 eb1 eb2:
  ok (ea1 & ea2) ->
  ok (eb1 & eb2) ->
  similar_with wa wb (ya ++ xa) (yb ++ xb) (ea1 & ea2) (eb1 & eb2) ->
  length ya = length yb ->
  disjoint (dom ea1) (from_list xa) ->
  disjoint (dom eb1) (from_list xb) ->
  similar_with wa wb (xa) (xb) (ea2) (eb2).
Proof.
  introv Hoka Hokb Hsi Hl Hdia Hdib.
  eapply similar_env_with_e_concat_invr.
  eauto.
  eauto.
  2: eapply similar_env_with_i_append_invr. 3: eauto.
  intros. split; intro.
  eapply simlist_relates_in_list_l in H.
  false. eapply disjoint_in_false. eapply Hdia. eauto. eauto.
  eapply simlist_relates_in_list_r in H.
  false. eapply disjoint_in_false. eapply Hdib. eauto. eauto. eauto.
Qed.

End SimilarEnvWith.




Instance similar_with_in_env{A: Set}(AS: Similar A){AF: FreeVar A}{ASO: SimilarOut AS}{ASC: SimilarCommon AS}: SimilarWithIn (EnvyEnv AF) (similar_env_with AS).
Proof.
  split; intros.
  - eapply similar_env_with_o_push; eauto.
  - eapply similar_env_with_i_append_invl; eauto.
  - eapply similar_env_with_i_reorder; eauto.
  - eapply similar_env_with_o_reorder; eauto.
  - eapply similar_env_notin; eauto.
Qed.

Section SimilarEnv.
Context {A: Set}{AS: Similar A}{AF: FreeVar A}{ASC: SimilarCommon AS}{ASO: SimilarOut AS}.

Section SimilarEnvWith.
Let SW: SimilarWith (env A) := similar_env_with AS.
Local Existing Instance SW.

Lemma similar_env_with_o_append_r zs1 ys1 zs2 ys2 (S1 S2: env A) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  length zs1 = length zs2 ->
  similar_with (zs1 ++ ys1) (zs2 ++ ys2) xs1 xs2 S1 S2.
Proof.
  introv Hs Hl.
  eapply similar_with_o_append_r; eauto.
  Unshelve. 2: eapply similar_with_in_env.
Qed.
Lemma similar_env_with_o_append_l zs1 ys1 zs2 ys2 (S1 S2: env A) xs1 xs2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  length zs1 = length zs2 ->
  similar_with (ys1 ++ zs1) (ys2 ++ zs2) xs1 xs2 S1 S2.
Proof.
  introv Hs Hl.
  eapply similar_env_with_o_append_swap; eauto.
  eapply similar_env_with_o_append_r; eauto.
Qed.
Lemma similar_env_with_o_append_m zs1 ys1 ws1 zs2 ys2 ws2 (S1 S2: env A) xs1 xs2:
  similar_with (ys1++ws1) (ys2++ws2) xs1 xs2 S1 S2 ->
  length zs1 = length zs2 ->
  length ys1 = length ys2 ->
  similar_with (ys1 ++ zs1 ++ ws1) (ys2 ++ zs2 ++ ws2) xs1 xs2 S1 S2.
Proof.
  introv Hs Hl Hly.
  eapply similar_env_with_o_append_swap; eauto.
  rewrite <- ! List.app_assoc.
  eapply similar_env_with_o_append_r; eauto.
  eapply similar_env_with_o_append_swap; eauto.
  forwards: similar_env_with_o_length Hs.
  rewrite ! List.app_length in *.
  rewrite Hly in H.
  apply -> PeanoNat.Nat.add_cancel_l in H.
  rewrite H. rewrite Hl. auto.
Qed.

Lemma similar_env_with_i_append_fresh_r  xs1 xs2 (S1 S2: env A) zs1 zs2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  fresh (dom S1) (length zs1) zs1 ->
  fresh (dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar_with ys1 ys2 (zs1++xs1) (zs2++xs2) S1 S2.
Proof.
  introv Hsi Hfr1 Hfr2 Hl.

  eapply similar_with_i_append_fresh_r; eauto.
  Unshelve. 4: eapply similar_with_in_env.
  simpl. auto.
  simpl. auto.
Qed.

Lemma similar_env_with_i_append_fresh_l  xs1 xs2 (S1 S2: env A) zs1 zs2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  fresh (dom S1) (length zs1) zs1 ->
  fresh (dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar_with ys1 ys2 (xs1++zs1) (xs2++zs2) S1 S2.
Proof.
  introv Hsi Hfr1 Hfr2 Hl.

  eapply similar_env_with_i_append_swap.
  eapply similar_env_with_i_append_fresh_r; eauto.
  auto.
Qed.

Lemma similar_env_with_i_append_fresh_m  xs1 xs2 (S1 S2: env A) zs1 zs2 ys1 ys2 ws1 ws2:
  similar_with ys1 ys2 (xs1++ws1) (xs2++ws2) S1 S2 ->
  fresh (dom S1) (length zs1) zs1 ->
  fresh (dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  length xs1 = length xs2 ->
  similar_with ys1 ys2 (xs1++zs1++ws1) (xs2++zs2++ws2) S1 S2.
Proof.
  introv Hsi Hfr1 Hfr2 Hl Hlx.

  eapply similar_env_with_i_append_swap; eauto.
  rewrite <- ! List.app_assoc.
  eapply similar_env_with_i_append_fresh_r; eauto.
  eapply similar_env_with_i_append_swap; eauto.
  forwards: similar_env_with_i_length Hsi.
  rewrite ! List.app_length in *.
  rewrite Hlx in H.
  apply -> PeanoNat.Nat.add_cancel_l in H.
  rewrite H. rewrite Hl. auto.

Qed.

End SimilarEnvWith.


Let SE: Similar (env A) := similar_env.
Local Existing Instance SE.

Lemma similar_env_push xs1 xs2 (S1 S2: env A) x1 x2 a1 a2:
  similar xs1 xs2 S1 S2 ->
  similar (x1::xs1) (x2::xs2) a1 a2  ->
  x1 \notin (from_list xs1) ->
  x2 \notin (from_list xs2) ->
  similar (x1::xs1) (x2::xs2) (S1 & x1 ~ a1) (S2 & x2 ~ a2).
Proof.
  introv HsS Hsa Hn1 Hn2.
  unfold similar.
  eapply similar_env_with_push_notin; eauto.
Qed.



Lemma similar_env_append_fresh_r xs1 xs2 (S1 S2: env A) zs1 zs2:
  similar xs1 xs2 S1 S2 ->
  fresh (dom S1) (length zs1) zs1 ->
  fresh (dom S2) (length zs2) zs2 ->
  length zs1 = length zs2 ->
  similar (zs1 ++ xs1) (zs2 ++ xs2) (S1) (S2).
Proof.
  introv HsS Hfr1 Hfr2 Hl.
  unfold similar, similar_env in *.
  eapply similar_with_i_append_fresh_r.
  eapply similar_with_o_append_r.
  all: eauto.
Qed.

Global Instance similar_common_env: SimilarCommon SE.
Proof.
  split; intros.
  - unfold similar, similar_env in *.
    eapply similar_env_with_comm; eauto.
  - (* similar_in_trans *)
    unfold similar, similar_env in *.
    eapply similar_env_with_trans; eauto.
  - unfold similar, similar_env in *.
    eapply similar_env_with_i_length; eauto.
  - (* similar_in_reorder *)
    unfold similar, SE, similar_env in *.
    assert (length ys1 = length ns). {
    rewrite H3. symmetry.
    eapply similar_env_with_i_length; eauto.
    }
    eapply similar_env_with_o_reorder; eauto.
    eapply similar_env_with_i_reorder; eauto.
Qed.

Global Instance similar_in_env: SimilarIn (EnvyEnv AF) SE.
Proof.
  split; intros.
  - (* similar_weaken *)
    unfold similar, similar_env in *.
    eapply similar_env_with_e_concat_invl; eauto.
  - (* similar_in_refl_by_fv *)
    unfold similar, similar_env in *.
    eapply similar_env_with_refl_by_fv; eauto.
  - eapply similar_env_append_fresh_r; eauto.
Qed.

Global Instance similar_env_in_env: SimilarEnvIn SE.
Proof.
  split; intros.
  - (* similar_by_nil *)
    eapply similar_env_empty. auto.
  - (* similar_extend_notin *)
    unfold similar, SE, similar_env in *.
    constructor; eauto using similar_env_with_o_push.
  - (* similar_extend_dom *)
    unfold similar, similar_env in *.
    inversions H0.
    eapply binds_single_inv in H10.
    eapply binds_single_inv in H13.
    destruct_all; subst.
    eapply similar_env_dom; eauto.
    eapply similar_env_with_o_push; eauto.
    simpl_dom. rewrite notin_singleton in H12. contradiction.
  - (* similar_swap *)
    unfold similar, similar_env in *.
    eapply similar_env_with_e_swap_after; eauto.

  - (* similar_concat *)
    unfold similar, similar_env in *.
    eapply similar_env_with_e_concat; eauto.

Qed.

End SimilarEnv.
