Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax
        Similarity Defs.
Require Import HeapUpdate Deref.

(** * Similarity of syntactic elements *)

(* Definition 45 in "Pure methods for roDOT" *)
Instance similar_var: Similar var := similar_by_subst.
Instance similar_avar: Similar avar := (@similar_by_subst avar _ _).
Instance similar_typ: Similar typ := (@similar_by_subst typ _ _).
Instance similar_dec: Similar dec := (@similar_by_subst dec _ _).
Instance similar_trm: Similar trm := (@similar_by_subst trm _ _).
Instance similar_defs: Similar defs := (@similar_by_subst defs _ _).
Instance similar_def: Similar def := (@similar_by_subst def _ _).
Instance similar_lit: Similar lit := (@similar_by_subst lit _ _).
Instance similar_item: Similar item := (@similar_by_subst item _ _).
Instance similar_frame: Similar frame := similar_by_subst.
Instance similar_stack: Similar stack := (@similar_by_subst stack _ _).

Instance similar_with_heap: SimilarWith heap := similar_env_with similar_item.
Instance similar_heap: Similar heap := similar_env.
Instance similar_with_renv: SimilarWith renv := similar_env_with similar_var.
Instance similar_renv: Similar renv := similar_env.

(* Lemmas for similarity and syntax transformations *)

Lemma var_subst_vars_Nth w1 xs1 xs2 w2 n:
  length xs1 = length xs2 ->
  fresh (from_list xs2) (length xs1) xs1 ->
  LibList.Nth n xs1 w1 ->
  LibList.Nth n xs2 w2 ->
  w1 /[xs1 --> xs2] = w2.
Proof.
  introv Hl Hfr Hw1 Hw2.
  reverts xs1 xs2.
  induction n; intros; subst.
  - inversions Hw1.  inversions Hw2.
    rewrite <- subst_vars_push_left.
    rewrite subst_vars_fresh.
    unfold subst_var, VarSubstVar. cases_if. auto.
    sympls.
    destruct (fresh_cons Hfr).
    unfold fv, VarFreeVar.
    fresh_solve.
  - inversions Hw1. inversions Hw2.
    sympls.
    rewrite from_list_cons in Hfr.
    rewrite IHn; eauto.
    unfold subst_var, VarSubstVar. cases_if.
    + false. eapply LibList.Nth_mem in H1. eapply list_mem_in_fromlist in H1. destruct Hfr. rewrite notin_union in H. destruct H. contradiction.
    + auto.
    + destruct_all. fresh_solve.
Qed.
Lemma simlist_relates_similar_var xs1 xs2 w1 w2:
  length xs1 = length xs2 ->
  simlist_relates xs1 xs2 w1 w2 ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_var, similar_by_subst, simlist_relates.
  introv Hl Hr; destruct_all.

  destructc (var_freshes (\{ w1 } \u \{w2} \u from_list xs1 \u from_list xs2) (length xs1)) as x0 Hx0.
  lets Hl0: fresh_length Hx0.
  rewrite <- Hl0 in Hx0.
  forwards Hw0: LibList.Nth_inbound_inv x x0.
  rewrite liblist_length.
  rewrite Hl0.
  rewrite <- liblist_length.
  eauto using LibList.Nth_inbound.
  destructc Hw0 as w0 Hw0.
  exists w0 x0.
  splits; eauto.
  - rewrite Hl0. auto.
  - unfold fv, VarFreeVar. intro. intro. rewrite in_singleton in H1; subst. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
  - eapply var_subst_vars_Nth; eauto.
  - eapply var_subst_vars_Nth; eauto. rewrite Hl0. eauto.
Qed.

Lemma similar_var_simlist_relates xs1 xs2 w1 w2:
  similar xs1 xs2 w1 w2 ->
  simlist_relates xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_var, similar_by_subst, simlist_relates.
  intro; destruct_all.

  reverts x xs1 xs2.
  induction x0; intros.
  - sympls. subst.
    false.
    rewrite from_list_nil in H2. unfold fv, VarFreeVar in H2. specialize (H2 w2). rewrite in_empty in H2. apply H2. rewrite in_singleton. auto.
  - sympls.
    destruct xs1; try discriminate.
    destruct xs2; try discriminate.
    sympls.
    destruct_all. rewrite from_list_cons in *.
    inversions H.
    inversions H0.
    rewrite from_list_cons in H1, H5.
    destruct (classicT (x = a)).
    + subst.
      rewrite subst_vars_fresh.
      rewrite subst_vars_fresh.
      unfold subst_var, VarSubstVar. cases_if.
      exists 0. splits; auto.
      all: unfold fv, VarFreeVar; eauto.
    +
      forwards: IHx0 x H7 H3; eauto.
      unfold fv, VarFreeVar.
      unfold fv, VarFreeVar in H2. intro. intro. specialize (H2 x1 H). rewrite in_singleton in H; subst. rewrite in_union in H2. destruct H2. rewrite in_singleton in H. contradiction. auto.
      rewrite subst_fresh; eauto. eapply notin_fv_subst_vars; eauto. unfold fv, VarFreeVar. rewrite notin_singleton. auto.
      fresh_solve.
      rewrite subst_fresh; eauto. eapply notin_fv_subst_vars; eauto. unfold fv, VarFreeVar. rewrite notin_singleton. auto.
      destruct_all.
      exists (S x1). eauto.
Qed.

Lemma similar_var_middle xs1 xs2 ys1 ys2 y1 y2:
  length xs1 = length xs2 ->
  length ys1 = length ys2 ->
  similar (xs1 ++ y1 :: ys1) (xs2 ++ y2 :: ys2) y1 y2.
Proof.
  intros Hlx Hly.
  eapply simlist_relates_similar_var.
  - rewrite ! List.app_length. simpls. rewrite Hlx. rewrite Hly. auto.
  - unfold simlist_relates.
    exists (length xs1).
    splits. rewrite <- liblist_app.
    replace (length xs1) with (0+(length xs1)).
    rewrite <- liblist_length.
    eapply (LibList.Nth_app_r 0). constructor. rewrite Nat.add_0_l. auto.

    rewrite Hlx.
    rewrite <- liblist_app.
    replace (length xs2) with (0+(length xs2)).
    rewrite <- liblist_length.
    eapply (LibList.Nth_app_r 0). constructor. rewrite Nat.add_0_l. auto.
Qed.

Lemma similar_var_push ys1 ys2 y1 y2:
  length ys1 = length ys2 ->
  similar (y1 :: ys1) (y2 :: ys2) y1 y2.
Proof.
  intros.
  forwards: similar_var_middle (@nil var) (@nil var) ys1 ys2; eauto.
Qed.

Section SimilarOpen.
Context {A: Set}
  {AS: SubstVar A}{AF: FreeVar A}{AO: Openable A}{AAS: AbstractSyntax AO AF AS}
  {ASC: SubstCommut AAS}{ASOC: SubstOpenCommut AAS}
  {AFOC: FvOpenCases AAS}{AFC: FvSubstCases AAS}{ASF: SubstFresh AAS}.

Let SA: Similar A := similar_by_subst.
Local Existing Instance SA.

Lemma similar_open_rec xs1 xs2 n (t1 t2:A) x1 x2:
  similar xs1 xs2 t1 t2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 (open_rec n x1 t1) (open_rec n x2 t2).
Proof.
  introv Hsimt Hsimx.
  destruct_similar_by_subst (\{x1} \u \{x2}) Hsimt t.
  destruct_similar_by_subst_by xt3 (fv t1 \u fv t2) Hsimx x.
  fresh_solve; unfold fv, VarFreeVar; fresh_solve.

  exists (open_rec n x3 t3) xt3.
  splits; eauto.
  - fresh_solve; eapply fresh_fv_open; fresh_solve.
  - eapply subset_trans. eapply fv_open_subset.
    eapply subset_union_l. unfold fv, VarFreeVar in H3; auto.
    auto.
  - rewrite subst_vars_open_commut.
    rewrite Hat. rewrite Hax. auto.
  - rewrite subst_vars_open_commut.
    rewrite Hbt. rewrite Hbx. auto.
Qed.
Lemma similar_open_rec_push xs1 xs2 n (t1 t2:A) x1 x2:
  similar xs1 xs2 t1 t2 ->
  similar (x1::xs1) (x2::xs2) (open_rec n x1 t1) (open_rec n x2 t2).
Proof.
  introv Hsimt.
  eapply similar_open_rec.
  eapply similar_by_subst_extend. eauto.
  eapply similar_var_push.
  eapply similar_by_subst_length in Hsimt.
  eauto.
Qed.



Lemma subst_vars_split_at (a: A) (xs ys: list var) (x y: var) (n: nat):
  length xs = length ys ->
  LibList.Nth n xs x ->
  LibList.Nth n ys y ->
  (a /[xs --> ys]) = (a /[(LibList.drop (S n) xs) --> (LibList.drop (S n) ys)] /[x -> y] /[(LibList.take (n) xs) --> (LibList.take (n) ys)]).
Proof.
  introv Hl Hnx Hny.
  rewrite (list_split_at Hnx) at 1.
  rewrite (list_split_at Hny) at 1.
  rewrite subst_vars_append.
  fequal.
  erewrite ! list_Nth_take_length by eauto. auto.
  rewrite ! list_length_cons. fequal.
  erewrite ! list_Nth_drop_S_length by eauto. auto.
Qed.



Lemma subst_vars_split_at_l (a: A) (xs ys: list var) (x y: var) (n: nat):
  length xs = length ys ->
  fresh (from_list ys) (length xs) xs ->
  fresh \{} (length ys) ys ->
  LibList.Nth n xs x ->
  LibList.Nth n ys y ->
  (a /[xs --> ys]) = (a /[x -> y] /[(LibList.take (n) xs) ++ (LibList.drop (S n) xs)  --> (LibList.take (n) ys) ++ (LibList.drop (S n) ys) ]).
Proof.
  introv Hl Hfrx Hfry Hnx Hny.
  rewrite (subst_vars_split_at a Hl Hnx Hny); eauto.
  rewrite subst_vars_commut.
  rewrite subst_vars_append.
  eauto.
  erewrite ! list_Nth_take_length by eauto. auto.
  erewrite ! list_Nth_drop_S_length by eauto. auto.

  eapply fresh_notin_drop_S; eauto.
  eapply notin_subset. eapply from_list_drop. eapply fresh_in_notin. eauto. eapply LibList.Nth_mem; eauto.
  eapply notin_subset. eapply from_list_drop. eapply fresh_in_notin_list. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
Qed.
Lemma subst_vars_split_at_r (a: A) (xs ys: list var) (x y: var) (n: nat):
  length xs = length ys ->
  fresh (from_list ys) (length xs) xs ->
  fresh \{} (length ys) ys ->
  LibList.Nth n xs x ->
  LibList.Nth n ys y ->
  (a /[xs --> ys]) = (a /[(LibList.take (n) xs) ++ (LibList.drop (S n) xs)  --> (LibList.take (n) ys) ++ (LibList.drop (S n) ys) ]  /[x -> y]).
Proof.
  introv Hl Hfrx Hfry Hnx Hny.
  rewrite (subst_vars_split_at a Hl Hnx Hny); eauto.
  rewrite <- subst_vars_commut.
  rewrite subst_vars_append.
  eauto.
  erewrite ! list_Nth_take_length by eauto. auto.
  erewrite ! list_Nth_drop_S_length by eauto. auto.

  eapply fresh_notin_take; eauto.
  eapply notin_subset. eapply from_list_take. eapply fresh_in_notin. eauto. eapply LibList.Nth_mem; eauto.
  eapply notin_subset. eapply from_list_take. eapply fresh_in_notin_list. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
Qed.

Lemma subst_twice (a: A) x1 y1 y2:
  x1 <> y1 ->
  (a /[ x1 -> y1]/[ x1 -> y2]) =(a /[ x1 -> y1]).
Proof.
  introv Hn.
  rewrite subst_fresh.
  auto.
  eapply subst_notfv; eauto.
Qed.

Lemma subst_vars_extract_at_l (a: A) (xs ys: list var) (x y: var) (n: nat):
  length xs = length ys ->
  fresh (from_list ys) (length xs) xs ->
  fresh \{} (length ys) ys ->
  LibList.Nth n xs x ->
  LibList.Nth n ys y ->
  (a /[xs --> ys]) = (a /[x -> y] /[xs --> ys]).
Proof.
  introv Hl Hfrx Hfry Hnx Hny.
  rewrite (subst_vars_split_at_l a Hl Hfrx Hfry Hnx Hny); eauto.
  rewrite (subst_vars_split_at_l (a /[ x -> y]) Hl Hfrx Hfry Hnx Hny); eauto.
  rewrite subst_twice. auto.

  eapply LibList.Nth_mem in Hnx; eauto.
  eapply list_mem_in_fromlist in Hnx.
  eapply LibList.Nth_mem in Hny; eauto.
  eapply list_mem_in_fromlist in Hny.
  intro. subst.
  eapply fresh_in_notin_list with (L:=from_list ys); eauto.
Qed.

Context {ASI: SubstId AAS}.

Lemma subst_twice_same (a: A) x1 y1:
  (a /[ x1 -> y1]/[ x1 -> y1]) =(a /[ x1 -> y1]).
Proof.
  destruct (classicT (x1 = y1)).
  - subst. rewrite subst_id. auto.
  - eapply subst_twice; eauto.
Qed.

  Lemma subst_merge_partial x y z (X:A):
    (subst_var y z (subst_var x y X)) = (subst_var y z (subst_var x z X)).
  Proof.
    intros.
    lets subst_commut': subst_commut.
    destruct (classicT(x=y)).
    - subst.
      rewrite subst_id. rewrite subst_twice_same. auto.
    -

    forwards Hr: (subst_commut' y z x y X); auto.
    cases_if.

    destruct (classicT(x = z)).
    + subst.
      rewrite subst_commut. cases_if. rewrite ! subst_id.
      auto. auto. auto.
    +
      destruct (classicT(y = z)).
      * subst. auto.
      * forwards Hr3: (subst_commut' y z x z X); auto.
        cases_if.

        rewrite Hr.
        rewrite Hr3.
        auto.
  Qed.

Lemma subst_vars_Nth_swap xs ys n1 n2 x1 x2 y1 y2 (a: A):
  length xs = length ys ->
  fresh (from_list ys) (length xs) xs ->
  fresh (fv a) (length ys) ys ->
  LibList.Nth n1 ys y1 ->
  LibList.Nth n2 ys y2 ->
  LibList.Nth n1 xs x1 ->
  LibList.Nth n2 xs x2 ->
  ((a /[ x1 -> x2]) /[ xs --> ys]) = ((a /[ xs --> ys]) /[ y1 -> y2]).
Proof.
  introv Hl Hfrx Hfry Hy1 Hy2 Hx1 Hx2.

  destruct (classicT (n1 = n2)).
  - (* Trivial substitution *)
    subst.
    forwards: LibList.Nth_functional Hy1 Hy2.
    forwards: LibList.Nth_functional Hx1 Hx2.
    subst.
    rewrite ! subst_id. auto.
  - (* Non -trivial *)
    forwards: fresh_nth_ne Hy1 Hy2; eauto.
    forwards: fresh_nth_ne Hx1 Hx2; eauto.
    rewrite subst_vars_extract_at_l with (n:=n2) (x:=x2) (y:=y2) (xs:=xs) by eauto.
    rewrite subst_merge_partial.
    rewrite <- subst_vars_extract_at_l with (n:=n2) (x:=x2) (y:=y2) (xs:=xs) by eauto.
    rewrite subst_vars_split_at_r with (a:=a) (n:=n1) (x:=x1) (y:=y1) (xs:=xs) by eauto.
    rewrite subst_vars_split_at_l with (n:=n1) (x:=x1) (y:=y1) (xs:=xs) by eauto.
    rewrite subst_twice.
    rewrite subst_merge.
    rewrite subst_vars_commut. auto.

    all: rewrite ? from_list_append, ? notin_union; try split.
    + eapply fresh_notin_take; eauto.
    + eapply fresh_notin_drop_S; eauto.
    + eapply notin_subset. eapply from_list_take; eauto.
      eapply fresh_in_notin. eauto. eapply LibList.Nth_mem; eauto.
    + eapply notin_subset. eapply from_list_drop; eauto.
      eapply fresh_in_notin. eauto. eapply LibList.Nth_mem; eauto.
    + eapply notin_subset. eapply from_list_take; eauto.
      eapply fresh_in_notin_list. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    + eapply notin_subset. eapply from_list_drop; eauto.
      eapply fresh_in_notin_list. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    +
      eapply LibList.Nth_mem in Hx1; eauto.
      eapply list_mem_in_fromlist in Hx1.
      eapply LibList.Nth_mem in Hy1; eauto.
      eapply list_mem_in_fromlist in Hy1.
      intro. subst.
      eapply fresh_in_notin_list with (L:=from_list ys); eauto.
    + eapply notin_fv_subst_vars.
      eapply fresh_in_notin. eauto. eapply LibList.Nth_mem; eauto.
      rewrite ? from_list_append, ? notin_union; try split.
      eapply fresh_notin_take; eauto.
      eapply fresh_notin_drop_S; eauto.
    + eapply LibList.Nth_mem in Hx1; eauto.
      eapply list_mem_in_fromlist in Hx1.
      eapply LibList.Nth_mem in Hy2; eauto.
      eapply list_mem_in_fromlist in Hy2.
      intro. subst.
      eapply fresh_in_notin_list with (L:=from_list ys); eauto.
Qed.

Lemma similar_by_subst_subst (xs ys: list var) (a b: A) (x1 x2 x3 x4: var):
  fresh \{} (length xs) xs ->
  fresh \{} (length ys) ys ->
  similar xs ys a b ->
  similar xs ys x1 x2 ->
  similar xs ys x3 x4 ->
  similar xs ys (a /[x1 -> x3]) (b /[x2 -> x4]).
Proof.
  introv Hfrx Hfry Hsia Hsix Hsiy.
  hnf in Hsia.
  destruct_all; subst.

  (* Locate the variables in the lists *)
  eapply similar_var_simlist_relates in Hsix.
  eapply similar_var_simlist_relates in Hsiy.
  hnf in Hsix.
  hnf in Hsiy.
  destruct_all.

  (* Find the corresponding variables in x0 *)
  forwards: LibList.Nth_inbound_inv x6 x0. rewrite liblist_length. rewrite <- H. rewrite <- liblist_length. eapply LibList.Nth_inbound; eauto.
  forwards: LibList.Nth_inbound_inv x5 x0. rewrite liblist_length. rewrite <- H0. rewrite <- liblist_length. eapply LibList.Nth_inbound; eauto.
  destruct_all.

  hnf.
  exists (x /[x8 -> x7]) x0.
  splits; eauto.
  - fresh_solve.
    eapply fresh_fv_subst_var. fresh_solve.
    eapply fresh_subset with (F:=(from_list xs)).
    eapply subset_singleton. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto. eauto.
    eapply fresh_fv_subst_var. fresh_solve.
    eapply fresh_subset with (F:=(from_list ys)).
    eapply subset_singleton. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto. eauto.
  - eapply subset_trans. eapply fv_subst_var_subset. eapply subset_union_l. eauto.
    eapply subset_singleton. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
  - eapply subst_vars_Nth_swap; eauto.
    eapply fresh_subset. eapply H2. eapply fresh_swap; eauto.
  -eapply subst_vars_Nth_swap; eauto.
    eapply fresh_subset. eapply H2. eapply fresh_swap; eauto.

Qed.



End SimilarOpen.

(* Lemmas for distributing similarity over syntax *)

(* Variables *)
Lemma similar_avar_f xs1 xs2 w1 w2:
  similar xs1 xs2 w1 w2  ->
  similar xs1 xs2 (avar_f w1) (avar_f w2).
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  intros.
  destruct_all.
  exists (avar_f x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

Lemma similar_avar_f_var xs1 xs2 w1 w2:
  similar xs1 xs2 (avar_f w1) (avar_f w2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists v x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.
Lemma similar_avar_f_case xs1 xs2 w1 a:
  similar xs1 xs2 (avar_f w1) a ->
  exists w2,
    a = (avar_f w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  intros.
  destruct_all.
  subst.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  exists  (v /[ x0 --> xs2]).
  splits; eauto.
  exists v x0; splits; eauto.
Qed.
Lemma similar_avar_b_case xs1 xs2 w1 a:
  similar xs1 xs2 (avar_b w1) a ->
  a = (avar_b w1).
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars_in H3.
  invert_subst_vars.
  autorewrite with simpl_subst_vars.
  auto.
Qed.

Lemma similar_avar_b_inv xs1 xs2 w1 w2:
  similar xs1 xs2 (avar_b w1) (avar_b w2) ->
  w1 = w2.
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  auto.
Qed.
Lemma similar_avar_b xs1 xs2 n:
  length xs1 = length xs2 ->
  similar xs1 xs2 (avar_b n) (avar_b n).
Proof.
  unfold similar, similar_avar, similar_var, similar_by_subst.
  introv Hl.
  destructc (var_freshes (from_list xs1 \u from_list xs2) (length xs1)) as x0 Hx0.
  exists (avar_b n) x0.

  pose proof (fresh_length _ _ _ Hx0).
  eapply fresh_resize_length in Hx0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  rewrite <- Hl. auto.
  intro. intro. rewrite in_empty in H0. contradiction.
Qed.

(* Types *)

Lemma similar_typ_nulary xs1 xs2 T1:
  typ_nulary T1 ->
  length xs1 = length xs2 ->
  similar xs1 xs2 T1 T1.
Proof.
  introv HT Hl.
  unfold similar, similar_typ, similar_by_subst.
  intros.
  destruct_all.
  destructc (var_freshes (from_list xs1 \u from_list xs2 \u fv T1) (length xs1)) as x0 Hx0.
  pose proof Hx0 as Hlx0.
  eapply fresh_length in Hlx0. eapply fresh_resize_length in Hx0.
  exists (T1) x0.
  splits; eauto; inversions HT; autorewrite with simpl_subst_vars; subst; sympls; try rewrite <- Hl; eauto using subset_empty_l.
Qed.
Lemma similar_typ_nulary_case xs1 xs2 T1 T2:
  typ_nulary T1 ->
  similar xs1 xs2 T1 T2 ->
  T2 = T1.
Proof.
  introv HT Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *; subst.
  all: inversions HT; subst; eauto.
Qed.
Lemma similar_typ_andor xs1 xs2 T T1 T2 T3 T4:
  typ_andor T ->
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 T2 T4 ->
  similar xs1 xs2 (T T1 T2) (T T3 T4).
Proof.
  introv HT Hsi1 Hsi2.
  destruct_similar_by_subst (fv T2 \u fv T4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3) Hsi2 d.
  exists (T ds3 d3) xds3.
  splits; inversions HT; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls; eauto.
  eapply subset_union_l; eauto.
  eapply subset_union_l; eauto.
Qed.
Lemma similar_typ_andor_case xs1 xs2 T T1 T2 T3:
  typ_andor T ->
  similar xs1 xs2 (T T1 T2) T3 ->
  exists T31 T32,
    T3 = (T T31 T32) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32.
Proof.
  introv HT Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions HT; inversions H3.
  - exists  (x1 /[ x0 --> xs2]) (x2 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists x1 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x2 x0; splits; eauto using subset_union_l1, subset_union_l2.
  - exists  (x1 /[ x0 --> xs2]) (x2 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists x1 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x2 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_typ_fun xs1 xs2 T1 T2 T3 T4:
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 T2 T4 ->
  similar xs1 xs2 (typ_fun T1 T2) (typ_fun T3 T4).
Proof.
  introv Hds Hd.
  destruct_similar_by_subst (fv T2 \u fv T4) Hds ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3) Hd d.
  exists (typ_fun ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto. sympls. eauto.
  eapply subset_union_l; eauto.
Qed.
Lemma similar_typ_fun_case xs1 xs2 T1 T2 T3:
  similar xs1 xs2 (typ_fun T1 T2) T3 ->
  exists T31 T32,
    T3 = (typ_fun T31 T32) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32.
Proof.
  introv Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (x1 /[ x0 --> xs2]) (x2 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists x1 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x2 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_typ_rcd xs1 xs2 D1 D2:
  similar xs1 xs2 D1 D2 ->
  similar xs1 xs2 (typ_rcd D1) (typ_rcd D2).
Proof.
  unfold similar, similar_trm, similar_dec, similar_by_subst.
  intros.
  destruct_all.
  exists (typ_rcd x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.
Lemma similar_typ_rcd_case xs1 xs2 D T3:
  similar xs1 xs2 (typ_rcd D) T3 ->
  exists D3,
    T3 = (typ_rcd D3) /\
    similar xs1 xs2 D D3.
Proof.
  introv Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (d /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_dec, similar_by_subst.
    exists d x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_dec_fld xs1 xs2 T1 T2 a T3 T4:
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 T2 T4 ->
  similar xs1 xs2 (dec_fld a T1 T2) (dec_fld a T3 T4).
Proof.
  introv Hds Hd.
  destruct_similar_by_subst (fv T2 \u fv T4) Hds ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3) Hd d.
  exists (dec_fld a ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  eapply subset_union_l; eauto.
Qed.
Lemma similar_dec_fld_case xs1 xs2 T1 T2 a T3:
  similar xs1 xs2 (dec_fld a T1 T2) T3 ->
  exists T31 T32,
    T3 = (dec_fld a T31 T32) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32.
Proof.
  introv Hsi.
  unfold similar, similar_dec, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (t /[ x0 --> xs2]) (t0 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists t x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_dec_typ xs1 xs2 T1 T2 a T3 T4:
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 T2 T4 ->
  similar xs1 xs2 (dec_typ a T1 T2) (dec_typ a T3 T4).
Proof.
  introv Hds Hd.
  destruct_similar_by_subst (fv T2 \u fv T4) Hds ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3) Hd d.
  exists (dec_typ a ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  eapply subset_union_l; eauto.
Qed.
Lemma similar_dec_typ_case xs1 xs2 T1 T2 a T3:
  similar xs1 xs2 (dec_typ a T1 T2) T3 ->
  exists T31 T32,
    T3 = (dec_typ a T31 T32) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32.
Proof.
  introv Hsi.
  unfold similar, similar_dec, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (t0 /[ x0 --> xs2]) (t1 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists t1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_dec_met xs1 xs2 T1 T2 a T3 T4 T5 T6:
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 T2 T4 ->
  similar xs1 xs2 T5 T6 ->
  similar xs1 xs2 (dec_met a T1 T2 T5) (dec_met a T3 T4 T6).
Proof.
  introv Hds Hd Hdr.
  destruct_similar_by_subst (fv T2 \u fv T4 \u fv T5 \u fv T6) Hds ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3 \u fv T5 \u fv T6) Hd d.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T3 \u fv T2 \u fv T4) Hdr dr.
  exists (dec_met a ds3 d3 dr3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  eapply subset_union_l; eauto.
  eapply subset_union_l; eauto.
Qed.
Lemma similar_dec_met_case xs1 xs2 T1 T2 T4 a T3:
  similar xs1 xs2 (dec_met a T1 T2 T4) T3 ->
  exists T31 T32 T34,
    T3 = (dec_met a T31 T32 T34) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32 /\
    similar xs1 xs2 T4 T34.
Proof.
  introv Hsi.
  unfold similar, similar_dec, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (t /[ x0 --> xs2]) (t0 /[ x0 --> xs2]) (t1 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_typ, similar_by_subst.
    exists t x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists t1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_typ_sel_case xs1 xs2 T1 T2 a T3:
  similar xs1 xs2 (typ_sel T1 a T2) T3 ->
  exists T31 T32,
    T3 = (typ_sel T31 a T32) /\
    similar xs1 xs2 T1 T31 /\
    similar xs1 xs2 T2 T32.
Proof.
  introv Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (a0 /[ x0 --> xs2]) (a1 /[ x0 --> xs2]). sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists a1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_typ_bnd xs1 xs2 T1 T3:
  similar xs1 xs2 T1 T3 ->
  similar xs1 xs2 (typ_bnd T1) (typ_bnd T3).
Proof.
  unfold similar, similar_typ, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  exists (typ_bnd x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.
Lemma similar_typ_bnd_case xs1 xs2 T1 T3:
  similar xs1 xs2 (typ_bnd T1) T3 ->
  exists T31,
    T3 = (typ_bnd T31) /\
    similar xs1 xs2 T1 T31.
Proof.
  introv Hsi.
  unfold similar, similar_typ, similar_by_subst in Hsi.
  destruct_all.
  destruct x; autorewrite with simpl_subst_vars in *.
  all: subst; inversions H3.
  - exists  (x /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists x x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_typ_sel xs1 xs2 T31 T32 T1 T2 a :
  similar xs1 xs2 T1 T31 ->
  similar xs1 xs2 T2 T32 ->
  similar xs1 xs2 (typ_sel T1 a T2) (typ_sel T31 a T32).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv T2 \u fv T32) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv T1 \u fv T31) Hsi2 d.
  exists (typ_sel ds3 a d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Terms *)

(* Forward for trm_var *)
Lemma similar_trm_var xs1 xs2 w1 w2:
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 (trm_var w1) (trm_var w2).
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  exists (trm_var x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

(* Inversion for trm_var *)
Lemma similar_trm_var_var xs1 xs2 w1 w2:
  similar xs1 xs2 (trm_var w1) (trm_var w2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

Lemma similar_trm_var_case xs1 xs2 w1 t:
  similar xs1 xs2 (trm_var w1) t ->
  exists w2,
    t = (trm_var w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists  (a /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_trm_read xs1 xs2 w1 a w2 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 (trm_read w1 a) (trm_read w2 a).
Proof.
  introv Hsi.
  destruct_similar_by_subst (from_list (@nil var)) Hsi ds.
  exists (trm_read ds3 a) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
Qed.

(* Inversion for trm_read *)
Lemma similar_trm_read_var xs1 xs2 w1 w2 m1 m2:
  similar xs1 xs2 (trm_read w1 m1) (trm_read w2 m2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.
Lemma similar_trm_read_case xs1 xs2 w1 a t:
  similar xs1 xs2 (trm_read w1 a) t ->
  exists w2,
    t = (trm_read w2 a) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (a0 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.


Lemma similar_trm_write xs1 xs2 w1 a w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (trm_write w1 a w3) (trm_write w2 a w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (trm_write ds3 a d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Inversion for trm_write *)
Lemma similar_trm_write_var1 xs1 xs2 w1 w2 m1 m2 y1 y2:
  similar xs1 xs2 (trm_write w1 m1 y1) (trm_write w2 m2 y2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.
Lemma similar_trm_write_var2 xs1 xs2 w1 w2 m1 m2 y1 y2:
  similar xs1 xs2 (trm_write w1 m1 y1) (trm_write w2 m2 y2) ->
  similar xs1 xs2 y1 y2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a0 x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l2.
Qed.
Lemma similar_trm_write_case xs1 xs2 w1 a w3 t:
  similar xs1 xs2 (trm_write w1 a w3) t ->
  exists w2 w4,
    t = (trm_write w2 a w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (a0 /[ x0 --> xs2]) (a1 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists a1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_trm_apply xs1 xs2 w1 w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (trm_apply w1 w3) (trm_apply w2 w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (trm_apply ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Inversion for trm_apply *)
Lemma similar_trm_apply_var1 xs1 xs2 w1 w2 y1 y2:
  similar xs1 xs2 (trm_apply w1 y1) (trm_apply w2 y2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.
Lemma similar_trm_apply_var2 xs1 xs2 w1 w2 y1 y2:
  similar xs1 xs2 (trm_apply w1 y1) (trm_apply w2 y2) ->
  similar xs1 xs2 y1 y2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a0 x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l2.
Qed.
Lemma similar_trm_apply_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (trm_apply w1 w3) t ->
  exists w2 w4,
    t = (trm_apply w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (a /[ x0 --> xs2]) (a0 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_trm_call xs1 xs2 w1 m w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (trm_call w1 m w3) (trm_call w2 m w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (trm_call ds3 m d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Inversion for trm_call *)
Lemma similar_trm_call_var1 xs1 xs2 w1 w2 m1 m2 y1 y2:
  similar xs1 xs2 (trm_call w1 m1 y1) (trm_call w2 m2 y2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.
Lemma similar_trm_call_var2 xs1 xs2 w1 w2 m1 m2 y1 y2:
  similar xs1 xs2 (trm_call w1 m1 y1) (trm_call w2 m2 y2) ->
  similar xs1 xs2 y1 y2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a0 x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l2.
Qed.
Lemma similar_trm_call_case xs1 xs2 w1 a w3 t:
  similar xs1 xs2 (trm_call w1 a w3) t ->
  exists w2 w4,
    t = (trm_call w2 a w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (a0 /[ x0 --> xs2]) (a1 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists a1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_trm_lit xs1 xs2 w1 w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (trm_lit w1 w3) (trm_lit w2 w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (trm_lit ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Inversion for trm_lit *)
Lemma similar_trm_lit_lit xs1 xs2 w1 w2 y1 y2:
  similar xs1 xs2 (trm_lit w1 y1) (trm_lit w2 y2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists l x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.
Lemma similar_trm_lit_trm xs1 xs2 w1 w2 y1 y2:
  similar xs1 xs2 (trm_lit w1 y1) (trm_lit w2 y2) ->
  similar xs1 xs2 y1 y2.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists x x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l2.
Qed.
Lemma similar_trm_lit_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (trm_lit w1 w3) t ->
  exists w2 w4,
    t = (trm_lit w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (l /[ x0 --> xs2]) (x /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists l x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_trm_let xs1 xs2 w1 w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (trm_let w1 w3) (trm_let w2 w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (trm_let ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

(* Inversion for trm_let *)
Lemma similar_trm_let_1 xs1 xs2 t11 t12 t21 t22:
  similar xs1 xs2 (trm_let t11 t12) (trm_let t21 t22)%list ->
  similar xs1 xs2 t11 t21.
Proof.
  unfold similar, similar_trm, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists x1 x0; splits; eauto using subset_union_l1.
Qed.
Lemma similar_trm_let_2 xs1 xs2 t11 t12 t21 t22:
  similar xs1 xs2 (trm_let t11 t12) (trm_let t21 t22)%list ->
  similar xs1 xs2 t12 t22.
Proof.
  unfold similar, similar_trm, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists x2 x0; splits; eauto using subset_union_l2.
Qed.
Lemma similar_trm_let_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (trm_let w1 w3) t ->
  exists w2 w4,
    t = (trm_let w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (x1 /[ x0 --> xs2]) (x2 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists x1 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x2 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.


(* Literals *)

Lemma similar_lit_fun xs1 xs2 w1 w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (lit_fun w1 w3) (lit_fun w2 w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (lit_fun ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

Lemma similar_lit_fun_trm xs1 xs2 ds1 ds2 T1 T2:
  similar xs1 xs2 (lit_fun T1 ds1) (lit_fun T2 ds2) ->
  similar xs1 xs2 ds1 ds2.
Proof.
  unfold similar, similar_trm, similar_lit, similar_by_subst.
  intros.
  destruct_all; invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists t0 x0; splits; eauto using subset_union_l2.
Qed.
Lemma similar_lit_fun_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (lit_fun w1 w3) t ->
  exists w2 w4,
    t = (lit_fun w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_lit, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (t0 /[ x0 --> xs2]) (t1 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists t1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_lit_obj xs1 xs2 w1 w3 w2 w4 :
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 w3 w4 ->
  similar xs1 xs2 (lit_obj w1 w3) (lit_obj w2 w4).
Proof.
  introv Hsi1 Hsi2.
  destruct_similar_by_subst (fv w3 \u fv w4) Hsi1 ds.
  destruct_similar_by_subst_by xds3 (fv w1 \u fv w2) Hsi2 d.
  exists (lit_obj ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto; sympls;
    eauto.
  eapply subset_union_l; eauto.
Qed.

Lemma similar_lit_obj_defs xs1 xs2 ds1 ds2 T1 T2:
  similar xs1 xs2 (lit_obj T1 ds1) (lit_obj T2 ds2) ->
  similar xs1 xs2 ds1 ds2.
Proof.
  unfold similar, similar_def, similar_lit, similar_by_subst.
  intros.
  destruct_all; invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists d x0; splits; eauto using subset_union_l2.
Qed.
Lemma similar_lit_obj_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (lit_obj w1 w3) t ->
  exists w2 w4,
    t = (lit_obj w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_lit, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (t0 /[ x0 --> xs2]) (d /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists d x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

(* Definitions *)

(* Forward for def_met *)
Lemma similar_def_met xs1 xs2 m t1 t2 :
  similar xs1 xs2 t1 t2 ->
  similar xs1 xs2 (def_met m t1) (def_met m t2).
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros. invert_subst_vars_in H.
  destruct_all. exists (def_met m x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

(* Forward for def_typ *)
Lemma similar_def_typ xs1 xs2 A T1 T2 :
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 (def_typ A T1) (def_typ A T2).
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros. invert_subst_vars_in H.
  destruct_all. exists (def_typ A x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

(* Forward for def_fld *)
Lemma similar_def_fld xs1 xs2 w1 w2 a:
  similar xs1 xs2 w1 w2 ->
  similar xs1 xs2 (def_fld a w1) (def_fld a w2).
Proof.
  unfold similar, similar_trm, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars_in H.
  exists (def_fld a x) x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

(* Inversion for def_fld *)
Lemma similar_def_fld_var xs1 xs2 w1 w2 m1 m2:
  similar xs1 xs2 (def_fld m1 w1) (def_fld m2 w2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_def, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists a x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

Lemma similar_def_fld_case xs1 xs2 a w1 t:
  similar xs1 xs2 (def_fld a w1) t ->
  exists w2,
    t = (def_fld a w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_def, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (a0 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

(* Inversion for def_met *)
Lemma similar_def_met_trm xs1 xs2 w1 w2 m1 m2:
  similar xs1 xs2 (def_met m1 w1) (def_met m2 w2) ->
  similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_def, similar_trm, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists t x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; auto.
Qed.

Lemma similar_def_met_case xs1 xs2 a w1 t:
  similar xs1 xs2 (def_met a w1) t ->
  exists w2,
    t = (def_met a w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_def, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (t0 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_def_typ_case xs1 xs2 a w1 t:
  similar xs1 xs2 (def_typ a w1) t ->
  exists w2,
    t = (def_typ a w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_def, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (t1 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists t1 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

(* Definition lists *)

Lemma similar_defs_nil xs1 xs2:
  length xs1 = length xs2 ->
  similar xs1 xs2 (defs_nil) (defs_nil).
Proof.
  introv Hl.
  unfold similar, similar_defs, similar_def, similar_by_subst.
  destructc (var_freshes (from_list xs1 \u from_list xs2) (length xs1)) as x0 Hx0.
  exists defs_nil x0.
  pose proof (fresh_length _ _ _ Hx0).
  eapply fresh_resize_length in Hx0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  rewrite <- Hl. auto.
  intro. intro. rewrite in_empty in H0. contradiction.
Qed.
Lemma similar_defs_nil_case xs1 xs2 t:
  similar xs1 xs2 (defs_nil) t ->
  t = defs_nil.
Proof.
  unfold similar, similar_defs, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  destruct x; simpl_subst_vars_in H3; try discriminate.
  simpl_subst_vars_in H4. auto.
Qed.



Lemma similar_defs_cons xs1 xs2 ds1 ds2 d1 d2:
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 d1 d2 ->
  similar xs1 xs2 (defs_cons ds1 d1) (defs_cons ds2 d2).
Proof.
  introv Hds Hd.
  destruct_similar_by_subst (fv d1 \u fv d2) Hds ds.
  destruct_similar_by_subst_by xds3 (fv ds1 \u fv ds2) Hd d.
  exists (defs_cons ds3 d3) xds3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  eapply subset_union_l; eauto.
Qed.

Lemma similar_defs_cons_def xs1 xs2 ds1 ds2 d1 d2:
  similar xs1 xs2 (defs_cons ds1 d1) (defs_cons ds2 d2) ->
  similar xs1 xs2 d1 d2.
Proof.
  unfold similar, similar_defs, similar_def, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists d x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l2.
Qed.
Lemma similar_defs_cons_defs xs1 xs2 ds1 ds2 d1 d2:
  similar xs1 xs2 (defs_cons ds1 d1) (defs_cons ds2 d2) ->
  similar xs1 xs2 ds1 ds2.
Proof.
  unfold similar, similar_defs, similar_def, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  invert_subst_vars_in H3.
  exists x x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.
Lemma similar_defs_cons_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (defs_cons w1 w3) t ->
  exists w2 w4,
    t = (defs_cons w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_defs, similar_avar, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (x /[ x0 --> xs2]) (d /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists x x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists d x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.



(* Stack *)

Lemma similar_stack_nil xs1 xs2:
  (length xs1 = length xs2) ->
  similar xs1 xs2 stack_nil stack_nil.
Proof.
  unfold similar, similar_stack, similar_by_subst.
  introv Hl.
  destructc (var_freshes (from_list xs1 \u from_list xs2) (length xs1)) as x0 Hx0.
  exists (stack_nil) x0.

  pose proof (fresh_length _ _ _ Hx0).
  eapply fresh_resize_length in Hx0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  rewrite <- Hl. auto.
  intro. intro. rewrite in_empty in H0. contradiction.
Qed.
Lemma similar_stack_nil_case xs1 xs2 a:
  similar xs1 xs2 (stack_nil) a ->
  a = (stack_nil).
Proof.
  unfold similar, similar_stack, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars_in H3.
  invert_subst_vars_of x.
  autorewrite with simpl_subst_vars.
  auto.
Qed.

Lemma similar_stack_cons_case xs1 xs2 w1 w3 t:
  similar xs1 xs2 (stack_cons w1 w3) t ->
  exists w2 w4,
    t = (stack_cons w2 w4) /\
    similar xs1 xs2 w1 w2 /\
    similar xs1 xs2 w3 w4.
Proof.
  unfold similar, similar_stack, similar_frame, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (f /[ x0 --> xs2]) (x /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists f x0; splits; eauto using subset_union_l1, subset_union_l2.
    exists x x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.
Lemma similar_frame_let_case xs1 xs2 w1 t:
  similar xs1 xs2 (frame_let w1) t ->
  exists w2,
    t = (frame_let w2) /\
    similar xs1 xs2 w1 w2.
Proof.
  unfold similar, similar_stack, similar_frame, similar_by_subst.
  intros.
  destruct_all.
  invert_subst_vars.
  autorewrite with simpl_subst_vars in *.
  - exists (t0 /[ x0 --> xs2]); sympls.
    splits; eauto; unfold similar, similar_avar, similar_by_subst.
    exists t0 x0; splits; eauto using subset_union_l1, subset_union_l2.
Qed.

Lemma similar_stack_let xs1 xs2 t1 s1 t2 s2:
  similar xs1 xs2 t1 t2 ->
  similar xs1 xs2 s1 s2 ->
  similar xs1 xs2 (stack_cons (frame_let t1) s1) (stack_cons (frame_let t2) s2).
Proof.
  introv Hst Hss.
  destruct_similar_by_subst (fv s1 \u fv s2) Hst t.
  destruct_similar_by_subst_by xt3 (fv t1 \u fv t2) Hss s.

  exists  (stack_cons (frame_let t3) s3)  xt3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto.
  unfold fv in *; eauto.
  eapply subset_union_l; unfold fv in *; eauto.
Qed.

Lemma similar_stack_let_trm xs1 xs2 t1 s1 t2 s2:
  similar xs1 xs2 (stack_cons (frame_let t1) s1) (stack_cons (frame_let t2) s2) ->
  similar xs1 xs2 t1 t2.
Proof.
  unfold similar, similar_trm, similar_stack, similar_by_subst.
  intros.
  destructc H as ? ? ? ? ? ? Ha31 Ha32.
  invert_subst_vars.
  invert_subst_vars_in H4.
  invert_subst_vars.
  invert_subst_vars_in Ha31.
  exists t x0.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto using subset_union_l1.
Qed.

Lemma similar_stack_let_stack xs1 xs2 t1 s1 t2 s2:
  similar xs1 xs2 (stack_cons (frame_let t1) s1) (stack_cons (frame_let t2) s2) ->
  similar xs1 xs2 s1 s2.
Proof.
  unfold similar, similar_trm, similar_stack, similar_by_subst.
  intros.
  destructc H as a3 xs3 ? ? ? ? ? ?.
  invert_subst_vars.
  invert_subst_vars_in H3.
  invert_subst_vars.
  invert_subst_vars_in H6.
  exists a3 xs3.
  splits; eauto; autorewrite with simpl_subst_vars; subst; sympls; eauto  using subset_union_l2.
Qed.

Lemma similar_stack_cons_inv xs1 xs2 fs' f fs:
  similar xs1 xs2 fs' (stack_cons f fs) ->
  exists f' fs0', fs' = stack_cons f' fs0' /\
               similar xs1 xs2 fs0' fs.
Proof.
  intros Hsim. destruct fs'.
  - false. hnf in Hsim. destruct_all. destruct x.
    rewrite stack_subst_vars_nil in H4. discriminate.
    rewrite stack_subst_vars_cons in H3. discriminate.
  - exists f0 fs'. split; try (apply eq_refl).
    destruct f0. destruct f. eapply similar_stack_let_stack. eauto.
Qed.

Lemma stack_similar_length xs1 xs2 fs1 fs2 :
  similar xs1 xs2 fs1 fs2 ->
  stack_length fs1 = stack_length fs2.
Proof.
  intros Hsim. hnf in Hsim. destruct_all.
  rename x into fs.
  pose proof (subst_vars_stack_length fs x0 xs1).
  pose proof (subst_vars_stack_length fs x0 xs2).
  congruence.
Qed.

(* Heap items *)

(* Forward for item_obj *)
Lemma similar_item_obj xs1 xs2 ds1 ds2:
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 (item_obj ds1) (item_obj ds2).
Proof.
  unfold similar, similar_defs, similar_item, similar_by_subst.
  intros.
  destruct_all.
  exists (item_obj x) x0.
  splits; eauto.
  autorewrite with simpl_subst_vars. rewrite H3. auto.
  autorewrite with simpl_subst_vars. rewrite H4. auto.
Qed.
(* Inversion for item_obj *)
Lemma similar_item_obj_defs xs1 xs2 ds1 ds2:
  similar xs1 xs2 (item_obj ds1) (item_obj ds2) ->
  similar xs1 xs2 ds1 ds2.
Proof.
  unfold similar, similar_defs, similar_item, similar_by_subst.
  intros.
  destruct_all; invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists d x0; splits; eauto using subset_union_l1.
Qed.
Lemma similar_item_fun xs1 xs2 ds1 ds2:
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 (item_fun ds1) (item_fun ds2).
Proof.
  unfold similar, similar_trm, similar_item, similar_by_subst.
  intros.
  destruct_all.
  exists (item_fun x) x0.
  splits; eauto.
  autorewrite with simpl_subst_vars. rewrite H3. auto.
  autorewrite with simpl_subst_vars. rewrite H4. auto.
Qed.
Lemma similar_item_fun_trm xs1 xs2 ds1 ds2:
  similar xs1 xs2 (item_fun ds1) (item_fun ds2) ->
  similar xs1 xs2 ds1 ds2.

Proof.
  unfold similar, similar_trm, similar_item, similar_by_subst.
  intros.
  destruct_all; invert_subst_vars.
  invert_subst_vars_in H3.
  sympls.
  exists t x0; splits; eauto using subset_union_l1.
Qed.


Ltac similar_syntax_invert :=
  match goal with
  | HU: typ_nulary ?TU, H: similar ?xs1 ?xs2 ?TU ?T |- _ => eapply (similar_typ_nulary_case HU) in H
  | HU: typ_andor ?TU, H: similar ?xs1 ?xs2 (?TU ?t1 ?t2) ?T |- _ => eapply (similar_typ_andor_case t1 t2 HU) in H
  | H: similar ?xs1 ?xs2 (typ_and ?t1 ?t2) ?T |- _ =>  eapply (similar_typ_andor_case t1 t2 typ_andor_and) in H
  | H: similar ?xs1 ?xs2 (typ_or ?t1 ?t2) ?T |- _ =>  eapply (similar_typ_andor_case t1 t2 typ_andor_or) in H
  | H: similar ?xs1 ?xs2 (typ_bot) ?T |- _ =>  eapply (similar_typ_nulary_case typ_nulary_bot) in H
  | H: similar ?xs1 ?xs2 (typ_top) ?T |- _ =>  eapply (similar_typ_nulary_case typ_nulary_top) in H
  | H: similar ?xs1 ?xs2 (typ_nocap) ?T |- _ =>  eapply (similar_typ_nulary_case typ_nulary_nocap) in H
  | H: similar ?xs1 ?xs2 (typ_fun _ _) ?T |- _ => eapply similar_typ_fun_case in H
  | H: similar ?xs1 ?xs2 (typ_rcd _) ?T |- _ => eapply similar_typ_rcd_case in H
  | H: similar ?xs1 ?xs2 (dec_fld _ _ _) ?T |- _ => eapply similar_dec_fld_case in H
  | H: similar ?xs1 ?xs2 (dec_typ _ _ _) ?T |- _ => eapply similar_dec_typ_case in H
  | H: similar ?xs1 ?xs2 (dec_met _ _ _ _) ?T |- _ => eapply similar_dec_met_case in H
  | H: similar ?xs1 ?xs2 (typ_bnd _) ?T |- _ => eapply similar_typ_bnd_case in H
  | H: similar ?xs1 ?xs2 (typ_sel _ _ _) ?T |- _ => eapply similar_typ_sel_case in H
  | H: similar ?xs1 ?xs2 (avar_f _) ?T |- _ => eapply similar_avar_f_case in H
  | H: similar ?xs1 ?xs2 (avar_b _) ?T |- _ => eapply similar_avar_b_case in H
  | H: similar ?xs1 ?xs2 (trm_var _) ?T |- _ => eapply similar_trm_var_case in H
  | H: similar ?xs1 ?xs2 (trm_read _ _) ?T |- _ => eapply similar_trm_read_case in H
  | H: similar ?xs1 ?xs2 (trm_write _ _ _) ?T |- _ => eapply similar_trm_write_case in H
  | H: similar ?xs1 ?xs2 (trm_call _ _ _) ?T |- _ => eapply similar_trm_call_case in H
  | H: similar ?xs1 ?xs2 (trm_apply _ _) ?T |- _ => eapply similar_trm_apply_case in H
  | H: similar ?xs1 ?xs2 (trm_let _ _) ?T |- _ => eapply similar_trm_let_case in H
  | H: similar ?xs1 ?xs2 (trm_lit _ _) ?T |- _ => eapply similar_trm_lit_case in H
  | H: similar ?xs1 ?xs2 (lit_fun _ _) ?T |- _ => eapply similar_lit_fun_case in H
  | H: similar ?xs1 ?xs2 (lit_obj _ _) ?T |- _ => eapply similar_lit_obj_case in H
  | H: similar ?xs1 ?xs2 (def_fld _ _) ?T |- _ => eapply similar_def_fld_case in H
  | H: similar ?xs1 ?xs2 (def_typ _ _) ?T |- _ => eapply similar_def_typ_case in H
  | H: similar ?xs1 ?xs2 (def_met _ _) ?T |- _ => eapply similar_def_met_case in H
  | H: similar ?xs1 ?xs2 (defs_nil) ?T |- _ => eapply similar_defs_nil_case in H
  | H: similar ?xs1 ?xs2 (item_fun _) ?T |- _ => eapply similar_item_fun_trm in H
  | H: similar ?xs1 ?xs2 (item_obj _) ?T |- _ => eapply similar_item_obj_defs in H
  | H: similar ?xs1 ?xs2 (defs_cons _ _) ?T |- _ => eapply similar_defs_cons_case in H
  | H: similar ?xs1 ?xs2 (stack_nil) ?T |- _ => eapply similar_stack_nil_case in H
  | H: similar ?xs1 ?xs2 (stack_cons _ _) ?T |- _ => eapply similar_stack_cons_case in H
  | H: similar ?xs1 ?xs2 (frame_let _) ?T |- _ => eapply similar_frame_let_case in H
  end.
Ltac similar_syntax_inverts := repeat (similar_syntax_invert; destruct_all; subst).



Lemma similar_noduplicates_var_eq xs1 xs2 (y1 y2 z2: var):
  fresh \{} (length xs1) xs1 ->
  similar xs1 xs2 y1 y2 ->
  similar xs1 xs2 y1 z2 ->
  (y2 = z2).
Proof.
  intros.
  eapply similar_var_simlist_relates in H0.
  eapply similar_var_simlist_relates in H1.
  eauto using simlist_relates_noduplicates_var_eq.
Qed.
Lemma similar_noduplicates_avar_eq xs1 xs2 (y1 y2 z2: avar):
  fresh \{} (length xs1) xs1 ->
  similar xs1 xs2 y1 y2 ->
  similar xs1 xs2 y1 z2 ->
  (y2 = z2).
Proof.
  intros.
  destruct y1.
  + eapply similar_avar_b_case in H0.
    eapply similar_avar_b_case in H1.
    subst; auto.
  + eapply similar_avar_f_case in H0.
    eapply similar_avar_f_case in H1.
    destruct_all.
    forwards: similar_noduplicates_var_eq H2 H3; eauto.
    subst; auto.
Qed.

Lemma similar_noduplicates_var_cases xs1 xs2 (y1 y2 z1 z2: var):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 y1 y2 ->
  similar xs1 xs2 z1 z2 ->
  (y1 = z1) <-> (y2 = z2).
Proof.
  intros.
  eapply similar_var_simlist_relates in H1.
  eapply similar_var_simlist_relates in H2.
  eauto using simlist_relates_noduplicates_var_cases.
Qed.

(* DEFS *)
Lemma similar_defs_has xs1 xs2 (ds1 ds2: defs) d1 d2:
  similar xs1 xs2 ds1 ds2 ->
  label_of d1 = label_of d2 ->
  defs_has ds1 d1 ->
  defs_has ds2 d2 ->
  similar xs1 xs2 d1 d2.
Proof.
  introv Hs Hlbl Hhas1 Hhas2.
  reverts Hs. reverts d2 ds2.
  eapply (defs_has_ind Hhas1); intros.
  - (* Eq case *)
    unfold similar, similar_defs, similar_def, similar_by_subst.
    unfold similar, similar_defs, similar_def, similar_by_subst in Hs.
    destruct_all.
    (* destruct definition lists *)
    invert_subst_vars_in H3.
    invert_subst_vars.
    invert_subst_vars_in Hhas1.
    invert_subst_vars_in Hhas2.
    (* Compare labels *)
    rewrite <- subst_vars_label_of_def in H4.
    rewrite <- subst_vars_label_of_def in Hlbl.
    rewrite <- subst_vars_label_of_def in H5.
    cases_if.
    inversions H5.
    clear C Hlbl.
    autorewrite with simpl_subst_vars in *.
    sympls.
    exists d0 x0; splits; eauto using subset_union_l2.
  - unfold similar, similar_defs, similar_def, similar_by_subst.
    unfold similar, similar_defs, similar_def, similar_by_subst in Hs.
    destruct_all.
    (* destruct definition lists *)
    invert_subst_vars_in H5.
    invert_subst_vars.
    invert_subst_vars_in Hhas1.
    invert_subst_vars_in Hhas2.
    rewrite <- subst_vars_label_of_def in H0.
    rewrite <- subst_vars_label_of_def in H7.
    cases_if.
    eapply H; eauto.
    unfold similar, similar_defs, similar_def, similar_by_subst.
    autorewrite with simpl_subst_vars in *.
    sympls.
    exists x x0; splits; eauto using subset_union_l1.
Qed.

Lemma similar_defs_hasnt xs1 xs2 (ds1 ds2: defs) d1 d2:
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 d1 d2 ->
  defs_hasnt ds1 (label_of d1) ->
  defs_hasnt ds2 (label_of d2).
Proof.
  introv Hsids Hsid Hhasnt.
  unfold defs_hasnt in *.
  reverts ds2.
  induction ds1; intros.
  - eapply similar_defs_nil_case in Hsids; subst. eauto.
  - forwards ds2' d2' Hds2 Hsids2 Hsid2: (similar_defs_cons_case Hsids).
    subst.
    assert (label_of d2' = label_of d).
    destruct d; similar_syntax_invert; destruct_all; subst; sympls; auto.
    assert (label_of d1 = label_of d2).
    destruct d1; similar_syntax_invert; destruct_all; subst; sympls; auto.
    inversions Hhasnt. rewrite H2. cases_if.
    forwards: IHds1; eauto.
    assert (label_of d2 <> label_of d2').
    rewrite H. rewrite <- H0. intro. symmetry in H3. contradiction.
    unfold get_def. fold get_def. cases_if. eauto.
Qed.

Lemma subst_heap_twice_same Sigma x1 x2 :
  subst_heap x1 x2 (subst_heap x1 x2 Sigma) = subst_heap x1 x2 Sigma.
Proof.
  induction Sigma.
  - unfold subst_heap. rewrite <- empty_def. repeat(rewrite map_empty). easy.
  - unfold subst_heap in *. destruct a.
    rewrite cons_to_push. repeat(rewrite map_push).
    repeat(rewrite subst_twice_same). rewrite IHSigma. easy.
Qed.

(* Updates *)

Lemma defs_update_similar ds1 ds2 a y1 y2 ds1' ds2' Lt1 Lt2 :
  defs_update ds1 a y1 ds1' ->
  defs_update ds2 a y2 ds2' ->
  similar Lt1 Lt2 ds1 ds2 ->
  similar Lt1 Lt2 y1 y2 ->
  similar Lt1 Lt2 ds1' ds2'.
Proof.
  introv Hupd1. reverts ds2 ds2'.
  induction Hupd1; introv Hupd2 Hsimds Hsimy; eauto.

  - reverts ds. induction Hupd2; intros.
    + similar_syntax_inverts. inversions H.
      eauto using similar_def_fld, similar_avar_f, similar_defs_cons.
    + destruct d; similar_syntax_inverts; try discriminate. inversions H0.
      simpl in H. false.

  - reverts ds d ds'. induction Hupd2; intros.
    + similar_syntax_inverts. inversions H0.
      destruct d; similar_syntax_inverts; try discriminate. inversions H0.
      simpl in H. false.
    + similar_syntax_inverts. inversions H1.
      eapply similar_defs_cons; eauto.

Qed.

Lemma heap_update_similar S1 S2 S1' S2' yObj1 yObj2 a yArg1 yArg2
  LS1 LS2 Lt1 Lt2 :
  fresh \{} (length LS1) LS1 ->
  fresh \{} (length LS2) LS2 ->
  heap_update S1 yObj1 a yArg1 S1' ->
  heap_update S2 yObj2 a yArg2 S2' ->
  similar Lt1 Lt2 yArg1 yArg2 ->
  similar LS1 LS2 yObj1 yObj2 ->
  similar_with Lt1 Lt2 LS1 LS2 S1  S2 ->
  similar_with Lt1 Lt2 LS1 LS2 S1' S2'.
Proof.
  introv Hnd1 Hnd2 Hupd1 Hupd2 HsimArg HsimObj HsimS.
  eapply similar_env_with_equiv in HsimS as [HlLt [HlS [Hdom Hsim]]].
  eapply similar_env_with_equiv. splits; eauto.

  { introv Hrel.
    destruct Hupd1 as [_ [_ [Heq1 _]]]; rewrite <- Heq1.
    destruct Hupd2 as [_ [_ [Heq2 _]]]; rewrite <- Heq2. eauto. }

  introv Hrel Hbi1 Hbi2.
  destruct (dup_lemma Hrel) as [n [Hxn1 Hxn2]]; eauto.
  eapply similar_var_simlist_relates in HsimObj.
  destruct (dup_lemma HsimObj) as [m [Hxm1 Hxm2]]; eauto.
  destruct (Peano_dec.dec_eq_nat n m) as [? | Hnm]; subst.

  - eapply (LibList.Nth_functional Hxn1) in Hxm1; subst.
    eapply (LibList.Nth_functional Hxn2) in Hxm2; subst.
    destruct Hupd1 as [_ [_ [_ [_ [ds1 [ds1' [HbiS1 [Hupd1 HbiS1']]]]]]]].
    destruct Hupd2 as [_ [_ [_ [_ [ds2 [ds2' [HbiS2 [Hupd2 HbiS2']]]]]]]].
    eapply (binds_functional Hbi1) in HbiS1'; subst.
    eapply (binds_functional Hbi2) in HbiS2'; subst.
    specialize (Hsim _ _ _ _ HsimObj HbiS1 HbiS2).
    similar_syntax_inverts. eapply similar_item_obj.
    eauto using defs_update_similar.

  - assert (Hne1 : x1 <> yObj1).
    { hnf; intros; subst. eapply (fresh_nth_eq _ Hnd1 Hxn1) in Hxm1; eauto. }
    assert (Hne2 : x2 <> yObj2).
    { hnf; intros; subst. eapply (fresh_nth_eq _ Hnd2 Hxn2) in Hxm2; eauto. }
    destruct Hupd1 as [_ [_ [Heq1 [Hupd1 _]]]].
    destruct Hupd2 as [_ [_ [Heq2 [Hupd2 _]]]].
    eapply binds_to_dom in Hbi1 as Hbi1'; rewrite <- Heq1 in Hbi1';
      eapply dom_to_binds in Hbi1' as [it1' Hbi1'];
      specialize (Hupd1 _ _ Hne1 Hbi1');
      eapply (binds_functional Hupd1) in Hbi1; subst.
    eapply binds_to_dom in Hbi2 as Hbi2'; rewrite <- Heq2 in Hbi2';
      eapply dom_to_binds in Hbi2' as [it2' Hbi2'];
      specialize (Hupd2 _ _ Hne2 Hbi2');
      eapply (binds_functional Hupd2) in Hbi2; subst.
    eauto.

Qed.

Lemma deref_def_similar e1 e2 Lt1 Lt2 d1 d2 :
  length Lt1 = length Lt2                            ->
  ok e1 -> ok e2                                      ->
  renv_idempotent e1 -> renv_idempotent e2            ->
  fields d1 \c dom e1 -> fields d2 \c dom e2          ->
  (forall w1 w2 y1 y2, similar Lt1 Lt2 w1 w2 ->
                  binds w1 y1 e1 ->
                  binds w2 y2 e2 ->
                  similar Lt1 Lt2 y1 y2)             ->
  similar Lt1 Lt2               d1                d2 ->
  similar Lt1 Lt2 (deref_def e1 d1) (deref_def e2 d2).
Proof.
  introv Hlen Hok1 Hok2 Hidm1 Hidm2 Hflds1 Hflds2 Hsime Hsimd.
  destruct d1; destruct d2; simpl in *; similar_syntax_inverts;
    try discriminate; inversions H; eauto;
    [ eapply similar_def_typ; eauto
    | eapply similar_def_fld; eauto
    | eapply similar_def_met; eauto ].

  rename a into w1, x into w2.
  destruct w1; destruct w2; similar_syntax_inverts; try discriminate.
  all: repeat(rewrite deref_avar_b).
  inversions H0. eapply similar_avar_b; try easy.
  inversions H. repeat(rewrite deref_avar_comm).
  eapply similar_avar_f. rename v into w1, x into w2.
  eapply deref_var_cases with (x:=w1) in Hidm1; eauto.
  destruct Hidm1 as [[_ Hc] | [y1 [? Hbi1]]]; subst.
  { false. unfold fields, FieldsDef, fields_def in Hflds1.
    specialize (Hflds1 _ (in_singleton_self _)). eauto. }
  eapply deref_var_cases with (x:=w2) in Hidm2; eauto.
  destruct Hidm2 as [[_ Hc] | [y2 [? Hbi2]]]; subst.
  { false. unfold fields, FieldsDef, fields_def in Hflds2.
    specialize (Hflds2 _ (in_singleton_self _)). eauto. }
  eauto.
Qed.

Lemma deref_defs_similar e1 e2 Lt1 Lt2 ds1 ds2 :
  length Lt1 = length Lt2                                ->
  ok e1 -> ok e2                                          ->
  renv_idempotent e1 -> renv_idempotent e2                ->
  fields ds1 \c dom e1 -> fields ds2 \c dom e2            ->
  (forall w1 w2 y1 y2, similar Lt1 Lt2 w1 w2 ->
                  binds w1 y1 e1 ->
                  binds w2 y2 e2 ->
                  similar Lt1 Lt2 y1 y2)                 ->
  similar Lt1 Lt2                ds1                 ds2 ->
  similar Lt1 Lt2 (deref_defs e1 ds1) (deref_defs e2 ds2).
Proof.
  introv Hlen Hok1 Hok2 Hidm1 Hidm2 Hflds1 Hflds2 Hsime Hsimd.
  reverts ds2.

  induction ds1; intros; destruct ds2; simpl in *; similar_syntax_inverts;
    try discriminate.
  eapply similar_defs_nil; eauto.

  inversions H.
  rename d into d1, x into ds2, x0 into d2, H0 into Hsimds, H1 into Hsimd.
  eapply similar_defs_cons. eapply IHds1; eauto.
  3: eapply deref_def_similar; eauto.
  all: eauto using subset_trans, subset_union_weak_l, subset_union_weak_r.
Qed.

Lemma similar_one_side_subst_var (x : var) xs ys Ltx Lty :
  length Ltx = length Lty ->
  length xs = length ys ->
  fresh (from_list ys) (length xs) xs ->
  similar Ltx Lty x x ->
  (forall n x y, LibList.Nth n xs x ->
            LibList.Nth n ys y ->
            similar Ltx Lty x y) ->
  similar Ltx Lty x (x /[ xs --> ys]).
Proof.
  introv HLt Hlen Hfr Hsimx Hsims.
  destruct (fv_decide x (from_list xs)) as [Hin | Hnin].

  - eapply list_in_fromlist_mem, LibList.mem_Nth in Hin as
        [n Hnx].
    forwards:LibList.Nth_inbound_inv.
    { eapply LibList.Nth_inbound in Hnx. rewrite liblist_length in *.
      rewrite Hlen in Hnx. exact Hnx. }
    destructc H as y Hny.
    erewrite var_subst_vars_Nth.
    4: { exact Hnx. } 4: { exact Hny. } all: eauto.

  - rewrite subst_vars_fresh; eauto.
    unfold fv, VarFreeVar. eapply fresh_from_list; eauto.
Qed.

Lemma similar_one_side_subst_mutind_typ :
  (forall (t : typ) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv t -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty t (t /[ xs --> ys]))
  /\ (forall (d : dec) xs ys Ltx Lty,
        length Ltx = length Lty ->
        length xs = length ys ->
        fresh (from_list ys) (length xs) xs ->
        (forall v, v \in fv d -> similar Ltx Lty v v) ->
        (forall n x y, LibList.Nth n xs x ->
                  LibList.Nth n ys y ->
                  similar Ltx Lty x y) ->
        similar Ltx Lty d (d /[ xs --> ys])).
Proof.
  apply typ_mutind; intros; simpl in *.

  all:
    try solve[ rewrite typ_subst_vars_nulary; eauto;
               eapply similar_typ_nulary; eauto
             | rewrite typ_subst_vars_rcd; eauto;
               eapply similar_typ_rcd; eauto
             | rewrite typ_subst_vars_bnd; eauto;
               eapply similar_typ_bnd; eauto ].

  - rewrite typ_subst_vars_andor; eauto.
    eapply similar_typ_andor; eauto.
    eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; repeat(rewrite in_union); tauto.

  - rewrite typ_subst_vars_andor; eauto.
    eapply similar_typ_andor; eauto.
    eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; repeat(rewrite in_union); tauto.

  - rewrite typ_subst_vars_sel; eauto.
    eapply similar_typ_sel.
    1: rename a into x. 2: rename a0 into x.
    all: destruct x; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    all: eapply similar_avar_f, similar_one_side_subst_var; eauto.
    all: eapply H2; simpl; rewrite in_union, in_singleton; eauto.

  - rewrite typ_subst_vars_fun; eauto.
    eapply similar_typ_fun; eauto.
    eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; repeat(rewrite in_union); tauto.

  - rewrite dec_subst_vars_typ; eauto.
    eapply similar_dec_typ; eauto.
    eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; repeat(rewrite in_union); tauto.

  - rewrite dec_subst_vars_fld; eauto.
    eapply similar_dec_fld; eauto.
    eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; repeat(rewrite in_union); tauto.

  - rewrite dec_subst_vars_met; eauto.
    eapply similar_dec_met; eauto.
    eapply H; eauto. 2: eapply H0; eauto. 3: eapply H1; eauto.
    all: intros; eapply H5; repeat(rewrite in_union); tauto.

Qed.

Lemma similar_one_side_subst_dec :
  (forall (d : dec) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv d -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty d (d /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_typ. easy. Qed.

Lemma similar_one_side_subst_typ :
  (forall (t : typ) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv t -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty t (t /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_typ. easy. Qed.

Lemma similar_one_side_subst_mutind_trm :
  (forall (t : trm) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv t -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty t (t /[ xs --> ys]))
  /\ (forall (l : lit) xs ys Ltx Lty,
        length Ltx = length Lty ->
        length xs = length ys ->
        fresh (from_list ys) (length xs) xs ->
        (forall v, v \in fv l -> similar Ltx Lty v v) ->
        (forall n x y, LibList.Nth n xs x ->
                  LibList.Nth n ys y ->
                  similar Ltx Lty x y) ->
        similar Ltx Lty l (l /[ xs --> ys]))
  /\ (forall (d : def) xs ys Ltx Lty,
        length Ltx = length Lty ->
        length xs = length ys ->
        fresh (from_list ys) (length xs) xs ->
        (forall v, v \in fv d -> similar Ltx Lty v v) ->
        (forall n x y, LibList.Nth n xs x ->
                  LibList.Nth n ys y ->
                  similar Ltx Lty x y) ->
        similar Ltx Lty d (d /[ xs --> ys]))
  /\ (forall (ds : defs) xs ys Ltx Lty,
        length Ltx = length Lty ->
        length xs = length ys ->
        fresh (from_list ys) (length xs) xs ->
        (forall v, v \in fv ds -> similar Ltx Lty v v) ->
        (forall n x y, LibList.Nth n xs x ->
                  LibList.Nth n ys y ->
                  similar Ltx Lty x y) ->
        similar Ltx Lty ds (ds /[ xs --> ys])).
Proof.
  eapply trm_mutind; intros; simpl in *.

  - rewrite trm_subst_vars_var; eauto.
    eapply similar_trm_var; eauto.
    destruct a; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    eapply similar_avar_f, similar_one_side_subst_var; eauto.
    eapply H2; simpl; rewrite in_singleton; eauto.

  - rewrite trm_subst_vars_read; eauto.
    eapply similar_trm_read; eauto.
    destruct a; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    eapply similar_avar_f, similar_one_side_subst_var; eauto.
    eapply H2; simpl; rewrite in_singleton; eauto.

  - rewrite trm_subst_vars_apply; eauto.
    eapply similar_trm_apply; eauto.
    1: rename a into x. 2: rename a0 into x.
    all: destruct x; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    all: eapply similar_avar_f, similar_one_side_subst_var; eauto.
    all: eapply H2; simpl; rewrite in_union, in_singleton; eauto.

  - rewrite trm_subst_vars_let; eauto.
    eapply similar_trm_let; eauto.
    1: eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; rewrite in_union; tauto.

  - rewrite trm_subst_vars_lit; eauto.
    eapply similar_trm_lit; eauto.
    1: eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; rewrite in_union; tauto.

  - rewrite trm_subst_vars_write; eauto.
    eapply similar_trm_write; eauto.
    1: rename a into x. 2: rename a0 into x.
    all: destruct x; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    all: eapply similar_avar_f, similar_one_side_subst_var; eauto.
    all: eapply H2; simpl; rewrite in_union, in_singleton; eauto.

  - rewrite trm_subst_vars_call; eauto.
    eapply similar_trm_call; eauto.
    1: rename a into x. 2: rename a0 into x.
    all: destruct x; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    all: eapply similar_avar_f, similar_one_side_subst_var; eauto.
    all: eapply H2; simpl; rewrite in_union, in_singleton; eauto.

  - rewrite lit_subst_vars_fun; eauto.
    eapply similar_lit_fun; eauto.
    1: eapply similar_one_side_subst_typ; eauto.
    2: eapply H; eauto.
    all: intros; eapply H3; rewrite in_union; tauto.

  - rewrite lit_subst_vars_obj; eauto.
    eapply similar_lit_obj; eauto.
    1: eapply similar_one_side_subst_typ; eauto.
    2: eapply H; eauto.
    all: intros; eapply H3; rewrite in_union; tauto.

  - rewrite def_subst_vars_typ; eauto.
    eapply similar_def_typ; eauto.
    eapply similar_one_side_subst_typ; eauto.

  - rewrite def_subst_vars_fld; eauto.
    eapply similar_def_fld; eauto.
    destruct a; try rewrite avar_subst_vars_b;
      try rewrite avar_subst_vars_f; try eapply similar_avar_b; eauto.
    eapply similar_avar_f, similar_one_side_subst_var; eauto.
    eapply H2; simpl; rewrite in_singleton; eauto.

  - rewrite def_subst_vars_met; eauto.
    eapply similar_def_met; eauto.

  - rewrite defs_subst_vars_nil; eauto.
    eapply similar_defs_nil; eauto.

  - rewrite defs_subst_vars_cons; eauto.
    eapply similar_defs_cons; eauto.
    1: eapply H; eauto. 2: eapply H0; eauto.
    all: intros; eapply H4; rewrite in_union; tauto.
Qed.

Lemma similar_one_side_subst_trm :
  (forall (t : trm) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv t -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty t (t /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_trm; easy. Qed.

Lemma similar_one_side_subst_lit :
  (forall (l : lit) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv l -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty l (l /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_trm; easy. Qed.

Lemma similar_one_side_subst_def :
  (forall (d : def) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv d -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty d (d /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_trm; easy. Qed.

Lemma similar_one_side_subst_defs :
  (forall (ds : defs) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv ds -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty ds (ds /[ xs --> ys])).
Proof. pose proof similar_one_side_subst_mutind_trm; easy. Qed.

Lemma similar_one_side_subst_item :
  (forall (itm : item) xs ys Ltx Lty,
      length Ltx = length Lty ->
      length xs = length ys ->
      fresh (from_list ys) (length xs) xs ->
      (forall v, v \in fv itm -> similar Ltx Lty v v) ->
      (forall n x y, LibList.Nth n xs x ->
                LibList.Nth n ys y ->
                similar Ltx Lty x y) ->
      similar Ltx Lty itm (itm /[ xs --> ys])).
Proof.
  intros. destruct itm; simpl in *.
  - rewrite item_subst_vars_obj. eapply similar_item_obj.
    eapply similar_one_side_subst_defs; eauto.
  - rewrite item_subst_vars_fun. eapply similar_item_fun.
    eapply similar_one_side_subst_trm; eauto.
Qed.
