Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax GeneralTyping Generalization Equivalence
        Similarity ContextSimilarity SyntaxSimilarity OperationalSemantics HeapUpdate Deref Defs ConfigValid.
Require Import UnboundSubstitution.

(** * Similarity of typing *)

Existing Instance similar_ctx_with.
Existing Instance similar_ctx.
Existing Instance similar_objctx.

(* Deal with objctx *)
Ltac similar_objctx_invert := (
  match goal with
  | H: similar _ _ objctx_empty _ |- _ =>
      inversions H
  | H: similar _ _ (objctx_lit _ _) _ |- _ =>
      inversions H
  | H: similar _ _ (objctx_heap _ _) _ |- _ =>
      inversions H
  end
).
Ltac similar_objctx_simplify G := (
  replace (G §§& objctx_empty) with G in * by (unfold ctx_obj; simpl; auto)
).

Class TypingSimilarity (A B: Set) (AS: Similar A)(BS: Similar B) (property: TypingProperty A B):=
  typing_similarity_obj xs1 xs2 G1 G2 Go1 Go2 e1 e2 (x1 x2: A) (T1 T2: B):
  ok e1 ->
  ctx_ok (G2 §§& Go2) ->
  disjoint (dom e1) (fv_env e1) ->
  dom e1 \c ctx_dom (G1 §§& Go1) ->
  dom e2 \c ctx_dom (G2 §§& Go2) ->
  ctx_dom (G1 §§& Go1) \c (from_list xs1) ->
  fresh \{} (length xs1) xs1 -> (* Needed for var equivalence in renv *)
  fresh \{} (length xs2) xs2 -> (* Needed for subtyping (and, or) *)
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 G1 G2->
  similar xs1 xs2 Go1 Go2 ->
  property G1 Go1 e1 x1 T1 ->
  property G2 Go2 e2 x2 T2.
Section TypingSimilarity.
  Context {A B: Set} {AS: Similar A}{BS: Similar B}{property: TypingProperty A B}.
  Context {TS: TypingSimilarity AS BS property}.
  Lemma typing_similarity xs1 xs2 G1 G2 e1 e2 (x1 x2: A) (T1 T2: B):
  ok e1 ->
  ctx_ok G2 ->
  disjoint (dom e1) (fv_env e1) ->
  dom e1 \c ctx_dom G1 ->
  dom e2 \c ctx_dom G2 ->
  ctx_dom G1 \c (from_list xs1) ->
  fresh \{} (length xs1) xs1 -> (* Needed for var equivalence in renv *)
  fresh \{} (length xs2) xs2 -> (* Needed for subtyping (and, or) *)
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 G1 G2->
  property G1 objctx_empty e1 x1 T1 ->
  property G2 objctx_empty e2 x2 T2.
  Proof.
  intros.
  eapply typing_similarity_obj; similar_objctx_simplify G1; similar_objctx_simplify G2; eauto.
  Qed.
End TypingSimilarity.
(* Notation for easy naming of hypotheses, should help save some renaming in proofs *)
Notation "x @@ n -> L" := (forall (n:x), L)  (only parsing, at level 99, L at level 200, n at level 98).

Definition TypingSimilarityInd (A B: Set) (AS: Similar A)(BS: Similar B) (property: TypingProperty A B):=
  forall G1 Go1 e1 x1 T1,
  property G1 Go1 e1 x1 T1 ->
  forall xs1 xs2 G2 Go2 e2 (x2: A) (T2: B),
  fresh \{} (length xs1) xs1 @@ Hfrxs1 ->
  fresh \{} (length xs2) xs2 @@ Hfrxs2 ->
  ok e1 @@ Hoke1 ->
  ctx_ok (G2 §§& Go2) @@ HokG2 ->
  disjoint (dom e1) (fv_env e1) @@ Hne1 ->
  dom e1 \c ctx_dom (G1 §§& Go1) @@ Hse1G1 ->
  dom e2 \c ctx_dom (G2 §§& Go2) @@ Hse2G2 ->
  ctx_dom (G1 §§& Go1) \c (from_list xs1) @@ HsG1xs1 ->
  similar xs1 xs2 e1 e2 @@ Hsie ->
  similar xs1 xs2 x1 x2 @@ Hsix ->
  similar xs1 xs2 T1 T2 @@ HsiT ->
  similar xs1 xs2 G1 G2 @@ HsiG ->
  similar xs1 xs2 Go1 Go2 @@ HsiGo ->
  property G2 Go2 e2 x2 T2.
Definition TypingSimilarityTInd (A: Set) (t: Typing A) (AS: Similar A) (TS: Similar typ) :=
  TypingSimilarityInd AS TS (obj_typing).
Definition TypingSimilarityIncapInd (_: ReferenceIncapacitate) (TS: Similar typ)  :=
  TypingSimilarityInd TS TS obj_ref_incap.
Definition TypingSimilarityCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C) (TS: Similar typ)  :=
  TypingSimilarityInd TS TS obj_cap_bnd.
Definition TypingSimilarityDInd (A: Set) (t: DecTyping A) (AS: Similar A) (TS: Similar dec)  :=
  TypingSimilarityInd AS TS (obj_dec_typing).


Require Import SimilarityPreservation.

Lemma similar_ctx_with_binds_to_binds xs1 xs2 ys1 ys2 (S1 S2: ctx) x1 x2 it1 it2:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 -> (* Perhaps not necessary, see subst_vars_unique *)
  similar_with xs1 xs2 ys1 ys2 S1 S2 ->
  simlist_relates ys1 ys2 x1 x2 ->
  ctx_binds x1 it1 S1 ->
  similar xs1 xs2 it1 it2 ->
  ctx_binds x2 it2 S2.
Proof.
  introv Hfrx1 Hfrx2.
  intros.
  forwards: (similar_ctx_cases H0 H); eauto.
  destruct_all.
  - eapply ctx_binds_dom in H1. contradiction.
  - ctx_binds_eq.
    unfold similar, similar_typ in H2, H5.
    assert (it2=x0).
    (eapply similar_by_subst_noduplicates_eq_l with (xs:=xs1) (ys:=xs2); eauto).
    subst; eauto.
Qed.

Lemma similar_ctx_binds_to_binds xs1 xs2 (S1 S2: ctx) x1 x2 it1 it2:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 -> (* Perhaps not necessary, see subst_vars_unique *)
  similar xs1 xs2 S1 S2 ->
  simlist_relates xs1 xs2 x1 x2 ->
  ctx_binds x1 it1 S1 ->
  similar xs1 xs2 it1 it2 ->
  ctx_binds x2 it2 S2.
Proof.
  introv Hfrx1 Hfrx2 HsiG.
  intros.
  induction HsiG.
  - eapply similar_ctx_with_binds_to_binds with (xs1:=x0) (xs2:=x3); eauto.
  - inversions H0.
    forwards: simlist_relates_noduplicates_var_eq H3 H; eauto. subst.

    forwards: (similar_by_subst_noduplicates_eq_l Hfrx2 Hfrx1 H2 H1).
    subst. eauto.
    forwards: simlist_relates_noduplicates_var_cases H3 H; eauto.
    forwards: IHHsiG; eauto. constructor. intro. symmetry in H6. apply <- H0 in H6. symmetry in H6. contradiction. eauto.
  - inversions H0.
    constructor; eauto.
Qed.

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma similar_tyq_var xs1 xs2 e1 e2 (x1 x2 T1 T2: var):
  fresh \{} (length xs1) xs1 ->
  ok e1 ->
  disjoint (dom e1) (fv_env e1) ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  typing_equiv e1 x1 T1 ->
  typing_equiv e2 x2 T2.
Proof.
  introv Hfrxs1 Hoke1 Hdie1 Hsie Hsiv HsiT Htyq.
  apply -> tyq_var_to_tyq_var_alt in Htyq; eauto.
  induction Htyq.
  - forwards: similar_noduplicates_var_eq Hsiv HsiT; subst; eauto.
  - forwards: similar_renv_binds_to_binds Hsiv HsiT; eauto.
  - forwards: similar_renv_binds_to_binds HsiT Hsiv; eauto.
  - forwards: similar_renv_binds_to HsiT; eauto.
    forwards: similar_renv_binds_to Hsiv; eauto.
    destruct_all.
    forwards: similar_noduplicates_var_eq H3 H4; subst; eauto.
Qed.

Instance typing_similarity_tyq_var: TypingSimilarity similar_var similar_var (fun G Go e (a b: var) => typing_equiv e a b).
Proof.
  hnf. intros.
  eapply similar_tyq_var with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.


Lemma similar_tyq_typ xs1 xs2 e1 e2 (x1 x2 T1 T2: typ):
  fresh \{} (length xs1) xs1 ->
  ok e1 ->
  disjoint (dom e1) (fv_env e1) ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  typing_equiv e1 x1 T1 ->
  typing_equiv e2 x2 T2.
Proof.
  hnf. introv Hfr Hoke Hdise Hsie Hsix HsiT Htyq.
  reverts e2 x2 T2.
  reverts Hoke Hdise Hfr.
  reverts xs1 xs2.
  induction Htyq; intros; similar_syntax_inverts.
  - (* nulary *)
    eauto.
  - (* andor *)
    econstructor. eauto. eapply IHHtyq; eauto.  eapply IHHtyq0; eauto.
  - (* fun *)
    eapply tyq_typ_fun.
    eapply IHHtyq; eauto.
    intro x2.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    applys tyq_typ_reopen z. sympls. notin_solve; eapply notin_fv_subst_vars; eauto.
    eapply (H1 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* fld *)
    eauto.
  - (* typ *)
    eapply tyq_typ_typ.
    intro y2.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    applys tyq_typ_reopen z. sympls. notin_solve; eapply notin_fv_subst_vars; eauto.
    eapply (H0 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
    intro y2.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    applys tyq_typ_reopen z. sympls. notin_solve; eapply notin_fv_subst_vars; eauto.
    eapply (H2 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* sel_normal *)
    forwards: similar_noduplicates_avar_eq H2 H4; eauto; subst.
    eapply tyq_typ_sel_normal.
    eapply similar_tyq_var; eauto.
  - (* sel *)
    forwards: similar_noduplicates_avar_eq H1 H3; eauto; subst.
    forwards: similar_noduplicates_avar_eq H0 H2; eauto; subst.
    eapply tyq_typ_sel.
  - (* bnd *)
    eapply tyq_typ_bnd.
    intro x2.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    applys tyq_typ_reopen z. sympls. notin_solve; eapply notin_fv_subst_vars; eauto.
    eapply (H0 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* met *)
    eapply tyq_typ_met.
    intro y2.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    applys tyq_typ_reopen z. sympls. notin_solve; eapply notin_fv_subst_vars; eauto.
    eapply (H0 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
    eapply IHHtyq; eauto.
    intros y2 y3.
    pick_fresh_with z ((from_list xs2) \u (from_list xs1)).
    pick_fresh_with z2 ((from_list xs2) \u (from_list xs1)).
    eapply tyq_typ_reopen_met with (x1:= z) (x2:=z2).
    eauto.
    eauto.
    eapply (H3 z z2 (z::z2::xs1)%list (z::z2::xs2)%list); eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_extend_notin; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
Qed.

Global Instance typing_similarity_tyq_typ: TypingSimilarity similar_typ similar_typ (fun G Go e (a b: typ) => typing_equiv e a b).
Proof.
  hnf. intros.
  eapply similar_tyq_typ with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.

Lemma similar_typ_indep xs1 xs2 n T1 T2:
  similar xs1 xs2 T1 T2 ->
  Indep.typ_indep (avar_b n) T1 ->
  Indep.typ_indep (avar_b n) T2.
Proof.
  introv Hsi Hidp.
  remember (avar_b n). reverts T2 n.
  induction Hidp; intros; subst.
  all: similar_syntax_inverts; eauto.
  -
    destruct v1; destruct v2; similar_syntax_inverts; eauto.
    all: eapply Indep.typ_indep_sel; eauto; discriminate.
  -
    destruct v1; destruct v2; similar_syntax_inverts; eauto.
    all: eapply Indep.typ_indep_sel_normal; eauto; discriminate.
Qed.

(* If a opened type is similar to a type, then the other is also opened *)
Lemma similar_open_case xs1 xs2 n x1 x2 (T1 T: typ):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 (open_rec n x1 T1) T ->
  exists T2,
    T = (open_rec n x2 T2) /\
    similar xs1 xs2 T1 T2.
Proof.
  introv Hfr1 Hfr2 Hsix HsiT.
  (* Reveal the substitution behind similarity *)
  pose proof HsiT as HsiT'.
  unfold similar, similar_typ, similar_by_subst in HsiT'.
  destruct_all; subst.
  eapply subst_vars_inverse in H3; eauto.
  subst.
  exists ((T1 /[ xs1 --> x0]) /[ x0 --> xs2]).
  rewrite subst_vars_open_commut in *.
  rewrite subst_vars_open_commut in *.
  splits.
  + (* variable *)
  fequal.

  eapply similar_var_simlist_relates in Hsix.
  unfold simlist_relates in Hsix.
  destruct_all.

  pose proof LibList.Nth_inbound H3.
  rewrite liblist_length, H, <- liblist_length in H5.
  eapply LibList.Nth_inbound_inv in H5.
  destruct_all.
  forwards Hx1xs1: var_subst_vars_Nth H3 H5; eauto.
  eapply fresh_swap; eauto. rewrite Hx1xs1.
  forwards Hx3x0: var_subst_vars_Nth H5 H4; eauto.
  + (* type *)

  unfold similar, similar_typ, similar_by_subst.
  exists (T1 /[ xs1 --> x0]) x0.

  pose proof (fv_open_cases x1 T1 n) as Hfv.
  assert (x1 \in from_list xs1). {
    eapply similar_out_fvl in Hsix.
    unfold fv, VarFreeVar in Hsix. apply -> subset_singleton. eauto.
  }
  assert (fv T1 \c from_list xs1) as HfvT1xs1. {
    eapply similar_out_fvl in HsiT.
    destruct Hfv.
    rewrite H4 in *; eauto.
    rewrite H4 in *; eauto.
    eapply subset_trans; eauto using subset_union_weak_r.
  }
  assert (fresh (fv T1) (length x0) x0). {
    destruct Hfv;
    rewrite H4 in *;
    fresh_solve.
  }
  assert (fv ((T1 /[ xs1 --> x0]) /[ x0 --> xs2]) \c from_list xs2) as HfvT1xs2. {
    eapply subst_vars_fv_subset.
    eapply subst_vars_fv_subset; eauto.
    eapply fresh_swap; eauto. eauto.
  }
  splits; eauto.
  fresh_solve. eapply fresh_subset; eauto.
  eapply subst_vars_fv_subset; eauto.
  eapply subst_vars_merge_same; eauto.
Qed.

Ltac similar_indep_find := (
  try match goal with
    H: Indep.typ_indep (avar_b 0) ?T, Hsi: similar ?xs1 ?xs2 ?T ?T2 |- _ => eapply (similar_typ_indep Hsi) in H
  end
).



Section SimilarBySubst.
Context {A: Set}{AO: Openable A}{AF: FreeVar A}{AV: SubstVar A}{AS: AbstractSyntax AO AF AV}
  {AI: SubstId AS}{SF : SubstFresh AS}{FS : FvSubstCases AS}{SC : SubstCommut AS}.
Let SA: Similar A := similar_by_subst.
Local Existing Instance SA.

Lemma similar_exists xs1 xs2 (T1: A) (L: fset var):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  (length xs1) = (length xs2) ->
  exists ys1 ys2 T2,
    similar (ys1++xs1) (ys2++xs2) T1 T2 /\
    length ys1 = length ys2 /\
    fresh \{} (length (ys1++xs1)) (ys1++xs1) /\
    fresh \{} (length (ys2++xs2)) (ys2++xs2) /\
    fresh L (length (ys2)) (ys2).
Proof.
  introv Hfr1 Hfr2 Hl.
  forwards L1 HL1: (fset_finite (fv T1 \- (from_list xs1))).
  set (ys1 := (LibList.remove_duplicates L1)).
  destruct (var_freshes (from_list xs2 \u fv T1 \u L) (length ys1)) as [ys2 Hys2].

  exists ys1 ys2 (T1 //[ys1 ++ xs1 --> ys2 ++ xs2]).

  forwards vs Hfrvs HT1vs: (subst_pvars_unfold T1 (ys1 ++ xs1) (ys2 ++ xs2)).
  rewrite HT1vs. clear HT1vs.

  pose proof Hys2 as Hlys. eapply fresh_length in Hlys.
  pose proof Hfrvs as Hlvs2. eapply fresh_length in Hlvs2.
  pose proof Hlvs2 as Hlvs1. rewrite List.app_length, Hlys, <-Hl, <-List.app_length in Hlvs1.

  eapply fresh_resize_length in Hys2.
  eapply fresh_resize_length in Hfrvs.

  assert (fv T1 \c from_list (ys1 ++ xs1)). {
    rewrite from_list_append.
    unfold ys1.
    rewrite from_list_remove_duplicates.
    rewrite <- HL1. rewrite remove_union_same. eapply subset_union_weak_l; eauto.
  }

  splits.
  - (* similar *)
    unfold similar, SA, similar_by_subst.
    exists (T1 /[ys1 ++ xs1 --> vs]) vs.
    splits.
    + eauto.
    + eauto.
    + fresh_solve.  eapply fresh_subset.
      eapply subst_vars_fv_subset; eauto.
      eapply subst_vars_fv_subset; eauto.
      eapply fresh_swap with (L:=\{}); eauto. eapply fresh_append. eauto. fresh_solve. eapply fresh_swap; eauto.
      eauto.
    + eapply subst_vars_fv_subset. eauto. eauto. eauto.
    + eapply subst_vars_merge_same; eauto.
    + eauto.
  - eauto.
  - eapply fresh_append. eapply fresh_remove_duplicates.
    fresh_solve. unfold ys1. rewrite from_list_remove_duplicates.
    rewrite <- HL1. eapply fresh_remove_from_list; eauto.
  -  eapply fresh_append. eauto.
    fresh_solve. eapply fresh_swap; eauto.
  - eauto.
Qed.

Lemma similar_exists_premises xs1 xs2 (T1: A) G G2 (e e2: renv):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  ctx_dom G \c from_list (xs1) ->
  dom e \c ctx_dom G ->
  dom e2 \c ctx_dom G2 ->
  similar xs1 xs2 e e2 ->
  similar xs1 xs2 G G2 ->
  exists (ys1 ys2: list var) (T: A),
  fresh \{} (length (ys1 ++ xs1)) (ys1++xs1) /\
  fresh \{} (length (ys2 ++ xs2)) (ys2++xs2) /\
  ctx_dom G \c from_list (ys1 ++ xs1) /\
  similar (ys1 ++ xs1) (ys2 ++ xs2) e e2 /\
  similar (ys1 ++ xs1) (ys2 ++ xs2) G G2 /\
  similar (ys1 ++ xs1) (ys2 ++ xs2) T1 T /\
  length ys1 = length ys2.
Proof.
  introv Hfrxs1 Hfrxs2 HdGxs1 HdeG HdeG2.
  introv Hsie HsiG.


  pose proof Hsie as Hl. eapply similar_length in Hl.

  forwards HT2: similar_exists T1 (ctx_dom G2) Hfrxs1 Hfrxs2. auto.
  destructc HT2 as ys1 ys2 Ty HsiTT Hly Hfryx1 Hfryx2 Hfry2.
  assert (fresh (ctx_dom G) (length ys1) ys1). {
    eapply fresh_append_fresh_l in Hfryx1.
    eapply fresh_subset. eauto. eauto.
  }
  assert (fresh (dom e) (length ys1) ys1). {
    eapply fresh_subset. eauto. eauto.
  }
  assert (fresh (dom e2) (length ys2) ys2). {
    eapply fresh_subset. eauto. eauto.
  }
  exists ys1 ys2 Ty.
  splits.
  eauto.
  eauto.

  rewrite from_list_append. eapply subset_union_rr; eauto.
  eapply similar_env_append_fresh_r; eauto.
  eapply similar_ctx_append_fresh; eauto.
  eauto. eauto.
Qed.

Lemma similar_fresh_cons_premises xs1 xs2 z G G2 (e e2: renv):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  ctx_dom G \c from_list (xs1) ->
  dom e \c ctx_dom G ->
  dom e2 \c ctx_dom G2 ->
  similar xs1 xs2 e e2 ->
  similar xs1 xs2 G G2 ->
  z \notin (ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u (from_list xs1) \u (from_list xs2)) ->
  fresh \{} (length (z :: xs1)) (z :: xs1) /\
  fresh \{} (length (z :: xs2)) (z :: xs2) /\
  ctx_dom G \c from_list (z :: xs1) /\
  similar (z :: xs1) (z :: xs2) e e2 /\
  similar (z :: xs1) (z :: xs2) G G2.
Proof.
  introv Hfrxs1 Hfrxs2 HdGxs1 HdeG HdeG2.
  introv Hsie HsiG Hfrz.
  splits; eauto.
  rewrite from_list_cons.
  eapply subset_union_rr; eauto.
  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_env_append_fresh_r; eauto.
  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_ctx_append_fresh; eauto.
Qed.

Lemma similar_push_cons_premises xs1 xs2 z vk T T2 G G2 (e e2: renv):
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  ctx_dom G \c from_list (xs1) ->
  dom e \c ctx_dom G ->
  dom e2 \c ctx_dom G2 ->
  ctx_ok G2 ->
  similar xs1 xs2 T T2 ->
  similar xs1 xs2 e e2 ->
  similar xs1 xs2 G G2 ->
  z \notin (ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u (from_list xs1) \u (from_list xs2)) ->
  fresh \{} (length (z :: xs1)) (z :: xs1) /\
  fresh \{} (length (z :: xs2)) (z :: xs2) /\
  ctx_dom (G § z / vk ~ T ) \c from_list (z :: xs1) /\
  dom e \c ctx_dom (G § z / vk ~ T) /\
  dom e2 \c ctx_dom (G2 § z / vk ~ T2) /\
  ctx_ok (G2 § z / vk ~ T2 )/\
  similar (z :: xs1) (z :: xs2) e e2 /\
  similar (z :: xs1) (z :: xs2) (G § z / vk ~ T ) (G2 § z / vk ~ T2 ).
Proof.
  introv Hfrxs1 Hfrxs2 HdGxs1 HdeG HdeG2 HokG2.
  introv HsiT Hsie HsiG Hfrz.
  splits; eauto.

  -
  ctx_simpl_dom. rewrite from_list_cons.
    eapply subset_union_2. eapply subset_refl. eauto.
  -
    ctx_simpl_dom. eapply subset_union_rr; eauto.
  -   ctx_simpl_dom. eapply subset_union_rr; eauto.
  -
  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_env_append_fresh_r; eauto.
  -
  eapply similar_ctx_push.
  eapply similar_extend; eauto.
  eapply simlist_relates_push_l; eauto.
  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_ctx_append_fresh; eauto.
Qed.

End SimilarBySubst.

Ltac similar_find_eq := (
  match goal with
    Hsi1: similar ?xs1 ?xs2 ?a ?b1,
    Hsi2: similar ?xs1 ?xs2 ?a ?b2 |- _ =>
    assert (b1=b2) by (eapply similar_by_subst_noduplicates_eq_l with (xs:=xs1) (ys:=xs2); eauto);
    clear Hsi2;
    subst; eauto
  end).

Lemma similar_ts_ind:
  TypingSimilarityTInd (ty_var typing_mode) similar_var similar_typ /\
  TypingSimilarityInd similar_typ similar_typ (subtyping) /\
  TypingSimilarityIncapInd (ty_incap _) similar_typ.
Proof.
  eapply ts_mutind; intros.
  (* Deal with objctx *)
  all: similar_objctx_invert; similar_objctx_simplify G; similar_objctx_simplify G2.
  (* Find structure of types *)
  all: similar_syntax_inverts.
  (* Solve cases depending on types being the same *)
  all: repeat similar_find_eq.
  all: similar_indep_find.

  - forwards: similar_ctx_binds_to_binds HsiG; eauto. eapply similar_var_simlist_relates; eauto.
  - eapply ty_rec_intro.
    eapply (H xs1 xs2); eauto.
    eapply similar_open_rec; eauto.
    eauto.
    intro Hm. eapply (H0 Hm xs1 xs2); eauto.
    eapply similar_open_rec; eauto.
    eapply similar_open_rec; eauto.
  - (* rec_elim *)
    eapply similar_open_case in HsiT; [>| eauto .. ].
    destruct_all; subst.
    eapply ty_rec_elim.
    eapply (H xs1 xs2); eauto.
    eapply similar_typ_bnd; eauto.
    similar_indep_find; eauto.
  - (* and *)
    eauto.
  - (* subtyp *)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.
    eapply ty_var_sub with (T:=Ty).
    eapply (H (ys1++xs1)%list (ys2++xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply (H0 (ys1++xs1)%list (ys2++xs2)%list); eauto.
    eapply similar_out_append; eauto.
  - (* mut-top*)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.
    eapply ty_var_mut_top. eauto.
    eapply (H (ys1++xs1)%list (ys2++xs2)%list); eauto.
    eapply similar_out_append; eauto.
  - eauto.
  - eauto.
  - (* trans *)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.
    eapply subtyp_trans with (T:=Ty).
    eapply (H (ys1++xs1)%list (ys2++xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply (H0 (ys1++xs1)%list (ys2++xs2)%list); eauto.
    eapply similar_out_append; eauto.
  - (* and *) eauto.
  - (* or *) eauto.
  - eauto.
  - (* dec_typ *)
    pick_fresh_with z (from_list xs1 \u from_list xs2).
    forwards HT: (similar_fresh_cons_premises (z:=z) Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG).
    eauto.
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2.

    eapply subtyp_typ; intros.
    eapply subtyp_reopen with (z0:= z). eauto. eauto.
    eapply (H z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.

    eapply subtyp_reopen with (z0:= z). eauto. eauto.
    eapply (H0 z (z::xs1)%list (z::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* sel *)
    eapply similar_open_case in Hsix; [>| eauto .. ].
    destruct_all; subst.

    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply subtyp_sel2 with (T:=Ty).
    eapply (H (ys1 ++ xs1)%list (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd.
    eapply similar_dec_typ; eauto.
    eapply similar_out_append; eauto.
  -
    eapply similar_open_case in HsiT; [>| eauto .. ].
    destruct_all; subst.

    pose proof (similar_exists_premises S Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply subtyp_sel1 with (S:=Ty).
    eapply (H (ys1 ++ xs1)%list (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd.
    eapply similar_dec_typ; eauto.
    eapply similar_out_append; eauto.
  - (* fun *)
    eapply subtyp_fun with (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2).
    eauto.
    eapply (H xs1 xs2); eauto.
    intros.

    forwards HT: (similar_push_cons_premises (z:=x2) v_par Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H2 Hsie HsiG).
    eauto.
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx Hsie2 HsiG2.
    assert (x2 \notin L) as HnL by eauto.
    eapply (H0 x2 HnL (x2::xs1)%list (x2::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* met *)
    eapply subtyp_met with (L:=L \u ctx_dom G\u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2).
    eapply (H xs1 xs2); eauto.

    intros.
    forwards HT: (similar_push_cons_premises (z:=x0) v_par Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H5 Hsie HsiG).
    eauto.
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx Hsie2 HsiG2.
    assert (x0 \notin L) as HnL by eauto.
    eapply (H0 x0 HnL (x0::xs1)%list (x0::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.

    (* Caused by modified method subtyping for sef *)
    {
      intros.
      forwards: t; eauto. eapply similar_typ_indep. eapply similar_symm; eauto. eauto.
      eapply similar_typ_indep; eauto.
    }

    intros.

    lets HT: (similar_push_cons_premises (z:=x0) v_par Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H5 Hsie HsiG).
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx Hsie2 HsiG2.
    eauto.
    assert (similar (x0::xs1) (x0::xs2) (open x0 TC2) (open x0 x1)) as H3x.
    eapply similar_open_rec_push; eauto.
    lets HT1: (similar_push_cons_premises (z:=y) v_recv Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx H3x Hsie2 HsiG2).
    destructc HT1 as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    rewrite ! from_list_cons.
    eauto.

    assert (x0 \notin L) as HnL by eauto.
    assert (y <> x0) as Hnx by eauto.
    assert (y \notin L) as HnL1 by eauto.
    eapply (H1 x0 HnL y Hnx HnL1 (y::x0::xs1)%list (y::x0::xs2)%list); eauto.

    rewrite ! (open_commut x0) by eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
    rewrite ! (open_commut x0) by eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* tyq *)
    eapply subtyp_tyq.
    eapply similar_tyq_typ with (xs1:=xs1); eauto.
  - (* nocap-bnd *)
    eauto.
  - (* nocap-typ *)
    eauto.
  - (* nocap-fld *)
    eauto.
  - (* nocap-met *)
    eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - (* icp-sel*)

    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    pose proof (similar_exists_premises S Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 Hsie2 HsiG2) as HS.
    destructc HS as zs1 zs2 Sy Hfrzs1 Hfrzs2 HsG1zs1 Hsie3 HsiG3 HsiS3 Hlzs.

    eapply ty_incap_sel with (S:=Sy)(T:=Ty).
    eapply (H (zs1 ++ ys1 ++ xs1)%list (zs2 ++ ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd. eapply similar_dec_typ; eauto.
    eapply similar_out_append; eauto.

    eapply (H0 (zs1 ++ ys1 ++ xs1)%list (zs2 ++ ys2 ++ xs2)%list); eauto.
    eapply similar_open_rec; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
  - eauto.
  - eauto.
  - eapply ty_incap_nocap. auto.
    eapply (H xs1 xs2); eauto.
    eapply similar_typ_nulary; eauto. eapply (similar_length _ _ T). eauto.
Qed.

Global Instance typing_similarity_ty_var: TypingSimilarity similar_var similar_typ (@ty_var typing_mode).
Proof.
  hnf. intros.
  eapply similar_ts_ind with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.
Global Instance typing_similarity_ty_incap: TypingSimilarity similar_typ similar_typ (@ty_incap typing_mode).
Proof.
  hnf. intros.
  eapply similar_ts_ind with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.
Global Instance typing_similarity_subtyp: TypingSimilarity similar_typ similar_typ (@subtyp typing_mode).
Proof.
  hnf. intros.
  eapply similar_ts_ind with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.

Global Instance typing_similarity_ty_capbnd C: TypingSimilarity similar_typ similar_typ (@ty_capbnd typing_mode C).
Proof.
  hnf. introv Hoke HokG2 Hne1 Hse1G1 Hse2G2 HsG1xs1. introv Hfrxs1 Hfrxs2 Hsie Hsix HsiT HsiG HsiGo. intros. reverts xs1 xs2 x2 T2.
  induction H; intros.
  (* Deal with objctx *)
  all: similar_objctx_invert.
  (* Find structure of types *)
  all: similar_syntax_inverts.
  (* Solve cases depending on types being the same *)
  all: repeat similar_find_eq.
  all: try solve [constructor; eauto].
  - pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    pose proof (similar_exists_premises S Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 Hsie2 HsiG2) as HS.
    destructc HS as zs1 zs2 Sy Hfrzs1 Hfrzs2 HsG1zs1 Hsie3 HsiG3 HsiS3 Hlzs.

    eapply ty_capbnd_sel with (S:=Sy)(T:=Ty).
    eapply typing_similarity with (xs3:= (zs1 ++ ys1 ++ xs1)%list) (xs4:=(zs2 ++ ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd. eapply similar_dec_typ; eauto.
    eapply similar_out_append; eauto.

    eapply (IHty_capbnd Hoke Hne1 Hse1G1 (zs1 ++ ys1 ++ xs1)%list HsG1zs1 Hfrzs1 (zs2 ++ ys2 ++ xs2)%list); eauto.
    eapply similar_open_rec; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
  -
    econstructor. eapply IHty_capbnd; eauto.  eapply IHty_capbnd0; eauto.
  -
  econstructor. eapply IHty_capbnd; eauto.  eapply IHty_capbnd0; eauto.
Qed.


Lemma typing_similarity_viewpoint_adapt xs1 xs2 G1 G2 e1 e2 (x1 x2: var) (T1 T2 S1 S2: typ):
  ok e1 ->
  ctx_ok G2 ->
  disjoint (dom e1) (fv_env e1) ->
  dom e1 \c ctx_dom G1 ->
  dom e2 \c ctx_dom G2 ->
  ctx_dom G1 \c (from_list xs1) ->
  fresh \{} (length xs1) xs1 -> (* Needed for var equivalence in renv *)
  fresh \{} (length xs2) xs2 -> (* Needed for subtyping (and, or) *)
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 G1 G2->
  viewpoint_adapt typing_mode x1 G1 objctx_empty e1 T1 S1 ->
  viewpoint_adapt typing_mode x2 G2 objctx_empty e2 T2 S2.
Proof.
  intros.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' typing_mode).
  - subst. similar_find_eq.
  - inversions H12.
    similar_syntax_inverts.
    similar_find_eq.
    econstructor.
    eapply typing_similarity; eauto.
    eapply typing_similarity; eauto.
Qed.



Lemma similar_vis xs1 xs2 x1 x2 G1 G2:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 G1 G2 ->
  ctx_vis G1 x1 ->
  ctx_vis G2 x2.
Proof.
  introv Hfr1 Hfr2 Hsix HsiG HvisG.
  eapply similar_var_simlist_relates in Hsix.
  induction HsiG.
  -
    forwards: (similar_ctx_cases Hsix H1); eauto.
    destruct_all.
    + eapply ctx_vis_in_dom in HvisG; contradiction.
    + eapply ctx_binds_weakens_vis; eauto.
  - inversions HvisG.
    forwards: simlist_relates_noduplicates_var_eq Hsix H0; eauto. subst. eauto.
    forwards Hcases: simlist_relates_noduplicates_var_cases Hsix H0; eauto.
    forwards Hvisx2: IHHsiG; eauto. constructor. intro Hxv2. apply <- Hcases in Hxv2. contradiction. eauto.
  - inversions HvisG.
Qed.

Lemma similar_ctx_binds xs1 xs2 (S1 S2: ctx) x1 x2 it1 it2 ys1 ys2:
  similar_with ys1 ys2 xs1 xs2 S1 S2 ->
  simlist_relates xs1 xs2 x1 x2 ->
  ctx_binds x1 it1 S1 ->
  ctx_binds x2 it2 S2 ->
  similar ys1 ys2 it1 it2.
Proof.
  introv HsiG Hsix Hbix1 Hbix2.
  induction HsiG.
  - false. eapply simlist_relates_nil_inv; eauto.
  - unfold simlist_relates in Hsix. destruct_all.
    inversions H5.
    + inversions H6. ctx_binds_eq. ctx_binds_eq. auto.
    + inversions H6. eapply IHHsiG; eauto. exists n; eauto.
  -
    unfold simlist_relates in Hsix. destruct_all.
    inversions H2.
    + inversions H3. eapply ctx_binds_dom in Hbix1. contradiction.
    + inversions H3. eapply IHHsiG; eauto. exists n; eauto.
Qed.

Lemma similar_ctx_obj G1 G2 Go1 Go2 xs1 xs2:
  similar xs1 xs2 G1 G2 ->
  similar xs1 xs2 Go1 Go2 ->
  similar xs1 xs2 (G1 §§& Go1) (G2 §§& Go2).
Proof.
  introv HsiG HsiGo.
  unfold ctx_obj.
  destruct Go1; try similar_objctx_invert.
  - eauto.
  - eapply similar_ctx_push.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var. eapply similar_length in H5. eauto.
    all: eauto.
  - eauto.
Qed.
Lemma similar_ctx_obj_met xs1 xs2 G1 G2 Go1 Go2 (xz1 xz2 xc1 xc2: var) (TZ1 TC1 TZ2 TC2: typ):
  similar xs1 xs2 xc1 xc2 ->
  similar xs1 xs2 xz1 xz2 ->
  similar xs1 xs2 TC1 TC2 ->
  similar xs1 xs2 TZ1 TZ2 ->
  similar xs1 xs2 G1 G2 ->
  similar xs1 xs2 Go1 Go2 ->
  similar (xs1) (xs2) (ctx_obj_met typing_mode G1 Go1 xz1 TZ1 xc1 TC1) (ctx_obj_met typing_mode G2 Go2 xz2 TZ2 xc2 TC2).
Proof.
  introv Hsic Hsiz HsiTC HsiTZ HsiG HsiGo.

  assert (length xs1 = length xs2) as Hl.
  eapply similar_length in Hsic; eauto.

  destruct Go1; similar_objctx_invert;
  unfold ctx_obj_met, ctx_obj in *; destruct (mode_visibility' typing_mode).
  - eapply similar_ctx_push; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.
  - eapply similar_ctx_push; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_hide; eauto.
  - eapply similar_ctx_push; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
  - eapply similar_ctx_push; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_hide; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
  - eapply similar_ctx_push; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.
  - eapply similar_ctx_push; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply similar_typ_andor. eauto.
    eapply similar_open_rec; eauto.
    eapply simlist_relates_similar_var; eauto.
    eapply similar_open_rec; eauto.
    eapply similar_var_simlist_relates; eauto.
    eapply similar_ctx_push; eauto.
    eapply similar_var_simlist_relates; eauto.

    eapply similar_ctx_hide; eauto.
Qed.


Lemma similar_rules_ind:
  TypingSimilarityTInd (ty_trm typing_mode) similar_trm similar_typ /\
  TypingSimilarityTInd (ty_lit typing_mode) similar_lit similar_typ /\
  TypingSimilarityDInd (ty_def typing_mode) similar_def similar_dec /\
  TypingSimilarityTInd (ty_defs typing_mode) similar_defs similar_typ.
Proof.
  eapply rules_mutind; intros.
  all: try similar_objctx_invert.
  all: similar_syntax_inverts.
  all: repeat similar_find_eq.

  - eapply ty_trm_var; eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    eapply typing_similarity; eauto.
  - (* apply *)
    eapply similar_open_case in HsiT; [>| eauto .. ].
    destruct_all; subst.

    pose proof (similar_exists_premises S Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HS.
    destructc HS as ys1 ys2 Sy Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply ty_trm_apply with (S:=Sy); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    eapply typing_similarity with (x3:=x) (T1:=typ_fun S T); eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_fun; eauto.
    eapply similar_out_append; eauto.
    eapply typing_similarity with (x3:=z) (T1:=S); eauto.
    eapply similar_out_append; eauto.
  - (* read *)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    pose proof (similar_exists_premises S Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 Hsie2 HsiG2) as HS.
    destructc HS as zs1 zs2 Sy Hfrzs1 Hfrzs2 HsG1zs1 Hsie3 HsiG3 HsiS3 Hlzs.

    eapply ty_trm_read with (S:=Sy)(T:=Ty); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.

    eapply typing_similarity with (x3:=x1) (T1:=typ_rcd (dec_fld a S T)); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd; eauto.
    eapply similar_dec_fld; eauto.
    eapply similar_out_append; eauto.

    eapply typing_similarity_viewpoint_adapt with (xs1:=(zs1 ++ ys1 ++ xs1)%list) (xs2:=(zs2 ++ ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
  - (* write *)

    pose proof (similar_exists_premises S Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HS.
    destructc HS as ys1 ys2 Sy Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    assert (length xs1 = length xs2) as Hlx. eapply similar_length with (a0:=x); eauto.

    eapply ty_trm_write with (S:=Sy); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    intro.
    eapply typing_similarity with (x3:=x) (T1:=(typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))); eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd; eauto.
    eapply similar_dec_typ; eauto.
    eapply similar_typ_nulary. eauto. rewrite ! List.app_length. rewrite Hlys. rewrite Hlx. eauto.
    eapply similar_typ_nulary. eauto. rewrite ! List.app_length. rewrite Hlys. rewrite Hlx. eauto.
    eapply typing_similarity with (x3:=x) (T1:=typ_rcd (dec_fld a S T)); eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd; eauto.
    eapply similar_dec_fld; eauto.
    eapply similar_out_append; eauto.
    eapply typing_similarity with (x3:=y) (T1:=S); eauto.
    eapply similar_out_append; eauto.
  - (* call *)
    eapply similar_open_case in HsiT; [>| eauto .. ].
    destruct_all; subst.
    eapply similar_open_case in H2; [>| eauto .. ].
    destruct_all; subst.

    pose proof (similar_exists_premises TC Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    pose proof (similar_exists_premises TZ Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 Hsie2 HsiG2) as HS.
    destructc HS as zs1 zs2 Sy Hfrzs1 Hfrzs2 HsG1zs1 Hsie3 HsiG3 HsiS3 Hlzs.

    eapply ty_trm_call with (TC:=Ty) (TZ:=Sy); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2); eauto.

    eapply typing_similarity with (x4:=x) (T1:=typ_rcd (dec_met m TC TZ TR)); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_typ_rcd; eauto.
    eapply similar_dec_met; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.

    eapply typing_similarity with (x4:=y) (T1:=TZ); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.

    eauto using similar_typ_indep.

    eapply typing_similarity with (x4:=x) (T1:=(open y TC)); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.

    eapply similar_open_rec; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
  - (* let *)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply ty_trm_let with (T:=Ty) (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list (ys1++xs1) \u from_list (ys2++xs2)).
    eapply (H (ys1 ++ xs1)%list (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    intros.
    forwards HT: (similar_push_cons_premises (z:=x1) v_let Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 HokG2 HsiT2 Hsie2 HsiG2).
    eauto.
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    assert (x1 \notin L) as HnL by eauto.
    eapply (H0 x1 HnL (x1::ys1++xs1)%list (x1::ys2++xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_out_append; eauto.
    eapply similar_extend; eauto.
    eapply similar_out_append; eauto.

  - pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply ty_trm_lit with (T:=Ty) (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list (ys1++xs1) \u from_list (ys2++xs2)).
    eapply (H (ys1 ++ xs1)%list (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    intros.
    forwards HT: (similar_push_cons_premises (z:=x1) v_let Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 HokG2 HsiT2 Hsie2 HsiG2).
    eauto.
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    assert (x1 \notin L) as HnL by eauto.
    eapply (H0 x1 HnL (x1::ys1++xs1)%list (x1::ys2++xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_out_append; eauto.
    eapply similar_extend; eauto.
    eapply similar_out_append; eauto.
  - (* sub *)
    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.
    eapply ty_trm_sub with (T:=Ty).

    eapply (H (ys1 ++ xs1)%list (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.

    eapply typing_similarity with  (xs3:=(ys1 ++ xs1)%list) (xs4:=(ys2 ++ xs2)%list) (x1:=T) (T1:=U); eauto.
    eapply similar_out_append; eauto.
  - (* lit fun *)
    eapply ty_lit_fun with (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2).
    eauto.
    intros.

    forwards HT: (similar_push_cons_premises (z:=x1) v_par Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H3 Hsie HsiG).
    eauto.
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx Hsie2 HsiG2.
    assert (x1 \notin L) as HnL by eauto.
    eapply (H x1 HnL (x1::xs1)%list (x1::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - (* lit obj *)
    pose proof (@ty_lit_obj typing_mode) as Hty_lit_obj. (*Adding to hypotheses to apply case analysis of typing_mode *)
    destruct (mode_objlit_refcaps' typing_mode).
    all: similar_syntax_inverts.
    all: repeat similar_find_eq.
    all: eapply Hty_lit_obj with (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2).
    2, 4: similar_indep_find; eauto.
    all: intros.

    forwards HT: (similar_push_cons_premises (z:=x) v_let Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H1 Hsie HsiG).
    eauto.
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    assert (x \notin L) as HnL by eauto.

    eapply (H x HnL) with (xs1:=(x::xs1)%list) (xs2:=(x::xs2)%list); eauto.

    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.

  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_ctx_append_fresh; eauto.

  constructor. eapply simlist_relates_push_l.
    eapply similar_extend; eauto.

    forwards HT: (similar_push_cons_premises (z:=x) v_let Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H1 Hsie HsiG).
    eauto.
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    assert (x \notin L) as HnL by eauto.

    eapply (H x HnL) with (xs1:=(x::xs1)%list) (xs2:=(x::xs2)%list); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.

  rewrite <- (List.app_nil_l xs1).
  rewrite <- (List.app_nil_l xs2).
  rewrite ! (List.app_comm_cons).
  eapply similar_ctx_append_fresh; eauto.

  constructor. eapply simlist_relates_push_l.
    eapply similar_extend; eauto.

  - eapply ty_def_typ with (L:=L \u ctx_dom (G §§& Go) \u ctx_dom (G2 §§& Go2) \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2); eauto.
    intros.
    lets HT: (similar_push_cons_premises (z:=x) v_let Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H2 Hsie).
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    eapply similar_ctx_obj; eauto.
    eauto.
    assert (x \notin L) as HnL by eauto.

    eapply @typing_similarity with (property:=(fun G Go e (a b: typ) => typing_equiv e a b)) (xs1:=(x::xs1)%list) (xs2:=(x::xs2)%list)  (T1:=(open x U)) (x1:=(open x T)); eauto.
    eapply typing_similarity_tyq_typ.
    eapply similar_open_rec_push; eauto.
    eapply similar_open_rec_push; eauto.
  - eapply ty_def_fld; eauto.
    intro. eapply similar_vis with (xs1:=xs1) (xs2:=xs2) (G1:= (G §§& Go)); eauto.
    eapply similar_ctx_obj; eauto.
    eapply typing_similarity with (x3:=x2) (T1:=T) (G1:= G§§&Go); eauto.
    eapply similar_ctx_obj; eauto.
  - eapply ty_def_met with (L:=L \u ctx_dom (G §§& Go) \u ctx_dom (G2 §§& Go2) \u fv_renv e \u fv_renv e2 \u from_list xs1 \u from_list xs2).
    intros.

    assert ((length xs1) = (length xs2)) as Hl. eapply similar_length in H1; eauto.

    forwards HsiGGo: similar_ctx_obj; eauto.

    pose proof (similar_push_cons_premises (z:=xz) v_par Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 HokG2 H2 Hsie HsiGGo) as HT.
    destructc HT as Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx Hsie2 HsiG2.
    eauto.
    assert (xz \notin L) as HnL by eauto.
    assert (similar (xz::xs1) (xz::xs2) (open xz TC) (open xz x)) as HsiTC.
    eapply similar_open_rec_push; eauto.

    lets HT1: (similar_push_cons_premises (z:=xc) v_recv Hfrys1 Hfrys2 HsG1ys1 Hse1Gx Hse2Gx HokGx HsiTC Hsie2 HsiG2).
    destructc HT1 as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    rewrite ! from_list_cons.
    eauto.

    assert (xc <> xz) as Hnx by eauto.
    assert (xc \notin L) as HnL1 by eauto.
    eapply (H xz HnL xc Hnx HnL1 (xc::xz::xs1)%list (xc::xz::xs2)%list); eauto.

    + unfold ctx_obj.
      rewrite ! ctx_obj_dom in *.
      rewrite <- ctx_concat_empty_l. eapply ctx_obj_met_ok; rewrite ? ctx_concat_empty_l; eauto.
    + unfold ctx_obj.
      rewrite ctx_obj_met_dom. eapply subset_trans. eapply Hse1G1. eapply subset_trans. eapply subset_union_weak_r. eapply subset_union_weak_r.
    + unfold ctx_obj.
      rewrite ctx_obj_met_dom. eapply subset_trans. eapply Hse2G2. eapply subset_trans. eapply subset_union_weak_r. eapply subset_union_weak_r.
    + unfold ctx_obj.
      rewrite ctx_obj_met_dom. rewrite ! from_list_cons.
      eapply subset_union_l. eauto using subset_union_weak_l, subset_union_weak_r, subset_trans.
      eapply subset_union_l. eauto using subset_union_weak_l, subset_union_weak_r, subset_trans.
      eauto using subset_union_weak_l, subset_union_weak_r, subset_trans.
    +
      rewrite ! (open_commut xz) by eauto.
      eapply similar_open_rec_push; eauto.
      eapply similar_open_rec_push; eauto.
    +
      rewrite ! (open_commut xz) by eauto.
      eapply similar_open_rec_push; eauto.
      eapply similar_open_rec_push; eauto.
    + eapply similar_ctx_obj_met; eauto.
      eapply similar_var_push. simpl. eauto.
      eapply similar_extend.
      eapply similar_var_push. simpl. eauto.

      eapply similar_extend; eauto.
      eapply similar_extend; eauto.

      eapply similar_extend; eauto.
      eapply similar_extend; eauto.

      rewrite <- (List.app_nil_l xs1).
      rewrite <- (List.app_nil_l xs2).
      rewrite ! (List.app_comm_cons).
      rewrite ! ctx_obj_dom in *.
      eapply similar_ctx_append_fresh; eauto.
      eapply similar_objctx_extend.
      eapply similar_objctx_extend; eauto.

  - eapply ty_defs_one; eauto.
  - eapply ty_defs_cons; eauto.
    eapply similar_defs_hasnt; eauto.
Qed.

Global Instance typing_similarity_ty_trm: TypingSimilarity similar_trm similar_typ (@ty_trm typing_mode).
Proof.
  hnf. intros.
  eapply similar_rules_ind with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.

Global Instance typing_similarity_ty_defs: TypingSimilarity similar_defs similar_typ (@ty_defs typing_mode).
Proof.
  hnf. intros.
  eapply similar_rules_ind with (xs1:=xs1)(xs2:=xs2); eauto.
Qed.

End WithTypingMode.


Lemma similarity_ctx_with_k xs1 xs2 ys1 ys2 G1 G2 (x1 x2: var) vk:
  similar_with xs1 xs2 ys1 ys2 G1 G2->
  simlist_relates ys1 ys2 x1 x2->
  ctx_k G1 x1 vk ->
  ctx_k G2 x2 vk.
Proof.
  introv HsiG Hsix Hk.
  induction HsiG.
  - inversions Hsix. destruct_all. inversions H0.
  - inversions Hsix. destruct x; destruct_all.
    + inversions H5. inversions H6.
      forwards: ctx_k_eq  H2 Hk; eauto.
      subst; eauto.
    + inversions H5. inversions H6.
      eapply IHHsiG; eauto. econstructor; eauto.
  - inversions Hsix. destruct x; destruct_all.
    + inversions H2. inversions H3.
      eapply ctx_k_in_dom in Hk; contradiction.
    + inversions H2. inversions H3.
      eapply IHHsiG; eauto. econstructor; eauto.
Qed.


Lemma similarity_ctx_k xs1 xs2 G1 G2 (x1 x2: var) vk:
  fresh \{} (length xs1) xs1 ->
  ctx_ok G2 ->
  similar xs1 xs2 G1 G2->
  similar xs1 xs2 x1 x2->
  ctx_k G1 x1 vk ->
  ctx_k G2 x2 vk.
Proof.
  introv Hfr HokG2 HsiG Hsix Hk.
  induction HsiG.
  - eapply similarity_ctx_with_k; eauto. eapply similar_var_simlist_relates; eauto.
  - eapply ctx_k_push_inv in Hk. destruct_all; subst.
    + eapply similar_var_simlist_relates in Hsix. forwards: simlist_relates_noduplicates_var_eq H0 Hsix; eauto. subst. eapply ctx_k_push_e.
    + forwards: IHHsiG; eauto. eapply ctx_k_push_n; eauto.  inversions HokG2. eauto.
  - false. eapply ctx_k_hide_inv; eauto.
Qed.



Lemma similar_ctx_with_refl_by_fv xs ys G:
  ctx_weakens G ->
  ctx_ok G ->
  ctx_fv G \c from_list xs ->
  similar_ctx_with xs xs ys ys G G.
Proof.
  introv Hwk Hok Hfv.
  induction ys.
  - eapply similar_ctx_empty. auto.
  - destruct (fv_decide a (ctx_dom G)).
    + forwards Hbi: ctx_dom_to_binds m; fweauto.
      destructecs Hbi.
      forwards Hk: ctx_dom_k; fweauto.
      destructecs Hk.
      eapply similar_ctx_dom; eauto.
      eapply similar_refl_by_fv.
      eapply subset_trans. 2: eauto.
      eapply ctx_binds_fv_subset; eauto.
    +
      eapply similar_ctx_notin; eauto.
Qed.


Lemma similar_ctx_refl_by_fv xs G:
  ctx_weakens G ->
  ctx_ok G ->
  ctx_fv G \c from_list xs ->
  similar_ctx xs xs G G.
Proof.
  introv Hwk Hok Hfv.
  eapply similar_ctx_global; eauto.
  eapply similar_ctx_with_refl_by_fv; eauto.
Qed.
