Set Implicit Arguments.

Require Import LibExtra Vars Terms Stack Heap Env SyntaxClasses.

(* * Machine Configurations. *)

(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** A machine configuration consists of:
- Term [t], the focus of execution;
- Stack [s];
- Heap [S];
- Environment [e];
- a set of variable names [Lv] that appeared at any step of the execution so far.
*)
Inductive config: Type := config_mk (t:trm) (s:stack) (S:heap) (e:renv) (Lv: fset var): config.
(* WARNING: the separator here ';' is a different character than normal ';' *)
Notation "t ; s ; Sigma ; e ; Lv" :=
  (config_mk t s Sigma e Lv) (at level 39, s, Sigma, e at level 38, Lv at level 38) : syntax_scope.

Definition config_heap C := (
  match C with
  _ ; _ ; Sc ; _ ; _ => Sc
  end ).

Definition config_env C := (
  match C with
  _ ; _ ; _ ; e ; _ => e
  end ).

Definition config_fv C := (
  match C with
  _ ; _ ; _ ; _ ; Lv => Lv
  end ).

(* heap and env live on the type level, not the set level, so they can't be
   used for FreeVar *)
(* Note: explicit application of dom, to infer the correct type for Sigma *)
Notation fv_renv e := ((@dom var e) \u (fv_env e)).

(* All free variables in a config (t, s, Sigma, e) is the union of the free
   variables in each component *)
Notation fv_config c :=
  match c with
  | (t ; s ; Sigma ; e ; Lv) =>
    (@fv trm _ t) \u (@fv stack _ s) \u (fv_heap Sigma)
      \u (fv_renv e) \u Lv
  end.

(* A list of variables l is fresh in the config c if each variable in l is
   fresh in c *)
Notation fresh_in_config l c := (fresh (fv_config c) (length l) l).

Lemma fresh_in_config_notin : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x \notin fv_config (t ; s ; Sigma ; e ; Lv).
Proof.
  induction l1; eauto using fresh_cons.
  introv Hfr. simpl in Hfr. destruct Hfr as [_ Hfr].
  apply fresh_union_r1 in Hfr. eauto.
Qed.

Lemma fresh_in_config_notin_e_dom : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x # e.
Proof. introv Hfr; eapply fresh_in_config_notin in Hfr. eauto. Qed.

Lemma fresh_in_config_notin_e_im : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x \notin fv_env e.
Proof. introv Hfr; eapply fresh_in_config_notin in Hfr. eauto. Qed.

Lemma fresh_in_config_notin_Sigma_dom : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x # Sigma.
Proof. introv Hfr; eapply fresh_in_config_notin in Hfr. eauto. Qed.

Lemma fresh_in_config_notin_Sigma_im : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x \notin fv_env Sigma.
Proof. introv Hfr; eapply fresh_in_config_notin in Hfr. eauto. Qed.

Lemma fresh_in_config_notin_t : forall l1 (x : var) l2 t s Sigma e Lv,
    fresh_in_config (l1 ++ x :: l2) (t ; s ; Sigma ; e ; Lv) ->
    x \notin fv t.
Proof. introv Hfr; eapply fresh_in_config_notin in Hfr. eauto. Qed.

Definition nothing_fresh_in_config A {AF: FreeVar A} (a:A) c :=
  (fv a) \c fv_config c.

Lemma fresh_in_config_notin_nothing_fresh A {AF: FreeVar A}  (a:A) l1 l2 c x:
  nothing_fresh_in_config a c ->
  fresh_in_config (l1 ++ x :: l2) c ->
  x \notin (fv a).
Proof.
  intros.  destruct c.
  unfold nothing_fresh_in_config in *.
  intro. apply H in H1. eapply fresh_in_config_notin; eauto.
Qed.

Tactic Notation "remember_dependent_config" hyp(t) hyp(s) hyp(S) hyp(e) hyp(Lv) "as" simple_intropattern(C) :=
  remember (t ; s ; S ; e; Lv) as C;
  revert dependent Lv;
  reverts t s S e.

(* All variables in xs are free in C *)
Notation vars_conf xs C := (
    from_list xs \c fv_config C
  ) (only parsing).

(* All variables in xs are free in C and no duplicates *)
Definition vars_conf_unique xs C := (
    fresh \{} (length xs) xs /\
    vars_conf xs C
  ).

Lemma vars_conf_unique_prepend t1 fs1 S1 e1 Lv1 t2 fs2 S2 e2 ys xs:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  fresh (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) (length ys) ys ->
  (from_list ys) \c (dom S2 \u dom e2) ->
  vars_conf_unique (ys ++ xs) (t2 ; fs2 ; S2 ; e2 ; fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)).
Proof.
  unfold vars_conf_unique.
  introv Hsf Hfr Hdom.
  destructecs Hsf.
  splits.
  - eapply fresh_append; eauto. rewrite union_empty_l. eapply fresh_swap; eauto. eapply fresh_subset; eauto.
  - rewrite from_list_append.
    eapply subset_union_l.
    + intro. intro Hx.
      repeat rewrite in_union.
      eapply Hdom in Hx. rewrite in_union in Hx. destruct Hx.
      eauto.
      right; eauto.
    + intro. intro Hx.
      repeat rewrite in_union.
      eapply Hsf in Hx. rewrite ! in_union in Hx.
      destructecs Hx; eauto 10.
Qed.


Lemma vars_conf_unique_prepend0 t1 fs1 S1 e1 Lv1 t2 fs2 S2 e2 xs:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  vars_conf_unique (xs) (t2 ; fs2 ; S2 ; e2 ; fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)).
Proof.
  introv Hsf.
  forwards: vars_conf_unique_prepend (@nil var); eauto.
  rewrite from_list_nil. intro. intro. rewrite in_empty in H. contradiction.
Qed.

Lemma vars_conf_unique_prepend1 t1 fs1 S1 e1 Lv1 t2 fs2 S2 e2 xs y:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  y \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y \in (dom S2 \u dom e2) ->
  vars_conf_unique (y::xs) (t2 ; fs2 ; S2 ; e2 ; fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)).
Proof.
  introv Hsf Hn Hdom.
  forwards: vars_conf_unique_prepend (y::(@nil var))%list; eauto.
  rewrite from_list_cons.
  rewrite from_list_nil. rewrite union_empty_r.
  intro. intro. rewrite in_singleton in H; subst. eapply Hdom.
Qed.
Lemma vars_conf_unique_prepend2 t1 fs1 S1 e1 Lv1 t2 fs2 S2 e2 xs y1 y2:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  y1 \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y1 \in (dom S2 \u dom e2) ->
  y2 \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y2 \in (dom S2 \u dom e2) ->
  y1 <> y2 ->
  vars_conf_unique (y1::y2::xs) (t2 ; fs2 ; S2 ; e2 ; fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)).
Proof.
  introv Hsf Hn1 Hdom1 Hn2 Hdom2 Hn.
  forwards: vars_conf_unique_prepend (y1::y2::(@nil var))%list; eauto.
  rewrite from_list_cons.
  rewrite from_list_cons.
  rewrite from_list_nil. rewrite union_empty_r.
  intro. intro Hx.
  rewrite ! in_union, ! in_singleton in Hx.
  destructecs Hx; subst; eauto.
Qed.
Lemma vars_conf_unique_prepend_concat t1 fs1 S1 e1 Lv1 S2 e2 ys xs:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  fresh (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) (length ys) ys ->
  (from_list ys) \c (dom S2 \u dom e2) ->
  vars_conf_unique (ys ++ xs) (t1 ; fs1 ; S1 & S2 ; e1 & e2 ; Lv1).
Proof.
  unfold vars_conf_unique.
  introv Hsf Hfr Hdom. 
  destructecs Hsf.
  splits.
  - eapply fresh_append; eauto. rewrite union_empty_l. eapply fresh_swap; eauto. eapply fresh_subset; eauto.
  - rewrite from_list_append.
    eapply subset_union_l.
    + rewrite ! dom_concat.
      intro. intro Hx.
      repeat rewrite in_union.
      eapply Hdom in Hx. rewrite in_union in Hx. destruct Hx.
      branch 3. branch 1. auto.
      branch 4. branch 1. auto.
    + rewrite ! fv_env_concat, ! dom_concat.
      intro. intro Hx.
      repeat rewrite in_union.
      eapply Hsf in Hx. rewrite ! in_union in Hx.
      destructecs Hx; eauto 10.
Qed.
Lemma vars_conf_unique_prepend2_concat t1 fs1 S1 e1 S2 e2 Lv1 xs y1 y2:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  y1 \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y1 \in (dom S2 \u dom e2) ->
  y2 \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y2 \in (dom S2 \u dom e2) ->
  y1 <> y2 ->
  vars_conf_unique (y1::y2::xs) (t1 ; fs1 ; S1 & S2 ; e1 & e2 ; Lv1).
Proof.
  introv Hsf Hn1 Hdom1 Hn2 Hdom2 Hn.
  forwards: vars_conf_unique_prepend_concat (y1::y2::(@nil var))%list; eauto.
  rewrite from_list_cons.
  rewrite from_list_cons.
  rewrite from_list_nil. rewrite union_empty_r.
  intro. intro Hx.
  rewrite ! in_union, ! in_singleton in Hx.
  destructecs Hx; subst; eauto.
Qed.
Lemma vars_conf_unique_prepend1_concat_e t1 fs1 S1 e1 e2 Lv1 xs y1:
  vars_conf_unique xs (t1 ; fs1 ; S1 ; e1 ; Lv1) ->
  y1 \notin (fv_config (t1 ; fs1 ; S1 ; e1 ; Lv1)) ->
  y1 \in (dom e2) ->
  vars_conf_unique (y1::xs) (t1 ; fs1 ; S1; e1 & e2 ; Lv1).
Proof.
  introv Hsf Hn1 Hdom1.
  forwards: vars_conf_unique_prepend_concat S1 (@empty item) (y1::(@nil var))%list ; eauto.
  rewrite from_list_cons.
  rewrite from_list_nil.
  rewrite dom_empty.
  rewrite union_empty_l.
  rewrite union_empty_r.
  eapply subset_singleton. eauto.
  rewrite concat_empty_r in H.
  rewrite <- List.app_comm_cons in H.
  rewrite List.app_nil_l in H. auto.
Qed.



Lemma vars_conf_unique_Lv_subset t fs S e Lv1 Lv2 xs:
  vars_conf_unique xs (t ; fs ; S ; e ; Lv1) ->
  Lv1 \c Lv2 ->
  vars_conf_unique xs (t ; fs ; S ; e ; Lv2).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  union_sub_solver.
Qed.

Lemma vars_conf_unique_t_subset t1 t2 fs S e Lv xs:
  vars_conf_unique xs (t1 ; fs ; S ; e ; Lv) ->
  fv t1 \c fv t2 ->
  vars_conf_unique xs (t2 ; fs ; S ; e ; Lv).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  union_sub_solver.
Qed.
Lemma vars_conf_unique_fss_subset t fs1 fs2 S e Lv xs:
  vars_conf_unique xs (t ; fs1 ; S ; e ; Lv) ->
  fv fs1 \c fv fs2 ->
  vars_conf_unique xs (t ; fs2 ; S ; e ; Lv).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  union_sub_solver.
Qed.

Lemma vars_conf_unique_S_subset t fs S1 S2 e Lv xs:
  vars_conf_unique xs (t ; fs ; S1 ; e ; Lv) ->
  fv_heap S1 \c fv_heap S2 ->
  vars_conf_unique xs (t ; fs ; S2 ; e ; Lv).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  forwards: subset_union_l2 Hsu.
  forwards: subset_union_l1 Hsu.
  union_sub_solver.
Qed.


Lemma vars_conf_unique_e_subset t fs S e1 e2 Lv xs:
  vars_conf_unique xs (t ; fs ; S ; e1 ; Lv) ->
  fv_renv e1 \c fv_renv e2 ->
  vars_conf_unique xs (t ; fs ; S ; e2 ; Lv).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  forwards: subset_union_l2 Hsu.
  forwards: subset_union_l1 Hsu.
  union_sub_solver.
Qed.

