Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import Vars TransformCollect Types.

Require Import Coq.Sets.Ensembles.
Require Import Coq.Sets.Constructive_sets.

(* Part of Figure 3 in "Reference mutability for DOT" *)
(** * Typing context ([Gamma]) *)
Inductive ctx: Set :=
(* Empty context with no variables *)
| ctx_empty
(* Context with a variable added at the end *)
| ctx_push: ctx -> var -> varkind -> typ -> ctx
(* Context making all previous variables inaccessible *)
| ctx_stop: ctx -> ctx
(* Context making all previous variables "invisible" *)
| ctx_hide: ctx -> ctx
.

Notation "G '§' v / vk ~ T" := (ctx_push G v vk T)
  (at level 30, vk at level 20, v at level 20, T at level 20).
Notation "v / vk ~ T" := (ctx_push ctx_empty v vk T)
  (at level 30, vk at level 20, T at level 20).
Notation "G '§ʸ' v ~ T" := (ctx_push G v v_loc T)
  (at level 30, v at level 20, T at level 20).
Notation "G '§ʷ' v ~ T" := (ctx_push G v v_ref T)
  (at level 30, v at level 20, T at level 20).
Notation "G '§ʳ' v ~ T" := (ctx_push G v v_recv T)
  (at level 30, v at level 20, T at level 20).
Notation "G '§ᶻ' v ~ T" := (ctx_push G v v_par T)
  (at level 30, v at level 20, T at level 20).
Notation "G '§ˢ' v ~ T" := (ctx_push G v v_self T)
  (at level 30, v at level 20, T at level 20).
Notation "G '§ˣ' v ~ T" := (ctx_push G v v_let T)
  (at level 30, v at level 20, T at level 20).


(** ** Context concatenation *)

Fixpoint ctx_concat(G1 G2: ctx): ctx :=
match G2 with
| ctx_empty => G1
| ctx_push c v vk t => ctx_push (ctx_concat G1 c) v vk t
| ctx_stop c => ctx_stop (ctx_concat G1 c)
| ctx_hide c => ctx_hide (ctx_concat G1 c)
end.

Notation "E §& F" := (ctx_concat E F)
  (at level 28, left associativity) : env_scope.

(* Concatenation properties *)
Lemma ctx_concat_empty_r(G1: ctx):
  G1 §& ctx_empty = G1.
Proof.
  unfold ctx_concat. apply eq_refl.
Qed.

Lemma ctx_concat_push v vk t G1 G2:
  ctx_concat G2 (ctx_push G1 v vk t) =
  ctx_push (ctx_concat G2 G1) v vk t.
Proof.
  unfold ctx_concat. apply eq_refl.
Qed.

Lemma ctx_concat_stop G1 G2:
  ctx_concat G2 (ctx_stop G1) =
  ctx_stop (ctx_concat G2 G1).
Proof.
  unfold ctx_concat. apply eq_refl.
Qed.

Lemma ctx_concat_hide G1 G2:
  ctx_concat G2 (ctx_hide G1) =
  ctx_hide (ctx_concat G2 G1).
Proof.
  unfold ctx_concat. apply eq_refl.
Qed.

Hint Rewrite ctx_concat_empty_r ctx_concat_push ctx_concat_stop ctx_concat_hide: ctx_flatten.
Tactic Notation "ctx_flatten" := (autorewrite with ctx_flatten).
Tactic Notation "ctx_flatten" "in" hyp(H) := (autorewrite with ctx_flatten in H).
Tactic Notation "ctx_flatten" "in" "*" := (autorewrite with ctx_flatten in *).

Lemma ctx_concat_empty_l G1:
  ctx_empty §& G1 = G1.
Proof.
  induction G1.
  all: ctx_flatten.
  all: try rewrite IHG1.
  all: auto.
Qed.

Lemma ctx_concat_empty_inv_l(G1 G2: ctx):
  ctx_empty = (G1 §& G2) ->
  G1 = ctx_empty.
Proof.
  intros.
  destruct G2.
  all: ctx_flatten in H.
  all: try discriminate H.
  all: auto.
Qed.

Lemma ctx_concat_empty_inv_r(G1 G2: ctx):
  ctx_empty = (G1 §& G2) ->
  G2 = ctx_empty.
Proof.
  intros.
  destruct G2.
  all: ctx_flatten in H.
  all: try discriminate H.
  all: auto.
Qed.

Lemma ctx_concat_assoc(G1 G2 G3: ctx):
  G1 §& (G2 §& G3) =
  G1 §& G2 §& G3.
Proof.
  revert G1 G2.
  induction G3.
  all: intros.
  all: ctx_flatten.
  all: try rewrite IHG3.
  all: apply eq_refl.
Qed.


Lemma ctx_concat_empty_r2(G1 G2: ctx):
  G1 §& G2 §& ctx_empty =
  G1 §& G2.
Proof.
  rewrite <- ctx_concat_assoc.
  rewrite ctx_concat_empty_r.
  apply eq_refl.
Qed.

Lemma ctx_concat_single G x vk T:
  G § x / vk ~ T =
  G §& (ctx_empty § x / vk ~ T).
Proof.
  ctx_flatten.
  apply eq_refl.
Qed.
Lemma ctx_concat_double1 G x vk T x2 vk2 T2:
  G § x / vk ~ T § x2 / vk2 ~ T2 =
  G §& ((ctx_empty § x / vk ~ T) § x2 / vk2 ~ T2).
Proof.
  ctx_flatten.
  apply eq_refl.
Qed.
Lemma ctx_concat_double2 G x vk T x2 vk2 T2:
  G § x / vk ~ T § x2 / vk2 ~ T2 =
  G §& (ctx_empty § x / vk ~ T) §& (ctx_empty § x2 / vk2 ~ T2).
Proof.
  ctx_flatten.
  apply eq_refl.
Qed.

Lemma ctx_push_concat G x vk T G1:
  (G § x / vk ~ T) §& G1 = G §& ((x / vk ~ T) §& G1).
Proof.
  rewrite ctx_concat_assoc.
  rewrite ctx_concat_push.
  rewrite ctx_concat_empty_r.
  auto.
Qed.

Lemma ctx_stop_concat G G1:
  (ctx_stop G) §& G1 = G §& ((ctx_stop ctx_empty) §& G1).
Proof.
  rewrite ctx_concat_assoc.
  rewrite ctx_concat_stop.
  rewrite ctx_concat_empty_r.
  auto.
Qed.

Lemma ctx_hide_concat G G1:
  (ctx_hide G) §& G1 =   (G) §& ((ctx_hide ctx_empty) §& G1).
Proof.
  rewrite ctx_concat_assoc.
  rewrite ctx_concat_hide.
  rewrite ctx_concat_empty_r.
  auto.
Qed.



(** ** Value map *)
(* used for substitution *)

Fixpoint ctx_map(fn: typ -> typ)(G: ctx): ctx :=
match G with
| ctx_empty => ctx_empty
| ctx_push G1 v vk t => ctx_push (ctx_map fn G1) v vk (fn t)
| ctx_stop G1 => ctx_stop (ctx_map fn G1)
| ctx_hide G1 => ctx_hide (ctx_map fn G1)
end.

Lemma ctx_map_empty f:
  ctx_map f ctx_empty =
  ctx_empty.
Proof.
  unfold ctx_map; auto.
Qed.

Lemma ctx_map_push(f: typ -> typ) x vk T G:
  ctx_map f (G § x / vk ~ T) =
  (ctx_map f G § x / vk ~ f T).
Proof.
  auto.
Qed.

Lemma ctx_map_stop(f: typ -> typ) G:
  ctx_map f (ctx_stop G) =
  ctx_stop (ctx_map f G).
Proof.
  auto.
Qed.

Lemma ctx_map_hide(f: typ -> typ) G:
  ctx_map f (ctx_hide G) =
  ctx_hide (ctx_map f G).
Proof.
  auto.
Qed.


Hint Rewrite ctx_map_empty ctx_map_push ctx_map_stop ctx_map_hide: ctx_flatten.

Lemma ctx_map_concat f G1 G2:
  ctx_map f (G1 §& G2) =
  ctx_map f G1 §& ctx_map f G2.
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: try rewrite IHG2.
  all: auto.
Qed.

(** Collect variables *)
(* used for free variables *)

Fixpoint ctx_collect_values(A: Set)(fn: typ -> fset A)(G: ctx): fset A :=
match G with
| ctx_empty => \{}
| ctx_push G1 v vk t => (fn t) \u (ctx_collect_values fn G1)
| ctx_stop G1 => (ctx_collect_values fn G1)
| ctx_hide G1 => (ctx_collect_values fn G1)
end.
Notation ctx_fv := (ctx_collect_values fv).
Instance CtxFreeVar : FreeVar ctx := ctx_fv.
(* FV *)

Lemma ctx_fv_empty:
  ctx_fv ctx_empty = \{}.
Proof.
  unfold ctx_fv. apply eq_refl.
Qed.

Lemma ctx_fv_push x vk v2 G:
  ctx_fv (G § x / vk ~ v2) =
  (fv v2) \u (ctx_fv G).
Proof.
  unfold ctx_fv.
  auto.
Qed.

Lemma ctx_fv_stop G:
  ctx_fv (ctx_stop G) =
  ctx_fv G.
Proof.
  unfold ctx_fv.
  auto.
Qed.

Lemma ctx_fv_hide G:
  ctx_fv (ctx_hide G) =
  ctx_fv G.
Proof.
  unfold ctx_fv.
  auto.
Qed.

Lemma ctx_fv_concat G1 G2:
  ctx_fv(G1 §& G2) = ctx_fv(G1) \u ctx_fv(G2).
Proof.
  induction G2.
  all: ctx_flatten.
  all: try rewrite union_empty_r.
  all: auto.
  rewrite_all ctx_fv_push.
  rewrite IHG2. auto.
Qed.

Ltac ctx_simpl_fv :=
  rewrite_all ctx_fv_concat in *;
  rewrite_all ctx_fv_push in *;
  rewrite_all ctx_fv_stop in *;
  rewrite_all ctx_fv_hide in *;
  rewrite_all ctx_fv_empty in *.

(** ** Keys *)

Fixpoint ctx_keys(G: ctx): list var :=
match G with
| ctx_empty => nil
| ctx_push G v vk t =>  v :: (ctx_keys G)
| ctx_stop G => ctx_keys G
| ctx_hide G => ctx_keys G
end.

Lemma ctx_keys_concat (G1 G2: ctx):
  ctx_keys (G1 §& G2) = ((ctx_keys G2) ++ (ctx_keys G1))%list.
Proof.
  induction G2; simpls; rewrite ? List.app_nil_r; eauto.
  rewrite IHG2. rewrite List.app_comm_cons. auto.
Qed.

(** ** Domain *)

Fixpoint ctx_dom(G: ctx): fset var :=
match G with
| ctx_empty => \{}
| ctx_push G v vk t =>  \{ v } \u (ctx_dom G)
| ctx_stop G => ctx_dom G
| ctx_hide G => ctx_dom G
end.

Lemma ctx_dom_empty:
  ctx_dom ctx_empty = \{}.
Proof.
  auto.
Qed.

Lemma ctx_dom_push G v vk t:
  ctx_dom (ctx_push G v vk t) = \{ v } \u (ctx_dom G).
Proof.
  auto.
Qed.

Lemma ctx_dom_stop G:
  ctx_dom (ctx_stop G) = ctx_dom G.
Proof.
  auto.
Qed.

Lemma ctx_dom_hide G:
  ctx_dom (ctx_hide G) = ctx_dom G.
Proof.
  auto.
Qed.

Hint Rewrite ctx_dom_empty ctx_dom_push ctx_dom_stop ctx_dom_hide: ctx_flatten.

Lemma ctx_in_dom_push G v vk t:
  v \in ctx_dom (ctx_push G v vk t).
Proof.
  unfold ctx_dom.
  rewrite in_union.
  left.
  apply in_singleton_self.
Qed.

Lemma ctx_dom_concat G1 G2:
  ctx_dom(G1 §& G2) = ctx_dom(G1) \u ctx_dom(G2).
Proof.
  induction G2.
  all: ctx_flatten.
  all: try rewrite union_empty_r.
  all: try rewrite IHG2.
  all: auto.
Qed.

Lemma ctx_dom_concat_r G1 G2 G3:
  ctx_dom(G2) = ctx_dom(G3) ->
  ctx_dom(G1 §& G2) = ctx_dom(G1 §& G3).
Proof.
  intros.
  rewrite ctx_dom_concat.
  rewrite ctx_dom_concat.
  rewrite H.
  auto.
Qed.

Lemma ctx_dom_map f G1:
  ctx_dom(ctx_map f G1) = ctx_dom(G1).
Proof.
  induction G1.
  all: ctx_flatten.
  all: try rewrite IHG1.
  all: auto.
Qed.

Lemma ctx_dom_concat_map_r f G1 G2:
  ctx_dom(G1 §& G2) = ctx_dom(G1 §& ctx_map f G2).
Proof.
  apply ctx_dom_concat_r.
  apply eq_sym.
  rewrite ctx_dom_map.
  auto.
Qed.

Lemma ctx_keys_dom G:
  from_list (ctx_keys G) = ctx_dom G.
Proof.
  induction G; simpls; rewrite ? from_list_nil, ? from_list_cons, ? IHG; auto.
Qed.


Ltac ctx_simpl_dom :=
  rewrite_all ctx_dom_map in *;
  rewrite_all ctx_dom_concat in *;
  rewrite_all ctx_dom_push in *;
  rewrite_all ctx_dom_stop in *;
  rewrite_all ctx_dom_hide in *;
  rewrite_all ctx_dom_empty in *.

(** ** Unbound variable *)

Notation "x '§#' E" := (x \notin (ctx_dom E)) (at level 67) : env_scope.

(* Solves goals `x §# G` by simplifying ctx_dom terms (unfolding ctx_dom) *)
Hint Extern 1 (_ §# _) => ctx_simpl_dom; notin_solve : core.

Lemma ctx_unbound_push_inv(x0 x: var) vk G T:
  x0 §# G § x / vk ~ T ->
  x0 §# G /\ x <> x0.
Proof.
  induction G;
  unfold ctx_dom;
  auto.
Qed.
Lemma ctx_unbound_push_inv_unbound(x0 x: var) vk G T:
  x0 §# G § x / vk ~ T ->
  x0 §# G.
Proof.
  intros.
  set (ctx_unbound_push_inv H).
  destruct a.
  assumption.
Qed.

Lemma ctx_unbound_push_inv_ne2(x0 x: var) vk G T:
  x0 §# G § x / vk ~ T ->
  x0 <> x.
Proof.
  intro.
  apply neq_sym.
  revert H.
  eapply ctx_unbound_push_inv.
Qed.

Lemma ctx_unbound_push(x0 x: var) vk G T:
  x <> x0 -> x0 §# G ->
  x0 §# G § x / vk ~ T.
Proof.
  induction G;
  unfold ctx_dom; auto.
Qed.

Lemma ctx_unbound_empty(x0: var):
  x0 §# ctx_empty.
Proof.
  unfold ctx_dom. auto.
Qed.

Hint Resolve ctx_unbound_empty : core.

Lemma ctx_unbound_stop_inv(x0: var) G:
  x0 §# (ctx_stop G) ->
  x0 §# G.
Proof.
  auto.
Qed.

Lemma ctx_unbound_stop(x0: var) G:
  x0 §# G ->
  x0 §# (ctx_stop G).
Proof.
  auto.
Qed.

Lemma ctx_unbound_hide_inv(x0: var) G:
  x0 §# (ctx_hide G) ->
  x0 §# G.
Proof.
  auto.
Qed.

Lemma ctx_unbound_hide(x0: var) G:
  x0 §# G ->
  x0 §# (ctx_hide G).
Proof.
  auto.
Qed.

Lemma ctx_unbound_concat(v: var)(G1 G2: ctx):
  v §# G1 ->
  v §# G2 ->
  v §# (G1 §& G2).
Proof.
  intros.
  rewrite ctx_dom_concat.
  auto.
Qed.

Lemma ctx_unbound_concat_l(v: var) G1 G2:
  v §# (G1 §& G2) ->
  v §# G1.
Proof.
  intros.
  rewrite ctx_dom_concat in H.
  auto.
Qed.

Lemma ctx_unbound_concat_r(v: var) G1 G2:
  v §# (G1 §& G2) ->
  v §# G2.
Proof.
  intros.
  rewrite ctx_dom_concat in H.
  auto.
Qed.

(* Solves goals `x §# G1` when there is `x §# G2` and they have the same domain *)
Hint Extern 1 (_ §# _) =>
match goal with
| [H : ctx_dom ?G1 = ctx_dom ?G2, _ : ?x §# ?G1 |- ?x §# ?G2 ] => rewrite <- H; assumption
| [H : ctx_dom ?G2 = ctx_dom ?G1, _ : ?x §# ?G1 |- ?x §# ?G2 ] => rewrite H; assumption
end : core.

(* Solves goals `x # e` when there is `x §# G` and they have the same domain *)
Hint Extern 1 (_ # _) =>
match goal with
              | H:ctx_dom ?G1 = dom ?G2, _:?x §# ?G1
                |- ?x # ?G2 => rewrite <- H; assumption
              | H:ctx_dom ?G2 = dom ?G1, _:?x §# ?G1
                |- ?x # ?G2 => rewrite H; assumption
              end : core.

(* OK *)
(* a context is ok if it does not have duplicate variables *)

Inductive ctx_ok: ctx -> Prop :=
| ctx_ok_empty:
  ctx_ok ctx_empty
| ctx_ok_push(G: ctx)(v: var)(vk: varkind)(t: typ):
  ctx_ok G ->
  v §# G ->
  ctx_ok (ctx_push G v vk t)
| ctx_ok_stop(G: ctx):
  ctx_ok G -> ctx_ok (ctx_stop G)
| ctx_ok_hide(G: ctx):
  ctx_ok G -> ctx_ok (ctx_hide G)
.

Hint Constructors ctx_ok : core.

Lemma ctx_ok_push_inv_ok G v vk t:
  ctx_ok (ctx_push G v vk t) ->
  ctx_ok G.
Proof.
  intros.
  inversion H.
  auto.
Qed.

Hint Extern 1 (ctx_ok (?E)) =>
  match goal with
    H: ctx_ok (ctx_push E _ _ _) |- _ =>
      apply (ctx_ok_push_inv_ok H)
  end : core.

Lemma ctx_ok_stop_inv G:
  ctx_ok (ctx_stop G) ->
  ctx_ok G.
Proof.
  intros.
  inversion H.
  auto.
Qed.
Lemma ctx_ok_hide_inv G:
  ctx_ok (ctx_hide G) ->
  ctx_ok G.
Proof.
  intros.
  inversion H.
  auto.
Qed.

(* Solves goals ctx_ok(ctx_stop _) and ctx_ok(ctx_hide _) *)
Hint Extern 1 (ctx_ok (?E)) =>
  match goal with
  | H: ctx_ok (ctx_stop E) |- _ =>
     apply (ctx_ok_stop_inv H)
  | H: ctx_ok (ctx_hide E) |- _ =>
     apply (ctx_ok_hide_inv H)
  end : core.

Lemma ctx_ok_push_inv_unbound G v vk t:
  ctx_ok (ctx_push G v vk t) ->
  v §# G.
Proof.
  intros.
  inversion H.
  auto.
Qed.


Lemma ctx_ok_concat_l G1 G2:
  ctx_ok (G1 §& G2) ->
  ctx_ok G1.
Proof.
  intros.
  induction G2.
  all: ctx_flatten in H.
  - assumption.
  - eauto using ctx_ok_push_inv_ok.
  - eauto using ctx_ok_stop_inv.
  - eauto using ctx_ok_hide_inv.
Qed.

Lemma ctx_ok_push_concat G1 x vk t G2:
  ctx_ok ((G1 § x / vk ~ t) §& G2) ->
  ctx_ok (G1 §& G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten in *; auto.
  - apply ctx_ok_push; auto.
    set (ctx_ok_push_inv_unbound H).
    ctx_simpl_dom.
    auto.
Qed.

Lemma ctx_ok_concat_r G1 G2:
  ctx_ok (G1 §& G2) ->
  ctx_ok G2.
Proof.
  intros.
  induction G2.
  all: ctx_flatten in H; eauto.
  inversions H; eauto.
Qed.


Lemma ctx_ok_middle G1 x vk T G2:
  ctx_ok (G1 §& G2) ->
  x §# G1 ->
  x §# G2 ->
  ctx_ok (G1 § x /vk ~ T §& G2).
Proof.
  introv Hok Hn1 Hn2.
  induction G2; ctx_flatten in *.
  - auto.
  - constructor. apply IHG2; auto. inversions Hok. ctx_simpl_dom. notin_solve.
  - auto.
  - auto.
Qed.
Lemma ctx_ok_weaken G1 G2 G3:
  ctx_ok (G1 §& G2 §& G3) ->
  ctx_ok (G1 §& G3).
Proof.
  intros.
  induction G3.
  - rewrite ctx_concat_empty_r in *. eapply ctx_ok_concat_l; eauto.
  - ctx_flatten in *. inversions H. constructor; eauto.
  - ctx_flatten in *. inversions H. constructor; eauto.
  - ctx_flatten in *. inversions H. constructor; eauto.
Qed.

Lemma ctx_ok_concat G1 G2:
  ctx_ok G1 -> ctx_ok G2 ->
  disjoint (ctx_dom G1) (ctx_dom G2) ->
  ctx_ok (G1 §& G2).
Proof.
  introv Hok1 Hok2 Hnmt. induction G2; simpl; eauto.
  constructor; simpl in Hnmt; hnf in *.
  - apply IHG2; eauto. eapply fset_extens, subset_empty_l.
    rewrite inter_union_r in Hnmt.
    hnf; intros x Hin. rewrite <- Hnmt, in_union; eauto.
  - intros Hin. rewrite ctx_dom_concat, in_union in Hin.
    inversions Hok2.  destruct Hin as [Hin1 | Hin2]; eauto.
    rewrite <- in_empty with (x := v), <- Hnmt, in_inter; split; eauto.
    rewrite in_union; eauto using in_singleton_self.
Qed.

Lemma ctx_ok_concat_rl G1 G2 G3:
  ctx_ok (G1 §& (G2 §& G3)) ->
  ctx_ok (G1 §& G2).
Proof.
  intros.
  rewrite ctx_concat_assoc in H.
  eapply ctx_ok_concat_l.
  eauto.
Qed.
Lemma ctx_ok_concat_rr G1 G2 G3:
  ctx_ok (G1 §& (G2 §& G3)) ->
  ctx_ok (G1 §& G3).
Proof.
  intros.
  rewrite ctx_concat_assoc in H.
  eapply ctx_ok_weaken; eauto.
Qed.
Lemma ctx_ok_middle_inv (G1 : ctx) (x : var) (vk : varkind) (T : typ) (G2 : ctx):
 ctx_ok ((G1 § x / vk ~ T) §& G2) ->
 ctx_ok (G1 §& G2) /\ x §# G1 /\ x §# G2.
Proof.
  introv Hok.
  induction G2; ctx_flatten in *; splits.
  - inversions Hok; eauto.
  - inversions Hok; eauto.
  - eauto.
  - inversions Hok. constructor. eapply IHG2; eauto. ctx_simpl_dom; notin_solve.
  -   inversions Hok. eapply ctx_ok_concat_l in H1. inversions H1. auto.
  -   inversions Hok. rewrite notin_union; split.  ctx_simpl_dom; notin_solve. eapply IHG2; eauto.
  -   inversions Hok. constructor.  eapply IHG2; eauto.
  -   inversions Hok. eapply ctx_ok_concat_l in H0. inversions H0. auto.
  -   inversions Hok.  eapply IHG2; eauto.
  -   inversions Hok. constructor.  eapply IHG2; eauto.
  -   inversions Hok. eapply ctx_ok_concat_l in H0. inversions H0. auto.
  -   inversions Hok.  eapply IHG2; eauto.
Qed.

Lemma ctx_ok_stop_l_concat_inv G1 G2 :
  ctx_ok (ctx_stop G1 §& G2) -> ctx_ok (G1 §& G2).
Proof.
  introv Hok. induction G2; simpl in *; eauto.
  inversions Hok. eauto.
Qed.

Lemma ctx_ok_hide_l_concat_inv G1 G2 :
  ctx_ok (ctx_hide G1 §& G2) -> ctx_ok (G1 §& G2).
Proof.
  introv Hok. induction G2; simpl in *; eauto.
  inversions Hok. eauto.
Qed.


Lemma ctx_ok_concat_inv G1 G2 :
  ctx_ok (G1 §& G2) ->
  ctx_ok G1 /\ ctx_ok G2 /\ disjoint (ctx_dom G1) (ctx_dom G2).
Proof.
  introv Hok. induction G1; simpl in *.
  - repeat(split); eauto.
    + rewrite ctx_concat_empty_l in Hok. auto.
    + hnf. rewrite inter_empty_l. auto.
  - apply ctx_ok_middle_inv in Hok. destruct Hok as [Hok [Hvnin1 Hvnin2]].
    specialize (IHG1 Hok). destruct IHG1 as [Hok1 [Hok2 HD]].
    repeat(split); eauto. hnf. rewrite inter_union_l, HD, union_empty_r.
    eauto using inter_singleton_notin.
  - apply ctx_ok_stop_l_concat_inv in Hok. specialize (IHG1 Hok).
    destruct_all. repeat(split); eauto.
  - apply ctx_ok_hide_l_concat_inv in Hok. specialize (IHG1 Hok).
    destruct_all. repeat(split); eauto.
Qed.




Lemma ctx_ok_fresh_keys G:
  ctx_ok G ->
  fresh \{} (length (ctx_keys G)) (ctx_keys G).
Proof.
  introv Hok.
  induction Hok; simpls; eauto.
  split. apply notin_empty.
  fresh_solve.
  eapply fresh_from_list.
  rewrite ctx_keys_dom; auto.
  eauto.
Qed.
Lemma ctx_fresh_keys_ok G:
  fresh \{} (length (ctx_keys G)) (ctx_keys G) ->
  ctx_ok G.
Proof.
  introv Hok.
  induction G; simpls; destructecs_all; eauto.
  constructor. eauto.
  rewrite union_empty_l in Hok.
  rewrite <- ctx_keys_dom.
  eapply fresh_single_notin; eauto.
Qed.

Lemma ctx_push_eq_inv Ga1 xa xb vka Sa Gb1 vkb Sb: 
  Ga1 § xa / vka ~ Sa = Gb1 § xb / vkb ~ Sb ->
  xa = xb /\ Ga1 = Gb1 /\ vka = vkb /\ Sa = Sb.
Proof.  
  introv Heq.
  inversions Heq.
  auto.
Qed.

Lemma ctx_ok_middle_eq_inv Ga1 x vka Sa Ga2 Gb1 vkb Sb Gb2:
  (Ga1 § x / vka ~ Sa) §& Ga2 = (Gb1 § x / vkb ~ Sb) §& Gb2 ->
  ctx_ok ((Gb1 § x / vkb ~ Sb) §& Gb2) ->
  Ga1 = Gb1 /\ vka = vkb /\ Sa = Sb /\ Ga2 = Gb2.
Proof.
  introv Heq Hok.
  reverts Ga2.
  induction Gb2; intros; simpls.
  -
  lets Hok': Hok.
  rewrite <- Heq in Hok'.
  forwards: ctx_ok_middle_inv Hok'.
  destruct Ga2; simpls; inversions Heq.
  + auto.
  + false. destructecs H. eapply H. rewrite in_union, in_singleton. auto.
  -
  destruct Ga2; simpls; inversions Heq.
  + inversions Hok. ctx_simpl_dom. false. apply H4. rewrite ! in_union, in_singleton. auto.
  + forwards: IHGb2. eauto. eapply H0. destructecs H; subst; auto.
  -
  destruct Ga2; simpls; inversions Heq.
 forwards: IHGb2. eauto. eapply H0. destructecs H; subst; auto.
  -
  destruct Ga2; simpls; inversions Heq.
 forwards: IHGb2. eauto. eapply H0. destructecs H; subst; auto.
Qed.



(** ** Weakens *)
(* a context G2 "weakens" if, when appended after context G1,
   does not affect variables in G1.
   G2 cannot concaint ctx_stop and ctx_hide *)

Inductive ctx_weakens: ctx -> Prop :=
| ctx_weakens_empty: ctx_weakens ctx_empty
| ctx_weakens_push G v vk t:
  ctx_weakens G ->
  ctx_weakens (ctx_push G v vk t)
.

Inductive ctx_weakens_hide: ctx -> Prop :=
| ctx_weakens_hide_empty: ctx_weakens_hide ctx_empty
| ctx_weakens_hide_push G v vk t:
    ctx_weakens_hide G ->
    ctx_weakens_hide (ctx_push G v vk t)
| ctx_weakens_hide_hide G:
    ctx_weakens_hide G ->
    ctx_weakens_hide (ctx_hide G)
.

Hint Constructors ctx_weakens ctx_weakens_hide : core.

Lemma ctx_weakens_concat G1 G2:
  ctx_weakens G1 ->
  ctx_weakens G2 ->
  ctx_weakens (G1 §& G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: inversion H0; auto.
Qed.

Lemma ctx_weakens_map G f:
  ctx_weakens G ->
  ctx_weakens (ctx_map f G).
Proof.
  intros.
  induction G.
  all: ctx_flatten.
  all: inversion H; auto.
Qed.

Lemma ctx_weakens_push_inv G v vk t:
  ctx_weakens (ctx_push G v vk t) ->
  ctx_weakens G.
Proof.
  intros.
  inversion H.
  auto.
Qed.

Lemma ctx_weakens_stop_false G:
  ctx_weakens (ctx_stop G) ->
  False.
Proof.
  intros.
  inversion H.
Qed.

Lemma ctx_weakens_hide_false G:
  ctx_weakens (ctx_hide G) ->
  False.
Proof.
  intros.
  inversion H.
Qed.

Lemma ctx_weakens_concat_left G1 G2:
  ctx_weakens (G1 §& G2) ->
  ctx_weakens G1.
Proof.
  intros.
  induction G2.
  - rewrite ctx_concat_empty_r in H. auto.
  - ctx_flatten in H.
    apply ctx_weakens_push_inv in H.
    auto.
  - rewrite ctx_concat_stop in H.
    apply ctx_weakens_stop_false in H.
    contradiction.
  - rewrite ctx_concat_hide in H.
    apply ctx_weakens_hide_false in H.
    contradiction.
Qed.

Lemma ctx_weakens_concat_right G1 G2:
  ctx_weakens (G1 §& G2) ->
  ctx_weakens G2.
Proof.
  intros.
  induction G2.
  - auto.
  - ctx_flatten in H.
    apply ctx_weakens_push_inv in H.
    auto.
  - rewrite ctx_concat_stop in H.
    apply ctx_weakens_stop_false in H.
    contradiction.
  - rewrite ctx_concat_hide in H.
    apply ctx_weakens_hide_false in H.
    contradiction.
Qed.


(** ** Binds *)
(* a context binds a variable if it contains it by ctx_push
   and after that, there is no ctx_stop
   or another occurrence of the same variable *)

Inductive ctx_binds: var -> typ -> ctx -> Prop :=
| ctx_binds_e(v: var)(vk: varkind)(t: typ)(G: ctx):
    ctx_binds v t (ctx_push G v vk t)
| ctx_binds_n(v v2: var)(vk: varkind)(t t2: typ)(G: ctx):
    v <> v2 ->
    ctx_binds v t G ->
    ctx_binds v t (ctx_push G v2 vk t2)
| ctx_binds_hide(v: var)(t: typ)(G: ctx):
    ctx_binds v t G ->
    ctx_binds v t (ctx_hide G)
.

Hint Resolve ctx_binds_e ctx_binds_n ctx_binds_hide: core.

Lemma ctx_binds_empty_inv v t:
  ctx_binds v t ctx_empty ->
  False.
Proof.
  intros. inversion H.
Qed.

Lemma ctx_binds_push_inv x1 x2 vk G v1 v2:
  ctx_binds x1 v1 (G § x2 / vk ~ v2 ) ->
  ((x1 = x2) /\ (v1 = v2)) \/
  ((x1 <> x2) /\ (ctx_binds x1 v1 G)).
Proof.
  intros.
  inversion H; subst; auto.
Qed.

Lemma ctx_binds_stop_inv v t G:
  ctx_binds v t (ctx_stop G) ->
  False.
Proof.
  intros. inversion H.
Qed.

Lemma ctx_binds_hide_inv v t G:
  ctx_binds v t (ctx_hide G) ->
  ctx_binds v t G.
Proof.
  intros. inversion H; subst. auto.
Qed.


Lemma ctx_binds_push_inv_n x1 x2 vk G v1 v2:
  ctx_binds x1 v1 (G § x2 / vk ~ v2 ) ->
  (x1 <> x2) ->
  (ctx_binds x1 v1 G).
Proof.
  intros.
  set (ctx_binds_push_inv H).
  destruct o.
  destruct H1.
  false.
  destruct H1.
  auto.
Qed.


Lemma ctx_binds_push_inv_e x1 vk G v1 v2:
  ctx_binds x1 v1 (G § x1 / vk ~ v2 ) ->
  v1 = v2.
Proof.
  intros.
  set (ctx_binds_push_inv H).
  destruct o.
  destruct H0. auto.
  destruct H0. contradiction.
Qed.

Lemma ctx_binds_push_vkchange x1 x2 vk1 vk2 G v1 v2:
  ctx_binds x1 v1 (G § x2 / vk1 ~ v2 ) ->
  ctx_binds x1 v1 (G § x2 / vk2 ~ v2 ).
Proof.
  intros.
  inversion H; subst; auto.
Qed.

Lemma ctx_binds_dom G v t:
  ctx_binds v t G ->
  (v \in (ctx_dom G)).
Proof.
  intros.
  induction G.
  - inversion H.
  - induction H.
    + apply ctx_in_dom_push.
    + unfold ctx_dom. rewrite in_union. right. fold ctx_dom. auto.
    + auto.
  - inversion H.
  - inversion H; auto.
Qed.

Lemma ctx_dom_to_binds G v:
  ctx_ok G ->
  ctx_weakens G ->
  (v \in (ctx_dom G)) ->
  exists t,
  ctx_binds v t G
.
Proof.
  intros.
  induction G.
  - rewrite ctx_dom_empty in H1.
    false.
    rewrite in_empty in H1.
    assumption.
  - rewrite ctx_dom_push in H1.
    rewrite in_union in H1.
    destruct H1.
    rewrite in_singleton in H1; subst.
    exists t. auto.

    pose proof (ctx_ok_push_inv_unbound H).
    pose proof (ctx_ok_push_inv_ok H).
    pose proof (ctx_weakens_push_inv H0).
    specialize (IHG H3 H4 H1).
    destruct IHG.
    exists x.
    apply ctx_binds_n.

    intro; subst.
    contradiction.
    auto.
  - rewrite ctx_dom_stop in H1. false.
    eapply ctx_weakens_stop_false; eauto.
  - false.
    eapply ctx_weakens_hide_false; eauto.
Qed.


Lemma ctx_binds_unbound_false G v t:
  ctx_binds v t G ->
  v §# G ->
  False.
Proof.
  intros.
  set (ctx_binds_dom H).
  contradict m.
  apply H0.
  eapply ctx_binds_dom. eauto.
Qed.

Lemma ctx_binds_concat_inv G1 G2 x a:
  ctx_binds x a (G1 §& G2) ->
  ctx_binds x a G1 \/ ctx_binds x a G2.
Proof.
  intros.
  induction G2.
  all: ctx_flatten in H.
  all: inversion H; subst; auto.
  - inversion H; subst.
    + right. apply ctx_binds_e.
    + specialize (IHG2 H7).
      destruct IHG2; auto.
  - specialize (IHG2 H3).
    destruct IHG2; auto.
Qed.

Lemma ctx_binds_concat_left(v: var) t G1 G2:
  ctx_binds v t G1 ->
  ctx_weakens G2 ->
  v §# G2 ->
  ctx_binds v t (G1 §& G2) .
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: inversion H0; auto.
  - apply ctx_binds_n.
    eapply ctx_unbound_push_inv_ne2. apply H1.
    apply IHG2.
    assumption.
    eapply ctx_unbound_push_inv. eauto.
Qed.

Lemma ctx_binds_concat_left_hide(v: var) t G1 G2:
  ctx_binds v t G1 ->
  ctx_weakens_hide G2 ->
  v §# G2 ->
  ctx_binds v t (G1 §& G2) .
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: inversion H0; auto.
  - apply ctx_binds_n.
    eapply ctx_unbound_push_inv_ne2. apply H1.
    apply IHG2.
    assumption.
    eapply ctx_unbound_push_inv. eauto.
Qed.

Lemma ctx_binds_concat_left_inv(v: var) t G1 G2:
  ctx_binds v t (G1 §& G2) ->
  v §# G2 ->
  ctx_binds v t G1.
Proof.
  intros.
  pose proof (ctx_binds_concat_inv G1 G2 H).
  destruct H1. auto.
  false. eapply ctx_binds_unbound_false; eauto.
Qed.


Lemma ctx_binds_concat_right(v: var) t G1 G2:
  ctx_binds v t G2 ->
  ctx_binds v t (G1 §& G2) .
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: inversion H; subst.
  + apply ctx_binds_e.
  + apply ctx_binds_n.
    assumption.
    apply IHG2. assumption.
  + apply ctx_binds_hide_inv; auto.
Qed.

Hint Resolve ctx_binds_concat_left ctx_binds_concat_right: core.

Lemma ctx_binds_map x v (f: typ -> typ) G:
  ctx_binds x v G ->
  ctx_binds x (f v) (ctx_map f G).
Proof.
  intros.
  induction H.
  all: ctx_flatten.
  all: auto.
Qed.


Lemma ctx_binds_ok_concat_unbound_l x T G1 G2:
    ctx_binds x T G2 ->
    ctx_ok (G1 §& G2) ->
    x §# G1.
Proof.
  intros.
  induction H.
  - inversion H0; subst.
    eauto.
  - apply IHctx_binds.
    inversion H0; subst; auto.
  - rewrite ctx_concat_hide in H0. auto.
Qed.


Lemma ctx_binds_ok_concat_unbound_r x T G1 G2:
    ctx_binds x T G1 ->
    ctx_ok (G1 §& G2) ->
    x §# G2.
Proof.
  intros.
  induction G2.
  - apply ctx_unbound_empty.
  - rewrite ctx_concat_push in H0.
    inversion H0; subst.
    apply ctx_unbound_push.
    + intro; subst.
      eapply ctx_binds_unbound_false.
      eauto.
      eauto.
    + auto.
  - apply ctx_unbound_stop.
    rewrite ctx_concat_stop in H0.
    auto.
  - apply ctx_unbound_hide.
    rewrite ctx_concat_hide in H0.
    auto.
Qed.

Lemma ctx_binds_weaken(x: var) (a: typ) (G1 G2 G3: ctx):
    ctx_binds x a (G1 §& G3) ->
    ctx_weakens G2 ->
    ctx_ok (G1 §& G2) ->
    ctx_binds x a (G1 §& G2 §& G3).
Proof.
  introv Bi Hwk Hok.

  induction G3.
  - rewrite ctx_concat_empty_r.
    rewrite ctx_concat_empty_r in Bi.
    apply ctx_binds_concat_left; auto.
    eapply ctx_binds_ok_concat_unbound_r;
    eauto.
  - rewrite ctx_concat_push.
    inversion Bi; subst.
    + apply ctx_binds_e.
    + apply ctx_binds_n; auto.
  - rewrite ctx_concat_stop.
    inversion Bi.
  - rewrite ctx_concat_hide.
    rewrite ctx_concat_hide in Bi.
    apply ctx_binds_hide.
    apply ctx_binds_hide_inv in Bi.
    auto.
Qed.

Lemma ctx_binds_middle_inv x1 x2 vk G1 G2 v1 v2:
  ctx_binds x1 v1 (G1 § x2 / vk ~ v2 §& G2) ->
  ctx_binds x1 v1 G2 \/
  ((x1 §# G2) /\ (x1 = x2) /\ (v1 = v2)) \/
  ((x1 §# G2) /\ (x1 <> x2) /\ (ctx_binds x1 v1 G1)).
Proof.
  intros.
  induction G2.
  - rewrite ctx_concat_empty_r in H.
    right.
    inversion H; auto.
  - inversion H; subst.
    + auto.
    + specialize (IHG2 H7).
      destruct IHG2.
      left.
        apply ctx_binds_n; auto.
      right.
      destruct H0.
      destruct H0.
      * left.
        split. apply ctx_unbound_push; auto. auto.
      * right. destruct H0. destruct H1.
        split. apply ctx_unbound_push; auto.
        auto.
  - inversion H.
  - inversion H; subst.
    ctx_flatten in *.
    apply ctx_binds_hide_inv in H.
    specialize (IHG2 H).
    destruct IHG2.
    + left.
      apply ctx_binds_hide. auto.
    + auto.
Qed.

Lemma ctx_binds_middle_concat x1 x2 vk G1 G2 v1 v2:
  ctx_binds x1 v1 (G1 § x2 / vk ~ v2 §& G2) ->
  x1 <> x2 ->
  ctx_binds x1 v1 (G1 §& G2).
Proof.
  intros.
  pose proof ctx_binds_middle_inv.
  specialize (H1 x1 x2 vk G1 G2 v1 v2 H).
  destruct H1.
  apply ctx_binds_concat_right; auto.
  destruct H1.
  destruct H1.
  destruct H2.
  false.

  induction G2.

  - eapply ctx_binds_push_inv_n; eauto.
  - rewrite ctx_concat_push in *.
    destruct H1.
    set (ctx_unbound_push_inv_ne2 H1).
    inversion H; subst.
    false.
    destruct H2.
    set (ctx_unbound_push_inv H1).
    destruct a.
    apply ctx_binds_n; auto.
  - rewrite ctx_concat_stop in *.
    inversion H.
  - rewrite ctx_concat_hide in *.
    inversion H; subst.
    apply ctx_binds_hide_inv in H.
    apply ctx_binds_hide. auto.
Qed.


Lemma ctx_binds_functional x v1 v2 G:
  ctx_binds x v1 G ->
  ctx_binds x v2 G ->
  v1 = v2.
Proof.
  intros.
  induction H.
  inversion H0; subst; auto.
  false.
  apply IHctx_binds.
  set (ctx_binds_push_inv  H0).
  destruct o.
  destruct H2.
  contradict H2; auto.
  destruct H2. assumption.
  apply ctx_binds_hide_inv in H0. auto.
Qed.

Lemma ctx_binds_middle_eq_inv x vk G1 G2 v1 v2:
  ctx_binds x v1 (G1 § x / vk ~ v2 §& G2) ->
  ctx_ok (G1 § x / vk ~ v2 §& G2) ->
  v1 = v2.
Proof.
  intros.
  set ctx_binds_middle_inv.
  specialize (o x x vk G1 G2 v1 v2 H).
  destruct o.
  set (ctx_binds_ok_concat_unbound_l).
  specialize (n x v1 _ G2 H1 H0).
  destruct n.
  apply ctx_in_dom_push.
  destruct H1.
  destruct H1.
  destruct H2.
  auto.
  destruct H1.
  destruct H2.
  contradict H2. auto.
Qed.

Lemma ctx_binds_middle_vkchange x1 x2 vk1 vk2 G1 G2 v1 v2:
  ctx_binds x1 v1 (G1 § x2 / vk1 ~ v2 §& G2) ->
  ctx_binds x1 v1 (G1 § x2 / vk2 ~ v2 §& G2 ).
Proof.
  intros.
  induction G2.
  + rewrite ctx_concat_empty_r in *; auto.
    eapply ctx_binds_push_vkchange; eauto.
  + rewrite ctx_concat_push in *; auto.
    inversion H; subst.
    - apply ctx_binds_e.
    - apply ctx_binds_n; auto.
  + rewrite ctx_concat_stop in *; auto.
    false.
    eapply ctx_binds_stop_inv; eauto.
  + rewrite ctx_concat_hide in *; auto.
    apply ctx_binds_hide.
    apply ctx_binds_hide_inv in H. eauto.
Qed.

Lemma ctx_binds_concat_hide_inv G1 G2 x a:
  ctx_binds x a ((ctx_hide G1) §& G2) ->
  ctx_binds x a (G1 §& G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten in *.
  all: inversion H; subst; auto.
Qed.

Lemma ctx_binds_concat_hide G1 G2 x a:
  ctx_binds x a (G1 §& G2) ->
  ctx_binds x a ((ctx_hide G1) §& G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten in *.
  all: inversion H; subst; auto.
Qed.

Lemma ctx_binds_concat_map_l G1 G2 x T f:
  ctx_ok (G1 §& G2) ->
  ctx_binds x T (G1 §& G2) ->
  ctx_binds x T (G1) ->
  ctx_binds x T (G1 §& ctx_map f G2).
Proof.
  introv Hok HbiG12 HbiG1.
  induction G2; simpls; auto.
  - inversions HbiG12. inversions Hok. eapply ctx_binds_dom in HbiG1. rewrite ctx_dom_concat in H4. rewrite notin_union in H4. destruct H4. contradiction.
    eapply ctx_binds_n; eauto.
  - inversions HbiG12.
  - inversions HbiG12. eauto.
Qed.

Lemma ctx_binds_concat_map_r G1 G2 x T f:
  ctx_binds x T (G2) ->
  ctx_binds x (f T) (G1 §& ctx_map f G2).
Proof.
  intros.
  eapply ctx_binds_concat_right.
  eapply ctx_binds_map.
  auto.
Qed.

(* If a variable is not free in a context, then it is not free in any type in the context *)
Lemma ctx_binds_notin_fv G y w T:
  ctx_binds y T G ->
  w \notin ctx_fv G ->
  w \notin fv T.
Proof.
  intros.
  intro.
  induction H.
  + ctx_simpl_fv.
  apply H0. rewrite in_union. auto.
  + ctx_simpl_fv. auto.
  + auto.
Qed.

Lemma ctx_binds_fv_subset G y T:
  ctx_binds y T G ->
  fv T \c ctx_fv G.
Proof.
  intros.
  intro.
  induction H.
  + ctx_simpl_fv.
    rewrite in_union. auto.
  + intro. ctx_simpl_fv. rewrite in_union.  auto.
  + auto.
Qed.


(** ** Binds-after
   If ctx_binds_k_after v vk t G G' then v is bound in G and G'
   is the longest initial sequence of G which v is unbound in. *)

Inductive ctx_binds_k_after: var -> varkind -> typ -> ctx -> ctx -> Prop :=
| ctx_binds_k_e(v: var)(vk: varkind)(t: typ)(G: ctx):
    ctx_binds_k_after v vk t (ctx_push G v vk t) G
| ctx_binds_k_n(v v2: var)(vk1 vk2: varkind)(t t2: typ)(G1 G2: ctx):
    v <> v2 ->
    ctx_binds_k_after v vk1 t G1 G2 ->
    ctx_binds_k_after v vk1 t (ctx_push G1 v2 vk2 t2) G2
.

Hint Resolve ctx_binds_k_e ctx_binds_k_n : core.

Lemma ctx_binds_k_after_binds v vk t G1 G2:
  ctx_binds_k_after v vk t G1 G2 ->
  ctx_binds v t G1.
Proof.
  intros.
  induction H; auto.
Qed.

Lemma ctx_binds_k_after_unbound_r v vk t G1 G2:
  ctx_ok G1 ->
  ctx_binds_k_after v vk t G1 G2 ->
  v §# G2.
Proof.
  intros.
  induction H0.
  - eapply ctx_ok_push_inv_unbound; eauto.
  - apply IHctx_binds_k_after.
    eapply ctx_ok_push_inv_ok; eauto.
Qed.

Lemma ctx_binds_k_after_concat v vk t G1 G2 G':
  ctx_binds_k_after v vk t G1 G' ->
  ctx_ok (G1 §& G2) ->
  ctx_weakens G2 ->
  ctx_binds_k_after v vk t (G1 §& G2) G'.
Proof.
  introv Hbin Hok.
  induction G2.
  - simpl. eauto.
  - intro.
    rewrite ctx_concat_push.
    rewrite ctx_concat_push in Hok.
    apply ctx_binds_k_n.
    intro; subst.
    assert (v0 §# (G1 §& G2)).
    eapply ctx_ok_push_inv_unbound. apply Hok.
    assert (ctx_binds v0 t G1).
    eapply ctx_binds_k_after_binds; eauto.
    eapply ctx_binds_unbound_false; eauto.
    apply IHG2.
    eapply ctx_ok_push_inv_ok; eauto.
    eapply ctx_weakens_push_inv; eauto.
  - intro. inversion H.
  - intro. inversion H.
Qed.

Lemma ctx_binds_k_after_eq v vk1 vk2 t1 t2 G G1 G2:
  ctx_binds_k_after v vk1 t1 G G1 ->
  ctx_binds_k_after v vk2 t2 G G2 ->
  vk1 = vk2 /\ t1 = t2 /\ G1 = G2.
Proof.
  introv H1 H2.
  induction H1.
  - inversion H2; subst.
    + auto.
    + contradiction.
  - inversion H2; subst.
    + contradiction.
    + auto.
Qed.

Lemma ctx_binds_k_after_ok : forall v vk T G1 G0,
    ctx_ok G1 ->
    (* G₀ is an initial sequence of G₁ *)
    ctx_binds_k_after v vk T G1 G0 ->
    (* and so G₀ must be well-formed as well *)
    ctx_ok G0.
Proof.
  introv Hok Hbk.
  induction Hbk; eauto.
Qed.

Lemma ctx_binds_k_after_weakens v vk t G1 G2:
  ctx_binds_k_after v vk t G1 G2 ->
  ctx_weakens G1 ->
  ctx_weakens G2.
Proof.
  introv Hbi Hwk.
  induction Hbi; inversions Hwk; auto.
Qed.

Lemma ctx_binds_k_after_initial_concat v vk t G1 G2:
  ctx_binds_k_after v vk t G1 G2 ->
  exists G3,
  G1 = (G2 §& G3).
Proof.
  introv Hbi.
  induction Hbi.
  - exists (v / vk ~ t). ctx_flatten. auto.
  - destructc IHHbi as G3 ?.
    subst.
    exists (G3 § v2 / vk2 ~ t2).
    ctx_flatten. auto.
Qed.

Lemma ctx_binds_k_after_initial_binds v vk t G1 G2 v2 t2:
  ctx_weakens G1 ->
  ctx_ok G1 ->
  ctx_binds_k_after v vk t G1 G2 ->
  ctx_binds v2 t2 G2 ->
  ctx_binds v2 t2 G1.
Proof.
  introv Hwk Hok Hbik Hbi.
  destructc (ctx_binds_k_after_initial_concat Hbik) as G3 ?.
  subst.
  eapply ctx_binds_concat_left.
  auto.
  eapply ctx_weakens_concat_right; eauto.
  eapply ctx_binds_ok_concat_unbound_r; eauto.
Qed.

(** ** Domain by variable kind *)
(* a set of variables with a given varkind *)

Fixpoint ctx_kdom(vk: varkind)(c: ctx): fset var :=
match c with
| ctx_empty => \{}
| ctx_push c v vk2 t => If (vk = vk2) then (\{ v } \u (ctx_kdom vk c)) else (ctx_kdom vk c)
| ctx_stop c => ctx_kdom vk c
| ctx_hide c => ctx_kdom vk c
end.

Lemma ctx_kdom_empty vk:
  ctx_kdom vk ctx_empty = \{}.
Proof.
  auto.
Qed.

Lemma ctx_kdom_push vk1 c v vk2 t:
  ctx_kdom vk1 (ctx_push c v vk2 t) =
  If (vk1 = vk2) then (\{ v } \u (ctx_kdom vk1 c)) else (ctx_kdom vk1 c)
.
Proof.
  auto.
Qed.

Lemma ctx_kdom_push_e c v vk t:
  ctx_kdom vk (ctx_push c v vk t) =
  \{ v } \u (ctx_kdom vk c).
Proof.
  pose proof (ctx_kdom_push vk c v vk t).
  destruct (classicT (vk=vk)).
  auto.
  contradiction.
Qed.
Lemma ctx_kdom_push_n vk1 c v vk2 t:
  vk1 <> vk2 ->
  ctx_kdom vk1 (ctx_push c v vk2 t) = (ctx_kdom vk1 c).
Proof.
  pose proof (ctx_kdom_push vk1 c v vk2 t).
  destruct (classicT (vk1=vk2)).
  contradiction.
  auto.
Qed.

Lemma ctx_in_kdom_push c v vk t:
  v \in ctx_kdom vk (ctx_push c v vk t).
Proof.
  rewrite ctx_kdom_push_e.
  rewrite in_union.
  left. apply in_singleton_self.
Qed.

Lemma ctx_kdom_stop vk c:
  ctx_kdom vk (ctx_stop c) = ctx_kdom vk c.
Proof.
  auto.
Qed.

Lemma ctx_kdom_hide vk c:
  ctx_kdom vk (ctx_hide c) = ctx_kdom vk c.
Proof.
  auto.
Qed.

Hint Rewrite ctx_kdom_empty ctx_kdom_push ctx_kdom_stop ctx_kdom_hide: ctx_flatten.

Lemma ctx_kdom_concat vk G1 G2:
  ctx_kdom vk (G1 §& G2) =
  ctx_kdom vk G1 \u ctx_kdom vk G2.
Proof.
  induction G2.
  all: ctx_flatten.
  - rewrite union_empty_r.
    apply eq_refl.
  - rewrite IHG2.
    destruct (classicT (vk=v0)).
    auto. auto.
  - rewrite IHG2.
    auto.
  - rewrite IHG2.
    auto.
Qed.

Lemma ctx_kdom_concat_r vk G1 G2 G3:
  ctx_kdom vk G2 = ctx_kdom vk G3 ->
  ctx_kdom vk (G1 §& G2) = ctx_kdom vk (G1 §& G3).
Proof.
  intros.
  rewrite ctx_kdom_concat.
  rewrite ctx_kdom_concat.
  rewrite H.
  auto.
Qed.

Lemma ctx_kdom_map f vk G1:
  ctx_kdom vk (ctx_map f G1) = ctx_kdom vk G1.
Proof.
  induction G1.
  all: ctx_flatten.
  all: try rewrite IHG1; auto.
Qed.

Lemma ctx_kdom_concat_map_r f vk G1 G2:
  ctx_kdom vk (G1 §& G2) = ctx_kdom vk (G1 §& ctx_map f G2).
Proof.
  apply ctx_kdom_concat_r.
  apply eq_sym.
  rewrite ctx_kdom_map.
  auto.
Qed.

Ltac ctx_simpl_kdom :=
  rewrite_all ctx_kdom_map in *;
  rewrite_all ctx_kdom_concat in *;
  rewrite_all ctx_kdom_push_e in *;
  rewrite_all ctx_kdom_stop in *;
  rewrite_all ctx_kdom_empty in *.


Lemma ctx_notin_dom_kdom x G vk:
  x §# G -> x \notin ctx_kdom vk G.
Proof.
  intros.
  induction G.
  all: auto.
  - ctx_simpl_dom.
    rewrite ctx_kdom_push.
    destruct (classicT (vk=v0)); auto.
Qed.

Lemma ctx_in_kdom_dom x G vk:
  x \in ctx_kdom vk G -> x \in ctx_dom G.
Proof.
  intros.
  induction G.
  - rewrite ctx_dom_empty. auto.
  - rewrite ctx_dom_push. rewrite ctx_kdom_push in H.
    destruct (classicT (vk=v0)); subst;
    rewrite in_union in *. destruct H; auto. right. auto.
  - rewrite ctx_dom_stop. auto.
  - rewrite ctx_dom_hide. auto.
Qed.

Lemma ctx_binds_kdom G G1 v vk t:
  ctx_binds_k_after v vk t G G1 ->
  (v \in (ctx_kdom vk G)).
Proof.
  intros.
  induction G.
  - inversion H.
  - induction H.
    + apply ctx_in_kdom_push.
    + rewrite ctx_kdom_push.
      destruct (classicT (vk1=vk2)); subst.
      * rewrite in_union. right. fold ctx_dom. auto.
      * auto.
  - inversion H.
  - inversion H.
Qed.

Lemma ctx_kdom_to_binds G v vk:
  ctx_ok G ->
  ctx_weakens G ->
  (v \in (ctx_kdom vk G)) ->
  exists t,
  ctx_binds v t G
.
Proof.
  intros.
  apply ctx_dom_to_binds; auto.
  eapply ctx_in_kdom_dom; eauto.
Qed.

Lemma ctx_dom_kdom G a:
  a \in ctx_dom G ->
  exists vk, a \in ctx_kdom vk G.
Proof.
  introv Hdom.
  induction G; simpls.
  - rewrite in_empty in Hdom. false.
  - rewrite in_union in Hdom.
    destruct Hdom.
    rewrite in_singleton in H; subst.
    + exists v0. cases_if. rewrite in_union, in_singleton. auto.
    + forwards Hvk: IHG; fweauto. destructecs Hvk.
      exists vk; cases_if; rewrite ? in_union; auto.
  - auto.
  - auto.
Qed.

(* VARKIND *)
Definition ctx_k G v vk:=
  exists t G2, ctx_binds_k_after v vk t G G2.

Lemma ctx_k_eq G v vk1 vk2:
  ctx_k G v vk1 ->
  ctx_k G v vk2 ->
  vk1 = vk2.
Proof.
  introv Hvk1 Hvk2.
  unfold ctx_k in Hvk1, Hvk2.
  destruct Hvk1 as [t1 [G1 Hba1]].
  destruct Hvk2 as [t2 [G2 Hba2]].
  destruct (ctx_binds_k_after_eq Hba1 Hba2).
  auto.
Qed.

Lemma ctx_k_ne_false G v vk1 vk2:
  vk1 <> vk2 ->
  ctx_k G v vk1 ->
  ctx_k G v vk2 ->
  False.
Proof.
  introv Hne Hvk1 Hvk2.
  unfold ctx_k in Hvk1, Hvk2.
  destruct Hvk1 as [t1 [G1 Hba1]].
  destruct Hvk2 as [t2 [G2 Hba2]].
  destruct (ctx_binds_k_after_eq Hba1 Hba2).
  contradiction.
Qed.

Lemma ctx_binds_k_after_k v vk t G1 G2:
  ctx_binds_k_after v vk t G1 G2 ->
  ctx_k G1 v vk.
Proof.
  intros.
  unfold ctx_k. eauto.
Qed.

Lemma ctx_k_in_kdom G v vk:
  ctx_k G v vk ->
  v \in ctx_kdom vk G.
Proof.
  intros.
  inversion H.
  destruct H0.
  apply (ctx_binds_kdom H0).
Qed.


Lemma ctx_in_kdom_k G v vk:
  ctx_ok G ->
  ctx_weakens G ->
  v \in ctx_kdom vk G ->
  ctx_k G v vk.
Proof.
  intros.
  unfold ctx_k.
  induction G; ctx_simpl_kdom.
  - rewrite in_empty in H1. contradiction.
  - rewrite ctx_kdom_push in H1. cases_if; destruct (classicT(v=v0)); subst.
    + exists t G. auto.
    + rewrite in_union in H1. destruct H1. rewrite in_singleton in H1. contradiction.
      assert (exists t0 G2, ctx_binds_k_after v v1 t0 G G2). eapply IHG; eauto. inversions H0. auto.
      destructc H2 as ? ? ?. exists x x0. constructor. auto. auto.
    + inversions H. eapply ctx_in_kdom_dom in H1. contradiction.
    + assert (exists t0 G2, ctx_binds_k_after v vk t0 G G2). inversions H0. eapply IHG; auto.
      destructc H2 as ? ? ?. exists x x0. constructor. auto. auto.
  - inversions H0.
  - inversions H0.
Qed.

Lemma ctx_k_in_dom G v vk:
  ctx_k G v vk ->
  v \in ctx_dom G.
Proof.
  intros.
  eauto using ctx_k_in_kdom, ctx_in_kdom_dom.
Qed.

Lemma ctx_k_push_n G y vk1 T v vk2:
  ctx_k G v vk2 ->
  y §# G ->
  ctx_k (G § y / vk1 ~ T) v vk2.
Proof.
  intros.
  destruct H as [T2 [G2 ?]].
    unfold ctx_k in *.
    exists T2 G2.
    eapply ctx_binds_k_n; auto.
  apply ctx_binds_kdom in H.
  intro. subst.
  eapply ctx_notin_dom_kdom in H0. eauto.
Qed.
Lemma ctx_k_push_nv G y vk1 T v vk2:
  ctx_k G v vk2 ->
  y <> v ->
  ctx_k (G § y / vk1 ~ T) v vk2.
Proof.
  intros.
  destruct H as [T2 [G2 ?]].
    unfold ctx_k in *.
    exists T2 G2.
    eapply ctx_binds_k_n; auto.
Qed.


Lemma ctx_k_push_e G y vk1 T:
  ctx_k (G § y / vk1 ~ T) y vk1.
Proof.
  unfold ctx_k.
  exists T G.
  apply ctx_binds_k_e.
Qed.
Lemma ctx_k_push_inv G x1 x2 vk1 vk2 T:
  ctx_k (G § x1 / vk1 ~ T) x2 vk2 ->
  (x1 = x2 /\ vk1 = vk2) \/ (x1 <> x2 /\ ctx_k G x2 vk2).
Proof.
  introv Hk.
  inversions Hk. destruct_all. inversions H. eauto. right. split. eauto.  econstructor. exists x0; eauto.
Qed.

Lemma ctx_k_hide_inv G x vk:
  ctx_k (ctx_hide G) x vk ->
  False.
Proof.
  introv Hk.
  inversions Hk. destruct_all.
  inversion H.
Qed.

Lemma ctx_kdom_disjoint G vk1 vk2:
  ctx_ok G ->
  ctx_weakens G ->
  vk1 <> vk2 ->
  disjoint (ctx_kdom vk1 G) (ctx_kdom vk2 G).
Proof.
  introv Hok Hwk Hvk.
  eapply in_notin_disjoint.
  intro. intro. intro.
  eapply ctx_in_kdom_k in H; eauto.
  eapply ctx_in_kdom_k in H0; eauto.
  eapply ctx_k_ne_false; eauto.
Qed.

Lemma ctx_dom_k G a:
  ctx_weakens G ->
  ctx_ok G ->
  a \in ctx_dom G ->
  exists vk, ctx_k G a vk.
Proof.
  introv Hwk Hok Hin.
  forwards: ctx_dom_kdom; fweauto.
  destructecs H.
  exists vk. eapply ctx_in_kdom_k; auto.
Qed.

(** ** Substitution in typing context *)

Notation subst_ctx x y := (ctx_map (subst_var x y)).
Notation "G '/[§' x '->' y ']'" :=
  (subst_ctx x y G) (x at level 69, at level 7, left associativity) : syntax_scope.
Notation subst_ctx_vars xs ys := (ctx_map (subst_vars xs ys)).
Notation "G '/[§' xs '-->' ys ']'" :=
  (subst_ctx_vars xs ys G) (xs at level 69, at level 7, left associativity) : syntax_scope.
Notation subst_ctx_pvars xs ys := (ctx_map (subst_pvars xs ys)).
Notation "G '//[§' xs '-->' ys ']'" :=
  (subst_ctx_pvars xs ys G) (xs at level 69, at level 7, left associativity) : syntax_scope.

Lemma notin_subst_ctx_fv_self x y G:
  x <> y -> x \notin ctx_fv (G /[§ x -> y ]).
Proof.
  introv Hneq. induction G; simpl; eauto.
  rewrite notin_union; split; eauto. eapply subst_notfv. auto.
Qed.

Lemma ctx_fv_subst_var_subset G x y :
  ctx_fv (subst_ctx x y G)  \c (fv G) \u \{y}.
Proof.
  induction G; simpl; unfold CtxFreeVar; eauto using subset_empty_l.
  rewrite ctx_fv_push. introv Hin.
  repeat(rewrite in_union in *). destruct Hin as [Hin | Hin].
  - apply fv_subst_var_subset in Hin.
    repeat(rewrite in_union in *). destruct_all; tauto.
  - apply IHG in Hin.
    repeat(rewrite in_union in *). destruct_all; tauto.
Qed.

Lemma ctx_subst_empty x y:
  subst_ctx x y ctx_empty = ctx_empty.
Proof.
  auto.
Qed.

Lemma ctx_subst_push x y G v vk t:
  subst_ctx x y (G § v / vk ~ t) =
  (subst_ctx x y G) § (v) / vk ~ (subst_var x y  t).
Proof.
  auto.
Qed.
Lemma ctx_subst_stop x y G:
  subst_ctx x y (ctx_stop G) =
  ctx_stop (subst_ctx x y G).
Proof.
  auto.
Qed.
Lemma ctx_subst_hide x y G:
  subst_ctx x y (ctx_hide G) =
  ctx_hide (subst_ctx x y G).
Proof.
  auto.
Qed.


Lemma ctx_subst_single x y v vk t:
  (ctx_empty § v / vk ~ t) /[§ x -> y ] =
  ctx_empty § v / vk ~ (t /[ x -> y ]).
Proof.
  intros.
  rewrite <- ctx_subst_empty with (x:=x) (y:=y) at 2.
  rewrite <- ctx_subst_push.
  apply eq_refl.
Qed.

Lemma ctx_binds_subst x y v t G:
  ctx_binds v t G ->
  ctx_binds v (t /[ x -> y ]) (G /[§ x -> y ]).
Proof.
  intros.
  induction H.
  - rewrite ctx_subst_push; auto.
  - rewrite ctx_subst_push; auto.
  - rewrite ctx_subst_hide; auto.
Qed.

Lemma subst_fresh_ctx x y G:
  x \notin ctx_fv G ->
  G /[§ x -> y ] = G.
Proof.
  induction G; intros.
  - unfold subst_ctx. auto.
  - rewrite ctx_fv_push, notin_union in H.
    rewrite ctx_subst_push.
    destruct H.
    rewrite subst_fresh by auto.
    rewrite IHG by auto.
    auto.
  - rewrite ctx_fv_stop in H.
    rewrite ctx_subst_stop.
    rewrite IHG by auto.
    auto.
  - rewrite ctx_fv_hide in H.
    rewrite ctx_subst_hide.
    rewrite IHG by auto.
    auto.
Qed.

(** ** Closed/sealed context *)

Fixpoint ctx_seal_rec n x G :=
  match G with
  | ctx_empty => ctx_empty
  | ctx_stop G => ctx_stop (ctx_seal_rec n x G)
  | ctx_hide G => ctx_hide (ctx_seal_rec n x G)
  | ctx_push G y vk t => ctx_push (ctx_seal_rec n x G) y vk (seal_rec n x t)
  end.

Instance CtxSealable: Sealable ctx := ctx_seal_rec.

Lemma ctx_seal_rec_binds G x T n y:
  ctx_binds x T G ->
  ctx_binds x (seal_rec n y T) (ctx_seal_rec n y G).
Proof.
  introv Hbi.
  induction G.
  - inversion Hbi.
  - inversions Hbi. simpls. auto.
    eapply ctx_binds_n; eauto.
  - inversion Hbi.
  - inversions Hbi.
    constructor.
    eauto.
Qed.

Lemma ctx_seal_map n y G:
 seal_rec n y G = ctx_map (seal_rec n y) G.
Proof.
  induction G; simpls; try rewrite <- IHG; eauto.
Qed.

Lemma ctx_sealed_empty:
  sealed ctx_empty.
Proof.
  unfold sealed. eauto.
Qed.
Hint Resolve ctx_sealed_empty: core.

Lemma ctx_sealed_hide G:
  sealed G ->
  sealed (ctx_hide G).
Proof.
  unfold sealed. simpl.
  intros.
  eapply f_equal.
  eauto.
Qed.
Hint Resolve ctx_sealed_hide: core.

Lemma ctx_sealed_stop G:
  sealed G ->
  sealed (ctx_stop G).
Proof.
  unfold sealed. simpl.
  intros.
  eapply f_equal.
  eauto.
Qed.
Hint Resolve ctx_sealed_stop: core.

Lemma ctx_sealed_push G x vk T:
    sealed G ->
    sealed T ->
    sealed (G § x/vk ~T).
Proof.
  unfold sealed. simpl.
  intros.
  repeat eapply fun_arg_eq_1; eauto.
Qed.
Hint Resolve ctx_sealed_push: core.

Lemma ctx_closed_hide_G G:
  sealed (ctx_hide G) ->
  sealed G.
Proof.
  unfold sealed. simpl.
  intros.
  specialize (H x).
  injection H; eauto.
Qed.

Lemma ctx_closed_stop_G G:
  sealed (ctx_stop G) ->
  sealed G.
Proof.
  unfold sealed. simpl.
  intros.
  specialize (H x).
  injection H; eauto.
Qed.

Lemma ctx_closed_push_T G x vk T:
    sealed (G § x/vk ~T) ->
    sealed T.
Proof.
  unfold sealed. simpl.
  intros.
  specialize (H x0).
  injection H; eauto.
Qed.
Lemma ctx_closed_push_G G x vk T:
    sealed (G § x/vk ~T) ->
    sealed G.
Proof.
  unfold sealed. simpl.
  intros.
  specialize (H x0).
  injection H; eauto.
Qed.

Lemma ctx_closed_binds G x T:
  sealed G ->
  ctx_binds x T G ->
  sealed T.
Proof.
  introv Hcl Hbi.
  induction Hbi.
  - eapply ctx_closed_push_T; eauto.
  - eapply ctx_closed_push_G in Hcl; eauto.
  - eapply ctx_closed_hide_G in Hcl; eauto.
Qed.
Lemma ctx_closed_binds_closed G x T:
  sealed G ->
  ctx_binds x T G ->
  closed T.
Proof.
  introv Hsl Hbi.
  forwards: ctx_closed_binds; eauto.
  replace T with (seal x T).
  eapply sealed_closed; eauto.
  eapply H.
Qed.
Lemma ctx_seal_push y G x vk T:
  seal y (G § x/vk ~ T) = (seal y G) § x/vk ~ (seal y T).
Proof.
  compute. eauto.
Qed.

Lemma ctx_seal_concat x G1 G2 :
  seal x (G1 §& G2) = seal x G1 §& seal x G2.
Proof.
  induction G2; eauto;
    try solve [simpl; unfold seal in *; rewrite IHG2; auto].
Qed.

Lemma ctx_sealed_concat G1 G2 :
  sealed (G1 §& G2) <-> (sealed G1 /\ sealed G2).
Proof.
  repeat(split).
  1,2:
    induction G2; simpl in *;
  eauto using ctx_closed_push_G, ctx_closed_hide_G, ctx_closed_stop_G,
    ctx_closed_push_T.
  intros [Hs1 Hs2]. intros x. specialize (Hs1 x). specialize (Hs2 x).
  rewrite ctx_seal_concat. congruence.
Qed.

Lemma ctx_seal_dom G n x:
  ctx_dom G =
  ctx_dom (seal_rec n x G).
Proof.
  induction G; simpl; try rewrite IHG; unfold seal_rec; eapply fset_extens; try union_sub_solver.
Qed.

Lemma ctx_seal_ok G n x:
  ctx_ok G ->
  ctx_ok (seal_rec n x G).
Proof.
  introv Hok.
  induction Hok; simpl; eauto.
  - econstructor. eapply IHHok.
    erewrite <- ctx_seal_dom. auto.
Qed.

(* OTHER *)

Lemma ctx_binds_weaken_hide (x: var) (a: typ) (G1 G2 G3: ctx):
  ctx_binds x a (ctx_hide G1 §& G3) ->
  ctx_weakens G2 ->
  ctx_ok (ctx_hide (G1 §& G2)) ->
  ctx_binds x a (ctx_hide (G1 §& G2) §& G3).
Proof.
  introv Bi Hwk Hok.

  induction G3.
  - rewrite ctx_concat_empty_r.
    rewrite ctx_concat_empty_r in Bi.
    inversions Bi. constructor.
    apply ctx_binds_concat_left; auto.
    inversions Hok.
    eapply ctx_binds_ok_concat_unbound_r;
    eauto.
  - rewrite ctx_concat_push.
    inversion Bi; subst.
    + apply ctx_binds_e.
    + apply ctx_binds_n; auto.
  - rewrite ctx_concat_stop.
    inversion Bi.
  - rewrite ctx_concat_hide.
    rewrite ctx_concat_hide in Bi.
    apply ctx_binds_hide.
    apply ctx_binds_hide_inv in Bi.
    auto.
Qed.

Lemma ctx_ok_concat_map G1 G2 (f: typ -> typ):
  ctx_ok (G1 §& G2) -> ctx_ok (G1 §& ctx_map f G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: ctx_flatten in H.
  all: auto.
  - apply ctx_ok_push.
    + auto.
    + rewrite <- (ctx_dom_concat_map_r f).
      eapply ctx_ok_push_inv_unbound; eauto.
Qed.
Hint Resolve ctx_ok_concat_map: core.

Lemma ctx_ok_map G (f: typ -> typ):
  ctx_ok G -> ctx_ok (ctx_map f G).
Proof.
  introv HokG.
  forwards: ctx_ok_concat_map ctx_empty G.
  rewrite ctx_concat_empty_l. eauto.
  rewrite ctx_concat_empty_l in H. eauto.
Qed.

Lemma ctx_ok_concat_map_inv G1 G2 (f: typ -> typ):
  ctx_ok (G1 §& ctx_map f G2) -> ctx_ok (G1 §& G2).
Proof.
  intros.
  induction G2.
  all: ctx_flatten.
  all: ctx_flatten in H.
  all: auto.
  - apply ctx_ok_push.
    + auto.
    + inversions H. rewrite <- (ctx_dom_concat_map_r f) in H5.
      auto.
Qed.

Lemma ctx_ok_map_inv G (f: typ -> typ):
  ctx_ok (ctx_map f G) -> ctx_ok G.
Proof.
  introv HokG.
  forwards: ctx_ok_concat_map_inv ctx_empty G.
  rewrite ctx_concat_empty_l. eauto.
  rewrite ctx_concat_empty_l in H. eauto.
Qed.

Ltac ctx_binds_eq :=
  match goal with
  | [Hb1: ctx_binds ?x _ ?G,
     Hb2: ctx_binds ?x _ ?G |- _] =>
     apply (ctx_binds_functional Hb1) in Hb2; inversions Hb2
  end.


(** ** Visibility *)
(* a variable is visible in the contex, if after it
   there is no ctx_stop or ctx_hide *)

Inductive ctx_vis: ctx -> var -> Prop :=
| ctx_vis_e(v: var)(vk: varkind)(t: typ)(G: ctx):
    ctx_vis (ctx_push G v vk t) v
| ctx_vis_n(v v2: var)(vk: varkind)(t: typ)(G: ctx):
    v <> v2 ->
    ctx_vis G v ->
    ctx_vis (ctx_push G v2 vk t) v
.

Hint Resolve ctx_vis_e ctx_vis_n: core.

Lemma ctx_vis_push_n_inv G v2 vk t v:
  ctx_vis (ctx_push G v2 vk t) v ->
  v <> v2 ->
  ctx_vis G v.
Proof.
  intros.
  inversions H.
  - contradiction.
  - auto.
Qed.

Lemma ctx_vis_empty_inv v:
  ctx_vis ctx_empty v ->
  False.
Proof.
  intros. inversion H.
Qed.

Lemma ctx_vis_stop_inv v G:
  ctx_vis (ctx_stop G) v ->
  False.
Proof.
  intros. inversion H.
Qed.

Lemma ctx_vis_hide_inv v G:
  ctx_vis (ctx_hide G) v ->
  False.
Proof.
  intros. inversion H.
Qed.

Lemma ctx_binds_weakens_vis G v t:
  ctx_binds v t G ->
  ctx_weakens G ->
  ctx_vis G v.
Proof.
  intros.
  induction H.
  - auto.
  - apply ctx_weakens_push_inv in H0. auto.
  - inversion H0.
Qed.

Lemma ctx_vis_in_dom G v:
  ctx_vis G v ->
  v \in ctx_dom G.
Proof.
  intros.
  induction H.
  - apply ctx_in_dom_push.
  - ctx_simpl_dom. rewrite in_union. right. auto.
Qed.

Lemma ctx_vis_hide_dom G1 G2 v:
  ctx_vis (ctx_hide G1 §& G2) v ->
  v \in ctx_dom G2.
Proof.
  intros.
  remember (ctx_hide G1 §& G2) as G.
  revert dependent G2.
  induction H; intros.
  - destruct G2; inversions HeqG. apply ctx_in_dom_push.
  - destruct G2; inversions HeqG.
    specialize (IHctx_vis _ (eq_refl _)).
    ctx_simpl_dom. rewrite in_union. right. auto.
Qed.

Lemma ctx_push_concat_inv (G G1 G2 : ctx) (v : var) (vk : varkind) (t : typ) :
  G § v / vk ~ t = G1 §& G2 ->
  (G2 = ctx_empty /\ G § v / vk ~ t = G1) \/
  (exists G2', G2 = G2' § v / vk ~ t).
Proof.
  introv Heq. induction G2; eauto; try solve[inversions Heq].
  simpl in Heq. inversions Heq; eauto.
Qed.

Lemma weaken_vis_push (G1 G2 : ctx) (x : var) :
  ctx_weakens G2 ->
  ctx_vis G1 x ->
  ctx_vis (G1 §& G2) x.
Proof.
  introv Hwk Hvis. induction Hwk; simpl; eauto.
  destruct (var_decide v x); subst; eauto.
Qed.

Lemma ctx_seal_vis G x v:
  ctx_vis G x ->
  ctx_vis (seal v G) x.
Proof.
  introv Hvis.
  induction Hvis; simpls; constructor; eauto.
Qed.
Hint Resolve ctx_seal_vis: core.

Lemma ctx_vis_weakens G1 G2 x :
  ctx_vis G1 x -> ctx_weakens G2 -> ctx_vis (G1 §& G2) x.
Proof.
  introv Hvis Hwk. induction Hwk; simpl; eauto.
  destruct (fv_decide v \{x}).
  - rewrite in_singleton in m. subst. eauto.
  - rewrite notin_singleton in n. subst. eauto.
Qed.


Lemma ctx_binds_k_after_concat_inv G1 G2 G0 v vk T :
  ctx_binds_k_after v vk T (G1 §& G2) G0 ->
  (ctx_binds_k_after v vk T G1 G0)
  \/ (exists G02, (G0 = G1 §& G02) /\ (ctx_binds_k_after v vk T G2 G02)).
Proof.
  introv Hbk. remember (G1 §& G2) as G. reverts G1 G2.
  induction Hbk; intros; subst.
  - apply ctx_push_concat_inv in HeqG as HeqG'.
    destruct HeqG' as [[? ?] | [G2' ?]]; subst; eauto.
    simpl in HeqG. inversions HeqG. eauto.
  - apply ctx_push_concat_inv in HeqG as HeqG'.
    destruct HeqG' as [[? ?] | [G2' ?]]; subst; eauto.
    simpl in HeqG. inversions HeqG.
    specialize (IHHbk _ _ eq_refl).
    destruct IHHbk as [? | [G02 [? IH]]]; eauto.
Qed.

Lemma ctx_binds_k_after_concat_map v vk T G1 G2 G02 f :
  ctx_binds_k_after v vk T G2 G02 ->
  ctx_binds_k_after v vk (f T) (G1 §& ctx_map f G2) (G1 §& ctx_map f G02).
Proof. introv Hbk. induction Hbk; simpl in *; eauto. Qed.

Lemma ctx_vis_concat_inv G1 G2 x :
  ctx_vis (G1 §& G2) x ->
  ctx_vis G1 x \/ ctx_vis G2 x.
Proof.
  introv Hvis. remember (G1 §& G2) as G. reverts G1 G2.
  induction Hvis; intros; subst; eauto.
  - apply ctx_push_concat_inv in HeqG. destruct_all; subst; eauto.
  - apply ctx_push_concat_inv in HeqG as ?. destruct_all; subst; eauto.
    inversions HeqG. rename x into G2. specialize (IHHvis _ _ eq_refl).
    destruct_all; eauto.
Qed.

Lemma ctx_vis_concat_map G1 G2 x f :
  ctx_vis (G1 §& G2) x ->
  ctx_vis (G1 §& ctx_map f G2) x.
Proof.
  introv Hvis. remember (G1 §& G2) as G. reverts G1 G2.
  induction Hvis; intros; subst; simpl in *; eauto.
  - apply ctx_push_concat_inv in HeqG. destruct_all; subst; simpl in *; eauto.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct_all; subst; simpl in *; eauto. inversions HeqG. eauto.
Qed.

Lemma ctx_vis_map G x f :
  ctx_vis G x ->
  ctx_vis (ctx_map f G) x.
Proof.
  pose proof (@ctx_vis_concat_map ctx_empty G x f).
  introv Hvis. repeat(rewrite ctx_concat_empty_l in *). auto.
Qed.

Lemma ctx_vis_weakens_left G1 G2 x :
  ctx_vis G2 x ->
  ctx_vis (G1 §& G2) x.
Proof. introv Hvis. induction Hvis; intros; subst; simpl in *; eauto. Qed.

Lemma ctx_binds_k_after_initial_segment v vk T G G1 :
  ctx_binds_k_after v vk T G G1 ->
  exists G2, G = (G1 § v / vk ~ T §& G2).
Proof.
  introv Hbk. induction Hbk.
  - exists ctx_empty. rewrite ctx_concat_empty_r. auto.
  - destruct_all; subst. rename x into G3.
    exists (G3 § v2 / vk2 ~ t2). simpl. auto.
Qed.

Lemma ctx_weakens_concat_inv G1 G2 G1' G2' :
  ctx_weakens G2' ->
  G1 §& G2 = G1' §& G2' ->
  ((exists G', G1 = G1' §& G' /\ G2' = G' §& G2)
   \/ (exists G', G1' = G1 §& G' /\ G2 = G' §& G2')).
Proof.
  reverts G1 G2 G1'.
  induction G2'; introv Hwk Heq; simpl in *; inversions Hwk.
  - subst. eauto.
  - apply eq_sym, ctx_push_concat_inv in Heq as Heq'.
    destruct Heq' as [[? ?] | [G3 ?]]; subst; simpl in *.
    + inversions Heq. left. eexists; split.
      2: reflexivity. simpl. eauto.
    + inversions Heq. apply eq_sym in H1.
      specialize (IHG2' _ _ _ H0 H1). destruct_all; subst; eauto.
Qed.

Lemma ctx_kdom_subset G vk :
  ctx_ok G -> ctx_weakens G -> ctx_kdom vk G \c ctx_dom G.
Proof.
  introv Hok Hwk. hnf. introv Hin. apply ctx_kdom_to_binds in Hin; eauto.
  destruct_all. eauto using ctx_binds_dom.
Qed.

Lemma subst_ctx_concat G1 G2 x y :
  subst_ctx x y (G1 §& G2) = (subst_ctx x y G1) §& (subst_ctx x y G2).
Proof. induction G2; simpl in *; eauto; try (rewrite IHG2; eauto). Qed.

Lemma ctx_binds_to_ctx_binds_k_after x T G :
  ctx_weakens G ->
  ctx_binds x T G ->
  exists Gx vk, ctx_binds_k_after x vk T G Gx.
Proof.
  induction 2; eauto.
  - rename H into Hwk, v into x, v2 into x', t2 into T', t into T,
      H0 into Hne, H1 into Hbi, vk into vk'.
    inversions Hwk. rename H0 into Hwk. specialize (IHctx_binds Hwk).
    destruct IHctx_binds as [Gx [vk Hbk]]. eauto.
  - inversions H.
Qed.

Lemma ctx_binds_k_after_ctx_concat_map x vk T G2 G2' G1 f :
  ctx_binds_k_after x vk    T                   G2                   G2' ->
  ctx_binds_k_after x vk (f T) (G1 §& ctx_map f G2) (G1 §& ctx_map f G2').
Proof. induction 1; simpl in *; eauto. Qed.

Lemma ctx_weakens_to_hide (G : ctx) (_ : ctx_weakens G) :
ctx_weakens_hide G.
Proof. induction H; eauto. Qed.

(** ** Tactics for rewriting contexts *)

(* First, a set of rewrite tactics with regular syntax is defined to
   work around Coq rewrite syntax irregularity *)
Local Ltac autorewrite_inG SC := autorewrite with SC.
Local Ltac autorewrite_inS SC := autorewrite with SC in *.
Local Ltac autorewrite_inH H SC := autorewrite with SC in H.

Local Ltac rewritewith1_inG E i v := rewrite E with (i := v).
Local Ltac rewritewith1_inS E i v:= rewrite E with (i := v) in *.
Local Ltac rewritewith1_inH H E i v := rewrite E with (i := v) in H.

Local Ltac rewriteb_inG E := rewrite <- E.
Local Ltac rewriteb_inS E := rewrite <- E in *.
Local Ltac rewriteb_inH H E := rewrite <- E in H.

(* A particular context rewrite tactic can be executed in 3 modes:
   1) rewrite in goal, 2) rewrite everywhere, 3) rewrite in hypothesis
 *)
Ltac rewritetac_inG tac := tac autorewrite_inG rewritewith1_inG rewriteb_inG.
Ltac rewritetac_inS tac := tac autorewrite_inS rewritewith1_inS rewriteb_inS.
Ltac rewritetac_inH H tac := tac ltac:(autorewrite_inH H) ltac:(rewritewith1_inH H) ltac:(rewriteb_inH H).

(* Tactic to split a context into a concatenation of 2 contexts,
   where the second one starts with xFrom. *)
Ltac ctx_split_from_tac xFrom AW W1 Wb :=
  AW ident:(ctx_flatten);
  W1 ctx_concat_single ident:(x) xFrom;
  repeat (Wb ctx_concat_push).

Tactic Notation "ctx_split_from" constr(xFrom) := rewritetac_inG ltac:(ctx_split_from_tac xFrom).
Tactic Notation "ctx_split_from" constr(xFrom) "in" "*" := rewritetac_inS ltac:(ctx_split_from_tac xFrom).
Tactic Notation "ctx_split_from" constr(xFrom) "in" hyp(H) := rewritetac_inH H ltac:(ctx_split_from_tac xFrom).

(* Tactic to split a context into a concatenation of 3 contexts,
   where the second one starts with xFrom
   and the third one starts with xTo. *)
Ltac ctx_split_between_tac xFrom xTo AW W1 Wb :=
  AW ident:(ctx_flatten);
  W1 ctx_concat_single ident:(x) xFrom;
  W1 ctx_concat_single ident:(x) xTo;
  repeat (Wb ctx_concat_push).

Tactic Notation "ctx_split_between" constr(xFrom) constr(xTo) := rewritetac_inG ltac:(ctx_split_between_tac xFrom xTo).
Tactic Notation "ctx_split_between" constr(xFrom) constr(xTo) "in" "*" := rewritetac_inS ltac:(ctx_split_between_tac xFrom xTo).
Tactic Notation "ctx_split_between" constr(xFrom) constr(xTo) "in" hyp(H) := rewritetac_inH H ltac:(ctx_split_between_tac xFrom xTo).
