Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import AbstractSyntax.

(** * Lemmas about definisions and definition lists. *)

(** Conjunctive property of definition list.
   A property that holds for defs exactly if it holds for all definitions. *)
Class DefsConjunctive (P: defs -> Prop) (Q: def -> Prop) := {
  (* Constructors *)
  defs_conj_nil: (P defs_nil) ;
  defs_conj_cons ds d: (P ds) -> (Q d) -> (P (defs_cons ds d)) ;
  (* Inversion *)
  defs_conj_cons_def ds d: (P (defs_cons ds d)) -> (Q d) ;
  defs_conj_cons_inv ds d: (P (defs_cons ds d)) -> (P ds)
}.

(** [ds = ... /\ {a = t} /\ ...]  #<br>#
    [ds = ... /\ {a = t'} /\ ...] #<br>#
    [―――――――――――――――――――――――――] #<br>#
    [t = t'] *)
Lemma defs_has_inv: forall ds a t t',
    defs_has ds (def_fld a t) ->
    defs_has ds (def_fld a t') ->
    t = t'.
Proof.
  intros. unfold defs_has, label_of in *.
  inversions H. inversions H0.
  rewrite H1 in H2. inversions H2.
  reflexivity.
Qed.

(* Finds two hypotheses def_has with def_fld of the same field and substitutes the value *)
Ltac defs_has_fld_eq := (
  match goal with
  | Hb1: defs_has ?x (def_fld ?a _), Hb2: defs_has ?x (def_fld ?a _) |- _ => apply (defs_has_inv Hb1) in Hb2; inversions Hb2
  end).

Lemma defs_has_met_inv: forall ds a t t',
    defs_has ds (def_met a t) ->
    defs_has ds (def_met a t') ->
    t = t'.
Proof.
  intros. unfold defs_has, label_of in *.
  inversions H. inversions H0.
  rewrite H1 in H2. inversions H2.
  reflexivity.
Qed.

Lemma defs_has_n d ds f:
  defs_has ds f ->
  label_of d <> label_of f ->
  defs_has (defs_cons ds d) f.
Proof.
  intros.
  unfold defs_has, get_def. cases_if. eauto.
Qed.

Lemma defs_has_cons_inv ds d d2:
  defs_has (defs_cons ds d) d2 ->
  d = d2 \/ defs_has ds d2.
Proof.
  introv Hhas.
  inversions Hhas. cases_if. inversions H0. eauto.
  right. eauto.
Qed.
Lemma defs_has_cons ds d:
  defs_has (defs_cons ds d) d.
Proof.
  unfold defs_has. destruct d; compute; cases_if; auto.
Qed.

(* Simpler induction on defs_has *)
Lemma defs_has_ind  (d1: def) (ds : defs) (Hhas:defs_has ds d1)(P : defs -> Prop):
  (forall d : defs, P (defs_cons d d1)) ->
  (forall d : defs, P d -> forall d0 : def, label_of d0 <> label_of d1 -> P (defs_cons d d0)) ->
   P ds.
Proof.
  intros.
  induction ds.
  - inversions Hhas.
  - inversions Hhas.
    cases_if.
    inversions H2. eauto.
    eauto.
Qed.

Lemma defs_has_open ds d x:
  defs_has ds d ->
  defs_has (open x ds) (open x d).
Proof.
  introv Hhas.
  eapply (defs_has_ind Hhas); intros.
  - sympls. eapply defs_has_cons.
  - sympls. eapply defs_has_n; eauto.
    destruct d1; destruct d; sympls; auto.
Qed.


Lemma defs_has_conjunctive (P: defs -> Prop) (Q: def -> Prop) (PC: DefsConjunctive P Q) ds d:
  P ds ->
  defs_has ds d ->
  Q d.
Proof.
  introv HP Hhas.
  induction ds.
  - inversions Hhas.
  - destruct (defs_has_cons_inv Hhas).
    + subst.
      eapply defs_conj_cons_def. eauto.
    + eapply IHds. eapply defs_conj_cons_inv. eauto. eauto.
Qed.

Lemma defs_has_fv yd a y2:
  defs_has yd (def_fld a (avar_f y2)) ->
  y2 \in (fv yd).
Proof.
  intros Hhas.
  eapply (defs_has_ind Hhas); intros.
  - sympls. rewrite in_union. rewrite in_singleton; auto.
  - sympls. rewrite in_union; auto.
Qed.

(** [d1 isin ds]             #<br>#
    [label(d2) \notin ds]     #<br>#
    [―――――――――――――――――――――]  #<br>#
    [label(d1) <> label(d2)]  *)
Lemma defs_has_hasnt_neq: forall ds d1 d2,
  defs_has ds d1 ->
  defs_hasnt ds (label_of d2) ->
  label_of d1 <> label_of d2.
Proof.
  introv Hhas Hhasnt.
  unfold defs_has in Hhas.
  unfold defs_hasnt in Hhasnt.
  induction ds.
  - simpl in Hhas. inversion Hhas.
  - simpl in Hhasnt. simpl in Hhas. case_if; case_if.
    + inversions Hhas. assumption.
    + apply IHds; eauto.
Qed.