Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import Vars Types Context Envs.

(** * Runtime environment *)

(** The runtime environment [e] is a part of the machine configuration.
   It maps references [w], which are used in the focus term, stack, and in methods and functions,
   to locations [y] which are names of object on the heap and are used in fields

  A new reference is created each time a reference is needed for some location.
  That happens when a new object is created and when a field is read.

  The environment figures in type equivalence, where types are considered equivalent
  if they only differ in prefixes of type selections, and the prefixes respectively are
  references to or location of the same objects.

  The environment always maps each reference to exactly one location,
  but many properties of equivalence should work even if it was allowed to relate any variables.
*)

(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
Notation renv := (env var) (only parsing).
Notation renv_empty := (empty) (only parsing).
Notation renv_bind e v1 v2 := (e & v1 ~ v2) (only parsing).

(* When a variable does not appear in the domain and the range of the environment *)
Definition renv_notin (v: var) (e: renv) := v \notin (dom e \u fv_env e).
(* The environment is idempotent when its dom and fv are disjoint *)
Definition renv_idempotent (e: renv) := forall x, x \in dom e -> x \notin fv_env e.

Notation renv_binds v1 v2 e := (binds v1 v2 e) (only parsing).


Lemma renv_push_in_fv_r v1 v2 e:
  v2 \in fv_env (e & (v1 ~ v2)).
Proof.
  rewrite env_push_fv.
  rewrite in_union.
  right.
  unfold fv.
  unfold VarFreeVar.
  rewrite in_singleton; auto.
Qed.

Lemma renv_bind_in_fv v1 v2 e:
  renv_binds v1 v2 e ->
  v2 \in fv_env e.
Proof.
  intros.
  induction e using env_ind.
  - false.
    eapply binds_empty_inv; eauto.
  -
    destruct (binds_push_inv H) as [[? ?] | [? ?]]; subst.
    + apply renv_push_in_fv_r.
    + apply env_push_in_fv_l. auto.
Qed.

Lemma renv_notin_dom_fv x e:
  x \notin dom e ->
  x \notin fv_env e ->
  renv_notin x e.
Proof.
  intros.
  unfold renv_notin.
  auto.
Qed.

Hint Resolve renv_notin_dom_fv : core.

Lemma renv_idempotent_concat_middle_inv e1 e2 e3:
  renv_idempotent (e1 & e2 & e3) ->
  renv_idempotent (e1 & e3).
Proof.
  unfold renv_idempotent.
  intro Hid.
  intro x.
  intro Hin.
  simpl_dom.
  forwards: (Hid x).
  - clear Hid.
    repeat rewrite in_union in *. destruct Hin; auto.
  - rewrite ! env_concat_fv in *. rewrite env_concat_fv in H. repeat rewrite notin_union in *. destruct_all; auto.
Qed.
Lemma renv_idempotent_concat_inv e1 e2:
  renv_idempotent (e1 & e2) ->
  renv_idempotent (e1).
Proof.
  rewrite <- (concat_empty_r e1) at 2.
  rewrite <- (concat_empty_r (e1 & e2)).
  eapply renv_idempotent_concat_middle_inv.
Qed.
Lemma renv_idempotent_push e x y:
  renv_idempotent e ->
  x \notin fv_env e ->
  y # e ->
  x <> y ->
  renv_idempotent (e & x ~ y).
Proof.
  unfold renv_idempotent.
  intros.
  simpl_dom.
  rewrite fv_env_push.
  rewrite notin_union.

  rewrite in_union in H3. destruct H3.
  - rewrite in_singleton in H3; subst. splits; auto. unfold fv, VarFreeVar. rewrite notin_singleton. auto.
  - splits; auto. unfold fv, VarFreeVar. rewrite notin_singleton. intro. subst. eauto.
Qed.

Lemma fv_env_to_binds y (e : renv) :
  ok e -> y \in fv_env e -> exists w, binds w y e.
Proof.
  introv Hok Hyin. induction Hok.
  - rewrite fv_env_empty, in_empty in Hyin. tauto.
  - rewrite fv_env_push, in_union in Hyin. destruct Hyin as [Hyin | Hyin].
    + specialize (IHHok Hyin). destruct IHHok as [w Hbi].
      exists w. apply binds_push_neq; auto. hnf. intros; subst.
      eauto using binds_to_dom.
    + unfold fv, VarFreeVar in Hyin. rewrite in_singleton in Hyin; subst.
      exists x; eauto.
Qed.
