(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import TypingHelpers.

Require Import FreshChain.

(** Tactics for working with freshness chain. *)

(* Initializes an empty freshness chain *)
Ltac fresh_chain_init Fr := pose proof fresh_chain_is_nil as Fr.
(* Adds a single known set into a known chain, detecting its type *)
Ltac fresh_chain_add_set Fr S := (
  match type of S with
    | fset var => apply (@fresh_chain_is_set (fresh_element_set S)) in Fr
    | var => apply (@fresh_chain_is_set (fresh_element_var S)) in Fr
    | list var => apply (@fresh_chain_is_set (fresh_element_list S)) in Fr
    | ctx => apply (@fresh_chain_is_set (fresh_element_ctx S)) in Fr; [>| constructor]
    | renv => apply (@fresh_chain_is_set (fresh_element_renv S)) in Fr; [>| constructor]
    | heap => apply (@fresh_chain_is_set (fresh_element_heap S)) in Fr; [>| constructor]
    | _ => let z := type of S in fail 1 "Unknown type of set" S ":"  z
  end
).

(* Adds a single known var into a known chain, detecting its type *)
Ltac fresh_chain_add_var_basic Fr S := (
  match type of S with
    | var => apply (@fresh_chain_is_var (fresh_element_var S)) in Fr
  end
).
Ltac fresh_chain_add_var Fr S := (
  match type of S with
    | var => apply fresh_chain_add_var_weak with (fe := (fresh_element_var S)) in Fr; [>|simpl]
  end
).

(* Takes a union of sets and prepends it to a chain value *)
Ltac fresh_chain_set_chain S fc := (
  match S with 
  | ?A \u ?B => (let Afs := (fresh_chain_set_chain A fc) in let BAfs := (fresh_chain_set_chain B Afs) in BAfs)
  | ?A => constr:(fresh_chain_set (fresh_element_set A) fc)
  end
).

Ltac fresh_chain_ctx_chain FC FE G fc := (
  match G with 
  | ?B §& ?A => (let Bs := (fresh_chain_ctx_chain FC FE B fc) in (fresh_chain_ctx_chain FC FE A Bs))
(*  | ?B § ?x / _ ~ _ => (let Bs := (fresh_chain_ctx_chain FC0 FC FE B fc) in constr:(FC (fresh_element_var x) Bs))*) (* This case is disabled, because it loses information about the type of the variable *)
  | ctx_empty => fresh_chain_nil
  | ?A => constr:(FC (FE A) fc)
  end
).
Ltac fresh_chain_ctxdom_var_chain G fc := fresh_chain_ctx_chain fresh_chain_var fresh_element_ctxdom G fc.

Ltac fresh_chain_renv_chain FC FE G fc := (
  match G with 
  | ?B & ?A => (let Bs := (fresh_chain_renv_chain FC FE B fc) in (fresh_chain_renv_chain FC FE A Bs))
  | ?A => constr:(FC (FE A) fc)
  end
).
Ltac fresh_chain_renvdom_varset_chain G fc := fresh_chain_renv_chain fresh_chain_var fresh_element_renvdom G fc.

Ltac fresh_chain_heap_chain FC FE G fc := (
  match G with 
  | ?B & ?A => (let Bs := (fresh_chain_heap_chain FC FE B fc) in (fresh_chain_heap_chain FC FE A Bs))
  | ?A => constr:(FC (FE A) fc)
  end
).
Ltac fresh_chain_heapdom_varset_chain G fc := fresh_chain_heap_chain fresh_chain_var fresh_element_heapdom G fc.


(* Solves one set prepends in a chain goal *)
Ltac fresh_chain_set_solve_r := 
match goal with
| H: ?G |- ?G => exact H
| |- fresh_chain_is fresh_chain_nil => constructor
| |- fresh_chain_is (fresh_chain_set ?fe ?fc) => apply fresh_chain_is_set; [>|constructor]
end.
(* Solves all set prepends in a chain goal *)
Ltac fresh_chain_set_solve := repeat fresh_chain_set_solve_r.

(* Solves one set prepends in a chain goal *)
Ltac fresh_chain_set_solve_by_r Fr fc := 
match goal with
| |- fresh_chain_is fc => exact Fr
| |- fresh_chain_is fresh_chain_nil => constructor
| |- fresh_chain_is (fresh_chain_set ?fe ?fc) => apply fresh_chain_is_set; [>|constructor]
end.
(* Solves all set prepends in a chain goal *)
Ltac fresh_chain_set_solve_by Fr :=
match (type of Fr) with
  | fresh_chain_is ?fc => repeat (fresh_chain_set_solve_by_r Fr fc)
end.




(* Gathers variables and syntax elements and prepends them to a chain value *)
Ltac gather_fresh_chain fc0 :=
let rec find fc :=
  match goal with
  | G: ctx |- _ => check_add fc (fresh_element_ctx G)
  | e: renv |- _ => check_add fc (fresh_element_renv e)
  | Si: heap |- _ => check_add fc (fresh_element_heap Si)
  | x: var |- _ => check_add fc (fresh_element_var x)
  | x: fset var |- _ => check_add fc (fresh_element_set x)
  | t: trm |- _ => check_add fc (fresh_element_element t)
  | t: def |- _ => check_add fc (fresh_element_element t)
  | t: defs |- _ => check_add fc (fresh_element_element t)
  | t: def |- _ => check_add fc (fresh_element_element t)
  | T: typ |- _ => check_add fc (fresh_element_element T)
  | T: dec |- _ => check_add fc (fresh_element_element T)
  | T: stack |- _ => check_add fc (fresh_element_element T)
  | _ => fc
  end
with check_add fc S := 
    match fc with
    | context [S] => fail 1
    | _ => find (fresh_chain_set S fc)
    end
in find fc0.

(* Gathers variables and syntax elements and prepends them to a known chain as sets *)
Ltac fresh_chain_add_gather Fr :=
  match type of Fr with 
  | fresh_chain_is ?fc =>
    let fcg := gather_fresh_chain fc in
    let Fr1 := fresh Fr in 
    assert (fresh_chain_is fcg) as Fr1; [> fresh_chain_set_solve| clear Fr; rename Fr1 into Fr]
  end.

Ltac fresh_chain_get_set Fr :=
match type of Fr with
fresh_chain_is ?fc => constr: (fresh_chain_to_set fc)
end.

(* Take a known freshness hypothesis and incorporates it into a known chain, detecting the type *)
Ltac fresh_chain_add_fresh Fr Hfr:= (
  
  match (type of Hfr) with
  | fresh ?S ?L ?xs =>
    let Sb := beautify_fset S in 
    let fc := fresh_chain_set_chain Sb fresh_chain_nil in 
    let Fr1 := fresh "Fr" in
    forwards Fr1: fresh_chain_add_fresh_list S fc Hfr;  [> simpls; union_sub_solver | fresh_chain_set_solve_by Fr | clear Fr Hfr; rename Fr1 into Fr]
  | ?x \notin ?S => (* Handling the case where the set is excatly as in Fr *)
    match type of Fr with 
    fresh_chain_is ?fc =>
    match (eval simpl in (fresh_chain_to_set fc)) with
    | S =>
      forwards Fr1: fresh_chain_add_fresh_var S fc Hfr;  [> apply subset_refl | apply Fr | clear Fr Hfr; rename Fr1 into Fr]
    end
    end
  | ?x \notin ?S => (* In this case the set in Fr is ignored *)
    let Sb := beautify_fset S in 
    let fc := fresh_chain_set_chain Sb fresh_chain_nil in 
    let Fr1 := fresh "Fr" in
    forwards Fr1: fresh_chain_add_fresh_var S fc Hfr;  [> simpls; union_sub_solver | fresh_chain_set_solve_by Fr | clear Fr Hfr; rename Fr1 into Fr]
  | (not (@eq var ?x ?y)) =>
    apply fresh_chain_is_var with (fe:=(fresh_element_var y)) in Fr; [> | auto | constructor];
    apply fresh_chain_is_var with (fe:=(fresh_element_var x)) in Fr; [> | auto | constructor];[>clear Hfr|apply fresh_element_fresh_var_var;apply Hfr|constructor]
  | ctx_ok ?G =>
    let Fr1 := fresh "Fr" in
    forwards Fr1: fresh_chain_add_fresh_ctx G Fr; (* Add freshness of the whole context *)
    [>  apply Hfr | (* The condition is determined by the guiding hypothesis *)
        clear Hfr Fr; (* Clear the ok hypothesis and the old chain *)
        rewrite ? fresh_chain_eq_merge_var_objctxdom, ? fresh_chain_eq_merge_var_ctxdom, ? fresh_chain_eq_merge_var_ctxdompush in Fr1 by (apply true); (* Split the context in the chain *)
        rename Fr1 into Fr (* Put the new chain in place of the old *)
    ]
  | (@ok var ?e) =>
    let Fr1 := fresh "Fr" in
    forwards Fr1: fresh_chain_add_fresh_renv e Fr; (* Add freshness of the whole context *)
    [>  apply Hfr | (* The condition is determined by the guiding hypothesis *)
        clear Hfr Fr; (* Clear the ok hypothesis and the old chain *)
        rewrite ? fresh_chain_eq_merge_var_renvdom in Fr1 by (apply true); (* Split the context in the chain *)
        (* there is not a renvdompush version like for contexts, because with env, the pushing is just concatenation *)
        rename Fr1 into Fr (* Put the new chain in place of the old *)
    ]
  | _ => fail 1 "Unknown type of freshness hypothesis" Hfr
  end
).

(* Picks a fresh variable for a chain *)
Ltac fresh_chain_pick Fr x :=
  match type of Fr with 
  | fresh_chain_is ?fc =>
    let L := eval simpl in (fresh_chain_to_set fc) in
    let Frx := fresh "Fr" in
    let Lb := beautify_fset L in
    destruct (var_fresh Lb) as [x Frx];
    let Fr1 := fresh "Fr" in
    forwards Fr1: fresh_chain_add_fresh_var Lb fc Frx;  [> simpls; union_sub_solver | fresh_chain_set_solve_by Fr | clear Fr Frx; rename Fr1 into Fr]
  end.




Ltac fresh_chain_cleanup Fr :=
   rewrite <- ? fresh_chain_eq_merge_set_set,
          ? fresh_chain_eq_merge_set_ctx,
          ? fresh_chain_eq_merge_set_renv,
          ? fresh_chain_eq_merge_set_heap,
          ? fresh_element_eq_element,
          ? fresh_element_eq_set_var,
          ? fresh_element_eq_ctxdom,
          ? fresh_element_eq_objctxdom,
          ? fresh_element_eq_renvdom,
          ? fresh_element_eq_heapdom,
          ? fresh_chain_eq_list_cons, ? fresh_chain_eq_list_nil in Fr by (try apply true; eauto).

Ltac fresh_chain_skip1 fcp :=
match fcp with
| fresh_chain_skip ?n ?fc1 => constr: (fresh_chain_skip (S n) fc1)
| _ => constr: (fresh_chain_skip 1 fcp)
end.

(* fc1 from context, fc2 target *)
Ltac fresh_element_match fc1 fc2 :=
match fc1 with
| fc2 => fresh_element_here
| (fresh_element_ctx ?G) =>
  match fc2 with
  | (fresh_element_ctxdom G) =>
    constr: (fresh_element_dom)
  end
| (fresh_element_renv ?G) =>
  match fc2 with
  | (fresh_element_renvdom G) =>
    constr: (fresh_element_dom)
  end
| (fresh_element_heap ?G) =>
  match fc2 with
  | (fresh_element_heapdom G) =>
    constr: (fresh_element_dom)
  end
| fresh_element_containing ?fcM ?fcS =>
  let fep := fresh_element_match fcM fc2 in
  constr: (fresh_element_left fep)
| fresh_element_containing ?fcM ?fcS =>
  let fep := fresh_element_match fcS fc2 in
  constr: (fresh_element_right fep)
end.

(* Find fe in fcCtx and return a path that leads to fe and ends there *)
Ltac fresh_chain_match_set fcCtx fe :=
match fcCtx with 
| (fresh_chain_set fe _) => (* exact match *)
    constr: (fresh_chain_here fresh_chain_end)
| (fresh_chain_set ?fe1 _) => (* inexact match *)
    let fem := (fresh_element_match fe1 fe) in
    constr: (fresh_chain_here_set fresh_chain_end fem)
| (fresh_chain_var ?fe1 _) => (* match set by var *)
    let fem := (fresh_element_match fe1 fe) in
    constr: (fresh_chain_here_set fresh_chain_end fem)
| (fresh_chain_subset ?fe1 _) => (* match set by subset *)
    let fem := (fresh_element_match fe1 fe) in
    constr: (fresh_chain_here_set fresh_chain_end fem)
| (fresh_chain_set _ ?fc1) => (* skip set *)
    let fcm := (fresh_chain_match_set fc1 fe) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
| (fresh_chain_var _ ?fc1) => (* skip var *)
    let fcm := (fresh_chain_match_set fc1 fe) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
| (fresh_chain_subset _ ?fc1) => (* skip subset *)
    let fcm := (fresh_chain_match_set fc1 fe) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
end.

(* Tests for fresh_chain_match_set *)
Goal True. Proof.
pick_fresh x.
pick_fresh y.
pick_fresh z.
Local Notation fcset x y := (fresh_chain_set (fresh_element_var x) y).
Local Notation fcvar x y := (fresh_chain_var (fresh_element_var x) y).
match fresh_chain_match_set (fcset x fresh_chain_nil) (fresh_element_var x) with (fresh_chain_here fresh_chain_end) => idtac end.
match fresh_chain_match_set (fcset x (fcset x (fcset y fresh_chain_nil))) (fresh_element_var y) with (fresh_chain_skip 2 (fresh_chain_here fresh_chain_end)) => idtac end.
auto. Qed.

(* Given a path fcpMain and a path to set fcpSet, make a path that goes to the set first and then continues with fcpMain *)
Ltac fresh_chain_match_prepend_set fcpSet fcpMain :=
match fcpMain with
| fresh_chain_end => fcpSet
| _ => constr: (fresh_chain_lookahead fcpSet fcpMain)
end.

(* fcCtx from context (known), fcTgt target (to prove) *)
Ltac fresh_chain_match fcCtx fcTgt :=
match fcTgt with
| (fresh_chain_var ?A ?fcTgt1) => (* Need var *)
  match fcCtx with
  | (fresh_chain_var A ?fcCtx1) => (* Found exact match in ctx *)
    let fcm := (fresh_chain_match fcCtx1 fcTgt1) in (* Solve rest of chains, restarting up to here *)
    constr: (fresh_chain_here fcm)
  | (fresh_chain_var ?B ?fcCtx1) => (* Found inexact match *)
    let fem := (fresh_element_match B A) in
    let fcm := (fresh_chain_match fcCtx1 fcTgt1) in (* Solve rest of chains, restarting up to here *)
    constr: (fresh_chain_here_var fcm fem)
  | (fresh_chain_var _ ?fcCtx1) => (* Skip var *)
    let fcm := (fresh_chain_match fcCtx1 fcTgt) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
  | (fresh_chain_set _ ?fcCtx1) => (* Skip set *)
    let fcm := (fresh_chain_match fcCtx1 fcTgt) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
  | (fresh_chain_subset _ ?fcCtx1) => (* Skip subset *)
    let fcm := (fresh_chain_match fcCtx1 fcTgt) in
    let fcmc := fresh_chain_skip1 fcm in
    fcmc
  end
| (fresh_chain_set ?A ?fcTgt1) => (* Need set *)
    let fcm := (fresh_chain_match_set fcCtx A) in
    let fcmc := (fresh_chain_match fcCtx fcTgt1) in
    let fcmp := fresh_chain_match_prepend_set fcm fcmc in
    fcmp
| fresh_chain_nil => fresh_chain_end
end.


Lemma fresh_chain_just_ne x y:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_var y) fresh_chain_nil)) ->
  x <> y.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. rewrite notin_singleton in H3. auto.
Qed.
Lemma fresh_chain_just_notin_fv (A: Set) {AF: FreeVar A} x (a: A):
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_element a) fresh_chain_nil)) ->
  x \notin fv a.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_ctxdom x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_ctxdom G) fresh_chain_nil)) ->
  x §# G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_objctxdom x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_objctxdom G) fresh_chain_nil)) ->
  x \notin (objctx_dom G).
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_ctxfv x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_ctx G) fresh_chain_nil)) ->
  x \notin ctx_fv G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_renvdom x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_renvdom G) fresh_chain_nil)) ->
  x # G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_heapdom x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_heapdom G) fresh_chain_nil)) ->
  x # G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_renvfv x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_renv G) fresh_chain_nil)) ->
  x \notin fv_env G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_heapfv x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_heap G) fresh_chain_nil)) ->
  x \notin fv_env G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.
Lemma fresh_chain_just_notin_set x G:
  fresh_chain_is (fresh_chain_var (fresh_element_var x) (fresh_chain_set (fresh_element_set G) fresh_chain_nil)) ->
  x \notin G.
Proof.  
  introv His.
  inversions His.
  inversions H3.
  inversions H5.
  simpl in H.
  destructecs H. auto.
Qed.



Ltac fresh_chain_before_var fc fe := (
  match fc with
  | fresh_chain_var fe ?fc1 => constr:(fresh_chain_nil)
  | fresh_chain_var ?fe1 ?fc1 =>
    let fc2 := (fresh_chain_before_var fc1 fe) in
    constr: (fresh_chain_var fe1 fc2)
  | fresh_chain_set ?fe1 ?fc1 => 
    let fc2 := (fresh_chain_before_var fc1 fe) in
    constr: (fresh_chain_set fe1 fc2)
  end).

Ltac fresh_chain_hyp_before_var Fr fe := (
  match type of Fr with
  | fresh_chain_is ?fc => fresh_chain_before_var fc fe
  end).

Ltac fresh_chain_front_var Fr x := (
  let fcLx := fresh_chain_hyp_before_var Fr (fresh_element_var x) in
  eapply fresh_chain_invert with (fcL:= fcLx) in Fr;
  simpl in Fr
).


Ltac fresh_chain_front_list_r Fr xs := (
  match xs with
  | (nil)%list => rewrite <- fresh_chain_eq_list_nil in Fr by (try apply true)
  | (?x::?xs1)%list => fresh_chain_front_list_r Fr xs1; fresh_chain_front_var Fr x
  end
).

(* Inverts the order in Fr so that there is var xs in the front. (All variables in xs must be previously in Fr in any order) *)
Ltac fresh_chain_front_list Fr xs := (
  fresh_chain_front_list_r Fr xs;
  rewrite <- ? fresh_chain_eq_list_cons in Fr by (try apply true)
).


Goal
  forall (X: fset var) (x y: var), fresh X 2 (x::y::nil) -> fresh X 2 (y::x::nil).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_cleanup FrG.
  fresh_chain_front_list FrG (y::x::nil)%list.
  eapply fresh_chain_fresh_list; [>eauto| simpl; union_sub_solver|eauto].
Qed.

Ltac simpl_fv_env := rewrite ? fv_env_concat, ? fv_env_single, ? fv_env_empty in *.

(* Can be used instead of idtac without printing *)
Ltac print_ignore x y := idtac.
Ltac print_idtac x y := idtac x y.


(* Transforms chain hypothesis Fr, by pulling fe to the front *)
Ltac fresh_chain_front_element Fr fe := (
  let fcLx := fresh_chain_hyp_before_var Fr fe in
  eapply fresh_chain_invert with (fcL:= fcLx) in Fr;
  simpl in Fr
).

(* Transforms chain hypothesis Fr, by pulling the vars in fc to the front *)
Ltac fresh_chain_front_chain_r Fr fc := (
  match fc with
  | (fresh_chain_var ?fe ?fc1) => fresh_chain_front_chain_r Fr fc1; fresh_chain_front_element Fr fe
  | (fresh_chain_set ?fe ?fc1) => fresh_chain_front_chain_r Fr fc1
  | fresh_chain_nil => idtac
  end
).

Ltac fresh_chain_front_chain Fr fc := (
  fresh_chain_front_chain_r Fr fc
).

(* Solve chain fcG in goal by matiching it in fcF in Fr, without reordering *)
Ltac fresh_chain_solve_chains_direct Fr fcF fcG print := (
    print "match " fcG;
    print "in " fcF;
    let fcp := fresh_chain_match fcF fcG in
    print "result " fcp;
    eapply (fresh_chain_is_select fcp Fr)
).

Ltac fresh_chain_solve_direct Fr print := (
  rewrite ? fresh_chain_eq_last_var_set by apply true; (* The last var can match a set *)
  match goal with
  | |- fresh_chain_is ?fcG =>
    match type of Fr with 
    | fresh_chain_is fcG => apply Fr
    | fresh_chain_is ?fcF =>
      fresh_chain_solve_chains_direct Fr fcF fcG print
    end
  end
).

(* Solve chain fcG in goal by matiching it in fcF in Fr, using reordering *)
Ltac fresh_chain_solve_chains_reorder Fr fcF fcG print := (
    print "reorder " fcF;
    print "by " fcG;
    fresh_chain_front_chain_r Fr fcG;
    fresh_chain_solve_direct Fr print
).


Ltac fresh_chain_solve_reorder Fr print := (
  match goal with
  | |- fresh_chain_is ?fcG =>
    match type of Fr with
    | fresh_chain_is fcG => apply Fr
    | fresh_chain_is ?fcF =>
      fresh_chain_solve_chains_reorder Fr fcF fcG print
    end
  end
).

(* Solve chain, allowing reordering *)
Ltac fresh_chain_solve_reorderable Fr print := (
    first [
      fresh_chain_solve_direct Fr print |
      fresh_chain_solve_reorder Fr print
    ]
).

(* FSR is tactic to solve chains while allowing reordering, 
   FSD when reordering should not be allowed *)
Ltac fresh_chain_solve_r Fr print FSR FSD :=
match goal with
(* Solving a chain *)
| |- fresh_chain_is ?fcG =>
  fresh_chain_solve_reorderable Fr print
(* Converting into a chain *)
| |- ctx_ok ?G =>
  let fcG := (fresh_chain_ctxdom_var_chain G fresh_chain_nil) in (* Generate a chain corresponding to the ctx *)
  print "ctx " fcG;
  eapply fresh_chain_cond_just_ctx with (fc := fcG);
  rewrite ? fresh_chain_eq_ctxdom_push, ? fresh_chain_eq_ctxdom_empty by apply true;
  FSR Fr print
| |- (@ok var ?e) =>
  let fce := (fresh_chain_renvdom_varset_chain e fresh_chain_nil) in (* Generate a chain corresponding to the renv *)
  print "env " fce;
  eapply fresh_chain_cond_just_renv with (fc := fce);
  FSR Fr print
| |- (@ok item ?Si) =>
  let fcSi := (fresh_chain_heapdom_varset_chain Si fresh_chain_nil) in (* Generate a chain corresponding to the renv *)
  print "heap " fcSi;
  eapply fresh_chain_cond_just_heap with (fc := fcSi);
  FSR Fr print
| |- (@notin var ?x (fv ?a)) =>
  apply fresh_chain_just_notin_fv;
  FSD Fr print
| |- (@notin var ?x (ctx_dom ?a)) =>
  apply fresh_chain_just_notin_ctxdom;
  FSD Fr print
| |- (@notin var ?x (objctx_dom ?a)) =>
  apply fresh_chain_just_notin_objctxdom;
  FSD Fr print
| |- (@notin var ?x (ctx_fv ?a)) =>
  apply fresh_chain_just_notin_ctxfv;
  FSD Fr print
| |- (@notin var ?x (@dom item ?a)) =>
  apply fresh_chain_just_notin_heapdom;
  FSD Fr print
| |- (@notin var ?x (@dom var ?a)) =>
  apply fresh_chain_just_notin_renvdom;
  FSD Fr print
| |- (@notin var ?x (@fv_in_values item (@fv item ItemFreeVar) ?a)) =>
  apply fresh_chain_just_notin_heapfv;
  FSD Fr print
| |- (@notin var ?x (@fv_in_values var (@fv var VarFreeVar) ?a)) =>
  apply fresh_chain_just_notin_renvfv;
  FSD Fr print
| |- (@notin var ?x ?a) =>
  apply fresh_chain_just_notin_set;
  FSD Fr print
| |- (not (@eq var ?x ?y)) =>
  eapply fresh_chain_just_ne;
  FSD Fr print
| |- (not (@eq var ?x ?y)) => (* Backtracing case. *)
  apply not_eq_sym;
  eapply fresh_chain_just_ne;
  FSD Fr print
(* Special handling *)
| |- fresh ?X ?l ?xs =>
  fresh_chain_front_list Fr xs;
  eapply fresh_chain_fresh_list_large;
  [>
    eauto |
    simpl;ctx_simpl_dom; ctx_simpl_fv;simpl_dom; simpl_fv_env; rewrite ? from_list_cons, ?from_list_nil; union_sub_solver |
    eauto
  ]
(* Directly solvable cases *)
| |- True => apply I
| |- (@notin var ?x \{}) => apply notin_empty
(* Simplification *)
| |- ?A /\ ?B => split;
   fresh_chain_solve_r Fr print FSR FSD 
| |- (@notin var ?x (?A \u ?B)) =>
  rewrite notin_union;
  fresh_chain_solve_r Fr print FSR FSD 
| |- (@notin var ?x (@singleton var ?y)) =>
  rewrite notin_singleton;
  fresh_chain_solve_r Fr print FSR FSD 
| |- (@notin var ?x (from_list (?y::?ys)%list)) =>
  rewrite from_list_cons;
  fresh_chain_solve_r Fr print FSR FSD 
| |- (@notin var ?x (from_list (nil)%list)) =>
  rewrite from_list_nil; 
  fresh_chain_solve_r Fr print FSR FSD 
| |- (@notin var ?x (@fv var VarFreeVar ?a)) =>
  change (@fv var VarFreeVar a) with \{ a};
  fresh_chain_solve_r Fr print FSR FSD 
end.
Ltac fresh_chain_solve Fr := fresh_chain_solve_r Fr print_ignore fresh_chain_solve_reorderable fresh_chain_solve_direct.
Ltac fresh_chain_apply Fr := fresh_chain_solve_r Fr print_ignore print_ignore print_ignore.
Ltac fresh_chain_solve_debug Fr := fresh_chain_solve_r Fr print_idtac fresh_chain_solve_reorderable fresh_chain_solve_direct.

Goal
  forall (X: fset var) (x y: var), fresh X 2 (x::y::nil) -> fresh X 2 (y::x::nil).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_cleanup FrG.
  fresh_chain_solve FrG.
Qed.

Goal
  forall (G1 G2 G3: ctx), ctx_ok (G1 §& G2 §& G3) -> ctx_ok (G1 §& G2 §& G3).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_solve FrG.
Qed.

Goal
  forall (X: fset var) (x y: var), fresh X 2 (x::y::nil) -> ctx_ok (ctx_empty § x / v_let ~ typ_top § y / v_let ~ typ_top).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_cleanup FrG.
  fresh_chain_solve FrG.
Qed.


Goal
  forall (G1 G2 G3 G4 G5: ctx), ctx_ok (G1 §& G2 §& G3 §& G4 §& G5) -> ctx_ok (G1 §& G4).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_solve FrG.
Qed.

Goal
  forall (e1 e2: renv), ok (e1 & e2) -> ok e1 /\ ok e2.
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_solve FrG.
Qed.

Goal
  forall (e1: renv), ok e1 -> ok e1.
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_gather FrG.
  fresh_chain_solve FrG.
Qed.

Goal
  forall (X Y Z: fset var) (x y: var), Z \c (X \u Y) -> fresh (X \u Y) 2 (x::y::nil) -> fresh (X \u Z) 2 (y::x::nil).
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H0.

  rewrite <- (fresh_chain_eq_prepend_susbet_subset) with (X:=Z) (fc:=(fresh_chain_set (fresh_element_set Y) _)) in FrG.
  2: { simpls. eapply subset_trans. eauto.  union_sub_solver. } 2, 3, 4: apply true.

  fresh_chain_cleanup FrG.
  fresh_chain_solve FrG.
Qed.


Ltac fresh_chain_include_in Fr Hin :=
  match type of Hin with 
  | ?x1 \in ctx_dom ?G1 =>
    rewrite fresh_element_eq_containing_var with (x:=x1) (feM := fresh_element_ctxdom G1) in Fr; [> | simpl; apply Hin | apply false .. ]
  | ?x1 \in ctx_dom ?G1 =>
    rewrite fresh_element_eq_containing_var with (x:=x1) (feM := fresh_element_ctx G1) in Fr; [> | simpl; union_in_solver | apply false .. ]
  | ?x1 \in @dom item ?G1 =>
    rewrite fresh_element_eq_containing_var with (x:=x1) (feM := fresh_element_heapdom G1) in Fr; [> | simpl; apply Hin | apply false .. ]
  | ?x1 \in @dom var ?G1 =>
    rewrite fresh_element_eq_containing_var with (x:=x1) (feM := fresh_element_renvdom G1) in Fr; [> | simpl; apply Hin | apply false .. ]
  end.

Ltac fresh_chain_include_typing Fr Hty := (
  match type of Hty with
  | ?G ⁏ ?e ⊢ ?x ∶ ?T =>
    let Hin := fresh "Hin" in
    assert (x \in ctx_dom G) as Hin by ( eapply typing_implies_dom; eapply Hty );
    fresh_chain_include_in Fr Hin; clear Hin
  | ?G ⁏ ?e ⊢ trm_read (avar_f ?x) ?f ∶ ?T =>
    let Hin := fresh "Hin" in
    let Htyx := fresh "Htyx" in
    assert (x \in ctx_dom G) as Hin by (eapply ty_trm_read_inv in Hty; destructc Hty as ? ? Htyx; eapply typing_implies_dom; eapply Htyx );
    fresh_chain_include_in Fr Hin; clear Hin
  | ?G ⁏ ?e ⊢ trm_call (avar_f ?x) ?f (avar_f ?y) ∶ ?T =>
    let Hin := fresh "Hin" in
    let Hin1 := fresh "Hin" in
    let Htyx := fresh "Htyx" in
    let Htyy := fresh "Htyy" in
    assert (x \in ctx_dom G /\ y \in ctx_dom G) as Hin by (eapply ty_trm_call_inv in Hty; destructc Hty as ? ? ? Htyx Htyy; split; eapply typing_implies_dom; [> eapply Htyx|eapply Htyy] );
    destructc Hin as Hin Hin1;
    fresh_chain_include_in Fr Hin; clear Hin;
    fresh_chain_include_in Fr Hin1; clear Hin1
  end
).

Ltac fresh_chain_simpl := (
  rewrite
    <- ? fresh_chain_eq_merge_set_set,
    ? fresh_element_eq_ctxdom_set,
    ? fresh_element_eq_objctxdom_set,
    ? fresh_element_eq_renvdom_set,
    <- ? fresh_chain_eq_last_var_set,
    ? fresh_chain_eq_merge_var_objctxdom,
    ? fresh_chain_eq_merge_var_ctxdom,
    ? fresh_chain_eq_merge_var_ctxdompush,
    ? fresh_chain_eq_last_var_set,
    <- ? fresh_element_eq_heapdom_single,
    <- ? fresh_element_eq_renvdom_single,
    ? fresh_element_eq_subst_ctxdom,
    ? fresh_element_eq_subst_objctxdom,
    ? fresh_chain_eq_merge_var_ctxdom_empty,
    ? fresh_element_eq_set_var,
    ? fresh_chain_eq_merge_set_empty
  by apply true
).
Ltac fresh_chain_solve2 FrG := (
fresh_chain_apply FrG; fresh_chain_simpl; fresh_chain_solve FrG
).

Section WithTypingMode.
Context (typing_mode: Mode).

Goal
  forall x G e f T,
  ctx_ok G ->
  G ⁏ e ⊢ trm_read (avar_f x) f ∶ T ->
  ctx_ok G.
Proof.
  intros.
  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H.
  fresh_chain_include_typing FrG H0.
  fresh_chain_solve FrG.
Qed.


End WithTypingMode.