Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax.

(** * Generalization *)
(**
  A syntax element can be generalized
  by replacing each occurence of a free variable
  with a fresh variable.

  This generalized element can, by substitution, reach any element that could reach the original element.
*)

Record generalization (A: Set) {AF: FreeVar A} {AS: SubstVar A} L (a: A) := {
  b: A ; (* The generalized element *)
  xs: list var ; (* Free variables of the original element *)
  ys: list var ; (* Free variables of the generalized element *)
  xs_fv: from_list xs = fv a;
  ys_fv: from_list ys = fv b;
  xsysl: length xs = length ys;
  ys_fresh: fresh (L) (length ys) ys ;
  subst: (b /[ys --> xs]) = a ; (* Reaching the original element *)
  mg: (forall (c:A) (zs vs: list var), (* Being maximally general *)
        (fv c) \c L ->
        c /[ zs --> vs] = a ->
        exists ws,
          length ws = length ys /\
          b /[ys --> ws] = c )
}.

(* Generalizable syntax allows generalizing any element with fresh variables *)
Class Generalizable (A: Set) (AF: FreeVar A) (AS: SubstVar A) :=
  generalize (X: A) (L: vars): generalization L X.

(* Tactics for instances *)

(* Unfold methods of generalization *)
Ltac generalization_unfold := (unfold b, xs, ys in *).
(* Rewrite sets of free variables to corresponding lists, using ys_fv methods, where right hand side contains a substitution. *)
Ltac generalization_rewrite_fv_list' := (
  repeat match goal with 
    Hxs: from_list ?xs1 = fv (?b1 /[ ?ys1 --> ?xs1]) |- _ => rewrite <- Hxs
  end
).
(* Rewrite sets of free variables to corresponding lists, using xs_fv and ys_fv methods *)
Ltac generalization_rewrite_fv_list := (
  repeat match goal with 
    Hxs: from_list ?xs1 = fv (?b1) |- _ => rewrite <- Hxs
  end
).
(* Rewrite lengths of variable list from xs to ys using the xsysl methods *)
Ltac generalization_rewrite_fv_length := (
  repeat match goal with 
    Hxs: length ?xs1 = length ?ys1 |- _ => rewrite Hxs
  end
).
(* Solve the xs_fv and ys_fv methods *)
Ltac generalize_solve_fv_list := (
  generalization_unfold; (* Unfold methods *)
  sympls; (* Simplify free variables of syntax element *)
  generalization_rewrite_fv_list; (* Rewrite fv sets to lists *)
  rewrite ? from_list_append; (* Separate appended lists *)
  auto (* Both sides are the same *)
).
(* Solve the xsysl method *)
Ltac generalize_solve_length := (
  generalization_unfold; (* Unfold methods *)
  rewrite ? List.app_length; (* Separate appended lists *)
  generalization_rewrite_fv_length; (* Rewrite xs to ys *)
  auto (* Both sides are the same *)
).
(* Solve the ys_fresh method *)
Ltac generalize_solve_fresh := (
  generalization_unfold; (* Unfold methods *)
  sympls;
  repeat eapply fresh_append; (* Separate appended lists *)
  eauto (* Solve freshness *)
).
(* Solve the subst method *)
Ltac generalize_solve_subst := (
    generalization_unfold; (* Unfold methods *)
    autorewrite with simpl_subst_vars; (* Simplify substitution in syntax element *)
    first [eapply f_equal | eapply f_equal2 | eapply f_equal3]; auto; (* Break equality goal into sub-elements *)
    rewrite ? subst_vars_append by (eauto; generalize_solve_length); subst; (* Break appended lists in substitutions *)
    repeat eapply f_equal; (* Strip same substitutions from the right *)
    rewrite subst_vars_fresh; eauto; (* Strip substitution of fresh variables *)
    generalization_rewrite_fv_list; (* In remaining freshness goals rewrite fv sets to lists *)
    fresh_solve
).


(* Destruct subset hypothesis of an union of sets on left side into subset hypotheses for each individual set from the left side *)
Ltac destruct_subset_union := (
    repeat match goal with 
    | H: ?a \u ?b \c ?c |- _ => pose proof (subset_union_l1 H); eapply subset_union_l2 in H
    end
).
(* Prepare for solving the mg method *)
Ltac generalite_mg_prepare c := (
  generalization_unfold; (* Unfold methods *)
  intros; (* Introduce the reaching element and variable list *)
  invert_subst_vars_of c; (* destruct the reaching element *)
  sympls;
  destruct_subset_union (* prepare hypotheses about free variables in the reaching element *)
).
(* Get a variable list from an sub-element, when solving a mg method *)
Ltac generalize_mg_get ws1 mg1 := (
    forwards ws1 ?Hws1 ?Hb1: mg1; eauto; eauto using subset_union_rl
).
(* Finish solving the mg method after the variable list was found *)
Ltac generalite_mg_finish := (
    split; (* Split the conjunction *)
    [> generalize_solve_length | (* Solve the length requirement *)
      generalize_solve_subst; (* Solve the substitution equality *)
      eapply fresh_subset;
      eauto
    ]
).

(* Instances *)

Program Definition generalize_var (v: var) L : (generalization L v) :=
  let (y, fr) := var_fresh (L \u \{v}) in
{|
  b := y ;
  xs := (v::nil) ;
  ys := (y::nil) ;
|}.
Next Obligation. rewrite from_list_cons. rewrite from_list_nil. unfold fv, VarFreeVar. rewrite union_empty_r. auto. Qed.
Next Obligation. rewrite from_list_cons. rewrite from_list_nil. unfold fv, VarFreeVar. rewrite union_empty_r. auto. Qed.
Next Obligation. unfold subst_var. unfold VarSubstVar. cases_if. auto. Qed.
Next Obligation. exists (c::nil)%list. unfold subst_var, VarSubstVar. simpls. cases_if. auto. Qed.

Instance VarGeneralizable: Generalizable VarFreeVar VarSubstVar := generalize_var.

Lemma generalize_avar_b L n: generalization L (avar_b n).
Proof.
  eapply {| b:= (avar_b n); xs:=(@nil var); ys:=(@nil var) |}.
  Unshelve.
  all: eauto.
  exists (@nil var)%list.
  simpls.
  destruct c.
  rewrite avar_subst_vars_b in H0. inversions H0. auto.
  rewrite avar_subst_vars_f in H0. inversions H0.
Qed.

Lemma generalize_avar_f L x: generalization L (avar_f x).
Proof.
  lets Gv: (generalize_var x L).
  eapply {| b:= (avar_f (b Gv)); xs:=(xs Gv); ys:=(ys Gv) |}.
  Unshelve.
  all: destruct Gv; generalization_unfold.
  - unfold fv, VarFreeVar. auto.
- rewrite ys_fv0.  unfold fv, VarFreeVar. auto.
- unfold fv, VarFreeVar in ys_fresh0.  auto.
- auto.
- rewrite avar_subst_vars_f. rewrite subst0.  auto.
- intros.
  destruct c; autorewrite with simpl_subst_vars in H0; inversions H0.
  specialize (mg0 _ _ _ H H2).
  destructc mg0 as ? ? ?.
  exists x.
  rewrite avar_subst_vars_f.
  subst. auto.
Qed.

Definition generalize_avar (v: avar) L  : (generalization L v) :=
match v with
| avar_b n => generalize_avar_b L n
| avar_f x => generalize_avar_f L x
end.
Instance AvarGeneralizable: Generalizable AvarFreeVar AvarSubstVar := generalize_avar.

Notation generalization_for T := (forall L', generalization L' T)(only parsing).
(* Notation generalize_for T := (forall L', generalization L' T)(only parsing). *)


Lemma generalize_dec_typ L B (T U: typ) (GT: generalization_for T) (GU: generalization_for U): generalization L (dec_typ B T U).
Proof.
  specialize (GT (L \u fv T \u fv U )).
  specialize (GU (L \u fv T \u fv U \u from_list (xs GT) \u from_list (ys GT))).
  eapply {| b:= dec_typ B (b GT) (b GU); xs:=(xs GT ++ xs GU); ys:=(ys GT ++ ys GU) |}.
  Unshelve.
  all: destruct GT; destruct GU.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_dec_fld L f (T U: typ) (GT: generalization_for T) (GU: generalization_for U): generalization L (dec_fld f T U).
Proof.
  specialize (GT (L \u fv T \u fv U )).
  specialize (GU (L \u fv T \u fv U \u from_list (xs GT) \u from_list (ys GT))).
  eapply {| b:= dec_fld f (b GT) (b GU); xs:=(xs GT ++ xs GU); ys:=(ys GT ++ ys GU) |}.
  Unshelve.
  all: destruct GT; destruct GU.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_dec_met L m (TC TZ TR: typ) (GTC: generalization_for TC) (GTZ: generalization_for TZ) (GTR: generalization_for TR): generalization L (dec_met m TC TZ TR).
Proof.
  specialize (GTC (L \u fv TC \u fv TZ \u fv TR)).
  specialize (GTZ (L \u fv TC \u fv TZ \u fv TR \u from_list (xs GTC) \u from_list (ys GTC) )).
  specialize (GTR (L \u fv TC \u fv TZ \u fv TR \u from_list (xs GTC) \u from_list (ys GTC)  \u from_list (xs GTZ) \u from_list (ys GTZ) )).
  eapply {| b:= dec_met m (b GTC) (b GTZ) (b GTR); xs:=(xs GTC ++ xs GTZ ++xs GTR); ys:=(ys GTC ++ ys GTZ ++ ys GTR) |}.
  Unshelve.
  all: destruct GTC; destruct GTZ; destruct GTR.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - 
    generalization_unfold;
    autorewrite with simpl_subst_vars;
    first [eapply f_equal2 | eapply f_equal3]; auto;
    rewrite ? subst_vars_append by (eauto; generalize_solve_length); subst;
    repeat eapply f_equal.

(*
all: ( rewrite subst_vars_fresh; eauto;
    generalization_rewrite_fv_list; fresh_solve ).

all: repeat eapply f_equal.

all: ( rewrite subst_vars_fresh; eauto;
    generalization_rewrite_fv_list; fresh_solve ).
*)
    rewrite subst_vars_fresh with (a:=b0); eauto.
    rewrite subst_vars_fresh with (a:=b0); eauto.
    rewrite <- ys_fv0.
    fresh_solve.
    rewrite <- ys_fv0.
    fresh_solve.
    rewrite subst_vars_fresh with (a:=b1); eauto.
    rewrite subst_vars_fresh; eauto.
    rewrite <- ys_fv1.
    fresh_solve.
    rewrite subst_vars_fresh; eauto.
    rewrite subst_vars_fresh; eauto.
    eauto.
    rewrite subst_vars_fresh; eauto.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    generalize_mg_get ws2 mg2.
    exists ((ws0 ++ ws1 ++ ws2)%list).
    split.
    + generalize_solve_length.
    +
    autorewrite with simpl_subst_vars;
    first [eapply f_equal2 | eapply f_equal3]; auto;
    rewrite ? subst_vars_append by (eauto; generalize_solve_length); subst;
    repeat eapply f_equal.

    rewrite subst_vars_fresh with (a:=b0); eauto.
    rewrite subst_vars_fresh with (a:=b0); eauto.
    rewrite <- ys_fv0.
    fresh_solve.
    rewrite <- ys_fv0.
    fresh_solve.
    rewrite subst_vars_fresh with (a:=b1); eauto.
    rewrite subst_vars_fresh; eauto.
    eapply fresh_subset; eauto.
    rewrite <- ys_fv1.
    fresh_solve.
    rewrite subst_vars_fresh; eauto.
    rewrite subst_vars_fresh; eauto.
    eapply fresh_subset; eauto.
    rewrite subst_vars_fresh; eauto.
    eapply fresh_subset; eauto.
    eapply fresh_subset; eauto.
Qed.

Lemma generalize_typ_andor (T: typ->typ->typ)(HT: typ_andor T) L (T1 T2: typ) (GT1: generalization_for T1) (GT2: generalization_for T2): generalization L (T T1 T2).
Proof.
  specialize (GT1 (L \u fv T1 \u fv T2) ).
  specialize (GT2 (L \u fv T1 \u fv T2 \u from_list (xs GT1) \u from_list (ys GT1) )).
  eapply {| b:= T (b GT1) (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2; inversions HT.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_typ_fun L (T1 T2: typ) (GT1: generalization_for T1) (GT2: generalization_for T2): generalization L (typ_fun T1 T2).
Proof.
  specialize (GT1 (L \u fv T1 \u fv T2) ).
  specialize (GT2 (L \u fv T1 \u fv T2 \u from_list (xs GT1) \u from_list (ys GT1) )).
  eapply {| b:= typ_fun (b GT1) (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_nulary T (HT: typ_nulary T) L: generalization L T. 
Proof.
  eapply  {| b := T; xs:= nil; ys:=nil |}.
  Unshelve.
  all: inversions HT; eauto.
  all:
    generalite_mg_prepare c;
    exists (@nil var);
    sympls; auto.
Qed.

Lemma generalize_typ_bnd L (T: typ) (GT: generalization_for T): generalization L (typ_bnd T).
Proof.
  specialize (GT (L \u fv T) ).
  eapply {| b:= typ_bnd (b GT); xs:=(xs GT); ys:=(ys GT) |}.
  Unshelve.
  all: destruct GT; eauto.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_typ_rcd L (D: dec) (GD: generalization_for D): generalization L (typ_rcd D).
Proof.
  specialize (GD (L \u fv D) ).
  eapply {| b:= typ_rcd (b GD); xs:=(xs GD); ys:=(ys GD) |}.
  Unshelve.
  all: destruct GD; eauto.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.

Ltac destruct_generalize_avar := (repeat match goal with |- context [generalize_avar ?L ?x] => destruct (generalize_avar L x) end).
Lemma generalize_typ_sel L (x: avar) B (y: avar): generalization L (typ_sel x B y ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x \u fv y) ).
  lets GT2: (generalize_avar y (L \u fv x \u fv y \u from_list (xs GT1) \u from_list (ys GT1) ) ) .
  eapply {| b:= typ_sel (b GT1) B (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2; destruct_generalize_avar.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.


Fixpoint generalize_typ (T : typ) L {struct T} : generalization L T :=
  match T with
  | typ_top      => generalize_nulary typ_nulary_top L
  | typ_bot       => generalize_nulary typ_nulary_bot L
  | typ_rcd D      => generalize_typ_rcd L (fun L' => generalize_dec D L')
  | typ_and T1 T2  => generalize_typ_andor typ_andor_and L (fun L' => generalize_typ T1 L') (fun L' => generalize_typ T2 L')
  | typ_or T1 T2  => generalize_typ_andor typ_andor_or L (fun L' => generalize_typ T1 L') (fun L' => generalize_typ T2 L')
  | typ_sel x B y => generalize_typ_sel L x B y
  | typ_bnd T =>  generalize_typ_bnd L (fun L' => generalize_typ T L')
  | typ_fun T1 T2  => generalize_typ_fun L (fun L' => generalize_typ T1 L') (fun L' => generalize_typ T2 L')
  | typ_nocap       => generalize_nulary typ_nulary_nocap L
  end with
generalize_dec (D : dec) L {struct D} : generalization L D :=
  match D with
  | dec_typ B T U => generalize_dec_typ L B (fun L' => generalize_typ T L') (fun L' => generalize_typ U L')
  | dec_fld f T U => generalize_dec_fld L f (fun L' => generalize_typ T L') (fun L' => generalize_typ U L')
  | dec_met m TC TZ TR => generalize_dec_met L m (fun L' => generalize_typ TC L') (fun L' => generalize_typ TZ L') (fun L' => generalize_typ TR L')
  end.

Instance TypGeneralizable: Generalizable TypFreeVar TypSubstVar := generalize_typ.


Lemma generalize_trm_var L (x: avar): generalization L (trm_var x ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x) ).
  eapply {| b:= trm_var (b GT1); xs:=(xs GT1); ys:=(ys GT1) |}.
  Unshelve.
  all: destruct GT1; destruct_generalize_avar; eauto.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_trm_write L (x: avar) m (y: avar): generalization L (trm_write x m y ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x \u fv y) ).
  lets GT2: (generalize_avar y (L \u fv x \u fv y \u from_list (xs GT1) \u from_list (ys GT1) ) ) .
  eapply {| b:= trm_write (b GT1) m (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2; destruct_generalize_avar.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.
Lemma generalize_trm_call L (x: avar) m (y: avar): generalization L (trm_call x m y ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x \u fv y) ).
  lets GT2: (generalize_avar y (L \u fv x \u fv y \u from_list (xs GT1) \u from_list (ys GT1) ) ) .
  eapply {| b:= trm_call (b GT1) m (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2; destruct_generalize_avar.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.
Lemma generalize_trm_apply L (x: avar)  (y: avar): generalization L (trm_apply x  y ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x \u fv y) ).
  lets GT2: (generalize_avar y (L \u fv x \u fv y \u from_list (xs GT1) \u from_list (ys GT1) ) ) .
  eapply {| b:= trm_apply (b GT1)  (b GT2); xs:=(xs GT1 ++ xs GT2); ys:=(ys GT1 ++ ys GT2) |}.
  Unshelve.
  all: destruct GT1; destruct GT2; destruct_generalize_avar.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.
Lemma generalize_trm_read L (x: avar) m : generalization L (trm_read x m ).
Proof.
  lets GT1: (generalize_avar x (L \u fv x) ).
  eapply {| b:= trm_read (b GT1) m; xs:=(xs GT1); ys:=(ys GT1) |}.
  Unshelve.
  all: destruct GT1; eauto; destruct_generalize_avar.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_defs_cons L (ds: defs) (d: def) (Gds: generalization_for ds) (Gd: generalization_for d): generalization L (defs_cons ds d).
Proof.
  specialize (Gds (L \u fv ds \u fv d) ).
  specialize (Gd (L \u fv ds \u fv d \u from_list (xs Gds) \u from_list (ys Gds) )).
  eapply {| b:= defs_cons (b Gds) (b Gd); xs:=(xs Gds ++ xs Gd); ys:=(ys Gds ++ ys Gd) |}.
  Unshelve.
  all: destruct Gds; destruct Gd.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_defs_nil L: generalization L defs_nil. 
Proof.
  eapply {| b := defs_nil; xs:= nil; ys:=nil |}.
  Unshelve.
  all: eauto.
  -
    generalite_mg_prepare c.
    exists (@nil var).
    sympl; eauto.
Qed.

Lemma generalize_trm_let L (t1: trm) (t2: trm) (Gt1: generalization_for t1) (Gt2: generalization_for t2): generalization L (trm_let t1 t2).
Proof.
  specialize (Gt1 (L \u fv t1 \u fv t2) ).
  specialize (Gt2 (L \u fv t1 \u fv t2 \u from_list (xs Gt1) \u from_list (ys Gt1) )).
  eapply {| b:= trm_let (b Gt1) (b Gt2); xs:=(xs Gt1 ++ xs Gt2); ys:=(ys Gt1 ++ ys Gt2) |}.
  Unshelve.
  all: destruct Gt1; destruct Gt2.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_trm_lit L (l: lit) (t2: trm) (Gt1: generalization_for l) (Gt2: generalization_for t2): generalization L (trm_lit l t2).
Proof.
  specialize (Gt1 (L \u fv l \u fv t2) ).
  specialize (Gt2 (L \u fv l \u fv t2 \u from_list (xs Gt1) \u from_list (ys Gt1) )).
  eapply {| b:= trm_lit (b Gt1) (b Gt2); xs:=(xs Gt1 ++ xs Gt2); ys:=(ys Gt1 ++ ys Gt2) |}.
  Unshelve.
  all: destruct Gt1; destruct Gt2.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_def_typ L B (x: typ) : generalization L (def_typ B x).
Proof.
  lets GT1: (generalize_typ x (L \u fv x) ).
  eapply {| b:= def_typ B (b GT1); xs:=(xs GT1); ys:=(ys GT1) |}.
  Unshelve.
  all: destruct GT1; eauto; destruct_generalize_avar.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.
Lemma generalize_def_fld L B (x: avar) : generalization L (def_fld B x).
Proof.
  lets GT1: (generalize_avar x (L \u fv x) ).
  eapply {| b:= def_fld B (b GT1); xs:=(xs GT1); ys:=(ys GT1) |}.
  Unshelve.
  all: destruct GT1; eauto; destruct_generalize_avar.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.
Lemma generalize_def_met L B (t: trm) (GT1: generalization_for t)  : generalization L (def_met B t).
Proof.
  specialize (GT1 (L \u fv t) ).
  eapply {| b:= def_met B (b GT1); xs:=(xs GT1); ys:=(ys GT1) |}.
  Unshelve.
  all: destruct GT1; eauto; destruct_generalize_avar.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    exists ((ws0)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_lit_fun L (T: typ) (t: trm) (Gt2: generalization_for t): generalization L (lit_fun T t).
Proof.
  lets Gt1: (generalize_typ T (L \u fv T \u fv t)).
  specialize (Gt2 (L \u fv T \u fv t \u from_list (xs Gt1) \u from_list (ys Gt1) )).
  eapply {| b:= lit_fun (b Gt1) (b Gt2); xs:=(xs Gt1 ++ xs Gt2); ys:=(ys Gt1 ++ ys Gt2) |}.
  Unshelve.
  all: destruct Gt1; destruct Gt2.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Lemma generalize_lit_obj L (T: typ) (t: defs) (Gt2: generalization_for t): generalization L (lit_obj T t).
Proof.
  lets Gt1: (generalize_typ T (L \u fv T \u fv t)).
  specialize (Gt2 (L \u fv T \u fv t \u from_list (xs Gt1) \u from_list (ys Gt1) )).
  eapply {| b:= lit_obj (b Gt1) (b Gt2); xs:=(xs Gt1 ++ xs Gt2); ys:=(ys Gt1 ++ ys Gt2) |}.
  Unshelve.
  all: destruct Gt1; destruct Gt2.
  - generalize_solve_fv_list.
  - generalize_solve_fv_list.
  - generalize_solve_length.
  - generalize_solve_fresh.
  - generalize_solve_subst.
  - generalite_mg_prepare c.
    generalize_mg_get ws0 mg0.
    generalize_mg_get ws1 mg1.
    exists ((ws0 ++ ws1)%list).
    generalite_mg_finish.
Qed.

Fixpoint generalize_trm (t : trm) L {struct t} : generalization L t :=
    match t with
    | trm_var a => generalize_trm_var L a
    | trm_read v m => generalize_trm_read L v m
    | trm_apply f' a => generalize_trm_apply L f' a
    | trm_let t1 t2 => generalize_trm_let L (fun L' => generalize_trm t1 L') (fun L' => generalize_trm t2 L')
    | trm_lit l t => generalize_trm_lit L (fun L' => generalize_lit l L') (fun L' => generalize_trm t L')
    | trm_write v m v2 => generalize_trm_write L v m v2
    | trm_call v m v2 => generalize_trm_call L v m v2
    end with
  generalize_lit (l : lit) L {struct l} :  generalization L l :=
    match l with
    | lit_fun T t =>  generalize_lit_fun L T (fun L' => generalize_trm t L')
    | lit_obj T ds => generalize_lit_obj L T (fun L' => generalize_defs ds L') 
    end with
  generalize_def (d : def) L {struct d} : generalization L d :=
    match d with
    | def_typ B T => generalize_def_typ L B T
    | def_fld m e => generalize_def_fld L m e
    | def_met m e => generalize_def_met L m (fun L' => generalize_trm e L')
    end with
  generalize_defs  (ds : defs) L {struct ds} :  generalization L ds :=
    match ds with
    | defs_nil       => generalize_defs_nil L
    | defs_cons ds d => generalize_defs_cons L (fun L' => generalize_defs ds L') (fun L' => generalize_def d L')
    end.

Instance TrmGeneralizable: Generalizable TrmFreeVar TrmSubstVar := generalize_trm.