Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Import LibExtra DotTactics SyntaxClasses.
Require Import Vars TransformCollect Types Terms Envs.

(* ************************************************************************** *)
(** * Heap Items
  - [item_obj ds] represents the object with member definitions [ds]
  - [item_fun T t] represents a function [λ(x: T)t]; [x] is bound in
    [t] and is omitted; we will denote lambda terms as [λ(T)t]. *)
Inductive item : Set :=
  | item_obj: defs -> item
  | item_fun: trm -> item.


(* ************************************************************************** *)
(** ** Heap Items are Abstract Syntax *)
Definition fv_item (v: item) : vars :=
  match v with
  | item_obj ds => (fv ds)
  | item_fun t => (fv t)
  end.
Instance ItemFreeVar : FreeVar item := fv_item.

Definition open_rec_item (n: nat) (x: var) (v: item) : item :=
  match v with
  | item_obj ds => item_obj (open_rec n x ds)
  | item_fun t => item_fun (open_rec (S n) x t)
  end.
Instance ItemOpenable : Openable item := open_rec_item.

Definition subst_item x y (v: item) : item :=
  match v with
  | item_obj ds => item_obj (subst_var x y ds)
  | item_fun t => item_fun (subst_var x y t)
  end.
Instance ItemSubstVar : SubstVar item := subst_item.

Instance ItemAbstractSyntax :
  AbstractSyntax ItemOpenable ItemFreeVar ItemSubstVar := {}.

Definition seal_rec_item (n: nat) (x: var) (v: item) : item :=
  match v with
  | item_obj ds => item_obj (seal_rec n x ds)
  | item_fun t => item_fun (seal_rec (S n) x t)
  end.
Instance ItemSealable : Sealable item := seal_rec_item.

Definition collect_item {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (it: item) : C :=
  match it with
  | item_obj ds => (collect df f comb ds)
  | item_fun t => (collect df f comb t)
  end.
Instance ItemCollect : Collect AvarAbstractSyntax ItemAbstractSyntax :=
  @collect_item.
Definition transform_item (f : avar -> nat -> avar) (it : item) (n : nat) : item :=
  match it with
  | item_obj ds => (item_obj (transform f ds n))
  | item_fun t => (item_fun (transform f t (S n)))
  end.
Instance ItemTransform : Transform AvarAbstractSyntax ItemAbstractSyntax :=
  @transform_item.

(** *** Operations are lifted from operations on Variables *)
Lemma fv_collect_item :
  (forall (T : item), fv T = collect (\{}) fv (@union var) T).
Proof.
  intros.
  induction T; simpls; rewrite <- fv_collect; auto.
Qed.

Lemma open_transform_item :
  (forall (T : item) n x, (open_depth_rec n x) T = transform (open_depth_rec n x) T).
Proof.
  unfold open_depth_rec.
  intros.
  induction T.
  all: repeat (apply fun_ext_dep; intro).
  all: simpl.
  all: fequal.
  all: assert (R: (fun (a : avar) (d0 : nat) => open_rec (n + d0) x a) = open_depth_rec n x)
      by auto; rewrite R; clear R.
  all:
  rewrite <- open_transform.
  all: unfold open_depth_rec; auto.
  rewrite plus_n_Sm; auto.
Qed.

Lemma subst_transform_item :
  (forall (T : item) x y, (ignore_second (subst_var x y)) T
                = transform (ignore_second (subst_var x y)) T).
Proof.

  intros; induction T.
  all: simpls; intros; intros_fun_ext.
  all: simpl.
  all: rewrite <- subst_transform.
  all: unfold ignore_second.
  all: simpls.
  all: auto.
Qed.

Lemma seal_transform_item :
  (forall (T : item) n x, (seal_depth_rec n x) T = transform (seal_depth_rec n x) T).
Proof.
  intros. induction T.
  all: unfold seal_depth_rec.
  all: simpls; unfold ItemTransform, transform_item.
  all: repeat (apply fun_ext_dep; intro).
  all: apply f_equal.
  all: assert (R: (fun (a : avar) (d0 : nat) => seal_rec (n + d0) x a) = seal_depth_rec n x)
      by auto; rewrite R; clear R.

  all: rewrite <- seal_transform.
  all: unfold seal_depth_rec; auto.
  rewrite plus_n_Sm; auto.
Qed.


Local Ltac solve_transform_syntax :=
  hnf; intros; (apply fun_ext_dep; intro);
    try apply fv_collect_item;
    try apply open_transform_item;
    try apply subst_transform_item.

Instance ItemFvCollect : FvCollect ItemCollect.
Proof. solve_transform_syntax. Qed.
Instance ItemOpenTransform : OpenTransform ItemTransform.
Proof. solve_transform_syntax. Qed.
Instance ItemSealTransform : SealTransform ItemTransform AvarSealable ItemSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro).
  apply seal_transform_item.
Qed.
Instance ItemSubstTransform : SubstTransform ItemTransform.
Proof. solve_transform_syntax. Qed.

Lemma fv_subst_cases_item : forall x y,
    (forall (T : item), FvSubstCasesOf x y T).
Proof.
  intros. induction T.
  all: simpls.
  all: eapply fv_subst_cases.
Qed.
Instance ItemFvSubstCases: FvSubstCases ItemAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_item.
Qed.

Lemma subst_id_item (v1: var):
  (forall (t: item),
  subst_var v1 v1 t = t).
Proof.
  intros. induction t.
  all: simpls. all: fequal. all: apply subst_id.
Qed.
Instance ItemSubstId : SubstId ItemAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite subst_id_item. compute. auto.
Qed.

Lemma transform_id_item :
  (forall (T : item) n, transform (ignore_second id) T n = T).
Proof.
  unfold ignore_second, id, transform; simpls.
  intros. induction T; simpl; intros; try f_equal; auto;
  rewrite transform_id; auto.
Qed.

Lemma transform_compose_item:
  (forall (T : item) n f g, ((transform f) ∘∘ (transform g)) T n = transform (f ∘∘ g) T n).
Proof.
  unfold compose12, transform; simpls.
  intros; induction T; simpl; intros; fequal.
  all:    rewrite ? (fold_compose12 (transform f));
      rewrite ? transform_compose; auto.
Qed.

Local Ltac solve_transform_sanity :=
  hnf; try apply transform_id_item;
    intros; intros_fun_ext; apply transform_compose_item.

Instance ItemTransformId : TransformId ItemTransform.
Proof. solve_transform_sanity. Qed.

Instance ItemTransformCompose : TransformCompose ItemTransform.
Proof. solve_transform_sanity. Qed.


Lemma transform_shift_item (f : avar -> nat -> avar) :
  (forall (X : item) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n).
Proof.
  intros. induction X; simpl; intros; try f_equal; auto.
  rewrite transform_shift. auto.
  rewrite <- Nat.add_succ_l. rewrite transform_shift. auto.
Qed.

Instance ItemTransformShift : TransformShift ItemTransform.
Proof.
  intro. eapply transform_shift_item.
Qed.

Lemma collect_compose_item :
  forall {C D} (f : C -> D) g comb1 comb2 i,
    (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
    (forall (T : item),
        (f ∘ collect i g comb1) T = collect (f i) (f ∘ g) comb2 T).
Proof.
  intros.
  induction T;
    unfold compose, collect, ItemCollect.
  all: simpls.
  all: rewrite  ? (fold_compose f (collect (C:=C) _ _ _)).
  all: erewrite (collect_compose); eauto.
Qed.

Lemma collect_reflect_item:
  forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall bl,
      (forall T, ((collect_item (Is_true bl) P or T)
             <-> (collect_item bl p orb T) = true)).
Proof.
  intros P p H bl.
  intro T.
  induction T; simpl; auto 2 using reflect_iff.
  all: unfold iff; auto 3 using Is_true_eq_left, Is_true_eq_true.
  all: split.
  all: destruct_ands; intros.

  all: rewrite ? collect_reflect; eauto.
  all: rewrite <- collect_reflect; eauto.
Qed.

Local Ltac solve_collect_compose :=
  unfold compose; apply fun_ext_dep;
  apply collect_compose_item; auto.

Local Ltac solve_collect_reflect :=
  apply collect_reflect_item; auto.

Local Ltac solve_collect_sanity :=
  hnf; intros; solve [solve_collect_compose || solve_collect_reflect].

Instance ItemCollectCompose : CollectCompose ItemCollect.
Proof. solve_collect_sanity. Qed.
Instance ItemCollectReflect : CollectReflect ItemCollect.
Proof. solve_collect_sanity. Qed.

Ltac fold_item_collect :=
  change (@collect_item) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            item ItemOpenable ItemFreeVar ItemSubstVar ItemAbstractSyntax
            ItemCollect).
Ltac fold_item_transform :=
  change (@transform_item) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            item ItemOpenable ItemFreeVar ItemSubstVar ItemAbstractSyntax
            ItemTransform).
Local Ltac sympl_item_transform :=
  simpl; unfold ItemTransform, ItemCollect; fold_item_transform; fold_item_collect.

Lemma collect_transform_item:
  (forall (t : item) bl p f g,
      (collect bl p andb t) = true ->
      transform (dec_apply p f g) t = transform f t).
Proof.
  intro t.
  induction t; intros.
  all: unfold transform, ItemTransform, transform_item.
  all: rewrite (collect_transform bl); eauto.
Qed.

Instance ItemCollectTransform :
  CollectTransform ItemCollect ItemTransform.
Proof. hnf; intros; eapply collect_transform_item; eauto. Qed.

(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** A heap, maps locations to heap items. *)
Notation heap := (env item).

(* heap and env live on the type level, not the set level, so they can't be
   used for FreeVar *)
(* Note: explicit application of dom, to infer the correct type for Sigma *)
Notation fv_heap Sigma := ((@dom item Sigma) \u (fv_env Sigma)).

Definition subst_heap x y : heap -> heap := map (subst_var x y).

Lemma heap_push_obj_in_fv_r v1 v2 ds e:
  v2 \in fv ds ->
  v2 \in fv_env (e & (v1 ~ (item_obj ds))).
Proof.
  rewrite env_push_fv.
  rewrite in_union.
  right.
  unfold fv.
  unfold fv_in_values.
  unfold ItemFreeVar.
  unfold fv_item.
  auto.
Qed.

Lemma notin_fv_s_binds_defs y s (ds: defs) x:
y \notin fv_env s ->
binds x (item_obj ds) s ->
y \notin fv ds.
Proof.
  intros.
  induction s using env_ind.
  - false.
  eapply binds_empty_inv; eauto.
  -
  destruct (binds_push_inv H0) as [[? ?] | [? ?]]; subst.
  + intro. apply H. apply heap_push_obj_in_fv_r. auto.
  + apply IHs; auto. intro. apply H. apply env_push_in_fv_l. auto.
Qed.


Lemma heap_bind_in_fv y1 yd S2 x:
  x \in fv yd ->
  binds y1 (item_obj yd) S2 ->
  x \in fv_env S2.
Proof.
  intros.
  induction S2 using env_ind.
  - false. eapply binds_empty_inv. eauto.
  - rewrite env_push_fv.
    destruct (binds_push_inv H0) as [[? ?] | [? ?]].
    all: subst; rewrite in_union; eauto.
Qed.
Lemma heap_bind_in_fv_heap y1 yd S2 x:
  x \in fv yd ->
  binds y1 (item_obj yd) S2 ->
  x \in fv_heap S2.
Proof.
  intros.
  assert (x \in fv_env S2).
  eapply heap_bind_in_fv; eauto.
  rewrite in_union. eauto.
Qed.

Lemma in_fv_subst_heap Sigma x x1 x2 :
  x <> x1 -> (x \in fv_heap Sigma) ->
  x \in fv_heap (subst_heap x1 x2 Sigma).
Proof.
  introv Hne Hsub.
  rewrite in_union in *.
  unfold subst_heap at 1. rewrite dom_map.
  destruct Hsub as [Hsub | Hsub]; try tauto. right.
  eapply in_fv_env_subst; eauto.
Qed.

Lemma subset_fv_subst_heap Sigma X x1 x2 :
  x1 \notin X -> X \c fv_heap Sigma ->
  X \c fv_heap (subst_heap x1 x2 Sigma).
Proof.
  introv Hn Hsub. introv Hin. eapply in_fv_subst_heap; eauto.
  hnf; intros; subst; firstorder.
Qed.

Lemma item_subst_vars_obj ds xs ys:
  (item_obj ds) /[xs --> ys] = (item_obj (ds /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma item_subst_vars_fun t xs ys:
  (item_fun t) /[xs --> ys] = (item_fun (t /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Notation subheap_unchanged S1 S2 :=
  (forall y itm, binds y itm S1 -> binds y itm S2) (only parsing).

(* Extract all fields from a heap *)

Class Fields (A: Type) :=
  fields: A -> vars.
Definition fields_def d: vars :=
  match d with
    | def_typ A T => \{}
    | def_met a t => \{}
    | def_fld a (avar_f x) => \{x}
    | def_fld a (avar_b n) => \{}
  end.
Instance FieldsDef: Fields def := fields_def.
Fixpoint fields_defs ds: vars :=
  match ds with
    | defs_nil => \{}
    | defs_cons ds d => (fields_defs ds) \u (fields d)
  end.
Instance FieldsDefs: Fields defs := fields_defs.
Fixpoint fields_item it: vars :=
  match it with
    | item_fun f => \{}
    | item_obj ds => (fields ds)
  end.
Instance FieldsItem: Fields item := fields_item.
Fixpoint fields_heap (S: heap) : vars :=
  match S with
    | nil%list => \{}
    | ((y,ds)::E)%list => (fields_heap E) \u (fields ds)
  end.
Instance FieldsHeap: Fields heap := fields_heap.

Class FieldsInFv (A: Set) {AF: Fields A}{AV: FreeVar A} :=
  fields_in_fv (a: A): fields a \c fv a.

Lemma fields_open_subset_def x n (d : def) :
  fields (open_rec n x d) \c \{x} \u fields d.
Proof.
  unfold fields, FieldsDef, fields_def. destruct d; simpl;
    eauto using subset_empty_l.
  destruct a; simpl; eauto using subset_union_weak_r.
  destruct (classicT (n=n0)); eauto using subset_union_weak_r,
    subset_union_weak_l.
Qed.

Lemma fields_open_subset_defs n x (ds : defs) :
  fields (open_rec n x ds) \c \{x} \u fields ds.
Proof.
  unfold fields, FieldsDefs in *.
  induction ds; simpl; eauto using subset_empty_l.
  eapply subset_union_l. eapply subset_trans; try eassumption.
  rewrite union_assoc; eapply subset_union_weak_l.
  rewrite union_assoc.
  forwards: fields_open_subset_def. eapply subset_trans; eauto.
  introv Hin. repeat(rewrite in_union in *). tauto.
Qed.
