(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.
Require Import Logctx.


(** * A path through and/or type constructors *)
Inductive logpath: Set :=
| logpath_nil: logpath (* The hole *)
| logpath_orl (oc: logpath): logpath (* Hole on the left of a type *)
| logpath_orr (oc: logpath): logpath (* Hole on the right of a type *)
| logpath_andl (oc: logpath): logpath (* Hole on the left of a type *)
| logpath_andr (oc: logpath): logpath (* Hole on the right of a type *)
.

(* Finds a part of a type under a path.
   Such a type is unique, but not guaranteed to exist.
 *)
Inductive logpath_typ: typ -> logpath -> typ -> Prop :=
| logpath_typ_nil T:
  logpath_typ T logpath_nil T
| logpath_typ_orl oc T S U:
  logpath_typ T oc U ->
  logpath_typ (typ_or T S) (logpath_orl oc) U
| logpath_typ_orr oc T S U:
  logpath_typ T oc U ->
  logpath_typ (typ_or S T) (logpath_orr oc) U
| logpath_typ_andl oc T S U:
  logpath_typ T oc U ->
  logpath_typ (typ_and T S) (logpath_andl oc) U
| logpath_typ_andr oc T S U:
  logpath_typ T oc U ->
  logpath_typ (typ_and S T) (logpath_andr oc) U
.
Hint Constructors logpath_typ: core.

(* Finds a logic context around a path.
   Such a context is unique, but not guaranteed to exist.
 *)
Inductive logpath_ctx: typ -> logpath -> logctx -> Prop :=
| logpath_ctx_nil T:
  logpath_ctx T logpath_nil logctx_nil
| logpath_ctx_orl oc T S oc1:
  logpath_ctx T oc oc1 ->
  logpath_ctx (typ_or T S) (logpath_orl oc) (logctx_orl oc1 S)
| logpath_ctx_orr oc T S  oc1:
  logpath_ctx T oc  oc1 ->
  logpath_ctx (typ_or S T) (logpath_orr oc) (logctx_orr S oc1)
| logpath_ctx_andl oc T S  oc1:
  logpath_ctx T oc  oc1 ->
  logpath_ctx (typ_and T S) (logpath_andl oc) (logctx_andl oc1 S)
| logpath_ctx_andr oc T S oc1:
  logpath_ctx T oc  oc1 ->
  logpath_ctx (typ_and S T) (logpath_andr oc) (logctx_andr S oc1)
.
Hint Constructors logpath_ctx: core.

(* The type and context, found by a path, together form the original type *)
Lemma logpath_ctx_typ T oc oc1 U:
  logpath_typ T oc U ->
  logpath_ctx T oc oc1 ->
  logctx_typ oc1 U = T.
Proof.
  introv Htyp Hctx.
  reverts U oc1 T.
  induction oc; intros; inversions Htyp; inversions Hctx; simpls; eauto; fequal; eauto.
Qed.

(* The path around which a logic context is formed *)
Fixpoint logctx_path(oc: logctx): logpath := 
  match oc with
  | logctx_nil => logpath_nil
  | logctx_orl ocl Tr => logpath_orl (logctx_path ocl)
  | logctx_orr Tl ocr => logpath_orr (logctx_path ocr)
  | logctx_andl ocl Tr => logpath_andl (logctx_path ocl)
  | logctx_andr Tl ocr => logpath_andr (logctx_path ocr)
  end.

(* Consistency of the path around which a context is found *)
Lemma logpath_ctx_path T op oc:
  logpath_ctx T op oc ->
  logctx_path oc = op.
Proof.
  introv Hp.
  reverts oc T.
  induction op; intros; inversions Hp; simpls; fequal; eauto.
Qed.
(* The type found under a path is unique *)
Lemma logpath_typ_functional T op U1 U2:
  logpath_typ T op U1 ->
  logpath_typ T op U2 ->
  U1 = U2.
Proof.
  introv Htyp1 Htyp2.
  reverts T U1 U2.
  induction op; intros; inversions Htyp1; inversions Htyp2; eauto.
Qed.

(* The context found around a path is unique - this also follows from logpath_ctx_path *)
Lemma logpath_ctx_functional T op oc1 oc2:
  logpath_ctx T op oc1 ->
  logpath_ctx T op oc2 ->
  oc1 = oc2.
Proof.
  introv Htyp1 Htyp2.
  reverts T oc1 oc2.
  induction op; intros; inversions Htyp1; inversions Htyp2; fequal; eauto.
Qed.

Lemma logpath_ctx_typ_exists T op oc:
  logpath_ctx T op oc ->
  exists S,
    logpath_typ T op S.
Proof.
  introv Hctx.
  induction Hctx; destruct_all.
  - exists T. constructor; eauto.
  - exists x. constructor; eauto.
  - exists x. constructor; eauto.
  - exists x. constructor; eauto.
  - exists x. constructor; eauto.
Qed.
Lemma logpath_typ_ctx_exists T op S:
  logpath_typ T op S ->
  exists oc,
    logpath_ctx T op oc.
Proof.
  introv Htyp.
  induction Htyp; destruct_all.
  - exists logctx_nil. constructor; eauto.
  - exists (logctx_orl x S). constructor; eauto.
  - exists (logctx_orr S x). constructor; eauto.
  - exists (logctx_andl x S). constructor; eauto.
  - exists (logctx_andr S x). constructor; eauto.
Qed.

(* Extend a path downwards *)
Fixpoint logpath_extend (oc oci: logpath): logpath := 
  match oc with
  | logpath_nil => oci
  | logpath_orl ocl => logpath_orl (logpath_extend ocl oci)
  | logpath_orr ocr => logpath_orr (logpath_extend ocr oci)
  | logpath_andl ocl => logpath_andl (logpath_extend ocl oci)
  | logpath_andr ocr => logpath_andr  (logpath_extend ocr oci)
  end.
(* Consistency of logpath_extend with logctx_path *)
Lemma logctx_path_extend oc1 oc2:
  logctx_path (logctx_extend oc1 oc2) = logpath_extend (logctx_path oc1) (logctx_path oc2).
Proof.
  induction oc1; simpls; fequal; eauto.
Qed.
Lemma logpath_extend_nil_l oc:
  logpath_extend oc logpath_nil = oc.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.

(* If a type is formed by a type and context, then that type is found under the path of the context *)
Lemma logctx_path_typ oc T:
  logpath_typ (logctx_typ oc T) (logctx_path oc) T.
Proof.
  reverts T.
  induction oc; intros; simpls; constructor; eauto.
Qed.
(* If a type is formed by a type and context, then that context is found around the path of the context *)
Lemma logctx_path_ctx oc T:
  logpath_ctx (logctx_typ oc T) (logctx_path oc) oc.
Proof.
  reverts T.
  induction oc; intros; simpls; constructor; eauto.
Qed.

Lemma logctx_path_eq_inv oc1 oc2 S T:
  logctx_path oc1 = logctx_path oc2 ->
  logctx_typ oc1 S = logctx_typ oc2 T ->
  oc1 = oc2 /\ S = T.
Proof.
  intros Heqp Heqt.
  forwards: logctx_path_ctx oc1 S.
  forwards: logctx_path_ctx oc2 T.
  rewrite Heqp, Heqt in H.
  forwards: logpath_ctx_functional H H0; eauto.

  forwards: logctx_path_typ oc1 S.
  forwards: logctx_path_typ oc2 T.
  rewrite Heqp, Heqt in H2.
  forwards: logpath_typ_functional H2 H3; eauto.
Qed.
Lemma logpath_typ_extend_case T op1 op2 U:
  logpath_typ T (logpath_extend op1 op2) U ->
  exists S,
  logpath_typ T op1 S /\ logpath_typ S op2 U.
Proof.
  introv Ht.
  forwards: logpath_typ_ctx_exists Ht; eauto; destruct_all.
  reverts T U x.
  induction op1; intros; simpls.
  exists T; eauto.
  all: inversions Ht; inversions H; forwards: IHop1; eauto; destruct_all; eauto.
Qed.
Lemma logctx_path_extend_case oc op1 op2:
  logctx_path oc = logpath_extend op1 op2 ->
  exists oc1 oc2, oc = logctx_extend oc1 oc2 /\ logctx_path oc1 = op1 /\ logctx_path oc2 = op2.
Proof.
  introv Hoc. reverts oc op2.
  induction op1; intros; simpls.
  - exists logctx_nil oc; splits; simpls; auto.
  - destruct oc; inversions Hoc. forwards oc1 oc2 Hoc Hop1 Hop2: IHop1; eauto.
    subst.
    exists (logctx_orl oc1 T) oc2; splits; simpls; auto.
  - destruct oc; inversions Hoc. forwards oc1 oc2 Hoc Hop1 Hop2: IHop1; eauto.
    subst.
    exists (logctx_orr T oc1) oc2; splits; simpls; auto.
  - destruct oc; inversions Hoc. forwards oc1 oc2 Hoc Hop1 Hop2: IHop1; eauto.
    subst.
    exists (logctx_andl oc1 T) oc2; splits; simpls; auto.
  - destruct oc; inversions Hoc. forwards oc1 oc2 Hoc Hop1 Hop2: IHop1; eauto.
    subst.
    exists (logctx_andr T oc1) oc2; splits; simpls; auto.
Qed.

(* Two paths which are not prefixes of each other *)
Inductive logpath_separate: logpath->logpath->Prop :=
| logpath_separate_orl_e oc1 oc2: (* Common prefix left *)
  logpath_separate oc1 oc2 -> 
  logpath_separate (logpath_orl oc1) (logpath_orl oc2)
| logpath_separate_orr_e oc1 oc2: (* Common prefix right *)
  logpath_separate oc1 oc2 -> 
  logpath_separate (logpath_orr oc1) (logpath_orr oc2)
| logpath_separate_andl_e oc1 oc2: (* Common prefix left *)
  logpath_separate oc1 oc2 -> 
  logpath_separate (logpath_andl oc1) (logpath_andl oc2)
| logpath_separate_andr_e oc1 oc2: (* Common prefix right *)
  logpath_separate oc1 oc2 -> 
  logpath_separate (logpath_andr oc1) (logpath_andr oc2)
| logpath_separate_orl_n oc1 oc2:
  (forall oc3, oc2 <> (logpath_orl oc3) /\ oc2 <> logpath_nil) ->
  logpath_separate (logpath_orl oc1) oc2
| logpath_separate_orr_n oc1 oc2:
  (forall oc3, oc2 <> (logpath_orr oc3) /\ oc2 <> logpath_nil) ->
  logpath_separate (logpath_orr oc1) oc2
| logpath_separate_andl_n oc1 oc2:
  (forall oc3, oc2 <> (logpath_andl oc3) /\ oc2 <> logpath_nil) ->
  logpath_separate (logpath_andl oc1) oc2
| logpath_separate_andr_n oc1 oc2:
  (forall oc3, oc2 <> (logpath_andr oc3) /\ oc2 <> logpath_nil) ->
  logpath_separate (logpath_andr oc1) oc2
.

Hint Resolve logpath_separate_orl_e logpath_separate_orr_e
  logpath_separate_andl_e logpath_separate_andr_e: core.


(* Consistency of logpath_separate with logctx_separate and logctx_path *)
Lemma logctx_separate_path_separate oc1 oc2:
  logctx_separate oc1 oc2 ->
  logpath_separate (logctx_path oc1) (logctx_path oc2).
Proof.
  introv Hsep.
  induction Hsep; simpls; constructor; eauto.
  all: intros.
  destruct oc2; forwards: H oc1 Tr; try forwards: H oc2 T; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1 Tl; try forwards: H oc2 T; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1 Tr; try forwards: H oc2 T; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1 Tl; try forwards: H oc2 T; destruct_all; simpls; try contradiction; split; discriminate.
Qed.

(* Symmetry of logpath_separate *)
Lemma logpath_separate_symm op1 op2:
  logpath_separate op1 op2 ->  
  logpath_separate op2 op1.
Proof.
  introv Hsep.
  induction Hsep.
  - eapply logpath_separate_orl_e; eauto.
  - eapply logpath_separate_orr_e; eauto.
  - eapply logpath_separate_andl_e; eauto.
  - eapply logpath_separate_andr_e; eauto.
  - destruct oc2; try forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction.
    eapply logpath_separate_orr_n; intros; split; discriminate.
    eapply logpath_separate_andl_n; intros; split; discriminate.
    eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct oc2; try forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction.
    eapply logpath_separate_orl_n; intros; split; discriminate.
    eapply logpath_separate_andl_n; intros; split; discriminate.
    eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct oc2; try forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction.
    eapply logpath_separate_orl_n; intros; split; discriminate.
    eapply logpath_separate_orr_n; intros; split; discriminate.
    eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct oc2; try forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction.
    eapply logpath_separate_orl_n; intros; split; discriminate.
    eapply logpath_separate_orr_n; intros; split; discriminate.
    eapply logpath_separate_andl_n; intros; split; discriminate.
Qed.

(* Separate paths can be extended and still will be separate *)
Lemma logpath_separate_extend op1 op2 op3:
  logpath_separate op1 op2 ->  
  logpath_separate op1 (logpath_extend op2 op3).
Proof.
  introv Hsep.
  induction Hsep;
  simpls;
  constructor; eauto.
  destruct oc2; forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction; split; discriminate.
  destruct oc2; forwards: H oc1; try forwards: H oc2; destruct_all; simpls; try contradiction; split; discriminate.
Qed.
Lemma logpath_separate_extend_left op1 op2 op3:
  logpath_separate op1 op2 ->  
  logpath_separate (logpath_extend op3 op1) (logpath_extend op3 op2).
Proof.
  introv Hsep.
  induction op3;
  simpls; eauto;
  constructor; eauto.
Qed.
Lemma logpath_separate_extend_inv op1 op2 op3:
  logpath_separate (logpath_extend op3 op1) (logpath_extend op3 op2) ->
  logpath_separate op1 op2.
Proof.
  introv Hsep.
  induction op3; simpls; eauto;
  inversions Hsep; eauto;
  forwards: H0 (logpath_extend op3 op2); destruct_all; try contradiction.
Qed.


(* If a type is formed from a context and a type,
   then types found under paths separate from that context
   will not be affected by changing the type under the context *)
Lemma logpath_logctx_typ_separate_preserve op oc T S U:
  logpath_separate op (logctx_path oc) ->
  logpath_typ (logctx_typ oc T) op S ->
  logpath_typ (logctx_typ oc U) op S.
Proof.
  introv Hsep Htyp.
  remember (logctx_path oc).
  reverts oc T S.
  induction Hsep; intros; subst.
  - inversions Htyp.
    destruct oc; simpls; inversions H; inversions Heql.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H; inversions Heql.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H; inversions Heql.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H; inversions Heql.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H0.
    forwards: H oc1; destruct_all; try contradiction.
    forwards: H (logctx_path oc); destruct_all; try contradiction.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H0.
    forwards: H oc1; destruct_all; try contradiction.
    constructor; eauto.
    forwards: H (logctx_path oc); destruct_all; try contradiction.
  - inversions Htyp.
    destruct oc; simpls; inversions H0.
    forwards: H oc1; destruct_all; try contradiction.
    forwards: H (logctx_path oc); destruct_all; try contradiction.
    constructor; eauto.
  - inversions Htyp.
    destruct oc; simpls; inversions H0.
    forwards: H oc1; destruct_all; try contradiction.
    constructor; eauto.
    forwards: H (logctx_path oc); destruct_all; try contradiction.
Qed.

(* 
This does not hold for any logctx_path,
because separate still requires the types on the shared path to be the same,
which is not guaranteed by just the path.
Therefore, logpath_ctx is required.
*)
Lemma logctx_path_separate T op1 op2 oc1 oc2:
  logpath_ctx T op1 oc1 ->
  logpath_ctx T op2 oc2 ->
  logpath_separate op1 op2 ->
  logctx_separate oc1 oc2.
Proof.
  introv Ho1 Ho2 Hsep.
  reverts oc1 oc2 T.
  induction Hsep; intros.
  - inversions Ho1; inversions Ho2. constructor; eauto.
  - inversions Ho1; inversions Ho2. constructor; eauto.
  - inversions Ho1; inversions Ho2. constructor; eauto.
  - inversions Ho1; inversions Ho2. constructor; eauto.
  - inversions Ho1. inversions Ho2; forwards: H oc1; try forwards: H oc; destruct_all; try contradiction. constructor; intros; split; discriminate.
  - inversions Ho1. inversions Ho2; forwards: H oc1; try forwards: H oc; destruct_all; try contradiction. constructor; intros; split; discriminate.
  - inversions Ho1. inversions Ho2; forwards: H oc1; try forwards: H oc; destruct_all; try contradiction. constructor; intros; split; discriminate.
  - inversions Ho1. inversions Ho2; forwards: H oc1; try forwards: H oc; destruct_all; try contradiction. constructor; intros; split; discriminate.
Qed.

Lemma logpath_separate_auto OS OT ocS ocT:
  (OS = logpath_orl \/ OS = logpath_orr \/ OS = logpath_andl \/ OS = logpath_andr) ->
  (OT = logpath_orl \/ OT = logpath_orr \/ OT = logpath_andl \/ OT = logpath_andr) ->
  (OT = OS -> logpath_separate ocS ocT) ->
  (logpath_separate (OS ocS) (OT ocT)).
Proof.
  introv HS HT Hsep.
  destruct_all; subst.
  all: constructor; eauto 2.
  all: intros; split; discriminate.
Qed.


Lemma logpath_intersect_cases ocS ocT:
  (ocS = ocT) \/
  (exists ocST, ocT = logpath_extend ocS ocST) \/
  (exists ocTS, ocS = logpath_extend ocT ocTS) \/
  logpath_separate ocS ocT.
Proof.
  reverts ocS.
  induction ocT; intros; simpls.
  - right. right. left. exists ocS; eauto.
  - destruct ocS; simpls; subst.
    + right. left.
      forwards HocST': IHocT logpath_nil; eauto.
    + (* Agreeing *) 
      forwards HocST: IHocT ocS; eauto.
      destruct_all; subst; eauto 6 using logpath_separate_orl_e.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orr_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andl_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT logpath_nil; eauto.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orl_n; intros; split; discriminate.
    + (* Agreeing *)
      forwards HocST: IHocT ocS; eauto.
      destruct_all; subst; eauto 6 using logpath_separate_orr_e.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andl_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT logpath_nil; eauto.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orl_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orr_n; intros; split; discriminate.
    + (* Agreeing *) 
      forwards HocST: IHocT ocS; eauto.
      destruct_all; subst; eauto 6 using logpath_separate_andl_e.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andr_n; intros; split; discriminate.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT logpath_nil; eauto.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orl_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_orr_n; intros; split; discriminate.
    + (* Disagreeing *)
      right. right. right.
      eapply logpath_separate_andl_n; intros; split; discriminate.
    + (* Agreeing *)
      forwards HocST: IHocT ocS; eauto.
      destruct_all; subst; eauto 6 using logpath_separate_andr_e.
Qed.

Lemma logpath_intersect_cases_nonnil ocS ocT:
  (ocS = ocT) \/
  (exists ocST, ocT = logpath_extend ocS ocST /\ ocST <> logpath_nil) \/
  (exists ocTS, ocS = logpath_extend ocT ocTS /\ ocTS <> logpath_nil) \/
  logpath_separate ocS ocT.
Proof.
  forwards: logpath_intersect_cases.
  destruct_all; eauto.
  -
  assert  (x = logpath_nil \/ x <> logpath_nil) as Hx. {
    destruct x; auto; right; discriminate.
  }
  destruct Hx; subst; simpls; rewrite ? logpath_extend_nil_l; auto ; right; left; eauto.
  -
  assert  (x = logpath_nil \/ x <> logpath_nil) as Hx. {
    destruct x; auto; right; discriminate.
  }
  destruct Hx; subst; simpls; rewrite ? logpath_extend_nil_l; auto ; right;  right ;left; eauto.
Qed.

Lemma logpath_separate_eq_false op:
  logpath_separate op op -> False.
Proof.
  introv Hsep.
  induction op; inversions Hsep; eauto.
  - lets: H0 op; destruct_all; contradiction.
  - lets: H0 op; destruct_all; contradiction.
  - lets: H0 op; destruct_all; contradiction.
  - lets: H0 op; destruct_all; contradiction.
Qed.
Lemma logpath_separate_extend_false op op2:
  logpath_separate op (logpath_extend op op2) -> False.
Proof.
  introv Hsep.
  induction op; inversions Hsep; eauto.
  all: lets: H0 (logpath_extend op op2); destruct_all; simpls; contradiction.
Qed.
Lemma logctx_follow_path_eq oc1 oc2:
  logctx_follow oc1 oc2 ->
  (logctx_path oc1) = (logctx_path oc2).
Proof.
  introv Hf.
  induction Hf; simpls; try rewrite IHHf; eauto.
Qed.
Lemma logctx_path_eq_follow oc1 oc2:
  (logctx_path oc1) = (logctx_path oc2) ->
  logctx_follow oc1 oc2.
Proof.
  introv Hf.
  reverts oc2.
  induction oc1; intros; simpls; destruct oc2; simpls; inversions Hf; try constructor; forwards: IHoc1; eauto.
Qed.

Inductive logpath_andor := logpath_and | logpath_or.
Definition logpath_andor_l TB op :=
  match TB with 
    | logpath_and => logpath_andl op
    | logpath_or => logpath_orl op
  end.
Definition logpath_andor_r TB op :=
  match TB with 
    | logpath_and => logpath_andr op
    | logpath_or => logpath_orr op
  end.
Definition logpath_andor_typ TB :=
  match TB with 
    | logpath_and => typ_and
    | logpath_or => typ_or
  end.
Definition logpath_andor_ctx_r TB Tl op :=
  match TB with 
    | logpath_and => logctx_andr Tl op
    | logpath_or => logctx_orr Tl op
  end.
Definition logpath_andor_ctx_l TB op Tr :=
  match TB with 
    | logpath_and => logctx_andl op Tr
    | logpath_or => logctx_orl op Tr
  end.

Lemma typ_andor_logpath_andor TB:
  typ_andor TB ->
  exists TB0,
    TB = logpath_andor_typ TB0.
Proof.
  introv HTB.
  inversions HTB.
  - exists logpath_and; simpls; auto.
  - exists logpath_or; simpls; auto.
Qed.

Lemma logpath_andor_typ_r TB U1 U2:
  (logctx_typ (logpath_andor_ctx_r TB U1 logctx_nil) U2) = (logpath_andor_typ TB U1 U2).
Proof.
  destruct TB; simpls; auto.
Qed.

Lemma logpath_andor_typ_l TB U1 U2:
  (logctx_typ (logpath_andor_ctx_l TB logctx_nil U2 ) U1) = (logpath_andor_typ TB U1 U2).
Proof.
  destruct TB; simpls; auto.
Qed.

Lemma logpath_andor_path_l TB U2:
  (logctx_path (logpath_andor_ctx_l TB logctx_nil U2 )) = (logpath_andor_l TB logpath_nil).
Proof.
  destruct TB; simpls; auto.
Qed.
Lemma logpath_andor_path_r TB U2:
  (logctx_path (logpath_andor_ctx_r TB U2 logctx_nil  )) = (logpath_andor_r TB logpath_nil).
Proof.
  destruct TB; simpls; auto.
Qed.
Lemma logpath_andor_typ_r' TB U1 U2 oc:
  (logctx_typ (logpath_andor_ctx_r TB U1 oc) U2) = (logpath_andor_typ TB U1 (logctx_typ oc U2)).
Proof.
  destruct TB; simpls; auto.
Qed.

Lemma logpath_andor_typ_l' TB U1 U2 oc:
  (logctx_typ (logpath_andor_ctx_l TB oc U2 ) U1) = (logpath_andor_typ TB (logctx_typ oc U1) U2).
Proof.
  destruct TB; simpls; auto.
Qed.

Lemma logpath_andor_path_l' TB U2 oc:
  (logctx_path (logpath_andor_ctx_l TB oc U2 )) = (logpath_andor_l TB (logctx_path oc)).
Proof.
  destruct TB; simpls; auto.
Qed.
Lemma logpath_andor_path_r' TB U2 oc:
  (logctx_path (logpath_andor_ctx_r TB U2 oc  )) = (logpath_andor_r TB (logctx_path oc)).
Proof.
  destruct TB; simpls; auto.
Qed.

Lemma logctx_follow_logpath_andor_l ocT TB oc S:
  logctx_follow ocT (logpath_andor_ctx_l TB oc S) ->
  exists oc2 S2,
    ocT = (logpath_andor_ctx_l TB oc2 S2) /\ logctx_follow oc2 oc.
Proof.
  introv Hf.
  destruct TB; simpls; inversions Hf; eauto.
Qed.
Lemma logctx_follow_logpath_andor_r ocT TB oc S:
  logctx_follow ocT (logpath_andor_ctx_r TB S oc) ->
  exists oc2 S2,
    ocT = (logpath_andor_ctx_r TB S2 oc2) /\ logctx_follow oc2 oc.
Proof.
  introv Hf.
  destruct TB; simpls; inversions Hf; eauto.
Qed.

Lemma logctx_andor_typ_inv TBS Sl Sr TBT Tl Tr:
  logpath_andor_typ TBT Tl Tr = logpath_andor_typ TBS Sl Sr ->
  TBT = TBS /\
  Tl = Sl /\ Tr = Sr.
Proof.
  destruct TBS; destruct TBT; introv H; simpls; inversions H; eauto.
Qed.


Lemma logctx_andor_ind:
    forall P : logctx -> Prop,
    P logctx_nil ->
    (forall TB (oc : logctx), P oc -> forall T : typ, P (logpath_andor_ctx_l TB oc T)) ->
    (forall TB (oc : logctx), P oc -> forall T : typ, P (logpath_andor_ctx_r TB T oc)) ->
    forall l : logctx, P l.
Proof.
  intros.
  induction l; auto 1;
  lets: H0 logpath_and;
  lets: H0 logpath_or;
  lets: H1 logpath_and;
  lets: H1 logpath_or;
  eauto.
Qed.
