(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.
Require Import Logctx OrctxLogctx Logpath.

(** * Ordering of paths through intersections and unions in a type *)

(* First argument is left or above second.

This orders the paths in DFS, prefix, left-to-right order.
logpath_nil is the lowest element,
left is lower than right.

This is not a total ordering - paths that disagree in the logical operator are not comparable.
Restricted to just ors or just ands, it is total.

This is a non-strict ordering.
 *)
Inductive logpath_le: logpath -> logpath -> Prop :=
| logpath_le_nil_l op:
    logpath_le logpath_nil op
| logpath_le_and opl opr:
    logpath_le (logpath_andl opl) (logpath_andr opr)
| logpath_le_or opl opr:
    logpath_le (logpath_orl opl) (logpath_orr opr)
| logpath_le_andl opl opr:
    logpath_le opl opr ->
    logpath_le (logpath_andl opl) (logpath_andl opr)
| logpath_le_andr opl opr:
    logpath_le opl opr ->
    logpath_le (logpath_andr opl) (logpath_andr opr)
| logpath_le_orl opl opr:
    logpath_le opl opr ->
    logpath_le (logpath_orl opl) (logpath_orl opr)
| logpath_le_orr opl opr:
    logpath_le opl opr ->
    logpath_le (logpath_orr opl) (logpath_orr opr)
.
Hint Constructors logpath_le: core. 

Lemma logpath_le_refl op:
  logpath_le op op.
Proof.
  induction op; eauto.
Qed.
Hint Resolve logpath_le_refl: core.

Lemma logpath_le_antisymm opl opr:
  logpath_le opl opr ->
  logpath_le opr opl ->
  opl = opr.
Proof.
  introv Hlel Hler.
  induction Hlel; inversions Hler; fequal;
  eauto.
Qed.

Lemma logpath_le_trans opl op opr:
  logpath_le opl op ->
  logpath_le op opr ->
  logpath_le opl opr.
Proof.
  introv Hlel Hler.
  reverts opr.
  induction Hlel; intros; inversions Hler;
  eauto.
Qed.

Lemma logpath_le_extend_l opl opr op:
  logpath_le opl opr ->
  logpath_le (logpath_extend op opl) (logpath_extend op opr).
Proof.
  introv Hle.
  induction op; simpls; eauto.
Qed.

Lemma logpath_le_extend_r_case opl opr:
  logpath_le opl opr ->
  (exists oplr, opr = (logpath_extend opr oplr)) \/
  (forall opl2 opr2, logpath_le (logpath_extend opl opl2) (logpath_extend opr opr2)).
Proof.
  introv Hle.
  induction Hle; intros; simpls; eauto.
  - left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
Qed.

Lemma logpath_le_extend_nil_l opl opr:
  logpath_le opl (logpath_extend opl opr).
Proof.
  forwards: logpath_le_extend_l logpath_nil opr opl; eauto.
  rewrite logpath_extend_nil_l in H. auto.
Qed.

Lemma logpath_extend_nil_invr opl opr:
  opl = (logpath_extend opl opr) ->
  opr = logpath_nil.
Proof.
  introv He.
  induction opl; simpls; try inversions He; auto.
Qed.

Lemma logpath_le_extend_nil_inv opl opr:
  logpath_le (logpath_extend opl opr) opl ->
  opr = logpath_nil.
Proof.
  introv Hle.
  forwards: logpath_le_extend_nil_l opl opr.
  forwards: logpath_le_antisymm; eauto.
  eapply logpath_extend_nil_invr; eauto.
Qed.

Lemma logpath_extend_overlap op1 op2 oc1 oc2:
  (logpath_extend op1 op2) = (logpath_extend oc1 oc2) ->
  (exists op1c2, op1 = logpath_extend oc1 op1c2 /\ oc2 = logpath_extend op1c2 op2) \/
  (exists op2c1, op2 = logpath_extend op2c1 oc2 /\ oc1 = logpath_extend op1 op2c1).
Proof.
  introv Heq.
  reverts oc1 oc2 op2.
  induction op1; intros; simpls; eauto 4; destruct oc1; simpls;
  try discriminate; eauto;
  inversions Heq; forwards: IHop1; eauto; destruct_all; subst; eauto.
Qed.
Lemma logpath_extend_nil_inv op1 op2:
  logpath_extend op1 op2 = logpath_nil ->
  op1 = logpath_nil /\ op2 = logpath_nil.
Proof.
  introv Heq.
  reverts op2;
  induction op1; intros; simpls; subst; eauto; discriminate.
Qed.

(* First argument is right or above seocnd *)
Inductive logpath_re: logpath -> logpath -> Prop :=
| logpath_re_nil_l op:
    logpath_re logpath_nil op
| logpath_re_and opl opr:
    logpath_re (logpath_andr opl) (logpath_andl opr)
| logpath_re_or opl opr:
    logpath_re (logpath_orr opl) (logpath_orl opr)
| logpath_re_andl opl opr:
    logpath_re opl opr ->
    logpath_re (logpath_andl opl) (logpath_andl opr)
| logpath_re_andr opl opr:
    logpath_re opl opr ->
    logpath_re (logpath_andr opl) (logpath_andr opr)
| logpath_re_orl opl opr:
    logpath_re opl opr ->
    logpath_re (logpath_orl opl) (logpath_orl opr)
| logpath_re_orr opl opr:
    logpath_re opl opr ->
    logpath_re (logpath_orr opl) (logpath_orr opr)
.
Hint Constructors logpath_re: core. 

Lemma logpath_re_refl op:
  logpath_re op op.
Proof.
  induction op; eauto.
Qed.
Hint Resolve logpath_re_refl: core.

Lemma logpath_re_antisymm opl opr:
  logpath_re opl opr ->
  logpath_re opr opl ->
  opl = opr.
Proof.
  introv Hlel Hler.
  induction Hlel; inversions Hler; fequal;
  eauto.
Qed.

Lemma logpath_re_trans opl op opr:
  logpath_re opl op ->
  logpath_re op opr ->
  logpath_re opl opr.
Proof.
  introv Hlel Hler.
  reverts opr.
  induction Hlel; intros; inversions Hler;
  eauto.
Qed.

Lemma logpath_re_extend_l opl opr op:
  logpath_re opl opr ->
  logpath_re (logpath_extend op opl) (logpath_extend op opr).
Proof.
  introv Hle.
  induction op; simpls; eauto.
Qed.

Lemma logpath_re_extend_r_case opl opr:
  logpath_re opl opr ->
  (exists oplr, opr = (logpath_extend opr oplr)) \/
  (forall opl2 opr2, logpath_re (logpath_extend opl opl2) (logpath_extend opr opr2)).
Proof.
  introv Hle.
  induction Hle; intros; simpls; eauto.
  - left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
  - destruct_all; subst.
    + left. exists logpath_nil; simpls. rewrite logpath_extend_nil_l. auto.
    + right. eauto.
Qed.

Lemma logpath_re_extend_nil_l opl opr:
  logpath_re opl (logpath_extend opl opr).
Proof.
  forwards: logpath_re_extend_l logpath_nil opr opl; eauto.
  rewrite logpath_extend_nil_l in H. auto.
Qed.


Lemma logpath_re_extend_nil_inv opl opr:
  logpath_re (logpath_extend opl opr) opl ->
  opr = logpath_nil.
Proof.
  introv Hle.
  forwards: logpath_re_extend_nil_l opl opr.
  forwards: logpath_re_antisymm; eauto.
  eapply logpath_extend_nil_invr; eauto.
Qed.

(* When opl is shorter or left and shorter or right, then it is shorter. *)
Lemma logpath_le_re_prefix opl opr:
  logpath_le opl opr ->
  logpath_re opl opr ->
  exists oplr, opr = logpath_extend opl oplr.
Proof.
  introv Hle Hre.
  induction Hle; inversions Hre; simpls; eauto; forwards: IHHle; destruct_all; subst; eauto.
Qed.

Lemma lopath_le_separate_re opl opr:
  logpath_le opl opr ->
  logpath_separate opl opr ->
  logpath_re opr opl.
Proof.
  introv Hle Hse.
  induction Hle; inversions Hse; econstructor; eauto; false (proj21 (H0 opr)); auto.
Qed.
Lemma lopath_re_separate_le opl opr:
  logpath_re opl opr ->
  logpath_separate opl opr ->
  logpath_le opr opl.
Proof.
  introv Hre Hse.
  induction Hre; inversions Hse; econstructor; eauto; false (proj21 (H0 opr)); auto.
Qed.

Lemma logpath_le_re_separate opl opr:
  logpath_le opl opr ->
  logpath_re opr opl ->
  opl = opr \/ logpath_separate opl opr.
Proof.
  introv Hle Hre.
  induction Hle; inversions Hre; simpls. eauto.
  all: try (right; constructor; intros; splits; discriminate).
  all: forwards: IHHle; destruct_all; subst; eauto.
Qed.
Lemma logpath_le_sep_or_extend ocl ocr:
  logpath_le ocl ocr ->
  logpath_separate ocl ocr \/ (exists ocp, ocr = logpath_extend ocl ocp).
Proof.
  introv Hle.
  induction Hle.
  - right. simpl. eauto.
  - left. constructor; intros; splits; discriminate.
  - left. constructor; intros; splits; discriminate.
  - destructecs IHHle.
    + left. constructor; eauto.
    + subst. right. exists ocp. simpl. auto.
  - destructecs IHHle.
    + left. constructor; eauto.
    + subst. right. exists ocp. simpl. auto.
  - destructecs IHHle.
    + left. constructor; eauto.
    + subst. right. exists ocp. simpl. auto.
  - destructecs IHHle.
    + left. constructor; eauto.
    + subst. right. exists ocp. simpl. auto.
Qed.

(* Ordering of contexts *)
Definition logctx_le (ocl ocr: logctx) := logpath_le (logctx_path ocl) (logctx_path ocr).
Definition logctx_re (ocl ocr: logctx) := logpath_re (logctx_path ocl) (logctx_path ocr).

Lemma logctx_le_cases ocl ocr Tl Tr:
  logctx_typ ocl Tl = logctx_typ ocr Tr ->
  logctx_le ocl ocr \/ logctx_le ocr ocl.
Proof.
  introv Heq. unfold logctx_le.
  reverts Tl Tr ocr.
  induction ocl; intros; simpls; subst. auto.
  all: destruct ocr; simpls; inversions Heq; eauto;
    forwards Hocl: IHocl; destruct_all; eauto.
Qed.

Lemma logctx_re_cases ocl ocr Tl Tr:
  logctx_typ ocl Tl = logctx_typ ocr Tr ->
  logctx_re ocl ocr \/ logctx_re ocr ocl.
Proof.
  introv Heq. unfold logctx_re.
  reverts Tl Tr ocr.
  induction ocl; intros; simpls; subst. auto.
  all: destruct ocr; simpls; inversions Heq; eauto;
    forwards Hocl: IHocl; destruct_all; eauto.
Qed.

Lemma logctx_le_nil oc:
  logctx_le logctx_nil oc.
Proof.
  unfold logctx_le.
  eauto.
Qed.

Lemma logctx_le_refl oc:
  logctx_le oc oc.
Proof.
  unfold logctx_le.
  eauto.
Qed.

Lemma logctx_le_trans ocl oc ocr:
  logctx_le ocl oc ->
  logctx_le oc ocr ->
  logctx_le ocl ocr.
Proof.
  introv Hlel Hler.
  unfold logctx_le in *.
  eapply logpath_le_trans; eauto.
Qed.

Lemma logctx_le_orl oc1 oc2 T1 T2:
  logctx_le oc1 oc2 ->
  logctx_le (logctx_orl oc1 T1) (logctx_orl oc2 T2).
Proof.
  unfold logctx_le in *. simpls.
    eapply logpath_le_orl.
Qed.
Lemma logctx_le_orr oc1 oc2 T1 T2:
  logctx_le oc1 oc2 ->
  logctx_le (logctx_orr T1 oc1) (logctx_orr T2 oc2).
Proof.
  unfold logctx_le in *. simpls.
    eapply logpath_le_orr.
Qed.
Lemma logctx_le_andl oc1 oc2 T1 T2:
  logctx_le oc1 oc2 ->
  logctx_le (logctx_andl oc1 T1) (logctx_andl oc2 T2).
Proof.
  unfold logctx_le in *. simpls.
    eapply logpath_le_andl.
Qed.
Lemma logctx_le_andr oc1 oc2 T1 T2:
  logctx_le oc1 oc2 ->
  logctx_le (logctx_andr T1 oc1) (logctx_andr T2 oc2).
Proof.
  unfold logctx_le in *. simpls.
    eapply logpath_le_andr.
Qed.
Lemma logctx_le_nil_inv oc:
  logctx_le oc logctx_nil ->
  oc = logctx_nil.
Proof.
  introv Hle.
  hnf in *; unfold logctx_path in *; destruct oc; simpls; auto; inversion Hle.
Qed.
Lemma logctx_le_extend_inv ocp oc1 oc2:
  logctx_le (logctx_extend ocp oc1) (logctx_extend ocp oc2) ->
  logctx_le oc1 oc2.
Proof.
  introv Hle.
  induction ocp; simpls; eauto; inversions Hle; eauto.
Qed.

Lemma logctx_le_extend_l_inv oc1 oc2:
  logctx_le (logctx_extend oc1 oc2) oc1 ->
  oc2 = logctx_nil.
Proof.
  introv Hle.
  forwards: logctx_le_extend_inv oc1 oc2 logctx_nil.
  rewrite logctx_extend_nil_l. eauto.
  eapply logctx_le_nil_inv; eauto.
Qed.
Lemma logctx_le_extend_l ocp oc1 oc2:
  logctx_le oc1 oc2 ->
  logctx_le (logctx_extend ocp oc1) (logctx_extend ocp oc2).
Proof.
  introv Hle.
  unfold logctx_le in *.
  rewrite ! logctx_path_extend.
  eapply logpath_le_extend_l. eauto.
Qed.

Lemma logctx_follow_le ocl ocr:
  logctx_follow ocl ocr ->
  logctx_le ocl ocr.
Proof.
  introv Hf.
  unfold logctx_le.
  eapply logctx_follow_path_eq in Hf. rewrite Hf.
  eapply logpath_le_refl.
Qed.

Lemma logctx_re_nil oc:
  logctx_re logctx_nil oc.
Proof.
  unfold logctx_re.
  eauto.
Qed.

Lemma logctx_re_refl oc:
  logctx_re oc oc.
Proof.
  unfold logctx_re.
  eauto.
Qed.


Lemma logctx_re_trans ocl oc ocr:
  logctx_re ocl oc ->
  logctx_re oc ocr ->
  logctx_re ocl ocr.
Proof.
  introv Hlel Hler.
  unfold logctx_re in *.
  eapply logpath_re_trans; eauto.
Qed.

Lemma logctx_follow_re ocl ocr:
  logctx_follow ocl ocr ->
  logctx_re ocl ocr.
Proof.
  introv Hf.
  unfold logctx_re.
  eapply logctx_follow_path_eq in Hf. rewrite Hf.
  eapply logpath_re_refl.
Qed.
Lemma logctx_re_extend_l ocp oc1 oc2:
  logctx_re oc1 oc2 ->
  logctx_re (logctx_extend ocp oc1) (logctx_extend ocp oc2).
Proof.
  introv Hle.
  unfold logctx_re in *.
  rewrite ! logctx_path_extend.
  eapply logpath_re_extend_l. eauto.
Qed.


(* Prefers shorter, left, or, first *)
Fixpoint logctx_leb (ocl ocr: logctx): logctx :=
match ocl, ocr with
| logctx_nil, _ => logctx_nil
| logctx_orl oc T, logctx_nil => logctx_nil
| logctx_orl oc T, logctx_orl ocp U => logctx_orl (logctx_leb oc ocp) T
| logctx_orl oc T, _ => logctx_orl oc T
| logctx_andl oc T, logctx_nil => logctx_nil
| logctx_andl oc T, logctx_orl ocp U => logctx_orl ocp U
| logctx_andl oc T, logctx_andl ocp U => logctx_andl (logctx_leb oc ocp) T
| logctx_andl oc T, _ => logctx_andl oc T
| logctx_orr T oc, logctx_nil => logctx_nil
| logctx_orr T oc, logctx_orl ocp U => logctx_orl ocp U
| logctx_orr T oc, logctx_andl ocp U => logctx_andl ocp U
| logctx_orr T oc, logctx_orr U ocp => logctx_orr T (logctx_leb oc ocp)
| logctx_orr T oc, _ => logctx_orr T oc
| logctx_andr T oc, logctx_nil => logctx_nil
| logctx_andr T oc, logctx_orl ocp U => logctx_orl ocp U
| logctx_andr T oc, logctx_andl ocp U => logctx_andl ocp U
| logctx_andr T oc, logctx_orr U ocp => logctx_orr U ocp
| logctx_andr T oc, logctx_andr U ocp => logctx_andr T (logctx_leb oc ocp)
end.

Lemma logctx_leb_le_l ocl ocr Tl Tr:
  logctx_typ ocl Tl = logctx_typ ocr Tr ->
  logctx_le (logctx_leb ocl ocr) ocl.
Proof.
  introv He.
  unfold logctx_le.
  reverts ocr Tl Tr.
  induction ocl; intros; simpls.
  - eapply logpath_le_refl.
  - destruct ocr; unfold logctx_path; simpls; inversions He; eauto.
  - destruct ocr; unfold logctx_path; simpls; inversions He; eauto.
  - destruct ocr; unfold logctx_path; simpls; inversions He; eauto.
  - destruct ocr; unfold logctx_path; simpls; inversions He; eauto.
Qed.

Lemma logctx_leb_nil_l oc:
  logctx_leb logctx_nil oc = logctx_nil.
Proof.
  simpls. auto.
Qed.
Lemma logctx_leb_nil_r oc:
  logctx_leb oc logctx_nil = logctx_nil.
Proof.
  destruct oc; simpls; auto.
Qed.

Lemma logctx_leb_le_comm ocl ocr Tl Tr:
  logctx_typ ocl Tl = logctx_typ ocr Tr ->
  logctx_le (logctx_leb ocl ocr) (logctx_leb ocr ocl).
Proof.
  introv He.
  reverts ocr Tl Tr; induction ocl; intros; destruct ocr; unfold logctx_le, logctx_path in *; simpls; inversions He; simpls; eauto using logpath_le_refl.
Qed.

Lemma logctx_leb_le_r ocl ocr Tl Tr:
  logctx_typ ocl Tl = logctx_typ ocr Tr ->
  logctx_le (logctx_leb ocl ocr) ocr.
Proof.
  introv He.
  eapply logctx_le_trans. eapply logctx_leb_le_comm. eauto. eapply logctx_leb_le_l.
  eauto.
Qed.


Lemma logctx_leb_le_b oc ocl ocr:
  logctx_le oc ocl ->
  logctx_le oc ocr ->
  logctx_le oc (logctx_leb ocl ocr).
Proof.
  unfold logctx_le.

  introv Hl Hr.
  reverts ocr ocl.
    induction oc; intros; unfold logctx_path in *; simpls; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
Qed.

Lemma logctx_leb_l oc1 oc2:
  logctx_le oc1 oc2 ->
  logctx_leb oc1 oc2 = oc1.
Proof.
  intros.
  unfold logctx_le in *.
  dependent induction H; unfold logctx_path in *; simpls.
  destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; destruct oc1; simpls; try discriminate; eauto.

  destruct oc2; destruct oc1; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.

  destruct oc2; destruct oc1; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.

  destruct oc2; destruct oc1; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.

  destruct oc2; destruct oc1; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.
Qed.

Hint Resolve logctx_le_extend_l logctx_re_extend_l: core.

Require Import Orctx.

Definition orctx_path oc := (logctx_path (orctx_logctx oc)).
Definition orctx_le (ocl ocr: orctx) := logpath_le (orctx_path ocl) (orctx_path ocr).
Definition orctx_re (ocl ocr: orctx) := logpath_re (orctx_path ocl) (orctx_path ocr).

Lemma orctx_le_cases ocl ocr:
  orctx_le ocl ocr \/ orctx_le ocr ocl.
Proof.
  unfold orctx_le.
  reverts ocr.
  induction ocl; intros; simpls; subst. auto.
  all: destruct ocr; unfold orctx_path; simpls; eauto;
    forwards Hocl: IHocl; destruct_all; eauto.
Qed.

Lemma orctx_follow_logctx ocl ocr:
  orctx_follow ocl ocr ->
  logctx_follow (orctx_logctx ocl) (orctx_logctx ocr).
Proof.
  introv Hf.
  induction Hf; simpls; eauto.
Qed.

Lemma orctx_follow_le ocl ocr:
  orctx_follow ocl ocr ->
  orctx_le ocl ocr.
Proof.
  introv Hf.
  eapply orctx_follow_logctx in Hf.
  unfold orctx_le, orctx_path.
  eapply logctx_follow_path_eq in Hf. rewrite Hf.
  eapply logpath_le_refl.
Qed.

Lemma orctx_le_trans ocl oc ocr:
  orctx_le ocl oc ->
  orctx_le oc ocr ->
  orctx_le ocl ocr.
Proof.
  introv Hlel Hler.
  unfold orctx_le in *.
  eapply logpath_le_trans; eauto.
Qed.

Lemma orctx_logctx_extend' oc x:
 (orctx_logctx (orctx_extend oc x)) = (logctx_extend (orctx_logctx oc) (orctx_logctx x)).
Proof.
  induction oc; simpls; eauto.
  rewrite IHoc. eauto.
  rewrite IHoc. eauto.
Qed.

Lemma orctx_le_extend oc x: orctx_le oc (orctx_extend oc x).
Proof.
  unfold orctx_le, orctx_path.
  rewrite orctx_logctx_extend'.
  rewrite logctx_path_extend.
  eapply logpath_le_extend_nil_l.
Qed.

Lemma orctx_le_refl oc:
  orctx_le oc oc.
Proof.
  unfold orctx_le. eapply logpath_le_refl; eauto.
Qed.

Lemma orctx_le_nil oc:
  orctx_le orctx_nil oc.
Proof.
  hnf; eauto.
Qed.

Lemma orctx_le_nil_inv oc:
  orctx_le oc orctx_nil ->
  oc = orctx_nil.
Proof.
  introv Hle.
  hnf in *; unfold orctx_path in *; destruct oc; simpls; auto; inversion Hle.
Qed.

Lemma orctx_le_extend_inv ocp oc1 oc2:
  orctx_le (orctx_extend ocp oc1) (orctx_extend ocp oc2) ->
  orctx_le oc1 oc2.
Proof.
  introv Hle.
  induction ocp; simpls; eauto; inversions Hle; eauto.
Qed.

Lemma orctx_le_eq_separate ocT T ocU U:
  typ_notor T ->
  orctx_typ ocT T = orctx_typ ocU U ->
  orctx_le ocT ocU ->
  (ocT = ocU /\ T = U) \/ orctx_separate ocT ocU.
Proof.
  introv HnT He Hle.
  forwards: orctx_intersect_cases_nonnil He.
  destruct_all; subst.
  - eauto.
  - false.
    destruct x; subst; simpls; try contradiction; eapply HnT; eauto.
  - false. rewrite <- orctx_extend_nil_l in Hle. eapply orctx_le_extend_inv in Hle. eapply orctx_le_nil_inv in Hle. contradiction.
  - eauto.
Qed.

Lemma orctx_le_l oc1 oc2 T1 T2:
  orctx_le oc1 oc2 ->
  orctx_le (orctx_l oc1 T1) (orctx_l oc2 T2).
Proof.
  unfold orctx_le, orctx_path in *. simpls.
    eapply logpath_le_orl.
Qed.
Lemma orctx_le_r oc1 oc2 T1 T2:
  orctx_le oc1 oc2 ->
  orctx_le (orctx_r T1 oc1) (orctx_r T2 oc2).
Proof.
  unfold orctx_le, orctx_path in *. simpls.
    eapply logpath_le_orr.
Qed.

Lemma orctx_le_extend_l_inv oc1 oc2:
  orctx_le (orctx_extend oc1 oc2) oc1 ->
  oc2 = orctx_nil.
Proof.
  introv Hle.
  forwards: orctx_le_extend_inv oc1 oc2 orctx_nil.
  rewrite orctx_extend_nil_l. eauto.
  eapply orctx_le_nil_inv; eauto.
Qed.

Lemma orctx_le_le_eq_inv (ocT : orctx) (T : typ) (ocU : orctx) (U : typ):
  orctx_typ ocT T = orctx_typ ocU U ->
  orctx_le ocT ocU ->
  orctx_le ocU ocT ->
  ocT = ocU /\ T = U.
Proof.
  introv He HleT HleU.
  reverts ocT; induction ocU; intros; simpls.
  - eapply orctx_le_nil_inv in HleT; subst; simpls; auto.
  - destruct ocT; simpls; subst.
    + inversions HleU.
    + inversions HleU.
      inversions HleT.
      inversions He.
      forwards: IHocU; eauto. destruct_all; subst; eauto.
    + inversions HleT.
  - destruct ocT; simpls; subst.
    + inversions HleU.
    + inversions HleU.
    + inversions HleU.
      inversions HleT.
      inversions He.
      forwards: IHocU; eauto. destruct_all; subst; eauto.
Qed.

Lemma orctx_le_cases_ne ocl ocr:
       orctx_le ocl ocr \/ (orctx_le ocr ocl /\ ocr <> ocl).
Proof.
  intros.
  destruct (orctx_le_cases ocl ocr).
  - auto.
  - destruct (classicT (ocr = ocl)).
    + subst. auto.
    + auto.
Qed.

(* Minimum of two contexts.
Prefers left and shorter. *)

Fixpoint orctx_leb (ocl ocr: orctx): orctx :=
match ocl, ocr with
| orctx_nil, oc => orctx_nil
| orctx_l oc T, orctx_nil => orctx_nil
| orctx_l oc T, orctx_r U ocp => orctx_l oc T
| orctx_l oc T, orctx_l ocp U => orctx_l (orctx_leb oc ocp) T
| orctx_r T oc, orctx_nil => orctx_nil
| orctx_r T oc, orctx_l ocp U => orctx_l ocp U
| orctx_r T oc, orctx_r U ocp => orctx_r T (orctx_leb oc ocp)
end.

Lemma orctx_leb_le_l ocl ocr:
  orctx_le (orctx_leb ocl ocr) ocl.
Proof.
  unfold orctx_le.
  reverts ocr.
  induction ocl; intros; simpls.
  - eapply logpath_le_refl.
  - destruct ocr; unfold orctx_path; simpls; eauto.
  - destruct ocr; unfold orctx_path; simpls; eauto.
Qed.

Lemma orctx_leb_nil_l oc:
  orctx_leb orctx_nil oc = orctx_nil.
Proof.
  simpls. auto.
Qed.
Lemma orctx_leb_nil_r oc:
  orctx_leb oc orctx_nil = orctx_nil.
Proof.
  destruct oc; simpls; auto.
Qed.

Lemma orctx_leb_le_comm ocl ocr:
  orctx_le (orctx_leb ocl ocr) (orctx_leb ocr ocl).
Proof.
  reverts ocr; induction ocl; intros; destruct ocr; unfold orctx_le, orctx_path in *; simpls; eauto using logpath_le_refl.
Qed.

Lemma orctx_leb_le_r ocl ocr:
  orctx_le (orctx_leb ocl ocr) ocr.
Proof.
  eapply orctx_le_trans. eapply orctx_leb_le_comm. eapply orctx_leb_le_l.
Qed.


Lemma orctx_leb_le_b oc ocl ocr:
  orctx_le oc ocl ->
  orctx_le oc ocr ->
  orctx_le oc (orctx_leb ocl ocr).
Proof.
  unfold orctx_le.

  introv Hl Hr.
  reverts ocr ocl.
    induction oc; intros; unfold orctx_path in *; simpls; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
  - destruct ocr; destruct ocl; simpls; inversions Hl; inversions Hr; eauto.
Qed.

Lemma orctx_leb_l oc1 oc2:
  orctx_le oc1 oc2 ->
  orctx_leb oc1 oc2 = oc1.
Proof.
  intros.
  unfold orctx_le in *.
  dependent induction H; unfold orctx_path in *; simpls.
  destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; simpls; try discriminate; eauto.
  destruct oc2; simpls; try discriminate; eauto.
  destruct oc1; simpls; try discriminate; eauto.

  destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; simpls; try discriminate; eauto.

  destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.

  destruct oc1; simpls; try discriminate; eauto.
  destruct oc2; simpls; try discriminate; eauto.
  inversions x0; inversions x. fequal; eauto.
Qed.
(*Warning: orctx_leb is not commutative, because they can have the same path but different types, so the types on the left get priority *)
