Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import Vars Types Context.

(** * Extension of typing context for typing objects *)

(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Extension of typing context for typing objects,
   with respect to the address of the object *)
Inductive objctx: Set :=
(* The address is irrelevant, for example when typing other terms than objects *)
| objctx_empty: objctx 
| objctx_lit(x: var)(T: typ): objctx
| objctx_heap(y: var)(T: typ): objctx
.

(* Typing context for method body *)
Notation ctx_met G xz TZ xc TC := (G § xz / v_par ~ TZ § xc / v_recv ~ (open xz TC)).
(* Opening method body and result type *)
Notation open_met xz xc T := (open xz (open_rec 1 xc T)) (only parsing).
Notation close_met z r T := (close z (close_rec 1 r T)) (only parsing).

(* Merges ctx and objctx into a context for typing the content of member definitions *)
Definition ctx_obj(G: ctx)(Go: objctx) :=
  match Go with
  | objctx_empty => G
  | objctx_lit x T => G § x / v_self ~ open x T
  | objctx_heap y T => G
  end.

Notation ctx_met_refined G xz TZ xc TC TC2 := (G § xz / v_par ~ TZ § xc / v_recv ~ (typ_and (open xz TC) TC2)).
Notation ctx_met_refined_heap G xz TZ xc TC y T := (ctx_met_refined G xz TZ xc TC (typ_and (open y T) (open xc T))).

Lemma ctx_met_concat G xz TZ xc TC :
  ctx_met G xz TZ xc TC
  = G §& (xz / v_par ~ TZ § xc / v_recv ~ (open xz TC)).
Proof. reflexivity. Qed.

Lemma ctx_met_refined_heap_concat G xz TZ xc TC y T :
  ctx_met_refined_heap G xz TZ xc TC y T
  = G §& ( xz / v_par ~ TZ
         § xc / v_recv ~ (typ_and (open xz TC)
                                  (typ_and (open y T) (open xc T)))).
Proof. reflexivity. Qed.

Lemma ctx_met_refined_heap_ok G xz TZ xc TC x U :
  ctx_ok G -> xz §# G -> xc §# G -> xc <> xz ->
  ctx_ok (ctx_met_refined_heap G xz TZ xc TC x U).
Proof. introv Hok Hcfr Hzfr Hneq. eauto. Qed.

Hint Transparent ctx_obj: core.

Definition subst_objctx vSrc vDst Go :=
  match Go with
  | objctx_empty => objctx_empty
  | objctx_lit x T => objctx_lit x (subst_var vSrc vDst T)
  | objctx_heap y T => objctx_heap (subst_var vSrc vDst y) (subst_var vSrc vDst T)
  end.
Notation "G '/[§§' x '->' y ']'" :=
  (subst_objctx x y G) (x at level 69, at level 7, left associativity) : syntax_scope.

Definition subst_objctx_pvars xs ys Go :=
  match Go with
  | objctx_empty => objctx_empty
  | objctx_lit x T => objctx_lit x (subst_pvars xs ys T)
  | objctx_heap y T => objctx_heap (subst_pvars xs ys y) (subst_pvars xs ys T)
  end.

Definition subst_objctx_varx vsSrc vsDst Go :=
  match Go with
  | objctx_empty => objctx_empty
  | objctx_lit x T => objctx_lit x (subst_vars vsSrc vsDst T)
  | objctx_heap y T => objctx_heap (subst_vars vsSrc vsDst y) (subst_vars vsSrc vsDst T)
  end.

Definition objctx_dom Go :=
  match Go with
  | objctx_empty => \{}
  | objctx_lit x T => \{ x }
  | objctx_heap y T => \{}
  end.

Definition objctx_keys Go: list var :=
  match Go with
  | objctx_empty => nil
  | objctx_lit x T => x::nil
  | objctx_heap y T => nil
  end.

Definition objctx_fv Go :=
  match Go with
  | objctx_empty => \{}
  | objctx_lit x T => fv T
  | objctx_heap y T => \{y} \u (fv T)
  end.

Notation "G '§§&' Go" := (ctx_obj G Go)  (at level 28, left associativity).

Lemma ctx_ok_ctx_obj_empty G:
  ctx_ok G ->
  ctx_ok (G §§& objctx_empty).
Proof.
  unfold ctx_obj.
  auto.
Qed.
Hint Resolve ctx_ok_ctx_obj_empty : core.

Lemma objctx_keys_dom Go:
  from_list (objctx_keys Go) = objctx_dom Go.
Proof.
  destruct Go; simpls; rewrite ? from_list_cons, ? from_list_nil; auto.
Qed.

Fixpoint objctx_seal_rec n y G {struct G} :=
  match G with
  | objctx_empty => objctx_empty
  | objctx_heap x T => objctx_heap x (seal_rec (S n) y T)
  | objctx_lit x T => objctx_lit x (seal_rec (S n) y T)
  end.

Instance ObjctxSealable: Sealable objctx := objctx_seal_rec.

Lemma ctx_obj_assoc G1 G2 Go:
  (G1 §& G2) §§& Go = G1 §& (G2 §§& Go).
Proof.
  unfold ctx_obj; destruct Go; ctx_flatten; auto.
Qed.



Lemma ctx_obj_dom G Go:
  ctx_dom (G §§& Go) = ctx_dom G \u objctx_dom Go.
Proof.
  destruct Go; unfold ctx_obj, objctx_dom; ctx_simpl_dom; eauto.
Qed.



Lemma ctx_ok_obj_inv_l G Go :
  ctx_ok (G §§& Go) ->
  ctx_ok G.
Proof.
  introv Hok. destruct Go; eauto.
  simpl in Hok. inversions Hok; eauto.
Qed.


Lemma objctx_dom_ctx_dom G Go :
  ctx_dom (G §§& Go) = ctx_dom G \u objctx_dom Go.
Proof. destruct Go; simpl; eauto. Qed.

Lemma ctx_ok_obj_disjoint G Go :
  ctx_ok (G §§& Go) ->
  disjoint (ctx_dom G) (objctx_dom Go).
Proof.
  introv Hok. destruct Go; simpl; hnf.
  1,3: rewrite inter_empty_r; tauto.
  simpl in Hok. inversions Hok.
  rewrite inter_comm. eapply inter_singleton_notin; eauto.
Qed.

Lemma ctx_ok_obj_disjoint_converse G Go :
  ctx_ok G ->
  disjoint (ctx_dom G) (objctx_dom Go) ->
  ctx_ok (G §§& Go).
Proof.
  introv Hok Hdis. destruct Go; simpl in *; eauto.
  constructor; eauto. hnf in Hdis. hnf. intros.
  erewrite <- in_empty, <- Hdis, in_inter, in_singleton. eauto.
Qed.

Lemma ctx_vis_objctx_inv G Go x :
  ctx_vis (G §§& Go) x ->
  ctx_vis G x \/ x \in objctx_dom Go.
Proof.
  introv Hvis. destruct Go; simpl in *; eauto.
  inversions Hvis; eauto. rewrite in_singleton; eauto.
Qed.

Lemma ctx_vis_objctx_concat_r G Go x :
  x \in objctx_dom Go -> ctx_vis (G §§& Go) x.
Proof.
  introv Hdom. destruct Go; simpl in *;
    try solve [ rewrite in_empty in *; tauto
              | rewrite in_singleton in *; subst; auto] .
Qed.

Lemma ctx_obj_subst G Go x y:
  x \notin objctx_dom Go ->
  subst_ctx x y G §§& subst_objctx x y Go = subst_ctx x y (G §§& Go).
Proof.
  introv Hn.
  destruct Go; simpls; auto 1. rewrite subst_open_commut. cases_if. rewrite notin_singleton in Hn. contradiction. auto.
Qed.
