(** remove printing ~ *)

(** This module contains lemmas related to invertible typing
    ([ty_var_inv], [⊢##] and [ty_item_inv], [⊢##v]). *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.
Require Import TypeInduction.


(** * Syntactic [\/]-context *)

(* A union context represent a union type with a single hole.
   It is constructed from top to bottom. *)
Inductive orctx: Set :=
| orctx_nil: orctx (* The hole *)
| orctx_l (oc: orctx) (T:typ): orctx (* Hole on the left of a type *)
| orctx_r (T:typ) (oc: orctx): orctx (* Hole on the right of a type *)
.

(* Filling the hole in a union context *)
Fixpoint orctx_typ (oc: orctx) (T:typ): typ := 
  match oc with
  | orctx_nil => T
  | orctx_l ocl Tr => (typ_or (orctx_typ ocl T) Tr)
  | orctx_r Tl ocr => (typ_or Tl (orctx_typ ocr T))
  end.

(* Appending an union context oci into the hole of oc *)
Fixpoint orctx_extend (oc oci: orctx): orctx := 
  match oc with
  | orctx_nil => oci
  | orctx_l ocl Tr => orctx_l (orctx_extend ocl oci) Tr
  | orctx_r Tl ocr => orctx_r Tl (orctx_extend ocr oci)
  end.

Lemma orctx_fold_r oc Tl Tr:
  typ_or Tl (orctx_typ oc Tr)  = orctx_typ (orctx_r Tl oc) Tr.
Proof.
  simpls; eauto.
Qed.
Lemma orctx_fold_l oc Tl Tr:
  typ_or (orctx_typ oc Tl) Tr = orctx_typ (orctx_l oc Tr) Tl.
Proof.
  simpls; eauto.
Qed.


(* Lemmas about orctx_extends *)

Lemma orctx_extend_l oc Tl Tr:
  orctx_typ oc (typ_or Tl Tr) = orctx_typ (orctx_extend oc (orctx_l orctx_nil Tr) ) Tl.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma orctx_extend_r oc Tl Tr:
  orctx_typ oc (typ_or Tl Tr) = orctx_typ (orctx_extend oc (orctx_r Tl orctx_nil) ) Tr.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma orctx_extend_nil_l oc:
  orctx_extend oc orctx_nil = oc.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma orctx_extend_nil_r oc:
  orctx_extend orctx_nil oc = oc.
Proof.
  induction oc; simpls; auto.
Qed.
Lemma orctx_extend_assoc oc1 oc2 oc3:
  orctx_extend (orctx_extend oc1 oc2 ) oc3 = orctx_extend oc1 (orctx_extend oc2 oc3).
Proof.
  reverts oc2 oc3.
  induction oc1; intros; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
Qed.


Lemma orctx_extend_typ oc1 oc2 T:
  orctx_typ (orctx_extend oc1 oc2 ) T = orctx_typ oc1 (orctx_typ oc2 T).
Proof.
  induction oc1; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
Qed.

(* Inversion lemmas *)

(* For orctx_recursive_inv, we will show that the union context increases the height of a type *)

(* The height of an union context *)
Fixpoint orctx_height (oc: orctx): nat := 
  match oc with
  | orctx_nil => 0
  | orctx_l ocl Tr => S (orctx_height ocl)
  | orctx_r Tl ocr => S (orctx_height ocr)
  end.

(* Union context increases the height of a type *)
Lemma orctx_typ_height oc T:
  typ_opening_ind_height T + orctx_height oc <= typ_opening_ind_height (orctx_typ oc T).
Proof.
  induction oc.
  - simpls. rewrite Nat.add_0_r. auto.
  - simpls. rewrite PeanoNat.Nat.add_succ_r. eapply Le.le_n_S. eapply Nat.le_trans. eapply IHoc. eapply Max.le_max_l.
  - simpls. rewrite PeanoNat.Nat.add_succ_r. eapply Le.le_n_S. eapply Nat.le_trans. eapply IHoc. eapply Max.le_max_r.
Qed.

Lemma orctx_extend_height oc1 oc2:
  orctx_height(orctx_extend oc1 oc2) =   orctx_height oc1 + orctx_height oc2.
Proof.
  intros.
  induction oc1;
  simpls; eauto.
Qed.

Lemma orctx_recursive_inv oc A:
  A = orctx_typ oc A ->
  oc = orctx_nil.
Proof.
  introv Heq.
  pose proof (f_equal typ_opening_ind_height Heq) as Heqh.

  pose proof (orctx_typ_height oc A).
  rewrite Heqh in H.
  rewrite <- PeanoNat.Nat.add_0_r in H.
  apply <- Nat.add_le_mono_l in H.
  destruct oc.
  - auto.
  - false. simpls. eapply Nat.nle_succ_0; eauto.
  - false. simpls. eapply Nat.nle_succ_0; eauto.
Qed.

Lemma orctx_typ_eq_inv oc A B:
  orctx_typ oc A = orctx_typ oc B ->
  A = B.
Proof.
  introv Heq.
  induction oc; simpls; inversions Heq; auto.
Qed.

Lemma orctx_notor_inv oc A B:
  typ_notor B ->
  orctx_typ oc A = B ->
  oc = orctx_nil.
Proof.
  introv Hne Heq.
  destruct oc.
  - auto.
  - false. simpls. eapply Hne; eauto.
  - false. simpls. eapply Hne; eauto.
Qed.

(* Separate union contexts are two contexts that have a common prefix (possibly trivial) and then go in different directions *)
(* If two union contexts construct the same type, then either one is a prefix of the other or they are separate *)
(* This will be useful for invertible typing, where we will have to align two context resulting in the same type,
   and there will always be a possibility that twe two contexts are separate *)

Inductive orctx_separate: orctx->orctx->Prop :=
| orctx_separate_l oc1 oc2 Tr: (* Common prefix left *)
  orctx_separate oc1 oc2 -> 
  orctx_separate (orctx_l oc1 Tr) (orctx_l oc2 Tr)
| orctx_separate_r oc1 oc2 Tl: (* Common prefix right *)
  orctx_separate oc1 oc2 -> 
  orctx_separate (orctx_r Tl oc1) (orctx_r Tl oc2)
| orctx_separate_lr oc1 oc2 Tl Tr: (* Separation point left and right *)
  orctx_separate (orctx_l oc1 Tr) (orctx_r Tl oc2)
| orctx_separate_rl oc1 oc2 Tl Tr: (* Separation point right and left *)
  orctx_separate (orctx_r Tl oc1) (orctx_l oc2 Tr).
Hint Constructors orctx_separate: core.

Lemma orctx_separate_symm oc1 oc2:
  orctx_separate oc1 oc2 ->
  orctx_separate oc2 oc1.
Proof.
  introv Hse.
  induction Hse; eauto.
Qed.

Inductive orctx_follow: orctx->orctx->Prop :=
| orctx_follow_nil :
  orctx_follow orctx_nil orctx_nil
| orctx_follow_l oc1 oc2 Tr1 Tr2:
  orctx_follow oc1 oc2 -> 
  orctx_follow (orctx_l oc1 Tr1) (orctx_l oc2 Tr2)
| orctx_follow_r oc1 oc2 Tl1 Tl2: (* Common prefix right *)
  orctx_follow oc1 oc2 -> 
  orctx_follow (orctx_r Tl1 oc1) (orctx_r Tl2 oc2)
.
Hint Constructors orctx_follow: core.

Lemma orctx_follow_refl oc:
  orctx_follow oc oc.
Proof.
  induction oc; eauto.
Qed.
Hint Resolve orctx_follow_refl: core.

Lemma orctx_follow_symm oc1 oc2:
  orctx_follow oc1 oc2 ->
  orctx_follow oc2 oc1.
Proof.
  introv Hfol. reverts oc2.
  induction oc1; intros; inversions Hfol; eauto.
Qed.


Lemma orctx_follow_eq_inv oc1 oc2 T1 T2:
  orctx_follow oc1 oc2 ->
  (orctx_typ oc1 T1) = (orctx_typ oc2 T2) ->
  T1 = T2 /\ oc1 = oc2.
Proof.
  introv Hf Heq.
  induction Hf; simpls; eauto; inversions Heq; forwards : IHHf; destruct_all; subst; eauto.
Qed.


Lemma orctx_follow_trans oc1 oc2 oc3:
  orctx_follow oc1 oc2 ->
  orctx_follow oc2 oc3 ->
  orctx_follow oc1 oc3.
Proof.
  introv Hf1 Hf2.
  reverts oc3; induction Hf1; intros; eauto.
  - inversions Hf2; eauto.
  - inversions Hf2; eauto.
Qed.



(* If two union contexts are separate,
  then we can change the types in the holes *)
Lemma orctx_separate_cross_r ocS S ocT T:
  orctx_typ ocS S = orctx_typ ocT T ->
  orctx_separate ocS ocT ->
  (forall S1, exists ocS1T,
    orctx_typ ocS S1 = orctx_typ ocS1T T  /\
    (forall T1, exists ocST1,
    orctx_typ ocST1 S = orctx_typ ocT T1 /\
    orctx_typ ocST1 S1 = orctx_typ ocS1T T1 /\
    orctx_follow ocST1 ocS) /\
    orctx_follow ocT ocS1T
  ).
Proof.
  introv Heq Hse.
  reverts S T.
  induction Hse; intros; simpls; inversions Heq.
  + (* Agreeing *)
    forwards ocS1T HocS1T Hrest HfS1T: (IHHse S T H0 S1); eauto.
    exists (orctx_l ocS1T Tr).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfST1: (Hrest T1).
    exists (orctx_l ocST1 Tr).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto. eauto.
  + (* Agreeing *)
    forwards ocS1T HocS1T Hrest HfS1T: (IHHse S T H0 S1); eauto.
    exists (orctx_r Tl ocS1T).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfST1: (Hrest T1).
    exists (orctx_r Tl ocST1).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto. eauto.
  +
    exists (orctx_r (orctx_typ oc1 S1) oc2). splits. auto.
    intros.
    exists (orctx_l oc1 (orctx_typ oc2 T1)). simpls. auto. auto.
  + (* Disagreeing *)
    exists (orctx_l oc2 (orctx_typ oc1 S1)). splits. auto.
    intros.
    exists (orctx_r (orctx_typ oc2 T1) oc1). simpls. auto. auto.
Qed.

Lemma orctx_separate_cross_l ocS S ocT T:
  orctx_typ ocS S = orctx_typ ocT T ->
  orctx_separate ocS ocT ->
  (forall T1, exists ocST1,
    orctx_typ ocST1 S = orctx_typ ocT T1 /\
    (forall S1, exists ocS1T,
    orctx_typ ocS S1 = orctx_typ ocS1T T  /\
    orctx_typ ocST1 S1 = orctx_typ ocS1T T1 /\
    orctx_follow ocS1T ocT) /\
    orctx_follow ocS ocST1
  ).
Proof.
  introv Heq Hse.
  eapply orctx_separate_symm in Hse.
  symmetry in Heq.
  intros.
  forwards H: orctx_separate_cross_r ocT T ocS S; eauto.
  destruct_all.
  exists x. splits. symmetry. eauto.
  intros.
  forwards: H0 S1.
  destruct_all.
  exists x0.
  splits; eauto. eauto.
Qed.

Lemma orctx_intersect_cases_any ocS S ocT T:
  orctx_typ ocS S = orctx_typ ocT T -> (
  (* equal *)
  (ocS = ocT /\ S = T) \/
  (* below *)
  (exists ocST, ocT = orctx_extend ocS ocST /\ S = orctx_typ ocST T) \/
  (* above *)
  (exists ocTS, ocS = orctx_extend ocT ocTS /\ T = orctx_typ ocTS S) \/
  (* aside *)
  orctx_separate ocS ocT).
Proof.
  introv Heq.
  reverts ocS S.
  induction ocT; intros; simpls.
  - right. right. left.
    exists ocS; split; eauto.
  - destruct ocS; simpls; subst.
    + (* S ends above *)
      forwards: IHocT; eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst;  eauto.
      right. right. left. exists x; eauto.
    + (* Disagreeing *) 
      eauto.
  - destruct ocS; simpls; subst.
    + (* S ends above *)
      forwards: IHocT; eauto.
    + (* Disagreeing *)
      eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst;  eauto.
      right. right. left. exists x; eauto.
Qed.

Lemma orctx_intersect_cases_nonnil ocS S ocT T:
  orctx_typ ocS S = orctx_typ ocT T -> (
  (* equal *)
  (ocS = ocT /\ S = T) \/
  (* below *)
  (exists ocST, ocT = orctx_extend ocS ocST /\ S = orctx_typ ocST T /\ ocST <> orctx_nil) \/
  (* above *)
  (exists ocTS, ocS = orctx_extend ocT ocTS /\ T = orctx_typ ocTS S /\ ocTS <> orctx_nil) \/
  (* aside *)
  orctx_separate ocS ocT).
Proof.
  introv Heq.
  forwards: orctx_intersect_cases_any; eauto.
  destruct_all; eauto.
  - destruct x.
    + simpls; rewrite orctx_extend_nil_l in H; eauto.
    + right. left. eexists; splits; eauto; discriminate.
    + right. left. eexists; splits; eauto; discriminate.
  - destruct x.
    + simpls; rewrite orctx_extend_nil_l in H; eauto.
    + right. right. left. eexists; splits; eauto; discriminate.
    + right. right. left. eexists; splits; eauto; discriminate.
Qed.

(* Intersecting two unions, where one has an intersection. Either the intersection is below S or aside. *)
Lemma orctx_intersect_cases_r ocS S ocT T:
  typ_notor T ->
  orctx_typ ocS S = orctx_typ ocT T ->
  (* below *)
  (exists ocST, ocT = orctx_extend ocS ocST /\ S = orctx_typ ocST T) \/
  (* aside *)
  orctx_separate ocS ocT.
Proof.
  introv Hno Heq.
  reverts ocS S.
  induction ocT; intros; simpls.
  - left. forwards: orctx_notor_inv Heq. auto. subst. exists orctx_nil. splits; simpls; eauto.
  - destruct ocS; simpls; subst.
    + (* S ends above *)
      forwards: IHocT; eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst;  eauto.
    + (* Disagreeing *) 
      eauto.
  - destruct ocS; simpls; subst.
    + (* S ends above *)
      forwards: IHocT; eauto.
    + (* Disagreeing *)
      eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst;  eauto.
Qed.

Lemma orctx_intersect_cases ocS S ocT T:
  typ_notor S ->
  typ_notor T ->
  orctx_typ ocS S = orctx_typ ocT T ->
  (* same *)
  (ocS = ocT /\ S = T) \/
  (* aside *)
  orctx_separate ocS ocT.
Proof.
  introv HnS HnT Heq.
  reverts ocS S.
  induction ocT; intros; simpls.
  - left. forwards: orctx_notor_inv Heq. auto. subst. simpls. auto.
  - destruct ocS; simpls; subst; try discriminate.
    + false. eapply HnS; eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto.
    + (* Disagreeing *)
      eauto.
  - destruct ocS; simpls; subst; try discriminate.
    + false. eapply HnS; eauto.
    + (* Disagreeing *)
      eauto.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto.
Qed.

(* Combines orctx_intersect_cases with orctx_separate_cross_r *)
Lemma orctx_intersect_cases_cross ocS S ocT T:
  typ_notor S ->
  typ_notor T ->
  orctx_typ ocS S = orctx_typ ocT T ->
  (* same *)
  (ocS = ocT /\ S = T) \/
  (* aside *)
  (forall S1, exists ocS1T,
    orctx_typ ocS S1 = orctx_typ ocS1T T /\
    (forall T1, exists ocST1,
    orctx_typ ocST1 S = orctx_typ ocT T1 /\
    orctx_typ ocST1 S1 = orctx_typ ocS1T T1/\
    orctx_follow ocST1 ocS) /\
    orctx_follow ocT ocS1T
  ).
Proof.
  introv HnS HnT Heq.
  forwards: orctx_intersect_cases ocS S ocT T; eauto.
  destruct_all; auto.
  right; intros.
  forwards: orctx_separate_cross_r S H S1; eauto.
Qed.

Lemma orctx_intersect_and_and_cases ocS S1 S2 ocT T1 T2:
  orctx_typ ocS (typ_and S1 S2) = orctx_typ ocT (typ_and T1 T2) ->
  (* same *)
  (ocS = ocT /\ S1 = T1 /\ S2 = T2) \/
  (* aside *)
  (exists ocS1T ocS2T,
    orctx_typ ocS S1 = orctx_typ ocS1T (typ_and T1 T2) /\
    orctx_typ ocS S2 = orctx_typ ocS2T (typ_and T1 T2) /\
    (
      forall T3, exists ocST3,
      orctx_typ ocST3 (typ_and S1 S2) = orctx_typ ocT T3 /\
      orctx_typ ocST3 S1 = orctx_typ ocS1T T3 /\
      orctx_typ ocST3 S2 = orctx_typ ocS2T T3 /\
      orctx_follow ocST3 ocS
    )
  ).
Proof.
  introv Heq. symmetry in Heq.
  forwards: orctx_intersect_cases; eauto.
  intros; discriminate.
  intros; discriminate.
  destruct_all.
  - inversions H0; auto.
  - right.
    intros.
    forwards Hcross1: orctx_separate_cross_l H S1; eauto.
    forwards ocsT31 HocsT31 Hex1 Hf1: (Hcross1).
    forwards Hcross2: orctx_separate_cross_l H S2; eauto.
    forwards ocsT32 HocsT32 Hex2 Hf2: (Hcross2).

    exists ocsT31 ocsT32.
    splits; auto.
    intros.
    forwards: Hex1 T3.
    forwards: Hex2 T3.
    destruct_all.
    exists x.
    splits; auto.

    rewrite H6.
    fequal.
    eapply orctx_follow_eq_inv.
    eapply orctx_follow_trans; eauto using orctx_follow_symm.
    rewrite <- H0; eauto.
Qed.

Fixpoint orctx_common_prefix (oc1 oc2: orctx): orctx := 
  match oc1 with
  | orctx_nil => orctx_nil
  | orctx_l ocl1 Tr1 =>
    match oc2 with
    | orctx_nil => orctx_nil
    | orctx_l ocl2 Tr2 => If Tr1 = Tr2 then (orctx_l (orctx_common_prefix ocl1 ocl2) Tr1) else orctx_nil
    | orctx_r Tl2 ocr2 => orctx_nil
    end
  | orctx_r Tl1 ocr1 =>
    match oc2 with
    | orctx_nil => orctx_nil
    | orctx_l ocl2 Tr2 => orctx_nil
    | orctx_r Tl2 ocr2 => If Tl1 = Tl2 then (orctx_r Tl1 (orctx_common_prefix ocr1 ocr2)) else orctx_nil
    end
  end.

Lemma orctx_common_l oc1 oc2:
  exists oc31,
  oc1 = orctx_extend (orctx_common_prefix oc1 oc2) oc31.
Proof.
  reverts oc2.
  induction oc1; intros; destruct oc2; simpls; eauto; cases_if; simpls; eauto.
  - destruct (IHoc1 oc2). exists x. rewrite H at 1. auto.
  - destruct (IHoc1 oc2). exists x. rewrite H at 1. auto.
Qed.

Lemma orctx_common_r oc1 oc2:
  exists oc31,
  oc2 = orctx_extend (orctx_common_prefix oc1 oc2) oc31.
Proof.
  reverts oc1.
  induction oc2; intros; destruct oc1; simpls; eauto; cases_if; simpls; eauto.
  - destruct (IHoc2 oc1). exists x. rewrite H at 1. auto.
  - destruct (IHoc2 oc1). exists x. rewrite H at 1. auto.
Qed.

Lemma orctx_separate_common_prefix ocS S ocT T:
  orctx_typ ocS S = orctx_typ ocT T ->
  orctx_separate ocS ocT ->
  (
    exists ocS1 ocT1,
    ocS = orctx_extend (orctx_common_prefix ocS ocT) (orctx_l ocS1 (orctx_typ ocT1 T)) /\
    ocT = orctx_extend (orctx_common_prefix ocS ocT) (orctx_r (orctx_typ ocS1 S) ocT1)
  ) \/ (
    exists ocS1 ocT1,
    ocS = orctx_extend (orctx_common_prefix ocS ocT) (orctx_r (orctx_typ ocT1 T) ocS1) /\
    ocT = orctx_extend (orctx_common_prefix ocS ocT) (orctx_l ocT1 (orctx_typ ocS1 S))
  ).
Proof.
  introv Heq Hse.
  reverts S T.
  induction Hse; intros; simpls; inversions Heq; eauto; cases_if.
  - forwards: IHHse H0.
    destruct_all; simpls.
    + left. exists x x0. rewrite H at 1. rewrite H1 at 3. auto.
    + right. exists x x0. rewrite H at 1. rewrite H1 at 3. auto.
  - forwards: IHHse H0.
    destruct_all; simpls.
    + left. exists x x0. rewrite H at 1. rewrite H1 at 3. auto.
    + right. exists x x0. rewrite H at 1. rewrite H1 at 3. auto.
Qed.

Lemma orctx_separate_common_prefix_length ocS ocT:
  orctx_separate ocS ocT ->
  orctx_height (orctx_common_prefix ocS ocT) < orctx_height ocS.
Proof.
  introv Hse.
  induction Hse; simpls; try cases_if; simpls; eauto using Lt.lt_n_S, Nat.lt_0_succ.
Qed.

Lemma typ_notor_ind (P: typ->Prop):
  (forall t, typ_notor t -> P t) ->
  (forall t1 t2, P t1 -> P t2 -> P (typ_or t1 t2)) ->
  forall t, P t.
Proof.
  intros.
  induction t; eauto 2; eapply H; intros; discriminate.
Qed.
Lemma orctx_prefix_ind (P: orctx -> Prop):
  P orctx_nil ->
  (forall oc : orctx, (forall ocl ocr, oc = orctx_extend ocl ocr -> P ocl) -> forall T : typ, P (orctx_l oc T)) ->
  (forall (T : typ) (oc : orctx), (forall ocl ocr, oc = orctx_extend ocl ocr -> P ocl) -> P (orctx_r T oc)) ->
  forall o : orctx, P o.
Proof.
  introv Hnil Hl Hr.
  intro.
  lets: (orctx_ind (fun oc => (forall ocl ocr, oc = orctx_extend ocl ocr -> P ocl))) o o; eauto.
  - intros. destruct ocl; simpls; try discriminate. auto.
  - intros. destruct ocl; simpls; try discriminate; auto. inversions H0. eapply Hl. intros; subst. eapply H. rewrite orctx_extend_assoc. eauto.
  - intros. destruct ocl; simpls; try discriminate; auto. inversions H0. eapply Hr. intros; subst. eapply H. rewrite orctx_extend_assoc. eauto.
  - intros. eapply (H orctx_nil). rewrite orctx_extend_nil_l. auto.
Qed.
Lemma orctx_reverse_destruct oc:
  oc = orctx_nil \/ (
    exists ocl T,
      oc = orctx_extend ocl (orctx_l orctx_nil T)
  ) \/ (
    exists ocr T,
      oc = orctx_extend ocr (orctx_r T orctx_nil)
  ).
Proof.
  induction oc; destruct_all; subst; eauto.
  -  right. left. exists orctx_nil T. simpls. eauto.
  -  right. left. exists (orctx_l x T). simpls. eauto.
  -  right. right. exists (orctx_l x T). simpls. eauto.
  -  right. right. exists (orctx_nil) T. simpls. eauto.
  -  right. left. exists (orctx_r T x). simpls. eauto.
  -  right. right. exists (orctx_r T x). simpls. eauto.
Qed.
(* Induction from below *)
Lemma orctx_reverse_ind (P : orctx -> Prop):
  P orctx_nil ->
  (forall oc : orctx, P oc -> forall T : typ, P (orctx_extend oc (orctx_l orctx_nil T ))) ->
  (forall (T : typ) (oc : orctx), P oc -> P (orctx_extend oc (orctx_r T orctx_nil))) ->
  forall o : orctx, P o.
Proof.
  intros.
  remember (orctx_height o) as i. 
  reverts o.
  induction i; intros; subst.
  -
  destruct o; inversions Heqi. eauto.
  -
  forwards: orctx_reverse_destruct o.
  destruct_all; subst; simpls.
  + inversions Heqi.
  + eapply H0. eapply IHi. rewrite orctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
  + eapply H1. eapply IHi. rewrite orctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
Qed.

Lemma orctx_shorter_ind (P: orctx -> Prop):
  P orctx_nil ->
  (forall oc : orctx, (forall ocl, orctx_height ocl < orctx_height oc -> P ocl) -> P oc) ->
  forall o : orctx, P o.
Proof.
  introv Hnil Hl.
  intro.
  lets: (orctx_ind (fun oc => (forall ocl, orctx_height ocl <= orctx_height oc -> P ocl))) o o; eauto.
  - intros. destruct ocl; simpls; try discriminate. auto. all: false; eapply PeanoNat.Nat.nle_succ_0; eauto.
  - intros. apply Hl; intros. eapply H. simpls. eapply le_S_n. unfold lt in H1. eapply Nat.le_trans; eauto.
  - intros. apply Hl; intros. eapply H. simpls. eapply le_S_n. unfold lt in H1. eapply Nat.le_trans; eauto.
Qed.

Ltac orctx_find_r := (
  match goal with |- context[typ_or ?A ?B] => replace (typ_or A B) with (orctx_typ (orctx_r A orctx_nil) B) by (simpls; apply eq_refl) end ).
Ltac orctx_find_r_in := (
  match goal with H: context[typ_or ?A ?B] |- _ =>  replace (typ_or A B) with (orctx_typ (orctx_r A orctx_nil) B) in H by (simpls; apply eq_refl) end).

Ltac orctx_find_l := (
  match goal with |- context[typ_or ?A ?B] =>  replace (typ_or A B) with (orctx_typ (orctx_l orctx_nil B) A) by (simpls; apply eq_refl) end ).
Ltac orctx_find_l_in := (
  match goal with H: context[typ_or ?A ?B] |- _ => replace (typ_or A B) with (orctx_typ (orctx_l orctx_nil B) A) in H by (simpls; apply eq_refl) end).
Ltac orctx_find_nil := (
  match goal with |- context[?B] =>  replace (B) with (orctx_typ orctx_nil B) by (simpls; apply eq_refl) end).
Ltac orctx_find_nil_in := (
  match goal with H: context[?B] |- _ =>  replace (B) with (orctx_typ orctx_nil B) in H by (simpls; apply eq_refl) end).

Ltac orctx_follow_solve := (eauto using orctx_follow_trans, orctx_follow_symm).
Ltac orctx_eqs :=
     repeat match goal with
        Heq1: orctx_typ ?ocL1 ?TL = ?TR,
        Heq2: orctx_typ ?ocL2 ?TL = ?TR
        |- _ => let Heqoc:= fresh "Heqoc" in let Heqoc1:= fresh "Heqoc1" in let Heqoc2:= fresh "Heqoc2" in
          forwards Heqoc: orctx_follow_eq_inv ocL1 ocL2;
          [> orctx_follow_solve | rewrite Heq1, Heq2; auto | destructc Heqoc as Heqoc1 Heqoc2; clear Heqoc1 Heq2; subst]
      end.