Set Implicit Arguments.

Require Import LibExtra DotTactics Vars Terms Heap Env SyntaxClasses TransformCollect Bool.

(** * Stack syntax *)

(** A stack frame stores the continuation of a let-in term.
  - [frame_let t] represents a let-in frame [let x = □ in t]
*)
Inductive frame : Set :=
  | frame_let : trm -> frame.

(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** A stack is a list of frames *)
Inductive stack: Set :=
  | stack_nil: stack
  | stack_cons (f: frame) (fs: stack): stack
.

(** ** Stack properties and operations *)

Fixpoint stack_length (fs : stack) : nat :=
  match fs with
  | stack_nil => 0
  | stack_cons _ fs' => (stack_length fs') + 1
  end.

Definition frame_fv fr := match fr with frame_let c => fv c end.
(* Free variables of a frame is just the free variable of the continuation *)
Instance FrameFreeVar : FreeVar frame := frame_fv.

(* Free variables of a stack is just the union of free variables in *)
Fixpoint stack_fv s: vars :=
  match s with
  | stack_nil => \{}
  | stack_cons f fs => (fv f) \u (@fv stack stack_fv fs)
  end.
Instance StackFreeVar : FreeVar stack := stack_fv.

Definition frame_subst_var x y fr :=
  match fr with frame_let c => frame_let (subst_var x y c) end.
Instance FrameSubstVar: SubstVar frame := frame_subst_var.

Fixpoint stack_subst_var x y st :=
  match st with
    | stack_nil => stack_nil
    | stack_cons f fs => stack_cons (subst_var x y f) (@subst_var stack stack_subst_var x y fs)
  end.

Instance StackSubstVar: SubstVar stack := stack_subst_var.

Definition frame_open n x fr :=
  match fr with frame_let c => frame_let (open_rec n x c) end.

Instance FrameOpenable: Openable frame := frame_open.

(* The stack alwas has one bound variable (placeholder for the result of the focus of execution) in each frame *)
(* A closed stack can contain #0 but not #1 *)
Fixpoint stack_open n x st :=
  match st with
    | stack_nil => stack_nil
    | stack_cons f fs => stack_cons (open_rec (S n) x f) (@open_rec stack stack_open n x fs)
  end.
Instance StackOpenable:  Openable stack := stack_open.


Instance FrameAbstractSyntax :
  AbstractSyntax FrameOpenable FrameFreeVar FrameSubstVar := {}.
Instance StackAbstractSyntax :
  AbstractSyntax StackOpenable StackFreeVar StackSubstVar := {}.

Definition collect_frame {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (fr: frame) : C :=
  match fr with
  | frame_let t => (collect df f comb t)
  end.
Instance FrameCollect : Collect AvarAbstractSyntax FrameAbstractSyntax :=
  @collect_frame.
(* Note: These explicit applications are here to deal with behavior of simpl tactic.
  A more straightforward way would be to use
  `collect_stack` instead of `@collect _ _ _ _ AvarAbstractSyntax _ _ _ _ StackAbstractSyntax (@collect_stack) C`
  If you look at how a similar definition looks for terms and types, it uses the simpler variant.
  However, the simpl tactic, when unfolding the fixpoint, replaces the fixpoint reference with the constant it is reducing.
  See https://coq.inria.fr/refman/proofs/writing-proofs/rewriting.html#coq:tacn.simpl
  That means that when reducing StackCollect, we would get StackCollect applied directly to fs.
  Then, the fold_stack_collect would not be able to fold this application, because it looks for collect_stack.
  For terms, this problem does not appear, maybe because the induction is mutual for terms, but not for stacks.
  To work around this problem, we use the class method `collect` here, applied not to a class instance StackCollect, but to collect_stack.
  Then, when StackCollect is unfolded, it will replace collect_stack with StackCollect and we will directly get `collect StackCollect`.
  On the other hand, there could be a problem if we reduce collect_stack. Try making sure that collect_stack does not appear by itself.
*)

Fixpoint collect_stack {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (st: stack) : C :=
  match st with
  | stack_nil => df
  | (stack_cons fr fs) => comb (collect df f comb fr) (@collect _ _ _ _ AvarAbstractSyntax _ _ _ _ StackAbstractSyntax (@collect_stack) C df f comb fs)
  end.
Instance StackCollect : Collect AvarAbstractSyntax StackAbstractSyntax :=
  @collect_stack.

Definition transform_frame (f : avar -> nat -> avar) (fr : frame) (n : nat) : frame :=
  match fr with
  | frame_let t => frame_let (transform f t n)
  end.
Instance FrameTransform : Transform AvarAbstractSyntax FrameAbstractSyntax :=
  @transform_frame.
Fixpoint transform_stack (f : avar -> nat -> avar) (st : stack) (n : nat) : stack :=
  match st with
  | stack_nil => stack_nil
  | (stack_cons fr fs) => (stack_cons (transform f fr (S n)) (@transform _ _ _ _ AvarAbstractSyntax _ _ _ _ StackAbstractSyntax transform_stack f fs n))
  end.
Instance StackTransform : Transform AvarAbstractSyntax StackAbstractSyntax :=
  @transform_stack.

Ltac fold_stack_open_rec :=
  change (@frame_open) with (@open_rec frame FrameOpenable);
  change (@stack_open) with (@open_rec stack StackOpenable).
Ltac fold_stack_fv :=
  change (@frame_fv) with (@fv frame FrameFreeVar);
  change (@stack_fv) with (@fv stack StackFreeVar).
Ltac fold_stack_subst :=
  change (@frame_subst_var) with (@subst_var frame FrameSubstVar);
  change (@stack_subst_var) with (@subst_var stack StackSubstVar).
Ltac fold_stack_abs :=
  fold_stack_open_rec; fold_stack_fv; fold_stack_subst.

Ltac fold_stack_collect :=
  change (@collect_frame) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            frame FrameOpenable FrameFreeVar FrameSubstVar FrameAbstractSyntax
            FrameCollect);
  change (@collect_stack) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            stack StackOpenable StackFreeVar StackSubstVar StackAbstractSyntax
            StackCollect).
Ltac fold_stack_transform :=
  change (@transform_frame) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            frame FrameOpenable FrameFreeVar FrameSubstVar FrameAbstractSyntax
            FrameTransform);
  change (@transform_stack) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            stack StackOpenable StackFreeVar StackSubstVar StackAbstractSyntax
            StackTransform).
Local Ltac sympl_stack_transform :=
  simpl; (* unfold FrameOpenable, StackOpenable, FrameTransform, StackTransform, FrameCollect, StackCollect, FrameSubstVar, StackSubstVar, FrameFreeVar, StackFreeVar;*)
  fold_stack_transform; fold_stack_collect; fold_stack_abs.


Lemma fv_collect_frame :
  (forall (T : frame), fv T = collect (\{}) fv (@union var) T).
Proof.
  intros.
  induction T; sympl_stack_transform; rewrite <- fv_collect; auto.
Qed.

Lemma open_transform_frame:
  (forall (T : frame) n x, (open_depth_rec n x) T = transform (open_depth_rec n x) T).
Proof.
  unfold open_depth_rec.
  intros.
  induction T.
  all: repeat (apply fun_ext_dep; intro).
  all: sympl_stack_transform.
  all: fequal.
  all: fold (@open_depth_rec avar _ n x).
  all: rewrite <- open_transform.
  all: unfold open_depth_rec; auto.
Qed.

Lemma subst_transform_frame :
  (forall (T : frame) x y, (ignore_second (subst_var x y)) T
                = transform (ignore_second (subst_var x y)) T).
Proof.
  intros; induction T.
  all: sympl_stack_transform; intros; intros_fun_ext.
  all: sympl_stack_transform.
  all: rewrite <- subst_transform.
  all: unfold ignore_second.
  all: simpls.
  all: auto.
Qed.

Local Ltac solve_transform_syntax_frame :=
  hnf; intros; (apply fun_ext_dep; intro);
    try apply fv_collect_frame;
    try apply open_transform_frame;
    try apply subst_transform_frame.

Instance FrameFvCollect : FvCollect FrameCollect.
Proof. solve_transform_syntax_frame. Qed.
Instance FrameOpenTransform : OpenTransform FrameTransform.
Proof. solve_transform_syntax_frame. Qed.
Instance FrameSubstTransform : SubstTransform FrameTransform.
Proof. solve_transform_syntax_frame. Qed.

Lemma fv_collect_stack :
  (forall (T : stack), fv T = collect (\{}) fv (@union var) T).
Proof.
  intros.
  induction T; sympl_stack_transform.
  - auto 1.
  - rewrite <- fv_collect; auto. rewrite IHT. unfold fv. auto.
Qed.

Lemma open_transform_stack :
  (forall (T : stack) n x, (open_depth_rec n x) T = transform (open_depth_rec n x) T).
Proof.
  unfold open_depth_rec.
  intros.
  induction T.
  all: sympl_stack_transform.
  all: intros_fun_ext.
  all: auto 1.
  all:  repeat match goal with
           | [ H :  _ = _ |- _] => rewrite <- H
           end;
    rewrite ? Nat.add_succ_r.
  all:  repeat match goal with
           | [ |- context[open_rec (?n + ?m) ?x ?t] ] =>
             change (open_rec (n + m) x t) with (open_depth_rec n x t m)
           end.
  all:
    rewrite ? open_transform;
    auto.
  all: fequal.
  all: fold (@open_depth_rec avar _ n x).
  all: rewrite <- open_transform.
  all: unfold open_depth_rec; rewrite plus_n_Sm; auto.
Qed.

Lemma subst_transform_stack :
  (forall (T : stack) x y, (ignore_second (subst_var x y)) T
                = transform (ignore_second (subst_var x y)) T).
Proof.
  intros; induction T.

  all: sympl_stack_transform; intros; intros_fun_ext;
    repeat match goal with
           | [ H : _ = _ |- _] => rewrite <- H
           end;
    rewrite <- ? subst_transform;
    unfold ignore_second; simpl; auto.
Qed.

Local Ltac solve_transform_syntax_stack :=
  hnf; intros; (apply fun_ext_dep; intro);
    try apply fv_collect_stack;
    try apply open_transform_stack;
    try apply subst_transform_stack.

Instance StackFvCollect : FvCollect StackCollect.
Proof. solve_transform_syntax_stack. Qed.
Instance StackOpenTransform : OpenTransform StackTransform.
Proof. solve_transform_syntax_stack. Qed.
Instance StackSubstTransform : SubstTransform StackTransform.
Proof. solve_transform_syntax_stack. Qed.

Lemma fv_subst_cases_frame : forall x y,
    (forall (T : frame), FvSubstCasesOf x y T).
Proof.
  intros. induction T.
  all: simpls.
  all: eapply fv_subst_cases.
Qed.
Instance FrameFvSubstCases: FvSubstCases FrameAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_frame.
Qed.


Lemma subst_id_frame (v1: var):
  (forall (t: frame),
  subst_var v1 v1 t = t).
Proof.
  intros. induction t.
  all: simpls. all: fequal. all: apply subst_id.
Qed.
Instance FrameSubstId : SubstId FrameAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite subst_id_frame. compute. auto.
Qed.
Lemma subst_id_stack (v1: var):
  (forall (t: stack),
  subst_var v1 v1 t = t).
Proof.
  intros. induction t.
  all: simpls; auto. all: fequal. eapply subst_id. unfold subst_var in IHt. auto.
Qed.
Instance StackSubstId : SubstId StackAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite subst_id_stack. compute. auto.
Qed.

Lemma fv_subst_cases_stack : forall x y,
    (forall (T : stack), FvSubstCasesOf x y T).
Proof.
  intros. induction T. all: simpls. all: fold_stack_abs.

  all: sympl_stack_transform.
  - destruct (classicT(x=y)); eauto.
  -
    destruct (fv_subst_cases x y f) as [[Hfva [Hx | Hna]] | [Hnxy [Hxa Hfva]]];
    destruct_all; subst; try rewrite H; try rewrite Hfva.
    all: sympl_stack_transform; eauto.
    + left. splits; eauto. rewrite subst_id. auto.
    + rewrite H1. right. splits; eauto. rewrite in_union; auto. rewrite union_remove_swap0 by auto. rewrite union_assoc. auto.
    + contradiction.
    + right. splits; eauto. rewrite in_union; auto. rewrite <- union_remove_swap2 by auto. rewrite union_assoc. auto.
    + rewrite H1. right. splits; eauto. rewrite in_union; auto. rewrite <- union_remove_dist.
  rewrite <- (union_comm \{y}).
  rewrite <- (union_comm \{y}).
  rewrite <- (union_comm \{y}).

  rewrite (union_assoc _ \{y}).
  rewrite <- (union_comm \{y}).
  rewrite  (union_assoc _ \{y}).
  rewrite union_same.
  rewrite  (union_assoc ). auto.
Qed.
Instance StackFvSubstCases: FvSubstCases StackAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_stack.
Qed.


Lemma transform_id_frame :
  (forall (T : frame) n, transform (ignore_second id) T n = T).
Proof.
  unfold ignore_second, id, transform; simpls.
  intros. induction T; simpl; intros; try f_equal; auto;
  rewrite transform_id; auto.
Qed.
Lemma transform_compose_frame:
  (forall (T : frame) n f g, ((transform f) ∘∘ (transform g)) T n = transform (f ∘∘ g) T n).
Proof.
  unfold compose12, transform; simpls.
  intros; induction T; simpl; intros; fequal.
  all:    rewrite ? (fold_compose12 (transform f));
      rewrite ? transform_compose; auto.
Qed.
Local Ltac solve_transform_sanity_frame :=
  hnf; try apply transform_id_frame;
    intros; intros_fun_ext; apply transform_compose_frame.
Instance FrameTransformId : TransformId FrameTransform.
Proof. solve_transform_sanity_frame. Qed.
Instance FrameTransformCompose : TransformCompose FrameTransform.
Proof. solve_transform_sanity_frame. Qed.

Lemma transform_id_stack :
  (forall (T : stack) n, transform (ignore_second id) T n = T).
Proof.
  unfold ignore_second, id, transform; simpls.
  intros. induction T; simpl; intros; try f_equal; auto;
  rewrite transform_id; auto.
Qed.
Lemma transform_compose_stack:
  (forall (T : stack) n f g, ((transform f) ∘∘ (transform g)) T n = transform (f ∘∘ g) T n).
Proof.
  unfold compose12, transform; simpls.
  intros; induction T; simpl; intros; fequal.
  all:    rewrite ? (fold_compose12 (transform f));
      rewrite ? transform_compose; auto.
Qed.
Local Ltac solve_transform_sanity_stack :=
  hnf; try apply transform_id_stack;
    intros; intros_fun_ext; apply transform_compose_stack.
Instance StackTransformId : TransformId StackTransform.
Proof. solve_transform_sanity_stack. Qed.
Instance StackTransformCompose : TransformCompose StackTransform.
Proof. solve_transform_sanity_stack. Qed.

Lemma collect_compose_frame :
  forall {C D} (f : C -> D) g comb1 comb2 i,
    (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
    (forall (T : frame),
        (f ∘ collect i g comb1) T = collect (f i) (f ∘ g) comb2 T).
Proof.
  intros.
  induction T;
    unfold compose, collect, FrameCollect.
  all: simpls.
  all: rewrite  ? (fold_compose f (collect (C:=C) _ _ _)).
  all: erewrite (collect_compose); eauto.
Qed.

Lemma collect_reflect_frame:
  forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall bl,
      (forall T, ((collect_frame (Bool.Is_true bl) P or T)
             <-> (collect_frame bl p orb T) = true)).
Proof.
  intros P p H bl.
  intro T.
  induction T; simpl; auto 2 using reflect_iff.
  all: unfold iff; auto 3 using Is_true_eq_left, Is_true_eq_true.
  all: split.
  all: destruct_ands; intros.

  all: rewrite ? collect_reflect; eauto.
  all: rewrite <- collect_reflect; eauto.
Qed.

Local Ltac solve_collect_compose_frame :=
  unfold compose; apply fun_ext_dep;
  apply collect_compose_frame; auto.

Local Ltac solve_collect_reflect_frame :=
  apply collect_reflect_frame; auto.

Local Ltac solve_collect_sanity_frame :=
  hnf; intros; solve [solve_collect_compose_frame || solve_collect_reflect_frame].

Instance FrameCollectCompose : CollectCompose FrameCollect.
Proof. solve_collect_sanity_frame. Qed.
Instance FrameCollectReflect : CollectReflect FrameCollect.
Proof. solve_collect_sanity_frame. Qed.

Lemma collect_transform_frame:
  (forall (t : frame) bl p f g,
      (collect bl p andb t) = true ->
      transform (dec_apply p f g) t = transform f t).
Proof.
  intro t.
  induction t; intros.
  all: unfold transform, FrameTransform, transform_frame.
  all: rewrite (collect_transform bl); eauto.
Qed.

Instance FrameCollectTransform :
  CollectTransform FrameCollect FrameTransform.
Proof. hnf; intros; eapply collect_transform_frame; eauto. Qed.


Lemma collect_compose_stack :
  forall {C D} (f : C -> D) g comb1 comb2 i,
    (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
    (forall (T : stack),
        (f ∘ collect i g comb1) T = collect (f i) (f ∘ g) comb2 T).
Proof.
  intros.
  induction T.
  all: unfold "∘" in *.
  all:    sympl_stack_transform;
      try congruence.
  all:  rewrite ? H,
              ? (fold_compose f (collect (C:=C) _ _ _)),
              ? (collect_compose _ _ _ _ _ H).
  all:    repeat match goal with
             | [ H : _ = _ |- _] => rewrite <- ? H; clear H
             end; auto.
Qed.

Lemma collect_reflect_stack:
  forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall bl,
      (forall T, ((collect_stack (Bool.Is_true bl) P or T)
             <-> (collect_stack bl p orb T) = true)).
Proof.
  intros P p H bl.
  intro T.
  induction T; sympl_stack_transform.
  all:
    auto 3 using Is_true_iff_eq_true; intros;
      repeat apply or_orb_iff_left; try apply collect_reflect; auto.
Qed.

Local Ltac solve_collect_compose_stack :=
  unfold compose; apply fun_ext_dep;
  apply collect_compose_stack; auto.

Local Ltac solve_collect_reflect_stack :=
  apply collect_reflect_stack; auto.

Local Ltac solve_collect_sanity_stack :=
  hnf; intros; solve [solve_collect_compose_stack || solve_collect_reflect_stack].

Instance StackCollectCompose : CollectCompose StackCollect.
Proof. solve_collect_sanity_stack. Qed.
Instance StackCollectReflect : CollectReflect StackCollect.
Proof. solve_collect_sanity_stack. Qed.

Lemma collect_transform_stack:
  (forall (t : stack) bl p f g,
      (collect bl p andb t) = true ->
      transform (dec_apply p f g) t = transform f t).
Proof.
 intro t.
  induction t.
  all: sympl_stack_transform; auto 2; intros;
    repeat (destruct_andbs; destruct_ands);
    repeat erewrite collect_transform by eauto;
    repeat
      match goal with
      | [ H : forall bl p f g,
            _ = true ->
            ?transform (dec_apply p f g) ?t = ?transform f ?t |- _] =>
        erewrite H by eauto
      end;
    repeat match goal with
           | [ H : _ = true |- _ ] => try unfold dec_apply; rewrite H
           end; auto.
Qed.

Instance StackCollectTransform :
  CollectTransform StackCollect StackTransform.
Proof. hnf; intros; eapply collect_transform_stack; eauto. Qed.


Lemma stack_closed_nil: closed stack_nil.
Proof.
  unfold closed. intros. simpls. auto.
Qed.
Lemma stack_closed_let t s:
  closed s ->
  (forall x, closed (open x t)) ->
  closed (stack_cons (frame_let t) s).
Proof.
  intros. unfold closed. intros. simpls. unfold closed in H. rewrite H. specialize (H0 x).
  eapply closed_open_down in H0. rewrite H0. auto.
Qed.
Lemma stack_closed_fs t s:
  closed (stack_cons (frame_let t) s) ->
  closed s.
Proof.
  unfold closed.
  intros.
  specialize (H n x).
  injection H; intros. eauto.
Qed.
Lemma stack_closed_t t s:
  closed (stack_cons (frame_let t) s)%list ->
  (forall x, closed (open x t)).
Proof.
  unfold closed.
  intros.
  destruct n.
  - rewrite open_twice. auto.
  - specialize (H n x0).
    simpls.
    injection H; intros.
    rewrite open_commut. rewrite H1. auto.
    auto.
Qed.

Lemma frame_subst_vars_let t xs ys:
  (frame_let t) /[xs --> ys] = (frame_let (t /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma stack_subst_vars_nil xs ys:
  (stack_nil /[xs --> ys]) = stack_nil.
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma stack_subst_vars_cons (fs: stack) (f: frame) xs ys:
  (stack_cons f fs) /[xs --> ys] = (stack_cons (f /[xs --> ys]) (fs /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Fixpoint stack_append (fs1 fs2: stack) :=
  match fs1 with
  | stack_nil => fs2
  | stack_cons f fs => stack_cons f (stack_append fs fs2)
  end.

Definition stack_suffix fs1 fs2 := exists fs3, fs2 = stack_append fs3 fs1.

Lemma stack_append_length (fs1 fs2 : stack) :
  stack_length (stack_append fs1 fs2) =
    (stack_length fs1) + (stack_length fs2).
Proof. induction fs1. simpl. auto. simpl. omega. Qed.

Lemma stack_nil_zero (fs : stack) :
  stack_length fs = 0 <-> fs = stack_nil.
Proof. split; intros. destruct fs. auto. simpl in H. omega. subst. auto. Qed.

Lemma stack_nil_pos (fs : stack) :
  stack_length fs > 0 <-> fs <> stack_nil.
Proof.
  split; intros. destruct fs. auto. simpl in H. omega. discriminate.
  destruct fs. false. simpl. omega.
Qed.

Lemma stack_append_longer_l (fs1 fs2 : stack) :
  fs2 <> stack_nil -> stack_length (stack_append fs1 fs2) > stack_length fs1.
Proof.
  intros. rewrite stack_append_length.
  pose proof (proj2 (stack_nil_pos fs2) H). omega.
Qed.

Lemma stack_append_longer_r (fs1 fs2 : stack) :
  fs1 <> stack_nil -> stack_length (stack_append fs1 fs2) > stack_length fs2.
Proof.
  intros. rewrite stack_append_length.
  pose proof (proj2 (stack_nil_pos fs1) H). omega.
Qed.

Lemma subst_vars_stack_length fs xs ys :
  stack_length (fs /[ xs --> ys]) = stack_length fs.
Proof.
  induction fs.
  - rewrite stack_subst_vars_nil. auto.
  - rewrite stack_subst_vars_cons. simpl. omega.
Qed.

Lemma stack_append_nil_r fs:
  stack_append fs stack_nil = fs.
Proof.
  induction fs; simpl; eauto; rewrite IHfs; eauto.
Qed.
Lemma stack_append_assoc fs1 fs2 fs3:
  stack_append (stack_append fs1 fs2) fs3 = stack_append fs1 (stack_append fs2 fs3).
Proof.
  induction fs1; simpl; eauto; rewrite IHfs1; eauto.
Qed.

Lemma stack_suffix_refl fs:
  stack_suffix fs fs.
Proof.
  hnf. exists stack_nil. simpl. auto.
Qed.

Lemma stack_suffix_trans fs1 fs2 fs3:
  stack_suffix fs2 fs1 ->
  stack_suffix fs3 fs2 ->
  stack_suffix fs3 fs1.
Proof.
  introv H12 H23.
  hnf in *.
  destruct_all; subst.
  rewrite <- stack_append_assoc. eauto.
Qed.

Lemma stack_suffix_nil fs:
  stack_suffix stack_nil fs.
Proof.
  hnf.
  exists fs. rewrite stack_append_nil_r. auto.
Qed.

Lemma stack_suffix_nil_inv fs:
  stack_suffix fs stack_nil ->
  fs = stack_nil.
Proof.
  introv Hpr.
  hnf in *.
  destruct_all.
  destruct fs. auto.
  induction x; simpls; discriminate.
Qed.

Lemma stack_suffix_append_r fs1 fs2 fs3:
  stack_suffix fs1 fs3 ->
  stack_suffix fs1 (stack_append fs2 fs3).
Proof.
  introv Hpr.
  hnf in *.
  destruct_all; subst.
  rewrite <- stack_append_assoc.
  eauto.
Qed.

Lemma stack_suffix_cons fs1 fs2 f:
  stack_suffix fs1 fs2 ->
  stack_suffix fs1 (stack_cons f fs2).
Proof.
  introv Hpr.
  hnf in *.
  destruct_all; subst.
  exists (stack_cons f x). simpl. auto.
Qed.
Lemma stack_suffix_cons_inv fs1 fs2 f2:
  stack_suffix fs1 (stack_cons f2 fs2) ->
  fs1 = (stack_cons f2 fs2) \/ stack_suffix fs1 fs2.
Proof.
  introv Hpr.
  hnf in *.
  destruct_all; subst.
  reverts fs1 fs2.
  induction x; intros; simpls.
  - eauto.
  - inversions H. right. eapply stack_suffix_append_r. eapply stack_suffix_refl.
Qed.

Lemma stack_suffix_cons_same_false fs f:
  stack_suffix (stack_cons f fs) fs ->
  False.
Proof.
  introv Hsu. hnf in Hsu; destruct_all.
  reverts x f; induction fs; intros.
  - induction x. simpls. discriminate.
    simpls. discriminate.
  - induction x. simpls. inversions H. specialize (IHfs stack_nil). simpls. eauto.
    simpls. inversions H.
    eapply (IHfs (stack_append x (stack_cons f0 stack_nil))).
    rewrite stack_append_assoc. simpls. eauto.
Qed.

Lemma stack_append_l_injective fs1 fs2 fs :
  stack_append fs1 fs = stack_append fs2 fs ->
  fs1 = fs2.
Proof.
  introv Heq. reverts fs2 fs. induction fs1; intros; simpl.
  - destruct fs2; eauto. false.
    forwards: stack_append_longer_r.
    2: { rewrite <- Heq in H. simpl in H. omega. }
    discriminate.
  - destruct fs2.
    + forwards: stack_append_longer_r.
      2: { rewrite Heq in H. simpl in H. omega. }
      discriminate.
    + inversions Heq; eauto. erewrite IHfs1; eauto.
Qed.

Lemma fv_stack_append fs1 fs2 :
  fv (stack_append fs1 fs2) = fv fs1 \u fv fs2.
Proof.
  induction fs1; simpl. rewrite union_empty_l; eauto.
  rewrite IHfs1. eauto using union_assoc.
Qed.
