Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import AbstractSyntax Context Defs Heap.

(** * T-free variables *)
(* T-free variables are free variables which occur used as values
   (excluding type selection prefixes).
*)

Class TFree(A : Set) := tfree: var -> A -> Prop.
Hint Unfold tfree: core.

Inductive tfree_trm: TFree trm :=
(* In a variable term, the variable is t-free *)
| tfree_trm_var v:
  tfree_trm v (trm_var (avar_f v))
(* In a read term, the object reference is t-free *)
| tfree_trm_read v1 l:
  tfree_trm v1 (trm_read (avar_f v1) l)
(* In an apply term, the function reference and the argument is t-free *)
| tfree_trm_apply_l v1 a2:
  tfree_trm v1 (trm_apply (avar_f v1) a2)
| tfree_trm_apply_r a1 v2:
  tfree_trm v2 (trm_apply a1 (avar_f v2))
(* In a let term, t-free variables are in the subterms *)
| tfree_trm_let_l v t1 t2:
  tfree_trm v t1 ->
  tfree_trm v (trm_let t1 t2)
| tfree_trm_let_r v t1 t2:
  tfree_trm v t2 ->
  tfree_trm v (trm_let t1 t2)
(* In a let term, t-free variables are in the subterms *)
| tfree_trm_lit_l v t1 t2:
  tfree_lit v t1 ->
  tfree_trm v (trm_lit t1 t2)
| tfree_trm_lit_r v t1 t2:
  tfree_trm v t2 ->
  tfree_trm v (trm_lit t1 t2)
(* In a write term, the object reference and the argument is t-free *)
| tfree_trm_write_l v1 a2 l:
  tfree_trm v1 (trm_write (avar_f v1) l a2)
| tfree_trm_write_r a1 v2 l:
  tfree_trm v2 (trm_write a1 l (avar_f v2))
(* In a call term, the object reference and the argument is t-free *)
| tfree_trm_call_l v1 a2 l:
  tfree_trm v1 (trm_call (avar_f v1) l a2)
| tfree_trm_call_r a1 v2 l:
  tfree_trm v2 (trm_call a1 l (avar_f v2))

with tfree_lit: TFree lit :=
(* In literals, t-free variables are in the subterms *)
| tfree_lit_fun v t T:
  tfree_trm v t ->
  tfree_lit v (lit_fun T t)
| tfree_lit_obj v t T:
  tfree_defs v t ->
  tfree_lit v (lit_obj T t)

with tfree_def: TFree def :=
(* In a field, the value is t-free *)
| tfree_def_trm v l:
  tfree_def v (def_fld l (avar_f v))
(* In methods, t-free variables are in the subterms *)
| tfree_def_met v t l:
  tfree_trm v t ->
  tfree_def v (def_met l t)

with tfree_defs: TFree defs :=
(* In definitions, t-free variables are in the subterms *)
| tfree_defs_cons_l v d ds:
  tfree_defs v ds ->
  tfree_defs v (defs_cons ds d)
| tfree_defs_cons_r v d ds:
  tfree_def v d ->
  tfree_defs v (defs_cons ds d)
.

Hint Constructors tfree_trm tfree_lit tfree_def tfree_defs : core.

Scheme tfree_trm_mut    := Induction for tfree_trm Sort Prop
with   tfree_lit_mut    := Induction for tfree_lit Sort Prop
with   tfree_def_mut    := Induction for tfree_def Sort Prop
with   tfree_defs_mut   := Induction for tfree_defs Sort Prop.
Combined Scheme tfree_mutind from tfree_trm_mut, tfree_lit_mut,
tfree_def_mut, tfree_defs_mut.

Inductive tfree_stack: TFree stack :=
(* In the stack, t-free variables are in the subterms *)
| tfree_stack_cons_l v t st:
  tfree_trm v t ->
  tfree_stack v (stack_cons (frame_let t) st)
| tfree_stack_cons_r v f st:
  tfree_stack v st ->
  tfree_stack v (stack_cons f st)
.

Existing Instances tfree_trm tfree_lit tfree_def tfree_defs tfree_stack.

(* Valid configurations require that all t-free variables
   in the focus, on the stack, and in methods are references.
*)
Definition all_tfree_k {A: Set} {AT: TFree A} G (t:A) vk :=
  forall v, tfree v t -> ctx_k G v vk.
Notation all_tfree_k_trm := (@all_tfree_k trm _).
Notation all_tfree_k_def := (@all_tfree_k def _).
Notation all_tfree_k_defs := (@all_tfree_k defs _).
Notation all_tfree_k_stack := (@all_tfree_k stack _).

(* T-free variables are free *)

Definition TFreeInFV (A: Set) (AT: TFree A) {AF: FreeVar A} :=
  forall v t,
    tfree v t ->
    v \in fv t.

Lemma tfree_in_fv:
  TFreeInFV tfree_trm /\
  TFreeInFV tfree_lit /\
  TFreeInFV tfree_def /\
  TFreeInFV tfree_defs.
Proof.
  unfold TFreeInFV.
  intros.
  apply tfree_mutind.
  all: intros.
  all: simpl.
  all: try rewrite in_union.
  all: auto using in_singleton_self.
Qed.

Lemma tfree_in_fv_trm v t:
  tfree_trm v t ->
  v \in fv t.
Proof.
  intros.
  apply tfree_in_fv.
  auto.
Qed.

Lemma avar_open_eq v k x a:
  (avar_f v) = (open_rec k x a) ->
  a = (avar_f v) \/ x = v.
Proof.
  intros.
  unfold open_rec in *.
  unfold AvarOpenable in *.
  destruct a.
  destruct (classicT (k = n)); subst.
  all: simplify_eq H; intros; subst; auto.
Qed.

(* Opening keeps t-free variables and may introduce another one *)
Local Definition TFreeOpenInvInd (A: Set) (AT: TFree A) {AO: Openable A} :=
  forall v t,
    tfree v t ->
    forall k x tR,
      t = (open_rec k x tR) ->
      tfree v tR \/ v = x.

Local Lemma tfree_open_inv_ind:
  TFreeOpenInvInd tfree_trm /\
  TFreeOpenInvInd tfree_lit /\
  TFreeOpenInvInd tfree_def /\
  TFreeOpenInvInd tfree_defs.
Proof.
  unfold TFreeOpenInvInd.
  unfold tfree.
  apply tfree_mutind.
  all: intros.
  all: destruct tR.
  all: match goal with
       | H: _ = _ |- _ => simplify_eq H; clear H
       end.
  all: intros.
  all: sympl in *.
  all: repeat try match goal with
       | H: (avar_f _) = _ |- _ => apply avar_open_eq in H; destruct H; subst
       end.
  all: auto.
  all: subst.
  all: match goal with
       | H: forall k0 x0 t0,
            open_rec ?k ?x ?t = open_rec k0 x0 t0 -> _
            |- _
            => destruct (H k x t); auto
       end.
Qed.

Class TFreeOpenInv(A : Set)(AT: TFree A){AO: Openable A} :=
  tfree_open_inv v x (t:A) n:
    tfree v (open_rec n x t) ->
    tfree v t \/ v = x
.

Instance tfree_open_inv_trm: TFreeOpenInv tfree_trm.
Proof.
  instance_by_ind TFreeOpenInv tfree_open_inv_ind.
Qed.
Hint Resolve tfree_open_inv_trm: core.

Instance tfree_open_inv_defs: TFreeOpenInv tfree_defs.
Proof.
  instance_by_ind TFreeOpenInv tfree_open_inv_ind.
Qed.
Hint Resolve tfree_open_inv_defs: core.

Local Definition TFreeOpenInd (A: Set) (AT: TFree A) {AO: Openable A} :=
  forall v t,
    tfree v t ->
    forall k x,
    tfree v (open_rec k x t).

Local Lemma tfree_open_ind:
  TFreeOpenInd tfree_trm /\
  TFreeOpenInd tfree_lit /\
  TFreeOpenInd tfree_def /\
  TFreeOpenInd tfree_defs.
Proof.
  unfold TFreeOpenInd.
  apply tfree_mutind.
  all: intros.

  all: sympl in *.
  all: eauto.
  all: solve [constructor; eapply H].
Qed.

Class TFreeOpen(A : Set)(AT: TFree A){AO: Openable A} :=
  tfree_open v x (t:A) n:
    tfree v t ->
    tfree v (open_rec n x t).

Instance tfree_open_trm: TFreeOpen tfree_trm.
Proof.
  instance_by_ind TFreeOpen tfree_open_ind.
Qed.
Hint Resolve tfree_open_trm: core.

Instance tfree_open_defs: TFreeOpen tfree_defs.
Proof.
  instance_by_ind TFreeOpen tfree_open_ind.
Qed.
Hint Resolve tfree_open_defs: core.

(* Lemmata about stacks and terms where all t-free variables are references *)
Lemma all_tfree_k_stack_nil G vk:
  all_tfree_k_stack G stack_nil vk.
Proof.
  unfold all_tfree_k.
  intros.
  inversion H.
Qed.
Hint Resolve all_tfree_k_stack_nil: core.

Lemma all_tfree_k_stack_push G st t vk:
  all_tfree_k G t vk ->
  all_tfree_k_stack G st vk ->
  all_tfree_k_stack G (stack_cons (frame_let t) st) vk.
Proof.
  unfold all_tfree_k.
  intros.
  inversion H1; auto.
Qed.
Hint Resolve all_tfree_k_stack_push: core.

Lemma all_tfree_k_stack_push_inv G st t vk:
  all_tfree_k_stack G (stack_cons (frame_let t) st) vk ->
  all_tfree_k G t vk /\
  all_tfree_k_stack G st vk.
Proof.
  unfold all_tfree_k.
  split; intros.
  apply H.
  apply tfree_stack_cons_l. auto.
  apply H.
  apply tfree_stack_cons_r. auto.
Qed.

Lemma all_tfree_k_Gpush (A: Set) {AT: TFree A} G (t:A) vk1 y vk2 T:
  all_tfree_k G t vk1 ->
  y §# G ->
  all_tfree_k (G § y / vk2 ~ T) t vk1.
Proof.
  unfold all_tfree_k.
  intros.
  apply ctx_k_push_n; auto.
Qed.
Hint Resolve all_tfree_k_Gpush: core.

Lemma all_tfree_k_open_trm G t x n vk:
  all_tfree_k_trm G t vk ->
  ctx_k G x vk ->
  all_tfree_k_trm G (open_rec n x t) vk.
Proof.
  unfold all_tfree_k.
  intros.
  destruct (tfree_open_inv_trm x t n H1);
  subst; auto.
Qed.

Hint Resolve all_tfree_k_open_trm: core.

Lemma all_tfree_k_trm_avar G x vk:
  ctx_k G x vk ->
  all_tfree_k G (trm_var (avar_f x)) vk.
Proof.
  unfold all_tfree_k.
  intros.
  inversion H0.
  subst; auto.
Qed.

Hint Resolve all_tfree_k_trm_avar: core.

Lemma all_tfree_k_trm_avar_inv G v vk:
  all_tfree_k G (trm_var (avar_f v)) vk ->
  ctx_k G v vk
.
Proof.
  unfold all_tfree_k.
  intros. apply H.
  apply tfree_trm_var.
Qed.

Lemma all_tfree_let_inv G t u vk:
  all_tfree_k G (trm_let t u) vk ->
  (all_tfree_k G t vk) /\ (all_tfree_k G u vk).
Proof.
  unfold all_tfree_k, tfree.
  split; intros; auto.
Qed.

Section TFreeOpenLemmata.
Context (A: Set) {AT: TFree A} {AO: Openable A} {ATO: TFreeOpenInv AT}.

Lemma all_tfree_k_open G v vk (t: A):
  all_tfree_k G t vk ->
  ctx_k G v vk ->
  all_tfree_k G (open v t) vk.
Proof.
  unfold all_tfree_k.
  intros.
  apply tfree_open_inv in H1.
  destruct H1; subst; auto.
Qed.

End TFreeOpenLemmata.



Lemma defs_has_tfree_lit d a w R t:
  defs_has d (def_fld a (avar_f w)) ->
  tfree w (trm_lit (lit_obj R d) t).
Proof.
  introv Hhas.
  induction d.
  - inversions Hhas.
  - destruct (defs_has_cons_inv Hhas). subst. eauto. specialize (IHd H). inversions IHd; eauto. inversions H2; eauto.
Qed.

Lemma defs_has_tfree d a w:
  defs_has d (def_fld a (avar_f w)) ->
  tfree w d.
Proof.
  introv Hhas.
  induction d.
  - inversions Hhas.
  - destruct (defs_has_cons_inv Hhas). subst. eauto. specialize (IHd H). inversions IHd; eauto.
Qed.

Lemma tfree_sbust_inv_mutind w y :
  (forall (v : var) (t0 : trm),
      tfree_trm v t0 ->
      forall t : trm,
        t0 = (t /[ w -> y]) -> tfree v t \/ tfree w t /\ v = y) /\
    (forall (v : var) (l0 : lit),
        tfree_lit v l0 ->
        forall l : lit,
          l0 = (l /[ w -> y]) -> tfree v l \/ tfree w l /\ v = y) /\
    (forall (v : var) (d0 : def),
        tfree_def v d0 ->
        forall d : def,
          d0 = (d /[ w -> y]) -> tfree v d \/ tfree w d /\ v = y) /\
    (forall (v : var) (ds0 : defs),
        tfree_defs v ds0 ->
        forall ds : defs,
          ds0 = (ds /[ w -> y]) -> tfree v ds \/ tfree w ds /\ v = y).
Proof.
  apply
    (tfree_mutind
       (fun v t0 _ => forall (t : trm), t0 = (t /[ w->y])
                                -> (tfree v t) \/ (tfree w t /\ v = y))
       (fun v l0 _ => forall (l : lit), l0 = (l /[ w->y])
                                -> (tfree v l) \/ (tfree w l /\ v = y))
       (fun v d0 _ => forall (d : def), d0 = (d /[ w->y])
                                -> (tfree v d) \/ (tfree w d /\ v = y))
       (fun v ds0 _ => forall (ds : defs), ds0 = (ds /[ w->y])
                                   -> (tfree v ds) \/ (tfree w ds /\ v = y))).
  1,2,3,4: introv Heq.
  5,6,7,8: introv Htf IH Heq.
  9,10,11,12: introv Heq.
  13,14: introv Htf IH Heq.
  15: introv Heq.
  16,17,18: introv Htf IH Heq.

  all: try(match goal with
           | [ |- tfree _ ?t \/ _ ] => destruct t; simpl in *; try discriminate
           end).
  all: inversions Heq.
  all: try solve [ destruct a; simpl in *; try discriminate; inversions H0;
                   destruct (classicT (eq v w)); subst; eauto
                 | destruct a; simpl in *; try discriminate; inversions H0;
                   destruct (classicT (eq v0 w)); subst; eauto
                 | destruct a; simpl in *; try discriminate; inversions H1;
                   destruct (classicT (eq v0 w)); subst; eauto
                 | destruct a0; simpl in *; try discriminate; inversions H1;
                   destruct (classicT (eq v w)); subst; eauto
                 | destruct a0; simpl in *; try discriminate; inversions H2;
                   destruct (classicT (eq v w)); subst; eauto
                 | specialize (IH _ eq_refl); destruct_all; eauto ].
Qed.

Lemma tfree_sbust_inv_trm v w y (t : trm) :
  tfree v (t /[ w -> y]) ->
  tfree v t \/ tfree w t /\ v = y.
Proof. destruct (tfree_sbust_inv_mutind w y); destruct_all; eauto. Qed.

Lemma tfree_sbust_inv_lit v w y (l : lit) :
  tfree v (l /[ w -> y]) ->
  tfree v l \/ tfree w l /\ v = y.
Proof. destruct (tfree_sbust_inv_mutind w y); destruct_all; eauto. Qed.

Lemma tfree_sbust_inv_defs v w y (ds : defs) :
  tfree v (ds /[ w -> y]) ->
  tfree v ds \/ tfree w ds /\ v = y.
Proof. destruct (tfree_sbust_inv_mutind w y); destruct_all; eauto. Qed.

Lemma tfree_sbust_inv_def v w y (d : def) :
  tfree v (d /[ w -> y]) ->
  tfree v d \/ tfree w d /\ v = y.
Proof. destruct (tfree_sbust_inv_mutind w y); destruct_all; eauto. Qed.

Lemma tfree_open_met xz xc (t : trm) v :
  tfree v t -> tfree v (open_met xz xc t).
Proof. introv Htf. eauto using tfree_open. Qed.

Lemma fields_tfree_def x (d : def) : x \in fields d -> tfree x d.
Proof.
  destruct d; unfold fields, FieldsDef, fields_def;
    try(rewrite in_empty); try tauto.
  destruct a; try(rewrite in_empty); try tauto.
  rewrite in_singleton; intros; subst. eauto.
Qed.

Lemma fields_tfree_defs x (ds : defs) : x \in fields ds -> tfree x ds.
Proof.
  unfold fields, FieldsDefs. induction ds;
    try(rewrite in_empty); try tauto. simpl.
  rewrite in_union. introv [Hin | Hin]; eauto.
  - eapply tfree_defs_cons_l, IHds. easy.
  - eapply tfree_defs_cons_r, fields_tfree_def. easy.
Qed.
