(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Import LibExtra DotTactics SyntaxClasses.
Require Import Vars Labels TransformCollect Types.

(* ************************************************************************** *)
(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** * Terms
  Terms ([trm], [t], [u]), member definitions ([def], [d] and [defs], [ds]):
  - [trm_var x] represents a variable [x];
  - [trm_read x a] represents a field selection [x.a];
  - [trm_apply x y] represents a function application [x y];
  - [trm_let t u] represents a let binding [let x = t in u]; since x is bound in [u],
    it is referred to in [u] using the de Bruijn index 0, and is therefore omitted from
    the let-term representation; we will denote let terms as [let t in u];
  - [trm_let t u] represents a let binding of a literal [let x = l in u];
  - [trm_write x a y] represents a field assignment [x.a := y];
  - [trm_call x m y] represents a method call application [x.m(y)];
  - [lit_fun T t] represents a function defintion [λ(x: T)t];
  - [lit_obj R ds] represents a obect literal [ν(s:R)ds];
 *)
Inductive trm : Set :=
  | trm_var  : avar -> trm
  | trm_read  : avar -> fld_label -> trm
  | trm_apply  : avar -> avar -> trm
  | trm_let  : trm -> trm -> trm
  | trm_lit  : lit -> trm -> trm
  | trm_write  : avar -> fld_label -> avar -> trm
  | trm_call: avar -> met_label -> avar -> trm
(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
with lit : Set :=
  | lit_fun : typ -> trm -> lit
  | lit_obj  : typ -> defs -> lit
(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(**
  - [def_typ A T] represents a type-member definition [{A(r) = T}];
  - [def_fld a x] represents a field definition [{a = x}];
  - [def_met m t] represents a method definition [{m(z,r) = t}]; *)
with def : Set :=
  | def_typ  : typ_label -> typ -> def
  | def_fld  : fld_label -> avar -> def
  | def_met  : met_label -> trm -> def
(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(**
  [defs] represents a list of definitions that are part of an intersection
  - [defs_nil] represents the empty list;
  - [defs_cons d ds] represents a concatenation of the definition [d] to the definitions [ds]. *)
with defs : Set :=
  | defs_nil : defs
  | defs_cons : defs -> def -> defs.
Hint Constructors trm def defs : core.

(** *** Mutual Induction Principles *)
Scheme trm_mut  := Induction for trm  Sort Prop
with   lit_mut  := Induction for lit  Sort Prop
with   def_mut  := Induction for def  Sort Prop
with   defs_mut := Induction for defs Sort Prop.
Combined Scheme trm_mutind from trm_mut, lit_mut, def_mut, defs_mut.


(* ************************************************************************** *)
(** ** Terms are Abstract Syntax *)
Fixpoint open_rec_trm (k: nat) (u: var) (t: trm): trm :=
  match t with
  | trm_var a      => trm_var (open_rec k u a)
  | trm_read v m    => trm_read (open_rec k u v) m
  | trm_apply f a    => trm_apply (open_rec k u f) (open_rec k u a)
  | trm_let t1 t2  => trm_let (open_rec_trm k u t1) (open_rec_trm (S k) u t2)
  | trm_lit l t => trm_lit (open_rec_lit k u l) (open_rec_trm (S k) u t)
  | trm_write v m v2 => trm_write (open_rec k u v) m (open_rec k u v2)
  | trm_call v m v2 => trm_call (open_rec k u v) m (open_rec k u v2)
  end with
open_rec_lit (k: nat) (u: var) (l: lit): lit :=
  match l with
  | lit_fun T t => lit_fun (open_rec k u T) (open_rec_trm (S k) u t)
  | lit_obj T ds =>
    lit_obj
      (open_rec (S k) u T)
      (open_rec_defs (S k) u ds)
      (* avar_n 0 refers to self/this *)
  end with
open_rec_def (k: nat) (u: var) (d: def): def :=
  match d with
  | def_typ L T => def_typ L (open_rec (S k) u T)
  | def_fld m v => def_fld m (open_rec k u v)
  | def_met m e => def_met m (open_rec_trm (S (S k)) u e)
  end with
open_rec_defs (k: nat) (u: var) (ds: defs): defs :=
  match ds with
  | defs_nil       => defs_nil
  | defs_cons tl d => defs_cons (open_rec_defs k u tl) (open_rec_def k u d)
  end.
Instance TrmOpenable : Openable trm := open_rec_trm.
Instance LitOpenable : Openable lit := open_rec_lit.
Instance DefOpenable : Openable def := open_rec_def.
Instance DefsOpenable : Openable defs := open_rec_defs.
Ltac fold_trm_open_rec :=
  change (open_rec_trm) with (@open_rec trm TrmOpenable);
  change (open_rec_lit) with (@open_rec lit LitOpenable);
  change (open_rec_def) with (@open_rec def DefOpenable);
  change (open_rec_defs) with (@open_rec defs DefsOpenable).

Fixpoint fv_trm (t: trm) : vars :=
  match t with
  | trm_var a         => (fv a)
  | trm_read x m       => (fv x)
  | trm_apply f a       => (fv f) \u (fv a)
  | trm_let t1 t2     => (fv_trm t1) \u (fv_trm t2)
  | trm_lit l t       => (fv_lit l) \u (fv_trm t)
  | trm_write x m y     => (fv x) \u (fv y)
  | trm_call x m y => (fv x) \u (fv y)
  end with
fv_lit (l: lit) : vars :=
  match l with
  | lit_fun T e    => (fv T) \u (fv_trm e)
  | lit_obj T ds => (fv T) \u (fv_defs ds)
  end with
fv_def (d: def) : vars :=
  match d with
  | def_typ _ T     => (fv T)
  | def_fld _ v     => (fv v)
  | def_met _ t     => (fv_trm t)
  end with
fv_defs(ds: defs) : vars :=
  match ds with
  | defs_nil         => \{}
  | defs_cons tl d   => (fv_defs tl) \u (fv_def d)
  end.
Instance TrmFreeVar : FreeVar trm := fv_trm.
Instance LitFreeVar : FreeVar lit := fv_lit.
Instance DefFreeVar : FreeVar def := fv_def.
Instance DefsFreeVar : FreeVar defs := fv_defs.
Ltac fold_trm_fv :=
  change (fv_trm) with (@fv trm TrmFreeVar);
  change (fv_lit) with (@fv lit LitFreeVar);
  change (fv_def) with (@fv def DefFreeVar);
  change (fv_defs) with (@fv defs DefsFreeVar).

Fixpoint subst_trm (z: var) (u: var) (t: trm) : trm :=
  match t with
  | trm_var x        => trm_var (subst_var z u x)
  | trm_read x1 L     => trm_read (subst_var z u x1) L
  | trm_apply x1 x2    => trm_apply (subst_var z u x1) (subst_var z u x2)
  | trm_let t1 t2    => trm_let (subst_trm z u t1) (subst_trm z u t2)
  | trm_lit l t      => trm_lit (subst_lit z u l) (subst_trm z u t)
  | trm_write x1 L x2  => trm_write (subst_var z u x1) L (subst_var z u x2)
  | trm_call x1 L x2  => trm_call (subst_var z u x1) L (subst_var z u x2)
  end with
subst_lit (z: var) (u: var) (l: lit) : lit :=
  match l with
  | lit_fun T t   => lit_fun (subst_var z u T) (subst_trm z u t)
  | lit_obj T ds =>
    lit_obj (subst_var z u T)
            (subst_defs z u ds)
  end with
subst_def (z: var) (u: var) (d: def) : def :=
  match d with
  | def_typ l t => def_typ l (subst_var z u t)
  | def_fld l v => def_fld l (subst_var z u v)
  | def_met l t => def_met l (subst_trm z u t)
  end with
subst_defs (z: var) (u: var) (ds: defs) : defs :=
  match ds with
  | defs_nil => defs_nil
  | defs_cons rest d => defs_cons (subst_defs z u rest) (subst_def z u d)
  end.
Instance TrmSubstVar : SubstVar trm := subst_trm.
Instance LitSubstVar : SubstVar lit := subst_lit.
Instance DefSubstVar : SubstVar def := subst_def.
Instance DefsSubstVar : SubstVar defs := subst_defs.
Ltac fold_trm_subst :=
  change (subst_trm) with (@subst_var trm TrmSubstVar);
  change (subst_lit) with (@subst_var lit LitSubstVar);
  change (subst_def) with (@subst_var def DefSubstVar);
  change (subst_defs) with (@subst_var defs DefsSubstVar).

Fixpoint seal_rec_trm (k: nat) (u: var) (t: trm): trm :=
  match t with
  | trm_var a      => trm_var (seal_rec k u a)
  | trm_read v m    => trm_read (seal_rec k u v) m
  | trm_apply f a    => trm_apply (seal_rec k u f) (seal_rec k u a)
  | trm_let t1 t2  => trm_let (seal_rec_trm k u t1) (seal_rec_trm (S k) u t2)
  | trm_lit l t => trm_lit (seal_rec_lit k u l) (seal_rec_trm (S k) u t)
  | trm_write v m v2 => trm_write (seal_rec k u v) m (seal_rec k u v2)
  | trm_call v m v2 => trm_call (seal_rec k u v) m (seal_rec k u v2)
  end with
seal_rec_lit (k: nat) (u: var) (l: lit): lit :=
  match l with
  | lit_fun T t => lit_fun (seal_rec k u T) (seal_rec_trm (S k) u t)
  | lit_obj T ds =>
    lit_obj
      (seal_rec (S k) u T)
      (seal_rec_defs (S k) u ds)
      (* avar_n 0 refers to self/this *)
  end with
seal_rec_def (k: nat) (u: var) (d: def): def :=
  match d with
  | def_typ L T => def_typ L (seal_rec (S k) u T)
  | def_fld m v => def_fld m (seal_rec k u v)
  | def_met m e => def_met m (seal_rec_trm (S (S k)) u e)
  end with
seal_rec_defs (k: nat) (u: var) (ds: defs): defs :=
  match ds with
  | defs_nil       => defs_nil
  | defs_cons tl d => defs_cons (seal_rec_defs k u tl) (seal_rec_def k u d)
  end.
Instance TrmSealable : Sealable trm := seal_rec_trm.
Instance LitSealable : Sealable lit := seal_rec_lit.
Instance DefSealable : Sealable def := seal_rec_def.
Instance DefsSealable : Sealable defs := seal_rec_defs.
Ltac fold_trm_seal_rec :=
  change (seal_rec_trm) with (@seal_rec trm TrmSealable);
  change (seal_rec_lit) with (@seal_rec lit LitSealable);
  change (seal_rec_def) with (@seal_rec def DefSealable);
  change (seal_rec_defs) with (@seal_rec defs DefsSealable).

Instance TrmAbstractSyntax :
  AbstractSyntax TrmOpenable TrmFreeVar TrmSubstVar := {}.
Instance LitAbstractSyntax :
  AbstractSyntax LitOpenable LitFreeVar LitSubstVar := {}.
Instance DefAbstractSyntax :
  AbstractSyntax DefOpenable DefFreeVar DefSubstVar := {}.
Instance DefsAbstractSyntax :
  AbstractSyntax DefsOpenable DefsFreeVar DefsSubstVar := {}.
Ltac fold_trm_abs :=
  fold_trm_open_rec; fold_trm_fv; fold_trm_subst.


(* ************************************************************************** *)
(** ** Definitions are labeled *)
Instance DefLabelOf : LabelOf def :=
  fun D =>
      match D with
      | def_typ L _ => label_typ L
      | def_fld m _ => label_fld m
      | def_met m _ => label_met m
      end.


(* ************************************************************************** *)
(** ** Terms Satisfy Syntax Lemmas *)
(** We will prove that the operations on Terms, Defintions and Definition Lists
    are lifted from locally nameless variables. The syntax lemmas will then by
    automatically lifted from the syntax lemmas of locally nameless variables
    through the type class mechanism *)

(** *** Collect Variables in Terms *)
Fixpoint collect_trm {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (t: trm) : C :=
  match t with
  | trm_var a       => (f a)
  | trm_read x m      => (f x)
  | trm_apply f' a      => comb (f f') (f a)
  | trm_let t1 t2    => comb (collect_trm df f comb t1) (collect_trm df f comb t2)
  | trm_lit l t      => comb (collect_lit df f comb l) (collect_trm df f comb t)
  | trm_write x m y    => comb (f x) (f y)
  | trm_call x m y    => comb (f x) (f y)
  end with
collect_lit {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (l: lit) : C :=
  match l with
  | lit_fun T e  => comb (collect df f comb T) (collect_trm df f comb e)
  | lit_obj T ds => comb (collect df f comb T) (collect_defs df f comb ds)
  end with
collect_def {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (d: def) : C :=
  match d with
  | def_typ _ T     => (collect df f comb T)
  | def_fld _ v     => (f v)
  | def_met _ t     => (collect_trm df f comb t)
  end with
collect_defs {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (ds : defs) : C  :=
  match ds with
  | defs_nil         => df
  | defs_cons tl d   => comb (collect_defs df f comb tl) (collect_def df f comb d)
  end.

Instance TrmCollect : Collect AvarAbstractSyntax TrmAbstractSyntax :=
  @collect_trm.

Instance LitCollect : Collect AvarAbstractSyntax LitAbstractSyntax :=
  @collect_lit.

Instance DefCollect : Collect AvarAbstractSyntax DefAbstractSyntax :=
  @collect_def.

Instance DefsCollect : Collect AvarAbstractSyntax DefsAbstractSyntax :=
  @collect_defs.

Ltac fold_trm_collect :=
  change (@collect_trm) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            trm TrmOpenable TrmFreeVar TrmSubstVar TrmAbstractSyntax
            TrmCollect);
  change (@collect_lit) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            lit LitOpenable LitFreeVar LitSubstVar LitAbstractSyntax
            LitCollect);
  change (@collect_def) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            def DefOpenable DefFreeVar DefSubstVar DefAbstractSyntax
            DefCollect);
  change (@collect_defs) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            defs DefsOpenable DefsFreeVar DefsSubstVar DefsAbstractSyntax
            DefsCollect).
Local Ltac sympl_trm_collect :=
  simpl; fold_trm_collect; fold_trm_abs.

(** *** Sanity for Collecting Variables in Terms *)
Lemma collect_compose_trm_def_defs :
  forall {C D} (f : C -> D) g comb1 comb2 i,
    (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
    (forall (t : trm),
        (f ∘ collect i g comb1) t = collect (f i) (f ∘ g) comb2 t) /\
    (forall (l : lit),
        (f ∘ collect i g comb1) l = collect (f i) (f ∘ g) comb2 l) /\
    (forall (d : def),
        (f ∘ collect i g comb1) d = collect (f i) (f ∘ g) comb2 d) /\
    (forall (ds : defs),
        (f ∘ collect i g comb1) ds = collect (f i) (f ∘ g) comb2 ds).
Proof.
  intros; apply trm_mutind; unfold "∘";
    sympl_trm_collect;
      try congruence;
      intros; rewrite ? H,
              ? (fold_compose f (collect (C:=C) _ _ _)),
              ? (collect_compose _ _ _ _ _ H);
      repeat match goal with
             | [ H : _ = _ |- _] => rewrite <- ? H; clear H
             end; auto.
Qed.

Local Hint Resolve collect_reflect : core.

Lemma collect_reflect_trm_def_defs :
  forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall bl,
      (forall (t : trm),
          ((collect (Is_true bl) P or t) <-> (collect bl p orb t) = true)) /\
      (forall (l : lit),
          ((collect (Is_true bl) P or l) <-> (collect bl p orb l) = true)) /\
      (forall (d : def),
          ((collect (Is_true bl) P or d) <-> (collect bl p orb d) = true)) /\
      (forall (ds : defs),
          ((collect (Is_true bl) P or ds) <-> (collect bl p orb ds) = true)).
Proof.
  intros P p H bl.
  apply trm_mutind; sympl_trm_collect;
    auto 3 using Is_true_iff_eq_true; intros;
      repeat apply or_orb_iff_left; try apply collect_reflect; auto.
Qed.

Local Ltac solve_collect_compose :=
  unfold compose; apply fun_ext_dep;
  apply collect_compose_trm_def_defs; auto.

Local Ltac solve_collect_reflect :=
  apply collect_reflect_trm_def_defs; auto.

Local Ltac solve_collect_sanity :=
  hnf; intros; solve [solve_collect_compose || solve_collect_reflect].

Instance TrmCollectCompose : CollectCompose TrmCollect.
Proof. solve_collect_sanity. Qed.

Instance LitCollectCompose : CollectCompose LitCollect.
Proof. solve_collect_sanity. Qed.

Instance DefCollectCompose : CollectCompose DefCollect.
Proof. solve_collect_sanity. Qed.

Instance DefsCollectCompose : CollectCompose DefsCollect.
Proof. solve_collect_sanity. Qed.

Instance TrmCollectReflect : CollectReflect TrmCollect.
Proof. solve_collect_sanity. Qed.

Instance LitCollectReflect : CollectReflect LitCollect.
Proof. solve_collect_sanity. Qed.

Instance DefCollectReflect : CollectReflect DefCollect.
Proof. solve_collect_sanity. Qed.

Instance DefsCollectReflect : CollectReflect DefsCollect.
Proof. solve_collect_sanity. Qed.

(** *** Transforming Variables in Terms *)
Fixpoint transform_trm (f : avar -> nat -> avar) (t : trm) (n : nat) : trm :=
  match t with
  | trm_var a      => trm_var (f a n)
  | trm_read v m    => trm_read (f v n) m
  | trm_apply f' a    => trm_apply (f f' n) (f a n)
  | trm_let t1 t2  => trm_let (transform_trm f t1 n) (transform_trm f t2 (S n))
  | trm_lit l t    => trm_lit (transform_lit f l n) (transform_trm f t (S n))
  | trm_write v m v2 => trm_write (f v n) m (f v2 n)
  | trm_call v m v2 => trm_call (f v n) m (f v2 n)
  end with
transform_lit (f : avar -> nat -> avar) (l : lit) (n : nat) : lit :=
  match l with
  | lit_fun T t => lit_fun (transform f T n) (transform_trm f t (S n))
  | lit_obj T ds =>
    lit_obj
      (transform f T (S n))
      (transform_defs f ds (S n)) (* avar_n (0) refers to self/this *)
  end with
transform_def (f : avar -> nat -> avar) (d : def) (n : nat) : def :=
  match d with
  | def_typ L T => def_typ L (transform f T (S n))
  | def_fld m e => def_fld m (f e n)
  | def_met m e => def_met m (transform_trm f e (S (S n)))
  end with
transform_defs (f : avar -> nat -> avar) (ds : defs) (n : nat) : defs :=
  match ds with
  | defs_nil       => defs_nil
  | defs_cons tl d => defs_cons (transform_defs f tl n) (transform_def f d n)
  end.

Instance TrmTransform : Transform AvarAbstractSyntax TrmAbstractSyntax :=
  @transform_trm.

Instance LitTransform : Transform AvarAbstractSyntax LitAbstractSyntax :=
  @transform_lit.

Instance DefTransform : Transform AvarAbstractSyntax DefAbstractSyntax :=
  @transform_def.

Instance DefsTransform : Transform AvarAbstractSyntax DefsAbstractSyntax :=
  @transform_defs.

Ltac fold_trm_transform :=
  change (@transform_trm) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            trm TrmOpenable TrmFreeVar TrmSubstVar TrmAbstractSyntax
            TrmTransform);
  change (@transform_lit) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            lit LitOpenable LitFreeVar LitSubstVar LitAbstractSyntax
            LitTransform);
  change (@transform_def) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            def DefOpenable DefFreeVar DefSubstVar DefAbstractSyntax
            DefTransform);
  change (@transform_defs) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            defs DefsOpenable DefsFreeVar DefsSubstVar DefsAbstractSyntax
            DefsTransform).
Local Ltac sympl_trm_transform :=
  simpl; fold_trm_transform; fold_trm_collect; fold_trm_abs.

(** *** Sanity for Transforming Variables in Terms *)
Lemma transform_id_trm_def_defs :
  (forall (t : trm) n, transform (ignore_second id) t n = t) /\
  (forall (l : lit) n, transform (ignore_second id) l n = l) /\
  (forall (d : def) n, transform (ignore_second id) d n = d) /\
  (forall (ds : defs) n, transform (ignore_second id) ds n = ds).
Proof.
  apply trm_mutind; sympl_trm_transform; intros; try f_equal;
    try rewrite transform_id; auto.
Qed.

Lemma transform_compose_trm_def_defs :
  (forall (t : trm) n f g,
      ((transform f) ∘∘ (transform g)) t n = transform (f ∘∘ g) t n) /\
  (forall (l : lit) n f g,
      ((transform f) ∘∘ (transform g)) l n = transform (f ∘∘ g) l n) /\
  (forall (d : def) n f g,
      ((transform f) ∘∘ (transform g)) d n = transform (f ∘∘ g) d n) /\
  (forall (ds : defs) n f g,
      ((transform f) ∘∘ (transform g)) ds n = transform (f ∘∘ g) ds n).
Proof.
  unfold "∘∘".
  apply trm_mutind; sympl_trm_transform; intros; try rewrite ? transform_length_s;
    repeat
      match goal with
      | [ H : forall n f g, _ = _ |- _] => rewrite <- H
      end; auto;
      f_equal;
      rewrite ? (fold_compose12 (transform f));
      rewrite ? transform_compose; auto.
Qed.

Local Ltac solve_transform_sanity :=
  hnf; try apply transform_id_trm_def_defs;
  intros; intros_fun_ext; apply transform_compose_trm_def_defs.

Instance TrmTransformId : TransformId TrmTransform.
Proof. solve_transform_sanity. Qed.

Instance LitTransformId : TransformId LitTransform.
Proof. solve_transform_sanity. Qed.

Instance DefTransformId : TransformId DefTransform.
Proof. solve_transform_sanity. Qed.

Instance DefsTransformId : TransformId DefsTransform.
Proof. solve_transform_sanity. Qed.

Instance TrmTransformCompose : TransformCompose TrmTransform.
Proof. solve_transform_sanity. Qed.

Instance LitTransformCompose : TransformCompose LitTransform.
Proof. solve_transform_sanity. Qed.

Instance DefTransformCompose : TransformCompose DefTransform.
Proof. solve_transform_sanity. Qed.

Instance DefsTransformCompose : TransformCompose DefsTransform.
Proof. solve_transform_sanity. Qed.

Lemma transform_shift_trm_defs (f : avar -> nat -> avar) :
  (forall (X : trm) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n) /\
  (forall (X : lit) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n) /\
  (forall (X : def) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n) /\
  (forall (X : defs) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n).
Proof.
  apply trm_mutind; simpl; intros; try f_equal; auto.
  - rewrite <- H0. rewrite Nat.add_succ_l. auto.
  - rewrite <- H0. rewrite Nat.add_succ_l. auto.
  - rewrite <- transform_shift. auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
  - rewrite <- transform_shift. auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
  - rewrite <- transform_shift. auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
Qed.

Instance TrmTransformShift : TransformShift TrmTransform.
Proof.
  intro. eapply transform_shift_trm_defs.
Qed.
Instance LitTransformShift : TransformShift LitTransform.
Proof.
  intro. eapply transform_shift_trm_defs.
Qed.
Instance DefTransformShift : TransformShift DefsTransform.
Proof.
  intro. eapply transform_shift_trm_defs.
Qed.
Instance DefsTransformShift : TransformShift DefsTransform.
Proof.
  intro. eapply transform_shift_trm_defs.
Qed.

Lemma collect_transform_term :
  (forall (t : trm) bl p f g,
      (collect bl p andb t) = true ->
      transform (dec_apply p f g) t = transform f t) /\
  (forall (l : lit) bl p f g,
      (collect bl p andb l) = true ->
      transform (dec_apply p f g) l = transform f l) /\
  (forall (d : def) bl p f g,
      (collect bl p andb d) = true ->
      transform (dec_apply p f g) d = transform f d) /\
  (forall (ds : defs) bl p f g,
      (collect bl p andb ds) = true ->
      transform (dec_apply p f g) ds = transform f ds).
Proof.
  apply trm_mutind; sympl_trm_transform; auto 2; intros;
    repeat (destruct_andbs; destruct_ands);
    repeat erewrite collect_transform by eauto;
    repeat
      match goal with
      | [ H : forall bl p f g,
            _ = true ->
            ?transform (dec_apply p f g) ?t = ?transform f ?t |- _] =>
        erewrite H by eauto
      end;
    repeat match goal with
           | [ H : _ = true |- _ ] => try unfold dec_apply; rewrite H
           end; auto.
Qed.

Instance TrmCollectTransform :
  CollectTransform TrmCollect TrmTransform.
Proof. hnf; intros; eapply collect_transform_term; eauto. Qed.

Instance LitCollectTransform :
  CollectTransform LitCollect LitTransform.
Proof. hnf; intros; eapply collect_transform_term; eauto. Qed.

Instance DefCollectTransform :
  CollectTransform DefCollect DefTransform.
Proof. hnf; intros; eapply collect_transform_term; eauto. Qed.

Instance DefsCollectTransform :
  CollectTransform DefsCollect DefsTransform.
Proof. hnf; intros; eapply collect_transform_term; eauto. Qed.

(** *** Sanity for Collecting and Transforming Variables *)
Lemma fv_collect_trm_def_defs :
  (forall (t : trm), fv t = collect (\{}) fv (@union var) t) /\
  (forall (l : lit), fv l = collect (\{}) fv (@union var) l) /\
  (forall (d : def), fv d = collect (\{}) fv (@union var) d) /\
  (forall (ds : defs), fv ds = collect (\{}) fv (@union var) ds).
Proof.
  apply trm_mutind; sympl_trm_transform; intros;
    repeat match goal with
    | [ H : _ = _ |- _ ] => rewrite <- ? H; clear H
    end; rewrite <- ? fv_collect; auto.
Qed.



Lemma open_transform_trm_def_defs :
  (forall (t : trm) n x, (open_depth_rec n x) t = transform (open_depth_rec n x) t) /\
  (forall (l : lit) n x, (open_depth_rec n x) l = transform (open_depth_rec n x) l) /\
  (forall (d : def) n x, (open_depth_rec n x) d = transform (open_depth_rec n x) d) /\
  (forall (ds : defs) n x, (open_depth_rec n x) ds = transform (open_depth_rec n x) ds).
Proof.
  unfold open_depth_rec.
  apply trm_mutind; sympl_trm_transform; intros; intros_fun_ext;
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    rewrite ? Nat.add_succ_r.
  all:  repeat match goal with
           | [ |- context[open_rec (?n + ?m) ?x ?t] ] =>
             change (open_rec (n + m) x t) with (open_depth_rec n x t m)
           end.
  all:
    rewrite ? open_transform;
    auto.
  all: fequal.
  all: fold (@open_depth_rec avar _ n x).
  all: rewrite <- open_transform.
  all: unfold open_depth_rec; rewrite plus_n_Sm; auto.
Qed.

Lemma seal_transform_trm_def_defs :
  (forall (t : trm) n x, (seal_depth_rec n x) t = transform (seal_depth_rec n x) t) /\
  (forall (l : lit) n x, (seal_depth_rec n x) l = transform (seal_depth_rec n x) l) /\
  (forall (d : def) n x, (seal_depth_rec n x) d = transform (seal_depth_rec n x) d) /\
  (forall (ds : defs) n x, (seal_depth_rec n x) ds = transform (seal_depth_rec n x) ds).
Proof.
  unfold seal_depth_rec.
  apply trm_mutind; sympl_trm_transform; intros; intros_fun_ext;
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    rewrite ? Nat.add_succ_r;
    repeat match goal with
           | [ |- context[seal_rec (?n + ?m) ?x ?t] ] =>
             change (seal_rec (n + m) x t) with (seal_depth_rec n x t m)
           end;
    rewrite ? seal_transform;
    auto.
  all: fequal.
  all: assert (R: (fun (a : avar) (d0 : nat) => seal_rec (n + d0) x a) = seal_depth_rec n x)
      by auto; rewrite R; clear R.
  all: rewrite <- seal_transform.
  all:  unfold seal_depth_rec; rewrite plus_n_Sm; auto.
Qed.

Lemma subst_transform_trm_def_defs :
  (forall (t : trm) x y, (ignore_second (subst_var x y)) t
                = transform (ignore_second (subst_var x y)) t) /\
  (forall (l : lit) x y, (ignore_second (subst_var x y)) l
                = transform (ignore_second (subst_var x y)) l) /\
  (forall (d : def) x y, (ignore_second (subst_var x y)) d
                = transform (ignore_second (subst_var x y)) d) /\
  (forall (ds : defs) x y, (ignore_second (subst_var x y)) ds
                = transform (ignore_second (subst_var x y)) ds).
Proof.
  apply trm_mutind; sympl_trm_transform; intros; intros_fun_ext;
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    rewrite <- ? subst_transform;
    unfold ignore_second; simpl; auto.
Qed.

Local Ltac solve_transform_syntax :=
  hnf; intros; (apply fun_ext_dep; intro);
    try apply fv_collect_trm_def_defs;
    try apply open_transform_trm_def_defs;
    try apply subst_transform_trm_def_defs.

Instance TrmFvCollect : FvCollect TrmCollect.
Proof. solve_transform_syntax. Qed.

Instance LitFvCollect : FvCollect LitCollect.
Proof. solve_transform_syntax. Qed.

Instance DefFvCollect : FvCollect DefCollect.
Proof. solve_transform_syntax. Qed.

Instance DefsFvCollect : FvCollect DefsCollect.
Proof. solve_transform_syntax. Qed.

Instance TrmOpenTransform : OpenTransform TrmTransform.
Proof. solve_transform_syntax. Qed.

Instance DefOpenTransform : OpenTransform DefTransform.
Proof. solve_transform_syntax. Qed.

Instance DefsOpenTransform : OpenTransform DefsTransform.
Proof. solve_transform_syntax. Qed.

Instance TrmSealTransform : SealTransform TrmTransform AvarSealable TrmSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro);
  apply seal_transform_trm_def_defs.
Qed.
Instance LitSealTransform : SealTransform LitTransform AvarSealable LitSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro).
  apply seal_transform_trm_def_defs.
Qed.
Instance DefSealTransform : SealTransform DefTransform AvarSealable DefSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro).
  apply seal_transform_trm_def_defs.
Qed.
Instance DefsSealTransform : SealTransform DefsTransform AvarSealable DefsSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro).
  apply seal_transform_trm_def_defs.
Qed.

Instance TrmSubstTransform : SubstTransform TrmTransform.
Proof. solve_transform_syntax. Qed.

Instance LitSubstTransform : SubstTransform LitTransform.
Proof. solve_transform_syntax. Qed.

Instance DefSubstTransform : SubstTransform DefTransform.
Proof. solve_transform_syntax. Qed.

Instance DefsSubstTransform : SubstTransform DefsTransform.
Proof. solve_transform_syntax. Qed.

(** Additional Lemmas that are not lifted automatically *)
Lemma fv_open_term_cases : forall x,
    (forall (t : trm) n,
        fv (open_rec n x t) = fv t \/
        fv (open_rec n x t) = \{ x} \u fv t) /\
    (forall (l : lit) n,
        fv (open_rec n x l) = fv l \/
        fv (open_rec n x l) = \{ x} \u fv l) /\
    (forall (d : def) n,
        fv (open_rec n x d) = fv d \/
        fv (open_rec n x d) = \{ x} \u fv d) /\
    (forall (ds : defs) n,
        fv (open_rec n x ds) = fv ds
        \/ fv (open_rec n x ds) = \{ x} \u fv ds).
Proof.
  intros x.
  apply trm_mutind; sympl_trm_transform; auto; intros;
    repeat apply union_eq_cases; auto; apply fv_open_cases.
Qed.

Instance TrmFvOpenCases : FvOpenCases TrmAbstractSyntax.
Proof.
  hnf; apply fv_open_term_cases.
Qed.

Instance LitFvOpenCases : FvOpenCases LitAbstractSyntax.
Proof.
  hnf; apply fv_open_term_cases.
Qed.

Instance DefFvOpenCases : FvOpenCases DefAbstractSyntax.
Proof.
  hnf; apply fv_open_term_cases.
Qed.

Instance DefsFvOpenCases : FvOpenCases DefsAbstractSyntax.
Proof.
  hnf; apply fv_open_term_cases.
Qed.

Lemma open_seal_inv_trm_defs:
  (forall (t: trm) n y m x, OpenSealInvOf n y m x t) /\
  (forall (l: lit) n y m x, OpenSealInvOf n y m x l) /\
  (forall (d: def) n y m x, OpenSealInvOf n y m x d) /\
  (forall (ds: defs) n y m x, OpenSealInvOf n y m x ds).
Proof.
  apply trm_mutind; intros; simpls; eauto.
  - inversions H0. erewrite open_seal_inv; eauto.
  - inversions H0. erewrite open_seal_inv; eauto.
  - inversions H0. erewrite open_seal_inv; eauto. erewrite open_seal_inv; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H0. erewrite open_seal_inv; eauto. erewrite open_seal_inv; eauto.
  - inversions H0. erewrite open_seal_inv; eauto. erewrite open_seal_inv; eauto.
  - inversions H1. erewrite H. 2: eapply Lt.lt_n_S; eauto. 2: eauto. erewrite open_seal_inv; eauto.
  - inversions H1. erewrite H. 2: eapply Lt.lt_n_S; eauto. 2: eauto. erewrite open_seal_inv. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H0. erewrite open_seal_inv. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H0. erewrite open_seal_inv; eauto.
  - inversions H1. erewrite H. 2: eapply Lt.lt_n_S; eapply Lt.lt_n_S; eauto. 2: eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
Qed.

Instance TrmOpenSealInv: OpenSealInv TrmAbstractSyntax TrmSealable.
Proof.
  hnf; intros.
  eapply open_seal_inv_trm_defs; eauto.
Qed.

Lemma open_twice_inv_trm_defs:
  (forall (t: trm) m n y1 y2, OpenTwiceInvOf m n y1 y2 t) /\
  (forall (t: lit) m n y1 y2, OpenTwiceInvOf m n y1 y2 t) /\
  (forall (t: def) m n y1 y2, OpenTwiceInvOf m n y1 y2 t) /\
  (forall (ds: defs) m n y1 y2, OpenTwiceInvOf m n y1 y2 ds).
Proof.
  apply trm_mutind; intros; simpls; eauto.
  - inversions H0. erewrite open_twice_inv; eauto.
  - inversions H0. erewrite open_twice_inv; eauto.
  - inversions H0. erewrite open_twice_inv; eauto. erewrite open_twice_inv; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H0. erewrite open_twice_inv; eauto. erewrite open_twice_inv; eauto.
  - inversions H0. erewrite open_twice_inv; eauto. erewrite open_twice_inv; eauto.
  - inversions H1. erewrite H. 2: eapply ne_n_S; eauto. 2: eauto. erewrite open_twice_inv; eauto.
  - inversions H1. erewrite H. 2: eapply ne_n_S; eauto. 2: eauto. erewrite open_twice_inv. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H0. erewrite open_twice_inv. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H0. erewrite open_twice_inv; eauto.
  - inversions H1. erewrite H. 2: eapply ne_n_S; eapply ne_n_S; eauto. 2: eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
Qed.

Instance TrmOpenTwiceInv : OpenTwiceInv TrmAbstractSyntax.
Proof.
  hnf; intros.
  eapply open_twice_inv_trm_defs; eauto.
Qed.



Lemma subst_id_trm (v1: var):
  (forall (t: trm),
  subst_var v1 v1 t = t) /\
  (forall (d: lit),
  subst_var v1 v1 d = d) /\
  (forall (d: def),
  subst_var v1 v1 d = d) /\
  (forall (d: defs),
  subst_var v1 v1 d = d).
Proof.
  apply trm_mutind; intros; auto.
  - rewrite <- (subst_id v1 a) at 2; auto.
  - rewrite <- (subst_id v1 a) at 2; auto.
  - rewrite <- (subst_id v1 a) at 2;rewrite <- (subst_id v1 a0) at 2; auto.
  - rewrite <- H, <- H0 at 2; auto.
  - rewrite <- H, <- H0 at 2; auto.
  - rewrite <- (subst_id v1 a) at 2;rewrite <- (subst_id v1 a0) at 2; auto.
  - rewrite <- (subst_id v1 a) at 2;rewrite <- (subst_id v1 a0) at 2; auto.
  - rewrite <- H, <- (subst_id v1 t) at 2; auto.
  - rewrite <- H, <- (subst_id v1 t) at 2; auto.
  - rewrite <- (subst_id v1 t0) at 2; auto.
  - rewrite <- (subst_id v1 a) at 2; auto.
  - rewrite <- H at 2; auto.
  - rewrite <- H, <- H0 at 2; auto.
Qed.

Instance TrmSubstId: SubstId TrmAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite (proj41 (subst_id_trm x)). compute. auto.
Qed.
Instance LitSubstId: SubstId LitAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite (proj42 (subst_id_trm x)). compute. auto.
Qed.
Instance DefSubstId: SubstId DefAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite (proj43 (subst_id_trm x)). compute. auto.
Qed.
Instance DefsSubstId: SubstId DefsAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite (proj44 (subst_id_trm x)). compute. auto.
Qed.

Local Lemma subst_trm_id y t0: (subst_trm y y t0 = t0). Proof. pose proof (subst_id y t0). eauto. Qed.
Local Lemma subst_lit_id y t0: (subst_lit y y t0 = t0). Proof. pose proof (subst_id y t0). eauto. Qed.

Lemma fv_subst_cases_trm_defs : forall x y,
    (forall (T : trm), FvSubstCasesOf x y T) /\
    (forall (T : lit), FvSubstCasesOf x y T) /\
    (forall (T : def), FvSubstCasesOf x y T) /\
    (forall (D : defs), FvSubstCasesOf x y D).
Proof.
  intros x y.
  apply trm_mutind; simpl in *; intros; fold_trm_subst; fold_trm_fv.
  all: autorewrite with sets in *; destruct_all; subst; rewrite_sets;autorewrite with sets in *; auto 7 using union_remove_swap2, union_remove_swap2r; try easy.
  all: auto 11 using fv_subst_cases with sets.
  all:
  try destruct (fv_subst_cases x y a) as [[Hfva [Hx | Hna]] | [Hnxy [Hxa Hfva]]];
  try destruct (fv_subst_cases x y a0) as [[Hfva0 [Hx0 | Hna0]] | [Hnxy0 [Hxa0 Hfva0]]];
  try destruct (fv_subst_cases x y t) as [[Hfvt [Hxt | Hnt]] | [Hnxyt [Hxt Hfvt]]];
  try destruct (fv_subst_cases x y t0) as [[Hfvt0 [Hxt0 | Hnt0]] | [Hnxyt0 [Hxt0 Hfvt0]]];
  subst;
  try rewrite Hfva; 
  try rewrite Hfva0;
  try rewrite Hfvt;
  try rewrite Hfvt0; repeat rewrite subst_id; eauto; try contradiction;
  rewrite_sets;autorewrite with sets in *; eauto 7 using union_remove_swap2, union_remove_swap2r; try easy.

  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto. 

  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto.

  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto. 
  


  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto. 


  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto.
Qed.

Instance TrmFvSubstCases: FvSubstCases TrmAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_trm_defs.
Qed.
Instance LitFvSubstCases: FvSubstCases LitAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_trm_defs.
Qed.
Instance DefFvSubstCases: FvSubstCases DefAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_trm_defs.
Qed.
Instance DefsFvSubstCases: FvSubstCases DefsAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_trm_defs.
Qed.

Lemma fv_close_trm:
    (forall (T : trm) n x, FvCloseOf n x T) /\
    (forall (D : lit) n x, FvCloseOf n x D) /\
    (forall (T : def) n x, FvCloseOf n x T) /\
    (forall (D : defs) n x, FvCloseOf n x D).
Proof.
  pose proof (fv_close (C:= TransformClosing (A:= typ))) as Htyp.
  pose proof (fv_close (C:= TransformClosing (A:= dec))) as Hdec.

  apply trm_mutind.
  all: sympl_trm_transform.
  all: fold_typ_transform; intros.
  all: rewrite ? transform_close_depth_1, ? transform_close_depth_2.
  all: unfold close_rec, TransformClosing in * |- .
  all: rewrite ? Htyp, ? Hdec.
  all: rewrite ? H, ? H0.
  all: unfold close_depth_rec in *.
  all: rewrite ? fv_close.
  all: autorewrite with sets in *; destruct_all; rewrite_sets;autorewrite with sets in *; auto 7 using union_remove_swap2, union_remove_swap2r; try easy.
  all: rewrite ? remove_empty_r; eauto.
  all: rewrite union_remove_dist; auto.
Qed.

Instance TrmFvClose: FvClose (TransformClosing (A:= trm)).
Proof.
  hnf. intros. apply fv_close_trm.
Qed.
Instance LitFvClose: FvClose (TransformClosing (A:= lit)).
Proof.
  hnf. intros. apply fv_close_trm.
Qed.
Instance DefFvClose: FvClose (TransformClosing (A:= def)).
Proof.
  hnf. intros. apply fv_close_trm.
Qed.
Instance DefsFvClose: FvClose (TransformClosing (A:= defs)).
Proof.
  hnf. intros. apply fv_close_trm.
Qed.

Lemma rewrite_transform_trm : transform_trm = transform.
Proof. trivial. Qed.
Lemma rewrite_transform_defs : transform_defs = transform.
Proof. trivial. Qed.
Lemma rewrite_transform_lit : transform_lit = transform.
Proof. trivial. Qed.
Lemma rewrite_transform_def : transform_def = transform.
Proof. trivial. Qed.


(* Lemmata for working with free variables in terms and definitions *)

Lemma notin_fv_defs y ds d:
  y \notin fv (defs_cons ds d) ->
  y \notin fv ds /\
  y \notin fv d.
Proof.
  intros.
  compute in H.
  split.
  apply notin_union_r1 in H.
  compute. auto.
  apply notin_union_r2 in H.
  compute. auto.
Qed.

Lemma fv_defs_eq ds d:
  fv (defs_cons ds d) = fv ds \u fv d.
Proof.
  intros.
  compute in *. auto.
Qed.

(* Rewriting *)

Ltac close_trm_solve := (
  unfold close_rec, TransformClosing, TrmTransform, DefsTransform;
  simpl;
  rewrite ? rewrite_transform_trm, ? rewrite_transform_defs, <- ? close_eq_close_depth, ? transform_close_depth_1, ? transform_close_depth_2;
  eauto using f_equal, f_equal2, f_equal3 
).

Lemma trm_close_var n x y:
  (trm_var y) /[#n <- x] = trm_var y /[#n <- x].
Proof.
  close_trm_solve.
Qed.
Lemma trm_close_read n x y a:
  (trm_read y a) /[#n <- x] = trm_read y /[#n <- x] a.
Proof.
  close_trm_solve.
Qed.
Lemma trm_close_write n x y a z:
  (trm_write y a z) /[#n <- x] = trm_write y /[#n <- x] a z /[#n <- x].
Proof.
  close_trm_solve.
Qed.
Lemma trm_close_call n x y a z:
  (trm_call y a z) /[#n <- x] = trm_call y /[#n <- x] a z /[#n <- x].
Proof.
  close_trm_solve.
Qed.
Lemma trm_close_let n x t1 t2:
  (trm_let t1 t2) /[#n <- x] = trm_let t1 /[#n <- x] t2 /[#S n <- x].
Proof.
  close_trm_solve.
Qed.
Lemma trm_close_lit n x t l:
  (trm_lit t l) /[#n <- x] = trm_lit t /[#n <- x] l /[#S n <- x].
Proof.
  close_trm_solve.
Qed.

Lemma lit_close_obj n x R ds:
  (lit_obj R ds) /[#n <- x] = lit_obj R /[#S n <- x] ds /[#S n <- x].
Proof.
  close_trm_solve.
Qed.
Lemma lit_close_fun n x T t:
  (lit_fun T t) /[#n <- x] = lit_fun T /[#n <- x] t /[#S n <- x].
Proof.
  close_trm_solve.
Qed.

Lemma defs_close_nil n x:
  (defs_nil) /[#n <- x] = defs_nil.
Proof.
  close_trm_solve.
Qed.

Lemma defs_close_cons n x d ds:
  (defs_cons d ds) /[#n <- x] = defs_cons d /[#n <- x] ds /[#n <- x].
Proof.
  close_trm_solve.
Qed.

Lemma def_close_fld n x A y:
  (def_fld A y) /[#n <- x] = (def_fld A y /[#n <- x]).
Proof.
  close_trm_solve.
Qed.
Lemma def_close_met n x t A:
  (def_met A t) /[#n <- x] = (def_met A t /[#S (S n) <- x]).
Proof.
  close_trm_solve.
Qed.
Lemma def_close_typ n x T A:
  (def_typ A T) /[#n <- x] = def_typ A T /[#S n <- x].
Proof.
  close_trm_solve.
Qed.

Hint Rewrite
  trm_close_var
  trm_close_read
  trm_close_write
  trm_close_call
  trm_close_let
  trm_close_lit
  lit_close_obj
  lit_close_fun
  defs_close_nil
  defs_close_cons
  def_close_fld
  def_close_met
  def_close_typ
  : simpl_close.

Lemma trm_subst_vars_var T1  xs ys:
  (trm_var T1 ) /[xs --> ys] = (trm_var (T1 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_var T1  xs ys:
  (trm_var T1 ) //[xs --> ys] = (trm_var (T1 //[xs --> ys])  ).
Proof.
  edestruct (subst_pvars_unfold (trm_var T1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (trm_subst_vars_var T1 xs zs); auto.
  rewrite trm_subst_vars_var; auto.
  simpls.

  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
Qed.


Lemma trm_subst_vars_read a T1  xs ys:
  (trm_read T1 a ) /[xs --> ys] = (trm_read (T1 /[xs --> ys]) a).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_read a T1  xs ys:
  (trm_read T1 a ) //[xs --> ys] = (trm_read (T1 //[xs --> ys]) a ).
Proof.
  edestruct (subst_pvars_unfold (trm_read T1 a)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (trm_subst_vars_read a T1 xs zs); auto.
  rewrite trm_subst_vars_read; auto.
  simpls.

  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
Qed.



Lemma trm_subst_vars_write a T1 T2 xs ys:
  (trm_write T1 a T2) /[xs --> ys] = (trm_write (T1 /[xs --> ys]) a (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_write a T1 T2 xs ys:
  (trm_write T1 a T2) //[xs --> ys] = (trm_write (T1 //[xs --> ys]) a (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (trm_write T1 a T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (trm_subst_vars_write a T1 T2 xs zs); auto.
  rewrite trm_subst_vars_write; auto.
  simpls.

  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.


Lemma trm_subst_vars_apply T1 T2 xs ys:
  (trm_apply T1 T2) /[xs --> ys] = (trm_apply (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_apply T1 T2 xs ys:
  (trm_apply T1 T2) //[xs --> ys] = (trm_apply (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (trm_apply T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (trm_subst_vars_apply T1 T2 xs zs); auto.
  rewrite trm_subst_vars_apply; auto.
  simpls.

  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.

Lemma trm_subst_vars_call a T1 T2 xs ys:
  (trm_call T1 a T2) /[xs --> ys] = (trm_call (T1 /[xs --> ys]) a (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_call a T1 T2 xs ys:
  (trm_call T1 a T2) //[xs --> ys] = (trm_call (T1 //[xs --> ys]) a (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (trm_call T1 a T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (trm_subst_vars_call a T1 T2 xs zs); auto.
  rewrite trm_subst_vars_call; auto.
  simpls.

  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.


Lemma trm_subst_vars_let T1 T2 xs ys:
  (trm_let T1 T2) /[xs --> ys] = (trm_let (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_let T1 T2 xs ys:
  (trm_let T1 T2) //[xs --> ys] = (trm_let (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (trm_let T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (trm_subst_vars_let T1 T2 xs zs); auto.
  rewrite trm_subst_vars_let; auto.
  simpls.
  change (fv_trm) with (@fv trm TrmFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.

Lemma trm_subst_vars_lit T1 T2 xs ys:
  (trm_lit T1 T2) /[xs --> ys] = (trm_lit (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma trm_subst_pvars_lit T1 T2 xs ys:
  (trm_lit T1 T2) //[xs --> ys] = (trm_lit (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (trm_lit T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (trm_subst_vars_lit T1 T2 xs zs); auto.
  rewrite trm_subst_vars_lit; auto.
  simpls.
  change (fv_trm) with (@fv trm TrmFreeVar) in Hfr.
  change (fv_lit) with (@fv lit LitFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.

Lemma lit_subst_vars_fun T1 T2 xs ys:
  (lit_fun T1 T2) /[xs --> ys] = (lit_fun (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma lit_subst_pvars_fun T1 T2 xs ys:
  (lit_fun T1 T2) //[xs --> ys] = (lit_fun (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (lit_fun T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (lit_subst_vars_fun T1 T2 xs zs); auto.
  rewrite lit_subst_vars_fun; auto.
  simpls.
  change (fv_trm) with (@fv trm TrmFreeVar) in Hfr.
  change (fv_lit) with (@fv lit LitFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.

Lemma lit_subst_vars_obj T1 T2 xs ys:
  (lit_obj T1 T2) /[xs --> ys] = (lit_obj (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma lit_subst_pvars_obj T1 T2 xs ys:
  (lit_obj T1 T2) //[xs --> ys] = (lit_obj (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (lit_obj T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (lit_subst_vars_obj T1 T2 xs zs); auto.
  rewrite lit_subst_vars_obj; auto.
  simpls.
  change (fv_trm) with (@fv trm TrmFreeVar) in Hfr.
  change (fv_defs) with (@fv defs DefsFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.

Lemma defs_subst_vars_cons T1 T2 xs ys:
  (defs_cons T1 T2) /[xs --> ys] = (defs_cons (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma defs_subst_pvars_cons T1 T2 xs ys:
  (defs_cons T1 T2) //[xs --> ys] = (defs_cons (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (defs_cons T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.
  rewrite (defs_subst_vars_cons T1 T2 xs zs); auto.
  rewrite defs_subst_vars_cons; auto.
  simpls.
  change (fv_def) with (@fv def DefFreeVar) in Hfr.
  change (fv_defs) with (@fv defs DefsFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
Qed.


Lemma defs_subst_vars_nil xs ys:
  (defs_nil) /[xs --> ys] = (defs_nil).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma defs_subst_pvars_nil xs ys:
  (defs_nil) //[xs --> ys] = (defs_nil).
Proof.
  edestruct (subst_pvars_unfold (defs_nil)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.

  rewrite (defs_subst_vars_nil xs zs); auto.
  rewrite (defs_subst_vars_nil zs ys); auto.
Qed.

Lemma def_subst_vars_fld a T1 xs ys:
  (def_fld a T1) /[xs --> ys] = (def_fld a  (T1 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma def_subst_pvars_fld a T1  xs ys:
  (def_fld a  T1 ) //[xs --> ys] = (def_fld a  (T1 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (def_fld a   T1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (def_subst_vars_fld a T1 xs zs); auto.
  rewrite def_subst_vars_fld; auto.
  simpls.
  change (fv_def) with (@fv def DefFreeVar) in Hfr.
  change (fv_defs) with (@fv defs DefsFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
Qed.

Lemma def_subst_vars_typ a T1 xs ys:
  (def_typ a T1) /[xs --> ys] = (def_typ a  (T1 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma def_subst_pvars_typ a T1  xs ys:
  (def_typ a  T1 ) //[xs --> ys] = (def_typ a  (T1 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (def_typ a   T1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (def_subst_vars_typ a T1 xs zs); auto.
  rewrite def_subst_vars_typ; auto.
  simpls.
  change (fv_def) with (@fv def DefFreeVar) in Hfr.
  change (fv_defs) with (@fv defs DefsFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
Qed.

Lemma def_subst_vars_met a T1 xs ys:
  (def_met a T1) /[xs --> ys] = (def_met a  (T1 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma def_subst_pvars_met a T1  xs ys:
  (def_met a  T1 ) //[xs --> ys] = (def_met a  (T1 //[xs --> ys])).
Proof.
  edestruct (subst_pvars_unfold (def_met a   T1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (def_subst_vars_met a T1 xs zs); auto.
  rewrite def_subst_vars_met; auto.
  simpls.
  change (fv_def) with (@fv def DefFreeVar) in Hfr.
  change (fv_defs) with (@fv defs DefsFreeVar) in Hfr.
  rewrite subst_vars_fresh_middle_change with  (ws:=zs0); eauto.
Qed.
Hint Rewrite
  trm_subst_vars_var
  trm_subst_vars_write
  trm_subst_vars_read
  trm_subst_vars_call
  trm_subst_vars_apply
  trm_subst_vars_let
  trm_subst_vars_lit
  lit_subst_vars_fun
  lit_subst_vars_obj
  def_subst_vars_fld
  def_subst_vars_typ
  def_subst_vars_met
  defs_subst_vars_nil
  defs_subst_vars_cons
  : simpl_subst_vars.

Lemma subst_vars_eq_inv_trmdefs:
  (forall (t: trm) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 t) /\
  (forall (l: lit) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 l) /\
  (forall (d: def) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 d) /\
  (forall (ds: defs) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 ds).
Proof.
  apply trm_mutind; intros; simpls; autorewrite with simpl_subst_vars in *.
  (* Solve simple cases where there are no free variables *)
  all: try match goal with
  | H: from_list ?xs \c \{} |- _ =>
    eapply subset_empty_inv in H;
    eapply from_list_empty_inv in H;
    subst;
    destruct ys2; inversions H1;
    destruct ys1; inversions H0;
    auto
  end.
  Ltac invert_eq T := match goal with H: @eq T _ _ |- _ => inversions H end.
  all: first [invert_eq trm| invert_eq lit| invert_eq def| invert_eq defs].
  all: try solve [eapply H; eauto | eapply (subst_vars_eq_inv xs ys1 ys2 a); eauto].
  - 
    forwards Hsrt: (subst_vars_eq_inv_step a) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_step a0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H0; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H0, H1. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step l) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_step a) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_step a0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H0; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H0, H1. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_step a) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_step a0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H0; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H0, H1. auto.
      
  - 
    forwards Hsrt: (subst_vars_eq_inv_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H1; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H1, H2. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step d)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H1; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H1, H2. auto.
  - eapply (subst_vars_eq_inv xs ys1 ys2 t0); eauto.
  -
    forwards Hsrt: (subst_vars_eq_inv_ind_step d) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step d0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
Qed.

Instance TrmSubstVarsEqInv: SubstVarsEqInv TrmAbstractSyntax.
Proof.
  hnf. intros. eapply (proj41 subst_vars_eq_inv_trmdefs); eauto.
Qed.
Instance DefsSubstVarsEqInv: SubstVarsEqInv DefsAbstractSyntax.
Proof.
  hnf. intros. eapply (proj44 subst_vars_eq_inv_trmdefs); eauto.
Qed.

Instance LitSubstVarsEqInv: SubstVarsEqInv LitAbstractSyntax.
Proof.
  hnf. intros. eapply (proj42 subst_vars_eq_inv_trmdefs); eauto.
Qed.
Instance DefSubstVarsEqInv: SubstVarsEqInv DefAbstractSyntax.
Proof.
  hnf. intros. eapply (proj43 subst_vars_eq_inv_trmdefs); eauto.
Qed.

Lemma trm_close_let_close_fresh_vars t1 t2 y1 y2:
  y1 \notin fv t1 ->
  y1 <> y2 ->
  ((trm_let t1 t2 /[<- y2]) /[<- y1]) = (trm_let t1 t2 /[<-- y2 :: y1 :: nil]).
Proof.
  introv Hfv Hny.
  autorewrite with simpl_close.
  rewrite close_unbound by auto.
  rewrite close_commut by auto.
  erewrite <- close_vars_nil with (t:=t2) (n:=2) at 1.
  rewrite <- close_vars_unfold_l. 2: rewrite from_list_nil. 2: auto.
  rewrite <- close_vars_unfold_l. 2: rewrite from_list_cons, from_list_nil. 2: auto.
  auto.
Qed.

Lemma trm_open_open_vars (t: trm) y1 y2:
  t /[#1 -> y1] /[-> y2] = t /[--> y2 :: y1 :: nil].
Proof.
  rewrite open_commut by auto.
  rewrite <- open_vars_single.
  rewrite open_vars_push_r.
  auto.
Qed.

