Set Implicit Arguments.
Require Import LibExtra DotTactics.
Require Import Types Orctx Logctx OrctxLogctx Logpath LogpathOrder.
Require Import Coq.Program.Equality.

(** * Additional induction schemes for types involving [/\] and [\/] *)

Lemma typ_logic_ind (P: typ->Prop):
  (forall T, typ_notlog T -> P T) ->
  (forall T U, P T -> P U -> P (typ_or T U)) ->
  (forall T U, P T -> P U -> P (typ_and T U)) ->
  (forall T, P T).
Proof.
  introv Hnl Ho Ha.
  intros.
  induction T; eauto; eapply Hnl; hnf; split; intros; discriminate.
Qed.

Inductive typ_notorand: typ-> Prop :=
| typ_notorand_not T:
  typ_notor T ->
  typ_notand T ->
  typ_notorand T
| typ_notorand_or T U:
  typ_notorand T ->
  typ_notorand U ->
  typ_notorand (typ_or T U)
.

Lemma typ_notorand_notand U:
  typ_notorand U ->
  typ_notand U.
Proof.
  introv Hn.
  dependent induction Hn; eauto.
  intros; discriminate.
Qed.

Lemma typ_notorand_oc_notand oc U:
  typ_notorand (orctx_typ oc U) ->
  typ_notand U.
Proof.
  introv Hn.
  reverts U.
  induction oc; intros; simpls.
  - eapply typ_notorand_notand; eauto.
  - inversions Hn.
    + false. eapply H. eauto.
    + eauto.
  - inversions Hn.
    + false. eapply H. eauto.
    + eauto.
Qed.

Lemma typ_notlog_notorand T:
  typ_notlog T ->
  typ_notorand T.
Proof.
  introv Hn.
  hnf in Hn.
  destruct Hn.
  eapply typ_notorand_not; eauto.
Qed.


(* PLAN:
induction that constructs and types from right to left 
*)

(* Types, which do not contain intersections left of a path of unions *)
Inductive typ_notorand_le: logpath -> typ -> Prop :=
| typ_notorand_le_nil T:
  typ_notorand_le logpath_nil T
| typ_notorand_orl op T U:
  typ_notorand_le op T ->
  typ_notorand_le (logpath_orl op) (typ_or T U)
| typ_notorand_orr op T U:
  typ_notorand T ->
  typ_notorand_le op U ->
  typ_notorand_le (logpath_orr op) (typ_or T U)
.
Hint Constructors typ_notorand_le: core.

Lemma typ_notorand_logctx_path op T:
  typ_notorand (logctx_typ op T) ->
  typ_notorand_le (logctx_path op) (logctx_typ op T).
Proof.
  introv Hn.
  reverts T; induction op; intros; simpls.
  - eapply typ_notorand_le_nil.
  - constructor.
    eapply IHop. inversions Hn; eauto.
    false. eapply H; eauto.
  - inversions Hn.
    false. eapply H; eauto.
    constructor. eauto. eauto.
  - false. eapply typ_notorand_notand; eauto.
  - false. eapply typ_notorand_notand; eauto.
Qed.

Inductive orctx_notorand_le: orctx->Prop :=
| orctx_notorand_le_nil:
  orctx_notorand_le orctx_nil
| orctx_notorand_le_l oc T:
  orctx_notorand_le oc ->
  orctx_notorand_le (orctx_l oc T)
| orctx_notorand_le_r oc T:
  typ_notorand T ->
  orctx_notorand_le oc ->
  orctx_notorand_le (orctx_r T oc).
Hint Constructors orctx_notorand_le: core.


(* Count of top-level logic constructors *)
Fixpoint typ_logic_count (T: typ): nat :=
  match T with
  | typ_and T1 T2 => 1 + (typ_logic_count T2) + (typ_logic_count T1) 
  | typ_or T1 T2 => 1 + (typ_logic_count T2) + (typ_logic_count T1)
  | _ => 0
  end.

Lemma typ_logic_count_notlog T:
  typ_notlog T <->
  typ_logic_count T = 0.
Proof.
  split.
  - introv Hn.
    hnf in Hn; destruct Hn.
    destruct T; simpls; eauto.
    false. eapply H0; eauto.
    false. eapply H; eauto.
  - introv He.
    destruct T; simpls; split; intros; discriminate.
Qed.


(* Count of and-under-or constructors *)
Fixpoint typ_orand_count (T: typ): nat :=
  match T with
  | typ_and T1 T2 => 1
  | typ_or T1 T2 => (typ_orand_count T2) + (typ_orand_count T1)
  | _ => 0
  end.

Lemma typ_orand_count_notlog T:
  typ_notlog T ->
  typ_orand_count T = 0.
Proof.
    introv Hn.
    hnf in Hn; destruct Hn.
    destruct T; simpls; eauto.
    false. eapply H0; eauto.
    false. eapply H; eauto.
Qed.

(* Analog of typ_logic_bndsel_count_notbndsel *) 
Lemma typ_orand_count_notorand T:
  typ_notorand T <->
  typ_orand_count T = 0.
Proof.
  split.
  - introv Hn.
    induction T using typ_logic_ind; simpls; eauto.
    + eapply typ_orand_count_notlog; eauto.
    + inversions Hn. false H; eauto.
      rewrite IHT1, IHT2; eauto.
    + false. eapply typ_notorand_notand; eauto.
  - introv He.
    induction T using typ_logic_ind; simpls; eauto.
    + eapply typ_notlog_notorand; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      eapply typ_notorand_or; eauto.
    + discriminate.
Qed.

(* Analog of typ_bndsel_count_sel, typ_bndsel_count_bnd *)
Lemma typ_orand_count_orand oc T1 T2:
  typ_orand_count (orctx_typ oc (typ_and T1 T2)) <> 0.
Proof.
  intro.
  apply  <- typ_orand_count_notorand in H.
  eapply typ_notorand_oc_notand in H.
  eapply H.
  fequal.
Qed.

(* Count of and constructors *)
Fixpoint typ_logic_and_count (T: typ): nat :=
  match T with
  | typ_and T1 T2 => 1 + (typ_logic_and_count T2) + (typ_logic_and_count T1)
  | typ_or T1 T2 => (typ_logic_and_count T2) + (typ_logic_and_count T1)
  | _ => 0
  end.

Lemma typ_logic_and_count_notlog T:
  typ_notlog T ->
  typ_logic_and_count T = 0.
Proof.
    introv Hn.
    hnf in Hn; destruct Hn.
    destruct T; simpls; eauto.
    false. eapply H0; eauto.
    false. eapply H; eauto.
Qed.

Lemma typ_logic_and_count_notorand T:
  typ_notorand T <->
  typ_logic_and_count T = 0.
Proof.
  split.
  - introv Hn.
    induction T using typ_logic_ind; simpls; eauto.
    + eapply typ_logic_and_count_notlog; eauto.
    + inversions Hn. false H; eauto.
      rewrite IHT1, IHT2; eauto.
    + false. eapply typ_notorand_notand; eauto.
  - introv He.
    induction T using typ_logic_ind; simpls; eauto.
    + eapply typ_notlog_notorand; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      eapply typ_notorand_or; eauto.
    + discriminate.
Qed.

(* Count of logic constructors right of the leftmost and-under-or (including then and and the ancestors) *)
Fixpoint typ_orand_index T: nat :=
  match T with
  | typ_and T1 T2 => 1
  | typ_or T1 T2 =>
      let nL := (typ_orand_index T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => 1 + (typ_orand_index T2) 
      (* In left -> go to left and add right *)
      | _ => 1 + nL + (typ_logic_count T2)
      end
  | _ => 0
  end.

Fixpoint typ_orand_leftmost_orctx T: orctx :=
  match T with
  | typ_or T1 T2 =>
      let nL := (typ_orand_count T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => orctx_r T1 (typ_orand_leftmost_orctx T2)
      (* In left -> go to left *)
      | _ => orctx_l (typ_orand_leftmost_orctx T1) T2
      end
  | _ => orctx_nil
  end.

Lemma typ_orand_leftmost_orctx_notlog T:
  typ_notlog T ->
  typ_orand_leftmost_orctx T = orctx_nil.
Proof.
  introv Hn.
  hnf in Hn. destruct Hn.
  destruct T; simpls; eauto.
  false H; eauto.
Qed.

Lemma typ_orand_leftmost_orctx_l_inv oc T U:
  orctx_l oc T = typ_orand_leftmost_orctx U ->
  exists U1,
    U = (typ_or U1 T) /\ oc = (typ_orand_leftmost_orctx U1) /\ typ_orand_count U1 <> 0.
Proof.
  introv Heq.
  induction U using typ_logic_ind.
  - eapply typ_orand_leftmost_orctx_notlog in H. rewrite H in Heq. discriminate.
  - exists U1. simpls. destruct (typ_orand_count U1). discriminate.
    inversions Heq. eauto.
  - simpls. discriminate.
Qed.
Lemma typ_orand_leftmost_orctx_r_inv oc T U:
  orctx_r T oc = typ_orand_leftmost_orctx U ->
  exists U2,
    U = (typ_or T U2) /\ oc = (typ_orand_leftmost_orctx U2) /\ typ_notorand T.
Proof.
  introv Heq.
  induction U using typ_logic_ind.
  - eapply typ_orand_leftmost_orctx_notlog in H. rewrite H in Heq. discriminate.
  - exists U2. simpls. remember (typ_orand_count U1). destruct n.
    inversions Heq. symmetry in Heqn. apply <- typ_orand_count_notorand in Heqn.  eauto.
 discriminate.
  - simpls. discriminate.
Qed.


Lemma typ_orand_leftmost_orctx_typ T:
  exists U, T = (orctx_typ (typ_orand_leftmost_orctx T) U).
Proof.
  remember (typ_orand_leftmost_orctx T) as oc.
  reverts T.
  induction oc; intros; simpls.
  - exists T. auto.
  - eapply typ_orand_leftmost_orctx_l_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all. exists x0. fequal; auto.
  - eapply typ_orand_leftmost_orctx_r_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all. exists x0. fequal; auto.
Qed.

Lemma typ_orand_leftmost_orctx_notorand T:
  orctx_notorand_le (typ_orand_leftmost_orctx T).
Proof.
  remember (typ_orand_leftmost_orctx T) as oc.
  reverts T.
  induction oc; constructor.
  - eapply typ_orand_leftmost_orctx_l_inv in Heqoc. destruct_all. eapply IHoc; eauto.
  - eapply typ_orand_leftmost_orctx_r_inv in Heqoc. destruct_all. eauto.
  - eapply typ_orand_leftmost_orctx_r_inv in Heqoc. destruct_all. eapply IHoc; eauto.
Qed.
Hint Resolve typ_orand_leftmost_orctx_notorand: core.

Lemma typ_orand_leftmost_orctx_orand_typ T:
  typ_orand_count T = 0 \/ exists U1 U2, T = (orctx_typ (typ_orand_leftmost_orctx T) (typ_and U1 U2)).
Proof.
  remember (typ_orand_leftmost_orctx T) as oc.
  reverts T.
  induction oc; intros; simpls.
  - destruct T; simpls; eauto.
    destruct (typ_orand_count T1); discriminate.
  - eapply typ_orand_leftmost_orctx_l_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + contradiction.
    + right. simpl. exists x0 x1. fequal; auto.
  - eapply typ_orand_leftmost_orctx_r_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + apply -> typ_orand_count_notorand in H1.
      left. simpl. rewrite H1, H. auto.
    + right. exists x0 x1. fequal; auto.
Qed.

Lemma typ_logic_and_count_and_l oc U1 U2:
  (typ_logic_and_count (orctx_typ oc U1)) < typ_logic_and_count (orctx_typ oc (typ_and U1 U2)).
Proof.
  induction oc; simpls.
  -
  apply Lt.le_lt_n_Sm.
  apply Plus.le_plus_r.
  - eapply Plus.plus_lt_compat_l. auto. 
  - eapply Plus.plus_lt_compat_r. auto.
Qed.
Lemma typ_logic_and_count_and_r oc U1 U2:
  (typ_logic_and_count (orctx_typ oc U2)) < typ_logic_and_count (orctx_typ oc (typ_and U1 U2)).
Proof.
 induction oc; simpls.
  -
  apply Lt.le_lt_n_Sm.
  apply Plus.le_plus_l.
  - eapply Plus.plus_lt_compat_l. auto. 
  - eapply Plus.plus_lt_compat_r. auto.
Qed.


Lemma typ_orand_ind (P: typ->Prop):
  (forall T, typ_notorand T -> P T) ->
  (forall oc T U, orctx_notorand_le oc -> P (orctx_typ oc T) -> P (orctx_typ oc U) -> P (orctx_typ oc (typ_and T U))) ->
  (forall T, P T).
Proof.
  introv Hn Ho.
  intros.
  remember (typ_logic_and_count T) as n.
  reverts T.
  induction n using nat_lt_induction; destruct n; intros.
  - (* No and-under ors *)
    symmetry in Heqn.
    eapply Hn. apply <- typ_logic_and_count_notorand. auto.
  - (* There is an and-under-or *)
    rename H into IHn.
    forwards: typ_orand_leftmost_orctx_orand_typ T.
    destruct H.
    + apply <- typ_orand_count_notorand in H.
      apply -> typ_logic_and_count_notorand in H.
      rewrite <- Heqn in H; discriminate.
    + destructc H as U1 U2 HU.
      rewrite HU.
      eapply Ho.
      * eauto.
      *
      eapply IHn. 2: eapply eq_refl.
      rewrite Heqn. rewrite HU at 2.
      eapply typ_logic_and_count_and_l.
      *
      eapply IHn. 2: eapply eq_refl.
      rewrite Heqn. rewrite HU at 2.
      eapply typ_logic_and_count_and_r.
Qed.

(* Analog of logctx_notbndsel_le_notbndsel *)
Lemma orctx_notorand_le_notorand ocp U:
  typ_notorand (orctx_typ ocp U) ->
  orctx_notorand_le ocp.
Proof.
  introv Hn.
  induction ocp; eauto.
  - inversions Hn; simpls. false H; eauto. eauto.
  - inversions Hn; simpls. false H; eauto. eauto.
Qed.

(* Analog of logctx_notbndsel_eq_sel_le, logctx_notbndsel_eq_bnd_le *)
Lemma orctx_notorand_eq_and_le oc T U S ocp:
  orctx_notorand_le oc ->
  orctx_typ ocp (typ_and T U) = orctx_typ oc S ->
  orctx_le oc ocp.
Proof.
  introv Hn He.
  reverts ocp.
  induction Hn; intros; simpls.
  - eapply orctx_le_nil.
  - destruct ocp; simpls; inversions He.
    + eapply orctx_le_l; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      eapply typ_notorand_oc_notand in H.
      eapply H. eauto.
    + eapply orctx_le_r; eauto.
Qed.



Lemma orctx_notorand_le_le ocL ocR TL TR:
  orctx_le ocL ocR ->
  orctx_typ ocL TL = orctx_typ ocR TR ->
  orctx_notorand_le ocR ->
  orctx_notorand_le ocL.
Proof.
  introv Hle He Hn.
  unfold orctx_le, orctx_path in *.
  reverts ocL TL TR.
  induction Hn; intros.
  - eapply orctx_le_nil_inv in Hle. subst. eauto.
  - destruct ocL; simpls; inversions Hle; eauto; inversions He; eauto.
  - destruct ocL; simpls.
    + inversions Hle; eauto.
    + inversions He. econstructor. eapply orctx_notorand_le_notorand; eauto.
    + inversions He. inversions Hle. eauto.
Qed.

Lemma typ_notorand_or_invl T U:
  typ_notorand (typ_or T U) ->
  typ_notorand T.
Proof.
  introv Hty.
  inversions Hty.
  - false. eapply H. eauto.
  - eauto.
Qed.
Lemma typ_notorand_or_invr T U:
  typ_notorand (typ_or T U) ->
  typ_notorand U.
Proof.
  introv Hty.
  inversions Hty.
  - false. eapply H. eauto.
  - eauto.
Qed.

Lemma typ_notorand_oc_notorand oc U:
  typ_notorand (orctx_typ oc U) ->
  typ_notorand U.
Proof.
  introv Hn.
  induction oc; simpls.
  - eauto.
  - eapply typ_notorand_or_invl in Hn. eauto.
  - eapply typ_notorand_or_invr in Hn. eauto.
Qed.


Lemma typ_notorand_eq_separate ocp oc S T U:
  typ_notorand S ->
  orctx_typ oc S = orctx_typ ocp (typ_and T U) ->
  orctx_separate oc ocp.
Proof.
  introv Hn He.
  forwards: orctx_intersect_cases_r He.
  - intros; discriminate.
  - destruct_all; subst.
    false. eapply typ_notorand_oc_notand in Hn. eapply Hn; eauto.
    auto.
Qed.

Lemma typ_notorand_separate_eq_oc oc ocp T U:
  orctx_le oc ocp ->
  orctx_typ oc T = orctx_typ ocp U ->
  orctx_notorand_le ocp ->
  orctx_separate oc ocp ->
  typ_notorand T.
Proof.
  introv Hle He Hn Hs.
  reverts ocp T U.
  induction oc; intros; simpls; subst.
  - inversions Hs.
  - destruct ocp; simpls; subst.
    + inversions Hs.
    + inversions Hle.
      inversions He.
      inversions Hs.
      inversions Hn.
      forwards: IHoc; eauto.
    + inversions Hle.
      inversions He.
      inversions Hs.
      inversions Hn.
      eapply typ_notorand_oc_notorand; eauto.
  - destruct ocp; simpls; subst.
    + inversions Hs.
    + inversions Hle.
    + inversions Hle.
      inversions He.
      inversions Hs.
      inversions Hn.
      forwards: IHoc; eauto.
Qed.


Lemma typ_notorand_eq_or_and oc T0 U0 T U:
  typ_notorand T0 ->
  typ_or T0 U0 = orctx_typ oc (typ_and T U) ->
  exists ocp,
    U0 = orctx_typ ocp (typ_and T U) /\
    oc = (orctx_r T0 ocp).
Proof.
  introv Hn He.
  destruct oc; simpls.
  - inversions He.
  - inversions He.
    eapply typ_notorand_oc_notand in Hn.
    false. eapply Hn; eauto.
  - inversions He.
    exists oc; eauto.
Qed.


Lemma typ_notorand_oc_replace oc T U:
  typ_notorand (orctx_typ oc T) ->
  typ_notorand U ->
  typ_notorand (orctx_typ oc U).
Proof.
  introv HnT HnU.
  induction oc; simpls; eauto.
  - forwards: typ_notorand_or_invr HnT.
    forwards: typ_notorand_or_invl HnT.
    eapply typ_notorand_or; eauto.
  - forwards: typ_notorand_or_invr HnT.
    forwards: typ_notorand_or_invl HnT.
    eapply typ_notorand_or; eauto.
Qed.

Lemma orctx_notorand_le_extend_invr oc1 oc2:
  orctx_notorand_le (orctx_extend oc1 oc2) ->
  orctx_notorand_le oc2.
Proof.
  introv H.
  induction oc1; simpls.
  - auto.
  - inversions H. eauto.
  - inversions H. eauto.
Qed.

Lemma orctx_notorand_replace_r oc ocp ocTU T U S:
  orctx_notorand_le ocp ->
  orctx_le ocp ocTU ->
  orctx_follow ocp oc ->
  orctx_typ ocTU T = orctx_typ ocp S ->
  orctx_typ ocTU U = orctx_typ oc S ->
  orctx_notorand_le oc.
Proof.
  introv Hn Hle HT HU.
  reverts ocTU.
  induction HT; intros; simpls.
  - eauto.
  - inversions Hn.
    destruct ocTU; simpls.
    + inversions Hle.
    + inversions Hle.
      inversions HU.
      inversions H.
      forwards: IHHT; eauto.
    + inversions HU.
      inversions H.
      eapply orctx_follow_eq_inv in H2. destruct H2; subst.
      eauto. eauto.
  - inversions Hn.
    destruct ocTU; simpls.
    + inversions Hle.
    + inversions Hle.
    + inversions Hle.
      inversions HU.
      inversions H.
      forwards: IHHT; eauto.
Qed.


Inductive ty_dist_eq (*(P: logctx -> typ -> Prop): logctx ->*): typ -> typ -> Prop :=
| ty_dist_eq_refl (*oc*) T:
(*  P oc T ->*)
  ty_dist_eq (*P oc*) T T
| ty_dist_eq_or (*oc*) T1 T2 U1 U2:
  ty_dist_eq (*P oc*) T1 U1 ->
  ty_dist_eq (*P oc*) T2 U2 ->
  ty_dist_eq (*P oc*) (typ_or T1 T2) (typ_or U1 U2)
| ty_dist_eq_and1 (*oc*) T1 T2 U:
  ty_dist_eq (*P oc*) T1 U ->
  ty_dist_eq (*P oc*) T2 U ->
  ty_dist_eq (*P oc*) (typ_and T1 T2) U
| ty_dist_eq_andl2 (*oc*) T1 T2 U:
  ty_dist_eq (*P oc*) U T1 ->
  ty_dist_eq (*P oc*) U (typ_and T1 T2)
| ty_dist_eq_andr2 (*oc*) T1 T2 U:
  ty_dist_eq (*P oc*) U T2 ->
  ty_dist_eq (*P oc*) U (typ_and T1 T2)
.
Hint Resolve ty_dist_eq_refl ty_dist_eq_or: core.


Lemma ty_dist_eq_and_invl_nil  T1 T2 U:
  ty_dist_eq (typ_and T1 T2) U ->
  ty_dist_eq T1 U.
Proof.
  introv He. remember (typ_and T1 T2). reverts T1 T2.
  induction He; intros; subst.
    + eapply ty_dist_eq_andl2. eapply ty_dist_eq_refl.
    + inversion Heqt.
    + inversions Heqt. auto.
    + eapply ty_dist_eq_andl2. eapply IHHe; eauto.
    + eapply ty_dist_eq_andr2. eapply IHHe; eauto.
Qed.
Lemma ty_dist_eq_and_invr_nil  T1 T2 U:
  ty_dist_eq (typ_and T1 T2) U ->
  ty_dist_eq T2 U.
Proof.
  introv He. remember (typ_and T1 T2). reverts T1 T2.
  induction He; intros; subst.
    + eapply ty_dist_eq_andr2. eapply ty_dist_eq_refl.
    + inversion Heqt.
    + inversions Heqt. auto.
    + eapply ty_dist_eq_andl2. eapply IHHe; eauto.
    + eapply ty_dist_eq_andr2. eapply IHHe; eauto.
Qed.



Lemma ty_dist_eq_notorand_eq T U:
  ty_dist_eq T U ->
  typ_notorand T ->
  typ_notorand U ->
  T = U.
Proof.
  introv He HT HU.
  induction He; eauto.
  - forwards: typ_notorand_or_invl HT.
    forwards: typ_notorand_or_invr HT.
    forwards: typ_notorand_or_invl HU.
    forwards: typ_notorand_or_invr HU.
    forwards: IHHe1; eauto.
    forwards: IHHe2; eauto.
    subst.
    eauto.
  - false. eapply typ_notorand_notand. eapply HT. eauto.
  - false. eapply typ_notorand_notand. eapply HU. eauto.
  - false. eapply typ_notorand_notand. eapply HU. eauto.
Qed.

Lemma  ty_dist_eq_logctx oc T U:
  ty_dist_eq T U ->
  ty_dist_eq (logctx_typ oc T) (logctx_typ oc U).

Proof.
  introv He.
  induction oc; simpls; eauto.

  - eapply ty_dist_eq_and1. eapply ty_dist_eq_andl2. eauto. eapply ty_dist_eq_andr2.  eapply ty_dist_eq_refl.
  - eapply ty_dist_eq_and1. eapply ty_dist_eq_andl2.  eapply ty_dist_eq_refl. eapply ty_dist_eq_andr2. eauto.
Qed.



Lemma ty_dist_eq_oc_and_cases T0 oc T U:
  typ_notorand T0 ->
  ty_dist_eq T0 (orctx_typ oc (typ_and T U)) ->
  ty_dist_eq T0 (orctx_typ oc T) \/ ty_dist_eq T0 (orctx_typ oc U).
Proof.
  introv Hn He.
  reverts T0.
  induction oc; intros; simpls.
  - inversions He.
    + eapply typ_notorand_notand in Hn. false Hn; eauto.
    + eapply typ_notorand_notand in Hn. false Hn; eauto.
    + eauto.
    + eauto.
  - inversions He.
    + eapply typ_notorand_or_invl in Hn. eapply typ_notorand_oc_notand in Hn. false Hn; eauto.
    + forwards: IHoc T2; eauto. eapply typ_notorand_or_invl; eauto.
      destruct_all; eauto.
    + eapply typ_notorand_notand in Hn. false Hn; eauto.
  - inversions He.
    + eapply typ_notorand_or_invr in Hn. eapply typ_notorand_oc_notand in Hn. false Hn; eauto.
    + forwards: IHoc T3; eauto. eapply typ_notorand_or_invr; eauto.
      destruct_all; eauto.
    + eapply typ_notorand_notand in Hn. false Hn; eauto.
Qed.

Lemma ty_dist_eq_oc T1 T2 oc:
  ty_dist_eq T1 T2 ->
  ty_dist_eq (orctx_typ oc T1) (orctx_typ oc T2).
Proof.
  introv He.
  induction oc; simpls; eauto.
Qed.

Lemma ty_dist_eq_oc_and_invl T1 oc T U:
  ty_dist_eq (orctx_typ oc (typ_and T U)) T1 ->
  ty_dist_eq (orctx_typ oc T) T1.
Proof.
  introv He.
  remember (orctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction He; intros; subst.
  - eapply ty_dist_eq_oc; constructor; eauto.
  - destruct oc; simpls; inversions Heqt.
    constructor; eauto.
    constructor; eauto.
  - destruct oc; simpls; inversions Heqt.
    eauto.
  - forwards: IHHe; eauto.
    eapply ty_dist_eq_andl2; eauto.
  - forwards: IHHe; eauto.
    eapply ty_dist_eq_andr2; eauto.
Qed.
Lemma ty_dist_eq_oc_and_invr T1 oc T U:
  ty_dist_eq (orctx_typ oc (typ_and T U)) T1 ->
  ty_dist_eq (orctx_typ oc U) T1.
Proof.
  introv He.
  remember (orctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction He; intros; subst.
  - eapply ty_dist_eq_oc; eapply ty_dist_eq_andr2; eauto.
  - destruct oc; simpls; inversions Heqt.
    constructor; eauto.
    constructor; eauto.
  - destruct oc; simpls; inversions Heqt.
    eauto.
  - forwards: IHHe; eauto.
    eapply ty_dist_eq_andl2; eauto.
  - forwards: IHHe; eauto.
    eapply ty_dist_eq_andr2; eauto.
Qed.

(* Analog of typ_bndsel_leftmost_logctx_sel_le, typ_bndsel_leftmost_logctx_bnd_le *)
Lemma typ_orand_leftmost_orctx_le oc T1 T2:
  orctx_le (typ_orand_leftmost_orctx (orctx_typ oc (typ_and T1 T2))) oc.
Proof.
  induction oc; simpls.
  - unfold orctx_le, orctx_path in *. simpls. constructor.
  - forwards: typ_orand_count_orand oc T1 T2.
    cases_if.
    eapply orctx_le_l; eauto.
  -
    cases_if.
    eapply orctx_le_r; eauto.
    constructor.
Qed.

(* Analog of typ_bndsel_leftmost_logctx_notbndsel_le *)
Lemma typ_orand_leftmost_orctx_notorand_le oc T:
  orctx_notorand_le oc ->
  orctx_le oc (typ_orand_leftmost_orctx (orctx_typ oc T)).
Proof.
  introv Hn.

  remember (typ_orand_leftmost_orctx (orctx_typ oc T)).
  reverts oc T.
  induction o; intros.
  - destruct oc; simpls.
    + constructor.
    + destruct (typ_orand_count (orctx_typ oc T)); discriminate.
    + destruct (typ_orand_count T0); discriminate.
  -
    eapply typ_orand_leftmost_orctx_l_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H. inversions Hn.
      eapply orctx_le_l; eauto.
    + inversions H. inversions Hn.
      apply typ_orand_count_notorand in H2. contradiction.
  -
    eapply typ_orand_leftmost_orctx_r_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + constructor. 
    + inversions H. inversions Hn.
      eapply orctx_le_r; eauto.
Qed.

Lemma typ_notlog_logctx_nil oc T:
  typ_notlog (logctx_typ oc T) ->
  oc = logctx_nil /\ typ_notlog T.
Proof.
  introv Hn.
  destruct oc; simpls; auto; hnf in Hn; destruct_all; solve [false H; eauto | false H0; eauto].
Qed.
