Set Implicit Arguments.
Require Import LibExtra DotTactics.
Require Import Types SyntaxClasses Objctx.

Require Import Lia.

(** * Additional induction schemes for types *)

(* Inductive property of types, which generates the induction scheme
   that is needed to prove reference_refl_incap_type *)
Inductive typ_icp_ind: typ -> Prop :=
(* Most types do not depend on anything *)
| typ_icp_ind_top: typ_icp_ind typ_top
| typ_icp_ind_bot: typ_icp_ind typ_bot
| typ_icp_ind_rcd D: typ_icp_ind (typ_rcd D)
(* Intersection and union types depend on their parts *)
| typ_icp_ind_and T1 T2:
    typ_icp_ind T1 ->
    typ_icp_ind T2 ->
    typ_icp_ind (typ_and T1 T2)
| typ_icp_ind_or T1 T2:
    typ_icp_ind T1 ->
    typ_icp_ind T2 ->
    typ_icp_ind (typ_or T1 T2)
| typ_icp_ind_sel x L y: typ_icp_ind (typ_sel x L y)
(* Recursive types depend on the inner part opened with an arbitrary variable *)
| typ_icp_ind_bnd T:
    (forall x, typ_icp_ind (open x T)) ->
    typ_icp_ind (typ_bnd T)
| typ_icp_ind_fun T1 T2: typ_icp_ind (typ_fun T1 T2)
| typ_icp_ind_nocap: typ_icp_ind (typ_nocap).

(* Defines a height of a type for the purpose of typ_icp_ind *)
Fixpoint typ_icp_ind_height (T: typ): nat :=
  match T with
  (* Recursive types are above the inner part *)
  | typ_bnd T => S (typ_icp_ind_height T)
  (* Intersection and union types are above both of the inner parts *)
  | typ_and T1 T2 => S (max (typ_icp_ind_height T1) (typ_icp_ind_height T2))
  | typ_or T1 T2 => S (max (typ_icp_ind_height T1) (typ_icp_ind_height T2))
  (* Most types have height 0 *)
  | _ => 0
  end.

(* Opening a type preserves its height *)
Lemma typ_icp_ind_height_open n x T:
  typ_icp_ind_height T = typ_icp_ind_height (open_rec n x T).
Proof.
  revert n.
  induction T; simpls; eauto.
Qed.


(* Allows using the inductive scheme generated by typ_icp_ind for all types *)
Lemma typ_icp_ind_all T:
  typ_icp_ind T.
Proof.
  remember (typ_icp_ind_height T).
  revert dependent T.
  (* The existence of a derivation of typ_icp_ind for all types
     is proven by induction on the height defined by typ_icp_ind_height.
     The types in the premises of every constructor in typ_icp_ind have lower height
     than the type in the conclusion.
     Therefore, an induction on natural numbers is used,
     in which the induction hypothesis assumes existence of derivation for all types with a lower height.
   *)
  
  set (P := (fun n => forall T : typ, n = typ_icp_ind_height T -> typ_icp_ind T)).
  pose proof (Nat.strong_right_induction P) as H.
  apply H with (z:=0).
  - (* P respects equality of numbers *)
    hnf.
    split; subst; eauto.
  - intros. hnf. intros.
    destruct T.
    all: try constructor. (* Cases with height 0 *)
    all: try (eapply H1; [ | | apply eq_refl]; simpls; lia). (* Intersections and unions *)
    (* Case typ_bnd *)
    intros.
    eapply H1.
    3: eapply typ_icp_ind_height_open.
    + lia.
    + subst; simpls; lia.
  - lia.
Qed.


(* Inductive property of types, which generates the induction scheme
   that opens nested types that are opened in typing *)
Inductive typ_opening_ind: typ -> Prop :=
(* Most types do not depend on anything *)
| typ_opening_ind_top: typ_opening_ind typ_top
| typ_opening_ind_bot: typ_opening_ind typ_bot
| typ_opening_ind_fun T1 T2: 
    typ_opening_ind T1 ->
    (forall x, typ_opening_ind (open x T2)) ->
    typ_opening_ind (typ_fun T1 T2)
| typ_opening_ind_fld a T1 T2: 
    typ_opening_ind T1 ->
    typ_opening_ind T2 ->
    typ_opening_ind (typ_rcd (dec_fld a T1 T2))
| typ_opening_ind_typ A T1 T2: 
    (forall x, typ_opening_ind (open x T1)) ->
    (forall x, typ_opening_ind (open x T2)) ->
    typ_opening_ind (typ_rcd (dec_typ A T1 T2))
| typ_opening_ind_met M T1 T2 T3: 
    (forall x, typ_opening_ind (open x T1)) ->
    typ_opening_ind T2 ->
    (forall x y, typ_opening_ind (open_met x y T1)) ->
    typ_opening_ind (typ_rcd (dec_met M T1 T2 T3))
(* Intersection and union types depend on their parts *)
| typ_opening_ind_and T1 T2:
    typ_opening_ind T1 ->
    typ_opening_ind T2 ->
    typ_opening_ind (typ_and T1 T2)
| typ_opening_ind_or T1 T2:
    typ_opening_ind T1 ->
    typ_opening_ind T2 ->
    typ_opening_ind (typ_or T1 T2)
| typ_opening_ind_sel x L y: typ_opening_ind (typ_sel x L y)
(* Recursive types depend on the inner part opened with an arbitrary variable *)
| typ_opening_ind_bnd T:
    (forall x, typ_opening_ind (open x T)) ->
    typ_opening_ind (typ_bnd T)
| typ_opening_ind_nocap: typ_opening_ind (typ_nocap).

(* Defines a height of a type for the purpose of typ_opening_ind *)
Fixpoint typ_opening_ind_height (T: typ): nat :=
  match T with
  (* Recursive types are above the inner part *)
  | typ_bnd T => S (typ_opening_ind_height T)
  (* Intersection and union types are above both of the inner parts *)
  | typ_and T1 T2 => S (max (typ_opening_ind_height T1) (typ_opening_ind_height T2))
  | typ_or T1 T2 => S (max (typ_opening_ind_height T1) (typ_opening_ind_height T2))
  | typ_fun T1 T2 => S (max (typ_opening_ind_height T1) (typ_opening_ind_height T2))
  | typ_rcd (dec_fld a T1 T2) => S (max (typ_opening_ind_height T1) (typ_opening_ind_height T2))
  | typ_rcd (dec_typ A T1 T2) => S (max (typ_opening_ind_height T1) (typ_opening_ind_height T2))
  | typ_rcd (dec_met m T1 T2 T3) => S (max (max (typ_opening_ind_height T1) (typ_opening_ind_height T2)) (typ_opening_ind_height T3))
  (* Most types have height 0 *)
  | _ => 0
  end.

(* Opening a type preserves its height *)
Lemma typ_opening_ind_height_open n x T:
  typ_opening_ind_height T = typ_opening_ind_height (open_rec n x T).
Proof.
  (* Mutual induction with declarations *)
  assert ((forall T n, typ_opening_ind_height T = typ_opening_ind_height (open_rec n x T)) /\ (forall D n, typ_opening_ind_height (typ_rcd D) = typ_opening_ind_height (open_rec n x (typ_rcd D)))).
  apply typ_mutind; intros; simpls; eauto.
  destruct H; eauto.
Qed.


(* Allows using the inductive scheme generated by typ_opening_ind for all types *)
Lemma typ_opening_ind_all T:
  typ_opening_ind T.
Proof.
  remember (typ_opening_ind_height T).
  revert dependent T.
  (* The existence of a derivation of typ_opening_ind for all types
     is proven by induction on the height defined by typ_opening_ind_height.
     The types in the premises of every constructor in typ_opening_ind have lower height
     than the type in the conclusion.
     Therefore, an induction on natural numbers is used,
     in which the induction hypothesis assumes existence of derivation for all types with a lower height.
   *)
  
  set (P := (fun n => forall T : typ, n = typ_opening_ind_height T -> typ_opening_ind T)).
  pose proof (Nat.strong_right_induction P) as H.
  apply H with (z:=0).
  - (* P respects equality of numbers *)
    hnf.
    split; subst; eauto.
  - intros. hnf. intros.
    destruct T.
    all: try constructor. (* Cases with height 0 *)
    all: try (eapply H1; [ | | apply eq_refl]; simpls; lia). (* Intersections and unions *)
    + destruct d; subst; simpls; constructor; intros; try eapply H1; try eapply typ_opening_ind_height_open; try eapply eq_refl; simpls; try lia.
      (* met result type *)
      rewrite <- typ_opening_ind_height_open . lia.
    + intros;
      eapply H1; try eapply typ_opening_ind_height_open; simpls; lia.
    + intros;
      eapply H1; try eapply typ_opening_ind_height_open; simpls; lia.
 - lia.
Qed.