(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Import LibExtra DotTactics SyntaxClasses.
Require Import Vars Labels TransformCollect.

(* ************************************************************************** *)
(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** * Types
    Types ([typ], [S], [T], [U]), type lists ([typs], [Ts]) and type
    declarations ([dec], [D]):
    - [typ_top] represents [⊤];
    - [typ_bot] represents [⊥];
    - [typ_rcd d] represents a record type [d], where [d] is either a type or
      field declaration;
    - [typ_and T U] represents an intersection type [T ∧ U];
    - [typ_or T U] represents a union type [T ∨ U];
    - [typ_sel x A y] represents type selection [x.A(y)];
    - [typ_bnd T] represents a recursive type [μ(x: T)]; however, since [x] is
      bound in the recursive type, it is referred to in [T] using the de Bruijn
      index 0, and is therefore omitted from the type representation; we will
      denote recursive types as [μ(T)];
    - [typ_fun T U] represents the dependent function type [∀(x: T)U]; as
      in the previous case, [x] represents a variable bound in [U], and is
      therefore omitted from the representation; we will denote function types
      as [∀(T)U];
    - [typ_nocap] represents the "read-only bottom" type [N]
 *)
Inductive typ : Set :=
  | typ_top  : typ
  | typ_bot  : typ
  | typ_rcd  : dec -> typ
  | typ_and  : typ -> typ -> typ
  | typ_or   : typ -> typ -> typ
  | typ_sel  : avar -> typ_label -> avar -> typ
  | typ_bnd  : typ -> typ
  | typ_fun  : typ -> typ -> typ
  | typ_nocap: typ (* subtype of types without reference capabilities *)
(* Part of Figure 8 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(**
  - [dec_typ A S T] represents a type declaraion [{A(r): S..T}];
  - [dec_fld a S T] represents a field declaration [{a: S..T}];
  - [dec_met m S T U] represents a field declaration [{m(r:S, z: T): U}] . *)
with dec : Set :=
  | dec_typ: typ_label -> typ -> typ -> dec
  | dec_fld: fld_label -> typ -> typ -> dec
  | dec_met: met_label ->  typ -> typ -> typ -> dec
.
Hint Constructors typ dec : core.

Notation typ_cap_marker C T := (typ_rcd (dec_typ (typ_label_refcap C) typ_bot T)) (only parsing).
(* The mutability marker type [{M(r) : ⊥..T}], where r is the receiver.
   Using de Bruijn indices, the receiver r is the 0-th index in T *)
Notation typ_mut_marker T := (typ_cap_marker typ_refcap_mutability T) (only parsing).

(** *** Mutual Induction Principles *)
Scheme typ_mut  := Induction for typ Sort Prop
with   dec_mut  := Induction for dec Sort Prop.
Combined Scheme typ_mutind from typ_mut, dec_mut.

(* ************************************************************************** *)
(** ** Types, Type Lists and Declarations are Abstract Syntax *)
Fixpoint open_rec_typ (k: nat) (u: var) (T: typ) {struct T} : typ :=
  match T with
  | typ_top        => typ_top
  | typ_bot        => typ_bot
  | typ_rcd D      => typ_rcd (open_rec_dec k u D)
  | typ_and T1 T2  => typ_and (open_rec_typ k u T1)
                             (open_rec_typ k u T2)
  | typ_or T1 T2  => typ_or (open_rec_typ k u T1)
                             (open_rec_typ k u T2)
  | typ_sel x L y => typ_sel (open_rec k u x) L (open_rec k u y)
  | typ_bnd T      => typ_bnd (open_rec_typ (S k) u T)
  | typ_fun T1 T2  => typ_fun (open_rec_typ k u T1)
                             (open_rec_typ (S k) u T2)
  | typ_nocap => typ_nocap
  end with
open_rec_dec (k: nat) (u: var) (D: dec): dec :=
  match D with
  | dec_typ L T U => dec_typ L (open_rec_typ (S k) u T)
                            (open_rec_typ (S k) u U)
  | dec_fld m T U => dec_fld m (open_rec_typ k u T)
                            (open_rec_typ k u U)
  | dec_met m TC TZ TR =>
    dec_met m (open_rec_typ (S k) u TC) (open_rec_typ k u TZ) (open_rec_typ (S (S k)) u TR)
  end.
Instance TypOpenable : Openable typ := open_rec_typ.
Instance DecOpenable : Openable dec := open_rec_dec.
Ltac fold_typ_open_rec :=
  change (open_rec_dec) with (@open_rec dec DecOpenable);
  change (open_rec_typ) with (@open_rec typ TypOpenable).

Fixpoint fv_typ (T: typ) : vars :=
  match T with
  | typ_top        => \{}
  | typ_bot        => \{}
  | typ_rcd D      => (fv_dec D)
  | typ_and T U    => (fv_typ T) \u (fv_typ U)
  | typ_or T U    => (fv_typ T) \u (fv_typ U)
  | typ_sel x L y   => (fv x) \u (fv y)
  | typ_bnd T      => (fv_typ T)
  | typ_fun T1 T2  => (fv_typ T1) \u (fv_typ T2)
  | typ_nocap => \{}
  end with
fv_dec (D: dec) : vars :=
  match D with
  | dec_typ L T U => (fv_typ T) \u (fv_typ U)
  | dec_fld m T U => (fv_typ T) \u (fv_typ U)
  | dec_met m TC TZ TR => (fv_typ TC) \u (fv_typ TZ) \u (fv_typ TR)
  end.
Instance TypFreeVar : FreeVar typ := fv_typ.
Instance DecFreeVar : FreeVar dec := fv_dec.
Ltac fold_typ_fv :=
  change (fv_dec) with (@fv dec DecFreeVar);
  change (fv_typ) with (@fv typ TypFreeVar).
Ltac folds_typ_fv :=
  change (fv_dec) with (@fv dec DecFreeVar) in *;
  change (fv_typ) with (@fv typ TypFreeVar) in *.

Fixpoint subst_typ (z: var) (u: var) (T: typ) { struct T } : typ :=
  match T with
  | typ_top        => typ_top
  | typ_bot        => typ_bot
  | typ_rcd D      => typ_rcd (subst_dec z u D)
  | typ_and T1 T2  => typ_and (subst_typ z u T1) (subst_typ z u T2)
  | typ_or T1 T2  => typ_or (subst_typ z u T1) (subst_typ z u T2)
  | typ_sel x L y    => typ_sel (subst_var z u x) L (subst_var z u y)
  | typ_bnd T      => typ_bnd (subst_typ z u T)
  | typ_fun T U    => typ_fun (subst_typ z u T) (subst_typ z u U)
  | typ_nocap => typ_nocap
  end with
subst_dec (z: var) (u: var) (D: dec) { struct D } : dec :=
  match D with
  | dec_typ L T U => dec_typ L (subst_typ z u T) (subst_typ z u U)
  | dec_fld L T U => dec_fld L (subst_typ z u T) (subst_typ z u U)
  | dec_met m TC TZ TR => dec_met m (subst_typ z u TC) (subst_typ z u TZ) (subst_typ z u TR)
  end.
Instance TypSubstVar : SubstVar typ := subst_typ.
Instance DecSubstVar : SubstVar dec := subst_dec.
Ltac fold_typ_subst :=
  change (subst_dec) with (@subst_var dec DecSubstVar);
  change (subst_typ) with (@subst_var typ TypSubstVar).

Fixpoint seal_rec_typ (k: nat) (u: var) (T: typ) {struct T} : typ :=
  match T with
  | typ_top        => typ_top
  | typ_bot        => typ_bot
  | typ_rcd D      => typ_rcd (seal_rec_dec k u D)
  | typ_and T1 T2  => typ_and (seal_rec_typ k u T1)
                             (seal_rec_typ k u T2)
  | typ_or T1 T2  => typ_or (seal_rec_typ k u T1)
                             (seal_rec_typ k u T2)
  | typ_sel x L y => typ_sel (seal_rec k u x) L (seal_rec k u y)
  | typ_bnd T      => typ_bnd (seal_rec_typ (S k) u T)
  | typ_fun T1 T2  => typ_fun (seal_rec_typ k u T1)
                             (seal_rec_typ (S k) u T2)
  | typ_nocap => typ_nocap
  end with
seal_rec_dec (k: nat) (u: var) (D: dec): dec :=
  match D with
  | dec_typ L T U => dec_typ L (seal_rec_typ (S k) u T)
                            (seal_rec_typ (S k) u U)
  | dec_fld m T U => dec_fld m (seal_rec_typ k u T)
                            (seal_rec_typ k u U)
  | dec_met m TC TZ TR =>
    dec_met m (seal_rec_typ (S k) u TC) (seal_rec_typ k u TZ) (seal_rec_typ (S (S k)) u TR)
  end.
Instance TypSealable : Sealable typ := seal_rec_typ.
Instance DecSealable : Sealable dec := seal_rec_dec.
Ltac fold_typ_seal_rec :=
  change (seal_rec_dec) with (@seal_rec dec DecSealable);
  change (seal_rec_typ) with (@seal_rec typ TypSealable).
Instance TypAbstractSyntax :
  AbstractSyntax TypOpenable TypFreeVar TypSubstVar := {}.
Instance DecAbstractSyntax :
  AbstractSyntax DecOpenable DecFreeVar DecSubstVar := {}.

Ltac fold_typ_abs :=
  fold_typ_open_rec; fold_typ_seal_rec; fold_typ_fv; fold_typ_subst.

(** *** Autorewrite lemmas *)
Lemma rewrite_open_rec_typ : open_rec_typ = open_rec.
Proof. trivial. Qed.
Lemma rewrite_open_rec_dec : open_rec_dec = open_rec.
Proof. trivial. Qed.
Lemma rewrite_fv_typ : fv_typ = fv.
Proof. trivial. Qed.
Lemma rewrite_fv_dec : fv_dec = fv.
Proof. trivial. Qed.
Lemma rewrite_subst_typ : subst_typ = subst_var.
Proof. trivial. Qed.
Lemma rewrite_subst_dec : subst_dec = subst_var.
Proof. trivial. Qed.
Hint Rewrite
     rewrite_open_rec_typ rewrite_fv_typ rewrite_subst_typ
     rewrite_open_rec_dec rewrite_fv_dec rewrite_subst_dec : typ_db.

(* ************************************************************************** *)
(** ** Declarations are labeled *)
Instance DecLabelOf : LabelOf dec :=
  fun D =>
      match D with
      | dec_typ L _ _ => label_typ L
      | dec_fld m _ _ => label_fld m
      | dec_met m _ _ _ => label_met m
      end.


(* ************************************************************************** *)
(** ** Types, Type Lists and Declarations Satisfy Syntax Lemmas *)
(** We will prove that the operations on Types, Type Lists and Declarations are
    lifted from locally nameless variables. The syntax lemmas will then by
    automatically lifted from the syntax lemmas of locally nameless variables
    through the type class mechanism *)

(** *** Collect Variables in Types *)
Fixpoint collect_typ {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (T: typ) : C :=
  match T with
  | typ_top        => df
  | typ_bot        => df
  | typ_rcd D      => (collect_dec df f comb D)
  | typ_and T U    => comb (collect_typ df f comb T) (collect_typ df f comb U)
  | typ_or T U    => comb (collect_typ df f comb T) (collect_typ df f comb U)
  | typ_sel x L y    => comb (f x) (f y)
  | typ_bnd T      => (collect_typ df f comb T)
  | typ_fun T1 T2  => comb (collect_typ df f comb T1) (collect_typ df f comb T2)
  | typ_nocap => df
  end with
collect_dec {C: Type} (df : C) (f : avar -> C) (comb: C -> C -> C) (D: dec) : C :=
  match D with
  | dec_typ L T U => comb (collect_typ df f comb T) (collect_typ df f comb U)
  | dec_fld m T U => comb (collect_typ df f comb T) (collect_typ df f comb U)
  | dec_met m TC TZ TR => comb (comb (collect_typ df f comb TC) (collect_typ df f comb TZ)) (collect_typ df f comb TR)
  end.

Instance TypCollect : Collect AvarAbstractSyntax TypAbstractSyntax
  := @collect_typ.

Instance DecCollect : Collect AvarAbstractSyntax DecAbstractSyntax :=
  @collect_dec.

Ltac fold_typ_collect :=
  change (@collect_typ) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            typ TypOpenable TypFreeVar TypSubstVar TypAbstractSyntax
            TypCollect);
  change (@collect_dec) with (@collect
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            dec DecOpenable DecFreeVar DecSubstVar DecAbstractSyntax
            DecCollect).

(** *** Autorewrite lemmas *)
Lemma rewrite_collect_typ : forall C, @collect_typ C = collect.
Proof. trivial. Qed.
Lemma rewrite_collect_dec : forall C, @collect_dec C = collect.
Proof. trivial. Qed.
Hint Rewrite
     rewrite_collect_typ rewrite_collect_dec : typ_db.

(** *** Sanity for Collecting Variables in Types, Type Lists and Declarations *)
Lemma collect_compose_typ_dec :
  forall {C D} (f : C -> D) g comb1 comb2 i,
    (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
    (forall (T : typ),
        (f ∘ collect i g comb1) T = collect (f i) (f ∘ g) comb2 T) /\
    (forall (D : dec),
        (f ∘ collect i g comb1) D = collect (f i) (f ∘ g) comb2 D).
Proof.
  intros.
  apply typ_mutind;
    unfold compose, collect, DecCollect, TypCollect;
    simpls; try congruence.
Qed.

Lemma collect_reflect_typ_dec :
  forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall bl,
      (forall T, ((collect_typ (Is_true bl) P or T)
             <-> (collect_typ bl p orb T) = true)) /\
      (forall D, ((collect_dec (Is_true bl) P or D)
             <-> (collect_dec bl p orb D) = true)).
Proof.
  intros P p H bl.
  apply typ_mutind; simpl; auto 2 using reflect_iff.
  all: unfold iff; auto 3 using Is_true_eq_left, Is_true_eq_true.
  all: split.
  all: destruct_ands; intros.
  all: destruct_orbs; destruct_ors.
  all: auto.
  
  all: forwards uu ui: H; auto.
Qed.

Local Ltac solve_collect_compose :=
  unfold compose; apply fun_ext_dep;
  apply collect_compose_typ_dec; auto.

Local Ltac solve_collect_reflect :=
  apply collect_reflect_typ_dec; auto.

Local Ltac solve_collect_sanity :=
  hnf; intros; solve [solve_collect_compose || solve_collect_reflect].

Instance TypCollectCompose : CollectCompose TypCollect.
Proof. solve_collect_sanity. Qed.

Instance TypCollectReflect : CollectReflect TypCollect.
Proof. solve_collect_sanity. Qed.

Instance DecCollectCompose : CollectCompose DecCollect.
Proof. solve_collect_sanity. Qed.

Instance DecCollectReflect : CollectReflect DecCollect.
Proof. solve_collect_sanity. Qed.

(** *** Transform Variables in Types *)
Fixpoint transform_typ (f : avar -> nat -> avar) (T : typ) (n : nat) : typ :=
  match T with
  | typ_top        => typ_top
  | typ_bot        => typ_bot
  | typ_rcd D      => typ_rcd (transform_dec f D n)
  | typ_and T1 T2  => typ_and (transform_typ f T1 n) (transform_typ f T2 n)
  | typ_or T1 T2  => typ_or (transform_typ f T1 n) (transform_typ f T2 n)
  | typ_sel x L y    => typ_sel (f x n) L (f y n)
  | typ_bnd T      => typ_bnd (transform_typ f T (S n))
  | typ_fun T1 T2  => typ_fun (transform_typ f T1 n) (transform_typ f T2 (S n))
  | typ_nocap => typ_nocap
  end with
transform_dec (f : avar -> nat -> avar) (D : dec) (n : nat) : dec :=
  match D with
  | dec_typ L T U => dec_typ L (transform_typ f T (S n)) (transform_typ f U (S n))
  | dec_fld m T U => dec_fld m (transform_typ f T n) (transform_typ f U n)
  | dec_met m TC TZ TR => dec_met m (transform_typ f TC (S n)) (transform_typ f TZ n) (transform_typ f TR (S (S n)))
  end.

Instance TypTransform : Transform AvarAbstractSyntax TypAbstractSyntax :=
  @transform_typ.

Instance DecTransform : Transform AvarAbstractSyntax DecAbstractSyntax :=
  @transform_dec.
Ltac fold_typ_transform :=
  change (@transform_typ) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            typ TypOpenable TypFreeVar TypSubstVar TypAbstractSyntax
            TypTransform);
  change (@transform_dec) with (@transform
            avar AvarOpenable AvarFreeVar AvarSubstVar AvarAbstractSyntax
            dec DecOpenable DecFreeVar DecSubstVar DecAbstractSyntax
            DecTransform).
Local Ltac fold_typ_all :=
  repeat (fold_typ_abs; fold_typ_collect; fold_typ_transform).
Local Ltac sympl_typ :=
  simpls; fold_typ_all.

(** *** Autorewrite lemmas *)
Lemma rewrite_transform_typ : transform_typ = transform.
Proof. trivial. Qed.
Lemma rewrite_transform_dec : transform_dec = transform.
Proof. trivial. Qed.
Hint Rewrite
     rewrite_transform_typ rewrite_transform_dec : typ_db.

(** *** Sanity for Transforming Variables *)
Lemma transform_id_typ_dec :
  (forall (T : typ) n, transform (ignore_second id) T n = T) /\
  (forall (D : dec) n, transform (ignore_second id) D n = D).
Proof.
  unfold ignore_second, id, transform; simpls.
  apply typ_mutind; simpl; intros; try f_equal; auto.
Qed.

Lemma transform_compose_typ_dec :
  (forall (T : typ) n f g, ((transform f) ∘∘ (transform g)) T n = transform (f ∘∘ g) T n) /\
  (forall (D : dec) n f g, ((transform f) ∘∘ (transform g)) D n = transform (f ∘∘ g) D n).
Proof.
  unfold compose12, transform; simpls.
  apply typ_mutind; simpl; intros;
  repeat
    match goal with
    | [ H : forall n f g, _ = _ |- _] => rewrite <- H
    end; auto.
Qed.

Local Ltac solve_transform_sanity :=
  hnf; try apply transform_id_typ_dec;
    intros; intros_fun_ext; apply transform_compose_typ_dec.

Instance TypTransformId : TransformId TypTransform.
Proof. solve_transform_sanity. Qed.

Instance DecTransformId : TransformId DecTransform.
Proof. solve_transform_sanity. Qed.

Instance TypTransformCompose : TransformCompose TypTransform.
Proof. solve_transform_sanity. Qed.

Instance DecTransformCompose : TransformCompose DecTransform.
Proof. solve_transform_sanity. Qed.

Lemma transform_shift_typ_dec (f : avar -> nat -> avar) :
  (forall (X : typ) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n) /\
  (forall (X : dec) (n m : nat), transform f X (n + m) = transform (fun (X' : avar) (n' : nat) => f X' (n' + m)) X n).
Proof.
  apply typ_mutind; simpl; intros; try f_equal; auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
  - rewrite <- H0. rewrite Nat.add_succ_l. auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
  - rewrite <- H0. rewrite Nat.add_succ_l. auto.
  - rewrite <- H. rewrite Nat.add_succ_l. auto.
  - rewrite <- H1. rewrite Nat.add_succ_l. auto.
Qed.

Instance TypTransformShift : TransformShift TypTransform.
Proof.
  intro. eapply transform_shift_typ_dec.
Qed.
Instance DecTransformShift : TransformShift TypTransform.
Proof.
  intro. eapply transform_shift_typ_dec.
Qed.


(** *** Sanity for Collecting and Transforming Variables *)
Lemma collect_transform_typ_typs_dec :
  (forall (T : typ) bl p f g,
      (collect bl p andb T) = true -> transform (dec_apply p f g) T = transform f T) /\
  (forall (D : dec) bl p f g,
      (collect bl p andb D) = true -> transform (dec_apply p f g) D = transform f D).
Proof.
  apply typ_mutind; simpl; auto 2; intros; destruct_andbs; destruct_ands;
    repeat
      match goal with
      | [ H : forall bl p f g,
            _ = true ->
            ?transform (dec_apply p f g) ?t = ?transform f ?t |- _] =>
        erewrite H by eauto
      end; auto.
  unfold dec_apply; rewrite H, H0; auto.
Qed.

Instance TypCollectTransform :
  CollectTransform TypCollect TypTransform.
Proof. hnf; intros; eapply collect_transform_typ_typs_dec; eauto. Qed.

Instance DecCollectTransform :
  CollectTransform DecCollect DecTransform.
Proof. hnf; intros; eapply collect_transform_typ_typs_dec; eauto. Qed.




(** *** Operations are lifted from operations on Variables *)
Lemma fv_collect_typ_dec :
  (forall (T : typ), fv T = collect (\{}) fv (@union var) T) /\
  (forall (D : dec), fv D = collect (\{}) fv (@union var) D).
Proof.
  apply typ_mutind; sympl_typ; try congruence.
  (* For ternary types, the unions are in different associativity, but rewrite union_assoc for some reason does not work *)
  intros.
  rewrite <- H, <- H0, <- H1.
  auto.
Qed.

Lemma open_transform_typ_dec :
  (forall (T : typ) n x, (open_depth_rec n x) T = transform (open_depth_rec n x) T) /\
  (forall (D : dec) n x, (open_depth_rec n x) D = transform (open_depth_rec n x) D).
Proof.
  unfold open_depth_rec.
  apply typ_mutind; sympl_typ; intros; repeat (apply fun_ext_dep; intro);
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    rewrite ? Nat.add_succ_r;
    auto.
Qed.

Lemma subst_transform_typ_dec :
  (forall (T : typ) x y, (ignore_second (subst_var x y)) T
                = transform (ignore_second (subst_var x y)) T) /\
  (forall (D : dec) x y, (ignore_second (subst_var x y)) D
                = transform (ignore_second (subst_var x y)) D).
Proof.
  unfold ignore_second.
  apply typ_mutind; sympl_typ; intros; intros_fun_ext;
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    unfold ignore_second; simpl; auto.
Qed.

Lemma seal_transform_typ_dec :
  (forall (T : typ) n x, (seal_depth_rec n x) T = transform (seal_depth_rec n x) T) /\
  (forall (D : dec) n x, (seal_depth_rec n x) D = transform (seal_depth_rec n x) D).
Proof.
  unfold seal_depth_rec.
  apply typ_mutind; sympl_typ; intros; repeat (apply fun_ext_dep; intro);
    repeat match goal with
           | [ H : forall _ _, _ = _ |- _] => rewrite <- H
           end;
    rewrite ? Nat.add_succ_r;
    auto.
Qed.

Local Ltac solve_transform_syntax :=
  hnf; intros; (apply fun_ext_dep; intro);
    try apply fv_collect_typ_dec;
    try apply open_transform_typ_dec;
    try apply subst_transform_typ_dec.

Instance TypFvCollect : FvCollect TypCollect.
Proof. solve_transform_syntax. Qed.

Instance DecFvCollect : FvCollect DecCollect.
Proof. solve_transform_syntax. Qed.

Instance TypOpenTransform : OpenTransform TypTransform.
Proof. solve_transform_syntax. Qed.

Instance DecOpenTransform : OpenTransform DecTransform.
Proof. solve_transform_syntax. Qed.

Instance TypSealTransform : SealTransform TypTransform AvarSealable TypSealable.
Proof.
  hnf; intros; (apply fun_ext_dep; intro).
  apply seal_transform_typ_dec.
Qed.

Instance TypSubstTransform : SubstTransform TypTransform.
Proof. solve_transform_syntax. Qed.

Instance DecSubstTransform : SubstTransform DecTransform.
Proof. solve_transform_syntax. Qed.

(** Additional Lemmas that are not lifted automatically *)
Lemma fv_open_typ_dec_cases : forall x,
  FvOpenCasesInd TypAbstractSyntax x /\
  FvOpenCasesInd DecAbstractSyntax x.
Proof.
  intros x.
  apply typ_mutind; sympl_typ; auto; intros;
    repeat apply union_eq_cases; auto using fv_open_cases.
Qed.

Instance TypFvOpenCases : FvOpenCases TypAbstractSyntax.
Proof.
  hnf. apply fv_open_typ_dec_cases.
Qed.

Instance DecFvOpenCases : FvOpenCases DecAbstractSyntax.
Proof.
  hnf. apply fv_open_typ_dec_cases.
Qed.

Lemma subst_id_typ (v1: var):
  (forall (t: typ),
  subst_var v1 v1 t = t) /\
  (forall (d: dec),
  subst_var v1 v1 d = d).
Proof.
  apply typ_mutind; intros; auto.
  - rewrite <- H at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; auto.
  - rewrite <- (subst_id v1 a) at 2. rewrite <- (subst_id v1 a0) at 2; auto.
  - rewrite <- H at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; auto.
  - rewrite <- H at 2; rewrite <- H0 at 2; rewrite <- H1 at 2; auto.
Qed.

Lemma subst_id_typ_typ (v1: var) (t: typ):
  subst_var v1 v1 t = t.
Proof.
  destruct (subst_id_typ v1).
  auto.
Qed.

Instance TypSubstId : SubstId TypAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite subst_id_typ_typ. compute. auto.
Qed.

Instance DecSubstId : SubstId DecAbstractSyntax. Proof.
  hnf. intros.
  eapply fun_ext_nondep_1. intro. rewrite (proj22 (subst_id_typ x)). compute. auto.
Qed.

Hint Extern 10 (@eq vars _ _) => apply fset_extens; hnf; intros; autorewrite with sets in *; destruct_all; auto with sets : sets.
Hint Extern 10 (not (@eq var _ _)) => match goal with
  Hx: ?x \in ?A, Hy: ?y \notin ?A |- ?x <> ?y => apply (in_notin_neq Hx Hy)
end : sets.

Lemma fv_subst_cases_typdec : forall x y,
    (forall (T : typ), FvSubstCasesOf x y T) /\
    (forall (D : dec), FvSubstCasesOf x y D).
Proof.
  intros x y.
  apply typ_mutind; sympl_typ; intros; eauto.

  all: destruct_all; subst.
  all: repeat rewrite subst_id_typ_typ.
  all: rewrite_sets.
  all: eauto.
  all: autorewrite with sets in *.

  all:  destruct_all; subst; rewrite_sets; autorewrite with sets in *; auto 7 using union_remove_swap2, union_remove_swap2r; try easy.

  all: auto 11 using fv_subst_cases with sets.

  destruct (fv_subst_cases x y a) as [[Hfva [Hx | Hna]] | [Hnxy [Hxa Hfva]]];
  destruct (fv_subst_cases x y a0) as [[Hfva0 [Hx0 | Hna0]] | [Hnxy0 [Hxa0 Hfva0]]];
  subst;
  rewrite Hfva;
  rewrite Hfva0; eauto; try contradiction;
  rewrite_sets;autorewrite with sets in *; eauto 7 using union_remove_swap2, union_remove_swap2r; try easy.
  right. splits; eauto.
  rewrite <- (union_comm \{y}).
  rewrite (union_assoc \{y}).
  rewrite union_same.

  rewrite union_comm_assoc.
  rewrite union_remove_dist.
  rewrite union_comm.
  auto.
Qed.

Instance TypFvSubstCases: FvSubstCases TypAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_typdec.
Qed.
Instance DecFvSubstCases: FvSubstCases DecAbstractSyntax.
Proof.
  hnf. apply fv_subst_cases_typdec.
Qed.

Lemma fv_close_typdec:
    (forall (T : typ) n x, FvCloseOf n x T) /\
    (forall (D : dec) n x, FvCloseOf n x D).
Proof.
  apply typ_mutind; sympl_typ; intros; unfold close_rec, TransformClosing in *; autorewrite with sets in *; destruct_all; rewrite_sets;autorewrite with sets in *; auto 7 using union_remove_swap2, union_remove_swap2r; try easy.
  all: try rewrite remove_empty_r; eauto.
  - rewrite H. rewrite H0.
    rewrite <- union_remove_dist. auto.
  - rewrite H. rewrite H0.
    rewrite <- union_remove_dist. auto.
  - unfold close_depth_rec. rewrite fv_close. rewrite fv_close.
    rewrite <- union_remove_dist. auto.
  - rewrite transform_close_depth_1. rewrite H. auto.
  - rewrite H. rewrite transform_close_depth_1. rewrite H0.
    rewrite <- union_remove_dist. auto.
  - rewrite transform_close_depth_1. rewrite H. rewrite transform_close_depth_1. rewrite H0.
    rewrite <- union_remove_dist. auto.
  - rewrite H. rewrite H0.
    rewrite <- union_remove_dist. auto.
  - rewrite transform_close_depth_1. rewrite H. rewrite H0. rewrite transform_close_depth_2. rewrite H1.
    rewrite <- union_remove_dist.
    rewrite <- union_remove_dist. auto.
Qed.

Instance TypFvClose: FvClose (TransformClosing (A:= typ)).
Proof.
  hnf. intros. apply fv_close_typdec.
Qed.
Instance DecFvClose: FvClose  (TransformClosing (A:= dec)).
Proof.
  hnf. intros. apply fv_close_typdec.
Qed.

Lemma open_seal_inv_typdec:
  (forall (t: typ) n y m x, OpenSealInvOf n y m x t) /\
  (forall (d: dec) n y m x, OpenSealInvOf n y m x d).
Proof.
  apply typ_mutind; intros; simpls; eauto.
  - inversions H1. erewrite H; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H0. erewrite open_seal_inv; eauto. erewrite open_seal_inv; eauto.
  - inversions H1. erewrite H. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H2. erewrite H. 2: eapply Lt.lt_n_S; eauto. 2: eauto. erewrite H0. eauto. eapply Lt.lt_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H3. erewrite H. 2: eapply Lt.lt_n_S; eauto. 2: eauto. erewrite H0; eauto. erewrite H1. 2: eapply Lt.lt_n_S; eapply Lt.lt_n_S; eauto. 2: eauto. eauto.
Qed.

Instance TypOpenSealInv: OpenSealInv TypAbstractSyntax TypSealable.
Proof.
  hnf; intros.
  eapply open_seal_inv_typdec; eauto.
Qed.

(* Solves union equality, where on each side there may be multiple occurences of \{y},
  otherwise they are the same. This tactic brings the singleton to the front and eliminates redundant ones. *)
Local Ltac solve_union_y y := (
  repeat rewrite <- union_assoc;
  repeat rewrite <- (union_comm_assoc \{y});
  repeat rewrite union_same_push;
  eauto).


Lemma fv_seal_cases_typdec:
  (forall (t: typ) n y, FvSealCasesOf y t n) /\
  (forall (d: dec) n y, FvSealCasesOf y d n).
Proof.
  apply typ_mutind; sympl_typ; intros; simpls; eauto.
  - specialize (H n y). specialize (H0 n y).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - specialize (H n y). specialize (H0 n y).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - pose proof (fv_seal_cases y a n). pose proof (fv_seal_cases y a0 n).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - specialize (H n y). specialize (H0 (S n) y).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - specialize (H (S n) y). specialize (H0 (S n) y).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - specialize (H n y). specialize (H0 n y).
    destruct_all; subst;
    rewrite H; rewrite H0; solve_union_y y.
  - specialize (H (S n) y). specialize (H0 n y). specialize (H1 (S (S n)) y).
    destruct_all; subst;
    rewrite H; rewrite H0; rewrite H1; solve_union_y y.
    right.
    solve_union_y y.
Qed.

Instance TypFvSealCases: FvSealCases TypAbstractSyntax TypSealable.
Proof.
  hnf; intros.
  eapply fv_seal_cases_typdec; eauto.
Qed.

Lemma open_twice_inv_typdec:
  (forall (t: typ) m n y1 y2, OpenTwiceInvOf m n y1 y2 t) /\
  (forall (d: dec) m n y1 y2, OpenTwiceInvOf m n y1 y2 d).
Proof.
  apply typ_mutind; intros; simpls; eauto.
  - inversions H1. erewrite H; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H0. erewrite open_twice_inv; eauto. erewrite open_twice_inv; eauto.
  - inversions H1. erewrite H. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0. eauto.  eapply ne_n_S; eauto.  eauto.
  - inversions H2. erewrite H. 2:  eapply ne_n_S; eauto.  2: eauto. erewrite H0. eauto. eapply ne_n_S; eauto. eauto.
  - inversions H2. erewrite H; eauto. erewrite H0; eauto.
  - inversions H3. erewrite H. 2: eapply ne_n_S; eauto. 2: eauto. erewrite H0; eauto. erewrite H1. 2: eapply ne_n_S; eapply ne_n_S; eauto. 2: eauto. eauto.
Qed.

Instance TypOpenTwiceInv : OpenTwiceInv TypAbstractSyntax.
Proof.
  hnf; intros.
  eapply open_twice_inv_typdec; eauto.
Qed.


(* Helper lemmata about free variables *)

(* A variable not occurring in bnd type does not occur in the type opened by another variable *)
Lemma notin_fv_open_bnd x y U:
  x <> y ->
  y \notin fv (typ_bnd U) ->
  y \notin fv (open x U).
Proof.
  intros.
  destruct (fv_open_cases x U 0) as [Hr | Hr]; rewrite Hr; eauto.
Qed.

Lemma fv_rcd_typ B T U:
  fv (typ_rcd (dec_typ B T U)) = fv T \u fv U.
Proof.
  unfold fv, TypFreeVar.
  eapply fset_extens.
  intro. intro.
  compute in *. eauto.
  intro. intro.
  compute in *. eauto.

Qed.


(* Makes it possible to unfold fv T *)
Lemma fv_typ_eq T:
  (fv T) = (fv_typ T).
Proof.
  eauto.
Qed.

Lemma open_rec_vars_sel_f n v B v2 xs:
  open_rec_vars n xs (typ_sel (avar_f v) B (avar_f v2)) = (typ_sel (avar_f v) B (avar_f v2)).
Proof.
  revert dependent n.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_sel n v B v2 xs:
  open_rec_vars n xs (typ_sel v B v2) = (typ_sel (open_rec_vars n xs v) B (open_rec_vars n xs v2)).
Proof.
  revert dependent n.
  revert dependent v.
  revert dependent v2.
  induction xs; simpls; eauto.
Qed.


Lemma open_rec_vars_and n xs A B:
  open_rec_vars n xs (typ_and A B) = (typ_and (open_rec_vars n xs A) (open_rec_vars n xs B)).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_or n xs A B:
  open_rec_vars n xs (typ_or A B) = (typ_or (open_rec_vars n xs A) (open_rec_vars n xs B)).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_fun n xs A B:
  open_rec_vars n xs (typ_fun A B) = (typ_fun (open_rec_vars n xs A) (open_rec_vars (S n) xs B)).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  induction xs; simpls; eauto.
Qed.


Lemma open_rec_vars_fld n xs A B a:
  open_rec_vars n xs (typ_rcd (dec_fld a A B)) = (typ_rcd (dec_fld a (open_rec_vars n xs A) (open_rec_vars n xs B))).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_typ n xs A B a:
  open_rec_vars n xs (typ_rcd (dec_typ a A B)) = (typ_rcd (dec_typ a (open_rec_vars (S n) xs A) (open_rec_vars (S n) xs B))).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_met n xs A B C a:
  open_rec_vars n xs (typ_rcd (dec_met a A B C)) = (typ_rcd (dec_met a (open_rec_vars (S n) xs A) (open_rec_vars n xs B) (open_rec_vars (S (S n)) xs C))).
Proof.
  revert dependent n.
  revert dependent A.
  revert dependent B.
  revert dependent C.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_bnd n xs A:
  open_rec_vars n xs (typ_bnd A) = (typ_bnd (open_rec_vars (S n) xs A)).
Proof.
  revert dependent n.
  revert dependent A.
  induction xs; simpls; eauto.
Qed.

Lemma open_rec_vars_top n xs:
  open_rec_vars n xs (typ_top) = typ_top.
Proof.
  revert dependent n.
  induction xs; simpls; eauto.
Qed.
Lemma open_rec_vars_bot n xs:
  open_rec_vars n xs (typ_bot) = typ_bot.
Proof.
  revert dependent n.
  induction xs; simpls; eauto.
Qed.
Lemma open_rec_vars_nocap n xs:
  open_rec_vars n xs (typ_nocap) = typ_nocap.
Proof.
  revert dependent n.
  induction xs; simpls; eauto.
Qed.

Hint Rewrite
  open_rec_vars_sel_f
  open_rec_vars_bnd
  open_rec_vars_sel
  open_rec_vars_and
  open_rec_vars_or
  open_rec_vars_fun
  open_rec_vars_fld
  open_rec_vars_typ
  open_rec_vars_met
  open_rec_vars_top
  open_rec_vars_bot
  open_rec_vars_nocap : rewrite_open.

Lemma typ_bot_open_inv T n x:
  (open_rec n x T) = typ_bot ->
  T = typ_bot.
Proof.
  intros.
  destruct T; simpls; inversions H; eauto.
Qed.

Lemma typ_and_open_inv T1 T2 T3 x n:
  typ_and T1 T2 = open_rec n x T3 ->
  exists T4 T5,
  T3 = typ_and T4 T5 /\
  T1 = open_rec n x T4 /\
  T2 = open_rec n x T5.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  eauto.
Qed.
Lemma typ_or_open_inv T1 T2 T3 x n:
  typ_or T1 T2 = open_rec n x T3 ->
  exists T4 T5,
  T3 = typ_or T4 T5 /\
  T1 = open_rec n x T4 /\
  T2 = open_rec n x T5.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  eauto.
Qed.
Lemma typ_fun_open_inv T1 T2 T3 x n:
  typ_fun T1 T2 = open_rec n x T3 ->
  exists T4 T5,
  T3 = typ_fun T4 T5 /\
  T1 = open_rec n x T4 /\
  T2 = open_rec (S n) x T5.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  eauto.
Qed.
Lemma typ_fld_open_inv T1 T2 T3 a x n:
  (typ_rcd (dec_fld a T1 T2)) = open_rec n x T3 ->
  exists T4 T5,
  T3 = (typ_rcd (dec_fld a T4 T5)) /\
  T1 = open_rec n x T4 /\
  T2 = open_rec n x T5.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  destruct d; simpls; inversions H1.
  eauto.
Qed.

Lemma typ_typ_open_inv T1 T2 T3 a x n:
  (typ_rcd (dec_typ a T1 T2)) = open_rec n x T3 ->
  exists T4 T5,
  T3 = (typ_rcd (dec_typ a T4 T5)) /\
  T1 = open_rec (S n) x T4 /\
  T2 = open_rec (S n) x T5.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  destruct d; simpls; inversions H1.
  eauto.
Qed.

Lemma typ_met_open_inv T1 T2 T3 TC1 a x n:
  (typ_rcd (dec_met a T1 T2 TC1)) = open_rec n x T3 ->
  exists T4 T5 TC2,
  T3 = (typ_rcd (dec_met a T4 T5 TC2)) /\
  T1 = open_rec (S n) x T4 /\
  T2 = open_rec n x T5 /\
  TC1 = open_rec (S (S n)) x TC2.
Proof.
  intros.
  destruct T3; simpls; inversions H.
  destruct d; simpls; inversions H1.
  exists t t0 t1.
  eauto.
Qed.

Lemma typ_and_open_vars_inv T1 T2 T3 xs n:
  typ_and T1 T2 = open_rec_vars n xs T3 ->
  exists T4 T5,
  T3 = typ_and T4 T5 /\
  T1 = open_rec_vars n xs T4 /\
  T2 = open_rec_vars n xs T5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 HT3 HT1 HT2: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_and_open_inv in HT3.
    lets T0 T1 HT33 HT4 HT5: HT3.
    subst.
    exists T0 T1.
    auto.
Qed.
Lemma typ_or_open_vars_inv T1 T2 T3 xs n:
  typ_or T1 T2 = open_rec_vars n xs T3 ->
  exists T4 T5,
  T3 = typ_or T4 T5 /\
  T1 = open_rec_vars n xs T4 /\
  T2 = open_rec_vars n xs T5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 HT3 HT1 HT2: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_or_open_inv in HT3.
    lets T0 T1 HT33 HT4 HT5: HT3.
    subst.
    exists T0 T1.
    auto.
Qed.
Lemma typ_fun_open_vars_inv T1 T2 T3 xs n:
  typ_fun T1 T2 = open_rec_vars n xs T3 ->
  exists T4 T5,
  T3 = typ_fun T4 T5 /\
  T1 = open_rec_vars n xs T4 /\
  T2 = open_rec_vars (S n) xs T5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 HT3 HT1 HT2: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_fun_open_inv in HT3.
    lets T0 T1 HT33 HT4 HT5: HT3.
    subst.
    exists T0 T1.
    auto.
Qed.

Lemma typ_fld_open_vars_inv T1 T2 T3 a xs n:
  (typ_rcd (dec_fld a T1 T2)) = open_rec_vars n xs T3 ->
  exists T4 T5,
  T3 = (typ_rcd (dec_fld a T4 T5)) /\
  T1 = open_rec_vars n xs T4 /\
  T2 = open_rec_vars n xs T5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 HT3 HT1 HT2: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_fld_open_inv in HT3.
    lets T0 T1 HT33 HT4 HT5: HT3.
    subst.
    exists T0 T1.
    auto.
Qed.


Lemma typ_typ_open_vars_inv T1 T2 T3 a xs n:
  (typ_rcd (dec_typ a T1 T2)) = open_rec_vars n xs T3 ->
  exists T4 T5,
  T3 = (typ_rcd (dec_typ a T4 T5)) /\
  T1 = open_rec_vars (S n) xs T4 /\
  T2 = open_rec_vars (S n) xs T5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 HT3 HT1 HT2: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_typ_open_inv in HT3.
    lets T0 T1 HT33 HT4 HT5: HT3.
    subst.
    exists T0 T1.
    auto.
Qed.

Lemma typ_top_open_vars_inv T3 xs n:
  typ_top = open_rec_vars n xs T3 ->
  T3 = typ_top.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. auto.
  - simpls.
    specialize (IHxs _ _ H).
    destruct T3; try discriminate. auto.
Qed.


Lemma typ_bot_open_vars_inv T3 xs n:
  typ_bot = open_rec_vars n xs T3 ->
  T3 = typ_bot.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. auto.
  - simpls.
    specialize (IHxs _ _ H).
    destruct T3; try discriminate. auto.
Qed.


Lemma typ_nocap_open_vars_inv T3 xs n:
  typ_nocap = open_rec_vars n xs T3 ->
  T3 = typ_nocap.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. auto.
  - simpls.
    specialize (IHxs _ _ H).
    destruct T3; try discriminate. auto.
Qed.

Lemma typ_rec_open_vars_inv T1 T3 xs n:
  typ_bnd T1 = open_rec_vars n xs T3 ->
  exists T4,
  T3 = typ_bnd T4 /\
  T1 = open_rec_vars (S n) xs T4.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 HT3 HT1: IHxs.
    subst.
    symmetry in HT3.
    destruct T3; try discriminate. simpls.
    inversions HT3.
    exists T3.
    split; eauto.
Qed.


Lemma typ_sel_open_vars_inv x1 x2 T3 A xs n:
  (typ_sel x1 A x2) = open_rec_vars n xs T3 ->
  exists x4 x5,
  T3 = (typ_sel x4 A x5) /\
  x1 = open_rec_vars n xs x4 /\
  x2 = open_rec_vars n xs x5.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists x1 x2; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets x4 x5 HT3 Hx1 Hx2: IHxs.
    subst.
    destruct T3; try discriminate. simpls.
    inversions HT3.
    eauto.
Qed.


Lemma typ_met_open_vars_inv T1 T2 T3 TC1 a xs n:
  (typ_rcd (dec_met a T1 T2 TC1)) = open_rec_vars n xs T3 ->
  exists T4 T5 TC2,
  T3 = (typ_rcd (dec_met a T4 T5 TC2)) /\
  T1 = open_rec_vars (S n) xs T4 /\
  T2 = open_rec_vars n xs T5 /\
  TC1 = open_rec_vars (S (S n)) xs TC2.
Proof.
  intros.
  revert dependent n.
  revert dependent T3.
  induction xs; intros.
  - simpls. subst. exists T1 T2 TC1; auto.
  - simpls.
    specialize (IHxs _ _ H).
    lets T4 T5 TC4 HT3 [HT1 [HT2 HTC]]: IHxs.
    subst.
    symmetry in HT3.
    eapply typ_met_open_inv in HT3.
    lets T0 T1 TC5 HT33 [HT4 [HT5 HTC2]]: HT3.
    subst.
    exists T0 T1 TC5.
    auto.
Qed.

Ltac invert_open := (
  match goal with
  | H: typ_top = open_rec_vars _ _ ?T |- _ => eapply typ_top_open_vars_inv in H; subst
  | H: typ_bot = open_rec_vars _ _ ?T |- _ => eapply typ_bot_open_vars_inv in H; subst
  | H: typ_nocap = open_rec_vars _ _ ?T |- _ => eapply typ_nocap_open_vars_inv in H; subst
  | H: (typ_bnd _) = open_rec_vars _ _ ?T |- _ => eapply typ_rec_open_vars_inv in H; destructc H as ? ? ?; subst
  | H: (typ_and _ _) = open_rec_vars _ _ ?T |- _ => eapply typ_and_open_vars_inv in H; destructc H as ? ? ? ? ?; subst
  | H: (typ_or _ _) = open_rec_vars _ _ ?T |- _ => eapply typ_or_open_vars_inv in H; destructc H as ? ? ? ? ?; subst
  | H: (typ_fun _ _) = open_rec_vars _ _ ?T |- _ => eapply typ_fun_open_vars_inv in H; destructc H as ? ? ? ? ?; subst
  | H: (typ_rcd (dec_fld _ _ _)) = open_rec_vars _ _ ?T |- _ => eapply typ_fld_open_vars_inv in H; destructc H as ? ? ? ? ?; subst
  | H: (typ_rcd (dec_typ _ _ _)) = open_rec_vars _ _ ?T |- _ => eapply typ_typ_open_vars_inv in H; destructc H as ?T ?T ?HT ?HT ?HT; subst
  | H: (typ_rcd (dec_met _ _ _ _)) = open_rec_vars _ _ ?T |- _ => eapply typ_met_open_vars_inv in H; destructc H as ?T ?T ?T ?HT ?HT ?HT ?HT; subst
  | H: (typ_sel _ _ _) = open_rec_vars _ _ ?T |- _ => eapply typ_sel_open_vars_inv in H; destructc H as ? ? ? ? ?; subst
  end
).

(* Types with similar syntactic properties *)

Inductive typ_nulary: typ -> Prop :=
| typ_nulary_top: typ_nulary typ_top
| typ_nulary_bot: typ_nulary typ_bot
| typ_nulary_nocap: typ_nulary typ_nocap
.
Hint Constructors typ_nulary : core.

Inductive typ_andor: (typ -> typ -> typ) -> Prop :=
| typ_andor_and: typ_andor typ_and
| typ_andor_or: typ_andor typ_or
.
Hint Constructors typ_andor : core.

Notation typ_binary_rcd_typ_ctor A :=
  (fun t1 t2 => typ_rcd (dec_typ A t1 t2)).
Notation typ_binary_rcd_fld_ctor A :=
  (fun t1 t2 => typ_rcd (dec_fld A t1 t2)).

Inductive typ_binary: (typ -> typ -> typ) -> Prop :=
| typ_binary_and: typ_binary typ_and
| typ_binary_or: typ_binary typ_or
| typ_binary_fun: typ_binary typ_fun
| typ_binary_rcd_typ(A: typ_label): typ_binary (typ_binary_rcd_typ_ctor A)
| typ_binary_rcd_fld(A: fld_label): typ_binary (typ_binary_rcd_fld_ctor A)
.
Hint Constructors typ_binary : core.

(* Handles inversions where tyq_typ introduces typ_binary and typ_nulary *)
Ltac inversions_tyq_core :=
 try match goal with
 |  H: typ_nulary _ |- _ => solve [inversion H]
 |  H: typ_andor ?T, Hb: (?T _ _) = _ |- _ => inversions H; inversions Hb
 |  H: typ_binary ?T, Hb: (?T _ _) = _ |- _ => inversions H; inversions Hb
 end.
Ltac inversions_tyq H := inversions H; inversions_tyq_core.

Lemma typ_andor_binary T:
  typ_andor T ->
 typ_binary T.
Proof.
  introv Hbi.
  inversions Hbi; eauto.
Qed.

Lemma typ_binary_inv T T2:
  typ_binary T ->
  typ_binary T2 ->
  forall t1 t2 t3 t4,
    (T t1 t2) = (T2 t3 t4) ->
    T = T2 /\ t1 = t3 /\ t2 = t4.
Proof.
  intros.
  inversion H;
  inversion H0; subst;
  simplify_eq H1; intros; subst; auto.
Qed.

Lemma typ_binary_inv_ctor T T2:
  typ_binary T ->
  typ_binary T2 ->
  forall t1 t2 t3 t4,
    (T t1 t2) = (T2 t3 t4) ->
    forall t5 t6,
      (T t5 t6) = (T2 t5 t6).
Proof.
  intros.
  inversion H; inversion H0; subst; simplify_eq H1; auto.
  intros; subst. auto.
  intros; subst. auto.
Qed.

(* Not-or types are all types that are not typ_or in the root *)
Notation typ_notor B := (forall T U, B <> typ_or T U) (only parsing).
Notation typ_notand T := (forall S1 S2, T <> typ_and S1 S2) (only parsing).
Definition typ_notlog T := typ_notor T /\ typ_notand T.


Ltac close_typ_solve := (
  unfold close_rec, TransformClosing, TypTransform;
  simpl;
  rewrite ? rewrite_transform_typ, <- ? close_eq_close_depth; (* Folds transform_typ into transform *)
  eauto using f_equal, f_equal2, f_equal3, transform_close_depth_1, transform_close_depth_2
).

Lemma typ_close_top n x:
  close_rec n x typ_top = typ_top.
Proof.
  close_typ_solve.
Qed.
Lemma typ_close_bot n x:
  close_rec n x typ_bot = typ_bot.
Proof.
  close_typ_solve.
Qed.
Lemma typ_close_nocap n x:
  close_rec n x typ_nocap = typ_nocap.
Proof.
  close_typ_solve.
Qed.

Lemma typ_close_and n x T1 T2:
  close_rec n x (typ_and T1 T2) = typ_and (close_rec n x T1) (close_rec n x T2).
Proof.
  close_typ_solve.
Qed.

Lemma typ_close_or n x T1 T2:
  close_rec n x (typ_or T1 T2) = typ_or (close_rec n x T1) (close_rec n x T2).
Proof.
  close_typ_solve.
Qed.

Lemma typ_close_rcd n x D:
  close_rec n x (typ_rcd D) = (typ_rcd (close_rec n x D)).
Proof.
  close_typ_solve.
Qed.

Lemma typ_close_bnd n x T1:
  close_rec n x (typ_bnd T1) = typ_bnd (close_rec (S n) x T1).
Proof.
  close_typ_solve.
Qed.


Lemma typ_close_fun n x T1 T2:
  close_rec n x (typ_fun T1 T2) = typ_fun (close_rec n x T1) (close_rec (S n) x T2).
Proof.
  close_typ_solve.
Qed.


Lemma typ_close_sel n x y B z:
  close_rec n x (typ_sel y B z) = (typ_sel (close_rec n x y) B (close_rec n x z)).
Proof.
  close_typ_solve.
Qed.

Lemma dec_close_fld n x A T1 T2:
  close_rec n x (dec_fld A T1 T2) = (dec_fld A (close_rec n x T1) (close_rec n x T2)).
Proof.
  close_typ_solve.
Qed.
Lemma dec_close_met n x T1 T2 T3 A:
  close_rec n x (dec_met A T1 T2 T3) = (dec_met A (close_rec (S n) x T1) (close_rec n x T2) (close_rec (S (S n)) x T3)).
Proof.
  close_typ_solve.
Qed.
Lemma dec_close_typ n x T1 T2 A:
  close_rec n x (dec_typ A T1 T2) = (dec_typ A (close_rec (S n) x T1) (close_rec (S n) x T2)).
Proof.
  close_typ_solve.
Qed.

Hint Rewrite
  typ_close_fun
  typ_close_rcd
  typ_close_sel
  typ_close_bnd
  typ_close_top
  typ_close_bot
  typ_close_nocap
  typ_close_and
  typ_close_or
  dec_close_typ
  dec_close_fld
  dec_close_met
  : simpl_close.


(* Solves simplification of subst_vars *)
Local Ltac solve_subst_vars_dist :=
  match goal with
    | |- (?T /[ ?xs --> ?ys]) = _ => (* Finds the substitution from xs to ys in the goal *)
      reverts ys;
      induction xs; (* induction on source list independent of target list *)
      intros;
      destruct ys; (* step in the target list *)
      simpl subst_vars; (* unfolds one substitution *)
      match goal with
        | IHxs: forall ys0, (_ /[ _ --> _ ] = _) |- _ => (* finds the inductive hypothesis *)
           rewrite IHxs (* rewrites the rest by induction *)
        | |- _ => idtac
      end;
      simple apply (eq_refl) (* not both sides are equal *)
  end.


Lemma typ_subst_nulary T (HT: typ_nulary T) x y:
  T /[x -> y] = T.
Proof.
  inversions HT; simpl; auto.
Qed.
Lemma typ_subst_vars_nulary T (HT: typ_nulary T) xs ys:
  T /[xs --> ys] = T.
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. eapply typ_subst_nulary. auto.
Qed.

Lemma typ_subst_pvars_nulary T (HT: typ_nulary T) xs ys:
  (T //[xs --> ys]) = T.
Proof.
  edestruct (subst_pvars_unfold T) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  rewrite (typ_subst_vars_nulary HT xs zs); auto.
  rewrite typ_subst_vars_nulary; auto.
Qed.

Lemma typ_subst_andor T (HT: typ_andor T) x y T1 T2:
  (T T1 T2) /[x -> y] = (T (T1 /[x -> y]) (T2 /[x -> y])).
Proof.
  inversions HT; simpl; auto.
Qed.
Lemma typ_subst_vars_andor T (HT: typ_andor T) xs ys T1 T2:
  (T T1 T2) /[xs --> ys] = (T (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. eapply typ_subst_andor. auto.
Qed.

Lemma typ_fv_andor T (HT: typ_andor T) T1 T2:
  fv (T T1 T2) = fv T1 \u fv T2.
Proof.
  inversions HT; simpls; auto.
Qed.

Lemma typ_subst_pvars_andor T (HT: typ_andor T) xs ys T1 T2:
  (T T1 T2) //[xs --> ys] = (T (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.

  edestruct (subst_pvars_unfold (T T1 T2)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (T1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  edestruct (subst_pvars_unfold (T2)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst1.

  rewrite typ_fv_andor in Hfr.

  rewrite (typ_subst_vars_andor HT xs zs); auto.
  rewrite typ_subst_vars_andor; auto.
  simpls.
  rewrite subst_vars_fresh_middle_change with (ws:=zs0); eauto.
  rewrite subst_vars_fresh_middle_change with (zs2:=zs) (ws:=zs1); eauto.
  auto.
Qed.


Lemma typ_subst_vars_fun T1 T2 xs ys:
  (typ_fun T1 T2) /[xs --> ys] = (typ_fun (T1 /[xs --> ys]) (T2 /[xs --> ys])).
Proof.
  solve_subst_vars_dist.
Qed.

(* Finds and unfolds parallel a substitution in the goal.
   Adds a fresh list to the hypotheses. *)
Ltac subst_pvars_find_unfold :=
  match goal with
    | |- context[(?T //[ ?xs --> ?ys])] => (* Finds any parallel substitution from xs to ys in the goal *)
        edestruct (subst_pvars_unfold T) as [?zs [?Hfr ?Hsubst]];
        rewrite Hsubst; clear Hsubst
  end.

(* Finds unfolded parallel substituitons which use a different list an merges them to use a single list.
   One of the lists (the primary) must be at least as fresh as the other (secondary) list.
   The secondary list is cleared.
 *)
Ltac subst_pvars_unfolded_merge :=
  match goal with
    | H1: fresh _ _ ?zs, (* Primary fresh list *)
      H2: fresh _ _ ?zs0 (* Secondary fresh list *)
      |- context[((?T /[ ?xs --> ?zs]) /[?zs --> ?ys])] => (* Finds unfolded parallel substitution from xs to ys through the secondary list in the goal *)
      rewrite (subst_vars_fresh_middle_change zs0 zs) by fresh_solve; (* Replaces the primary list by the secondary *)
      clear H2 zs0 (* Clears the secondary list *)
  end.

Ltac solve_subst_pvars_dist fold_tactic subs_vars_lemma :=
  repeat subst_pvars_find_unfold; (* Unfold parallel substituions *)
  simpl fv in *; (* Simplify free variable set in freshness *)
  fold_tactic; (* Put free variable set in the usual form *)
  repeat rewrite subs_vars_lemma; auto; (* Distributes substitution in the syntax *)
  repeat subst_pvars_unfolded_merge; (* Merge lists of intermediate variables into the primary one *)
  auto.

Lemma typ_subst_pvars_fun T1 T2 xs ys:
  (typ_fun T1 T2) //[xs --> ys] = (typ_fun (T1 //[xs --> ys]) (T2 //[xs --> ys])).
Proof.
  solve_subst_pvars_dist folds_typ_fv typ_subst_vars_fun.
Qed.

Lemma typ_subst_vars_bnd T1 xs ys:
  (typ_bnd T1) /[xs --> ys] = (typ_bnd (T1 /[xs --> ys])).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma typ_subst_pvars_bnd T1 xs ys:
  (typ_bnd T1) //[xs --> ys] = (typ_bnd (T1 //[xs --> ys])).
Proof.
  solve_subst_pvars_dist folds_typ_fv typ_subst_vars_bnd.
Qed.

Lemma typ_subst_vars_sel T1 B T2 xs ys:
  (typ_sel T1 B T2) /[xs --> ys] = (typ_sel (T1 /[xs --> ys]) B (T2 /[xs --> ys])).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma typ_subst_pvars_sel T1 B T2 xs ys:
  (typ_sel T1 B T2) //[xs --> ys] = (typ_sel (T1 //[xs --> ys]) B (T2 //[xs --> ys])).
Proof.
  solve_subst_pvars_dist folds_typ_fv typ_subst_vars_sel.
Qed.

Lemma dec_subst_vars_fld a T1 T2 xs ys:
  ( (dec_fld a T1 T2)) /[xs --> ys] = ( (dec_fld a (T1 /[xs --> ys]) (T2 /[xs --> ys]))).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma dec_subst_pvars_fld a T1 T2 xs ys:
  ( (dec_fld a  T1 T2)) //[xs --> ys] = ( (dec_fld a (T1 //[xs --> ys]) (T2 //[xs --> ys]))).
Proof.
  solve_subst_pvars_dist folds_typ_fv dec_subst_vars_fld.
Qed.

Lemma dec_subst_vars_typ a T1 T2 xs ys:
  ( (dec_typ a T1 T2)) /[xs --> ys] = ( (dec_typ a (T1 /[xs --> ys]) (T2 /[xs --> ys]))).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma dec_subst_pvars_typ a T1 T2 xs ys:
  ( (dec_typ a  T1 T2)) //[xs --> ys] = ( (dec_typ a (T1 //[xs --> ys]) (T2 //[xs --> ys]))).
Proof.
  solve_subst_pvars_dist folds_typ_fv dec_subst_vars_typ.
Qed.

Lemma dec_subst_vars_met a T1 T2 T3 xs ys:
  ( (dec_met a T1 T2 T3)) /[xs --> ys] = ( (dec_met a (T1 /[xs --> ys]) (T2 /[xs --> ys]) (T3 /[xs --> ys]))).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma dec_subst_pvars_met a T1 T2 T3 xs ys:
  ( (dec_met a  T1 T2 T3)) //[xs --> ys] = ( (dec_met a (T1 //[xs --> ys]) (T2 //[xs --> ys]) (T3 //[xs --> ys]))).
Proof.
  solve_subst_pvars_dist folds_typ_fv dec_subst_vars_met.
Qed.

Lemma typ_subst_vars_rcd D xs ys:
  (typ_rcd D) /[xs --> ys] = (typ_rcd (D /[xs --> ys]) ).
Proof.
  solve_subst_vars_dist.
Qed.

Lemma typ_subst_pvars_rcd D xs ys:
  (typ_rcd D) //[xs --> ys] = (typ_rcd (D //[xs --> ys])).
Proof.
  solve_subst_pvars_dist folds_typ_fv typ_subst_vars_rcd.
Qed.

Lemma subst_open_commut_typ : forall (T : typ) (x y z : var) (n : nat),
    (open_rec n z T) /[ x -> y] = open_rec n (z /[ x -> y]) (T /[ x -> y]).
Proof.
  intros.
  rewrite subst_open_commut. eapply f_equal. auto.
Qed.

Hint Rewrite
  typ_subst_vars_fun
  typ_subst_vars_rcd
  typ_subst_vars_sel
  typ_subst_vars_bnd
  (typ_subst_vars_nulary typ_nulary_top)
  (typ_subst_vars_nulary typ_nulary_bot)
  (typ_subst_vars_nulary typ_nulary_nocap)
  (typ_subst_vars_andor typ_andor_and)
  (typ_subst_vars_andor typ_andor_or)
  dec_subst_vars_typ
  dec_subst_vars_fld
  dec_subst_vars_met
  : simpl_subst_vars.

Lemma subst_vars_eq_inv_typdec:
  (forall (t: typ) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 t) /\
  (forall (d: dec) xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 d).
Proof.
  apply typ_mutind; intros; simpls; autorewrite with simpl_subst_vars in *.
  (* Solve simple cases where there are no free variables *)
  all: try match goal with
  | H: from_list ?xs \c \{} |- _ =>
    eapply subset_empty_inv in H;
    eapply from_list_empty_inv in H;
    subst;
    destruct ys2; inversions H1;
    destruct ys1; inversions H0;
    auto
  end.
  Ltac invert_eq T := match goal with H: @eq T _ _ |- _ => inversions H end.
  all: first [invert_eq typ| invert_eq dec].
  all: try solve [eapply H; eauto].
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_step a) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_step a0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H0; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H0, H1. auto.
      
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t0) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t1)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    forwards Hci: list_cover_interval_2 Hsrt1 Hsrt5. eapply fresh_noduplicates; eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H2; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.
    rewrite H2, H3. auto.
  - 
    forwards Hsrt: (subst_vars_eq_inv_ind_step t) xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt0: (subst_vars_eq_inv_ind_step t0)  xs ys1 ys2; try assumption; fresh_solve.
    forwards Hsrt1: (subst_vars_eq_inv_ind_step t1)  xs ys1 ys2; try assumption; fresh_solve.
    destructecs Hsrt.
    destructecs Hsrt0.
    destructecs Hsrt1.
    forwards Hci: list_cover_interval_3 Hsrt2 Hsrt6 Hsrt10. eapply fresh_noduplicates; eauto. eauto. eauto. eauto. eauto.
    eapply list_cover_interval_reorder_eq. 3: rewrite H3; eauto. eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.  eapply list_reorder_app; eauto.   eapply list_reorder_app; eauto.
    rewrite H3, H4. auto.
Qed.

Instance TypSubstVarsEqInv: SubstVarsEqInv TypAbstractSyntax.
Proof.
  hnf. intros. eapply (proj21 subst_vars_eq_inv_typdec); eauto.
Qed.
Instance DecSubstVarsEqInv: SubstVarsEqInv DecAbstractSyntax.
Proof.
  hnf. intros. eapply (proj22 subst_vars_eq_inv_typdec); eauto.
Qed.

