(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

(* ************************************************************************** *)
(** * Locally Nameless Variables
    The proof represents variables using the
    #<a href="http://www.chargueraud.org/softs/ln/">locally nameless representation</a>#:
    - [avar_b n] represents a variable using the de Bruijn index [n];
    - [avar_f x] represents a free variable with name [x].
    de Bruijn-indexed variables represent bound variables, whereas named variables represent free variables
    that are in the evaluation context/type environment.  *)

Inductive avar : Set :=
  | avar_b : nat -> avar
  | avar_f : var -> avar.
Hint Constructors avar : core.
Notation avars := (list avar).

Instance VarSubstVar : SubstVar var :=
  fun x y z => If z = x then y else z.

Instance VarSealable: Sealable var :=
  fun n x y => y.

Instance VarOpenable : Openable var :=
  fun n x a => a.

Instance VarFreeVar : FreeVar var := singleton (A:=var).

Instance VarAbstractSyntax :
  AbstractSyntax VarOpenable VarFreeVar VarSubstVar := {}.

Lemma seal_var n (x y: var):
  seal_rec n x y = y.
Proof.
  unfold seal_rec, VarSealable. auto.
Qed.

Instance VarSubstId : SubstId VarAbstractSyntax.
Proof.
  hnf; intros.
  apply fun_ext_dep; intros a. unfold subst_var, VarSubstVar. cases_if; auto.
Qed.
Instance VarSubstCommut : SubstCommut VarAbstractSyntax.
Proof.
  hnf; intros.
  apply fun_ext_dep; intros a. unfold compose, subst_var, VarSubstVar; destruct H0;
  repeat cases_if; auto.
Qed.
Instance VarSubstFresh : SubstFresh VarAbstractSyntax.
Proof.
  hnf; intros.
  unfold subst_var, VarSubstVar, fv, VarFreeVar in *.
  repeat cases_if; auto.
  rewrite notin_singleton in H. contradiction.
Qed.
Instance VarFvSubstCases : FvSubstCases VarAbstractSyntax.
Proof.
  hnf; intros.
  unfold subst_var, VarSubstVar, fv, VarFreeVar in *.
  repeat cases_if; auto.
  rewrite in_singleton. rewrite notin_singleton.
  destruct (classicT(x=y)); subst; auto.
  rewrite remove_same.
  rewrite union_empty_l. auto.
Qed.

(* ************************************************************************** *)
(** ** Locally Nameless Variables are Abstract Syntax *)

Instance AvarOpenable : Openable avar :=
  fun n x a =>
    match a with
    | avar_b i => If n = i then avar_f x else avar_b i
    | avar_f y => avar_f y
    end.

Instance AvarFreeVar : FreeVar avar :=
  fun a =>
    match a with
    | avar_b i => \{}
    | avar_f x => \{x}
    end.

Instance AvarSubstVar : SubstVar avar :=
  fun (x: var) (y: var) (a: avar) =>
    match a with
    | avar_b i => avar_b i
    | avar_f z => (avar_f (If z = x then y else z))
    end.

Instance AvarAbstractSyntax :
  AbstractSyntax AvarOpenable AvarFreeVar AvarSubstVar := {}.

Instance AvarSealable : Sealable avar :=
  fun n x a =>
    match a with
    | avar_b i => If n <= i then avar_f x else avar_b i
    | avar_f y => avar_f y
    end.

Instance AvarDecideFv : DecideFv AvarAbstractSyntax :=
  fun x a =>
    match a with
    | avar_b i => false
    | avar_f y => (If x = y then true else false)
    end.

Instance AvarClosing : Closing AvarAbstractSyntax :=
  fun (n : nat) (x : var) (a : avar) =>
    match a with
    | avar_b i => avar_b i
    | avar_f y => (If y = x then avar_b n else avar_f y)
    end.

(* ************************************************************************** *)
(** ** Tactics for Locally Nameless Variables *)
Local Ltac unfold_avar_syntax :=
  unfold VarSubstVar, AvarOpenable, AvarFreeVar, AvarSubstVar, AvarSealable, AvarDecideFv, AvarClosing.

Local Ltac avar_unfold :=
  intros; unfold_syntax; unfold compose;
  unfold_avar_syntax; cbn in *.

Ltac destruct_avars :=
  repeat match goal with
         | [H : avar |- _ ] => destruct H
         end; cbn in *; repeat cases_if.

Ltac var_notin :=
  match goal with
  | [ H : ?x \notin \{ ?x} |- _ ] => exfalso; apply H; rewrite in_singleton; auto
  end.

Local Ltac unfold_rest :=
  unfold dec_apply, dec_notin_fv, ignore_second, id;
  cbn in *; repeat cases_if; cbn in *.

Local Ltac destruct_solve :=
  destruct_avars; unfold_rest; auto; try var_notin.

Local Ltac avar_solver :=
  avar_unfold; destruct_solve.

Local Ltac avar_fun_ext_solver :=
  avar_unfold;
  apply fun_ext_dep; intros a;
  destruct_solve.

Local Ltac avar_fun_ext_2_solver :=
  avar_unfold;
  apply fun_ext_dep; intros a;
  apply fun_ext_dep; intros m;
  destruct_solve.


(* ************************************************************************** *)
(** ** Locally Nameless Variables Satisfies Syntax Lemmas *)
Instance AvarReflectFv : ReflectFv AvarDecideFv.
Proof.
  hnf; avar_solver; constructor;
    try rewrite ? in_empty, ? in_singleton;
    intros; try congruence; contradiction.
Qed.

Instance AvarOpenClose : OpenClose AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarCloseOpen : CloseOpen AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarSubstOpenClose : SubstOpenClose AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.
Instance AvarCloseSubst : CloseSubst AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.
Instance AvarCloseCommut : CloseCommut AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.
Instance AvarOpenCloseCommut : OpenCloseCommut AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.
Instance AvarSubstCloseCommut : SubstCloseCommut AvarClosing.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarOpenTwice : OpenTwice AvarOpenable.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarOpenCommut : OpenCommut AvarOpenable.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarSubstFresh : SubstFresh AvarAbstractSyntax.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarSubstOpenCommut : SubstOpenCommut AvarAbstractSyntax.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarSubstCommut : SubstCommut AvarAbstractSyntax.
Proof.
  hnf; avar_fun_ext_solver.
  destruct H0; subst; contradiction.
Qed.

Instance AvarSubstId : SubstId AvarAbstractSyntax.
Proof.
  hnf; avar_fun_ext_solver.
Qed.

Instance AvarFvOpenCases : FvOpenCases AvarAbstractSyntax.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarSealOpen: SealOpen AvarAbstractSyntax AvarSealable.
Proof.
  split.
  - hnf; avar_fun_ext_solver.
    + eapply Le.le_Sn_le in C. contradiction.
    + false. apply C0. eapply Nat.le_refl.
    + false. eapply C. eapply Lt.le_lt_or_eq in C1. destruct C1; contradiction.
  - hnf; avar_fun_ext_solver.
    + false. eapply Lt.lt_not_le in H. contradiction.
Qed.
Instance AvarSealClose: SealClose AvarAbstractSyntax AvarSealable AvarClosing.
Proof.
  constructor.
  hnf; avar_fun_ext_solver.
  false. eapply Lt.lt_not_le in H. contradiction.
Qed.


Instance AvarSealClosed: SealClosed AvarAbstractSyntax AvarSealable.
Proof.
  hnf; avar_solver.
  - false. specialize (H n0 y C). simpls. cases_if.
Qed.

Instance AvarOpenSealInv: OpenSealInv AvarAbstractSyntax AvarSealable.
Proof.
  hnf; avar_solver.
  - false. eapply Lt.lt_not_le in H. contradiction.
Qed.

Instance AvarFvSealCases: FvSealCases AvarAbstractSyntax AvarSealable.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarOpenTwiceInv : OpenTwiceInv AvarAbstractSyntax.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarCloseUnbound : CloseUnbound AvarClosing.
Proof.
  hnf; avar_solver.
Qed.

Instance AvarClosingInner : ClosingInner AvarReflectFv AvarClosing.
Proof.
  hnf; avar_fun_ext_2_solver.
Qed.

Instance AvarSubstitutionInner : SubstitutionInner AvarReflectFv.
Proof.
  hnf; avar_fun_ext_2_solver.
Qed.

Instance AvarFvSubstCases: FvSubstCases AvarAbstractSyntax.
Proof.
  hnf. intros.
  unfold subst_var. unfold AvarSubstVar.
  destruct a.
  - simpl. auto.
  - simpl. cases_if.
    destruct (classicT (x = y)).
    + subst. eauto.
    + right. splits; eauto.
      rewrite in_singleton. auto.
      eapply fset_extens; hnf; intros.
      * rewrite in_singleton in H; subst.
        rewrite in_union. right. rewrite in_singleton. auto.
      * rewrite in_union in H. destruct H.
        rewrite in_remove in H. destruct H.
        contradiction.
        auto.
  + auto.
Qed.

Instance AvarFvClose: FvClose AvarClosing.
Proof.
  hnf; avar_solver.
  - symmetry. eapply remove_empty_r.
  - symmetry. eapply remove_same.
  - eapply fset_extens; intro; intro.
    rewrite in_singleton in H; subst.
    rewrite in_remove. rewrite in_singleton. rewrite notin_singleton. auto.
    rewrite in_remove in H. rewrite in_singleton in H. destruct H; subst. rewrite in_singleton. auto.
Qed.

(** The rest of the lemmas are automatically satisfied as a consequence of
    [AvarClosingUnbound] *)

Lemma open_rec_vars_avar_f n xs x:
  (avar_f x) /[#n --> xs] = (avar_f x).
Proof.
  revert dependent n.
  induction xs; simpls; eauto.
Qed.

Lemma avar_f_open_cases v n xs x:
  avar_f v = x /[#n -> xs] ->
  x = (avar_f v) \/ (x = (avar_b n) /\ xs = v).
Proof.
  introv Heq.
  destruct x. destruct (classicT (n=n0)).
  - subst. simpls. cases_if. inversions Heq. auto.
  - simpls. cases_if.
  - simpls. inversions Heq. auto.
Qed.

Lemma avar_b_open_vars_inv v xs x m:
  avar_b v = x /[#m --> xs] ->
  x= (avar_b v).
Proof.
  introv Hv.
  reverts m x.
  induction xs; intros; simpls.
  - auto.
  - forwards: IHxs; eauto. destruct x. destruct (classicT (n=v)). subst. auto. simpls. cases_if. simpls. discriminate.
Qed.

Lemma avar_b_open_inv_l y x n m:
  avar_b n = open_rec m y x ->
  n <> m.
Proof.
  introv Hn.
  intro. subst.
  destruct x.
  - simpls. cases_if.
  - simpls. discriminate.
Qed.

Lemma avar_b_open_vars_inv_l v xs x m:
  avar_b v = x /[#m --> xs] ->
  length xs + m <= v \/ v < m.
Proof.
  introv Hv.
  reverts m x.
  induction xs; intros; simpls.
  - destruct (PeanoNat.Nat.lt_ge_cases v m); auto.
  - forwards: IHxs; eauto. destruct H. rewrite Nat.add_succ_r in H.
    auto.
    apply -> Nat.lt_succ_r in H.
    apply -> Nat.le_lteq in H.
    destruct H; auto.
    subst.
    eapply avar_b_open_vars_inv in Hv.
    symmetry in Hv.
    eapply avar_b_open_inv_l in Hv.
    contradiction.
Qed.

Lemma avar_b_open_vars_inv_l0 v xs x:
  avar_b v = open_vars xs x ->
  length xs <= v.
Proof.
  introv Hv.
  forwards: avar_b_open_vars_inv_l; eauto.
  destruct H; simpls.
  rewrite Nat.add_0_r in H. eauto.
  eapply Nat.nlt_0_r in H. contradiction.
Qed.

Lemma avar_b_open_vars_add ys1 n m p:
  (n < length ys1) ->
  open_rec_vars m ys1 (avar_b (m + n)) = open_rec_vars p ys1 (avar_b (n + p)).
Proof.
  introv Hn.
  reverts m n p.
  induction ys1; intros.
  - simpls. false. eapply Nat.nlt_0_r. eauto.
  - simpls. cases_if. cases_if. rewrite open_rec_vars_avar_f. rewrite open_rec_vars_avar_f. auto.
    false.
    assert (n = 0).
    eapply Plus.plus_reg_l. rewrite Nat.add_0_r. eauto.
    subst. contradiction.
    cases_if.
    +
    assert ( 0 = n).
    eapply Plus.plus_reg_l. rewrite Nat.add_0_r. rewrite Plus.plus_comm. eauto.
    subst. rewrite Nat.add_0_r in C. contradiction.
    +
    assert (0<> n). intro. subst. rewrite Nat.add_0_r in C. contradiction.
    eassert _. eapply IHys1 with (m:=(S m)) (p:=(S p)) (n:=(Nat.pred n)).
    eapply Lt.lt_S_n. rewrite Nat.succ_pred; eauto.
    rewrite Plus.plus_Snm_nSm in X.
    rewrite Nat.succ_pred in X by eauto.
    rewrite <- Plus.plus_Snm_nSm in X.
    rewrite Nat.succ_pred in X by eauto.
    auto.
Qed.

Lemma avar_b_open_vars_addS ys1 n:
  (n < length ys1) ->
  open_rec_vars 1 ys1 (avar_b (S n)) = open_vars ys1 (avar_b n).
Proof.
  pose proof (@avar_b_open_vars_add ys1 n 1 0).
  simpls. rewrite Nat.add_0_r in H. auto.
Qed.

Lemma avar_b_open_vars_l xs n m:
  (length xs + m <= n) ->
  (avar_b n) = (avar_b n)/[#m --> xs].
Proof.
  introv Hnl.
  reverts n m.
  induction xs; intros; simpls.
  - auto.
  - cases_if.
    + false. rewrite <- Nat.add_succ_l in Hnl.
      assert ( S (length xs) + n <= n + 0). rewrite Nat.add_0_r. auto.
      rewrite Nat.add_comm in H.
      eapply Plus.plus_le_reg_l in H.
      eapply Nat.nle_succ_0 in H. contradiction.
    + eapply IHxs.
      rewrite Nat.add_succ_r. auto.
Qed.

Lemma avar_b_open_vars_l0 xs n:
  (length xs <= n) ->
  (avar_b n) = open_vars xs (avar_b n).
Proof.
  introv Hl.
  eapply avar_b_open_vars_l.
  rewrite Nat.add_0_r. auto.
Qed.

Lemma avar_f_open_vars_cases v xs x m:
  avar_f v = x /[#m --> xs] ->
  x = (avar_f v) \/ (
    exists n,
    x = (avar_b (m + n)) /\
    LibList.Nth n xs v /\
    (n < length xs)
  ).
Proof.
  introv Heq.
  reverts m.
  induction xs; intros; simpls.
  - auto.
  - rewrite open_rec_vars_open_rec_commut in Heq by auto.
    destruct (avar_f_open_cases _ _ _ Heq).
    + forwards: IHxs; eauto.
      destruct H0.
      * auto.
      * destructc H0 as n Hx Hv Hl. right. exists (S n).
        splits.
        ** rewrite Hx. eapply f_equal. eapply Plus.plus_Snm_nSm.
        ** eapply LibList.Nth_succ; auto.
        ** unfold lt in *. eapply Le.le_n_S. auto.
    + destruct H. subst. right. exists 0. splits.
      * eapply avar_b_open_vars_inv. symmetry. rewrite Nat.add_0_r. eauto.
      * eapply LibList.Nth_zero.
      * eapply Nat.lt_0_succ.
Qed.
Lemma avar_b_open_rec_vars ys n m:
  (n < length ys) ->
  exists v, (open_rec_vars m ys (avar_b (m + n))) = (avar_f v) /\ (LibList.Nth n ys v).
Proof.
  introv Hl.
  reverts ys.
  induction n; intros; simpls.
  - destruct ys. simpls. false. eapply Nat.nlt_0_r. eauto.
    exists v. split.
    pose proof (@avar_b_open_vars_add (v::ys) 0 m 0).
    rewrite H by auto.
    rewrite Nat.add_0_l.
    unfold open_rec_vars.  simpls. cases_if. eapply open_rec_vars_avar_f. apply LibList.Nth_zero.
  - destruct ys. simpls. false. eapply Nat.nlt_0_r. eauto.
    simpls. eapply Lt.lt_S_n in Hl.
    destructc (IHn ys Hl) as v' Hv Hn.
    exists v'.
    cases_if.
    + rewrite <- plus_n_Sm in C.
      rewrite Nat.add_comm in C.
      eapply Nat.succ_add_discr in C. contradiction.
    +
    split.
    rewrite <- Hv.
    pose proof (@avar_b_open_vars_add  ys n m 0) as Hmn.
    rewrite Hmn by auto.
    rewrite Nat.add_0_r.
    pose proof (@avar_b_open_vars_add  ys n (S m) 0) as Hmn1.
    rewrite <- Nat.add_succ_comm.
    rewrite Hmn1 by auto.
    rewrite Nat.add_0_r.
    auto.
    eapply LibList.Nth_succ. eauto.
Qed.
Lemma avar_b_open_vars ys n:
  (n < length ys) ->
  exists v, (open_vars ys (avar_b n)) = (avar_f v) /\ (LibList.Nth n ys v).
Proof.
  eapply avar_b_open_rec_vars.
Qed.


Lemma avar_f_reopen_vars_cases v xs x ys m:
  length ys = length xs ->
  avar_f v = open_rec_vars m xs x ->
  (
    x = (avar_f v) /\
    avar_f v = open_rec_vars m ys x
  ) \/ (
    exists n w,
      x = (avar_b (m + n)) /\
      LibList.Nth n xs v /\
      LibList.Nth n ys w /\
      (avar_f w = open_rec_vars m ys x)
  ).
Proof.
  introv Hl Heq.
  destruct (avar_f_open_vars_cases xs x m Heq); subst.
  - left. rewrite open_rec_vars_avar_f. auto.
  - right. destructc H as n Hb Hn Hnl. subst.
    pose proof Hnl.
    rewrite <- Hl in Hnl.
    eapply (avar_b_open_rec_vars ys m) in Hnl.
    destructc Hnl as w How Hny.
    exists n w.
    splits; eauto.
Qed.

Lemma open_rec_avar_cases n x a1 a2:
  open_rec n x a1 = open_rec n x a2 ->
  x \notin (fv a1 \u fv a2) ->
  (exists w, a1 = (avar_f w) /\ a2 = (avar_f w)) \/
  (exists m, a1 = (avar_b m) /\ a2 = (avar_b m)). (* m does not need to be n *)
Proof.
  introv Heq Hn.
  remember (open_rec n x a1).
  destruct a1; destruct a2; simpls.
  - cases_if; cases_if. right; exists n0; auto. subst. inversions Heq. right; exists n1; auto.
  - cases_if. subst. inversions Heq. autorewrite with sets in Hn. destruct Hn. contradiction.
  - cases_if. subst. inversions Heq. autorewrite with sets in Hn. destruct Hn. contradiction.
  - subst. inversions Heq. left. exists v0; auto.
Qed.

Lemma open_vars_avar_cases n x a1 a2:
  open_rec_vars n x a1 = open_rec_vars n x a2 ->
  fresh (fv a1 \u fv a2) (length x) x ->
  (exists w, a1 = (avar_f w) /\ a2 = (avar_f w)) \/
  (exists m, a1 = (avar_b m) /\ a2 = (avar_b m)). (* m does not need to be n *)
Proof.
  introv Heq Hn.
  remember (open_rec_vars n x a1).
  destruct a; simpls.
  - right; exists n0. split.
    + eapply avar_b_open_vars_inv in Heqa. auto.
    + eapply avar_b_open_vars_inv in Heq. auto.
  - destruct (avar_f_open_vars_cases x a1 n Heqa);
    destruct (avar_f_open_vars_cases x a2 n Heq).
    + eauto.
    + destruct_all; subst. simpls.
      eapply LibList.Nth_mem in H1.
      false. eapply fresh_mem_false with (l:=(length x)); eauto. fresh_solve.
    + destruct_all; subst. simpls.
      eapply LibList.Nth_mem in H1.
      false. eapply fresh_mem_false with (l:=(length x)); eauto. fresh_solve.
    + destruct_all. subst. right.
      exists (n+x1).
      assert (x0=x1). eapply fresh_nth_eq; eauto. eapply fresh_empty; eauto. subst. auto.
Qed.



Lemma reopen_vars_eq (v: var) n xs ys (x1 x2: avar):
  fresh (fv x1 \u fv x2) (length ys) xs ->
  avar_f v = open_rec_vars n xs x1 ->
  avar_f v = open_rec_vars n xs x2 ->
  exists w,
  avar_f w = open_rec_vars n ys x1 /\
  avar_f w = open_rec_vars n ys x2.
Proof.
  introv Hfr Hv1 Hv2.
  pose proof Hfr as Hl. eapply fresh_length in Hl. symmetry in Hl.

  destruct (avar_f_reopen_vars_cases xs x1 ys n Hl Hv1); subst;
  destruct (avar_f_reopen_vars_cases xs x2 ys n Hl Hv2); subst.
  - destruct H. destruct H0. exists v. subst. rewrite open_rec_vars_avar_f in *. auto.
  - destruct_all; subst. simpls.
    eapply LibList.Nth_mem in H1.
    false. eapply fresh_mem_false with (l:=(length ys)); eauto. fresh_solve.
  - destruct_all; subst. simpls.
    eapply LibList.Nth_mem in H2.
    false. eapply fresh_mem_false with (l:=(length ys)); eauto. fresh_solve.
  - destruct_all; subst. simpls.

    assert (x3 = x). {
      eapply fresh_nth_eq with (l:=length ys) (v:=v); eauto. fresh_solve.
    }
    subst.
    exists x4. eauto.
Qed.

Lemma reopen_vars_unique n xs ys (x1 x2: avar):
  fresh (fv x1 \u fv x2 ) (length ys) xs ->
  open_rec_vars n xs x1 = open_rec_vars n xs x2 ->
  open_rec_vars n ys x1 = open_rec_vars n ys x2.
Proof.
  introv Hfr Heq.
  pose proof Hfr as Hl. eapply fresh_length in Hl. symmetry in Hl.

  remember (open_rec_vars n xs x2).
  symmetry in Heq.
  destruct a.
  -
  pose proof Heq.
  eapply avar_b_open_vars_inv_l in H.
  eapply avar_b_open_vars_inv in Heq.
  eapply avar_b_open_vars_inv in Heqa.
  subst.
  destruct H.
  rewrite <- avar_b_open_vars_l. auto.
  rewrite Hl. auto.
  eauto.
  - forwards w Hw1 Hw2: reopen_vars_eq; eauto.
    rewrite <- Hw1. rewrite <- Hw2.
    auto.
Qed.

Lemma avar_close_b n x m:
  (avar_b m) /[#n <- x] = avar_b m.
Proof.
  unfold close_rec, AvarClosing.
  auto.
Qed.

Lemma avar_close_f n x y:
  (avar_f y) /[#n <- x] = (If y = x then avar_b n else avar_f y).
Proof.
  unfold close_rec, AvarClosing.
  auto.
Qed.

Hint Rewrite
  avar_close_b
  avar_close_f: simpl_close.

Lemma avar_subst_vars_f v1 xs ys:
  (avar_f v1) /[ xs --> ys] = avar_f v1 /[ xs --> ys].
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.

Lemma avar_subst_pvars_f v1 x y:
  (avar_f v1) //[ x --> y] = avar_f v1 //[ x --> y].
Proof.
   edestruct (subst_pvars_unfold (avar_f v1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  edestruct (subst_pvars_unfold (v1)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst0.
  rewrite (avar_subst_vars_f v1 x zs); auto.
  rewrite (avar_subst_vars_f); auto.
  unfold fv, AvarFreeVar in Hfr.
  rewrite subst_vars_fresh_middle_change with (zs1:=zs) (ws:=zs0); eauto.
Qed.
Lemma avar_subst_vars_b v1 xs ys:
  (avar_b v1) /[ xs --> ys] = avar_b v1.
Proof.
  reverts ys.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite IHxs. simpls. auto.
Qed.
Lemma avar_subst_pvars_b v1 x y:
  (avar_b v1) //[ x --> y] = avar_b v1.
Proof.
   edestruct (subst_pvars_unfold (avar_b v1)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
rewrite avar_subst_vars_b.
rewrite avar_subst_vars_b. auto.
Qed.

(* With syntax instances defined for var, we can define lemmas for how subsitution commutes with opening *)

Section VarSubstVar.

  Context `{AS: AbstractSyntax}.
  Context {SF: SubstFresh AS}.
  Context {FS: FvSubstCases AS}.
  Context {SC: SubstCommut AS}.
  Context {ASOC: SubstOpenCommut AS}.
  Context {SI: SubstId AS}.


  Lemma subst_vars_open_commut n x (X: A) xs ys:
     X /[#n -> x] /[ xs --> ys] = X /[ xs --> ys] /[#n -> x/[ xs --> ys]].
  Proof.
    revert dependent ys.
    induction xs; intros; simpls.
    - auto.
    - destruct ys. auto.
      rewrite IHxs.
      rewrite subst_open_commut.
       eauto.
  Qed.

  Lemma subst_pvars_open_commut n x (T: A) xs ys:
     T /[#n -> x] //[ xs --> ys] = T //[ xs --> ys] /[#n -> x //[ xs --> ys]].
  Proof.

    lets vs0 Hf0 Heq0: (subst_pvars_unfold_L (open_rec n x T) xs ys (\{x} \u fv T)).
    rewrite Heq0.

    lets vs1 Hf1 Heq1: (subst_pvars_unfold_L  T xs ys \{x}).
    rewrite Heq1.

    lets vs2 Hf2 Heq2: (subst_pvars_unfold_L x xs ys (fv T)).
    rewrite Heq2.

    rewrite subst_vars_fresh_middle_change with (zs:=vs2) (ws:=vs0); eauto.

    rewrite subst_vars_fresh_middle_change with (zs:=vs1) (ws:=vs0); eauto.

    rewrite subst_vars_open_commut.
    rewrite subst_vars_open_commut.
    eauto.
    unfold fv, VarFreeVar.
    fresh_solve.
  Qed.
End VarSubstVar.

(* Rewritings for subst_var substitution *)

Lemma subst_eq_var (v1 v2: var):
  v1 /[v1 -> v2] = v2.
Proof.
  unfold subst_var, VarSubstVar.
  apply substi_eq_var.
Qed.

Lemma subst_toeq_var (v1 v2: var):
  v2 /[v1 -> v2] = v2.
Proof.
  unfold subst_var, VarSubstVar.
  apply substi_toeq_var.
Qed.

Lemma subst_ne_var (v1 v2 v3: var):
  v1 <> v2 ->
  v1 /[v2 -> v3] = v1.
Proof.
  intros. unfold subst_var, VarSubstVar. cases_if. auto.
Qed.

(* Case analysis for variable substitution *)
Lemma subst_cases (v1 v2 v3: var):
  v3 /[v1 -> v2] = v2 \/
  v3 /[v1 -> v2] = v3.
Proof.
  unfold subst_var, VarSubstVar.
  cases_if; auto.
Qed.

Lemma subst_cases2(v1 v2 v3: var):
  (v3 = v1 /\ v3 /[v1 -> v2] = v2) \/
  (v3 <> v1 /\ v3 /[v1 -> v2] = v3).
Proof.
  unfold subst_var, VarSubstVar.
  cases_if; auto.
Qed.

Lemma subst_cases2_avar(v1 v2: var)(v3: avar):
  (v3 = (avar_f v1) /\ v3 /[v1 -> v2] = (avar_f v2)) \/
  (v3 <> (avar_f v1) /\ v3 /[v1 -> v2] = v3).
Proof.
  unfold subst_var.
  unfold AvarSubstVar.
  destruct v3.
  - right. split. simplify_eq. auto.
  - destruct (classicT (v=v1)).
    + subst; auto.
    + right. split. simplify_eq. auto. auto.
Qed.

Lemma subst_back (v1 v2 v3: var):
  v3 /[v2 -> v1] /[v1 -> v2] = v3 /[v1 -> v2].
Proof.
  unfold subst_var.
  unfold VarSubstVar.
  destruct (classicT (v3=v2)); subst; auto.
  rewrite substi_eq_var, substi_toeq_var.
  auto.
Qed.

Lemma subst_trans (v1 v2 v3 v4: var):
  v3 <> v2 ->
  v3 /[v1 -> v2] /[v2 -> v4] = v3 /[v1 -> v4].
Proof.
  unfold subst_var, VarSubstVar.
  intros.
  destruct (classicT (v3=v1));
  cases_if; auto.
Qed.


Lemma avar_open_cases (a: avar) n x:
  (a /[#n -> x]  = a /\ a <> (avar_b n)) \/
  (a /[#n -> x]  = (avar_f x) /\ a = (avar_b n)).
Proof.
  unfold open_rec, AvarOpenable.
  destruct a.
  cases_if.
  auto.
  left. split; auto. intro. inversions H. contradiction.
  left. split; auto. discriminate.
Qed.

Lemma avar_subst_cases (a: avar) x y:
  (a /[x -> y]  = a /\ a <> (avar_f x)) \/
  (a /[x -> y] = (avar_f y) /\ a = (avar_f x)).
Proof.
  unfold subst_var, AvarSubstVar.
  destruct a.
  left. split; auto. discriminate.
  cases_if.
  auto.
  left. split; auto. intro. inversions H. contradiction.
Qed.

Instance VarSubstVarsEqInv: SubstVarsEqInv VarAbstractSyntax.
Proof.
  hnf; intros.
  reverts ys1 ys2.
  induction xs; intros.
  - (* Trivial for empty list *)
    destruct ys2; inversions H1.
    destruct ys1; inversions H0.
    auto.
  - (* Non-empty lists. *)
    destruct ys2; inversions H1.
    destruct ys1; inversions H0.
    simpls.
    (* Because from_list xs \c fv a, then xs must be a::nil *)
    rewrite from_list_cons in H2.
    forwards: subset_union_l2 H2.
    forwards: subset_union_l1 H2.
    apply subset_singleton in H1.
    unfold fv, VarFreeVar in H1.
    rewrite in_singleton in H1. subst.
    
    destruct xs.
    +
      destruct ys2; inversions H5.
      destruct ys1; inversions H4.
      (* xs = a::nil *)
      rewrite ! subst_vars_fresh in H3 by auto.
      rewrite ! subst_eq_var in H3. subst.
      fequal.
    + (* xs has more than one varianble - contradiction. *)
      false.
    rewrite from_list_cons in H0.
    forwards: subset_union_l1 H0.
    apply subset_singleton in H1.
    unfold fv, VarFreeVar in H1.
    rewrite in_singleton in H1. subst.
    destructecs H.
    inversions H.
    rewrite ! notin_union, notin_singleton in H6; destructecs H6.
    contradiction.
Qed.

Instance AvarSubstVarsEqInv: SubstVarsEqInv AvarAbstractSyntax.
Proof.
  hnf; intros.
  destruct a.
  - simpl in H2.
    eapply subset_empty_inv in H2.
    eapply from_list_empty_inv in H2.
    subst.
    destruct ys2; inversions H1.
    destruct ys1; inversions H0.
    auto.
  - rewrite ! avar_subst_vars_f in H3.
    inversions H3.
    eapply subst_vars_eq_inv; eauto.
Qed.


(**
  * Variable kind.
*)

(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(**
Variable kind.
[v_loc] indicates a location variable (keys in the heap).
[v_ref] indicates a reference variable (keys in the renv).
[v_par] indicates a method parameter.
[v_self] indicates an object self-reference.
[v_let] indicates let-bound variable.
[v_recv] indicates a method receiver parameter.
*)
Inductive varkind: Set :=
| v_loc: varkind (* Location *)
| v_ref: varkind (* Reference *)
| v_par: varkind (* Parameter *)
| v_self: varkind (* Self-ref *)
| v_let: varkind (* Let local *)
| v_recv: varkind (* Receiver *)
.

Inductive varkind_abstract: varkind -> Prop :=
| v_abstract_let: varkind_abstract v_let
| v_abstract_par: varkind_abstract v_par
| v_abstract_self: varkind_abstract v_self
| v_abstract_recv: varkind_abstract v_recv
.

Lemma vk_dec (vk1 vk2 : varkind) : vk1 = vk2 \/ vk1 <> vk2.
Proof.
  destruct vk1; destruct vk2;
    try solve [ left; eauto | right; eauto
              | left; easy | right; easy ].
Qed.