(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties.

Generalizable Variables A B.

(* ************************************************************************** *)
(** ** List-Like Syntax *)
(** To better use some of Coq's built-in features for generating induction
    schemes for mutually recursive structures we will often define specialized
    versions of list. In the following typeclass we think of [A] being a list
    containing [B]s. *)

Class ToList (B A : Set) :=
  to_list : A -> list B.
Arguments to_list _ _ _ !_ /.

Class OfList (B A : Set) :=
  of_list : list B -> A.
Arguments of_list _ _ _ !_ /.

Class ListLike {B A : Set} `(ToList B A) `(OfList B A) :=
  { to_list_of_list : forall x, (of_list (to_list x)) = x;
    of_list_to_list : forall l, (to_list (of_list l)) = l;
  }.
Bind Scope syntax_scope with ToList OfList.

Section About_ListLike.
  Context `{ListLike B A}.

  Definition app_s xs ys := of_list ((to_list xs) ++ (to_list ys)).

  Infix "++" := app_s (right associativity, at level 60).

  Lemma app_s_assoc : forall xs ys zs,
      xs ++ ys ++ zs = (xs ++ ys) ++ zs.
  Proof.
    intros. unfold app_s. rewrite ? of_list_to_list.
    apply f_equal. rewrite List.app_assoc.
    auto.
  Qed.

  Definition map_f f x := of_list (List.map f (to_list x)).

  Lemma map_f_app_s : forall xs ys f,
      map_f f (xs ++ ys) = (map_f f xs) ++ (map_f f ys).
  Proof.
    intros. unfold app_s, map_f. rewrite ? of_list_to_list.
    apply f_equal. rewrite List.map_app.
    auto.
  Qed.

  Definition length_s xs := length (to_list xs).

  Lemma app_s_length_s : forall xs ys,
      length_s (xs ++ ys) = length_s xs + length_s ys.
  Proof.
    intros. unfold length_s, app_s. rewrite ? of_list_to_list.
    rewrite List.app_length; auto.
  Qed.

  Lemma singles_map_to_list : forall f ys a,
      length ys = length_s a ->
      ys ~* (to_list (map_f f a)) = map f (ys ~* to_list a).
  Proof.
    intros. rewrite map_singles.
    - rewrite liblist_map.
      unfold map_f;rewrite of_list_to_list.
      auto.
    - rewrite liblist_length. auto.
  Qed.

  Lemma singles_rev_map_to_list : forall f ys a,
      length ys = length_s a ->
      ys ~** (to_list (map_f f a)) = map f (ys ~** to_list a).
  Proof.
    intros. rewrite map_singles.
    - rewrite liblist_map. rewrite List.map_rev.
      unfold map_f; rewrite of_list_to_list; auto.
    - rewrite liblist_length. rewrite ? List.rev_length.
      auto.
  Qed.

  Lemma of_list_app : forall xs ys,
      of_list (xs ++ ys) = (of_list xs) ++ (of_list ys).
  Proof.
    intros. unfold app_s.
    rewrite ? of_list_to_list; auto.
  Qed.

  Lemma listlike_rev_ind : forall (P : A -> Prop),
       P (of_list nil) ->
       (forall (b : B) (a : A),
           P a ->
           P (a ++ (of_list (b :: nil)))) ->
       forall a : A, P a.
  Proof.
    intros. rewrite <- to_list_of_list.
    induction (to_list a) using List.rev_ind; auto.
    rewrite of_list_app; auto.
  Qed.

End About_ListLike.

Infix "++" := app_s (right associativity, at level 60) : syntax_scope.
Arguments app_s _ _ _ _ _ !_ /.
Arguments length_s _ _ _ !_ /.





(** *** List of Syntax *)
Section ListOfSyntax.
  Context {A: Set}.
  Context {O: Openable A}.
  Context {F: FreeVar A}.
  Context {S: SubstVar A}.
  Context {AS: AbstractSyntax O F S}.

  Lemma list_map_compose : forall {A B C : Type} (f : A -> B) (g : B -> C),
      List.map g ∘ List.map f = List.map (g ∘ f).
  Proof.
    intros. unfold compose.
    apply fun_ext_dep; intros xs.
    rewrite List.map_map; auto.
  Qed.

  Instance ListOpenable : Openable (list A) :=
    (fun n x => List.map (open_rec n x)).
  Instance ListFreeVar : FreeVar (list A) :=
    (List.fold_right (fun a acc => union (fv a) acc) \{}).
  Instance ListSubstVar : SubstVar (list A) :=
    (fun x y => List.map (subst_var x y)).
  Instance ListSealable {FX: Sealable A} : Sealable (list A) :=
    (fun n x => List.map (seal_rec n x)).
  Instance ListAbstractSyntax : AbstractSyntax ListOpenable ListFreeVar ListSubstVar := {}.

  Section OpenCompose.
    Context {OT: OpenTwice O}.

    Instance ListOpenTwice : OpenTwice ListOpenable.
    Proof.
      hnf. unfold open_rec, ListOpenable; cbn.
      intros; rewrite list_map_compose.
      rewrite open_twice_comp; auto.
    Qed.

    Context {OC: OpenCommut O}.
    Instance ListOpenCommut : OpenCommut ListOpenable.
    Proof.
      hnf; unfold open_rec, ListOpenable; simpl.
      intros; rewrite ? list_map_compose.
      rewrite open_commut_comp by auto; auto.
    Qed.
  End OpenCompose.

  Section SubstFresh.
    Context {SF: SubstFresh AS}.


    Instance ListSubstFresh : SubstFresh ListAbstractSyntax.
    Proof.
      unfold SubstFresh. induction X; intros.
      - simpl; auto.
      - cbn in *. rewrite notin_union in H.
        destruct H. apply IHX in H0. rewrite H0.
        rewrite subst_fresh; auto.
    Qed.
  End SubstFresh.

  Section SubstOpenCommut.
    Context {SF: SubstOpenCommut AS}.

    Instance ListSubstOpenCommut : SubstOpenCommut ListAbstractSyntax.
    Proof.
      hnf; intros.
      unfold open_rec, subst_var; simpl.
      induction X; auto.
      simpl. rewrite subst_open_commut.
      f_equal; auto.
    Qed.
  End SubstOpenCommut.

  Section FvOpenCases.
    Context {FOC: FvOpenCases AS}.

    Instance ListFvOpenCases : FvOpenCases ListAbstractSyntax.
    Proof.
      hnf; intros x a n. induction a; auto.
      intros; simpl. apply union_eq_cases; auto using fv_open_cases.
    Qed.
  End FvOpenCases.

  Section Closing.
    Context {C: Closing AS}.

    Definition ListClosing : Closing ListAbstractSyntax :=
      (fun n x => List.map (close_rec n x)).

    Local Ltac lift_closing lem :=
      hnf;
      unfold open_rec, close_rec, subst_var,
      ListOpenable, ListClosing, ListSubstVar;
      simpl; intros;
      rewrite ? list_map_compose, lem; auto.

    Context {OC: OpenClose C}.
    Instance ListOpenClose : OpenClose ListClosing.
    Proof.
      hnf;
      unfold open_rec, close_rec, subst_var; simpl; intros;
      lift_closing open_close_comp.
    Qed.

    Context {CO: CloseOpen C}.
    Instance ListCloseOpen : CloseOpen ListClosing.
    Proof.
      lift_closing close_open_comp.
    Qed.

    Context {SOC: SubstOpenClose C}.
    Instance ListSubstOpenClose : SubstOpenClose ListClosing.
    Proof.
      lift_closing subst_open_close_comp.
    Qed.
    Context {CS: CloseSubst C}.
    Instance ListCloseSubst : CloseSubst ListClosing.
    Proof.
      lift_closing close_subst_comp.
    Qed.
    Context {SCCo: SubstCloseCommut C}.
    Instance ListSubstCloseCommut : SubstCloseCommut ListClosing.
    Proof.
      lift_closing subst_close_commut_comp.
    Qed.
    Context {CCo: CloseCommut C}.
    Instance ListCloseCommut : CloseCommut ListClosing.
    Proof.
      lift_closing close_commut_comp.
    Qed.
    Context {OCCo: OpenCloseCommut C}.
    Instance ListOpenCloseCommut : OpenCloseCommut ListClosing.
    Proof.
      lift_closing open_close_commut_comp.
    Qed.

    Context {CU: CloseUnbound C}.

    Instance ListCloseUnbound : CloseUnbound ListClosing.
    Proof.
      hnf; intros. unfold close_rec; simpl.
      induction a; auto.
      simpl in H. apply notin_union in H. destruct H.
      simpl. rewrite close_unbound by auto.
      rewrite IHa by auto. auto.
    Qed.
    (** From above, this implies OpenInjective and SubstIntro *)
  End Closing.


End ListOfSyntax.
Existing Instances
         ListOpenable
         ListFreeVar
         ListSubstVar
         ListAbstractSyntax
         ListOpenTwice
         ListOpenCommut
         ListSubstFresh
         ListSubstOpenCommut
         ListClosing
         ListOpenClose
         ListCloseOpen
         ListSubstCloseCommut
         ListSubstOpenClose
         ListCloseUnbound
         ListOpenCloseCommut.

