(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties.

Generalizable Variables A B.


(* ************************************************************************** *)
(** ** Extra Definitions and Simple Lemmas *)


Section Open.
  Context `{AS: AbstractSyntax A}.
  Lemma open_vars_closed n xs (T: A):
    closed T ->
    T /[#n --> xs] = T.
  Proof.
    intros.
    revert dependent n.
    revert dependent T.
    induction xs; simpls; eauto.
    intros.
    rewrite H2.
    rewrite IHxs; eauto.
  Qed.
  Lemma open_vars_nil n (T: A):
    T /[#n --> nil] = T.
  Proof.
    eauto.
  Qed.
  Lemma open_vars_single n (T: A) x:
     T /[#n --> x::nil] = T /[#n -> x].
  Proof.
    eauto.
  Qed.
  Lemma open_vars_push_r x n xs (T: A):
    T /[#n -> x] /[# S n --> xs] = T /[# n --> x::xs].
  Proof.
    unfold open_rec_vars. auto.
  Qed.
End Open.




Section Fv_env.
  Context `{AS: AbstractSyntax A}.

  Lemma fv_env_single: forall x a,
    fv_env (x ~ a) = fv a.
  Proof.
    intros. unfold fv_in_values.
    rewrite single_def, values_def.
    rewrite liblist_fold_right, liblist_map.
    simpl. rewrite union_empty_r. auto.
  Qed.

  Lemma fv_env_empty:
    fv_env (@empty A) = \{}.
  Proof.
    intros. unfold fv_in_values.
    rewrite empty_def, values_def.
    rewrite liblist_fold_right, liblist_map.
    simpl. auto.
  Qed.

  Lemma fv_env_push: forall E x a,
      fv_env (E & x ~ a) = fv_env E \u fv a.
  Proof.
    intros. rewrite fv_in_values_cons.
    rewrite fv_env_single. auto.
  Qed.

  Lemma fv_env_concat (e1 e2 : env A) :
    fv_env (e1 & e2) = (fv_env e1) \u (fv_env e2).
  Proof.
    induction e2.
    - rewrite <- empty_def, concat_empty_r, fv_env_empty, union_empty_r. auto.
    - destruct a. rewrite cons_to_push, concat_assoc.
      repeat(rewrite fv_env_push). rewrite IHe2. rewrite union_assoc. auto.
  Qed.

End Fv_env.


Section About_OpenTwice.
  Context `{OpenTwice}.

  Lemma open_twice n x y T:
    T /[#n -> x] /[#n -> y] = T /[#n -> x].
  Proof.
    pose proof (open_twice_comp n x y).
    remember (open_rec n y (open_rec n x T) ) as T'. rewrite <- H0. subst. compute. auto.
  Qed.

  Lemma open_exists x y n T:
    (exists T2, T = T2 /[#n -> x]) ->
    (exists T2, T = T2 /[#n -> y]).
  Proof.
    intros.
    destruct H0 as [T2 HT].

    rewrite <- (open_twice n x y) in HT.
    exists (open_rec n x T2). auto.
  Qed.

  Lemma closed_reopen n x y (T: A):
    T /[#n -> x] = T ->
    T /[#n -> y] = T.
  Proof.
    introv Hx.
    rewrite <- Hx.
    rewrite open_twice. auto.
  Qed.

End About_OpenTwice.

Section About_OpenCommut.
  Context `{OpenCommut A}.

  Lemma open_commut a x n m t:
    m <> n ->
    t /[#n -> x] /[#m -> a] =
    t /[#m -> a] /[#n -> x].
  Proof.
    intros.
    rewrite (fold_compose (open_rec _ _)).
    rewrite (fold_compose (open_rec _ _)).
    rewrite open_commut_comp; auto; Coq.omega.Omega.omega.
  Qed.


  Lemma open_rec_vars_open_rec_commut: forall x xs n m a,
      n < m \/ (length xs + m) < S n ->
      a /[#n -> x] /[# m --> xs] = a /[# m --> xs] /[#n -> x].
  Proof.
    intros x. induction xs; simpl; intros.
    - auto.
    - rewrite <- IHxs; try Coq.omega.Omega.omega.
      fequal.
      rewrite (fold_compose (open_rec _ _)).
      rewrite open_commut; auto; Coq.omega.Omega.omega.
  Qed.

  Lemma open_vars_S_commut : forall x xs n a,
      a /[#n -> x] /[# S n --> xs] = a /[# S n --> xs] /[#n -> x].
  Proof.
    intros. apply open_rec_vars_open_rec_commut.
    Coq.omega.Omega.omega.
  Qed.

  Lemma open_vars_app : forall xs ys n a,
      a /[#n --> xs ++ ys] =
      a /[#(n + length xs) --> ys] /[#n --> xs].
  Proof.
    induction xs; simpl; intros.
    - rewrite Nat.add_0_r; auto.
    - rewrite IHxs. rewrite open_rec_vars_open_rec_commut by Coq.omega.Omega.omega.
      rewrite Nat.add_succ_comm; auto.
  Qed.
  Lemma open_vars_push x n xs (T: A):
    T /[# S n --> xs] /[#n -> x] = T /[# n --> x::xs].
  Proof.
    rewrite <- open_vars_S_commut.
    unfold open_rec_vars. auto.
  Qed.

End About_OpenCommut.

Section About_OpenInjective.
  Context `{AS: AbstractSyntax A}.
  Context {AO: OpenInjective AS}.
  Context {AF: FvOpenCases AS}.
  Lemma open_vars_injective (X Y: A) k zs:
    fresh (fv X \u fv Y) (length zs) zs ->
    X /[# k --> zs] = Y /[# k --> zs] ->
    X = Y.
  Proof.
    intros.
    revert dependent k.
    revert dependent X.
    revert dependent Y.
    induction zs; intros.
    - simpls. auto.
    - simpls.
      destruct H2.

    assert((open_rec k a X) = (open_rec k a Y)).
      eapply IHzs; eauto.
      destruct (fv_open_cases a X k) as [Hfo | Hfo]; rewrite Hfo;
      destruct (fv_open_cases a Y k) as [Hfoy | Hfoy]; rewrite Hfoy;
      fresh_solve.

    rewrite notin_union in H2.
    destruct H2.

    eapply open_fresh_injective in H5; eauto.
  Qed.
End About_OpenInjective.


Section About_FvOpenCases.
  Context `{FvOpenCases}.

  Lemma notin_fv_open n x y (U: A):
    x <> y ->
    y \notin fv U ->
    y \notin fv (U /[#n -> x]).
  Proof.
    intros.
    destruct (fv_open_cases x U n) as [Hr | Hr]; rewrite Hr; eauto.
  Qed.

  Lemma notin_open_rec_vars : forall x ys T n,
    x \notin fv T ->
    fresh (\{ x} \u (fv T)) (length ys) ys ->
    x \notin fv T /[# n --> ys].
  Proof.
    induction ys; introv Hn Hf; simpls; auto.
    assert (H' : fv (open_rec n a T) = fv T \/
                 fv (open_rec n a T) = \{ a} \u fv T)
      by auto using fv_open_cases. destruct Hf.
    apply IHys; destruct H' as [H' | H']; rewrite H'; auto.
  Qed.

  Lemma notin_open_rec_vars_list : forall x ys T n,
    x \notin fv T ->
    x \notin from_list ys ->
    x \notin fv T /[# n --> ys].
  Proof.
    intros x ys.
    induction ys; introv Hn Hf; simpls; auto.
    rewrite from_list_cons in Hf.
    apply IHys; auto.
    assert (H' : fv (open_rec n a T) = fv T \/
                 fv (open_rec n a T) = \{ a} \u fv T)
      by auto using fv_open_cases.
    destruct H' as [H' | H']; rewrite H'; auto.
  Qed.

  Lemma fresh_fv_open n x ws a:
    fresh ((fv a) \u \{ x }) (length ws) ws ->
    fresh (fv a /[#n -> x]) (length ws) ws.
  Proof.
    introv Hfr.
    destruct (fv_open_cases x a n).
    -
      rewrite H3. fresh_solve.
    - rewrite H3.
      fresh_solve.
  Qed.
  Lemma fv_open_subset n x a:
    (fv a /[#n -> x]) \c (\{x} \u (fv a)).
  Proof.
    destruct (fv_open_cases x a n).
    rewrite H3. eapply subset_union_weak_r.
    rewrite H3. eapply subset_refl.
  Qed.
  Lemma fv_open_superset n x a:
     (fv a) \c (fv a /[#n -> x]).
  Proof.
    destruct (fv_open_cases x a n).
    rewrite H3. eapply subset_refl. 
    rewrite H3. eapply subset_union_weak_r.
  Qed.
  Lemma fv_open_vars_superset n xs a:
     (fv a) \c (fv a /[#n --> xs]).
  Proof.
    reverts n a.
    induction xs; intros; simpls.
    - apply subset_refl.
    - eapply subset_trans. eapply fv_open_superset.  apply IHxs.
  Qed.


End About_FvOpenCases.

Section About_ClosingSanity.
  Section OpenClose.
    Context `{OpenClose}.

    Lemma open_close n x (T: A):
        T /[#n <- x] /[#n -> x] = T /[#n -> x].
    Proof.
      pose proof (open_close_comp (A:=A) n x ).
      unfold compose in H3.
      eapply (fun_eq_1) in H3.
      eauto.
    Qed.


  End OpenClose.

  Section CloseOpen.
    Context `{CloseOpen}.

    Lemma close_open n x (T: A):
        T /[#n -> x] /[#n <- x] = T /[#n <- x].
    Proof.
      pose proof (close_open_comp (A:=A) n x).
      unfold compose in H3.
      eapply (fun_eq_1) in H3.
      eauto.
    Qed.


  End CloseOpen.

  Section OpenCloseCommut.
    Context `{OpenCloseCommut}.

    Lemma open_close_commut n m x y (T: A):
        x <> y -> m <> n ->
        T /[#m <- x] /[#n -> y] = T /[#n -> y] /[#m <- x].
    Proof.
      introv Hxy Hmn.
      pose proof (open_close_commut_comp (A:=A) Hxy Hmn).
      unfold compose in H3.
      eapply (fun_eq_1) in H3.
      eauto.
    Qed.
  End OpenCloseCommut.


  Section CloseCommut.
  Context `{CloseCommut A}.

  Lemma close_commut a x n m t:
    a <> x ->
    t /[#n <- x] /[#m <- a] = t /[#m <- a] /[#n <- x].
  Proof.
    intros.
    rewrite (fold_compose (close_rec _ _)).
    rewrite (fold_compose (close_rec _ _)).
    rewrite close_commut_comp; auto; Coq.omega.Omega.omega.
  Qed.

  End CloseCommut.
End About_ClosingSanity.



Section CloseVars.
  Context `{AC: Closing A}.
  Lemma close_vars_unfold_r (t: A) x (xs: list var) n:
    t /[#n <-- x::xs] = t /[#n <- x] /[# S n <-- xs] .
  Proof.
    eauto.
  Qed.

  Lemma close_vars_nil (t: A) n:
    t /[#n <-- nil] = t.
  Proof.
    unfold close_vars. auto.
  Qed.
  

  Context {ACC: CloseCommut AC}.

  (* Analogy to open_rec_vars_open_rec_commut *)
  Lemma close_vars_commut (t: A) x (xs: list var) m n:
    x \notin (from_list xs) ->
    t /[# m <-- xs] /[#n <- x] = t /[#n <- x] /[# m <-- xs].
  Proof.
    reverts m t.
    induction xs; introv Hn.
    - unfold close_rec_vars. auto.
    - rewrite from_list_cons in Hn.
      rewrite ! close_vars_unfold_r.
      rewrite close_commut.
      rewrite IHxs;
      eauto.
      notin_solve.
  Qed.

  Lemma close_vars_unfold_l (t: A) x (xs: list var) n:
    x \notin from_list xs ->
    t /[#n <-- x::xs] = t /[# S n <-- xs]  /[#n <- x].
  Proof.
    intros.
    rewrite close_vars_unfold_r.
    rewrite close_vars_commut; eauto.
  Qed.

  Context {ACU: CloseUnbound AC}.
  Lemma close_vars_fresh (t: A) n (ys: list var):
    fresh (fv t) (length ys) ys ->
    t /[# n <-- ys]  = t.
  Proof.
    introv Hfr.
    reverts t n.
    induction ys; intros.
    - simpls. auto.
    - rewrite close_vars_unfold_r.
      rewrite close_unbound; eauto.
      eapply IHys.
      simpl in Hfr. destruct_all. eauto.
  Qed.


End CloseVars.




Section About_SealOpen.
  Context `{SealOpen}.

  (* Sealing is sealing at a level above and opening *)
  (* This is an applied version of open_seal_fold_comp. *)
  Lemma open_seal_fold n x X:
    X /[# S n ->> x] /[#n -> x] = X /[#n ->> x].
  Proof.
    pose proof open_seal_fold_comp.
    unfold compose in H3.
    specialize (H3 n x).
    apply fun_eq_1 with (x1:=X) in H3.
    auto.
  Qed.

  (* Sealing commutes with opening at a lower level *)
  (* This is an applied version of open_seal_commut *)
  Lemma open_seal_commut n m x y X:
    n < m ->
    X /[#m ->> y] /[#n -> x] = X /[#n -> x] /[#m ->> y].
  Proof.
    intros.
    pose proof open_seal_commut_comp as Hcomp.
    unfold compose in Hcomp.
    specialize (Hcomp n m x y).
    apply fun_eq_1 with (x1:=X) in Hcomp;
    auto.
  Qed.

End About_SealOpen.

Section About_SealClose.
  Context `{SealClose}.

  (* Sealing commutes with opening at a lower level *)
  (* This is an applied version of open_seal_commut *)
  Lemma close_seal_commut n m x y X:
    n < m ->
    x <> y ->
    X /[#m ->> y] /[#n <- x] = X /[#n <- x] /[#m ->> y].
  Proof.
    intros.
    pose proof close_seal_commut_comp as Hcomp.
    unfold compose in Hcomp.
    specialize (Hcomp n m x y).
    apply fun_eq_1 with (x1:=X) in Hcomp;
    auto.
  Qed.

  Lemma close_vars_seal_commut n m xs y X:
    (n + length xs) <= m ->
    y \notin (from_list xs) ->
    X /[#m ->> y] /[#n <-- xs] = X /[#n <-- xs] /[#m ->> y].
  Proof.
    intros.
    reverts n X.
    induction xs; intros.
    -
    rewrite ! close_vars_nil. auto.
    -
    simpl in H6.
rewrite from_list_cons, notin_union, notin_singleton in H7.
destruct H7.
    unfold close_rec_vars; fold close_rec_vars.
    rewrite <- IHxs.
    rewrite close_seal_commut. auto.

eapply Nat.lt_le_trans. 2: eauto.
eapply Nat.lt_add_pos_r. eapply Nat.lt_0_succ.
auto.
auto.


rewrite Nat.add_succ_comm. auto.
  Qed.


End About_SealClose.

Section SealOpenTwice.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {OC: OpenCommut H}.
  Context {SA: Sealable A}.
  Context {SC: SealOpen AS SA}.

  (* After sealing, opening with big indices does nothing *)
  Lemma open_sealed n x m y X:
    n <= m ->
    X /[#n ->> x] /[#m -> y] = X /[#n ->> x].
  Proof.
    introv Hle.

    destructc (PeanoNat.Nat.le_exists_sub n m Hle) as p Hm Hp.
    clear Hp.
    subst.
    reverts n.
    induction p; intros.
    -
      rewrite Nat.add_0_l.
      rewrite <- (open_seal_fold n x X).
      rewrite open_twice. auto.
    -
      rewrite <- (open_seal_fold n x X).
      rewrite open_commut.
      eapply f_equal.
      rewrite Nat.add_succ_comm.
      eapply IHp.
      forwards Hp: (Plus.plus_le_compat_r 0 p (S n)); auto.
      apply Nat.le_0_l.
      intro.
      rewrite Nat.add_succ_l in H2.
      symmetry in H2.
      apply Nat.succ_add_discr in H2.
      auto.
  Qed.

  (* Opening largest valid index *)
  Lemma open_seal_down n x y X:
    X /[# S n ->> x] /[#n -> y] = X /[#n -> y] /[#n ->> x].
  Proof.
    rewrite <- (open_seal_fold n x).
    rewrite open_seal_commut.
    rewrite (open_seal_commut x x).
    rewrite open_twice. auto.
    eapply Nat.lt_succ_diag_r.
    eapply Nat.lt_succ_diag_r.
  Qed.

  Lemma sealed_closed y X:
    closed X /[->> y].
  Proof.
    hnf. intros.
    rewrite open_sealed; auto.
    eapply Nat.le_0_l.
  Qed.

  Lemma sealed_closed_n n y (R: A):
    R /[#n ->> y] = R ->
    R /[#n -> y] = R.
  Proof.
    introv Hsl. replace (open_rec n y R) with (open_rec n y ( seal_rec n y R )).
    erewrite (open_sealed); eauto.
    rewrite Hsl. auto.
  Qed.

  Lemma sealed_closed0 X:
    sealed X ->
    closed X.
  Proof.
    introv Hsl.
    hnf in *. intros.
    specialize (Hsl x).
    rewrite <- Hsl.
    eapply sealed_closed.
  Qed.
End SealOpenTwice.

Section SealOpenClosed.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {OC: OpenCommut H}.
  Context {SA: Sealable A}.
  Context {SO: SealOpen AS SA}.
  Context {SC: SealClosed AS SA}.

  Lemma closed_sealed n y X:
    closed X ->
    X /[#n ->> y] = X.
  Proof.
    introv Hcl.
    unfold closed in Hcl.
    eapply closed_sealed_rec.
    intros. eauto.
  Qed.


  Lemma closed_sealed0 X:
    closed X ->
    sealed X.
  Proof.
    introv Hcl.
    intro.
    eapply closed_sealed in Hcl.
    eapply Hcl.
  Qed.


  Lemma sealed_up n x y X:
    X /[->> x] = X ->
    X /[#n ->> y] = X.
  Proof.
    introv Hsl.
    assert (closed X).
    rewrite <- Hsl.
    eapply sealed_closed; eauto.
    eapply closed_sealed; eauto.
  Qed.

  Lemma sealed_sealed_rec n x (U: A):
    sealed U ->
    U /[#n ->> x] = U.
  Proof.
    intros.
    eapply sealed_up with (x:= x); eauto.
  Qed.
End SealOpenClosed.

Section OpenTwiceInv.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {OC: OpenCommut H}.
  Context {OTI: OpenTwiceInv AS}.

  Lemma reopen_closed x y (a: A):
    closed a /[-> x] ->
    closed a /[-> y].
  Proof.
    unfold closed.
    introv Hcl.
    intros.
    specialize (Hcl n x0).
    destruct (classicT (n=0)).
    - subst. rewrite open_twice. auto.
    - rewrite open_commut by auto.
      rewrite open_commut in Hcl by auto.
      eapply open_twice_inv in Hcl; eauto.
      rewrite Hcl. auto.
  Qed.

  Lemma closed_open_down x n y (R: A):
    closed R /[-> x] ->
    R /[# S n -> y] = R.
  Proof.
    introv Hcl.
    unfold closed in Hcl.
    specialize (Hcl (S n) y).
    rewrite open_commut in Hcl.
    eapply  open_twice_inv in Hcl.
    auto.
    auto.
    auto.
Qed.

End OpenTwiceInv.



Section SealOpenClosedInv.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {SOC: SubstOpenCommut AS}.
  Context {CA: Closing AS}.
  Context {SOCA: SubstOpenClose CA}.
  Context {SA: Sealable A}.
  Context {SO: SealOpen AS SA}.
  Context {SC: SealClosed AS SA}.
  Context {OSI: OpenSealInv AS SA}.

  Lemma closed_seal_down x y (R: A):
    closed R /[-> x] ->
    R /[#1 ->> y] = R.
  Proof.
    introv Hclt.
    eapply closed_sealed with (n:=0) (y0:=y) in Hclt.
    rewrite <- open_seal_fold in Hclt.
    rewrite open_seal_commut in Hclt.
    rewrite open_twice in Hclt.
    rewrite <- open_seal_commut in Hclt.
    eapply open_seal_inv in Hclt.
    auto.
    auto.
    auto.
    auto.
  Qed.
End SealOpenClosedInv.

Section AboutFvClose.
  Context `{AS: AbstractSyntax A}.
  Context {CA: Closing AS}.
  Context {FC: FvClose CA}.
  Lemma notin_fv_close x (T: A) n:
    x \notin fv T /[#n <- x].
  Proof.
    rewrite fv_close.
    intro Hr.
    rewrite in_remove in Hr. destruct Hr.
    rewrite notin_singleton in H3. contradiction.
  Qed.
  Lemma notin_fv_close_notin x y (T: A) n:
    x \notin fv T ->
    x \notin fv T /[#n <- y].
  Proof.
    introv Hn.
    rewrite fv_close.
    intro Hr.
    rewrite in_remove in Hr. destruct Hr.
    contradiction.
  Qed.

  Lemma fv_close_subset x (T: A) n:
    fv T /[#n <- x] \c fv T.
  Proof.
    rewrite fv_close.
    intro. intro. rewrite in_remove in H2. destruct H2. auto.
  Qed.

  Lemma fv_close_vars_subset xs (T: A) n:
    fv T /[#n <-- xs] \c fv T.
  Proof.
    reverts T n.
    induction xs; intros; simpls.
    - rewrite close_vars_nil. apply subset_refl.
    - rewrite close_vars_unfold_r.
      eapply subset_trans.
      eapply IHxs.
      eapply fv_close_subset.
  Qed.

  Lemma fv_close_vars xs (T: A) n:
    fv T /[#n <-- xs] = fv T \- (from_list xs).
  Proof.
    reverts T n.
    induction xs; intros; simpls.
    - rewrite close_vars_nil, from_list_nil, remove_empty_l. auto.
    - rewrite close_vars_unfold_r, from_list_cons. rewrite IHxs.
      rewrite remove_union. rewrite fv_close. auto.
  Qed.

End AboutFvClose.


(* ************************************************************************** *)
(** ** Typeclass Implications *)
Section About_CloseUnbound.
  Context `{AS: AbstractSyntax A}.
  Context {C: Closing AS}.
  Context {CU: CloseUnbound C}.
  Context {CO: CloseOpen C}.

  (** Every CloseUnbound leads to a OpenInjective *)
  Instance CloseUnboundOpenInjective : OpenInjective AS.
  Proof.
    hnf. introv HnX HnY Heq.
    apply (f_equal (close_rec k z)) in Heq.
    rewrite ? (fold_compose (close_rec _ _)) in Heq.
    rewrite close_open_comp in Heq.
    rewrite (close_unbound _ _ HnY), (close_unbound _ _ HnX) in Heq.
    auto.
  Qed.

  Context {SOC: SubstOpenClose C}.
  Instance CloseUnboundSubstIntro : SubstIntro AS.
  Proof.
    hnf. intros. rewrite (fold_compose (subst_var x y)).
    rewrite subst_open_close_comp. unfold compose.
    apply f_equal. symmetry. auto using close_unbound.
  Qed.
End About_CloseUnbound.
Existing Instances
         CloseUnboundOpenInjective
         CloseUnboundSubstIntro.
