Set Implicit Arguments.

Require Import LibExtra.
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties SyntaxClassLemmas SyntaxClassSubstLemmas SyntaxClassSubstVarsLemmas.

Generalizable Variables A B.

Section Subst_PVars.
  Context `{AS: SubstVar}.
  Context `{AF: FreeVar A}.

  (* Parallel substitution is equivalent to two list substitutions *)
  Lemma subst_pvars_unfold a xs ys:
    exists vs,
      fresh ((from_list xs) \u (from_list ys) \u fv a ) (length ys) vs /\
      a //[ xs --> ys] = a /[xs --> vs] /[vs --> ys].
  Proof.
    unfold subst_pvars.
    remember (var_freshes (from_list xs \u from_list ys \u fv a)  (length ys) ) as s.
    destruct s.
    clear Heqs.
    exists x.
    auto.
  Qed.
  (* Substituting no variables is identity *)
  Lemma subst_pvars_nil_r (a: A) ys:
     a //[ ys --> nil] = a.
  Proof.
    intros.
    lets vs Hf Heq: (@subst_pvars_unfold a ys nil).
    eapply fresh_length in Hf.
    destruct vs; simpls; try discriminate.
    induction ys; simpls; auto.
  Qed.
End Subst_PVars.
Section Subst_PVars.
  Context `{AS: AbstractSyntax}.
  Context {SF: SubstFresh AS}.
  Context {FS: FvSubstCases AS}.
  Context {SC: SubstCommut AS}.
  Context {SI: SubstId AS}.
  Context {OC: OpenCommut H}.

  (* Substituting no variables is identity *)
  Lemma subst_pvars_nil_l (a: A) ys:
    a //[ nil --> ys ] = a.
  Proof.
    intros.
    lets vs Hf Heq: (subst_pvars_unfold a nil ys).
    rewrite Heq. simpls.
    rewrite subst_vars_fresh; eauto.
    eapply fresh_resize_length in Hf.
    fresh_solve.
  Qed.

  (* If domain has no duplicates and disjoint from codomain,
     parallel substitution is the smae as list substitution *)
  Lemma subst_vars_pvars xs ys (X: A):
    (length xs) = (length ys) ->
    fresh (from_list ys) (length xs) xs ->
    X /[xs --> ys] = X //[xs --> ys].
  Proof.
    introv Hlx Hfrx.
    forwards vs Hfrv Hs: (subst_pvars_unfold X xs ys ); eauto.
    forwards: fresh_length; eauto.
    rewrite <- H2 in Hfrv.
    rewrite Hs.
    rewrite subst_vars_merge; eauto.
    rewrite H2. auto.
  Qed.

  (* Substituting fresh variables is identity *)
  Lemma subst_pvars_fresh xs ys (a: A):
    fresh (fv a) (length xs) xs ->
    a //[ xs --> ys] = a.
  Proof.
    introv Hfr.

    lets vs0 Hf0 Heq0: (subst_pvars_unfold a xs ys).
    rewrite Heq0.

    rewrite subst_vars_fresh with (xs0:=xs).
    rewrite subst_vars_fresh.
    auto.
    eapply fresh_resize_length in Hf0.
    fresh_solve.
    auto.
  Qed.

  (* Parallel substitution is equivalent to two list substitutions,
     and we can choose which variables cannot occur in the middle list.
     Stronger than subst_pvars_unfold. *)
  Lemma subst_pvars_unfold_L a xs ys L:
    exists vs,
      fresh ((from_list xs) \u (from_list ys) \u fv a \u L) (length ys) vs /\
      a //[ xs --> ys] = a /[xs --> vs] /[vs --> ys].
  Proof.
    forwards vs Hfr Heq: (subst_pvars_unfold a xs ys); eauto.
    set (F:= var_freshes ((from_list xs) \u (from_list ys) \u fv a \u L) (length ys)).
    destructc F as fs ?.
    exists fs.
    split. auto.
    rewrite Heq.
    eapply subst_vars_fresh_middle_change.
    fresh_solve.
    fresh_solve.
  Qed.


  Lemma subst_pvars_push (a: A) x xs y ys:
    x \notin (from_list ys) ->
    a //[ x::xs --> y::ys] = a //[ xs --> ys] /[x -> y].
  Proof.
    introv Hnys.

    lets x0 f Heq0: (subst_pvars_unfold a (x :: xs) (y :: ys)).
    rewrite Heq0.
    lets x1 f0 Heq1: (subst_pvars_unfold a xs ys).
    rewrite Heq1.

    lets: (fresh_length _ _ _ f).
    destruct x0.
    simpls. discriminate.
    simpls. inversions H2.
    destruct f.
    rewrite notin_union in H2.
    destruct H2.
    rewrite from_list_cons in H2.
    rewrite notin_union in H2.
    destruct H2. rewrite notin_singleton in H2.
    rewrite notin_union in H5.
    destruct H5.
    rewrite from_list_cons in H5.
    rewrite notin_union in H5.
    destruct H5. rewrite notin_singleton in H5.
    rewrite from_list_cons in H3.
    rewrite from_list_cons in H3.

    rewrite <- subst_vars_commut with (x2:=x) (y0:=v).
    rewrite subst_merge.
    eapply f_equal.
    eapply subst_vars_fresh_middle_change.
    fresh_solve.
    fresh_solve.
    auto.
    eapply notin_fv_subst_vars.
    eapply notin_fv_subst_vars.
    notin_solve.
    eapply from_list_fresh. fresh_solve.
    auto.
    eapply from_list_fresh. fresh_solve.
    auto.
    eapply from_list_fresh. fresh_solve.
  Qed.

  (* Substituting a variable to itself is identity *)
  Lemma subst_pvars_push_same (a: A) x xs ys:
    x \notin from_list ys ->
    a //[ x::xs --> x::ys ] = a //[ xs --> ys ].
  Proof.
    introv Hfr.
    rewrite subst_pvars_push. rewrite subst_id. eauto. eauto.
  Qed.
  (* Substituting a list of vars to itself is identity *)
  Lemma subst_pvars_append_same (a: A) zs xs ys:
    fresh (from_list ys) (length zs) zs ->
    subst_pvars (zs ++ xs) (zs ++ ys) a = (subst_pvars xs ys a).
  Proof.
    introv Hfr.
    induction zs.
    rewrite ! List.app_nil_l. eauto.
    rewrite <- ! List.app_comm_cons.
    rewrite <- IHzs.
    eapply subst_pvars_push_same.
    rewrite from_list_append. notin_solve. eapply from_list_fresh; eauto.
    simpl in Hfr; destruct Hfr; eauto.
  Qed.

  Lemma subst_pvars_one (a: A) x y :
    a //[ x::nil --> y::nil] = a /[x -> y].
  Proof.
    rewrite subst_pvars_push.
    rewrite subst_pvars_nil_l. auto. rewrite from_list_nil. notin_solve.
  Qed.

  Lemma subst_pvars_push_lr_ex (a: A) x xs y ys L:
    x \notin (from_list xs) ->
    exists v,
    v \notin ((from_list xs) \u (from_list ys) \u \{x} \u \{y} \u (fv a) \u L) /\
    a //[ x::xs --> y::ys ] = a /[x -> v] //[ xs --> ys] /[v -> y].
  Proof.
    introv Hnx.
    (* common proof start *)

    lets x0 f Heq0: (@subst_pvars_unfold_L a (x :: xs) (y :: ys) L).
    rewrite Heq0.

    lets: (fresh_length _ _ _ f).
    destruct x0.
    simpls. discriminate.

    exists v.

    lets x1 f0 Heq1: (subst_pvars_unfold (a /[x -> v]) xs ys).
    rewrite Heq1.

    simpls. inversions H2.
    destruct f.
    rewrite notin_union in H2.
    destruct H2.
    rewrite from_list_cons in H2.
    rewrite notin_union in H2.
    destruct H2. rewrite notin_singleton in H2.
    rewrite notin_union in H5.
    destruct H5.
    rewrite from_list_cons in H5.
    rewrite notin_union in H5.
    destruct H5. rewrite notin_singleton in H5.
    rewrite from_list_cons in H3.
    rewrite from_list_cons in H3.
    (* common proof end *)
    split.
    - notin_solve.
    - rewrite subst_vars_commut with (x2:=x) (y0:=v).
      eapply f_equal.
      eapply subst_vars_fresh_middle_change.
      fresh_solve.
      rewrite <- H4.
      eapply fresh_fv_subst_var; fresh_solve.
      fresh_solve.
      auto.
      eapply fresh_resize_length in H3.
      assert (fresh \{ x} (length x0) x0) by fresh_solve.
      eapply fresh_single_notin. eauto.
      auto.
  Qed.

  Lemma subst_pvars_push_lr (a: A) x xs y ys v:
    x \notin (from_list xs) ->
    v \notin ((from_list xs) \u (from_list ys) \u \{x} \u \{y} \u (fv a)) ->
    a //[x::xs --> y::ys] = a /[x -> v] //[ xs --> ys] /[v -> y].
  Proof.
    introv Hnx Hnv.

    lets x0 f Heq0: (@subst_pvars_unfold_L a (x :: xs) (y :: ys) \{v}).
    rewrite Heq0.

    lets: (fresh_length _ _ _ f).
    destruct x0.
    simpls. discriminate.

    replace a /[ x :: xs --> v0 :: x0] /[ v0 :: x0 --> y :: ys] with a /[ x :: xs --> v :: x0] /[ v :: x0 --> y :: ys].

    lets x1 f0 Heq1: (subst_pvars_unfold (a /[x -> v]) xs ys).
    rewrite Heq1.

    simpls. inversions H2.
    destruct f.
    rewrite notin_union in H2.
    destruct H2.
    rewrite from_list_cons in H2.
    rewrite notin_union in H2.
    destruct H2. rewrite notin_singleton in H2.
    rewrite notin_union in H5.
    destruct H5.
    rewrite from_list_cons in H5.
    rewrite notin_union in H5.
    destruct H5. rewrite notin_singleton in H5.
    rewrite from_list_cons in H3.
    rewrite from_list_cons in H3.
    (* common proof end *)
      rewrite subst_vars_commut with (x2:=x) (y0:=v).
      eapply f_equal.
      eapply subst_vars_fresh_middle_change.
      fresh_solve.
      rewrite <- H4.
      eapply fresh_fv_subst_var; fresh_solve.
      fresh_solve.
      auto.
      eapply fresh_resize_length in H3.
      assert (fresh \{ x} (length x0) x0) by fresh_solve.
      eapply fresh_single_notin. eauto.
      auto.
    eapply subst_vars_fresh_middle_change.
    rewrite from_list_cons in *.
    rewrite from_list_cons in *.
    fresh_solve; fresh_solve.
    fresh_solve; fresh_solve.
  Qed.

  Context {SOC: SubstOpenCommut AS}.

  Lemma subst_pvars_open_fresh_commut n x T xs ys:
      fresh \{x} (length xs) xs ->
      T /[#n -> x] //[ xs --> ys] = T //[ xs --> ys] /[#n -> x].
  Proof.
    introv Hfr.

    lets vs0 Hf0 Heq0: (@subst_pvars_unfold_L (open_rec n x T) xs ys (\{x} \u fv T)).
    rewrite Heq0.

    lets vs1 Hf1 Heq1: (@subst_pvars_unfold_L T xs ys \{x}).
    rewrite Heq1.

    rewrite subst_vars_open_fresh_commut by auto.
    rewrite subst_vars_open_fresh_commut.
    eapply f_equal.
    eapply subst_vars_fresh_middle_change.

    fresh_solve.
    fresh_solve.
    eapply fresh_resize_length in Hf0.
    fresh_solve.
  Qed.


  Lemma notin_fv_subst_pvars x xs ys (X: A):
    x \notin (fv X) ->
    x \notin (from_list ys) ->
    x \notin (fv X //[ xs --> ys]).
  Proof.
    edestruct (subst_pvars_unfold_L X xs ys \{x}) as [zs [Hfr Hsubst]].
    rewrite Hsubst.
    intros.
    eapply notin_fv_subst_vars.
    eapply notin_fv_subst_vars.
    eauto.
    repeat eapply fresh_union_r2 in Hfr. eapply fresh_resize_length in Hfr.  eapply from_list_fresh in Hfr.
    notin_solve.
    eauto.
  Qed.
  Lemma subst_pvars_open_rec_nth_commut m n xs ys x y T:
    fresh \{} (length xs) xs ->
    LibList.Nth n xs x ->
    LibList.Nth n ys y ->
    T /[#m -> x] //[ xs --> ys] = T //[ xs --> ys] /[#m -> y].
  Proof.
    introv Hfr Hnx Hny.
    reverts ys m T.
    induction Hnx; intros; inversions Hny.
    - simpls.
      destruct Hfr.

      forwards v Hnv Heq: (@subst_pvars_push_lr_ex (open_rec m x T) x l y l0 (fv T)).
      eapply fresh_single_notin. eapply fresh_union_r2. eauto.
      rewrite Heq.

      rewrite (subst_open_commut x). rewrite substi_eq_var.

      rewrite subst_pvars_open_fresh_commut.
      rewrite (subst_open_commut v).
      rewrite substi_eq_var.
      eapply f_equal.
      symmetry.
      eapply subst_pvars_push_lr.
      eapply fresh_single_notin. eapply fresh_union_r2. eauto.
      notin_solve.
      eapply fresh_from_list.
      notin_solve. eauto.
    - simpls.
      destruct Hfr.

      forwards v Hnv Heq: (@subst_pvars_push_lr_ex (open_rec m x T) y0 l y1 l0 (fv T)).
      eapply fresh_single_notin. eapply fresh_union_r2. eauto.
      rewrite Heq.

      assert (x <> y0). {
        assert(y0 \notin (from_list l)).
        eapply fresh_single_notin. eapply fresh_union_r2. eauto.
        eapply LibList.Nth_mem in Hnx. eapply list_mem_in_fromlist in Hnx. intro; subst. contradiction.
      }
      assert (y <> v). {
        repeat rewrite notin_union in Hnv. destructc Hnv as ? ? ?.
        eapply LibList.Nth_mem in H3. eapply list_mem_in_fromlist in H3.
        intro; subst. contradiction.
      }
      rewrite (subst_open_commut). rewrite substi_ne_var by auto.
      rewrite IHHnx; eauto.
      rewrite (subst_open_commut v).
      rewrite substi_ne_var by auto.
      eapply f_equal.
      symmetry.
      eapply subst_pvars_push_lr with (v:=v).
      eapply fresh_single_notin. eapply fresh_union_r2. eauto.
      notin_solve.
  Qed.

  Lemma reorder_subst (xs ys xs1 ys1: list var) ns n (T: A):
    fresh (fv T) (length xs) xs ->
    LibList.noduplicates ns ->
    list_reorder xs ns xs1 ->
    list_reorder ys ns ys1 ->
    T /[#n --> xs1] //[xs --> ys] = T /[#n --> ys1].
  Proof.
    introv Hfr Hnd Hre1 Hre2.
    reverts ys n T ys1.
    induction Hre1; intros; inversions Hre2.
    - simpls. eapply subst_pvars_fresh. auto.
    -
      rewrite <- open_vars_push.
      rewrite <- open_vars_push.
      erewrite  subst_pvars_open_rec_nth_commut; eauto.
      eapply f_equal.
      erewrite IHHre1; eauto.
      inversions Hnd. auto.
  Qed.

  Lemma open_vars_reorder_eq xs0 ys0 xs1 ys1 xs2 ys2 (T1 T2: A) n:
  (* It should be sufficient that xs1 not in T2, xs2 not in T1 *)
    list_reordering  xs0 ys0 xs1 ys1 xs2 ys2 ->
    fresh (fv T1 \u fv T2) (length xs0) xs0 ->
    T1 /[#n --> xs1] = T2 /[#n --> xs2] ->
    T1 /[#n --> ys1] = T2 /[#n --> ys2].
  Proof.
    introv Hre Hfr Heq.
    inversions Hre.

    rewrite <- reorder_subst with (xs:=xs0) (ys:=ys0) (xs1:=xs1) (ys1:=ys1) (ns:=ns1).
    rewrite <- reorder_subst with (xs:=xs0) (ys:=ys0) (xs1:=xs2) (ys1:=ys2) (ns:=ns2).

    eapply f_equal. auto.

    all: eauto.
  Qed.


End Subst_PVars.