Set Implicit Arguments.

Require Import LibExtra.
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties SyntaxClassLemmas SyntaxClassSubstLemmas.

Generalizable Variables A B.

(* Basic rewriting of substitution of lists of vars *)
Section SubstVars.
  Context `{AS: AbstractSyntax A}.

  (* Substituting single variable after a list of variables, is prepending it to the list *)
  Lemma subst_vars_push_left (x y: var) (xs ys: list var) (a: A):
    a /[xs --> ys] /[x -> y] = a /[ x::xs --> y::ys].
  Proof.
    simpl; auto.
  Qed.

  (* Substituting list of variables after a single variable, is appending it to the list *)
  Lemma subst_vars_push_right (x y: var) (xs ys: list var) (a:A):
    (length xs) = (length ys) -> (* If lengths were different, the variables would not align *)
    a /[x -> y] /[xs --> ys] = a /[xs++(x::nil) --> ys++(y::nil)].
  Proof.
    introv Hl.
    reverts ys.
    induction xs; intros; simpls.
    - destruct ys; try discriminate.
      simpls. auto.
    - destruct ys; try discriminate. simpls.
      inversions Hl.
      rewrite IHxs by auto.
      auto.
  Qed.
  (* Substituting no variables is identity *)
  Lemma subst_vars_nil_r (xs: list var) (a:A):
    a /[xs --> nil] = a.
  Proof.
    induction xs; simpls; auto.
  Qed.

  (* Substituting list of variables after another list, is appending the lists *)
  Lemma subst_vars_append (X: A) xs1 xs2 ys1 ys2:
    length xs1 = length ys1 ->
    length xs2 = length ys2 ->
    X /[ xs1 ++ xs2 --> ys1 ++ ys2] = X /[ xs2 --> ys2] /[ xs1 --> ys1].
  Proof.
    introv Hl1 Hl2.
    reverts ys1.
    induction xs1; intros.
    - destruct ys1; simpls; try discriminate. auto.
    - destruct ys1; simpls; try discriminate.
      eapply f_equal. eapply IHxs1. auto.
  Qed.
End SubstVars.

(* Substitution of fresh lists of vars *)
Section SubstFresh.
  Context `{SubstFresh A}.

  (* Substituting fresh variables is identity *)
  Lemma subst_vars_fresh xs ys (a: A):
    fresh (fv a) (length xs) xs ->
    a /[xs --> ys] = a.
  Proof.
    introv Hfr.
    revert dependent a.
    revert dependent ys.
    induction xs; intros; simpls.
    - auto.
    - destruct ys.
      + auto.
      + destruct Hfr.
        rewrite IHxs.
        rewrite subst_fresh; eauto.
        fresh_solve.
  Qed.

End SubstFresh.

(* Free variables after substitution of lists of vars *)
Section SubstVarsFv.
  Context `{AS: AbstractSyntax A}.
  Context {AFS: FvSubstCases AS}.

  (* Which variables may be free after substitution *)
  Lemma fv_subst_vars_subset a xs ys:
    (fv (a /[ xs --> ys])) \c ((fv a) \u (from_list ys)).
  Proof.
    introv.
    revert dependent ys.
    induction xs; intros.
    - simpls. rewrite in_union. auto.
    - simpls. destruct ys.
      + rewrite in_union. auto.
      + rewrite from_list_cons.
        pose proof (fv_subst_var_subset) as Hs.
        eapply Hs in H2.
        rewrite in_union in H2.
        destructecs_all.
        forwards: IHxs; eauto. rewrite ! in_union in *; destruct_all; auto.
        rewrite ! in_union in *; destruct_all; auto.
  Qed.
  (* A free variable will be free if it is not in domain *)
  Lemma in_fv_subst_vars x xs ys (X: A):
    x \in (fv X) ->
    x \notin (from_list xs) ->
    x \in (fv X /[xs --> ys]).
  Proof.
    introv HnX Hnys.
    revert dependent ys.
    induction xs; intros.
    - simpls. auto.
    - destruct ys.
      + simpls. auto.
      + simpls.
        rewrite from_list_cons, notin_union in Hnys.
        destructecs Hnys.
        rewrite notin_singleton in Hnys0.
        eapply in_fv_subst_var; auto.
  Qed.
  (* A non-free variable will not be free if it is not in codomain.
     Equivalent to fv_subst_vars_subset. *)
  Lemma notin_fv_subst_vars x xs ys (X: A):
    x \notin (fv X) ->
    x \notin (from_list ys) ->
    x \notin (fv X /[xs --> ys]).
  Proof.
      introv HnX Hnys.
      revert dependent ys.
      induction xs; intros.
      - simpls. auto.
      - destruct ys.
        + simpls. auto.
        + simpls.
          rewrite from_list_cons, notin_union in Hnys.
          destructec Hnys.
          rewrite notin_singleton in Hnys0.
          eapply notin_fv_subst_var; auto.
  Qed.
  (* A variable will not be free if it is in domain and not in codomain *)
  Lemma notin_fv_subst_vars_in x xs ys (X: A):
    x \in (from_list xs) ->
    x \notin (from_list ys) ->
    length xs = length ys ->
    x \notin (fv X /[xs --> ys]).
  Proof.
    introv Hnx Hnys Hl.
    revert dependent ys.
    induction xs; intros.
    - rewrite from_list_nil in Hnx. rewrite in_empty in Hnx. contradiction.
    - destruct ys.
      + simpls. discriminate.
      + simpls.
        rewrite from_list_cons, notin_union in Hnys.
        destructecs Hnys.
        rewrite notin_singleton in Hnys0.
        rewrite from_list_cons, in_union in Hnx.
        destructec Hnx.
        * rewrite in_singleton in Hnx. subst. eapply subst_notfv. auto.
        * eapply notin_fv_subst_var; auto.
  Qed.

  Lemma fresh_fv_subst_vars xs ys ws a:
    fresh ((fv a) \u from_list ys) (length ws) ws ->
    fresh (fv (a /[ xs --> ys])) (length ws) ws.
  Proof.
    introv Hfr.
    induction ws.
    - unfold fresh. simpls. auto.
    - simpls.
      destruct Hfr.
      fresh_solve.
      + eapply notin_fv_subst_vars; notin_solve.
      + fresh_solve.
        apply IHws. fresh_solve.
  Qed.
  Lemma fresh_fv_subst_vars_in xs ys zs (X: A):
    fresh (from_list ys) (length xs) xs ->
    (from_list xs) \c (from_list zs) ->
    length zs = length ys ->
    fresh (fv X /[zs --> ys]) (length xs) xs.
  Proof.
    introv Hfr Hsu Hl.
    reverts X.
    induction xs; intros.
    - eauto.
    - rewrite from_list_cons in *. inversions Hfr.
      constructor.
      eapply notin_fv_subst_vars_in.
      eapply subset_union_l1 in Hsu.
      apply -> subset_singleton in Hsu. eauto.
      eauto. eauto.
      eapply subset_union_l2 in Hsu.
      forwards: IHxs; eauto.
  Qed.
End SubstVarsFv.

Section OpenTwiceClosed.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {SOC: SubstOpenCommut AS}.
  Context {CA: Closing AS}.

  Lemma closed_subst_vars x y (T: A):
    closed T ->
    closed (T /[ x --> y]).
  Proof.
    introv Ho.
    reverts y .
    induction x; intros; simpls; auto; destruct y; auto.
    eapply closed_subst_var. intro.
    eapply IHx.
  Qed.
  Lemma closed0_subst_vars (T2: A) x y:
    (forall z : var, open z T2 = T2) ->
    (forall z : var, open z (T2 /[x --> y]) = (T2/[x --> y])).
  Proof.
    introv Ho.
    reverts y .
    induction x; intros; simpls; auto; destruct y; auto.
    eapply closed0_subst_var. intro.
    eapply IHx.
  Qed.

End OpenTwiceClosed.

Section SubstVars.
  Context `{AC: Closing A}.
  Context {ACS: CloseSubst AC} {AFC: FvClose AC}  {AFS: FvSubstCases AS}.
  Context {SC: SubstCommut AS}.
  Context {SI: SubstId AS}.
  Context {SF: SubstFresh AS}.

  Lemma subst_vars_swap2 (t: A) x1 x2 y1 y2:
    x2 <> x1 ->
    x2 <> y1 ->
    y2 <> x1 ->
    (t /[ x2 :: x1 :: nil --> y2 :: y1 :: nil]) = (t /[ x1 :: x2 :: nil --> y1 :: y2 :: nil]).
  Proof.
    intros.
    unfold subst_vars.
    rewrite subst_commut_ne; eauto.
  Qed.

  (* Substitution of a variable does not conflict with a list of variables if
     the source variable is not in any list and
     the target variable is not in the source list.
  *)
  Lemma subst_vars_commut x y ws zs (X:A):
    x \notin (from_list ws) ->
    x \notin (from_list zs) ->
    y \notin (from_list ws) ->
    X /[ws --> zs] /[x -> y] = X /[x -> y] /[ws --> zs].
  Proof.
    intros.
    reverts x y zs.
    induction ws; intros; simpls.
    - auto.
    -

      Ltac notin_simpl_hyp H :=
        rewrite !from_list_cons, !notin_union, !notin_singleton in H.

      notin_simpl_hyp H2.
      destructecs H2.

      notin_simpl_hyp H4.
      destructecs H4.

      destruct zs. auto.
      simpls.
      notin_simpl_hyp H3.
      destruct H3.
      rewrite subst_commut_ne; eauto.
      rewrite IHws; eauto.
  Qed.


 Lemma subst_vars_merge_to_same_in_l (X: A) xs y z:
    z \in (from_list xs) ->
    let ys := (LibList.make (length xs) y) in
    X /[z -> y] /[xs --> ys] = X /[xs --> ys].
  Proof.
    introv Hz. intros. (*unfold ys; clear ys. *)
    eapply list_in_fromlist_mem in Hz.
    eapply LibList.mem_Nth in Hz.
    destructecs Hz.
    forwards Hlt: LibList.Nth_inbound Hz.
    assert (LibList.length xs = LibList.length ys) as Hl. {
      
      forwards Hl: LibList.length_make (length xs) y. rewrite liblist_length. rewrite <- Hl. unfold ys.  rewrite liblist_length.   auto.
    } 
    forwards Htdxs: list_Nth_cut_take_drop xs Hz.
    forwards Htdys: list_Nth_cut_take_drop ys.
    eapply LibList.Nth_make. rewrite <- liblist_length. eauto.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. rewrite <- Hl. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (z :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. rewrite <- Hl. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }

    rewrite ! subst_vars_append by auto.
    rewrite <- ! subst_vars_push_left.
    fequal.
    assert (LibList.drop (S n) ys = LibList.make (length (LibList.drop (S n) xs) ) y) as Heqdr. {
      rewrite <- liblist_length.
      rewrite LibList.length_drop.
      unfold ys.
      rewrite list_make_drop. rewrite liblist_length. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. rewrite <- liblist_length. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }
    rewrite ! Heqdr.
    rewrite ! subst_vars_commut_to_same.
    rewrite subst_twice. auto.
  Qed.
  Lemma subst_vars_merge_to_same_in_r (X: A) xs y z:
    z \in (from_list xs) ->
    let ys := (LibList.make (length xs) y) in
    X /[xs --> ys] = X /[xs --> ys] /[z -> y].
  Proof.
    introv Hz. intros. (*unfold ys; clear ys. *)
    eapply list_in_fromlist_mem in Hz.
    eapply LibList.mem_Nth in Hz.
    destructecs Hz.
    forwards Hlt: LibList.Nth_inbound Hz.
    assert (LibList.length xs = LibList.length ys) as Hl. {
      
      forwards Hl: LibList.length_make (length xs) y. rewrite liblist_length. rewrite <- Hl. unfold ys.  rewrite liblist_length.   auto.
    } 
    forwards Htdxs: list_Nth_cut_take_drop xs Hz.
    forwards Htdys: list_Nth_cut_take_drop ys.
    eapply LibList.Nth_make. rewrite <- liblist_length. eauto.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. rewrite <- Hl. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (z :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. rewrite <- Hl. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }

    rewrite ! subst_vars_append by auto.
    rewrite <- ! subst_vars_push_left.
    assert (LibList.take n ys = LibList.make (length (LibList.take n xs) ) y) as Heqdr. {
      rewrite <- liblist_length.
      rewrite LibList.length_take.
      unfold ys.
      rewrite list_make_take. auto.
      rewrite <- LibNat.le_peano.
      eapply LibOrder.lt_to_le.
       rewrite <- liblist_length. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    rewrite ! Heqdr.
    rewrite ! subst_vars_commut_to_same.
    rewrite subst_twice. auto.
  Qed.


  
  Lemma subst_to_same_remove (X: A) xs y:
    (X /[xs --> (LibList.make (length xs) y)]) =     (X /[(LibList.remove y xs) --> (LibList.make (length (LibList.remove y xs)) y)]).
  Proof.
    induction xs.
    - simpls. auto.
    - destruct (classicT (a = y)).
      + subst. rewrite list_remove_cons_eq. rewrite <- IHxs.
        simpls. rewrite subst_id. auto.
      + rewrite list_remove_cons_ne by auto. simpls.
        rewrite IHxs. auto.
  Qed.

  Lemma subst_vars_in_to_same (X: A) xs y z x:
    z \in (from_list xs) ->
    let ys := (LibList.make (length xs) y) in
    X /[x -> z] /[xs --> ys] =
    X /[x -> y] /[xs --> ys]. (* To get ((X /[xs --> ys]) /[x -> y]), use subst_vars_commut_to_same *)
  Proof.
    introv Hz. intros. (*unfold ys; clear ys. *)

    destruct (classicT(z = y)) as [|Hzy].
    subst. auto.

    destruct (classicT(z = x)) as [|Hzx].
    subst. rewrite subst_id. unfold ys. rewrite subst_vars_merge_to_same_in_l; auto.

    eapply list_in_fromlist_mem in Hz.
    eapply list_mem_Nth_last in Hz.
    destructecs Hz.
    forwards Hlt: LibList.Nth_inbound Hz0.
    assert (LibList.length xs = LibList.length ys) as Hl. {
      
      forwards Hl: LibList.length_make (length xs) y. rewrite liblist_length. rewrite <- Hl. unfold ys.  rewrite liblist_length.   auto.
    } 
    forwards Htdxs: list_Nth_cut_take_drop xs Hz0.
    forwards Htdys: list_Nth_cut_take_drop ys.
    eapply LibList.Nth_make. rewrite <- liblist_length. eauto.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. rewrite <- Hl. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (z :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. rewrite <- Hl. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }

    rewrite ! subst_vars_append by auto.
    rewrite <- ! subst_vars_push_left.

    unfold ys at 2 4.
  
    assert (S n <= length xs). {
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano.
      rewrite <- liblist_length. auto.
    }
    
    rewrite list_make_drop by auto.
    rewrite <- liblist_length.
    rewrite <- LibList.length_drop by (rewrite liblist_length, LibNat.le_peano; auto).
    rewrite liblist_length.
    rewrite ! subst_to_same_remove with (xs:=LibList.drop (S n) xs).

    assert (z \notin from_list (LibList.remove y (LibList.drop (S n) xs))). {
      eapply notin_from_list_remove_from_list. auto.
    }
    assert (z \notin from_list (LibList.make (length (LibList.remove y (LibList.drop (S n) xs))) y)). { 
      eapply notin_from_list_make.
      auto.
    }
    assert (y \notin from_list (LibList.remove y (LibList.drop (S n) xs))). {
      eapply notin_from_list_remove.
    }

    rewrite ! subst_vars_commut with (x:=z) (y:= y) by auto.

    rewrite subst_commut with (x0:=z) by auto. cases_if.
    destruct (classicT (x = y)). 
    * subst. rewrite 2 subst_id. auto.
    * rewrite subst_commut with (w:=x) by auto. cases_if. auto.
  Qed.

  Lemma subst_vars_neutralize xs ys (X:A) z:
    (length xs) = (length ys) ->
    disjoint (fv X) (from_list ys) ->
    z \notin (from_list ys) ->
    let zs := (LibList.make (length xs) z) in
    X /[xs --> ys] /[ys --> zs] = X/[xs --> zs].
  Proof.
    introv Hlx HfvX Hz. intros; unfold zs in *. clear zs.
    reverts ys.
    induction xs; intros; destruct ys; simpls; try discriminate.
    - auto.
    - inversions Hlx.
      rewrite from_list_cons in HfvX.
      rewrite from_list_cons in Hz.
      eapply disjoint_union_r_inv in HfvX.
      destructecs HfvX.
      eapply disjoint_singleton_inv_r in HfvX0.

      rewrite <- (IHxs ys) by auto.
      rewrite ! H3.
      rewrite subst_vars_commut_to_same.

      destruct (fv_decide v (from_list ys)).
      + rewrite subst_vars_merge_to_same_in_l by auto.
        rewrite subst_vars_in_to_same by auto.
        rewrite subst_vars_commut_to_same. auto.
      + destruct (classicT(a=v)).
        * subst. rewrite subst_id.
          rewrite subst_vars_commut_to_same. auto.
        *
          rewrite subst_merge.
          rewrite subst_vars_commut_to_same. auto. auto.
          eapply notin_fv_subst_vars; auto.
  Qed.

  Lemma subst_vars_merge xs ys zs (X:A):
    (length zs) = (length ys) ->
    (length xs) = (length ys) ->
    fresh (fv X \u (from_list xs) \u (from_list zs)) (length ys) ys ->
    fresh ((from_list zs)) (length xs) xs ->
    X /[xs --> ys] /[ys --> zs] = X /[xs --> zs].
  Proof.
    introv Hlz Hlx Hfry Hfrx.
    reverts ys zs.
    induction xs; intros; destruct ys; destruct zs; simpls; try discriminate.
    - auto.
    - rewrite ? from_list_cons in *.
      destruct_all.
      rewrite <- subst_vars_commut with (x:=a) (y:= v).
      rewrite subst_merge.
      eapply f_equal.
      eapply IHxs; eauto.
      all: notin_solve.
      eapply notin_fv_subst_vars; eauto.
      eapply notin_fv_subst_vars; eauto.
      all: eapply fresh_single_notin with (n:=(length ys)); fresh_solve.
  Qed.
  Lemma subst_vars_merge_same xs ys (X:A):
    (length xs) = (length ys) ->
    fresh (fv X \u (from_list xs)) (length ys) ys ->
    X /[xs --> ys] /[ys --> xs] = X.
  Proof.
    introv Hlx Hfry.
    reverts ys.
    induction xs; intros; destruct ys; simpls; try discriminate.
    - auto.
    - rewrite ? from_list_cons in *.
      destruct_all.

      (* Apply induction *)
      inversions Hlx.
      forwards Hind: IHxs H5; eauto.

      (* There is no freshness condition on xs, so xs can contain duplicates*)
      destruct (fv_decide a (from_list xs)).
      + (* a occurs in xs *)
        rewrite (subst_fresh v).
        rewrite Hind.
        rewrite (subst_fresh).
        auto.
        notin_solve.
        eapply notin_fv_subst_vars_in; eauto. notin_solve. eapply from_list_fresh; eauto.
     + (* does not occur *)
      rewrite <- subst_vars_commut with (x:=a) (y:= v).
      rewrite Hind.
      rewrite subst_merge_same; eauto.
      eapply from_list_fresh; eauto.
      auto.
      eapply from_list_fresh; eauto.
  Qed.
  Lemma subst_vars_fv_subset_E xs a ys E:
    (fv a) \c ((from_list xs) \u E) ->
    fresh (from_list xs) (length ys) ys ->
    length xs = length ys ->
    (fv (a /[xs --> ys])) \c ((from_list ys) \u E).
  Proof.
    introv Hfva Hfr Hl.

    reverts ys a E. induction xs; intros; destruct ys; try discriminate.
    - rewrite subst_vars_nil_r. auto.
    - rewrite from_list_cons in *.
      simpls.
      destruct Hfr.
      rewrite notin_union in H2. destruct H2.
      rewrite notin_singleton in H2.
      inversions Hl.

      destruct (fv_decide a (from_list xs)).
      + assert( a \notin (fv (a0 /[ xs --> ys]))). eapply notin_fv_subst_vars_in; eauto. eapply from_list_fresh; eauto.
        rewrite subst_fresh by auto. intro. intro. rewrite ! in_union. rewrite or_assoc. right. rewrite <- in_union. eapply IHxs; eauto. intro. intro. eapply Hfva in H8. rewrite ! in_union in H8. destruct H8.  destruct H8.
        rewrite in_singleton in H8. subst. rewrite in_union. auto. rewrite in_union. auto. rewrite in_union. auto.
      +

      rewrite subst_vars_commut; eauto. 2: eapply from_list_fresh; eauto. intro. intro.
      rewrite union_comm.
      rewrite union_assoc.
      rewrite union_comm.
      eapply IHxs; eauto. intro. intro.

      destruct (fv_decide x0 (fv a0)).
      * (* x0 \in fv a0 *)
        eapply Hfva in m. rewrite <- union_assoc in m. rewrite in_union in m. destruct m. rewrite in_singleton in H8.
        (* x0 = a *)
        subst. eapply subst_notfv in H7; eauto. contradiction.
        (* x0 <> a *)
        rewrite union_assoc. rewrite in_union. auto.
     * (* x0 \notin fv a0 *)
       destruct (classicT(x0 = v)).
       (* x0 = v *)
       subst. rewrite union_assoc. rewrite in_union. right. rewrite in_singleton. auto.
       (* x0 <> v *)
       false. eapply notin_fv_subst_var; eauto.
  Qed.
  (* Same as previous lemma, but the freshness is swapped *)
  Lemma subst_vars_fv_subset_E' xs a ys E:
    (fv a) \c ((from_list xs) \u E) ->
    fresh (from_list ys) (length xs) xs ->
    length xs = length ys ->
    (fv (a /[xs --> ys])) \c ((from_list ys) \u E).
  Proof.
    introv Hfva Hfr Hl.

    reverts ys a E. induction xs; intros; destruct ys; try discriminate.
    - rewrite subst_vars_nil_r. auto.
    - rewrite from_list_cons in *.
      simpls.
      destruct Hfr.
      rewrite notin_union in H2. destruct H2.
      rewrite notin_singleton in H2.
      inversions Hl.

      destruct (fv_decide a (from_list xs)).
      + assert( a \notin (fv (a0 /[ xs --> ys]))). eapply notin_fv_subst_vars_in; eauto.
        rewrite subst_fresh by auto. intro. intro. rewrite ! in_union. rewrite or_assoc. right. rewrite <- in_union. eapply IHxs; eauto. intro. intro. eapply Hfva in H8. rewrite ! in_union in H8. destruct H8.  destruct H8.
        rewrite in_singleton in H8. subst. rewrite in_union. auto. rewrite in_union. auto. rewrite in_union. auto.
      +

      rewrite subst_vars_commut; eauto. 2: eapply from_list_fresh; eauto. intro. intro.
      rewrite union_comm.
      rewrite union_assoc.
      rewrite union_comm.
      eapply IHxs; eauto. intro. intro.

      destruct (fv_decide x0 (fv a0)).
      * (* x0 \in fv a0 *)
        eapply Hfva in m. rewrite <- union_assoc in m. rewrite in_union in m. destruct m. rewrite in_singleton in H8.
        (* x0 = a *)
        subst. eapply subst_notfv in H7; eauto. contradiction.
        (* x0 <> a *)
        rewrite union_assoc. rewrite in_union. auto.
     * (* x0 \notin fv a0 *)
       destruct (classicT(x0 = v)).
       (* x0 = v *)
       subst. rewrite union_assoc. rewrite in_union. right. rewrite in_singleton. auto.
       (* x0 <> v *)
       false. eapply notin_fv_subst_var; eauto.
  Qed.
  Lemma subst_vars_fv_subset xs a ys:
    (fv a) \c (from_list xs) ->
    fresh (from_list xs) (length ys) ys ->
    length xs = length ys ->
    (fv (a /[xs --> ys])) \c (from_list ys).
  Proof.
    intros. rewrite <- union_empty_r in H2.
    forwards: subst_vars_fv_subset_E; eauto.
    rewrite union_empty_r in H5. eauto.
  Qed.
  Lemma subst_vars_fv_subset' xs a ys:
    (fv a) \c (from_list xs) ->
    fresh (from_list ys) (length xs) xs ->
    length xs = length ys ->
    (fv (a /[xs --> ys])) \c (from_list ys).
  Proof.
    intros. rewrite <- union_empty_r in H2.
    forwards: subst_vars_fv_subset_E'; eauto.
    rewrite union_empty_r in H5. eauto.
  Qed.
  Lemma subst_vars_fresh_exists (a: A) xs:
    (fv a) \c (from_list xs) ->
    exists b ys,
    fresh (from_list xs) (length ys) ys /\
    length xs = length ys /\
    (fv b) \c (from_list ys) /\
    b /[ys --> xs] = a.
  Proof.
    introv Hfv.
    destruct (var_freshes (from_list xs)  (length xs) ) as [ys Hfr].
    lets Hl: fresh_length Hfr.
    eapply fresh_resize_length in Hfr.
    exists (a /[xs --> ys]) ys.
    splits.
    - eauto.
    - eauto.
    - eapply subst_vars_fv_subset; eauto.
    - rewrite subst_vars_merge_same; eauto.
      eapply fresh_subset; eauto. intro. intro. rewrite in_union in H2. destruct H2; auto.
  Qed.

  Lemma subst_vars_permutation (X: A) xs1 ys1 xs2 ys2 ns:
    list_reorder xs1 ns xs2 ->
    list_reorder ys1 ns ys2 ->
    LibList.noduplicates ns ->
    length xs1 = length ns ->
    length ys1 = length ns ->
    fresh (from_list ys1) (length xs1) xs1 ->
    (X /[xs1 --> ys1]) = (X /[xs2 --> ys2]).
  Proof.
    eapply (list_permutation_ind (  fun (zs1 ws1 zs2 ws2: list var)=>   fresh (from_list ws1) (length zs1) zs1 ->
      (X /[zs1 --> ws1]) = (X /[zs2 --> ws2]))); simpls; eauto.
    introv Hl1 Hl2 Hind Hfr.

    rewrite ! from_list_append in *.
    rewrite ! from_list_cons in *.

    rewrite subst_vars_append.
    rewrite <- subst_vars_push_left.
    rewrite <- subst_vars_commut.
    eapply f_equal.
    rewrite <- subst_vars_append.
    eapply Hind.
    +
      eapply fresh_insert_inv in Hfr. auto.
    + auto.
    + auto.
    + forwards: fresh_append_fresh_r. eapply fresh_empty. eauto. notin_solve.
    + forwards: fresh_append_r; eauto.
    + forwards: fresh_append_l; eauto. eapply fresh_single_notin with (n:=length zsl). fresh_solve.
    + auto.
    + simpl. eapply f_equal. auto.
  Qed.
 Lemma subst_vars_twice_Nth_l (X: A) xs ys x y n:
    LibList.Nth n xs x ->
    LibList.Nth n ys y ->
    length xs = length ys ->
    x \notin from_list (LibList.drop (S n) xs) ->
    disjoint (from_list xs) (from_list ys) ->
    X /[x -> y] /[xs --> ys] = X /[xs --> ys].
  Proof.
    introv Hxs Hys Hl Hxxs Hxsys.
    forwards Hltx: LibList.Nth_inbound Hxs.
    forwards Hlty: LibList.Nth_inbound Hys.
    assert (LibList.length xs = LibList.length ys) as Hl'. {
      rewrite liblist_length. auto.
    }
    forwards Htdxs: list_Nth_cut_take_drop xs Hxs.
    forwards Htdys: list_Nth_cut_take_drop ys Hys.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (x :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl'. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }
    assert (x \notin from_list (LibList.drop (S n) ys)). {
      eapply notin_from_list_drop_from_list. 
      eapply disjoint_in_notin. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    assert (y \notin from_list (LibList.drop (S n) xs)). {
      eapply notin_from_list_drop_from_list. 
      eapply disjoint_in_notin. eapply disjoint_comm. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    rewrite ! subst_vars_append by auto.
    rewrite <- ! subst_vars_push_left.
    rewrite ! subst_vars_commut by auto.
    rewrite subst_twice. auto.
  Qed.


 Lemma subst_vars_twice_Nth_r (X: A) xs ys x y n:
    LibList.Nth n xs x ->
    LibList.Nth n ys y ->
    length xs = length ys ->
    x \notin from_list (LibList.take n xs) ->
    disjoint (from_list xs) (from_list ys) ->
    X /[xs --> ys] /[x -> y] = X /[xs --> ys].
  Proof.
    introv Hxs Hys Hl Hxxs Hxsys.
    forwards Hltx: LibList.Nth_inbound Hxs.
    forwards Hlty: LibList.Nth_inbound Hys.
    assert (LibList.length xs = LibList.length ys) as Hl'. {
      rewrite liblist_length. auto.
    }
    forwards Htdxs: list_Nth_cut_take_drop xs Hxs.
    forwards Htdys: list_Nth_cut_take_drop ys Hys.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (x :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl'. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }
    assert (x \notin from_list (LibList.take n ys)). {
      eapply notin_from_list_take_from_list. 
      eapply disjoint_in_notin. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    assert (y \notin from_list (LibList.take n xs)). {
      eapply notin_from_list_take_from_list. 
      eapply disjoint_in_notin. eapply disjoint_comm. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    rewrite ! subst_vars_append by auto.
    rewrite <- ! subst_vars_push_left.
    rewrite ! subst_vars_commut by auto.
    rewrite subst_twice. auto.
  Qed.
  Lemma subst_vars_commut_Nth_r (X: A) xs ys x y n:
    LibList.Nth n xs x ->
    LibList.Nth n ys y ->
    length xs = length ys ->
    x \notin from_list (LibList.take n xs) ->
    disjoint (from_list xs) (from_list ys) ->
    X /[list_removeat n xs --> list_removeat n ys] /[x -> y] = X /[xs --> ys].
  Proof.
    introv Hxs Hys Hl Hxxs Hxsys.
    forwards Hltx: LibList.Nth_inbound Hxs.
    forwards Hlty: LibList.Nth_inbound Hys.
    assert (LibList.length xs = LibList.length ys) as Hl'. {
      rewrite liblist_length. auto.
    }
    forwards Htdxs: list_Nth_cut_take_drop xs Hxs.
    forwards Htdys: list_Nth_cut_take_drop ys Hys.
    rewrite Htdxs, Htdys.
    
    assert (length (LibList.take n xs) = length (LibList.take n ys)). {
      rewrite <- liblist_length.
      rewrite ! LibList.length_take. auto.
      eapply LibOrder.lt_to_le. auto.
      eapply LibOrder.lt_to_le. auto.
    }
    assert (length (x :: LibList.drop (S n) xs) = length (y :: LibList.drop (S n) ys)). {
      rewrite !list_length_cons.
      fequal.
      rewrite <- liblist_length.
      rewrite ! LibList.length_drop.
      rewrite Hl'. auto.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
      rewrite LibNat.le_peano.
      eapply Lt.lt_le_S.
      rewrite <- LibNat.lt_peano. auto.
    }
    assert (x \notin from_list (LibList.take n ys)). {
      eapply notin_from_list_take_from_list. 
      eapply disjoint_in_notin. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    assert (y \notin from_list (LibList.take n xs)). {
      eapply notin_from_list_take_from_list. 
      eapply disjoint_in_notin. eapply disjoint_comm. eauto. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    }
    rewrite ! subst_vars_append; rewrite <- ? Htdxs, <- ? Htdys; auto.
    rewrite <- ! subst_vars_push_left.
    rewrite ! subst_vars_commut by auto.
    auto.
  Qed.


  Lemma subst_vars_append_fresh_l (t: A) xs1 xs2 ys1 ys2:
    fresh (fv t) (length (xs2)) (xs2) ->
    length xs1 = length ys1 ->
    length xs2 = length ys2 ->
    t /[ xs1++xs2 --> ys1++ys2] = t /[ xs1 --> ys1].
  Proof.
    introv Hfrx1y2 Hl1 Hl2.
    rewrite subst_vars_append by auto.
    rewrite subst_vars_fresh with (xs:=xs2).
    auto.
    auto.
  Qed.
  Lemma subst_vars_append_fresh_r (t: A) xs1 xs2 ys1 ys2:
    fresh (from_list ys2) (length (xs1)) (xs1) ->
    length xs1 = length ys1 ->
    length xs2 = length ys2 ->
    ((from_list xs1) \n fv t) \c (from_list xs2) ->
    t /[ xs1++xs2 --> ys1++ys2] = t /[ xs2 --> ys2].
  Proof.
    introv Hfrx1y2 Hl1 Hl2 Hfv.
    rewrite subst_vars_append by auto.
    rewrite subst_vars_fresh with (xs:=xs1).
    auto.
    eapply fresh_forall_in_from_list. fresh_solve.
    introv Hinx.
    forwards: fresh_in_from_list_notin; fweauto.
    destruct (fv_decide x (fv t)).
    - forwards: Hfv x. rewrite in_inter; auto.
      eapply notin_fv_subst_vars_in; eauto.
    - eapply notin_fv_subst_vars; eauto.
  Qed.

  Lemma subst_vars_cover_reorder (X: A) xs1 ys1 xs2 ys2 ns:
    list_reorder xs1 ns xs2 ->
    list_reorder ys1 ns ys2 ->
    length xs1 = length ys1 ->
    list_cover_interval ns (length xs1) ->
    fresh (from_list ys1) (length xs1) xs1 ->
    X /[xs1 --> ys1] = X /[xs2 --> ys2].
  Proof.
    introv Hrex Hrey Hlxy Hci Hfr.
    reverts xs1 xs2 ys1 ys2.
    remember (length ns) as m.
    reverts ns.
    induction m; intros.
    - destruct ns; inversions Heqm.
      inversions Hrex. inversions Hrey.
      eapply list_cover_interval_nil_inv in Hci.
      destruct xs1; inversions Hci.
      destruct ys1; inversions Hlxy.
      auto.
    - destruct ns; inversions Heqm.
      inversions Hrex. inversions Hrey.
      simpls.

      assert (disjoint (from_list xs1) (from_list ys1)) as Hdi. {
        apply disjoint_comm.
        eapply fresh_disjoint; eauto.
      }
      assert (LibOrder.lt n (LibList.length xs1)) as Hlt. {
        eapply LibList.Nth_inbound; eauto.
      }
      assert (LibOrder.le n (LibList.length xs1)) as Hle. {
        eapply LibOrder.lt_to_le; auto.
      }

      destruct (classicT (n \in (from_list ns))).
      + (* Duplicate variable *)
        (* Find the closest duplicate and eliminate it *)
        eapply list_in_fromlist_mem in m.
        forwards Hmf: list_mem_Nth_first m.
        destructecs Hmf.
        forwards Hys: list_reorder_nth_xs_nth_ys H5 H7 Hmf0.
        forwards Hys0: list_reorder_nth_xs_nth_ys H6 H9 Hmf0.
        forwards Hr: subst_vars_twice_Nth_r Hys Hys0.
        *
          forwards: list_reorder_lengths H5.
          forwards: list_reorder_lengths H6.
          rewrite H2, H3. auto.
        *
          forwards Hren0: list_reorder_take n0 H5.
          intro. apply Hmf.
          eapply list_in_fromlist_mem in H2.
          forwards Hmfy: list_mem_Nth_first H2. destructecs Hmfy.
          forwards Hnn0: list_reorder_nth_Nth Hren0 Hmfy0.
          destructecs Hnn0.
          forwards: LibList.noduplicates_Nth_same_inv Hnn0 H7.
          eapply fresh_noduplicates; eauto.
          subst. eapply list_mem_in_fromlist.
          eapply LibList.Nth_mem; eauto.
        * eapply list_reorder_disjoint; eauto.
        * rewrite Hr.
          eapply IHm; eauto.
          eapply list_cover_interval_cons_mem_inv; eauto.

      + (* Non-duplicate variable *)
          assert(length (LibList.take n xs1) = n) as Htlx. {
            rewrite <- liblist_length.
            rewrite LibList.length_take; auto.
          }
          assert(length (LibList.take n ys1) = n) as Htly. {
            rewrite <- liblist_length.
            rewrite LibList.length_take; auto.
            rewrite <- liblist_length in Hlxy. rewrite <- Hlxy. auto.
          }
          erewrite <- subst_vars_commut_Nth_r; rweauto.
          fequal.
          eapply IHm. 
          3: {
            eapply list_reorder_remove_middle_Nth.
            eauto. eauto.
            intro. eapply n0. 
            eapply list_mem_in_fromlist; eauto.
          }
          5: {
            eapply list_reorder_remove_middle_Nth.
            eauto. eauto.
            intro. eapply n0. eapply list_mem_in_fromlist; eauto.
          }
          * rewrite <- liblist_length. rewrite LibList.length_map. auto.
          * rewrite list_removeat_length. eapply list_cover_interval_remove_n; eauto. rewrite <- liblist_length, <- LibNat.lt_peano; auto.
          * rewrite 2 list_removeat_length. rewrite Hlxy; auto. rewrite <- Hlxy. rewrite <- liblist_length, <- LibNat.lt_peano; auto. rewrite <- liblist_length, <- LibNat.lt_peano; auto.
          * eapply fresh_removeat_rl; eauto. rewrite <- liblist_length, <- LibNat.lt_peano. eapply LibList.Nth_inbound; eauto.
          * eapply list_fresh_Nth_notin_take; eauto.
  Qed.

  Lemma subst_vars_split_subset_r (t: A) xs ys  ns1 ns2 xs1 xs2 ys1 ys2:
    list_cover_interval (ns1 ++ ns2) (length xs) ->
    fresh (from_list ys) (length xs) xs ->
    list_reorder xs (ns1 ++ ns2) (xs1 ++ xs2) ->
    list_reorder ys (ns1 ++ ns2) (ys1 ++ ys2) ->
    length xs1 = length ns1 ->
    length ys1 = length ns1 ->
    length ys = length xs ->
    ((from_list xs1) \n fv t) \c (from_list xs2) ->
    LibList.noduplicates ns1 ->
    (t /[ xs --> ys]) = (t /[ xs2 --> ys2]).
  Proof.
    introv Hci Hfrx Hrex Hrey Hlxs Hlys Hlxy Hfv.
    introv Hnd1.
    forwards Heq: (subst_vars_cover_reorder t) Hrex Hrey Hci; eauto.
    rewrite Heq.
    eapply subst_vars_append_fresh_r.
    - forwards Hrex1: list_reorder_app_case Hrex. auto.
      forwards Hrey1: list_reorder_app_case Hrey. auto.
      destructecs Hrex1.
      destructecs Hrey1.
      eapply fresh_reorder; eauto.
      eapply fresh_subset.
      eapply list_reorder_from_list_subset; eauto.
      eauto.
    - rewrite Hlxs; auto.
    - forwards Hrex1: list_reorder_app_case Hrex. auto.
      forwards Hrey1: list_reorder_app_case Hrey. auto.
      destructecs Hrex1.
      destructecs Hrey1.
      forwards Hlx1: list_reorder_lengths Hrex1.
      forwards Hly1: list_reorder_lengths Hrey1.
      rewrite Hlx1, Hly1. auto. 
    - auto.
  Qed.

  Lemma subst_vars_reorder_fv_inter (t: A) xs ys  ns2 xs2 ys2:
    fresh (from_list ys) (length xs) xs ->
    list_reorder xs (ns2) (xs2) ->
    list_reorder ys (ns2) (ys2) ->
    length ys = length xs ->
    (from_list xs2) = (from_list xs \n fv t) ->
    LibList.noduplicates ns2 ->
    (t /[ xs --> ys]) = (t /[ xs2 --> ys2]).
  Proof.
    introv Hfrx Hrex Hrey Hlxy Hfv Hnd.

    lets Hrr: list_reorder_refl_n (length xs).
    destructc Hrr as nsr Hndr Hlr Hgtr Hrr.
    forwards Hrrx: Hrr xs. auto.
    forwards Hrry: Hrr ys. auto.

    forwards Hrax: list_reorder_app Hrrx Hrex .
    forwards Hray: list_reorder_app  Hrry Hrey.
   
    forwards: subst_vars_split_subset_r t Hrax Hray.
    eapply list_cover_interval_app_l.
    rewrite <- Hlr.
    eapply list_noduplicates_gt_cover_interval; eauto.
    rewrite Hlr. eauto.
    auto.
    eapply list_reorder_lengths; eauto.
    eapply list_reorder_lengths; eauto.
    auto.
    rewrite <- Hfv. eapply subset_refl.
    auto.
    auto.
  Qed.

  Lemma subst_vars_reorder_fv_inter_exists (t: A) xs :
    LibList.noduplicates xs ->
    exists ns xs1,
    list_reorder xs ns xs1 /\
    LibList.noduplicates ns /\ 
    from_list xs1 = fv t \n from_list xs /\(
    forall ys,
    length ys = length xs ->
    fresh (from_list ys) (length xs) xs ->
    exists ys1,
    list_reorder ys ns ys1 /\
    (t /[ xs --> ys]) = (t /[ xs1 --> ys1])).
  Proof.
    introv Hnd.
    forwards Hret: list_reorder_in_set xs (fv t). auto.
    destructecs Hret.

    exists ns ys. splits.
    - auto.
    - auto.
    - auto.
    - introv Hl Hfr.

      forwards Hre11: list_reorder_same_length_exists ys0 Hret0. auto.
      destructc Hre11 as ys11 Hre11.

      forwards: subst_vars_reorder_fv_inter Hret0 Hre11; eauto.
      rewrite inter_comm; eauto.
  Qed.
  Lemma subst_vars_reorder_fv_inter_2 (t: A) xs ys1 ys2 :
    length ys1 = length xs ->
    length ys2 = length xs ->
    fresh (from_list ys1) (length xs) xs ->
    fresh (from_list ys2) (length xs) xs ->
    (t /[ xs --> ys1]) = (t /[ xs --> ys2]) ->
    exists ns xs1 ys11 ys21,
    list_reorder xs ns xs1 /\
    LibList.noduplicates ns /\ 
    from_list xs1 = fv t \n from_list xs /\
    list_reorder ys1 ns ys11 /\
    list_reorder ys2 ns ys21 /\
    (t /[ xs1 --> ys11]) = (t /[ xs1 --> ys21]).
  Proof.
    introv l1 Hl2 Hfr1 Hfr2 He.

    forwards Hsrt: subst_vars_reorder_fv_inter_exists t. eapply fresh_noduplicates; eauto.
    destructecs Hsrt.  
    forwards Hstr1: Hsrt ys1; fweauto.
    forwards Hstr2: Hsrt ys2; fweauto.
    destructecs Hstr1.
    destructecs Hstr2.

    exists ns xs1 ys0 ys3.  splits; eauto.
    rewrite <- Hstr1, <- Hstr2. auto.
  Qed.

  (* Applies induction step in proofs of SubstVarsEqInv *)
  Lemma subst_vars_eq_inv_ind_step (t: A) xs (ys1 ys2: list var):
    (forall xs ys1 ys2, SubstVarsEqInvOf xs ys1 ys2 t) ->
    fresh (from_list ys1) (length xs) xs ->
    fresh (from_list ys2) (length xs) xs ->
    length ys1 = length xs ->
    length ys2 = length xs ->
    (t /[ xs --> ys1]) = (t /[ xs --> ys2]) ->
    exists ns xs1 ys11,
    list_reorder xs ns xs1 /\
    LibList.noduplicates ns /\ 
    from_list xs1 = fv t \n from_list xs /\
    list_reorder ys1 ns ys11 /\
    list_reorder ys2 ns ys11.
  Proof.
    introv Hind Hfr1 Hfr2 Hl1 Hl2 He.

    forwards Hsrt: subst_vars_reorder_fv_inter_2 t Hfr1 Hfr2; fweauto.
    destructecs Hsrt.
    forwards: Hind Hsrt. eapply fresh_reorder; eauto. eapply fresh_subset. eapply subset_union_2.
    eapply list_reorder_from_list_subset; eauto.
    eapply list_reorder_from_list_subset; eauto. eauto.

    eapply eq_trans. eapply list_reorder_lengths; eauto. symmetry. eapply list_reorder_lengths; eauto. 
    eapply eq_trans. eapply list_reorder_lengths; eauto. symmetry. eapply list_reorder_lengths; eauto.
    rewrite Hsrt2. eapply inter_subset_l. subst.
    
    exists ns xs1 ys21. splits; eauto.
  Qed.
  (* Same as subst_vars_eq_inv_ind_step, but applies already proven SubstVarsEqInv *)
  Lemma subst_vars_eq_inv_step {SVE: SubstVarsEqInv AS} (t: A) xs (ys1 ys2: list var):
    fresh (from_list ys1) (length xs) xs ->
    fresh (from_list ys2) (length xs) xs ->
    length ys1 = length xs ->
    length ys2 = length xs ->
    (t /[ xs --> ys1]) = (t /[ xs --> ys2]) ->
    exists ns xs1 ys11,
    list_reorder xs ns xs1 /\
    LibList.noduplicates ns /\ 
    from_list xs1 = fv t \n from_list xs /\
    list_reorder ys1 ns ys11 /\
    list_reorder ys2 ns ys11.
  Proof.
    introv Hfr1 Hfr2 Hl1 Hl2 He.

    forwards Hsrt: subst_vars_reorder_fv_inter_2 t Hfr1 Hfr2; fweauto.
    destructecs Hsrt.
    forwards: subst_vars_eq_inv Hsrt. eapply fresh_reorder; eauto. eapply fresh_subset. eapply subset_union_2.
    eapply list_reorder_from_list_subset; eauto.
    eapply list_reorder_from_list_subset; eauto. eauto.

    eapply eq_trans. eapply list_reorder_lengths; eauto. symmetry. eapply list_reorder_lengths; eauto. 
    eapply eq_trans. eapply list_reorder_lengths; eauto. symmetry. eapply list_reorder_lengths; eauto.
    rewrite Hsrt2. eapply inter_subset_l. subst.
    
    exists ns xs1 ys21. splits; eauto.
  Qed.

 (* Parallel substitution uses a list of fresh vars in the middle.
     This shows that this list can be replaced by any list of fresh vars *)
  Lemma subst_vars_fresh_middle_change_ind xs1 ys1 zs1 ws1 xs2 ys2 zs2 (a: A):
    fresh (from_list(xs1) \u from_list(xs2) \u from_list(ys1) \u from_list(ys2) \u fv a) (length (ys1++ys2)) (zs1++zs2) ->
    fresh (from_list(xs1) \u from_list(xs2) \u from_list(ys1) \u from_list(ys2) \u from_list(zs1) \u from_list(zs2) \u fv a) (length ys1) ws1 -> (* Using length ys here to make ws and ys same length *)
    (length zs1) = (length ys1) ->
    (length xs2) = (length zs2) ->
    (a /[xs1 --> zs1] /[xs2 --> zs2] /[ zs1 --> ys1] /[zs2 --> ys2]) = (a /[xs1 --> ws1] /[xs2 --> zs2] /[ ws1 --> ys1] /[zs2 --> ys2]).
  Proof.
    introv Hfrz Hfrw Hlzy Hlzx.


    reverts a xs1 ys1 zs1.
    reverts xs2 ys2 zs2.
    induction ws1; intros; simpls.
    - destruct ys1; simpls; try contradiction.
      destruct zs1; simpls; try discriminate.
      simpls.
      rewrite subst_vars_nil_r.
      auto.
    -
      destruct ys1; simpls; try contradiction.

      lets Hfrz1: (fresh_append_l _ _ _ _ Hfrz).
      lets Hfrz2: (fresh_append_r _ _ _ _ Hfrz).
      destructecs Hfrw.

      destruct xs1; simpls.
      + (* Empty xs2 *)
        eapply fresh_resize_length in Hfrw.
        eapply fresh_resize_length in Hfrz.

        rewrite (subst_vars_fresh zs1).
        rewrite (subst_vars_fresh ws1).
        rewrite subst_fresh.
        auto.
        eapply notin_fv_subst_vars; notin_solve.
        eapply fresh_fv_subst_vars; fresh_solve.
        eapply fresh_fv_subst_vars; fresh_solve.
        eapply fresh_append_fresh_l with (n:=(length (zs1 ++ zs2))). fresh_solve.
      + (* Non-empty xs2 *)
        destruct zs1; simpls; try discriminate.
        injection Hlzy as Hlzy.

        (* Simplify freshness hypotheses *)
        rewrite ! from_list_cons in *.
        rewrite ! notin_union in Hfrw0.
        rewrite ! notin_union in Hfrz1.
        destructecs_all.
        (*rewrite ! notin_union in Hfrz9.
        destructc Hfrz9 as Htmp Hfrz9.*) clear Hfrz9. (* The first part just repeats what is in Hfrz1 *)
        assert(a \notin from_list ws1) as Hnaws1. {
          eapply fresh_single_notin. eapply fresh_union_r2. eauto.
        }
        assert(v1 \notin from_list ws1) as Hnv1ws1. {
          eapply fresh_single_notin. eapply fresh_union_r2. eauto.
        }
        (* Compare lengths - needed to append substitution to list *)
        assert(length zs2 = length ys2) as Hlzy2. {
          eapply fresh_length in Hfrz.
          rewrite List.app_length in Hfrz.
          rewrite List.app_length in Hfrz.
          rewrite Hlzy in Hfrz.
          eapply Plus.plus_reg_l in Hfrz.
          auto.
        }
        eapply fresh_resize_length in Hfrz.
        assert(length ws1 = length ys1) as Hlwy1. {
          eapply fresh_length in Hfrw. auto.
        }
        eapply fresh_resize_length in Hfrw.

        (* Insert v1 from zs to right side *)
        rewrite <- subst_merge with (x:=v0) (y:=v1) (z:=a).
        rewrite <- subst_merge with (x:=a) (y:=v1) (z:=v).
        (* Eliminate a from ws from right side *)
        rewrite subst_vars_commut with (x:=a) (y:= v1).
        rewrite subst_vars_commut with (x:=a) (y:= v1).
        rewrite subst_merge_same with (x:=v1) (y:=a).
        (* Merge v1 into zs2 *)
        rewrite subst_vars_push_right.
        rewrite subst_vars_push_right.
        rewrite subst_vars_push_right.
        rewrite subst_vars_push_right.
        (* Invoke induction *)
        eapply IHws1.
        all: notin_solve.
        all: try assumption.
        * rewrite List.app_length. rewrite List.app_length. simpl. rewrite Hlzx. auto.
        * (* Fix length *)
          replace (length (ys1 ++ ys2 ++ v :: nil)) with (length (zs1 ++ zs2 ++ v1 :: nil)).
          repeat rewrite from_list_append, from_list_cons, from_list_nil.
          (* Simplify rhs *)
          rewrite List.app_assoc.
          eapply fresh_cons_app.
          fresh_solve; fresh_solve.
          repeat rewrite List.app_length. simpl.
          rewrite Hlzy2, Hlzy. auto.
        * replace (length (ys1)) with (length (ws1)).
          repeat rewrite from_list_append, from_list_cons, from_list_nil.
          fresh_solve.
        * eapply notin_fv_subst_var. eapply notin_fv_subst_vars. notin_solve.
          auto. notin_solve.
        * eapply notin_fv_subst_vars.
          eapply notin_fv_subst_vars.
          eapply subst_notfv.
          notin_solve.
          eapply fresh_append_r in Hfrz. eapply fresh_single_notin. eapply fresh_union_r2. eauto.
          auto.
        * eapply notin_fv_subst_vars. auto. auto.
  Qed.
  Lemma subst_vars_fresh_middle_change zs ws xs ys (a: A):
    fresh (from_list xs \u from_list ys \u fv a) (length ys) zs ->
    fresh (from_list xs \u from_list ys \u fv a) (length ys) ws -> (* Using length zs here to make zs and ws same length *)
    a /[xs --> zs] /[ zs --> ys] = a /[xs --> ws] /[ ws --> ys].
  Proof.
    introv Hfrzs Hfrws.

    set (F:= var_freshes (from_list xs \u from_list ys \u from_list zs \u from_list ws \u fv a) (length ys)).
    destructc F as fs ?.

    assert(length zs = length ys) as Hlzy. eapply fresh_length in Hfrzs. auto.
    eapply fresh_resize_length in Hfrzs.
    assert(length fs = length ys) as Hlfy. eapply fresh_length in f. auto.
    eapply fresh_resize_length in f.
    assert(length ws = length ys) as Hlwy. eapply fresh_length in Hfrws. auto.
    eapply fresh_resize_length in Hfrws.

    pose proof (subst_vars_fresh_middle_change_ind) as Hi.

    forwards Hizf: (Hi xs ys zs fs nil nil nil a); eauto.

      rewrite ! List.app_nil_r.
      rewrite ! from_list_nil.
      rewrite <- Hlzy.
      fresh_solve.

      repeat rewrite from_list_nil.
      rewrite <- Hlfy.
      fresh_solve.


    forwards Hiwf: (Hi xs ys ws fs nil nil nil a); eauto.

      repeat rewrite List.app_nil_r.
      repeat rewrite from_list_nil.
      rewrite <- Hlwy.
      fresh_solve.

      repeat rewrite from_list_nil.
      rewrite <- Hlfy.
      fresh_solve.
    rewrite Hizf.
    rewrite Hiwf.
    auto.
  Qed.



  Lemma subst_vars_inverse (a b: A) (xs ys: list var):
    fresh \{} (length ys) ys ->
    fresh (from_list ys) (length xs) xs ->
    fv a \c from_list xs ->
    length ys = length xs ->
    a /[ xs --> ys] = b ->
    a = b /[ys --> xs].
  Proof.
    introv Hfrys Hfrxs Hfv Hl Hbab.
      rewrite <- Hbab.
      rewrite subst_vars_merge_same; eauto.
      replace (fv a \u from_list xs) with (from_list xs).
      eapply fresh_swap; eauto. rewrite union_subset_r; auto.
  Qed.
  Lemma subst_vars_unique (a1 a2 b: A) (xs ys: list var):
    fresh \{} (length ys) ys ->
    fresh (from_list ys) (length xs) xs ->
    fv a1 \c from_list xs ->
    fv a2 \c from_list xs ->
    length ys = length xs ->
    a1 /[ xs --> ys] = b ->
    a2 /[ xs --> ys] = b ->
    a1 = a2.
  Proof.
    introv Hfrys Hfrxs Hfv1 Hfv2 Hl Hs1 Hs2.
    eapply subst_vars_inverse in Hs1; eauto.
    eapply subst_vars_inverse in Hs2; eauto.
    rewrite Hs1, Hs2; eauto.
  Qed.

  Context {ACU: CloseUnbound AC}.
  Context {ASCC: SubstCloseCommut AC}.

  Lemma close_vars_subst_vars (t: A) (xs ys: list var) n:
    length xs = length ys ->
    fresh \{} (length xs) xs ->
    fresh (from_list xs \u fv t) (length ys) ys ->
    t /[#n <-- xs] = t /[xs --> ys] /[#n <-- ys].
  Proof.
    introv Hl Hfrx Hfry.
    reverts ys t n.
    induction xs; intros; destruct ys; try discriminate.
    - simpls. eauto.
    - simpls.
      destructc Hfry as Hnv Hfry.
      destructc Hfrx as Hna Hfrx.
      rewrite from_list_cons in Hnv, Hfry.
      rewrite close_vars_unfold_r. rewrite close_vars_unfold_r.
      rewrite close_subst.
      pose proof (fresh_union_r1 Hfrx) as Hfrxs2.
      pose proof (fresh_union_r2 Hfrx) as Hfrxsa.
      eapply from_list_fresh in Hfrxsa.
        rewrite (IHxs Hfrxs2 ys); eauto.
      rewrite (close_unbound) with (x:=v).
      rewrite close_subst_vars. eauto. eauto.
      repeat eapply fresh_union_r1 in Hfry.
      eapply from_list_fresh in Hfry.
      notin_solve.
      rewrite fv_close.
      eapply notin_remove. left.
      eapply notin_fv_subst_vars; eauto. auto.
      repeat eapply fresh_union_r2 in Hfry.
      eapply from_list_fresh in Hfry. auto.
      fresh_solve.
      rewrite fv_close.
      eapply fresh_remove. fresh_solve.
  Qed.

  Lemma close_vars_subst_vars_fresh (t: A) (xs ys: list var) n:
    length xs = length ys ->
    fresh (fv t) (length xs) xs ->
    fresh (from_list xs \u fv t) (length ys) ys ->
    t /[#n <-- xs] = t /[#n <-- ys].
  Proof.
    introv Hl Hfrx Hfry.
    rewrite ! close_vars_fresh; eauto.
  Qed.

  Context {ASOC: SubstOpenClose AC}.
  Context {ACC: CloseCommut AC}.
  Context {AOC: OpenCommut H}.
  Context {AOT: OpenTwice H}.
  Context {ASOCo: SubstOpenCommut AS}.


    Lemma subst_open_close_vars n xs ys (a: A):
      length ys = length xs ->
      fresh (from_list ys) (length xs) xs ->
      a /[# n --> xs] /[xs --> ys] = a /[#n <-- xs] /[# n --> ys].
    Proof.
      introv Hl Hfr.
      reverts ys n a.
      induction xs; intros; destruct ys; inversions Hl.
      - (* nil *)
        simpl.
        rewrite close_vars_nil.
        auto.
      - simpl.
        inversions Hfr.
        rewrite from_list_cons in H2, H4.
        rewrite notin_union, notin_singleton in H2.
        destructecs H2.
        (* Rewrite *)
        rewrite close_vars_unfold_l.
        rewrite <- subst_open_close.
        rewrite <- subst_open_vars_commut_notin by auto.
        rewrite open_vars_S_commut with (xs0:=ys).
        (* Solve tail by induction *)
        forwards Hxs: IHxs ys (S n) (a0); fweauto.
        rewrite <- Hxs.
        rewrite <- subst_vars_open_fresh_commut.
        rewrite open_vars_S_commut.
        auto.
        eauto.
        eapply from_list_fresh; eauto.
    Qed.
    Lemma subst_open_close_vars_closed n xs ys (a: A):
      length ys = length xs ->
      fresh (from_list ys) (length xs) xs ->
      closed a ->
      a /[xs --> ys] = a /[#n <-- xs] /[# n --> ys].
    Proof.
      introv Hl Hfr Hcl.
      reverts ys n a.
      induction xs; intros; destruct ys; inversions Hl.
      - (* nil *)
        simpl.
        rewrite close_vars_nil.
        auto.
      - simpl.
        inversions Hfr.
        rewrite from_list_cons in H2, H4.
        rewrite notin_union, notin_singleton in H2.
        destructecs H2.
        (* Rewrite *)
        rewrite close_vars_unfold_l.
        rewrite <- subst_open_close.
        rewrite <- subst_open_vars_commut_notin by auto.
        rewrite open_vars_S_commut.
        (* Solve tail by induction *)
        forwards Hxs: IHxs ys (S n) (a0); fweauto.
        rewrite <- Hxs.
        rewrite closed_subst_vars by auto.
        auto.
        eapply from_list_fresh; eauto.
    Qed.


End SubstVars.


