(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Import LibExtra.
Require Import
  SyntaxClasses SyntaxClassProperties SyntaxClassDepth
  ListSyntaxClasses SyntaxClassLemmas.

Generalizable Variables A B.

(* ************************************************************************** *)
(** ** Collecting *)
(** Many operations on a syntax class involves collecting information about
    all instances of a sub-syntax. *)
Class Collect `(AS : AbstractSyntax B) `(AS': AbstractSyntax A) :=
  collect : forall {C : Type}, C -> (B -> C) -> (C -> C -> C) -> A -> C.
Arguments collect _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ !_ /.

Class ListLikeCollect `{ASB : AbstractSyntax B} `{ASA: AbstractSyntax A}
      {C: Set} `{ASC: AbstractSyntax C} {TL: ToList C A}
      (CBC: Collect ASB ASC) (CBA: Collect ASB ASA) :=
  collect_fold_right_map : forall (C' : Type) (df : C') f comb (a : A),
    collect df f comb a =
    List.fold_right comb df (List.map (collect df f comb) (to_list a)).

(** Sanity for Collect *)
Class CollectCompose `(C: Collect B A) :=
  collect_compose :
    forall {C D} (f : C -> D) g comb1 comb2 i,
      (forall x y, f (comb1 x y) = comb2 (f x) (f y)) ->
      f ∘ (collect i g comb1) = collect (f i) (f ∘ g) comb2.

Class CollectReflect `(C: Collect B A) :=
  collect_reflect : forall P p,
    (forall b, (P b) <-> (p b) = true) ->
    forall {bl} a,
      ((collect (Is_true bl) P or a) <-> (collect bl p orb a) = true).


(** Lifting Syntax Operations *)
Class FvCollect `(Collect) :=
  fv_collect : fv = collect (\{}) fv (@union var).

Section AboutCollectCompose.
  Context `{C: Collect B A}.

  Context {CC: CollectCompose C}.
  Lemma collect_compose_same :
      forall {C} (f : C -> C) g comb i,
        (forall x y, f (comb x y) = comb (f x) (f y)) ->
               f ∘ (collect i g comb) = collect (f i) (f ∘ g) comb.
  Proof. auto using collect_compose. Qed.

  Lemma collect_compose_negb_orb : forall b f,
    negb ∘ (collect b f orb) = collect (negb b) (negb ∘ f) andb.
  Proof.
    intros. apply collect_compose.
    apply negb_orb.
  Qed.

  Lemma in_collect : forall x a,
      (x \in collect \{} fv (union (A:=var)) a) =
      collect (x \in \{}) (fun b => x \in fv b) or a.
  Proof.
    intros.
    remember (fun s => x \in s) as ff.
    assert (H': forall y z, ff (y \u z) = (ff y \/ ff z)).
    { intros. rewrite Heqff. apply in_union. }
    rewrite (fold_compose ff).
    rewrite (collect_compose ff fv (@union var) or \{}); auto.
  Qed.

  Lemma notin_collect : forall x a,
      (x \notin collect \{} fv (union (A:=var)) a) =
      collect (x \notin \{}) (fun b => x \notin fv b) and a.
  Proof.
    intros.
    remember (fun s => x \notin s) as ff.
    assert (H': forall y z, ff (y \u z) = (ff y /\ ff z)).
    { intros. apply prop_ext. rewrite Heqff. apply notin_union. }
    rewrite (fold_compose ff).
    rewrite (collect_compose ff fv (@union var) and \{}); auto.
  Qed.

End AboutCollectCompose.


(* ************************************************************************** *)
(** ** Transforming *)
(** Many operations on a syntax class involves transforming sub-syntaxes. *)
Class Transform `(AS : AbstractSyntax B) `(AS': AbstractSyntax A) :=
  transform : (B -> nat -> B) -> A -> nat -> A.
Arguments transform _ _ _ _ _ _ _ _ _ _ _ _ !_ /.

Class ListLikeTransform `{ASB : AbstractSyntax B} `{ASA: AbstractSyntax A}
      {C: Set} `{ASC: AbstractSyntax C}
      {TL: ToList C A} {FL: OfList C A} {LL: ListLike TL FL}
      (TBC: Transform ASB ASC) (TBA: Transform ASB ASA) :=
  transform_map_f : forall f n a,
    transform f a n = map_f (fun c => transform f c n) a.

(** Sanity for Transform *)
Class TransformId `(T: Transform B A) :=
  transform_id : forall x n,
    transform (ignore_second id) x n = x.

Class TransformCompose `(T: Transform B A) :=
  transform_compose : forall f g,
    (transform f) ∘∘ (transform g) = transform (f ∘∘ g).

Class TransformShift `(AS : AbstractSyntax B) `(AS': AbstractSyntax A) (T: Transform AS AS') :=
  transform_shift f X n m: transform f X (n + m) = transform (fun X' n' => f X' (n' + m)) X n.

(** Lifting Syntax Operations *)
Class OpenTransform `(Transform) :=
  open_transform : forall n x,
    (open_depth_rec n x) = transform (open_depth_rec n x).

Class SubstTransform `(Transform) :=
  subst_transform : forall x y,
    (ignore_second (subst_var x y)) = transform (ignore_second (subst_var x y)).

Class SealTransform `(Transform B A)  `(Sealable B)  `(Sealable A) :=
  seal_transform : forall n x,
    (seal_depth_rec n x) = transform (seal_depth_rec n x).

(* ************************************************************************** *)
(** ** Links Between Syntax, Collect and Transform *)
Class CollectTransform
      `{ASB : AbstractSyntax B} `{ASA : AbstractSyntax A}
      (C : Collect ASB ASA) (T : Transform ASB ASA) :=
  collect_transform: forall bl p a f g,
    (collect bl p andb a) = true ->
    transform (dec_apply p f g) a = transform f a.


(* ************************************************************************** *)
(** ** Transformed Syntax Lemmas *)
Section AboutTransformSyntax.
  Context `{OA: Openable A} {FA: FreeVar A} {SA: SubstVar A}.
  Context `{OB: Openable B} {FB: FreeVar B} {SB: SubstVar B}.
  Context {ASB: AbstractSyntax OB FB SB} {ASA: AbstractSyntax OA FA SA}.
  Context {C: Collect ASB ASA} {CC: CollectCompose C} {CR: CollectReflect C}
          {FC: FvCollect C}.
  Context {T: Transform ASB ASA} {TI: TransformId T} {TC: TransformCompose T}
          {OT: OpenTransform T} {ST: SubstTransform T}.
  Context {CT: CollectTransform C T}.

  (** Every TransformSyntax leads to OpenCompose *)
  Section OpenCompose.
    Context {OTA: OpenTwice OB}.
    Instance TransformOpenTwice : OpenTwice OA.
    Proof.
      hnf.
      intros. unfold compose.
      apply fun_ext_dep. intros a.
      rewrite ? open_eq_open_depth.
      rewrite (open_transform n x). rewrite (open_transform n y).
      rewrite fold_compose12. rewrite transform_compose.
      rewrite open_depth_twice_comp.
      auto.
    Qed.

    Context {OTC: OpenCommut OB}.
    Instance TransformOpenCommut : OpenCommut OA.
    Proof.
      hnf.
      intros.
      apply fun_ext_dep. intros a.
      unfold compose. rewrite ? open_eq_open_depth.
      rewrite ? (open_transform m x). rewrite ? (open_transform n y).
      rewrite ? fold_compose12. rewrite ? transform_compose.
      rewrite open_depth_commut_comp; auto.
    Qed.
  End OpenCompose.

  (** Every TransformSyntax leads to SubstOpenCommut *)
  Section SubstOpenCommut.
    Context {SF: SubstOpenCommut ASB}.

    Instance TransformSubstOpenCommut : SubstOpenCommut ASA.
    Proof.
      hnf. intros.
      rewrite ? open_eq_open_depth, ? (subst_depth 0).
      rewrite ? subst_transform.
      rewrite (open_transform n z), (open_transform n (z \[ x -> y])).
      rewrite ? fold_compose12. rewrite ? transform_compose.
      rewrite subst_open_commut_depth. auto.
    Qed.
  End SubstOpenCommut.
  Section SubstCommut.
    Context {OTC: SubstCommut ASB}.
    Instance TransformSubstCommut : SubstCommut ASA.
    Proof.
      hnf.
      intros.

      apply fun_ext_dep. intros a.
      unfold compose.
      rewrite ? (subst_depth 0).
      rewrite subst_transform.
      rewrite (subst_transform w z).
      rewrite (subst_transform w (z \[ x -> y])).

      rewrite ? fold_compose12. rewrite ? transform_compose.

      eapply f_equal3; auto.
      eapply fun_ext_nondep_2. intros.
      rewrite <- fold_compose12.
      rewrite <- fold_compose12.

      rewrite <- (subst_depth).
      rewrite <- (subst_depth).
      rewrite <- (subst_depth).
      rewrite <- (subst_depth).

      eapply subst_commut; eauto.
    Qed.
  End SubstCommut.

  (** ListLike transform leads to length preserving opening *)
  Section ListLike.
    Context {SC: Set}.
    Context `{ASSC: AbstractSyntax SC}.
    Context {CSC: Collect ASB ASSC}.
    Context {CCSC: CollectCompose CSC} {CRSC: CollectReflect CSC}
            {FCSC: FvCollect CSC}.
    Context {TSC: Transform ASB ASSC}.
    Context {TISC: TransformId TSC} {TCSC: TransformCompose TSC}
            {OTSC: OpenTransform TSC} {STSC: SubstTransform TSC}.
    Context {TCSSC: CollectTransform CSC TSC}.
    Context {TL: ToList SC A} {FL: OfList SC A} {LL: ListLike TL FL}.

    Section ListLikeCollect.
      Context {LLC: ListLikeCollect CSC C}.

      Lemma fv_one : forall c,
          fv (of_list (c :: nil)) = fv c.
      Proof.
        intros. rewrite fv_collect.
        rewrite collect_fold_right_map.
        rewrite <- fv_collect. rewrite of_list_to_list.
        simpl. auto.
      Qed.

      Lemma fv_rev : forall a,
          fv a = fv (of_list (List.rev (to_list a))).
      Proof.
        intros.
        assert (H': a = (of_list (to_list a)))
          by (symmetry; apply to_list_of_list).
        rewrite H' at 1; clear H'.
        rewrite fv_collect. rewrite ? collect_fold_right_map, ? of_list_to_list.
        rewrite List.map_rev. rewrite List.fold_left_rev_right.
        symmetry.
        assert (H': forall A, (fun (x y : fset A) => y \u x) = (fun x y => x \u y))
          by (intros; intros_fun_ext; apply union_comm); rewrite H'; clear H'.
        apply List.fold_symmetric; auto.
      Qed.

      Lemma fv_app_union : forall xs ys,
          fv (of_list xs ++ of_list ys)
          = fv (of_list xs) \u fv (of_list ys).
      Proof.
        intros. induction xs.
        + rewrite fv_collect, collect_fold_right_map.
          unfold app_s. rewrite ? of_list_to_list. unfold List.app.
          rewrite <- (of_list_to_list ys), <- collect_fold_right_map at 1.
          rewrite <- union_empty_l at 1. f_equal.
          rewrite collect_fold_right_map, of_list_to_list. auto.
        + unfold app_s. rewrite ? of_list_to_list.
          rewrite <- List.app_comm_cons.
          rewrite ? fv_collect, ? collect_fold_right_map, 2 of_list_to_list.
          simpl. rewrite <- union_assoc.
          rewrite <- (of_list_to_list (xs ++ ys)), <- (of_list_to_list xs).
          rewrite <- ? collect_fold_right_map, <- ? fv_collect, of_list_to_list.
          rewrite of_list_app. rewrite IHxs. auto.
      Qed.

      Lemma fv_env_singles_forward : forall ys a,
          length ys = length (to_list a) ->
          fv_env (ys ~* to_list a) = fv a.
      Proof.
        intros.
        assert (R: fv a
                   = collect \{} fv (@union var) a)
          by (rewrite fv_collect; auto); rewrite R; clear R.
        rewrite fv_collect. rewrite collect_fold_right_map.
        unfold fv_in_values. rewrite liblist_fold_right.
        rewrite values_singles;
          rewrite ? liblist_length; auto.
        clear H2. induction (to_list a); simpl; auto.
        rewrite IHl; auto.
      Qed.

      Lemma fv_env_singles : forall ys a,
          length ys = length_s a ->
          fv_env (ys ~** to_list a) = fv a.
      Proof.
        unfold length_s. intros.
        rewrite fv_rev.
        assert (R: (List.rev (to_list a))
                   = (to_list (of_list (List.rev (to_list a)))))
          by (rewrite of_list_to_list; auto); rewrite R.
        rewrite fv_env_singles_forward
          by (rewrite of_list_to_list, ? List.rev_length; auto).
        rewrite to_list_of_list; auto.
      Qed.
    End ListLikeCollect.

    Section ListLikeTransform.
      Context {LLT: ListLikeTransform TSC T}.

      Lemma transform_length_s :
        forall f n a, length_s (transform f a n) = length_s a.
      Proof.
        intros. rewrite transform_map_f. unfold length_s, map_f.
        rewrite of_list_to_list. rewrite List.map_length; auto.
      Qed.

      Lemma length_s_open :
        forall n x a, length_s (open_rec n x a) = length_s a.
      Proof.
        intros. rewrite open_eq_open_depth, open_transform.
        rewrite transform_length_s; auto.
      Qed.

      Lemma length_s_subst :
        forall x y a, length_s (subst_var x y a) = length_s a.
      Proof.
        intros. rewrite (subst_depth 0), subst_transform, transform_length_s.
        auto.
      Qed.

      Lemma subst_map_f : forall x y a,
          subst_var x y a = map_f (subst_var x y) a.
      Proof.
        intros. rewrite (subst_depth 0), subst_transform, transform_map_f.
        rewrite <- subst_transform. auto.
      Qed.

      Lemma subst_singles_to_list : forall x y ys a,
          length ys = length_s a ->
          (ys ~* to_list (subst_var x y a)) = subst_env x y (ys ~* to_list a).
      Proof.
        intros. rewrite map_singles. rewrite liblist_map.
        rewrite subst_map_f. unfold map_f. rewrite of_list_to_list; auto.
        rewrite liblist_length; auto.
      Qed.

      Lemma subst_singles_rev_to_list : forall x y ys a,
          length ys = length_s a ->
          (ys ~** to_list (subst_var x y a)) = subst_env x y (ys ~** to_list a).
      Proof.
        intros. rewrite subst_map_f. rewrite map_singles, liblist_map.
        rewrite List.map_rev. unfold map_f. rewrite of_list_to_list; auto.
        rewrite liblist_length; auto.
      Qed.
    End ListLikeTransform.
  End ListLike.

  (** Every TransformSyntax leads to a  and Closing *)
  Section DecideFvClosing.
    Context {DFv: DecideFv ASB}.
    Context {DFvS: ReflectFv DFv}.

    Instance CollectDecideFv : DecideFv ASA :=
      (fun x => collect false (fun a => dec_fv x a) orb).

    Instance CollectReflectFv : ReflectFv CollectDecideFv.
    Proof.
      hnf. unfold dec_notin_fv, dec_fv, CollectDecideFv. simpl.
      rewrite fv_collect.
      intros. rewrite in_collect.
      assert (x \in \{} = (Bool.Is_true false)) by (auto using in_empty).
      rewrite H. apply collect_reflect.
      auto using reflect_fv.
    Qed.

    Lemma notin_fv_collect : forall x a,
        x \notin fv a ->
        collect (negb false) (dec_notin_fv x) andb a = true.
    Proof.
      intros x a H.
      pose proof ((reflect_fv x a)) as [H1 H2].
      assert (H3: negb (dec_fv x a) = true) by (destruct (dec_fv x a); auto).
      unfold dec_fv, CollectDecideFv in H3; simpl in H3.
      rewrite (fold_compose negb) in H3.
      rewrite collect_compose_negb_orb in H3.
      auto.
    Qed.

    Section SubstFresh.
      Context {SI: SubstitutionInner DFvS}.
      Context {SF: SubstFresh ASB}.

      Instance TransformSubstFresh : SubstFresh ASA.
      Proof.
        hnf. intros.
        apply notin_fv_collect in H.
        rewrite (subst_depth 0), subst_transform.
        rewrite subst_inner.
        rewrite (collect_transform _ _ _ _ _ H).
        apply transform_id.
      Qed.
    End SubstFresh.

    Section Closing.
      Context {Cl: Closing ASB}.
      Context {OC: OpenClose Cl} {CO: CloseOpen Cl} {SOC: SubstOpenClose Cl} {CS: CloseSubst Cl} {CCo: CloseCommut Cl} {SCCo: SubstCloseCommut Cl} {OCCo: OpenCloseCommut Cl}.

      Instance TransformClosing : Closing ASA :=
        (fun n x a => transform (close_depth_rec n x) a 0).

      Instance TransformOpenClose : OpenClose TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose. rewrite ? open_eq_open_depth.
        rewrite open_transform. rewrite fold_compose12.
        rewrite transform_compose. rewrite open_depth_close_depth.
        auto.
      Qed.

      Instance TransformCloseOpen : CloseOpen TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose. rewrite ? open_eq_open_depth.
        rewrite open_transform. rewrite fold_compose12.
        rewrite transform_compose. rewrite close_depth_open_depth.
        auto.
      Qed.

      Instance TransformSubstOpenClose : SubstOpenClose TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose. rewrite ? (subst_depth 0), ? open_eq_open_depth.
        rewrite (subst_transform x y).
        rewrite (open_transform n x), (open_transform n y).
        rewrite ? fold_compose12. rewrite ? transform_compose.
        rewrite subst_open_close_depth. auto.
      Qed.
      Instance TransformCloseSubst : CloseSubst TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose. rewrite ? (subst_depth 0).
        rewrite (subst_transform x y).
        rewrite ? fold_compose12. rewrite ? transform_compose.
        rewrite close_subst_depth. auto.
      Qed.
      Instance TransformSubstCloseCommut : SubstCloseCommut TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose. rewrite ? (subst_depth 0).
        rewrite (subst_transform x y).
        rewrite ? fold_compose12. rewrite ? transform_compose.
        rewrite subst_close_commut_depth. auto. auto. auto.
      Qed.
      Instance TransformCloseCommut : CloseCommut TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose.
        rewrite ? fold_compose12. rewrite ? transform_compose.
        rewrite close_commut_depth. auto. auto.
      Qed.
      Instance TransformOpenCloseCommut : OpenCloseCommut TransformClosing.
      Proof.
        hnf.
        unfold close_rec, TransformClosing. simpl. intros.
        apply fun_ext_dep. intros a.
        unfold compose.
        rewrite ? open_eq_open_depth.
        rewrite (open_transform n y).
        rewrite ? fold_compose12. rewrite ? transform_compose.
        rewrite open_close_commut_depth. auto. auto. auto.
      Qed.

      Section CloseUnbound.
        Context {CU: CloseUnbound Cl}.
        Context {CI: ClosingInner DFvS Cl}.

        Instance TransformCloseUnbound : CloseUnbound TransformClosing.
        Proof.
          hnf.
          simpl. intros.
          apply notin_fv_collect in H.
          unfold close_rec, TransformClosing; simpl.
          rewrite close_inner.
          rewrite (collect_transform _ _ _ _ _ H).
          apply transform_id.
        Qed.

      End CloseUnbound.

      Section TransformCloseDepth.
      Context {TS : TransformShift T}.

      Lemma transform_close_depth_1 n x (T1: A):
        transform (close_depth_rec n x) T1 1 = transform (close_depth_rec (S n) x) T1 0.
      Proof.
        replace 1 with (0 + 1). (* Puts the number into a form that transform_shift understands *)

        rewrite transform_shift.
        eapply f_equal3; auto. (* Strips transform *)
        eapply fun_ext_nondep_2.
          intros. unfold close_depth_rec.
          eapply f_equal3; auto. (* Strips close_rec *)
          rewrite Nat.add_succ_l.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_0_r.
          auto.
        (* 0 + 1 *)
        rewrite Nat.add_0_l.
        auto.
      Qed.
      Lemma transform_close_depth_2 n x (T1: A):
        transform (close_depth_rec n x) T1 2 = transform (close_depth_rec (S (S n)) x) T1 0.
      Proof.
        replace 2 with (0 + 2). (* Puts the number into a form that transform_shift understands *)

        rewrite transform_shift.
        eapply f_equal3; auto. (* Strips transform *)
        eapply fun_ext_nondep_2.
          intros. unfold close_depth_rec.
          eapply f_equal3; auto. (* Strips close_rec *)
          rewrite Nat.add_succ_l.
          rewrite Nat.add_succ_l.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_succ_r.
          rewrite Nat.add_0_r.
          auto.
        (* 0 + 1 *)
        rewrite Nat.add_0_l.
        auto.
      Qed.

      End TransformCloseDepth.

    End Closing.

  End DecideFvClosing.

  Section SealOpen.
    Context {XA: Sealable A} {XB: Sealable B}.
    Context {OTA: SealOpen ASB XB} {OTSC: SealTransform T XB XA}.
    Instance TransformSealOpen : SealOpen ASA XA.
    Proof.
      split; intros.
      - (* fold *)
        unfold compose.
        apply fun_ext_dep. intros a.
        rewrite (open_eq_open_depth).
        rewrite (seal_eq_seal_depth n).
        rewrite (seal_eq_seal_depth (S n)).
        rewrite (open_transform n x).
        rewrite (seal_transform (S n) x).
        replace (seal_depth_rec n x) with (transform (seal_depth_rec n x)).
        2: symmetry; apply (seal_transform n x).
        rewrite fold_compose12.
        rewrite transform_compose.
        eapply f_equal3; eauto.
        unfold open_depth_rec, seal_depth_rec.
        unfold compose12.
        apply fun_ext_dep. intro.
        apply fun_ext_dep. intro.
        eapply open_seal_fold.
      - (* commut *)
        unfold compose.
        apply fun_ext_dep. intros a.
        rewrite (open_eq_open_depth n x).
        rewrite (open_eq_open_depth n x).
        rewrite (seal_eq_seal_depth m y).
        rewrite (seal_eq_seal_depth m y).
        rewrite (seal_transform m y).
        rewrite (open_transform n x).
        rewrite fold_compose12.
        rewrite fold_compose12.
        rewrite transform_compose.
        rewrite transform_compose.
        eapply f_equal3; eauto.
        unfold open_depth_rec, seal_depth_rec.
        unfold compose12.
        apply fun_ext_dep. intro.
        apply fun_ext_dep. intro.
        eapply open_seal_commut.
        eapply Plus.plus_lt_compat_r; auto.
    Qed.
  End SealOpen.

  Section SealClose.
    Context {XA: Sealable A} {XB: Sealable B}.
    Context {CB: Closing ASB}.
    Context {OTA: SealClose ASB XB CB} {OTSC: SealTransform T XB XA}.
    Instance TransformSealClose : SealClose ASA XA TransformClosing.
    Proof.
      split; intros.
      - (* commut *)
        unfold compose.
        apply fun_ext_dep. intros a.
        rewrite (seal_eq_seal_depth m y).
        rewrite (seal_eq_seal_depth m y).
        rewrite (seal_transform m y).
        unfold close_depth_rec, close_rec, TransformClosing.
        rewrite fold_compose12.
        rewrite fold_compose12.
        rewrite transform_compose.
        rewrite transform_compose.
        eapply f_equal3; eauto.
        unfold close_depth_rec, seal_depth_rec.
        unfold compose12.
        apply fun_ext_dep. intro.
        apply fun_ext_dep. intro.
        eapply close_seal_commut.
        eapply Plus.plus_lt_compat_r; auto.
        auto.
    Qed.
  End SealClose.

End AboutTransformSyntax.

Existing Instances
         TransformOpenTwice
         TransformOpenCommut
         TransformSubstOpenCommut
         TransformSubstCommut
         CollectDecideFv
         CollectReflectFv
         TransformSubstFresh
         TransformClosing
         TransformOpenClose
         TransformCloseOpen
         TransformSubstOpenClose
         TransformCloseSubst
         TransformSubstCloseCommut
         TransformOpenCloseCommut
         TransformCloseCommut
         TransformCloseUnbound
         TransformSealOpen.


