(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Weakening.

(** * Transfomrations *)

(** A syntactic element in a tuple with a typing context and a type *)
#[ universes(notemplate) ]
Inductive TypedElement (AC AT A: Type): Type := typed_elem (G: AC) (T: AT) (a: A).
Notation "'〈' G '⊢' a '∶' T '〉'" :=
  (typed_elem G T a) (at level 41, G at level 32).

(* Definition 23 in "Pure methods for roDOT" *)
(** A general transformation of a typed syntactic element *)
Definition Transform (AC AT A: Type) := (TypedElement AC AT A) -> (TypedElement AC AT A) -> Prop.

(* Transform a into b with Tr *)
Definition transform (AC AT A: Type) (Tr: Transform AC AT A) a b := Tr a b.
Notation " TEa '∼' trf '∼>' TEb " :=
  (transform trf TEa TEb) (at level 42).

(** A property of a transformation *)
Definition TrfCond TC T A := TypedElement TC T A -> Prop.

(** Algebraic operators on transformations *)
Definition trf_concat TC T A (trf1 trf2: Transform TC T A): Transform TC T A :=
    fun TEa TEb =>
      exists TEc,
      TEa ∼trf1∼> TEc /\
      TEc ∼trf2∼> TEb.
Definition trf_union TC T A (trf1 trf2: Transform TC T A): Transform TC T A :=
    fun TEa TEb =>
      TEa ∼trf1∼> TEb \/
      TEa ∼trf2∼> TEb.
Definition trf_inverse TC T A (trf: Transform TC T A): Transform TC T A :=
    fun TEa TEb =>
      TEb ∼trf∼> TEa.

(** ** Classes *)

(** Algebraic classes of transformations *)
(* This reflexivity is unconidtional, which makes it less useful. See ReflexiveIfTransform *)
Class ReflexiveTransform TC T A (trf: Transform TC T A) :=
  trf_reflexive TEa: TEa ∼trf∼> TEa.
Class SymmetricTransform TC T A (trf: Transform TC T A) :=
  trf_symmetric TEa TEb:
    TEa ∼trf∼> TEb ->
    TEb ∼trf∼> TEa.
Class TransitiveTransform C T A (trf: Transform C T A) :=
  trf_transitive TEa TEb TEc:
    TEa ∼trf∼> TEb ->
    TEb ∼trf∼> TEc ->
    TEa ∼trf∼> TEc.
Class ReflexiveIfTransform TC T A (Cond: TrfCond TC T A) (trf: Transform TC T A) :=
  trf_reflexive_if TEa: Cond TEa -> TEa ∼trf∼> TEa.

(* Algebraic classes of transformation operators *)
Class IdempotentTransformOp TC T A (trfop: Transform TC T A -> Transform TC T A) :=
  trf_op_idempotent trf: (trfop trf) = (trfop (trfop trf)).

(** ** Conditions *)

(** Transformation requires a certain condition to hold on the left side *)
Class TransformRequires TC T A (Cond: TrfCond TC T A) (trf: Transform TC T A) := 
  trf_requires TEa TEb:
    TEa ∼trf∼> TEb ->
    Cond TEa.
(* Part of Definition 54 in "Pure methods for roDOT" *)
(** Transformation preserves a certain condition if it hold on the left side *)
Class TransformPreserves TC T A (Cond: TrfCond TC T A) (trf: Transform TC T A) := 
  trf_preserves TEa TEb:
    TEa ∼trf∼> TEb ->
    Cond TEa ->
    Cond TEb.
(** Transformation ensures certain binary condition *)
Class TransformEnsures TC T A (Cond: _ -> _ -> Prop) (trf: Transform TC T A) := 
  trf_ensures TEa TEb:
    TEa ∼trf∼> TEb ->
    Cond TEa TEb.
(** Transformation which applies to all left sides related by certain binary condition *)
Class TransformClosedOver TC T A (Cond: _ -> _ -> Prop) (trf: Transform TC T A) := 
  trf_closer_over TEa TEb TEc:
    TEa ∼trf∼> TEb ->
    Cond TEa TEc ->
    TEc ∼trf∼> TEb.
(** Transformation where all right sides are related by a certain binary condition *)
Class TransformShares TC T A (Cond: _ -> _ -> Prop) (trf: Transform TC T A) := 
  trf_shares TEa TEb TEc:
    TEa ∼trf∼> TEb ->
    TEa ∼trf∼> TEc ->
    Cond TEb TEc.
(** Transformation where the right side is determined by the left side *)
Notation TransformFunctional TC T A := (@TransformShares TC T A eq).

(** ** Relationships *)

(** Two transformations that are independent and can be swapped *)
Class TransformCommute TC T A (trf1 trf2: Transform TC T A) :=
  trf_commute:
    trf_concat trf1 trf2 = trf_concat trf2 trf1.
(** Transformation trf1 is preserved when both sides are transformed by trf2 *)
Class TransformCompat TC T A (trf1 trf2: Transform TC T A) :=
  trf_compat TEa1 TEb1 TEa2 TEb2:
    TEa1 ∼trf1∼> TEb1 ->
    TEa1 ∼trf2∼> TEa2 ->
    TEb1 ∼trf2∼> TEb2 ->
    TEa2 ∼trf1∼> TEb2.
(** Transformation trf1 is preserved when both sides are transformed by trf2 and the targets are related by a certain condition *)
Class TransformCondCompat TC T A (Cond: _ -> _ -> Prop) (trf1 trf2: Transform TC T A) :=
  trf_cond_compat TEa1 TEb1 TEa2 TEb2:
    TEa1 ∼trf1∼> TEb1 ->
    TEa1 ∼trf2∼> TEa2 ->
    TEb1 ∼trf2∼> TEb2 ->
    Cond TEa2 TEb2 ->
    TEa2 ∼trf1∼> TEb2.

(** A symmetric transformation is an inverse of itself *)
Lemma trf_symmetric_inverse TC T A (trf: Transform TC T A) (TS: SymmetricTransform trf):
  trf_inverse trf = trf.
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  split; intro; eapply TS; eauto.
Qed.

(* Inverse of inverse *)
Lemma trf_inverse_inverse TC T A (trf: Transform TC T A):
  trf_inverse (trf_inverse trf) = trf.
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  split; intros; destruct_all; eauto.
Qed.

(** Swap the sides of a transformation *)
Lemma trf_inverse_intro TC T A (trf: Transform TC T A) TEa TEb:
  TEa ∼trf∼> TEb ->
  TEb ∼(trf_inverse trf)∼> TEa.
Proof.
  intros.
  unfold trf_inverse, transform. auto.
Qed.

(** Inverse of concatenation of transformations *)
Lemma trf_inverse_concat TC T A (trf1 trf2: Transform TC T A):
  trf_inverse (trf_concat trf1 trf2) = (trf_concat (trf_inverse trf2) (trf_inverse trf1)).
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse, trf_concat, transform.
  split; intros; destruct_all; eauto.
Qed.
(** Inverse of union of transformations *)
Lemma trf_inverse_union TC T A (trf1 trf2: Transform TC T A):
  trf_inverse (trf_union trf1 trf2) = (trf_union (trf_inverse trf1) (trf_inverse trf2)).
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse, trf_union, transform.
  split; intros; destruct_all; eauto.
Qed.
(** Union of concatenation is commutative *)
Lemma trf_union_swap TC T A (trf1 trf2: Transform TC T A):
  (trf_union trf1 trf2) = (trf_union trf2 trf1).
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_union, transform.
  split; intros; destruct_all; eauto.
Qed.
(** Union with itself *)
Lemma trf_union_diag TC T A (trf: Transform TC T A):
  (trf_union trf trf) = trf.
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_union, transform.
  split; intros; destruct_all; eauto.
Qed.
(** Concatenation of a transitive transformation with itself *)
Lemma trf_concat_diag TC T A Cond (trf: Transform TC T A) {TT: TransitiveTransform trf} {TR: ReflexiveIfTransform Cond trf} {TQ: TransformRequires Cond trf}:
  (trf_concat trf trf) = trf.
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_concat, transform.
  split; intros; destruct_all.
  - eapply trf_transitive; eauto.
  - exists x1; split.
    eapply trf_reflexive_if; eauto.
    eauto.
Qed.


(** ** Instances *)

(* Concat and union of transitive is not always transitive *)
(* Concat of symmetric is not always symmetric, but union is *)

(** Union of symmetric is symmetric *)
Global Instance SymmetricTransformUnion TC T A (trf1 trf2: Transform TC T A) (TI1: SymmetricTransform trf1) (TI2: SymmetricTransform trf2):
  SymmetricTransform (trf_union trf1 trf2).
Proof.
  hnf. introv Htrf.
  inversions Htrf.
  eapply TI1 in H; eauto.
  left; eauto.
  eapply TI2 in H; eauto.
  right; eauto.
Qed.

Instance TransformRequiresConcat TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformRequires Cond trf1):
  TransformRequires Cond (trf_concat trf1 trf2).
Proof.
  hnf.
  introv Htrf.
  hnf in Htrf.
  destructecs Htrf.
  eapply TI1 in Htrf0; eauto.
Qed.

Global Instance TransformPreservesConcat TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformPreserves Cond trf1) (TI2: TransformPreserves Cond trf2):
  TransformPreserves Cond (trf_concat trf1 trf2).
Proof.
  hnf.
  introv Htrf.
  hnf in Htrf.
  destructecs Htrf.
  eapply TI1 in Htrf0; eauto.
Qed.

Global Instance TransformClosedOverConcat TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformClosedOver Cond trf1):
  TransformClosedOver Cond (trf_concat trf1 trf2).
Proof.
  hnf.
  introv Htrf Hc.
  hnf in Htrf.
  destructecs Htrf.
  eapply TI1 in Htrf0. eapply Htrf0 in Hc.
  econstructor; eauto.
Qed.

Global Instance TransformRequiresUnion TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformRequires Cond trf1) (TI2: TransformRequires Cond trf2):
  TransformRequires Cond (trf_union trf1 trf2).
Proof.
  hnf.
  introv Htrf.
  hnf in Htrf.
  destructecs Htrf.
  - eapply TI1 in Htrf; eauto.
  - eapply TI2 in Htrf; eauto.
Qed.

Global Instance TransformPreservesUnion TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformPreserves Cond trf1) (TI2: TransformPreserves Cond trf2):
  TransformPreserves Cond (trf_union trf1 trf2).
Proof.
  hnf.
  introv Htrf.
  hnf in Htrf.
  destructecs Htrf.
  - eapply TI1 in Htrf; eauto.
  - eapply TI2 in Htrf; eauto.
Qed.

Global Instance TransformClosedOverUnion TC T A Cond (trf1 trf2: Transform TC T A) (TI1: TransformClosedOver Cond trf1) (TI2: TransformClosedOver Cond trf2):
  TransformClosedOver Cond (trf_union trf1 trf2).
Proof.
  hnf.
  introv Htrf Hc.
  hnf in Htrf.
  destructecs Htrf.
  -
  eapply TI1 in Htrf. eapply Htrf in Hc.
  unfold trf_union, transform; eauto.
  -
  eapply TI2 in Htrf. eapply Htrf in Hc.
  unfold trf_union, transform; eauto.
Qed.

(** Reflexive and transitive closure *)
Inductive trf_iterate TC T A (Cond: _ -> Prop) (trf: Transform TC T A): Transform TC T A :=
| trf_iterate_0 TEa:
  Cond TEa ->
  TEa ∼(trf_iterate Cond trf)∼> TEa
| trf_iterate_S TEa TEb TEc:
  TEa ∼(trf_iterate Cond trf)∼> TEb ->
  TEb ∼trf∼> TEc ->
  TEa ∼(trf_iterate Cond trf)∼> TEc.

Lemma trf_iterate_1 TC T A (Cond: _ -> Prop) (trf: Transform TC T A) TEa TEb:
  Cond TEa ->
  TEa ∼trf∼> TEb ->
  TEa ∼(trf_iterate Cond trf)∼> TEb.
Proof.
  intros.
  eapply trf_iterate_S; eauto.
  eapply trf_iterate_0; eauto.
Qed.
Lemma trf_iterate_r TC T A (Cond: _ -> Prop) (trf: Transform TC T A) TEa TEb TEc:
  Cond TEa ->
  TEa ∼ trf ∼> TEb ->
  TEb ∼(trf_iterate Cond trf)∼> TEc ->
  TEa ∼(trf_iterate Cond trf)∼> TEc.
Proof.
  introv Hc Ht Hti.
  reverts TEa.
  induction Hti; intros.
  eapply trf_iterate_1; eauto.
  eapply trf_iterate_S; eauto.
Qed.

Global Instance TransformRequiresIterate TC T A Cond (trf: Transform TC T A):
  TransformRequires Cond (trf_iterate Cond trf).
Proof.
  hnf.
  introv Htrf.
  induction Htrf.
  - auto.
  - eauto.
Qed.
Global Instance TransformPreservesIterate TC T A Cond Cond2 (trf: Transform TC T A) {TP: TransformPreserves Cond trf}:
  TransformPreserves Cond (trf_iterate Cond2 trf).
Proof.
  hnf.
  introv Htrf.
  induction Htrf; eauto.
Qed.

(** Inversion of iterated transformation *)
Lemma trf_inverse_iterate TC T A Cond (trf: Transform TC T A) {TRi: TransformRequires Cond trf} {TR: TransformRequires Cond (trf_inverse trf)}:
  trf_inverse (trf_iterate Cond trf) = trf_iterate Cond (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros.
  split.
  - intros.
    induction H.
    + eapply trf_iterate_0; eauto.
    + eapply trf_iterate_r; eauto.
  - intros.
    induction H.
    + eapply trf_iterate_0; eauto.
    + eapply trf_iterate_r; eauto.
Qed.


Global Instance TransformIterateSymmetric TC T A Cond (trf: Transform TC T A) (TS: SymmetricTransform trf) {TR: TransformRequires Cond trf}:
  SymmetricTransform (trf_iterate Cond trf).
Proof.
  hnf. introv Htrf.
  rewrite <- (trf_symmetric_inverse TS).
  lets TRi: TR.
  rewrite <- (trf_symmetric_inverse TS) in TR.
  rewrite <- trf_inverse_iterate.
  eauto.
Qed.

Global Instance TransformIterateTransitive TC T A Cond (trf: Transform TC T A):
  TransitiveTransform (trf_iterate Cond trf).
Proof.
  hnf. introv Htrfa Htrfb.
  reverts TEa.
  induction Htrfb; intros.
  - auto.
  - eapply trf_iterate_S. 
    eapply IHHtrfb; eauto.
    eauto.
Qed.

(** Iterated transformation is reflexive up to Cond *)
Global Instance TransformIterateReflexive TC T A Cond (trf: Transform TC T A):
  ReflexiveIfTransform Cond (trf_iterate Cond trf).
Proof.
  hnf. introv Hc.
  eapply trf_iterate_0; eauto. 
Qed.

Lemma trf_iterate_eq TC T A Cond (trf: Transform TC T A)
  {TT: TransitiveTransform trf} {TR: ReflexiveIfTransform Cond trf} {TQ: TransformRequires Cond trf}:
  trf_iterate Cond trf = trf.
Proof.
  eapply pred_ext_2.
  intros.
  split.
  - intros.
    induction H.
    + eapply trf_reflexive_if; auto.
    + eapply trf_transitive; eauto.
  - intros.
    eapply trf_iterate_1; eauto.
Qed.

(** Iteration is idempotent *)
Lemma trf_iterate_iterate TC T A Cond (trf: Transform TC T A):
  trf_iterate Cond (trf_iterate Cond trf) = trf_iterate Cond trf.
Proof.
  eapply trf_iterate_eq.
Qed.

(** Lift a unary condition to a binary condition which requires both sides to be the same *)
Definition trf_id_cond TC T A (Cond: _ -> Prop): Transform TC T A :=
  fun TEa TEb =>
    TEa = TEb /\ Cond TEa.

Global Instance TransformIdReflexive TC T A (Cond: _ -> Prop): @ReflexiveIfTransform TC T A Cond (trf_id_cond Cond).
Proof.
  hnf; intros.
  hnf; auto.
Qed.

Global Instance TransformIdSymmetric TC T A (Cond: _ -> Prop): @SymmetricTransform TC T A (trf_id_cond Cond).
Proof.
  hnf; intros.
  hnf in H.
  destructecs H.
  subst.
  hnf; auto.
Qed.
Global Instance TransformIdTransitive TC T A (Cond: _ -> Prop): @TransitiveTransform TC T A (trf_id_cond Cond).
Proof.
  hnf; intros.
  hnf in H.
  destructecs H.
  subst.
  hnf; auto.
Qed.
Global Instance TransformIdRequires TC T A (Cond: _ -> Prop): @TransformRequires TC T A Cond (trf_id_cond Cond).
Proof.
  hnf; intros.
  hnf in H; destructecs H; auto.
Qed.
Global Instance TransformIdFunctional TC T A (Cond: _ -> Prop): TransformFunctional TC T A (trf_id_cond Cond).
Proof.
  hnf; intros.
  hnf in H, H0. destructecs H; destructecs H0; subst; auto.
Qed.

(* Definition 49 in "Pure methods for roDOT" *)
Definition TransformCondTypeIdentical TC T A (TEa TEb: TypedElement TC T A) :=
  match TEa,TEb with
    〈Ga ⊢ a ∶ Ta〉,〈Gb ⊢ b ∶ Tb〉 => Ga = Gb /\ Ta = Tb
  end.

Global Instance TransformIdTypeIdentical TC T A (Cond: _ -> Prop): @TransformEnsures TC T A (@TransformCondTypeIdentical TC T A ) (trf_id_cond Cond) .
Proof.
  hnf; intros.
  hnf in H.
  destructecs H; subst.
  hnf.
  destruct TEb; auto.
Qed.

(** ** Definitions for conditional lifting *)

(** Conditional application of a transformation. *)
Notation " TEa '∼' Cond '∼∼' trf '∼>' TEb " :=
  ((Cond TEa TEb) -> (transform trf TEa TEb)) (at level 42, no associativity).

(** A condition that applies regardless of types *)
Definition LiftCond := forall TC T A : Type, TypedElement TC T A -> TypedElement TC T A -> Prop.

Definition trf_liftcond_id: LiftCond := fun TC T A => (@eq (TypedElement TC T A)).
Definition trf_liftcond_true: LiftCond := fun TC T A TEa TEb => True.
Definition trf_liftcond_inverse (Cond: LiftCond): LiftCond := fun TC T A TEa TEb => Cond TC T A TEb TEa.
Class TransformLiftCondSymmetric (Cond: LiftCond):=
  trf_liftcond_symmetric TC T A TEa TEb:
    Cond TC T A TEa TEb ->
    Cond TC T A TEb TEa.

Hint Unfold trf_liftcond_true: core.

Instance TransformCondIdSymmetric: TransformLiftCondSymmetric trf_liftcond_id.
Proof.
  hnf; intros.
  hnf in *.
  auto.
Qed.
Instance TransformCondTrueSymmetric: TransformLiftCondSymmetric trf_liftcond_true.
Proof.
  hnf; intros.
  hnf in *.
  auto.
Qed.
Instance TransformCondTypeIdenticalSymmetric: TransformLiftCondSymmetric TransformCondTypeIdentical.
Proof.
  hnf; intros.
  destruct TEa, TEb.
  hnf in *.
  destructecs H; auto.
Qed.

(** Definitions for mutually inductive proofs *)
Definition ImplyTransformInd (TC T A: Type) (trf1 trf2: Transform TC T A) :=
  forall TEa TEb,
  trf1 TEa TEb ->
  trf2 TEa TEb.

Definition InverseTransformInd (TC1 TC2 T1 T2 A1 A2: Type) (trf: Transform TC1 T1 A1) (trfop: Transform TC1 T1 A1 -> Transform TC2 T2 A2) :=
  ImplyTransformInd (trfop trf) (trf_inverse (trfop (trf_inverse trf))).

Definition InverseTransformIndCond (TC1 TC2 T1 T2 A1 A2: Type) (Cond: LiftCond) (trf: Transform TC1 T1 A1) (trfop: LiftCond -> Transform TC1 T1 A1 -> Transform TC2 T2 A2) :=
  ImplyTransformInd (trfop Cond trf) (trf_inverse (trfop (trf_liftcond_inverse Cond) (trf_inverse trf))).

Definition CondTransformInd (TC T A: Type) (Cond: _ -> Prop) (trf1: Transform TC T A) :=
  forall TEa TEb,
  trf1 TEa TEb ->
  Cond TEa.

Definition ClosedTransformInd (TC T A: Type) (Cond: _ -> _ -> Prop)  (trf: Transform TC T A) :=
  forall TEa TEb,
  trf TEa TEb ->
  forall TEc,
  Cond TEa TEc ->
  trf TEc TEb.

Definition CondCompatTransformInd (TC T A: Type) (Cond: _ -> _ -> Prop)  (trf trf2: Transform TC T A) :=
  forall TEa TEb,
  trf TEa TEb ->
  forall TEc TEd,
  trf2 TEa TEc ->
  trf2 TEb TEd ->
  Cond TEc TEd ->
  trf TEc TEd.