(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers TransformationConfig ConfigTyping.
Require Import Similarity ConfigSimilarity ContextSimilarity SyntaxSimilarity SimilarityPreservation TypingSimilarity
 TypedReductionEquivalence OperationalSemantics TypedOperationalSemantics TypedPreservation ConfigValid TransformationPreservation.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Similarity transformation.
Transforms a well-typed configuration into a similar configuration of the same type.
*)

(* Used in Definition 24 in "Pure methods for roDOT" *)
Inductive trf_similarity: TransformConfig :=
| trf_similarity_c Ga Ca T Gb Cb:
  ty_config Ga Ca T ->
  ty_config Gb Cb T ->
  similar_conf_1to1 Ca Cb ->
  〈Ga ⊢ Ca ∶ T〉 ∼trf_similarity∼> 〈Gb ⊢ Cb ∶ T〉.

(** Similarity transformation is symmetric *)
Global Instance TransformSimilaritySymmetric: SymmetricTransform trf_similarity.
Proof.
  hnf; intros.
  inversions H.
  econstructor; eauto.
  eapply similar_conf_1to1_symm; eauto.
Qed.

(** Similarity transformation transforms answers into answers *)
Global Instance TransformSimilarityPreservesIsAnswer: TransformPreserves TransformIsAnswer trf_similarity.
Proof.
  hnf; intros.
  inversions H.
  inversions H3. destruct_all.
  inversions H0.
  destruct Cb.
  destructc H5 as ? ? ? ?.
  similar_syntax_inverts.
  econstructor.
Qed.

(** Similarity transformation is type safe *)
Global Instance TransformSimilarityTyped: TransformRequires (TransformCondConfTyped typing_mode) trf_similarity.
Proof.
  hnf; intros.
  inversions H.
  eauto.
Qed.

(* Lemma 57 in "Pure methods for roDOT" *)
Global Instance TransformConfigReducesTo1Similarity: TransformConfigReducesTo1 typing_mode trf_similarity trf_similarity.
Proof.
  hnf; intros.
  inversions H.
  econstructor; eauto.
  eapply typed_preservation in H0; eauto.
  eapply typed_preservation in H1; eauto.
  eapply similar_conf_1to1_preservation.
  eauto.
  eapply config_typed_valid; eauto.
  eapply config_typed_valid; eauto.
  eapply typed_red_to_untyped; eauto.
  eapply typed_red_to_untyped; eauto.
Qed.

Global Instance TransformSimilarityReflexive: ReflexiveIfTransform (TransformCondConfTyped typing_mode) trf_similarity.
Proof.
  hnf; intros.
  destruct TEa.
  eapply trf_similarity_c; eauto.
  eapply similar_conf_1to1_refl.
  eapply config_typed_valid; eauto.
Qed.

Lemma trf_similarity_step G1 C1 T Ga Ca Gb Cb:
  ty_config G1 C1 T ->
  G1 ⊢ C1 ∶ T ↦ Ga ⊢ Ca ->
  G1 ⊢ C1 ∶ T ↦ Gb ⊢ Cb ->
  〈Ga ⊢ Ca ∶ T〉 ∼trf_similarity∼> 〈Gb ⊢ Cb ∶ T〉.
Proof.
  introv Hty Hreda Hredb.
  assert(〈 G1 ⊢ C1 ∶ T 〉 ∼ trf_similarity ∼> 〈 G1 ⊢ C1 ∶ T 〉).
  eapply TransformSimilarityReflexive; eauto.
  eapply TransformConfigReducesTo1Similarity; eauto.
Qed.

Global Instance TransformSimilarityTransitive: TransitiveTransform trf_similarity.
Proof.
  hnf; intros.
  destruct TEa, TEb, TEc.
  inversions H.
  inversions H0.
  econstructor; eauto.
  forwards: config_typed_valid H8.
  forwards: config_typed_valid H5.
  forwards: config_typed_valid H10.
  eapply similar_conf_1to1_trans with (C2:=a0); eauto.
Qed.

Global Instance TransformSimilarityPreservesTermination: TransformPreserves TransformCondTerminates trf_similarity.
Proof.
  eapply trf_reduces_to_self_terminates.
  rewrite trf_symmetric_inverse.
  eapply TransformConfigReducesTo1Similarity.
  eapply TransformSimilaritySymmetric.
  eapply TransformSimilarityPreservesIsAnswer.
  rewrite trf_symmetric_inverse.
  eapply TransformSimilarityPreservesIsAnswer.
  eapply TransformSimilaritySymmetric.
  eapply TransformSimilarityTyped.
  rewrite trf_symmetric_inverse.
  eapply TransformSimilarityTyped.
  eapply TransformSimilaritySymmetric.
Qed.

Lemma trf_similarity_retype T Ga Gb C:
  ty_config Ga C T ->
  ty_config Gb C T ->
  〈Ga ⊢ C ∶ T〉 ∼trf_similarity∼> 〈Gb ⊢ C ∶ T〉.
Proof.
  introv Htya Htyb.

  hnf. constructor; eauto.
  eapply similar_conf_1to1_refl. eapply config_typed_valid; eauto.
Qed.

Lemma trf_similarity_retype_Lv T Ga Gb t fs S e Lva Lvb:
  ty_config Ga (t;fs;S;e;Lva) T ->
  ty_config Gb (t;fs;S;e;Lvb) T ->
  Lva \c Lvb ->
  〈Ga ⊢ (t;fs;S;e;Lva) ∶ T〉 ∼trf_similarity∼> 〈Gb ⊢ (t;fs;S;e;Lvb) ∶ T〉.
Proof.
  introv Htya Htyb Hlv.

  hnf. constructor; eauto.
  forwards: similar_conf_1to1_refl. eapply config_typed_valid; eapply Htya.
  hnf in *.
  destructecs H.
  exists xs1 xs2.
  splits; eauto.
  eapply vars_conf_unique_Lv_subset. eauto.
  eauto.
Qed.



End WithTypingMode.
