(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationTerm TransformationLifting Weakening EWeakening Substitution Narrowing Subenvironments.
Require Import TypingHelpers.
Require Import ReadableSyntax FreshChain FreshChainTactics.
Require Import Visibility.
Require Import Lia.
Require Import Sealing.
Require Import Closedness.

(** * A transformation that swaps to consecutive field reads. *)

(* Let-chain transformations.
   Due to A-normal form, practical transformations have the form of changing a chain of let terms.
   An example of that is swapping two reads.
   In the continuation, variable references must be preserved, so due to a locally nameless representation,
   the indices 0 and 1 must be swapped.
   To avoid working with indices directly, we can assume a closed variant of the term, and replace variables by indices using close_vars (t/[<--ys]).
   Then in the definition of transformation, we only work with closed terms.
   The variables in the read terms are assumed to be closed.
   If they were allowed to be open, they would have to be shifted by 1 up and down because they move in and out of scope of the other let statement.
   However, the lifting of local transformations opens nested terms so local transformation will only work with closed terms.
*)

(* A template for transformations that swaps two terms in a let-chain *)
Definition transfrom_let_swap (B: Type) T1 T2 y1 y2 t (C: trm -> trm -> B): B :=
  C (trm_let T1 (trm_let (T2/[<-- y1::nil ]) (t/[<-- y2::y1::nil]))) (trm_let (T2) (trm_let (T1/[<--y2::nil]) (t/[<--y1::y2::nil]))).

Section WithTypingMode.
Context (typing_mode: Mode).

(* A local transformation which swaps two reads in a let-chain *)
Inductive trf_swap_reads_local: TermTransform :=
  trf_swap_reads_local_at x1 f1 x2 f2 t y1 y2 G e T1 T2 T:
    ctx_ok G ->
    ok e ->
    closed t ->
    fresh (ctx_dom G \u ctx_fv G \u fv_renv e \u fv T \u fv T1 \u fv T2) 2 (y1::y2::nil)%list -> (* Fresh variable names to use for the locals *)
    G ⁏ e ⊢ ❪Read $x1․f1❫ ∶ T1 -> (* Requiring typing of both reads in the original context *)
    G ⁏ e ⊢ ❪Read $x2․f2❫ ∶ T2 ->
    G §ˣ y2 ~ T2 §ˣ y1 ~ T1 ⁏ e ⊢ t ∶ T ->
    〈G ⁏ e ⊢ ❪Let y1=Read $x1․f1 In Let y2=Read $x2․f2 In t❫ ∶ T〉 ∼trf_swap_reads_local∼> 〈G ⁏ e ⊢ ❪Let y2=Read $x2․f2 In Let y1=Read $x1․f1 In t❫ ∶ T〉.

Global Instance TransformSwapReadsIdtyp:
  TransformEnsures (@TransformCondTypeIdentical _ _ _) (trf_swap_reads_local).
Proof.
  hnf; intros.
  hnf in *; intros.
  inversions H.
  auto.
Qed.


Global Instance TransformSwapReadsRequiresOk: TransformRequires (@TransformCondOk _ _) trf_swap_reads_local.
Proof.
  hnf; intros.
  inversions H.
  hnf.
  eauto.
Qed.

Global Instance TransformSwapReadsSymmetric: SymmetricTransform trf_swap_reads_local.
Proof.
  hnf; intros.
  hnf; intros.
  inversions H.
  eapply trf_swap_reads_local_at; eauto.
  fresh_solve; eauto.
  eapply typing_swap; eauto.
Qed.



Global Instance TransformSwapReadsTyped: TransformRequires (TransformCondTermTyped typing_mode) (trf_swap_reads_local).
Proof.
  hnf in *.
  introv Htr.
  inversions Htr.
  rename H into HokG, H0 into Hoke.
  rename H3 into Htyread1, H4 into Htyread2.
  rename H2 into Hfr, H1 into Hcl.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.
  fresh_chain_include_typing Fr Htyread1.
  fresh_chain_include_typing Fr Htyread2.

  let LFr := (fresh_chain_get_set Fr) in
  eapply ty_trm_let with (L:=LFr).
  eapply Htyread1.
  intros. simpl in H.
  fresh_chain_add_fresh Fr H.
  (* Rewrite the closing into list *)
  rewrite trm_close_let_close_fresh_vars by (simpl;fresh_chain_solve Fr).
  sympls. 

  let LFr := (fresh_chain_get_set Fr) in
  eapply ty_trm_let with (L:=LFr).

  eapply weaken_push. eapply Htyread2.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  intros.
  simpl in H.
  fresh_chain_add_fresh Fr H.


  rewrite trm_open_open_vars.
  rewrite <- subst_open_close_vars_closed.

  unfold subst_vars.
  rewrite subst_commut_ne by (fresh_chain_solve Fr).
  eapply rename_subst_trm_2; [> fresh_chain_solve Fr..|].
  eapply typing_swap; [> fresh_chain_solve Fr..|].
  auto.
  auto.
  fresh_chain_solve Fr.

  auto.
Unshelve.
  eapply weaken_ty_trm.
Qed.




Global Instance TransformSwapReadsWeakening: TransformWeakeningCompat _ _ trf_swap_reads_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H5.
  rename ea0 into ea.
  rename H into Hwk, H0 into HokG, H1 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG HokG.

  inversions Htrf.
  clears Ga0 Ga4 Ga5.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H12.
  fresh_chain_cleanup Fr.

  clear H9 H10 H4 Hoke.

  rename H11 into Hclt, H13 into Hty1, H14 into Hty2, H15 into Htyt.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_reads_local_at with (T1:=T1) (T2:=T2).
  - fresh_chain_solve FrG.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply weaken_ty. ctx_flatten.
    eapply rename_subst_trm_2.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    auto 1.
    fresh_chain_solve FrG.

Unshelve.
  eapply weaken_ty_trm.
  eapply weaken_ty_trm.
  eapply weaken_ty_trm.
Qed.

Global Instance TransformSwapReadsEWeakening: TransformEWeakeningCompat _ _ trf_swap_reads_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H3.
  rename Ga0 into Ga.
  rename H1 into HokG, H0 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG Hoke.

  inversions Htrf.
  clears e0 e4 e5.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H11.
  fresh_chain_cleanup Fr.

  clear H8 H9 HokG.

  rename H10 into Hclt, H12 into Hty1, H13 into Hty2, H14 into Htyt.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_reads_local_at with (T1:=T1) (T2:=T2).
  - fresh_chain_solve Fr.
  - fresh_chain_solve FrG.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  -

    eapply ewkn.
    eapply rename_subst_trm_2.
    fresh_chain_solve Fr.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    auto 1.
    fresh_chain_solve FrG.

Unshelve.
  eapply ewkn_ty_trm.
  eapply ewkn_ty_trm.
  eapply ewkn_ty_trm.
Qed.


Global Instance TransformSwapReadsNarrowing: TransformNarrowingCompat _ _ _ trf_swap_reads_local.
Proof.
  hnf in *.
  introv Htrf Hna1 Hna2 Hti.
  inversions Hna1.
  inversions Hna2.
  inversions Hti. inversions H5.
  rename e0 into e.
  rename H into HokG, H0 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG HokG.

  inversions Htrf.
  clears Ga0 x0 Txa0 Ga3.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H13.
  fresh_chain_cleanup Fr.

  clear H10 H3 H11 Hoke.

  rename H12 into Hclt, H14 into Hty1, H15 into Hty2, H16 into Htyt.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.

  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  fresh_chain_add_var FrG z1. 2: fresh_chain_solve Fr.
  fresh_chain_add_var FrG z2. 2: fresh_chain_solve Fr.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_reads_local_at with (T1:=T1) (T2:=T2).
  - fresh_chain_solve FrG.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  -
    eapply narrow; eauto.
    eapply Hty1. unfold ctx_obj; sympl. fresh_chain_solve FrG.
    eapply subctx_concat. eapply subctx_grow. eapply subctx_refl.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
    eauto.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
  - eapply narrow; eauto.
    eapply Hty2. unfold ctx_obj; sympl. fresh_chain_solve FrG.
    eapply subctx_concat. eapply subctx_grow. eapply subctx_refl.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
    eauto.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply narrow.
    3: {
      ctx_flatten.
      eapply subctx_push.
      eapply subctx_push.
      eapply subctx_concat.
      eapply subctx_grow.
      eapply subctx_refl.

      fresh_chain_solve FrG.
      fresh_chain_solve FrG.
      eauto.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    }
    eapply rename_subst_trm_2.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    unfold ctx_obj.
    fresh_chain_solve FrG.

Unshelve.
  eapply narrow_ty_trm.
  eapply narrow_ty_trm.
  eapply narrow_ty_trm.
Qed.

Global Instance TransformSwapReadsPreservesIsAnswer: TransformPreserves TransformIsAnswerTerm trf_swap_reads_local.
Proof. (* easy, this transformation only works on non-answers *)
  hnf. intros.
  hnf in *.
  destruct TEa, TEb.
  inversions H.
  inversions H0.
  inversions H.
Qed.

(* Rewriting lemma for trm_let_find_open_with *)
Lemma trm_let_open_close t1 t2 x:
  x \notin (fv t2) ->
  trm_let t1 t2 = trm_let t1 (t2/[->x]/[<-x]).
Proof.
  intros.
  rewrite close_open.
  rewrite close_unbound; auto.
Qed.

(* Finds a trm_let subterm in the goal and rewrites it so that the second subterm has a form of (close t) 
   replaces the corresponding de bruijn index by the variable y *)
Ltac trm_let_find_open_with y := (
  match goal with 
  | |- context[ (trm_let ?tx1 ?tx2) ] =>
    match tx2 with 
    | ?t /[<-?x] => fail 1
    | _ =>
      rewrite (@trm_let_open_close tx1 tx2 y);
      let a := fresh "a" in
      set (a := (tx2/[->y]));
      sympl in a;
      subst a;
      repeat cases_if
    end
end).

Lemma if_apply_1 (A B: Type) (P1 P2: Prop) (P: sumbool P1 P2) a b (f: A -> B):
  f (if P then a else b) = if P then f a else f b.
Proof.
  cases_if; auto.
Qed.

Lemma TEST_fresh_chain_is_containing G x:
 fresh_chain_is (fresh_chain_var (fresh_element_var x ) (fresh_chain_set (fresh_element_ctx G) fresh_chain_nil)) ->
 fresh_chain_is (fresh_chain_var (fresh_element_var x ) (fresh_chain_set (fresh_element_containing (fresh_element_ctx G) (fresh_element_ctx G)) fresh_chain_nil)).
Proof.
  introv Fr.

  rewrite fresh_element_eq_containing_refl with (fe := (fresh_element_ctx G))  in Fr by apply true.
  auto.
Qed.


Lemma trf_swap_reads_local_subst_rewrite x1 f1 x2 f2 (t: trm) (y1 y2 x0 y: var):
  fresh (\{x1} \u \{x2} \u \{x0} \u \{y}) 2 (y1 :: y2 :: nil) ->
  (
    (trm_let (trm_read (avar_f x1) f1) (trm_let (trm_read (avar_f x2) f2) t /[<- y2]) /[<- y1]) /[ x0 -> y]
  ) = (
    (trm_let (trm_read (avar_f x1 /[ x0 -> y]) f1) (trm_let (trm_read (avar_f x2 /[ x0 -> y]) f2) t /[ x0 -> y] /[<- y2]) /[<- y1])
  ).
Proof.
  introv Hfr.
  
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.

    autorewrite with simpl_close. (* Push closing down *)
    sympl. (* Push substitution down *)
    rewrite ! subst_close_commut. (* Swap closing and subst *)
    autorewrite with simpl_close.
    fequal. (* Solve first trm_let *)
    fequal. (* Solve second trm_let *)
    fequal. (* Solve second trm_read *)
    (* Variable of the second read *)
    rewrite if_apply_1 with (f := subst_var x0 y). (* Push substitution down the case analysis *)
    cases_if; sympl. rewrite subst_fresh. cases_if; auto.

    fresh_chain_solve Fr.
    
    assert (x2 \[ x0 -> y] <> y1).
    cases_if.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.

    rewrite If_r with (P:= x2 /[ x0 -> y] = y1). auto. auto.

    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
Qed.

Local Instance typing_mode':Mode := typing_mode.

Global Instance TransformSwapReadsSubstCompat: TransformSubstCompat typing_mode _ _ trf_swap_reads_local.
Proof.
  intros.
  introv Htrf Hsu1 Hsu2 Hti.
  hnf in *.
  destruct TEa2, TEb2.
  inversions Hti.
  inversions Hsu1.
  inversions Hsu2.

  rewrite <- H in H6.

  inversions Htrf.

  rewrite H15 in *.
  eapply eq_push_inv in H16. destructecs H16; subst.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H8.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H22.
  fresh_chain_cleanup Fr.

  forwards Htysu1: (substitution_epush ) (@subst_ty_trm typing_mode)    H24; fweauto.
  forwards Htysu2: (substitution_epush ) (@subst_ty_trm typing_mode)    H23; fweauto.

  fresh_chain_include_typing Fr H24.
  fresh_chain_include_typing Fr H23.

  rewrite fresh_element_eq_containing_renv_push_dom in Fr by apply true.
  rewrite fresh_element_eq_containing_renv_push_fv in Fr by apply true.
  rewrite fresh_element_eq_containing_renv_push_env in Fr by apply true.

  rewrite fresh_element_eq_containing_ctx_concat_r in Fr by apply true.
  rewrite fresh_element_eq_containing_ctx_concat_l in Fr by apply true.
  rewrite fresh_element_eq_containing_ctx_push_ctx in Fr by apply true.

  fresh_chain_add_var FrG y2. 2: fresh_chain_solve Fr.
  fresh_chain_add_var FrG y1. 2: fresh_chain_solve Fr.


  (* Make it easier to solve ctx_ok with substitution *)
  rewrite fresh_element_eq_containing_refl with (fe:=(fresh_element_ctxdom G3)) in FrG by apply true .
  rewrite fresh_element_eq_ctxdom_map with (f:=subst_var x0 y) in FrG at 2 by apply true.



  assert (ctx_ok (G0 §& subst_ctx x0 y G3)) as HokGs. {
  eapply ctx_ok_concat_map. fresh_chain_apply FrG. fresh_chain_solve FrG.
}

(* Rewrite one side into the form needed by trf_swap_reads_local_at *)

  rewrite trf_swap_reads_local_subst_rewrite. 2: fresh_chain_solve Fr.
  rewrite trf_swap_reads_local_subst_rewrite.

  eapply trf_swap_reads_local_at.
  5: eauto. (* Solve first read typing, this resolves ?T1 *)
  5: eauto. (* Solve first second typing, this resolves ?T2 *)
  - auto.
  - auto.
  - eapply closed_subst_var; eauto.
  - fresh_solve;
    rewrite ? ctx_fv_concat; try apply notin_fv_subst_var;
    try fresh_chain_solve Fr; try fresh_chain_solve FrG;  try fresh_chain_solve2 FrG.
    rewrite notin_union; split;  try fresh_chain_solve Fr.
    eapply notin_subst_ctx_fv;  try fresh_chain_solve Fr.
    rewrite notin_union; split;  try fresh_chain_solve Fr.
    eapply notin_subst_ctx_fv;  try fresh_chain_solve Fr.
  -

  rewrite <- ! ctx_concat_push.
  rewrite <- ! ctx_subst_push.

  eapply substitution_epush. eapply (@subst_ty_trm typing_mode).
  +
  ctx_flatten; eauto.
  +
    fresh_chain_solve FrG.
  + auto.
+
  eauto.
+
  ctx_flatten; eauto.
  eapply weaken_push.
  eapply weaken_push.
  eauto.

  fresh_chain_solve FrG.
  fresh_chain_apply FrG.
  fresh_chain_simpl; fresh_chain_solve FrG.

  fresh_chain_apply FrG. 
  fresh_chain_simpl; fresh_chain_solve FrG.

  fresh_chain_apply FrG. rewrite <- fresh_chain_eq_last_var_set by apply true.
   rewrite fresh_chain_eq_merge_var_ctxdompush by apply true.
   rewrite fresh_chain_eq_merge_var_ctxdom by apply true.
  fresh_chain_solve FrG.
  + intros.
    ctx_flatten in H8.
    eapply ctx_vis_push_n_inv in H8.
    eapply ctx_vis_push_n_inv in H8.
    ctx_flatten.
    eapply ctx_vis_n.   fresh_chain_solve Fr.
    eapply ctx_vis_n.   fresh_chain_solve Fr.
    eauto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
  - fresh_chain_solve Fr.
Qed.

Lemma seal_trf_swap_reads_local x1 m1 x2 m2 t x:
  (trm_let (trm_read (avar_f x1) m1 ) (trm_let (trm_read (avar_f x2) m2) t ) ) /[# 0 ->> x] =
  (trm_let (trm_read (avar_f x1) m1 ) (trm_let (trm_read (avar_f x2) m2 ) t /[# 2 ->> x] ) )  .
Proof.  
  sympl. auto.
Qed.


Global Instance TransformSwapReadsSealing: TransformSealingCompat _ _ trf_swap_reads_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H.
  rename e0 into e.

  inversions Htrf.

  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H11.
  fresh_chain_cleanup Fr.


  rename H10 into Hclt, H12 into Hty1, H13 into Hty2, H14 into Htyt.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  rewrite seal_trf_swap_reads_local.
  rewrite seal_trf_swap_reads_local.

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  erewrite <- ! close_vars_seal_commut.
  2: simpl; lia.
  2: fresh_chain_solve Fr.
  2: simpl; lia.
  2: fresh_chain_solve Fr.

  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  assert (closed t /[ y1 -> z1] /[ y2 -> z2]) as Hclts. {
    eapply closed_subst_var; eapply closed_subst_var; eauto.
  }
  eapply trf_swap_reads_local_at with (T1:=T1/[# 0 ->> x]) (T2:=T2/[# 0 ->> x]).
  - apply ctx_seal_ok. auto.
  - auto.
  - erewrite closed_sealed. eapply Hclts. eapply Hclts.
  - rewrite <- ctx_seal_dom. fresh_solve; try eapply fv_seal_notin; try eapply ctx_fv_seal_notin; try fresh_chain_solve Fr.
  - replace (trm_read (avar_f x1) f1) with ((trm_read (avar_f x1) f1)/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.
  - replace (trm_read (avar_f x2) f2) with ((trm_read (avar_f x2) f2)/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.
  - erewrite closed_sealed with (X:=(t /[ y1 -> z1] /[ y2 -> z2])) by eapply Hclts.
    erewrite <- closed_sealed with (X:=(t /[ y1 -> z1] /[ y2 -> z2])) (n:=0) (y:=x) by eapply Hclts.
    rewrite <- ctx_seal_push.
    rewrite <- ctx_seal_push.

    eapply sealing_obj with (Go:=objctx_empty).
    eapply rename_subst_trm_2.
    auto.
    auto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.

Unshelve.

5: eapply TransformSealClose.
5: eapply TransformSealClose.
all: eapply sealing_ty_trm.
Qed.


Global Instance TransformSwapReadsUnhiding: TransformUnhidingCompat _ _ trf_swap_reads_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H.
  rename e0 into e.

  inversions Htrf.
  clears G0 G3.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H9.
  fresh_chain_cleanup Fr.

  rename H8 into Hclt, H10 into Hty1, H11 into Hty2, H12 into Htyt.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.
 
  fresh_chain_add_gather Fr. eapply ctx_ok_concat_inv in H6. destructecs H6. auto. eapply ctx_ok_concat_inv in H6. destructecs H6. inversion H0. auto.

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_reads_local_at with (T1:=T1) (T2:=T2).
  - eapply ctx_ok_hide_l_concat_inv; eauto.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.

  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply unhiding with (Go:=objctx_empty); eauto.
    ctx_flatten.
    eapply rename_subst_trm_2.
    auto.
    auto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
Unshelve.
  eapply unhiding_ty_trm.
  eapply unhiding_ty_trm.
  eapply unhiding_ty_trm.
Qed.


Section Tests.

(* Example instance of transform_swap_reads *)
Lemma test_local_transfrom_swap_reads x1 f1 x2 f2 m1:
  x1 <> x2 ->
  〈 (ctx_empty §ʷ x1 ~ (typ_rcd (dec_fld f1 typ_top typ_top)) §ʷ x2 ~ (typ_rcd (dec_fld f2 typ_top (typ_rcd (dec_met m1 typ_top typ_top typ_top))))) ⁏ empty ⊢ (
    trm_let (trm_read (avar_f x1) f1) (trm_let (trm_read (avar_f x2) f2) (trm_call (avar_b 0) m1 (avar_b 1)))
  ) ∶ typ_top 〉 ∼ trf_swap_reads_local ∼>
  〈 (ctx_empty §ʷ x1 ~ (typ_rcd (dec_fld f1 typ_top typ_top)) §ʷ x2 ~ (typ_rcd (dec_fld f2 typ_top (typ_rcd (dec_met m1 typ_top typ_top typ_top))))) ⁏ empty ⊢  (
    trm_let (trm_read (avar_f x2)  f2) (trm_let (trm_read (avar_f x1) f1) (trm_call (avar_b 1) m1 (avar_b 0)))
  ) ∶ typ_top 〉.
Proof.
  introv Hnx12.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx12.
  fresh_chain_pick Fr x.
  fresh_chain_pick Fr y.

  assert (x <> y) by fresh_chain_solve Fr.
  assert (x1 <> x) by fresh_chain_solve Fr.
  assert (x2 <> y) by fresh_chain_solve Fr.
  assert (x1 <> x2) by fresh_chain_solve Fr.

  trm_let_find_open_with x. 2: sympls; fresh_chain_solve Fr.
  trm_let_find_open_with y. 2: sympls; fresh_chain_solve Fr.
  trm_let_find_open_with y. 2: sympls; fresh_chain_solve Fr.
  trm_let_find_open_with x. 2: sympls; fresh_chain_solve Fr.

  eapply trf_swap_reads_local_at with (T1 := typ_top) (T2:= (typ_rcd (dec_met m1 typ_top typ_top typ_top))).
  + sympls.
    fresh_chain_solve Fr.
  + eauto.
  + unfold closed. intros; sympls; eauto.
  + fresh_chain_solve Fr.
  + destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva.
    * eapply ty_trm_read. eauto. eapply ty_var_var; eauto. unfold viewpoint_adapt. rewrite Hmva. eauto.
    * eapply ty_trm_sub.
      eapply ty_trm_read. intro. constructor; eauto. eapply ty_var_var. eauto. unfold viewpoint_adapt. rewrite Hmva. constructor. eauto. eauto. eauto.
  + destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva.
    * eapply ty_trm_read. eauto. eapply ty_var_var; eauto. unfold viewpoint_adapt. rewrite Hmva. eauto.
    * eapply ty_trm_sub.
      eapply ty_trm_read. intro. constructor; eauto. eapply ty_var_var. eauto. unfold viewpoint_adapt. rewrite Hmva. eauto. constructor. eauto. eauto. eauto.
  + eapply ty_trm_sub. eapply ty_trm_call. eauto. eauto. eauto. eauto. sympls. eauto. sympls. eauto. simpl. eauto.
Qed.

(* A bigger test with swapping inside a method *)

(* Field 1 - some value *)
Local Notation Tf1 f1 := (typ_rcd (dec_fld f1 typ_top typ_top)) (only parsing).
Local Notation df1 f1 := (def_fld f1 (avar_b 0)) (only parsing).
(* Field 2 - a method *)
Local Notation Tf2 f2 m2 := (typ_rcd (dec_fld f2 typ_top (typ_rcd (dec_met m2 typ_top typ_top typ_top)))) (only parsing).
(* Method which will contain the swapping *)
Local Notation Tm1 m1 f2 m2 := (typ_rcd (dec_met m1 typ_top (Tf2 f2 m2) typ_top)) (only parsing).
(* Object which will contain m1 and f1 *)
Local Notation To1 f1 m1 f2 m2 := (typ_and (Tm1 m1 f2 m2) (Tf1 f1)) (only parsing).

Lemma test_transfrom_swap_reads f1 f2 m1 m2:
  〈 ctx_empty ⁏ empty ⊢ (
    trm_lit (lit_obj (To1 f1 m1 f2 m2) (defs_cons (defs_cons defs_nil (def_met m1
      (trm_let (trm_read (avar_b 1 (*recv of m1*)) f1) (trm_let (trm_read (avar_b 1 (*arg of m1*)) f2) (trm_call (avar_b 0 (*read f2*)) m2 (avar_b 1 (*read f1*)))))
    )) (df1 f1))) (trm_var (avar_b 0))
  ) ∶ typ_top 〉 ∼ (trf_local_idtyp_trm typing_mode trf_swap_reads_local) ∼>
  〈 ctx_empty ⁏ empty ⊢ (
    trm_lit (lit_obj (To1 f1 m1 f2 m2) (defs_cons (defs_cons defs_nil (def_met m1
      (trm_let (trm_read (avar_b 0 (*arg of m1*)) f2) (trm_let (trm_read (avar_b 2 (*recv of m1*)) f1) (trm_call (avar_b 1 (*read f2*)) m2 (avar_b 0 (*read f1*)))))
    )) (df1 f1))) (trm_var (avar_b 0))
  ) ∶ typ_top 〉.
Proof.
  eapply trf_local_idtyp_trm_litl with (L:= \{}) (Tl := (typ_objlit typing_mode (To1 f1 m1 f2 m2))).
  2: {
    intros. sympls. repeat cases_if.
    eapply ty_trm_var. eauto. eauto.
    eapply ty_trm_var. eauto. eauto.
  }
  eapply trf_local_idtyp_lit_obj with (L:=\{}).
  2: eapply typ_indep_and; eauto 10.
  intros; sympls. repeat cases_if.
  sympls.
  eapply trf_local_idtyp_defs_consl.
  {
    eapply trf_local_idtyp_defs_one.
    eapply trf_local_idtyp_def_met with (L:=\{x}).
    intros. sympls. ctx_split_from xz.
    match goal with |- context[?G1 §& ?G2] => set (Gobj:=G1) end.
    ctx_flatten.

    repeat (cases_if; sympls).

    eapply trf_local_idtyp_trm_local.
    pick_fresh y1.
    pick_fresh y2.
    assert (y1 <> y2). notin_solve.
    assert (xc <> y2). notin_solve.
    assert (xz <> y1). notin_solve.


  trm_let_find_open_with y1. 2: sympls; eauto.
  trm_let_find_open_with y2.  2: sympls; eauto.
  trm_let_find_open_with y2.  2: sympls; eauto.
  trm_let_find_open_with y1.  2: sympls; eauto.


    eapply trf_swap_reads_local_at with (T1:=typ_top) (T2:=(typ_rcd (dec_met m2 typ_top typ_top typ_top))) (T:=typ_top).
    +  unfold Gobj. destruct ( mode_visibility' typing_mode ); constructor; try constructor; eauto.
    + eauto.
    +
      unfold closed.
    intros.
    simpls. eauto.
    + unfold Gobj. cases_if; ctx_simpl_dom; ctx_simpl_fv; fresh_solve; eauto; rewrite fv_env_empty; notin_solve.
    + destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva; try typing_mode_contradiction typing_mode.
      * eapply ty_trm_read. eauto.
        eapply ty_var_sub. eapply ty_var_var; eauto. eapply subtyp_trans. eapply subtyp_and12.
        eapply subtyp_trans. eapply subtyp_and11. eapply subtyp_and12. unfold viewpoint_adapt. rewrite Hmva. eauto.
      * eapply ty_trm_sub.
        eapply ty_trm_read. intro. constructor; eauto.
        eapply ty_var_sub. eapply ty_var_var; eauto. eapply subtyp_trans. eapply subtyp_and12.
        eapply subtyp_trans. eapply subtyp_and11. eapply subtyp_and12. unfold viewpoint_adapt. rewrite Hmva. constructor. eauto. eauto.  eauto.
    + destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva; try typing_mode_contradiction typing_mode.
      * eapply ty_trm_read. eauto. eapply ty_var_var; eauto. unfold viewpoint_adapt. rewrite Hmva. eauto.
      * eapply ty_trm_sub.
        eapply ty_trm_read. intro. constructor; eauto.
        eapply ty_var_var; eauto.
        unfold viewpoint_adapt. rewrite Hmva. constructor. eauto. eauto.  eauto.
    + eapply ty_trm_sub. eapply ty_trm_call. eauto. eauto. eauto. eauto. sympls. eauto. sympls. eauto. simpl. eauto.
    + eapply subtyp_refl.
  }
  eauto.
  unfold defs_hasnt; sympls; cases_if; auto.
  unfold defs_hasnt; sympls; cases_if; auto.
  eapply subtyp_top.
  Qed.
End Tests.


End WithTypingMode.
