Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
(** Freshness chain. *)
(*
This is designed in order to represent multiple assertions about sets of variables (mainly freshness)
while avoiding duplicate occurrences of variables and sets.

The idea is that often, there is machine configuration or parts of it,
and there are variables fresh for those parts.
(That can be expressed by a `fresh` assertion.)

If there are more parts, such as an additional term, there may be variables that are fresh for both the original parts and the new parts too.
If expressed by a freshness assertion, this assertion would also contain all the elements from the original freshness assertion,
usually in a form of a set union.

If this process is repeated multiple times, there would be many freshness assertions which would have a lot of parts in common.
This would make it necessary to work with multiple assertions involving large unions, which has these disadvantages:
- many hypotheses in the proof context
- long hypotheses (long unions)
- repetitive hypotheses (the same unions appearing in multiple hypotheses)
- the unions may have different associativity
- to get the intended goal, the unions have to be rewritten into an disjunction
- to get the intended goal, unions have to be destructed in multiple steps

This system allows compacting multiple assertions into a single "freshness chain".
The chain is built right-to-left and contains "sets" and "variables", where each "variable" is fresh for all the elements to the right of it.
Aditionally, environments and context included in the chain are ok.

The avoidance of duplicates is achieved by having the chain represent two things:
a) A set of variables.
b) A set of freshness assertions about variables within that set.

The chain works best if it deals with individual variables, sets and syntax elements.
We want to avoid set unions or environment concatenations, or lists of variables, which would hinder accessing their individual parts.
To keep a chain in such form, we may have to split an element of the chain into multiple elements.
This can be achieved by setoid rewriting.

The rewriting is facilitated by a concept of equivalence, where two chains are equivalent if they represent the same set of variables and the same set of assertions.
However, it may also be useful to rewrite a chain and add or remove some assertions or variables.
For example, if we have a chain as a hypothesis, it is safe to remove any variables from it (which will make the sets of variables and assertions smaller).

Therefore, the equivalence is split into four properties - two directions of subset for variables and two for assertions - and the properties can be arbitrarily combined.

*)

Inductive fresh_element: Type :=
| fresh_element_var (x: var): fresh_element
| fresh_element_list (x: list var): fresh_element
| fresh_element_set (X: fset var): fresh_element
| fresh_element_element (A: Set) {AF: FreeVar A} (a: A): fresh_element
| fresh_element_ctx (G: ctx): fresh_element
| fresh_element_ctxdom (G: ctx): fresh_element
| fresh_element_renv (e: renv): fresh_element
| fresh_element_renvdom (e: renv): fresh_element
| fresh_element_heap (Si: heap): fresh_element
| fresh_element_heapdom (Si: heap): fresh_element
| fresh_element_objctx (Go: objctx): fresh_element
| fresh_element_objctxdom (Go: objctx): fresh_element
| fresh_element_containing (feM feSub: fresh_element): fresh_element (* Like just feM, but additionally asserts that feSub \c feM *)
.
Inductive fresh_chain: Type :=
| fresh_chain_nil: fresh_chain
| fresh_chain_set (fe: fresh_element) (fc: fresh_chain): fresh_chain
| fresh_chain_var (fe: fresh_element) (fc: fresh_chain) : fresh_chain
| fresh_chain_subset (fe: fresh_element) (fc: fresh_chain): fresh_chain
| fresh_chain_branch (fcBr fcM: fresh_chain): fresh_chain
| fresh_chain_parallel (fcBr1 fcBr2 fcM: fresh_chain): fresh_chain
| fresh_chain_union (fcA fcB: fresh_chain): fresh_chain
.

Fixpoint fresh_element_to_set (fe: fresh_element): fset var :=
  match fe with
  | fresh_element_var x => \{x}
  | fresh_element_list xs => from_list xs
  | fresh_element_set X => X
  | fresh_element_element a => fv a
  | fresh_element_ctx G => ctx_dom G \u ctx_fv G
  | fresh_element_ctxdom G => ctx_dom G
  | fresh_element_renv e => dom e \u fv_env e
  | fresh_element_renvdom e => dom e
  | fresh_element_heap Si => dom Si \u fv_env Si
  | fresh_element_heapdom Si => dom Si
  | fresh_element_objctx Go => objctx_dom Go \u objctx_fv Go
  | fresh_element_objctxdom Go => objctx_dom Go
  | fresh_element_containing feM feSub => fresh_element_to_set feM 
  end.

Fixpoint fresh_chain_to_set (fc: fresh_chain): fset var :=
  match fc with
  | fresh_chain_nil => \{}
  | fresh_chain_set fe fc => (fresh_element_to_set fe) \u (fresh_chain_to_set fc)
  | fresh_chain_var fe fc => (fresh_element_to_set fe) \u (fresh_chain_to_set fc)
  | fresh_chain_subset fe fc => (fresh_chain_to_set fc) 
  | fresh_chain_branch fcBr fcM => (fresh_chain_to_set fcM)
  (* Avoid combining branches and unions because of potential size explosion *)
  | fresh_chain_union fc1 fc2 => (fresh_chain_to_set fc1) \u (fresh_chain_to_set fc2) 
  | fresh_chain_parallel fcBr1 fcBr2 fcM => (fresh_chain_to_set fcBr1) \u (fresh_chain_to_set fcBr2) \u (fresh_chain_to_set fcM)
  end.

Fixpoint fresh_element_to_set_large (fe: fresh_element): fset var :=
  match fe with
  | fresh_element_var x => \{x}
  | fresh_element_list xs => from_list xs
  | fresh_element_set X => X
  | fresh_element_element a => fv a
  | fresh_element_ctx G => ctx_dom G \u ctx_fv G
  | fresh_element_ctxdom G => ctx_dom G
  | fresh_element_renv e => dom e \u fv_env e
  | fresh_element_renvdom e => dom e
  | fresh_element_heap Si => dom Si \u fv_env Si
  | fresh_element_heapdom Si => dom Si
  | fresh_element_objctx Go => objctx_dom Go \u objctx_fv Go
  | fresh_element_objctxdom Go => objctx_dom Go
  | fresh_element_containing feM feSub => fresh_element_to_set_large feM \u fresh_element_to_set_large feSub
  end.

Fixpoint fresh_chain_to_set_large (fc: fresh_chain): fset var :=
  match fc with
  | fresh_chain_nil => \{}
  | fresh_chain_set fe fc => (fresh_element_to_set_large fe) \u (fresh_chain_to_set_large fc)
  | fresh_chain_var fe fc => (fresh_element_to_set_large fe) \u (fresh_chain_to_set_large fc)
  | fresh_chain_subset fe fc => (fresh_element_to_set_large fe) \u (fresh_chain_to_set_large fc) (* fe is included here although it follows from the condition, to make *)
  | fresh_chain_branch fcBr fcM => (fresh_chain_to_set_large fcM)
  (* Avoid combining branches and unions because of potential size explosion *)
  | fresh_chain_union fc1 fc2 => (fresh_chain_to_set_large fc1) \u (fresh_chain_to_set_large fc2) 
  | fresh_chain_parallel fcBr1 fcBr2 fcM => (fresh_chain_to_set_large fcBr1) \u (fresh_chain_to_set_large fcBr2) \u (fresh_chain_to_set_large fcM)
  end.

Fixpoint fresh_chain_append (fcA fcB: fresh_chain): fresh_chain :=
  match fcA with
  | fresh_chain_nil => fcB
  | fresh_chain_set fe fc => fresh_chain_set fe (fresh_chain_append fc fcB)
  | fresh_chain_var fe fc => fresh_chain_var fe (fresh_chain_append fc fcB)
  | fresh_chain_subset fe fc => fresh_chain_subset fe (fresh_chain_append fc fcB)
  | fresh_chain_branch fcBr fcM => fresh_chain_branch fcBr (fresh_chain_append fcM fcB)
  (* Avoid combining branches and unions because of potential size explosion *)
  | fresh_chain_union fc1 fc2 => fresh_chain_union (fresh_chain_append fc1 fcB) (fresh_chain_append fc2 fcB)
  | fresh_chain_parallel fcBr1 fcBr2 fcM => fresh_chain_parallel fcBr1 fcBr2 (fresh_chain_append fcM fcB)
  end.

Fixpoint fresh_chain_out (fc: fresh_chain): fresh_chain := (* Excluding subset assertions *)
  match fc with
  | fresh_chain_nil => fc
  | fresh_chain_set fe fc => fresh_chain_set fe (fresh_chain_out fc)
  | fresh_chain_var fe fc => fresh_chain_var fe (fresh_chain_out fc)
  | fresh_chain_subset fe fc => fresh_chain_set fe (fresh_chain_out fc)
  | fresh_chain_branch fcBr fcM => fresh_chain_branch (fresh_chain_out fcBr) (fresh_chain_out fcM)
  (* Avoid combining branches and unions because of potential size explosion *)
  | fresh_chain_union fc1 fc2 => fresh_chain_union (fresh_chain_out fc1) (fresh_chain_out fc2)
  | fresh_chain_parallel fcBr1 fcBr2 fcM => fresh_chain_parallel (fresh_chain_out fcBr1) (fresh_chain_out fcBr2) (fresh_chain_out fcM)
  end.

(* Appending rewriting *)
Lemma fresh_chain_append_nil_l fc:
  fresh_chain_append fresh_chain_nil fc = fc.
Proof. simpl. auto. Qed.
Lemma fresh_chain_append_nil_r fc:
  fresh_chain_append fc fresh_chain_nil = fc.
Proof.
  induction fc; simpl; rewrite ? IHfc, ? IHfc1, ? IHfc2, ? IHfc3; auto.
Qed.
Lemma fresh_chain_append_assoc fc1 fc2 fc3:
  fresh_chain_append (fresh_chain_append fc1 fc2) fc3 = fresh_chain_append fc1 (fresh_chain_append fc2 fc3).
Proof.
  induction fc1; simpl; rewrite ? IHfc1, ? IHfc1_1, ? IHfc1_2, ? IHfc1_3; auto.
Qed.
Lemma fresh_chain_append_to_set fc1 fc2:
  fresh_chain_to_set (fresh_chain_append fc1 fc2) = (fresh_chain_to_set fc1) \u (fresh_chain_to_set fc2).
Proof.
  induction fc1; simpls.
  - rewrite union_empty_l. auto.
  - rewrite IHfc1. rewrite union_assoc. auto.
  - rewrite IHfc1. rewrite union_assoc. auto.
  - rewrite IHfc1. auto.
  - rewrite IHfc1_2. auto.
  - rewrite IHfc1_3. rewrite ! union_assoc. auto.
  - rewrite IHfc1_1. rewrite IHfc1_2. rewrite union_zip. rewrite union_same. auto.
Qed.
Lemma fresh_chain_append_to_set_large fc1 fc2:
  fresh_chain_to_set_large (fresh_chain_append fc1 fc2) = (fresh_chain_to_set_large fc1) \u (fresh_chain_to_set_large fc2).
Proof.
  induction fc1; simpls.
  - rewrite union_empty_l. auto.
  - rewrite IHfc1. rewrite union_assoc. auto.
  - rewrite IHfc1. rewrite union_assoc. auto.
  - rewrite IHfc1. auto.
  - rewrite IHfc1_2. auto.
  - rewrite IHfc1_3. rewrite ! union_assoc. auto.
  - rewrite IHfc1_1. rewrite IHfc1_2. rewrite union_zip. rewrite union_same. auto.
Qed.

(* Element decomposition *)
Fixpoint fresh_element_part_list (fe: fresh_element): list var :=
  match fe with
  | fresh_element_var x => x::nil
  | fresh_element_list xs => xs
  | fresh_element_set X => nil
  | fresh_element_element a => nil
  | fresh_element_ctx G => ctx_keys G
  | fresh_element_ctxdom G => ctx_keys G
  | fresh_element_renv e => keys e
  | fresh_element_renvdom e => keys e
  | fresh_element_heap Si => keys Si
  | fresh_element_heapdom Si => keys Si
  | fresh_element_objctx Go => objctx_keys Go
  | fresh_element_objctxdom Go => objctx_keys Go
  | fresh_element_containing feM feSub => fresh_element_part_list feM
  end.
Fixpoint fresh_element_part_set (fe: fresh_element): fset var :=
  match fe with
  | fresh_element_var x => \{}
  | fresh_element_list xs => \{}
  | fresh_element_set X => X
  | fresh_element_element a => fv a
  | fresh_element_ctx G => ctx_fv G
  | fresh_element_ctxdom G => \{}
  | fresh_element_renv e => fv_env e
  | fresh_element_renvdom e => \{}
  | fresh_element_heap Si => fv_env Si
  | fresh_element_heapdom Si => \{}
  | fresh_element_objctx Go => objctx_fv Go
  | fresh_element_objctxdom Go => \{}
  | fresh_element_containing feM feSub => fresh_element_part_set feM
  end.

Lemma fresh_element_parts_to_set fe:
  (from_list (fresh_element_part_list fe)) \u (fresh_element_part_set fe) = (fresh_element_to_set fe).
Proof.
  induction fe; simpls; rewrite ? from_list_cons, ? from_list_nil, ? union_empty_r, ? union_empty_l; auto.
  -
  rewrite ctx_keys_dom; auto.
  -
  rewrite ctx_keys_dom; auto.
  -
  rewrite env_from_list_keys_dom; auto.
  -
  rewrite env_from_list_keys_dom; auto.
  -
  rewrite env_from_list_keys_dom; auto.
  -
  rewrite env_from_list_keys_dom; auto.
  -
  rewrite objctx_keys_dom; auto.
  -
  rewrite objctx_keys_dom; auto.
Qed.



Inductive fresh_element_fresh: fresh_element -> fresh_element -> Prop :=
  fresh_element_fresh_c fey fex:
    fresh (fresh_element_to_set fex) (length (fresh_element_part_list fey)) (fresh_element_part_list fey) ->
    disjoint (fresh_element_part_set fey) (fresh_element_to_set fex) ->
    fresh_element_fresh fey fex
.
Lemma fresh_element_fresh_var_var y x:
    y <> x ->
    fresh_element_fresh (fresh_element_var y) (fresh_element_var x).
Proof.
  introv Hne.
  constructor; simpls.
  - rewrite notin_singleton; auto.
  - eapply disjoint_empty_l.
Qed.

Hint Constructors fresh_element_fresh: fresh_element.

Inductive fresh_chain_fresh: fresh_element -> fresh_chain -> Prop :=
| fresh_chain_fresh_nil x:
    fresh_chain_fresh x fresh_chain_nil
| fresh_chain_fresh_set x fc fe:
    fresh_element_fresh x fe ->
    fresh_chain_fresh x fc ->
    fresh_chain_fresh x (fresh_chain_set fe fc)
| fresh_chain_fresh_var x fc fe:
    fresh_element_fresh x fe ->
    fresh_chain_fresh x fc ->
    fresh_chain_fresh x (fresh_chain_var fe fc)
| fresh_chain_fresh_subset x fc fe:
    fresh_element_fresh x fe ->
    fresh_chain_fresh x fc ->
    fresh_chain_fresh x (fresh_chain_subset fe fc)
| fresh_chain_fresh_branch x fcBr fcM: (* do not go into the branch *)
    fresh_chain_fresh x fcM ->
    fresh_chain_fresh x (fresh_chain_branch fcBr fcM)
| fresh_chain_fresh_union x fcA fcB:
    fresh_chain_fresh x fcA ->
    fresh_chain_fresh x fcB ->
    fresh_chain_fresh x (fresh_chain_union fcA fcB)
| fresh_chain_fresh_parallel x fcA fcB fcM:
    fresh_chain_fresh x (fresh_chain_append fcA fcM) ->
    fresh_chain_fresh x fcB ->
    fresh_chain_fresh x (fresh_chain_parallel fcA fcB fcM)
.

Hint Constructors fresh_chain_fresh: core.


Inductive fresh_element_is: fresh_element -> Prop :=
| fresh_element_is_var x: fresh_element_is (fresh_element_var x)
| fresh_element_is_list xs:
    fresh \{} (length xs) xs ->
    fresh_element_is (fresh_element_list xs)
| fresh_element_is_set X: fresh_element_is (fresh_element_set X)
| fresh_element_is_element (A: Set) {AF: FreeVar A} (a: A): fresh_element_is (fresh_element_element a)
| fresh_element_is_ctx (G: ctx):
    ctx_ok G ->
    fresh_element_is (fresh_element_ctx G)
| fresh_element_is_renv (e: renv):
    ok e ->
    fresh_element_is (fresh_element_renv e)
| fresh_element_is_heap (Si: heap):
    ok Si ->
    fresh_element_is (fresh_element_heap Si)
| fresh_element_is_objctx (Go: objctx):
    ctx_ok (ctx_empty §§& Go) ->
    fresh_element_is (fresh_element_objctx Go)
| fresh_element_is_ctxdom (G: ctx):
    ctx_ok G ->
    fresh_element_is (fresh_element_ctxdom G)
| fresh_element_is_renvdom (e: renv):
    ok e ->
    fresh_element_is (fresh_element_renvdom e)
| fresh_element_is_heapdom (Si: heap):
    ok Si ->
    fresh_element_is (fresh_element_heapdom Si)
| fresh_element_is_objctxdom (Go: objctx):
    ctx_ok (ctx_empty §§& Go) ->
    fresh_element_is (fresh_element_objctxdom Go)
| fresh_element_is_containing feM feSub:
  fresh_element_is feM ->
  fresh_element_is feSub ->
  fresh_element_to_set feSub \c fresh_element_to_set feM ->
  fresh_element_is (fresh_element_containing feM feSub)
.

Hint Constructors fresh_element_is: core.

Inductive fresh_chain_is: fresh_chain -> Prop :=
| fresh_chain_is_nil: fresh_chain_is fresh_chain_nil
| fresh_chain_is_var fe fc:
    fresh_chain_is fc ->
    fresh_element_is fe ->
    fresh_chain_fresh fe fc ->
    fresh_chain_is (fresh_chain_var fe fc)
| fresh_chain_is_set fe fc:
    fresh_chain_is fc ->
    fresh_element_is fe ->
    fresh_chain_is (fresh_chain_set fe fc)
| fresh_chain_is_subset fe fc:
    fresh_chain_is fc ->
    fresh_element_is fe ->
    fresh_element_to_set fe \c fresh_chain_to_set fc ->
    fresh_chain_is (fresh_chain_subset fe fc)
| fresh_chain_is_branch fcBr fcM:
    fresh_chain_is (fresh_chain_append fcBr fcM) ->
    fresh_chain_is (fresh_chain_branch fcBr fcM)
| fresh_chain_is_union fcA fcB:
    fresh_chain_is fcA ->
    fresh_chain_is fcB ->
    fresh_chain_is (fresh_chain_union fcA fcB)
| fresh_chain_is_parallel fcA fcB fcM:
    fresh_chain_is (fresh_chain_append fcA fcM) ->
    fresh_chain_is (fresh_chain_append fcB fcM) ->
    fresh_chain_is (fresh_chain_parallel fcA fcB fcM)
.

Hint Constructors fresh_chain_is: core.


Declare Custom Entry fresh_chain_custom.
Notation "'•'" := (fresh_chain_nil) (in custom fresh_chain_custom at level 1).
Notation "A" := (fresh_element_set A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_var A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_list A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_element A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_ctx A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "'∂' A" := (fresh_element_ctxdom A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_objctx A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "'∂' A" := (fresh_element_objctxdom A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_renv A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "'∂' A" := (fresh_element_renvdom A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "A" := (fresh_element_heap A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "'∂' A" := (fresh_element_heapdom A) (in custom fresh_chain_custom at level 0, A global, only printing).
Notation "fe ; fc" := (fresh_chain_set fe fc) (in custom fresh_chain_custom at level 1, right associativity, fc custom fresh_chain_custom, fe custom fresh_chain_custom).
Notation "fe ∶ fc" := (fresh_chain_var fe fc) (in custom fresh_chain_custom at level 1, right associativity, fc custom fresh_chain_custom, fe custom fresh_chain_custom).
Notation "fe ⊂ fc" := (fresh_chain_subset fe fc) (in custom fresh_chain_custom at level 1, right associativity, fc custom fresh_chain_custom, fe custom fresh_chain_custom).
Notation "'❴' fcA '|' fcB '❵'" := (fresh_chain_union fcA fcB) (in custom fresh_chain_custom at level 1, right associativity, fcA custom fresh_chain_custom, fcB custom fresh_chain_custom).
Notation "'⟦' fcBr '⟧' fcM" := (fresh_chain_branch fcBr fcM) (in custom fresh_chain_custom at level 1, right associativity, fcBr custom fresh_chain_custom, fcM custom fresh_chain_custom).
Notation "feS ∈ feM" := (fresh_element_containing feM feS) (in custom fresh_chain_custom at level 1, right associativity, feM custom fresh_chain_custom, feS custom fresh_chain_custom).

Notation "'❰' fc '❱'" := (fresh_chain_is fc) (fc custom fresh_chain_custom at level 2).


Record fresh_chain_eq_flags := {
  fresh_chain_eq_flag_impl: bool; (* The assertions of the left chain imply the assertions of the right chain *)
  fresh_chain_eq_flag_rev_impl: bool; (* The assertions of the right chain imply the assertions of the left chain *)
  fresh_chain_eq_flag_subset: bool; (* The set of variables represented by the left chain is a subset of the set represented by the right chain *)
  fresh_chain_eq_flag_superset: bool
}.

Declare Scope fresh_chain_flags_scope.
Open Scope fresh_chain_flags_scope.
Notation "⚐ a b c d ⚐" := ({|
  fresh_chain_eq_flag_impl := a;
  fresh_chain_eq_flag_rev_impl := b;
  fresh_chain_eq_flag_subset := c;
  fresh_chain_eq_flag_superset := d |}) (at level 8, a at level 5, b at level 5, c at level 5, d at level 5): fresh_chain_flags_scope.


(* The relation has multiple flags so that all the properties can be shown without branching. *)
Inductive fresh_chain_eq flags fcA fcB: Prop :=
fresh_chain_eq_c:
  (if fresh_chain_eq_flag_impl flags then fresh_chain_is fcA -> fresh_chain_is fcB else True) ->
  (if fresh_chain_eq_flag_rev_impl flags then fresh_chain_is fcB -> fresh_chain_is fcA else True) ->
  (if fresh_chain_eq_flag_subset flags then fresh_chain_to_set fcA \c fresh_chain_to_set fcB else True) ->
  (if fresh_chain_eq_flag_superset flags then fresh_chain_to_set fcB \c fresh_chain_to_set fcA else True) ->
  fresh_chain_eq flags fcA fcB.
Inductive fresh_element_eq flags fcA fcB: Prop :=
fresh_element_eq_c:
  (if fresh_chain_eq_flag_impl flags then fresh_element_is fcA -> fresh_element_is fcB else True) ->
  (if fresh_chain_eq_flag_rev_impl flags then fresh_element_is fcB -> fresh_element_is fcA else True) ->
  (if fresh_chain_eq_flag_subset flags then fresh_element_to_set fcA \c fresh_element_to_set fcB else True) ->
  (if fresh_chain_eq_flag_superset flags then fresh_element_to_set fcB \c fresh_element_to_set fcA else True) ->
  fresh_element_eq flags fcA fcB.


Lemma fresh_chain_fresh_append fe fc1 fc2:
  fresh_chain_fresh fe fc1 ->
  fresh_chain_fresh fe fc2 ->
  fresh_chain_fresh fe (fresh_chain_append fc1 fc2).
Proof.
  introv Hfc1 Hfc2.
  induction Hfc1; simpls.
  - eauto.
  - constructor; eauto.
  - constructor; eauto.
  - constructor; eauto.
  - constructor; eauto.
  - constructor; eauto.
  - constructor.
    rewrite <- fresh_chain_append_assoc.
    eauto.
    eauto.
Qed.

Lemma fresh_chain_fresh_append_invr fe fc1 fc2:
  fresh_chain_fresh fe (fresh_chain_append fc1 fc2) ->
  fresh_chain_fresh fe fc2.
Proof.
  reverts fc2.
  induction fc1; intros; simpls; try inversions H; eauto.
Qed.

Lemma fresh_chain_fresh_append_invl fe fc1 fc2:
  fresh_chain_fresh fe (fresh_chain_append fc1 fc2) ->
  fresh_chain_fresh fe fc1.
Proof.
  reverts fc2.
  induction fc1; intros; simpls; try inversions H; eauto.
  econstructor.
  eapply fresh_chain_fresh_append.
  eapply IHfc1_1. eauto.
  eapply IHfc1_3. eapply fresh_chain_fresh_append_invr; eauto.
  eauto.
Qed.

Lemma fresh_chain_is_append_invr fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_is fc2.
Proof.
  reverts fc2.
  induction fc1; intros; simpls.
  - eauto.
  - inversions H. eauto.
  - inversions H. eauto using fresh_chain_fresh_append_invl.
  - inversions H. eauto.
  - inversions H.
    eapply IHfc1_2.
    eapply IHfc1_1. eauto.
  - inversions H.
    eapply IHfc1_3.
    eapply IHfc1_1. eauto.
  - inversions H.
    eapply IHfc1_2.
    eauto.
Qed.

(* Large and small sets *)
Lemma fresh_element_to_set_subset_large fe:
  fresh_element_to_set fe \c fresh_element_to_set_large fe.
Proof.
  induction fe; simpls; union_sub_solver.
Qed.
Lemma fresh_chain_to_set_subset_large fe:
  fresh_chain_to_set fe \c fresh_chain_to_set_large fe.
Proof.
  induction fe; simpls; try union_sub_solver.
  - eapply subset_union_2. eapply fresh_element_to_set_subset_large. auto.
  - eapply subset_union_2. eapply fresh_element_to_set_subset_large. auto.
Qed.

Lemma fresh_element_to_set_large_subset fe:
  fresh_element_is fe ->
  fresh_element_to_set_large fe \c fresh_element_to_set fe.
Proof.
  introv Hfe.
  induction Hfe; simpls; try union_sub_solver.
  - eapply subset_union_l. auto. eapply subset_trans; eauto.
Qed.

Lemma fresh_chain_to_set_large_subset_app fc fc2:
  fresh_chain_is (fresh_chain_append fc fc2) ->
  fresh_chain_to_set_large fc2 \c fresh_chain_to_set fc2 ->
  fresh_chain_to_set_large (fresh_chain_append fc fc2) \c fresh_chain_to_set (fresh_chain_append fc fc2).
Proof.
  introv His Hsu2.
  reverts fc2.
  induction fc; intros; simpls; try union_sub_solver.
  - inversions His. eapply subset_union_2. eapply fresh_element_to_set_large_subset. auto. auto.
  - inversions His. eapply subset_union_2. eapply fresh_element_to_set_large_subset. auto. auto.
  - inversions His. eapply subset_union_l.
    eapply subset_trans. eapply fresh_element_to_set_large_subset; eauto. auto. eauto.
  - inversions His. eapply IHfc2. eapply fresh_chain_is_append_invr; eauto. auto.
  - inversions His.
    forwards: fresh_chain_is_append_invr H1.
    forwards: IHfc3; fweauto.
    forwards: IHfc2; fweauto.
    forwards: IHfc1; fweauto.
    rewrite ! fresh_chain_append_to_set_large in *.
    rewrite ! fresh_chain_append_to_set in *.
    forwards: subset_union_2 H0 H2.
    forwards: subset_union_2 H5 H4.
    eapply subset_trans. eapply subset_trans. 2: eapply H6. union_sub_solver. union_sub_solver.
  - inversions His.
    forwards: fresh_chain_is_append_invr H1.
    forwards: IHfc2; fweauto.
    forwards: IHfc1; fweauto.
    rewrite ! fresh_chain_append_to_set_large in *.
    rewrite ! fresh_chain_append_to_set in *.
    forwards: subset_union_2 H0 H3.
    eapply subset_trans. eapply subset_trans. 2: eapply H4. union_sub_solver. union_sub_solver.
Qed.

Lemma fresh_chain_to_set_large_subset fc:
  fresh_chain_is ( fc) ->
  fresh_chain_to_set_large ( fc ) \c fresh_chain_to_set (fc).
Proof.
  introv His.
  forwards: fresh_chain_to_set_large_subset_app fc fresh_chain_nil; rewrite ? fresh_chain_append_nil_r in *; simpl; auto using subset_refl.
Qed.

Lemma fresh_element_to_set_large_eq fe:
  fresh_element_is fe ->
  fresh_element_to_set_large fe = fresh_element_to_set fe.
Proof.
  introv Hfe. eauto using fset_extens, fresh_element_to_set_large_subset, fresh_element_to_set_subset_large.
Qed.

Lemma fresh_chain_to_set_large_eq fe:
  fresh_chain_is fe ->
  fresh_chain_to_set_large fe = fresh_chain_to_set fe.
Proof.
  introv Hfe. eauto using fset_extens, fresh_chain_to_set_large_subset, fresh_chain_to_set_subset_large.
Qed.



(* DISJOINT SETS *)
Lemma fresh_element_disjoint_set fc1 fc2:
  fresh_element_fresh fc1 fc2 ->
  disjoint (fresh_element_to_set fc1) (fresh_element_to_set fc2).
Proof.
  introv Hfr.
  induction Hfr; simpl.
  rewrite <- fresh_element_parts_to_set.
  eapply disjoint_union_l.
  - eapply disjoint_comm. eapply fresh_disjoint. eauto.
  - auto.
Qed.

Lemma fresh_chain_disjoint_set fe fc:
  fresh_chain_fresh fe fc ->
  disjoint (fresh_element_to_set fe) (fresh_chain_to_set fc).
Proof.
  introv Hfr.
  induction Hfr; simpls.
  - (* nil *) apply disjoint_empty_r.
  - (* set *) apply disjoint_union_r. eapply fresh_element_disjoint_set; eauto. auto.
  - (* var *) apply disjoint_union_r. eapply fresh_element_disjoint_set; eauto. auto.
  - (* sub *) auto. (*apply disjoint_union_r. eapply fresh_element_disjoint_set; eauto. auto.*)
  - auto.
  - apply disjoint_union_r; auto.
  - rewrite fresh_chain_append_to_set in IHHfr1. 
    apply disjoint_union_r_inv in IHHfr1. destruct IHHfr1.
    apply disjoint_union_r; auto.
    apply disjoint_union_r; auto.
Qed.

Lemma fresh_element_is_fresh fe:
  fresh_element_is fe ->
  fresh \{} (length (fresh_element_part_list fe)) (fresh_element_part_list fe).
Proof.
  introv Hi.
  induction Hi; simpls; eauto using ok_fresh_keys, ctx_ok_fresh_keys.
  - destruct Go; unfold ctx_obj; simpls; auto.
  - destruct Go; unfold ctx_obj; simpls; auto.
Qed.

Lemma fresh_element_from_disjoint_set fc1 fc2:
  disjoint (fresh_element_to_set fc1) (fresh_element_to_set fc2) ->
  fresh_element_is fc1 ->
  fresh_element_is fc2 ->
  fresh_element_fresh fc1 fc2.
Proof.
  introv Hdi Hi1 Hi2.
  rewrite <- fresh_element_parts_to_set in Hdi.
  eapply disjoint_union_l_inv in Hdi. destruct Hdi.
  constructor.
  eapply disjoint_fresh.
  eapply fresh_element_is_fresh; eauto.
  eapply disjoint_comm. eauto.
  auto.
Qed.

Fixpoint fresh_chain_branching (fc: fresh_chain): nat :=
  match fc with
  | fresh_chain_nil => 1
  | fresh_chain_set fe fc => (fresh_chain_branching fc)
  | fresh_chain_var fe fc => (fresh_chain_branching fc)
  | fresh_chain_subset fe fc => (fresh_chain_branching fc)
  | fresh_chain_branch fcBr fcM => (S (fresh_chain_branching fcBr) ) * (fresh_chain_branching fcM)
  | fresh_chain_parallel fcBr1 fcBr2 fcM => ((fresh_chain_branching fcBr1) + (fresh_chain_branching fcBr2)) * (fresh_chain_branching fcM)
  | fresh_chain_union fc1 fc2 => ((fresh_chain_branching fc1) + (fresh_chain_branching fc2))
  end.
Fixpoint fresh_chain_size (fc: fresh_chain): nat :=
  match fc with
  | fresh_chain_nil => 0
  | fresh_chain_set fe fc => S (fresh_chain_size fc)
  | fresh_chain_var fe fc => S (fresh_chain_size fc)
  | fresh_chain_subset fe fc => S (fresh_chain_size fc)
  | fresh_chain_branch fcBr fcM => S ((fresh_chain_size fcBr) + (fresh_chain_size fcM) + (fresh_chain_branching fcBr) * (fresh_chain_size fcM))
  | fresh_chain_parallel fcBr1 fcBr2 fcM => S ((fresh_chain_size fcBr1) + (fresh_chain_size fcBr2) + (((fresh_chain_branching fcBr1) + (fresh_chain_branching fcBr2)) * (fresh_chain_size fcM)))
  | fresh_chain_union fc1 fc2 => S ((fresh_chain_size fc1) + (fresh_chain_size fc2))
  end.


Require Import Lia.

Lemma fresh_chain_branching_gt0 (fc: fresh_chain):
  0 < fresh_chain_branching fc.
Proof.
  induction fc; simpl; try lia.
  rewrite Nat.mul_add_distr_r.
  apply Nat.lt_lt_add_l.
  apply Nat.mul_pos_pos; auto.
Qed.

Lemma fresh_chain_append_size (fc1 fc2: fresh_chain):
  fresh_chain_size (fresh_chain_append fc1 fc2) = fresh_chain_size fc1 + (fresh_chain_branching fc1) * (fresh_chain_size fc2).
Proof.
  reverts fc2.
  induction fc1; intros; simpls.
  - eauto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - rewrite IHfc1_2.  lia.
  - rewrite IHfc1_3.  lia.
  - rewrite IHfc1_1. rewrite IHfc1_2.  lia.
Qed.


Lemma fresh_chain_out_append fc1 fc2:
  fresh_chain_out (fresh_chain_append fc1 fc2) =   fresh_chain_append (fresh_chain_out fc1) (fresh_chain_out fc2).
Proof.
  introv.
  induction fc1; intros; simpls; rewrite ? IHfc1, ?IHfc1_1, ?IHfc1_2, ?IHfc1_3; eauto.
Qed.

Lemma fresh_chain_out_to_set_large_sub fc1 :
  fresh_chain_to_set_large (fresh_chain_out fc1) \c fresh_chain_to_set_large (fc1).
Proof.
  induction fc1; intros; simpls; try union_sub_solver.
Qed.

Lemma fresh_chain_out_sub_to_set fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_to_set (fresh_chain_append fc1 fc2) \c fresh_chain_to_set (fresh_chain_append (fresh_chain_out fc1) fc2).
Proof.
  introv Hfc.
  reverts fc2.
  induction fc1; intros; simpls. 
  -  union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. forwards: subset_trans H3 H. union_sub_solver.
  -  inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_2; fweauto. auto.
  -  inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_1; fweauto. forwards: IHfc1_2; fweauto.  forwards: IHfc1_3; fweauto.
    rewrite ! fresh_chain_append_to_set in H0.
    rewrite ! fresh_chain_append_to_set in H2.
    rewrite ! fresh_chain_append_to_set in H4.
    rewrite ! fresh_chain_append_to_set.

Ltac union_sub_destruct_hyp := (
  match goal with
  | H: _ \u _ \c _ |- _ => forwards: subset_union_l2 H; forwards: subset_union_l1 H; clear H
  end
).

repeat union_sub_destruct_hyp.
union_sub_solver_destruct_ls;
[> eapply subset_trans;[>eauto|union_sub_solver]..].

-
 inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_1; fweauto. forwards: IHfc1_2; fweauto.
    rewrite ! fresh_chain_append_to_set in H0.
    rewrite ! fresh_chain_append_to_set in H3.
    rewrite ! fresh_chain_append_to_set.

repeat union_sub_destruct_hyp.
union_sub_solver_destruct_ls;
[> eapply subset_trans;[>eauto|union_sub_solver]..].

Qed.

Lemma fresh_chain_out_sub_to_set_large fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_to_set_large (fresh_chain_append fc1 fc2) \c fresh_chain_to_set_large (fresh_chain_append (fresh_chain_out fc1) fc2).
Proof.
  introv Hfc.
  reverts fc2.
  induction fc1; intros; simpls. 
  -  union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. union_sub_solver.
  -  inversions Hfc. forwards: IHfc1; fweauto. union_sub_solver.
  -  inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_2; fweauto. auto.
  -  inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_1; fweauto. forwards: IHfc1_2; fweauto.  forwards: IHfc1_3; fweauto.
    rewrite ! fresh_chain_append_to_set_large in H0.
    rewrite ! fresh_chain_append_to_set_large in H2.
    rewrite ! fresh_chain_append_to_set_large in H4.
    rewrite ! fresh_chain_append_to_set_large.


repeat union_sub_destruct_hyp.
union_sub_solver_destruct_ls;
[> eapply subset_trans;[>eauto|union_sub_solver]..].

-
 inversions Hfc.
     forwards: fresh_chain_is_append_invr; fweauto. forwards: IHfc1_1; fweauto. forwards: IHfc1_2; fweauto.
    rewrite ! fresh_chain_append_to_set_large in H0.
    rewrite ! fresh_chain_append_to_set_large in H3.
    rewrite ! fresh_chain_append_to_set_large.

repeat union_sub_destruct_hyp.
union_sub_solver_destruct_ls;
[> eapply subset_trans;[>eauto|union_sub_solver]..].

Qed.


Lemma fresh_chain_out_to_set_large fc:
  fresh_chain_is fc ->
  fresh_chain_to_set_large (fresh_chain_out fc) = fresh_chain_to_set_large fc.
Proof.
  introv Hfc.

  eapply fset_extens.
  - eapply fresh_chain_out_to_set_large_sub.
  - forwards: fresh_chain_out_sub_to_set_large fc fresh_chain_nil.
    rewrite fresh_chain_append_nil_r. auto.
    rewrite ! fresh_chain_append_nil_r in H. auto.
Qed.


Lemma fresh_chain_fresh_out fe fc1:
  fresh_chain_fresh fe fc1 ->
  fresh_chain_fresh fe (fresh_chain_out fc1).
Proof.
  induction fc1; intros; simpls; try inversions H; eauto.
  econstructor.
  eapply fresh_chain_fresh_append.
  eapply IHfc1_1. eauto.
  forwards: fresh_chain_fresh_append_invl H3; eauto.
  forwards: fresh_chain_fresh_append_invr H3; eauto.
  eauto.
Qed.

Lemma fresh_chain_out_fresh_app fe fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_fresh fe (fresh_chain_out fc1) ->
  fresh_chain_fresh fe fc1.
Proof.
  introv Hfc Hfr.
  reverts fc2.
  induction fc1; intros; simpls; try inversions H; try inversions Hfr; try inversions Hfc; eauto.
  econstructor; eauto. eapply IHfc1_2. eauto. eapply fresh_chain_is_append_invr; eauto.
  econstructor; eauto.

  eapply fresh_chain_fresh_append.
  eapply IHfc1_1.
  eapply fresh_chain_fresh_append_invl; eauto.
  eauto.
  eapply IHfc1_3.
  eapply fresh_chain_fresh_append_invr; eauto.
  eauto.
  eapply fresh_chain_is_append_invr; eauto.

Qed.

Lemma fresh_chain_out_fresh fe fc1 :
  fresh_chain_is fc1 ->
  fresh_chain_fresh fe (fresh_chain_out fc1) ->
  fresh_chain_fresh fe fc1.
Proof.
  introv Hfc Hfr.
  eapply fresh_chain_out_fresh_app with (fc2:=fresh_chain_nil); eauto.
  rewrite fresh_chain_append_nil_r; eauto.
Qed.

Lemma fresh_chain_fresh_subset_subset x fc X:
  X \c fresh_chain_to_set fc ->
  fresh_element_is x ->
  fresh_chain_fresh x fc ->
  fresh_chain_fresh x (fresh_chain_subset (fresh_element_set X) fc).
Proof.
  introv Hsu Hfe Hfc.
  econstructor.
  eapply fresh_element_from_disjoint_set; eauto.
  simpls.
  eapply disjoint_subset_l. 2: eapply Hsu.
  eapply fresh_chain_disjoint_set; eauto.
  eauto.
Qed.

Lemma fresh_chain_append_out_fresh fe0 fc1 fc2:
  fresh_chain_is fc1 ->
  fresh_chain_fresh fe0 (fresh_chain_append fc2 (fresh_chain_out fc1)) ->
  fresh_chain_fresh fe0 (fresh_chain_append fc2 fc1).
Proof.
  introv Hfc Hfr.  
  eapply fresh_chain_fresh_append.
  eapply fresh_chain_fresh_append_invl. eauto.
  eapply fresh_chain_out_fresh. eauto.
  eapply fresh_chain_fresh_append_invr; eauto.
Qed.

Lemma fresh_chain_append_branch_lt1 fc1 fc2 fc3:
  fresh_chain_size fc1 + fresh_chain_branching fc1 * fresh_chain_size fc3 <
  S
  (fresh_chain_size fc1 + fresh_chain_size fc2 +
   (fresh_chain_branching fc1 + fresh_chain_branching fc2) * fresh_chain_size fc3).
Proof.
    unfold lt. eapply le_n_S.
    rewrite <- ! Nat.add_assoc.
    eapply Plus.plus_le_compat_l.
    rewrite Nat.add_comm.
    eapply Plus.le_plus_trans.
    eapply Nat.mul_le_mono_r. lia.
Qed.
Lemma fresh_chain_append_branch_lt2 fc1 fc2 fc3:
  fresh_chain_size fc2 + fresh_chain_branching fc2 * fresh_chain_size fc3 <
  S
  (fresh_chain_size fc1 + fresh_chain_size fc2 +
   (fresh_chain_branching fc1 + fresh_chain_branching fc2) * fresh_chain_size fc3).
Proof.
 unfold lt. eapply le_n_S.
    rewrite <- ! Nat.add_assoc.
    rewrite Nat.add_comm with (n:=(fresh_chain_size fc1 )).
    eapply Plus.le_plus_trans.
    eapply Plus.plus_le_compat_l.
    eapply Nat.mul_le_mono_r. lia.
Qed.

Lemma fresh_chain_is_append_invl fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_is (fresh_chain_out fc1).
Proof.
  remember (fresh_chain_size fc1) as fS2.
  reverts fc1 fc2.
  induction fS2 using nat_lt_induction; intros; subst.
  destruct fc1; simpls.
  - eauto.
  - inversions H0. eauto.
  - inversions H0.
    forwards: H; fweauto.
    constructor;
    eauto using fresh_chain_fresh_append_invl, fresh_chain_fresh_out.
  - inversions H0. eauto.
  - inversions H0.
    rewrite <- fresh_chain_append_assoc in H2.
    econstructor.
    rewrite <- fresh_chain_out_append.
    eapply H. 2: eauto.
    rewrite fresh_chain_append_size. lia.
    eauto.
  - inversions H0.
    rewrite <- fresh_chain_append_assoc in H3.
    rewrite <- fresh_chain_append_assoc in H5.
    econstructor.
    +
    rewrite <- fresh_chain_out_append.
    eapply H. 2: eauto.
    rewrite fresh_chain_append_size.
    eapply fresh_chain_append_branch_lt1.
    eauto.
    +
    rewrite <- fresh_chain_out_append.
    eapply H. 2: eauto.
    rewrite fresh_chain_append_size.
    eapply fresh_chain_append_branch_lt2.
    eauto.
  - inversions H0.
    econstructor.
    +
    eapply H. 2: eauto. lia.
    eauto.
    +
    eapply H. 2: eauto. lia.
    eauto.
Qed.

Lemma fresh_chain_out_idem fc:
  fresh_chain_out (fresh_chain_out fc) = (fresh_chain_out fc).
Proof.
  induction fc; simpl; rewrite ? IHfc, ? IHfc1, ? IHfc2, ? IHfc3; auto.
Qed.

Lemma fresh_chain_is_out fc:
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_out fc).
Proof.

  remember (fresh_chain_size fc) as fS.
  reverts fc.
  induction fS using nat_lt_induction; intros; subst.
  destruct fc; simpls.
  - auto.
  - inversions H0. forwards: H; fweauto. eauto. 
  - inversions H0. forwards: H; fweauto. eauto using fresh_chain_fresh_out.
  - inversions H0. forwards: H; fweauto. eauto using fresh_chain_fresh_out.
  - inversions H0. constructor. forwards: H. 3: eauto. 2: eauto. rewrite fresh_chain_append_size. lia.
    rewrite fresh_chain_out_append in H0. auto.
  - inversions H0. constructor.
    forwards: H. 3: eapply H3. 2: eauto. rewrite fresh_chain_append_size. rewrite Nat.mul_add_distr_r. lia.
    rewrite fresh_chain_out_append in H0. auto.
    forwards: H. 3: eapply H5. 2: eauto. rewrite fresh_chain_append_size. rewrite Nat.mul_add_distr_r. lia.
    rewrite fresh_chain_out_append in H0. auto.
  - inversions H0. constructor.
    forwards: H. 3: eapply H3. 2: eauto.  lia.
    auto.
    forwards: H. 3: eapply H4. 2: eauto.  lia.
    auto.
Qed.

Lemma fresh_chain_out_to_set fc:
  fresh_chain_is fc ->
  fresh_chain_to_set (fresh_chain_out fc) = fresh_chain_to_set fc.
Proof.
  introv Hfc.
  rewrite <- ! fresh_chain_to_set_large_eq.
  eapply fresh_chain_out_to_set_large.
  eauto.
  eauto.
  eapply fresh_chain_is_out; eauto.
Qed.



Lemma fresh_element_from_disjoint_set_large fc1 fc2:
  disjoint (fresh_element_to_set fc1) (fresh_element_to_set_large fc2) ->
  fresh_element_is fc1 ->
  fresh_element_is fc2 ->
  fresh_element_fresh fc1 fc2.
Proof.
  introv Hdi Hi1 Hi2.
  forwards: fresh_element_from_disjoint_set.
  eapply disjoint_subset_l.
  2:eapply fresh_element_to_set_subset_large.
  all: eauto.
Qed.


Lemma fresh_chain_from_disjoint_set_out fe fc:
  disjoint (fresh_element_to_set fe) (fresh_chain_to_set_large fc) ->
  fresh_element_is fe ->
  fresh_chain_is (fresh_chain_out fc) ->
  fresh_chain_fresh fe fc.
Proof.
  introv Hdi Hi1 Hi2.

  remember (fresh_chain_size fc) as fS.
  reverts fc fe.
  induction fS using nat_lt_induction; intros; subst.
  destruct fc; simpls.
  - eauto.
  -
    eapply disjoint_union_r_inv in Hdi. destruct Hdi.
    inversions Hi2.
    econstructor.
    eapply fresh_element_from_disjoint_set_large; eauto.
    eauto.
  -
    eapply disjoint_union_r_inv in Hdi. destruct Hdi.
    inversions Hi2.
    econstructor.
    eapply fresh_element_from_disjoint_set_large; eauto.
    eauto.
  -
   eapply disjoint_union_r_inv in Hdi. destruct Hdi.
    inversions Hi2.
    econstructor.
    eapply fresh_element_from_disjoint_set_large; eauto.
    eauto.
  -
    inversions Hi2.
    econstructor.
    eapply H. 3: eauto. lia.
    eapply fresh_chain_is_append_invr; eauto. eauto. eauto.
  -
    inversions Hi2.
    eapply disjoint_union_r_inv in Hdi. destructecs Hdi.
    eapply disjoint_union_r_inv in Hdi. destructecs Hdi.
    econstructor.
    +
    eapply H. 3: eauto.
    rewrite fresh_chain_append_size.
    eapply fresh_chain_append_branch_lt1.
    rewrite fresh_chain_out_append.
    eauto.
    rewrite fresh_chain_append_to_set_large.
    eapply disjoint_union_r; eauto.
    eauto.
    +
    eapply H. 3: eauto.
    eapply Nat.le_lt_trans.
    2:
    eapply fresh_chain_append_branch_lt2.
    eapply Plus.le_plus_l.
    eapply fresh_chain_is_append_invl in H4.  rewrite fresh_chain_out_idem in H4. eauto.
    eauto.
    eauto.
  -
    inversions Hi2.
    eapply disjoint_union_r_inv in Hdi. destructecs Hdi.
    econstructor.
    eapply H. 3: eauto.
    unfold lt. eapply le_n_S.
    eapply Plus.le_plus_l.
    auto.
    auto.
    auto.
    eapply H. 3: eauto.
    unfold lt. eapply le_n_S.
    eapply Plus.le_plus_r.
    auto.
    auto.
    auto.
Qed.

Lemma fresh_chain_from_disjoint_set fe fc:
  disjoint (fresh_element_to_set fe) (fresh_chain_to_set fc) ->
  fresh_element_is fe ->
  fresh_chain_is fc ->
  fresh_chain_fresh fe fc.
Proof.
  intros.
  
  eapply fresh_chain_from_disjoint_set_out.
  rewrite fresh_chain_to_set_large_eq; eauto.
  eauto.
  eauto using fresh_chain_is_out.
Qed.

Lemma fresh_chain_fresh_subset_fresh fe fc1 fc2:
  fresh_chain_fresh fe fc1 ->
  fresh_chain_to_set fc2 \c fresh_chain_to_set fc1 ->
  fresh_element_is fe ->
  fresh_chain_is fc1 ->
  fresh_chain_is fc2 ->
  fresh_chain_fresh fe fc2.
Proof.
  introv Hfr Hsu Hie Hic1 Hic2.
  eapply fresh_chain_from_disjoint_set; auto 1.
  eapply disjoint_subset_l.
  eapply fresh_chain_disjoint_set; eauto.
  auto.
Qed.
Lemma fresh_chain_append_fresh_out fe0 fc1 fc2:
  fresh_chain_is fc1 ->
  fresh_chain_fresh fe0 (fresh_chain_append fc2 fc1) ->
  fresh_chain_fresh fe0 (fresh_chain_append fc2 (fresh_chain_out fc1)).
Proof.
  introv Hfc Hfr.  
  eapply fresh_chain_fresh_append.
  eapply fresh_chain_fresh_append_invl. eauto.
  eapply fresh_chain_out_fresh. eauto.
  eapply fresh_chain_is_out; eauto.
  eapply fresh_chain_fresh_out.
  eapply fresh_chain_fresh_out.
  eapply fresh_chain_fresh_append_invr; eauto.
Qed.

Lemma fresh_chain_append_out_is fc1 fc2:
  fresh_chain_is fc1 ->
  fresh_chain_is (fresh_chain_append fc2 (fresh_chain_out fc1)) ->
  fresh_chain_is (fresh_chain_append fc2 fc1).
Proof.
  introv Hfc1 Hfc2.
  remember ((fresh_chain_append fc2 (fresh_chain_out fc1))).
  reverts fc1 fc2.
  induction Hfc2; intros; subst; destruct fc2; simpls; inversions Heqf; auto 2.

  - econstructor; eauto. eapply fresh_chain_append_out_fresh; eauto.
  - econstructor; eauto.
  - econstructor; eauto. rewrite fresh_chain_append_to_set in *. rewrite fresh_chain_out_to_set in H0. eauto. eauto.
  - econstructor; eauto. rewrite <- fresh_chain_append_assoc in *. eapply IHHfc2; eauto.
  - econstructor; eauto.
  - econstructor; eauto. rewrite <- fresh_chain_append_assoc in *. eapply IHHfc2_1; eauto.
    rewrite <- fresh_chain_append_assoc in *. eapply IHHfc2_2; eauto.
Qed.


Lemma fresh_chain_append_is_out fc1 fc2:
  fresh_chain_is (fresh_chain_append fc2 fc1) ->
  fresh_chain_is (fresh_chain_append fc2 (fresh_chain_out fc1)).
Proof.

  introv Hfc1.
  remember ((fresh_chain_append fc2 fc1)).
  reverts fc1 fc2.
  induction Hfc1; intros; subst; destruct fc2; simpls; inversions Heqf; simpls; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out. eapply fresh_chain_append_fresh_out; eauto. eapply fresh_chain_is_append_invr; eauto.
  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
    rewrite fresh_chain_append_to_set in *. rewrite fresh_chain_out_to_set. eauto.  eapply fresh_chain_is_append_invr; eauto.
  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
    forwards: IHHfc1; fweauto.
    eapply fresh_chain_is_out in H0. rewrite fresh_chain_out_append in H0. rewrite fresh_chain_out_idem in H0. auto.
  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
    forwards: IHHfc1 fc1; fweauto. rewrite fresh_chain_append_assoc. eauto.
    rewrite fresh_chain_append_assoc in H. auto.

  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
    forwards: IHHfc1_1; fweauto.
    eapply fresh_chain_is_out in H0. rewrite fresh_chain_out_append in H0. rewrite fresh_chain_out_idem in H0. auto.
    forwards: IHHfc1_2; fweauto.
    eapply fresh_chain_is_out in H0. rewrite fresh_chain_out_append in H0. rewrite fresh_chain_out_idem in H0. auto.
  - econstructor; eauto using fresh_chain_is_out, fresh_chain_fresh_out.
    forwards: IHHfc1_1 fc1; fweauto.
    rewrite fresh_chain_append_assoc. eauto.
    rewrite fresh_chain_append_assoc in H. auto.
    forwards: IHHfc1_2 fc1; fweauto.
    rewrite fresh_chain_append_assoc. eauto.
    rewrite fresh_chain_append_assoc in H. auto.
Qed.


(* EQUIVALENCE RELATIONS *)

Lemma fresh_element_eq_trans flags a b c:
  fresh_element_eq flags a b ->
  fresh_element_eq flags b c ->
  fresh_element_eq flags a c.
Proof.
  intros.
  destruct flags.
  inversions H. simpls.
  inversions H0. simpls.
  constructor; simpl; cases_if; auto.
  eapply subset_trans; eauto.
  eapply subset_trans; eauto.
Qed.
Lemma fresh_element_eq_refl flags a:
  fresh_element_eq flags a a.
Proof.
  intros.
  destruct flags.
  constructor; simpl; cases_if; auto.
  apply subset_refl.
  apply subset_refl.
Qed.

Lemma fresh_element_eq_inv fl1 fl2 fl3 fl4 a b:
  fresh_element_eq ⚐ fl1 fl2 fl3 fl4 ⚐ a b ->
  fresh_element_eq ⚐ fl2 fl1 fl4 fl3 ⚐ b a.
Proof.
  intros.
  inversions H.
  constructor; auto.
Qed.

Add Parametric Relation flags: fresh_element (@fresh_element_eq flags)
  reflexivity proved by (@fresh_element_eq_refl flags)
  transitivity proved by (@fresh_element_eq_trans flags)
  as fresh_element_eq_rel.

Lemma fresh_chain_eq_trans flags a b c:
  fresh_chain_eq flags a b ->
  fresh_chain_eq flags b c ->
  fresh_chain_eq flags a c.
Proof.
  intros.
  destruct flags.
  inversions H. simpls.
  inversions H0. simpls.
  constructor; simpl; cases_if; auto.
  eapply subset_trans; eauto.
  eapply subset_trans; eauto.
Qed.
Lemma fresh_chain_eq_refl flags a:
  fresh_chain_eq flags a a.
Proof.
  intros.
  destruct flags.
  constructor; simpl; cases_if; auto.
  apply subset_refl.
  apply subset_refl.
Qed.

Lemma fresh_chain_eq_inv fl1 fl2 fl3 fl4 a b:
  fresh_chain_eq ⚐ fl1 fl2 fl3 fl4 ⚐ a b ->
  fresh_chain_eq ⚐ fl2 fl1 fl4 fl3 ⚐ b a.
Proof.
  intros.
  inversions H.
  constructor; simpls; cases_if; eauto using fresh_element_eq_refl.
Qed.

Add Parametric Relation flags: fresh_chain (@fresh_chain_eq flags)
  reflexivity proved by (@fresh_chain_eq_refl flags)
  transitivity proved by (@fresh_chain_eq_trans flags)
  as fresh_chain_eq_rel.

(* BASIC EQUIVALENCES FOR REWRITING *)

(* Enlarging a set has more elements and assertions *)
Lemma fresh_element_eq_set_subset fl1 fl2 fl3 X Y:
  Y \c X ->
  fresh_element_eq ⚐ fl1 fl2 false fl3 ⚐ (fresh_element_set X) (fresh_element_set Y).
Proof.
  introv Hsu.
  constructor; simpls; auto 1; cases_if; auto.
Qed.
(* Singleton set is the same as a var *)
Lemma fresh_element_eq_set_var flags x:
  fresh_element_eq flags (fresh_element_set \{x}) (fresh_element_var x).
Proof.
  constructor; simpls; auto 1; cases_if; auto; try apply subset_refl.
Qed.
(* Singleton list is the same as a var *)
Lemma fresh_element_eq_list_var flags x:
  fresh_element_eq flags (fresh_element_list (x::nil)) (fresh_element_var x).
Proof.
  constructor; simpls; auto 1; cases_if; auto; try (rewrite from_list_cons, from_list_nil, ?union_empty_l, ?union_empty_r; apply subset_refl).
Qed.

Ltac fresh_sets_destruct := (
  match goal with
  | H: fresh (?A \u ?B) ?L ?xs |- _ =>
    let Hr := fresh H "r" in
    let Hl := fresh H "l" in
      forwards [Hl Hr]: fresh_union_r H;
      clear H
  | H: fresh ?A ?L (?xs ++ ?ys) |- _ =>
    let Hr := fresh H "r" in
    let Hl := fresh H "l" in
      forwards Hl: fresh_append_l H;
      forwards Hr: fresh_append_r H;
      clear H
  | H: disjoint (?A \u ?B) ?C |- _ =>
    let Hr := fresh H "r" in
    let Hl := fresh H "l" in
      forwards [Hl Hr]: disjoint_union_l_inv H;
      clear H
  | H: disjoint ?C (?A \u ?B) |- _ =>
    let Hr := fresh H "r" in
    let Hl := fresh H "l" in
      forwards [Hl Hr]: disjoint_union_r_inv H;
      clear H
  end
).

Ltac fresh_sets_ctxs := (
  match goal with 
  | H: fresh ?A (length (ctx_keys ?G)) (ctx_keys ?G)
    |- fresh ?B (length (ctx_keys ?G)) (ctx_keys ?G) =>
      eapply (disjoint_fresh A (ctx_keys G) H); rewrite ctx_keys_dom
  | H: fresh ?A (length (keys ?G)) (keys ?G)
    |- fresh ?B (length (keys ?G)) (keys ?G) =>
      eapply (disjoint_fresh A (keys G) H); rewrite env_from_list_keys_dom
  | H: fresh ?A (length (ctx_keys ?G)) (ctx_keys ?G)
    |- disjoint (ctx_dom ?G) ?B =>
      rewrite <- ctx_keys_dom
  | H: fresh ?A (length (keys ?G)) (keys ?G)
    |- disjoint (dom ?G) ?B =>
      rewrite <- env_from_list_keys_dom
  end
).

Ltac fresh_sets_construct := repeat first [eapply disjoint_union_l | eapply disjoint_union_r | eapply fresh_union_l].

Ltac fresh_element_fresh_set_solver := let Hfr := fresh "Hfr" in (
    introv Hfr;
    inversions Hfr; constructor; simpls; auto;
    rewrite <- ! fresh_element_parts_to_set in *;
    repeat fresh_sets_destruct; fresh_sets_construct; repeat fresh_sets_ctxs;
    eauto using fresh_disjoint, disjoint_empty_l, disjoint_empty_r, disjoint_comm
).

(* Converting to a set may lose assertions, but preserves elements *)
Lemma fresh_element_eq_to_set fl1 fl2 fl3 fe:
  fresh_element_eq ⚐ fl1 false fl2 fl3 ⚐ fe (fresh_element_set (fresh_element_to_set fe)).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; solve[ apply subset_refl ].
Qed.

Lemma fresh_element_eq_list_app_l fl1 fl2 xs ys:
  fresh_element_eq ⚐ fl1 false false fl2 ⚐ (fresh_element_list (xs++ys)) (fresh_element_list xs).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - introv His.
    inversions His.
    eapply fresh_append_l in H0.
    eauto.
  - rewrite from_list_append. union_sub_solver.
Qed.

Lemma fresh_element_eq_list_app_r fl1 fl2 xs ys:
  fresh_element_eq ⚐ fl1 false false fl2 ⚐ (fresh_element_list (xs++ys)) (fresh_element_list ys).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - introv His.
    inversions His.
    eapply fresh_append_r in H0.
    eauto.
  - rewrite from_list_append. union_sub_solver.
Qed.

Lemma fresh_element_eq_list_app_swap flags xs ys:
  fresh_element_eq flags (fresh_element_list (xs++ys)) (fresh_element_list (ys++xs)).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - introv His.
    inversions His.
    eapply fresh_app in H0.
    eauto.
  - introv His.
    inversions His.
    eapply fresh_app in H0.
    eauto.
  - rewrite ! from_list_append. union_sub_solver.
  - rewrite ! from_list_append. union_sub_solver.
Qed.

Lemma fresh_element_eq_ctx flags G:
  ctx_ok G ->
  fresh_element_eq flags (fresh_element_set (ctx_fv G \u ctx_dom G)) (fresh_element_ctx G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.
Lemma fresh_element_eq_ctxdom flags G:
  ctx_ok G ->
  fresh_element_eq flags (fresh_element_set (ctx_dom G)) (fresh_element_ctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; union_sub_solver.
Qed.

Lemma fresh_element_eq_ctxdom_set fl1 fl2 fl3 G:
  fresh_element_eq ⚐ false fl1 fl2 fl3 ⚐ (fresh_element_set (ctx_dom G)) (fresh_element_ctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; union_sub_solver.
Qed.
Lemma fresh_element_eq_objctxdom_set fl1 fl2 fl3 G:
  fresh_element_eq ⚐ false fl1 fl2 fl3 ⚐ (fresh_element_set (objctx_dom G)) (fresh_element_objctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; union_sub_solver.
Qed.
Lemma fresh_element_eq_renvdom_set fl1 fl2 fl3 G:
  fresh_element_eq ⚐ false fl1 fl2 fl3 ⚐ (fresh_element_set (dom G)) (fresh_element_renvdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; union_sub_solver.
Qed.



Lemma fresh_element_eq_objctxdom flags G:
  ctx_ok (ctx_empty §§& G) ->
  fresh_element_eq flags (fresh_element_set (objctx_dom G)) (fresh_element_objctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; union_sub_solver.
Qed.
Lemma fresh_element_eq_ctxdom_list flags G:
  fresh_element_eq flags (fresh_element_list (ctx_keys G)) (fresh_element_ctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; rewrite ? ctx_keys_dom in *; try union_sub_solver.
  - 
    introv Hfr;
    inversions Hfr. constructor. eapply ctx_fresh_keys_ok; eauto.
  -
    introv Hfr;
    inversions Hfr. constructor. eapply ctx_ok_fresh_keys; eauto.
Qed.
Lemma fresh_element_eq_subst_ctxdom flags G x y:
  fresh_element_eq flags (fresh_element_ctxdom (subst_ctx x y G)) (fresh_element_ctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; intros.
  inversions H. constructor. eapply ctx_ok_map_inv; eauto.
  inversions H. constructor. eapply ctx_ok_map; eauto.
  ctx_simpl_dom. apply subset_refl.
  ctx_simpl_dom. apply subset_refl.
Qed.
Lemma fresh_element_eq_subst_objctxdom flags G x y:
  fresh_element_eq flags (fresh_element_objctxdom (subst_objctx x y G)) (fresh_element_objctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; intros.
  inversions H. constructor. destruct G; sympls; unfold ctx_obj in *; auto.
  inversions H. constructor. destruct G; sympls; unfold ctx_obj in *; auto.
  destruct G; simpl; apply subset_refl.
  destruct G; simpl; apply subset_refl.
Qed.


Lemma ctx_obj_keys (G1: ctx) (G2: objctx):
  ctx_keys (G1 §§& G2) = ((objctx_keys G2) ++ (ctx_keys G1))%list.
Proof.
  destruct G2; unfold ctx_obj; simpls; rewrite ? List.app_nil_r; auto 1.
Qed.
Lemma objctx_fresh_keys_ok G:
  fresh \{} (length (objctx_keys G)) (objctx_keys G) ->
  ctx_ok (ctx_empty §§& G).
Proof.
  introv Hfr.
  eapply ctx_fresh_keys_ok. rewrite ctx_obj_keys. simpl. rewrite ! List.app_nil_r. auto.
Qed.
Lemma objctx_ok_fresh_keys G:
  ctx_ok (ctx_empty §§& G) ->
  fresh \{} (length (objctx_keys G)) (objctx_keys G).
Proof.
  introv Hol.
  eapply ctx_ok_fresh_keys in Hol. rewrite ctx_obj_keys in Hol. simpl in Hol. rewrite ! List.app_nil_r in Hol. auto.
Qed.
Lemma fresh_element_eq_objctxdom_list flags G:
  fresh_element_eq flags (fresh_element_list (objctx_keys G)) (fresh_element_objctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; rewrite ? objctx_keys_dom in *; try union_sub_solver.
  - 
    introv Hfr;
    inversions Hfr. constructor. eapply objctx_fresh_keys_ok; eauto.
  -
    introv Hfr;
    inversions Hfr. constructor. eapply objctx_ok_fresh_keys; eauto.
Qed.
Lemma fresh_element_eq_ctxdom_map flags f G:
  fresh_element_eq flags (fresh_element_ctxdom G) (fresh_element_ctxdom (ctx_map f G)).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; intros.
  - inversions H. constructor.
    eapply ctx_ok_map. auto.
  - inversions H. constructor.
    eapply ctx_ok_map_inv; eauto.
  - rewrite ctx_dom_map.
    union_sub_solver.
  - rewrite ctx_dom_map.
    union_sub_solver.
Qed.
Lemma fresh_element_eq_ctx_ctxdom fl1 fl2 fl3 G:
  fresh_element_eq ⚐ fl1 fl2 false fl3 ⚐ (fresh_element_ctx G) (fresh_element_ctxdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; rewrite ? ctx_keys_dom in *; try union_sub_solver.
  - 
    introv Hfr;
    inversions Hfr. constructor. auto.
  - introv Hfr;
    inversions Hfr. constructor. auto.
Qed.

Lemma fresh_element_eq_renv flags (e: renv):
  ok e ->
  fresh_element_eq flags (fresh_element_set (fv_env e \u dom e)) (fresh_element_renv e).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.
Lemma fresh_element_eq_renvdom flags (e: renv):
  ok e ->
  fresh_element_eq flags (fresh_element_set (dom e)) (fresh_element_renvdom e).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.
Lemma fresh_element_eq_renvdom_single flags (x y: var):
  fresh_element_eq flags (fresh_element_var x) (fresh_element_renvdom (x~y)).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - intro. constructor. rewrite <- concat_empty_l. eauto.
  - rewrite dom_single.
   union_sub_solver.
  - rewrite dom_single.
   union_sub_solver.
Qed.

Lemma fresh_element_eq_renvdom_list flags G:
  fresh_element_eq flags (fresh_element_list (keys G)) (fresh_element_renvdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; rewrite ? env_from_list_keys_dom in *; try union_sub_solver.
  - 
    introv Hfr;
    inversions Hfr. constructor. eapply fresh_keys_ok; eauto.
  -
    introv Hfr;
    inversions Hfr. constructor. eapply ok_fresh_keys; eauto.
Qed.

Lemma fresh_element_eq_heap flags (Si: heap):
  ok Si ->
  fresh_element_eq flags (fresh_element_set (fv_env Si \u dom Si)) (fresh_element_heap Si).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.
Lemma fresh_element_eq_heapdom flags (Si: heap):
  ok Si ->
  fresh_element_eq flags (fresh_element_set (dom Si)) (fresh_element_heapdom Si).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.
Lemma fresh_element_eq_heapdom_list flags G:
  fresh_element_eq flags (fresh_element_list (keys G)) (fresh_element_heapdom G).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; simpls; rewrite ? env_from_list_keys_dom in *; try union_sub_solver.
  - 
    introv Hfr;
    inversions Hfr. constructor. eapply fresh_keys_ok; eauto.
  -
    introv Hfr;
    inversions Hfr. constructor. eapply ok_fresh_keys; eauto.
Qed.
Lemma fresh_element_eq_heapdom_single flags (x: var) y:
  fresh_element_eq flags (fresh_element_var x) (fresh_element_heapdom (x~y)).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - intro. constructor. rewrite <- concat_empty_l. eauto.
  - rewrite dom_single.
   union_sub_solver.
  - rewrite dom_single.
   union_sub_solver.
Qed.


Lemma fresh_element_eq_element flags A {AS: FreeVar A} (a: A):
  fresh_element_eq flags (fresh_element_set (fv a)) (fresh_element_element a).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto; union_sub_solver.
Qed.

Lemma fresh_element_eq_list_cons_l fl1 fl2 x ys:
  fresh_element_eq ⚐ fl1 false false fl2 ⚐ (fresh_element_list (x :: ys)) (fresh_element_var x).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - rewrite from_list_cons; union_sub_solver.
Qed.

Lemma fresh_element_eq_list_cons_r fl1 fl2 x ys:
  fresh_element_eq ⚐ fl1 false false fl2 ⚐ (fresh_element_list (x :: ys)) (fresh_element_list ys).
Proof.
  constructor; simpls; auto 1; cases_if; subst; auto.
  - introv Hfr. inversions Hfr. inversions H0. constructor; auto.
  - rewrite from_list_cons; union_sub_solver.
Qed.

Lemma fresh_element_eq_containing fl2 fl3 fl4 fl5 fl6 fl7 feM feS:
  fresh_element_eq ⚐ fl4 fl2 fl3 fl4 ⚐ feM feS ->
  fresh_element_eq ⚐ fl4 fl5 fl6 fl7 ⚐ feM (fresh_element_containing feM feS).
Proof.
  introv Heq.
  inversions Heq.
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - constructor; auto.
  - inversions H3; auto.
  - union_sub_solver.
  - union_sub_solver.
Qed.

Lemma fresh_element_eq_containing_refl flags fe:
  fresh_element_eq flags fe (fresh_element_containing fe fe).
Proof.
  destruct flags.
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - constructor; auto. union_sub_solver.
  - inversions H; auto.
  - union_sub_solver.
  - union_sub_solver.
Qed.


Lemma fresh_element_eq_containing_sub fl1 feM feS:
  fresh_element_eq ⚐ fl1 false false false ⚐ (fresh_element_containing feM feS) feS.
Proof.
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H; auto.
  (* Note the subset relation between feM and feS cannot be used in equivalence, because it depeneds on validity *)
Qed.

Lemma fresh_element_eq_containing_renv_push_dom flags e x y:
  fresh_element_eq flags (fresh_element_renv (e & x ~ y)) (fresh_element_containing (fresh_element_renv (e & x ~ y)) (fresh_element_var x)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  rewrite dom_concat, dom_single; union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_renv_push_fv flags e x y:
  fresh_element_eq flags (fresh_element_renv (e & x ~ y)) (fresh_element_containing (fresh_element_renv (e & x ~ y)) (fresh_element_var y)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  rewrite fv_env_concat, fv_env_single. unfold fv, VarFreeVar. union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_renv_push_env flags e f:
  fresh_element_eq flags (fresh_element_renv (e & f)) (fresh_element_containing (fresh_element_renv (e & f)) (fresh_element_renv e)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ok_concat_inv_l in H1. constructor. auto.
  - rewrite fv_env_concat, dom_concat. union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_heap_push_dom flags e x y:
  fresh_element_eq flags (fresh_element_heap (e & x ~ y)) (fresh_element_containing (fresh_element_heap (e & x ~ y)) (fresh_element_var x)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  rewrite dom_concat, dom_single; union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_heap_push_fv flags e x y:
  fresh_element_eq flags (fresh_element_heap (e & x ~ y)) (fresh_element_containing (fresh_element_heap (e & x ~ y)) (fresh_element_element y)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  rewrite fv_env_concat, fv_env_single. unfold fv, VarFreeVar. union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_heap_push_env flags e f:
  fresh_element_eq flags (fresh_element_heap (e & f)) (fresh_element_containing (fresh_element_heap (e & f)) (fresh_element_heap e)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ok_concat_inv_l in H1. constructor. auto.
  - rewrite fv_env_concat, dom_concat. union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctx_push_dom flags G x vk T:
  fresh_element_eq flags (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_containing (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_var x)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctx_push_fv flags G x vk T:
  fresh_element_eq flags (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_containing (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_element T)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctx_push_ctx flags  G x vk T:
  fresh_element_eq flags (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_containing (fresh_element_ctx (G § x / vk ~ T)) (fresh_element_ctx G)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. inversions H1. auto.
  -
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctx_concat_l flags G1 G2:
  fresh_element_eq flags (fresh_element_ctx (G1 §& G2)) (fresh_element_containing (fresh_element_ctx (G1 §& G2)) (fresh_element_ctx G1)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ctx_ok_concat_inv in H1. destructecs H1. constructor. auto.
  - rewrite ctx_dom_concat. rewrite ctx_fv_concat.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctx_concat_r flags G1 G2:
  fresh_element_eq flags (fresh_element_ctx (G1 §& G2)) (fresh_element_containing (fresh_element_ctx (G1 §& G2)) (fresh_element_ctx G2)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ctx_ok_concat_inv in H1. destructecs H1. constructor. auto.
  - rewrite ctx_dom_concat. rewrite ctx_fv_concat.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_concat_l flags G1 G2:
  fresh_element_eq flags (fresh_element_ctxdom (G1 §& G2)) (fresh_element_containing (fresh_element_ctxdom (G1 §& G2)) (fresh_element_ctxdom G1)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ctx_ok_concat_inv in H1. destructecs H1. constructor. auto.
  - rewrite ctx_dom_concat.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_concat_r flags G1 G2:
  fresh_element_eq flags (fresh_element_ctxdom (G1 §& G2)) (fresh_element_containing (fresh_element_ctxdom (G1 §& G2)) (fresh_element_ctxdom G2)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ctx_ok_concat_inv in H1. destructecs H1. constructor. auto.
  - rewrite ctx_dom_concat.
    union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_push_dom flags G x vk T:
  fresh_element_eq flags (fresh_element_ctxdom (G § x / vk ~ T)) (fresh_element_containing (fresh_element_ctxdom (G § x / vk ~ T)) (fresh_element_var x)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_push_ctx flags  G x vk T:
  fresh_element_eq flags (fresh_element_ctxdom (G § x / vk ~ T)) (fresh_element_containing (fresh_element_ctxdom (G § x / vk ~ T)) (fresh_element_ctxdom G)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. inversions H1. auto.
  -
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_obj_l flags G1 G2:
  fresh_element_eq flags (fresh_element_ctxdom (G1 §§& G2)) (fresh_element_containing (fresh_element_ctxdom (G1 §§& G2)) (fresh_element_ctxdom G1)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. eapply ctx_ok_obj_inv_l in H1. constructor. auto.
  - rewrite ctx_obj_dom.
  union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_ctxdom_obj_r flags G1 G2:
  fresh_element_eq flags (fresh_element_ctxdom (G1 §§& G2)) (fresh_element_containing (fresh_element_ctxdom (G1 §§& G2)) (fresh_element_objctxdom G2)).
Proof.
  destruct flags.
  eapply fresh_element_eq_containing with (fl2:=false) (fl3:=false).
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1.
  - inversions H. constructor. rewrite <- ctx_concat_empty_r with (G1:=G1) in H1. rewrite ctx_obj_assoc in H1. 
    eapply ctx_ok_concat_inv in H1. destructecs H1.
    auto.
  - rewrite ctx_obj_dom.
    union_sub_solver.
Qed.
Lemma fresh_element_eq_containing_var flags feM x:
  x \in (fresh_element_to_set feM) ->
  fresh_element_eq flags feM (fresh_element_containing feM (fresh_element_var x)).
Proof.
  introv Hin.
  destruct flags.
  eapply fresh_element_eq_containing. Unshelve. all: try apply false.
  constructor; simpls; auto 1; cases_if; subst; intros; auto 1. 
  rewrite subset_singleton. auto.
Qed.

(* HELPER PROPERTIES FOR REWRITING *)

Lemma fresh_element_fresh_symm fc1 fc2:
  fresh_element_fresh fc1 fc2 ->
  fresh_element_is fc1 ->
  fresh_element_is fc2 ->
  fresh_element_fresh fc2 fc1.
Proof.
  introv Hfr Hi1 Hi2.
  induction Hfr.
  rewrite <- fresh_element_parts_to_set in H.
  rewrite <- fresh_element_parts_to_set in H0.
  eapply fresh_union_r in H. destructecs H.
  eapply disjoint_union_r_inv in H0. destructecs H0.
  econstructor.
  -
  eapply disjoint_fresh.
  eapply fresh_element_is_fresh; eauto.
  rewrite <- fresh_element_parts_to_set.
  eapply disjoint_union_l.
  +
  eapply disjoint_comm.
  eapply fresh_disjoint; eauto.
  +
  eauto.
  -
  rewrite <- fresh_element_parts_to_set.
  eapply disjoint_union_r.
  eapply fresh_disjoint; eauto.
  eapply disjoint_comm.
  eauto.
Qed.

Lemma fresh_element_element_eq_fresh fl1 fl2 fl3 fc1 fc2 fc:
  fresh_element_is fc ->
  fresh_element_is fc1 ->
  fresh_element_is fc2 ->
  fresh_element_fresh fc1 fc ->
  fresh_element_eq ⚐ fl1 fl2 fl3 true ⚐ fc1 fc2 ->
  fresh_element_fresh fc2 fc.
Proof.
  introv His His1 His2 Hfr Hfe.
  inversions Hfe; simpls.
  forwards: fresh_element_disjoint_set Hfr.
  eapply fresh_element_from_disjoint_set; auto 1.
  eapply disjoint_subset_r; eauto.
Qed.
Lemma fresh_chain_element_eq_fresh fl1 fl2 fl3 fc1 fc2 fc:
  fresh_chain_is fc ->
  fresh_element_is fc1 ->
  fresh_element_is fc2 ->
  fresh_chain_fresh fc1 fc ->
  fresh_element_eq ⚐ fl1 fl2 fl3 true ⚐ fc1 fc2 ->
  fresh_chain_fresh fc2 fc.
Proof.
  introv His His1 His2 Hfr Hfe.
  inversions Hfe; simpls.
  forwards: fresh_chain_disjoint_set Hfr.
  eapply fresh_chain_from_disjoint_set; auto 1.
  eapply disjoint_subset_r; eauto.
Qed.

(* MORPHISMS FOR REWRITING *)
Require Export Setoid.
Require Export Relation_Definitions.

(* Implications for assertions *)
Add Parametric Morphism fl1 fl2 fl3: fresh_element_is
  with signature fresh_element_eq ⚐ true fl1 fl2 fl3 ⚐ ==> impl
  as fresh_element_eq_is_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

Add Parametric Morphism fl1 fl2 fl3: fresh_element_to_set
  with signature fresh_element_eq ⚐ fl1 fl2 true fl3 ⚐ ==> (@subset var)
  as fresh_element_subset_set_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

Add Parametric Morphism fl1 fl2 fl3: fresh_chain_is
  with signature fresh_chain_eq ⚐ true fl1 fl2 fl3 ⚐ ==> impl
  as fresh_chain_eq_is_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

Add Parametric Morphism fl1 fl2 fl3: fresh_chain_to_set
  with signature fresh_chain_eq ⚐ fl1 fl2 true fl3 ⚐ ==> (@subset var)
  as fresh_chain_subset_set_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

Add Parametric Morphism fl1 fl2 fl3: fresh_element_is
  with signature fresh_element_eq ⚐ fl1 true fl2 fl3 ⚐ ==> flip impl
  as fresh_element_eq_is_rev_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

Add Parametric Morphism fl1 fl2 fl3: fresh_chain_is
  with signature fresh_chain_eq ⚐ fl1 true fl2 fl3 ⚐ ==> flip impl
  as fresh_chain_eq_is_rev_mor.
Proof. introv Hfce. inversions Hfce. simpls. auto. Qed.

(* BASIC EQUIVALENCES FOR REWRITING *)

(* Warning: this rewriting is less permissive than it might seem to be possible,
  because the fresh_element_containing guarantees that one part is subset of the other.
  However, this guarantee is not available when reasoning about subsets - these two concerns are separated
*)
Add Parametric Morphism fl1 fl2 fl3 fl4: fresh_element_containing
  with signature fresh_element_eq ⚐ fl1 fl2 (orb fl1 fl3) (orb fl2 fl4) ⚐ ==>  fresh_element_eq ⚐ fl1 fl2 (orb fl2 fl3) (orb fl1 fl4) ⚐ ==> fresh_element_eq ⚐ fl1 fl2 fl3 fl4 ⚐
  as fresh_element_eq_containing_mor.
Proof.
  introv Hfce.
  inversions Hfce; simpl; econstructor; cases_if;   inversions H3;  simpls;subst;rewrite ? Bool.orb_true_r in *; simpls; auto 1.
  - introv His.
    inversions His.
    constructor; eauto using subset_trans.
  - introv His.
    inversions His.
    constructor; eauto using subset_trans.
Qed.

(* Union with nil chain does not change anything *)
Lemma fresh_chain_eq_union_nil_l fcA flags:
  fresh_chain_eq flags fcA (fresh_chain_union fresh_chain_nil fcA).
Proof.  
  constructor; cases_if; simpl; auto.
  - introv Hfc.
    inversions Hfc; auto.
  - union_sub_solver.
  - union_sub_solver.
Qed.

(* Union with some chain has more elements and assertions *)
Lemma fresh_chain_eq_union_l fl1 fl2 fcA fcB:
  fresh_chain_eq ⚐ fl1 false false fl2 ⚐ (fresh_chain_union fcA fcB) fcA.
Proof.
  constructor; cases_if; auto.
  - introv Hfc. inversions Hfc; auto.
  - simpl.
    eapply subset_union_weak_l; eauto.
Qed.

(* The "var" elements can be swapped with each other - they are fresh with respect to each other *)
Lemma fresh_chain_eq_swap_var_var flags fe1 fe2 fc:
  fresh_chain_eq flags (fresh_chain_var fe1 (fresh_chain_var fe2 fc)) (fresh_chain_var fe2 (fresh_chain_var fe1 fc)).
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    inversions H3.
    constructor; auto.
    constructor; auto.
    eapply fresh_element_fresh_symm; eauto.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    inversions H3.
    constructor; auto.
    constructor; auto.
    eapply fresh_element_fresh_symm; eauto.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
Qed.

(* Union of sets can be split, adjacent sets can be merged *)
Lemma fresh_chain_eq_merge_set_set flags fc X Y:
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set X) (fresh_chain_set (fresh_element_set Y) fc)) (fresh_chain_set (fresh_element_set (X \u Y)) fc).
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv His. inversions His. inversions H1. eauto.
  - introv His. inversions His. inversions H2. eauto.
  - rewrite union_assoc. apply subset_refl.
  - rewrite union_assoc. apply subset_refl.
Qed.

Add Parametric Morphism flags: fresh_chain_set
  with signature fresh_element_eq flags ==> fresh_chain_eq flags ==> fresh_chain_eq flags
  as fresh_chain_impl_set.
Proof.
  introv Hfce.
  destruct flags.
  inversions Hfce; simpl; econstructor; cases_if;simpls;subst; simpls; auto.
  - introv Hfc.
    inversions Hfc.
    inversions H3; simpls.
    econstructor; auto.  
  - introv Hfc.
    inversions Hfc.
    inversions H3; simpls.
    econstructor; auto.
  - apply subset_union_2. auto.    inversions H3; simpls. auto.
  - apply subset_union_2. auto.    inversions H3; simpls. auto.
Qed.

Lemma fresh_chain_eq_merge_set_ctx flags fc G:
  ctx_ok G ->
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set (ctx_fv G)) (fresh_chain_set (fresh_element_set (ctx_dom G)) fc)) (fresh_chain_set (fresh_element_ctx G) fc).
Proof.
  introv Hok.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_merge_set_set.
  rewrite fresh_element_eq_ctx.
  eapply fresh_chain_eq_refl.
  auto.
Qed.
Lemma fresh_chain_eq_merge_set_renv flags fc e:
  ok e ->
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set (fv_env e)) (fresh_chain_set (fresh_element_set (dom e)) fc)) (fresh_chain_set (fresh_element_renv e) fc).
Proof.
  introv Hok.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_merge_set_set.
  rewrite fresh_element_eq_renv.
  eapply fresh_chain_eq_refl.
  auto.
Qed.
Lemma fresh_chain_eq_merge_set_heap flags fc e:
  ok e ->
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set (fv_env e)) (fresh_chain_set (fresh_element_set (dom e)) fc)) (fresh_chain_set (fresh_element_heap e) fc).
Proof.
  introv Hok.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_merge_set_set.
  rewrite fresh_element_eq_heap.
  eapply fresh_chain_eq_refl.
  auto.
Qed.


Ltac fresh_chain_eq_solve_flags := try apply true.

(* Test for rewriting *)
Goal forall(A B C: fset var),
  fresh_chain_is (fresh_chain_set (fresh_element_set A) (fresh_chain_set (fresh_element_set B) (fresh_chain_set (fresh_element_set C) fresh_chain_nil))) ->
  fresh_chain_is (fresh_chain_set (fresh_element_set A) (fresh_chain_set (fresh_element_set (B \u C)) fresh_chain_nil)).
Proof.
  introv Hfc.
  rewrite fresh_chain_eq_merge_set_set with (X:=B) in Hfc; fresh_chain_eq_solve_flags. auto.
Qed.

Add Parametric Morphism  fl1 fl2 fl3 fl4: fresh_chain_var
  with signature fresh_element_eq ⚐ fl1 fl2 (orb fl2 fl3) (orb fl1 fl4) ⚐ ==>  fresh_chain_eq ⚐ fl1 fl2 (orb fl2 fl3) (orb fl1 fl4) ⚐ ==> fresh_chain_eq ⚐ fl1 fl2 fl3 fl4 ⚐
  as fresh_chain_impl_var.
Proof.
  introv Hfce.
  inversions Hfce; simpl; econstructor; cases_if;simpls;subst; simpls; auto 1.
  - introv His.
    inversions His.
    inversions H3; simpls.
    econstructor; auto.
    forwards: fresh_chain_disjoint_set H8.
    eapply fresh_chain_from_disjoint_set; eauto.
    eapply disjoint_subset_l; eauto.
    eapply disjoint_subset_r; eauto.
  - introv His.
    inversions His.
    inversions H3; simpls.
    econstructor; auto.
    forwards: fresh_chain_disjoint_set H8.
    eapply fresh_chain_from_disjoint_set; eauto.
    eapply disjoint_subset_l; eauto.
    eapply disjoint_subset_r; eauto.
  - inversions H3; simpls.  rewrite Bool.orb_true_r in H6, H1; simpl in H6.
    apply subset_union_2; auto.
  - inversions H3; simpls.  rewrite Bool.orb_true_r in H7, H2; simpl in H7.
    apply subset_union_2; auto.
Qed.


(* Test for rewriting *)
Goal forall(A B C: fset var),
  fresh_chain_is (fresh_chain_var (fresh_element_set A) (fresh_chain_set (fresh_element_set B) (fresh_chain_set (fresh_element_set C) fresh_chain_nil))) ->
  fresh_chain_is (fresh_chain_var (fresh_element_set A) (fresh_chain_set (fresh_element_set (B \u C)) fresh_chain_nil)).
Proof.
  introv Hfc.
  rewrite fresh_chain_eq_merge_set_set with (X:=B) in Hfc; fresh_chain_eq_solve_flags. auto.
Qed.

Add Parametric Morphism flags: (fresh_chain_union)
  with signature fresh_chain_eq flags ==> fresh_chain_eq flags ==> fresh_chain_eq flags
  as fresh_chain_impl_union.
Proof.
  introv Hfce.
  inversions Hfce; destruct flags; simpl; econstructor; cases_if;simpls;subst; simpls; auto.
  - introv Hfc.
    inversions Hfc.
    destruct H3; simpls.
    econstructor; auto.
  - introv Hfc.
    inversions Hfc.
    destruct H3; simpls.
    econstructor; auto.
  - destruct H3; simpls.
    apply subset_union_2; auto.
  - destruct H3; simpls.
    apply subset_union_2; auto.
Qed.


(* The "set" elements can be swapped with each other - they are not fresh with respect to each other *)
Lemma fresh_chain_eq_swap_set_set flags fe1 fe2 fc:
  fresh_chain_eq flags (fresh_chain_set fe1 (fresh_chain_set fe2 fc)) (fresh_chain_set fe2 (fresh_chain_set fe1 fc)).
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    constructor; auto.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    constructor; auto.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
Qed.

Lemma fresh_chain_eq_swap_set_var fl1 fl3 fl4 fe1 fe2 fc:
  fresh_chain_eq ⚐ fl1 false fl3 fl4 ⚐ (fresh_chain_var fe2 (fresh_chain_set fe1 fc)) (fresh_chain_set fe1 (fresh_chain_var fe2 fc)).
Proof.
 constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    inversions H3.
    eauto.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
  - rewrite ! union_assoc. rewrite union_comm with (E:=_ fe1). apply subset_refl.
Qed.

Lemma fresh_chain_eq_merge_list_list flags xs ys fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_list xs) (fresh_chain_var (fresh_element_list ys) fc)) (fresh_chain_var (fresh_element_list (xs++ys)) fc).
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    inversions H2.
    inversions H5.
    inversions H3.
    inversions H7.
    simpls.
    assert (fresh \{} (length (xs ++ ys)) (xs ++ ys)) as Hfrxy. {
    eapply fresh_append. eauto.
    rewrite union_empty_l.
    eapply fresh_swap; eauto.
    }
    constructor; auto.
    eapply fresh_chain_from_disjoint_set.
    *
    forwards: fresh_chain_disjoint_set H6.
    forwards: fresh_chain_disjoint_set H8.
    simpls. rewrite from_list_append.
    eapply disjoint_union_l; eauto.
    * eauto.
    * eauto.
  - introv Hfci.
    inversions Hfci.

    forwards H3r: fresh_chain_element_eq_fresh H3 fresh_element_eq_list_app_r; eauto. rewrite fresh_element_eq_list_app_r in H2; auto. apply true.
    forwards H3l: fresh_chain_element_eq_fresh H3 fresh_element_eq_list_app_l; eauto. rewrite fresh_element_eq_list_app_l in H2; auto. apply true.

    inversions H2.
    forwards: fresh_append_l H0.
    forwards: fresh_append_r H0.
    forwards: fresh_append_fresh_l H0.
    constructor; auto.
    constructor.
    constructor; simpls. auto. eapply fresh_disjoint; eauto. eauto.
  - rewrite ! union_assoc. rewrite from_list_append. apply subset_refl.
  - rewrite ! union_assoc. rewrite from_list_append. apply subset_refl.
  Unshelve. all: fresh_chain_eq_solve_flags.
Qed.

Add Parametric Morphism fl1 fl2 fl3 fl4: (fresh_chain_eq ⚐ fl1 fl2 fl3 fl4 ⚐)
  with signature fresh_chain_eq ⚐ fl2 fl1 fl4 fl3 ⚐ ==> fresh_chain_eq ⚐ fl1 fl2 fl3 fl4 ⚐ ==> impl
  as fresh_chain_eq_eq_mor.
Proof. introv Hfce1 Hfce2. intro. eapply fresh_chain_eq_trans. eapply fresh_chain_eq_inv; eauto.  eapply fresh_chain_eq_trans; eauto.  Qed.

Lemma fresh_chain_eq_merge_var_ctxdom flags fc G1 G2:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (G1 §& G2)) fc) (fresh_chain_var (fresh_element_ctxdom G2) (fresh_chain_var (fresh_element_ctxdom G1) fc)) .
Proof.
  destruct flags.
  rewrite <- ! fresh_element_eq_ctxdom_list.
  rewrite ctx_keys_concat.
  rewrite fresh_element_eq_list_app_swap.
  eapply fresh_chain_eq_inv.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_swap_var_var.
  eapply fresh_chain_eq_merge_list_list.
Qed.
Lemma fresh_chain_eq_merge_var_ctxdom_empty flags fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (ctx_empty)) fc) fc .
Proof.
  constructor; cases_if; simpls; auto 1; intros.
  inversions H; auto.
  econstructor; eauto.
  eapply fresh_chain_from_disjoint_set. sympls. eapply disjoint_empty_l.
  econstructor; eauto.
  auto.
  union_sub_solver.
  union_sub_solver.
Qed.
Lemma fresh_chain_eq_merge_set_empty flags fc:
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set (\{})) fc) fc .
Proof.
  constructor; cases_if; simpls; auto 1; intros.
  inversions H; auto.
  econstructor; eauto.

  union_sub_solver.
  union_sub_solver.
Qed.




Lemma fresh_chain_eq_merge_var_ctxdompush flags fc G x vk T:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (G § x / vk ~ T)) fc) (fresh_chain_var (fresh_element_var x) (fresh_chain_var (fresh_element_ctxdom G) fc)) .
Proof.
  destruct flags.
  rewrite <- ! fresh_element_eq_ctxdom_list.
  simpl. rewrite <- List.app_nil_l with (l := (ctx_keys G)) at 1.
  rewrite <- fresh_element_eq_list_var.

  eapply fresh_chain_eq_inv.
  eapply fresh_chain_eq_merge_list_list.
Qed.
Lemma fresh_chain_eq_merge_var_ctxdom_rev flags fc G1 G2:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (G1 §& G2)) fc) (fresh_chain_var (fresh_element_ctxdom G1) (fresh_chain_var (fresh_element_ctxdom G2) fc)) .
Proof.
  destruct flags.
  rewrite <- ! fresh_element_eq_ctxdom_list.
  rewrite ctx_keys_concat.
  eapply fresh_chain_eq_inv.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_swap_var_var.
  eapply fresh_chain_eq_merge_list_list.
Qed.

Lemma fresh_chain_eq_merge_var_objctxdom flags fc G1 G2:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (G1 §§& G2)) fc) (fresh_chain_var (fresh_element_objctxdom G2) (fresh_chain_var (fresh_element_ctxdom G1) fc)) .
Proof.
  destruct flags.
  rewrite <- ! fresh_element_eq_ctxdom_list.
  rewrite <- ! fresh_element_eq_objctxdom_list.
  rewrite ctx_obj_keys.
  rewrite fresh_element_eq_list_app_swap.
  eapply fresh_chain_eq_inv.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_swap_var_var.
  eapply fresh_chain_eq_merge_list_list.
Qed.


Lemma fresh_chain_eq_merge_var_renvdom flags fc G1 G2:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_renvdom (G1 & G2)) fc) (fresh_chain_var (fresh_element_renvdom G2) (fresh_chain_var (fresh_element_renvdom G1) fc)) .
Proof.
  destruct flags.
  rewrite <- ! fresh_element_eq_renvdom_list.
  rewrite env_concat_keys.
  rewrite fresh_element_eq_list_app_swap.
  eapply fresh_chain_eq_inv.
  eapply fresh_chain_eq_trans.
  eapply fresh_chain_eq_swap_var_var.
  eapply fresh_chain_eq_merge_list_list.
Qed.


Lemma fresh_chain_eq_list_cons flags x xs fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_list (x::xs)%list) fc) (fresh_chain_var (fresh_element_var x) (fresh_chain_var (fresh_element_list xs) fc)).
Proof.
  constructor; cases_if; simpls; auto 1.
- introv Hfci.
    inversions Hfci.

    forwards H3r: fresh_chain_element_eq_fresh H2 fresh_element_eq_list_cons_r; fweauto. rewrite fresh_element_eq_list_cons_r in H2; fweauto. apply true.
    forwards H3l: fresh_chain_element_eq_fresh H3 fresh_element_eq_list_cons_l; fweauto.
    inversions H2. simpls. destructecs H0.
    rewrite union_empty_l in H0.

    constructor; auto.

    constructor; auto.
    constructor; auto.
    +
    simpls. split; eauto. eapply from_list_fresh; eauto.
    + simpls. eapply disjoint_empty_l; eauto.
  - introv Hfci.
    inversions Hfci.
    inversions H1.
    inversions H2.
    inversions H5.
    inversions H3.
    inversions H5.
    simpls.
    destructecs H.
    constructor; auto.
    constructor; auto. simpl. split. apply notin_empty. rewrite union_empty_l. eapply fresh_from_list; eauto.
    eapply fresh_chain_from_disjoint_set.
    *
    forwards: fresh_chain_disjoint_set H6.
    forwards: fresh_chain_disjoint_set H7.
    simpls. rewrite from_list_cons.
    eapply disjoint_union_l; eauto.
    * eauto.
    * eauto.
  - rewrite ! union_assoc. rewrite from_list_cons. apply subset_refl.
  - rewrite ! union_assoc. rewrite from_list_cons. apply subset_refl.
  Unshelve. all: fresh_chain_eq_solve_flags.
Qed.

Lemma fresh_chain_eq_list_nil flags fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_list (nil)%list) fc) fc.
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci; auto.

  - introv Hfci.
    constructor; auto.
    eapply fresh_chain_from_disjoint_set; simpls.
    rewrite from_list_nil. apply disjoint_empty_l.
    constructor; auto.
    auto.
  - rewrite from_list_nil. union_sub_solver.
  - rewrite from_list_nil. union_sub_solver.
Qed.


Lemma fresh_chain_eq_ctx_empty flags fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctx ctx_empty ) fc) fc.
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci; auto.

  - introv Hfci.
    constructor; auto.
    eapply fresh_chain_from_disjoint_set; simpls.
    rewrite union_empty_l. apply disjoint_empty_l.
    constructor; auto.
    auto.
  - union_sub_solver.
  - union_sub_solver.
Qed.
Lemma fresh_chain_eq_ctxdom_empty flags fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom ctx_empty ) fc) fc.
Proof.
  constructor; cases_if; simpls; auto 1.
  - introv Hfci.
    inversions Hfci; auto.

  - introv Hfci.
    constructor; auto.
    eapply fresh_chain_from_disjoint_set; simpls.
    apply disjoint_empty_l.
    constructor; auto.
    auto.
  - union_sub_solver.
  - union_sub_solver.
Qed.


Lemma fresh_chain_eq_ctxdom_push flags x vk T G fc:
  fresh_chain_eq flags (fresh_chain_var (fresh_element_ctxdom (G § x / vk ~ T)) fc) (fresh_chain_var (fresh_element_var x) (fresh_chain_var (fresh_element_ctxdom G) fc)).
Proof.
  destruct flags.
  rewrite <- fresh_element_eq_ctxdom_list.
  rewrite <- fresh_element_eq_ctxdom_list.
  eapply fresh_chain_eq_list_cons.
Qed.

Lemma fresh_chain_eq_last_var_set flags fe:
  fresh_chain_eq flags (fresh_chain_var fe fresh_chain_nil) (fresh_chain_set fe fresh_chain_nil).
Proof.
  destruct flags.
  constructor; cases_if; simpls; auto 1; intros.
  - inversions H. constructor; eauto.
  - inversions H. constructor; eauto.
  - union_sub_solver.
  - union_sub_solver.
Qed.


Lemma fresh_chain_ctx_ok G fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 (fresh_chain_set (fresh_element_ctx G) fc2)) ->
  ctx_ok G.
Proof.
  introv Hfc.
  eapply fresh_chain_is_append_invr in Hfc.
  inversions Hfc.
  inversions H2.
  auto.
Qed.
Lemma fresh_chain_ctxdom_ok G fc1 fc2:
  fresh_chain_is (fresh_chain_append fc1 (fresh_chain_set (fresh_element_ctxdom G) fc2)) ->
  ctx_ok G.
Proof.
  introv Hfc.
  eapply fresh_chain_is_append_invr in Hfc.
  inversions Hfc.
  inversions H2.
  auto.
Qed.


(* Weak version of the chain, containing only sets *)
Fixpoint fresh_chain_weak (fc: fresh_chain): fresh_chain :=
  match fc with 
  | fresh_chain_var fe fc => fresh_chain_set fe (fresh_chain_weak fc)
  | fresh_chain_set fe fc => fresh_chain_set fe (fresh_chain_weak fc)
  | fresh_chain_subset fe fc => fresh_chain_set fe (fresh_chain_weak fc)
  | fresh_chain_nil => fresh_chain_nil
  | fresh_chain_branch fcBr fcM => fresh_chain_branch (fresh_chain_weak fcBr) (fresh_chain_weak fcM)
  | fresh_chain_union fc1 fc2 => fresh_chain_union (fresh_chain_weak fc1) (fresh_chain_weak fc2)
  | fresh_chain_parallel fcBr1 fcBr2 fcM => fresh_chain_parallel (fresh_chain_weak fcBr1) (fresh_chain_weak fcBr2) (fresh_chain_weak fcM)
  end.


Lemma fresh_chain_weak_to_set_large_sub fc1 :
  fresh_chain_to_set_large (fresh_chain_weak fc1) \c fresh_chain_to_set_large (fc1).
Proof.
  induction fc1; intros; simpls; union_sub_solver.
Qed.


Lemma fresh_chain_weak_to_set_large_append_eq fc1 fc2 :
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_chain_to_set_large (fresh_chain_append (fresh_chain_weak fc1) fc2) = fresh_chain_to_set_large ((fresh_chain_append fc1 fc2)).
Proof.
  introv His. reverts fc2.
  induction fc1; intros; simpls; inversions His; simpls; rewrite ? IHfc1, ? IHfc1_1, ? IHfc1_2, ? IHfc1_3; rewrite ? fresh_chain_append_to_set_large in *; simpl; eauto.
  - eapply fresh_chain_is_append_invr; eauto.
  -
    forwards: IHfc1_1 H1.
    forwards: IHfc1_2 H3.
    forwards: IHfc1_3 fc2. eapply fresh_chain_is_append_invr; eauto.
    rewrite ? fresh_chain_append_to_set_large in *.
    forwards Heq1: f_equal2 (@union var) H H0.

    eapply eq_trans. eapply eq_trans. 2: eapply Heq1.
    eapply fset_extens;
    union_sub_solver.
    eapply fset_extens;
    union_sub_solver.
  - eapply fresh_chain_is_append_invr; eauto.
Qed.

Lemma fresh_chain_weak_append fc1 fc2:
  fresh_chain_weak (fresh_chain_append fc1 fc2) =   fresh_chain_append (fresh_chain_weak fc1) (fresh_chain_weak fc2).
Proof.
  introv.
  reverts fc2.
  induction fc1; intros; simpls; rewrite ? IHfc1, ?IHfc1_1, ?IHfc1_2, ?IHfc1_3; eauto.
Qed.

Lemma fresh_chain_weak_is fc:
  fresh_chain_is fc -> fresh_chain_is (fresh_chain_weak fc).
Proof.
  introv His.
  induction His; simpl; eauto.
    econstructor. rewrite <- fresh_chain_weak_append. eauto.
    econstructor; rewrite <- fresh_chain_weak_append; eauto.
Qed.

Lemma fresh_chain_weak_to_set_eq fc1 :
  fresh_chain_is fc1 ->
  fresh_chain_to_set (fresh_chain_weak fc1) = fresh_chain_to_set (fc1).
Proof.
  introv His.
  forwards: fresh_chain_weak_is; fweauto.
  rewrite <- ! fresh_chain_to_set_large_eq by auto. 
  forwards: fresh_chain_weak_to_set_large_append_eq fc1 fresh_chain_nil.
  rewrite fresh_chain_append_nil_r; auto.
  rewrite ! fresh_chain_append_nil_r in H0; auto.
Qed.


Lemma fresh_chain_weak_idem fc:
  fresh_chain_weak (fresh_chain_weak fc) = (fresh_chain_weak fc).
Proof.
  induction fc; simpl; rewrite ? IHfc, ? IHfc1, ? IHfc2, ? IHfc3; auto.
Qed.

Lemma fresh_chain_fresh_weak fe fc1:
  fresh_chain_fresh fe fc1 ->
  fresh_chain_fresh fe (fresh_chain_weak fc1).
Proof.
  induction fc1; intros; simpls; try inversions H; eauto.
  econstructor.
  eapply fresh_chain_fresh_append.
  eapply IHfc1_1. eauto.
  forwards: fresh_chain_fresh_append_invl H3; eauto.
  forwards: fresh_chain_fresh_append_invr H3; eauto.
  eauto.
Qed.
Lemma fresh_chain_fresh_weak_inv fe fc1:
  fresh_chain_is fc1 ->
  fresh_element_is fe ->
  fresh_chain_fresh fe (fresh_chain_weak fc1) ->
  fresh_chain_fresh fe fc1.
Proof.
  introv Hfc Hfe Hfr.
  forwards Hdi: fresh_chain_disjoint_set Hfr.
  eapply fresh_chain_from_disjoint_set.
  rewrite fresh_chain_weak_to_set_eq in Hdi by auto.
  auto.
  auto.  auto.
Qed.

Lemma fresh_chain_eq_weak fl1  fc:
  fresh_chain_eq ⚐ fl1 false false false ⚐ fc (fresh_chain_weak fc).
Proof.
  constructor; cases_if; auto.
  - eapply fresh_chain_weak_is; eauto.

Qed.

Lemma fresh_chain_eq_weak_is fl1 fl2 fl3 fc:
  fresh_chain_is fc ->
  fresh_chain_eq ⚐ fl1 false fl2 fl3 ⚐ fc (fresh_chain_weak fc).
Proof. 
  introv His.
  constructor; cases_if; auto.
  - eapply fresh_chain_weak_is; eauto.
  - rewrite fresh_chain_weak_to_set_eq. eapply subset_refl. auto.
  - rewrite fresh_chain_weak_to_set_eq. eapply subset_refl. auto.
Qed.

Lemma fresh_chain_add_var_weak fc fe:
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_var fe (fresh_chain_weak fc)) ->
  fresh_chain_is (fresh_chain_var fe fc).
Proof.
  introv Hfc Hfe.
  inversions Hfe.
  constructor; auto.
  eapply fresh_chain_fresh_weak_inv; eauto.
Qed.

(* "Inverting" a chain, from left of a variable to the right - throws away all sets.
    Branches are ignored (not implemented) yet. *)
Fixpoint fresh_chain_inversion (fc: fresh_chain): fresh_chain :=
  match fc with 
  | fresh_chain_var fe fc => fresh_chain_var fe (fresh_chain_inversion fc)
  | fresh_chain_set fe fc => fresh_chain_inversion fc
  | fresh_chain_subset fe fc => fresh_chain_inversion fc
  | _ => fresh_chain_nil
  end.

Lemma fresh_chain_fresh_inversion fe fc:
  fresh_chain_fresh fe fc ->
  fresh_chain_fresh fe (fresh_chain_inversion fc).
Proof.
  introv Hfr.
  induction Hfr; simpls; auto.
Qed.

Lemma fresh_chain_invert fcL fe fcR:
  fresh_chain_is (fresh_chain_append fcL (fresh_chain_var fe fcR)) ->
  fresh_chain_is (fresh_chain_var fe (fresh_chain_append (fresh_chain_inversion fcL) (fcR))).
Proof.
  introv His.
  induction fcL; simpls; eauto 1.
  - inversion His. eauto.
  - inversion His; subst.
    forwards: IHfcL; fweauto.
    inversions H.
    forwards: fresh_chain_fresh_append_invl H7.
    forwards: fresh_chain_fresh_append_invr H7.
    forwards: fresh_chain_fresh_append_invl H3.
    forwards: fresh_chain_fresh_append_invr H3.
    inversions H8.
    constructor; eauto.
    constructor; eauto.
    eapply fresh_chain_fresh_append.
    eapply fresh_chain_fresh_inversion; auto.
    auto.
    constructor; eauto.
    eapply fresh_element_fresh_symm; eauto.
  - inversion His. eauto.
  - inversions His.
    forwards: fresh_chain_is_append_invr H0.
    forwards: fresh_chain_is_append_invr H.
    auto.
  - inversions His.
    forwards: fresh_chain_is_append_invr H1.
    forwards: fresh_chain_is_append_invr H.
    auto.
  - inversions His.
    forwards: fresh_chain_is_append_invr H1.
    auto.
Qed.


Inductive fresh_chain_indifferent: fresh_chain -> Prop :=
| fresh_chain_indifferent_set fe fc: fresh_chain_indifferent fc -> fresh_chain_indifferent (fresh_chain_set fe fc)
| fresh_chain_indifferent_branch fcBr fcM: fresh_chain_indifferent fcBr -> fresh_chain_indifferent fcM -> fresh_chain_indifferent (fresh_chain_branch fcBr fcM)
| fresh_chain_indifferent_union fc1 fc2: fresh_chain_indifferent fc1 -> fresh_chain_indifferent fc2 -> fresh_chain_indifferent (fresh_chain_union fc1 fc2)
| fresh_chain_indifferent_parallel fc1 fc2 fc3: fresh_chain_indifferent fc1 -> fresh_chain_indifferent fc2 -> fresh_chain_indifferent fc3 -> fresh_chain_indifferent (fresh_chain_parallel fc1 fc2 fc3)
| fresh_chain_indifferent_nil: fresh_chain_indifferent (fresh_chain_nil)
.
Hint Constructors fresh_chain_indifferent: core.

Lemma fresh_chain_weak_indifferent fc:
  fresh_chain_indifferent (fresh_chain_weak fc).
Proof.
  induction fc; simpl; eauto.
Qed.

Lemma fresh_chain_append_nil_inv fc1 fc2:
  (fresh_chain_append fc1 fc2) = fresh_chain_nil ->
  fc1 = fresh_chain_nil /\ fc2 = fresh_chain_nil.
Proof.
  introv Heq.
  induction fc1; simpls; try discriminate; split;  eauto.
Qed.

Lemma fresh_chain_indifferent_append fc1 fc2:
  fresh_chain_indifferent fc1 ->
  fresh_chain_indifferent fc2 ->
  fresh_chain_indifferent (fresh_chain_append fc1 fc2).
Proof.
  introv Hindf1 Hindf2.
  induction Hindf1; simpls; eauto.
Qed.
Lemma fresh_chain_indifferent_append_inv fc1 fc2:
  fresh_chain_indifferent (fresh_chain_append fc1 fc2) ->
  fresh_chain_indifferent fc1 /\
  fresh_chain_indifferent fc2.
Proof.
  introv Hindf.
  induction fc1; inversions Hindf; simpls; subst; split; eauto.
  - forwards ?H ?H: IHfc1; eauto.
  - forwards ?H ?H: IHfc1; eauto.
  - forwards ?H ?H: IHfc1_2; eauto.
  - forwards ?H ?H: IHfc1_2; eauto.
  - forwards ?H ?H: IHfc1_3; eauto.
  - forwards ?H ?H: IHfc1_3; eauto.
  - forwards ?H ?H: IHfc1_2; eauto.
    forwards ?H ?H: IHfc1_1; eauto.
  - forwards ?H ?H: IHfc1_2; eauto.
Qed.


Lemma fresh_chain_is_indifferent_append fc1 fc2 fc3:
  fresh_chain_is (fresh_chain_append fc1 fc3)->
  fresh_chain_is fc2 ->
  fresh_chain_indifferent fc1 ->
  fresh_chain_is (fresh_chain_append fc1 fc2).
Proof.
  introv His1 His2 Hindf.
  reverts fc2 fc3.
  induction Hindf; intros; simpls; inversion His1; simpls; subst; try discriminate; eauto.
  - constructor.
    eapply IHHindf1. eapply IHHindf2. eauto.
    eapply fresh_chain_is_append_invr; eauto. eauto.
  - constructor.
    eapply IHHindf1. eapply IHHindf3. eauto.
    eapply fresh_chain_is_append_invr; eauto. eauto.
    eapply IHHindf2. eapply IHHindf3. eauto.
    eapply fresh_chain_is_append_invr; eauto. eauto.
Qed.

(* COLLECTING

  These definitions allow defining a function from chains to props,
  which allows defining lemmas where the argument is a the chain and the premises are derived by evaluation from the chain.

  For example, a ctx_ok premise can be generated from a chain, by collecting the ctxdom parts in the chain.
  To get a prop that is equivalent to validity of the chain, the other kinds of chain elements are collected as a set,
  which is used to construct another condition.
  To avoid generating useless conditions, the props are collected under an option.
*)


(* Takes two options, and if both are present, applies a binary function *)
Definition option_map_either (A: Type) (f : A -> A -> option A) (o : option A) (o' : option A): option A :=
match o with
| Some x => match o' with
            | Some x' => f x x'
            | None => Some x
            end
| None => o'
end.

(* Merges two sets into an optional set, using set union *)
Definition fresh_chain_reduce_set (X Y: fset var): option (fset var) := Some (X \u Y).
(* Merges two optional sets into an optional set, using set union *)
Definition fresh_chain_reduce_opt_set: forall (X Y: option(fset var)), option (fset var) := option_map_either fresh_chain_reduce_set.
(* Merges two elements into an optional element using R *)
Definition fresh_chain_reduce_elem(A: Type) (R: A -> A -> A) (a b: A) := Some (R a b).
(* Merges two optional elements into an optional element using R *)
Definition fresh_chain_reduce_opt_elem(A: Type) (R: A -> A -> A): forall (a b: option A), option A := option_map_either (fresh_chain_reduce_elem R).

(* Merges two pairs of an element and an set, into an optional pair, using R and set union *)
Definition fresh_chain_reduce_pair(A: Type) (R: A -> A -> A) (o1 o2: (option A) * (option (fset var))): (option A) * (option (fset var)) :=
  match o1 with (a1, X1) =>
  match o2 with (a2, X2) =>
  ((fresh_chain_reduce_opt_elem R a1 a2), fresh_chain_reduce_opt_set X1 X2)
  end end.

(* Collects a pair of optional element and optional set over a chain *)
Fixpoint fresh_chain_collect_opt (A: Type) (FE: fresh_element -> (option A * option (fset var)))  (R: A -> A -> A) (fc: fresh_chain) :  (option A * (option (fset var))) :=
  match fc with
  | fresh_chain_nil => (None, None)
  | fresh_chain_set fe fc => fresh_chain_reduce_pair R (fresh_chain_collect_opt FE R fc) (FE fe)
  | fresh_chain_var fe fc => fresh_chain_reduce_pair R (fresh_chain_collect_opt FE R fc) (FE fe)
  | fresh_chain_subset fe fc => (fresh_chain_collect_opt FE R fc)
  | fresh_chain_branch fcBr fcM => (fresh_chain_collect_opt FE R fcM)
  (* Avoid combining branches and unions because of potential size explosion *)
  | fresh_chain_union fc1 fc2 =>  fresh_chain_reduce_pair R (fresh_chain_collect_opt FE R fc2) (fresh_chain_collect_opt FE R fc1)
  | fresh_chain_parallel fcBr1 fcBr2 fcM =>  fresh_chain_reduce_pair R (fresh_chain_reduce_pair R (fresh_chain_collect_opt FE R fcM) (fresh_chain_collect_opt FE R fcBr2)) (fresh_chain_collect_opt FE R fcBr1) 
  end.

(* Converts a pair of optional element and optional set into a condition*)
Definition fresh_collect_cond (A: Type) (R: A -> A -> A) (CA: A->Prop) (CS: fset var -> Prop) (CA0: option Prop) AX : Prop :=
  match AX with 
  | (None, None) => match CA0 with | Some P => P | None => True end
  | (Some a, None) => CA a
  | (None, Some X) => match CA0 with | Some P => P /\ CS X | None => CS X end
  | (Some a, Some X) => CA a /\ CS X
  end.

(* Collection spec:
  In order to state lemmas about fresh_chain_collect_opt and fresh_collect_cond,
  we need the arguments to behave sanely.
  All these operations and their relations are stored in this record.
*)

Record fresh_chain_collect_spec:= {
  (* The type of the element part of the pair *)
  fresh_chain_collect_spec_A: Type;
  (* Converts a chain element into a pair of optional element and optional set *)
  fresh_chain_collect_spec_FE: fresh_element -> (option fresh_chain_collect_spec_A * option (fset var));
  (* Merges two elements into one *)
  fresh_chain_collect_spec_R: fresh_chain_collect_spec_A -> fresh_chain_collect_spec_A -> fresh_chain_collect_spec_A;
  (* Condition for an element *)
  fresh_chain_collect_spec_CA: fresh_chain_collect_spec_A -> Prop;
  (* The condition distributes into element parts *)
  fresh_chain_collect_spec_CA1 a b: fresh_chain_collect_spec_CA (fresh_chain_collect_spec_R a b) -> fresh_chain_collect_spec_CA a;
  fresh_chain_collect_spec_CA2 a b: fresh_chain_collect_spec_CA (fresh_chain_collect_spec_R a b) -> fresh_chain_collect_spec_CA b;
  (* Condition for a set *)
  fresh_chain_collect_spec_CS: fset var -> Prop;
  (* The condition distributes into set parts *)
  fresh_chain_collect_spec_CS1 a b: fresh_chain_collect_spec_CS (a \u b) -> fresh_chain_collect_spec_CS a;
  fresh_chain_collect_spec_CS2 a b: fresh_chain_collect_spec_CS (a \u b) -> fresh_chain_collect_spec_CS b;
  (* An optional condition in case there is no element *)
  fresh_chain_collect_spec_C0: option Prop;
  (* The no-element condition holds holds even if there is an element *)
  fresh_chain_collect_spec_CA0:  match fresh_chain_collect_spec_C0 with | Some P => forall a, fresh_chain_collect_spec_CA a -> P | None => True end;
}.

(* Condition over a single element *)
Definition fresh_element_collect_cond (SP: fresh_chain_collect_spec) (fe: fresh_element) : Prop :=
  fresh_collect_cond (fresh_chain_collect_spec_R SP) (fresh_chain_collect_spec_CA SP) (fresh_chain_collect_spec_CS SP) (fresh_chain_collect_spec_C0 SP) ((fresh_chain_collect_spec_FE SP) fe).

(* Condition over the chain *)
Definition fresh_chain_collect_cond (SP: fresh_chain_collect_spec) (fc: fresh_chain) : Prop :=
  fresh_collect_cond
    (fresh_chain_collect_spec_R SP)
    (fresh_chain_collect_spec_CA SP)
    (fresh_chain_collect_spec_CS SP) 
    (fresh_chain_collect_spec_C0 SP)
    (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc).

Class fresh_chain_collect_spec_fw  (SP: fresh_chain_collect_spec) := {
    fresh_chain_collect_spec_CA12 a b:
      fresh_chain_collect_spec_CA SP a ->
      fresh_chain_collect_spec_CA SP b ->
      fresh_chain_collect_spec_CA SP (fresh_chain_collect_spec_R SP a b);
    fresh_chain_collect_spec_CS12 a b:
      fresh_chain_collect_spec_CS SP a ->
      fresh_chain_collect_spec_CS SP b ->
      fresh_chain_collect_spec_CS SP (a \u b)
}.

(* Collecting context domains *)
Fixpoint fresh_chain_ctx_map_cond (fe: fresh_element): (option ctx * option (fset var)) :=
  match fe with 
  | fresh_element_ctxdom G => ((Some G), None)
  | fresh_element_ctx G => ((Some G), (Some (ctx_fv G)))
  | _ => (None, Some (fresh_element_to_set fe))
  end.
Fixpoint fresh_chain_ctx_map_nofv_cond (fe: fresh_element): (option ctx * option (fset var)) :=
  match fe with 
  | fresh_element_ctxdom G => ((Some G), None)
  | fresh_element_ctx G => ((Some G), None)
  | _ => (None, None)
  end.
Fixpoint fresh_chain_renv_map_cond (fe: fresh_element): (option renv * option (fset var)) :=
  match fe with 
  | fresh_element_renvdom G => ((Some G), None)
  | fresh_element_renv G => ((Some G), (Some (fv_env G)))
  | _ => (None, Some (fresh_element_to_set fe))
  end.
Fixpoint fresh_chain_renv_map_nofv_cond (fe: fresh_element): (option renv * option (fset var)) :=
  match fe with 
  | fresh_element_renvdom G => ((Some G), None)
  | fresh_element_renv G => ((Some G), None)
  | _ => (None, None)
  end.
Fixpoint fresh_chain_heap_map_cond (fe: fresh_element): (option heap * option (fset var)) :=
  match fe with 
  | fresh_element_heapdom G => ((Some G), None)
  | fresh_element_heap G => ((Some G), (Some (fv_env G)))
  | _ => (None, Some (fresh_element_to_set fe))
  end.
Fixpoint fresh_chain_heap_map_nofv_cond (fe: fresh_element): (option heap * option (fset var)) :=
  match fe with 
  | fresh_element_heapdom G => ((Some G), None)
  | fresh_element_heap G => ((Some G), None)
  | _ => (None, None)
  end.


Lemma fresh_chain_collect_cond_var_inv (SP: fresh_chain_collect_spec) (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP (fresh_chain_var fe fc) ->
  fresh_chain_collect_cond SP fc /\ fresh_element_collect_cond SP fe.
Proof.
  introv Hfc.
  hnf in Hfc. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc) as fcA fcX.
  destructc ((fresh_chain_collect_spec_FE SP) fe) as feA feX.
  all: destruct SP; simpls.
  destruct fcA, fcX, feA, feX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; splits; eauto 8.
Qed.
Lemma fresh_chain_collect_cond_var(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP fc ->
  fresh_element_collect_cond SP fe ->
  fresh_chain_collect_cond SP (fresh_chain_var fe fc).
Proof.
  introv Hfc Hfe.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc) as fcA fcX.
  destructc ((fresh_chain_collect_spec_FE SP) fe) as feA feX.
  all: destruct SP; simpls.
  destruct fcA, fcX, feA, feX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 2.
Qed.

Lemma fresh_chain_collect_cond_set_inv (SP: fresh_chain_collect_spec) (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP (fresh_chain_set fe fc) ->
  fresh_chain_collect_cond SP fc /\ fresh_element_collect_cond SP fe.
Proof.
  eapply fresh_chain_collect_cond_var_inv; eauto.
Qed.
Lemma fresh_chain_collect_cond_set(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP fc ->
  fresh_element_collect_cond SP fe ->
  fresh_chain_collect_cond SP (fresh_chain_set fe fc).
Proof.
  introv Hfc Hfe.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc) as fcA fcX.
  destructc ((fresh_chain_collect_spec_FE SP) fe) as feA feX.
  all: destruct SP; simpls.
  destruct fcA, fcX, feA, feX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 2.
Qed.

Lemma fresh_chain_collect_cond_branch_inv (SP: fresh_chain_collect_spec) (fcBr fcM: fresh_chain):
  fresh_chain_collect_cond SP (fresh_chain_branch fcBr fcM) ->
  fresh_chain_collect_cond SP fcM.
Proof.
  eauto.
Qed.
Lemma fresh_chain_collect_cond_branch(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fcBr fcM: fresh_chain):
  fresh_chain_collect_cond SP fcM ->
  fresh_chain_collect_cond SP (fresh_chain_branch fcBr fcM).
Proof.
  introv Hfc.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fcM) as fcA fcX.
  all: destruct SP; simpls.
  destruct fcA, fcX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 2.
Qed.

Lemma fresh_chain_collect_cond_subset_inv (SP: fresh_chain_collect_spec) (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP (fresh_chain_subset fe fc) ->
  fresh_chain_collect_cond SP fc.
Proof.
  introv Hfc.
  hnf in Hfc. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc) as fcA fcX.
  all: destruct SP; simpls.
  destruct fcA, fcX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; eauto 8.
Qed.
Lemma fresh_chain_collect_cond_subset(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fc: fresh_chain) (fe: fresh_element):
  fresh_chain_collect_cond SP fc ->
  fresh_chain_collect_cond SP (fresh_chain_subset fe fc).
Proof.
  introv Hfc.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc) as fcA fcX.
  destructc ((fresh_chain_collect_spec_FE SP) fe) as feA feX.
  all: destruct SP; simpls.
  destruct fcA, fcX, feA, feX,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 2.
Qed.


Lemma fresh_chain_collect_cond_union_inv (SP: fresh_chain_collect_spec) (fc1 fc2: fresh_chain):
  fresh_chain_collect_cond SP (fresh_chain_union fc1 fc2) ->
  fresh_chain_collect_cond SP fc1 /\ fresh_chain_collect_cond SP fc2.
Proof.
  introv Hfc.
  hnf in Hfc. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc1) as fcA1 fcX1.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc2) as fcA2 fcX2.
  simpls.
  all: destruct SP; simpls.
  destruct fcA1, fcX1, fcA2, fcX2,  fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; splits; eauto 4.
Qed.


Lemma fresh_chain_collect_cond_union(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fc1 fc2: fresh_chain):
  fresh_chain_collect_cond SP fc1 ->
  fresh_chain_collect_cond SP fc2 ->
  fresh_chain_collect_cond SP (fresh_chain_union fc1 fc2).
Proof.
  introv Hfc1 Hfc2.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc1) as fcA fcX.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc2) as fcA2 fcX2.
  all: destruct SP; simpls.
  destruct fcA, fcX, fcA2, fcX2, fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 2.
Qed.

Lemma fresh_chain_collect_cond_parallel_inv (SP: fresh_chain_collect_spec) (fc1 fc2 fcM: fresh_chain):
  fresh_chain_collect_cond SP (fresh_chain_parallel fc1 fc2 fcM) ->
  fresh_chain_collect_cond SP fc1 /\ fresh_chain_collect_cond SP fc2 /\ fresh_chain_collect_cond SP fcM.
Proof.
  introv Hfc.
  hnf in Hfc. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc1) as fcA1 fcX1.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc2) as fcA2 fcX2.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fcM) as fcAM fcXM.
  simpls.
  all: destruct SP; simpls.
  destruct fcA1, fcX1, fcA2, fcX2, fcAM, fcXM, fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; splits; eauto 6.
Qed.


Lemma fresh_chain_collect_cond_parallel(SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP} (fc1 fc2 fcM: fresh_chain):
  fresh_chain_collect_cond SP fc1 ->
  fresh_chain_collect_cond SP fc2 ->
  fresh_chain_collect_cond SP fcM ->
  fresh_chain_collect_cond SP (fresh_chain_parallel fc1 fc2 fcM).
Proof.
  introv Hfc1 Hfc2 HfcM.
  pose proof (fresh_chain_collect_spec_CA12).
  pose proof (fresh_chain_collect_spec_CS12).
  hnf in *. simpls.
  unfold fresh_chain_collect_cond, fresh_element_collect_cond in *.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc1) as fcA fcX.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fc2) as fcA2 fcX2.
  destructc (fresh_chain_collect_opt (fresh_chain_collect_spec_FE SP) (fresh_chain_collect_spec_R SP) fcM) as fcA3 fcX3.
  all: destruct SP; simpls.
  destruct fcA, fcX, fcA2, fcX2, fcA3, fcX3, fresh_chain_collect_spec_C1.
  all: simpls; destructecs_all; try split; eauto 3.
Qed.


Lemma fresh_chain_collect_cond_nil_inv (SP: fresh_chain_collect_spec) (fc: fresh_chain):
  fresh_chain_collect_cond SP fc ->
  match fresh_chain_collect_spec_C0 SP with | Some P => P | None => True end.
Proof.
  introv Hfc.
  destruct SP. destruct fresh_chain_collect_spec_C1; simpl.
  induction fc; simpls.
  - eauto.
  - forwards Hfcs: fresh_chain_collect_cond_set_inv Hfc.
    destructecs Hfcs. eauto.
  - forwards Hfcs: fresh_chain_collect_cond_var_inv Hfc.
    destructecs Hfcs. eauto.
  - forwards Hfcs: fresh_chain_collect_cond_subset_inv Hfc.
    eauto.
  - forwards Hfcs: fresh_chain_collect_cond_branch_inv Hfc.
     eauto.
  - forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hfc.
    destructecs Hfcs. eauto.
  - forwards Hfcs: fresh_chain_collect_cond_union_inv Hfc.
    destructecs Hfcs. eauto.
  - auto.
Qed.

Lemma fresh_chain_collect_cond_append (SP: fresh_chain_collect_spec) {SPF: fresh_chain_collect_spec_fw SP}  (fc1 fc2: fresh_chain):
    fresh_chain_collect_cond SP fc1 ->
  fresh_chain_collect_cond SP fc2 ->
  fresh_chain_collect_cond SP (fresh_chain_append fc1 fc2).
Proof.
  introv Hfc1 Hfc2.
  reverts fc2.
  induction fc1; intros; simpls.
  - auto.
  - forwards Hfcs: fresh_chain_collect_cond_set_inv Hfc1.
    destructecs Hfcs.
    eapply fresh_chain_collect_cond_set; eauto.
  - forwards Hfcs: fresh_chain_collect_cond_var_inv Hfc1.
    destructecs Hfcs.
    eapply fresh_chain_collect_cond_var; eauto.
  - forwards Hfcs: fresh_chain_collect_cond_subset_inv Hfc1.
    eapply fresh_chain_collect_cond_subset; eauto.
  - forwards Hfcs: fresh_chain_collect_cond_branch_inv Hfc1.
    eapply fresh_chain_collect_cond_branch; eauto.
  - forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hfc1.
    destructecs Hfcs.
    eapply fresh_chain_collect_cond_parallel; eauto.

  - forwards Hfcs: fresh_chain_collect_cond_union_inv Hfc1.
    destructecs Hfcs.
    eapply fresh_chain_collect_cond_union; eauto.


Qed.

(* INSTANCES for COLLECTING *)

(* Collecting condition of the form x §# G *)
Definition fresh_chain_ctx_notin_spec (x: var) : fresh_chain_collect_spec := {|
    fresh_chain_collect_spec_A := ctx;
    fresh_chain_collect_spec_FE := fresh_chain_ctx_map_cond;
    fresh_chain_collect_spec_R := ctx_concat;
    fresh_chain_collect_spec_CA := (fun G => x §# G);
    fresh_chain_collect_spec_CA1 := (@ctx_unbound_concat_l x);
    fresh_chain_collect_spec_CA2 := (@ctx_unbound_concat_r x);
    fresh_chain_collect_spec_CS := (notin x);
    fresh_chain_collect_spec_CS1 := (@notin_union_r1 x);
    fresh_chain_collect_spec_CS2 := (@notin_union_r2 x);
    fresh_chain_collect_spec_C0 := None;
    fresh_chain_collect_spec_CA0 := I |}.

(* Collecting condition of the form ctx_ok G1 §# G *)
Definition fresh_chain_ctx_ok_spec (G1: ctx) : fresh_chain_collect_spec := {|
    fresh_chain_collect_spec_A := ctx;
    fresh_chain_collect_spec_FE := fresh_chain_ctx_map_cond;
    fresh_chain_collect_spec_R := ctx_concat;
    fresh_chain_collect_spec_CA := (fun G => ctx_ok (G1 §& G));
    fresh_chain_collect_spec_CA1 := (@ctx_ok_concat_rl G1);
    fresh_chain_collect_spec_CA2 := (@ctx_ok_concat_rr G1);
    fresh_chain_collect_spec_CS := (fun X => disjoint (ctx_dom G1) X);
    fresh_chain_collect_spec_CS1 := (fun X Y CS => proj21 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_CS2 := (fun X Y CS => proj22 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_C0 := Some (ctx_ok G1);
    fresh_chain_collect_spec_CA0 := ( @ctx_ok_concat_l G1 ) |}.

(* Collecting condition of the form ctx_ok G *)
Definition fresh_chain_ctx_just_ok_spec : fresh_chain_collect_spec := {|
    fresh_chain_collect_spec_A := ctx;
    fresh_chain_collect_spec_FE := fresh_chain_ctx_map_nofv_cond;
    fresh_chain_collect_spec_R := ctx_concat;
    fresh_chain_collect_spec_CA := ctx_ok;
    fresh_chain_collect_spec_CA1 := ctx_ok_concat_l;
    fresh_chain_collect_spec_CA2 := ctx_ok_concat_r;
    fresh_chain_collect_spec_CS := (fun X => True);
    fresh_chain_collect_spec_CS1 := (fun X Y CS => I);
    fresh_chain_collect_spec_CS2 := (fun X Y CS => I);
    fresh_chain_collect_spec_C0 := None;
    fresh_chain_collect_spec_CA0 := I |}.

Definition fresh_chain_renv_ok_spec (G1: renv) : fresh_chain_collect_spec := {|
    fresh_chain_collect_spec_A := renv;
    fresh_chain_collect_spec_FE := fresh_chain_renv_map_cond;
    fresh_chain_collect_spec_R := (@concat var);
    fresh_chain_collect_spec_CA := (fun G => ok (G1 & G));
    fresh_chain_collect_spec_CA1 := (@ok_concat_rl var G1);
    fresh_chain_collect_spec_CA2 := (@ok_concat_rr var G1);
    fresh_chain_collect_spec_CS := (fun X => disjoint (dom G1) X);
    fresh_chain_collect_spec_CS1 := (fun X Y CS => proj21 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_CS2 := (fun X Y CS => proj22 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_C0 := Some (ok G1);
    fresh_chain_collect_spec_CA0 := ( @ok_concat_inv_l var G1 ) |}.

Definition fresh_chain_heap_ok_spec (G1: heap) : fresh_chain_collect_spec := {|
    fresh_chain_collect_spec_A := heap;
    fresh_chain_collect_spec_FE := fresh_chain_heap_map_cond;
    fresh_chain_collect_spec_R := (@concat item);
    fresh_chain_collect_spec_CA := (fun G => ok (G1 & G));
    fresh_chain_collect_spec_CA1 := (@ok_concat_rl item G1);
    fresh_chain_collect_spec_CA2 := (@ok_concat_rr item G1);
    fresh_chain_collect_spec_CS := (fun X => disjoint (dom G1) X);
    fresh_chain_collect_spec_CS1 := (fun X Y CS => proj21 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_CS2 := (fun X Y CS => proj22 (disjoint_union_r_inv CS));
    fresh_chain_collect_spec_C0 := Some (ok G1);
    fresh_chain_collect_spec_CA0 := ( @ok_concat_inv_l item G1 ) |}.

Inductive fresh_element_path: Set :=
| fresh_element_here: fresh_element_path
| fresh_element_dom: fresh_element_path
| fresh_element_left (fp: fresh_element_path): fresh_element_path 
| fresh_element_right (fp: fresh_element_path): fresh_element_path.

(* Specifies a linear part of a chain, where branches and some elements can be skipped *)
Inductive fresh_chain_path: Set :=
| fresh_chain_end: fresh_chain_path
| fresh_chain_here (fp: fresh_chain_path): fresh_chain_path 
| fresh_chain_here_set (fp: fresh_chain_path) (fep: fresh_element_path): fresh_chain_path 
| fresh_chain_here_var (fp: fresh_chain_path) (fep: fresh_element_path): fresh_chain_path 
| fresh_chain_skip (n: nat) (fp: fresh_chain_path): fresh_chain_path
| fresh_chain_left (fp: fresh_chain_path): fresh_chain_path 
| fresh_chain_right (fp: fresh_chain_path): fresh_chain_path
| fresh_chain_enter (fp: fresh_chain_path): fresh_chain_path
| fresh_chain_lookahead (fp1 fp2: fresh_chain_path): fresh_chain_path. (* Look back using fp1 and continue with fp2 *)

(* Skips first n elements in the chain *)
Fixpoint fresh_chain_skip_in n (fc: fresh_chain) :=
  match n, fc with
  | 0, _ => fc
  | S n, fresh_chain_set X fc1 => (fresh_chain_skip_in n fc1)
  | S n, fresh_chain_var X fc1 => (fresh_chain_skip_in n fc1)
  | S n, fresh_chain_subset X fc1 => (fresh_chain_skip_in n fc1)
  | S n, (fresh_chain_branch fcBr fcM) => (fresh_chain_skip_in n fcM)
  | S n, fresh_chain_parallel fc1 fc2 fcM => (fresh_chain_skip_in n fcM)
  | _, _ => fresh_chain_nil  (* Cannot skip, for example union *)
  end.

Fixpoint fresh_element_select (fcp: fresh_element_path) (fc: fresh_element) : fresh_element :=
  match fcp, fc with
  | (fresh_element_left fcp1), (fresh_element_containing fcM fcS) => fresh_element_select fcp1 fcM
  | (fresh_element_right fcp1), (fresh_element_containing fcM fcS) => fresh_element_select fcp1 fcS
  | (fresh_element_dom), (fresh_element_ctx G) => (fresh_element_ctxdom G)
  | (fresh_element_dom), (fresh_element_renv G) => (fresh_element_renvdom G)
  | (fresh_element_dom), (fresh_element_heap G) => (fresh_element_heapdom G)
  | _,_ => fc
  end.

Fixpoint fresh_chain_select_p (fcp: fresh_chain_path) (fc: fresh_chain) : fresh_chain * bool :=
  match fcp, fc with
  | fresh_chain_here fcp1, fresh_chain_set X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_set X fc2, true)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_here fcp1, fresh_chain_var X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_var X fc2, true)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_here_set fcp1 fep, fresh_chain_set X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_set (fresh_element_select fep X) fc2, true)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_here_set fcp1 fep, fresh_chain_var X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_set (fresh_element_select fep X) fc2, true) (* converting var to set *)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_here_set fcp1 fep, fresh_chain_subset X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_set (fresh_element_select fep X) fc2, true) (* converting subset to set *)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_here_var fcp1 fep, fresh_chain_var X fc1 =>
    match (fresh_chain_select_p fcp1 fc1) with
    | (fc2, true) => (fresh_chain_var (fresh_element_select fep X) fc2, true) (* selecting a var *)
    | (fc2, false) => (fc2, false)
    end
  | fresh_chain_skip n fcp1, _ =>
    (fresh_chain_select_p fcp1 (fresh_chain_skip_in n fc))
  | fresh_chain_enter fcp1, (fresh_chain_branch fcBr fcM) =>
    match (fresh_chain_select_p fcp1 (fresh_chain_append fcBr fcM)) with
    | (fc2, _) => (fc2, false)
    end
  | fresh_chain_left fcp1, (fresh_chain_union fc1 fc2) =>
    (fresh_chain_select_p fcp1 fc1)
  | fresh_chain_right fcp1, (fresh_chain_union fc1 fc2) =>
    (fresh_chain_select_p fcp1 fc2)
  | fresh_chain_left fcp1, (fresh_chain_parallel fc1 fc2 fcM) =>
    (fresh_chain_select_p fcp1 (fresh_chain_append fc1 fcM))
  | fresh_chain_right fcp1, (fresh_chain_parallel fc1 fc2 fcM) =>
    (fresh_chain_select_p fcp1 (fresh_chain_append fc2 fcM))
  | fresh_chain_lookahead fcp1 fcp2, _ =>
    match (fresh_chain_select_p fcp1 (fresh_chain_weak fc)), (fresh_chain_select_p fcp2 fc) with (* The lookahead cannot have vars, only sets *)
    | (fc1, true), (fc2, true) => (fresh_chain_append fc1 fc2, true) (* If no branches are used, prepend the lookahead *)
    | _, X => X (* If any branches are used, just ignore the lookahead *)
    end
  | _, _ => (fresh_chain_nil, true)
  end.

Definition fresh_chain_select (fcp: fresh_chain_path) (fc: fresh_chain) : fresh_chain :=
  fst (fresh_chain_select_p fcp fc).

Local Lemma TEST_fresh_chain_select (fe1 fe2: fresh_element):
  fresh_chain_select (fresh_chain_lookahead (fresh_chain_here fresh_chain_end) (fresh_chain_here fresh_chain_end)) (fresh_chain_set fe1 (fresh_chain_set fe2 fresh_chain_nil)) = 
  (fresh_chain_set fe1 (fresh_chain_set fe1 fresh_chain_nil)).
Proof.
  auto.
Qed.

Lemma fresh_chain_fresh_skip n fe (fc: fresh_chain):
  fresh_chain_fresh fe fc ->
  fresh_chain_fresh fe (fresh_chain_skip_in n fc).
Proof.
  introv Hfr.
  reverts fc;
  induction n; intros; simpls.
  - auto.
  - destruct fc; simpls; auto 1.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
      eapply fresh_chain_fresh_append_invr in H2; auto.
Qed.

Lemma fresh_element_fresh_containing_inv2l fe1 fe2_1 fe2_2:
  fresh_element_fresh fe1 (fresh_element_containing fe2_1 fe2_2) ->
  fresh_element_fresh fe1 fe2_1.
Proof.
  introv Hfr.
  inversions Hfr; simpls.
  constructor; eauto.
Qed.
Lemma fresh_element_fresh_containing_inv2r fe1 fe2_1 fe2_2:
  fresh_element_is (fresh_element_containing fe2_1 fe2_2) ->
  fresh_element_fresh fe1 (fresh_element_containing fe2_1 fe2_2) ->
  fresh_element_fresh fe1 fe2_2.
Proof.
  introv Hi Hfr.
  inversions Hi.
  inversions Hfr; simpls.
  constructor; eauto. eapply fresh_subset; eauto. eapply disjoint_subset_l; eauto.
Qed.

Lemma fresh_element_fresh_containing_inv1l fe1_1 fe1_2 fe2:
  fresh_element_fresh (fresh_element_containing fe1_1 fe1_2) fe2 ->
  fresh_element_fresh fe1_1 fe2.
Proof.
  introv Hfr.
  inversions Hfr; simpls.
  constructor; eauto.
Qed.
Lemma fresh_element_fresh_containing_inv1r fe1_1 fe1_2 fe2:
  fresh_element_is fe2 ->
  fresh_element_is (fresh_element_containing fe1_1 fe1_2) ->
  fresh_element_fresh (fresh_element_containing fe1_1 fe1_2) fe2 ->
  fresh_element_fresh fe1_2 fe2.
Proof.
  introv Hi2 Hi Hfr.
  inversions Hi.
  eapply fresh_element_disjoint_set in Hfr.
  simpls.
  eapply fresh_element_from_disjoint_set.
  eapply disjoint_subset_r; eauto.
  eauto.
  eauto.
Qed.

Lemma fresh_element_fresh_select fep fe1 (fe2: fresh_element):
  fresh_element_is fe2 ->
  fresh_element_fresh fe1 fe2 ->
  fresh_element_fresh fe1 (fresh_element_select fep fe2).
Proof.
  introv Hi Hfr.
  reverts fe2;
  induction fep; intros; simpls.
  - auto.
  - destruct fe2; simpls; auto 1;
    inversions Hi; inversions Hfr; constructor; simpls; eauto using disjoint_union_r;
    apply disjoint_union_r_inv in H1; destruct H1; auto.
 
  - destruct fe2; simpls; auto 1.
    inversions Hi.
    eapply fresh_element_fresh_containing_inv2l in Hfr; eauto. 
  - destruct fe2; simpls; auto 1.
    inversions Hi.
    eapply fresh_element_fresh_containing_inv2r in Hfr; eauto.
Qed.

Lemma fresh_element_is_select fep (fe2: fresh_element):
  fresh_element_is fe2 ->
  fresh_element_is (fresh_element_select fep fe2).
Proof.
  introv Hi.
  reverts fe2;
  induction fep; intros; simpls.
  - auto.
  -  destruct fe2; simpls; auto 1;
    inversions Hi;  
  eauto.
  - destruct fe2; simpls; auto 1.
    inversions Hi.  
  eauto.
  - destruct fe2; simpls; auto 1.
    inversions Hi.  
  eauto.
Qed.


Lemma fresh_chain_is_skip n (fc: fresh_chain):
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_skip_in n fc).
Proof.
  introv Hfr.
  reverts fc;
  induction n; intros; simpls.
  - auto.
  - destruct fc; simpls; auto 1.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
    + inversions Hfr; eauto.
      eapply fresh_chain_is_append_invr in H0; auto.
    + inversions Hfr; eauto.
      eapply fresh_chain_is_append_invr in H1; auto.
Qed.

Lemma fresh_chain_fresh_select_p (fcp: fresh_chain_path) fe (fc: fresh_chain):
  fresh_chain_is fc ->
  fresh_chain_fresh fe fc ->
  match (fresh_chain_select_p fcp fc) with
  | (fc2, true) => fresh_chain_fresh fe fc2
  | _ => True
  end.
Proof.
  introv Hi Hfr.
  reverts fc.
  induction fcp; intros; unfold fresh_chain_select in *; simpls.
  - (* nil *)
    auto.
  - destruct fc; auto 1; inversions Hi; inversions Hfr.
    + (forwards: IHfcp H1; fweauto).
       destruct (fresh_chain_select_p fcp fc) eqn: Hb; destruct b; eauto.
    + (forwards: IHfcp H1; fweauto).
       destruct (fresh_chain_select_p fcp fc) eqn: Hb; destruct b; eauto.
  - destruct fc; auto 1; inversions Hi; inversions Hfr;
    (forwards: IHfcp H1; fweauto);
    destruct (fresh_chain_select_p fcp fc) eqn: Hb; destruct b; eauto.
    + constructor. eapply fresh_element_fresh_select; eauto. eauto.
    + constructor. eapply fresh_element_fresh_select; eauto. eauto.
    + constructor. eapply fresh_element_fresh_select; eauto. eauto.
  - destruct fc; auto 1; inversions Hi; inversions Hfr;
    (forwards: IHfcp H1; fweauto);
    destruct (fresh_chain_select_p fcp fc) eqn: Hb; destruct b; eauto.
    constructor. eapply fresh_element_fresh_select; eauto. eauto.
  - forwards: IHfcp. 2: eapply fresh_chain_fresh_skip with (n:=n) (fc:=fc); eauto. eapply fresh_chain_is_skip; eauto. 
    destruct (fresh_chain_select_p fcp (fresh_chain_skip_in n fc)); destruct b; eauto. 
  - destruct fc; auto 1.
  +
    inversions Hi; inversions Hfr; eauto.
    forwards: IHfcp H1 H4 . destruct (fresh_chain_select_p fcp (fresh_chain_append fc1 fc3)); destruct b; eauto.
  +
    inversions Hi; inversions Hfr; eauto.
    forwards: IHfcp H1 H4  . destruct (fresh_chain_select_p fcp fc1); destruct b; eauto.
  - destruct fc; auto 1.
+
    inversions Hi; inversions Hfr; eauto.
    forwards H2r: fresh_chain_fresh_append_invr; fweauto.
    forwards H2l: fresh_chain_fresh_append_invr; fweauto.
    forwards: fresh_chain_fresh_append H6 H2r.
    forwards: IHfcp H.  
    eauto. 
    destruct (fresh_chain_select_p fcp (fresh_chain_append fc2 fc3)); destruct b; eauto.
+
    inversions Hi; inversions Hfr; eauto.
    forwards: IHfcp H2 H5. destruct (fresh_chain_select_p fcp fc2); destruct b; eauto.
  - destruct fc; auto 1;
    inversions Hi; eauto.
    (* Outside branch - no consequence *)
    destruct (fresh_chain_select_p fcp (fresh_chain_append fc1 fc2) ); eauto.
  - 
    forwards Hfcp1: IHfcp1 (fresh_chain_weak fc); eauto.
    rewrite <- fresh_chain_eq_weak; try apply true. auto.
    eapply fresh_chain_fresh_weak; eauto.
    forwards Hfcp2: IHfcp2; eauto.

    destruct (fresh_chain_select_p fcp1  (fresh_chain_weak fc)  ).
    destruct (fresh_chain_select_p fcp2 fc ).
    destruct b, b0; eauto.
    eapply fresh_chain_fresh_append; eauto.

Qed.

Lemma fresh_chain_fresh_select (fcp: fresh_chain_path) fe (fc fc2: fresh_chain):
  fresh_chain_is fc ->
  fresh_chain_fresh fe fc ->
  (fresh_chain_select_p fcp fc) = (fc2, true)  ->
  fresh_chain_fresh fe fc2.
Proof.
  introv Hi Hfr Heq.
  forwards: fresh_chain_fresh_select_p fcp fe Hi Hfr ; fweauto.
  destruct (fresh_chain_select_p fcp  fc).
  inversions Heq. auto.
Qed.

Lemma fresh_element_fresh_select_element (fep: fresh_element_path) fe (fc: fresh_element):
  fresh_element_is fc ->
  fresh_element_is fe ->
  fresh_element_fresh fe fc ->
  fresh_element_fresh (fresh_element_select fep fe) fc.
Proof.
  introv Hi2 Hi Hfr.
  reverts fe;
  induction fep; intros; simpls.
  - auto.
  - destruct fe; auto 1; inversions Hfr; constructor; simpls; eauto using disjoint_empty_l.
  - destruct fe; auto 1.
    inversions Hi.
    eapply IHfep. auto.
    eapply fresh_element_fresh_containing_inv1l; eauto. 
  - destruct fe; auto 1.
    inversions Hi.
    eapply IHfep. auto.
    eapply fresh_element_fresh_containing_inv1r; eauto.
Qed.

Lemma fresh_chain_fresh_select_element (fcp: fresh_element_path) fe (fc1 fc2: fresh_chain):
  fresh_chain_is (fresh_chain_append fc1 fc2) ->
  fresh_element_is fe ->
  fresh_chain_fresh fe fc1 ->
  fresh_chain_fresh (fresh_element_select fcp fe) fc1.
Proof.
  introv Hi Hi2 Hfr.
  reverts fc2.
  induction Hfr; intros; simpls; inversions Hi; eauto.
  - econstructor. eauto.
    eapply fresh_element_fresh_select_element; eauto.
    eauto.
  - econstructor. eauto.
    eapply fresh_element_fresh_select_element; eauto.
    eauto.
  - econstructor. eauto.
    eapply fresh_element_fresh_select_element; eauto.
    eauto.
  - econstructor. eapply IHHfr; eauto using fresh_chain_is_append_invr.
  - econstructor. eapply IHHfr1 with (fc2:=fc2). eauto.
    rewrite <- fresh_chain_append_assoc in H1. eauto.
    eapply IHHfr2; eauto using fresh_chain_is_append_invr.
Qed.



Lemma fresh_chain_indifferent_skip n (fc: fresh_chain):
  fresh_chain_indifferent fc ->
  fresh_chain_indifferent (fresh_chain_skip_in n fc).
Proof.
  introv Hfr.
  reverts fc;
  induction n; intros; simpls.
  - auto.
  - destruct fc; simpls; auto 1;
     inversions Hfr; eauto.
Qed.

Lemma fresh_chain_indifferent_select (fcp: fresh_chain_path) (fc: fresh_chain):
  fresh_chain_indifferent fc ->
  fresh_chain_indifferent (fresh_chain_select fcp fc).
Proof.
  introv Hi.
  reverts fc.
  induction fcp; intros; simpls.
  - auto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H0.
    destruct (fresh_chain_select_p fcp fc); destruct b; simpls; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H0.
    destruct (fresh_chain_select_p fcp fc); destruct b; simpls; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
  -
    eapply IHfcp. eapply fresh_chain_indifferent_skip; eauto.
  - destruct fc; auto 1.
  +
    inversions Hi; eauto.
    eapply IHfcp; eauto.
    eapply fresh_chain_indifferent_append; eauto.
  +
    inversions Hi; eauto.
  - destruct fc; auto 1.
  +
    inversions Hi; eauto.
    eapply IHfcp; eauto.
    eapply fresh_chain_indifferent_append; eauto.
  +
    inversions Hi; eauto.
  - destruct fc; auto 1.
    inversions Hi.

    unfold fresh_chain_select in *; simpls.
    forwards: fresh_chain_indifferent_append H1 H2.
    forwards: IHfcp H.
    destruct (fresh_chain_select_p fcp (fresh_chain_append fc1 fc2)) eqn: Heq; auto.
-
    unfold fresh_chain_select in *; simpls.

    forwards: IHfcp1  (fresh_chain_weak fc).
    eapply fresh_chain_weak_indifferent.

    forwards: IHfcp2 fc; eauto.

    destruct (fresh_chain_select_p fcp1 (fresh_chain_weak fc)) eqn: Hfcp1.
    destruct (fresh_chain_select_p fcp2 fc) eqn: Hfcp2.
    destruct b, b0; simpls; eauto.
    eapply fresh_chain_indifferent_append; eauto.
Qed.


Lemma fresh_chain_is_select (fcp: fresh_chain_path) (fc: fresh_chain):
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_select fcp fc).
Proof.
  introv Hi.
  reverts fc.
  induction fcp; intros; simpls.
  - auto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc); destruct b; simpls; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc) eqn: Hf; destruct b; simpls; eauto.
    constructor; eauto.
    eapply fresh_chain_fresh_select. 3: eauto. eauto. eauto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc); destruct b; simpls; eauto.
    constructor; eauto.
    eapply fresh_element_is_select; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc) eqn: Hf; destruct b; simpls; eauto.
    constructor; eauto.
    eapply fresh_element_is_select; eauto.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc) eqn: Hf; destruct b; simpls; eauto.
    constructor; eauto.
    eapply fresh_element_is_select; eauto.
  - destruct fc; auto 1.
    +
    inversions Hi; unfold fresh_chain_select in *; simpls; eauto.
    forwards: IHfcp H1.
    destruct (fresh_chain_select_p fcp fc) eqn: Hf; destruct b; simpls; eauto.
    constructor; eauto.
    eapply fresh_element_is_select; eauto.
    forwards: fresh_chain_fresh_select Hf; eauto.
    eapply fresh_chain_fresh_select. 3: eauto. eauto. eauto.
    eapply fresh_chain_fresh_select_element with (fc2:=fresh_chain_nil). rewrite fresh_chain_append_nil_r. auto. auto. auto.

  -
    eapply IHfcp. eapply fresh_chain_is_skip; eauto.
  - destruct fc; auto 1.
  +
    inversions Hi; eauto.
  +
    inversions Hi; eauto.

  - destruct fc; auto 1.
  +
    inversions Hi; eauto.
  +
    inversions Hi; eauto.

  - destruct fc; auto 1.
    inversions Hi.
    unfold fresh_chain_select in *; simpls.
    forwards: IHfcp H0.
    destruct (fresh_chain_select_p fcp (fresh_chain_append fc1 fc2)) eqn: Heq; auto.
-
    forwards: IHfcp1  (fresh_chain_weak fc).
    rewrite <- fresh_chain_eq_weak; try apply true. eauto. 
    forwards: IHfcp2 fc; eauto.

    unfold fresh_chain_select in *.
    simpls.
    destruct (fresh_chain_select_p fcp1 (fresh_chain_weak fc)) eqn: Hfcp1.
    simpls.
    destruct (fresh_chain_select_p fcp2 fc) eqn: Hfcp2.
    simpls.
    destruct b, b0; eauto; simpls.
    forwards: fresh_chain_indifferent_select (fresh_chain_weak fc). eapply fresh_chain_weak_indifferent.
    forwards: fresh_chain_is_indifferent_append f f0 fresh_chain_nil.
    rewrite fresh_chain_append_nil_r. auto. auto. unfold fresh_chain_select in H1. erewrite Hfcp1 in H1. simpl in H1. auto.
    auto.
Qed.


Lemma fresh_element_collect_cond_ctx_disjoint G fe:
  fresh_element_collect_cond (fresh_chain_ctx_ok_spec G) fe ->
  disjoint (ctx_dom G) (fresh_element_to_set fe).
Proof.
  introv Hc.
  induction fe; unfold fresh_element_collect_cond in Hc; simpls; destructecs_all; eauto.
  - eapply ctx_ok_concat_inv in Hc0; destructecs_all; eauto using disjoint_union_r.
  - eapply ctx_ok_concat_inv in Hc; destructecs_all; eauto.
Qed.

Lemma fresh_chain_collect_cond_ctx_disjoint_append G fc fc2:
  fresh_chain_is (fresh_chain_append fc fc2) ->
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  disjoint (ctx_dom G) (fresh_chain_to_set fc2) ->
  disjoint (ctx_dom G) (fresh_chain_to_set fc).
Proof.
  introv His Hc Hd.

  remember (fresh_chain_size fc) as fS2.
  reverts fc fc2.
  induction fS2 using nat_lt_induction; intros; subst.
  destruct fc; simpls.
  - eapply disjoint_empty_r.
  - forwards Hfcs: fresh_chain_collect_cond_var_inv Hc.
    destructecs_all.
    inversions His.
    eapply disjoint_union_r.
    eapply fresh_element_collect_cond_ctx_disjoint; eauto.
    eauto.
  - forwards Hfcs: fresh_chain_collect_cond_set_inv Hc.
    destructecs_all.
    inversions His.
    eapply disjoint_union_r.
    eapply fresh_element_collect_cond_ctx_disjoint; eauto.
    eauto.
  - forwards Hfcs: fresh_chain_collect_cond_subset_inv Hc.
    destructecs_all.
    inversions His. eauto.
  - forwards Hfcs: fresh_chain_collect_cond_branch_inv Hc.
    destructecs_all.
    inversions His.
    forwards: fresh_chain_is_append_invr H1.
    eapply H. 4: eauto. 3: eauto. lia. auto. auto.
  - forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hc.
    destructecs_all.
    inversions His.
    forwards: fresh_chain_is_append_invr H2.

    forwards: H Hfcs H0. 2: eauto.
    rewrite Nat.mul_add_distr_r.
    forwards: fresh_chain_branching_gt0 fc1.

    rewrite <- (Nat.succ_pred_pos _ H1).
    rewrite Nat.mul_succ_l. lia.
    auto.
    
    forwards: H Hfcs1 H4. 2: eauto.
    rewrite Nat.mul_add_distr_r. lia.
    rewrite fresh_chain_append_to_set.
    eapply disjoint_union_r; auto.

    forwards: H Hfcs0 H2. 2: eauto.
    rewrite Nat.mul_add_distr_r. lia.
    rewrite fresh_chain_append_to_set.
    eapply disjoint_union_r; auto.
  
    eauto using disjoint_union_r.

  - forwards Hfcs: fresh_chain_collect_cond_union_inv Hc.
    destructecs_all.
    inversions His.
    forwards: fresh_chain_is_append_invr H2.

    forwards: H Hfcs H3. 2: eauto.
    lia.    auto.
    
    forwards: H Hfcs0 H2. 2: eauto.
    lia. auto.
    
    eauto using disjoint_union_r.
Qed.

Lemma fresh_chain_collect_cond_ctx_disjoint G fc:
  fresh_chain_is fc ->
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  disjoint (ctx_dom G) (fresh_chain_to_set fc).
Proof.
  introv His Hc.
  eapply fresh_chain_collect_cond_ctx_disjoint_append.
  rewrite fresh_chain_append_nil_r. 
  all: auto.
  simpls. apply disjoint_empty_r.
Qed.

Lemma fresh_element_collect_cond_ctx_ok G fe:
  fresh_element_collect_cond (fresh_chain_ctx_ok_spec G) fe ->
  ctx_ok G.
Proof.
  introv Hc.
  induction fe; unfold fresh_element_collect_cond in Hc; simpls; destructecs_all; eauto.
  - eapply ctx_ok_concat_inv in Hc0; destructecs_all; auto.
  - eapply ctx_ok_concat_inv in Hc; destructecs_all; auto.
Qed.
Lemma fresh_chain_collect_cond_ctx_ok G fe:
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fe ->
  ctx_ok G.
Proof.
  introv Hc.
  forwards: fresh_chain_collect_cond_nil_inv Hc.
  simpls. auto.
Qed.

Lemma fresh_chain_collect_cond_renv_ok G fe:
  fresh_chain_collect_cond (fresh_chain_renv_ok_spec G) fe ->
  ok G.
Proof.
  introv Hc.
  forwards: fresh_chain_collect_cond_nil_inv Hc.
  simpls. auto.
Qed.

Lemma fresh_chain_collect_cond_heap_ok G fe:
  fresh_chain_collect_cond (fresh_chain_heap_ok_spec G) fe ->
  ok G.
Proof.
  introv Hc.
  forwards: fresh_chain_collect_cond_nil_inv Hc.
  simpls. auto.
Qed.




Lemma fresh_element_disjoint_collect_cond_ctx G fe:
  ctx_ok G ->
  disjoint (ctx_dom G) (fresh_element_to_set fe) ->
  fresh_element_is fe ->
  fresh_element_collect_cond (fresh_chain_ctx_ok_spec G) fe.
Proof.
  introv HokG Hd His.
  induction fe; unfold fresh_element_collect_cond; simpls; destructecs_all; eauto.
  - inversions His.
    forwards: disjoint_union_r_inv Hd. destructecs H.
    split.
    eapply ctx_ok_concat; eauto. eauto.
  - inversions His.
    eapply ctx_ok_concat; eauto.
Qed.

Lemma fresh_element_collect_cond_renv_disjoint G fe:
  fresh_element_collect_cond (fresh_chain_renv_ok_spec G) fe ->
  disjoint (dom G) (fresh_element_to_set fe).
Proof.
  introv Hc.
  induction fe; unfold fresh_element_collect_cond in Hc; simpls; destructecs_all; eauto.
  - forwards: ok_concat_disjoint Hc0.
    eapply ok_concat_inv in Hc0; destructecs_all; eauto using disjoint_union_r.
  - forwards: ok_concat_disjoint Hc. eapply ok_concat_inv in Hc; destructecs_all; eauto.
Qed.
Lemma fresh_element_collect_cond_heap_disjoint G fe:
  fresh_element_collect_cond (fresh_chain_heap_ok_spec G) fe ->
  disjoint (dom G) (fresh_element_to_set fe).
Proof.
  introv Hc.
  induction fe; unfold fresh_element_collect_cond in Hc; simpls; destructecs_all; eauto.
  - forwards: ok_concat_disjoint Hc0.
    eapply ok_concat_inv in Hc0; destructecs_all; eauto using disjoint_union_r.
  - forwards: ok_concat_disjoint Hc. eapply ok_concat_inv in Hc; destructecs_all; eauto.
Qed.

Lemma fresh_element_add_fresh_element_ctxdom G fe:
  fresh_element_collect_cond (fresh_chain_ctx_ok_spec G) fe ->
  fresh_element_is fe ->
  fresh_element_is (fresh_element_ctxdom G) ->
  fresh_element_fresh (fresh_element_ctxdom G) fe.
Proof. 
  introv Hc Hfe HfeG.
  eapply fresh_element_from_disjoint_set; eauto.
  simpls.
  eapply fresh_element_collect_cond_ctx_disjoint; eauto.
Qed.

Lemma fresh_element_add_fresh_element_renvdom G fe:
  fresh_element_collect_cond (fresh_chain_renv_ok_spec G) fe ->
  fresh_element_is fe ->
  fresh_element_is (fresh_element_renvdom G) ->
  fresh_element_fresh (fresh_element_renvdom G) fe.
Proof. 
  introv Hc Hfe HfeG.
  eapply fresh_element_from_disjoint_set; eauto.
  simpls.
  eapply fresh_element_collect_cond_renv_disjoint; eauto.
Qed.

Lemma fresh_element_add_fresh_element_heapdom G fe:
  fresh_element_collect_cond (fresh_chain_heap_ok_spec G) fe ->
  fresh_element_is fe ->
  fresh_element_is (fresh_element_heapdom G) ->
  fresh_element_fresh (fresh_element_heapdom G) fe.
Proof. 
  introv Hc Hfe HfeG.
  eapply fresh_element_from_disjoint_set; eauto.
  simpls.
  eapply fresh_element_collect_cond_heap_disjoint; eauto.
Qed.

Lemma fresh_chain_add_fresh_element_ctx_subset fc fe G:
  ctx_ok G ->
  fresh_chain_is fc ->
  fresh_element_is fe ->
  fresh_element_to_set fe \c fresh_chain_to_set fc ->
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  fresh_element_collect_cond (fresh_chain_ctx_ok_spec G) fe.
Proof.
  introv HokG Hifc Hife Hsu Hcnfc.
  eapply fresh_element_disjoint_collect_cond_ctx.
  auto.
  eapply disjoint_subset_l.
  eapply fresh_chain_collect_cond_ctx_disjoint; eauto.
  eauto.
  auto.
Qed.

(*
Lemma fresh_chain_fresh_collect_ctx_append G fc:
  ctx_ok G ->
  fresh_chain_is fc ->
  fresh_chain_fresh (fresh_element_ctxdom G) fc ->
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc.
Proof.
  introv HokG His Hfr.
  induction His; inversions Hfr.
  -
    unfold fresh_chain_collect_cond, fresh_collect_cond in *; intros; simpls; auto 1.
  - Check fresh_chain_collect_cond_var.
  destruct (fresh_chain_collect_opt fresh_chain_ctx_map_cond ctx_concat fc); simpls.
  destruct (fresh_chain_ctx_map_cond fe).

  destruct o, o0, o1.
simpls. destruct o2.

unfold fresh_chain_reduce_set.

  all: rewrite ? ctx_concat_empty_l; eauto using disjoint_empty_l.
*)

Lemma fresh_chain_add_fresh_element_ctx_append G fc fc2:
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  fresh_chain_is (fresh_chain_append fc fc2) ->
  fresh_chain_fresh (fresh_element_ctxdom G) fc2 ->
  fresh_chain_fresh (fresh_element_ctxdom G) fc.
Proof. 
  introv Hc Hfc Hfc2.
  reverts fc2;
  induction fc; intros.
  - constructor.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_set_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_ctxdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_var_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_ctxdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_subset_inv Hc.
    forwards Hfc1: IHfc; fweauto.
    forwards Hfc12: fresh_chain_fresh_append Hfc1 Hfc2.

    constructor.
    eapply fresh_element_from_disjoint_set.
    eapply disjoint_subset_l. 2: eauto.
    eapply fresh_chain_disjoint_set; eauto.
    constructor.
    eapply fresh_chain_collect_cond_ctx_ok; eauto.
    auto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_branch_inv Hc.
    constructor.
    eapply IHfc2; eauto.
    eapply fresh_chain_is_append_invr; eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc3; fweauto.
    forwards: IHfc1; fweauto. eapply fresh_chain_fresh_append; eauto.
    forwards: IHfc2; fweauto. eapply fresh_chain_fresh_append; eauto.
    constructor.
    eapply fresh_chain_fresh_append; eauto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_union_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc1; fweauto.
    forwards: IHfc2; fweauto.
    eauto.
Qed.


Lemma fresh_chain_add_fresh_element_ctx G fc:
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  fresh_chain_is fc ->
  fresh_chain_fresh (fresh_element_ctxdom G) fc.
Proof. 
  introv Hc Hfc.

  eapply fresh_chain_add_fresh_element_ctx_append; eauto.
  rewrite fresh_chain_append_nil_r; eauto.
Qed.


Lemma fresh_chain_add_fresh_element_renv_append G fc fc2:
  fresh_chain_collect_cond (fresh_chain_renv_ok_spec G) fc ->
  fresh_chain_is (fresh_chain_append fc fc2) ->
  fresh_chain_fresh (fresh_element_renvdom G) fc2 ->
  fresh_chain_fresh (fresh_element_renvdom G) fc.
Proof. 
  introv Hc Hfc Hfc2.
  reverts fc2;
  induction fc; intros.
  - constructor.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_set_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_renvdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_var_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_renvdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_subset_inv Hc.
    forwards Hfc1: IHfc; fweauto.
    forwards Hfc12: fresh_chain_fresh_append Hfc1 Hfc2.

    constructor.
    eapply fresh_element_from_disjoint_set.
    eapply disjoint_subset_l. 2: eauto.
    eapply fresh_chain_disjoint_set; eauto.
    constructor.
    eapply fresh_chain_collect_cond_renv_ok; eauto.
    auto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_branch_inv Hc.
    constructor.
    eapply IHfc2; eauto.
    eapply fresh_chain_is_append_invr; eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc3; fweauto.
    forwards: IHfc1; fweauto. eapply fresh_chain_fresh_append; eauto.
    forwards: IHfc2; fweauto. eapply fresh_chain_fresh_append; eauto.
    constructor.
    eapply fresh_chain_fresh_append; eauto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_union_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc1; fweauto.
    forwards: IHfc2; fweauto.
    eauto.
Qed.


Lemma fresh_chain_add_fresh_element_renv G fc:
  fresh_chain_collect_cond (fresh_chain_renv_ok_spec G) fc ->
  fresh_chain_is fc ->
  fresh_chain_fresh (fresh_element_renvdom G) fc.
Proof. 
  introv Hc Hfc.

  eapply fresh_chain_add_fresh_element_renv_append; eauto.
  rewrite fresh_chain_append_nil_r; eauto.
Qed.


Lemma fresh_chain_add_fresh_element_heap_append G fc fc2:
  fresh_chain_collect_cond (fresh_chain_heap_ok_spec G) fc ->
  fresh_chain_is (fresh_chain_append fc fc2) ->
  fresh_chain_fresh (fresh_element_heapdom G) fc2 ->
  fresh_chain_fresh (fresh_element_heapdom G) fc.
Proof. 
  introv Hc Hfc Hfc2.
  reverts fc2;
  induction fc; intros.
  - constructor.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_set_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_heapdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_var_inv Hc.
    destructecs Hfcs.
    constructor.
    eapply fresh_element_add_fresh_element_heapdom; eauto.
    forwards: fresh_chain_collect_cond_nil_inv Hfcs0; simpls.
    constructor; auto.
    eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_subset_inv Hc.
    forwards Hfc1: IHfc; fweauto.
    forwards Hfc12: fresh_chain_fresh_append Hfc1 Hfc2.

    constructor.
    eapply fresh_element_from_disjoint_set.
    eapply disjoint_subset_l. 2: eauto.
    eapply fresh_chain_disjoint_set; eauto.
    constructor.
    eapply fresh_chain_collect_cond_heap_ok; eauto.
    auto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_branch_inv Hc.
    constructor.
    eapply IHfc2; eauto.
    eapply fresh_chain_is_append_invr; eauto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_parallel_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc3; fweauto.
    forwards: IHfc1; fweauto. eapply fresh_chain_fresh_append; eauto.
    forwards: IHfc2; fweauto. eapply fresh_chain_fresh_append; eauto.
    constructor.
    eapply fresh_chain_fresh_append; eauto.
    auto.
  - inversions Hfc.
    forwards Hfcs: fresh_chain_collect_cond_union_inv Hc.
    destructecs Hfcs.
    forwards His3: fresh_chain_is_append_invr H1; fweauto.
    forwards: IHfc1; fweauto.
    forwards: IHfc2; fweauto.
    eauto.
Qed.


Lemma fresh_chain_add_fresh_element_heap G fc:
  fresh_chain_collect_cond (fresh_chain_heap_ok_spec G) fc ->
  fresh_chain_is fc ->
  fresh_chain_fresh (fresh_element_heapdom G) fc.
Proof. 
  introv Hc Hfc.

  eapply fresh_chain_add_fresh_element_heap_append; eauto.
  rewrite fresh_chain_append_nil_r; eauto.
Qed.




Lemma fresh_chain_add_fresh_ctx G fc:
  fresh_chain_collect_cond (fresh_chain_ctx_ok_spec G) fc ->
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_var (fresh_element_ctxdom G) fc).
Proof.
  introv Hc Hfc.
  constructor.
  - auto.
  - apply fresh_chain_collect_cond_nil_inv in Hc. simpls. eauto.
  - eapply fresh_chain_add_fresh_element_ctx; eauto.
Qed.
Lemma fresh_chain_add_fresh_renv G fc:
  fresh_chain_collect_cond (fresh_chain_renv_ok_spec G) fc ->
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_var (fresh_element_renvdom G) fc).
Proof.
  introv Hc Hfc.
  constructor.
  - auto.
  - apply fresh_chain_collect_cond_nil_inv in Hc. simpls. eauto.
  - eapply fresh_chain_add_fresh_element_renv; eauto.
Qed.


Fixpoint fresh_chain_collect_just_ctx_opt (fc: fresh_chain): option ctx :=
match fc with
| fresh_chain_var (fresh_element_ctxdom G) fc1
| fresh_chain_var (fresh_element_ctx G) fc1 =>
  match (fresh_chain_collect_just_ctx_opt fc1) with | None => Some G | Some G1 => Some (G1 §& G) end
| fresh_chain_set (fresh_element_ctxdom G) fc1
| fresh_chain_set (fresh_element_ctx G) fc1 =>
  match (fresh_chain_collect_just_ctx_opt fc1) with | None => Some G | Some G1 => Some G1 end
| _ => None
end.
Definition fresh_chain_collect_just_ctx fc: ctx := match fresh_chain_collect_just_ctx_opt fc with | None => ctx_empty | Some G => G end.

Fixpoint fresh_chain_collect_just_renv_opt (fc: fresh_chain): option renv :=
match fc with
| fresh_chain_var (fresh_element_renvdom e) fc1
| fresh_chain_var (fresh_element_renv e) fc1 =>
  match (fresh_chain_collect_just_renv_opt fc1) with | None => Some e | Some e1 => Some (e1 & e) end
| fresh_chain_set (fresh_element_renvdom e) fc1
| fresh_chain_set (fresh_element_renv e) fc1 =>
  match (fresh_chain_collect_just_renv_opt fc1) with | None => Some e | Some e1 => Some e1 end
| _ => None
end.
Definition fresh_chain_collect_just_renv fc: renv := match fresh_chain_collect_just_renv_opt fc with | None => empty | Some e => e end.

Fixpoint fresh_chain_collect_just_heap_opt (fc: fresh_chain): option heap :=
match fc with
| fresh_chain_var (fresh_element_heapdom e) fc1 => match (fresh_chain_collect_just_heap_opt fc1) with | None => Some e | Some e1 => Some (e1 & e) end
| _ => None
end.
Definition fresh_chain_collect_just_heap fc: heap := match fresh_chain_collect_just_heap_opt fc with | None => empty | Some e => e end.





Lemma fresh_chain_cond_just_ctx_subset fc:
  ctx_dom (fresh_chain_collect_just_ctx fc) \c fresh_chain_to_set fc.
Proof.
  induction fc; simpls; (* Induction on the chain *)
  try eapply subset_empty_l; (* Solves invalid cases (= except set and var) *)
  unfold fresh_chain_collect_just_ctx in *; simpl; (* Expand the matching *)
  destruct fe; simpls; (* Cases on the elements *)
  try eapply subset_empty_l; (* Solves invalid cases *)
  destruct (fresh_chain_collect_just_ctx_opt fc) eqn: Heq; (* Cases on the value of the rest of the chain *)
  rewrite ? ctx_dom_concat;
  union_sub_solver.
Qed.

Lemma fresh_chain_cond_just_ctx fc:
  fresh_chain_is fc ->
  ctx_ok (fresh_chain_collect_just_ctx fc).
Proof.
  introv Hfc.
  induction Hfc; unfold fresh_chain_collect_just_ctx in *; simpls; auto.
  - (* Var *)
  forwards: fresh_chain_cond_just_ctx_subset fc.
  unfold fresh_chain_collect_just_ctx in *.
  destruct ( fresh_chain_collect_just_ctx_opt fc) eqn: Hefc; destruct fe; inversions H; auto.
  +
  eapply ctx_ok_concat; auto.
  eapply fresh_chain_disjoint_set in H0.
  eapply disjoint_union_l_inv in H0. destruct H0.
  eapply disjoint_comm.
  eapply disjoint_subset_l; eauto.
  +
  eapply ctx_ok_concat; auto.
  eapply fresh_chain_disjoint_set in H0.
  simpl in H0.
  eapply disjoint_comm.
  eapply disjoint_subset_l; eauto.
  - (* Set *)
  forwards: fresh_chain_cond_just_ctx_subset fc.
  unfold fresh_chain_collect_just_ctx in *.
  destruct ( fresh_chain_collect_just_ctx_opt fc) eqn: Hefc; destruct fe; auto;
  inversions H; auto.
Qed.

Lemma fresh_chain_cond_just_renv_subset fc:
  dom (fresh_chain_collect_just_renv fc) \c fresh_chain_to_set fc.
Proof.
  induction fc; (* Induction on the chain *)
  unfold fresh_chain_collect_just_renv in *; simpls; rewrite ? dom_empty; (* Expand the matching *)
  try eapply subset_empty_l; (* Solves invalid cases (= except set and var) *)
  destruct fe; simpls; rewrite ? dom_empty;  (* Cases on the elements *)
  try eapply subset_empty_l;   (* Solves invalid cases *)
  destruct (fresh_chain_collect_just_renv_opt fc) eqn: Heq; (* Cases on the value of the rest of the chain *)
  rewrite ? dom_concat;
  union_sub_solver.
Qed.



Lemma fresh_chain_cond_just_renv fc:
  fresh_chain_is fc ->
  ok (fresh_chain_collect_just_renv fc).
Proof.
  introv Hfc.
  induction Hfc; unfold fresh_chain_collect_just_renv in *; simpls; auto.
  - (* Var *)
  forwards: fresh_chain_cond_just_renv_subset fc.
  unfold fresh_chain_collect_just_renv in *.
  destruct ( fresh_chain_collect_just_renv_opt fc) eqn: Hefc; destruct fe; inversions H; auto.
  +
  eapply ok_concat; auto.
  eapply fresh_chain_disjoint_set in H0.
  eapply disjoint_union_l_inv in H0. destruct H0.
  eapply disjoint_comm.
  eapply disjoint_subset_l; eauto.
  +
  eapply ok_concat; auto.
  eapply fresh_chain_disjoint_set in H0.
  simpl in H0.
  eapply disjoint_comm.
  eapply disjoint_subset_l; eauto.
  - (* Set *)
  forwards: fresh_chain_cond_just_renv_subset fc.
  unfold fresh_chain_collect_just_renv in *.
  destruct ( fresh_chain_collect_just_renv_opt fc) eqn: Hefc; destruct fe; auto;
  inversions H; auto.
Qed.

Lemma fresh_chain_cond_just_heap_subset fc:
  dom (fresh_chain_collect_just_heap fc) \c fresh_chain_to_set fc.
Proof.
  induction fc;
  unfold fresh_chain_collect_just_heap in *; simpl; rewrite ? dom_empty;
  simpls; try eapply subset_empty_l; eauto.
  destruct fe; simpls;  rewrite ? dom_empty; try eapply subset_empty_l; eauto.
  destruct (fresh_chain_collect_just_heap_opt fc) eqn: Heq.
  rewrite dom_concat.
  union_sub_solver.
  union_sub_solver.
Qed.



Lemma fresh_chain_cond_just_heap fc:
  fresh_chain_is fc ->
  ok (fresh_chain_collect_just_heap fc).
Proof.
  introv Hfc.
  induction Hfc; unfold fresh_chain_collect_just_heap in *; simpls; auto.
  forwards: fresh_chain_cond_just_heap_subset fc.
  unfold fresh_chain_collect_just_heap in *.
  destruct ( fresh_chain_collect_just_heap_opt fc) eqn: Hefc; destruct fe; auto.
  inversions H.
  eapply ok_concat; auto.
  eapply fresh_chain_disjoint_set in H0.
  simpl in H0.
  eapply disjoint_comm.
  eapply disjoint_subset_l; eauto. 
  inversions H; auto.
Qed.





Lemma fresh_chain_add_fresh_list S l xs fc:
  fresh_chain_to_set fc \c S ->
  fresh S l xs ->
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_var (fresh_element_list xs) fc).
Proof.
  introv HS Hfr Hfc.
  subst.
  eapply fresh_resize_length in Hfr.
  constructor.
  auto.
  constructor; eauto.
  eapply fresh_chain_from_disjoint_set.
  simpls.
  apply disjoint_comm.
  apply fresh_disjoint; auto.
  eapply fresh_subset; eauto.
  constructor; eauto.
  auto.
Qed.

Lemma fresh_chain_add_fresh_var S x fc:
  fresh_chain_to_set fc \c S ->
  x \notin S ->
  fresh_chain_is fc ->
  fresh_chain_is (fresh_chain_var (fresh_element_var x) fc).
Proof.
  introv HS Hfr Hfc.
  subst.
  constructor.
  auto.
  constructor; eauto.
  eapply fresh_chain_from_disjoint_set.
  simpls.
  apply disjoint_singleton.
  intro. eauto.
  constructor; eauto.
  auto.
Qed.

(* Opposite of fresh_chain_add_fresh_list *)
Lemma fresh_chain_fresh_list S l xs fc:
  fresh_chain_is (fresh_chain_var (fresh_element_list xs) fc) ->
  S \c fresh_chain_to_set fc ->
  l = length xs ->
  fresh S l xs.
Proof.
  introv His Hsu Hl.
  subst.
  inversions His.
  inversions H2.
  forwards: fresh_chain_disjoint_set H3.
  eapply disjoint_fresh; eauto.
  apply disjoint_comm.
  simpl in H.
  eapply disjoint_subset_l; eauto.
Qed.
Lemma fresh_chain_fresh_list_large S l xs fc:
  fresh_chain_is (fresh_chain_var (fresh_element_list xs) fc) ->
  S \c fresh_chain_to_set_large fc ->
  l = length xs ->
  fresh S l xs.
Proof.
  introv His Hsu Hl.
  subst.
  inversions His.
  rewrite fresh_chain_to_set_large_eq in Hsu.
  eapply fresh_chain_fresh_list; eauto. eauto.
Qed.

Lemma fresh_chain_eq_prepend_susbet_set flags fc X:
  X \c fresh_chain_to_set fc ->
  fresh_chain_eq flags (fresh_chain_set (fresh_element_set X) fc) fc.
Proof.
  introv Hok.
  destruct flags; econstructor; simpls; cases_if; auto 1. 
  - intros. inversions H. auto.
  - intros. econstructor; eauto.
  - union_sub_solver.
  - union_sub_solver.
Qed.

Lemma fresh_chain_eq_prepend_susbet_subset flags fc X:
  X \c fresh_chain_to_set fc ->
  fresh_chain_eq flags (fresh_chain_subset (fresh_element_set X) fc) fc.
Proof.
  introv Hok.
  destruct flags; econstructor; simpls; cases_if; auto 1. 
  - intros. inversions H. auto.
  - intros. econstructor; eauto.
  - union_sub_solver.
  - union_sub_solver.
Qed.

